/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg__do_cmd_checkout.c
 *
 * @details Handle 'vv checkout'.
 *
 */

//////////////////////////////////////////////////////////////////

#include <sg.h>

#include "sg_typedefs.h"
#include "sg_prototypes.h"

#include "sg_cmd_util_prototypes.h"

//////////////////////////////////////////////////////////////////

void do_cmd_checkout(SG_context * pCtx,
					 const char * pszRepoName,
					 SG_rev_spec* pRevSpec,
					 const char * pszFolder,
					 const char* psz_attach,
					 SG_bool bSubmodule,
					 const SG_vector * pvec_use_submodules,
					 int * pExitStatus)
{
	char* psz_hid_cs = NULL;
	char* psz_branch_name = NULL;
	SG_vhash * pvhResults = NULL;
	SG_int64 status;
	SG_repo * pRepo = NULL;
	SG_vhash * pvh_pile = NULL;

	if(psz_attach!=NULL && SG_rev_spec__is_empty(pCtx, pRevSpec))
		SG_ERR_THROW2(SG_ERR_USAGE, (pCtx, "When using the --attach option, you must also specify a revision."));

	SG_vv_utils__rev_spec_to_csid__no_wd(pCtx, pszRepoName, pRevSpec, &psz_hid_cs, &psz_branch_name);

	if(SG_context__err_equals(pCtx, SG_ERR_BRANCH_NEEDS_MERGE))
	{
		SG_uint32 branchesSpecified;

		SG_context__err_reset(pCtx);

		SG_ERR_CHECK(  SG_rev_spec__count_branches(pCtx, pRevSpec, &branchesSpecified)  );

		if(!branchesSpecified)
		{
			SG_ERR_CHECK(  SG_console(pCtx, SG_CS_STDERR, "The master branch has multiple heads.\n\n")  );
			SG_ERR_CHECK(  SG_console(pCtx, SG_CS_STDERR, "Check out using \"--attach=master\" to attach your working folder to it.\n")  );
			SG_ERR_CHECK(  SG_console(pCtx, SG_CS_STDERR, "Use the '--rev' or '--tag' option to pick a specific revision.\n")  );

			SG_ERR_CHECK(  SG_rev_spec__add_branch(pCtx, pRevSpec, "master")  );
		}
		else
		{
			SG_ERR_CHECK(  SG_console(pCtx, SG_CS_STDERR, "The branch has multiple heads.\n")  );
			SG_ERR_CHECK(  SG_console(pCtx, SG_CS_STDERR, "Check out using the \"--attach\" option rather than \"--branch\" to attach your working folder to it.\n")  );
			SG_ERR_CHECK(  SG_console(pCtx, SG_CS_STDERR, "You must also use the \"--rev\" or \"--tag\" option to pick a specific revision.\n")  );
		}
		SG_ERR_CHECK(  SG_console(pCtx, SG_CS_STDERR, "\nCurrent heads are:\n")  );

		SG_ERR_CHECK(  SG_REPO__OPEN_REPO_INSTANCE(pCtx, pszRepoName, &pRepo)  );
		SG_ERR_CHECK(  SG_vc_branches__cleanup(pCtx, pRepo, &pvh_pile)  );
		SG_ERR_CHECK(  SG_cmd_util__dump_log__revspec(pCtx, SG_CS_STDERR, pRepo, pRevSpec, pvh_pile, SG_FALSE, SG_FALSE)  );

		SG_VHASH_NULLFREE(pCtx, pvh_pile);
		SG_REPO_NULLFREE(pCtx, pRepo);

		SG_ASSERT(pExitStatus!=NULL);
		*pExitStatus=1;
		goto fail;
	}
	else
		SG_ERR_CHECK_CURRENT;

	// Try to create the WD and populate it.

	SG_ERR_CHECK(  SG_vv_verbs__checkout(pCtx,
										 pszRepoName, psz_hid_cs, pszFolder,
										 SG_TRUE, bSubmodule,
										 psz_attach ? psz_attach : psz_branch_name,
										 pvec_use_submodules,
										 &pvhResults)  );
	SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvhResults, "status", &status)  );
	switch (status)
	{
	default:
	case SG_VV_VERBS__STATUS__OK:
		break;

	case SG_VV_VERBS__STATUS__SUBMODULE_PROBLEM:
		{
			SG_varray * pvaMissingDetails;	// we don't own this

			SG_ERR_CHECK(  SG_vhash__get__varray(pCtx, pvhResults, "sub", &pvaMissingDetails)  );

			SG_ERR_IGNORE(  sg_report_submodule_missing_details(pCtx, pvaMissingDetails)  );
			SG_ERR_THROW(  SG_ERR_SUBMODULE_NESTED_UPDATE_CHECKOUT_PROBLEM  ); // throw so that exe exits with non-zero status
		}
		break;
	}
	
fail:
	// TODO 2010/10/25 Decide which error message we should try to provide a better error message for.
	// TODO            Candidates include
	// TODO SG_ERR_INVALID_REPO_NAME
	// TODO SG_ERR_AMBIGUOUS_ID_PREFIX
	// TODO SG_ERR_MULTIPLE_HEADS_FROM_DAGNODE
	// TODO
	// TODO            But we should look and see what the lower layers are throwing now.  Some of
	// TODO            of the messages that were here are now being handled.

	SG_NULLFREE(pCtx, psz_hid_cs);
	SG_NULLFREE(pCtx, psz_branch_name);
	SG_VHASH_NULLFREE(pCtx, pvhResults);
	SG_VHASH_NULLFREE(pCtx, pvh_pile);
	SG_REPO_NULLFREE(pCtx, pRepo);
}
