/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include <sg.h>
#include "sg_typedefs.h"
#include "sg_prototypes.h"

static void x_parse_prefixed_path(
        SG_context* pCtx,
        const char* psz_path,
        SG_string** ppstr_path, 
        SG_treenode_entry_type* pt
        )
{
    SG_ERR_CHECK_RETURN(  SG_string__alloc__sz(pCtx, ppstr_path, 1 + psz_path)  );
    if ('D' == psz_path[0])
    {
        *pt = SG_TREENODEENTRY_TYPE_DIRECTORY;
    }
    else if ('F' == psz_path[0])
    {
        *pt = SG_TREENODEENTRY_TYPE_REGULAR_FILE;
    }
    else if ('L' == psz_path[0])
    {
        *pt = SG_TREENODEENTRY_TYPE_SYMLINK;
    }
    else if ('S' == psz_path[0])
    {
        *pt = SG_TREENODEENTRY_TYPE_SUBMODULE;
    }
    else
    {
        SG_ERR_THROW_RETURN(  SG_ERR_UNSPECIFIED  );
    }
}

static void x_which_path(
        SG_context* pCtx,
        SG_tncache* ptn_cache,
        SG_treenode* ptn, 
        SG_varray* pva,
        const char* psz_gid,
        SG_string** ppPathOfFoundObject, 
        SG_treenode_entry_type* pt
        )
{
    SG_uint32 count = 0;
    SG_uint32 i = 0;
    SG_string* pstr_path = NULL;
    SG_treenode_entry* ptne = NULL;
    char* psz_found_gid = NULL;

	SG_ERR_CHECK(  SG_varray__count(pCtx, pva, &count )  );
	for (i = 0; i < count-1; i++)
	{
        const char* psz_path = NULL;

        SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pva, i, &psz_path)  );
		SG_ERR_CHECK(  SG_treenode__find_treenodeentry_by_path__cached(pCtx, ptn_cache, ptn, 1+psz_path, &psz_found_gid, &ptne)  );
        if (
                ptne
                && (0 == strcmp(psz_gid, psz_found_gid))
           )
        {
            SG_ERR_CHECK(  SG_string__alloc__sz(pCtx, &pstr_path, 1+psz_path)  );
            break;
        }
        SG_NULLFREE(pCtx, psz_found_gid);
        SG_TREENODE_ENTRY_NULLFREE(pCtx, ptne);
	}

    if (pstr_path)
    {
        *ppPathOfFoundObject = pstr_path;
        pstr_path = NULL;
        SG_ERR_CHECK(  SG_treenode_entry__get_entry_type(pCtx, ptne, pt)  );
    }
    else
    {
        const char* psz_last_path = NULL;

        SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pva, count-1, &psz_last_path)  );
        SG_ERR_CHECK(  x_parse_prefixed_path(pCtx, psz_last_path, ppPathOfFoundObject, pt)  );
    }

fail:
    SG_NULLFREE(pCtx, psz_found_gid);
    SG_TREENODE_ENTRY_NULLFREE(pCtx, ptne);
    SG_STRING_NULLFREE(pCtx, pstr_path);
}

static void x_get_audits(
        SG_context * pCtx,
        SG_varray* pva,
        SG_vhash** ppvh
        )
{
    SG_vhash* pvh_audits = NULL;
    SG_uint32 count = 0;
    SG_uint32 i = 0;

    SG_ERR_CHECK(  SG_vhash__alloc(pCtx, &pvh_audits)  );
    SG_ERR_CHECK(  SG_varray__count(pCtx, pva, &count)  );
    for (i=0; i<count; i++)
    {
        SG_vhash* pvh = NULL;
        const char* psz_csid = NULL;
        const char* psz_userid = NULL;
        SG_int64 timestamp = -1;
        SG_bool b_already = SG_FALSE;

        SG_ERR_CHECK(  SG_varray__get__vhash(pCtx, pva, i, &pvh)  );
        SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh, "csid", &psz_csid)  );
        SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh, "userid", &psz_userid)  );
        SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvh, "timestamp", &timestamp)  );
        SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh_audits, psz_csid, &b_already)  );
        if (!b_already)
        {
            SG_vhash* pvh_a = NULL;

            SG_ERR_CHECK(  SG_vhash__addnew__vhash(pCtx, pvh_audits, psz_csid, &pvh_a)  );
            SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvh_a, "userid", psz_userid)  );
            SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh_a, "timestamp", timestamp)  );
        }
    }

    *ppvh = pvh_audits;
    pvh_audits = NULL;

fail:
    SG_VHASH_NULLFREE(pCtx, pvh_audits);
}

static void x_get_comments(
        SG_context * pCtx,
        SG_varray* pva,
        SG_vhash** ppvh
        )
{
    SG_vhash* pvh_comments = NULL;
    SG_uint32 count = 0;
    SG_uint32 i = 0;

    SG_ERR_CHECK(  SG_vhash__alloc(pCtx, &pvh_comments)  );
    SG_ERR_CHECK(  SG_varray__count(pCtx, pva, &count)  );
    for (i=0; i<count; i++)
    {
        SG_vhash* pvh = NULL;
        const char* psz_val = NULL;
        const char* psz_csid = NULL;
        SG_bool b_already = SG_FALSE;

        SG_ERR_CHECK(  SG_varray__get__vhash(pCtx, pva, i, &pvh)  );
        SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh, "text", &psz_val)  );
        SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh, "csid", &psz_csid)  );
        SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh_comments, psz_csid, &b_already)  );
        if (!b_already)
        {
            SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvh_comments, psz_csid, psz_val)  );
        }
    }

    *ppvh = pvh_comments;
    pvh_comments = NULL;

fail:
    SG_VHASH_NULLFREE(pCtx, pvh_comments);
}

static void x_get_users(
        SG_context * pCtx,
        SG_varray* pva,
        SG_vhash** ppvh
        )
{
    SG_vhash* pvh_users = NULL;
    SG_uint32 count = 0;
    SG_uint32 i = 0;

    SG_ERR_CHECK(  SG_vhash__alloc(pCtx, &pvh_users)  );
    SG_ERR_CHECK(  SG_varray__count(pCtx, pva, &count)  );
    for (i=0; i<count; i++)
    {
        SG_vhash* pvh = NULL;
        const char* psz_username = NULL;
        const char* psz_userid = NULL;
        SG_bool b_already = SG_FALSE;

        SG_ERR_CHECK(  SG_varray__get__vhash(pCtx, pva, i, &pvh)  );
        SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh, "name", &psz_username)  );
        SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh, SG_ZING_FIELD__RECID, &psz_userid)  );
        SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh_users, psz_userid, &b_already)  );
        if (!b_already)
        {
            SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvh_users, psz_userid, psz_username)  );
        }
    }

    *ppvh = pvh_users;
    pvh_users = NULL;

fail:
    SG_VHASH_NULLFREE(pCtx, pvh_users);
}

void do_cmd_fast_export(
    SG_context * pCtx,
    SG_option_state * pOptSt,
	const char * psz_repo,
	const char * psz_fi
	)
{
	SG_dagnode* pdn = NULL;
    SG_repo* pRepo = NULL;
	char** paszHids = NULL;
	const char* pszHid = NULL;
	SG_repo_fetch_dagnodes_handle* pdh = NULL;
    SG_uint32 count_nodes;
    SG_uint32 count_tags;
    SG_uint32 i;
    SG_uint32 i_changeset;
    SG_uint32 i_tag;
    SG_vhash* pvh_result = NULL;
    SG_file* pfi = NULL;
    SG_pathname* pPath_fi = NULL;
    SG_varray* pva_csid_list__branch_names = NULL;

    SG_varray* pva_all_audits = NULL;
    SG_varray* pva_all_comments = NULL;
    SG_varray* pva_all_tags = NULL;
    SG_varray* pva_all_users = NULL;

    SG_vhash* pvh_branch_names = NULL;
    SG_vhash* pvh_audits = NULL;
    SG_vhash* pvh_comments = NULL;
    SG_vhash* pvh_users = NULL;

    SG_vhash* pvh_csid_to_commit_mark = NULL;
    SG_vhash* pvh_blobid_to_mark = NULL;
    SG_uint32 next_mark = 1;
    SG_changeset* pcs = NULL;
    SG_string* pstr_repo_path = NULL;
    SG_string* pstr_branch_name = NULL;
    SG_vhash* pvh_items = NULL;
    SG_tncache* ptn_cache = NULL;

    SG_vhash* pvh_paths = NULL;
    SG_blobset* pbs = NULL;

    SG_NULLARGCHECK_RETURN(pOptSt);
    SG_NULLARGCHECK_RETURN(psz_repo);
    SG_NULLARGCHECK_RETURN(psz_fi);

    SG_ERR_CHECK(  SG_repo__open_repo_instance(pCtx, psz_repo, &pRepo)  );

    SG_ERR_CHECK(  SG_repo__list_blobs(pCtx, pRepo, 0, 0, 0, &pbs)  );

    SG_ERR_CHECK(  SG_repo__fetch_chrono_dagnode_list(pCtx, pRepo, SG_DAGNUM__VERSION_CONTROL, 0, 0, &count_nodes, &paszHids)  );
    
    SG_ERR_CHECK(  SG_vhash__alloc__params(pCtx, &pvh_result, count_nodes, NULL, NULL)  );
    SG_ERR_CHECK(  SG_varray__alloc__params(pCtx, &pva_csid_list__branch_names, count_nodes, NULL, NULL)  );

    SG_ERR_CHECK(  SG_repo__fetch_dagnodes__begin(pCtx, pRepo, SG_DAGNUM__VERSION_CONTROL, &pdh)  );
    pszHid = (const char*)paszHids;
    for (i = 0; i < count_nodes; i++)
    {
        SG_int32 gen = -1;
        SG_rbtree* prb_parents = NULL;
        SG_vhash* pvh_dn = NULL;

        SG_ERR_CHECK(  SG_repo__fetch_dagnodes__one(pCtx, pRepo, pdh, pszHid, &pdn)  );
        SG_ERR_CHECK(  SG_dagnode__get_generation(pCtx, pdn, &gen)  );
        SG_ERR_CHECK(  SG_dagnode__get_parents__rbtree_ref(pCtx, pdn, &prb_parents)  );

        SG_ERR_CHECK(  SG_varray__append__string__sz(pCtx, pva_csid_list__branch_names, pszHid)  );
        SG_ERR_CHECK(  SG_vhash__addnew__vhash(pCtx, pvh_result, pszHid, &pvh_dn)  );
        SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh_dn, "g", gen)  ); 
        if (prb_parents)
        {
            SG_varray* pva_parents = NULL;

            SG_ERR_CHECK(  SG_vhash__addnew__varray(pCtx, pvh_dn, "p", &pva_parents)  );
            SG_ERR_CHECK(  SG_rbtree__copy_keys_into_varray(pCtx, prb_parents, pva_parents)  );
        }

        pszHid += SG_HID_MAX_BUFFER_LENGTH;

        SG_DAGNODE_NULLFREE(pCtx, pdn);
    }
    SG_ERR_CHECK(  SG_repo__fetch_dagnodes__end(pCtx, pRepo, &pdh)  );
    SG_NULLFREE(pCtx, paszHids);

    //SG_VHASH_STDOUT(pvh_result);

    SG_ERR_CHECK(  SG_vc_branches__get_branch_names_for_given_changesets(pCtx, pRepo, &pva_csid_list__branch_names, &pvh_branch_names)  );
    SG_ERR_CHECK(  SG_user__list_all(pCtx, pRepo, &pva_all_users)  );
    SG_ERR_CHECK(  SG_audit__list_all(pCtx, pRepo, SG_DAGNUM__VERSION_CONTROL, &pva_all_audits)  );

    SG_ERR_CHECK(  SG_vc_comments__list_all(pCtx, pRepo, &pva_all_comments)  );
    SG_ERR_CHECK(  SG_vc_tags__list_all__with_history(pCtx, pRepo, &pva_all_tags)  );

    SG_ERR_CHECK(  x_get_users(pCtx, pva_all_users, &pvh_users)  );
    SG_ERR_CHECK(  x_get_comments(pCtx, pva_all_comments, &pvh_comments)  );
    SG_ERR_CHECK(  x_get_audits(pCtx, pva_all_audits, &pvh_audits)  );

    //SG_VHASH_STDOUT(pvh_branch_names);

    SG_ERR_CHECK(  SG_vhash__alloc__params(pCtx, &pvh_csid_to_commit_mark, count_nodes, NULL, NULL)  );
    SG_ERR_CHECK(  SG_vhash__alloc(pCtx, &pvh_blobid_to_mark)  );

    SG_ERR_CHECK(  SG_pathname__alloc__sz(pCtx, &pPath_fi, psz_fi)  );
	SG_ERR_CHECK(  SG_file__open__pathname(pCtx, pPath_fi, SG_FILE_WRONLY | SG_FILE_CREATE_NEW, 0644, &pfi)  );

    /*

       'commit' SP <ref> LF
       mark?
       ('author' (SP <name>)? SP LT <email> GT SP <when> LF)?
       'committer' (SP <name>)? SP LT <email> GT SP <when> LF
       data
       ('from' SP <committish> LF)?
       ('merge' SP <committish> LF)?
       (filemodify | filedelete | filecopy | filerename | filedeleteall | notemodify)*
       LF?


    */

    SG_ERR_CHECK(  SG_vhash__sort__vhash_field_int__asc(pCtx, pvh_result, "g")  );
    SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_result, &count_nodes)  );
    for (i_changeset = 1; i_changeset < count_nodes; i_changeset++)
    {
        SG_vhash* pvh_dn = NULL;
        const char* psz_csid = NULL;
        const char* psz_branch_name = NULL;
        const char* psz_comment = NULL;
        SG_vhash* pvh_a = NULL;
        SG_vhash* pvh_bn = NULL;
        SG_varray* pva_parents = NULL;
        SG_vhash* pvh_changes = NULL;
        const char* psz_root = NULL;
        SG_treenode* ptn_root = NULL;

        printf("Doing changeset %7d of %7d\n", i_changeset+1, count_nodes);

        SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvh_items)  );
        SG_ERR_CHECK(  SG_vhash__get_nth_pair__vhash(pCtx, pvh_result, i_changeset, &psz_csid, &pvh_dn)  );
        SG_ERR_CHECK(  SG_changeset__load_from_repo(pCtx, pRepo, psz_csid, &pcs)  );
        SG_ERR_CHECK(  SG_changeset__tree__get_changes(pCtx, pcs, &pvh_changes)  );
        SG_ERR_CHECK(  SG_changeset__tree__get_root(pCtx, pcs, &psz_root)  );

        SG_ERR_CHECK(  SG_tncache__alloc(pCtx, pRepo, &ptn_cache)  );
        SG_ERR_CHECK(  SG_tncache__load(pCtx, ptn_cache, psz_root, &ptn_root)  );

        if (pvh_changes)
        {
            SG_uint32 count_parents = 0;
            SG_uint32 i_parent = 0;

            SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_changes, &count_parents)  );
            for (i_parent=0; i_parent<count_parents; i_parent++)
            {
                const char* psz_csid_parent = NULL;
                SG_vhash* pvh_chgs_for_one_parent = NULL;
                SG_uint32 count_gids = 0;
                SG_uint32 i_gid = 0;

                SG_ERR_CHECK(  SG_vhash__get_nth_pair__vhash(pCtx, pvh_changes, i_parent, &psz_csid_parent, &pvh_chgs_for_one_parent)  );

                SG_ERR_CHECK(  SG_repo__treendx__get_paths(pCtx, pRepo, SG_DAGNUM__VERSION_CONTROL, pvh_chgs_for_one_parent, &pvh_paths)  );
                SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_chgs_for_one_parent, &count_gids)  );
                for (i_gid=0; i_gid<count_gids; i_gid++)
                {
                    const char* psz_gid = NULL;
                    const SG_variant* pv = NULL;
                    SG_uint32 count_paths = 0;
                    SG_treenode_entry_type t = SG_TREENODEENTRY_TYPE__INVALID;
                    SG_varray* pva_paths = NULL;

                    SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvh_chgs_for_one_parent, i_gid, &psz_gid, &pv)  );
                    SG_ERR_CHECK(  SG_vhash__get__varray(pCtx, pvh_paths, psz_gid, &pva_paths)  );
                    SG_ERR_CHECK(  SG_varray__count(pCtx, pva_paths, &count_paths)  );
                    if (1 == count_paths)
                    {
                        const char* psz_path = NULL;

                        SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pva_paths, 0, &psz_path)  );
                        SG_ERR_CHECK(  x_parse_prefixed_path(pCtx, psz_path, &pstr_repo_path, &t)  );
                    }
                    else
                    {
                        SG_uint32 i;
                        SG_bool b_always_a_dir = SG_TRUE;

                        for (i=0; i<count_paths; i++)
                        {
                            const char* psz_path = NULL;

                            SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pva_paths, i, &psz_path)  );
                            if ('D' != psz_path[0])
                            {
                                b_always_a_dir = SG_FALSE;
                                break;
                            }
                        }

                        if (b_always_a_dir)
                        {
                            t = SG_TREENODEENTRY_TYPE_DIRECTORY;
                            pstr_repo_path = NULL;
                        }
                        else
                        {
                            SG_ERR_CHECK(  x_which_path(pCtx, ptn_cache, ptn_root, pva_paths, psz_gid, &pstr_repo_path, &t)  );
                        }
                    }

                    if (SG_VARIANT_TYPE_NULL == pv->type)
                    {
                        // deleted

                        char buf[1024];

                        if (
                                (SG_TREENODEENTRY_TYPE_REGULAR_FILE == t)
                                || (SG_TREENODEENTRY_TYPE_SYMLINK == t)
                           )
                        {
                            SG_ERR_CHECK(  SG_sprintf(pCtx, buf, sizeof(buf), "D %s", 2 + SG_string__sz(pstr_repo_path))  );
                            SG_ERR_CHECK(  SG_vhash__update__null(pCtx, pvh_items, buf)  );
                        }
                        SG_STRING_NULLFREE(pCtx, pstr_repo_path);
                    }
                    else
                    {
                        SG_vhash* pvh_chgs_for_one_gid = NULL;
                        const char* psz_hid = NULL;

                        SG_ERR_CHECK(  SG_variant__get__vhash(pCtx, pv, &pvh_chgs_for_one_gid)  );

                        if (
                                (SG_TREENODEENTRY_TYPE_REGULAR_FILE == t)
                                || (SG_TREENODEENTRY_TYPE_SYMLINK == t)
                           )
                        {
                            SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pvh_chgs_for_one_gid, "hid", &psz_hid)  );


                            if (psz_hid)
                            {
                                SG_uint64 len_full = 0;
                                SG_uint64 len_check = 0;
                                SG_bool b_already = SG_FALSE;
                                SG_int64 my_mark = -1;
                                char buf[1024];

                                SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh_blobid_to_mark, psz_hid, &b_already)  );
                                if (b_already)
                                {
                                    SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvh_blobid_to_mark, psz_hid, &my_mark)  );
                                }
                                else
                                {
                                    SG_bool b_found = SG_FALSE;

                                    my_mark = next_mark++;
                                    SG_ERR_CHECK(  SG_file__write__format(pCtx, pfi, "blob\n")  );
                                    SG_ERR_CHECK(  SG_file__write__format(pCtx, pfi, "mark :%d\n", (int) my_mark)  );
                                    SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh_blobid_to_mark, psz_hid, my_mark)  );
                                    SG_ERR_CHECK(  SG_blobset__lookup(
                                                pCtx, 
                                                pbs, 
                                                psz_hid, 
                                                &b_found, 
                                                NULL, 
                                                NULL, 
                                                NULL, 
                                                NULL, 
                                                &len_full, 
                                                NULL,
                                                NULL, 
                                                NULL
                                                )  );
                                    SG_ERR_CHECK(  SG_file__write__format(pCtx, pfi, "data %d\n", (SG_uint32) len_full)  ); // TODO 64 bit int
                                    SG_ERR_CHECK(  SG_repo__fetch_blob_into_file(
                                                pCtx, 
                                                pRepo, 
                                                psz_hid, 
                                                pfi,
                                                &len_check
                                                )  );
                                    SG_ERR_CHECK(  SG_file__write__format(pCtx, pfi, "\n")  );
                                }

                                if (SG_TREENODEENTRY_TYPE_SYMLINK == t)
                                {
                                    SG_ERR_CHECK(  SG_sprintf(pCtx, buf, sizeof(buf), "M %s :%d %s", "120000", (int) my_mark, 2 + SG_string__sz(pstr_repo_path))  );
                                }
                                else
                                {
                                    // TODO 100755 for executables
                                    SG_ERR_CHECK(  SG_sprintf(pCtx, buf, sizeof(buf), "M %s :%d %s", "100644", (int) my_mark, 2 + SG_string__sz(pstr_repo_path))  );
                                }
                                SG_ERR_CHECK(  SG_vhash__update__null(pCtx, pvh_items, buf)  );
                            }
                        }
                        SG_STRING_NULLFREE(pCtx, pstr_repo_path);
                    }
                }
                SG_VHASH_NULLFREE(pCtx, pvh_paths);
            }
        }
        SG_ERR_CHECK(  SG_vhash__check__vhash(pCtx, pvh_branch_names, psz_csid, &pvh_bn)  );
        if (pvh_bn)
        {
            SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvh_bn, 0, &psz_branch_name, NULL)  );
        }

        if (psz_branch_name)
        {
            // hmmph.  the git-fast-import spec says that a branch name can
            // have spaces in it, but git itself barfs.

            SG_ERR_CHECK(  SG_string__alloc__sz(pCtx, &pstr_branch_name, psz_branch_name)  );
            SG_ERR_CHECK( SG_string__replace_bytes(
                pCtx,
                pstr_branch_name,
                (const SG_byte *)" ",1,
                (const SG_byte *)"_",1,
                SG_UINT32_MAX,SG_TRUE,
                NULL
                )  );
            SG_ERR_CHECK(  SG_file__write__format(pCtx, pfi, "commit refs/heads/%s\n", SG_string__sz(pstr_branch_name))  );
            SG_STRING_NULLFREE(pCtx, pstr_branch_name);
        }
        else
        {
            SG_ERR_CHECK(  SG_file__write__format(pCtx, pfi, "commit refs/heads/%s\n", "TODO")  );
        }
        SG_ERR_CHECK(  SG_file__write__format(pCtx, pfi, "mark :%d\n", next_mark)  );
        SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh_csid_to_commit_mark, psz_csid, next_mark)  );
        next_mark++;

        SG_ERR_CHECK(  SG_vhash__check__vhash(pCtx, pvh_audits, psz_csid, &pvh_a)  );
        if (pvh_a)
        {
            const char* psz_userid = NULL;
            const char* psz_username = NULL;
            SG_int64 timestamp = -1;

            SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh_a, "userid", &psz_userid)  );
            SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pvh_users, psz_userid, &psz_username)  );
            if (!psz_username)
            {
                psz_username = SG_NOBODY__USERNAME;
            }

            SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvh_a, "timestamp", &timestamp)  );
            SG_ERR_CHECK(  SG_file__write__format(pCtx, pfi, "committer <%s> %d +0000\n", psz_username, (SG_int32) (timestamp / 1000))  );
        }
        else
        {
            // TODO error no audit
        }

        SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pvh_comments, psz_csid, &psz_comment)  );
        if (psz_comment)
        {
            SG_ERR_CHECK(  SG_file__write__format(pCtx, pfi, "data %d\n", SG_STRLEN(psz_comment))  );
            SG_ERR_CHECK(  SG_file__write__format(pCtx, pfi, "%s\n", psz_comment)  );
        }
        else
        {
            SG_ERR_CHECK(  SG_file__write__format(pCtx, pfi, "data 0\n")  );
        }

        SG_ERR_CHECK(  SG_vhash__check__varray(pCtx, pvh_dn, "p", &pva_parents)  );
        if (pva_parents)
        {
            SG_uint32 count_parents = 0;
            SG_uint32 i_parent = 0;

            SG_ERR_CHECK(  SG_varray__count(pCtx, pva_parents, &count_parents)  );
            for (i_parent=0; i_parent<count_parents; i_parent++)
            {
                const char* psz_csid_parent = NULL;
                SG_int32 mark = -1;
                SG_bool b_has = SG_FALSE;

                SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pva_parents, i_parent, &psz_csid_parent)  );
                SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh_csid_to_commit_mark, psz_csid_parent, &b_has)  );
                if (b_has)
                {
                    SG_ERR_CHECK(  SG_vhash__get__int32(pCtx, pvh_csid_to_commit_mark, psz_csid_parent, &mark)  );
                    if (0 == i_parent)
                    {
                        SG_ERR_CHECK(  SG_file__write__format(pCtx, pfi, "from :%d\n", mark)  );
                    }
                    else
                    {
                        SG_ERR_CHECK(  SG_file__write__format(pCtx, pfi, "merge :%d\n", mark)  );
                    }
                }
            }
        }

        if (pvh_items)
        {
            SG_uint32 count = 0;
            SG_uint32 i = 0;

            SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_items, &count)  );
            for (i=0; i<count; i++)
            {
                const char* psz_line = NULL;

                SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvh_items, i, &psz_line, NULL)  );
                SG_ERR_CHECK(  SG_file__write__format(pCtx, pfi, "%s\n", psz_line)  );
            }
        }

        SG_VHASH_NULLFREE(pCtx, pvh_items);

        SG_ERR_CHECK(  SG_file__write__format(pCtx, pfi, "\n")  );

        SG_CHANGESET_NULLFREE(pCtx, pcs);
        SG_TNCACHE_NULLFREE(pCtx, ptn_cache);
    }

    SG_ERR_CHECK(  SG_varray__count(pCtx, pva_all_tags, &count_tags)  );
    for (i_tag = 1; i_tag < count_tags; i_tag++)
    {
        SG_vhash* pvh_tag = NULL;
        const char* psz_tag = NULL;
        const char* psz_csid = NULL;
        SG_int32 mark = -1;
        SG_int64 timestamp = 0;
        const char* psz_username = NULL;
        SG_vhash* pvh_a = NULL;
        const char* psz_userid = NULL;

        SG_ERR_CHECK(  SG_varray__get__vhash(pCtx, pva_all_tags, i_tag, &pvh_tag)  );
        SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh_tag, "tag", &psz_tag)  );
        SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh_tag, "csid", &psz_csid)  );
        SG_ERR_CHECK(  SG_vhash__get__int32(pCtx, pvh_csid_to_commit_mark, psz_csid, &mark)  );
        {
            SG_varray* pva_history = NULL;
            SG_vhash* pvh_history_1 = NULL;
            SG_varray* pva_audits = NULL;

            SG_ERR_CHECK(  SG_vhash__get__varray(pCtx, pvh_tag, SG_ZING_FIELD__HISTORY, &pva_history)  );
            SG_ERR_CHECK(  SG_varray__get__vhash(pCtx, pva_history, 0, &pvh_history_1)  );
            SG_ERR_CHECK(  SG_vhash__get__varray(pCtx, pvh_history_1, "audits", &pva_audits)  );
            SG_ERR_CHECK(  SG_varray__get__vhash(pCtx, pva_audits, 0, &pvh_a)  );
        }
        SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh_a, "userid", &psz_userid)  );
        SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pvh_users, psz_userid, &psz_username)  );
        if (!psz_username)
        {
            psz_username = SG_NOBODY__USERNAME;
        }

        SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvh_a, "timestamp", &timestamp)  );

        SG_ERR_CHECK(  SG_file__write__format(pCtx, pfi, "tag %s\n", psz_tag)  );
        SG_ERR_CHECK(  SG_file__write__format(pCtx, pfi, "from :%d\n", mark)  );
        SG_ERR_CHECK(  SG_file__write__format(pCtx, pfi, "tagger <%s> %d +0000\n", psz_username, (SG_int32) (timestamp / 1000))  );
        SG_ERR_CHECK(  SG_file__write__format(pCtx, pfi, "data 0\n")  );
        SG_ERR_CHECK(  SG_file__write__format(pCtx, pfi, "\n")  );
    }


fail:
    SG_ERR_IGNORE(  SG_file__close(pCtx, &pfi)  );
    SG_TNCACHE_NULLFREE(pCtx, ptn_cache);
    SG_BLOBSET_NULLFREE(pCtx, pbs);
    SG_STRING_NULLFREE(pCtx, pstr_branch_name);
    SG_VHASH_NULLFREE(pCtx, pvh_items);
    SG_STRING_NULLFREE(pCtx, pstr_repo_path);
    SG_PATHNAME_NULLFREE(pCtx, pPath_fi);
    SG_CHANGESET_NULLFREE(pCtx, pcs);
    SG_VHASH_NULLFREE(pCtx, pvh_csid_to_commit_mark);
    SG_VHASH_NULLFREE(pCtx, pvh_blobid_to_mark);
    SG_VHASH_NULLFREE(pCtx, pvh_result);
    SG_DAGNODE_NULLFREE(pCtx, pdn);
    SG_NULLFREE(pCtx, paszHids);
    SG_REPO_NULLFREE(pCtx, pRepo);

    SG_VHASH_NULLFREE(pCtx, pvh_branch_names);
    SG_VHASH_NULLFREE(pCtx, pvh_comments);
    SG_VHASH_NULLFREE(pCtx, pvh_audits);
    SG_VHASH_NULLFREE(pCtx, pvh_users);

    SG_VARRAY_NULLFREE(pCtx, pva_all_tags);
    SG_VARRAY_NULLFREE(pCtx, pva_all_audits);
    SG_VARRAY_NULLFREE(pCtx, pva_all_comments);
    SG_VARRAY_NULLFREE(pCtx, pva_all_users);
    SG_VHASH_NULLFREE(pCtx, pvh_paths);
}

/*

   filerename

   add comment lines with more info?

*/

