/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg__do_cmd_merge.c
 *
 * @details
 *
 */

//////////////////////////////////////////////////////////////////

#include <sg.h>
#include "sg_typedefs.h"
#include "sg_prototypes.h"

//////////////////////////////////////////////////////////////////

/**
 * Handle the MERGE command.
 *
 *
 */
void do_cmd_merge(SG_context * pCtx,
				  SG_rev_spec * pRevSpec,
				  const char * pszToolOverride,
				  SG_bool bVerbose,
				  SG_bool bTest,
				  const SG_vector * pvec_use_submodules)
{
	SG_string * pStrOutput = NULL;
	SG_wd_plan * pPlan = NULL;
	const SG_wd_plan_stats * pStats;
	SG_int64 nrUnresolvedIssues = 0;
    char * psz_other_csid = NULL;
	SG_vhash * pvhResults = NULL;
	SG_int64 status;

	// use the current directory to find the pending-tree and
	// determine the target changeset using --rev/--tag or the
	// unique other head of the tied branch.  WE DO NOT TAKE A
	// BRANCH ARGUMENT.

	SG_ERR_CHECK(  SG_vv_utils__rev_spec_to_csid_for_merge(pCtx, NULL,
														   pRevSpec,
														   &psz_other_csid)  );

	// try to do the merge using the desired cset.

	SG_ERR_CHECK(  SG_vv_verbs__merge(pCtx,
									  NULL,
									  psz_other_csid,
									  pszToolOverride,
									  bTest,
									  pvec_use_submodules,
									  &pvhResults,
									  &pPlan)  );
	SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvhResults, "status", &status)  );
	switch (status)
	{
	default:
		SG_ERR_THROW(  SG_ERR_NOTIMPLEMENTED  );
		break;
		
	case SG_VV_VERBS__STATUS__SUBMODULE_PROBLEM:
		{
			SG_varray * pvaMissingDetails;	// we don't own this

			SG_ERR_CHECK(  SG_vhash__get__varray(pCtx, pvhResults, "sub", &pvaMissingDetails)  );

			SG_ERR_IGNORE(  sg_report_submodule_missing_details(pCtx, pvaMissingDetails)  );
			SG_ERR_THROW(  SG_ERR_SUBMODULE_NESTED_UPDATE_CHECKOUT_PROBLEM  ); // throw so that exe exits with non-zero status
		}
		break;

	case SG_VV_VERBS__STATUS__PORTABILITY_WARNING:
		{
			SG_varray * pvaWarnings;	// we don't own this
			
			SG_ERR_CHECK(  SG_vhash__get__varray(pCtx, pvhResults, "port", &pvaWarnings)  );
			SG_ERR_IGNORE(  sg_report_portability_warnings2(pCtx, pvaWarnings)  );
			SG_ERR_THROW(  SG_ERR_PORTABILITY_WARNINGS  ); // throw so that exe exits with non-zero status
		}
		break;

	case SG_VV_VERBS__STATUS__OK:
		SG_ERR_CHECK(  SG_wd_plan__get_stats__ref(pCtx, pPlan, &pStats)  );
		if (bTest || bVerbose)
		{
			SG_varray * pva_wd_parents = NULL;
			const char * pszHidBaseline = NULL;
			const char * pszHidOther = NULL;

			// Do a pretty display of the wd_plan and unresolved issues.

			SG_ERR_CHECK(  SG_vhash__get__varray(pCtx, pvhResults, "parents", &pva_wd_parents)  );
			SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pva_wd_parents, 0, &pszHidBaseline)  );
			SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pva_wd_parents, 1, &pszHidOther)  );
			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDOUT, "Merging changeset: %s\n", pszHidOther)  );
			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDOUT, "    into baseline: %s\n", pszHidBaseline)  );

			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDOUT, "Steps required to perform the merge:\n")  );
			SG_ERR_CHECK(  SG_wd_plan__format__to_string(pCtx, pPlan, &pStrOutput)  );
			SG_ERR_IGNORE(  SG_console__raw(pCtx, SG_CS_STDOUT, SG_string__sz(pStrOutput))  );
			SG_ERR_IGNORE(  SG_console__flush(pCtx, SG_CS_STDOUT)  );
		}
		SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvhResults, "nr_unresolved_issues", &nrUnresolvedIssues)  );
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDOUT,
								   ("%d updated, %d deleted, %d added, %d merged, %d unresolved\n"),
								   pStats->nrDirsChanged + pStats->nrFilesChanged + pStats->nrSubmodulesChanged,
								   pStats->nrDirsDeleted + pStats->nrFilesDeleted + pStats->nrSubmodulesDeleted,
								   pStats->nrDirsAdded   + pStats->nrFilesAdded   + pStats->nrSubmodulesAdded,
								   pStats->nrFilesAutoMerged,
								   (SG_uint32)nrUnresolvedIssues)  );
		break;
	}

fail:
	SG_STRING_NULLFREE(pCtx, pStrOutput);
	SG_VHASH_NULLFREE(pCtx, pvhResults);
	SG_WD_PLAN_NULLFREE(pCtx, pPlan);
	SG_NULLFREE(pCtx, psz_other_csid);
}
