/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg__do_cmd_resolve.c
 *
 * @details
 *
 */

//////////////////////////////////////////////////////////////////

#include <sg.h>
#include "sg_typedefs.h"
#include "sg_prototypes.h"

//////////////////////////////////////////////////////////////////

struct _resolve_data
{
	SG_pathname *		pPathCwd;
	SG_pendingtree *	pPendingTree;
	SG_stringarray *	psaGids;			// list of the GIDs of the issues that we want to operate on.
	SG_bool				bForceContentReMerge;
	const char *		pszUseTool;
};

//////////////////////////////////////////////////////////////////

static void _resolve__lookup_issue(SG_context * pCtx,
								   struct _resolve_data * pData,
								   const char * pszGid,
								   const SG_vhash ** ppvhIssue)
{
	const SG_vhash * pvhIssue;
	SG_bool bFound;

	if (!pData->pPendingTree)		// previous iteration probably freed it after a SAVE
		SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pData->pPathCwd, &pData->pPendingTree)  );
			
	SG_ERR_CHECK(  SG_pendingtree__find_wd_issue_by_gid(pCtx, pData->pPendingTree, pszGid, &bFound, &pvhIssue)  );
	if (!bFound)
	{
		// Since we may have released the VFILE lock during the outer loop,
		// this could technically happen but only if they did something
		// like 'vv revert --all' in another shell while the external merge
		// tool was running.
		//
		// TODO 2010/07/11 I don't like this error message, but I doubt anybody
		// TODO            will ever see it.  Think about taking the GID and
		// TODO            finding a repo-path for the item using just the
		// TODO            PTNODEs and printing it.

		SG_ERR_THROW2(  SG_ERR_ASSERT,
						(pCtx, "RESOLVE failed to find ISSUE for GID %s.", pszGid)  );
	}

	*ppvhIssue = pvhIssue;

fail:
	return;
}

/**
 * We need to check/recheck the RESOLVED/UNRESOLVED status each time
 * after we release the VFILE lock (because the user could have done
 * another RESOLVE in another shell window while the external merge
 * tool was running).
 */
static void _resolve__is_resolved(SG_context * pCtx,
								  const SG_vhash * pvhIssue,
								  SG_bool * pbIsResolved)
{
	SG_int64 s64;
	SG_pendingtree_wd_issue_status status;

	SG_ERR_CHECK_RETURN(  SG_vhash__get__int64(pCtx, pvhIssue, "status", &s64)  );
	status = (SG_pendingtree_wd_issue_status)s64;

	*pbIsResolved = ((status & SG_ISSUE_STATUS__MARKED_RESOLVED) == SG_ISSUE_STATUS__MARKED_RESOLVED);
}

//////////////////////////////////////////////////////////////////

/**
 * Print the detailed output for this ISSUE.
 * This is called by "vv resolve --list" and
 * "vv resolve --fix" prior to making any changes.
 * 
 * Optionally return the current repo-path of the entry.
 */
static void _resolve__list(SG_context * pCtx,
						   struct _resolve_data * pData,
						   const SG_vhash * pvhIssue,
						   SG_string ** ppStrRepoPath)
{
	SG_string * pStrOutput = NULL;

	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pStrOutput)  );
	SG_ERR_CHECK(  SG_pendingtree__format_issue(pCtx, pData->pPendingTree, pvhIssue, pStrOutput, ppStrRepoPath)  );
	SG_ERR_IGNORE(  SG_console__raw(pCtx, SG_CS_STDOUT, SG_string__sz(pStrOutput))  );

fail:
	SG_STRING_NULLFREE(pCtx, pStrOutput);
}

/**
 * Print detailed info for all of the ISSUES in the array.
 * For a "vv resolve --list" this is the set of unresolved
 * issues; for a "vv resolve --list-all" this is the complete
 * set of all issues.
 */
static void _resolve__do_list(SG_context * pCtx, struct _resolve_data * pData)
{
	SG_uint32 k, kLimit;

	SG_ERR_CHECK(  SG_stringarray__count(pCtx, pData->psaGids, &kLimit)  );
	for (k=0; k<kLimit; k++)
	{
		const char * pszGid_k;
		const SG_vhash * pvhIssue_k;
		SG_bool bFound;

		SG_ERR_CHECK(  SG_stringarray__get_nth(pCtx, pData->psaGids, k, &pszGid_k)  );
		SG_ERR_CHECK(  SG_pendingtree__find_wd_issue_by_gid(pCtx, pData->pPendingTree, pszGid_k, &bFound, &pvhIssue_k)  );
		if (!bFound)
		{
			// This should never happen because we are still holding the
			// pendingtree lock and still have the original pendingtree
			// structure in memory.

			SG_ERR_THROW2(  SG_ERR_ASSERT,
							(pCtx, "RESOLVE failed to find ISSUE for GID %s.", pszGid_k)  );
		}

		SG_ERR_CHECK(  _resolve__list(pCtx, pData, pvhIssue_k, NULL)  );
	}

fail:
	return;
}

//////////////////////////////////////////////////////////////////

/**
 * Update the RESOLVED/UNRESOLVED status for this ISSUE and do an
 * incremental save on the pendingtree.
 */
static void _resolve__mark(SG_context * pCtx,
						   struct _resolve_data * pData,
						   const SG_vhash * pvhIssue)
{
	SG_int64 s;
	SG_pendingtree_wd_issue_status status;

	SG_ERR_CHECK_RETURN(  SG_vhash__get__int64(pCtx, pvhIssue, "status", &s)  );
	status = (SG_pendingtree_wd_issue_status)s;

	status |= SG_ISSUE_STATUS__MARKED_RESOLVED;

	// update the status on the ISSUE and save the pendingtree now.
	// since this trashes stuff within in it, go ahead and free it
	// so no one trips over the trash.

	SG_ERR_CHECK_RETURN(  SG_pendingtree__set_wd_issue_status(pCtx, pData->pPendingTree, pvhIssue, status)  );
	SG_PENDINGTREE_NULLFREE(pCtx, pData->pPendingTree);
}

static void _resolve__do_mark_1(SG_context * pCtx,
								struct _resolve_data * pData,
								const char * pszGid)
{
	const SG_vhash * pvhIssue;

	SG_ERR_CHECK(  _resolve__lookup_issue(pCtx, pData, pszGid, &pvhIssue)  );
	SG_ERR_CHECK(  _resolve__mark(pCtx, pData, pvhIssue)  );

fail:
	return;
}

/**
 * Update the RESOLVED/UNRESOLVED status for each ISSUE in the array
 * and then SAVE the pendingtree.
 *
 * TODO 2010/09/28 This saves the pendingtree after *each* issue.
 * TODO            Maybe think about having a _dont_save version of _1 and
 * TODO            doing a batch.
 */
static void _resolve__do_mark(SG_context * pCtx, struct _resolve_data * pData)
{
	SG_uint32 k, kLimit;

	SG_ERR_CHECK(  SG_stringarray__count(pCtx, pData->psaGids, &kLimit)  );
	for (k=0; k<kLimit; k++)
	{
		const char * pszGid_k;

		SG_ERR_CHECK(  SG_stringarray__get_nth(pCtx, pData->psaGids, k, &pszGid_k)  );
		SG_ERR_CHECK(  _resolve__do_mark_1(pCtx, pData, pszGid_k)  );
	}

fail:
	return;
}

//////////////////////////////////////////////////////////////////

enum _fix_status
{
	FIX_ZERO = 0,
	FIX_USER_MERGED = 1,
	FIX_ALREADY_MERGED = 2,
	FIX_MERGETOOL_ERROR = 3,
	FIX_MERGE_NOT_ATTEMPTED = 4,
};

/**
 * MERGE left the TEMP files for 3 inputs in the TEMP DIR and
 * put REPO-PATHS for each of them in the ISSUE.  It also
 * included a REPO-PATH for the RESULT-FILE.
 *
 * Re-create absolute pathnames for all for files and return them.
 *
 * You own all of them.
 */
static void _resolve__fetch_temp_pathnames(SG_context * pCtx,
										   struct _resolve_data * pData,
										   const SG_vhash * pvhIssue,
										   SG_pathname ** ppPathAncestor,
										   SG_pathname ** ppPathBaseline,
										   SG_pathname ** ppPathOther,
										   SG_pathname ** ppPathResult)
{
	SG_pathname * pPathAncestor = NULL;
	SG_pathname * pPathBaseline = NULL;
	SG_pathname * pPathOther = NULL;
	SG_pathname * pPathResult = NULL;
	const SG_pathname * pPathWorkingDirTop;
	const char * pszRepoPath;

	SG_ERR_CHECK(  SG_pendingtree__get_working_directory_top__ref(pCtx, pData->pPendingTree, &pPathWorkingDirTop)  );

	SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhIssue, "repopath_tempfile_ancestor", &pszRepoPath)  );
	SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path(pCtx, pPathWorkingDirTop, pszRepoPath, &pPathAncestor)  );

	SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhIssue, "repopath_tempfile_baseline", &pszRepoPath)  );
	SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path(pCtx, pPathWorkingDirTop, pszRepoPath, &pPathBaseline)  );

	SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhIssue, "repopath_tempfile_other", &pszRepoPath)  );
	SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path(pCtx, pPathWorkingDirTop, pszRepoPath, &pPathOther)  );

	SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhIssue, "repopath_tempfile_result", &pszRepoPath)  );
	SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path(pCtx, pPathWorkingDirTop, pszRepoPath, &pPathResult)  );

	// TODO 2010/09/29 Verify that the TEMP DIR and the 3 input files still
	// TODO            exist on disk. Maybe silently re-create them.

	*ppPathAncestor = pPathAncestor;
	*ppPathBaseline = pPathBaseline;
	*ppPathOther    = pPathOther;
	*ppPathResult   = pPathResult;
	return;

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathAncestor);
	SG_PATHNAME_NULLFREE(pCtx, pPathBaseline);
	SG_PATHNAME_NULLFREE(pCtx, pPathOther);
	SG_PATHNAME_NULLFREE(pCtx, pPathResult);
}

static void _resolve__fetch_labels(SG_context * pCtx,
								   const SG_vhash * pvhIssue,
								   SG_string ** ppStringAncestorLabel,
								   SG_string ** ppStringBaselineLabel,
								   SG_string ** ppStringOtherLabel,
								   SG_string ** ppStringResultLabel)
{
	const char * psz;
	SG_string * pStringAncestorLabel = NULL;
	SG_string * pStringBaselineLabel = NULL;
	SG_string * pStringOtherLabel = NULL;
	SG_string * pStringResultLabel = NULL;

	SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhIssue, "label_ancestor", &psz)  );
	SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &pStringAncestorLabel, psz)  );

	SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhIssue, "label_baseline", &psz)  );
	SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &pStringBaselineLabel, psz)  );

	SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhIssue, "label_other", &psz)  );
	SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &pStringOtherLabel, psz)  );

	SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhIssue, "label_result", &psz)  );
	SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &pStringResultLabel, psz)  );

	*ppStringAncestorLabel = pStringAncestorLabel;
	*ppStringBaselineLabel = pStringBaselineLabel;
	*ppStringOtherLabel    = pStringOtherLabel;
	*ppStringResultLabel   = pStringResultLabel;
	return;

fail:
	SG_STRING_NULLFREE(pCtx, pStringAncestorLabel);
	SG_STRING_NULLFREE(pCtx, pStringBaselineLabel);
	SG_STRING_NULLFREE(pCtx, pStringOtherLabel);
	SG_STRING_NULLFREE(pCtx, pStringResultLabel);
}

static void _resolve__fix__set_attrbits_after_content_merge(SG_context * pCtx,
															const SG_vhash * pvhIssue_In,
															SG_mrg_cset_entry_conflict_flags conflict_flags,
															const SG_pathname * pPathFinal,
															const SG_pathname * pPathBackup)
{
	SG_int64 attrbits;
	SG_bool bHave_Previously_Fixed_Structure = SG_FALSE;

	// A content merge was successful (meaning a "merge-tool" was successful
	// and created a "result" file in TEMP).  Our caller MOVED this file to
	// the correct location in the WD.  The merge-tool is not responsible for
	// setting attrbits or xattrs (since it could be *any* external editor).
	//
	// We must set the appropriate attrbits on the newly created file (which
	// just happens to now be in the WD).

	// Get the value that MERGE thought was appropriate.  These are either the
	// final result (when no conflict) or (probably) the baseline/interim value.

	SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvhIssue_In, "attrbits", &attrbits)  );

	// If there was a attrbit conflict *and* a previous RESOLVE already
	// asked the user what the final result should be (and set it on the
	// previous merge-result file), we should be able to get the correct
	// attrbit value from the live backup file and not have to bother
	// the user again.

	if (conflict_flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_ATTRBITS)
	{
		SG_ERR_CHECK(  SG_vhash__has(pCtx, pvhIssue_In, "resolve_structural_fixed", &bHave_Previously_Fixed_Structure)  );
		if (bHave_Previously_Fixed_Structure)
		{
			SG_ERR_CHECK(  SG_vhash__get__bool(pCtx, pvhIssue_In, "resolve_structural_fixed", &bHave_Previously_Fixed_Structure)  );
			if (bHave_Previously_Fixed_Structure)
			{
				// This was a --REMERGE after a successful RESOLVE (or a --MARK).
				// So everything in the WD (content, attrbits, xattrs) was set to
				// something (and the user answered any "which attr" questions)
				// but the user wanted to re-do the content merge.

				SG_ERR_CHECK(  SG_attributes__bits__read(pCtx, pPathBackup, attrbits, &attrbits)  );
			}
		}
	}

	SG_ERR_CHECK(  SG_attributes__bits__apply(pCtx, pPathFinal, attrbits)  );

	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDOUT,
							   "RESOLVE: Set file attributes to '%d' on merge result.\n",
							   (SG_uint32)attrbits)  );

fail:
	return;
}
	
/**
 * Release VFILE lock and invoke the proper mergetool for this file.
 *
 * This should set *pFixContentStatus to one of:
 *     FIX_USER_MERGED,
 *     FIX_ALREADY_MERGED, or
 *     FIX_MERGETOOL_ERROR.
 *
 */
static void _resolve__fix__merge_content_if_necessary(SG_context * pCtx,
													  struct _resolve_data * pData,
													  const char * pszGid,
													  const SG_vhash * pvhIssue_In,
													  SG_string * pStrRepoPath,
													  SG_mrg_cset_entry_conflict_flags conflict_flags,
													  enum _fix_status * pFixContentStatus)
{
	SG_pathname * pPathAncestor = NULL;
	SG_pathname * pPathBaseline = NULL;
	SG_pathname * pPathOther = NULL;
	SG_pathname * pPathResultInTemp = NULL;
	SG_pathname * pPathFinalResult = NULL;
	SG_pathname * pPathBackup = NULL;
	const char * szToolName = NULL;
	SG_filetool * pTool = NULL;
	SG_string * pStringAncestorLabel = NULL;
	SG_string * pStringBaselineLabel = NULL;
	SG_string * pStringOtherLabel = NULL;
	SG_string * pStringResultLabel = NULL;
	const SG_vhash * pvhIssue_Out;
	SG_int32 statusMergeToolResult;
	SG_bool bHas_Resolve_Status = SG_FALSE;
	SG_bool bPreviouslyMerged = SG_FALSE;
	SG_repo * pRepo;
	SG_varray * pvaExternal = NULL;

	// We have 2 types of status fields.  the __DIVERGENT_FILE_EDIT__ flags
	// in "conflict_flags" which get set during the MERGE (and never changed).
	// And an optional resolve-mergetool status (from a previous RESOLVE).
	//
	// If a resolve-merge has been attempted, the issue should have been updated
	// with the status.
	//
	// If there is a resolve-merge status, we need to disregard the status
	// in the conflict_flags (since we don't know what state the result-file
	// is in.

	SG_ERR_CHECK(  SG_vhash__has(pCtx, pvhIssue_In, "resolve_mergetool_status", &bHas_Resolve_Status)  );
	if (bHas_Resolve_Status)
	{
		SG_int64 i64;
		SG_int32 statusPreviousResolve;

		SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvhIssue_In, "resolve_mergetool_status", &i64)  );
		statusPreviousResolve = (SG_int32)i64;

		bPreviouslyMerged = (statusPreviousResolve == SG_MERGETOOL__STATUS__SUCCESSFUL);
	}
	else
	{
		SG_ASSERT(  (conflict_flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__MASK)  );

		bPreviouslyMerged = ((conflict_flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__AUTO_OK) != 0);
	}

	if (bPreviouslyMerged && !pData->bForceContentReMerge)
	{
		*pFixContentStatus = FIX_ALREADY_MERGED;
		goto done;
	}

	// reconstruct the pathnames to the TEMP files that MERGE left for us

	SG_ERR_CHECK(  _resolve__fetch_temp_pathnames(pCtx, pData, pvhIssue_In,
												  &pPathAncestor, &pPathBaseline, &pPathOther,
												  &pPathResultInTemp)  );

	// get gui-friendly labels for each panel.

	SG_ERR_CHECK(  _resolve__fetch_labels(pCtx, pvhIssue_In,
										  &pStringAncestorLabel, &pStringBaselineLabel, &pStringOtherLabel, &pStringResultLabel)  );

	// determine the name of the mergetool to use.  if they gave us one,
	// use it.  if not, use the suffix on the ancestor version of the
	// file to pick one.

	SG_ERR_CHECK(  SG_pendingtree__get_repo(pCtx, pData->pPendingTree, &pRepo)  );
	SG_ERR_CHECK(  SG_mergetool__select(pCtx,
										SG_MERGETOOL__CONTEXT__RESOLVE,
										NULL,
										pPathAncestor,
										pData->pszUseTool,
										pRepo,
										&pTool)  );
	if (!pTool)
		SG_ERR_THROW2(  SG_ERR_NOT_FOUND,
						(pCtx, "Could not determine the proper MERGETOOL for '%s'.", SG_string__sz(pStrRepoPath))  );

	//////////////////////////////////////////////////////////////////
	// Free the PENDINGTREE so that we release the VFILE lock for
	// the duration of the external tool.

	pvhIssue_In = NULL;
	SG_PENDINGTREE_NULLFREE(pCtx, pData->pPendingTree);

	//////////////////////////////////////////////////////////////////
	// Invoke the external tool.  We set pPathResultInTemp to a TEMP file
	// in the TEMP dir with the 3 input files.  This way if the user
	// does incremental saves and aborts the merge or the tool has
	// an error, we haven't altered anything in the WD.

	SG_ERR_CHECK(  SG_filetool__get_name(pCtx, pTool, &szToolName)  );
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDOUT,
							   "RESOLVE: Starting mergetool '%s' for '%s'...\n",
							   szToolName,
							   SG_string__sz(pStrRepoPath))  );

	SG_ERR_CHECK(  SG_mergetool__invoke(pCtx, pTool,
										pPathAncestor, pStringAncestorLabel,
										pPathBaseline, pStringBaselineLabel,
										pPathOther,    pStringOtherLabel,
										pPathResultInTemp, pStringResultLabel,
										&statusMergeToolResult)  );

	//////////////////////////////////////////////////////////////////

	if (statusMergeToolResult == SG_MERGETOOL__STATUS__SUCCESSFUL)
	{
		// move the merge-result from TEMP to where it should currently
		// be in the WD.  that is, in step [1] we update the content;
		// in step [2] we deal with any divergent MOVES/RENAMES or other
		// structural conflicts.
		//
		// if there is already a file in the destination back it up.
		// this may be the original baseline version or the auto-merge
		// result from the MERGE or the result from a previous RESOLVE
		// attempt.  in any case, back it up.

		const SG_pathname * pPathWorkingDirTop;
		SG_bool bExists;

		// re-open the pendingtree and get the VFILE lock and lookup this ISSUE.

		SG_ERR_CHECK(  _resolve__lookup_issue(pCtx, pData, pszGid, &pvhIssue_Out)  );

		SG_ERR_CHECK(  SG_pendingtree__get_working_directory_top__ref(pCtx,
																	  pData->pPendingTree,
																	  &pPathWorkingDirTop)  );
		SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path(pCtx,
																			 pPathWorkingDirTop,
																			 SG_string__sz(pStrRepoPath),
																			 &pPathFinalResult)  );
		SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pPathFinalResult, &bExists, NULL, NULL)  );
		if (bExists)
		{
			SG_ERR_CHECK(  SG_workingdir__calc_backup_name(pCtx, pPathFinalResult, &pPathBackup)  );
			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDOUT,
									   "RESOLVE: Creating backup file: '%s'\n",
									   SG_pathname__sz(pPathBackup))  );
			SG_ERR_CHECK(  SG_fsobj__move__pathname_pathname(pCtx, pPathFinalResult, pPathBackup)  );
		}
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDOUT,
								   "RESOLVE: Installing merge result in working copy: '%s'\n",
								   SG_pathname__sz(pPathFinalResult))  );
		SG_ERR_CHECK(  SG_fsobj__move__pathname_pathname(pCtx, pPathResultInTemp, pPathFinalResult)  );

		// figure out what attrbits and/or xattrs should be on the final result file in the WD.

		SG_ERR_CHECK(  _resolve__fix__set_attrbits_after_content_merge(pCtx, pvhIssue_Out, conflict_flags, pPathFinalResult, pPathBackup)  );
#ifdef SG_BUILD_FLAG_FEATURE_XATTR
		// TODO 2010/11/28 Deal with XATTRS.
		// SG_ERR_CHECK(  _resolve__fix__set_xattrs_after_content_merge(pCtx, pData, pvhIssue_Out, conflict_flags, pPathFinalResult, pPathBackup)  );
#endif

		// update the resolve_mergetool fields in the issue and save the
		// pendingtree, so once again our pendingtree/pvhIssue variables
		// are trashed after we do this.
		//
		// we DO NOT set these fields when != __SUCCESSFUL because we don't
		// copy the possibly incomplete or missing result file to the WD.
		// so if there was a previous resolve which was successful, we want
		// to keep it.

		SG_ERR_CHECK(  SG_pendingtree__set_wd_issue_resolve_mergetool_status(pCtx, pData->pPendingTree,
																			 pvhIssue_Out,
																			 szToolName,
																			 statusMergeToolResult)  );
		SG_PENDINGTREE_NULLFREE(pCtx, pData->pPendingTree);
		pvhIssue_Out = NULL;

		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDOUT,
								   "RESOLVE: File merge was successful using mergetool '%s'.\n",
								   szToolName)  );

		// we don't update the overall (resolved/unresolved) "status" for this issue here
		// because the caller may have to deal with part [2].

		*pFixContentStatus = FIX_USER_MERGED;
	}
	else if (statusMergeToolResult == SG_MERGETOOL__STATUS__CONFLICT)
	{
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDOUT,
								   "RESOLVE: The mergetool '%s' found conflicts; working copy version not changed.\n",
								   szToolName)  );
		*pFixContentStatus = FIX_MERGETOOL_ERROR;
	}
	else if (statusMergeToolResult == SG_MERGETOOL__STATUS__ERROR)
	{
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDOUT,
								   "RESOLVE: The mergetool '%s' exited with an error; working copy version not changed.\n",
								   szToolName)  );
		*pFixContentStatus = FIX_MERGETOOL_ERROR;
	}
	else if (statusMergeToolResult == SG_MERGETOOL__STATUS__NOT_ATTEMPTED)
	{
//		// re-open the pendingtree and get the VFILE lock and lookup this ISSUE.
//
//		SG_ERR_CHECK(  _resolve__lookup_issue(pCtx, pData, pszGid, &pvhIssue_Out)  );
//
//		// This is somewhat of a lie -- we successfully didn't do anything.
//		// We didn't attempt the merge; we didn't create the temp-result-file;
//		// we didn't move it to the WD; we didn't overwrite whatever is in the WD now.
//		//
//		// All we do here is update the edit-status.
//
//		SG_ERR_CHECK(  SG_pendingtree__set_wd_issue_resolve_mergetool_status(pCtx, pData->pPendingTree,
//																			 pvhIssue_Out,
//																			 SG_string__sz(pStringToolName),
//																			 statusMergeToolResult)  );
//		SG_PENDINGTREE_NULLFREE(pCtx, pData->pPendingTree);
//		pvhIssue_Out = NULL;

		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDOUT,
								   "RESOLVE: The mergetool '%s' did not attempt to merge the files; working copy version not changed.\n",
								   szToolName)  );
		*pFixContentStatus = FIX_MERGE_NOT_ATTEMPTED;
	}
	else if (statusMergeToolResult == SG_MERGETOOL__STATUS__EXEC_ERROR)
	{
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDOUT,
								   "RESOLVE: The mergetool '%s' could not be launched; working copy version not changed.\n",
								   szToolName)  );
		*pFixContentStatus = FIX_MERGETOOL_ERROR;
	}
	else
	{
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDOUT,
								   "RESOLVE: The mergetool '%s' did not create the expected result file; working copy version not changed.\n",
								   szToolName)  );
		*pFixContentStatus = FIX_MERGETOOL_ERROR;
	}

done:
	;
fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathAncestor);
	SG_PATHNAME_NULLFREE(pCtx, pPathBaseline);
	SG_PATHNAME_NULLFREE(pCtx, pPathOther);
	SG_PATHNAME_NULLFREE(pCtx, pPathResultInTemp);
	SG_FILETOOL_NULLFREE(pCtx, pTool);
	SG_PATHNAME_NULLFREE(pCtx, pPathFinalResult);
	SG_PATHNAME_NULLFREE(pCtx, pPathBackup);
	SG_STRING_NULLFREE(pCtx, pStringAncestorLabel);
	SG_STRING_NULLFREE(pCtx, pStringBaselineLabel);
	SG_STRING_NULLFREE(pCtx, pStringOtherLabel);
	SG_STRING_NULLFREE(pCtx, pStringResultLabel);
	SG_VARRAY_NULLFREE(pCtx, pvaExternal);
}

static void _resolve__fix__redo(SG_context * pCtx,
								struct _resolve_data * pData,
								const char * pszGid,
								const SG_vhash * pvhIssue,
								SG_string * pStrRepoPath,
								SG_mrg_cset_entry_conflict_flags conflict_flags,
								enum _fix_status * pFixContentStatus)
{
	SG_bool bIsResolved;
	SG_bool bHaveFileContentConflict;
	
	SG_ERR_CHECK(  _resolve__is_resolved(pCtx, pvhIssue, &bIsResolved)  );
	SG_ASSERT(  (bIsResolved)  );
	
	// we have completely resolved the issue, but yet we're back here.
	// the only REDO we support is --remerge on the content.

	bHaveFileContentConflict = ((conflict_flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__MASK) != 0);

	if (!pData->bForceContentReMerge || !bHaveFileContentConflict)
	{
		// Skip the issue if it is already resolved.  In theory, we should not
		// get this (because we filtered the pData->psaGids by status as we
		// parsed the command line arguments), but if they did another resolve
		// in another shell while we didn't have the lock, it could happen.

		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDOUT,
								   "Issue already resolved; nothing to be done for '%s'.\n",
								   SG_string__sz(pStrRepoPath))  );

		*pFixContentStatus = FIX_ALREADY_MERGED;
		return;
	}

	// they want to re-do the file content merge. (either auto-merge
	// was successful or they did a previous RESOLVE and successfully
	// merge the file content and now want to retry it.)

	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDOUT,
							   "Issue already resolved; re-merging file content for '%s'.\n",
							   SG_string__sz(pStrRepoPath))  );

	// do just the re-merge and don't worry about any structual issues
	// because they should have already been addressed.

	SG_ERR_CHECK(  _resolve__fix__merge_content_if_necessary(pCtx, pData, pszGid,
															 pvhIssue, pStrRepoPath,
															 conflict_flags,
															 pFixContentStatus)  );
	pvhIssue = NULL;

fail:
	return;
}

//////////////////////////////////////////////////////////////////

/**
 * Try to resolve structural problems for this ISSUE which
 * includes one branch deleting the entry and the other
 * branch altering it in some way.
 *
 * (If it was a delete vs unaltered, it wouldn't be a conflict
 * now would it.)
 *
 * The merge-apply code temporarily un-deleted the entry and
 * may have given it a temporary name/location.
 *
 * Ask the user if they want to delete it or take the altered
 * version.
 * 
 * TODO 2010/07/14 For a 2-way merge we cannot have both
 * TODO            DELETE-vs-x *AND* DIVERGENT_x
 * TODO            problems (because there are only 2 terms).
 * TODO            However, when we fully support n-way merges, 
 * TODO            we have to allow for both at the same time
 * TODO            (since we may have 3 or more terms).
 */
static void _resolve__fix__structural__delete(SG_context * pCtx,
											  struct _resolve_data * pData,
											  const char * pszGid,
											  const SG_vhash * pvhIssue,
											  SG_string * pStrRepoPath,
											  SG_mrg_cset_entry_conflict_flags conflict_flags,
											  SG_bool bCollisions,
											  SG_portability_flags portability_flags,
											  SG_bool * pbDidFix)
{
	SG_bool bDidFix = SG_FALSE;

	SG_UNUSED( pszGid );
	SG_UNUSED( conflict_flags );
	SG_UNUSED( bCollisions );
	SG_UNUSED( portability_flags );

	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDOUT,
							   "TODO _resolve__fix: handle DELETE-vs-x structural issues for '%s'.  For now, fix manually and use 'vv resolve --mark'.\n",
							   SG_string__sz(pStrRepoPath))  );

	if (bDidFix)
	{
		SG_ERR_CHECK(  SG_pendingtree__set_wd_issue_resolve_structural_status(pCtx, pData->pPendingTree,
																			  pvhIssue, SG_TRUE)  );
		SG_PENDINGTREE_NULLFREE(pCtx, pData->pPendingTree);
		pvhIssue = NULL;
	}
	
	*pbDidFix = bDidFix;

fail:
	return;
}

/**
 * Try to resolve non-delete-related structural problems
 * for this ISSUE.
 */
static void _resolve__fix__structural__non_delete(SG_context * pCtx,
												  struct _resolve_data * pData,
												  const char * pszGid,
												  const SG_vhash * pvhIssue,
												  SG_string * pStrRepoPath,
												  SG_mrg_cset_entry_conflict_flags conflict_flags,
												  SG_bool bCollisions,
												  SG_portability_flags portability_flags,
												  SG_bool * pbDidFix)
{
	SG_bool bDidFix = SG_FALSE;

	SG_UNUSED( pszGid );
	SG_UNUSED( conflict_flags );
	SG_UNUSED( bCollisions );
	SG_UNUSED( portability_flags );

	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDOUT,
							   "TODO _resolve__fix: handle non-delete structural issues for '%s'.  For now, fix manually and use 'vv resolve --mark'.\n",
							   SG_string__sz(pStrRepoPath))  );

	if (bDidFix)
	{
		SG_ERR_CHECK(  SG_pendingtree__set_wd_issue_resolve_structural_status(pCtx, pData->pPendingTree,
																			  pvhIssue, SG_TRUE)  );
		SG_PENDINGTREE_NULLFREE(pCtx, pData->pPendingTree);
		pvhIssue = NULL;
	}
	
	*pbDidFix = bDidFix;

fail:
	return;
}

static void _resolve__fix__structural(SG_context * pCtx,
									  struct _resolve_data * pData,
									  const char * pszGid,
									  const SG_vhash * pvhIssue,
									  SG_string * pStrRepoPath,
									  SG_mrg_cset_entry_conflict_flags conflict_flags,
									  SG_bool * pbDidFix)
{
	SG_bool bHave_Previously_Fixed_Structure = SG_FALSE;
	SG_portability_flags portability_flags;
	SG_bool bCollisions = SG_FALSE;
	SG_int64 i64;

	SG_ERR_CHECK(  SG_vhash__has(pCtx, pvhIssue, "resolve_structural_fixed", &bHave_Previously_Fixed_Structure)  );
	if (bHave_Previously_Fixed_Structure)
	{
		SG_ERR_CHECK(  SG_vhash__get__bool(pCtx, pvhIssue, "resolve_structural_fixed", &bHave_Previously_Fixed_Structure)  );
		if (bHave_Previously_Fixed_Structure)
		{
			*pbDidFix = SG_TRUE;
			return;
		}
	}

	SG_ERR_CHECK(  SG_vhash__get__bool(pCtx, pvhIssue, "collision_flags", &bCollisions)  );
	SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvhIssue, "portability_flags", &i64)  );
	portability_flags = (SG_portability_flags)i64;

	if (conflict_flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__UNDELETE__MASK)
	{
		SG_ERR_CHECK(  _resolve__fix__structural__delete(pCtx, pData,
														 pszGid, pvhIssue, pStrRepoPath,
														 conflict_flags, bCollisions, portability_flags,
														 pbDidFix)  );
		return;
	}
	else if (bCollisions || (portability_flags) || (conflict_flags & ~SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__MASK))
	{
		SG_ERR_CHECK(  _resolve__fix__structural__non_delete(pCtx, pData,
															 pszGid, pvhIssue, pStrRepoPath,
															 conflict_flags, bCollisions, portability_flags,
															 pbDidFix)  );
		return;
	}
	else
	{
		// we don't have a structural problem.

		*pbDidFix = SG_TRUE;
		return;
	}

fail:
	return;
}

//////////////////////////////////////////////////////////////////

/**
 * Try to FIX the ISSUE.
 *
 * Alter something in the pendingtree/issue/WD and then SAVE the pendingtree.
 * We allow this to be an incremental save after just this issue.  We also
 * allow the VFILE lock to be released while the external mergetool (such as
 * DiffMerge) is running.  (Not because DiffMerge needs it, but rather so that they
 * could do other STATUS/DIFF commands in another shell while doing the text
 * merge.)
 */	
static void _resolve__fix(SG_context * pCtx,
						  struct _resolve_data * pData,
						  const char * pszGid,
						  enum _fix_status * pFixContentStatus, SG_bool * pbFixedStructural)
{
	const SG_vhash * pvhIssue = NULL;
	SG_string * pStrRepoPath = NULL;
	SG_int64 i64;
	SG_mrg_cset_entry_conflict_flags conflict_flags;
	SG_bool bIsResolved = SG_FALSE;
	SG_bool bHaveFileContentConflict = SG_FALSE;

	// Fetch the ISSUE using the current pendingtree (allocating one if
	// necessary) and print detailed info about the ISSUE on the console.

	SG_ERR_CHECK(  _resolve__lookup_issue(pCtx, pData, pszGid, &pvhIssue)  );
	SG_ERR_CHECK(  _resolve__list(pCtx, pData, pvhIssue, &pStrRepoPath)  );

	// TODO 2010/07/12 We should have a --prompt option to allow them to
	// TODO            skip an issue.  Like "/bin/rm -i *".

	// see if we might have to use a mergetool.  here we only care if
	// there is/was a (potential) conflict and don't care whether or
	// not auto-merge was originally successful.

	SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvhIssue, "conflict_flags", &i64)  );
	conflict_flags = (SG_mrg_cset_entry_conflict_flags)i64;
	bHaveFileContentConflict = ((conflict_flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__MASK) != 0);

	SG_ERR_CHECK(  _resolve__is_resolved(pCtx, pvhIssue, &bIsResolved)  );
	if (bIsResolved)
	{
		SG_ERR_CHECK(  _resolve__fix__redo(pCtx, pData, pszGid,
										   pvhIssue, pStrRepoPath,
										   conflict_flags,
										   pFixContentStatus)  );
		goto done;
	}

	// There are 2 main types of problems:
	// [1] File content conflicts that (may) need attention.
	//
	// [2] Structural changes, including: MOVEs, RENAMEs, CHMODs, XATTRs,
	//     entryname collisions, potential entryname collisions, and etc.
	//
	// We could also have both -- both edit conflicts and rename conflicts,
	// for example.
	//
	// Do these in 2 steps so that we can release the VFILE lock while they
	// are editing the file.

	//////////////////////////////////////////////////////////////////
	// [1] 
	//////////////////////////////////////////////////////////////////

	if (bHaveFileContentConflict)
	{
		SG_ERR_CHECK(  _resolve__fix__merge_content_if_necessary(pCtx, pData, pszGid,
																 pvhIssue, pStrRepoPath,
																 conflict_flags,
																 pFixContentStatus)  );
		pvhIssue = NULL;

		if (*pFixContentStatus == FIX_MERGETOOL_ERROR)
			goto done;
	}

	//////////////////////////////////////////////////////////////////
	// [2]
	//////////////////////////////////////////////////////////////////

	if (!pvhIssue)
		SG_ERR_CHECK(  _resolve__lookup_issue(pCtx, pData, pszGid, &pvhIssue)  );

	SG_ERR_CHECK(  _resolve__fix__structural(pCtx, pData, pszGid,
											 pvhIssue, pStrRepoPath,
											 conflict_flags,
											 pbFixedStructural)  );
	if (!*pbFixedStructural)
		goto done;

	//////////////////////////////////////////////////////////////////

	if (((*pFixContentStatus == FIX_USER_MERGED) || (*pFixContentStatus == FIX_ALREADY_MERGED))
		&& (*pbFixedStructural))
	{
		// only when both parts are fixed do we mark the issue as RESOLVED.

		SG_ERR_CHECK(  _resolve__mark(pCtx, pData, pvhIssue)  );
		pvhIssue = NULL;
	}

done:
	;
fail:
	SG_STRING_NULLFREE(pCtx, pStrRepoPath);
}

/**
 * Try to FIX each ISSUE in the array and update the WD and allow
 * incremental saves of the pendingtree.
 */
static void _resolve__do_fix(SG_context * pCtx, struct _resolve_data * pData)
{
	SG_uint32 k, kLimit;

	SG_ERR_CHECK(  SG_stringarray__count(pCtx, pData->psaGids, &kLimit)  );
	for (k=0; k<kLimit; k++)
	{
		const char * pszGid_k;
		enum _fix_status fixContentStatus = FIX_ZERO;
		SG_bool bFixedStructural = SG_FALSE;

		SG_ERR_CHECK(  SG_stringarray__get_nth(pCtx, pData->psaGids, k, &pszGid_k)  );

		SG_ERR_CHECK(  _resolve__fix(pCtx, pData, pszGid_k, &fixContentStatus, &bFixedStructural)  );

		// If everything went well, we go on to the next ISSUE in the list.
		// If the user aborted the merge (or if we had a problem of some kind
		// we should probably just stop).

		if ((fixContentStatus == FIX_MERGETOOL_ERROR) || (!bFixedStructural))
		{
			if (k+1 < kLimit)
			{
				// TODO 2010/07/12 Ask if they want to continue anyway.

				break;
			}
		}
	}

fail:
	return;
}

//////////////////////////////////////////////////////////////////

/**
 * Assuming that we have something of the form:
 * 
 *     vv resolve [--foo] <arg_0> [<arg_1> [<arg_2> ...]]
 *
 * where each <arg_x> is an absolute or relative path in the WD
 * (probably not a repo-path).
 *
 * Use the PENDINGTREE to lookup each path and get the entry's GID.
 * Use the GID to search for an ISSUE in the list of issues.  If we
 * find it, add the GID to the stringarray we are building.  If not,
 * throw an error.
 */
static void _resolve__map_args_to_gids(SG_context * pCtx,
									   struct _resolve_data * pData,
									   SG_uint32 count_args, const char ** paszArgs,
									   SG_bool bWantResolved,
									   SG_bool bWantUnresolved)
{
	SG_pathname * pPath_k = NULL;
	char * pszGid_k = NULL;
	SG_uint32 kArg;
	SG_bool bWantBoth = (bWantResolved && bWantUnresolved);

	SG_ERR_CHECK(  SG_STRINGARRAY__ALLOC(pCtx, &pData->psaGids, count_args)  );

	for (kArg=0; kArg<count_args; kArg++)
	{
		const SG_vhash * pvhIssue_k;
		SG_bool bFound;
		SG_bool bDuplicate;
		SG_bool bWantThisOne;

		// take each <arg_k> and get a full pathname for it and
		// search for it in the pendingtree and get its GID.
		// in theory, if an entry has an issue, it is dirty and
		// should have a ptnode.

		if (paszArgs[kArg][0] == '@')
			SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path2(pCtx, pData->pPendingTree, paszArgs[kArg], &pPath_k)  );
		else
			SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPath_k, paszArgs[kArg])  );

		SG_ERR_CHECK(  SG_pendingtree__get_gid_from_local_path(pCtx, pData->pPendingTree, pPath_k, &pszGid_k)  );
#if 0 && defined(DEBUG)
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   ("Mapped arg[%d] '%s' to:\n"
									"\t%s\n"
									"\t[gid %s]\n"),
								   kArg, paszArgs[kArg],
								   SG_pathname__sz(pPath_k),
								   pszGid_k)  );
#endif

		// see if there is an ISSUE for this GID.

		SG_ERR_CHECK(  SG_pendingtree__find_wd_issue_by_gid(pCtx, pData->pPendingTree, pszGid_k, &bFound, &pvhIssue_k)  );
		if (!bFound)
			SG_ERR_THROW2(  SG_ERR_ISSUE_NOT_FOUND,
							(pCtx, "No issue found for '%s': %s",
							 paszArgs[kArg],
							 SG_pathname__sz(pPath_k))  );

		if (bWantBoth)
			bWantThisOne = SG_TRUE;
		else
		{
			SG_int64 s;
			SG_pendingtree_wd_issue_status status;
			SG_bool bResolved;

			SG_ERR_CHECK_RETURN(  SG_vhash__get__int64(pCtx, pvhIssue_k, "status", &s)  );
			status = (SG_pendingtree_wd_issue_status)s;
			bResolved = ((status & SG_ISSUE_STATUS__MARKED_RESOLVED) == SG_ISSUE_STATUS__MARKED_RESOLVED);

			bWantThisOne = ((bWantResolved && bResolved) || (bWantUnresolved && !bResolved));
		}

		if (bWantThisOne)
		{
			// check for duplicate args on command line. (or rather, args that
			// map to the same GID.)

			SG_ERR_CHECK(  SG_stringarray__find(pCtx, pData->psaGids, pszGid_k, 0, &bDuplicate, NULL)  );
			if (bDuplicate)
				SG_ERR_THROW2(  SG_ERR_DUPLICATE_ISSUE,
								(pCtx, "Argument '%s' maps to an issue already named.", paszArgs[kArg])  );

			SG_ERR_CHECK(  SG_stringarray__add(pCtx, pData->psaGids, pszGid_k)  );
		}

		SG_NULLFREE(pCtx, pszGid_k);
		SG_PATHNAME_NULLFREE(pCtx, pPath_k);
	}

	return;

fail:
	SG_NULLFREE(pCtx, pszGid_k);
	SG_PATHNAME_NULLFREE(pCtx, pPath_k);
}

//////////////////////////////////////////////////////////////////

/**
 * build an ordered stringarray of the GIDs of all of the issues.
 * i need to have the list of GIDs be independent of pPendingTree
 * and pvaIssues so that FIX can do incremental SAVES of the
 * pendingtree (which trashes the ptnodes and/or requires a reload).
 */
static void _resolve__get_all_issue_gids(SG_context * pCtx,
										 struct _resolve_data * pData,
										 SG_bool bWantResolved,
										 SG_bool bWantUnresolved)
{
	const SG_varray * pvaIssues;			// varray[pvhIssue *] of all issues -- we do not own this
	SG_bool bHaveIssues;
	SG_uint32 k;
	SG_uint32 nrIssues = 0;
	SG_bool bWantBoth = (bWantResolved && bWantUnresolved);

	SG_ERR_CHECK(  SG_pendingtree__get_wd_issues__ref(pCtx, pData->pPendingTree, &bHaveIssues, &pvaIssues)  );
	if (bHaveIssues)
		SG_ERR_CHECK(  SG_varray__count(pCtx, pvaIssues, &nrIssues)  );

	SG_ERR_CHECK(  SG_STRINGARRAY__ALLOC(pCtx, &pData->psaGids, nrIssues)  );

	for (k=0; k<nrIssues; k++)
	{
		const SG_vhash * pvhIssue_k;
		const char * pszGid_k;
		SG_bool bWantThisOne;

		SG_ERR_CHECK(  SG_varray__get__vhash(pCtx, pvaIssues, k, (SG_vhash **)&pvhIssue_k)  );

		if (bWantBoth)
			bWantThisOne = SG_TRUE;
		else
		{
			SG_int64 s;
			SG_pendingtree_wd_issue_status status;
			SG_bool bResolved;

			SG_ERR_CHECK_RETURN(  SG_vhash__get__int64(pCtx, pvhIssue_k, "status", &s)  );
			status = (SG_pendingtree_wd_issue_status)s;
			bResolved = ((status & SG_ISSUE_STATUS__MARKED_RESOLVED) == SG_ISSUE_STATUS__MARKED_RESOLVED);

			bWantThisOne = ((bWantResolved && bResolved) || (bWantUnresolved && !bResolved));
		}

		if (bWantThisOne)
		{
			SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhIssue_k, "gid", &pszGid_k)  );
			SG_ERR_CHECK(  SG_stringarray__add(pCtx, pData->psaGids, pszGid_k)  );
		}
	}

fail:
	return;
}

//////////////////////////////////////////////////////////////////

/**
 * Handle the RESOLVE command.
 *
 *
 */
void do_cmd_resolve(SG_context * pCtx,
					SG_option_state * pOptSt,
					SG_uint32 count_args, const char** paszArgs)
{
	struct _resolve_data data;
	SG_uint32 sum = 0;
	SG_bool bAll = SG_FALSE;
	SG_bool bWantResolved = SG_FALSE;
	SG_bool bWantUnresolved = SG_FALSE;
	SG_bool bReqArg = SG_FALSE;
	SG_bool bFix = SG_FALSE;

	memset(&data, 0, sizeof(data));
	data.pPathCwd = NULL;
	data.pPendingTree = NULL;
	data.psaGids = NULL;
	data.bForceContentReMerge = pOptSt->bRemerge;
	data.pszUseTool = pOptSt->psz_tool;

	// allow at most ONE of the "command" options:
	//
	// the --list-all options do not allow ARGs.
	// the --mark-all options do not allow ARGs.
	// the --Mark requires at least one ARG.
	// the --List allows 0 or more ARGs.
	// 
	// if no "command" option, allow 0 or more ARGs
	// and implictly try to FIX things.
	// 
	// allow either of the following "modifier" options
	// when fixing:
	//
	// the --remerge option (re-)runs the mergetool on files
	// (either all or the ones listed in args) even if they
	// were already auto-merged (or manually merged by a
	// previous RESOLVE).
	//
	// the --tool=<toolname> names an alterate mergetool to
	// use instead of default or the one normally selected
	// based on suffix.
	//
	// WARNING: We set sg_cl_options[].has_arg to 0 for all of
	//          our "commands" options so that we get all of the
	//          pathnames in ARGs rather than bound to the option.
	//          That is, I want to be able to say:
	//               vv resolve --mark foo bar
	//          rather than:
	//               vv resolve --mark foo --mark bar
	//
	//          It also allows me to have:
	//               vv resolve --list
	//          and
	//               vv resolve --list foo

	if (pOptSt->bListAll)		{ sum++; bAll = SG_TRUE;  bWantResolved = SG_TRUE;  bWantUnresolved = SG_TRUE;  bReqArg = SG_FALSE; }
	if (pOptSt->bMarkAll)		{ sum++; bAll = SG_TRUE;  bWantResolved = SG_TRUE;  bWantUnresolved = SG_TRUE;  bReqArg = SG_FALSE; }
	if (pOptSt->bList)
	{
		if (count_args == 0)	{ sum++; bAll = SG_FALSE; bWantResolved = SG_FALSE; bWantUnresolved = SG_TRUE;  bReqArg = SG_FALSE; }
		else					{ sum++; bAll = SG_FALSE; bWantResolved = SG_TRUE;  bWantUnresolved = SG_TRUE;  bReqArg = SG_FALSE; }
	}
	if (pOptSt->bMark)			{ sum++; bAll = SG_FALSE; bWantResolved = SG_FALSE; bWantUnresolved = SG_TRUE;  bReqArg = SG_TRUE;  }

	if (sum == 0)
	{
		bFix = SG_TRUE;
		bAll = SG_FALSE;
		bWantResolved = data.bForceContentReMerge;
		bWantUnresolved = SG_TRUE;
		bReqArg = SG_FALSE;
	}

	if (sum > 1)
		SG_ERR_THROW(  SG_ERR_USAGE  );

	if (bReqArg && (count_args == 0))
		SG_ERR_THROW(  SG_ERR_USAGE  );

	if (bAll && (count_args > 0))
		SG_ERR_THROW(  SG_ERR_USAGE  );

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC(pCtx, &data.pPathCwd)  );
	SG_ERR_CHECK(  SG_pathname__set__from_cwd(pCtx, data.pPathCwd)  );

	// Do a complete scan first.  This ensures that the pendingtree knows
	// about everything that is dirty in the WD and helps ensure that every
	// issue in the issues list has a ptnode in the pendingtree.
	// 
	// TODO 2010/07/16 Technically, this should NOT be required.  But it
	// TODO            helps.  The problem is that when a file is edited
	// TODO            we don't automatically get the notification, rather
	// TODO            we do a status aka scan (and/or use the timestamp
	// TODO            cache) when various commands start which detect
	// TODO            file content changes.  So the fact that the MERGE
	// TODO            may have written a bunch of merged/edited files
	// TODO            doesn't necessarily mean that they are listed in
	// TODO            the pendingtree -- because the user may have edited
	// TODO            them again (or edited other files) since the merge
	// TODO            completed.  So we scan.
	// TODO
	// TODO            See also the comment in sg.c:do_cmd_commit() for sprawl-809.
	// TODO
	// TODO            What this scan is helping to hide is a problem where
	// TODO            we're hitting the issues list for GIDs and then
	// TODO            using SG_pendingtree__find_repo_path_by_gid() to
	// TODO            dynamically convert it into a "live/current" repo-path.
	// TODO            and it assumes that it is only called for dirty entries
	// TODO            (or rather, for entries that have a ptnode).  We need
	// TODO            to fix that.

	SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, data.pPathCwd, &data.pPendingTree)  );
	SG_ERR_CHECK(  SG_pendingtree__scan(pCtx, data.pPendingTree, SG_TRUE, SG_TRUE, SG_TRUE, gpBlankFilespec)  );
	SG_PENDINGTREE_NULLFREE(pCtx, data.pPendingTree);

	// Now load the pendingtree for real.

	SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, data.pPathCwd, &data.pPendingTree)  );

	if (data.pszUseTool)
	{
		SG_repo * pRepo;
		SG_bool bDefined;

		SG_ERR_CHECK(  SG_pendingtree__get_repo(pCtx, data.pPendingTree, &pRepo)  );
		SG_ERR_CHECK(  SG_mergetool__exists(pCtx, SG_MERGETOOL__CONTEXT__RESOLVE, data.pszUseTool, pRepo, &bDefined)  );
		if (!bDefined)
			SG_ERR_THROW2(  SG_ERR_INVALIDARG,
							(pCtx, "Unknown mergetool '%s'.", data.pszUseTool)  );
	}

	if (count_args > 0)
		SG_ERR_CHECK(  _resolve__map_args_to_gids(pCtx, &data, count_args, paszArgs, bWantResolved, bWantUnresolved)  );
	else
		SG_ERR_CHECK(  _resolve__get_all_issue_gids(pCtx, &data, bWantResolved, bWantUnresolved)  );

	//////////////////////////////////////////////////////////////////

	if (pOptSt->bListAll || pOptSt->bList)
	{
		SG_ERR_CHECK(  _resolve__do_list(pCtx, &data)  );
	}
	else if (pOptSt->bMarkAll || pOptSt->bMark)
	{
		SG_ERR_CHECK(  _resolve__do_mark(pCtx, &data)  );
	}
	else if (bFix)		// no command option given -- assume we want to FIX the issues
	{
		SG_ERR_CHECK(  _resolve__do_fix(pCtx, &data)  );
	}

fail:
	SG_PATHNAME_NULLFREE(pCtx, data.pPathCwd);
	SG_PENDINGTREE_NULLFREE(pCtx, data.pPendingTree);
	SG_STRINGARRAY_NULLFREE(pCtx, data.psaGids);
}
