/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg__do_cmd_resolve2.c
 *
 * @details Implements the resolve command.
 *
 */


/*
**
** Includes
**
*/

#include <sg.h>
#include "sg_typedefs.h"
#include "sg_prototypes.h"


/*
**
** Types
**
*/

/**
 * Type of function that handles a resolve subcommand.
 */
typedef void _subcommand__command_handler(
	SG_context*            pCtx,     //< [in] [out] Error and context info.
	const SG_option_state* pOptions, //< [in] Options parsed from the command line.
	SG_uint32              uArgs,    //< [in] Number of elements in ppArgs.
	const char**           ppArgs,   //< [in] Free arguments remaining on the command line.
	                                 //<      Does NOT include the subcommand name itself.
	const SG_pathname*     pCwd,     //< [in] CWD the command is running in.
	SG_uint32              uIndent   //< [in] Base amount to indent all messages.
	);

/**
 * Specification of a subcommand that can be passed to the main resolve command.
 */
typedef struct _subcommand__command_spec
{
	const char*                   szName;   //< Name of the subcommand, as used on the command line.
	_subcommand__command_handler* fHandler; //< Function that implements the subcommand.
}
_subcommand__command_spec;

/**
 * Options and state data used by _interactive__next and its helpers.
 */
typedef struct _interactive__session
{
	// options/configuration
	const SG_pathname* pCwd;            //< CWD where the prompt is running.
	char*              szGid;           //< Only prompt for conflicts on this item.
	                                    //< If NULL, prompt for conflicts on all items.
	SG_bool            bOnlyUnresolved; //< If true, only prompt for unresolved conflicts.
	SG_uint32          uMax;            //< Prompt for this many conflicts at most.
	                                    //< If zero, there is no maximum.
	const char*        szViewTool;      //< Use this tool by default for view commands.
	                                    //< May be NULL to not provide a default.
	const char*        szDiffTool;      //< Use this tool by default for diff commands.
	                                    //< May be NULL to not provide a default.
	const char*        szMergeTool;     //< Use this tool by default for merge commands.
	                                    //< May be NULL to not provide a default.
	SG_uint32          uIndent;         //< Indent all messages this many levels.

	// internal state preserved across calls
	SG_uint32          uChoice;         //< Index of the choice currently being prompted for.
	                                    //< Should be initialized to zero.
	SG_bool            bActive;         //< Whether or not the session is active.
	                                    //< Should be initialized to SG_TRUE.
	SG_bool            bWrite;          //< Whether or not the next prompt should write the choice's details.
	                                    //< Should be initialized to SG_TRUE.
	SG_bool            bAutomatic;      //< Whether or not the current conflict has used an automatic command.
	                                    //< Should be initialized to SG_FALSE.
}
_interactive__session;

/**
 * A parsed command from the interactive prompt.
 */
typedef struct _interactive__command
{
	const char*        szCommand; //< The command/verb, or NULL if not specified.
	const char*        szValue1;  //< The first value name, or NULL if not specified.
	const char*        szValue2;  //< The second value name, or NULL if not specified.
	const char*        szTool;    //< The tool name, or NULL if not specified.
	const char*        szLabel;   //< The label name, or NULL if not specified.
	SG_resolve__value* pValue1;   //< szValue1 parsed into an actual value, or NULL if it couldn't be.
	SG_resolve__value* pValue2;   //< szValue2 parsed into an actual value, or NULL if it couldn't be.
}
_interactive__command;

/**
 * Requirement specifications for a single argument to an interactive command.
 */
typedef enum _interactive__command_argument
{
	ARGUMENT_REQUIRED,  //< The argument is required for the command.
	ARGUMENT_OPTIONAL,  //< The argument is optional for the command.
	ARGUMENT_TOGETHER,  //< A pair of arguments are optional as a unit (they must both be specified or not).
	ARGUMENT_FORBIDDEN, //< The argument cannot be used with the command.
	ARGUMENT_COUNT      //< Number of values in the enumeration.
}
_interactive__command_argument;

/**
 * Type of function that handles an interactive user command.
 */
typedef void _interactive__command_handler(
	SG_context*                  pCtx,     //< [in] [out] Error and context info.
	_interactive__session*       pSession, //< [in] The interactive session data.
	SG_resolve__choice*          pChoice,  //< [in] The choice that the command targets.
	const _interactive__command* pCommand  //< [in] The entered command.
	);

/**
 * Specification of a command that can be entered at the interactive prompt.
 */
typedef struct _interactive__command_spec
{
	const char*                    szName;         //< Name of the command, as it must be entered/used.
	const char*                    szAbbreviation; //< Optional shortcut for the command, usually one letter.
	SG_bool                        bMergeableOnly; //< True if the command is only valid for mergeable choices.
	_interactive__command_handler* fHandler;       //< Function that implements the command.
	_interactive__command_argument eValue1;        //< Command's requirement for a first VALUE argument.
	_interactive__command_argument eValue2;        //< Command's requirement for a second VALUE argument.
	_interactive__command_argument eTool;          //< Command's requirement for a --tool argument.
	_interactive__command_argument eLabel;         //< Command's requirement for a --label argument.
	const char*                    szPrompt;       //< String to use for the command in the interactive prompt.
	const char*                    szUsage;        //< One line message to display for the command's usage.
	const char*                    szUsageExtra1;  //< Additional line to display indented beneath the command's usage.
	const char*                    szUsageExtra2;  //< Additional line to display indented beneath the command's usage.
	const char*                    szUsageExtra3;  //< Additional line to display indented beneath the command's usage.
	const char*                    szUsageExtra4;  //< Additional line to display indented beneath the command's usage.
	const char*                    szUsageExtra5;  //< Additional line to display indented beneath the command's usage.
}
_interactive__command_spec;

/**
 * Data about possible diff/status flags.
 */
typedef struct _status__data
{
	SG_diffstatus_flags eStatus; //< The flag this data is about.
	const char*         szTitle; //< A title/name for the flag.
}
_status__data;


/*
**
** Globals
**
*/

/**
 * String used as a single indent in output.
 */
static const char* gszIndent = "  ";

/**
 * String used as the path for an item that no longer exists.
 */
static const char* gszNonExistentPath = "(non-existent)";

/*
 * Forward declarations for all of the interactive command handlers.
 */
static _subcommand__command_handler _subcommand__handler__list;
static _subcommand__command_handler _subcommand__handler__next;
static _subcommand__command_handler _subcommand__handler__filename;
//static _subcommand__command_handler _subcommand__handler__view;
static _subcommand__command_handler _subcommand__handler__diff;
static _subcommand__command_handler _subcommand__handler__merge;
static _subcommand__command_handler _subcommand__handler__accept;

/**
 * Specification of all the available subcommands.
 */
static const _subcommand__command_spec gaSubcommands[] =
{
	{ "list",     _subcommand__handler__list },
	{ "next",     _subcommand__handler__next },
//	{ "view",     _subcommand__handler__view },
	{ "diff",     _subcommand__handler__diff },
	{ "merge",    _subcommand__handler__merge },
	{ "accept",   _subcommand__handler__accept },
};

/**
 * The default subcommand to try when we get an argument
 * that doesn't match any other defined subcommands.
 */
static const _subcommand__command_spec gcDefaultSubcommand = { "filename", _subcommand__handler__filename };

/*
 * Forward declarations for all of the interactive command handlers.
 */
//static _interactive__command_handler _interactive__handler__view;
static _interactive__command_handler _interactive__handler__diff;
static _interactive__command_handler _interactive__handler__merge;
static _interactive__command_handler _interactive__handler__accept;
static _interactive__command_handler _interactive__handler__repeat;
static _interactive__command_handler _interactive__handler__skip;
static _interactive__command_handler _interactive__handler__quit;
static _interactive__command_handler _interactive__handler__help;

/**
 * Specification of all the available interactive commands.
 */
static const _interactive__command_spec gaInteractiveCommands[] =
{
/*	{
		"view", "v", SG_TRUE, _interactive__handler__view,
		ARGUMENT_REQUIRED, ARGUMENT_FORBIDDEN, ARGUMENT_OPTIONAL, ARGUMENT_FORBIDDEN,
		"(v)iew",
		"v[iew] VALUE [--tool TOOL]",
		"View a specified value in a viewer tool (not implemented).",
		"The default tool for this situation my optionally be overridden.",
		"This command is only available for mergeable conflicts.",
		NULL, NULL
	},
*/	{
		"diff", "d", SG_TRUE, _interactive__handler__diff,
		ARGUMENT_REQUIRED, ARGUMENT_REQUIRED, ARGUMENT_OPTIONAL, ARGUMENT_FORBIDDEN,
		"(d)iff",
		"d[iff] VALUE1 VALUE2 [--tool TOOL]",
		"Diff two specified values in a diff tool.",
		"The default tool for this situation my optionally be overridden.",
		"This command is only available for mergeable conflicts.",
		NULL, NULL
	},
	{
		"merge", "m", SG_TRUE, _interactive__handler__merge,
		ARGUMENT_TOGETHER, ARGUMENT_TOGETHER, ARGUMENT_OPTIONAL, ARGUMENT_OPTIONAL,
		"(m)erge",
		"m[erge] [VALUE1 VALUE2] [--tool TOOL] [--label LABEL]",
		"Merge two specified values into a new value using a merge tool.",
		"Values optional if exactly two leaf values exist to be inferred.",
		"The default tool for this situation my optionally be overridden.",
		"The label for the new value may optionally be specified.",
		"This command is only available for mergeable conflicts.",
	},
	{
		"accept", "a", SG_FALSE, _interactive__handler__accept,
		ARGUMENT_REQUIRED, ARGUMENT_FORBIDDEN, ARGUMENT_FORBIDDEN, ARGUMENT_FORBIDDEN,
		"(a)ccept",
		"a[ccept] VALUE",
		"Accept a specified value as the resolution for this conflict.",
		NULL, NULL, NULL, NULL
	},
	{
		"repeat", "r", SG_FALSE, _interactive__handler__repeat,
		ARGUMENT_FORBIDDEN, ARGUMENT_FORBIDDEN, ARGUMENT_FORBIDDEN, ARGUMENT_FORBIDDEN,
		"(r)epeat",
		"r[epeat]",
		"Display the details of the current conflict again.",
		NULL, NULL, NULL, NULL
	},
	{
		"skip", "s", SG_FALSE, _interactive__handler__skip,
		ARGUMENT_FORBIDDEN, ARGUMENT_FORBIDDEN, ARGUMENT_FORBIDDEN, ARGUMENT_FORBIDDEN,
		"(s)kip",
		"s[kip]",
		"Skip this conflict without resolving it.",
		NULL, NULL, NULL, NULL
	},
	{
		"quit", "q", SG_FALSE, _interactive__handler__quit,
		ARGUMENT_FORBIDDEN, ARGUMENT_FORBIDDEN, ARGUMENT_FORBIDDEN, ARGUMENT_FORBIDDEN,
		"(q)uit",
		"q[uit]",
		"Stop resolving conflicts.",
		NULL, NULL, NULL, NULL
	},
	{
		"help", "h", SG_FALSE, _interactive__handler__help,
		ARGUMENT_FORBIDDEN, ARGUMENT_FORBIDDEN, ARGUMENT_FORBIDDEN, ARGUMENT_FORBIDDEN,
		"(h)elp",
		"h[elp]",
		"Display usage information.",
		NULL, NULL, NULL, NULL
	},
};

/**
 * Data about different diff/status flags.
 */
static const _status__data gcStatuses[] =
{
	{ SG_DIFFSTATUS_FLAGS__ADDED,            "Added" },
	{ SG_DIFFSTATUS_FLAGS__DELETED,          "Deleted" },
	{ SG_DIFFSTATUS_FLAGS__MODIFIED,         "Modified" },
	{ SG_DIFFSTATUS_FLAGS__LOST,             "Lost" },
	{ SG_DIFFSTATUS_FLAGS__FOUND,            "Found" },
	{ SG_DIFFSTATUS_FLAGS__RENAMED,          "Renamed" },
	{ SG_DIFFSTATUS_FLAGS__MOVED,            "Moved" },
	{ SG_DIFFSTATUS_FLAGS__CHANGED_ATTRBITS, "Attributes" },
	{ SG_DIFFSTATUS_FLAGS__CHANGED_XATTRS,   "ExAttributes" },
};

static const _status__data gcUnchangedStatus = { 0u, "Unchanged" };


/*
**
** Helpers
**
*/

/**
 * A thread-safe and SG-like implementation of strtok.
 * TODO: Move this to sg_str_utils and write tests for it.
 */
static void _SG_strtok(
	SG_context*  pCtx,         //< [in] [out] Error and context info.
	char*        szString,     //< [in] The string to tokenize.
	                           //<      If NULL, tokenization is continued on ppHandle.
	const char*  szDelimiters, //< [in] A string of characters that will delimit the next token.
	                           //<      If NULL, then whitespace delimiters will be used.
	SG_bool      bSkipEmpty,   //< [in] If true, zero-length tokens will be skipped instead of returned.
	const char** ppToken,      //< [out] The next token in the string.
	                           //<       NULL if there are no more tokens.
	void**       ppHandle      //< [in] [out] A handle to this tokenization string/session/instance.
	)
{
	static const char* szWhitespaceDelimiters = " \t\r\n";

	const char* szToken = NULL;

	SG_NULLARGCHECK(ppToken);
	SG_NULLARGCHECK(ppHandle);

	// if they didn't supply a string, start with the handle
	if (szString == NULL)
	{
		szString = (char*)*ppHandle;
	}

	// if they didn't supply delimiters, use whitespace
	if (szDelimiters == NULL)
	{
		szDelimiters = szWhitespaceDelimiters;
	}

	// if we have a string to search, do it
	// we won't if the handle is NULL, meaning we found the last token last time
	if (szString != NULL)
	{
		// if we're skipping empty tokens
		// then advance until we find a non-delimiter
		while (
			   *szString != '\0'
			&& bSkipEmpty == SG_TRUE
			&& strchr(szDelimiters, *szString) != NULL
			)
		{
			++szString;
		}

		// store off the start of the string
		// so we can return it as the token
		szToken = szString;

		// advance until we find a delimiter
		while (
			   *szString != '\0'
			&& strchr(szDelimiters, *szString) == NULL
			)
		{
			++szString;
		}

		// setup the handle to keep going with the next call
		if (*szString == '\0')
		{
			// we found the end of the string, so this is the last token
			// set the handle to NULL, so next time we'll know that we're done
			*ppHandle = NULL;
		}
		else
		{
			// set the handle to the character after the delimiter
			// that's where we'll pick up next time
			*ppHandle = (void*)(szString + 1u);
		}

		// replace the delimiter with a NULL-terminator
		*szString = '\0';
	}

	// return the token
	*ppToken = szToken;

fail:
	return;
}

/**
 * Prints an indented formatted line to STDOUT.
 */
static void _console__stdout(
	SG_context* pCtx,     //< [in] [out] Error and context info.
	SG_uint32   uIndent,  //< [in] Number of times to indent the string.
	const char* szFormat, //< [in] Format string to print, or NULL to only print the indent.
	...                   //< [in] Data to substitute into format string.
	)
{
	while (uIndent > 0u)
	{
		SG_ERR_CHECK(  SG_console(pCtx, SG_CS_STDOUT, "%s", gszIndent)  );
		--uIndent;
	}

	if (szFormat != NULL)
	{
		va_list pArgs;

		va_start(pArgs, szFormat);
		SG_console__va(pCtx, SG_CS_STDOUT, szFormat, pArgs);
		va_end(pArgs);
		SG_ERR_CHECK_CURRENT;
	}

	SG_ERR_CHECK(  SG_console(pCtx, SG_CS_STDOUT, "\n")  );

fail:
	return;
}

/**
 * Prompts the user for a line of input and returns it.
 */
static void _console__prompt(
	SG_context* pCtx,     //< [in] [out] Error and context info.
	SG_uint32   uIndent,  //< [in] Number of times to indent the prompt.
	SG_string** ppInput,  //< [out] The input entered by the user.
	const char* szPrompt, //< [in] A prompt to display to the user, or NULL to not bother.
	...                   //< [in] Data to substitute into the prompt.
	)
{
	SG_string* sInput = NULL;

	SG_NULLARGCHECK(ppInput);

	while (uIndent > 0u)
	{
		SG_ERR_CHECK(  SG_console(pCtx, SG_CS_STDOUT, "%s", gszIndent)  );
		--uIndent;
	}

	if (szPrompt != NULL)
	{
		va_list pArgs;

		va_start(pArgs, szPrompt);
		SG_console__va(pCtx, SG_CS_STDOUT, szPrompt, pArgs);
		va_end(pArgs);
		SG_ERR_CHECK_CURRENT;
	}

	// Note: Unlike almost every other SG function that takes an object
	//       double pointer, this one expects that we've already allocated
	//       it, rather than allocating it itself and returning it to us.
	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &sInput)  );
	SG_ERR_CHECK(  SG_console__readline_stdin(pCtx, &sInput)  );

	*ppInput = sInput;
	sInput = NULL;

fail:
	SG_STRING_NULLFREE(pCtx, sInput);
	return;
}

/**
 * Runs a complete pendingtree scan on a folder.
 */
static void _pendingtree__scan(
	SG_context*        pCtx,   //< [in] [out] Error and context info.
	const SG_pathname* pFolder //< [in] The folder containing a pendingtree to scan.
	)
{
	SG_pendingtree* pPendingTree = NULL;

	SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pFolder, &pPendingTree)  );
	SG_ERR_CHECK(  SG_pendingtree__scan(pCtx, pPendingTree, SG_TRUE, SG_TRUE, SG_TRUE, gpBlankFilespec)  );
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);

fail:
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	return;
}

/**
 * An SG_resolve__value__callback that appends the value to a vector.
 */
static void _resolve__value_callback__append_to_vector(
	SG_context*        pCtx,      //< [in] [out] Error and context info.
	SG_resolve__value* pValue,    //< [in] The current value.
	void*              pUserData, //< [in] The SG_vector* to append values to.
	SG_bool*           pContinue  //< [in] Whether or not iteration should continue.
	)
{
	SG_vector* pVector = (SG_vector*)pUserData;

	SG_NULLARGCHECK(pValue);
	SG_NULLARGCHECK(pUserData);
	SG_NULLARGCHECK(pContinue);

	SG_ERR_CHECK(  SG_vector__append(pCtx, pVector, pValue, NULL)  );

	*pContinue = SG_TRUE;

fail:
	return;
}

/**
 * Gets the two leaf values of a given resolve choice.
 * Returns NULLs if the choice doesn't have exactly two leaves.
 */
static void _resolve__choice__get_two_leaves(
	SG_context*         pCtx,    //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice, //< [in] The choice to get two leaves from.
	SG_resolve__value** ppLeaf1, //< [out] The first of the two leaves.
	                             //<       NULL if the choice doesn't have exactly two leaves.
	SG_resolve__value** ppLeaf2  //< [out] The second of the two leaves.
	                             //<       NULL if the choice doesn't have exactly two leaves.
	)
{
	SG_vector*         pLeaves = NULL;
	SG_uint32          uCount  = 0u;
	SG_resolve__value* pLeaf1  = NULL;
	SG_resolve__value* pLeaf2  = NULL;

	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(ppLeaf1);
	SG_NULLARGCHECK(ppLeaf2);

	SG_ERR_CHECK(  SG_VECTOR__ALLOC(pCtx, &pLeaves, 3u)  );
	SG_ERR_CHECK(  SG_resolve__choice__foreach_value(pCtx, pChoice, SG_TRUE, _resolve__value_callback__append_to_vector, (void*)pLeaves)  );
	SG_ERR_CHECK(  SG_vector__length(pCtx, pLeaves, &uCount)  );
	if (uCount == 2u)
	{
		SG_ERR_CHECK(  SG_vector__get(pCtx, pLeaves, 0u, (void**)&pLeaf1)  );
		SG_ERR_CHECK(  SG_vector__get(pCtx, pLeaves, 1u, (void**)&pLeaf2)  );
	}

	*ppLeaf1 = pLeaf1;
	pLeaf1 = NULL;
	*ppLeaf2 = pLeaf2;
	pLeaf2 = NULL;

fail:
	SG_VECTOR_NULLFREE(pCtx, pLeaves);
	return;
}

/**
 * User data used by _resolve__get_choice__index__choice.
 */
typedef struct _resolve__get_choice__index__choice__data
{
	SG_uint32           uIndex;  //< [in] Index of the choice to retrieve.
	SG_resolve__choice* pChoice; //< [out] Choice at the given index.
}
_resolve__get_choice__index__choice__data;

/**
 * SG_resolve__choice__callback that stops iterating after N calls.
 * It then returns whichever choice it stopped at.
 */
void _resolve__get_choice__index__choice(
	SG_context*         pCtx,      //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice,   //< [in] The current choice.
	void*               pUserData, //< [in] A _resolve__get_choice__index__choice__data instance with additional parameters.
	SG_bool*            pContinue  //< [out] Whether or not iteration should continue.
	)
{
	_resolve__get_choice__index__choice__data* pData = (_resolve__get_choice__index__choice__data*)pUserData;

	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pUserData);
	SG_NULLARGCHECK(pContinue);

	if (pData->uIndex == 0u)
	{
		pData->pChoice = pChoice;
		*pContinue = SG_FALSE;
	}
	else
	{
		pData->uIndex -= 1u;
		*pContinue = SG_TRUE;
	}

fail:
	return;
}

/**
 * Retrieve the choice at a specified index,
 * either in a whole resolve or a single item.
 */
void _resolve__get_choice__index(
	SG_context*          pCtx,            //< [in] [out] Error and context info.
	SG_resolve*          pResolve,        //< [in] The resolve to get a choice from.
	SG_resolve__item*    pItem,           //< [in] If non-NULL, only consider choices on this item.
	SG_uint32            uIndex,          //< [in] The index of the choice to get.
	SG_bool              bOnlyUnresolved, //< [in] Whether or not to only consider unresolved choices.
	SG_resolve__choice** pChoice          //< [out] The retrieved choice.
	)
{
	_resolve__get_choice__index__choice__data cData;

	SG_NULLARGCHECK(pCtx);
	SG_NULLARGCHECK(pResolve);
	SG_NULLARGCHECK(pChoice);

	cData.uIndex  = uIndex;
	cData.pChoice = NULL;

	if (pItem == NULL)
	{
		SG_ERR_CHECK(  SG_resolve__foreach_choice(pCtx, pResolve, bOnlyUnresolved, _resolve__get_choice__index__choice, (void*)&cData)  );
	}
	else
	{
		SG_ERR_CHECK(  SG_resolve__item__foreach_choice(pCtx, pItem, bOnlyUnresolved, _resolve__get_choice__index__choice, (void*)&cData)  );
	}

	*pChoice = cData.pChoice;

fail:
	return;
}


/*
**
** Private Functionality
**
*/

/**
 * Parses a path specified by the user into the GID of an item.
 * The input path can be an absolute path, relative path, repo path, or NULL.
 */
static void _parse_path(
	SG_context*     pCtx,         //< [in] [out] Error and context info.
	SG_pendingtree* pPendingTree, //< [in] Pendingtree to use to convert repo paths into GIDs.
	const char*     szPath,       //< [in] The path to parse, or NULL if the user didn't specify a path.
	char**          ppGid         //< [out] The parsed repo path, or NULL if szPath is NULL.
	)
{
	SG_pathname* pAbsolutePath = NULL;
	char*        szGid         = NULL;

	SG_NULLARGCHECK(pPendingTree);
	SG_NULLARGCHECK(szPath);
	SG_NULLARGCHECK(ppGid);

	if (szPath != NULL)
	{
		if (szPath[0] != '\0' && szPath[0] == '@' && szPath[1] == '/')
		{
			// if it's a repo path, we need to first convert it to an absolute path
			SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path2(pCtx, pPendingTree, szPath, &pAbsolutePath)  );
		}
		else
		{
			// otherwise just allocate whatever they gave us as a path
			// pathnames are always absolute
			// it will interpret the path as relative to the CWD if necessary
			SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pAbsolutePath, szPath)  );
		}

		// get the GID of the item at the path
		SG_ERR_CHECK(  SG_pendingtree__get_gid_from_local_path(pCtx, pPendingTree, pAbsolutePath, &szGid)  );
	}

	// return the GID
	*ppGid = szGid;
	szGid = NULL;

fail:
	SG_PATHNAME_NULLFREE(pCtx, pAbsolutePath);
	SG_NULLFREE(pCtx, szGid);
	return;
}

/**
 * Gets the resolve item and choice that corresond to a given path and state.
 * Prints error messages and returns NULLs if the item/choice cannot be found.
 */
static void _find_choice(
	SG_context*          pCtx,      //< [in] [out] Error and context info.
	SG_uint32            uIndent,   //< [in] Base amount to indent error messages.
	SG_resolve*          pResolve,  //< [in] Resolve data to find the item/choice in.
	const char*          szGid,     //< [in] GID of the item to find.
	SG_resolve__state    eState,    //< [in] State of the choice to find on the item.
	                                //<      If SG_RESOLVE__STATE__COUNT and the item only has one choice,
	                                //<      then that choice is returned, otherwise an error will be thrown.
	SG_resolve__item**   ppItem,    //< [out] The found item, or NULL if not found.
	SG_resolve__choice** ppChoice   //< [out] The found choice, or NULL if not found.
	)
{
	SG_resolve__item*   pItem   = NULL;
	SG_resolve__choice* pChoice = NULL;

	SG_NULLARGCHECK(pResolve);
	SG_NULLARGCHECK(szGid);

	SG_ERR_CHECK(  SG_resolve__get_item__gid(pCtx, pResolve, szGid, &pItem)  );
	if (eState == SG_RESOLVE__STATE__COUNT)
	{
		// they want us to find the item's only choice
		SG_int32 iState = 0u;

		// run through each possible state
		for (iState = 0u; iState < SG_RESOLVE__STATE__COUNT; ++iState)
		{
			SG_resolve__choice* pCurrentChoice = NULL;

			// see if the item has a choice for this state
			SG_ERR_CHECK(  SG_resolve__item__check_choice__state(pCtx, pItem, (SG_resolve__state)iState, &pCurrentChoice)  );
			if (pCurrentChoice != NULL)
			{
				if (pChoice == NULL)
				{
					// this is the first choice we've found on the item
					// store it for later and possible returning
					pChoice = pCurrentChoice;
				}
				else
				{
					// this is the second choice we've found on the item
					SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "The specified item has multiple conflicts, so one must be explicitly specified.")  );
				}
			}
		}

		// if we didn't find a choice, throw an error
		if (pChoice == NULL)
		{
			SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "The specified item has no conflicts.")  );
		}
	}
	else
	{
		// they want a specific choice on the item
		SG_ERR_CHECK(  SG_resolve__item__check_choice__state(pCtx, pItem, eState, &pChoice)  );
		if (pChoice == NULL)
		{
			const char* szState = NULL;
			
			SG_ERR_CHECK(  SG_resolve__state__get_info(pCtx, eState, &szState, NULL)  );
			SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "No '%s' conflict was found on the given item.", szState)  );
		}
	}

	// return our data
	if (ppItem != NULL)
	{
		*ppItem   = pItem;
	}
	if (ppChoice != NULL)
	{
		*ppChoice = pChoice;
	}

fail:
	return;
}

/**
 * Builds a string that lists the current statuses of an item.
 */
static void _write_choice__build_status(
	SG_context*       pCtx,    //< [in] [out] Error and context info.
	SG_resolve__item* pItem,   //< [in] The item to build a status list for.
	SG_string**       ppStatus //< [out] The built list of status strings.
	)
{
	static const char* szSeparator = ", ";

	SG_string* sStatus = NULL;
	SG_uint32  uIndex  = 0u;
	SG_bool    bEmpty  = SG_TRUE;

	// start with an empty string
	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &sStatus)  );

	// run through each status
	for (uIndex = 0u; uIndex < SG_NrElements(gcStatuses); ++uIndex)
	{
		const _status__data* pStatus    = gcStatuses + uIndex;
		SG_bool              bHasStatus = SG_FALSE;

		// check if the item has this status
		SG_ERR_CHECK(  SG_resolve__item__has_status(pCtx, pItem, pStatus->eStatus, &bHasStatus)  );
		if (bHasStatus == SG_TRUE)
		{
			// append a separator if the string already contains statuses
			if (bEmpty == SG_FALSE)
			{
				SG_ERR_CHECK(  SG_string__append__sz(pCtx, sStatus, szSeparator)  );
			}

			// append the status string
			SG_ERR_CHECK(  SG_string__append__sz(pCtx, sStatus, pStatus->szTitle)  );
			bEmpty = SG_FALSE;
		}
	}

	// if it's still empty, its status must be unchanged
	if (bEmpty == SG_TRUE)
	{
		SG_string__append__sz(pCtx, sStatus, gcUnchangedStatus.szTitle);
	}

	// return the string
	*ppStatus = sStatus;
	sStatus = NULL;

fail:
	SG_STRING_NULLFREE(pCtx, sStatus);
	return;
}

/**
 * An SG_resolve__item__callback that writes basic information about the item to stdout.
 */
static void _write_choice__item(
	SG_context*       pCtx,      //< [in] [out] Error and context info.
	SG_resolve__item* pItem,     //< [in] The current item.
	void*             pUserData, //< [in] Pointer to a base indent level (SG_uint32).
	                             //<      NULL is equivalent to zero.
	SG_bool*          pContinue  //< [out] Whether or not iteration should continue.
	)
{
	SG_uint32*             pData              = (SG_uint32*)pUserData;
	SG_uint32              uIndent            = 0u;
	SG_treenode_entry_type eType              = SG_TREENODEENTRY_TYPE__INVALID;
	const char*            szType             = NULL;
	SG_string*             sStatus            = NULL;
	const char*            szWorkingRepoPath  = NULL;
	const char*            szBaselineRepoPath = NULL;

	SG_NULLARGCHECK(pItem);
	SG_NULLARGCHECK(pContinue);

	// if they passed an indent, use it
	if (pData != NULL)
	{
		uIndent = *pData;
	}

	// get data from the item
	SG_ERR_CHECK(  SG_resolve__item__get_type(pCtx, pItem, &eType)  );
	szType = SG_treenode_entry_type__type_to_label(eType);
	SG_ERR_CHECK(  _write_choice__build_status(pCtx, pItem, &sStatus)  );
	SG_ERR_CHECK(  SG_resolve__item__get_repo_path__working(pCtx, pItem, &szWorkingRepoPath)  );
	SG_ERR_CHECK(  SG_resolve__item__get_repo_path__baseline(pCtx, pItem, &szBaselineRepoPath)  );
	if (szWorkingRepoPath == NULL)
	{
		szWorkingRepoPath = gszNonExistentPath;
	}
	if (szBaselineRepoPath == NULL)
	{
		szBaselineRepoPath = gszNonExistentPath;
	}

	// write data about the item
	SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "%s: %s", szType, szWorkingRepoPath)  );
	SG_ERR_CHECK(  _console__stdout(pCtx, uIndent + 1u, "Status: %s", SG_string__sz(sStatus))  );
	SG_ERR_CHECK(  _console__stdout(pCtx, uIndent + 1u, "Baseline Path: %s", szBaselineRepoPath)  );

fail:
	SG_STRING_NULLFREE(pCtx, sStatus);
	return;
}

/**
 * An SG_resolve__value__callback that retrieves information about the value.
 */
static void _write_choice__get_value(
	SG_context*        pCtx,      //< [in] [out] Error and context info.
	SG_resolve__value* pValue,    //< [in] The current value.
	void*              pUserData, //< [in] SG_vhash to add the value label/summary to.
	SG_bool*           pContinue  //< [out] Whether or not iteration should continue.
	)
{
	SG_vhash*   pData    = (SG_vhash*)pUserData;
	const char* szLabel  = NULL;
	SG_string*  sSummary = NULL;
	SG_bool     bLeaf    = SG_FALSE;
	SG_string*  sLabel   = NULL;

	SG_NULLARGCHECK(pValue);
	SG_NULLARGCHECK(pUserData);
	SG_NULLARGCHECK(pContinue);

	// get data from the value
	SG_ERR_CHECK(  SG_resolve__value__get_label(pCtx, pValue, &szLabel)  );
	SG_ERR_CHECK(  SG_resolve__value__get_summary(pCtx, pValue, &sSummary)  );
	SG_ERR_CHECK(  SG_resolve__value__has_flag(pCtx, pValue, SG_RESOLVE__VALUE__LEAF, &bLeaf)  );

	// if the value is a leaf, generate an annotated label
	if (bLeaf == SG_TRUE)
	{
		SG_ERR_CHECK(  SG_string__alloc__format(pCtx, &sLabel, "%s*", szLabel)  );
		szLabel = SG_string__sz(sLabel);
	}

	// add the data to the hash
	if (sSummary == NULL)
	{
		SG_ERR_CHECK(  SG_vhash__add__null(pCtx, pData, szLabel)  );
	}
	else
	{
		SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pData, szLabel, SG_string__sz(sSummary))  );
	}

fail:
	SG_STRING_NULLFREE(pCtx, sSummary);
	SG_STRING_NULLFREE(pCtx, sLabel);
	return;
}

/**
 * Writes information about a choice to stdout.
 */
static void _write_choice(
	SG_context*         pCtx,    //< [in] [out] Error and context info.
	SG_uint32           uIndent, //< [in] Base level to indent the display.
	SG_resolve__choice* pChoice, //< [in] The choice to write.
	SG_resolve__item*   pItem    //< [in] The item that owns the choice, if known.
	                             //<      If NULL, this function will look it up.
	)
{
	static const char* aValueHeaders[] =
	{
		"Possible Existence States:",
		"Possible Names:",
		"Possible Locations: (parent folders)",
		"Possible Attributes:",
		"Possible Extended Attributes:",
		"Possible Contents: (use 'diff' to examine, * indicates merge leaf)",
	};

	SG_treenode_entry_type           eItemType          = SG_TREENODEENTRY_TYPE__INVALID;
	const char*                      szItemType         = NULL;
	SG_string*                       sStatus            = NULL;
	const char*                      szWorkingRepoPath  = NULL;
	const char*                      szBaselineRepoPath = NULL;
	SG_resolve__state                eState             = SG_RESOLVE__STATE__COUNT;
	const char*                      szStateName        = NULL;
	const char*                      szStateDescription = NULL;
	SG_mrg_cset_entry_conflict_flags eConflictCauses    = SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__ZERO;
	SG_bool                          bCollisionCause    = SG_FALSE;
	SG_vhash*                        pCauseDescriptions = NULL;
	SG_bool                          bResolved          = SG_FALSE;
	SG_bool                          bDeleted           = SG_FALSE;
	SG_resolve__value*               pAcceptedValue     = NULL;
	SG_uint32                        uCount             = 0u;
	SG_uint32                        uIndex             = 0u;
	SG_vhash*                        pValues            = NULL;
	SG_uint32                        uLabelWidth        = 0u;
	SG_string*                       sLabel             = NULL;

	SG_NULLARGCHECK(pChoice);

	// if we weren't given an item, look it up
	if (pItem == NULL)
	{
		SG_ERR_CHECK(  SG_resolve__choice__get_item(pCtx, pChoice, &pItem)  );
	}

	// get data about the item
	SG_ERR_CHECK(  SG_resolve__item__get_type(pCtx, pItem, &eItemType)  );
	szItemType = SG_treenode_entry_type__type_to_label(eItemType);
	SG_ERR_CHECK(  _write_choice__build_status(pCtx, pItem, &sStatus)  );
	SG_ERR_CHECK(  SG_resolve__item__get_repo_path__working(pCtx, pItem, &szWorkingRepoPath)  );
	SG_ERR_CHECK(  SG_resolve__item__get_repo_path__baseline(pCtx, pItem, &szBaselineRepoPath)  );
	if (szWorkingRepoPath == NULL)
	{
		szWorkingRepoPath = gszNonExistentPath;
	}
	if (szBaselineRepoPath == NULL)
	{
		szBaselineRepoPath = gszNonExistentPath;
	}

	// get data about the choice
	SG_ERR_CHECK(  SG_resolve__choice__get_state(pCtx, pChoice, &eState)  );
	SG_ERR_CHECK(  SG_resolve__state__get_info(pCtx, eState, &szStateName, &szStateDescription)  );
	SG_ERR_CHECK(  SG_resolve__choice__get_causes(pCtx, pChoice, &eConflictCauses, &bCollisionCause, &pCauseDescriptions)  );
	SG_ERR_CHECK(  SG_resolve__choice__get_resolved(pCtx, pChoice, &bResolved, &bDeleted)  );
	SG_ERR_CHECK(  SG_resolve__choice__get_value__accepted(pCtx, pChoice, &pAcceptedValue)  );

	// header
	SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "%s %s conflict on %s: %s", (bResolved == SG_FALSE ? "Unresolved" : "Resolved"), szStateName, szItemType, szWorkingRepoPath)  );
	SG_ERR_CHECK(  _console__stdout(pCtx, uIndent + 1u, "Status: %s", SG_string__sz(sStatus))  );
	SG_ERR_CHECK(  _console__stdout(pCtx, uIndent + 1u, "Baseline Path: %s", szBaselineRepoPath)  );
	SG_ERR_CHECK(  _console__stdout(pCtx, uIndent + 1u, "Problem: %s", szStateDescription)  );

	// causes
	SG_ERR_CHECK(  _console__stdout(pCtx, uIndent + 1u, "Cause(s):")  );
	SG_ERR_CHECK(  SG_vhash__count(pCtx, pCauseDescriptions, &uCount)  );
	for (uIndex = 0u; uIndex < uCount; ++uIndex)
	{
		const char* szCauseName = NULL;
		const char* szCauseDescription = NULL;

		SG_ERR_CHECK(  SG_vhash__get_nth_pair__sz(pCtx, pCauseDescriptions, uIndex, &szCauseName, &szCauseDescription)  );
		SG_ERR_CHECK(  _console__stdout(pCtx, uIndent + 2u, "%s: %s", szCauseName, szCauseDescription)  );
	}

	// cycle
	if ((eConflictCauses & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__MOVES_CAUSED_PATH_CYCLE) == SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__MOVES_CAUSED_PATH_CYCLE)
	{
		const char* szCyclePath  = NULL;
		SG_uint32   uItemsIndent = uIndent + 2u;

		SG_ERR_CHECK(  SG_resolve__choice__get_cycle_info(pCtx, pChoice, &szCyclePath)  );
		SG_ERR_CHECK(  _console__stdout(pCtx, uIndent + 1u, "Cycling Path: %s", szCyclePath)  );
		SG_ERR_CHECK(  _console__stdout(pCtx, uIndent + 1u, "Other Cycling Items:")  );
		SG_ERR_CHECK(  SG_resolve__choice__foreach_cycling_item(pCtx, pChoice, _write_choice__item, (void*)&uItemsIndent)  );
	}

	// collision
	if (bCollisionCause == SG_TRUE)
	{
		SG_uint32   uItemsIndent = uIndent + 2u;

		SG_ERR_CHECK(  _console__stdout(pCtx, uIndent + 1u, "Other Colliding Items:")  );
		SG_ERR_CHECK(  SG_resolve__choice__foreach_colliding_item(pCtx, pChoice, _write_choice__item, (void*)&uItemsIndent)  );
	}

	// values
	SG_ERR_CHECK(  _console__stdout(pCtx, uIndent + 1u, "%s", aValueHeaders[eState])  );
	SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pValues)  );
	SG_ERR_CHECK(  SG_resolve__choice__foreach_value(pCtx, pChoice, SG_FALSE, _write_choice__get_value, (void*)pValues)  );
	SG_ERR_CHECK(  SG_vhash__count(pCtx, pValues, &uCount)  );
	uLabelWidth = 0u;
	for (uIndex = 0u; uIndex < uCount; ++uIndex)
	{
		const char* szLabel = NULL;
		SG_uint32   uLabel  = 0u;

		SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pValues, uIndex, &szLabel, NULL)  );
		uLabel = SG_STRLEN(szLabel);
		uLabelWidth = SG_MAX(uLabelWidth, uLabel);
	}
	for (uIndex = 0u; uIndex < uCount; ++uIndex)
	{
		const char*       szLabel  = NULL;
		const SG_variant* pSummary = NULL;

		SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pValues, uIndex, &szLabel, &pSummary)  );
		if (pSummary->type == SG_VARIANT_TYPE_NULL)
		{
			SG_ERR_CHECK(  _console__stdout(pCtx, uIndent + 2u, "%s", szLabel)  );
		}
		else
		{
			SG_ASSERT(pSummary->type == SG_VARIANT_TYPE_SZ);
			SG_ERR_CHECK(  SG_string__alloc__format(pCtx, &sLabel, "%s:", szLabel)  );
			SG_ERR_CHECK(  _console__stdout(pCtx, uIndent + 2u, "%-*s %s", uLabelWidth + 1u, SG_string__sz(sLabel), pSummary->v.val_sz)  );
			SG_STRING_NULLFREE(pCtx, sLabel);
		}
	}

	// resolution
	if (bResolved == SG_TRUE)
	{
		// if we're resolved, it should either be via deletion or an accepted value
		SG_ASSERT(bDeleted == SG_TRUE || pAcceptedValue != NULL);

		if (pAcceptedValue != NULL)
		{
			const char* szAcceptedLabel = NULL;

			SG_ERR_CHECK(  SG_resolve__value__get_label(pCtx, pAcceptedValue, &szAcceptedLabel)  );
			SG_ERR_CHECK(  _console__stdout(pCtx, uIndent + 1u, "Resolution: %s", szAcceptedLabel)  );
		}
		else //if (bDeleted == SG_TRUE)
		{
			SG_ERR_CHECK(  _console__stdout(pCtx, uIndent + 1u, "Resolution: Item Deleted")  );
		}
	}

	// empty line
	SG_ERR_CHECK(  _console__stdout(pCtx, 0u, NULL)  );

fail:
	SG_STRING_NULLFREE(pCtx, sStatus);
	SG_VHASH_NULLFREE(pCtx, pCauseDescriptions);
	SG_VHASH_NULLFREE(pCtx, pValues);
	SG_STRING_NULLFREE(pCtx, sLabel);
	return;
}

/**
 * User data used by _write_choice__callback.
 */
typedef struct _write_choice__callback__data
{
	SG_uint32 uIndent; //< [in] Base amount to indent the display.
	SG_uint32 uMax;    //< [in] Maximum number of choices to handle.
	                   //<      If zero, all available choices will be handled.
	SG_uint32 uCount;  //< [in] Number of choices handled so far.
}
_write_choice__callback__data;

/**
 * An SG_resolve__choice__callback that writes information about the item to stdout.
 */
static void _write_choice__callback(
	SG_context*         pCtx,      //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice,   //< [in] The current choice.
	void*               pUserData, //< [in] A _write_choice__callback__data structure with additional parameters.
	SG_bool*            pContinue  //< [out] Whether or not iteration should continue.
	)
{
	_write_choice__callback__data* pData = (_write_choice__callback__data*)pUserData;

	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pUserData);
	SG_NULLARGCHECK(pContinue);

	// write the choice out
	SG_ERR_CHECK(  _write_choice(pCtx, pData->uIndent, pChoice, NULL)  );

	// if we've hit our choice limit, stop iterating
	pData->uCount += 1u;
	if (
		   pData->uMax != 0u
		&& pData->uCount >= pData->uMax
		)
	{
		*pContinue = SG_FALSE;
	}
	else
	{
		*pContinue = SG_TRUE;
	}

fail:
	return;
}

/**
 * Writes the result of an accept to stdout.
 */
static void _write_accept(
	SG_context*         pCtx,    //< [in] [out] Error and context info.
	SG_uint32           uIndent, //< [in] Base amount to indent the messages.
	SG_resolve__value*  pValue,  //< [in] The value that was accepted.
	SG_resolve__choice* pChoice, //< [in] The choice that owns the value.
	                             //<      If NULL, this function will look it up from the value.
	SG_resolve__item*   pItem    //< [in] The item that owns the choice.
	                             //<      If NULL, this function will look it up from the choice.
	)
{
	SG_treenode_entry_type eType      = SG_TREENODEENTRY_TYPE__INVALID;
	const char*            szType     = NULL;
	const char*            szRepoPath = NULL;
	SG_resolve__state      eState     = SG_RESOLVE__STATE__COUNT;
	const char*            szState    = NULL;
	const char*            szLabel    = NULL;

	SG_NULLARGCHECK(pValue);

	if (pChoice == NULL)
	{
		SG_ERR_CHECK(  SG_resolve__value__get_choice(pCtx, pValue, &pChoice)  );
	}
	if (pItem == NULL)
	{
		SG_ERR_CHECK(  SG_resolve__choice__get_item(pCtx, pChoice, &pItem)  );
	}

	SG_ERR_CHECK(  SG_resolve__item__get_type(pCtx, pItem, &eType)  );
	szType = SG_treenode_entry_type__type_to_label(eType);
	SG_ERR_CHECK(  SG_resolve__item__get_repo_path__working(pCtx, pItem, &szRepoPath)  );
	SG_ERR_CHECK(  SG_resolve__choice__get_state(pCtx, pChoice, &eState)  );
	SG_ERR_CHECK(  SG_resolve__state__get_info(pCtx, eState, &szState, NULL)  );
	SG_ERR_CHECK(  SG_resolve__value__get_label(pCtx, pValue, &szLabel)  );

	SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Accepted '%s' value for '%s' conflict on %s:", szLabel, szState, szType)  );
	SG_ERR_CHECK(  _console__stdout(pCtx, uIndent + 1u, "%s", szRepoPath)  );

fail:
	return;
}

/**
 * Writes the result of a merge to stdout.
 *
 * TODO: Add a message about overwriting an old value?
 */
static void _write_merge(
	SG_context*         pCtx,           //< [in] [out] Error and context info.
	SG_uint32           uIndent,        //< [in] Base amount to indent the messages.
	SG_resolve__value*  pBaselineValue, //< [in] The baseline value that was merged.
	SG_resolve__value*  pOtherValue,    //< [in] The other value that was merged.
	SG_int32            iResult,        //< [in] The result status from the merge.
	SG_resolve__value*  pResultValue,   //< [in] The result value from the merge.
	SG_resolve__choice* pChoice,        //< [in] The choice that owns the values.
	                                    //<      If NULL, this function will look it up from the values.
	SG_resolve__item*   pItem           //< [in] The item that owns the choice.
	                                    //<      If NULL, this function will look it up from the choice.
	)
{
	const char*        szResultLabel  = NULL;
	const char*        szResultTool   = NULL;
	SG_resolve__value* pAcceptedValue = NULL;

	SG_NULLARGCHECK(pBaselineValue);
	SG_NULLARGCHECK(pOtherValue);

	// if they didn't supply a choice, look it up
	if (pChoice == NULL)
	{
		SG_resolve__choice* pBaselineChoice = NULL;
		SG_resolve__choice* pOtherChoice    = NULL;

		SG_ERR_CHECK(  SG_resolve__value__get_choice(pCtx, pBaselineValue, &pBaselineChoice)  );
		SG_ERR_CHECK(  SG_resolve__value__get_choice(pCtx, pOtherValue, &pOtherChoice)  );
		SG_ARGCHECK(pBaselineChoice == pOtherChoice, pBaselineValue|pOtherValue);
		pChoice = pBaselineChoice;
	}

	// if they didn't supply an item, look it up
	if (pItem == NULL)
	{
		SG_ERR_CHECK(  SG_resolve__choice__get_item(pCtx, pChoice, &pItem)  );
	}

	// check what happened and display appropriate output
	if (pResultValue != NULL)
	{
		SG_bool bMerged = SG_FALSE;

		SG_ERR_CHECK(  SG_resolve__value__get_label(pCtx, pResultValue, &szResultLabel)  );
		SG_ERR_CHECK(  SG_resolve__value__get_merge(pCtx, pResultValue, &bMerged, NULL, NULL, NULL, &szResultTool, NULL)  );
		SG_ASSERT(bMerged == SG_TRUE);
	}
	switch (iResult)
	{
	case SG_MERGETOOL__STATUS__SUCCESSFUL:
		SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Used merge tool '%s' to create value '%s' with no conflicts.", szResultTool, szResultLabel)  );
		break;
	case SG_MERGETOOL__STATUS__CONFLICT:
		SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Used merge tool '%s' to create merged value '%s' with some conflicts remaining.", szResultTool, szResultLabel)  );
		break;
	case SG_MERGETOOL__STATUS__ERROR:
		SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Merge tool '%s' was canceled or exited with an error.", szResultTool)  );
		SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Created merged value '%s' with partial result.", szResultLabel)  );
		break;
	case SG_MERGETOOL__STATUS__NOT_ATTEMPTED:
		SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Merge tool did not attempt the merge, no value created.")  );
		break;
	case SG_MERGETOOL__STATUS__EXEC_ERROR:
		SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "There was an error launching the merge tool, no value created.")  );
		break;
	case SG_MERGETOOL__STATUS__NO_OUTPUT:
		SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "The merge tool did not generate any output, no value created.")  );
		break;
	default:
		SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Unknown merge tool result: %i", iResult)  );
		break;
	}

	// if the new value was automatically accepted as the choice's resolution
	// then let the user know
	SG_ERR_CHECK(  SG_resolve__choice__get_value__accepted(pCtx, pChoice, &pAcceptedValue)  );
	if (pAcceptedValue != NULL && pAcceptedValue == pResultValue)
	{
		SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "New merged value automatically accepted as its choice's resolution.")  );
	}

fail:
	return;
}


/*
**
** Private Functionality - Command-line Resolve Extensions
**
*/

/**
 * Ensures that a value can be viewed and then does it.
 * This is the common "view" code shared by interactive and non-interactive commands.
 *//*
static void _resolve__view(
	SG_context*         pCtx,    //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice, //< [in] The choice to view a value on.
	SG_resolve__value*  pValue,  //< [in] The value to view.
	const char*         szTool,  //< [in] Name of the tool to use to view the value.
	SG_uint32           uIndent  //< [in] Base number of times to indent console messages.
	)
{
	SG_bool bMergeable = SG_FALSE;

	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pValue);

	// make sure this is a mergeable choice
	SG_ERR_CHECK(  SG_resolve__choice__get_mergeable(pCtx, pChoice, &bMergeable)  );
	if (bMergeable == SG_FALSE)
	{
		SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Cannot view values on non-mergeable conflicts.")  );
	}
	else
	{
		SG_ERR_THROW(SG_ERR_NOTIMPLEMENTED);
		SG_UNUSED(szTool);
		//SG_ERR_CHECK(  SG_resolve__view_value(pCtx, pValue, szTool)  );
	}

fail:
	return;
}
*/
/**
 * Ensures that values can be diffed and then does it.
 * This is the common "diff" code shared by interactive and non-interactive commands.
 */
static void _resolve__diff(
	SG_context*         pCtx,    //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice, //< [in] The choice to diff values on.
	SG_resolve__value*  pLeft,   //< [in] The left value to diff.
	SG_resolve__value*  pRight,  //< [in] The right value to diff.
	const char*         szTool,  //< [in] Name of the tool to use to diff the values.
	SG_uint32           uIndent  //< [in] Base number of times to indent console messages.
	)
{
	SG_bool bMergeable = SG_FALSE;

	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pLeft);
	SG_NULLARGCHECK(pRight);

	// make sure this is a mergeable choice
	SG_ERR_CHECK(  SG_resolve__choice__get_mergeable(pCtx, pChoice, &bMergeable)  );
	if (bMergeable == SG_FALSE)
	{
		SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Cannot diff values on non-mergeable conflicts.")  );
	}
	else
	{
		SG_ERR_CHECK(  SG_resolve__diff_values(pCtx, pLeft, pRight, szTool)  );
	}

fail:
	return;
}

/**
 * Ensures that values can be merged and then does it.
 * This is the common "merge" code shared by interactive and non-interactive commands.
 */
static void _resolve__merge(
	SG_context*         pCtx,       //< [in] [out] Error and context info.
	SG_resolve__item*   pItem,      //< [in] The item that owns pChoice.
	                                //<      If NULL, it will be looked up from pChoice.
	SG_resolve__choice* pChoice,    //< [in] The choice to merge values on.
	SG_resolve__value*  pBaseline,  //< [in] The baseline value to merge.
	                                //<      May be NULL if pOther is also NULL and the conflict has exactly two leaf values.
	                                //<      In that case, the two leaf values are merged.
	SG_resolve__value*  pOther,     //< [in] The other value to merge.
	                                //<      May be NULL if pBaseline is also NULL and the conflict has exactly two leaf values.
	                                //<      In that case, the two leaf values are merged.
	const char*         szTool,     //< [in] Name of the tool to use to merge the values.
	const char*         szLabel,    //< [in] Label for the newly merged value.
	                                //<      If NULL, a label will be automatically generated.
	SG_int32            iOverwrite, //< [in] How to handle szLabel already existing.
	                                //<      Unused if szLabel is NULL.
	                                //<      If positive, overwrite the existing resolution.
	                                //<      If zero, display an error.
	                                //<      If negative, ask the user with a y/n prompt.
	SG_uint32           uIndent,    //< [in] Base number of times to indent console messages.
	SG_resolve__value** ppResult    //< [out] The value that resulted from the merge.
	                                //<       NULL if the merge wasn't performed or failed.
	)
{
	SG_resolve__value* pResult    = NULL;
	SG_bool            bMergeable = SG_FALSE;
	SG_string*         sOverwrite = NULL;

	SG_NULLARGCHECK(pChoice);
	SG_ARGCHECK(SG_BOTH_NULL_OR_NOT(pBaseline, pOther), pBaseline+pOther);

	// make quick easy checks
	SG_ERR_CHECK(  SG_resolve__choice__get_mergeable(pCtx, pChoice, &bMergeable)  );
	if (bMergeable == SG_FALSE)
	{
		SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Cannot merge values on non-mergeable conflicts.")  );
	}
	else if (iOverwrite > 0 && szLabel == NULL)
	{
		SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Overwrite option cannot be used without specifying the label to overwrite.")  );
	}
	else
	{
		// if they didn't specify values, try using the choice's two leaves
		if (pBaseline == NULL)
		{
			SG_ERR_CHECK(  _resolve__choice__get_two_leaves(pCtx, pChoice, &pBaseline, &pOther)  );
			SG_ASSERT(SG_BOTH_NULL_OR_NOT(pBaseline, pOther));
			if (pBaseline != NULL)
			{
				const char* szBaseline = NULL;
				const char* szOther    = NULL;

				SG_ERR_CHECK(  SG_resolve__value__get_label(pCtx, pBaseline, &szBaseline)  );
				SG_ERR_CHECK(  SG_resolve__value__get_label(pCtx, pOther, &szOther)  );
				SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Merging leaf values '%s' and '%s'.", szBaseline, szOther)  );
			}
		}

		// make sure we have a value now
		if (pBaseline == NULL)
		{
			SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Values must be specified unless there are exactly two leaves to infer.")  );
		}
		else
		{
			SG_resolve__value* pLabeledValue = NULL;
			SG_bool            bOverwrite    = SG_FALSE;

			// if they specified a label, check if we can handle it
			if (szLabel != NULL)
			{
				// check if the choice already has a value with this label
				SG_ERR_CHECK(  SG_resolve__choice__check_value__label(pCtx, pChoice, szLabel, 0u, &pLabeledValue)  );
				if (pLabeledValue != NULL)
				{
					// check if the caller wants to try overwriting it
					if (iOverwrite > 0)
					{
						// overwrite existing value
						bOverwrite = SG_TRUE;
					}
					else if (iOverwrite == 0)
					{
						// don't overwrite existing value
						SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Value '%s' is already in use, use --overwrite to overwrite it.", szLabel)  );
						bOverwrite = SG_FALSE;
					}
					else if (iOverwrite < 0)
					{
						// prompt to overwrite existing value
						SG_ERR_CHECK(  _console__prompt(pCtx, uIndent, &sOverwrite, "Value '%s' already in use, overwrite? (y/n): ", szLabel)  );
						bOverwrite = SG_parse_bool(SG_string__sz(sOverwrite));
					}
					else
					{
						// this should never happen
						bOverwrite = SG_FALSE;
					}

					// if they want to overwrite, check if it's possible
					if (bOverwrite == SG_TRUE)
					{
						SG_uint32 uFlags = 0u;

						SG_ERR_CHECK(  SG_resolve__value__get_flags(pCtx, pLabeledValue, &uFlags)  );
						if ((uFlags & SG_RESOLVE__VALUE__OVERWRITABLE) == 0)
						{
							bOverwrite = SG_FALSE;
							if ((uFlags & SG_RESOLVE__VALUE__AUTOMATIC) != 0)
							{
								// can't overwrite an automatic value
								SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Cannot overwrite automatic value: %s", szLabel)  );
							}
							else if ((uFlags & SG_RESOLVE__VALUE__RESULT) != 0)
							{
								// can't overwrite the current resolution
								SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Cannot overwrite resolution value: %s", szLabel)  );
							}
							else
							{
								// this should never happen
								SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Cannot overwrite value: %s", szLabel)  );
							}
						}
					}
				}
			}

			// run the merge
			if (pLabeledValue == NULL || bOverwrite == SG_TRUE)
			{
				SG_int32 iResult = SG_MERGETOOL__STATUS__MUST_BE_LAST;

				SG_ERR_CHECK(  SG_resolve__merge_values(pCtx, pBaseline, pOther, szTool, szLabel, &iResult, &pResult)  );
				SG_ERR_CHECK(  _write_merge(pCtx, uIndent, pBaseline, pOther, iResult, pResult, pChoice, pItem)  );
			}
		}
	}

	// return whatever result we had
	*ppResult = pResult;
	pResult = NULL;

fail:
	SG_STRING_NULLFREE(pCtx, sOverwrite);
	return;
}

/**
 * Ensures that a value can be accepted and then does it.
 * This is the common "accept" code shared by interactive and non-interactive commands.
 */
static void _resolve__accept(
	SG_context*         pCtx,       //< [in] [out] Error and context info.
	SG_resolve__item*   pItem,      //< [in] The item that owns pChoice.
	                                //<      If NULL, it will be looked up from pChoice.
	SG_resolve__choice* pChoice,    //< [in] The choice to accept a value on.
	SG_resolve__value*  pValue,     //< [in] The value to accept.
	SG_int32            iOverwrite, //< [in] How to handle the choice already having an accepted value.
	                                //<      If positive, overwrite the existing resolution.
	                                //<      If zero, display an error.
	                                //<      If negative, ask the user with a y/n prompt.
	SG_uint32           uIndent,    //< [in] Base number of times to indent console messages.
	SG_bool*            pAccepted   //< [out] Whether or not the value was accepted.
	)
{
	SG_resolve__value* pAcceptedValue = NULL;
	SG_bool            bOverwrite     = SG_FALSE;
	const char*        szLabel        = NULL;
	SG_string*         sOverwrite     = NULL;

	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pValue);
	SG_NULLARGCHECK(pAccepted);

	// check if the choice already has a resolution
	SG_ERR_CHECK(  SG_resolve__choice__get_value__accepted(pCtx, pChoice, &pAcceptedValue)  );
	if (pValue == pAcceptedValue)
	{
		// already resolved with given value
		SG_ERR_CHECK(  SG_resolve__value__get_label(pCtx, pValue, &szLabel)  );
		SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "The conflict is already resolved using the specified value: %s", szLabel)  );
		bOverwrite = SG_FALSE;
	}
	else if (pAcceptedValue != NULL)
	{
		if (iOverwrite > 0)
		{
			// overwrite existing value
			bOverwrite = SG_TRUE;
		}
		else if (iOverwrite == 0)
		{
			// don't overwrite existing value
			SG_ERR_CHECK(  SG_resolve__value__get_label(pCtx, pAcceptedValue, &szLabel)  );
			SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "The conflict was already resolved by accepting value: %s", szLabel)  );
			SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Use --overwrite to accept a different value.")  );
			bOverwrite = SG_FALSE;
		}
		else if (iOverwrite < 0)
		{
			// prompt to overwrite existing value
			SG_ERR_CHECK(  SG_resolve__value__get_label(pCtx, pAcceptedValue, &szLabel)  );
			SG_ERR_CHECK(  _console__prompt(pCtx, uIndent, &sOverwrite, "Choice already already resolved using value '%s', overwrite? (y/n): ", szLabel)  );
			bOverwrite = SG_parse_bool(SG_string__sz(sOverwrite));
		}
	}

	// run the accept
	if (pAcceptedValue == NULL || bOverwrite == SG_TRUE)
	{
		SG_ERR_CHECK(  SG_resolve__accept_value(pCtx, pValue)  );
		SG_ERR_CHECK(  _write_accept(pCtx, uIndent, pValue, pChoice, pItem)  );
		*pAccepted = SG_TRUE;
	}
	else
	{
		*pAccepted = SG_FALSE;
	}

fail:
	SG_STRING_NULLFREE(pCtx, sOverwrite);
	return;
}


/*
**
** Private Functionality - Interactive Helpers and Command Handlers
**
*/

/**
 * Builds the user prompt for an interactive session.
 * TODO: Update to use global command spec data.
 */
static void _interactive__build_prompt(
	SG_context*            pCtx,     //< [in] [out] Error and context info.
	_interactive__session* pSession, //< [in] Data for the current interactive session.
	SG_resolve__choice*    pChoice,  //< [in] The choice being prompted for.
	SG_string**            ppPrompt  //< [out] The built prompt.
	)
{
	SG_bool    bMergeable = SG_FALSE;
	SG_string* sPrompt    = NULL;
	SG_uint32  uIndex     = 0u;
	SG_bool    bSeparator = SG_FALSE;

	SG_NULLARGCHECK(pSession);
	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(ppPrompt);

	// check if this is a mergeable choice
	SG_ERR_CHECK(  SG_resolve__choice__get_mergeable(pCtx, pChoice, &bMergeable)  );

	// start with an empty prompt
	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &sPrompt)  );

	// run through each command and append it to the prompt
	for (uIndex = 0u; uIndex < SG_NrElements(gaInteractiveCommands); ++uIndex)
	{
		const _interactive__command_spec* pSpec = gaInteractiveCommands + uIndex;

		// only include the command if its "mergeability" matches the current choice
		if (pSpec->bMergeableOnly == SG_FALSE || bMergeable == SG_TRUE)
		{
			// if we need a separator, add one
			if (bSeparator == SG_TRUE)
			{
				SG_ERR_CHECK(  SG_string__append__sz(pCtx, sPrompt, " ")  );
			}

			// append the command's prompt string
			SG_ERR_CHECK(  SG_string__append__sz(pCtx, sPrompt, pSpec->szPrompt)  );
			bSeparator = SG_TRUE;
		}
	}

	// finish it off
	SG_ERR_CHECK(  SG_string__append__sz(pCtx, sPrompt, ": ")  );

	// return the prompt
	*ppPrompt = sPrompt;
	sPrompt = NULL;

fail:
	SG_STRING_NULLFREE(pCtx, sPrompt);
	return;
}

/**
 * Builds an automatic command to use for an interactive prompt, if one is relevent.
 */
static void _interactive__build_automatic_command(
	SG_context*            pCtx,      //< [in] [out] Error and context info.
	_interactive__session* pSession,  //< [in] Data for the current interactive session.
	SG_resolve__choice*    pChoice,   //< [in] The choice being prompted for.
	SG_string**            ppCommand, //< [out] The built command.
	                                  //<       Set to NULL if no automatic command is relevent.
	SG_string**            ppMessage  //< [out] An explanatory message to display.
	                                  //<       Set to NULL if none is necessary.
	)
{
	SG_bool    bMergeable = SG_FALSE;
	SG_string* sCommand   = NULL;
	SG_string* sMessage   = NULL;

	SG_NULLARGCHECK(pSession);
	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(ppCommand);

	// check if we should run an automatic merge
	SG_ERR_CHECK(  SG_resolve__choice__get_mergeable(pCtx, pChoice, &bMergeable)  );
	if (bMergeable == SG_TRUE)
	{
		SG_resolve__value* pLeaf1 = NULL;
		SG_resolve__value* pLeaf2 = NULL;

		SG_ERR_CHECK(  _resolve__choice__get_two_leaves(pCtx, pChoice, &pLeaf1, &pLeaf2)  );
		SG_ASSERT(SG_BOTH_NULL_OR_NOT(pLeaf1, pLeaf2));
		if (pLeaf1 != NULL)
		{
			const char* szLabel1 = NULL;
			const char* szLabel2 = NULL;

			// get the labels of the two leaves
			SG_ERR_CHECK(  SG_resolve__value__get_label(pCtx, pLeaf1, &szLabel1)  );
			SG_ERR_CHECK(  SG_resolve__value__get_label(pCtx, pLeaf2, &szLabel2)  );

			// setup the command and message
			SG_ERR_CHECK(  SG_string__alloc__format(pCtx, &sCommand, "merge %s %s", szLabel1, szLabel2)  );
			SG_ERR_CHECK(  SG_string__alloc__sz(pCtx, &sMessage, "Automatically merging last two leaves, close tool without saving to cancel.")  );
		}
	}

	// return whatever we built
	*ppCommand = sCommand;
	sCommand = NULL;
	*ppMessage = sMessage;
	sMessage = NULL;

fail:
	SG_STRING_NULLFREE(pCtx, sCommand);
	SG_STRING_NULLFREE(pCtx, sMessage);
	return;
}

/**
 * Parses a command entered by the user at the interactive prompt.
 */
static void _interactive__parse_command(
	SG_context*            pCtx,       //< [in] [out] Error and context info.
	char*                  szCommand,  //< [in] The command entered by the user.
	                                   //<      Will be tokenized during parsing.
	SG_resolve__choice*    pChoice,    //< [in] The choice to use to parse value names.
	                                   //<      If NULL, value name parsing is skipped.
	_interactive__command* pCommand,   //< [out] The parsed command.
	                                   //<       Data points to tokens within szInput.
	SG_bool*               pUsageError //< [out] Whether or not there was a usage error.
	)
{
	void*   pHandle = NULL;
	SG_bool bUsage  = SG_FALSE;

	SG_NULLARGCHECK(szCommand);
	SG_NULLARGCHECK(pCommand);
	SG_NULLARGCHECK(pUsageError);

	// initialize the command with default data
	pCommand->szCommand = NULL;
	pCommand->szValue1  = NULL;
	pCommand->szValue2  = NULL;
	pCommand->szTool    = NULL;
	pCommand->szLabel   = NULL;
	pCommand->pValue1   = NULL;
	pCommand->pValue2   = NULL;

	// get the first token they entered, which will be the command
	SG_ERR_CHECK(  _SG_strtok(pCtx, szCommand, NULL, SG_TRUE, &pCommand->szCommand, &pHandle)  );
	if (pCommand->szCommand != NULL)
	{
		const char*  szToken = NULL;
		const char** ppNext  = NULL;

		// keep getting tokens until we run out
		SG_ERR_CHECK(  _SG_strtok(pCtx, NULL, NULL, SG_TRUE, &szToken, &pHandle)  );
		while (szToken != NULL)
		{
			// check what the token means
			if (SG_stricmp(szToken, "--tool") == 0)
			{
				if (ppNext == NULL)
				{
					// expect a tool name next
					ppNext = &pCommand->szTool;
				}
				else
				{
					// usage error
					bUsage = SG_TRUE;
				}
			}
			else if (SG_stricmp(szToken, "--label") == 0)
			{
				if (ppNext == NULL)
				{
					// expect a label name next
					ppNext = &pCommand->szLabel;
				}
				else
				{
					// usage error
					bUsage = SG_TRUE;
				}
			}
			else if (ppNext == NULL)
			{
				// we're not expecting anything
				// we'll assume the token is a value name
				if (pCommand->szValue1 == NULL)
				{
					// it's the first value so far
					pCommand->szValue1 = szToken;
				}
				else if (pCommand->szValue2 == NULL)
				{
					// it's the second value so far
					pCommand->szValue2 = szToken;
				}
				else
				{
					// usage error
					bUsage = SG_TRUE;
				}
			}
			else
			{
				// we're expecting something
				// assign the token to that
				*ppNext = szToken;
				ppNext = NULL;
			}

			// get the next token
			SG_ERR_CHECK(  _SG_strtok(pCtx, NULL, NULL, SG_TRUE, &szToken, &pHandle)  );
		}
	}

	// try parsing non-NULL value names into values
	if (pChoice != NULL)
	{
		if (pCommand->szValue1 != NULL)
		{
			SG_ERR_CHECK(  SG_resolve__choice__check_value__label(pCtx, pChoice, pCommand->szValue1, 0u, &pCommand->pValue1)  );
		}
		if (pCommand->szValue2 != NULL)
		{
			SG_ERR_CHECK(  SG_resolve__choice__check_value__label(pCtx, pChoice, pCommand->szValue2, 0u, &pCommand->pValue2)  );
		}
	}

	// return data
	*pUsageError = bUsage;

fail:
	return;
}

/**
 * Writes the usage of a command spec to the console.
 */
static void _interactive__write_usage(
	SG_context*                       pCtx,     //< [in] [out] Error and context info.
	_interactive__session*            pSession, //< [in] The current interactive session.
	const _interactive__command_spec* pSpec     //< [in] The command spec to write usage for.
	)
{
	SG_ERR_CHECK(  _console__stdout(pCtx, pSession->uIndent, "%s", pSpec->szUsage)  );
	if (pSpec->szUsageExtra1 != NULL)
	{
		SG_ERR_CHECK(  _console__stdout(pCtx, pSession->uIndent + 1u, "%s", pSpec->szUsageExtra1)  );
	}
	if (pSpec->szUsageExtra2 != NULL)
	{
		SG_ERR_CHECK(  _console__stdout(pCtx, pSession->uIndent + 1u, "%s", pSpec->szUsageExtra2)  );
	}
	if (pSpec->szUsageExtra3 != NULL)
	{
		SG_ERR_CHECK(  _console__stdout(pCtx, pSession->uIndent + 1u, "%s", pSpec->szUsageExtra3)  );
	}
	if (pSpec->szUsageExtra4 != NULL)
	{
		SG_ERR_CHECK(  _console__stdout(pCtx, pSession->uIndent + 1u, "%s", pSpec->szUsageExtra4)  );
	}
	if (pSpec->szUsageExtra5 != NULL)
	{
		SG_ERR_CHECK(  _console__stdout(pCtx, pSession->uIndent + 1u, "%s", pSpec->szUsageExtra5)  );
	}

fail:
	return;
}

/**
 * Checks a command spec's argument requirement against actual argument values.
 */
static SG_bool _interactive__match_command__valid(
	_interactive__command_argument eArgument, //< [in] The type of argument requirement.
	void*                          pValue,    //< [in] The value to check against the requirement.
	void*                          pValue2    //< [in] The second value to check.  Only used with ARGUMENT_TOGETHER.
	)
{
	SG_ASSERT(eArgument != ARGUMENT_COUNT);

	switch (eArgument)
	{
	case ARGUMENT_REQUIRED:
		return pValue == NULL ? SG_FALSE : SG_TRUE;
	case ARGUMENT_FORBIDDEN:
		return pValue == NULL ? SG_TRUE : SG_FALSE;
	case ARGUMENT_TOGETHER:
		return SG_BOTH_NULL_OR_NOT(pValue, pValue2) ? SG_TRUE : SG_FALSE;
	case ARGUMENT_OPTIONAL:
		return SG_TRUE;
	default:
		return SG_FALSE;
	}
}

/**
 * Checks if a parsed command from the user matches a command spec.
 */
static void _interactive__match_command(
	SG_context*                       pCtx,     //< [in] [out] Error and context info.
	_interactive__session*            pSession, //< [in] The current interactive session.
	SG_resolve__choice*               pChoice,  //< [in] The choice that the command was entered for.
	_interactive__command*            pCommand, //< [in] The parsed command from the user.
	const _interactive__command_spec* pSpec,    //< [in] The spec to match the parsed command against.
	SG_bool*                          pMatch,   //< [out] True if the command itself matches.
	SG_bool*                          pValid    //< [out] True if the arguments are valid.
	                                            //<       Ignored if pMatch is false.
	)
{
	SG_NULLARGCHECK(pSession);
	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pCommand);
	SG_NULLARGCHECK(pSpec);
	SG_NULLARGCHECK(pMatch);
	SG_NULLARGCHECK(pValid);

	// check if the command name matches the spec
	if (
		   SG_stricmp(pCommand->szCommand, pSpec->szName) == 0
		|| (
			   pSpec->szAbbreviation != NULL
			&& SG_stricmp(pCommand->szCommand, pSpec->szAbbreviation) == 0
			)
		)
	{
		SG_bool bMergeable = SG_FALSE;

		*pMatch = SG_TRUE;

		// check the command against the spec for validity
		SG_ERR_CHECK(  SG_resolve__choice__get_mergeable(pCtx, pChoice, &bMergeable)  );
		if (pSpec->bMergeableOnly == SG_TRUE && bMergeable == SG_FALSE)
		{
			*pValid = SG_FALSE;
		}
		else if (
			   _interactive__match_command__valid(pSpec->eValue1, (void*)pCommand->pValue1, (void*)pCommand->pValue2) == SG_FALSE
			|| _interactive__match_command__valid(pSpec->eValue2, (void*)pCommand->pValue2, (void*)pCommand->pValue1) == SG_FALSE
			|| _interactive__match_command__valid(pSpec->eTool,   (void*)pCommand->szTool,  NULL) == SG_FALSE
			|| _interactive__match_command__valid(pSpec->eLabel,  (void*)pCommand->szLabel, NULL) == SG_FALSE
			)
		{
			*pValid = SG_FALSE;
		}
		else
		{
			*pValid = SG_TRUE;
		}
	}
	else
	{
		*pMatch = SG_FALSE;
	}

fail:
	return;
}

/**
 * Advances an interactive session to the next choice/conflict.
 */
static void _interactive__advance_choice(
	SG_context*            pCtx,      //< [in] [out] Error and context info.
	_interactive__session* pSession,  //< [in] The interactive session to advance.
	SG_bool                bResolved, //< [in] Whether or not the current choice was resolved.
	SG_bool                bQuit      //< [in] Whether or not the session should be stopped.
	)
{
	SG_UNUSED(pCtx);

	// check if we're quitting altogether
	if (bQuit == SG_TRUE)
	{
		// quitting, mark the session as inactive
		pSession->bActive = SG_FALSE;
	}

	// if we have a max choice count,
	// decrement it and stop if we hit zero
	if (pSession->uMax > 0u)
	{
		pSession->uMax -= 1u;
		if (pSession->uMax == 0u)
		{
			pSession->bActive = SG_FALSE;
		}
	}

	// if the session will still be active, move it to the next choice
	if (pSession->bActive)
	{
		// increment the choice index if necessary
		if (bResolved == SG_FALSE || pSession->bOnlyUnresolved == SG_FALSE)
		{
			++pSession->uChoice;
		}

		// reset choice-specific data
		pSession->bAutomatic = SG_FALSE;
		pSession->bWrite     = SG_TRUE;
	}
}

/**
 * _interactive__command_handler for the "view" command.
 *//*
static void _interactive__handler__view(
	SG_context*                  pCtx,
	_interactive__session*       pSession,
	SG_resolve__choice*          pChoice,
	const _interactive__command* pCommand
	)
{
	const char* szTool = pCommand->szTool;

	SG_NULLARGCHECK(pSession);
	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pCommand);

	// if no tool was specified at the prompt, fall back to our default
	if (szTool == NULL)
	{
		szTool = pSession->szViewTool;
	}

	// view the value
	SG_ERR_CHECK(  _resolve__view(pCtx, pChoice, pCommand->pValue1, szTool, pSession->uIndent)  );

fail:
	return;
}
*/
/**
 * _interactive__command_handler for the "diff" command.
 */
static void _interactive__handler__diff(
	SG_context*                  pCtx,
	_interactive__session*       pSession,
	SG_resolve__choice*          pChoice,
	const _interactive__command* pCommand
	)
{
	const char* szTool = pCommand->szTool;

	SG_NULLARGCHECK(pSession);
	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pCommand);

	// if no tool was specified at the prompt, fall back to our default
	if (szTool == NULL)
	{
		szTool = pSession->szDiffTool;
	}

	// diff the values
	SG_ERR_CHECK(  _resolve__diff(pCtx, pChoice, pCommand->pValue1, pCommand->pValue2, szTool, pSession->uIndent)  );

fail:
	return;
}

/**
 * _interactive__command_handler for the "merge" command.
 */
static void _interactive__handler__merge(
	SG_context*                  pCtx,
	_interactive__session*       pSession,
	SG_resolve__choice*          pChoice,
	const _interactive__command* pCommand
	)
{
	const char*        szTool  = pCommand->szTool;
	SG_resolve__value* pResult = NULL;

	SG_NULLARGCHECK(pSession);
	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pCommand);

	// if no tool was specified at the prompt, fall back to our default
	if (szTool == NULL)
	{
		szTool = pSession->szMergeTool;
	}

	// run the merge
	SG_ERR_CHECK(  _resolve__merge(pCtx, NULL, pChoice, pCommand->pValue1, pCommand->pValue2, szTool, pCommand->szLabel, -1, pSession->uIndent, &pResult)  );
	if (pResult != NULL)
	{
		SG_resolve__value* pAccepted = NULL;

		// if the result value was automatically accepted
		// then advance to the next choice
		SG_ERR_CHECK(  SG_resolve__choice__get_value__accepted(pCtx, pChoice, &pAccepted)  );
		if (pAccepted != NULL && pAccepted == pResult)
		{
			SG_ERR_CHECK(  _interactive__advance_choice(pCtx, pSession, SG_TRUE, SG_FALSE)  );
		}
	}

fail:
	return;
}

/**
 * _interactive__command_handler for the "accept" command.
 */
static void _interactive__handler__accept(
	SG_context*                  pCtx,
	_interactive__session*       pSession,
	SG_resolve__choice*          pChoice,
	const _interactive__command* pCommand
	)
{
	SG_bool bAccepted = SG_FALSE;

	SG_NULLARGCHECK(pSession);
	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pCommand);

	// accept the value
	SG_ERR_CHECK(  _resolve__accept(pCtx, NULL, pChoice, pCommand->pValue1, -1, pSession->uIndent, &bAccepted)  );
	if (bAccepted == SG_TRUE)
	{
		// if the accept succeeded, advance to the next choice
		SG_ERR_CHECK(  _interactive__advance_choice(pCtx, pSession, SG_TRUE, SG_FALSE)  );
	}

fail:
	return;
}

/**
 * _interactive__command_handler for the "repeat" command.
 */
static void _interactive__handler__repeat(
	SG_context*                  pCtx,
	_interactive__session*       pSession,
	SG_resolve__choice*          pChoice,
	const _interactive__command* pCommand
	)
{
	SG_NULLARGCHECK(pSession);
	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pCommand);

	// write the choice out again next time
	pSession->bWrite = SG_TRUE;

fail:
	return;
}

/**
 * _interactive__command_handler for the "skip" command.
 */
static void _interactive__handler__skip(
	SG_context*                  pCtx,
	_interactive__session*       pSession,
	SG_resolve__choice*          pChoice,
	const _interactive__command* pCommand
	)
{
	SG_NULLARGCHECK(pSession);
	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pCommand);

	// advance to the next choice
	SG_ERR_CHECK(  _interactive__advance_choice(pCtx, pSession, SG_FALSE, SG_FALSE)  );

fail:
	return;
}

/**
 * _interactive__command_handler for the "quit" command.
 */
static void _interactive__handler__quit(
	SG_context*                  pCtx,
	_interactive__session*       pSession,
	SG_resolve__choice*          pChoice,
	const _interactive__command* pCommand
	)
{
	SG_NULLARGCHECK(pSession);
	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pCommand);

	// end the session
	SG_ERR_CHECK(  _interactive__advance_choice(pCtx, pSession, SG_FALSE, SG_TRUE)  );

fail:
	return;
}

/**
 * _interactive__command_handler for the "help" command.
 */
static void _interactive__handler__help(
	SG_context*                  pCtx,
	_interactive__session*       pSession,
	SG_resolve__choice*          pChoice,
	const _interactive__command* pCommand
	)
{
	SG_bool   bMergeable = SG_FALSE;
	SG_uint32 uIndex     = 0u;

	SG_NULLARGCHECK(pSession);
	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pCommand);

	// check if this is a mergeable choice
	SG_ERR_CHECK(  SG_resolve__choice__get_mergeable(pCtx, pChoice, &bMergeable)  );

	// run through each command and write its usage
	for (uIndex = 0u; uIndex < SG_NrElements(gaInteractiveCommands); ++uIndex)
	{
		const _interactive__command_spec* pSpec = gaInteractiveCommands + uIndex;

		// only include the command if its "mergeability" matches the current choice
		if (pSpec->bMergeableOnly == SG_FALSE || bMergeable == SG_TRUE)
		{
			SG_ERR_CHECK(  _interactive__write_usage(pCtx, pSession, pSpec)  );
		}
	}

fail:
	return;
}

/**
 * Displays and handles a single interactive prompt/command for a given choice.
 */
static void _interactive__choice(
	SG_context*            pCtx,     //< [in] [out] Error and context info.
	_interactive__session* pSession, //< [in] Data for the current interactive session.
	SG_resolve__choice*    pChoice   //< [in] The choice being prompted about.
	)
{
	SG_string*                 sPrompt    = NULL;
	SG_string*                 sCommand   = NULL;
	SG_string*                 sMessage   = NULL;
	char*                      szCommand  = NULL;
	_interactive__command cCommand;
	SG_bool                    bUsage     = SG_FALSE;

	SG_NULLARGCHECK(pSession);
	SG_NULLARGCHECK(pChoice);

	// if we need to write the current choice, do that
	if (pSession->bWrite == SG_TRUE)
	{
		SG_ERR_CHECK(  _write_choice(pCtx, pSession->uIndent, pChoice, NULL)  );
		pSession->bWrite = SG_FALSE;
	}

	// build the prompt
	SG_ERR_CHECK(  _interactive__build_prompt(pCtx, pSession, pChoice, &sPrompt)  );

	// if we haven't used an automatic command for this choice
	// then build any automatic command that's relevent
	if (pSession->bAutomatic == SG_FALSE)
	{
		SG_ERR_CHECK(  _interactive__build_automatic_command(pCtx, pSession, pChoice, &sCommand, &sMessage)  );
	}

	// check if we have an automatic command
	if (sCommand == NULL)
	{
		// no automatic command, ask the user for one
		SG_ERR_CHECK(  _console__prompt(pCtx, pSession->uIndent, &sCommand, "%s", SG_string__sz(sPrompt))  );
	}
	else
	{
		// automatic command, write it out with the prompt
		SG_ERR_CHECK(  _console__stdout(pCtx, pSession->uIndent, "%s%s", SG_string__sz(sPrompt), SG_string__sz(sCommand))  );

		// if there's a message to go with it, write that as well
		if (sMessage != NULL)
		{
			SG_ERR_CHECK(  _console__stdout(pCtx, pSession->uIndent, "%s", SG_string__sz(sMessage))  );
		}

		// mark that we've used an automatic command
		pSession->bAutomatic = SG_TRUE;
	}

	// unwrap the command string, we just need the buffer
	SG_ERR_CHECK(  SG_string__sizzle(pCtx, &sCommand, (SG_byte**)&szCommand, NULL)  );

	// parse the command
	SG_ERR_CHECK(  _interactive__parse_command(pCtx, szCommand, pChoice, &cCommand, &bUsage)  );

	// decide what to do
	if (cCommand.szValue1 != NULL && cCommand.pValue1 == NULL)
	{
		// value1 failed to parse
		SG_ERR_CHECK(  _console__stdout(pCtx, pSession->uIndent, "Unknown value: %s", cCommand.szValue1)  );
	}
	else if (cCommand.szValue2 != NULL && cCommand.pValue2 == NULL)
	{
		// value2 failed to parse
		SG_ERR_CHECK(  _console__stdout(pCtx, pSession->uIndent, "Unknown value: %s", cCommand.szValue2)  );
	}
	else
	{
		SG_uint32 uIndex = 0u;
		SG_bool   bMatch = SG_FALSE;

		// look for the command to execute
		for (uIndex = 0u; uIndex < SG_NrElements(gaInteractiveCommands); ++uIndex)
		{
			const _interactive__command_spec* pSpec  = gaInteractiveCommands + uIndex;
			SG_bool                           bValid = SG_FALSE;

			SG_ERR_CHECK(  _interactive__match_command(pCtx, pSession, pChoice, &cCommand, pSpec, &bMatch, &bValid)  );
			if (bMatch == SG_TRUE)
			{
				if (bUsage == SG_FALSE && bValid == SG_TRUE)
				{
					// right command with valid arguments, execute it
					SG_ERR_CHECK(  pSpec->fHandler(pCtx, pSession, pChoice, &cCommand)  );
				}
				else
				{
					// right command with bad arguments, print its usage
					SG_ERR_CHECK(  _interactive__write_usage(pCtx, pSession, pSpec)  );
				}
				break;
			}
		}

		// if we didn't find a command, write an error
		if (bMatch == SG_FALSE)
		{
			SG_ERR_CHECK(  _console__stdout(pCtx, pSession->uIndent, "Unknown command: %s", cCommand.szCommand)  );
		}
	}

	// add a blank line after whatever output we had
	SG_ERR_CHECK(  _console__stdout(pCtx, 0u, NULL)  );

fail:
	SG_STRING_NULLFREE(pCtx, sPrompt);
	SG_STRING_NULLFREE(pCtx, sCommand);
	SG_STRING_NULLFREE(pCtx, sMessage);
	SG_NULLFREE(pCtx, szCommand);
	return;
}

/**
 * Displays and handles a single interactive prompt/command.
 */
static void _interactive__next(
	SG_context*            pCtx,    //< [in] [out] Error and context info.
	_interactive__session* pSession //< [in] An initialized _interactive__session structure.
	)
{
	SG_pendingtree*     pPendingTree = NULL;
	SG_resolve*         pResolve     = NULL;
	SG_resolve__item*   pItem        = NULL;
	SG_resolve__choice* pChoice      = NULL;

	SG_NULLARGCHECK(pSession);

	// allocate a fresh pendingtree and resolve data
	SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pSession->pCwd, &pPendingTree)  );
	SG_ERR_CHECK(  SG_RESOLVE__ALLOC(pCtx, pPendingTree, &pResolve)  );

	// if we have a GID, find the corresponding resolve item
	if (pSession->szGid != NULL)
	{
		SG_ERR_CHECK(  SG_resolve__check_item__gid(pCtx, pResolve, pSession->szGid, &pItem)  );
	}

	// find the choice at the current index
	SG_ERR_CHECK(  _resolve__get_choice__index(pCtx, pResolve, pItem, pSession->uChoice, pSession->bOnlyUnresolved, &pChoice)  );
	if (pChoice == NULL)
	{
		// if we ran out of choices, then stop the session
		pSession->bActive = SG_FALSE;
	}
	else
	{
		// display/handle a prompt for this choice
		SG_ERR_CHECK(  _interactive__choice(pCtx, pSession, pChoice)  );

		// save any data that changed
		SG_ERR_CHECK(  SG_resolve__save(pCtx, pResolve, SG_TRUE)  );
	}

fail:
	SG_RESOLVE_NULLFREE(pCtx, pResolve);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	return;
}


/*
**
** Private Functionality - Action Implementations
**
*/

/**
 * Lists conflicts one by one and asks the user what to do with each.
 */
void _action__interactive(
	SG_context*        pCtx,        //< [in] [out] Error and context info.
	const SG_pathname* pCwd,        //< [in] The current working directory to allocate a pendingtree from.
	const char*        szPath,      //< [in] If non-NULL, only list conflicts on this item (as supplied by user).
	SG_bool            bAll,        //< [in] If true, include already resolved conflicts in the list.
	SG_uint32          uMax,        //< [in] Maximum number of conflicts to list.
	const char*        szViewTool,  //< [in] If non-NULL, use this viewer tool for view commands where the user doesn't specify one.
	const char*        szDiffTool,  //< [in] If non-NULL, use this viewer tool for view commands where the user doesn't specify one.
	const char*        szMergeTool, //< [in] If non-NULL, use this viewer tool for view commands where the user doesn't specify one.
	SG_uint32          uIndent      //< [in] Base amount to indent all messages.
	)
{
	_interactive__session cSession;
	SG_pendingtree*       pPendingTree = NULL;

	// initialize szGid so that NULLFREE behaves properly in the "fail" label,
	// if we end up jumping there before it's otherwise initialized
	cSession.szGid = NULL;

	SG_NULLARGCHECK(pCwd);

	// if they specified a path, get the GID of the item at that path
	if (szPath != NULL)
	{
		SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pCwd, &pPendingTree)  );
		SG_ERR_CHECK(  _parse_path(pCtx, pPendingTree, szPath, &cSession.szGid)  );
		SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	}

	// initialize the rest of the prompt data
	cSession.pCwd            = pCwd;
	cSession.bOnlyUnresolved = (bAll == SG_FALSE ? SG_TRUE : SG_FALSE);
	cSession.uMax            = uMax;
	cSession.szViewTool      = szViewTool;
	cSession.szDiffTool      = szDiffTool;
	cSession.szMergeTool     = szMergeTool;
	cSession.uIndent         = uIndent;
	cSession.uChoice         = 0u;
	cSession.bActive         = SG_TRUE;
	cSession.bWrite          = SG_TRUE;
	cSession.bAutomatic      = SG_FALSE;

	// run a prompt loop until it tells us to stop
	while (cSession.bActive != SG_FALSE)
	{
		SG_ERR_CHECK(  _interactive__next(pCtx, &cSession)  );
	}

fail:
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_NULLFREE(pCtx, cSession.szGid);
	return;
}

/**
 * Lists conflicts.
 */
void _action__list(
	SG_context*        pCtx,   //< [in] [out] Error and context info.
	const SG_pathname* pCwd,   //< [in] The current working directory to allocate a pendingtree from.
	const char*        szPath, //< [in] If non-NULL, only list conflicts on this item (as supplied by user).
	SG_bool            bAll,   //< [in] If true, include already resolved conflicts in the list.
	SG_uint32          uMax,   //< [in] Maximum number of conflicts to list.
	SG_uint32          uIndent //< [in] Base amount to indent all messages.
	)
{
	SG_pendingtree*              pPendingTree    = NULL;
	SG_resolve*                  pResolve        = NULL;
	SG_bool                      bOnlyUnresolved = SG_FALSE;
	char*                        szGid           = NULL;
	_write_choice__callback__data cData;

	// load the pendingtree and resolve data
	SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pCwd, &pPendingTree)  );
	SG_ERR_CHECK(  SG_RESOLVE__ALLOC(pCtx, pPendingTree, &pResolve)  );

	// setup the callback's data
	cData.uIndent = uIndent;
	cData.uMax    = uMax;
	cData.uCount  = 0u;

	// decide what we're listing
	if (bAll == SG_FALSE)
	{
		bOnlyUnresolved = SG_TRUE;
	}
	else
	{
		bOnlyUnresolved = SG_FALSE;
	}

	// decide which set of choices to iterate
	if (szPath == NULL)
	{
		// no item specified, iterate all of them
		SG_ERR_CHECK(  SG_resolve__foreach_choice(pCtx, pResolve, bOnlyUnresolved, _write_choice__callback, (void*)&cData)  );
	}
	else
	{
		SG_resolve__item* pItem = NULL;

		// an item was specified, just iterate that one
		SG_ERR_CHECK(  _parse_path(pCtx, pPendingTree, szPath, &szGid)  );
		SG_ERR_CHECK(  SG_resolve__get_item__gid(pCtx, pResolve, szGid, &pItem)  );
		SG_ERR_CHECK(  SG_resolve__item__foreach_choice(pCtx, pItem, bOnlyUnresolved, _write_choice__callback, (void*)&cData)  );
	}

fail:
	SG_RESOLVE_NULLFREE(pCtx, pResolve);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_NULLFREE(pCtx, szGid);
	return;
}

/**
 * Accepts a value on multiple conflicts.
 */
void _action__accept_all(
	SG_context*        pCtx,       //< [in] [out] Error and context info.
	const SG_pathname* pCwd,       //< [in] The current working directory to allocate a pendingtree from.
	const char*        szValue,    //< [in] Name of the value to accept.
	const char*        szPath,     //< [in] Path of the file to accept values on, as specified by the user.
	                               //<      If NULL, the accept will not be restricted by file.
	SG_resolve__state  eState,     //< [in] The conflict state/type to accept values on.
	                               //<      If SG_RESOLVE__STATE__COUNT, the accept will not be restricted on conflict state/type.
	SG_bool            bOverwrite, //< [in] If true, overwrite the resolutions of matching conflicts that are already resolved.
	                               //<      If false, ignore matching conflicts that are already resolved.
	SG_uint32          uIndent     //< [in] Base amount to indent all messages.
	)
{
	SG_bool         bOnlyUnresolved = bOverwrite == SG_FALSE ? SG_TRUE : SG_FALSE;
	SG_bool         bContinue       = SG_TRUE;
	SG_uint32       uChoice         = 0u;
	SG_pendingtree* pPendingTree    = NULL;
	SG_resolve*     pResolve        = NULL;
	char*           szGid           = NULL;

	// keep going until we're done
	while (bContinue == SG_TRUE)
	{
		SG_resolve__item*   pItem   = NULL;
		SG_resolve__choice* pChoice = NULL;

		// allocate a fresh pendingtree and resolve data
		SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pCwd, &pPendingTree)  );
		SG_ERR_CHECK(  SG_RESOLVE__ALLOC(pCtx, pPendingTree, &pResolve)  );

		// if we have a path, get the corresponding item
		if (szPath != NULL)
		{
			// if we haven't parsed the path into a GID yet, do that
			// otherwise reuse the GID from last time, because the item
			// may no longer be at its original path from szPath
			if (szGid == NULL)
			{
				SG_ERR_CHECK(  _parse_path(pCtx, pPendingTree, szPath, &szGid)  );
			}

			// get the item to operate on from the GID
			SG_ERR_CHECK(  SG_resolve__get_item__gid(pCtx, pResolve, szGid, &pItem)  );
		}

		// find the choice at the current index
		SG_ERR_CHECK(  _resolve__get_choice__index(pCtx, pResolve, pItem, uChoice, bOnlyUnresolved, &pChoice)  );
		if (pChoice == NULL)
		{
			// if we ran out of choices, then we're done
			bContinue = SG_FALSE;
		}
		else
		{
			SG_resolve__state eCurrentState = SG_RESOLVE__STATE__COUNT;

			// if we only care about one type of choice,
			// then check if this one matches
			SG_ERR_CHECK(  SG_resolve__choice__get_state(pCtx, pChoice, &eCurrentState)  );
			if (
				   eState != SG_RESOLVE__STATE__COUNT
				&& eState != eCurrentState
				)
			{
				// skip this choice
				uChoice += 1u;
			}
			else
			{
				SG_resolve__value* pValue = NULL;

				// check if this choice has a value with the specified label
				SG_ERR_CHECK(  SG_resolve__choice__check_value__label(pCtx, pChoice, szValue, 0u, &pValue)  );
				if (pValue == NULL)
				{
					// skip this choice
					uChoice += 1u;
				}
				else
				{
					// accept the value
					SG_ERR_CHECK(  SG_resolve__accept_value(pCtx, pValue)  );
					SG_ERR_CHECK(  _write_accept(pCtx, uIndent, pValue, pChoice, pItem)  );

					// advance to the next choice
					if (bOnlyUnresolved == SG_FALSE)
					{
						// This is only necessary if we're iterating all choices.
						// If we're only iterating unresolved choices, it will happen
						// automatically because the choice we were on is now resolved.
						uChoice += 1u;
					}
				}
			}
		}

		// save any data that changed
		SG_ERR_CHECK(  SG_resolve__save(pCtx, pResolve, SG_TRUE)  );

		// free everything so we can reload it for the next choice
		// Note: don't free szGid, we want to reuse it next time.
		SG_RESOLVE_NULLFREE(pCtx, pResolve);
		SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	}

fail:
	SG_RESOLVE_NULLFREE(pCtx, pResolve);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_NULLFREE(pCtx, szGid);
	return;
}

/**
 * Accepts a value on a conflict.
 */
void _action__accept_one(
	SG_context*        pCtx,       //< [in] [out] Error and context info.
	const SG_pathname* pCwd,       //< [in] The current working directory to allocate a pendingtree from.
	const char*        szValue,    //< [in] Name of the value to accept.
	const char*        szPath,     //< [in] Path of the file to accept a value on, as specified by the user.
	SG_resolve__state  eState,     //< [in] The state to accept a value on.
	                               //<      May be SG_RESOLVE__STATE__COUNT if the specified item only has a single conflict.
	SG_bool            bOverwrite, //< [in] If true, the item's existing resolution may be overwritten.
	                               //<      If false, an already resolved item will produce an error.
	SG_uint32          uIndent     //< [in] Base amount to indent all messages.
	)
{
	SG_pendingtree*     pPendingTree = NULL;
	SG_resolve*         pResolve     = NULL;
	char*               szGid        = NULL;
	SG_resolve__item*   pItem        = NULL;
	SG_resolve__choice* pChoice      = NULL;

	SG_NULLARGCHECK(pCwd);
	SG_NULLARGCHECK(szValue);
	SG_NULLARGCHECK(szPath);

	// load the pendingtree and resolve data
	SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pCwd, &pPendingTree)  );
	SG_ERR_CHECK(  SG_RESOLVE__ALLOC(pCtx, pPendingTree, &pResolve)  );

	// parse the specified path
	SG_ERR_CHECK(  _parse_path(pCtx, pPendingTree, szPath, &szGid)  );

	// find a single item/choice to accept the value on
	SG_ERR_CHECK(  _find_choice(pCtx, 0u, pResolve, szGid, eState, &pItem, &pChoice)  );
	if (pItem != NULL && pChoice != NULL)
	{
		SG_resolve__value* pValue = NULL;

		// get the named value
		SG_ERR_CHECK(  SG_resolve__choice__check_value__label(pCtx, pChoice, szValue, 0u, &pValue)  );
		if (pValue == NULL)
		{
			SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Unknown value: %s", szValue)  );
		}
		else
		{
			SG_bool bAccepted = SG_FALSE;

			// run the accept
			SG_ERR_CHECK(  _resolve__accept(pCtx, pItem, pChoice, pValue, (SG_int32)bOverwrite, uIndent, &bAccepted)  );
			if (bAccepted == SG_TRUE)
			{
				// if the accept succeeded, save the changes
				SG_ERR_CHECK(  SG_resolve__save(pCtx, pResolve, SG_TRUE)  );
			}
		}
	}

fail:
	SG_RESOLVE_NULLFREE(pCtx, pResolve);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_NULLFREE(pCtx, szGid);
	return;
}

/**
 * Views a value on a conflict.
 *//*
static void _action__view(
	SG_context*        pCtx,    //< [in] [out] Error and context info.
	const SG_pathname* pCwd,    //< [in] The current working directory to allocate a pendingtree from.
	const char*        szValue, //< [in] Name of the value to view.
	const char*        szPath,  //< [in] Path of the file to view a value on, as specified by the user.
	SG_resolve__state  eState,  //< [in] The state to view a value on.
	                            //<      May be SG_RESOLVE__STATE__COUNT if the specified item only has a single conflict.
	const char*        szTool,  //< [in] Name of a viewer tool to use.
	                            //<      May be NULL to use the default.
	SG_uint32          uIndent  //< [in] Base amount to indent all messages.
	)
{
	SG_pendingtree*     pPendingTree = NULL;
	SG_resolve*         pResolve     = NULL;
	char*               szGid        = NULL;
	SG_resolve__choice* pChoice      = NULL;

	SG_NULLARGCHECK(pCwd);
	SG_NULLARGCHECK(szValue);
	SG_NULLARGCHECK(szPath);

	// load the pendingtree and resolve data
	SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pCwd, &pPendingTree)  );
	SG_ERR_CHECK(  SG_RESOLVE__ALLOC(pCtx, pPendingTree, &pResolve)  );

	// find the requested choice
	SG_ERR_CHECK(  _parse_path(pCtx, pPendingTree, szPath, &szGid)  );
	SG_ERR_CHECK(  _find_choice(pCtx, uIndent, pResolve, szGid, eState, NULL, &pChoice)  );
	if (pChoice != NULL)
	{
		SG_resolve__value* pValue = NULL;

		// get the named value
		SG_ERR_CHECK(  SG_resolve__choice__check_value__label(pCtx, pChoice, szValue, 0u, &pValue)  );
		if (pValue == NULL)
		{
			SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Unknown value: %s", szValue)  );
		}
		else
		{
			// run the view
			SG_ERR_CHECK(  _resolve__view(pCtx, pChoice, pValue, szTool, uIndent)  );
		}
	}

fail:
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_RESOLVE_NULLFREE(pCtx, pResolve);
	SG_NULLFREE(pCtx, szGid);
	return;
}
*/
/**
 * Diffs two values on a conflict.
 */
static void _action__diff(
	SG_context*        pCtx,    //< [in] [out] Error and context info.
	const SG_pathname* pCwd,    //< [in] The current working directory to allocate a pendingtree from.
	const char*        szLeft,  //< [in] Name of the left value to diff.
	const char*        szRight, //< [in] Name of the right value to diff.
	const char*        szPath,  //< [in] Path of the file to diff values on, as specified by the user.
	SG_resolve__state  eState,  //< [in] The state to diff values on.
	                            //<      May be SG_RESOLVE__STATE__COUNT if the specified item only has a single conflict.
	const char*        szTool,  //< [in] Name of a diff tool to use.
	                            //<      May be NULL to use the default.
	SG_uint32          uIndent  //< [in] Base amount to indent all messages.
	)
{
	SG_pendingtree*     pPendingTree = NULL;
	SG_resolve*         pResolve     = NULL;
	char*               szGid        = NULL;
	SG_resolve__choice* pChoice      = NULL;

	SG_NULLARGCHECK(pCwd);
	SG_NULLARGCHECK(szLeft);
	SG_NULLARGCHECK(szRight);
	SG_NULLARGCHECK(szPath);

	// load the pendingtree and resolve data
	SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pCwd, &pPendingTree)  );
	SG_ERR_CHECK(  SG_RESOLVE__ALLOC(pCtx, pPendingTree, &pResolve)  );

	// find the requested choice
	SG_ERR_CHECK(  _parse_path(pCtx, pPendingTree, szPath, &szGid)  );
	SG_ERR_CHECK(  _find_choice(pCtx, uIndent, pResolve, szGid, eState, NULL, &pChoice)  );
	if (pChoice != NULL)
	{
		SG_resolve__value* pLeft  = NULL;
		SG_resolve__value* pRight = NULL;

		// get the named values
		SG_ERR_CHECK(  SG_resolve__choice__check_value__label(pCtx, pChoice, szLeft, 0u, &pLeft)  );
		SG_ERR_CHECK(  SG_resolve__choice__check_value__label(pCtx, pChoice, szRight, 0u, &pRight)  );
		if (pLeft == NULL)
		{
			SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Unknown value: %s", szLeft)  );
		}
		else if (pRight == NULL)
		{
			SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Unknown value: %s", szRight)  );
		}
		else
		{
			// run the diff
			SG_ERR_CHECK(  _resolve__diff(pCtx, pChoice, pLeft, pRight, szTool, uIndent)  );
		}
	}

fail:
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_RESOLVE_NULLFREE(pCtx, pResolve);
	SG_NULLFREE(pCtx, szGid);
	return;
}

/**
 * Merges two values on a conflict.
 */
static void _action__merge(
	SG_context*        pCtx,       //< [in] [out] Error and context info.
	const SG_pathname* pCwd,       //< [in] The current working directory to allocate a pendingtree from.
	const char*        szBaseline, //< [in] Name of the baseline value to merge.
	                               //<      May be NULL if szOther is also NULL and the conflict has exactly two leaf values.
	                               //<      In that case, the two leaf values are merged.
	const char*        szOther,    //< [in] Name of the other value to diff.
	                               //<      May be NULL if szBaseline is also NULL and the conflict has exactly two leaf values.
	                               //<      In that case, the two leaf values are merged.
	const char*        szPath,     //< [in] Path of the file to merge values on, as specified by the user.
	SG_resolve__state  eState,     //< [in] The state to merge values on.
	                               //<      May be SG_RESOLVE__STATE__COUNT if the specified item only has a single conflict.
	const char*        szTool,     //< [in] Name of a merge tool to use.
	                               //<      May be NULL to use the default.
	const char*        szLabel,    //< [in] Label to assign to the new value.
	                               //<      May be NULL to use an auto-generated one.
	SG_bool            bOverwrite, //< [in] If true, szLabel may specify an existing label.
	                               //<      In that case, the existing label is overwritten with the new value.
	SG_uint32          uIndent     //< [in] Base amount to indent all messages.
	)
{
	SG_pendingtree*     pPendingTree = NULL;
	SG_resolve*         pResolve     = NULL;
	char*               szGid        = NULL;
	SG_resolve__item*   pItem        = NULL;
	SG_resolve__choice* pChoice      = NULL;

	SG_NULLARGCHECK(pCwd);
	SG_ARGCHECK(SG_BOTH_NULL_OR_NOT(szBaseline, szOther), szBaseline+szOther);
	SG_NULLARGCHECK(szPath);

	// load the pendingtree and resolve data
	SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pCwd, &pPendingTree)  );
	SG_ERR_CHECK(  SG_RESOLVE__ALLOC(pCtx, pPendingTree, &pResolve)  );

	// find the requested choice
	SG_ERR_CHECK(  _parse_path(pCtx, pPendingTree, szPath, &szGid)  );
	SG_ERR_CHECK(  _find_choice(pCtx, uIndent, pResolve, szGid, eState, &pItem, &pChoice)  );
	if (pItem != NULL && pChoice != NULL)
	{
		SG_resolve__value* pBaseline = NULL;
		SG_resolve__value* pOther    = NULL;
		SG_resolve__value* pResult   = NULL;

		// if they provided value names, get the actual values
		// if they didn't, we'll just pass NULLs down to _resolve__merge
		if (szBaseline != NULL)
		{
			SG_ERR_CHECK(  SG_resolve__choice__check_value__label(pCtx, pChoice, szBaseline, 0u, &pBaseline)  );
			SG_ERR_CHECK(  SG_resolve__choice__check_value__label(pCtx, pChoice, szOther, 0u, &pOther)  );
			if (pBaseline == NULL)
			{
				SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Unknown value: %s", szBaseline)  );
			}
			else if (pOther == NULL)
			{
				SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Unknown value: %s", szOther)  );
			}
		}

		// run the merge
		if (SG_BOTH_NULL_OR_NOT(pBaseline, pOther))
		{
			SG_ERR_CHECK(  _resolve__merge(pCtx, pItem, pChoice, pBaseline, pOther, szTool, szLabel, (SG_int32)bOverwrite, uIndent, &pResult)  );
			if (pResult != NULL)
			{
				// save the changed resolve/pendingtree data
				SG_ERR_CHECK(  SG_resolve__save(pCtx, pResolve, SG_TRUE)  );
			}
		}
	}

fail:
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_RESOLVE_NULLFREE(pCtx, pResolve);
	SG_NULLFREE(pCtx, szGid);
	return;
}


/*
**
** Private Functionality - Subcommand Handlers
**
*/

/**
 * _subcommand__command_handler for "list" subcommand.
 */
static void _subcommand__handler__list(
	SG_context*            pCtx,
	const SG_option_state* pOptions,
	SG_uint32              uArgs,
	const char**           ppArgs,
	const SG_pathname*     pCwd,
	SG_uint32              uIndent
	)
{
	const char* szFilename = NULL;

	SG_NULLARGCHECK(pOptions);
	SG_NULLARGCHECK(ppArgs);
	SG_NULLARGCHECK(pCwd);

	switch (uArgs)
	{
	default:
		SG_ERR_THROW(SG_ERR_USAGE);
	case 1u:
		szFilename = ppArgs[0];
	case 0u:
		;
	}

	if (pOptions->bInteractive == SG_TRUE)
	{
		SG_ERR_CHECK(  _action__interactive(pCtx, pCwd, szFilename, pOptions->bAll, 0u, NULL, NULL, pOptions->psz_tool, uIndent)  );
	}
	else
	{
		SG_ERR_CHECK(  _action__list(pCtx, pCwd, szFilename, pOptions->bAll, 0u, uIndent)  );
	}

fail:
	return;
}

/**
 * _subcommand__command_handler for "next" subcommand.
 */
static void _subcommand__handler__next(
	SG_context*            pCtx,
	const SG_option_state* pOptions,
	SG_uint32              uArgs,
	const char**           ppArgs,
	const SG_pathname*     pCwd,
	SG_uint32              uIndent
	)
{
	const char* szFilename = NULL;

	SG_NULLARGCHECK(pOptions);
	SG_NULLARGCHECK(ppArgs);
	SG_NULLARGCHECK(pCwd);

	switch (uArgs)
	{
	default:
		SG_ERR_THROW(SG_ERR_USAGE);
	case 1u:
		szFilename = ppArgs[0];
	case 0u:
		;
	}

	if (pOptions->bInteractive == SG_TRUE)
	{
		SG_ERR_CHECK(  _action__interactive(pCtx, pCwd, szFilename, pOptions->bAll, 1u, NULL, NULL, pOptions->psz_tool, uIndent)  );
	}
	else
	{
		SG_ERR_CHECK(  _action__list(pCtx, pCwd, szFilename, pOptions->bAll, 1u, uIndent)  );
	}

fail:
	return;
}

/**
 * _subcommand__command_handler for "FILENAME" subcommand.
 */
static void _subcommand__handler__filename(
	SG_context*            pCtx,
	const SG_option_state* pOptions,
	SG_uint32              uArgs,
	const char**           ppArgs,
	const SG_pathname*     pCwd,
	SG_uint32              uIndent
	)
{
	const char*  szFilename = NULL;
	SG_pathname* pFilename  = NULL;
	SG_bool      bExists    = SG_FALSE;

	SG_NULLARGCHECK(pOptions);
	SG_NULLARGCHECK(ppArgs);
	SG_NULLARGCHECK(pCwd);

	switch (uArgs)
	{
	default:
		SG_ERR_THROW(SG_ERR_USAGE);
	case 1u:
		szFilename = ppArgs[0];
	}

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pFilename, szFilename)  );
	SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pFilename, &bExists, NULL, NULL)  );
	if (bExists == SG_FALSE)
	{
		SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "File doesn't exist: %s", szFilename)  );
	}
	else
	{
		SG_ERR_CHECK(  _action__interactive(pCtx, pCwd, szFilename, pOptions->bAll, 0u, NULL, NULL, pOptions->psz_tool, uIndent)  );
	}

fail:
	return;
}

/**
 * _subcommand__command_handler for "view" subcommand.
 *//*
static void _subcommand__handler__view(
	SG_context*            pCtx,
	const SG_option_state* pOptions,
	SG_uint32              uArgs,
	const char**           ppArgs,
	const SG_pathname*     pCwd,
	SG_uint32              uIndent
	)
{
	const char*         szValue    = NULL;
	const char*         szFilename = NULL;
	const char*         szConflict = NULL;
	SG_resolve__state   eState     = SG_RESOLVE__STATE__COUNT;

	SG_NULLARGCHECK(pOptions);
	SG_NULLARGCHECK(ppArgs);
	SG_NULLARGCHECK(pCwd);

	switch (uArgs)
	{
	default:
		SG_ERR_THROW(SG_ERR_USAGE);
	case 3u:
		szConflict = ppArgs[2];
	case 2u:
		szFilename = ppArgs[1];
		szValue    = ppArgs[0];
	}

	if (szConflict != NULL)
	{
		SG_ERR_CHECK(  SG_resolve__state__check_name(pCtx, szConflict, &eState)  );
		if (eState == SG_RESOLVE__STATE__COUNT)
		{
			SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Unknown conflict type: %s", szConflict)  );
			goto fail;
		}
	}

	SG_ERR_CHECK(  _action__view(pCtx, pCwd, szValue, szFilename, eState, pOptions->psz_tool, uIndent)  );

fail:
	return;
}
*/
/**
 * _subcommand__command_handler for "diff" subcommand.
 */
static void _subcommand__handler__diff(
	SG_context*            pCtx,
	const SG_option_state* pOptions,
	SG_uint32              uArgs,
	const char**           ppArgs,
	const SG_pathname*     pCwd,
	SG_uint32              uIndent
	)
{
	const char*         szLeft     = NULL;
	const char*         szRight    = NULL;
	const char*         szFilename = NULL;
	const char*         szConflict = NULL;
	SG_resolve__state   eState     = SG_RESOLVE__STATE__COUNT;

	SG_NULLARGCHECK(pOptions);
	SG_NULLARGCHECK(ppArgs);
	SG_NULLARGCHECK(pCwd);

	switch (uArgs)
	{
	default:
		SG_ERR_THROW(SG_ERR_USAGE);
	case 4u:
		szConflict = ppArgs[3];
	case 3u:
		szFilename = ppArgs[2];
		szRight    = ppArgs[1];
		szLeft     = ppArgs[0];
	}

	if (szConflict != NULL)
	{
		SG_ERR_CHECK(  SG_resolve__state__check_name(pCtx, szConflict, &eState)  );
		if (eState == SG_RESOLVE__STATE__COUNT)
		{
			SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Unknown conflict type: %s", szConflict)  );
			goto fail;
		}
	}

	SG_ERR_CHECK(  _action__diff(pCtx, pCwd, szLeft, szRight, szFilename, eState, pOptions->psz_tool, uIndent)  );

fail:
	return;
}

/**
 * _subcommand__command_handler for "merge" subcommand.
 */
static void _subcommand__handler__merge(
	SG_context*            pCtx,
	const SG_option_state* pOptions,
	SG_uint32              uArgs,
	const char**           ppArgs,
	const SG_pathname*     pCwd,
	SG_uint32              uIndent
	)
{
	const char*       szLeft     = NULL;
	const char*       szRight    = NULL;
	const char*       szFilename = NULL;
	const char*       szConflict = NULL;
	SG_resolve__state eState     = SG_RESOLVE__STATE__COUNT;

	SG_NULLARGCHECK(pOptions);
	SG_NULLARGCHECK(ppArgs);
	SG_NULLARGCHECK(pCwd);

	if (uArgs == 1u)
	{
		szFilename = ppArgs[0];
	}
	else if (uArgs == 2u)
	{
		szFilename = ppArgs[0];
		szConflict = ppArgs[1];
	}
	else if (uArgs == 3u)
	{
		szLeft     = ppArgs[0];
		szRight    = ppArgs[1];
		szFilename = ppArgs[2];
	}
	else if (uArgs == 4u)
	{
		szLeft     = ppArgs[0];
		szRight    = ppArgs[1];
		szFilename = ppArgs[2];
		szConflict = ppArgs[3];
	}
	else
	{
		SG_ERR_THROW(SG_ERR_USAGE);
	}

	if (szConflict != NULL)
	{
		SG_ERR_CHECK(  SG_resolve__state__check_name(pCtx, szConflict, &eState)  );
		if (eState == SG_RESOLVE__STATE__COUNT)
		{
			SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Unknown conflict type: %s", szConflict)  );
			goto fail;
		}
	}

	SG_ERR_CHECK(  _action__merge(pCtx, pCwd, szLeft, szRight, szFilename, eState, pOptions->psz_tool, pOptions->psz_label, pOptions->bOverwrite, uIndent)  );

fail:
	return;
}

/**
 * _subcommand__command_handler for "accept" subcommand.
 */
static void _subcommand__handler__accept(
	SG_context*            pCtx,
	const SG_option_state* pOptions,
	SG_uint32              uArgs,
	const char**           ppArgs,
	const SG_pathname*     pCwd,
	SG_uint32              uIndent
	)
{
	const char*       szValue    = NULL;
	const char*       szFilename = NULL;
	const char*       szConflict = NULL;
	const char*       szUnknown  = NULL;
	SG_resolve__state eState     = SG_RESOLVE__STATE__COUNT;

	SG_NULLARGCHECK(pOptions);
	SG_NULLARGCHECK(ppArgs);
	SG_NULLARGCHECK(pCwd);

	// parse the arguments
	if (uArgs < 1u || uArgs > 3u)
	{
		SG_ERR_THROW(SG_ERR_USAGE);
	}
	else if (uArgs == 1u)
	{
		szValue = ppArgs[0];
	}
	else if (uArgs == 2u)
	{
		szValue   = ppArgs[0];
		szUnknown = ppArgs[1];
	}
	else if (uArgs == 3u)
	{
		szValue    = ppArgs[0];
		szFilename = ppArgs[1];
		szConflict = ppArgs[2];
	}

	// process the parsed arguments
	if (szUnknown != NULL)
	{
		// szUnknown might be a conflict or a filename
		// try parsing a conflict out of it
		SG_ERR_CHECK(  SG_resolve__state__check_name(pCtx, szUnknown, &eState)  );
		if (eState == SG_RESOLVE__STATE__COUNT)
		{
			// didn't parse correctly, assume it's a filename
			szFilename = szUnknown;
		}
		else
		{
			// parsed fine, it's a conflict
			szConflict = szUnknown;
		}
		szUnknown = NULL;
	}
	else if (szConflict != NULL)
	{
		// parse the conflict into an enumerated value
		SG_ERR_CHECK(  SG_resolve__state__check_name(pCtx, szConflict, &eState)  );
		if (eState == SG_RESOLVE__STATE__COUNT)
		{
			SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Unknown conflict type: %s", szConflict)  );
			goto fail;
		}
	}

	// check for problems
	if (szConflict != NULL && eState == SG_RESOLVE__STATE__COUNT)
	{
		// if a conflict was specified, it would be correctly parsed by now
		SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "Unknown conflict type: %s", szConflict)  );
	}
	else
	{
		// call the appropriate command
		if (pOptions->bAll == SG_FALSE)
		{
			if (szFilename == NULL)
			{
				SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "To resolve one conflict specify the file (and conflict if it has several).")  );
				SG_ERR_CHECK(  _console__stdout(pCtx, uIndent, "To resolve multiple conflicts use --all.")  );
			}
			else
			{
				SG_ERR_CHECK(  _action__accept_one(pCtx, pCwd, szValue, szFilename, eState, pOptions->bOverwrite, uIndent)  );
			}
		}
		else
		{
			SG_ERR_CHECK(  _action__accept_all(pCtx, pCwd, szValue, szFilename, eState, pOptions->bOverwrite, uIndent)  );
		}
	}

fail:
	return;
}


/*
**
** Public Functionality
**
*/

/**
 * Entry point for resolve command implementation.
 */
void do_cmd_resolve2(
	SG_context*      pCtx,     //< [in] [out] Error and context info.
	SG_option_state* pOptions, //< [in] Parsed option state.
	SG_uint32        uArgs,    //< [in] Number of unparsed arguments.
	const char**     ppArgs    //< [in] Unparsed arguments.
	)
{
	SG_uint32    uIndent = 0u;
	SG_pathname* pCwd    = NULL;

	SG_NULLARGCHECK(pOptions);
	SG_ARGCHECK(ppArgs != NULL || uArgs == 0u, ppArgs);

	// get the current working directory
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC(pCtx, &pCwd)  );
	SG_ERR_CHECK(  SG_pathname__set__from_cwd(pCtx, pCwd)  );

	// do a complete pendingtree scan
	SG_ERR_CHECK(  _pendingtree__scan(pCtx, pCwd)  );

	// if we don't have any arguments, do the default action
	if (uArgs == 0u)
	{
		// default action is an unfiltered interactive prompt
		SG_ERR_CHECK(  _action__interactive(pCtx, pCwd, NULL, SG_FALSE, 0u, NULL, NULL, pOptions->psz_tool, uIndent)  );
	}
	else
	{
		SG_uint32                        uIndex = 0u;
		const _subcommand__command_spec* pSpec  = NULL;

		// run through the defined subcommands and look for a match
		for (uIndex = 0u; uIndex < SG_NrElements(gaSubcommands); ++uIndex)
		{
			if (SG_stricmp(ppArgs[0], gaSubcommands[uIndex].szName) == 0)
			{
				pSpec = gaSubcommands + uIndex;
				break;
			}
		}

		// if we didn't find a match, use the default subcommand
		if (pSpec == NULL)
		{
			pSpec = &gcDefaultSubcommand;
		}

		// run the command's handler
		SG_ERR_CHECK(  pSpec->fHandler(pCtx, pOptions, uArgs - 1u, ppArgs + 1, pCwd, uIndent)  );
	}

fail:
	SG_PATHNAME_NULLFREE(pCtx, pCwd);
	return;
}
