/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include <sg.h>
#include <signal.h>
#include "sg_typedefs.h"
#include "sg_cmd_util_prototypes.h"
#include "../libraries/mongoose/sg_mongoose.h"

#define SG_CORE_TEMPLATE_PATH "core/templates"
#define SG_STATIC_UI_PATH "ui"

static void cmd_serve_mg_uri_callback(struct mg_connection *conn, const struct mg_request_info *request_info, void *user_data)
{
	SG_context * pCtx = NULL; // This comes from the mg_connection, it's not ours to free.
	SG_uridispatchcontext * pDispatchContext = NULL;

	const char * szResponseStatusCode = "500 Internal Server Error";
	SG_uint64 responseContentLength = 0;
	SG_vhash * pResponseHeaders = NULL;

	SG_UNUSED(user_data);

	pCtx = mg_get_sg_context(conn);

	// Receive the request's headers.
	{
		if(pCtx==NULL)
		{
			pDispatchContext = SG_uridispatch__begin_request(pCtx,
				request_info->request_method,
				request_info->uri,
				request_info->query_string,
				(request_info->is_ssl?"https:":"http:"),
				NULL);
		}
		else
		{
			// Convert headers from a mongoose data structure to an SG_vhash.
			SG_vhash * pRequestHeaders = NULL;
			int i;
			SG_VHASH__ALLOC__PARAMS(pCtx, &pRequestHeaders, request_info->num_headers, NULL, NULL);
			for(i = 0; !SG_context__has_err(pCtx) && i<request_info->num_headers; ++i)
				SG_vhash__add__string__sz(pCtx, pRequestHeaders, request_info->http_headers[i].name, request_info->http_headers[i].value);

			pDispatchContext = SG_uridispatch__begin_request(pCtx,
				request_info->request_method,
				request_info->uri,
				request_info->query_string,
				(request_info->is_ssl?"https:":"http:"),
				&pRequestHeaders);

			SG_VHASH_NULLFREE(pCtx, pRequestHeaders); // In case SG_uridispatch__begin_request() failed to take ownership for whatever reason.
		}
	}

	// Receive the request's message body (if applicable).
	{
		while(!SG_uridispatch__get_response_headers(pDispatchContext, &szResponseStatusCode, &responseContentLength, &pResponseHeaders))
		{
			SG_byte request_chunk[4*1024];
			int chunk_len = 0;
			chunk_len = mg_get_chunk(conn, (char*)request_chunk, sizeof(request_chunk));
			if(chunk_len>0)
				SG_uridispatch__chunk_request_body(pDispatchContext, request_chunk, (SG_uint32)chunk_len);
			else
			{
				SG_uridispatch__abort(&pDispatchContext); // See comments on mg_get_chunk in sg_mongoose.h.
				if(pCtx!=NULL)
					SG_ERR_IGNORE(  SG_log__report_warning(pCtx, "Failed to fetch data from the client. Perhaps they closed the connection.")  );
				return;
			}
		}
	}

	// Send the response's headers.
	{
		SG_int_to_string_buffer tmp;

		mg_printf(conn, "HTTP/1.1 %s\r\n", szResponseStatusCode);
		mg_printf(conn, "Content-Length: %s\r\n", SG_uint64_to_sz(responseContentLength, tmp));

		if (pResponseHeaders!=NULL)
		{
			SG_uint32 numHeaders = 0;
			SG_uint32 j;

			SG_vhash__count(pCtx, pResponseHeaders, &numHeaders);
			SG_ASSERT(!SG_context__has_err(pCtx)); // The only errors SG_vhash__count throws are INVALIDARG errors on NULL arguments.

			for(j = 0; j<numHeaders; ++j)
			{
				const char * szName = NULL;
				const SG_variant * pValue = NULL;

				SG_vhash__get_nth_pair(pCtx, pResponseHeaders, j, &szName, &pValue);
				SG_ASSERT(!SG_context__has_err(pCtx)); // pResponseHeaders is non-NULL and j is in the valid range so we should be fine

				// Some headers are allowed to be set multiple times (namely, Set-Cookie)
				if (pValue->type==SG_VARIANT_TYPE_VARRAY)
				{
					SG_varray * pValues = pValue->v.val_varray;
					SG_uint32 numValues = 0;
					SG_uint32 k;

					SG_varray__count(pCtx, pValues, &numValues);
					SG_ASSERT(!SG_context__has_err(pCtx)); // The only errors SG_vhash__count throws are INVALIDARG errors on NULL arguments.

					for (k = 0; k < numValues; k++)
					{
						const SG_variant * pNestedValue = NULL;
						SG_varray__get__variant(pCtx, pValues, k, &pNestedValue);
						SG_ASSERT(!SG_context__has_err(pCtx)); // k is in the valid range so we should be fine

						// Only strings are supported in the header hash for varrays
						if (pNestedValue->type==SG_VARIANT_TYPE_SZ)
							mg_printf(conn, "%s: %s\r\n", szName, pNestedValue->v.val_sz);
						else
							SG_ERR_IGNORE(  SG_log__report_warning(pCtx, "Skipped a value for HTTP header \"%s\". Unsupported variant type.", szName)  );
					}
				}
				else if (pValue->type == SG_VARIANT_TYPE_SZ)
					mg_printf(conn, "%s: %s\r\n", szName, pValue->v.val_sz);
				else if (pValue->type == SG_VARIANT_TYPE_INT64)
					mg_printf(conn, "%s: %s\r\n", szName, SG_uint64_to_sz(pValue->v.val_int64, tmp));
				else
					SG_ERR_IGNORE(  SG_log__report_warning(pCtx, "Skipped HTTP header \"%s\". Unsupported variant type.", szName)  );
			}
			SG_VHASH_NULLFREE(pCtx, pResponseHeaders);

		}
		else
		{
			if(responseContentLength>0)
				mg_printf(conn, "Content-Type: text/plain;charset=UTF-8\r\n");
		}

		mg_printf(conn, "\r\n");
	}

	// Send the response's message body (if applicable).
	{
		const SG_byte * pResponseBuffer = NULL;
		SG_uint32 responseBufferLength = 0;

		while(pDispatchContext!=NULL)
		{
			SG_uridispatch__chunk_response_body(
				&pDispatchContext,
				&pResponseBuffer,
				&responseBufferLength);
			if(responseBufferLength>0)
				mg_write(conn, pResponseBuffer, responseBufferLength);
		}
	}
}


static SG_bool shut_down_mg_flag = SG_FALSE;

void shut_down_mg(int SG_UNUSED_PARAM(sig))
{
	SG_UNUSED(sig);
	shut_down_mg_flag = SG_TRUE;
}

extern SG_bool _sg_uridispatch__debug_remote_shutdown;

static void _requireDirectory(
	SG_context *pCtx,
	SG_pathname *path)
{
	SG_bool ssiExists = SG_FALSE;
	SG_fsobj_type fot;
	SG_fsobj_perms perms;

	SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, path, &ssiExists, &fot, &perms)  );

	if (! ssiExists)
	{
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "The required subfolder %s (inferred from localsetting %s) does not exist.\n\n",
			SG_pathname__sz(path), SG_LOCALSETTING__SERVER_FILES)  );
		SG_ERR_THROW(SG_ERR_NOT_A_DIRECTORY);
	}
	else if (fot != SG_FSOBJ_TYPE__DIRECTORY)
	{
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "%s (inferred from localsetting %s) is not a directory.\n\n",
			SG_pathname__sz(path), SG_LOCALSETTING__SERVER_FILES)  );
		SG_ERR_THROW(SG_ERR_NOT_A_DIRECTORY);
	}

fail:
	;
}

void do_cmd_serve(SG_context * pCtx, SG_option_state* pOptSt, int * pExitStatus)
{
	struct mg_context * ctx = NULL;
	char* psz_collateral_root = NULL;
	char * psz_enable_diagnostics = NULL;
	SG_pathname *collateralRoot = NULL;
	SG_pathname * staticPath = NULL;
	SG_pathname * templatesPath = NULL;
	SG_string * aliases = NULL;
	SG_string * ports = NULL;
	SG_repo* pRepo = NULL;
	SG_fsobj_type fot;
	SG_fsobj_perms perms;
	SG_bool ssiExists = SG_FALSE;
	const char * sz_ssl_cert_pem = NULL;
	SG_pathname * p_ssl_cert_pem = NULL;
	SG_bool exists_ssl_cert_pem = SG_FALSE;
	SG_bool bEnableDiagnostics = SG_FALSE;
	const char * sz_port_prefix = "";
	int ok = 0;

	SG_ASSERT(pExitStatus!=NULL);

	// if we're in a working copy, then we want to grab
	// options from this repo scope

	SG_cmd_util__get_repo_from_cwd(pCtx, &pRepo, NULL);
	SG_ERR_CHECK_CURRENT_DISREGARD(SG_ERR_NOT_A_WORKING_COPY);

	SG_ERR_CHECK(  SG_localsettings__get__sz(pCtx, SG_LOCALSETTING__SERVER_FILES, pRepo, &psz_collateral_root, NULL)  );
	SG_ERR_CHECK(  SG_localsettings__get__sz(pCtx, SG_LOCALSETTING__SERVER_ENABLE_DIAGNOSTICS, pRepo, &psz_enable_diagnostics, NULL)  );
	bEnableDiagnostics = (psz_enable_diagnostics!=NULL && SG_stricmp(psz_enable_diagnostics, "true")==0);
	SG_NULLFREE(pCtx, psz_enable_diagnostics);
	SG_REPO_NULLFREE(pCtx, pRepo);

	if ((! psz_collateral_root) || (strlen(psz_collateral_root) == 0))
	{
		SG_NULLFREE(pCtx, psz_collateral_root);
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "The localsetting %s is missing or empty.\n\n%s must be set to run '%s serve'.\n\n", 
			SG_LOCALSETTING__SERVER_FILES, SG_LOCALSETTING__SERVER_FILES, WHATSMYNAME
			)  );
		goto fail;
	}

	SG_ERR_CHECK(  SG_pathname__alloc__sz(pCtx, &collateralRoot, psz_collateral_root)  );
	SG_NULLFREE(pCtx, psz_collateral_root);

	SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, collateralRoot, &ssiExists, &fot, &perms)  );

	if (! ssiExists)
	{
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "The path %s specified in localsetting %s does not exist.\n\n",
			SG_pathname__sz(collateralRoot), SG_LOCALSETTING__SERVER_FILES)  );
		goto fail;
	}
	else if (fot != SG_FSOBJ_TYPE__DIRECTORY)
	{
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "The path %s specified in localsetting %s is not a directory.\n\n",
			SG_pathname__sz(collateralRoot), SG_LOCALSETTING__SERVER_FILES)  );
		goto fail;
	}

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx, &staticPath, collateralRoot, SG_STATIC_UI_PATH)  );
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx, &templatesPath, collateralRoot, SG_CORE_TEMPLATE_PATH)  );

	SG_ERR_CHECK(  _requireDirectory(pCtx, staticPath)  );
	SG_ERR_CHECK(  _requireDirectory(pCtx, templatesPath)  );

	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &aliases)  );
	SG_ERR_CHECK(  SG_string__sprintf(pCtx, aliases, "/%s=%s", SG_STATIC_UI_PATH, SG_pathname__sz(staticPath))  );

	SG_ERR_CHECK(  SG_string__append__sz(pCtx, aliases, ",/favicon.ico=")  );
	SG_ERR_CHECK(  SG_string__append__sz(pCtx, aliases, SG_pathname__sz(staticPath))  );
	SG_ERR_CHECK(  SG_string__append__sz(pCtx, aliases, "/favicon.ico")  );

	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &ports)  );

	if(pOptSt->psz_cert!=NULL && pOptSt->psz_cert[0]!='\0')
		sz_ssl_cert_pem = pOptSt->psz_cert;
	else
		sz_ssl_cert_pem = "ssl_cert.pem";
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &p_ssl_cert_pem, sz_ssl_cert_pem)  );
	SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, p_ssl_cert_pem, &exists_ssl_cert_pem, NULL, NULL)  );
	SG_PATHNAME_NULLFREE(pCtx, p_ssl_cert_pem);

	if((pOptSt->bNoPlain || pOptSt->iSPort!=0) && !exists_ssl_cert_pem)
		SG_ERR_THROW2(SG_ERR_NOTAFILE, (pCtx, "SSL certificate file '%s' does not exist", sz_ssl_cert_pem));

	if(exists_ssl_cert_pem && pOptSt->iSPort==0)
		pOptSt->iSPort = 8081;

	if(!pOptSt->bPublic)
	{
		sz_port_prefix = "127.0.0.1:";
		SG_ERR_CHECK(  SG_console(pCtx, SG_CS_STDOUT, "Binding to localhost (127.0.0.1). Use '--public' option to allow remote connections.\n")  );
	}	

	if (pOptSt->bNoPlain)
	{
		SG_ERR_CHECK(  SG_string__sprintf(pCtx, ports, "%s%lis", sz_port_prefix, (long)pOptSt->iSPort)  );
		SG_ERR_CHECK(  SG_console(pCtx, SG_CS_STDOUT, "Serving encrypted on port %d.\n", pOptSt->iSPort)  );
	}
	else if(pOptSt->iSPort==0)
	{
		SG_ERR_CHECK(  SG_string__sprintf(pCtx, ports, "%s%li", sz_port_prefix, (long)pOptSt->iPort)  );
		SG_ERR_CHECK(  SG_console(pCtx, SG_CS_STDOUT, "Serving unencrypted on port %d.\n", pOptSt->iPort)  );
	}
	else
	{
		SG_ERR_CHECK(  SG_string__sprintf(pCtx, ports, "%s%li,%s%lis", sz_port_prefix, (long)pOptSt->iPort, sz_port_prefix, (long)pOptSt->iSPort)  );
		SG_ERR_CHECK(  SG_console(pCtx, SG_CS_STDOUT, "Serving on ports %d (unencrypted) and %d (encrypted).\n", pOptSt->iPort, pOptSt->iSPort)  );
	}

	SG_uridispatch__init(NULL);
	ctx = mg_start();
	if(exists_ssl_cert_pem)
	{
		ok =  mg_set_option(ctx, "ssl_cert", sz_ssl_cert_pem);
		if(!ok)
			SG_ERR_THROW2(SG_ERR_UNSPECIFIED,  (pCtx, "An error occured when configuring the server to use the SSL certificate '%s'.", sz_ssl_cert_pem));
	}

	ok = mg_set_option(ctx, "aliases", SG_string__sz(aliases));
	if(ok!=1)
		SG_ERR_THROW2(SG_ERR_UNSPECIFIED,  (pCtx, "An error occured when configuring aliases on the server ('%s').",SG_string__sz(aliases)));

	ok = mg_set_option(ctx, "ports", SG_string__sz(ports));
	if(ok!=1)
		SG_ERR_THROW2(SG_ERR_UNSPECIFIED,  (pCtx, "An error occured when binding server to port(s) '%s'.",SG_string__sz(ports)));

	ok = mg_set_option(ctx, "mime_types", ".css=text/css;charset=UTF-8,.js=application/javascript;charset=UTF-8");
	if(ok!=1)
		SG_ERR_THROW2(SG_ERR_UNSPECIFIED,  (pCtx, "An error occured when configuring mime types on the server."));

	SG_STRING_NULLFREE(pCtx, aliases);
	SG_STRING_NULLFREE(pCtx, ports);

	mg_set_uri_callback(ctx, "*", &cmd_serve_mg_uri_callback, NULL);

	(void)signal(SIGINT,shut_down_mg);
	while(!shut_down_mg_flag && !(_sg_uridispatch__debug_remote_shutdown && bEnableDiagnostics))
		SG_sleep_ms(10);
	(void)signal(SIGINT,SIG_DFL);
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDOUT, "Shutting down server...\n")  );

fail:
	if(ctx!=NULL)
	{
		if(SG_context__has_err(pCtx))
		{
			// On Windows mg_stop() seems to close the stderr stream. Print the error while we still can.
			SG_string * pStrContextDump = NULL;
			if(SG_ERR_OK==SG_context__err_to_string(pCtx, SG_TRUE, &pStrContextDump) && pStrContextDump!=NULL)
			{
				SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "%s", SG_string__sz(pStrContextDump))  );
				SG_STRING_NULLFREE(pCtx, pStrContextDump);
			}
			else
			{
				SG_error err;
				char bufErrorMessage[SG_ERROR_BUFFER_SIZE+1];
				SG_context__get_err(pCtx,&err);
				SG_error__get_message(err, SG_TRUE, bufErrorMessage, SG_ERROR_BUFFER_SIZE);
				SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "%s", bufErrorMessage)  );
			}
			SG_context__err_reset(pCtx);
			*pExitStatus = 1;
		}
		
		mg_stop(ctx);
	}
	SG_NULLFREE(pCtx, psz_collateral_root);
	SG_STRING_NULLFREE(pCtx, aliases);
	SG_STRING_NULLFREE(pCtx, ports);
	SG_PATHNAME_NULLFREE(pCtx, collateralRoot);
	SG_PATHNAME_NULLFREE(pCtx, staticPath);
	SG_PATHNAME_NULLFREE(pCtx, templatesPath);
	SG_REPO_NULLFREE(pCtx, pRepo);
	SG_PATHNAME_NULLFREE(pCtx, p_ssl_cert_pem);
}
