/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg__do_cmd_update.c
 *
 * @details Handle the UPDATE command.
 *
 */

//////////////////////////////////////////////////////////////////

#include <sg.h>
#include "sg_typedefs.h"
#include "sg_prototypes.h"
#include "sg_cmd_util_prototypes.h"

//////////////////////////////////////////////////////////////////

/**
 * Handle the UPDATE command.
 *
 *
 */
void do_cmd_update(SG_context * pCtx,
				   SG_rev_spec* pRevSpec,
				   const char * psz_attach,
				   SG_bool bVerbose,
				   SG_bool bTest,
				   const SG_vector * pvec_use_submodules,
				   int * pExitStatus)
{
	SG_vhash * pvhResults = NULL;
	SG_int64 status;
	char* pszTargetChangeset = NULL;
	char* psz_target_branch_name = NULL;
	char* psz_current_attachment_name = NULL;
	SG_repo * pRepo = NULL;
	SG_vhash * pvh_pile = NULL;

	if(psz_attach!=NULL && SG_rev_spec__is_empty(pCtx, pRevSpec))
		SG_ERR_THROW2(SG_ERR_USAGE, (pCtx, "When using the --attach option, you must also specify a revision."));

	// use the current directory to find the pendingtree and then
	// determine the target changeset using --rev/--tag/--branch or unique
	// branch head (when tied) or unique descendant head from current baseline.
	// This may return (hid,name), (hid,null), (null,null)

	SG_vv_utils__rev_spec_to_csid__wd(pCtx, NULL,
													pRevSpec,
													&pszTargetChangeset, &psz_target_branch_name);

	if(SG_context__err_equals(pCtx, SG_ERR_BRANCH_NEEDS_MERGE))
	{
		SG_uint32 branchesSpecified;

		SG_context__err_reset(pCtx);

		SG_ERR_CHECK(  SG_rev_spec__count_branches(pCtx, pRevSpec, &branchesSpecified)  );

		if(!branchesSpecified)
		{
			SG_ERR_CHECK(  SG_cmd_util__get_branch_name_from_cwd(pCtx, &psz_current_attachment_name)  );
			SG_ERR_CHECK(  SG_console(pCtx, SG_CS_STDERR, "The current branch (%s) has multiple heads.\n", psz_current_attachment_name)  );
			SG_ERR_CHECK(  SG_console(pCtx, SG_CS_STDERR, "Update using the \"--rev\" or \"--tag\" option to pick a specific revision.\n")  );

			SG_ERR_CHECK(  SG_rev_spec__add_branch(pCtx, pRevSpec, psz_current_attachment_name)  );
			SG_NULLFREE(pCtx, psz_current_attachment_name);
		}
		else
		{
			SG_ERR_CHECK(  SG_console(pCtx, SG_CS_STDERR, "The target branch has multiple heads.\n")  );
			SG_ERR_CHECK(  SG_console(pCtx, SG_CS_STDERR, "Update using the \"--attach\" option rather than \"--branch\" to attach your working folder to it.\n")  );
			SG_ERR_CHECK(  SG_console(pCtx, SG_CS_STDERR, "You must also use the \"--rev\" or \"--tag\" option to pick a specific revision.\n")  );
		}
		SG_ERR_CHECK(  SG_console(pCtx, SG_CS_STDERR, "\nCurrent heads are:\n")  );

		SG_ERR_CHECK(  SG_cmd_util__get_repo_from_cwd(pCtx, &pRepo, NULL)  );
		SG_ERR_CHECK(  SG_vc_branches__cleanup(pCtx, pRepo, &pvh_pile)  );
		SG_ERR_CHECK(  SG_cmd_util__dump_log__revspec(pCtx, SG_CS_STDERR, pRepo, pRevSpec, pvh_pile, SG_FALSE, SG_FALSE)  );

		SG_VHASH_NULLFREE(pCtx, pvh_pile);
		SG_REPO_NULLFREE(pCtx, pRepo);

		SG_ASSERT(pExitStatus!=NULL);
		*pExitStatus=1;
		goto fail;
	}
	else
		SG_ERR_CHECK_CURRENT;

	// use the current directory and the computed target
	// values to update the WD on disk.

	SG_ERR_CHECK(  SG_vv_verbs__update2(pCtx, NULL,
										pszTargetChangeset, 
                                        psz_attach ? psz_attach : psz_target_branch_name,
										bVerbose, bTest,
										pvec_use_submodules,
										&pvhResults)  );
	SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvhResults, "status", &status)  );
	switch (status)
	{
	default:
	case SG_VV_VERBS__STATUS__OK:
		if (bVerbose || bTest)
		{
			SG_varray * pvaActionLog;	// we don't own this

			SG_ERR_CHECK(  SG_vhash__get__varray(pCtx, pvhResults, "log", &pvaActionLog)  );
			SG_ERR_IGNORE(  sg_report_action_log2(pCtx, pvaActionLog)  );
		}
		break;

	case SG_VV_VERBS__STATUS__PORTABILITY_WARNING:
		{
			SG_varray * pvaWarnings;	// we don't own this
			
			SG_ERR_CHECK(  SG_vhash__get__varray(pCtx, pvhResults, "port", &pvaWarnings)  );
			SG_ERR_IGNORE(  sg_report_portability_warnings2(pCtx, pvaWarnings)  );
			SG_ERR_THROW(  SG_ERR_PORTABILITY_WARNINGS  ); // throw so that exe exits with non-zero status
		}
		break;

	case SG_VV_VERBS__STATUS__SUBMODULE_PROBLEM:
		{
			SG_varray * pvaMissingDetails;	// we don't own this

			SG_ERR_CHECK(  SG_vhash__get__varray(pCtx, pvhResults, "sub", &pvaMissingDetails)  );

			SG_ERR_IGNORE(  sg_report_submodule_missing_details(pCtx, pvaMissingDetails)  );
			SG_ERR_THROW(  SG_ERR_SUBMODULE_NESTED_UPDATE_CHECKOUT_PROBLEM  ); // throw so that exe exits with non-zero status
		}
		break;
	}

fail:
	SG_NULLFREE(pCtx, pszTargetChangeset);
	SG_NULLFREE(pCtx, psz_target_branch_name);
	SG_NULLFREE(pCtx, psz_current_attachment_name);
	SG_VHASH_NULLFREE(pCtx, pvhResults);
	SG_VHASH_NULLFREE(pCtx, pvh_pile);
	SG_REPO_NULLFREE(pCtx, pRepo);
}
