/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include <sg.h>
#include "sg_typedefs.h"
#include "sg_prototypes.h"
#include "sg_cmd_util_prototypes.h"

#if defined(WINDOWS)
#define DEFAULT_EDITOR "notepad"
#else
#define DEFAULT_EDITOR "vi"
#endif

#define EDITOR_CONFIG_PATH "/machine/editor"

#define COMMENT_FILE__LINE_ENDING SG_PLATFORM_NATIVE_EOL_STR
#define COMMENT_FILE__LINE_PREFIX "#"

#define COMMENT_FILE__BOILERPLATE	COMMENT_FILE__LINE_ENDING COMMENT_FILE__LINE_ENDING \
									COMMENT_FILE__LINE_PREFIX " Enter commment.  Lines beginning with '" COMMENT_FILE__LINE_PREFIX "' are removed." COMMENT_FILE__LINE_ENDING \
									COMMENT_FILE__LINE_PREFIX " Leave message empty to cancel." COMMENT_FILE__LINE_ENDING \
									COMMENT_FILE__LINE_PREFIX " --" COMMENT_FILE__LINE_ENDING

static void _get_editor(
	SG_context* pCtx,
	char** ppszEditor);

static void _get_comment_from_file(SG_context* pCtx, const SG_pathname* pFilePathname, char** ppszComment)
{
	SG_string* pstrCommentFile = NULL;
	SG_string* pstrReturn = NULL;

	SG_file__read_into_string(pCtx, pFilePathname, &pstrCommentFile);

	if (SG_context__has_err(pCtx))
	{
		if (SG_context__err_equals(pCtx, SG_ERR_EOF)) // empty file
		{
			SG_context__err_reset(pCtx);
			SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pstrCommentFile)  );
		}
		else
			SG_ERR_CHECK_CURRENT;
	}

	SG_ERR_CHECK(  SG_STRING__ALLOC__RESERVE(pCtx, &pstrReturn, SG_string__length_in_bytes(pstrCommentFile))  );
	{
		const SG_uint32 lenPrefix = (const SG_uint32)strlen(COMMENT_FILE__LINE_PREFIX);
		const SG_byte* pos = (SG_byte*)SG_string__sz(pstrCommentFile);
		const SG_uint32 len = SG_string__length_in_bytes(pstrCommentFile);
		const SG_byte* lastByte = pos + len - 1; /* There's no contract that SG_string's buffer is null-terminated. Use this instead. */

		/* Copy lines that don't start with COMMENT_FILE__LINE_PREFIX */
		while (pos <= lastByte)
		{
			SG_bool bSkipLine = SG_FALSE;
			const SG_byte* szLineStart;

			if ( pos + lenPrefix <= lastByte && !strncmp((const char*)pos, COMMENT_FILE__LINE_PREFIX, lenPrefix) )
				bSkipLine = SG_TRUE;

			/* Find the end of the line or the end of the buffer */
			szLineStart = pos;
			while ( (pos <= lastByte) && (*pos != SG_CR) && (*pos != SG_LF) )
				pos++;

			if (pos < lastByte)
			{
				/* Include the line ending character(s), handle different types. */
				if (*pos == SG_CR && *(pos+1) == SG_LF)
					pos += 2;
				else
					pos++;
			}

			if (!bSkipLine && pos != szLineStart)
				SG_ERR_CHECK(  SG_string__append__buf_len(pCtx, pstrReturn, szLineStart, (SG_uint32)(pos-szLineStart))  );

			if (pos == lastByte)
				break;
		}
	}

	/* Trim trailing line breaks */
	{
		SG_byte myByte;
		while (SG_string__length_in_bytes(pstrReturn))
		{
			SG_ERR_CHECK(  SG_string__get_byte_r(pCtx, pstrReturn, 0, &myByte)  );
			if (myByte == SG_CR || myByte == SG_LF)
				SG_ERR_CHECK(  SG_string__remove(pCtx, pstrReturn, SG_string__length_in_bytes(pstrReturn) - 1, 1)  );
			else
				break;
		}
	}
	
	if (SG_string__length_in_bytes(pstrReturn))
		SG_ERR_CHECK(  SG_string__sizzle(pCtx, &pstrReturn, (SG_byte**)ppszComment, NULL)  );

	/* Fall through to common cleanup */
fail:
	SG_STRING_NULLFREE(pCtx, pstrCommentFile);
	SG_STRING_NULLFREE(pCtx, pstrReturn);
}

void SG_cmd_util__build_comment_boilerplate(
	SG_context* pCtx, 
	SG_bool bForCommit,
	SG_uint32 count_args, const char ** paszArgs, 
	SG_option_state* pOptSt, 
	SG_string** ppstrBoilerplate)
{
	SG_string* pstrBoilerplate = NULL;
	SG_repo* pRepo = NULL;
	SG_pathname* pPathCwd = NULL;
	char* pszUserFromCwd = NULL;
	const char* pszUser_ref = NULL;
	char* pszBranchFromCwd = NULL;
	
	SG_pendingtree* pPendingTree = NULL;
	SG_treediff2 * pTreeDiff = NULL;
	SG_vhash* pvhStatus = NULL;
	SG_string* pstrStatus = NULL;

	SG_ERR_CHECK(  SG_string__alloc(pCtx, &pstrBoilerplate)  );

	/* Static stuff explaining what to do */
	SG_ERR_CHECK(  SG_string__append__sz(pCtx, pstrBoilerplate, COMMENT_FILE__BOILERPLATE)  );

	/* vv: user: ian@sourcegear.com */
	if (pOptSt->psz_username && *(pOptSt->psz_username))
	{
		pszUser_ref = pOptSt->psz_username;
		SG_ERR_CHECK(  SG_PATHNAME__ALLOC(pCtx, &pPathCwd)  );
		SG_ERR_CHECK(  SG_pathname__set__from_cwd(pCtx, pPathCwd)  );
	}
	else
	{
		SG_ERR_CHECK(  SG_cmd_util__get_repo_from_cwd(pCtx, &pRepo, &pPathCwd)  );
		SG_ERR_CHECK(  SG_user__get_username_for_repo(pCtx, pRepo, &pszUserFromCwd)  );
		pszUser_ref = pszUserFromCwd;
	}
	SG_ERR_CHECK(  SG_string__append__format(pCtx, pstrBoilerplate, COMMENT_FILE__LINE_PREFIX "   User: %s%s", pszUser_ref, COMMENT_FILE__LINE_ENDING)  );

	if (bForCommit)
	{
		/* vv: branch 'master' */
		SG_ERR_CHECK(  SG_cmd_util__get_branch_name_from_cwd(pCtx, &pszBranchFromCwd)  );
		if (pszBranchFromCwd)
			SG_ERR_CHECK(  SG_string__append__format(pCtx, pstrBoilerplate, COMMENT_FILE__LINE_PREFIX " Branch: %s%s", pszBranchFromCwd, COMMENT_FILE__LINE_ENDING)  );
		else
			SG_ERR_CHECK(  SG_string__append__format(pCtx, pstrBoilerplate, COMMENT_FILE__LINE_PREFIX " Branch: *None*%s", COMMENT_FILE__LINE_ENDING)  );

		SG_ERR_CHECK(  SG_string__append__sz(pCtx, pstrBoilerplate, COMMENT_FILE__LINE_PREFIX "[ASSOCIATIONS]" COMMENT_FILE__LINE_ENDING)  );

		/* Status */
		{
			const SG_varray * pvaMergeIssues_ref = NULL;
			SG_ERR_CHECK(  SG_pendingtree__alloc(pCtx, pPathCwd, &pPendingTree)  );
			SG_ERR_CHECK(  SG_vv_verbs__status(pCtx, pPendingTree, NULL, NULL, count_args, paszArgs, 
				pOptSt->pFilespec, pOptSt->bRecursive, pOptSt->bRecursive, &pvaMergeIssues_ref, &pvhStatus)  );
			SG_ERR_CHECK(  SG_status_formatter__vhash_to_string(pCtx, pvhStatus, pvaMergeIssues_ref, pPendingTree,
				COMMENT_FILE__LINE_PREFIX " ", COMMENT_FILE__LINE_ENDING, &pstrStatus)  );
		}

		SG_ERR_CHECK(  SG_string__append__format(pCtx, pstrBoilerplate, COMMENT_FILE__LINE_PREFIX "%s", COMMENT_FILE__LINE_ENDING)  );
		SG_ERR_CHECK(  SG_string__append__sz(pCtx, pstrBoilerplate, SG_string__sz(pstrStatus))  );
	}

	SG_RETURN_AND_NULL(pstrBoilerplate, ppstrBoilerplate);

	/* Fall through to common cleanup */
fail:
	SG_STRING_NULLFREE(pCtx, pstrBoilerplate);
	SG_REPO_NULLFREE(pCtx, pRepo);
	SG_PATHNAME_NULLFREE(pCtx, pPathCwd);
	SG_NULLFREE(pCtx, pszUserFromCwd);
	SG_NULLFREE(pCtx, pszBranchFromCwd);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_TREEDIFF2_NULLFREE(pCtx, pTreeDiff);
	SG_VHASH_NULLFREE(pCtx, pvhStatus);
	SG_STRING_NULLFREE(pCtx, pstrStatus);
}

static void _sg_cmd_util__insert_associations_into_boilerplate(
	SG_context *pCtx,
	SG_string *boilerplate,
	const SG_varray *descs)
{
	SG_string *newString = NULL;
	const char *suffix = NULL;
	const char *sz = SG_string__sz(boilerplate);
	const char *placeholder = COMMENT_FILE__LINE_PREFIX "[ASSOCIATIONS]" COMMENT_FILE__LINE_ENDING;
	const char *phpos = strstr(sz, placeholder);

	SG_NULLARGCHECK(boilerplate);

	if (! phpos)
		return;

	SG_ERR_CHECK(  SG_STRING__ALLOC__COPY(pCtx, &newString, boilerplate)  );
	SG_ERR_CHECK(  SG_string__remove(pCtx, newString, (SG_uint32)(phpos - sz), (SG_uint32)(strlen(sz) - (phpos - sz)))  );
	suffix = phpos + strlen(placeholder);

	if (descs)
	{
		SG_uint32 i, count;

		SG_ERR_CHECK(  SG_varray__count(pCtx, descs, &count)  );

		if (count)
			SG_ERR_CHECK(  SG_string__append__sz(pCtx, newString, COMMENT_FILE__LINE_PREFIX COMMENT_FILE__LINE_ENDING)  );

		for ( i = 0; i < count; ++i )
		{
			const char *asz, *t;
			SG_uint32 len;
			const char *rest = COMMENT_FILE__LINE_ENDING;

			SG_ERR_CHECK(  SG_varray__get__sz(pCtx, descs, i, &asz)  );
			len = (SG_uint32)strlen(asz);

			for ( t = asz; *t; ++t )
				if ((*t == '\r') || (*t == '\n'))
				{
					len = (SG_uint32)(t - asz);
					rest = "..." COMMENT_FILE__LINE_ENDING;
					break;
				}

			// todo: line feeds? 
			SG_ERR_CHECK(  SG_string__append__sz(pCtx, newString, COMMENT_FILE__LINE_PREFIX " ")  );
			SG_ERR_CHECK(  SG_string__append__buf_len(pCtx, newString, (const SG_byte *)asz, len)  );
			SG_ERR_CHECK(  SG_string__append__sz(pCtx, newString, rest)  );
		}
	}

	SG_ERR_CHECK(  SG_string__append__sz(pCtx, newString, suffix)  );
	SG_ERR_CHECK(  SG_string__set__string(pCtx, boilerplate, newString)  );

fail:
	SG_STRING_NULLFREE(pCtx, newString);
}


void SG_cmd_util__get_comment_from_editor(
	SG_context* pCtx, 
	SG_string *pstrBoilerplate,
	char** ppszComment,
	const SG_varray *descs)
{
	SG_tempfile* pTempFile = NULL;
	char* pszEditor = NULL;
	SG_string* pstrFullCmdLine = NULL;
	SG_exit_status editorExitStatus;
	char* pszComment = NULL;

	SG_NULLARGCHECK_RETURN(ppszComment);

	SG_ERR_CHECK(  _sg_cmd_util__insert_associations_into_boilerplate(pCtx, pstrBoilerplate, descs)  );

	/* Create temp file, write boilerplate to it. */
	SG_ERR_CHECK(  SG_tempfile__open_new(pCtx, &pTempFile)  );

	/* It's not clear whether writing a UTF8 signature is a good idea or not. It's mostly a Windows
	   convention. For now, we don't. */
	/* SG_ERR_CHECK(  SG_file__write(pCtx, pTempFile->pFile, sizeof(UTF8_BOM), UTF8_BOM, NULL)  ); */

	SG_ERR_CHECK(  SG_file__write__string(pCtx, pTempFile->pFile, pstrBoilerplate)  );
	SG_ERR_CHECK(  SG_tempfile__close(pCtx, pTempFile)  );

	/* Open temp file with editor */
	SG_ERR_CHECK(  _get_editor(pCtx, &pszEditor)  );
	SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &pstrFullCmdLine, pszEditor)  );
	SG_ERR_CHECK(  SG_string__append__format(pCtx, pstrFullCmdLine, " %s", SG_pathname__sz(pTempFile->pPathname))  );
	SG_ERR_CHECK(  SG_exec__system(pCtx, SG_string__sz(pstrFullCmdLine), &editorExitStatus)  );
	
	/* Read from file. */
	if (!editorExitStatus)
		SG_ERR_CHECK(  _get_comment_from_file(pCtx, pTempFile->pPathname, &pszComment)  );

	*ppszComment = pszComment;
	pszComment = NULL;

	/* Fall through to common cleanup */
fail:
	SG_NULLFREE(pCtx, pszComment);
	SG_ERR_IGNORE(  SG_tempfile__close_and_delete(pCtx, &pTempFile)  );
	SG_STRING_NULLFREE(pCtx, pstrFullCmdLine);
	SG_NULLFREE(pCtx, pszEditor);
}


static void _get_editor(
	SG_context* pCtx,
	char** ppszEditor)
{
	char* psz = NULL;
	SG_string* pstr = NULL;
	SG_uint32 len = 0;

	/* Check config setting first. */
	SG_ERR_CHECK(  SG_localsettings__get__sz(pCtx, EDITOR_CONFIG_PATH, NULL, &psz, NULL)  );
	if (psz)
	{
		*ppszEditor = psz;
		return;
	}

	/* Check environment variables in order of preference. */
	SG_ERR_CHECK(  SG_environment__get__str(pCtx, "VVEDITOR", &pstr, &len)  );
	if (len)
	{
		SG_ERR_CHECK(  SG_string__sizzle(pCtx, &pstr, (SG_byte**)ppszEditor, NULL)  );
		return;
	}
	SG_STRING_NULLFREE(pCtx, pstr);
	SG_ERR_CHECK(  SG_environment__get__str(pCtx, "HGEDITOR", &pstr, &len)  );
	if (len)
	{
		SG_ERR_CHECK(  SG_string__sizzle(pCtx, &pstr, (SG_byte**)ppszEditor, NULL)  );
		return;
	}
	SG_STRING_NULLFREE(pCtx, pstr);
	SG_ERR_CHECK(  SG_environment__get__str(pCtx, "GIT_EDITOR", &pstr, &len)  );
	if (len)
	{
		SG_ERR_CHECK(  SG_string__sizzle(pCtx, &pstr, (SG_byte**)ppszEditor, NULL)  );
		return;
	}
	SG_STRING_NULLFREE(pCtx, pstr);
	SG_ERR_CHECK(  SG_environment__get__str(pCtx, "BZR_EDITOR", &pstr, &len)  );
	if (len)
	{
		SG_ERR_CHECK(  SG_string__sizzle(pCtx, &pstr, (SG_byte**)ppszEditor, NULL)  );
		return;
	}
	SG_STRING_NULLFREE(pCtx, pstr);
	SG_ERR_CHECK(  SG_environment__get__str(pCtx, "SVN_EDITOR", &pstr, &len)  );
	if (len)
	{
		SG_ERR_CHECK(  SG_string__sizzle(pCtx, &pstr, (SG_byte**)ppszEditor, NULL)  );
		return;
	}
	SG_STRING_NULLFREE(pCtx, pstr);
	SG_ERR_CHECK(  SG_environment__get__str(pCtx, "VISUAL", &pstr, &len)  );
	if (len)
	{
		SG_ERR_CHECK(  SG_string__sizzle(pCtx, &pstr, (SG_byte**)ppszEditor, NULL)  );
		return;
	}
	SG_STRING_NULLFREE(pCtx, pstr);
	SG_ERR_CHECK(  SG_environment__get__str(pCtx, "EDITOR", &pstr, &len)  );
	if (len)
	{
		SG_ERR_CHECK(  SG_string__sizzle(pCtx, &pstr, (SG_byte**)ppszEditor, NULL)  );
		return;
	}

	/* No relevant settings or environment variables set. Return the platform default. */
	SG_ERR_CHECK(  SG_STRDUP(pCtx, DEFAULT_EDITOR, ppszEditor)  );

	return;

fail:
	SG_NULLFREE(pCtx, psz);
	SG_STRING_NULLFREE(pCtx, pstr);
}
