
File Classes
----

A "file class" is just a type of file, such as "text" or "binary".  This is called a "class" instead of a "type" to avoid confusion with tool types.  Known classes are defined in localsettings, including information about how to recognize files of each class.  Files are divided into classes so that different tools can be used on different classes of files.  This way you can, for example, use a different merge tool for code files vs. XML files vs. binary files.

File Class Configuration Settings
---

The following settings are used to configure file classes.

fileclasses/CLASS
: The settings for a class named CLASS are stored under this tree.  A class' name is basically arbitrary, but by convention the names of built-in classes start with a colon.
fileclasses/CLASS/patterns
: A list/array of patterns that are used to recognize files of the class CLASS.  For example, for a file class named "code", the patterns list might contain `*.c`, `*.h`, `*.js`, and so forth.  These patterns are matched against candidate files using the same pattern matcher used by the include/exclude code.  The patterns should never include path information, as they're only ever matched against plain file names without paths.  When setting this value using `vv config`, remember to use `add-to` instead of `set` to create a list value instead of a string value.  Additionally, if you're adding any patterns that contain wildcards which your shell can expand, remember to escape them appropriately to get them past the shell.
fileclasses/CLASS/tools/TYPE/CONTEXT
: The name of the tool of type TYPE to invoke for files of class CLASS in context CONTEXT.  For example, to define the merge tool to use from the resolve command for code files to `foo`, you would set `fileclasses/code/tools/merge/resolve` to `foo`.
fileclasses/CLASS/tools/TYPE
: The name of the tool of type TYPE to invoke for files of class CLASS in all contexts.  You can use this when you want to use one tool across all contexts so that you don't have to specify the same tool for each context individually.

Built In Classes
---

There are two built in classes called `:text` and `:binary`, which are defined in the `default` config scope.  If the file classification code is unable to match a file to any of the classes in the config using their defined patterns, then it will run a binary file detection algorithm and use the result to assign the file to one of these two built in classes.

Note that the `:text` and `:binary` classes also have a number of filename patterns associated with them, like any other user-defined file class.  This is a performance shortcut, as it allows the classification code to identify a file as text or binary using simple standard pattern matching on its filename.  This is faster than falling back on the binary detection algorithm because that algorithm would require loading the file from disk.

Tools
---

A "tool" is a single program that filetool knows how to invoke in a particular manner.  A tool may be defined by the user in the localsettings (an "external" tool) or it may be compiled into Veracity (an "internal" tool).  Each external tool's configuration includes the arguments that are passed to the tool on the command line, so if you want to be able to call one external program with different command line parameters, you'll need to define a different tool for each set of parameters.  Each tool must have a unique name among all tools of the same type.

Tools are divided into "types", based on their purpose.  Code calling filetool must specify the type of tool that they're looking for.  The types of tools that are currently implemented (used by existing code) are listed below, along with more details about each.  New types of tools can be implemented relatively easily.

Some types of tools can be called in different "contexts".  For example, merge tools might be called from the `merge` command or the `resolve` command, and it can be helpful to configure different tools to be used in each context (like you probably want an interactive merge tool for `resolve`, but an automated one for `merge`).  It's up to the implementation of the tool type as to whether or not it uses contexts at all, and if so which contexts are available/used and what they mean.

Tool Configuration Settings
---

The following settings are used to configure tools:

filetools/TYPE/TOOL
: The settings for a tool named TOOL of type TYPE are stored under this tree.  For example, the settings for a merge tool called `diffmerge` are stored under `filetools/merge/diffmerge`.  A tool's name is basically arbitrary, but by convention the names of internal tools start with a colon.
filetools/TYPE/TOOL/path
: The full path to the tool's executable on disk.
filetools/TYPE/TOOL/args
: A list of command line arguments to pass to the tool when invoked.  These arguments may contain special tokens that will be replaced with actual data when the tool is actually invoked.  The set of tokens is different per type of tool; see the documentation for each type of tool (later in this document) for a list of supported tokens.  When adding items to this list using the `vv config` command, remember to use `add-to` instead of `set` to create a list value rather than a string value.  You will also need to escape flags, otherwise `vv` will interpret the flag as an argument to itself, rather than interpreting it as config data.  To escape the flag, prepend a backslash.  Note that if your shell recognizes backslashes, you may have to escape it with another backslash.  For example: 

    vv config add-to filetools/TYPE/TOOL/args \--flag # for shells that don't interpret \

    vv config add-to filetools/TYPE/TOOL/args \\--flag # for shells where the \ must itself be escaped

filetools/TYPE/TOOL/stdout
: A filename that should be bound to the tool's STDOUT stream when it is invoked.  This value may contain the same special tokens as the tool's args.  This might be used for tools that can only write to STDOUT, and can't be configured with a command line argument to write to a file instead.
filetools/TYPE/TOOL/stderr
: A filename that should be bound to the tool's STDERR stream when it is invoked.  This value may contain the same special tokens as the tool's args.
filetools/TYPE/TOOL/stdin
: A filename that should be bound to the tool's STDIN stream when it is invoked.  This value may contain the same special tokens as the tool's args.  This might be used for tools that can only read from STDIN, and can't be configured with a command line argument to read from a file instead.

Implemented Tool Types
---
This section explains the tool types currently implemented and used.

merge
---
Tools of type `merge` are used by the merge and resolve code/commands to perform 3-way file merges.

Contexts
---
Merge tools are always invoked in one of the following contexts:

merge
: The tool is being run as the result of a `vv merge` command or equivalent API call.  In other words, a collection of files is being merged into a working copy.  Since this implies a potentially large number of merges happening at once, using interactive tools in this context is unusual.  Generally the tools selected for this context will do the best that they can without user input, and leave complicated situations for later when a user can direct the merge.
resolve
: The tool is being run as the result of a `vv resolve` command or equivalent API call.  In other words, the user is resolving different changes to a single file.  This is usually the context where user interactive tools are used to guide the user through resolving conflicting changes.

Argument Tokens
---
Merge tools are invoked with values for the following argument tokens:

@ANCESTOR@
: The full path and filename of the "ancestor" version of the file involved in the merge.  This is the most recent common version of the file, before the "baseline" and "other" changesets were created.
@BASELINE@
: The full path and filename of the "baseline" version of the file involved in the merge.  This is a copy of the version of the file this is present in the "baseline" changeset.
@OTHER@
: The full path and filename of the "other" file involved in the merge.  This is the version of the file that is present in the "other" changeset in the merge.
@RESULT@
: The full path and filename where the tool should write the merged version of the file.
@ANCESTOR_LABEL@
: A user-friendly label for the "ancestor" file.  Sometimes this might just be the filename.
@BASELINE_LABEL@
: A user-friendly label for the "baseline" file.  Sometimes this might just be the filename.
@OTHER_LABEL@
: A user-friendly label for the "other" file.  Sometimes this might just be the filename.
@RESULT_LABEL@
: A user-friendly label for the "result" file.

Internal Merge Tools
---

The following merge tools are compiled into sglib:

\:skip
: This tool makes no attempt to merge the file, it simply returns that the merge was skipped/not attempted.  This tool is generally used for file classes that cannot be meaningfully merged, like most binary files.
\:merge
: This is a compiled-in automatic 3-way merge tool, which is functionally the same as GNU diff3.  It also ignores end-of-line differences.
\:merge_strict
: This is like `:merge` except that it respects end-of-line differences.
\:baseline
: This tool resolves merges by simply accepting the "baseline" file as the result, rejecting any changes in the "other" file.
\:other
: This tool resolves merges by simply accepting the "other" file as the result, rejecting any changes in the "baseline" file.
\:gnu_diff3
: This is a tool that has hard-coded logic to find and execute an external installation of GNU's diff3 program.  This may eventually be replaced by equivalent localsettings in the "default" (hard-coded, read-only) scope.
\:diffmerge
: This is a tool that has hard-coded logic to find and execute an external installation of SourceGear's DiffMerge program.  This may eventually be replaced by equivalent localsettings in the "default" (hard-coded, read-only) scope.
