/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

load("vscript_test_lib.js");

//////////////////////////////////////////////////////////////////
// Test for sprawl-912.
// Create a merge with a delete-vs-edit file conflict (where the
// delete is overruled).
// Try to force the delete (without using RESOLVE) so that you can
// commit without the file.  (Previously, you'd have to commit with
// the file and then delete it in a subsequent commit.)
//////////////////////////////////////////////////////////////////

function st_merge_delete_vs_edit__SPRAWL_912()
{
    load("update_helpers.js");          // load the helper functions
    initialize_update_helpers(this);    // initialize helper functions

    //////////////////////////////////////////////////////////////////

    this.setUp = function()
    {
        this.workdir_root = sg.fs.getcwd();     // save this for later

        //////////////////////////////////////////////////////////////////
        // fetch the HID of the initial commit.

        var h = sg.pending_tree.history();
        this.csets.push( h[0].changeset_id );
        this.names.push( "*Initial Commit*" );

        //////////////////////////////////////////////////////////////////
	// create:
	//
	//      0
	//      |
	//      1      create a file in (1)
	//      |
	//      2
	//     / \
	//    3   4    edit the file in (3); delete the file in (4)

        this.do_fsobj_mkdir("dir_0");
        this.do_addremove_scan();
        this.do_commit("CSET 0");

        this.do_fsobj_mkdir("dir_1");
        this.do_fsobj_create_file_exactly("dir_1/file_11", "Version 1");
        this.do_addremove_scan();
        this.do_commit("CSET 1");

        this.do_fsobj_mkdir("dir_2");
        this.do_addremove_scan();
        this.do_commit("CSET 2");

        this.do_fsobj_mkdir("dir_3");
        this.do_addremove_scan();
	this.do_fsobj_append_file("dir_1/file_11", "Hello from [3].");
        this.do_commit("CSET 3");

	this.rev_trunk_head = this.csets[ this.csets.length - 1 ];

	//////////////////////////////////////////////////////////////////
	// update back to '2' and start the other branch.

	this.do_update_when_clean_by_name("CSET 2");
        this.do_fsobj_mkdir("dir_4");
	this.do_vc_remove_file("dir_1/file_11");
        this.do_addremove_scan();
        this.do_commit_in_branch("CSET 4");

	this.rev_branch_head = this.csets_in_branch[ this.csets_in_branch.length - 1 ];

	//////////////////////////////////////////////////////////////////

	var o = sg.exec(vv, "heads");
	print("Heads are:");
	print(o.stdout);
    }

    //////////////////////////////////////////////////////////////////
    //
    //////////////////////////////////////////////////////////////////

    this.do_update_by_rev = function(rev)
    {
	print("vv update --rev=" + rev);
	var o = sg.exec(vv, "update", "--rev="+rev);

	//print("STDOUT:\n" + o.stdout);
	//print("STDERR:\n" + o.stderr);

	testlib.ok( o.exit_status == 0, "Exit status should be 0" );
	this.assert_current_baseline(rev);
    }

    this.do_exec_merge = function(rev, bExpectFail, msgExpected)
    {
	print("vv merge --rev=" + rev);
	var o = sg.exec(vv, "merge", "--rev="+rev, "--verbose");
	
	print("STDOUT:\n" + o.stdout);
	print("STDERR:\n" + o.stderr);

	if (bExpectFail)
	{
	    testlib.ok( o.exit_status == 1, "Exit status should be 1" );
	    testlib.ok( o.stderr.search(msgExpected) >= 0, "Expecting error message [" + msgExpected + "]. Received:\n" + o.stderr);
	}
	else
	{
	    testlib.ok( o.exit_status == 0, "Exit status should be 0" );
	}
    }

    //////////////////////////////////////////////////////////////////
    // NOTE: this.confirm() and this.confirm_when_ignoring_ignores()
    //       compute a STATUS and compare that with the expected results.
    //       A STATUS reflects the changes made to the WD from the
    //       BASELINE.  So the STATUS should show how the FINAL RESULT
    //       changed from the BASELINE.
    //
    //       This feels a little odd because we are accustomed to
    //       think about the total net changes from the ANCESTOR to
    //       the FINAL RESULT.
    //
    //       So, for example, if a file was renamed between the
    //       ancestor and the baseline branch (but not in the other
    //       changeset), it will not show up in the status.  It will
    //       have the new name in the final result, but that is not
    //       a change relative to the baseline.
    //////////////////////////////////////////////////////////////////

    this.test_simple_tb_1 = function()
    {
	this.force_clean_WD();

	// set WD to the trunk.
	// merge (fold) in the branch.
	// make sure basic clean merge works as expected.

	this.do_update_by_rev(this.rev_trunk_head);
	this.do_exec_merge(this.rev_branch_head, 0, "");

	// dir_1/file_11 is still in trunk. it was renamed by merge 
	// due to delete-vs-edit conflict.

        var expect_test = new Array;
        expect_test["Added"] = [ "@/dir_4" ];
	expect_test["Renamed"] = [ "@/dir_1/file_11~L0~g[0-9a-f]*~" ];
        this.confirm_when_ignoring_ignores(expect_test);

	// try to let the DELETE win.
	//
	// CURRENTLY 'vv remove' (remove-safely) does not allow REMOVES
	// on things that have been MOVED or RENAMED.  So we need to
	// manually undo the rename first.
	// See W8935.

	var dir = sg.fs.readdir("dir_1");
	for (var k in dir)
	{
	    var name_k = dir[k].name;
	    if (name_k.search("file_11") != -1)
	    {
		this.do_vc_rename( ("dir_1/" + name_k), "file_11" );
	    }
	}

	var result = sg.exec(vv, "remove", "dir_1/file_11");
	if (result.exit_status != 0)
	    print(dumpObj(result, "sg.exec()", "", 0));
	testlib.ok( (result.exit_status == 0), "Remove should pass");

        var expect_test = new Array;
        expect_test["Added"] = [ "@/dir_4" ];
	expect_test["Removed"] = [ "@/dir_1/file_11" ];
        this.confirm_when_ignoring_ignores(expect_test);

	var result = sg.exec(vv, "status", "--no-ignores");
	print(dumpObj(result, "sg.exec()", "", 0));
	testlib.ok( (result.exit_status == 0), "STATUS should pass");
    }

    this.test_simple_bt_1 = function()
    {
	this.force_clean_WD();

	// set WD to the branch.
	// merge (fold) in the trunk.
	// make sure basic clean merge works as expected.

	this.do_update_by_rev(this.rev_branch_head);
	this.do_exec_merge(this.rev_trunk_head, 0, "");

	// dir_1/file_11 was deleted from branch.  it should appear
	// as added because of the delete-vs-edit conflict.  that is,
	// it appears as added relative to the branch-baseline.

        var expect_test = new Array;
        expect_test["Added"] = [ "@/dir_3",
				 "@/dir_1/file_11~A~g[0-9a-f]*~" ];
        this.confirm_when_ignoring_ignores(expect_test);

	// try to let the DELETE win.
	// this is a long way of saying "/bin/rm dir_1/file_11*"...

	var dir = sg.fs.readdir("dir_1");
	for (var k in dir)
	{
	    var name_k = dir[k].name;
	    if (name_k.search("file_11") != -1)
	    {
		var result = sg.exec(vv, "remove", ("dir_1/" + name_k));
		print(dumpObj(result, "sg.exec()", "", 0));
		testlib.ok( (result.exit_status == 0), "Remove should pass");
	    }
	}

        var expect_test = new Array;
        expect_test["Added"] = [ "@/dir_3" ];
	expect_test["Removed"] = [ "@/dir_1/file_11~A~g[0-9a-f]*~" ];
        this.confirm_when_ignoring_ignores(expect_test);

	var result = sg.exec(vv, "status", "--no-ignores");
	print(dumpObj(result, "sg.exec()", "", 0));
	testlib.ok( (result.exit_status == 0), "STATUS should pass");

    }

}
