/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include "sg_fs3__private.h"

struct _SG_dbndx_update
{
	SG_repo* pRepo;
    SG_uint64 iDagNum;

    SG_vhash* pvh_schema;

    sqlite3* psql;
	SG_bool			bInTransaction;

    struct
    {
        SG_bool b_in_update;
        SG_uint32 count_so_far;
        SG_uint32 count_changesets;
        sqlite3_stmt* pStmt_history;
        SG_rbtree* prb_stmt_cache;
    } update;
};

#define MY_SLEEP_MS      100
#define MY_TIMEOUT_MS  30000

void SG_dbndx_update__open(
	SG_context* pCtx,
	SG_repo* pRepo,
	SG_uint64 iDagNum,
    sqlite3* psql,
	SG_dbndx_update** ppNew
	)
{
	SG_dbndx_update* pndx = NULL;
    sqlite3_stmt* pStmt_schema = NULL;
    int rc = -1;

	SG_ERR_CHECK(  SG_alloc1(pCtx, pndx)  );

    pndx->pRepo = pRepo;
    pndx->iDagNum = iDagNum;
	pndx->bInTransaction = SG_FALSE;

    if (SG_DAGNUM__IS_AUDIT(iDagNum))
    {
        SG_uint64 iBaseDagNum = SG_DAGNUM__AUDIT_TO_BASE(iDagNum);

        if (SG_DAGNUM__IS_DB(iBaseDagNum))
        {
            SG_ERR_THROW(  SG_ERR_NOTIMPLEMENTED  );
        }
    }

    pndx->psql = psql;

    SG_ERR_CHECK(  sg_sqlite__prepare(pCtx, pndx->psql, &pStmt_schema, "SELECT json FROM cur_schema")  );
    if ((rc = sqlite3_step(pStmt_schema)) == SQLITE_ROW)
    {
        const char* psz_json = (const char*) sqlite3_column_text(pStmt_schema,0);

        SG_ERR_CHECK(  SG_VHASH__ALLOC__FROM_JSON__SZ(pCtx, &pndx->pvh_schema, psz_json)  );
    }
    else if (SQLITE_DONE != rc)
    {
        SG_ERR_THROW(SG_ERR_SQLITE(rc));
    }
    SG_ERR_CHECK(  sg_sqlite__nullfinalize(pCtx, &pStmt_schema)  );

	*ppNew = pndx;
    pndx = NULL;

fail:
    SG_dbndx_update__free(pCtx, pndx);
    SG_ERR_IGNORE(  sg_sqlite__nullfinalize(pCtx, &pStmt_schema)  );
}

void SG_dbndx_update__free(SG_context* pCtx, SG_dbndx_update* pdbc)
{
	if (!pdbc)
		return;

    SG_VHASH_NULLFREE(pCtx, pdbc->pvh_schema);

	SG_NULLFREE(pCtx, pdbc);
}

static void sg_dbndx__add_record_to_history_table(
        SG_context* pCtx,
        SG_dbndx_update* pdbc,
        SG_dbrecord* prec,
        const char* psz_csid,
        SG_int32 generation
        )
{
    if (SG_DAGNUM__HAS_NO_RECID(pdbc->iDagNum))
    {
        const char* psz_hid = NULL;

        SG_ERR_CHECK(  SG_dbrecord__get_hid__ref(pCtx, prec, &psz_hid)  );

        SG_ERR_CHECK(  sg_sqlite__reset(pCtx, pdbc->update.pStmt_history)  );
        SG_ERR_CHECK(  sg_sqlite__clear_bindings(pCtx, pdbc->update.pStmt_history)  );
        SG_ERR_CHECK(  sg_sqlite__bind_text(pCtx, pdbc->update.pStmt_history, 1, psz_hid)  );
        SG_ERR_CHECK(  sg_sqlite__bind_text(pCtx, pdbc->update.pStmt_history, 2, psz_csid)  );
        SG_ERR_CHECK(  sg_sqlite__bind_int(pCtx, pdbc->update.pStmt_history, 3, generation)  );

        SG_ERR_CHECK(  sg_sqlite__step(pCtx, pdbc->update.pStmt_history, SQLITE_DONE)  );
    }
    else
    {
        const char* psz_recid = NULL;

        SG_ERR_CHECK(  SG_dbrecord__check_value(pCtx, prec, SG_ZING_FIELD__RECID, &psz_recid)  );
        if (psz_recid)
        {
            const char* psz_hid = NULL;

            SG_ASSERT(psz_recid);

            SG_ERR_CHECK(  SG_dbrecord__get_hid__ref(pCtx, prec, &psz_hid)  );

            SG_ERR_CHECK(  sg_sqlite__reset(pCtx, pdbc->update.pStmt_history)  );
            SG_ERR_CHECK(  sg_sqlite__clear_bindings(pCtx, pdbc->update.pStmt_history)  );
            SG_ERR_CHECK(  sg_sqlite__bind_text(pCtx, pdbc->update.pStmt_history, 1, psz_recid)  );
            SG_ERR_CHECK(  sg_sqlite__bind_text(pCtx, pdbc->update.pStmt_history, 2, psz_hid)  );
            SG_ERR_CHECK(  sg_sqlite__bind_text(pCtx, pdbc->update.pStmt_history, 3, psz_csid)  );
            SG_ERR_CHECK(  sg_sqlite__bind_int(pCtx, pdbc->update.pStmt_history, 4, generation)  );

            SG_ERR_CHECK(  sg_sqlite__step(pCtx, pdbc->update.pStmt_history, SQLITE_DONE)  );
        }
    }

fail:
    ;
}

static void sg_dbndx__add_record_to_full_text_search(
        SG_context* pCtx,
        SG_dbndx_update* pndx,
        SG_dbrecord* prec,
        SG_vhash* pvh_schema
        )
{
	SG_uint32 count_fields = 0;
	SG_uint32 i;
    const char* psz_hid = NULL;
    const char* psz_rectype = NULL;
    SG_vhash* pvh_fields = NULL;
    sqlite3_stmt* pStmt = NULL;

    if (SG_DAGNUM__HAS_SINGLE_RECTYPE(pndx->iDagNum))
    {
        SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvh_schema, 0, &psz_rectype, NULL)  );
    }
    else
    {
        SG_ERR_CHECK(  SG_dbrecord__check_value(pCtx, prec, SG_ZING_FIELD__RECTYPE, &psz_rectype)  );
        SG_ASSERT(psz_rectype);
    }

    SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvh_schema, psz_rectype, &pvh_fields)  );
    SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_fields, &count_fields)  );

    SG_ERR_CHECK(  SG_dbrecord__get_hid__ref(pCtx, prec, &psz_hid)  );
    
	for (i=0; i<count_fields; i++)
	{
        SG_vhash* pvh_one_field = NULL;
        const char* psz_field_name = NULL;
        const char* psz_val = NULL;
        SG_bool b_has = SG_FALSE;
        SG_bool b_full_text_search = SG_FALSE;

        SG_ERR_CHECK(  SG_vhash__get_nth_pair__vhash(pCtx, pvh_fields, i, &psz_field_name, &pvh_one_field)  );

        SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh_one_field, SG_DBNDX_SCHEMA__FIELD__FULL_TEXT_SEARCH, &b_has)  );
        if (b_has)
        {
            SG_ERR_CHECK(  SG_vhash__get__bool(pCtx, pvh_one_field, SG_DBNDX_SCHEMA__FIELD__FULL_TEXT_SEARCH, &b_full_text_search)  );
        }

        if (b_full_text_search)
        {
            SG_ERR_CHECK(  SG_dbrecord__check_value(pCtx, prec, psz_field_name, &psz_val)  );

            if (psz_val)
            {
                sqlite3_int64 rowid = -1;

                SG_ERR_CHECK(  sg_sqlite__prepare(pCtx, pndx->psql, &pStmt, "INSERT INTO \"%s_%s_%s\" (text) VALUES (?)", SG_DBNDX_FTS_TABLE_PREFIX, psz_rectype, psz_field_name)  );

                SG_ERR_CHECK(  sg_sqlite__bind_text(pCtx, pStmt, 1, psz_val)  );
                SG_ERR_CHECK(  sg_sqlite__step(pCtx, pStmt, SQLITE_DONE)  );
                SG_ERR_CHECK(  sg_sqlite__nullfinalize(pCtx, &pStmt)  );
                rowid = sqlite3_last_insert_rowid(pndx->psql);

                SG_ERR_CHECK(  sg_sqlite__prepare(pCtx, pndx->psql, &pStmt, "INSERT INTO \"%s_map_%s_%s\" (ftsrowid, hidrec) VALUES (?, ?)", SG_DBNDX_FTS_TABLE_PREFIX, psz_rectype, psz_field_name)  );
                SG_ERR_CHECK(  sg_sqlite__bind_int64(pCtx, pStmt, 1, rowid)  );
                SG_ERR_CHECK(  sg_sqlite__bind_text(pCtx, pStmt, 2, psz_hid)  );
                SG_ERR_CHECK(  sg_sqlite__step(pCtx, pStmt, SQLITE_DONE)  );
                SG_ERR_CHECK(  sg_sqlite__nullfinalize(pCtx, &pStmt)  );
            }
        }
	}

fail:
    ;
}

static void sg_dbndx__add_record_to_records_table(
        SG_context* pCtx,
        SG_dbndx_update* pndx,
        SG_dbrecord* prec,
        SG_vhash* pvh_schema
        )
{
	SG_uint32 count_fields = 0;
	SG_uint32 i;
	SG_uint32 ndx_bind = 0;
    const char* psz_hid = NULL;
    const char* psz_rectype = NULL;
    sqlite3_stmt* pStmt = NULL;
    SG_vhash* pvh_fields = NULL;
    SG_string* pstr_sql = NULL;
    SG_uint32 count_extra_ndx_columns = 0;
    SG_bool b_stmt_already_in_cache = SG_FALSE;

    if (SG_DAGNUM__HAS_SINGLE_RECTYPE(pndx->iDagNum))
    {
        SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvh_schema, 0, &psz_rectype, NULL)  );
    }
    else
    {
        SG_ERR_CHECK(  SG_dbrecord__check_value(pCtx, prec, SG_ZING_FIELD__RECTYPE, &psz_rectype)  );
        SG_ASSERT(psz_rectype);
    }

    SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvh_schema, psz_rectype, &pvh_fields)  );
    SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_fields, &count_fields)  );

    // check to see if we already have the insert stmt cached
    b_stmt_already_in_cache = SG_FALSE;
    pStmt = NULL;
    SG_ERR_CHECK(  SG_rbtree__find(pCtx, pndx->update.prb_stmt_cache, psz_rectype, &b_stmt_already_in_cache, (void**) &pStmt)  );

    if (!pStmt)
    {
        // build the stmt

        SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pstr_sql)  );
        // we use OR IGNORE here because the record tables have hidrec as
        // the primary key, and it should be safe, because if the hidrec
        // matches, then every other field should match exactly as well.
        SG_ERR_CHECK(  SG_string__sprintf(pCtx, pstr_sql, "INSERT OR IGNORE INTO '%s_%s' (hidrec",
                    SG_DBNDX_RECORD_TABLE_PREFIX,
                    psz_rectype
                    )  );

        for (i=0; i<count_fields; i++)
        {
            const char* psz_field_name = NULL;
            SG_vhash* pvh_field = NULL;
            SG_bool b_has = SG_FALSE;
            SG_bool b_sort_by_allowed = SG_FALSE;

            SG_ERR_CHECK(  SG_vhash__get_nth_pair__vhash(pCtx, pvh_fields, i, &psz_field_name, &pvh_field)  );

            SG_ERR_CHECK(  SG_string__append__format(
                        pCtx, 
                        pstr_sql, 
                        ", \"%s\"",
                        psz_field_name
                        )  );
            SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh_field, SG_DBNDX_SCHEMA__FIELD__SORT_BY_ALLOWED, &b_has)  );
            if (b_has)
            {
                SG_ERR_CHECK(  SG_vhash__get__bool(pCtx, pvh_field, SG_DBNDX_SCHEMA__FIELD__SORT_BY_ALLOWED, &b_sort_by_allowed)  );
            }
            if (b_sort_by_allowed)
            {
                count_extra_ndx_columns++;
                SG_ERR_CHECK(  SG_string__append__format(
                            pCtx, 
                            pstr_sql, 
                            ", \"ndx$_%s\"",
                            psz_field_name
                            )  );
            }
        }

        SG_ERR_CHECK(  SG_string__append__sz(
                    pCtx, 
                    pstr_sql, 
                    ") VALUES (?"
                    )  );

        for (i=0; i<count_fields + count_extra_ndx_columns; i++)
        {
            SG_ERR_CHECK(  SG_string__append__sz(
                        pCtx, 
                        pstr_sql, 
                        ", ?"
                        )  );
        }

        SG_ERR_CHECK(  SG_string__append__sz(
                    pCtx, 
                    pstr_sql, 
                    ")"
                    )  );

        SG_ERR_CHECK(  sg_sqlite__prepare(pCtx, pndx->psql, &pStmt, "%s", SG_string__sz(pstr_sql))  );
        
        SG_ERR_CHECK(  SG_rbtree__add__with_assoc(pCtx, pndx->update.prb_stmt_cache, psz_rectype, pStmt)  );

        SG_STRING_NULLFREE(pCtx, pstr_sql);
    }

    SG_ERR_CHECK(  sg_sqlite__reset(pCtx, pStmt)  );
    SG_ERR_CHECK(  sg_sqlite__clear_bindings(pCtx, pStmt)  );

    SG_ERR_CHECK(  SG_dbrecord__get_hid__ref(pCtx, prec, &psz_hid)  );
    SG_ERR_CHECK(  sg_sqlite__bind_text(pCtx, pStmt, 1, psz_hid)  );
    
    // Now bind everything into the stmt
    ndx_bind = 2; // bind indexes are 1-based, and the hid was 1
	for (i=0; i<count_fields; i++)
	{
        SG_vhash* pvh_one_field = NULL;
        const char* psz_field_name = NULL;
        const char* psz_field_type = NULL;
        SG_bool b_integerish = SG_FALSE;
        const char* psz_val = NULL;
        SG_bool b_has = SG_FALSE;
        SG_bool b_sort_by_allowed = SG_FALSE;

        SG_ERR_CHECK(  SG_vhash__get_nth_pair__vhash(pCtx, pvh_fields, i, &psz_field_name, &pvh_one_field)  );
        SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh_one_field, SG_DBNDX_SCHEMA__FIELD__TYPE, &psz_field_type)  );
        SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh_one_field, SG_DBNDX_SCHEMA__FIELD__SORT_BY_ALLOWED, &b_has)  );
        if (b_has)
        {
            SG_ERR_CHECK(  SG_vhash__get__bool(pCtx, pvh_one_field, SG_DBNDX_SCHEMA__FIELD__SORT_BY_ALLOWED, &b_sort_by_allowed)  );
        }
        SG_ERR_CHECK(  SG_zing__field_type_is_integerish__sz(pCtx, psz_field_type, &b_integerish)  );

        SG_ERR_CHECK(  SG_dbrecord__check_value(pCtx, prec, psz_field_name, &psz_val)  );

        if (psz_val)
        {
            if (b_integerish)
            {
                SG_int64 intvalue = 0;

                SG_ERR_CHECK(  SG_int64__parse__strict(pCtx, &intvalue, psz_val)  );
                SG_ERR_CHECK(  sg_sqlite__bind_int64(pCtx, pStmt, ndx_bind++, intvalue)  );
                SG_ASSERT(!b_sort_by_allowed);
            }
            else
            {
                SG_ERR_CHECK(  sg_sqlite__bind_text(pCtx, pStmt, ndx_bind++, psz_val)  );
                if (b_sort_by_allowed)
                {
                    SG_vhash* pvh_allowed = NULL;
                    SG_int64 ndx_val = -1;

                    SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvh_one_field, SG_DBNDX_SCHEMA__FIELD__ALLOWED, &pvh_allowed)  );
                    SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvh_allowed, psz_val, &ndx_val)  );
                    SG_ERR_CHECK(  sg_sqlite__bind_int64(pCtx, pStmt, ndx_bind++, ndx_val)  );
                }
            }
        }
        else
        {
            SG_ERR_CHECK(  sg_sqlite__bind_null(pCtx, pStmt, ndx_bind++)  );
            if (b_sort_by_allowed)
            {
                SG_ERR_CHECK(  sg_sqlite__bind_null(pCtx, pStmt, ndx_bind++)  );
            }
        }
	}

    SG_ERR_CHECK(  sg_sqlite__step(pCtx, pStmt, SQLITE_DONE)  );

fail:
    SG_STRING_NULLFREE(pCtx, pstr_sql);
}

static void sg_dbndx__create_indexes(SG_context* pCtx, SG_dbndx_update* pndx)
{
#if 0
    SG_NULLARGCHECK_RETURN(pndx);
#else
    SG_uint32 count_rectypes = 0;
    SG_uint32 count_fields = 0;
    SG_uint32 i_rectype = 0;
    SG_uint32 i_field = 0;

    SG_NULLARGCHECK_RETURN(pndx);

    // TODO this is called after every update.  the idea is that
    // on a big clone, we want to leave all the record tables
    // un-indexed until the end, so the bulk inserts will be
    // faster.  And this can be called over and over, since
    // all the CREATE INDEX calls use IF NOT EXISTS.  But is
    // there a perf penalty for this when we're calling it so
    // often?  Is the tradeoff worth it?

    SG_ERR_CHECK(  SG_vhash__count(pCtx, pndx->pvh_schema, &count_rectypes)  );
    for (i_rectype=0; i_rectype<count_rectypes; i_rectype++)
    {
        const char* psz_rectype = NULL;
        SG_vhash* pvh_rectype = NULL;

        SG_ERR_CHECK(  SG_vhash__get_nth_pair__vhash(pCtx, pndx->pvh_schema, i_rectype, &psz_rectype, &pvh_rectype)  );

        SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_rectype, &count_fields)  );
        for (i_field=0; i_field<count_fields; i_field++)
        {
            const char* psz_field = NULL;
            SG_vhash* pvh_field = NULL;
            SG_bool b_has = SG_FALSE;
            SG_bool b_sort_by_allowed = SG_FALSE;
            SG_bool b_index = SG_FALSE;
            SG_uint16 type = 0;
            const char* psz_field_type = NULL;

            SG_ERR_CHECK(  SG_vhash__get_nth_pair__vhash(pCtx, pvh_rectype, i_field, &psz_field, &pvh_field)  );

            SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh_field, SG_DBNDX_SCHEMA__FIELD__TYPE, &psz_field_type)  );
            SG_ERR_CHECK(  SG_zing__field_type__sz_to_uint16(pCtx, psz_field_type, &type)  );
            // never index a bool field
            if (SG_ZING_TYPE__BOOL != type)
            {
                SG_ERR_CHECK(  SG_vhash__get__bool(pCtx, pvh_field, SG_DBNDX_SCHEMA__FIELD__INDEX, &b_index)  );

                if (b_index)
                {
                    SG_ERR_CHECK(  sg_sqlite__exec__va(pCtx, pndx->psql, "CREATE INDEX IF NOT EXISTS \"index$_%s_%s\" ON \"%s_%s\" (\"%s\")",
                                psz_rectype,
                                psz_field,
                                SG_DBNDX_RECORD_TABLE_PREFIX,
                                psz_rectype,
                                psz_field
                                )  );

                    SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh_field, SG_DBNDX_SCHEMA__FIELD__SORT_BY_ALLOWED, &b_has)  );
                    if (b_has)
                    {
                        SG_ERR_CHECK(  SG_vhash__get__bool(pCtx, pvh_field, SG_DBNDX_SCHEMA__FIELD__SORT_BY_ALLOWED, &b_sort_by_allowed)  );
                    }
                    if (b_sort_by_allowed)
                    {
                        SG_ERR_CHECK(  sg_sqlite__exec__va(pCtx, pndx->psql, "CREATE INDEX IF NOT EXISTS \"index$_%s_ndx$_%s\" ON \"%s_%s\" (\"ndx$_%s\")",
                                    psz_rectype,
                                    psz_field,
                                    SG_DBNDX_RECORD_TABLE_PREFIX,
                                    psz_rectype,
                                    psz_field)  );
                    }
                }
            }
        }
    }

fail:
    ;
#endif
}

static void SG_dbndx_update__do_audits(
	SG_context* pCtx,
	SG_dbndx_update* pdbc,
    SG_varray* pva_audits
	)
{
    SG_uint32 count = 0;
    SG_uint32 i = 0;
	sqlite3_stmt* pStmt_add = NULL;

    if (pva_audits)
    {
        SG_ERR_CHECK(  sg_sqlite__prepare(pCtx, pdbc->psql, &pStmt_add, "INSERT INTO db_audits (csid, userid, timestamp, hidrec) VALUES (?, ?, ?, ?)")  );

        SG_ERR_CHECK(  SG_varray__count(pCtx, pva_audits, &count)  );
        for (i=0; i<count; i++)
        {
            SG_vhash* pvh = NULL;
            const char* psz_csid = NULL;
            const char* psz_who = NULL;
            SG_int64 i_when = 0;
            const char* psz_hid_audit = NULL;

            SG_ERR_CHECK(  SG_varray__get__vhash(pCtx, pva_audits, i, &pvh)  );

            SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh, "hidrec", &psz_hid_audit)  );

            SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh, "csid", &psz_csid)  );
            SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh, SG_AUDIT__USERID, &psz_who)  );
            SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvh, SG_AUDIT__TIMESTAMP, &i_when)  );

            SG_ERR_CHECK(  sg_sqlite__reset(pCtx, pStmt_add)  );
            SG_ERR_CHECK(  sg_sqlite__clear_bindings(pCtx, pStmt_add)  );

            SG_ERR_CHECK(  sg_sqlite__bind_text(pCtx, pStmt_add, 1, psz_csid)  );
            SG_ERR_CHECK(  sg_sqlite__bind_text(pCtx, pStmt_add, 2, psz_who)  );
            SG_ERR_CHECK(  sg_sqlite__bind_int64(pCtx, pStmt_add, 3, i_when)  );
            SG_ERR_CHECK(  sg_sqlite__bind_text(pCtx, pStmt_add, 4, psz_hid_audit)  );

            SG_ERR_CHECK(  sg_sqlite__step__retry(pCtx, pStmt_add, SQLITE_DONE, MY_SLEEP_MS, MY_TIMEOUT_MS)  );
        }

        SG_ERR_CHECK(  sg_sqlite__nullfinalize(pCtx, &pStmt_add)  );
    }

fail:
    SG_ERR_IGNORE(  sg_sqlite__nullfinalize(pCtx, &pStmt_add)  );
}

static void SG_dbndx_update__begin(SG_context* pCtx, SG_dbndx_update* pdbc, SG_uint32 count_changesets)
{
    SG_ASSERT(!pdbc->update.b_in_update);

    memset(&pdbc->update, 0, sizeof(pdbc->update));

    pdbc->update.b_in_update = SG_TRUE;
    pdbc->update.count_changesets = count_changesets;

    SG_ERR_CHECK(  sg_sqlite__exec__retry(pCtx, pdbc->psql, "BEGIN TRANSACTION", MY_SLEEP_MS, MY_TIMEOUT_MS)  );
	pdbc->bInTransaction = SG_TRUE;

    if (SG_DAGNUM__HAS_NO_RECID(pdbc->iDagNum))
    {
        SG_ERR_CHECK(  sg_sqlite__prepare(pCtx, pdbc->psql, &pdbc->update.pStmt_history, "INSERT INTO db_history (hidrec, csid, generation) VALUES (?, ?, ?)")  );
    }
    else
    {
        SG_ERR_CHECK(  sg_sqlite__prepare(pCtx, pdbc->psql, &pdbc->update.pStmt_history, "INSERT INTO db_history (recid, hidrec, csid, generation) VALUES (?, ?, ?, ?)")  );
    }

    SG_ERR_CHECK(  SG_rbtree__alloc(pCtx, &pdbc->update.prb_stmt_cache)  );

fail:
    ;
}

static void my_finalize_cached_stmts (
        SG_context* pCtx,
        SG_rbtree* prb
       )
{
    SG_rbtree_iterator* pit = NULL;
    SG_bool b = SG_FALSE;
    const char* psz_rectype = NULL;
    sqlite3_stmt* pStmt = NULL;

    SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pit, prb, &b, &psz_rectype, (void**) &pStmt)  );
    while (b)
    {
        SG_ERR_CHECK(  sg_sqlite__finalize(pCtx, pStmt)  );

        SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pit, &b, &psz_rectype, (void**) &pStmt)  );
    }

fail:
    SG_RBTREE_ITERATOR_NULLFREE(pCtx, pit);
}

static void SG_dbndx_update__end(SG_context* pCtx, SG_dbndx_update* pdbc)
{
    SG_ASSERT(pdbc->update.b_in_update);

    SG_ASSERT(pdbc->update.count_changesets == pdbc->update.count_so_far);

	SG_ERR_CHECK(  sg_sqlite__nullfinalize(pCtx, &pdbc->update.pStmt_history)  );

    SG_ERR_CHECK(  my_finalize_cached_stmts(pCtx, pdbc->update.prb_stmt_cache)  );
    SG_RBTREE_NULLFREE(pCtx, pdbc->update.prb_stmt_cache);

    SG_ERR_CHECK(  sg_dbndx__create_indexes(pCtx, pdbc)  );

    SG_ERR_CHECK(  sg_sqlite__exec__retry(pCtx, pdbc->psql, "COMMIT TRANSACTION", MY_SLEEP_MS, MY_TIMEOUT_MS)  );
	pdbc->bInTransaction = SG_FALSE;

    pdbc->update.b_in_update = SG_FALSE;

fail:
    ;
}

static void SG_dbndx_update__abort(SG_context* pCtx, SG_dbndx_update* pdbc)
{
	SG_ERR_CHECK(  sg_sqlite__nullfinalize(pCtx, &pdbc->update.pStmt_history)  );

    SG_ERR_CHECK(  my_finalize_cached_stmts(pCtx, pdbc->update.prb_stmt_cache)  );
    SG_RBTREE_NULLFREE(pCtx, pdbc->update.prb_stmt_cache);

    if (pdbc->update.b_in_update)
    {
        if (pdbc->bInTransaction)
        {
            SG_ERR_CHECK(  sg_sqlite__exec__retry(pCtx, pdbc->psql, "ROLLBACK TRANSACTION", MY_SLEEP_MS, MY_TIMEOUT_MS)  );
            pdbc->bInTransaction = SG_FALSE;
        }
    }

    pdbc->update.b_in_update = SG_FALSE;

fail:
    ;
}

void SG_dbndx_update__do(
        SG_context* pCtx, 
        SG_dbndx_update* pdbc, 
        SG_varray* pva_audits,
        SG_vhash* pvh_changesets
        )
{
    SG_changeset* pcs = NULL;
	SG_dbrecord* pRec = NULL;
	sqlite3_stmt* pStmt = NULL;
    SG_uint32 i_changeset = 0;
    SG_uint32 count_changesets = 0;
    SG_vhash* pvh_schema = NULL;
    sqlite3_stmt* pStmt_schema = NULL;
    SG_vhash* pvh_new = NULL;
    SG_vhash* pvh_intersection = NULL;

	SG_NULLARGCHECK_RETURN(pdbc);
	SG_ARGCHECK_RETURN( pdbc->pRepo, pdbc );

    //SG_ASSERT(pdbc->update.b_in_update);

    if (pvh_changesets)
    {
        int rc = 0;

        SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_changesets, &count_changesets)  );
        SG_ERR_CHECK(  SG_dbndx_update__begin(pCtx, pdbc, count_changesets)  );

        SG_ERR_CHECK(  sg_sqlite__prepare(pCtx, pdbc->psql, &pStmt_schema, "SELECT json FROM cur_schema")  );
        if ((rc = sqlite3_step(pStmt_schema)) == SQLITE_ROW)
        {
            const char* psz_json = (const char*) sqlite3_column_text(pStmt_schema,0);

            SG_ERR_CHECK(  SG_VHASH__ALLOC__FROM_JSON__SZ(pCtx, &pvh_schema, psz_json)  );
        }
        else if (SQLITE_DONE != rc)
        {
            SG_ERR_THROW(SG_ERR_SQLITE(rc));
        }
        SG_ERR_CHECK(  sg_sqlite__nullfinalize(pCtx, &pStmt_schema)  );

        for (i_changeset=0; i_changeset<count_changesets; i_changeset++)
        {
            const char* psz_csid = NULL;
            SG_vhash* pvh_changes = NULL;
            SG_uint32 count_parents = 0;
            SG_int32 generation = 0;

            SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvh_changesets, i_changeset, &psz_csid, NULL)  );

            SG_ERR_CHECK(  SG_changeset__load_from_repo(pCtx, pdbc->pRepo, psz_csid, &pcs)  );
            SG_ERR_CHECK(  SG_changeset__get_generation(pCtx, pcs, &generation)  );
            SG_ERR_CHECK(  SG_changeset__db__get_changes(pCtx, pcs, &pvh_changes)  );

            if (pvh_changes)
            {
                // TODO fix this to eliminate dupes
                SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_changes, &count_parents)  );

                if (count_parents > 0)
                {
                    const char* psz_csid_parent = NULL;
                    SG_vhash* pvh_changes_one_parent = NULL;
                    SG_vhash* pvh_record_adds = NULL;
                    SG_uint32 count_adds = 0;

                    SG_ERR_CHECK(  SG_vhash__get_nth_pair__vhash(pCtx, pvh_changes, 0, &psz_csid_parent, &pvh_changes_one_parent)  );

                    SG_ERR_CHECK(  SG_vhash__check__vhash(pCtx, pvh_changes_one_parent, "add", &pvh_record_adds)  );
                    if (pvh_record_adds)
                    {
                        SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_record_adds, &count_adds)  );
                        if (count_adds)
                        {
                            SG_ERR_CHECK(  SG_vhash__alloc__copy(pCtx, &pvh_new, pvh_record_adds)  );
                        }
                    }
                }

                if (pvh_new)
                {
                    SG_uint32 i_parent = 0;

                    for (i_parent=0; i_parent<count_parents; i_parent++)
                    {
                        const char* psz_csid_parent = NULL;
                        SG_vhash* pvh_changes_one_parent = NULL;
                        SG_vhash* pvh_record_adds = NULL;
                        SG_uint32 count_adds = 0;

                        SG_ERR_CHECK(  SG_vhash__get_nth_pair__vhash(pCtx, pvh_changes, i_parent, &psz_csid_parent, &pvh_changes_one_parent)  );

                        SG_ERR_CHECK(  SG_vhash__check__vhash(pCtx, pvh_changes_one_parent, "add", &pvh_record_adds)  );
                        if (pvh_record_adds)
                        {
                            SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_record_adds, &count_adds)  );
                            if (count_adds)
                            {
                                SG_uint32 count_new = 0;

                                SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_new, &count_new)  );
                                if (count_new)
                                {
                                    SG_uint32 i = 0;

                                    SG_ERR_CHECK(  SG_vhash__alloc(pCtx, &pvh_intersection)  );
                                    for (i=0; i<count_new; i++)
                                    {
                                        const char* psz_hid_rec = NULL;
                                        SG_bool b_has = SG_FALSE;

                                        SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvh_new, i, &psz_hid_rec, NULL)  );
                                        SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh_record_adds, psz_hid_rec, &b_has)  );
                                        if (b_has)
                                        {
                                            SG_ERR_CHECK(  SG_vhash__add__null(pCtx, pvh_intersection, psz_hid_rec)  );
                                        }
                                    }

                                    SG_VHASH_NULLFREE(pCtx, pvh_new);
                                    pvh_new = pvh_intersection;
                                    pvh_intersection = NULL;

                                    SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_new, &count_new)  );
                                    if (0 == count_new)
                                    {
                                        SG_VHASH_NULLFREE(pCtx, pvh_new);
                                    }
                                }
                            }
                        }
                    }
                }

                if (pvh_new)
                {
                    SG_uint32 count_new = 0;

                    SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_new, &count_new)  );
                    if (count_new)
                    {
                        SG_uint32 i;

                        for (i=0; i<count_new; i++)
                        {
                            const char* psz_hid_rec = NULL;

                            SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvh_new, i, &psz_hid_rec, NULL)  );
                            SG_ERR_CHECK(  SG_dbrecord__load_from_repo(pCtx, pdbc->pRepo, psz_hid_rec, &pRec)  );
                            SG_ERR_CHECK(  sg_dbndx__add_record_to_records_table(pCtx, pdbc, pRec, pvh_schema)  );
                            SG_ERR_CHECK(  sg_dbndx__add_record_to_full_text_search(pCtx, pdbc, pRec, pvh_schema)  );

                            SG_ERR_CHECK(  sg_dbndx__add_record_to_history_table(pCtx, pdbc, pRec, psz_csid, generation)  );
                            SG_DBRECORD_NULLFREE(pCtx, pRec);
                        }
                    }

                    SG_VHASH_NULLFREE(pCtx, pvh_new);
                }
            }

            SG_CHANGESET_NULLFREE(pCtx, pcs);

            pdbc->update.count_so_far++;
        }

        SG_ERR_CHECK(  SG_dbndx_update__end(pCtx, pdbc)  );
    }

    if (pva_audits)
    {
        SG_ERR_CHECK(  SG_dbndx_update__do_audits(pCtx, pdbc, pva_audits)  );
    }

fail:
    SG_CHANGESET_NULLFREE(pCtx, pcs);
    if (pdbc->update.b_in_update)
    {
        SG_ERR_IGNORE(  SG_dbndx_update__abort(pCtx, pdbc)  );
    }

    if (pStmt_schema)
    {
        SG_ERR_IGNORE(  sg_sqlite__nullfinalize(pCtx, &pStmt_schema)  );
    }
    if (pStmt)
    {
        SG_ERR_IGNORE(  sg_sqlite__nullfinalize(pCtx, &pStmt)  );
    }

    SG_DBRECORD_NULLFREE(pCtx, pRec);
    SG_VHASH_NULLFREE(pCtx, pvh_schema);
    SG_VHASH_NULLFREE(pCtx, pvh_new);
    SG_VHASH_NULLFREE(pCtx, pvh_intersection);
}

