/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
 * This module defines a "filetool" object and associated functionality.
 *
 * A filetool is a specification of a program/function that operates on one or more files.
 * Some tools are "built in" or "internal", which is to say they're compiled into the code.
 * Other tools are "external", meaning they execute an external program.
 *
 * This module is responsible for:
 * - Classifying files by type (such as text, binary, etc).
 * - The list of file types/classes and how types are defined in localsettings/config.
 * - The set of known tool types.
 * - The list of external tools and how they are defined in localsettings/config.
 * - Invoking tools using a set of arguments.
 *
 * There is currently no code for creating/manipulating/managing tools or file types.
 * It's expected that all of that work will occur via localsettings/config.
 * This module just finds/loads that information and puts it to use.
 *
 * In the future, it is hoped that functionality involving file types and
 * classification can be factored out into a seperate "filetype" module.
 *
 * This module is generally used as follows:
 * - Call get_class to determine the type/class of the file(s) you're working with.
 * - Call find_tool to allocate a tool for your needs.
 *   - To include some of your own internal tools in the search, implement and pass a "plugin" callback
 * - Build a vhash of arguments to pass to the tool.
 *   - The argument keys to use depends on the tool type.
 *   - See the types enumeration for more information.
 * - Call invoke_tool to run the tool.
 */

#ifndef H_SG_FILETOOL_PROTOTYPES_H
#define H_SG_FILETOOL_PROTOTYPES_H

BEGIN_EXTERN_C;

/**
 * Determines the class of a file (i.e. text, binary, etc).
 */
void SG_filetool__get_class(
	SG_context*        pCtx,      //< [in] [out] Error and context info.
	const char*        szPath,    //< [in] The repo path or disk path of the file to get the class of.
	                              //<      If NULL, pDiskPath will be used.
	                              //<      This is passed separately because sometimes the actual file on disk will have a mangled name.
	const SG_pathname* pDiskPath, //< [in] If pattern-matching on szPath fails, we'll use this path to decide based on file contents.
	SG_repo*           pRepo,     //< [in] Repo to consider file classes from.
	                              //<      If NULL, then repo-specific file classes will not be considered.
	SG_string**        ppClass    //< [out] The class name of the given file/path.
	                              //<       Never returns NULL, returns SG_FILETOOL__CLASS__DEFAULT if no appropriate class was found.
	                              //<       Caller owns this.
	);

/**
 * Finds the most appropriate tool for a given set of circumstances:
 * - Type of tool needed.
 * - Context the tool is being used in (if the tool type uses contexts).
 * - Class of file the tool will be used on.
 */
void SG_filetool__find_tool(
	SG_context*         pCtx,      //< [in] [out] Error and context info.
	const char*         szType,    //< [in] The type of file tool to load.
	const char*         szContext, //< [in] The type-specific context to find a tool for.
	                               //<      Can be NULL to ignore context-specific tools (or if the type doesn't use contexts).
	const char*         szClass,   //< [in] The class of file to load a tool for.
	SG_repo*            pRepo,     //< [in] The repo to look up tools in.
	                               //<      If NULL, then only built-in and machine-wide tools will be considered.
	SG_string**         ppName     //< [out] The name of the found tool, or NULL if none was found.
	                               //<       Caller owns this.
	);

/**
 * Loads a file tool by its name.
 */
void SG_filetool__get_tool(
	SG_context*         pCtx,      //< [in] [out] Error and context info.
	const char*         szType,    //< [in] The type of file tool to load.
	const char*         szContext, //< [in] The type-specific context to load the tool for.
	                               //<      Can be NULL if no specific context is being used.
	const char*         szName,    //< [in] The name of the tool to retrieve.
	SG_filetool__plugin fPlugin,   //< [in] Callback to use to find invokers for built-in tools implemented by the caller.
	                               //<      Can be NULL if no built-in tools are provided by the caller.
	SG_repo*            pRepo,     //< [in] The repo to look up tools in.
	                               //<      If NULL, then only built-in and machine-wide tools will be considered.
	SG_filetool**       ppTool     //< [out] The loaded tool, or NULL if none was found.
	                               //<       Caller owns this.
	);

/**
 * Frees a filetool.
 */
void SG_filetool__free(
	SG_context*  pCtx, //< [in] [out] Error and context info.
	SG_filetool* pTool //< The tool to free.
	);

/**
 * Frees a filetool outside of the current error level.
 * Also NULLs the pointer after freeing.
 */
#define SG_FILETOOL_NULLFREE(pCtx, pTool)      \
	SG_STATEMENT(                              \
		SG_context__push_level(pCtx);          \
		SG_filetool__free(pCtx, pTool);        \
		SG_ASSERT(!SG_context__has_err(pCtx)); \
		SG_context__pop_level(pCtx);           \
		pTool = NULL;                          \
	)

/**
 * Gets the name of a tool.
 */
void SG_filetool__get_name(
	SG_context*        pCtx,  //< [in] [out] Error and context info.
	const SG_filetool* pTool, //< [in] Tool to get the name of.
	const char**       ppName //< [out] Name of the given tool.
	                          //<       Caller does NOT own this.
	);

/**
 * Gets the executable path of a tool.
 */
void SG_filetool__get_executable(
	SG_context*        pCtx,        //< [in] [out] Error and context info.
	const SG_filetool* pTool,       //< [in] Tool to get the executable path of.
	const char**       ppExecutable //< [out] Executable path of the given tool.
	                                //<       Set to NULL if the tool is built-in.
	                                //<       Caller does NOT own this.
	);

/**
 * Invokes a tool.
 */
void SG_filetool__invoke(
	SG_context*        pCtx,         //< [in] [out] Error and context info.
	const SG_filetool* pTool,        //< [in] The tool to invoke.
	const SG_vhash*    pTokenValues, //< [in] Values for placeholder tokens in the tool's arguments.
	                                 //<      Keys are the names of the tokens to replace with the corresponding value.
	                                 //<      May be NULL if there are no values to substitute.
	SG_int32*          pExitCode     //< [out] The exit code returned by the tool.
	                                 //<       May be NULL if the exit code is unneeded.
	);

/**
 * An internal filetool that does nothing and always succeeds.
 */
void SG_filetool__invoke__null(
	SG_context*        pCtx,
	const SG_filetool* pTool,
	const SG_vhash*    pTokenValues,
	SG_int32*          pExitCode
	);

/**
 * An SG_filetool__invoker that invokes an external program.
 *
 * This function is mainly useful within custom external invokers,
 * when you just want to do something before/after the program is run.
 * Just do the extra thing yourself and call this function to handle executing the program.
 */
void SG_filetool__invoke__external(
	SG_context*        pCtx,
	const SG_filetool* pTool,
	const SG_vhash*    pTokenValues,
	SG_int32*          pExitCode
	);

/**
 * Function to simply fetch the path where SourceGear Diffmerge exists on the machine.
 */
void SG_filetool__get_builtin_diffmerge_exe(
	SG_context * pCtx,          //< [in] [out] Error and context info.
	SG_pathname ** ppPathCmd    //< [out] Path to diffmerge executable. Caller owns it.
	);

END_EXTERN_C;

#endif
