/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg_pendingtree_prototypes.h
 *
 * @details The Pending Tree is an object for managing the pending
 * changes to a version control tree.  It could also be called the
 * pending changeset.
 *
 * This header file describes the API for a pending tree.  The
 * implementation of this class is actually in sg_closet.c, since the
 * pending tree is stored in the closet.
 *
 * The pending tree is initialized with a root directory HID obtained
 * from a dag node.  After that, you can call various methods to make
 * changes to the tree.  This class demand-loads any treenodes
 * necessary to keep an accurate list of everything that is dirty.
 *
 * When the time comes, you can tell a pending tree to commit itself
 * to the repo.
 *
 * A pending tree is specific to a given working directory.  Its main
 * job is to tell us what the repo would look like if all the changes
 * in the working directory were committed.  This is the main reason
 * we call this a pending tree instead of a pending changeset.  It is
 * more than just a list of changes.  You can ask it questions about
 * any part the tree, not just the changed parts.
 *
 * The pending tree handles "bubble up" automatically.
 *
 * The pending tree handles the "two level lookup" needed to make sure
 * that operations on a working directory happen to the right objects
 * even after renames and moves may have happened.
 *
 * The pending tree API offers functions that allow individual
 * operations to be reverted.
 *
 * As noted above, the pending tree API can commit itself to the repo,
 * but it also can perform this operation at finer granularity,
 * committing only a portion of its pending changes instead of the
 * entire pending changeset.
 *
 * We have spoken in the past of managing the pending changeset as two
 * lists (a list of dirty tree nodes and a list of new file blobs).
 * The pending tree conceptually encapsulates and manages both of
 * these lists.
 *
 */

//////////////////////////////////////////////////////////////////

#ifndef H_SG_PENDINGTREE_PROTOTYPES_H
#define H_SG_PENDINGTREE_PROTOTYPES_H

BEGIN_EXTERN_C;

/**
 * This function creates the in-memory pending changeset object for
 * the given working directory.  If a pending changeset already exists
 * (in the drawer) for this working directory, it will be loaded into
 * the in-memory object.  Otherwise, a new one is created, empty.
 *
 */
void SG_pendingtree__alloc(
	SG_context* pCtx,
	const SG_pathname* pPathWorkingDirectory,
	SG_pendingtree** ppThis
	);

/**
 * This function creates the in-memory pending changeset object for
 * the current working directory.  If a pending changeset already exists
 * (in the drawer) for this working directory, it will be loaded into
 * the in-memory object.  Otherwise, a new one is created, empty.
 *
 */
void SG_pendingtree__alloc_from_cwd(
	SG_context* pCtx,
	SG_pendingtree** ppThis
	);

#if defined(DEBUG)
#define SG_PENDINGTREE__ALLOC(pCtx,pPathWorkingDirectory,ppThis)	SG_STATEMENT(	SG_pendingtree * _pNew = NULL;												\
																					SG_pendingtree__alloc(pCtx,pPathWorkingDirectory,&_pNew); 					\
																					_sg_mem__set_caller_data(_pNew,__FILE__,__LINE__,"SG_pendingtree");			\
																					*(ppThis) = _pNew;															)
#define SG_PENDINGTREE__ALLOC_FROM_CWD(pCtx,ppThis)					SG_STATEMENT(	SG_pendingtree * _pNew = NULL; 												\
																					SG_pendingtree__alloc_from_cwd(pCtx,&_pNew); 								\
																					_sg_mem__set_caller_data(_pNew,__FILE__,__LINE__,"SG_pendingtree");			\
																					*(ppThis) = _pNew;															)
#else
#define SG_PENDINGTREE__ALLOC(pCtx,pPathWorkingDirectory,ppThis)	SG_pendingtree__alloc(pCtx,pPathWorkingDirectory,ppThis)
#define SG_PENDINGTREE__ALLOC_FROM_CWD(pCtx,ppThis)					SG_pendingtree__alloc_from_cwd(pCtx,ppThis)
#endif

void SG_pendingtree__free(SG_context* pCtx, SG_pendingtree* pThis);

//////////////////////////////////////////////////////////////////

/**
 * Re-read the wd.json and rebuild the PTNODEs and etc.
 * This is ***ONLY*** necessary because I needed to do 2 operations
 * while holding the file lock when doing a MERGE.
 * [1] first i need to do a STATUS (aka scan_dir) to see if there is
 *     any dirt in the WD.  The STATUS modifies the in-memory PTNODES/
 *     pendingtree.  But normally, a STATUS does not get saved back
 *     to disk.
 * [2] Then the MERGE wants to combine the CSETS (using just Treenodes
 *     and treenode-entries) and then update the WD and the PTNODES
 *     and SAVE the pendingtree back to wd.json.
 * Because [1] calls scan_dir with "addremove=1, markimplicit=1" (and
 * causes all of the LOST/FOUND items to be added to the pendingtree
 * (and because the IMPLICIT bit is in temp_flags rather than saved_flags)
 * the SAVE does more than it thinks it does.
 *
 * This routine can be used between [1] and [2] to discard the dirt
 * created by [1] and give [2] a fresh view -- all without releasing
 * the VFILE lock.
 *
 * TODO 2010/07/02 Refactor everything that made the above comment
 * TODO            necessary.
 *
 */
void SG_pendingtree__reload_pendingtree_from_vfile(SG_context * pCtx,
												   SG_pendingtree * pPendingTree);

//////////////////////////////////////////////////////////////////

/**
 * Scan is a bad idea in general.  It's useful for testing, and we
 * can probably end up with a version which is useful for some
 * end-user scenarios.  But it's not a very scalable concept.
 *
 * One of the reasons that Git is considered to be so fast is that it
 * doesn't scan by default.  All files are explicitly added to the
 * pending changeset.
 *
 * TODO 2010/06/05 The above comment is very old.  Review and see if
 * TODO            we still think it is bad thing.
 *
 * This is primarily an external API for scan and used by the CLC.
 * It takes care of loading the IGNORES list.  It SAVES the result
 * of the scan to the wd.json.
 */
void SG_pendingtree__scan(SG_context* pCtx,
						  SG_pendingtree* pPendingTree,
						  SG_bool bRecursive,
						  SG_bool bIncludeLostItems,
						  SG_bool bIncludeFoundItems,
						  const SG_file_spec* pFilespec);

/**
 * just scan, don't unload non-dirty ptnodes, don't save.
 */
void SG_pendingtree__scan__no_unload__no_save(SG_context* pCtx,
											  SG_pendingtree* pPendingTree,
											  SG_bool bRecursive,
											  SG_bool bIncludeLostItems,
											  SG_bool bIncludeFoundItems,
											  const SG_file_spec* pFilespec);

void SG_pendingtree__get_working_directory_top(
	SG_context* pCtx,
	const SG_pendingtree* pPendingTree,
	SG_pathname** ppPath);

/**
 * borrow the pathname to the "@" directory.
 */
void SG_pendingtree__get_working_directory_top__ref(
	SG_context* pCtx,
	const SG_pendingtree* pPendingTree,
	const SG_pathname** ppPath);

/**
 * This is an external API to scan the directory and automatically
 * ADD newly found items and REMOVE missing items.  It SAVES the
 * result of the scan to the wd.json.
 */
void SG_pendingtree__addremove(SG_context* pCtx,
							   SG_pendingtree* pPendingTree,
							   SG_uint32 count_items, const char** paszItems,
							   SG_bool bRecursive,
							   const SG_file_spec* pFilespec,
							   SG_bool bTest);

/**
 * The list of INCLUDES/EXCLUDES apply to the things to be actually committed
 * (and not to any internal scans that we might need to make to ensure that we
 * are up to date before or afterwards).
 * Commit DOES NOT take an IGNORES list.
 */
void SG_pendingtree__commit(SG_context* pCtx,
							SG_pendingtree* pPendingTree,
							const SG_audit* pq,
							SG_uint32 count_items, const char** paszItems,
							SG_bool bRecursive,
							const SG_file_spec* pFilespec,
							const char* const* paszAssocs, SG_uint32 count_assocs,
							const SG_stringarray * psaStamps,
							SG_bool bNoSubmodules,
							const char * psz_comment,
							SG_dagnode** ppdn);

void SG_pendingtree__precommit_check(SG_context* pCtx,
							SG_pendingtree* pPendingTree,
							const SG_audit* pq,
							SG_uint32 count_items, const char** paszItems,
							SG_bool bRecursive,
							const SG_file_spec* pFilespec,
							const char* const* paszAssocs, SG_uint32 count_assocs,
							SG_varray **descs);


void SG_pendingtree__get_repo_path(
	SG_context* pCtx,
	SG_pendingtree* pPendingTree,
	const char* pszgid,
	SG_string** ppResult
	);

void SG_pendingtree__find_gids_for_locking__must_be_files(
    SG_context* pCtx,
    SG_pendingtree* pPendingTree,
    const SG_pathname* pPathRelativeTo,
    SG_uint32 count_items, 
    const char** paszItems,
    SG_vhash** ppvh_gids
    );

void SG_pendingtree__revert(SG_context* pCtx,
							SG_pendingtree* pPendingTree,
							SG_pendingtree_action_log_enum eActionLog,
							SG_uint32 count_items, const char** paszItems,
							SG_bool bRecursive,
							SG_bool bNoSubmodules,
							SG_bool bNoBackups,
							const SG_vector * pvec_use_submodules,
							const SG_file_spec* pFilespec,
							SG_bool bPrecomputeOnly,
							SG_varray * pvaMissingDetails);

void SG_pendingtree__remove(SG_context* pCtx,
							SG_pendingtree* pPendingTree,
							SG_uint32 count_items, const char** paszItems,
							const SG_file_spec* pFilespec,
							SG_bool bTest, SG_bool bForce, SG_bool bNoBackup);

void SG_pendingtree__remove_dont_save_pendingtree(SG_context* pCtx,
												  SG_pendingtree* pPendingTree,
												  SG_uint32 count_items, const char** paszItems,
												  const SG_file_spec* pFilespec,
												  SG_bool bTest, SG_bool bForce, SG_bool bNoBackup);

/**
 * See note above the function source to see why you probably
 * don't want to call this.
 */
void SG_pendingtree__remove_parked__dont_save_pendingtree(SG_context * pCtx,
														  SG_pendingtree * pPendingTree,
														  const SG_pathname * pPathParked,
														  const char * pszGidEntry);

/**
 * See note above the function source to see why you probably
 * don't want to call this.
 */
void SG_pendingtree__hack__unadd_dont_save_pendingtree(SG_context * pCtx,
													   SG_pendingtree * pPendingTree,
													   const char * pszGid,
													   const SG_pathname * pPath_Input);

void SG_pendingtree__rename(
	SG_context* pCtx,
	SG_pendingtree* pPendingTree,
	const SG_pathname* pPathLocal,
	const char* pszNewName,
	SG_bool bIgnoreErrorNotFoundOnDisk
	);

void SG_pendingtree__rename_dont_save_pendingtree(
	SG_context* pCtx,
	SG_pendingtree* pPendingTree,
	const SG_pathname* pPathLocal,
	const char* pszNewName,
	SG_bool bIgnoreErrorNotFoundOnDisk
	);

void SG_pendingtree__move(
	SG_context* pCtx,
	SG_pendingtree* pPendingTree,
	const SG_pathname** pArrayPathsToMove,
	SG_uint32 nCountMoves,
	const SG_pathname* pPathMoveTo
	);

void SG_pendingtree__move_dont_save_pendingtree(
	SG_context* pCtx,
	SG_pendingtree* pPendingTree,
	const SG_pathname* pPathMoveMe,
	const SG_pathname* pPathMoveTo,
	SG_bool bAllowDestinationDirInDrawer	// only MERGE gets to set this to TRUE
	);

void SG_pendingtree__add(SG_context* pCtx,
						 SG_pendingtree* pPendingTree,
						 SG_uint32 count_items, const char** paszItems,
						 SG_bool bRecursive,
						 const SG_file_spec* pFilespec,
						 SG_bool bTest);

void SG_pendingtree__add_dont_save_pendingtree(SG_context* pCtx,
											   SG_pendingtree* pPendingTree,
											   SG_uint32 count_items, const char** paszItems,
											   SG_bool bRecursive,
											   const SG_file_spec* pFilespec);

/**
 * Add an entry to the pendingtree using the given GID.
 * This is needed for MERGE.
 */
void SG_pendingtree__add_dont_save_pendingtree__with_gid(
	SG_context * pCtx,
	SG_pendingtree * pPendingTree,
	const char * pszGid,
	const SG_pathname * pPath);

/**
 * SG_pendingtree__update_baseline() is used primarily to update the contents
 * of the Working Directory to be relative to a different changeset -- almost
 * as if a fresh WD were created using the named changeset.  I say almost because
 * we do allow for dirt in the current WD and try to "move it forward" if possible.
 * This works best when your local dirt is completely independent of any changes
 * that have been made to a descendant changeset of your current baseline.
 *
 * When there is dirt in the WD, we do place restrictions on the target changeset
 * (such as being a descendant of the current baselin) so that we have
 * some hope of the dirt being carried forward successfully.  (If you want to
 * warp to an arbitrary changeset, do a COMMIT or REVERT first and then warp
 * with a clean WD.)
 *
 * The caller should have already used SG_vv_utils__rev_arg_to_csid__wd() to
 * convert hid-prefixes and other --rev/--tag/--branch/default conveniences
 * into an actual full target HID (and optionally the name of named branch
 * to be tied to after the update).  If the given target changeset (HID) is
 * NULL, we look for a single/unique head/tip that is a descendant of the
 * current baseline.  This fall-back is a last resort because __rev_arg_to_csid__wd
 * has already tried to find the unique head of the current named branch and etc.
 * If we cannot determine a unique/well-defined target HID, we fail.
 *
 * We try to preserve the WD dirt as best as possible during the update.  This
 * means performing individual merges for files that are dirty in the WD and
 * that were modified in any of the changesets between the current baseline and
 * the target changeset.  As always, this can produce conflicts.
 *
 * pvaMissingDetails is passed in to collect problem information for
 * nested updates/checkouts for submodules.  This will accumulate a
 * series of vhashes that can be displayed to the user to help them
 * do CLONES/PULLS before retrying the UPDATE.
 */
void SG_pendingtree__update_baseline__precompute(SG_context* pCtx,
												 SG_pendingtree* pPendingTree,
												 const char* pszHIDTargetChangeset,
												 const SG_vector * pvec_use_submodules,
												 SG_varray * pvaMissingDetails);

void SG_pendingtree__update_baseline__apply(SG_context* pCtx,
											SG_pendingtree* pPendingTree,
											SG_pendingtree_action_log_enum eActionLog,
											const char* psz_branch_name);

void SG_pendingtree_update_data__free(SG_context * pCtx,
									  SG_pendingtree_update_data * pud);

//////////////////////////////////////////////////////////////////

void SG_pendingtree__get_repo(
        SG_context* pCtx,
		SG_pendingtree* pPendingTree,
        SG_repo** ppRepo
        );

/**
 * Return a VArray containing the HIDs of the parent changeset(s)
 * for the Working Directory.  Normally, there is only one parent
 * and any changes (dirt) in the WD is considered to be relative
 * to it.  This has historically been called the "baseline".
 *
 * However, when we have an uncommitted merge, we will have
 * multiple parents.  Parents are generally considered to be
 * equivalent and they may be in random order in the pendingtree.
 * When the merge is committed, the parents will be sorted for HID
 * computation reasons.
 *
 * However, for user consistency, Parent[0] in the uncommitted merge
 * will be the baseline before the merge (so that any DIFF commands
 * will show the user all of the changes relative to their last baseline).
 *
 * The returned VArray is an allocated copy; you own it and must free it.
 * It remains valid after the pendingtree is unlocked/freed.
 */
void SG_pendingtree__get_wd_parents(SG_context* pCtx,
									SG_pendingtree * pPendingTree,
									SG_varray ** ppvaParents);

/**
 * Like __get_wd_parents() but without the alloc.  The VArray is
 * within the pendingtree and is only valid as long as the pendingtree
 * is valid.
 */
void SG_pendingtree__get_wd_parents__ref(SG_context* pCtx,
										 SG_pendingtree * pPendingTree,
										 const SG_varray ** ppvaParents);

/**
 * Set the changeset parent(s) (aka baseline) of the Working Directory.
 * Normally this is only updated after a COMMIT.  It may also be
 * set during an UPDATE, MERGE, or REVERT.
 *
 * We STEAL YOUR VARRAY and NULL your pointer to it.
 *
 * You must call __save at some point to get the new set of parent(s)
 * written to disk (along with the rest of the pendingtree stuff)
 */
void SG_pendingtree__set_wd_parents(SG_context * pCtx,
									SG_pendingtree * pPendingTree,
									SG_varray ** ppva_new_wd_parents);

/**
 * Convenience wrapper for __set_wd_parents() when you only have 1 parent.
 *
 * As with it, you must still call __save at some point.
 */
void SG_pendingtree__set_single_wd_parent(SG_context * pCtx,
										  SG_pendingtree * pPendingTree,
										  const char * psz_hid_parent);

/**
 * Store a varray[vhash] os unresolved issues with the pendingtree.
 * These are problems that MERGE had and could not automatically
 * resolve.  We put them in the wd.json so that the user can later
 * use the RESOLVE command and addess each issue.  We also use this
 * to prevent them from doing a COMMIT with outstanding issues.
 *
 * We STEAL YOUR VARRAY and NULL your pointer to it.
 *
 * You must call __save at some point to get the new set of issues
 * written to disk (along with the rest of the pendingtree stuff).
 */
void SG_pendingtree__set_wd_issues(SG_context * pCtx,
								   SG_pendingtree * pPendingTree,
								   SG_varray ** ppva_new_wd_issues);

/**
 * Clear the set of unresolved issues from the pendingtree.  This
 * removes the varray[vhash] from the pendingtree (and after a __save()
 * from the wd.json).
 *
 * You must call __save at some point to get the new set of issues
 * written to disk (along with the rest of the pendingtree stuff).
 */
void SG_pendingtree__clear_wd_issues(SG_context * pCtx,
									 SG_pendingtree * pPendingTree);

/**
 * Return a VARRAY containing the UNRESOLVED ISSUES caused by a MERGE.
 * You DO NOT own this.  This is the live version within the pendingtree
 * and only valid as long as the pendingtree is in memory.
 */
void SG_pendingtree__get_wd_issues__ref(SG_context * pCtx,
										SG_pendingtree * pPendingTree,
										SG_bool * pbHaveIssues,
										const SG_varray ** ppva_wd_issues);

/**
 * Store a new set of choices in the pendingtree.
 *
 * We STEAL YOUR VARRAY and NULL your pointer to it.
 *
 * You must call __save at some point to get the new set of choices
 * written to disk (along with the rest of the pendingtree stuff).
 *
 * Note: Don't set this directly.  Create and manipulate the choices
 *       by allocating an SG_resolve instance with SG_resolve__alloc.
 *       Once configured, use SG_resolve__save (which calls this function)
 *       to store the choices into the pendingtree.
 */
void SG_pendingtree__set_wd_choices(
	SG_context*     pCtx,         //< [in] [out] Error and context info.
	SG_pendingtree* pPendingTree, //< [in] The pendingtree to store choices in.
	SG_vhash**      ppChoices     //< [in] [out] The choices to store.
	                              //<            If NULL or empty, any existing choices are cleared.
	                              //<            Set to NULL after successful call.
	);

/**
 * Return a VARRAY containing the CHOICES created by RESOLVE.
 * You DO NOT own this.  This is the live version within the pendingtree
 * and only valid as long as the pendingtree is in memory.
 *
 * Note: Don't access this directly.  Instead allocate an SG_resolve
 *       with SG_resolve__alloc and access/manipulate the data through it.
 *       When finished, use SG_resolve__save to store the data back into
 *       the pendingtree.
 */
void SG_pendingtree__get_wd_choices__ref(
	SG_context*      pCtx,         //< [in] [out] Error and context info.
	SG_pendingtree*  pPendingTree, //< [in] The pendingtree to load choices from.
	SG_bool*         pHaveChoices, //< [out] Whether or not any choices exist in the pendingtree.
	const SG_vhash** ppChoices     //< [out] The set of loaded choices.
	                               //<       Set to NULL if no choices exist.
	);

/**
 * Set the resolved/unresolved STATUS on an individual ISSUE in the
 * array of ISSUES.
 *
 * THIS ONLY SETS THE STATUS OF THE ISSUE; IT DOES NOT ALTER THE WD
 * IN ANY WAY.  This is to allow the user to MANUALLY resolve/unresolve
 * an issue after MANUALLY making whatever changes were needed to the WD.
 *
 * This update the live version of the VARRAY within the pendingtree.
 *
 * You must call __save at some point to get the new set of issues
 * written to disk (along with the rest of the pendingtree stuff).
 */
void SG_pendingtree__set_wd_issue_status__dont_save_pendingtree(SG_context * pCtx,
																SG_pendingtree * pPendingTree,
																const SG_vhash * pvhIssue,
																SG_pendingtree_wd_issue_status status);

/**
 * Set the STATUS on an individual ISSUE and SAVE the pendingtree to disk.
 */
void SG_pendingtree__set_wd_issue_status(SG_context * pCtx,
										 SG_pendingtree * pPendingTree,
										 const SG_vhash * pvhIssue,
										 SG_pendingtree_wd_issue_status status);

/**
 * Set the RESOLVE MERGETOOL fields on the ISSUE.
 */
void SG_pendingtree__set_wd_issue_resolve_mergetool_status(SG_context * pCtx,
														   SG_pendingtree * pPendingTree,
														   const SG_vhash * pvhIssue,
														   const char * pszResolveMergeToolName,
														   SG_int32 statusResolveMergeTool);

/**
 * Indicate whether RESOLVE has dealt with the structural part of the ISSUE.
 */
void SG_pendingtree__set_wd_issue_resolve_structural_status(SG_context * pCtx,
															SG_pendingtree * pPendingTree,
															const SG_vhash * pvhIssue,
															SG_bool bValue);

/**
 * When we have a pending MERGE, record the MERGE-TEMP-DIR in the wd.json
 * so that RESOLVE can reuse the directory when doing manual merges using
 * a different mergetool.
 *
 * Pass NULL to clear the field in the wd.json.
 *
 * You must call __save at some point to get this value and wd.json
 * written to disk.
 *
 * This value should be cleared by COMMIT or a REVERT --ALL.
 *
 */
void SG_pendingtree__set_mrg_temp_dir(SG_context * pCtx,
									  SG_pendingtree * pPendingTree,
									  const SG_pathname * pPathMergeTempDir);

/**
 * Fetch the value of MERGE-TEMP-DIR.  This is only defined when
 * we have a pending MERGE.
 *
 * You own the returned pathname.
 */
void SG_pendingtree__get_mrg_temp_dir(SG_context * pCtx,
									  SG_pendingtree * pPendingTree,
									  SG_bool * pbHasOne,
									  SG_pathname ** ppPathMergeTempDir);

/**
 * See if there are any unresolved issues.
 */
void SG_pendingtree__count_unresolved_wd_issues(SG_context * pCtx,
												SG_pendingtree * pPendingTree,
												SG_uint32 * pNrUnresolved);

/**
 * Fetch the ISSUE associated with the given GID, if present.
 */
void SG_pendingtree__find_wd_issue_by_gid(SG_context * pCtx,
										  SG_pendingtree * pPendingTree,
										  const char * pszGidToFind,
										  SG_bool * pbFound,
										  const SG_vhash ** ppvhIssue);

/**
 * Save the in-memory pendingtree to disk.  This updates both the
 * "pending" and the "parents" parts of the "wd.json" file.
 */
void SG_pendingtree__save(
	SG_context* pCtx,
	SG_pendingtree* pPendingTree
	);

//////////////////////////////////////////////////////////////////

/**
 * TODO HACK This routine is a hack/compromise because the MERGE
 * TODO      code needs to do a diff in the middle of the computation
 * TODO      and we want to keep as much of the ptnodes in-memory
 * TODO      because stuff from the merge will likely touch clean
 * TODO      stuff.  ALSO, we need to PREVENT the diff code from
 * TODO      nulling the super-root (which is part of that
 * TODO      one-operation-and-save problem).
 * TODO
 * TODO      See SPRAWL-660.
 */
void SG_pendingtree__do_diff__keep_clean_objects(SG_context* pCtx, SG_pendingtree * pPendingTree,
														 const char * szHidArbitraryCSet,		/* optional */
														 SG_treediff2 * pTreeDiff_cache_owner,	/* optional */
														 SG_treediff2 ** ppTreeDiff);

/**
 * TODO HACK HACK This routine builds upon the previous hack and
 * TODO           reloads the VFILE/VHASH after the diff has been
 * TODO           computed because the caller doesn't want the
 * TODO           side-effects of the diff dirtying up the PTNODES
 * TODO           but can't just throw-away the pendingtree and
 * TODO           create a new one because of the file locking.
 */
void SG_pendingtree__do_diff__keep_clean_objects__reload(SG_context* pCtx, SG_pendingtree * pPendingTree,
														 const char * szHidArbitraryCSet,		/* optional */
														 SG_treediff2 * pTreeDiff_cache_owner,	/* optional */
														 SG_treediff2 ** ppTreeDiff);

//////////////////////////////////////////////////////////////////

void SG_pendingtree__diff_or_status(
		SG_context * pCtx,
		SG_pendingtree * pTree,
		const char * psz_cset_1,
		const char * psz_cset_2,
		SG_treediff2** ppTreeDiff);

void SG_pendingtree__get_gid_from_local_path(
		SG_context* pCtx,
		SG_pendingtree* pTree,
		const SG_pathname* pLocalPathPathname,
		char** ppszgid);

void SG_pendingtree__get_gids_for_paths(SG_context * pCtx,
										SG_pendingtree * pPendingTree,
										SG_uint32 count_args,
										const char ** paszArgs,
										SG_stringarray* pReturnGIDs,
										SG_bool * bRootIsOneOfThem);


//////////////////////////////////////////////////////////////////

/**
 * This version caused the created treediff to share the treenode cache
 * with the given treediff.  if the cache_owner is null, it behaves like
 * the other version.
 */
void SG_pendingtree__diff__shared(SG_context* pCtx, SG_pendingtree* pPendingTree,
									  const char * szHidArbitraryCSet,		/* optional */
									  SG_treediff2 * pTreeDiff_cache_owner,	/* optional */
									  SG_treediff2 ** ppTreeDiff
									 );

void SG_pendingtree__history(SG_context * pCtx, SG_pendingtree * pPendingTree,
						 SG_uint32 count_args, const char ** paszArgs,
						 const char ** pasz_changesets,
						 SG_uint32 nCountChangesets,
						 const char* pStrUser,
						 SG_uint32 nResultLimit,
						 SG_int64 nFromDate,
						 SG_int64 nToDate,
						 SG_bool bAllLeaves,
						 SG_varray ** ppVArrayResults);



/**
 * Get the "Action Log" of the low-level working-directory operations
 * that need to be / were performed.  This is only valid for verbs
 * that have a --test or --verbose option.
 *
 * You DO NOT own the returned varray.
 *
 * The returned varray pointer is NULL when empty.
 */
void SG_pendingtree__get_action_log(SG_context * pCtx,
									SG_pendingtree * pPendingTree,
									SG_varray ** ppResult);

void SG_pendingtree__steal_action_log(SG_context * pCtx,
									  SG_pendingtree * pPendingTree,
									  SG_varray ** ppResult);

//////////////////////////////////////////////////////////////////

/**
 * Get handle to VARRAY of portability warnings issued so far.
 *
 * You CAN append new warnings to the end.  These will
 * get returned to the creator of the pendingtree as if
 * generated by pendingtree-proper.
 *
 * You DO NOT own this.
 */
void SG_pendingtree__get_warnings(SG_context* pCtx,
								  SG_pendingtree* pPendingTree,
								  SG_varray** ppResult);

void SG_pendingtree__steal_warnings(SG_context* pCtx,
									SG_pendingtree* pPendingTree,
									SG_varray** ppResult);

/**
 * Get various portability-warning settings.
 *
 * For pvaWarnings, this is a handle to VARRAY of portability
 * warnings issued so far.  You CAN append new warnings to the
 * end.  These will get returned to the creator of the
 * pendingtree as if generated by pendingtree-proper.
 * You DO NOT own this VARRAY.
 */
void SG_pendingtree__get_port_settings(SG_context * pCtx,
									   SG_pendingtree * pPendingTree,
									   SG_utf8_converter ** ppConverter,
									   SG_varray ** ppvaWarnings,
									   SG_portability_flags * pPortMask);

//////////////////////////////////////////////////////////////////

/**
 * Use GID to find entry and return repo-path.
 * You own the returned string.
 */
void SG_pendingtree__find_repo_path_by_gid(SG_context * pCtx,
										   SG_pendingtree * pPendingTree,
										   const char * pszGid,
										   SG_string ** ppStrRepoPath);

//////////////////////////////////////////////////////////////////

/**
 * Format a RESOLVED/UNRESOLVED ISSUE from a MERGE and append to
 * the given string.
 *
 * Optionally also return the current repo-path for the issue.
 * You own this and must free it.
 */
void SG_pendingtree__format_issue(SG_context * pCtx,
								  SG_pendingtree * pPendingTree,
								  const SG_vhash * pvhIssue,
								  SG_string * pStrOutput,
								  SG_string ** ppStrRepoPath_Returned);

//////////////////////////////////////////////////////////////////

/**
 * load/build a temporary pendingtree, diff it against the baseline
 * and dump it to vhash (like we write to the vfile, but also include extra info).
 */
void SG_pendingtree_debug__export(SG_context * pCtx, SG_pathname * pPathWorkingDir,
								  SG_bool bProcessIgnores,
								  SG_treediff2 ** ppTreeDiff,
								  SG_vhash ** ppvhExport);

/**
 * dump already in-memory pendingtree to vhash and write it to the console.
 */
void SG_pendingtree_debug__dump_existing_to_console(SG_context * pCtx, SG_pendingtree * pPendingTree);

/**
 * Inspect the "wd.json" file.
 */
void SG_pendingtree_debug__get_wd_dot_json_stats(SG_context * pCtx,
												 const SG_pathname * pPathWorkingDir,
												 SG_uint32 * pNrParents,
												 SG_bool * pbHavePending);


void SG_pendingtree__walk_tree(SG_context * pCtx, SG_pendingtree * pPendingTree);

void SG_pendingtree__get_timestamp_cache(SG_context * pCtx,
										 SG_pendingtree * pPendingTree,
										 SG_timestamp_cache ** ppTSC);

//////////////////////////////////////////////////////////////////

/**
 * Remember the HID of the file/symlink content at the time
 * that we altered it (in MERGE) ***IF AND ONLY IF*** we are
 * sure that the content came from the repo and/or was generated.
 *
 * If the user does a REVERT after the MERGE, we usually backup
 * any dirty files/symlinks (because that might be their only
 * version of the file/symlink and they may have spent a lot of
 * time in a manual merge or something) before we overwrite them
 * with the reverted/baseline version.
 *
 * The only problem with this is that it tends to leave a lot
 * of ~sg00~ files in the directory when it really doesn't need
 * to.
 *
 * I added a "disposable-if" HID that can (optionally) be set
 * during MERGE.  It represents the value of the HID of the
 * file/symlink that we created (added by other branch) or
 * generated (auto-merge result).  The idea is that, at the
 * time of the REVERT, if the file content matches this HID,
 * then the user HAS NOT edited the file after the MERGE and
 * we are free to delete it.
 */
void SG_pendingtree__set_disposable_hid(SG_context * pCtx,
										SG_pendingtree * pPendingTree,
										const char * pszGid,
										const char * pszHid);

/**
 * For directories which were created by MERGE (because
 * they were created in the other branch), set a "disposable"
 * bit on the ptnode so that if they later do REVERT rather
 * than a COMMIT, we can try to delete them rather than leaving
 * them around to trash up the place.  We don't need _hid
 * version because the HID on the directory is in flux.
 */
void SG_pendingtree__set_disposable_dir(SG_context * pCtx,
										SG_pendingtree * pPendingTree,
										const char * pszGid);

/**
 * MERGE/RESOLVE sometimes need to cause an otherwise clean ptnode
 * to stay in the pendingtree so that find_by_gid can get it.
 *
 * Yes, this is a hack.
 *
 */
void SG_pendingtree__hack__set_sticky_dir(SG_context * pCtx,
										  SG_pendingtree * pPendingTree,
										  const char * pszGid);

/**
 * Clear ALL of the sticky bits in the pendingtree.
 */
void SG_pendingtree__hack__clear_all_sticky_bits(SG_context * pCtx,
												 SG_pendingtree * pPendingTree);

//////////////////////////////////////////////////////////////////

END_EXTERN_C;

#endif //H_SG_PENDINGTREE_PROTOTYPES_H
