/*
Copyright 2010 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License"	);
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg_resolve_prototypes.h
 *
 * @details A module that helps the user resolve issues that merge could not solve automatically.
 *
 * Resolve works by considering each item (file, directory, symlink, submodule)
 * as a collection of state values.  The possible types of state values are
 * defined by the SG_resolve__state enumeration.  Each item has a value for each
 * of these states, and conflicts arise when merge is unable to determine the
 * correct value for one or more states on an item.
 *
 * Resolve starts by generating a set of choices.  Each choice regards a single
 * conflicting state on a single item.  A choice contains data that is useful
 * to the user when deciding what the final value for that state on that item
 * should be.  Most notably, the choice contains the various values assigned
 * to that state by changesets involved in the merge (the parent revisions, the
 * common ancestor revision, and the current working directory, usually).
 * Resolving the choice just means telling the choice which of these values
 * the item should use.
 *
 * Some choices are "mergeable", meaning that the type of state they're
 * associated with has values that can be merged, rather than being mutually
 * exclusive (this is most often the case with file contents).  In this case,
 * any already existing values (i.e. from parent revisions) can be merged
 * together to create new values using a merge tool (from SG_mergetool).
 * The newly created values can be accepted or merged further, just like any
 * of the original values, or when there is only a single leaf value remaining,
 * it is automatically accepted as the resolution.
 */

#ifndef H_SG_RESOLVE_PROTOTYPES_H
#define H_SG_RESOLVE_PROTOTYPES_H

BEGIN_EXTERN_C;

/**
 * Gets a printable name for a state.
 */
void SG_resolve__state__get_info(
	SG_context*       pCtx,         //< [in] [out] Error and context info.
	SG_resolve__state eState,       //< [in] The state to get the name of.
	const char**      ppName,       //< [out] The name of the state.
	const char**      ppDescription //< [out] A brief description of the state.
	);

/**
 * Parses a state's name back into its enumerated value.
 * Returns SG_RESOLVE__STATE__COUNT if the name is invalid.
 */
void SG_resolve__state__check_name(
	SG_context*        pCtx,   //< [in] [out] Error and context info.
	const char*        szName, //< [in] The name to parse into a state.
	SG_resolve__state* pState  //< [out] The parsed state.
	                           //<       SG_RESOLVE__STATE__COUNT if it could not be parsed.
	);

/**
 * Parses a state's name back into its enumerated value.
 * Throws SG_ERR_NOT_FOUND if the name is invalid.
 */
void SG_resolve__state__parse_name(
	SG_context*        pCtx,   //< [in] [out] Error and context info.
	const char*        szName, //< [in] The name to parse into a state.
	SG_resolve__state* pState  //< [out] The parsed state.
	);

/**
 * Allocates the resolve data structure and populates it with choices based
 * on the state of the given pendingtree.
 *
 * Note: The given pendingtree must remain valid as long as the allocated
 *       resolve data is in use.
 * Note: Any changes made to the pendingtree AFTER this allocation may render
 *       the allocated resolve data obsolete.
 */
void SG_resolve__alloc(
	SG_context*     pCtx,         //< [in] [out] Error and context info.
	SG_pendingtree* pPendingTree, //< [in] The pendingtree to generate choices from.
	SG_resolve**    ppResolve     //< [out] The allocated resolve structure.
	);

/**
 * ALLOC macro for SG_resolve__alloc.
 */
#if defined(DEBUG)
#define SG_RESOLVE__ALLOC(pCtx, pPendingTree, ppResolve)                       \
	SG_STATEMENT(                                                              \
		SG_resolve* _pResolve = NULL;                                          \
		SG_resolve__alloc(pCtx, pPendingTree, &_pResolve);                     \
		_sg_mem__set_caller_data(_pResolve, __FILE__, __LINE__, "SG_resolve"); \
		*(ppResolve) = _pResolve;                                              \
	);
#else
#define SG_RESOLVE__ALLOC(pCtx, pPendingTree, ppResolve) \
	SG_resolve__alloc(pCtx, pPendingTree, ppResolve);
#endif

/**
 * Frees a resolve data structure.
 */
void SG_resolve__free(
	SG_context* pCtx,     //< [in] [out] Error and context info.
	SG_resolve* ppResolve //< [in] [out] The resolve structure to free.
	);

/**
 * NULLFREE macro for SG_resolve__free.
 */
#define SG_RESOLVE_NULLFREE(pCtx, pResolve) _sg_generic_nullfree(pCtx, pResolve, SG_resolve__free)

/**
 * Gets a count of how many choices are currently unresolved.
 * This is meant to be an easy way to check if there are unresolved conflicts,
 * for instance when attempting to commit.
 *
 * If you have an SG_resolve*, pass it.  If you don't and won't need one for
 * anything else, you can supply a pendingtree instead and this function will
 * handle the SG_resolve for you.  It is an error to pass both.
 */
void SG_resolve__get_unresolved_count(
	SG_context*     pCtx,         //< [in] [out] Error and context info.
	SG_resolve*     pResolve,     //< [in] Resolve instance to count unresolved choices in.
	                              //<      Must be NULL if pPendingTree is non-NULL.
	SG_pendingtree* pPendingTree, //< [in] Pendingtree to count the unresolved choices in.
	                              //<      Must be NULL if pResolve is non-NULL.
	SG_uint32*      pCount        //< [out] Number of unresolved choices/conflicts remaining.
	                              //<       If non-zero, things like commits should not be allowed.
	);

/**
 * Saves the state of the resolve back to the pendingtree it was allocated from.
 * Optionally, you can also save the pendingtree back to disk.
 */
void SG_resolve__save(
	SG_context* pCtx,            //< [in] [out] Error and context info.
	SG_resolve* pResolve,        //< [in] The resolve to save back to its pendingtree.
	SG_bool     bSavePendingTree //< [in] Whether or not the underlying pendingtree should also be saved.
	);

/**
 * Runs a viewer tool on a single value of a choice.
 * This is not implemented yet because we lack a system for viewer/editor tools.
 *//*
void SG_resolve__view_value(
	SG_context*        pCtx,           //< [in] [out] Error and context info.
	SG_resolve__value* pValue,         //< [in] The value to view.
	const char*        szSuggestedTool //< [in] Name of the viewer tool to use, or NULL if unimportant.
	                                   //<      Note that this is only a suggestion, and may not be used.
	);
*/
/**
 * Runs a diff tool on two values.
 */
void SG_resolve__diff_values(
	SG_context*        pCtx,           //< [in] [out] Error and context info.
	SG_resolve__value* pLeft,          //< [in] The value to treat as the left value in the diff tool.
	SG_resolve__value* pRight,         //< [in] The value to treat as the right value in the diff tool.
	const char*        szSuggestedTool //< [in] Name of the diff tool to use, or NULL if unimportant.
	                                   //<      Note that this is only a suggestion, and may not be used.
	);

/**
 * Runs a merge tool on two values in a mergeable choice.
 * If the resulting value's ancestry contains all of the other values in
 * the same choice, and that choice is unresolved, then the new value is
 * accepted as the choice's resolved value.
 */
void SG_resolve__merge_values(
	SG_context*         pCtx,            //< [in] [out] Error and context info.
	SG_resolve__value*  pBaseline,       //< [in] The value to treat as the baseline value in the merge tool.
	SG_resolve__value*  pOther,          //< [in] The value to treat as the other value in the merge tool.
	const char*         szSuggestedTool, //< [in] Name of the merge tool to use, or NULL if unimportant.
	                                     //<      Note that this is only a suggestion, and may not be used.
	const char*         szLabel,         //< [in] A label to assign to the result value, or NULL to use a default.
	                                     //<      If a value already exists with this label, then:
	                                     //<      1) If it's an automatically generated value, it is an error.
	                                     //<      2) If it's a manually generated value, it is overwritten.
	SG_int32*           pStatus,         //< [out] The SG_MERGETOOL__STATUS__* code returned by the merge tool.
	SG_resolve__value** ppResult         //< [out] The value resulting from the merge.  May be NULL if unneeded.
	                                     //<       Set to NULL if the merge is unsuccessful.
	);

/**
 * Resolves a choice by accepting a given value.
 * This will overwrite any already accepted value on the choice.
 */
void SG_resolve__accept_value(
	SG_context*        pCtx,  //< [in] [out] Error and context info.
	SG_resolve__value* pValue //< [in] The value to accept for the choice.
	);

/**
 * Gets the pendingtree that the resolve data is associated with.
 */
void SG_resolve__get_pendingtree(
	SG_context*      pCtx,         //< [in] [out] Error and context info.
	SG_resolve*      pResolve,     //< [in] The resolve to get the associated pendingtree of.
	SG_pendingtree** ppPendingTree //< [out] The pendingtree associated with the resolve.
	);

/**
 * Finds an item in a resolve by its working repo path.
 */
void SG_resolve__check_item__path(
	SG_context*        pCtx,       //< [in] [out] Error and context info.
	SG_resolve*        pResolve,   //< [in] The resolve structure to find an item in.
	const char*        szRepoPath, //< [in] The repo path of the item to find.
	SG_resolve__item** ppItem      //< [out] The requested item.
	                               //<       NULL if the item wasn't found in the resolve.
	);

/**
 * Finds an item in a resolve by its working repo path.
 * Throws SG_ERR_NOT_FOUND if the item isn't found.
 */
void SG_resolve__get_item__path(
	SG_context*        pCtx,       //< [in] [out] Error and context info.
	SG_resolve*        pResolve,   //< [in] The resolve structure to find an item in.
	const char*        szRepoPath, //< [in] The repo path of the item to find.
	SG_resolve__item** ppItem      //< [out] The requested item.
	);

/**
 * Finds an item in a resolve by GID.
 */
void SG_resolve__check_item__gid(
	SG_context*        pCtx,     //< [in] [out] Error and context info.
	SG_resolve*        pResolve, //< [in] The resolve structure to find an item in.
	const char*        szGid,    //< [in] The GID of the item to find.
	SG_resolve__item** ppItem    //< [out] The requested item.
	                             //<       NULL if the item wasn't found in the resolve.
	);

/**
 * Finds an item in a resolve by GID.
 * Throws SG_ERR_NOT_FOUND if the item isn't found.
 */
void SG_resolve__get_item__gid(
	SG_context*        pCtx,     //< [in] [out] Error and context info.
	SG_resolve*        pResolve, //< [in] The resolve structure to find an item in.
	const char*        szGid,    //< [in] The GID of the item to find.
	SG_resolve__item** ppItem    //< [out] The requested item.
	                             //<       NULL if the item wasn't found in the resolve.
	);

/**
 * Iterates through items involved in the resolve.
 * Items are returned in order of preferred resolution.
 */
void SG_resolve__foreach_item(
	SG_context*                 pCtx,            //< [in] [out] Error and context info.
	SG_resolve*                 pResolve,        //< [in] The resolve structure to iterate items in.
	SG_bool                     bOnlyUnresolved, //< [in] If SG_TRUE, only iterate unresolved items.
	SG_resolve__item__callback* fCallback,       //< [in] Callback to supply each item to.
	void*                       pUserData        //< [in] User data to supply to the callback.
	);

/**
 * Iterates through choices on items involved in the resolve.
 * This is a shortcut for a nested loop using SG_resolve__foreach_item
 * and then SG_resolve__item__foreach_choice on each item.
 */
void SG_resolve__foreach_choice(
	SG_context*                   pCtx,            //< [in] [out] Error and context info.
	SG_resolve*                   pResolve,        //< [in] The resolve structure to iterate choices in.
	SG_bool                       bOnlyUnresolved, //< [in] If SG_TRUE, only iterate unresolved items/choices.
	SG_resolve__choice__callback* fCallback,       //< [in] Callback to supply each choice to.
	void*                         pUserData        //< [in] User data to supply to the callback.
	);

/**
 * Gets the resolve structure that contains an item.
 */
void SG_resolve__item__get_resolve(
	SG_context*       pCtx,     //< [in] [out] Error and context info.
	SG_resolve__item* pItem,    //< [in] The item to get the containing resolve for.
	SG_resolve**      ppResolve //< [out] The resolve that contains the item.
	);

/**
 * Gets the GID of an item.
 */
void SG_resolve__item__get_gid(
	SG_context*       pCtx,  //< [in] [out] Error and context info.
	SG_resolve__item* pItem, //< [in] The item to get the GID of.
	const char**      ppGid  //< [out] The GID of the given item.
	);

/**
 * Gets the type of an item.
 */
void SG_resolve__item__get_type(
	SG_context*             pCtx,  //< [in] [out] Error and context info.
	SG_resolve__item*       pItem, //< [in] The item to get the type of.
	SG_treenode_entry_type* pType  //< [out] The type of the item.
	);

/**
 * Gets the current status of an item.
 */
void SG_resolve__item__get_status(
	SG_context*          pCtx,   //< [in] [out] Error and context info.
	SG_resolve__item*    pItem,  //< [in] The item to get the status of.
	SG_diffstatus_flags* pStatus //< [out] The item's current status flags.
	);

/**
 * Checks whether or not an item has a given status flag.
 * If multiple flags are passed, it checks if the item has ALL of them.
 */
void SG_resolve__item__has_status(
	SG_context*         pCtx,      //< [in] [out] Error and context info.
	SG_resolve__item*   pItem,     //< [in] The item to check the status of.
	SG_diffstatus_flags eStatus,   //< [in] The status flag(s) to check for.
	SG_bool*            pHasStatus //< [out] True if the item has all the given status flags, false otherwise.
	);

/**
 * Gets the HID of this item's common ancestor changeset.
 */
void SG_resolve__item__get_ancestor(
	SG_context*       pCtx,         //< [in] [out] Error and context info.
	SG_resolve__item* pItem,        //< [in] The item to get the ancestor of.
	const char**      ppAncestorHid //< [out] The HID of the item's ancestor changeset.
	);

/**
 * Gets the repo path of an item in a given changeset related to the resolve.
 */
void SG_resolve__item__get_repo_path(
	SG_context*       pCtx,        //< [in] [out] Error and context info.
	SG_resolve__item* pItem,       //< [in] The item to get the repo path of.
	const char*       szChangeset, //< [in] The HID of the changeset to get the item's repo path in.
	                               //<      Pass SG_RESOLVE__HID__WORKING to get the repo path in the working directory,
	                               //<      but use the __working overload to get the path of a currently deleted item.
	const char**      ppRepoPath   //< [out] The item's repo path in the given changeset.
	                               //<       Set to NULL if the item doesn't exist in the changeset.
	);

/**
 * Gets the repo path of an item in the working directory.
 */
void SG_resolve__item__get_repo_path__working(
	SG_context*       pCtx,      //< [in] [out] Error and context info.
	SG_resolve__item* pItem,     //< [in] The item to get the repo path of.
	const char**      ppRepoPath //< [out] The item's repo path in the working directory.
	                             //<       Set to NULL if the item doesn't exist in the changeset.
	);

/**
 * Gets the repo path of an item in its common ancestor changeset.
 */
void SG_resolve__item__get_repo_path__ancestor(
	SG_context*       pCtx,      //< [in] [out] Error and context info.
	SG_resolve__item* pItem,     //< [in] The item to get the repo path of.
	const char**      ppRepoPath //< [out] The item's repo path in its common ancestor changeset.
	                             //<       Set to NULL if the item doesn't exist in the changeset.
	);

/**
 * Gets the repo path of an item in the baseline parent changeset of the merge.
 */
void SG_resolve__item__get_repo_path__baseline(
	SG_context*       pCtx,      //< [in] [out] Error and context info.
	SG_resolve__item* pItem,     //< [in] The item to get the repo path of.
	const char**      ppRepoPath //< [out] The item's repo path in the baseline changeset.
	                             //<       Set to NULL if the item doesn't exist in the changeset.
	);

/**
 * Gets the repo path of an item in one of the "other" parent changesets of the merge.
 * Throws SG_ERR_NOT_FOUND if there is no "other" parent at the given index.
 */
void SG_resolve__item__get_repo_path__other(
	SG_context*       pCtx,      //< [in] [out] Error and context info.
	SG_resolve__item* pItem,     //< [in] The item to get the repo path of.
	SG_uint32         uIndex,    //< [in] 1-based index of the "other" parent to get the item's repo path in.
	const char**      ppRepoPath //< [out] The item's repo path in the given other changeset.
	                             //<       Set to NULL if the item doesn't exist in the changeset.
	);

/**
 * Gets the resolved status of an item.
 *
 * An item is considered resolved in two circumstances:
 * 1) All of its choices are resolved via accepting a value.
 * 2) Its file has been deleted.
 *
 * Note: Currently #2 is limited to cases where the item has
 *       been deleted due to its EXISTENCE choice being resolved
 *       as SG_FALSE.  In cases where the item has been deleted
 *       via other means, pDeleted will still return SG_FALSE.
 */
void SG_resolve__item__get_resolved(
	SG_context*       pCtx,      //< [in] [out] Error and context info.
	SG_resolve__item* pItem,     //< [in] The item to get the resolved status of.
	SG_bool*          pResolved, //< [out] Whether or not the given item has been resolved.
	                             //<       This output is required.
	SG_bool*          pDeleted   //< [out] If pResolved is set to SG_TRUE, then this indicates
	                             //<       if the item's resolution is to be deleted.  If pResolved
	                             //<       is set to SG_FALSE, this output is ignored and meaningless.
	                             //<       This output is optional.
	);

/**
 * Gets a choice from the item by state.
 */
void SG_resolve__item__check_choice__state(
	SG_context*          pCtx,    //< [in] [out] Error and context info.
	SG_resolve__item*    pItem,   //< [in] The item to get a choice from.
	SG_resolve__state    eState,  //< [in] The state of the choice to retrieve.
	SG_resolve__choice** ppChoice //< [out] The requested choice.
	                              //<       NULL if the item has no choice for that state.
	);

/**
 * Gets a choice from the item by state.
 * Throws SG_ERR_NOT_FOUND if there is no choice for that state on the item.
 */
void SG_resolve__item__get_choice__state(
	SG_context*          pCtx,    //< [in] [out] Error and context info.
	SG_resolve__item*    pItem,   //< [in] The item to get a choice from.
	SG_resolve__state    eState,  //< [in] The state of the choice to retrieve.
	SG_resolve__choice** ppChoice //< [out] The requested choice.
	);

/**
 * Iterates through choices on a single item involved in the resolve.
 * Choices are returned in order of preferred resolution.
 */
void SG_resolve__item__foreach_choice(
	SG_context*                   pCtx,            //< [in] [out] Error and context info.
	SG_resolve__item*             pItem,           //< [in] The item to iterate the choices on.
	SG_bool                       bOnlyUnresolved, //< [in] If SG_TRUE, only iterate unresolved choices.
	SG_resolve__choice__callback* fCallback,       //< [in] Callback to supply each choice to.
	void*                         pUserData        //< [in] User data to supply to the callback.
	);

/**
 * Gets the item that contains a choice.
 */
void SG_resolve__choice__get_item(
	SG_context*         pCtx,    //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice, //< [in] The choice to get the containing item for.
	SG_resolve__item**  ppItem   //< [out] The item that contains the choice.
	);

/**
 * Gets the item state that a choice is about.
 */
void SG_resolve__choice__get_state(
	SG_context*         pCtx,    //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice, //< [in] The choice to get the state of.
	SG_resolve__state*  pState   //< [out] The item state that the choice is about.
	);

/**
 * Gets information about the cause(s) of the choice.
 */
void SG_resolve__choice__get_causes(
	SG_context*                       pCtx,          //< [in] [out] Error and context info.
	SG_resolve__choice*               pChoice,       //< [in] The choice to get the cause(s) of.
	SG_mrg_cset_entry_conflict_flags* pConflicts,    //< [out] Set of conflict flags that caused this choice.
	                                                 //<       May be NULL if unneeded.
	SG_bool*                          pCollision,    //< [out] Whether or not a collision caused this choice.
	                                                 //<       May be NULL if unneeded.
	SG_vhash**                        ppDescriptions //< [out] A hash of descriptions of the choice's causes.
	                                                 //<       Each key is a single name of a cause (i.e. "Divergent Edit").
	                                                 //<       Each value is a brief description of that cause.
	);

/**
 * Gets whether or not a choice has mergeable state values.
 * This is generally only the case with file content state values.
 */
void SG_resolve__choice__get_mergeable(
	SG_context*         pCtx,      //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice,   //< [in] The choice to get the mergeable state of.
	SG_bool*            pMergeable //< [out] Whether or not the choice's state values are mergeable.
	);

/**
 * Gets whether or not a choice is considered resolved.
 *
 * A choice is considered resolved in two circumstances:
 * 1) One of its values has been accepted as its result.
 * 2) Its item has been deleted.
 *
 * Note: Currently #2 is limited to cases where the item has
 *       been deleted due to its EXISTENCE choice being resolved
 *       as SG_FALSE.  In cases where the item has been deleted
 *       via other means, pDeleted will still return SG_FALSE.
 */
void SG_resolve__choice__get_resolved(
	SG_context*         pCtx,     //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice,  //< [in] The choice to get the resolved state of.
	SG_bool*            pResolved, //< [out] Whether or not the given choice has been resolved.
	                               //<       This output is required.
	SG_bool*            pDeleted   //< [out] If pResolved is set to SG_TRUE, then this indicates
	                               //<       if the choice's resolution is for its item to be
	                               //<       deleted.  If pResolved is set to SG_FALSE, this
	                               //<       output is ignored and meaningless.
	                               //<       This output is optional.
	);

/**
 * Iterates through the set of other items involved in the same collision.
 * The item that owns the choice will not be returned.
 * This function only returns data for choices caused by a collision.
 */
void SG_resolve__choice__foreach_colliding_item(
	SG_context*                 pCtx,      //< [in] [out] Error and context info.
	SG_resolve__choice*         pChoice,   //< [in] The choice to iterate colliding items in.
	SG_resolve__item__callback* fCallback, //< [in] The callback to supply colliding items to.
	void*                       pUserData  //< [in] User data to supply to the callback.
	);

/**
 * Gets data specific to choices that are caused by path cycles.
 */
void SG_resolve__choice__get_cycle_info(
	SG_context*         pCtx,    //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice, //< [in] The choice to retrieve cycle data from.
	const char**        ppPath   //< [out] The cycling pseudo-path that merge wanted to create.
	                             //<       i.e. dir1/dir2/dir3/dir1
	);

/**
 * Iterates through the set of other items involved in the same cycle.
 * The item that owns the choice will not be returned.
 * This function only returns data for choices caused by a path cycle.
 */
void SG_resolve__choice__foreach_cycling_item(
	SG_context*                 pCtx,      //< [in] [out] Error and context info.
	SG_resolve__choice*         pChoice,   //< [in] The choice to iterate cycling items in.
	SG_resolve__item__callback* fCallback, //< [in] The callback to supply cycling items to.
	void*                       pUserData  //< [in] User data to supply to the callback.
	);

/**
 * Gets a value from a choice by its label.
 */
void SG_resolve__choice__check_value__label(
	SG_context*         pCtx,    //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice, //< [in] The choice to find a value in.
	const char*         szLabel, //< [in] The label of the value to retrieve.
	                             //<      Well-known value labels are defined by SG_RESOLVE__LABEL__*
	SG_int32            iIndex,  //< [in] Index to use for indexed labels (i.e. merge1, merge2, other3)
	                             //<      See well-known label definitions for which are indexed.
	                             //<      Ignored if 0, label indices are always 1-based.
	                             //<      Also ignored if 1, labels at index 1 don't actually have the suffix.
	                             //<      This allows iteration starting at 1 with no special case for the first label with no suffix.
	                             //<      A negative index starts from the last available index and works forward.
	                             //<      So -1 retrieves the value with the highest index, -2 the second highest, etc.
	SG_resolve__value** ppValue  //< [out] The requested value, or NULL if it wasn't found.
	);

/**
 * Gets a value from a choice by its label.
 * Throws SG_ERR_NOT_FOUND if the value isn't found.
 */
void SG_resolve__choice__get_value__label(
	SG_context*         pCtx,    //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice, //< [in] The choice to find a value in.
	const char*         szLabel, //< [in] The label of the value to retrieve.
	                             //<      Well-known value labels are defined by SG_RESOLVE__LABEL__*
	SG_int32            iIndex,  //< [in] Index to use for indexed labels (i.e. merge1, merge2, other3)
	                             //<      See well-known label definitions for which are indexed.
	                             //<      Ignored if 0, label indices are always 1-based.
	                             //<      Also ignored if 1, labels at index 1 don't actually have the suffix.
	                             //<      This allows iteration starting at 1 with no special case for the first label with no suffix.
	                             //<      A negative index starts from the last available index and works forward.
	                             //<      So -1 retrieves the value with the highest index, -2 the second highest, etc.
	SG_resolve__value** ppValue  //< [out] The requested value.
	);

/**
 * Gets a value from a choice by its associated changeset's HID.
 *
 * This function won't accept SG_RESOLVE__HID__WORKING as the HID to
 * look up because one choice can have multiple values associated with the
 * working HID under some circumstances (see below).  To get one of the
 * working values from the choice use SG_resolve__choice__check_value__label,
 * passing SG_RESOLVE__LABEL__WORKING and the index of the working value you
 * want (generally -1 to get the most recent one).
 *
 * One choice may end up with multiple values associated with the "working"
 * changeset if the user merges a working value with another value.  Since it
 * is now the parent of another value, the working value must remain unchanged
 * after such a merge for the historical record of the choice to make sense.
 * Therefore, when such a merge occurs an additional working value is created
 * to represent the current state of the WD so that the old working 
 * value no longer has to.  The working value that is currently directly
 * associated with the WD therefore is always the one with the highest index.
 */
void SG_resolve__choice__check_value__changeset(
	SG_context*         pCtx,    //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice, //< [in] The choice to find a value in.
	const char*         szHid,   //< [in] The HID of the changeset to retrieve the associated value from.
	                             //<      Cannot be SG_RESOLVE__HID__WORKING.
	SG_resolve__value** ppValue  //< [out] The requested value, or NULL if it wasn't found.
	);

/**
 * Gets a value from a choice by its associated changeset's HID.
 * Throws SG_ERR_NOT_FOUND if the value isn't found.
 *
 * Note: This function won't accept SG_RESOLVE__HID__WORKING.  See
 *       SG_resolve__choice__check_value__changeset for more info.
 */
void SG_resolve__choice__get_value__changeset(
	SG_context*         pCtx,    //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice, //< [in] The choice to find a value in.
	const char*         szHid,   //< [in] The HID of the changeset to retrieve the associated value from.
	                             //<      Cannot be SG_RESOLVE__HID__WORKING.
	SG_resolve__value** ppValue  //< [out] The requested value.
	);

/**
 * Gets the value that was accepted as the resolution of the choice.
 *
 * Note that this function can return NULL even if the choice is resolved.
 * This might occur if the choice was resolved using a method other than
 * accepting one of its values.  See SG_resolve__choice__get_resolved.
 */
void SG_resolve__choice__get_value__accepted(
	SG_context*         pCtx,    //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice, //< [in] The choice to get the result from.
	SG_resolve__value** ppValue  //< [out] The value accepted as the choice's resolution.
	                             //<       NULL if none has been accepted.
	);

/**
 * Iterates through each value associated with a choice.
 */
void SG_resolve__choice__foreach_value(
	SG_context*                  pCtx,        //< [in] [out] Error and context info.
	SG_resolve__choice*          pChoice,     //< [in] The choice to iterate values in.
	SG_bool                      bOnlyLeaves, //< [in] If SG_TRUE, only iterate leaf values for mergeable choices.
	SG_resolve__value__callback* fCallback,   //< [in] The callback to supply values to.
	void*                        pUserData    //< [in] User data to supply to the callback.
	);

/**
 * Gets the choice that contains a value.
 */
void SG_resolve__value__get_choice(
	SG_context*          pCtx,    //< [in] [out] Error and context info.
	SG_resolve__value*   pValue,  //< [in] The value to get the containing choice for.
	SG_resolve__choice** ppChoice //< [out] The choice that contains the value.
	);

/**
 * Gets the label and/or changeset HID that identifies a value.
 */
void SG_resolve__value__get_label(
	SG_context*        pCtx,   //< [in] [out] Error and context info.
	SG_resolve__value* pValue, //< [in] The value to get data about.
	const char**       ppLabel //< [out] The value's label.
	);

/**
 * Gets a value structure's actual data value.
 * Use SG_resolve__value__get_summary if you're looking for something more suitable for display.
 *
 * Throws SG_ERR_USAGE if the data value isn't a bool.
 * Values for the following choice types have boolean data:
 * - EXISTENCE
 */
void SG_resolve__value__get_data__bool(
	SG_context*        pCtx,   //< [in] [out] Error and context info.
	SG_resolve__value* pValue, //< [in] The value to get the data of.
	SG_bool*           pData   //< [out] The value structure's data as a bool.
	);

/**
 * Gets a value structure's actual data value.
 * Use SG_resolve__value__get_summary if you're looking for something more suitable for display.
 *
 * Throws SG_ERR_USAGE if the data value isn't an integer.
 * Values for the following choice types have integral data:
 * - ATTRIBUTES
 */
void SG_resolve__value__get_data__int64(
	SG_context*        pCtx,   //< [in] [out] Error and context info.
	SG_resolve__value* pValue, //< [in] The value to get the data of.
	SG_int64*          pData   //< [out] The value structure's data as an integer.
	);

/**
 * Gets a value structure's actual data value.
 * Use SG_resolve__value__get_summary if you're looking for something more suitable for display.
 *
 * Throws SG_ERR_USAGE if the data value isn't a string.
 * Values for the following choice types have textual data:
 * - NAME
 * - LOCATION
 * - XATTRIBUTES
 * - CONTENTS
 */
void SG_resolve__value__get_data__sz(
	SG_context*        pCtx,   //< [in] [out] Error and context info.
	SG_resolve__value* pValue, //< [in] The value to get the data of.
	const char**       ppData  //< [out] The value structure's data as a string.
	);

/**
 * Gets a summary of the value that is suitable for display.
 * In many cases, this is the actual value in a human-readable form.
 * In other cases, this is a shorter and/or more readable version of the actual value.
 * Some values have no good way to summarize, and so return NULL.
 */
void SG_resolve__value__get_summary(
	SG_context*         pCtx,     //< [in] [out] Error and context info.
	SG_resolve__value*  pValue,   //< [in] The value to get a summary of.
	SG_string**         ppSummary //< [out] A summary of the value suitable for display.
	                              //<       NULL if the value cannot be well summarized.
	);

/**
 * Gets a set of flags that describe the value.
 */
void SG_resolve__value__get_flags(
	SG_context*        pCtx,   //< [in] [out] Error and context info.
	SG_resolve__value* pValue, //< [in] The value to get flags for.
	SG_uint32*         pFlags  //< [out] The flags for the value.
	);

/**
 * Checks whether or not a value has a given flag.
 * If multiple flags are passed, it checks if the value has ALL of them.
 */
void SG_resolve__value__has_flag(
	SG_context*        pCtx,    //< [in] [out] Error and context info.
	SG_resolve__value* pValue,  //< [in] The value to check for a flag on.
	SG_uint32          uFlag,   //< [in] The flag(s) to check for.
	SG_bool*           pHasFlag //< [out] True if the value has the given flag(s), false otherwise.
	);

/**
 * Gets data about the value's association with a changeset.
 */
void SG_resolve__value__get_changeset(
	SG_context*        pCtx,          //< [in] [out] Error and context info.
	SG_resolve__value* pValue,        //< [in] The value to get info about.
	SG_bool*           pHasChangeset, //< [out] Whether or not the value is associated with a changeset.
	                                  //<       This output is required.
	const char**       ppChangeset    //< [out] The HID of the changeset the value is associated with.
	                                  //<       Set to SG_RESOLVE__HID__WORKING if the value is associated with the working directory.
	                                  //<       This output is optional, and meaningless/ignored if pHasChangeset is false.
	);

/**
 * Gets data about the value's creation via merge.
 */
void SG_resolve__value__get_merge(
	SG_context*         pCtx,       //< [in] [out] Error and context info.
	SG_resolve__value*  pValue,     //< [in] The value to get info about.
	SG_bool*            pWasMerged, //< [out] Whether or not the value was created via merging.
	                                //<       This output is required.
	SG_resolve__value** ppBaseline, //< [out] The value's "baseline" parent.
	                                //<       This output is optional, and meaningless/ignored if pWasMerged is false.
	SG_resolve__value** ppOther,    //< [out] The value's "other" parent.
	                                //<       This output is optional, and meaningless/ignored if pWasMerged is false.
	SG_bool*            pAutomatic, //< [out] Whether the merge was automatic (done by the merge system)
	                                //<       or manual (done by the user with the resolve system).
	                                //<       This output is optional, and meaningless/ignored if pWasMerged is false.
	const char**        ppTool,     //< [out] The name of the merge tool used to create the value.
	                                //<       This output is optional, and meaningless/ignored if pWasMerged is false.
	SG_int32*           pToolResult //< [out] The result of running ppTool to create the value.
	                                //<       One of the SG_MERGETOOL__STATUS__* values.
	                                //<       This output is optional, and meaningless/ignored if pWasMerged is false.
	);

END_EXTERN_C;

#endif
