/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg_vv_verbs_prototypes.h
 *
 * @details Routines to (eventually) perform most major vv commands.
 *
 */

//////////////////////////////////////////////////////////////////

#ifndef H_SG_VV_VERBS_PROTOTYPES_H
#define H_SG_VV_VERBS_PROTOTYPES_H

BEGIN_EXTERN_C;

//////////////////////////////////////////////////////////////////

/**
 * Create a new REPO with the given name and
 * return the HID of the initial CSet.
 *
 * This DOES NOT create a Working Directory.
 * You own the returned HID.
 *
 * TODO 2010/10/21 Deal with admin_id.
 */
void SG_vv_verbs__init_new_repo(SG_context * pCtx,
								const char * pszRepoName,
								const char * pszFolder,      //< [in] arbitrary path (abs or rel) for init WD (if desired); defaults to CWD if blank or null
								const char * pszStorage,     //< [in] If null, use default config: new_repo/driver
								const char * pszHashMethod,	 //< [in] If null, use default config: new_repo/hash_method
								SG_bool bNoWD,               //< [in] Request a new repo only (no WD)
								SG_bool bSubmodule,          //< [in] Request repo/WD be created as a submodule within an existing repo/WD
								const char * psz_shared_users,	 //< [in] The new repo should share users accounts with this one 
								char ** ppszGidRepoId,       //< [out] repo-id of the created repo (gid)
								char ** ppszHidCSetFirst);   //< [out] HID of the initial CSET

/**
 * Create a new WD that references the named REPO
 * at the given CSET HID.
 */
void SG_vv_verbs__checkout(SG_context * pCtx,
						   const char * pszRepoName,
						   const char * pszHidCSet,		//< [in] Full HID of CSET (caller should use SG_vv_utils__lookup_cset()).
						   const char * pszFolder,		//< [in] arbitrary path (abs or rel); defaults to CWD if blank or null.
						   SG_bool bRequireNewDirOrEmpty,
						   SG_bool bSubmodule,
                           const char* psz_branch_name,
						   const SG_vector * pvec_use_submodules,
						   SG_vhash ** ppvhResult);

/**
 * Export the contents of the version control tree
 * as the given CSET HID.
 */
void SG_vv_verbs__export(SG_context * pCtx,
						 const char * pszRepoName,
						 const char * pszHidCSet,		//< [in] Full HID of CSET (caller should use SG_vv_utils__lookup_cset()).
						 const char * pszFolder,		//< [in] arbitrary path (abs or rel); defaults to CWD if blank or null.
						 SG_bool bNoSubmodules,			//< [in] do not export contained submodules
						 const SG_vector * pvec_use_submodules,	//< [in] optional list of hints for repo-names of referenced submodules
						 SG_vhash ** ppvhResult);		//< [out] list of submodule problems

/**
 * Get a vhash summary of version control changes between:
 *  1) two arbitrary changesets (provide both pszFromRev and pszToRev)
 *  2) the working folder and an arbitrary changeset (provide only pszFromRev)
 *  3) the working folder and its baseline (provide neither pszFromRev and pszToRev)
 *
 * In all cases, a valid pPendingTree must be provided. "Well," you might say, "it's
 * "quite strange that I would provide a pending tree when comparing two arbitrary changesets!"
 * "Yes," I would nod, "that is quite strange. In that case, the SG_repo pointer will be plucked
 * from pPendingTree, but it is strange nonetheless."
 */
void SG_vv_verbs__status(SG_context * pCtx,
						 SG_pendingtree* pPendingTree,		// [in]
						 const char* pszFromRev,			// [in] Optional. Provide only for cases 2 and 3.
						 const char* pszToRev,				// [in] Optional. Provide only for case 1.
						 SG_uint32 count_file_args,			// [in] The number of elements in paszFileArgs.
						 const char** paszFileArgs,			// [in] File(s) whose changes should be determined 
						 const SG_file_spec* pFilespec,		// [in] the set of --includes/--excludes/etc.
						 SG_bool bRecursive,
						 SG_bool bVerbose,
						 const SG_varray** ppvaMergeIssues, // [out] list of merge issues. You don't own this, and it is valid only as long as pPendingTree is in memory.
						 SG_vhash** ppvhStatus);			// [out] status data

/**
 * Do a version control checkin.
 */
void SG_vv_verbs__commit(SG_context * pCtx,
						 const char * pszFolder,			//<[in] arbitrary path (abs or rel) to a WD or a location within a WD; defaults to CWD if blank or null.
						 const char * pszComment,			//<[in] the checkin comment (optional); can be added later using 'vv comment'.
						 const SG_file_spec * pFilespec,	//<[in] the set of --includes/--excludes/etc.
						 SG_bool bRecursive,
						 SG_bool bNoSubmodules,				//<[in] do not do nested commits for dirty submodules.
						 const char * pszUser,				//<[in] credit this user (useful for import mainly); defaults to the current user.
						 SG_int64 i64Date,					//<[in] date to assign to checkin (useful for import mainly); see SG_AUDIT__WHEN__NOW.
						 SG_uint32 count_args, const char ** paszArgs,	//<[in] (optional) list of files_or_folders (relative to pszFolder).
						 SG_uint32 count_assocs, const char * const * ppszAssocs,
						 const SG_stringarray * psaStamps,
						 void (*promptfn)(SG_context *, SG_string *,char **,const SG_varray *),
						 SG_string *boilerplate,

						 char ** ppsz_hid_cset);			//<[out] (optional) the HID of the newly created CSET.

/**
 * Update the working directory to a different changeset.
 */
void SG_vv_verbs__update(SG_context * pCtx,
						 const char * pszFolder,			//<[in] path (abs or rel) to some folder within the WD
						 const char * pszTargetChangeset,	//<[in] full HID of desired cset (may be null if for absolute/unique head)
						 const char * pszTargetBranchName,	//<[in] name of branch to tie to after update (may be null if no tie desired)
						 SG_bool bVerbose,					//<[in] generate log of steps performed or to be performed
						 SG_bool bTest,						//<[in] test-only; do not modify WD.  (verbose is implied when testing)
						 const SG_vector * pvec_use_submodules,	//<[in] (optional) array of hints for nested checkouts within the update
						 SG_varray ** ppvaActionLog,		//<[out] (optional) array of operations required (valid when verbose)
						 SG_varray ** ppvaWarnings,			//<[out] (optional) array of portability problems (valid when not ignoring)
						 SG_varray ** ppvaMissingDetails);	//<[out] (optional) array of nested update/checkout problems


#define SG_VV_VERBS__STATUS__OK						0	// action performed ( __DO_IT and/or __LOG_IT ); "log" may be present
#define SG_VV_VERBS__STATUS__PORTABILITY_WARNING	1	// action not performed because of portability warning; "port" should have details
#define SG_VV_VERBS__STATUS__SUBMODULE_PROBLEM		2	// action not performed because of submodule problems; "sub" should have details

/**
 * Wrapper for SG_vv_verbs__update() to gather up various
 * exit/warning results and build a vhash wrapping them
 * to make it easier to return the results to JS.
 */
void SG_vv_verbs__update2(SG_context * pCtx,
						  const char * pszFolder,
						  const char * pszTargetChangeset,
						  const char * pszTargetBranchName,
						  SG_bool bVerbose,
						  SG_bool bTest,
						  const SG_vector * pvec_use_submodules,
						  SG_vhash ** ppvhResult);

/**
 * Wrapper for pull.  This does not include any option 
 * to trigger an update after the pull.  You must manually
 * call update yourself.  
 */
void SG_vv_verbs__pull(SG_context * pCtx,
					   const char* pszSrcRepoSpec,				// [in] The repo to pull from.  This may be a local repo name, or a URL of the remote server (http://SERVER/repos/REPONAME).
					   const char * pszDestRepoDescriptorName,  // [in] The local repo descriptor name.
					   SG_vhash * pvh_area_names,				// [in] The list of dag areas to pull.  It is an error to specify multiple entries in pva_area_names and any criteria in ppRevSpec.
					   SG_rev_spec** ppRevSpec,				    // [in] The nodes to pull. On succes, ownership is taken.
																//		The remote repo will look them up: the state of these branches in the destination repo is irrelevant.
					   SG_vhash** ppvhStats);					// [out] Optional stats about the work done by pull.
/**
 * Wrapper for push.
 */
void SG_vv_verbs__push(SG_context * pCtx,
						   const char * pszDestRepoSpec,			// [in]  The repo to push to.  This may be a local repo name, or a URL of the remote server (http://SERVER/repos/REPONAME).
						   const char* pszSrcRepoDescriptorName,	// [in]  The local repo descriptor name.
						   SG_vhash * pvh_area_names,				// [in]  The list of dags to push.  It is an error to specify multiple areas and either of prbTags or prbDagnodePrefixes.
						   SG_rev_spec* pRevSpec,
						   SG_bool bForce,             				// [in]  If true, do not perform the branch/lock pull-and-check before doing the push
						   SG_vhash** ppvhStats);					// [out] Optional stats about the work done by push.
/**
 * Wrapper for incoming.
 */
void SG_vv_verbs__incoming(SG_context * pCtx,
						   const char* pszSrcRepoSpec,					// [in]  The repo to pull from.  This may be a local repo name, or a URL of the remote server (http://SERVER/repos/REPONAME).
						   const char * pszDestRepoDescriptorName,		// [in]  The local repo descriptor name.
						   SG_history_result** pp_incomingChangesets,	// [out] The changesets that need to be transferred from src to dest.
						   SG_vhash** ppvhStats);						// [out] Optional stats about the work done by incoming.

void SG_vv_verbs__outgoing(SG_context* pCtx,
							const char* pszDestRepoSpec,				// [in]  The repo to push to.  This may be a local repo name, or a URL of the remote server (http://SERVER/repos/REPONAME).
							const char* pszSrcRepoDescriptorName,		// [in]  The local repo descriptor name.
							SG_history_result** pp_OutgoingChangesets,  // [out] The changesets that need to be transferred from src to dest.
							SG_vhash** ppvhStats);						// [out] Optional stats about the work done by outgoing.

/**
 *
 */
void SG_vv_verbs__merge(SG_context * pCtx,
						const char * pszFolder,
						const char * psz_other_csetid,
						const char * pszToolOverride,
						SG_bool bTest,
						const SG_vector * pvec_use_submodules,
						SG_vhash ** ppvhResults,
						SG_wd_plan ** ppPlan);

/**
 *
 *
 */
void SG_vv_verbs__revert(SG_context * pCtx,
						 const char * pszFolder,
						 SG_bool bRecursive,
						 SG_bool bNoSubmodules,
						 SG_bool bVerbose,
						 SG_bool bTest,
						 SG_bool bNoBackups,
						 const SG_vector * pvec_use_submodules,
						 const SG_file_spec * pFilespec,
						 SG_uint32 count_items, const char ** paszItems,
						 SG_vhash ** ppvhResult);

void SG_vv_verbs__history__repo2(
        SG_context * pCtx,
        SG_repo *, 
        SG_uint32 count_args, const char ** paszArgs,  //The arguments can be repo paths
		SG_rev_spec* pRevSpec, // optional
		const char* pszUser,
        const char* pszStamp,
        SG_uint32 nResultLimit,
		SG_bool bHideObjectMerges,
        SG_int64 nFromDate,
        SG_int64 nToDate,
        SG_bool* pbHasResult,
        SG_history_result ** ppResult,
		SG_history_token ** ppHistoryToken
);

void SG_vv_verbs__history__repo(
        SG_context * pCtx,
        const char * pszRepoDescriptorName, //The local descriptor name.
        SG_uint32 count_args, const char ** paszArgs,  //The arguments can be repo paths
		SG_rev_spec* pRevSpec, // optional
		const char* pszUser,
        const char* pszStamp,
        SG_uint32 nResultLimit,
		SG_bool bHideObjectMerges,
        SG_int64 nFromDate,
        SG_int64 nToDate,
        SG_bool* pbHasResult,
		SG_vhash** ppvhBranchPile,
		SG_history_result ** ppResult,
		SG_history_token ** ppHistoryToken
);


void SG_vv_verbs__history__workingFolder(
        SG_context * pCtx,
        const char * pszPendingTreePath, //The local path that will be used to lookup the pending tree.  If this is null, the pendingtree will be looked up by the current working folder.
        SG_uint32 count_args, const char ** paszArgs,  //The arguments can be either local paths or repo paths
		SG_rev_spec* pRevSpec, // optional
		const char* pszUser,
        const char* pszStamp,
		SG_bool bDetectCurrentBranch,
        SG_uint32 nResultLimit,
		SG_bool bHideObjectMerges,
		SG_int64 nFromDate,
        SG_int64 nToDate,
        SG_bool* pbHasResult,
		SG_vhash** ppvhBranchPile,
		SG_history_result ** ppResult,
		SG_history_token ** ppHistoryToken
);

void SG_vv_verbs__history__fetch_more__repo(SG_context * pCtx,
											const char * pszRepoDescriptorName, //The local descriptor name.
											SG_history_token * pHistoryToken,
											SG_uint32 nResultLimit,
											SG_history_result ** ppHistoryResults,
											SG_history_token **ppNewToken);

void SG_vv_verbs__history__fetch_more__repo2(SG_context * pCtx,
											SG_repo * pRepo, 
											SG_history_token * pHistoryToken,
											SG_uint32 nResultLimit,
											SG_history_result ** ppHistoryResults,
											SG_history_token **ppNewToken);


void SG_vv_verbs__history__fetch_more__working_folder(SG_context * pCtx,
											const char * pszPendingTreePath,  //The local path that will be used to lookup the pending tree.  If this is null, the pendingtree will be looked up by the current working folder.
											SG_history_token * pHistoryToken,
											SG_uint32 nResultLimit,
											SG_history_result ** ppHistoryResults,
											SG_history_token **ppNewToken);

//////////////////////////////////////////////////////////////////

END_EXTERN_C;

#endif//H_SG_VV_VERBS_PROTOTYPES_H
