/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include <sg.h>

#define SG_BLOBSET_MAX_IN_MEMORY (100 * 1000)

struct _sg_blobset
{
    SG_uint32 count;
    SG_bool b_memory;
    sqlite3* psql;
    sqlite3_stmt* pStmt;
};

static void sg_blobset__move_to_temporary_db(
	SG_context * pCtx,
    SG_blobset* pbs
    );

void SG_blobset__insert(
	SG_context * pCtx,
    SG_blobset* pbs,
    const char* psz_hid,
    const char* psz_filename,
    SG_uint64 offset,
    SG_blob_encoding encoding,
    SG_uint64 len_encoded,
    SG_uint64 len_full,
    const char* psz_hid_vcdiff,
    const char* psz_objectid,
    SG_int64 i_objectid_timestamp
    )
{
    if (
            pbs->b_memory
            && ((pbs->count + 1) > SG_BLOBSET_MAX_IN_MEMORY)
       )
    {
        SG_ERR_CHECK(  sg_blobset__move_to_temporary_db(pCtx, pbs)  );
    }

    SG_ERR_CHECK(  sg_sqlite__reset(pCtx, pbs->pStmt)  );
    SG_ERR_CHECK(  sg_sqlite__clear_bindings(pCtx, pbs->pStmt)  );

    if (SG_BLOBENCODING__VCDIFF == encoding)
    {
        SG_ASSERT(psz_hid_vcdiff);
    }

    SG_ERR_CHECK(  sg_sqlite__bind_text(pCtx, pbs->pStmt,1,psz_hid)  );
    if (psz_filename)
    {
        SG_ERR_CHECK(  sg_sqlite__bind_text(pCtx, pbs->pStmt,2,psz_filename)  );
    }
    else
    {
        SG_ERR_CHECK(  sg_sqlite__bind_null(pCtx, pbs->pStmt,2)  );
    }
    SG_ERR_CHECK(  sg_sqlite__bind_int64(pCtx, pbs->pStmt,3,offset)  );
    SG_ERR_CHECK(  sg_sqlite__bind_int(pCtx, pbs->pStmt,4,encoding)  );
    SG_ERR_CHECK(  sg_sqlite__bind_int64(pCtx, pbs->pStmt,5,len_encoded)  );
    SG_ERR_CHECK(  sg_sqlite__bind_int64(pCtx, pbs->pStmt,6,len_full)  );
    if (psz_hid_vcdiff)
    {
        SG_ERR_CHECK(  sg_sqlite__bind_text(pCtx, pbs->pStmt,7,psz_hid_vcdiff)  );
    }
    else
    {
        SG_ERR_CHECK(  sg_sqlite__bind_null(pCtx, pbs->pStmt,7)  );
    }
    if (psz_objectid)
    {
        SG_ERR_CHECK(  sg_sqlite__bind_text(pCtx, pbs->pStmt,8,psz_objectid)  );
    }
    else
    {
        SG_ERR_CHECK(  sg_sqlite__bind_null(pCtx, pbs->pStmt,8)  );
    }
    SG_ERR_CHECK(  sg_sqlite__bind_int64(pCtx, pbs->pStmt,9,i_objectid_timestamp)  );
    sg_sqlite__step(pCtx,pbs->pStmt,SQLITE_DONE);
    pbs->count++;

fail:
    ;
}

void SG_blobset__create_blobs_table(
        SG_context* pCtx, 
        sqlite3* psql
        )
{
	SG_ERR_CHECK_RETURN(  sg_sqlite__exec__va(pCtx, psql, "CREATE TABLE \"blobs\""
                              "   ("
                              "     \"hid\"                VARCHAR UNIQUE NOT NULL,"
                              "     \"filename\"           VARCHAR            NULL,"
                              "     \"offset\"             INTEGER            NULL,"
                              "     \"encoding\"           INTEGER            NULL,"
                              "     \"len_encoded\"        INTEGER            NULL,"
                              "     \"len_full\"           INTEGER            NULL,"
                              "     \"hid_vcdiff\"         VARCHAR            NULL,"
                              "     \"objectid\"           VARCHAR            NULL,"
                              "     \"objectid_timestamp\" INTEGER            NULL"
                              "   )")  );
}

static void _my_selectall__prepare(
        SG_context* pCtx, 
        sqlite3* psql, 
        sqlite3_stmt** pp
        )
{
    sqlite3_stmt* pStmt = NULL;

	SG_ERR_CHECK(  sg_sqlite__prepare(pCtx, psql, &pStmt,
									  "SELECT \"hid\", \"filename\", \"offset\", \"encoding\", \"len_encoded\", \"len_full\", \"hid_vcdiff\", \"objectid\", \"objectid_timestamp\" FROM \"blobs\"")  );

    *pp = pStmt;

    return;
fail:
    return;
}

static void _my_insert__prepare(
        SG_context* pCtx, 
        sqlite3* psql, 
        sqlite3_stmt** pp
        )
{
    sqlite3_stmt* pStmt = NULL;

    SG_ERR_CHECK(  sg_sqlite__prepare(pCtx, psql,&pStmt,
                                      "INSERT OR IGNORE INTO \"blobs\" (\"hid\", \"filename\", \"offset\", \"encoding\", \"len_encoded\", \"len_full\", \"hid_vcdiff\", \"objectid\", \"objectid_timestamp\") VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)")  );

    *pp = pStmt;

    return;
fail:
    return;
}

void SG_blobset__create(
	SG_context * pCtx,
    SG_blobset** pp
    )
{
    SG_blobset* pbs = NULL;

	SG_ERR_CHECK(  SG_alloc1(pCtx, pbs)  );

    pbs->b_memory = SG_TRUE;
	SG_ERR_CHECK(  sg_sqlite__open__memory(pCtx, &pbs->psql)  );
    SG_ERR_CHECK(  SG_blobset__create_blobs_table(pCtx, pbs->psql)  );

    *pp = pbs;
    pbs = NULL;

fail:
    // TODO free it
    ;
}

void SG_blobset__begin_inserting(
	SG_context * pCtx,
    SG_blobset* pbs
    )
{
    if (pbs->psql)
    {
        SG_ERR_CHECK(  _my_insert__prepare(pCtx, pbs->psql, &pbs->pStmt)  );
        SG_ERR_CHECK(  sg_sqlite__exec(pCtx, pbs->psql, "BEGIN TRANSACTION")  );
    }

fail:
    ;
}

void SG_blobset__finish_inserting(
	SG_context * pCtx,
    SG_blobset* pbs
    )
{
    if (pbs->psql)
    {
        SG_ERR_CHECK(  sg_sqlite__nullfinalize(pCtx, &pbs->pStmt)  );
        SG_ERR_CHECK(  sg_sqlite__exec(pCtx, pbs->psql, "COMMIT TRANSACTION")  );
    }

fail:
    ;
}

void SG_blobset__free(
	SG_context * pCtx,
    SG_blobset* pbs
    )
{
    if (!pbs)
    {
        return;
    }

    if (pbs->psql)
    {
        SG_ERR_CHECK(  sg_sqlite__nullfinalize(pCtx, &pbs->pStmt)  );
        SG_ERR_CHECK(  sg_sqlite__close(pCtx, pbs->psql)  );
    }
    SG_NULLFREE(pCtx, pbs);

fail:
    ;
}

static void sg_blobset__move_to_temporary_db(
	SG_context * pCtx,
    SG_blobset* pbs
    )
{
    sqlite3* psql_new = NULL;
    SG_bool b_inserting = SG_FALSE;

    if (pbs->pStmt)
    {
        SG_ERR_CHECK(  SG_blobset__finish_inserting(pCtx, pbs)  );
        b_inserting = SG_TRUE;
    }

	SG_ERR_CHECK(  sg_sqlite__open__temporary(pCtx, &psql_new)  );
    SG_ERR_CHECK(  SG_blobset__create_blobs_table(pCtx, psql_new)  );
    SG_ERR_CHECK(  SG_blobset__copy_into_sqlite(pCtx, pbs, psql_new)  );
    SG_ERR_CHECK(  sg_sqlite__close(pCtx, pbs->psql)  );
    pbs->psql = psql_new;
    pbs->b_memory = SG_FALSE;
    psql_new = NULL;

    if (b_inserting)
    {
        SG_ERR_CHECK(  SG_blobset__begin_inserting(pCtx, pbs)  );
    }

fail:
    if (psql_new)
    {
        SG_ERR_IGNORE(  sg_sqlite__close(pCtx, psql_new)  );
    }
}

void SG_blobset__copy_into_sqlite(
	SG_context * pCtx,
    SG_blobset* pbs,
    sqlite3* psql_to
    )
{
    int rc = 0;
    sqlite3_stmt* pStmt_from = NULL;
    sqlite3_stmt* pStmt_to = NULL;
    SG_uint32 count_columns = 0;

    SG_ERR_CHECK(  _my_selectall__prepare(pCtx, pbs->psql, &pStmt_from)  );
    SG_ERR_CHECK(  _my_insert__prepare(pCtx, psql_to, &pStmt_to)  );

    while ((rc=sqlite3_step(pStmt_from)) == SQLITE_ROW)
    {
        SG_uint32 i= 0;

        count_columns = sqlite3_column_count(pStmt_from);

        SG_ERR_CHECK(  sg_sqlite__reset(pCtx, pStmt_to)  );
        SG_ERR_CHECK(  sg_sqlite__clear_bindings(pCtx, pStmt_to)  );

        for (i=0; i<count_columns; i++)
        {
            int t = sqlite3_column_type(pStmt_from, i);

            if (SQLITE_INTEGER == t)
            {
                SG_ERR_CHECK(  sg_sqlite__bind_int64(pCtx, pStmt_to, i+1, sqlite3_column_int64(pStmt_from, i))  );
            }
            else if (SQLITE_FLOAT == t)
            {
                SG_ERR_CHECK(  sg_sqlite__bind_double(pCtx, pStmt_to, i+1, sqlite3_column_double(pStmt_from, i))  );
            }
            else if (SQLITE_TEXT == t)
            {
                SG_ERR_CHECK(  sg_sqlite__bind_text(pCtx, pStmt_to, i+1, (char*) sqlite3_column_text(pStmt_from, i))  );
            }
            else if (SQLITE_BLOB == t)
            {
                SG_ERR_THROW(  SG_ERR_NOTIMPLEMENTED  );
            }
            else if (SQLITE_NULL == t)
            {
                SG_ERR_CHECK(  sg_sqlite__bind_null(pCtx, pStmt_to, i+1)  );
            }
        }

        SG_ERR_CHECK(  sg_sqlite__step(pCtx, pStmt_to, SQLITE_DONE)  );
    }
    if (rc != SQLITE_DONE)
    {
        SG_ERR_THROW(  SG_ERR_SQLITE(rc)  );
    }

fail:
    SG_ERR_IGNORE(  sg_sqlite__nullfinalize(pCtx, &pStmt_from)  );
    SG_ERR_IGNORE(  sg_sqlite__nullfinalize(pCtx, &pStmt_to)  );
}

void SG_blobset__get(
        SG_context * pCtx,
        SG_blobset* pbs,
        SG_uint32 limit,
        SG_uint32 skip,
        SG_vhash** ppvh
        )
{
    SG_vhash* pvh_result = NULL;
    sqlite3_stmt* pStmt = NULL;
    int rc = 0;

    SG_ERR_CHECK(  SG_vhash__alloc(pCtx, &pvh_result)  );
	SG_ERR_CHECK(  sg_sqlite__prepare(pCtx, pbs->psql, &pStmt,
									  "SELECT \"hid\", \"encoding\", \"len_encoded\", \"len_full\", \"hid_vcdiff\", \"objectid\", \"objectid_timestamp\" FROM \"blobs\" ORDER BY \"hid\" ASC LIMIT %d OFFSET %d", limit ? (SG_int32) limit : -1, skip)  );
    while ((rc=sqlite3_step(pStmt)) == SQLITE_ROW)
    {
        SG_vhash* pvh = NULL;

        const char* psz_hid = (const char *)sqlite3_column_text(pStmt,0);
        SG_blob_encoding encoding = (SG_blob_encoding) sqlite3_column_int(pStmt, 1);
        SG_uint64 len_encoded = sqlite3_column_int64(pStmt, 2);
        SG_uint64 len_full = sqlite3_column_int64(pStmt, 3);
        const char* psz_hid_vcdiff = (const char *)sqlite3_column_text(pStmt,4);
        const char* psz_objectid = (const char *)sqlite3_column_text(pStmt,5);
        SG_int64 objectid_timestamp = sqlite3_column_int64(pStmt, 6);

        SG_ERR_CHECK(  SG_vhash__addnew__vhash(pCtx, pvh_result, psz_hid, &pvh)  );

        SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh, "encoding", encoding)  );
        SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh, "len_encoded", len_encoded)  );
        SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh, "len_full", len_full)  );

        if (psz_hid_vcdiff)
        {
            SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvh, "hid_vcdiff", psz_hid_vcdiff)  );
        }
        if (psz_objectid)
        {
            SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvh, "objectid", psz_objectid)  );
            SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh, "objectid_timestamp", objectid_timestamp)  );
        }
    }
    if (rc != SQLITE_DONE)
    {
        SG_ERR_THROW(  SG_ERR_SQLITE(rc)  );
    }

    *ppvh = pvh_result;
    pvh_result = NULL;

fail:
    SG_VHASH_NULLFREE(pCtx, pvh_result);
    SG_ERR_IGNORE(  sg_sqlite__nullfinalize(pCtx, &pStmt)  );
}

void SG_blobset__lookup__vhash(
        SG_context * pCtx,
        SG_blobset* pbs,
        const char* psz_hid,
        SG_vhash** ppvh
        )
{
    SG_bool b_found = SG_FALSE;
    char* psz_filename = NULL;
    SG_uint64 offset = 0;
    SG_blob_encoding encoding = 0;
    SG_uint64 len_encoded = 0;
    SG_uint64 len_full = 0;
    char* psz_hid_vcdiff = NULL;
    char* psz_objectid = NULL;
    SG_int64 objectid_timestamp = 0;
    SG_vhash* pvh = NULL;

    SG_ERR_CHECK(  SG_blobset__lookup(
                pCtx,
                pbs,
                psz_hid,
                &b_found,
                &psz_filename,
                &offset,
                &encoding,
                &len_encoded,
                &len_full,
                &psz_hid_vcdiff,
                &psz_objectid,
                &objectid_timestamp
                )  );

    if (b_found)
    {
        SG_ERR_CHECK(  SG_vhash__alloc(pCtx, &pvh)  );

        SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvh, "hid", psz_hid)  );
        if (psz_filename)
        {
            SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvh, "filename", psz_filename)  );
        }
        SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh, "offset", offset)  );
        SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh, "encoding", encoding)  );
        SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh, "len_encoded", len_encoded)  );
        SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh, "len_full", len_full)  );

        if (psz_hid_vcdiff)
        {
            SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvh, "hid_vcdiff", psz_hid_vcdiff)  );
        }
        if (psz_objectid)
        {
            SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvh, "objectid", psz_objectid)  );
        }
        SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh, "objectid_timestamp", objectid_timestamp)  );
    }

    *ppvh = pvh;
    pvh = NULL;

fail:
    SG_VHASH_NULLFREE(pCtx, pvh);
    SG_NULLFREE(pCtx, psz_filename);
    SG_NULLFREE(pCtx, psz_objectid);
    SG_NULLFREE(pCtx, psz_hid_vcdiff);
}

void SG_blobset__get_stats(
        SG_context * pCtx,
        SG_blobset* pbs,
        SG_blob_encoding encoding,
        SG_uint32* p_count,
        SG_uint64* p_len_encoded,
        SG_uint64* p_len_full,
        SG_uint64* p_max_len_encoded,
        SG_uint64* p_max_len_full
        )
{
	sqlite3_stmt * pStmt = NULL;
    SG_uint32 count = 0;
    SG_uint64 len_encoded = 0;
    SG_uint64 len_full = 0;
    SG_uint64 max_len_encoded = 0;
    SG_uint64 max_len_full = 0;

    if (encoding)
    {
        SG_ERR_CHECK(  sg_sqlite__prepare(pCtx, pbs->psql, &pStmt,
                                          "SELECT COUNT(\"hid\"), SUM(\"len_encoded\"), SUM(\"len_full\"), MAX(\"len_encoded\"), MAX(\"len_full\") FROM \"blobs\" WHERE \"encoding\" = ?")  );
        SG_ERR_CHECK(  sg_sqlite__bind_int(pCtx, pStmt,1,encoding)  );
    }
    else
    {
        SG_ERR_CHECK(  sg_sqlite__prepare(pCtx, pbs->psql, &pStmt,
                                          "SELECT COUNT(\"hid\"), SUM(\"len_encoded\"), SUM(\"len_full\"), MAX(\"len_encoded\"), MAX(\"len_full\") FROM \"blobs\"")  );
    }
    SG_ERR_CHECK(  sg_sqlite__step(pCtx, pStmt, SQLITE_ROW)  );

    count = sqlite3_column_int(pStmt, 0);
    len_encoded = sqlite3_column_int64(pStmt, 1);
    len_full = sqlite3_column_int64(pStmt, 2);
    max_len_encoded = sqlite3_column_int64(pStmt, 3);
    max_len_full = sqlite3_column_int64(pStmt, 4);
    SG_ERR_CHECK(  sg_sqlite__nullfinalize(pCtx, &pStmt)  );

    if (p_count)
    {
        *p_count = count;
    }
    if (p_len_encoded)
    {
        *p_len_encoded = len_encoded;
    }
    if (p_len_full)
    {
        *p_len_full = len_full;
    }
    if (p_max_len_encoded)
    {
        *p_max_len_encoded = max_len_encoded;
    }
    if (p_max_len_full)
    {
        *p_max_len_full = max_len_full;
    }

fail:
    SG_ERR_IGNORE(  sg_sqlite__nullfinalize(pCtx, &pStmt)  );
}

void SG_blobset__lookup(
        SG_context * pCtx,
        SG_blobset* pbs,
        const char* psz_hid,
        SG_bool* pb_found,
        char** ppsz_filename,
        SG_uint64* p_offset,
        SG_blob_encoding* p_blob_encoding,
        SG_uint64* p_len_encoded,
        SG_uint64* p_len_full,
        char** ppsz_hid_vcdiff_reference,
        char** ppsz_objectid,
        SG_int64* pi_objectid_timestamp
        )
{
	sqlite3_stmt * pStmt = NULL;
    int rc;

    const char* psz_filename = NULL;
    SG_uint64 offset = 0;
    SG_blob_encoding blob_encoding = 0;
    SG_uint64 len_encoded = 0;
    SG_uint64 len_full = 0;
    const char * psz_hid_vcdiff_reference = NULL;
    const char * psz_objectid = NULL;
    SG_int64 i_objectid_timestamp = -1;

	SG_ERR_CHECK(  sg_sqlite__prepare(pCtx, pbs->psql, &pStmt,
									  "SELECT \"filename\", \"offset\", \"encoding\", \"len_encoded\", \"len_full\", \"hid_vcdiff\", \"objectid\", \"objectid_timestamp\" FROM \"blobs\" WHERE \"hid\" = ?")  );
	SG_ERR_CHECK(  sg_sqlite__bind_text(pCtx, pStmt,1,psz_hid)  );

    rc = sqlite3_step(pStmt);
    if (SQLITE_ROW == rc)
    {
        psz_filename = (const char *)sqlite3_column_text(pStmt,0);
        offset = sqlite3_column_int64(pStmt, 1);

        blob_encoding = (SG_blob_encoding) sqlite3_column_int(pStmt, 2);

        len_encoded = sqlite3_column_int64(pStmt, 3);
        len_full = sqlite3_column_int64(pStmt, 4);

        if (SG_IS_BLOBENCODING_FULL(blob_encoding))
        {
            SG_ASSERT(len_encoded == len_full);
        }

        psz_hid_vcdiff_reference = (const char *)sqlite3_column_text(pStmt,5);

        if (ppsz_hid_vcdiff_reference)
        {
            if (psz_hid_vcdiff_reference)
            {
                SG_ERR_CHECK(  SG_STRDUP(pCtx, psz_hid_vcdiff_reference, ppsz_hid_vcdiff_reference)  );
            }
            else
            {
                *ppsz_hid_vcdiff_reference = NULL;
            }
        }


        psz_objectid = (const char *)sqlite3_column_text(pStmt,6);

        if (ppsz_objectid)
        {
            if (psz_objectid)
            {
                SG_ERR_CHECK(  SG_STRDUP(pCtx, psz_objectid, ppsz_objectid)  );
            }
            else
            {
                *ppsz_objectid = NULL;
            }
        }

        i_objectid_timestamp = sqlite3_column_int64(pStmt, 7);

        if (p_blob_encoding)
        {
            *p_blob_encoding = blob_encoding;
        }
        if (p_len_encoded)
        {
            *p_len_encoded = len_encoded;
        }
        if (p_len_full)
        {
            *p_len_full = len_full;
        }
        if (ppsz_filename)
        {
            if (psz_filename)
            {
                SG_ERR_CHECK(  SG_STRDUP(pCtx, psz_filename, ppsz_filename)  );
            }
            else
            {
                *ppsz_filename = NULL;
            }
        }
        if (p_offset)
        {
            *p_offset = offset;
        }
        if (pi_objectid_timestamp)
        {
            *pi_objectid_timestamp = i_objectid_timestamp;
        }

        SG_ERR_CHECK(  sg_sqlite__step(pCtx, pStmt, SQLITE_DONE)  );

        SG_ERR_CHECK(  sg_sqlite__nullfinalize(pCtx, &pStmt)  );

        *pb_found = SG_TRUE;
    }
    else if (SQLITE_DONE == rc)
    {
        *pb_found = SG_FALSE;
        SG_ERR_IGNORE(  sg_sqlite__nullfinalize(pCtx, &pStmt)  );
    }
    else
    {
        SG_ERR_THROW(  SG_ERR_SQLITE(rc)  );
    }

	return;

fail:
	SG_ERR_IGNORE(  sg_sqlite__nullfinalize(pCtx, &pStmt)  );
}

void SG_blobset__list_blobs_for_objectid(
	SG_context* pCtx,
    SG_blobset* pbs,
	const char* psz_objectid,
	SG_vhash** ppvh
    )
{
	sqlite3_stmt * pStmt = NULL;
	int rc;
    SG_vhash* pvh_results = NULL;

	SG_NULLARGCHECK_RETURN(pbs);

    SG_ERR_CHECK(  SG_vhash__alloc(pCtx, &pvh_results)  );

    // TODO is the following query going to be slow?  perhaps something with
    // a join?  or with groupby?
    SG_ERR_CHECK(  sg_sqlite__exec(pCtx, pbs->psql, "CREATE INDEX IF NOT EXISTS \"blobs_objectid\" ON \"blobs\" (\"objectid\")")  );
    SG_ERR_CHECK(  sg_sqlite__exec(pCtx, pbs->psql, "CREATE INDEX IF NOT EXISTS \"blobs_hid_vcdiff\" ON \"blobs\" (\"hid_vcdiff\")")  );
	SG_ERR_CHECK(  sg_sqlite__prepare(
                pCtx, 
                pbs->psql,
                &pStmt,
                "SELECT \"objectid_timestamp\", \"len_encoded\", \"encoding\", \"len_full\", \"hid_vcdiff\", \"hid\", (SELECT count(*) FROM \"blobs\" q WHERE q.\"hid_vcdiff\"=z.\"hid\") FROM \"blobs\" z WHERE \"objectid\" = ? ORDER BY \"objectid_timestamp\" ASC, \"len_full\" ASC")  );
	SG_ERR_CHECK(  sg_sqlite__bind_text(pCtx, pStmt,1,psz_objectid)  );
	while ((rc = sqlite3_step(pStmt)) == SQLITE_ROW)
	{
        SG_uint64 len_encoded = 0;
        SG_uint64 len_full = 0;
        SG_uint64 timestamp = 0;
        SG_blob_encoding encoding = 0;
        const char * psz_hid_vcdiff_reference = NULL;
        const char * psz_hid = NULL;
        SG_vhash* pvh_rev = NULL;
        SG_int64 count_blobs_referencing_me = 0;

		timestamp = (SG_uint32) sqlite3_column_int(pStmt, 0);
		len_encoded = (SG_uint64) sqlite3_column_int64(pStmt, 1);
		encoding = (SG_blob_encoding) sqlite3_column_int(pStmt, 2);
		len_full = (SG_uint64) sqlite3_column_int64(pStmt, 3);
		psz_hid_vcdiff_reference = (const char *)sqlite3_column_text(pStmt, 4);
		psz_hid = (const char *)sqlite3_column_text(pStmt, 5);
		count_blobs_referencing_me = (SG_int64) sqlite3_column_int64(pStmt, 6);

        SG_ERR_CHECK(  SG_vhash__addnew__vhash(pCtx, pvh_results, psz_hid, &pvh_rev)  );
		SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh_rev, "len_encoded", len_encoded)  );
		SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh_rev, "len_full", len_full)  );
		SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh_rev, "timestamp", timestamp)  );
		SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh_rev, "encoding", encoding)  );
        if (SG_BLOBENCODING__VCDIFF == encoding)
        {
            SG_ASSERT(psz_hid_vcdiff_reference);
        }
        if (psz_hid_vcdiff_reference)
        {
            SG_ASSERT(SG_BLOBENCODING__VCDIFF == encoding);
            SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvh_rev, "hid_vcdiff_reference", psz_hid_vcdiff_reference)  );
        }
		SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh_rev, "count_blobs_referencing_me", count_blobs_referencing_me)  );
	}
    if (rc != SQLITE_DONE)
    {
        SG_ERR_THROW(  SG_ERR_SQLITE(rc)  );
    }

    *ppvh = pvh_results;
    pvh_results = NULL;

fail:
    SG_VHASH_NULLFREE(pCtx, pvh_results);
	SG_ERR_IGNORE(  sg_sqlite__nullfinalize(pCtx, &pStmt)  );
}

void SG_blobset__list_pack_candidates(
	SG_context* pCtx,
    SG_blobset* pbs,
	SG_uint32 nMinRevisionCount,
	SG_varray ** ppvArrayResults
    )
{
	sqlite3_stmt * pStmt = NULL;
	int rc;
	SG_varray * pvaResults = NULL;
	SG_vhash * pvhCurrent = NULL;

	SG_NULLARGCHECK_RETURN(pbs);

    SG_ERR_CHECK(  sg_sqlite__exec(pCtx, pbs->psql, "CREATE INDEX IF NOT EXISTS \"blobs_objectid\" ON \"blobs\" (\"objectid\")")  );

	SG_ERR_CHECK(  SG_VARRAY__ALLOC(pCtx, &pvaResults)  );
    SG_ERR_CHECK(  sg_sqlite__prepare(pCtx, pbs->psql,&pStmt,
                            "SELECT COUNT(\"hid\") n, SUM(\"len_encoded\"), SUM(\"len_full\"), \"objectid\" FROM \"blobs\" WHERE \"objectid\" IS NOT NULL GROUP BY \"objectid\" ORDER BY \"n\" DESC")  );

	rc = sqlite3_step(pStmt);
	while (SQLITE_ROW == rc)
	{
        SG_uint32 count = 0;

		count = (SG_uint32) sqlite3_column_int(pStmt, 0);

        if (count >= nMinRevisionCount)
        {
            SG_uint64 len_encoded = 0;
            SG_uint64 len_full = 0;
            const char * objectid = NULL;

            len_encoded = (SG_uint64) sqlite3_column_int64(pStmt, 1);
            len_full = (SG_uint64) sqlite3_column_int64(pStmt, 2);
            objectid = (const char *)sqlite3_column_text(pStmt, 3);

            SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvhCurrent)  );
            SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvhCurrent, "len_encoded", len_encoded)  );
            SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvhCurrent, "len_full", len_full)  );
            SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvhCurrent, "revision_count", count)  );
            SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhCurrent, "objectid", objectid)  );
            SG_ERR_CHECK(  SG_varray__append__vhash(pCtx, pvaResults, &pvhCurrent)  );
        }

		rc = sqlite3_step(pStmt);
	}
	SG_RETURN_AND_NULL(pvaResults, ppvArrayResults);
	SG_ERR_IGNORE(  sg_sqlite__nullfinalize(pCtx, &pStmt)  );
	return;

fail:
	SG_VARRAY_NULLFREE(pCtx, pvaResults);
	SG_ERR_IGNORE(  sg_sqlite__nullfinalize(pCtx, &pStmt)  );
}

void SG_blobset__inserting(
	SG_context* pCtx,
    SG_blobset* pbs,
    SG_bool* pb
    )
{
	SG_NULLARGCHECK_RETURN(pbs);

    if (pbs->pStmt)
    {
        *pb = SG_TRUE;
    }
    else
    {
        *pb = SG_FALSE;
    }
}

void SG_blobset__count(
	SG_context* pCtx,
    SG_blobset* pbs,
    SG_uint32* pi_count
    )
{
	SG_NULLARGCHECK_RETURN(pbs);

    *pi_count = pbs->count;
}

void SG_blobset__compare_to(
	SG_context* pCtx,
    SG_blobset* pbs_me,
    SG_blobset* pbs_other,
    SG_bool* pb
    )
{
    sqlite3_stmt* pStmt_me = NULL;
    sqlite3_stmt* pStmt_other = NULL;
    int rc_me = 0;
    int rc_other = 0;

	SG_NULLARGCHECK_RETURN(pbs_me);
	SG_NULLARGCHECK_RETURN(pbs_other);
	SG_NULLARGCHECK_RETURN(pb);

    if (pbs_me->count != pbs_other->count)
    {
        *pb = SG_FALSE;
        goto done;
    }

	SG_ERR_CHECK(  sg_sqlite__prepare(pCtx, pbs_me->psql, &pStmt_me,
									  "SELECT \"hid\", \"len_full\" FROM \"blobs\" ORDER BY \"hid\" ASC")  );
	SG_ERR_CHECK(  sg_sqlite__prepare(pCtx, pbs_other->psql, &pStmt_other,
									  "SELECT \"hid\", \"len_full\" FROM \"blobs\" ORDER BY \"hid\" ASC")  );

    while (1)
    {
        rc_me=sqlite3_step(pStmt_me);
        rc_other=sqlite3_step(pStmt_other);

        if (
                (SQLITE_ROW == rc_me)
                && (SQLITE_ROW == rc_other)
           )
        {
            const char* psz_hid_me = (const char *)sqlite3_column_text(pStmt_me,0);
            SG_uint64 len_full_me = sqlite3_column_int64(pStmt_me, 1);

            const char* psz_hid_other = (const char *)sqlite3_column_text(pStmt_other,0);
            SG_uint64 len_full_other = sqlite3_column_int64(pStmt_other, 1);

            if (len_full_me != len_full_other)
            {
                *pb = SG_FALSE;
                goto done;
            }

            if (0 != strcmp(psz_hid_me, psz_hid_other))
            {
                *pb = SG_FALSE;
                goto done;
            }
        }
        else
        {
            break;
        }
    }

    if (
            (SQLITE_ROW == rc_me)
            || (SQLITE_ROW == rc_other)
       )
    {
        *pb = SG_FALSE;
        goto done;
    }

    if (rc_me != SQLITE_DONE)
    {
        SG_ERR_THROW(  SG_ERR_SQLITE(rc_me)  );
    }
    if (rc_other != SQLITE_DONE)
    {
        SG_ERR_THROW(  SG_ERR_SQLITE(rc_other)  );
    }

    *pb = SG_TRUE;

done:
fail:
    SG_ERR_IGNORE(  sg_sqlite__nullfinalize(pCtx, &pStmt_me)  );
    SG_ERR_IGNORE(  sg_sqlite__nullfinalize(pCtx, &pStmt_other)  );
}

