/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include <sg.h>
#include <zlib.h>

#define MY_BUSY_TIMEOUT_MS      (30000)

#define MY_CHUNK_SIZE			(16*1024)

static void _sg_clone__validate_new_repo_name(SG_context* pCtx, const char* psz_new_descriptor_name)
{
	char * szNormalizedDescriptorName = NULL;
	SG_bool alreadyExists = SG_FALSE;
	SG_ERR_CHECK(  SG_repo__validate_repo_name(pCtx, psz_new_descriptor_name, &szNormalizedDescriptorName)  );
	SG_ERR_CHECK(  SG_closet__descriptors__exists(pCtx, szNormalizedDescriptorName, &alreadyExists)  );
	if(alreadyExists)
		SG_ERR_THROW(SG_ERR_REPO_ALREADY_EXISTS);
	SG_NULLFREE(pCtx, szNormalizedDescriptorName);

	return;
fail:
	SG_NULLFREE(pCtx, szNormalizedDescriptorName);
}

struct _sg_clone
{
    SG_blobset* pbs;
	SG_pathname* pPath;
	SG_repo* pRepo;
    SG_vhash* pvh_tempfiles;
    char* psz_fragball_name;
    SG_vhash* pvh_repo_info;

	SG_byte buf1[MY_CHUNK_SIZE];
	SG_byte buf2[MY_CHUNK_SIZE];
};
typedef struct _sg_clone sg_clone;

static void SG_clone__get_blob_full(
	SG_context* pCtx,
	sg_clone* pMe,
    const char* psz_hid,
    SG_pathname** ppPath,
    SG_uint64* pi_offset,
    SG_uint64* pi_len_full
    );

static void sg_clone__free(SG_context* pCtx, SG_clone* pClone)
{
	if (pClone)
	{
		sg_clone* ps = (sg_clone*)pClone;
        SG_NULLFREE(pCtx, ps->psz_fragball_name);
        SG_VHASH_NULLFREE(pCtx, ps->pvh_repo_info);
		SG_PATHNAME_NULLFREE(pCtx, ps->pPath);
		SG_VHASH_NULLFREE(pCtx, ps->pvh_tempfiles);
		SG_REPO_NULLFREE(pCtx, ps->pRepo);
        SG_BLOBSET_NULLFREE(pCtx, ps->pbs);
		SG_NULLFREE(pCtx, ps);
	}
}

static void SG_clone__remove_tempfile(
        SG_context* pCtx,
        sg_clone* pMe,
        const char* psz_hid
        )
{
    SG_bool b_has = SG_FALSE;
    SG_pathname* pPath = NULL;

    SG_ERR_CHECK(  SG_vhash__has(pCtx, pMe->pvh_tempfiles, psz_hid, &b_has)  );
    if (b_has)
    {
        SG_bool b_exists = SG_FALSE;
        const char* psz_name = NULL;
        
        SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pMe->pvh_tempfiles, psz_hid, &psz_name)  );
        SG_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx, &pPath, pMe->pPath, psz_name)  );
        SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pPath, &b_exists, NULL, NULL)  );
        if (b_exists)
        {
            SG_ERR_CHECK(  SG_fsobj__remove__pathname(pCtx, pPath)  );
        }
        SG_ERR_CHECK(  SG_vhash__remove(pCtx, pMe->pvh_tempfiles, psz_hid)  );
    }

fail:
    SG_PATHNAME_NULLFREE(pCtx, pPath);
}

static void SG_clone__get_tempfile_pathname(
        SG_context* pCtx,
        sg_clone* pMe,
        const char* psz_hid,
        SG_pathname** pp,
        SG_bool* pb_exists
        )
{
    SG_bool b_has = SG_FALSE;
    SG_pathname* pPath = NULL;
    char buf[SG_TID_MAX_BUFFER_LENGTH];
    SG_bool b_exists = SG_FALSE;

    if (psz_hid)
    {
        SG_NULLARGCHECK(pb_exists);
        SG_ERR_CHECK(  SG_vhash__has(pCtx, pMe->pvh_tempfiles, psz_hid, &b_has)  );
        if (b_has)
        {
            const char* psz_name = NULL;
            
            SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pMe->pvh_tempfiles, psz_hid, &psz_name)  );
            SG_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx, &pPath, pMe->pPath, psz_name)  );
            SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pPath, &b_exists, NULL, NULL)  );
            goto done;
        }
    }
    else
    {
        SG_ARGCHECK(!pb_exists, pb_exists);
    }

    SG_ERR_CHECK(  SG_tid__generate(pCtx, buf, sizeof(buf))  );
    SG_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx, &pPath, pMe->pPath, buf)  );
    b_exists = SG_FALSE;
    if (psz_hid)
    {
        SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pMe->pvh_tempfiles, psz_hid, buf)  );
    }

done:
    *pp = pPath;
    pPath = NULL;

    if (pb_exists)
    {
        *pb_exists = b_exists;
    }

fail:
    SG_PATHNAME_NULLFREE(pCtx, pPath);
}

struct found_blob
{
    SG_uint64 offset;
    SG_blob_encoding encoding;
    SG_uint64 len_encoded;
    SG_uint64 len_full;
    char objectid[SG_GID_BUFFER_LENGTH];
    SG_int64 objectid_timestamp;
    char hid_vcdiff[SG_HID_MAX_BUFFER_LENGTH];
};

void SG_clone__find_blob(
        SG_context* pCtx,
        sg_clone* pMe,
        const char* psz_hid,
        struct found_blob* pfb
        )
{
    SG_bool b_found = SG_FALSE;
    char* psz_hid_vcdiff = NULL;
    char* psz_objectid = NULL;
    char* psz_filename = NULL;

    SG_ERR_CHECK(  SG_blobset__lookup(
                pCtx, 
                pMe->pbs, 
                psz_hid, 
                &b_found, 
                &psz_filename, 
                &pfb->offset, 
                &pfb->encoding, 
                &pfb->len_encoded, 
                &pfb->len_full, 
                &psz_hid_vcdiff,
                &psz_objectid, 
                &pfb->objectid_timestamp
                )  );

    if (!b_found)
	{
		// blob doesn't exist
		SG_ERR_THROW(SG_ERR_BLOB_NOT_FOUND);
	}

    if (psz_objectid)
    {
        SG_ERR_CHECK(  SG_strcpy(pCtx, pfb->objectid, sizeof(pfb->objectid), psz_objectid)  );
    }
    else
    {
        pfb->objectid[0] = 0;
    }
    if (psz_hid_vcdiff)
    {
        SG_ERR_CHECK(  SG_strcpy(pCtx, pfb->hid_vcdiff, sizeof(pfb->hid_vcdiff), psz_hid_vcdiff)  );
    }
    else
    {
        pfb->hid_vcdiff[0] = 0;
    }

fail:
    SG_NULLFREE(pCtx, psz_filename);
    SG_NULLFREE(pCtx, psz_objectid);
    SG_NULLFREE(pCtx, psz_hid_vcdiff);
}

void SG_clone__list_pack_candidates(
	SG_context* pCtx,
	SG_clone* pClone,
	SG_uint32 nMinRevisionCount,
	SG_varray ** ppvArrayResults
    )
{
	sg_clone* pMe = (sg_clone*)pClone;

    SG_ERR_CHECK_RETURN(  SG_blobset__list_pack_candidates(pCtx, pMe->pbs, nMinRevisionCount, ppvArrayResults)  );
}

void SG_clone__list_revisions_for_object(
	SG_context* pCtx,
	SG_clone* pClone,
	const char * pszgid,
	SG_vhash** ppvh
    )
{
	sg_clone* pMe = (sg_clone*)pClone;

    SG_ERR_CHECK_RETURN(  SG_blobset__list_blobs_for_objectid(pCtx, pMe->pbs, pszgid, ppvh)  );
}

static void _store_one_blob(
	SG_context* pCtx,
	sg_clone* pMe,
	SG_file* pFile,
    const char* psz_objectid,
    SG_int64 i_objectid_timestamp,
	SG_repo_tx_handle* ptx,
	const char* psz_hid,
	SG_blob_encoding encoding,
	const char* psz_hid_vcdiff_reference,
	SG_uint64 len_full,
	SG_uint64 len_encoded
	)
{
	SG_uint32 sofar = 0;
	SG_uint32 got = 0;
	SG_repo_store_blob_handle* pbh = NULL;

	SG_NULLARGCHECK_RETURN(ptx);

	SG_ERR_CHECK(  SG_repo__store_blob__begin(
                pCtx, 
                pMe->pRepo, 
                ptx, 
                psz_objectid, 
                i_objectid_timestamp, 
                encoding, 
                psz_hid_vcdiff_reference, 
                len_full, 
                len_encoded, 
                psz_hid, 
                &pbh)  );

	while (sofar < (SG_uint32) len_encoded)
	{
		SG_uint32 want = 0;
		if ((len_encoded - sofar) > sizeof(pMe->buf1))
		{
			want = sizeof(pMe->buf1);
		}
		else
		{
			want = (SG_uint32) (len_encoded - sofar);
		}
		SG_ERR_CHECK(  SG_file__read(pCtx, pFile, want, pMe->buf1, &got)  );
		SG_ERR_CHECK(  SG_repo__store_blob__chunk(pCtx, pMe->pRepo, pbh, got, pMe->buf1, NULL)  );

		sofar += got;
	}

	SG_ERR_CHECK(  SG_repo__store_blob__end(pCtx, pMe->pRepo, ptx, &pbh, NULL)  );

	return;

fail:
	if (pbh)
		SG_ERR_IGNORE(  SG_repo__store_blob__abort(pCtx, pMe->pRepo, ptx, &pbh)  );
}

static void _store_one_blob__pathname(
	SG_context* pCtx,
	sg_clone* pMe,
    SG_pathname* pPath,
    const char* psz_objectid,
    SG_int64 i_objectid_timestamp,
	SG_repo_tx_handle* ptx,
	const char* psz_hid,
	SG_blob_encoding encoding,
	const char* psz_hid_vcdiff_reference,
	SG_uint64 len_full
	)
{
    SG_file* pFile = NULL;
    SG_uint64 len_encoded = 0;

    SG_ERR_CHECK(  SG_fsobj__length__pathname(pCtx, pPath, &len_encoded, NULL)  );

    if (SG_IS_BLOBENCODING_FULL(encoding))
    {
        SG_ASSERT(len_full == len_encoded);
    }

    SG_ERR_CHECK(  SG_file__open__pathname(pCtx, pPath, SG_FILE_RDONLY|SG_FILE_OPEN_EXISTING, SG_FSOBJ_PERMS__UNUSED, &pFile)  );
    SG_ERR_CHECK(  _store_one_blob(
        pCtx,
        pMe,
        pFile,
        psz_objectid,
        i_objectid_timestamp,
        ptx,
        psz_hid,
        encoding,
        psz_hid_vcdiff_reference,
        len_full,
        len_encoded
        )  );
    SG_ERR_CHECK(  SG_file__close(pCtx, &pFile)  );

fail:
    if (pFile)
    {
        SG_ERR_IGNORE(  SG_file__close(pCtx, &pFile)  );
    }
}

static void _save_to_tempfile__zlib_to_full(
	SG_context* pCtx,
	sg_clone* pMe,
	SG_file* pFile_zlib,
	const char* psz_hid,
	SG_uint64 len_encoded,
	SG_uint64 len_full,
    SG_pathname** ppPath
	)
{
	SG_uint64 sofar_encoded = 0;
	SG_uint64 sofar_full = 0;
    SG_pathname* pPath = NULL;
    int zError = -1;
	z_stream zStream;
    SG_file* pFile_full = NULL;
    SG_bool b_done = SG_FALSE;
    SG_bool b_already = SG_FALSE;

    SG_ERR_CHECK(  SG_clone__get_tempfile_pathname(pCtx, pMe, psz_hid, &pPath, &b_already)  );

    if (b_already)
    {
        SG_uint64 off = 0;

		SG_ERR_CHECK(  SG_file__tell(pCtx, pFile_zlib, &off)  );
        SG_ERR_CHECK(  SG_file__seek(pCtx, pFile_zlib, off + len_encoded)  );
    }
    else
    {
        SG_ERR_CHECK(  SG_file__open__pathname(pCtx, pPath, SG_FILE_WRONLY|SG_FILE_CREATE_NEW, 0644, &pFile_full)  );

        memset(&zStream,0,sizeof(zStream));
        zError = inflateInit(&zStream);
        if (zError != Z_OK)
        {
            SG_ERR_THROW(  SG_ERR_ZLIB(zError)  );
        }

        while (!b_done)
        {
            SG_uint32 got_full = 0;

            if (0 == zStream.avail_in)
            {
                SG_uint32 want = 0;
                SG_uint32 got_encoded = 0;

                if ((len_encoded - sofar_encoded) > sizeof(pMe->buf1))
                {
                    want = sizeof(pMe->buf1);
                }
                else
                {
                    want = (SG_uint32) (len_encoded - sofar_encoded);
                }
                SG_ERR_CHECK(  SG_file__read(pCtx, pFile_zlib, want, pMe->buf1, &got_encoded)  );

                zStream.next_in = pMe->buf1;
                zStream.avail_in = got_encoded;

                sofar_encoded += got_encoded;
            }

            zStream.next_out = pMe->buf2;
            zStream.avail_out = sizeof(pMe->buf2);

            while (1)
            {
                // let decompressor decompress what it can of our input.  it may or
                // may not take all of it.  this will update next_in, avail_in,
                // next_out, and avail_out.
                //
                // WARNING: we get Z_STREAM_END when the decompressor is finished
                // WARNING: processing all of the input.  it gives this immediately
                // WARNING: (with the last partial buffer) rather than returning OK
                // WARNING: for the last partial buffer and then returning END with
                // WARNING: zero bytes transferred.

                zError = inflate(&zStream,Z_NO_FLUSH);
                if ((zError != Z_OK)  &&  (zError != Z_STREAM_END))
                {
                    SG_ERR_THROW(  SG_ERR_ZLIB(zError)  );
                }

                if (zError == Z_STREAM_END)
                {
                    b_done = SG_TRUE;
                    break;
                }

                // if decompressor didn't take all of our input, let it continue
                // with the existing buffer (next_in was advanced).

                if (zStream.avail_in == 0)
                {
                    break;
                }

                if (zStream.avail_out == 0)
                {
                    /* no room left */
                    break;
                }
            }

            if (zStream.avail_out < sizeof(pMe->buf2))
            {
                got_full = sizeof(pMe->buf2) - zStream.avail_out;
            }
            else
            {
                got_full = 0;
            }

            if (got_full)
            {
                SG_ERR_CHECK(  SG_file__write(pCtx, pFile_full, got_full, (SG_byte*) pMe->buf2, NULL)  );
                sofar_full += got_full;
            }
        }

        if (sofar_full != len_full)
        {
            SG_ERR_THROW(  SG_ERR_INCOMPLETEWRITE  );
        }
        if (sofar_encoded != len_encoded)
        {
            SG_ERR_THROW(  SG_ERR_INCOMPLETEWRITE  );
        }

        SG_ASSERT( (zStream.avail_in == 0) );

        inflateEnd(&zStream);

        SG_ERR_CHECK(  SG_file__close(pCtx, &pFile_full)  );
    }

    *ppPath = pPath;
    pPath = NULL;

fail:
    if (pFile_full)
    {
        SG_ERR_IGNORE(  SG_file__close(pCtx, &pFile_full)  );
    }
    SG_PATHNAME_NULLFREE(pCtx, pPath);
}

static void _save_to_tempfile__full_to_vcdiff(
	SG_context* pCtx,
	sg_clone* pMe,
	SG_file* pFile,
	SG_uint64 len_full,
    SG_pathname* pPath_ref,
    SG_uint64 offset_ref,
    SG_uint64 len_full_ref,
    SG_pathname** ppPath_temp
	)
{
    SG_seekreader* psr_reference = NULL;
    SG_readstream* pstrm_target = NULL;
    SG_writestream* pstrm_delta = NULL;
    SG_pathname* pPath = NULL;

    SG_UNUSED(len_full_ref);

    SG_ERR_CHECK(  SG_clone__get_tempfile_pathname(pCtx, pMe, NULL, &pPath, NULL)  );

    SG_ERR_CHECK(  SG_seekreader__alloc__for_file(pCtx, pPath_ref, offset_ref, &psr_reference)  );

    SG_ERR_CHECK(  SG_readstream__alloc(
					   pCtx,
					   pFile,
					   (SG_stream__func__read*) SG_file__read,
                       NULL,
                       len_full,
					   &pstrm_target)  );

    SG_ERR_CHECK(  SG_writestream__alloc__for_file(pCtx, pPath, &pstrm_delta)  );

    SG_ERR_CHECK(  SG_vcdiff__deltify__streams(pCtx, psr_reference, pstrm_target, pstrm_delta)  );

    SG_ERR_CHECK(  SG_readstream__close(pCtx, pstrm_target)  );
    pstrm_target = NULL;
    SG_ERR_CHECK(  SG_writestream__close(pCtx, pstrm_delta)  );
    pstrm_delta = NULL;
    SG_ERR_CHECK(  SG_seekreader__close(pCtx, psr_reference)  );
    psr_reference = NULL;

    *ppPath_temp = pPath;
    pPath = NULL;

fail:
    if (psr_reference)
    {
        SG_ERR_IGNORE(  SG_seekreader__close(pCtx, psr_reference)  );
    }
}

static void _save_to_tempfile__full_to_zlib(
	SG_context* pCtx,
	sg_clone* pMe,
	SG_file* pFile,
	SG_uint64 len_full,
    SG_pathname** ppPath_temp
	)
{
	SG_uint32 sofar = 0;
	SG_uint32 got = 0;
	z_stream zStream;
    int zError = -1;
    SG_pathname* pPath = NULL;
    SG_file* pFile_temp = NULL;

    SG_ERR_CHECK(  SG_clone__get_tempfile_pathname(pCtx, pMe, NULL, &pPath, NULL)  );

    // init the compressor
    memset(&zStream,0,sizeof(zStream));
    zError = deflateInit(&zStream,Z_DEFAULT_COMPRESSION);
    if (zError != Z_OK)
    {
        SG_ERR_THROW(  SG_ERR_ZLIB(zError)  );
    }

    SG_ERR_CHECK(  SG_file__open__pathname(pCtx, pPath, SG_FILE_WRONLY|SG_FILE_CREATE_NEW, 0644, &pFile_temp)  );

	while (sofar < (SG_uint32) len_full)
	{
		SG_uint32 want = 0;
		if ((len_full - sofar) > sizeof(pMe->buf1))
		{
			want = sizeof(pMe->buf1);
		}
		else
		{
			want = (SG_uint32) (len_full - sofar);
		}
		SG_ERR_CHECK(  SG_file__read(pCtx, pFile, want, pMe->buf1, &got)  );

        // give this chunk to compressor (it will update next_in and avail_in as
        // it consumes the input).

        zStream.next_in = (SG_byte*) pMe->buf1;
        zStream.avail_in = got;

        while (1)
        {
            // give the compressor the complete output buffer

            zStream.next_out = pMe->buf2;
            zStream.avail_out = sizeof(pMe->buf2);

            // let compressor compress what it can of our input.  it may or
            // may not take all of it.  this will update next_in, avail_in,
            // next_out, and avail_out.

            zError = deflate(&zStream,Z_NO_FLUSH);
            if (zError != Z_OK)
            {
                SG_ERR_THROW(  SG_ERR_ZLIB(zError)  );
            }

            // if there was enough input to generate a compression block,
            // we can write it to our output file.  the amount generated
            // is the delta of avail_out (or equally of next_out).

            if (zStream.avail_out < sizeof(pMe->buf2))
            {
                SG_uint32 nOut = sizeof(pMe->buf2) - zStream.avail_out;
                SG_ASSERT ( (nOut == (SG_uint32)(zStream.next_out - pMe->buf2)) );

                SG_ERR_CHECK(  SG_file__write(pCtx, pFile_temp, nOut, (SG_byte*) pMe->buf2, NULL)  );
            }

            // if compressor didn't take all of our input, let it
            // continue with the existing buffer (next_in was advanced).

            if (zStream.avail_in == 0)
                break;
        }

		sofar += got;
	}

    // we reached end of the input file.  now we need to tell the
    // compressor that we have no more input and that it needs to
    // compress and flush any partial buffers that it may have.

    SG_ASSERT( (zStream.avail_in == 0) );
    while (1)
    {
        zStream.next_out = pMe->buf2;
        zStream.avail_out = sizeof(pMe->buf2);

        zError = deflate(&zStream,Z_FINISH);
        if ((zError != Z_OK) && (zError != Z_STREAM_END))
        {
            SG_ERR_THROW(  SG_ERR_ZLIB(zError)  );
        }

        if (zStream.avail_out < sizeof(pMe->buf2))
        {
            SG_uint32 nOut = sizeof(pMe->buf2) - zStream.avail_out;
            SG_ASSERT ( (nOut == (SG_uint32)(zStream.next_out - pMe->buf2)) );

            SG_ERR_CHECK(  SG_file__write(pCtx, pFile_temp, nOut, (SG_byte*) pMe->buf2, NULL)  );
        }

        if (zError == Z_STREAM_END)
            break;
    }

    SG_ERR_CHECK(  SG_file__close(pCtx, &pFile_temp)  );

	deflateEnd(&zStream);

    *ppPath_temp = pPath;
    pPath = NULL;

fail:
    if (pFile_temp)
    {
        SG_ERR_IGNORE(  SG_file__close(pCtx, &pFile_temp)  );
    }
}

static void _save_to_tempfile__full_to_vcdiff__pathname(
	SG_context* pCtx,
	sg_clone* pMe,
    SG_pathname* pPath_full,
    SG_pathname* pPath_ref,
    SG_uint64 offset_ref,
    SG_uint64 len_full_ref,
    SG_pathname** ppPath_vcdiff
	)
{
    SG_file* pFile = NULL;
    SG_pathname* pPath_temp_vcdiff = NULL;
    SG_uint64 len_full = 0;

    SG_ERR_CHECK(  SG_fsobj__length__pathname(pCtx, pPath_full, &len_full, NULL)  );

    SG_ERR_CHECK(  SG_file__open__pathname(pCtx, pPath_full, SG_FILE_RDONLY|SG_FILE_OPEN_EXISTING, SG_FSOBJ_PERMS__UNUSED, &pFile)  );
    SG_ERR_CHECK(  _save_to_tempfile__full_to_vcdiff(
                pCtx,
                pMe,
                pFile,
                len_full,
                pPath_ref,
                offset_ref,
                len_full_ref,
                &pPath_temp_vcdiff
                )  );
    SG_ERR_CHECK(  SG_file__close(pCtx, &pFile)  );

    *ppPath_vcdiff = pPath_temp_vcdiff;
    pPath_temp_vcdiff = NULL;

fail:
    if (pFile)
    {
        SG_ERR_IGNORE(  SG_file__close(pCtx, &pFile)  );
    }
}

static void _save_to_tempfile__full_to_zlib__pathname(
	SG_context* pCtx,
	sg_clone* pMe,
    SG_pathname* pPath_full,
    SG_pathname** ppPath_zlib
	)
{
    SG_file* pFile = NULL;
    SG_pathname* pPath_temp_zlib = NULL;
    SG_uint64 len_full = 0;

    SG_ERR_CHECK(  SG_fsobj__length__pathname(pCtx, pPath_full, &len_full, NULL)  );

    SG_ERR_CHECK(  SG_file__open__pathname(pCtx, pPath_full, SG_FILE_RDONLY|SG_FILE_OPEN_EXISTING, SG_FSOBJ_PERMS__UNUSED, &pFile)  );
    SG_ERR_CHECK(  _save_to_tempfile__full_to_zlib(
                pCtx,
                pMe,
                pFile,
                len_full,
                &pPath_temp_zlib
                )  );
    SG_ERR_CHECK(  SG_file__close(pCtx, &pFile)  );

    *ppPath_zlib = pPath_temp_zlib;
    pPath_temp_zlib = NULL;

fail:
    if (pFile)
    {
        SG_ERR_IGNORE(  SG_file__close(pCtx, &pFile)  );
    }
}

static void _save_to_tempfile__vcdiff_to_full(
	SG_context* pCtx,
	sg_clone* pMe,
	SG_file* pFile,
    const char* psz_hid,
    SG_uint64 len_encoded,
    SG_uint64 len_full,
    SG_pathname* pPath_ref,
    SG_uint64 offset_ref,
    SG_uint64 len_full_ref,
    SG_pathname** ppPath_temp
    )
{
    SG_seekreader* psr_reference = NULL;
    SG_readstream* pstrm_delta = NULL;
    SG_file* pFile_full = NULL;
    SG_bool b_done = SG_FALSE;
    SG_pathname* pPath = NULL;
    SG_vcdiff_undeltify_state* pst = NULL;
    SG_uint64 sofar_full = 0;
    SG_bool b_already = SG_FALSE;

    SG_UNUSED(len_full_ref);

    SG_ERR_CHECK(  SG_clone__get_tempfile_pathname(pCtx, pMe, psz_hid, &pPath, &b_already)  );
    if (b_already)
    {
        SG_uint64 off = 0;

		SG_ERR_CHECK(  SG_file__tell(pCtx, pFile, &off)  );
        SG_ERR_CHECK(  SG_file__seek(pCtx, pFile, off + len_encoded)  );
    }
    else
    {
        SG_ERR_CHECK(  SG_file__open__pathname(pCtx, pPath, SG_FILE_WRONLY|SG_FILE_CREATE_NEW, 0644, &pFile_full)  );

        SG_ERR_CHECK(  SG_seekreader__alloc__for_file(pCtx, pPath_ref, offset_ref, &psr_reference)  );

        SG_ERR_CHECK(  SG_readstream__alloc(
                           pCtx,
                           pFile,
                           (SG_stream__func__read*) SG_file__read,
                           NULL,
                           len_encoded,
                           &pstrm_delta)  );

        /* Now begin the undeltify operation */
        SG_ERR_CHECK(  SG_vcdiff__undeltify__begin(pCtx, psr_reference, pstrm_delta, &pst)  );
        SG_ASSERT(pst);

        while (!b_done)
        {
            SG_uint32 got_full = 0;
            SG_byte* p_buf = NULL;

            SG_ERR_CHECK(  SG_vcdiff__undeltify__chunk(pCtx, pst, &p_buf, &got_full)  );
            if (
                    !p_buf
                    || (0 == got_full)
               )
            {
                b_done = SG_TRUE;
            }

            if (got_full)
            {
                SG_ERR_CHECK(  SG_file__write(pCtx, pFile_full, got_full, (SG_byte*) p_buf, NULL)  );
                sofar_full += got_full;
            }
        }

        // end

        SG_ERR_CHECK(  SG_vcdiff__undeltify__end(pCtx, pst)  );
        pst = NULL;

        SG_ERR_CHECK(  SG_readstream__close(pCtx, pstrm_delta)  );
        pstrm_delta = NULL;

        SG_ERR_CHECK(  SG_seekreader__close(pCtx, psr_reference)  );
        psr_reference = NULL;

        SG_ERR_CHECK(  SG_file__close(pCtx, &pFile_full)  );

        if (sofar_full != len_full)
        {
            SG_ERR_THROW(  SG_ERR_INCOMPLETEWRITE  );
        }
    }

    *ppPath_temp = pPath;
    pPath = NULL;

fail:
    if (pFile_full)
    {
        SG_ERR_IGNORE(  SG_file__close(pCtx, &pFile_full)  );
    }
    SG_PATHNAME_NULLFREE(pCtx, pPath);
}

static void SG_clone__get_blob_full(
	SG_context* pCtx,
	sg_clone* pMe,
    const char* psz_hid,
    SG_pathname** ppPath,
    SG_uint64* pi_offset,
    SG_uint64* pi_len_full
    )
{
    SG_bool b_has = SG_FALSE;
    SG_pathname* pPath_data = NULL;
    SG_pathname* pPath_temp = NULL;
    SG_pathname* pPath_ref = NULL;
    SG_file* pFile_data = NULL;

    SG_ERR_CHECK(  SG_vhash__has(pCtx, pMe->pvh_tempfiles, psz_hid, &b_has)  );
    if (b_has)
    {
        const char* psz_filename = NULL;
        SG_uint64 len = 0;

        SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pMe->pvh_tempfiles, psz_hid, &psz_filename)  );
        SG_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx, &pPath_temp, pMe->pPath, psz_filename)  );

        SG_ERR_CHECK(  SG_fsobj__length__pathname(pCtx, pPath_temp, &len, NULL)  );

        *ppPath = pPath_temp;
        pPath_temp = NULL;
        *pi_offset = 0;
        *pi_len_full = len;

        goto done;
    }
    else
    {
        struct found_blob fb;

        memset(&fb, 0, sizeof(fb));

        SG_ERR_CHECK(  SG_clone__find_blob(pCtx, pMe, psz_hid, &fb)  );
        SG_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx, &pPath_data, pMe->pPath, pMe->psz_fragball_name)  );

        if (SG_IS_BLOBENCODING_FULL(fb.encoding))
        {
            *ppPath = pPath_data;
            pPath_data = NULL;
            *pi_offset = fb.offset;
            *pi_len_full = fb.len_full;

            goto done;
        }
        else if (SG_IS_BLOBENCODING_ZLIB(fb.encoding))
        {
			SG_ERR_CHECK(  SG_file__open__pathname(pCtx, pPath_data, SG_FILE_RDONLY|SG_FILE_OPEN_EXISTING, SG_FSOBJ_PERMS__UNUSED, &pFile_data)  );
            SG_ERR_CHECK(  SG_file__seek(pCtx, pFile_data, fb.offset)  );
            SG_ERR_CHECK(  _save_to_tempfile__zlib_to_full(
                        pCtx,
                        pMe,
                        pFile_data,
                        psz_hid,
                        fb.len_encoded,
                        fb.len_full,
                        &pPath_temp
                        )  );
            SG_ERR_CHECK(  SG_file__close(pCtx, &pFile_data)  );

            *ppPath = pPath_temp;
            pPath_temp = NULL;
            *pi_offset = 0;
            *pi_len_full = fb.len_full;

            goto done;
        }
        else if (SG_IS_BLOBENCODING_VCDIFF(fb.encoding))
        {
            SG_uint64 offset_ref = 0;
            SG_uint64 len_full_ref = 0;

            SG_ERR_CHECK(  SG_clone__get_blob_full(
                        pCtx,
                        pMe,
                        fb.hid_vcdiff,
                        &pPath_ref,
                        &offset_ref,
                        &len_full_ref
                        )  );
			SG_ERR_CHECK(  SG_file__open__pathname(pCtx, pPath_data, SG_FILE_RDONLY|SG_FILE_OPEN_EXISTING, SG_FSOBJ_PERMS__UNUSED, &pFile_data)  );
            SG_ERR_CHECK(  SG_file__seek(pCtx, pFile_data, fb.offset)  );
            SG_ERR_CHECK(  _save_to_tempfile__vcdiff_to_full(
                        pCtx,
                        pMe,
                        pFile_data,
                        psz_hid,
                        fb.len_encoded,
                        fb.len_full,
                        pPath_ref,
                        offset_ref,
                        len_full_ref,
                        &pPath_temp
                        )  );
            SG_ERR_CHECK(  SG_file__close(pCtx, &pFile_data)  );
            SG_PATHNAME_NULLFREE(pCtx, pPath_ref);

            *ppPath = pPath_temp;
            pPath_temp = NULL;
            *pi_offset = 0;
            *pi_len_full = fb.len_full;

            goto done;
        }
        else
        {
            SG_ERR_THROW(SG_ERR_NOTIMPLEMENTED);
        }
    }

done:
fail:
    SG_PATHNAME_NULLFREE(pCtx, pPath_temp);
    SG_PATHNAME_NULLFREE(pCtx, pPath_data);
    SG_PATHNAME_NULLFREE(pCtx, pPath_ref);
    if (pFile_data)
    {
        SG_ERR_IGNORE(  SG_file__close(pCtx, &pFile_data)  );
    }
}

static void _convert_one_blob(
	SG_context* pCtx,
	sg_clone* pMe,
	SG_file* pFile_given,
    const char* psz_objectid,
    SG_int64 i_objectid_timestamp,
	SG_repo_tx_handle* ptx,
	const char* psz_hid,
	SG_blob_encoding encoding_CURRENT,
	const char* psz_hid_vcdiff_reference,
	SG_uint64 len_full,
	SG_uint64 len_encoded,
	SG_blob_encoding encoding_DESIRED,
	const char* psz_hid_vcdiff_reference_DESIRED,
    SG_bool* pb_vcdiff_result_was_ok,
    SG_uint32 demand_vcdiff_savings_over_full,
    SG_uint32 demand_vcdiff_savings_over_zlib,
    SG_uint32 demand_zlib_savings_over_full
	)
{
    SG_pathname* pPath_temp_full = NULL;
    SG_pathname* pPath_temp_zlib = NULL;
    SG_pathname* pPath_ref = NULL;
    SG_pathname* pPath_temp_vcdiff = NULL;
    SG_uint64 len_encoded_zlib = 0;
    SG_uint64 remember_the_offset = 0;

    SG_ARGCHECK_RETURN( !SG_ARE_BLOBENCODINGS_BASICALLY_EQUAL(encoding_CURRENT, encoding_DESIRED) || SG_IS_BLOBENCODING_VCDIFF(encoding_DESIRED), encoding_DESIRED );

    SG_ARGCHECK_RETURN( ( (demand_vcdiff_savings_over_full <= 100)), demand_vcdiff_savings_over_full);
    SG_ARGCHECK_RETURN( ( (demand_vcdiff_savings_over_zlib <= 100)), demand_vcdiff_savings_over_zlib);
    SG_ARGCHECK_RETURN( ( (demand_zlib_savings_over_full <= 100)), demand_zlib_savings_over_full);

    SG_ARGCHECK_RETURN(
            (
             SG_IS_BLOBENCODING_FULL(encoding_CURRENT)
            || SG_IS_BLOBENCODING_ZLIB(encoding_CURRENT)
            || SG_IS_BLOBENCODING_VCDIFF(encoding_CURRENT)
            ), 
            encoding_CURRENT
            );
    SG_ARGCHECK_RETURN(
            (
             SG_IS_BLOBENCODING_FULL(encoding_DESIRED)
            || SG_IS_BLOBENCODING_ZLIB(encoding_DESIRED)
            || SG_IS_BLOBENCODING_VCDIFF(encoding_DESIRED)
            ), 
            encoding_DESIRED
            );

    SG_ERR_CHECK(  SG_file__tell(pCtx, pFile_given, &remember_the_offset)  );

    if (SG_IS_BLOBENCODING_FULL(encoding_CURRENT))
    {
        SG_ASSERT(len_full == len_encoded);
        SG_ASSERT(!psz_hid_vcdiff_reference);

        // We'll need the file in zlib form no matter what, so convert it.
        SG_ERR_CHECK(  _save_to_tempfile__full_to_zlib(
                    pCtx,
                    pMe,
                    pFile_given,
                    len_full,
                    &pPath_temp_zlib
                    )  );
        SG_ERR_CHECK(  SG_fsobj__length__pathname(pCtx, pPath_temp_zlib, &len_encoded_zlib, NULL)  );
    }
    else 
    {
        if (SG_IS_BLOBENCODING_ZLIB(encoding_CURRENT))
        {
            len_encoded_zlib = len_encoded;
            SG_ERR_CHECK(  _save_to_tempfile__zlib_to_full(
                        pCtx,
                        pMe,
                        pFile_given,
                        psz_hid,
                        len_encoded,
                        len_full,
                        &pPath_temp_full
                        )  );
        }
        else if (SG_IS_BLOBENCODING_VCDIFF(encoding_CURRENT))
        {
            SG_uint64 offset_ref = 0;
            SG_uint64 len_full_ref = 0;

            SG_ERR_CHECK(  SG_clone__get_blob_full(
                        pCtx,
                        pMe,
                        psz_hid_vcdiff_reference,
                        &pPath_ref,
                        &offset_ref,
                        &len_full_ref
                        )  );
            SG_ERR_CHECK(  _save_to_tempfile__vcdiff_to_full(
                        pCtx,
                        pMe,
                        pFile_given,
                        psz_hid,
                        len_encoded,
                        len_full,
                        pPath_ref,
                        offset_ref,
                        len_full_ref,
                        &pPath_temp_full
                        )  );
            SG_PATHNAME_NULLFREE(pCtx, pPath_ref);

            if (!SG_IS_BLOBENCODING_FULL(encoding_DESIRED))
            {
                SG_ERR_CHECK(  _save_to_tempfile__full_to_zlib__pathname(
                            pCtx,
                            pMe,
                            pPath_temp_full,
                            &pPath_temp_zlib
                            )  );
                SG_ERR_CHECK(  SG_fsobj__length__pathname(pCtx, pPath_temp_zlib, &len_encoded_zlib, NULL)  );
            }
        }
    }

    if (SG_IS_BLOBENCODING_FULL(encoding_DESIRED))
    {
        SG_ERR_CHECK(  _store_one_blob__pathname(
            pCtx,
            pMe,
            pPath_temp_full,
            psz_objectid,
            i_objectid_timestamp,
            ptx,
            psz_hid,
            encoding_DESIRED,
            NULL,
            len_full
            )  );
    }
    else if (SG_IS_BLOBENCODING_ZLIB(encoding_DESIRED))
    {
        SG_bool b_zlib = SG_TRUE;

        // decide whether to use zlib or not
        if (demand_zlib_savings_over_full)
        {
            SG_uint64 threshold = len_full - (len_full * demand_zlib_savings_over_full / 100);
            if (len_encoded_zlib > threshold)
            {
                b_zlib = SG_FALSE;
            }
        }

        if (b_zlib)
        {
            SG_ERR_CHECK(  _store_one_blob__pathname(
                pCtx,
                pMe,
                pPath_temp_zlib,
                psz_objectid,
                i_objectid_timestamp,
                ptx,
                psz_hid,
                encoding_DESIRED,
                NULL,
                len_full
                )  );
        }
        else
        {
            // store it full

            if (SG_IS_BLOBENCODING_FULL(encoding_CURRENT))
            {
                SG_ERR_CHECK(  SG_file__seek(pCtx, pFile_given, remember_the_offset)  );
                SG_ERR_CHECK(  _store_one_blob(
                    pCtx,
                    pMe,
                    pFile_given,
                    psz_objectid,
                    i_objectid_timestamp,
                    ptx,
                    psz_hid,
                    SG_BLOBENCODING__KEEPFULLFORNOW,
                    NULL,
                    len_full,
                    len_full
                    )  );
            }
            else
            {
                SG_ERR_CHECK(  _store_one_blob__pathname(
                    pCtx,
                    pMe,
                    pPath_temp_full,
                    psz_objectid,
                    i_objectid_timestamp,
                    ptx,
                    psz_hid,
                    SG_BLOBENCODING__KEEPFULLFORNOW,
                    NULL,
                    len_full
                    )  );
            }
        }
    }
    else if (SG_IS_BLOBENCODING_VCDIFF(encoding_DESIRED))
    {
        SG_uint64 offset_ref = 0;
        SG_uint64 len_full_ref = 0;
        SG_uint64 len_encoded_vcdiff = 0;
        SG_bool b_keep_the_vcdiff_result = SG_TRUE;

        SG_ASSERT(psz_hid_vcdiff_reference_DESIRED);

        SG_ERR_CHECK(  SG_clone__get_blob_full(
                    pCtx,
                    pMe,
                    psz_hid_vcdiff_reference_DESIRED,
                    &pPath_ref,
                    &offset_ref,
                    &len_full_ref
                    )  );
        if (SG_IS_BLOBENCODING_FULL(encoding_CURRENT))
        {
            SG_ERR_CHECK(  SG_file__seek(pCtx, pFile_given, remember_the_offset)  );
            SG_ERR_CHECK(  _save_to_tempfile__full_to_vcdiff(
                        pCtx,
                        pMe,
                        pFile_given,
                        len_full,
                        pPath_ref,
                        offset_ref,
                        len_full_ref,
                        &pPath_temp_vcdiff
                        )  );
        }
        else
        {
            SG_ERR_CHECK(  _save_to_tempfile__full_to_vcdiff__pathname(
                        pCtx,
                        pMe,
                        pPath_temp_full,
                        pPath_ref,
                        offset_ref,
                        len_full_ref,
                        &pPath_temp_vcdiff
                        )  );
        }
        SG_ERR_CHECK(  SG_fsobj__length__pathname(pCtx, pPath_temp_vcdiff, &len_encoded_vcdiff, NULL)  );

        b_keep_the_vcdiff_result = SG_TRUE;

        SG_ASSERT(len_encoded_zlib);
        if (len_encoded_zlib < len_full)
        {
            if (demand_vcdiff_savings_over_zlib)
            {
                SG_uint64 threshold = len_encoded_zlib - (len_encoded_zlib * demand_vcdiff_savings_over_zlib / 100);
                if (len_encoded_vcdiff > threshold)
                {
                    b_keep_the_vcdiff_result = SG_FALSE;
                }
            }
        }

        if (b_keep_the_vcdiff_result)
        {
            if (demand_vcdiff_savings_over_full)
            {
                SG_uint64 threshold = len_full - (len_full * demand_vcdiff_savings_over_full / 100);

                if (len_encoded_vcdiff > threshold)
                {
                    b_keep_the_vcdiff_result = SG_FALSE;
                }
            }
        }

        if (b_keep_the_vcdiff_result)
        {
            SG_ERR_CHECK(  _store_one_blob__pathname(
                pCtx,
                pMe,
                pPath_temp_vcdiff,
                psz_objectid,
                i_objectid_timestamp,
                ptx,
                psz_hid,
                encoding_DESIRED,
                psz_hid_vcdiff_reference_DESIRED,
                len_full
                )  );
            *pb_vcdiff_result_was_ok = SG_TRUE;
        }
        else
        {
            SG_bool b_zlib = SG_TRUE;

            // decide whether to use zlib or not
            if (demand_zlib_savings_over_full)
            {
                SG_uint64 threshold = len_full - (len_full * demand_zlib_savings_over_full / 100);
                if (len_encoded_zlib > threshold)
                {
                    b_zlib = SG_FALSE;
                }
            }

            if (b_zlib)
            {
                // store it ZLIB

                if (SG_IS_BLOBENCODING_ZLIB(encoding_CURRENT))
                {
                    SG_ERR_CHECK(  SG_file__seek(pCtx, pFile_given, remember_the_offset)  );
                    SG_ERR_CHECK(  _store_one_blob(
                        pCtx,
                        pMe,
                        pFile_given,
                        psz_objectid,
                        i_objectid_timestamp,
                        ptx,
                        psz_hid,
                        SG_BLOBENCODING__ZLIB,
                        NULL,
                        len_full,
                        len_encoded_zlib
                        )  );
                }
                else
                {
                    SG_ERR_CHECK(  _store_one_blob__pathname(
                        pCtx,
                        pMe,
                        pPath_temp_zlib,
                        psz_objectid,
                        i_objectid_timestamp,
                        ptx,
                        psz_hid,
                        SG_BLOBENCODING__ZLIB,
                        NULL,
                        len_full
                        )  );
                }
            }
            else
            {
                // store it FULL

                if (SG_IS_BLOBENCODING_FULL(encoding_CURRENT))
                {
                    SG_ERR_CHECK(  SG_file__seek(pCtx, pFile_given, remember_the_offset)  );
                    SG_ERR_CHECK(  _store_one_blob(
                        pCtx,
                        pMe,
                        pFile_given,
                        psz_objectid,
                        i_objectid_timestamp,
                        ptx,
                        psz_hid,
                        SG_BLOBENCODING__KEEPFULLFORNOW,
                        NULL,
                        len_full,
                        len_full
                        )  );
                }
                else
                {
                    SG_ERR_CHECK(  _store_one_blob__pathname(
                        pCtx,
                        pMe,
                        pPath_temp_full,
                        psz_objectid,
                        i_objectid_timestamp,
                        ptx,
                        psz_hid,
                        SG_BLOBENCODING__KEEPFULLFORNOW,
                        NULL,
                        len_full
                        )  );
                }
            }
            *pb_vcdiff_result_was_ok = SG_FALSE;
        }
    }

    if (pPath_temp_vcdiff)
    {
        SG_ERR_CHECK(  SG_fsobj__remove__pathname(pCtx, pPath_temp_vcdiff)  );
    }
    if (pPath_temp_zlib)
    {
        SG_ERR_CHECK(  SG_fsobj__remove__pathname(pCtx, pPath_temp_zlib)  );
    }
    SG_ERR_CHECK(  SG_clone__remove_tempfile(pCtx, pMe, psz_hid)  );

fail:
    SG_PATHNAME_NULLFREE(pCtx, pPath_ref);
    SG_PATHNAME_NULLFREE(pCtx, pPath_temp_full);
    SG_PATHNAME_NULLFREE(pCtx, pPath_temp_vcdiff);
    SG_PATHNAME_NULLFREE(pCtx, pPath_temp_zlib);
}

static void _slurp__version_1(
	SG_context* pCtx,
	sg_clone* pMe,
	SG_file* pfb
	)
{
	SG_vhash* pvh = NULL;
	const char* psz_op = NULL;
	SG_uint64 iPos = 0;

    SG_ERR_CHECK(  SG_blobset__create(pCtx, &pMe->pbs)  );

    SG_ERR_CHECK(  SG_blobset__begin_inserting(pCtx, pMe->pbs)  );

	SG_NULLARGCHECK_RETURN(pfb);

	while (1)
	{
		SG_uint64 offset_begin_record = 0;

		SG_ERR_CHECK(  SG_file__tell(pCtx, pfb, &offset_begin_record)  );
		SG_ERR_CHECK(  SG_fragball__read_object_header(pCtx, pfb, &pvh)  );
		if (!pvh)
		{
			break;
		}
		SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh, "op", &psz_op)  );
		if (0 == strcmp(psz_op, "blob"))
		{
			SG_int64 i64;
			const char* psz_hid = NULL;
			SG_blob_encoding encoding = 0;
			const char* psz_hid_vcdiff_reference = NULL;
			const char* psz_objectid = NULL;
            SG_int64 i_objectid_timestamp = 0;
			SG_uint64 len_full  = 0;
			SG_uint64 len_encoded  = 0;

			SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh, "hid", &psz_hid)  );
			SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvh, "encoding", &i64)  );
			encoding = (SG_blob_encoding) i64;
			SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pvh, "vcdiff_ref", &psz_hid_vcdiff_reference)  );
			SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pvh, "objectid", &psz_objectid)  );
            if (psz_objectid)
            {
                SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvh, "objectid_timestamp", &i_objectid_timestamp)  );
            }
			SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvh, "len_full", &i64)  );
			len_full = (SG_uint64) i64;
			SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvh, "len_encoded", &i64)  );
			len_encoded = (SG_uint64) i64;
			SG_ERR_CHECK(  SG_file__tell(pCtx, pfb, &iPos)  );

            SG_ERR_CHECK(  SG_blobset__insert(
                        pCtx,
                        pMe->pbs,
                        psz_hid,
                        NULL,
                        iPos,
                        encoding,
                        len_encoded,
                        len_full,
                        psz_hid_vcdiff_reference,
                        psz_objectid,
                        i_objectid_timestamp
                        )  );

			/* seek ahead to skip the blob */
			SG_ERR_CHECK(  SG_file__seek(pCtx, pfb, iPos + len_encoded)  );
		}
		SG_VHASH_NULLFREE(pCtx, pvh);
	}

    SG_ERR_CHECK(  SG_blobset__finish_inserting(pCtx, pMe->pbs)  );

	SG_VHASH_NULLFREE(pCtx, pvh);

	return;

fail:
    SG_BLOBSET_NULLFREE(pCtx, pMe->pbs);
	SG_VHASH_NULLFREE(pCtx, pvh);
}

static void _commit__version_1(
	SG_context* pCtx,
	sg_clone* pMe,
    SG_vhash* pvh_specific,
    struct SG_clone__params__all_to_something* pats,
    struct SG_clone__demands* p_demands,
	SG_repo_tx_handle* pTx,
	SG_file* pfb
	)
{
	SG_vhash* pvh = NULL;
	const char* psz_op = NULL;
	SG_dagfrag* pFrag = NULL;

    SG_uint32 demand_vcdiff_savings_over_full = SG_CLONE_DEFAULT__DEMAND_VCDIFF_SAVINGS_OVER_FULL;
    SG_uint32 demand_vcdiff_savings_over_zlib = SG_CLONE_DEFAULT__DEMAND_VCDIFF_SAVINGS_OVER_ZLIB;
    SG_uint32 demand_zlib_savings_over_full =   SG_CLONE_DEFAULT__DEMAND_ZLIB_SAVINGS_OVER_FULL;

    if (p_demands)
    {
        if (p_demands->vcdiff_savings_over_full >= 0)
        {
            demand_vcdiff_savings_over_full = p_demands->vcdiff_savings_over_full;
        }
        if (p_demands->vcdiff_savings_over_zlib >= 0)
        {
            demand_vcdiff_savings_over_zlib = p_demands->vcdiff_savings_over_zlib;
        }
        if (p_demands->zlib_savings_over_full >= 0)
        {
            demand_zlib_savings_over_full = p_demands->zlib_savings_over_full;
        }
    }

	while (1)
	{
		SG_vhash* pvh_frag = NULL;
		SG_uint64 offset_begin_record = 0;

		SG_ERR_CHECK(  SG_file__tell(pCtx, pfb, &offset_begin_record)  );
		SG_ERR_CHECK(  SG_fragball__read_object_header(pCtx, pfb, &pvh)  );
		if (!pvh)
		{
			break;
		}
		SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh, "op", &psz_op)  );
		if (0 == strcmp(psz_op, "frag"))
		{
			SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvh, "frag", &pvh_frag)  );
			SG_ERR_CHECK(  SG_dagfrag__alloc__from_vhash(pCtx, &pFrag, pvh_frag)  );
			SG_ERR_CHECK(  SG_repo__store_dagfrag(pCtx, pMe->pRepo, pTx, pFrag)  );
			pFrag = NULL;
		}
		else if (0 == strcmp(psz_op, "blob"))
		{
			SG_int64 i64 = 0;
			const char* psz_hid = NULL;
			const char* psz_objectid = NULL;
            SG_int64 i_objectid_timestamp = -1;
			SG_blob_encoding encoding_CURRENT = 0;
			const char* psz_hid_vcdiff_reference = NULL;
			SG_uint64 len_full  = 0;
			SG_uint64 len_encoded  = 0;
			SG_blob_encoding encoding_DESIRED = 0;
			const char* psz_hid_vcdiff_reference_DESIRED = NULL;
            SG_bool b_omit = SG_FALSE;
			SG_uint64 offset_begin_blob  = 0;

            SG_ERR_CHECK(  SG_file__tell(pCtx, pfb, &offset_begin_blob)  );

			SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh, "hid", &psz_hid)  );
			SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvh, "encoding", &i64)  );
			encoding_CURRENT = (SG_blob_encoding) i64;
			SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pvh, "vcdiff_ref", &psz_hid_vcdiff_reference)  );
			SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pvh, "objectid", &psz_objectid)  );
            if (psz_objectid)
            {
                SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvh, "objectid_timestamp", &i_objectid_timestamp)  );
            }
			SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvh, "len_full", &i64)  );
			len_full = (SG_uint64) i64;
			SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvh, "len_encoded", &i64)  );
			len_encoded = (SG_uint64) i64;

            if (SG_IS_BLOBENCODING_VCDIFF(encoding_CURRENT))
            {
                SG_ASSERT(psz_hid_vcdiff_reference);
            }
            else
            {
                SG_ASSERT(!psz_hid_vcdiff_reference);
                if (SG_IS_BLOBENCODING_FULL(encoding_CURRENT))
                {
                    SG_ASSERT(len_full == len_encoded);
                }
                else
                {
                    SG_ASSERT(SG_IS_BLOBENCODING_ZLIB(encoding_CURRENT));
                }
            }

            if (pvh_specific)
            {
                SG_vhash* pvh_chg = NULL;

                SG_ERR_CHECK(  SG_vhash__check__vhash(pCtx, pvh_specific, psz_hid, &pvh_chg)  );
                if (pvh_chg)
                {
                    SG_bool b_has = SG_FALSE;
                    SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh_chg, SG_CLONE_CHANGES__OMIT, &b_has)  );
                    if (b_has)
                    {
                        SG_ERR_CHECK(  SG_vhash__get__bool(pCtx, pvh_chg, SG_CLONE_CHANGES__OMIT, &b_omit)  );
                    }

                    if (!b_omit)
                    {
                        SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvh_chg, SG_CLONE_CHANGES__ENCODING, &i64)  );
                        encoding_DESIRED = (SG_blob_encoding)  i64;

                        if (SG_IS_BLOBENCODING_VCDIFF(encoding_DESIRED))
                        {
                            SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh_chg, SG_CLONE_CHANGES__HID_VCDIFF, &psz_hid_vcdiff_reference_DESIRED)  );
                        }
                    }
                }
            }

            if (!encoding_DESIRED && pats)
            {
                if (
                        (SG_BLOBENCODING__ALWAYSFULL == encoding_CURRENT)
                        && !(pats->flags & SG_CLONE_FLAG__OVERRIDE_ALWAYSFULL)
                   )
                {
                    // leave it alone
                }
                else
                {
                    SG_bool b_do = SG_TRUE;

                    if (
                            pats->low_pass
                            && (len_full < pats->low_pass)
                       )
                    {
                        b_do = SG_FALSE;
                    }
                    else if (
                            pats->high_pass
                            && (len_full < pats->high_pass)
                       )
                    {
                        b_do = SG_FALSE;
                    }

                    if (b_do)
                    {
                        encoding_DESIRED = pats->new_encoding;
                    }
                }
            }

            if (b_omit)
            {
                // just seek the file past this blob
                SG_ERR_CHECK(  SG_file__seek(pCtx, pfb, (offset_begin_blob + len_encoded))  );
            }
            else if (encoding_DESIRED)
            {
                if (SG_IS_BLOBENCODING_VCDIFF(encoding_DESIRED))
                {
                    SG_ASSERT(psz_hid_vcdiff_reference_DESIRED);

                    if (
                            SG_IS_BLOBENCODING_VCDIFF(encoding_CURRENT) 
                            && (0 == strcmp(psz_hid_vcdiff_reference, psz_hid_vcdiff_reference_DESIRED))
                       )
                    {
                        // same.  just store it.
                        SG_ERR_CHECK(  _store_one_blob(
                                    pCtx,
                                    pMe,
                                    pfb,
                                    psz_objectid,
                                    i_objectid_timestamp,
                                    pTx,
                                    psz_hid,
                                    encoding_CURRENT,
                                    psz_hid_vcdiff_reference,
                                    len_full,
                                    len_encoded)  );
                    }
                    else
                    {
                        SG_bool b_vcdiff_result_was_ok = SG_FALSE;

                        SG_ERR_CHECK(  _convert_one_blob(
                                    pCtx,
                                    pMe,
                                    pfb,
                                    psz_objectid,
                                    i_objectid_timestamp,
                                    pTx,
                                    psz_hid,
                                    encoding_CURRENT,
                                    psz_hid_vcdiff_reference,
                                    len_full,
                                    len_encoded,
                                    encoding_DESIRED,
                                    psz_hid_vcdiff_reference_DESIRED,
                                    &b_vcdiff_result_was_ok,
                                    demand_vcdiff_savings_over_full,
                                    demand_vcdiff_savings_over_zlib,
                                    demand_zlib_savings_over_full
                                    )  );
                        if (
                                SG_IS_BLOBENCODING_VCDIFF(encoding_DESIRED)
                                && !b_vcdiff_result_was_ok
                                )
                        {
                            // TODO use this as a keyframe?
                        }
                    }
                }
                else
                {
                    SG_ASSERT(!psz_hid_vcdiff_reference_DESIRED);

                    if (SG_ARE_BLOBENCODINGS_BASICALLY_EQUAL(encoding_DESIRED, encoding_CURRENT))
                    {
                        SG_ERR_CHECK(  _store_one_blob(
                                    pCtx,
                                    pMe,
                                    pfb,
                                    psz_objectid,
                                    i_objectid_timestamp,
                                    pTx,
                                    psz_hid,
                                    encoding_DESIRED,
                                    psz_hid_vcdiff_reference,
                                    len_full,
                                    len_encoded)  );
                    }
                    else
                    {
                        SG_ERR_CHECK(  _convert_one_blob(
                                    pCtx,
                                    pMe,
                                    pfb,
                                    psz_objectid,
                                    i_objectid_timestamp,
                                    pTx,
                                    psz_hid,
                                    encoding_CURRENT,
                                    psz_hid_vcdiff_reference,
                                    len_full,
                                    len_encoded,
                                    encoding_DESIRED,
                                    NULL,
                                    NULL,
                                    0,
                                    0,
                                    demand_zlib_savings_over_full
                                    )  );
                    }
                }
            }
            else
            {
                // no requests made.  just store this blob
                // the way we got it.

                SG_ERR_CHECK(  _store_one_blob(
                            pCtx,
                            pMe,
                            pfb,
                            psz_objectid,
                            i_objectid_timestamp,
                            pTx,
                            psz_hid,
                            encoding_CURRENT,
                            psz_hid_vcdiff_reference,
                            len_full,
                            len_encoded)  );
            }

#if defined(DEBUG)
            {
                SG_uint64 offset_end_blob  = 0;

                SG_ERR_CHECK(  SG_file__tell(pCtx, pfb, &offset_end_blob)  );
                SG_ASSERT(offset_end_blob == (offset_begin_blob + len_encoded));
            }
#endif
		}
		else
		{
			SG_ERR_THROW(  SG_ERR_FRAGBALL_INVALID_OP  );
		}
		SG_VHASH_NULLFREE(pCtx, pvh);
	}

	/* fall through */

fail:
	SG_VHASH_NULLFREE(pCtx, pvh);
	SG_DAGFRAG_NULLFREE(pCtx, pFrag);
}

static void _make_clone_area_path(SG_context* pCtx, const char* psz_name, SG_pathname** ppPath)
{
	SG_pathname* pPath_tempdir = NULL;
	SG_pathname* pPath_staging = NULL;

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__USER_TEMP_DIRECTORY(pCtx, &pPath_tempdir)  );
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx, &pPath_staging, pPath_tempdir, psz_name)  );

	*ppPath = pPath_staging;
	pPath_staging = NULL;

	/* fallthru */

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPath_tempdir);
	SG_PATHNAME_NULLFREE(pCtx, pPath_staging);
}

///////////////////////////////////////////////////////////////////////////////

static void sg_clone__slurp(
	SG_context* pCtx,
    sg_clone* pMe
	);

static void my_create(
        SG_context* pCtx,
        const char* psz_existing,
        SG_bool b_slurp,
        SG_clone** ppClone
        )
{
	char* psz_staging_area_name = NULL;
	SG_pathname* pPath_staging = NULL;
	sg_clone* pMe = NULL;
	SG_clone* pClone = NULL;
	SG_sync_client* pClient = NULL;

    SG_ERR_CHECK(  SG_alloc1(pCtx, pMe)  );
    pClone = (SG_clone*)pMe;

	SG_ERR_CHECK(  SG_tid__alloc(pCtx, &psz_staging_area_name)  );		// TODO consider SG_tid__alloc2(...)
	SG_ERR_CHECK(  _make_clone_area_path(pCtx, (const char*)psz_staging_area_name, &pPath_staging)  );
    pMe->pPath = pPath_staging;
    pPath_staging = NULL;
    SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pMe->pvh_tempfiles)  );

	SG_ERR_CHECK(  SG_sync_client__open(pCtx, psz_existing, &pClient)  );

	SG_ERR_CHECK(  SG_sync_client__get_repo_info(pCtx, pClient, SG_FALSE, &pMe->pvh_repo_info)  );

	SG_ERR_CHECK(  SG_fsobj__mkdir__pathname(pCtx, pMe->pPath)  );

	SG_ERR_CHECK(  SG_sync_client__pull_clone(pCtx, pClient, pMe->pPath, &pMe->psz_fragball_name)  );

    if (b_slurp)
    {
        SG_ERR_CHECK(  sg_clone__slurp(pCtx, pMe)  );
    }

    *ppClone = pClone;
    pClone = NULL;

	/* fallthru */

fail:
    if (pClone)
    {
        SG_ERR_IGNORE(  SG_clone__cleanup(pCtx, &pClone)  );
    }
	SG_SYNC_CLIENT_NULLFREE(pCtx, pClient);
	SG_NULLFREE(pCtx, psz_staging_area_name);
	SG_PATHNAME_NULLFREE(pCtx, pPath_staging);
}

void SG_clone__create(
        SG_context* pCtx,
        const char* psz_existing,
        SG_clone** ppClone
        )
{
    SG_ERR_CHECK_RETURN(  my_create(pCtx, psz_existing, SG_TRUE, ppClone)  );
}

void SG_clone__cleanup(SG_context* pCtx, SG_clone** ppStaging)
{
	SG_pathname* pPath = NULL;

    if (ppStaging && *ppStaging)
    {
        sg_clone* pMe = (sg_clone*)*ppStaging;

		pPath = pMe->pPath;
		if (pPath)
			pMe->pPath = NULL;

		SG_ERR_CHECK(  sg_clone__free(pCtx, *ppStaging)  );
		
		if (pPath)
			SG_ERR_CHECK(  SG_fsobj__rmdir_recursive__pathname(pCtx, pPath) );

        *ppStaging = NULL;
    }
	
fail:
	SG_PATHNAME_NULLFREE(pCtx, pPath);
}

static void sg_clone__slurp(
	SG_context* pCtx,
    sg_clone* pMe
	)
{
	SG_vhash* pvh_version = NULL;
	const char* psz_version = NULL;
	SG_uint32 version = 0;
	SG_file* pfb = NULL;
	SG_pathname* pPath = NULL;

	SG_NULLARGCHECK_RETURN(pMe);

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx, &pPath, pMe->pPath, pMe->psz_fragball_name)  );

	SG_ERR_CHECK(  SG_file__open__pathname(pCtx, pPath, SG_FILE_RDONLY | SG_FILE_OPEN_EXISTING, SG_FSOBJ_PERMS__UNUSED, &pfb)  );
	SG_PATHNAME_NULLFREE(pCtx, pPath);

	SG_ERR_CHECK(  SG_fragball__read_object_header(pCtx, pfb, &pvh_version)  );
	SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh_version, "version", &psz_version)  );
	version = atoi(psz_version);
	SG_VHASH_NULLFREE(pCtx, pvh_version);

	switch (version)
	{
		case 1:
			SG_ERR_CHECK(  _slurp__version_1(pCtx, pMe, pfb)  );
			break;

		default:
			SG_ERR_THROW(  SG_ERR_FRAGBALL_INVALID_VERSION  );
			break;
	}

	SG_ERR_CHECK(  SG_file__close(pCtx, &pfb)  );

	return;

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_FILE_NULLCLOSE(pCtx, pfb);
	SG_VHASH_NULLFREE(pCtx, pvh_version);
}

void SG_clone__commit(
        SG_context* pCtx, 
        SG_clone* pClone, 
        const char* psz_new_repo,
        SG_vhash* pvh_specific,
        struct SG_clone__params__all_to_something* pats,
        struct SG_clone__demands* p_demands
        )
{
	SG_repo_tx_handle* ptx = NULL;
	sg_clone* pMe = (sg_clone*)pClone;
	SG_bool bInRepoTx = SG_FALSE;
	SG_vhash* pvh_version = NULL;
	const char* psz_version = NULL;
	SG_uint32 version = 0;
	SG_pathname* pPathFragball = NULL;
	SG_file* pfb = NULL;
    const char* psz_repo_id = NULL;
    const char* psz_admin_id = NULL;
    const char* psz_hash_method = NULL;

	SG_ERR_CHECK(  SG_log__push_operation(pCtx, "Saving new repository", SG_LOG__FLAG__NONE)  );
	SG_ERR_CHECK(  SG_log__set_steps(pCtx, 2u, NULL)  );

	SG_NULLARGCHECK(pClone);

    SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pMe->pvh_repo_info, SG_SYNC_REPO_INFO_KEY__REPO_ID, &psz_repo_id)  );
    SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pMe->pvh_repo_info, SG_SYNC_REPO_INFO_KEY__ADMIN_ID, &psz_admin_id)  );
    SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pMe->pvh_repo_info, SG_SYNC_REPO_INFO_KEY__HASH_METHOD, &psz_hash_method)  );
    SG_ERR_CHECK(  SG_repo__create_empty_clone_from_parts(
                pCtx, 
                psz_repo_id,
                psz_admin_id,
                psz_hash_method,
                psz_new_repo)
            );
    SG_ERR_CHECK(  SG_REPO__OPEN_REPO_INSTANCE(pCtx, psz_new_repo, &pMe->pRepo)  );

	SG_ERR_CHECK(  SG_repo__begin_tx(pCtx, pMe->pRepo, &ptx)  );
	bInRepoTx = SG_TRUE;

	SG_ERR_CHECK(  SG_log__set_step(pCtx, "Copying data")  );
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx, &pPathFragball, pMe->pPath, pMe->psz_fragball_name)  );
	SG_ERR_CHECK(  SG_file__open__pathname(pCtx, pPathFragball, SG_FILE_RDONLY | SG_FILE_OPEN_EXISTING, SG_FSOBJ_PERMS__UNUSED, &pfb)  );
	SG_PATHNAME_NULLFREE(pCtx, pPathFragball);
	SG_ERR_CHECK(  SG_fragball__read_object_header(pCtx, pfb, &pvh_version)  );
	SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh_version, "version", &psz_version)  );
	version = atoi(psz_version);
	SG_VHASH_NULLFREE(pCtx, pvh_version);
	switch (version)
	{
	case 1:
		SG_ERR_CHECK(  _commit__version_1(pCtx, pMe, pvh_specific, pats, p_demands, ptx, pfb)  );
		break;

	default:
		SG_ERR_THROW(  SG_ERR_FRAGBALL_INVALID_VERSION  );
		break;
	}
	SG_ERR_CHECK(  SG_log__finish_step(pCtx)  );

	SG_ERR_CHECK(  SG_log__set_step(pCtx, "Committing changes")  );
	SG_ERR_CHECK(  SG_repo__commit_tx(pCtx, pMe->pRepo, &ptx)  );
	bInRepoTx = SG_FALSE;
	SG_ERR_CHECK(  SG_log__finish_step(pCtx)  );

	/* fall through */
fail:
	if (bInRepoTx)
		SG_ERR_IGNORE(  SG_repo__abort_tx(pCtx, pMe->pRepo, &ptx)  );
	SG_VHASH_NULLFREE(pCtx, pvh_version);
	SG_PATHNAME_NULLFREE(pCtx, pPathFragball);
	SG_FILE_NULLCLOSE(pCtx, pfb);
	SG_log__pop_operation(pCtx);
}

void SG_clone__exact(
	SG_context* pCtx,
    const char* psz_existing,
    const char* psz_new
    )
{
    SG_clone* pClone = NULL;

    SG_NULLARGCHECK_RETURN(psz_existing);
    SG_NULLARGCHECK_RETURN(psz_new);

	SG_ERR_CHECK(  _sg_clone__validate_new_repo_name(pCtx, psz_new)  );

    SG_ERR_CHECK(  my_create(pCtx, psz_existing, SG_FALSE, &pClone)  );

	SG_ERR_CHECK(  SG_clone__commit(pCtx, pClone, psz_new, NULL, NULL, NULL)  );

fail:
    SG_ERR_IGNORE(  SG_clone__cleanup(pCtx, &pClone)  );
}

static void sg_clone__pack__vcdiff__one_object(
    SG_context* pCtx,
    SG_clone * pClone,
    const char * pszObjectID,
    SG_uint32 nKeyframeDensity,
    SG_uint32 nRevisionsToLeaveFull,
    SG_uint64 low_pass,
    SG_uint64 high_pass,
    SG_vhash* pvh_specific
    )
{
	SG_vhash* pvh_revisions = NULL;
	SG_uint32 i_cur_revision = 0;
	SG_uint32 count_all_revs = 0;
	SG_uint32 count_net_revs = 0;
    char* flags = NULL;

    SG_ARGCHECK_RETURN(nKeyframeDensity > 0, nKeyframeDensity);

	SG_ERR_CHECK(  SG_clone__list_revisions_for_object(pCtx, pClone, pszObjectID, &pvh_revisions)  );
	SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_revisions, &count_all_revs)  );

    if (count_all_revs < 2)
    {
        goto done;
    }

    // mark the last N revisions ZLIB (usually 1)

    if (nRevisionsToLeaveFull > 0)
    {
        SG_uint32 i = 0;

        for (i=0; i<nRevisionsToLeaveFull; i++)
        {
            SG_vhash* pvh_rev = NULL;
            const char* psz_hid = NULL;
            SG_uint32 ndx = 0;
            SG_vhash* pvh_chg = NULL;

            if ((count_all_revs - i) == 0)
            {
                break;
            }

            ndx = count_all_revs - 1 - i;
            SG_ERR_CHECK(  SG_vhash__get_nth_pair__vhash(pCtx, pvh_revisions, ndx, &psz_hid, &pvh_rev)  );

            SG_ERR_CHECK(  SG_vhash__addnew__vhash(pCtx, pvh_specific, psz_hid, &pvh_chg)  );
            SG_ERR_CHECK(  SG_vhash__add__int64(
                        pCtx, 
                        pvh_chg, 
                        SG_CLONE_CHANGES__ENCODING, 
                        (SG_int64) SG_BLOBENCODING__ZLIB
                        )  );
        }
    }

    if (count_all_revs <= nRevisionsToLeaveFull)
	{
		goto done; // early return
	}

    count_net_revs = count_all_revs - nRevisionsToLeaveFull;

    SG_ERR_CHECK_RETURN(  SG_allocN(pCtx, count_net_revs, flags)  );

    // For pass one, find stuff that is already eligible to be
    // used as a reference blob.

	for (i_cur_revision = 0; i_cur_revision < count_net_revs; i_cur_revision++)
	{
        SG_int64 i64 = 0;
        SG_blob_encoding encoding = 0;
        SG_vhash* pvh_rev = NULL;
        const char* psz_hid = NULL;
        SG_uint64 len_full = 0;
        SG_int64 count_blobs_referencing_me = 0;

		SG_ERR_CHECK(  SG_vhash__get_nth_pair__vhash(pCtx, pvh_revisions, i_cur_revision, &psz_hid, &pvh_rev)  );

		SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvh_rev, "encoding", &i64)  );
        encoding = (SG_blob_encoding) i64;
		SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvh_rev, "len_full", &i64)  );
        len_full = (SG_uint64) len_full;

		SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvh_rev, "count_blobs_referencing_me", &count_blobs_referencing_me)  );

		if (
                (encoding == SG_BLOBENCODING__ALWAYSFULL)
                )
		{
            // TODO shouldn't we check override here?
            flags[i_cur_revision] = SG_TRUE;
		}
        else if (
                (
                    low_pass
                    && (len_full < low_pass)
                )
                || (
                    high_pass 
                    && (len_full > high_pass)
                )
                )
        {
            // stuff that falls outside the low/high pass range is not
            // supposed to be vcdiffed

            flags[i_cur_revision] = SG_TRUE;
        }
	}

    // Now add keyframes according to the specified density.  Basically
    // we just walk the revisions and keep track of how many consecutive
    // non-keyframes we have seen.  When that number reaches the specified
    // density, we add a keyframe right there.

    {
        SG_uint32 count_non_keyframes_seen = 0;
        for (i_cur_revision=0; i_cur_revision < count_net_revs; i_cur_revision++)
        {
            if (flags[i_cur_revision])
            {
                count_non_keyframes_seen = 0;
            }
            else
            {
                count_non_keyframes_seen++;
            }

            if (
                    (count_non_keyframes_seen >= nKeyframeDensity)
                    || (i_cur_revision == (count_net_revs-1))
               )
            {
                flags[i_cur_revision] = SG_TRUE;
                count_non_keyframes_seen = 0;
            }
        }
    }

    // now set up the delta chains
	for (i_cur_revision = 0; i_cur_revision < count_net_revs; i_cur_revision++)
	{
        const char* psz_hid_ref = NULL;
        const char* psz_hid = NULL;
        SG_vhash* pvh_ref = NULL;

        if (flags[i_cur_revision])
        {
            // This is a keyframe.  Don't deltify it.
            continue;
        }

		SG_ERR_CHECK(  SG_vhash__get_nth_pair__vhash(pCtx, pvh_revisions, i_cur_revision, &psz_hid, NULL)  );

        SG_ERR_CHECK(  SG_vhash__get_nth_pair__vhash(pCtx, pvh_revisions, i_cur_revision + 1, &psz_hid_ref, &pvh_ref)  );

		if (psz_hid_ref)
		{
            SG_vhash* pvh_chg = NULL;
            SG_int64 ref_len_full = -1;

            SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvh_ref, "len_full", &ref_len_full)  );
            SG_ERR_CHECK(  SG_vhash__addnew__vhash(pCtx, pvh_specific, psz_hid, &pvh_chg)  );
            if (0 == ref_len_full)
            {
                SG_ERR_CHECK(  SG_vhash__add__int64(
                            pCtx, 
                            pvh_chg, 
                            SG_CLONE_CHANGES__ENCODING, 
                            (SG_int64) SG_BLOBENCODING__ZLIB
                            )  );
            }
            else
            {
                SG_ERR_CHECK(  SG_vhash__add__int64(
                            pCtx, 
                            pvh_chg, 
                            SG_CLONE_CHANGES__ENCODING, 
                            (SG_int64) SG_BLOBENCODING__VCDIFF
                            )  );

                SG_ERR_CHECK(  SG_vhash__add__string__sz(
                            pCtx,
                            pvh_chg,
                            SG_CLONE_CHANGES__HID_VCDIFF,
                            psz_hid_ref
                            )  );
            }
		}
    }

done:
fail:
    SG_NULLFREE(pCtx, flags);
	SG_VHASH_NULLFREE(pCtx, pvh_revisions);
}

void SG_clone__pack(
    SG_context* pCtx,
    const char* psz_existing,
    const char* psz_new,
    SG_int32 nKeyframeDensity,
    SG_int32 nRevisionsToLeaveFull,
    SG_uint32 nMinRevisions,
    SG_uint64 low_pass,
    SG_uint64 high_pass,
    struct SG_clone__demands* p_demands
    )
{
    // Look for the objects that take up the most
    // space, and vcdiff them.

    SG_varray * pva_candidates = NULL;
    SG_uint32 count_candidates = 0;
    SG_uint32 i = 0;
    SG_vhash * pvhResult = NULL;
    const char * pszgid = NULL;
	SG_int64 nRevisionCount = 0;
    SG_clone* pClone = NULL;
    SG_vhash* pvh_specific = NULL;

	SG_ARGCHECK(nKeyframeDensity > 0, keyframe-density);

	SG_ERR_CHECK(  _sg_clone__validate_new_repo_name(pCtx, psz_new)  );

    SG_ERR_CHECK(  SG_clone__create(pCtx, psz_existing, &pClone)  );

    SG_ERR_CHECK(  SG_clone__list_pack_candidates(pCtx, pClone, nMinRevisions, &pva_candidates)  );

    // TODO when packing a large repo, will it use too much RAM
    // to keep pvh_specific all in memory?

    SG_ERR_CHECK(  SG_vhash__alloc(pCtx, &pvh_specific)  );
    SG_ERR_CHECK(  SG_varray__count(pCtx, pva_candidates, &count_candidates)  );
    for (i=0; i < count_candidates; i++)
    {
    	SG_ERR_CHECK(  SG_varray__get__vhash(pCtx, pva_candidates, i, &pvhResult)  );
    	SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhResult, "objectid", &pszgid)  );
    	SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvhResult, "revision_count", &nRevisionCount)  );

    	if (nRevisionCount  < nMinRevisions)
		{
    		continue;
		}

  		SG_ERR_CHECK( sg_clone__pack__vcdiff__one_object(pCtx, pClone, pszgid, nKeyframeDensity, nRevisionsToLeaveFull, low_pass, high_pass, pvh_specific)  );
    }

	SG_ERR_CHECK(  SG_clone__commit(pCtx, pClone, psz_new, pvh_specific, NULL, p_demands)  );

fail:
    SG_ERR_IGNORE(  SG_clone__cleanup(pCtx, &pClone)  );
	SG_VARRAY_NULLFREE(pCtx, pva_candidates);
    SG_VHASH_NULLFREE(pCtx, pvh_specific);
}

void SG_clone__all_to_something(
    SG_context* pCtx,
    const char* psz_existing,
    const char* psz_new,
    struct SG_clone__params__all_to_something* pats,
    struct SG_clone__demands* p_demands
    )
{
    SG_clone* pClone = NULL;

	SG_ERR_CHECK(  _sg_clone__validate_new_repo_name(pCtx, psz_new)  );

    SG_ERR_CHECK(  SG_clone__create(pCtx, psz_existing, &pClone)  );

	SG_ERR_CHECK(  SG_clone__commit(pCtx, pClone, psz_new, NULL, pats, p_demands)  );

fail:
    SG_ERR_IGNORE(  SG_clone__cleanup(pCtx, &pClone)  );
}

void SG_clone__init_demands(
	SG_context* pCtx,
    struct SG_clone__demands* p_demands
    )
{
    SG_NULLARGCHECK_RETURN(p_demands);

    p_demands->vcdiff_savings_over_full = -1;
    p_demands->vcdiff_savings_over_zlib = -1;
    p_demands->zlib_savings_over_full = -1;
}


