/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include <sg.h>

#define CLOSET_VERSION 1
#define CLOSET_VERSION_FILE "version"

static void _write_closet_version(SG_context* pCtx, const SG_pathname* pPathCloset)
{
	SG_pathname* pPathVersionFile = NULL;
	SG_file* pFile = NULL;

	SG_ERR_CHECK(  SG_pathname__alloc__pathname_sz(pCtx, &pPathVersionFile, pPathCloset, CLOSET_VERSION_FILE)  );
	SG_ERR_CHECK(  SG_file__open__pathname(pCtx, pPathVersionFile, SG_FILE_OPEN_OR_CREATE|SG_FILE_WRONLY|SG_FILE_TRUNC, 0644, &pFile)  );

	/* The version is represented by the length of the file. We write null bytes to discourage edits. */
	{
		SG_uint32 i;
		SG_byte buf[CLOSET_VERSION];
		SG_uint32 lenWritten = 0;

		for (i = 0; i < sizeof(buf); i++)
			buf[i] = 0;
		SG_ERR_CHECK(  SG_file__write(pCtx, pFile, sizeof(buf), buf, &lenWritten)  );
		if (lenWritten != sizeof(buf))
			SG_ERR_THROW(SG_ERR_INCOMPLETEWRITE);
	}

	SG_ERR_CHECK(  SG_file__close(pCtx, &pFile)  );
	SG_PATHNAME_NULLFREE(pCtx, pPathVersionFile);

	return;

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathVersionFile);
	SG_FILE_NULLCLOSE(pCtx, pFile);
}

static void _verify_closet_version(SG_context* pCtx, const SG_pathname* pPathCloset)
{
	SG_pathname* pPathVersionFile = NULL;
	SG_fsobj_type fsobjType;
	SG_uint64 len = 0;

	SG_ERR_CHECK(  SG_pathname__alloc__pathname_sz(pCtx, &pPathVersionFile, pPathCloset, CLOSET_VERSION_FILE)  );

	SG_fsobj__length__pathname(pCtx, pPathVersionFile, &len, &fsobjType);
	if (SG_context__has_err(pCtx))
	{
		SG_error err = SG_ERR_UNSPECIFIED;
		SG_context__get_err(pCtx, &err);

		switch (err)
		{
		default:
			SG_ERR_RETHROW;

#if defined(MAC) || defined(LINUX)
		case SG_ERR_ERRNO(ENOENT):
#endif
#if defined(WINDOWS)
		case SG_ERR_GETLASTERROR(ERROR_FILE_NOT_FOUND):
		case SG_ERR_GETLASTERROR(ERROR_PATH_NOT_FOUND):
#endif
			/* If the file is not found, pre-1.0 we "grandfather it in" and write the version 1 file.
			 * In the future, we probably want to fail with the unsupported closet version error. */
			SG_context__err_reset(pCtx);
			SG_ERR_CHECK(  _write_closet_version(pCtx, pPathCloset)  );
		}
	}
	else
	{
		if (fsobjType != SG_FSOBJ_TYPE__REGULAR)
		{
			SG_ERR_THROW2(SG_ERR_UNSUPPORTED_CLOSET_VERSION, (pCtx, "the version file is invalid"));
		}

		/* The version is represented by the length of the file so we don't have to open and parse it. */
		if (len != CLOSET_VERSION)
		{
			SG_int_to_string_buffer buf;
			SG_ERR_CHECK(  SG_int64_to_sz(len, buf)  );
			SG_ERR_THROW2(SG_ERR_UNSUPPORTED_CLOSET_VERSION, (pCtx, "%s", buf));
		}
	}

	/* Common cleanup */
fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathVersionFile);
}

static void sg_closet__open(SG_context * pCtx, SG_pathname** ppPath)
{
	SG_string* pstrEnv = NULL;
	SG_uint32 len;
	SG_pathname* pPath = NULL;

	SG_ERR_CHECK(  SG_environment__get__str(pCtx, "SGCLOSET", &pstrEnv, &len)  );
	if (len)
	{
		SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPath, SG_string__sz(pstrEnv))  );
	}
	else
	{
		SG_ERR_CHECK(  SG_PATHNAME__ALLOC__USER_APPDATA_DIRECTORY(pCtx, &pPath)  );
		SG_ERR_CHECK(  SG_pathname__append__from_sz(pCtx, pPath, ".sgcloset")  );
	}
	SG_STRING_NULLFREE(pCtx, pstrEnv);


	SG_fsobj__verify_directory_exists_on_disk__pathname(pCtx, pPath);
	if (!SG_context__has_err(pCtx))
	{
		SG_ERR_CHECK(  _verify_closet_version(pCtx, pPath)  );
		*ppPath = pPath;
		return;
	}

	if (SG_context__err_equals(pCtx, SG_ERR_NOT_A_DIRECTORY))
		SG_ERR_RETHROW;

	if (SG_context__err_equals(pCtx, SG_ERR_NOT_FOUND))
	{
		SG_context__err_reset(pCtx);
		SG_ERR_CHECK(  SG_fsobj__mkdir__pathname(pCtx, pPath)  );
		SG_ERR_CHECK(  _write_closet_version(pCtx, pPath)  );
	}

	*ppPath = pPath;

	return;
fail:
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_STRING_NULLFREE(pCtx, pstrEnv);
}

#define sg_CLOSET_CONFIG_DESCRIPTORS        1
#define sg_CLOSET_CONFIG_LOCAL_SETTINGS     2

static void sg_closet__config__get(
	SG_context * pCtx,
	SG_uint32 which,
	SG_jsondb** ppJsondb
    )
{
	SG_pathname* pPath = NULL;
	SG_jsondb* pJsondb = NULL;
	SG_bool bHasRoot;

	SG_ERR_CHECK(  sg_closet__open(pCtx, &pPath)  );

	switch (which)
	{
	case sg_CLOSET_CONFIG_DESCRIPTORS:
		SG_ERR_CHECK(  SG_pathname__append__from_sz(pCtx, pPath, "descriptors.jsondb")  );
		SG_ERR_CHECK(  SG_jsondb__create_or_open(pCtx, pPath, "descriptors", &pJsondb)  );
		break;
	case sg_CLOSET_CONFIG_LOCAL_SETTINGS:
		SG_ERR_CHECK(  SG_pathname__append__from_sz(pCtx, pPath, "settings.jsondb")  );
		SG_ERR_CHECK(  SG_jsondb__create_or_open(pCtx, pPath, "settings", &pJsondb)  );
		break;
	default:
		SG_ERR_THROW(  SG_ERR_INVALIDARG  );
	}

	SG_ERR_CHECK(  SG_jsondb__has(pCtx, pJsondb, "/", &bHasRoot)  );
	if (!bHasRoot)
		SG_ERR_CHECK(  SG_jsondb__add__vhash(pCtx, pJsondb, "/", SG_FALSE, NULL)  );

	SG_PATHNAME_NULLFREE(pCtx, pPath);

	*ppJsondb = pJsondb;

	return;
fail:
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_JSONDB_NULLFREE(pCtx, pJsondb);
}

void SG_closet__get_localsettings(
	SG_context * pCtx,
	SG_jsondb** ppJsondb
    )
{
    SG_ERR_CHECK_RETURN(  sg_closet__config__get(pCtx, sg_CLOSET_CONFIG_LOCAL_SETTINGS, ppJsondb)  );
}

void SG_closet__get_server_state_path(
	SG_context * pCtx,
	SG_pathname ** ppPath
    )
{
	SG_pathname* pPath = NULL;
	
	SG_ERR_CHECK(  sg_closet__open(pCtx, &pPath)  );

	SG_ERR_CHECK(  SG_pathname__append__from_sz(pCtx, pPath, "server-state")  );

	SG_fsobj__verify_directory_exists_on_disk__pathname(pCtx, pPath);
	if (SG_context__err_equals(pCtx, SG_ERR_NOT_FOUND))
	{
		SG_context__err_reset(pCtx);
		SG_ERR_CHECK(  SG_fsobj__mkdir__pathname(pCtx, pPath)  );
	}

	SG_RETURN_AND_NULL(pPath, ppPath);
	
fail:
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	
}

static void _name_to_path(SG_context* pCtx, const char* pszName, char** ppszPath)
{
	char* pszEscapedName = NULL;
	char* pszPath = NULL;
	SG_uint32 len;

	SG_ERR_CHECK(  SG_jsondb__escape_keyname(pCtx, pszName, &pszEscapedName)  );

	len = SG_STRLEN(pszEscapedName) + 2;

	SG_ERR_CHECK(  SG_allocN(pCtx, len, pszPath)  );
	pszPath[0] = '/';
	pszPath[1] = 0;
	SG_ERR_CHECK(  SG_strcat(pCtx, pszPath, len, pszEscapedName)  );

	SG_NULLFREE(pCtx, pszEscapedName);

	*ppszPath = pszPath;

	return;

fail:
	SG_NULLFREE(pCtx, pszPath);
	SG_NULLFREE(pCtx, pszEscapedName);
}
void SG_closet__descriptors__list(SG_context * pCtx, SG_vhash** ppResult)
{
	SG_jsondb* p = NULL;
	SG_vhash * pResult = NULL;

	SG_ERR_CHECK_RETURN(  sg_closet__config__get(pCtx, sg_CLOSET_CONFIG_DESCRIPTORS, &p)  );
	SG_ERR_CHECK(  SG_jsondb__get__vhash(pCtx, p, "/", &pResult)  );
	SG_ERR_CHECK(  SG_vhash__sort(pCtx, pResult, SG_FALSE, SG_vhash_sort_callback__increasing)  );
	SG_RETURN_AND_NULL(pResult, ppResult);

	/* fall through */
fail:
	SG_JSONDB_NULLFREE(pCtx, p);
	SG_VHASH_NULLFREE(pCtx, pResult);
}

void SG_closet__descriptors__add(SG_context * pCtx, const char* pszName, const SG_vhash* pvhDescriptor)
{
	SG_jsondb* p = NULL;
	char * szName = NULL;
	char* pszPath = NULL;

	SG_ERR_CHECK(  SG_repo__validate_repo_name(pCtx, pszName, &szName)  );

	SG_ERR_CHECK(  sg_closet__config__get(pCtx, sg_CLOSET_CONFIG_DESCRIPTORS, &p)  );
	SG_ERR_CHECK(  _name_to_path(pCtx, szName, &pszPath)  );
	SG_ERR_CHECK(  SG_jsondb__add__vhash(pCtx, p, pszPath, SG_TRUE, pvhDescriptor)  );

	/* fall through */
fail:
	SG_JSONDB_NULLFREE(pCtx, p);
	SG_NULLFREE(pCtx, pszPath);
	SG_NULLFREE(pCtx, szName);
}

void SG_closet__descriptors__exists(SG_context* pCtx, const char* pszName, SG_bool* pbExists)
{
	SG_jsondb* p = NULL;
	char * szName = NULL;
	char* pszPath = NULL;

	SG_NULLARGCHECK_RETURN(pbExists);

	SG_repo__validate_repo_name(pCtx, pszName, &szName);
	if(SG_context__err_equals(pCtx, SG_ERR_INVALID_REPO_NAME))
	{
		SG_context__err_reset(pCtx);
		*pbExists=SG_FALSE;
	}
	else
	{
		SG_ERR_CHECK_CURRENT;

		SG_ERR_CHECK(  sg_closet__config__get(pCtx, sg_CLOSET_CONFIG_DESCRIPTORS, &p)  );
		SG_ERR_CHECK(  _name_to_path(pCtx, szName, &pszPath)  );
		SG_ERR_CHECK(  SG_jsondb__has(pCtx, p, pszPath, pbExists)  );
	}

	/* fall through */
fail:
	SG_JSONDB_NULLFREE(pCtx, p);
	SG_NULLFREE(pCtx, pszPath);
	SG_NULLFREE(pCtx, szName);
}

void SG_closet__descriptors__get(SG_context * pCtx, const char* pszName, SG_vhash** ppResult)
{
	SG_jsondb* p = NULL;
	char * szName = NULL;
	char* pszPath = NULL;

	SG_ERR_CHECK(  SG_repo__validate_repo_name(pCtx, pszName, &szName)  );
	SG_ERR_CHECK(  sg_closet__config__get(pCtx, sg_CLOSET_CONFIG_DESCRIPTORS, &p)  );
	SG_ERR_CHECK(  _name_to_path(pCtx, szName, &pszPath)  );

	SG_jsondb__get__vhash(pCtx, p, pszPath, ppResult);
	if(SG_context__err_equals(pCtx, SG_ERR_NOT_FOUND))
		SG_ERR_RESET_THROW2(SG_ERR_NOTAREPOSITORY, (pCtx, "%s", szName));
	else
		SG_ERR_CHECK_CURRENT;

	/* fall through */
fail:
	SG_JSONDB_NULLFREE(pCtx, p);
	SG_NULLFREE(pCtx, pszPath);
	SG_NULLFREE(pCtx, szName);
}

void SG_closet__descriptors__remove(SG_context * pCtx, const char* pszName)
{
	SG_jsondb* p = NULL;
	char * szName = NULL;
	char* pszPath = NULL;

	SG_ERR_CHECK(  SG_repo__validate_repo_name(pCtx, pszName, &szName)  );
	SG_ERR_CHECK(  sg_closet__config__get(pCtx, sg_CLOSET_CONFIG_DESCRIPTORS, &p)  );
	SG_ERR_CHECK(  _name_to_path(pCtx, szName, &pszPath)  );
	SG_ERR_CHECK(  SG_jsondb__remove(pCtx, p, pszPath)  );

	/* fall through */
fail:
	SG_JSONDB_NULLFREE(pCtx, p);
	SG_NULLFREE(pCtx, pszPath);
	SG_NULLFREE(pCtx, szName);
}

void SG_closet__get_partial_repo_instance_descriptor_for_new_local_repo(SG_context * pCtx,
																		const char * pszStorage,	// use config: new_repo/driver if null
																		const char * pszRepoId,		// optional
																		const char * pszAdminId,	// optional
																		SG_vhash** ppvhDescriptor)
{
	SG_vhash* pvhPartialDescriptor = NULL;
	SG_pathname* pPath = NULL;
	char* pszRepoImpl = NULL;

	SG_ERR_CHECK(  sg_closet__open(pCtx, &pPath)  );

	SG_ERR_CHECK(  SG_pathname__append__from_sz(pCtx, pPath, "repo")  );

	SG_fsobj__verify_directory_exists_on_disk__pathname(pCtx, pPath);
	if (SG_context__err_equals(pCtx, SG_ERR_NOT_FOUND))
	{
		SG_context__err_reset(pCtx);
		SG_ERR_CHECK(  SG_fsobj__mkdir__pathname(pCtx, pPath)  );
	}

	SG_ERR_CHECK_CURRENT;


	SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvhPartialDescriptor)  );

	if (pszStorage && *pszStorage)
	{
		SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhPartialDescriptor, SG_RIDESC_KEY__STORAGE, pszStorage)  );
	}
	else
	{
		SG_ERR_CHECK(  SG_localsettings__get__sz(pCtx, SG_LOCALSETTING__NEWREPO_DRIVER, NULL, &pszRepoImpl, NULL)  );
        if (pszRepoImpl)
        {
            SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhPartialDescriptor, SG_RIDESC_KEY__STORAGE, pszRepoImpl)  );
        }
		SG_NULLFREE(pCtx, pszRepoImpl);
	}
	
	/* This is used by non-filesystem repo implementations because the sqlite dbndx files go here. */
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhPartialDescriptor, SG_RIDESC_FSLOCAL__PATH_PARENT_DIR, SG_pathname__sz(pPath))  );

	// cache these in the descriptor if we have them.  they are optional and
	// can be back-filled later if necessary.

	if (pszRepoId && *pszRepoId)
		SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhPartialDescriptor, SG_RIDESC_KEY__REPO_ID, pszRepoId)  );

	if (pszAdminId && *pszAdminId)
		SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhPartialDescriptor, SG_RIDESC_KEY__ADMIN_ID, pszAdminId)  );
	
	SG_PATHNAME_NULLFREE(pCtx, pPath);

	*ppvhDescriptor = pvhPartialDescriptor;

	return;
fail:
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_VHASH_NULLFREE(pCtx, pvhPartialDescriptor);
    SG_NULLFREE(pCtx, pszRepoImpl);
}

//////////////////////////////////////////////////////////////////

struct _rev_map_data
{
	SG_vhash * pvhResult;
	const char * pszRepoIdWanted;
};

static void _rev_map_cb__eval(SG_context * pCtx,
							  struct _rev_map_data * pData,
							  const char * psz_repo_id,
							  const char * pszKey_DescName)
{
	SG_string * pString = NULL;
	SG_vhash * pvhBindings = NULL;

	if (strcmp(psz_repo_id, pData->pszRepoIdWanted) != 0)
		return;

	// this repo matches the requested repo-id, so we want to add it to the result vhash.
	// 
	// see if it also has any bindings (urls for pulls/pushes) associated with it in the
	// localsettings.  if so, add them too.  normally, we only have the paths/default
	// binding that was (probably) created when repo was cloned, but the user could have
	// deleted it or added other aliases to other servers.  we allow for there to be zero
	// or many bindings by copying the "paths" sub-vhash.

	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pString)  );
	SG_ERR_CHECK(  SG_string__sprintf(pCtx, pString, "%s/%s/%s",
									  SG_LOCALSETTING__SCOPE__INSTANCE,
									  pszKey_DescName,
									  SG_LOCALSETTING__PATHS)  );

	SG_ERR_IGNORE(  SG_localsettings__get__vhash(pCtx,
												 SG_string__sz(pString),
												 NULL,		// null repo because we only want /instance/ settings.
												 &pvhBindings,
												 NULL)  );
	// pvhBindings may or may not be null.  it will be null afterwards.
    if (pvhBindings)
    {
        SG_ERR_CHECK(  SG_vhash__add__vhash(pCtx, pData->pvhResult, pszKey_DescName, &pvhBindings)  );
    }
    else
    {
        // code elsewhere depends on the old behavior of add__vhash, which
        // added a null if given a null.
        SG_ERR_CHECK(  SG_vhash__add__null(pCtx, pData->pvhResult, pszKey_DescName)  );
    }
	SG_ASSERT(  (pvhBindings == NULL)  );

fail:
	SG_STRING_NULLFREE(pCtx, pString);
	SG_VHASH_NULLFREE(pCtx, pvhBindings);
}
							  
static SG_vhash_foreach_callback _rev_map_cb;
static void _rev_map_cb(SG_context * pCtx,
						void * pVoidData,
						const SG_vhash * pvhForeach,
						const char * pszKey_DescName,
						const SG_variant * pvValue)
{
	struct _rev_map_data * pData = (struct _rev_map_data *)pVoidData;
	const SG_vhash * pvhValue;
	SG_repo * pRepo = NULL;
	char * psz_repo_id = NULL;
	SG_bool bHasRepoId;

	SG_UNUSED( pvhForeach );

	// pvValue contains a vhash with
	//      {
	//        storage
	//        path_parent_dir
	//        dir_name
	// and optionally
	//        repo_id
	//      }
	// 
	// if we have a repo_id field, just use it and avoid having to open the repo to get it.

	SG_ERR_CHECK(  SG_variant__get__vhash(pCtx, pvValue, (SG_vhash **)&pvhValue)  );
	SG_ERR_CHECK(  SG_vhash__has(pCtx, pvhValue, SG_RIDESC_KEY__REPO_ID, &bHasRepoId)  );
	if (bHasRepoId)
	{
		const char * psz_repo_id_ref;

		SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhValue, SG_RIDESC_KEY__REPO_ID,  &psz_repo_id_ref )  );	// we don't own this string
		SG_ERR_CHECK(  _rev_map_cb__eval(pCtx, pData, psz_repo_id_ref, pszKey_DescName)  );
	}
	else
	{
		// use the given descriptor name to try to open the repo.
		// if this fails (because the user deleted the actual repo
		// without deleting the record in the descriptor list),
		// then ignore this entry in the table.

		SG_ERR_IGNORE(  SG_REPO__OPEN_REPO_INSTANCE(pCtx, pszKey_DescName, &pRepo)  );
		if (pRepo)
		{
			SG_ERR_CHECK(  SG_repo__get_repo_id( pCtx, pRepo, &psz_repo_id )  );

			SG_ERR_CHECK(  _rev_map_cb__eval(pCtx, pData, psz_repo_id, pszKey_DescName)  );

			// TODO 2010/01/07 Consider backfilling the repo-id into
			// TODO            the descriptor list so that a subsequent invocation
			// TODO            won't have to stumble here again.
		}
	}

fail:
	SG_REPO_NULLFREE(pCtx, pRepo);
	SG_NULLFREE(pCtx, psz_repo_id);
}

/**
 * Use the array of hints (a list of repo-names for 'preferred'
 * submodules).  Lookup the (repo_id) of the hint by
 * name.  See if that matches the (repo_id) that we are
 * searching for.
 *
 * NOTE: The user can gives us a SINGLE/FLAT list of repo-names
 * as suggestions.  We pass this down through whatever nested
 * hierarchy of submodules as a flat list.  We do not force them
 * to try to match the submodule structure and nest the hints
 * nor bother peeling off outer matches as we dive.  This has
 * the odd property that if they reference a library submodule
 * in more than one place in the tree (and it has multiple
 * repo-instances), we'll bind both submodules to the same
 * instance.
 *
 * NOTE: Also note that we use the hints in the order given
 * and stop searching when we find a match.  WE DO NOT check
 * to see that the hints refer to distinct repo_id's.
 * This also means that we do not look to see if one repo-instance
 * has more csets than another.
 */
static void _rev_map__find_using_hints(SG_context * pCtx,
									   struct _rev_map_data * pData,
									   const SG_vector * pvec_use_submodules,
									   const SG_vhash * pvhDescList,
									   SG_bool * pbFoundMatch)
{
	SG_uint32 nrHints, kHint;
	SG_uint32 count_results_before, count_results_after;

	SG_ERR_CHECK(  SG_vhash__count(pCtx, pData->pvhResult, &count_results_before)  );
	SG_ASSERT(  (count_results_before == 0)  );

	SG_ERR_CHECK(  SG_vector__length(pCtx, pvec_use_submodules, &nrHints)  );
	for (kHint=0; kHint < nrHints; kHint++)
	{
		char * pszRepoNameHint_k;
		SG_bool bHas_k;

		SG_ERR_CHECK(  SG_vector__get(pCtx, pvec_use_submodules, kHint, (void **)&pszRepoNameHint_k)  );
		SG_ERR_CHECK(  SG_vhash__has(pCtx, pvhDescList, pszRepoNameHint_k, &bHas_k)  );
		if (bHas_k)
		{
			const SG_variant * pvDesc_k;

			SG_ERR_CHECK(  SG_vhash__get__variant(pCtx, pvhDescList, pszRepoNameHint_k, &pvDesc_k)  );
			SG_ERR_CHECK(  _rev_map_cb(pCtx, pData, pvhDescList, pszRepoNameHint_k, pvDesc_k)  );
			SG_ERR_CHECK(  SG_vhash__count(pCtx, pData->pvhResult, &count_results_after)  );
			if (count_results_after > count_results_before)
			{
				*pbFoundMatch = SG_TRUE;
				return;
			}
		}
	}

fail:
	*pbFoundMatch = SG_FALSE;
}

void SG_closet__reverse_map_repo_id_to_descriptor_names(SG_context * pCtx,
														const char * pszRepoId,
														const SG_vector * pvec_use_submodules,
														SG_vhash ** ppvhList)
{
	SG_vhash * pvhDescList = NULL;
	struct _rev_map_data data;
	SG_bool bFoundMatch = SG_FALSE;

	memset(&data, 0, sizeof(data));

	SG_NONEMPTYCHECK_RETURN( pszRepoId );
	// pvec_use_submodules is optional
	SG_NULLARGCHECK_RETURN( ppvhList );

	data.pszRepoIdWanted = pszRepoId;
	SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &data.pvhResult)  );

	SG_ERR_CHECK(  SG_closet__descriptors__list(pCtx, &pvhDescList)  );
	if (pvec_use_submodules)
	{
		// try to find a match using the hints given.

		SG_ERR_CHECK(  _rev_map__find_using_hints(pCtx, &data, pvec_use_submodules, pvhDescList, &bFoundMatch)  );
	}

	if (!bFoundMatch)
	{
		// no hints or no match.  do full linear search on
		// the list of all named repo-instances.

		SG_ERR_CHECK(  SG_vhash__foreach(pCtx, pvhDescList, _rev_map_cb, (void *)&data)  );
	}

#if 0 && defined(DEBUG)
	SG_ERR_IGNORE(  SG_vhash_debug__dump_to_console__named(pCtx, data.pvhResult, pszRepoId)  );
#endif

	*ppvhList = data.pvhResult;
	data.pvhResult = NULL;

fail:
	SG_VHASH_NULLFREE(pCtx, data.pvhResult);
	SG_VHASH_NULLFREE(pCtx, pvhDescList);
}
