/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg_dagquery.c
 *
 * @details An odd collection of routines to ask some hard DAG relationship questions.
 *
 */

//////////////////////////////////////////////////////////////////

#include <sg.h>

//////////////////////////////////////////////////////////////////

void SG_dagquery__how_are_dagnodes_related(SG_context * pCtx,
										   SG_repo * pRepo,
                                           SG_uint64 dagnum,
										   const char * pszHid1,
										   const char * pszHid2,
										   SG_bool bSkipDescendantCheck,
										   SG_bool bSkipAncestorCheck,
										   SG_dagquery_relationship * pdqRel)
{
	SG_dagnode * pdn1 = NULL;
	SG_dagnode * pdn2 = NULL;
	SG_dagfrag * pFrag = NULL;
	SG_dagquery_relationship dqRel = SG_DAGQUERY_RELATIONSHIP__UNKNOWN;
	SG_int32 gen1, gen2;
	SG_bool bFound;

	SG_NULLARGCHECK_RETURN(pRepo);
	SG_NONEMPTYCHECK_RETURN(pszHid1);
	SG_NONEMPTYCHECK_RETURN(pszHid2);
	SG_NULLARGCHECK_RETURN(pdqRel);

	if (strcmp(pszHid1, pszHid2) == 0)
	{
		dqRel = SG_DAGQUERY_RELATIONSHIP__SAME;
		goto cleanup;
	}

	// fetch the dagnode for both HIDs.  this throws when the HID is not found.

	SG_ERR_CHECK(  SG_repo__fetch_dagnode(pCtx, pRepo, dagnum, pszHid1, &pdn1)  );
	SG_ERR_CHECK(  SG_repo__fetch_dagnode(pCtx, pRepo, dagnum, pszHid2, &pdn2)  );

	// we say that 2 nodes are either:
	// [1] ancestor/descendant of each other;
	// [2] or that they are peers (cousins) of each other (no matter
	//     how distant in the DAG).  (that have an LCA, but we don't
	//     care about it.)

	// get the generation of both dagnodes.  if they are the same, then they
	// cannot have an ancestor/descendant relationship and therefore must be
	// peers/cousins (we don't care how close/distant they are).

	SG_ERR_CHECK(  SG_dagnode__get_generation(pCtx, pdn1, &gen1)  );
	SG_ERR_CHECK(  SG_dagnode__get_generation(pCtx, pdn2, &gen2)  );
	if (gen1 == gen2)
	{
		dqRel = SG_DAGQUERY_RELATIONSHIP__PEER;
		goto cleanup;
	}

	// see if one is an ancestor of the other.  since we only have PARENT
	// edges in our DAG, we start with the deeper one and walk backwards
	// until we've visited all ancestors at the depth of the shallower one.
	//
	// i'm going to be lazy here and not reinvent a recursive-ish parent-edge
	// graph walker.  instead, i'm going to create a DAGFRAG using the
	// deeper one and request the generation difference as the "thickness".
	// in theory, if we have an ancestor/descendant relationship, the
	// shallower one should be in the END-FRINGE of the DAGFRAG.
	//
	// i'm going to pick an arbitrary direction "cs1 is R of cs2".

	SG_ERR_CHECK(  SG_dagfrag__alloc_transient(pCtx, dagnum, &pFrag)  );
	if (gen1 > gen2)		// cs1 is *DEEPER* than cs2
	{
		if (bSkipDescendantCheck)
		{
			dqRel = SG_DAGQUERY_RELATIONSHIP__UNKNOWN;
		}
		else
		{
			SG_ERR_CHECK(  SG_dagfrag__load_from_repo__one(pCtx, pFrag, pRepo, pszHid1, (gen1 - gen2))  );
			SG_ERR_CHECK(  SG_dagfrag__query(pCtx, pFrag, pszHid2, NULL, NULL, &bFound, NULL)  );

			if (bFound)			// pszHid2 is an ancestor of pszHid1.  READ pszHid1 is a descendent of pszHid2.
				dqRel = SG_DAGQUERY_RELATIONSHIP__DESCENDANT;
			else				// they are *distant* peers.
				dqRel = SG_DAGQUERY_RELATIONSHIP__PEER;
		}
		goto cleanup;
	}
	else
	{
		if (bSkipAncestorCheck)
		{
			dqRel = SG_DAGQUERY_RELATIONSHIP__UNKNOWN;
		}
		else
		{
			SG_ERR_CHECK(  SG_dagfrag__load_from_repo__one(pCtx, pFrag, pRepo, pszHid2, (gen2 - gen1))  );
			SG_ERR_CHECK(  SG_dagfrag__query(pCtx, pFrag, pszHid1, NULL, NULL, &bFound, NULL)  );

			if (bFound)			// pszHid1 is an ancestor of pszHid2.
				dqRel = SG_DAGQUERY_RELATIONSHIP__ANCESTOR;
			else				// they are *distant* peers.
				dqRel = SG_DAGQUERY_RELATIONSHIP__PEER;
		}
		goto cleanup;
	}

	/*NOTREACHED*/

cleanup:
	*pdqRel = dqRel;

fail:
	SG_DAGNODE_NULLFREE(pCtx, pdn1);
	SG_DAGNODE_NULLFREE(pCtx, pdn2);
	SG_DAGFRAG_NULLFREE(pCtx, pFrag);
}

//////////////////////////////////////////////////////////////////

void SG_dagquery__find_descendant_heads(SG_context * pCtx,
										SG_repo * pRepo,
                                        SG_uint64 iDagNum,
										const char * pszHidStart,
										SG_bool bStopIfMultiple,
										SG_dagquery_find_head_status * pdqfhs,
										SG_rbtree ** pprbHeads)
{
	SG_rbtree * prbLeaves = NULL;
	SG_rbtree * prbHeadsFound = NULL;
	SG_rbtree_iterator * pIter = NULL;
	const char * pszKey_k = NULL;
	SG_bool b;
	SG_dagquery_find_head_status dqfhs;
	SG_dagquery_relationship dqRel;
	SG_uint32 nrFound;

	SG_NULLARGCHECK_RETURN(pRepo);
	SG_NONEMPTYCHECK_RETURN(pszHidStart);
	SG_NULLARGCHECK_RETURN(pdqfhs);
	SG_NULLARGCHECK_RETURN(pprbHeads);

	SG_ERR_CHECK(  SG_RBTREE__ALLOC(pCtx, &prbHeadsFound)  );

	// fetch a list of all of the LEAVES in the DAG.
	// this rbtree only contains keys; no assoc values.

	SG_ERR_CHECK(  SG_repo__fetch_dag_leaves(pCtx, pRepo, iDagNum, &prbLeaves)  );

	// if the starting point is a leaf, then we are done (we don't care how many
	// other leaves are in the rbtree because none will be a child of ours because
	// we are a leaf).

	SG_ERR_CHECK(  SG_rbtree__find(pCtx, prbLeaves, pszHidStart, &b, NULL)  );
	if (b)
	{
		SG_ERR_CHECK(  SG_rbtree__add(pCtx, prbHeadsFound, pszHidStart)  );

		dqfhs = SG_DAGQUERY_FIND_HEAD_STATUS__IS_LEAF;
		goto done;
	}

	// inspect each leaf and qualify it; put the ones that pass
	// into the list of actual heads.

	nrFound = 0;
	SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pIter, prbLeaves, &b, &pszKey_k, NULL)  );
	while (b)
	{
		// is head[k] a descendant of start?
		SG_ERR_CHECK(  SG_dagquery__how_are_dagnodes_related(pCtx, pRepo, iDagNum, pszKey_k, pszHidStart,
															 SG_FALSE, // we care about descendants, so don't skip
															 SG_TRUE,  // we don't care about ancestors, so skip them
															 &dqRel)  );

		if (dqRel == SG_DAGQUERY_RELATIONSHIP__DESCENDANT)
		{
			nrFound++;

			if (bStopIfMultiple && (nrFound > 1))
			{
				// they wanted a unique answer and we've found too many answers
				// (which they won't be able to use anyway) so just stop and
				// return the status.  (we delete prbHeadsFound because it is
				// incomplete and so that they won't be tempted to use it.)

				SG_RBTREE_NULLFREE(pCtx, prbHeadsFound);
				dqfhs = SG_DAGQUERY_FIND_HEAD_STATUS__MULTIPLE;
				goto done;
			}

			SG_ERR_CHECK(  SG_rbtree__add(pCtx, prbHeadsFound, pszKey_k)  );
		}

		SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pIter, &b, &pszKey_k, NULL)  );
	}

	switch (nrFound)
	{
	case 0:
		// this should NEVER happen.  we should always be able to find a
		// leaf/head for a node.
		//
		// TODO the only case where this might happen is if named branches
		// TODO cause the leaf to be disqualified.  so i'm going to THROW
		// TODO here rather than ASSERT.

		SG_ERR_THROW2(  SG_ERR_DAG_NOT_CONSISTENT,
						(pCtx, "Could not find head/leaf for changeset [%s]", pszHidStart)  );
		break;

	case 1:
		dqfhs = SG_DAGQUERY_FIND_HEAD_STATUS__UNIQUE;
		break;

	default:
		dqfhs = SG_DAGQUERY_FIND_HEAD_STATUS__MULTIPLE;
		break;
	}

done:
	*pprbHeads = prbHeadsFound;
	prbHeadsFound = NULL;

	*pdqfhs = dqfhs;

fail:
	SG_RBTREE_NULLFREE(pCtx, prbLeaves);
	SG_RBTREE_NULLFREE(pCtx, prbHeadsFound);
	SG_RBTREE_ITERATOR_NULLFREE(pCtx, pIter);
}

//////////////////////////////////////////////////////////////////

void SG_dagquery__find_single_descendant_head(SG_context * pCtx,
											  SG_repo * pRepo,
                                              SG_uint64 dagnum,
											  const char * pszHidStart,
											  SG_dagquery_find_head_status * pdqfhs,
											  char * bufHidHead,
											  SG_uint32 lenBufHidHead)
{
	SG_rbtree * prbHeads = NULL;
	const char * pszKey_0;
	SG_dagquery_find_head_status dqfhs;
	SG_bool b;

	SG_NULLARGCHECK_RETURN(pRepo);
	SG_NONEMPTYCHECK_RETURN(pszHidStart);
	SG_NULLARGCHECK_RETURN(pdqfhs);
	SG_NULLARGCHECK_RETURN(bufHidHead);

	SG_ERR_CHECK(  SG_dagquery__find_descendant_heads(pCtx, pRepo, dagnum, pszHidStart, SG_TRUE, &dqfhs, &prbHeads)  );

	switch (dqfhs)
	{
	case SG_DAGQUERY_FIND_HEAD_STATUS__IS_LEAF:
	case SG_DAGQUERY_FIND_HEAD_STATUS__UNIQUE:
		SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, NULL, prbHeads, &b, &pszKey_0, NULL)  );
		SG_ASSERT(  (b)  );
		SG_ERR_CHECK(  SG_strcpy(pCtx, bufHidHead, lenBufHidHead, pszKey_0)  );
		break;

	case SG_DAGQUERY_FIND_HEAD_STATUS__MULTIPLE:
	default:
		// don't throw, just return the status
		bufHidHead[0] = 0;
		break;
	}

	*pdqfhs = dqfhs;

fail:
	SG_RBTREE_NULLFREE(pCtx, prbHeads);
}

typedef struct
{
    SG_uint64 dagnum;
	const char* pszNewNodeHid;
	const char* pszOldNodeHid;
	SG_rbtree* prbNewNodeHids;
} new_since_context;


static void _dagquery__new_since__callback(SG_context* pCtx, 
										   SG_repo *pRepo, 
										   void *pVoidData, 
										   SG_dagnode *pCurrentDagnode, 
										   SG_rbtree* pDagnodeCache, 
										   SG_dagwalker_continue* pContinue)
{
	SG_dagquery_relationship rel;
	new_since_context* pcb = (new_since_context*)pVoidData;
	const char* szCurrentNodeHid = NULL;

	SG_UNUSED(pDagnodeCache);

	SG_ERR_CHECK(  SG_dagnode__get_id_ref(pCtx, pCurrentDagnode, &szCurrentNodeHid)  );

	if (strcmp(szCurrentNodeHid, pcb->pszOldNodeHid) == 0)
	{
		*pContinue = SG_DAGWALKER_CONTINUE__EMPTY_QUEUE;
		return;
	}

	// TODO: Consider looking at revision numbers to make this faster.
	SG_ERR_CHECK(  SG_dagquery__how_are_dagnodes_related(pCtx, pRepo, pcb->dagnum, szCurrentNodeHid, pcb->pszOldNodeHid, SG_FALSE, SG_TRUE, &rel)  );
	if (rel != SG_DAGQUERY_RELATIONSHIP__DESCENDANT)
	{
		SG_ERR_CHECK(  SG_dagquery__how_are_dagnodes_related(pCtx, pRepo, pcb->dagnum, szCurrentNodeHid, pcb->pszOldNodeHid, SG_TRUE, SG_FALSE, &rel)  );
		if (rel == SG_DAGQUERY_RELATIONSHIP__ANCESTOR)
		{
			*pContinue = SG_DAGWALKER_CONTINUE__EMPTY_QUEUE;
			return;
		}
	}

	SG_ERR_CHECK(  SG_rbtree__add(pCtx, pcb->prbNewNodeHids, szCurrentNodeHid)  );

fail:
	;
}

void SG_dagquery__new_since(SG_context * pCtx,
							SG_repo* pRepo,
							SG_uint64 iDagNum,
							const char* pszOldNodeHid,
							const char* pszNewNodeHid,
							SG_rbtree** pprbNewNodeHids)
{
	SG_dagquery_relationship rel;
	new_since_context cbCtx;

    cbCtx.dagnum = iDagNum;
	cbCtx.prbNewNodeHids = NULL;
	cbCtx.pszNewNodeHid = pszNewNodeHid;
	cbCtx.pszOldNodeHid = pszOldNodeHid;

	SG_ERR_CHECK(  SG_dagquery__how_are_dagnodes_related(pCtx, pRepo, iDagNum, pszNewNodeHid, pszOldNodeHid, SG_FALSE, SG_FALSE, &rel)  );

	if (rel != SG_DAGQUERY_RELATIONSHIP__DESCENDANT)
		SG_ERR_THROW2_RETURN(SG_ERR_UNSPECIFIED, (pCtx, "pszNewNodeHid must be a descendant of pszOldNodeHid"));

	SG_ERR_CHECK(  SG_RBTREE__ALLOC(pCtx, &cbCtx.prbNewNodeHids)  );

	SG_ERR_CHECK(  SG_dagwalker__walk_dag_single(pCtx, pRepo, iDagNum, pszNewNodeHid, _dagquery__new_since__callback, (void*)&cbCtx)  );

	SG_RETURN_AND_NULL(cbCtx.prbNewNodeHids, pprbNewNodeHids);

	/* Fall through to common cleanup */
fail:
	SG_RBTREE_NULLFREE(pCtx, cbCtx.prbNewNodeHids);
}

void SG_repo__dag__find_direct_path_from_root(
        SG_context * pCtx,
        SG_repo* pRepo,
        SG_uint64 dagnum,
        const char* psz_csid,
        SG_varray** ppva
        )
{
    SG_dagnode* pdn = NULL;
    char* psz_cur = NULL;
	const char** a_parents = NULL;
	SG_uint32 count_parents = 0;
    SG_varray* pva = NULL;

    SG_ERR_CHECK(  SG_VARRAY__ALLOC(pCtx, &pva)  );
    SG_ERR_CHECK(  SG_STRDUP(pCtx, psz_csid, &psz_cur)  );
    while (1)
    {
        SG_ERR_CHECK(  SG_repo__fetch_dagnode(pCtx, pRepo, dagnum, psz_cur, &pdn)  );
        SG_ERR_CHECK(  SG_varray__append__string__sz(pCtx, pva, psz_cur)  );
        SG_ERR_CHECK(  SG_dagnode__get_parents(pCtx, pdn, &count_parents, &a_parents)  );
        if (0 == count_parents)
        {
            break;
        }
        SG_NULLFREE(pCtx, psz_cur);
        SG_ERR_CHECK(  SG_STRDUP(pCtx, a_parents[0], &psz_cur)  );
        SG_NULLFREE(pCtx, a_parents);
        SG_DAGNODE_NULLFREE(pCtx, pdn);
    }

    SG_ERR_CHECK(  SG_varray__append__string__sz(pCtx, pva, "")  );

    *ppva = pva;
    pva = NULL;

fail:
    SG_VARRAY_NULLFREE(pCtx, pva);
    SG_DAGNODE_NULLFREE(pCtx, pdn);
    SG_NULLFREE(pCtx, psz_cur);
}

void SG_repo__dag__find_direct_backward_path(
        SG_context * pCtx,
        SG_repo* pRepo,
        SG_uint64 dagnum,
        const char* psz_csid_from,
        const char* psz_csid_to,
        SG_varray** ppva
        )
{
    SG_ERR_CHECK_RETURN(  SG_repo__find_dag_path(pCtx, pRepo, dagnum, psz_csid_to, psz_csid_from, ppva)  );
}

static void loop_innards_tree_lock_violation(
    SG_context* pCtx,
    SG_repo* pRepo,
    const char* psz_gid,
    const char* psz_userid,
    SG_varray* pva_path,
    SG_uint32 i_path_step,
    SG_bool* pb
    )
{
    SG_changeset* pcs = NULL;
    const char* psz_csid_cur = NULL;
    const char* psz_csid_parent = NULL;
    SG_vhash* pvh_changes = NULL;
    SG_vhash* pvh_one_parent_changes = NULL;
    SG_bool b_result = SG_FALSE;
    SG_varray* pva_audits = NULL;

    SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pva_path, i_path_step, &psz_csid_cur)  );
    SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pva_path, i_path_step + 1, &psz_csid_parent)  );
    SG_ERR_CHECK(  SG_changeset__load_from_repo(pCtx, pRepo, psz_csid_cur, &pcs)  );

    SG_ERR_CHECK(  SG_changeset__tree__get_changes(pCtx, pcs, &pvh_changes)  );

    SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvh_changes, psz_csid_parent, &pvh_one_parent_changes)  );

    SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh_one_parent_changes, psz_gid, &b_result)  );
    if (b_result)
    {
        // if the change was done by psz_userid, it doesn't count as a lock violation

        SG_uint32 count_audits = 0;
        SG_uint32 i = 0;

		SG_ERR_CHECK(  SG_audit__lookup(pCtx, pRepo, psz_csid_cur, SG_DAGNUM__VERSION_CONTROL, &pva_audits)  );
		SG_ERR_CHECK(  SG_varray__count(pCtx, pva_audits, &count_audits)  );
		for (i=0; i<count_audits; i++)
		{
			SG_vhash* pvh = NULL;  
            const char* psz_who = NULL;

			SG_ERR_CHECK(  SG_varray__get__vhash(pCtx, pva_audits, i, &pvh)  );

            SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh, SG_AUDIT__USERID, &psz_who)  );
            if (0 == strcmp(psz_who, psz_userid))
            {
                b_result = SG_FALSE;
                break;
            }
		}	
    }

    *pb = b_result;

fail:
    SG_VARRAY_NULLFREE(pCtx, pva_audits);
    SG_CHANGESET_NULLFREE(pCtx, pcs);
}

static void loop_innards_make_delta(
    SG_context* pCtx,
    SG_repo* pRepo,
    SG_varray* pva_path,
    SG_uint32 i_path_step,
    SG_bool b_normal,
    SG_vhash* pvh_add,
    SG_vhash* pvh_remove
    )
{
    SG_changeset* pcs = NULL;

    const char* psz_csid_cur = NULL;
    const char* psz_csid_parent = NULL;
    SG_vhash* pvh_changes = NULL;
    SG_vhash* pvh_one_parent_changes = NULL;
    SG_vhash* pvh_cs_add = NULL;
    SG_vhash* pvh_cs_remove = NULL;

    SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pva_path, i_path_step, &psz_csid_cur)  );
    SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pva_path, i_path_step + 1, &psz_csid_parent)  );
    SG_ERR_CHECK(  SG_changeset__load_from_repo(pCtx, pRepo, psz_csid_cur, &pcs)  );

    SG_ERR_CHECK(  SG_changeset__db__get_changes(pCtx, pcs, &pvh_changes)  );

    SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvh_changes, psz_csid_parent, &pvh_one_parent_changes)  );

    if (b_normal)
    {
        SG_ERR_CHECK(  SG_vhash__check__vhash(pCtx, pvh_one_parent_changes, "add", &pvh_cs_remove)  );
        SG_ERR_CHECK(  SG_vhash__check__vhash(pCtx, pvh_one_parent_changes, "remove", &pvh_cs_add)  );
    }
    else
    {
        SG_ERR_CHECK(  SG_vhash__check__vhash(pCtx, pvh_one_parent_changes, "remove", &pvh_cs_remove)  );
        SG_ERR_CHECK(  SG_vhash__check__vhash(pCtx, pvh_one_parent_changes, "add", &pvh_cs_add)  );
    }

    if (pvh_cs_add)
    {
        SG_uint32 count = 0;
        SG_uint32 i = 0;

        SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_cs_add, &count)  );
        for (i=0; i<count; i++)
        {
            const char* psz_hid_rec = NULL;
            SG_bool b = SG_FALSE;

            SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvh_cs_add, i, &psz_hid_rec, NULL)  );
            SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh_remove, psz_hid_rec, &b)  );
            if (b)
            {
                SG_ERR_CHECK(  SG_vhash__remove(pCtx, pvh_remove, psz_hid_rec)  );
            }
            else
            {
                SG_ERR_CHECK(  SG_vhash__update__null(pCtx, pvh_add, psz_hid_rec)  );
            }
        }
    }

    if (pvh_cs_remove)
    {
        SG_uint32 count = 0;
        SG_uint32 i = 0;

        SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_cs_remove, &count)  );
        for (i=0; i<count; i++)
        {
            const char* psz_hid_rec = NULL;
            SG_bool b = SG_FALSE;

            SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvh_cs_remove, i, &psz_hid_rec, NULL)  );
            SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh_add, psz_hid_rec, &b)  );
            if (b)
            {
                SG_ERR_CHECK(  SG_vhash__remove(pCtx, pvh_add, psz_hid_rec)  );
            }
            else
            {
                SG_ERR_CHECK(  SG_vhash__update__null(pCtx, pvh_remove, psz_hid_rec)  );
            }
        }
    }


fail:
    SG_CHANGESET_NULLFREE(pCtx, pcs);
}

void ensure_sub_vhashes(
    SG_context* pCtx,
    SG_vhash* pvh_delta,
    SG_vhash** ppvh_add,
    SG_vhash** ppvh_remove
    )
{
    SG_ERR_CHECK(  SG_vhash__check__vhash(pCtx, pvh_delta, "add", ppvh_add)  );
    if (!*ppvh_add)
    {
        SG_ERR_CHECK(  SG_vhash__addnew__vhash(pCtx, pvh_delta, "add", ppvh_add)  );
    }
    SG_ERR_CHECK(  SG_vhash__check__vhash(pCtx, pvh_delta, "remove", ppvh_remove)  );
    if (!*ppvh_remove)
    {
        SG_ERR_CHECK(  SG_vhash__addnew__vhash(pCtx, pvh_delta, "remove", ppvh_remove)  );
    }

fail:
    ;
}

void SG_db__make_delta_from_path__NEG(
    SG_context* pCtx,
    SG_repo* pRepo,
    SG_varray* pva_path,
    SG_vhash* pvh_delta
    )
{
    SG_uint32 count = 0;
    SG_uint32 i_path_step = 0;
    SG_vhash* pvh_add = NULL;
    SG_vhash* pvh_remove = NULL;

    SG_NULLARGCHECK_RETURN(pRepo);

    SG_ERR_CHECK(  ensure_sub_vhashes(pCtx, pvh_delta, &pvh_add, &pvh_remove)  );

    SG_ERR_CHECK(  SG_varray__count(pCtx, pva_path, &count)  );
    SG_ASSERT(count >= 2);

    i_path_step = count-2;
    while (1)
    {
        SG_ERR_CHECK(  loop_innards_make_delta(pCtx, pRepo, pva_path, i_path_step, SG_FALSE, pvh_add, pvh_remove)  );

        if (0 == i_path_step)
        {
            break;
        }

        i_path_step--;
    }

fail:
    ;
}

void SG_tree__check_path_for_lock_violation(
    SG_context* pCtx,
    SG_repo* pRepo,
    const char* psz_gid,
    const char* psz_userid,
    SG_varray* pva_path,
    SG_bool* pb
    )
{
    SG_uint32 count = 0;
    SG_uint32 i_path_step = 0;

    SG_NULLARGCHECK_RETURN(pRepo);
    SG_NULLARGCHECK_RETURN(psz_gid);
    SG_NULLARGCHECK_RETURN(pva_path);
    SG_NULLARGCHECK_RETURN(pb);

    SG_ERR_CHECK(  SG_varray__count(pCtx, pva_path, &count)  );
    SG_ASSERT(count >= 2);

    for (i_path_step=0; i_path_step<(count-1); i_path_step++)
    {
        SG_bool b = SG_FALSE;

        SG_ERR_CHECK(  loop_innards_tree_lock_violation(pCtx, pRepo, psz_gid, psz_userid, pva_path, i_path_step, &b)  );

        if (b)
        {
            *pb = SG_TRUE;
            goto done;
        }
    }

    *pb = SG_FALSE;

done:
fail:
    ;
}

void SG_db__make_delta_from_path(
    SG_context* pCtx,
    SG_repo* pRepo,
    SG_varray* pva_path,
    SG_vhash* pvh_delta
    )
{
    SG_uint32 count = 0;
    SG_uint32 i_path_step = 0;
    SG_vhash* pvh_add = NULL;
    SG_vhash* pvh_remove = NULL;

    SG_NULLARGCHECK_RETURN(pRepo);

    SG_ERR_CHECK(  ensure_sub_vhashes(pCtx, pvh_delta, &pvh_add, &pvh_remove)  );

    SG_ERR_CHECK(  SG_varray__count(pCtx, pva_path, &count)  );
    SG_ASSERT(count >= 2);

    for (i_path_step=0; i_path_step<(count-1); i_path_step++)
    {
        SG_ERR_CHECK(  loop_innards_make_delta(pCtx, pRepo, pva_path, i_path_step, SG_TRUE, pvh_add, pvh_remove)  );
    }

fail:
    ;
}

void SG_repo__db__calc_delta(
        SG_context * pCtx,
        SG_repo* pRepo,
        SG_uint64 dagnum,
        const char* psz_csid_from,
        const char* psz_csid_to,
        SG_vhash** ppvh
        )
{
    SG_dagnode* pdn_from = NULL;
    SG_dagnode* pdn_to = NULL;
    SG_int32 gen_from = -1;
    SG_int32 gen_to = -1;
    SG_varray* pva_direct_backward_path = NULL;
    SG_varray* pva_direct_forward_path = NULL;
    SG_vhash* pvh_delta = NULL;
    SG_rbtree* prb_temp = NULL;
    SG_daglca* plca = NULL;
    char* psz_csid_ancestor = NULL;

    SG_NULLARGCHECK_RETURN(psz_csid_from);
    SG_NULLARGCHECK_RETURN(psz_csid_to);
    SG_NULLARGCHECK_RETURN(pRepo);
    SG_NULLARGCHECK_RETURN(ppvh);

    SG_ERR_CHECK(  SG_repo__fetch_dagnode(pCtx, pRepo, dagnum, psz_csid_from, &pdn_from)  );
    SG_ERR_CHECK(  SG_dagnode__get_generation(pCtx, pdn_from, &gen_from)  );
    SG_ERR_CHECK(  SG_repo__fetch_dagnode(pCtx, pRepo, dagnum, psz_csid_to, &pdn_to)  );
    SG_ERR_CHECK(  SG_dagnode__get_generation(pCtx, pdn_to, &gen_to)  );

    if (gen_from > gen_to)
    {
        SG_ERR_CHECK(  SG_repo__dag__find_direct_backward_path(
                    pCtx,
                    pRepo,
                    dagnum,
                    psz_csid_from,
                    psz_csid_to,
                    &pva_direct_backward_path
                    )  );
        if (pva_direct_backward_path)
        {
            SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvh_delta)  );
            SG_ERR_CHECK(  SG_db__make_delta_from_path(
                        pCtx,
                        pRepo,
                        pva_direct_backward_path,
                        pvh_delta
                        )  );
        }
    }
    else if (gen_from < gen_to)
    {
        SG_ERR_CHECK(  SG_repo__dag__find_direct_backward_path(
                    pCtx,
                    pRepo,
                    dagnum,
                    psz_csid_to,
                    psz_csid_from,
                    &pva_direct_forward_path
                    )  );
        if (pva_direct_forward_path)
        {
            SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvh_delta)  );
            SG_ERR_CHECK(  SG_db__make_delta_from_path__NEG(
                        pCtx,
                        pRepo,
                        pva_direct_forward_path,
                        pvh_delta
                        )  );
        }
    }

    if (!pvh_delta)
    {
        SG_ERR_CHECK(  SG_RBTREE__ALLOC(pCtx, &prb_temp)  );
        SG_ERR_CHECK(  SG_rbtree__add(pCtx,prb_temp,psz_csid_from)  );
        SG_ERR_CHECK(  SG_rbtree__add(pCtx,prb_temp,psz_csid_to)  );
        SG_ERR_CHECK(  SG_repo__get_dag_lca(pCtx,pRepo,dagnum,prb_temp,&plca)  );
        {
            const char* psz_hid = NULL;
            SG_daglca_node_type node_type = 0;
            SG_int32 gen = -1;

            SG_ERR_CHECK(  SG_daglca__iterator__first(pCtx,
                                                      NULL,
                                                      plca,
                                                      SG_FALSE,
                                                      &psz_hid,
                                                      &node_type,
                                                      &gen,
                                                      NULL)  );
            SG_ERR_CHECK(  SG_STRDUP(pCtx, psz_hid, &psz_csid_ancestor)  );
        }

        SG_ERR_CHECK(  SG_repo__dag__find_direct_backward_path(
                    pCtx,
                    pRepo,
                    dagnum,
                    psz_csid_from,
                    psz_csid_ancestor,
                    &pva_direct_backward_path
                    )  );
        SG_ERR_CHECK(  SG_repo__dag__find_direct_backward_path(
                    pCtx,
                    pRepo,
                    dagnum,
                    psz_csid_to,
                    psz_csid_ancestor,
                    &pva_direct_forward_path
                    )  );
        SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvh_delta)  );
        SG_ERR_CHECK(  SG_db__make_delta_from_path(
                    pCtx,
                    pRepo,
                    pva_direct_backward_path,
                    pvh_delta
                    )  );
        SG_ERR_CHECK(  SG_db__make_delta_from_path__NEG(
                    pCtx,
                    pRepo,
                    pva_direct_forward_path,
                    pvh_delta
                    )  );
    }

    *ppvh = pvh_delta;
    pvh_delta = NULL;

fail:
    SG_NULLFREE(pCtx, psz_csid_ancestor);
    SG_RBTREE_NULLFREE(pCtx, prb_temp);
    SG_DAGLCA_NULLFREE(pCtx, plca);
    SG_VHASH_NULLFREE(pCtx, pvh_delta);
    SG_VARRAY_NULLFREE(pCtx, pva_direct_backward_path);
    SG_VARRAY_NULLFREE(pCtx, pva_direct_forward_path);
    SG_DAGNODE_NULLFREE(pCtx, pdn_from);
    SG_DAGNODE_NULLFREE(pCtx, pdn_to);
}

void SG_repo__db__calc_delta_from_root(
        SG_context * pCtx,
        SG_repo* pRepo,
        SG_uint64 dagnum,
        const char* psz_csid_to,
        SG_vhash** ppvh
        )
{
    SG_varray* pva_direct_forward_path = NULL;
    SG_vhash* pvh_delta = NULL;

    SG_NULLARGCHECK_RETURN(psz_csid_to);
    SG_NULLARGCHECK_RETURN(pRepo);
    SG_NULLARGCHECK_RETURN(ppvh);

    SG_ERR_CHECK(  SG_repo__dag__find_direct_path_from_root(
                pCtx,
                pRepo,
                dagnum,
                psz_csid_to,
                &pva_direct_forward_path
                )  );
    if (pva_direct_forward_path)
    {
        SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvh_delta)  );
        SG_ERR_CHECK(  SG_db__make_delta_from_path__NEG(
                    pCtx,
                    pRepo,
                    pva_direct_forward_path,
                    pvh_delta
                    )  );
    }

    *ppvh = pvh_delta;
    pvh_delta = NULL;

fail:
    SG_VHASH_NULLFREE(pCtx, pvh_delta);
    SG_VARRAY_NULLFREE(pCtx, pva_direct_forward_path);
}

void SG_repo__tree__check_for_lock_violation(
        SG_context * pCtx,
        SG_repo* pRepo,
        const char* psz_gid,
        const char* psz_userid,
        const char* psz_csid_from,
        const char* psz_csid_to,
        SG_bool* pb
        )
{
    SG_dagnode* pdn_from = NULL;
    SG_dagnode* pdn_to = NULL;
    SG_int32 gen_from = -1;
    SG_int32 gen_to = -1;
    SG_varray* pva_direct_backward_path = NULL;
    SG_varray* pva_direct_forward_path = NULL;
    SG_rbtree* prb_temp = NULL;
    SG_daglca* plca = NULL;
    char* psz_csid_ancestor = NULL;
    SG_bool b_changed = SG_FALSE;

    SG_NULLARGCHECK_RETURN(psz_csid_from);
    SG_NULLARGCHECK_RETURN(psz_csid_to);
    SG_NULLARGCHECK_RETURN(pRepo);
    SG_NULLARGCHECK_RETURN(pb);

    SG_ERR_CHECK(  SG_repo__fetch_dagnode(pCtx, pRepo, SG_DAGNUM__VERSION_CONTROL, psz_csid_from, &pdn_from)  );
    SG_ERR_CHECK(  SG_dagnode__get_generation(pCtx, pdn_from, &gen_from)  );
    SG_ERR_CHECK(  SG_repo__fetch_dagnode(pCtx, pRepo, SG_DAGNUM__VERSION_CONTROL, psz_csid_to, &pdn_to)  );
    SG_ERR_CHECK(  SG_dagnode__get_generation(pCtx, pdn_to, &gen_to)  );

    if (gen_from > gen_to)
    {
        SG_ERR_CHECK(  SG_repo__dag__find_direct_backward_path(
                    pCtx,
                    pRepo,
                    SG_DAGNUM__VERSION_CONTROL,
                    psz_csid_from,
                    psz_csid_to,
                    &pva_direct_backward_path
                    )  );
        if (pva_direct_backward_path)
        {
            SG_ERR_CHECK(  SG_tree__check_path_for_lock_violation(
                        pCtx,
                        pRepo,
                        psz_gid,
                        psz_userid,
                        pva_direct_backward_path,
                        &b_changed
                        )  );
            goto done;
        }
    }
    else if (gen_from < gen_to)
    {
        SG_ERR_CHECK(  SG_repo__dag__find_direct_backward_path(
                    pCtx,
                    pRepo,
                    SG_DAGNUM__VERSION_CONTROL,
                    psz_csid_to,
                    psz_csid_from,
                    &pva_direct_forward_path
                    )  );
        if (pva_direct_forward_path)
        {
            SG_ERR_CHECK(  SG_tree__check_path_for_lock_violation(
                        pCtx,
                        pRepo,
                        psz_gid,
                        psz_userid,
                        pva_direct_forward_path,
                        &b_changed
                        )  );
            goto done;
        }
    }

    SG_ERR_CHECK(  SG_RBTREE__ALLOC(pCtx, &prb_temp)  );
    SG_ERR_CHECK(  SG_rbtree__add(pCtx,prb_temp,psz_csid_from)  );
    SG_ERR_CHECK(  SG_rbtree__add(pCtx,prb_temp,psz_csid_to)  );
    SG_ERR_CHECK(  SG_repo__get_dag_lca(pCtx,pRepo,SG_DAGNUM__VERSION_CONTROL,prb_temp,&plca)  );
    {
        const char* psz_hid = NULL;
        SG_daglca_node_type node_type = 0;
        SG_int32 gen = -1;

        SG_ERR_CHECK(  SG_daglca__iterator__first(pCtx,
                                                  NULL,
                                                  plca,
                                                  SG_FALSE,
                                                  &psz_hid,
                                                  &node_type,
                                                  &gen,
                                                  NULL)  );
        SG_ERR_CHECK(  SG_STRDUP(pCtx, psz_hid, &psz_csid_ancestor)  );
    }

    SG_ERR_CHECK(  SG_repo__dag__find_direct_backward_path(
                pCtx,
                pRepo,
                SG_DAGNUM__VERSION_CONTROL,
                psz_csid_from,
                psz_csid_ancestor,
                &pva_direct_backward_path
                )  );
    SG_ERR_CHECK(  SG_repo__dag__find_direct_backward_path(
                pCtx,
                pRepo,
                SG_DAGNUM__VERSION_CONTROL,
                psz_csid_to,
                psz_csid_ancestor,
                &pva_direct_forward_path
                )  );
    SG_ERR_CHECK(  SG_tree__check_path_for_lock_violation(
                pCtx,
                pRepo,
                psz_gid,
                psz_userid,
                pva_direct_backward_path,
                &b_changed
                )  );
    if (b_changed)
    {
        SG_ERR_CHECK(  SG_tree__check_path_for_lock_violation(
                    pCtx,
                    pRepo,
                    psz_gid,
                    psz_userid,
                    pva_direct_forward_path,
                    &b_changed
                    )  );
    }

done:
    *pb = b_changed;

fail:
    SG_NULLFREE(pCtx, psz_csid_ancestor);
    SG_RBTREE_NULLFREE(pCtx, prb_temp);
    SG_DAGLCA_NULLFREE(pCtx, plca);
    SG_VARRAY_NULLFREE(pCtx, pva_direct_backward_path);
    SG_VARRAY_NULLFREE(pCtx, pva_direct_forward_path);
    SG_DAGNODE_NULLFREE(pCtx, pdn_from);
    SG_DAGNODE_NULLFREE(pCtx, pdn_to);
}

