/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg_difftool.c
 *
 * @details Code to handle how we choose which diff-tool to use
 * to diff a set of files.
 * 
 * Created by PaulE, copied off of sg_mergetool* by Andy.
 *
 */

#include <sg.h>

///////////////////////////////////////////////////////////////////////////////


/*
**
** Types
**
*/

/**
 * Definition of an internal diff tool.
 */
typedef struct
{
	const char*           szName;    //< Unique name of the internal diff tool.
	SG_filetool__invoker* pInvoker;  //< Function that implements the tool.
}
_internal_tool;

/**
 * Functions that translate filetool invocations into difftool invocations.
 */
static SG_filetool__invoker _invoke__skip;
static SG_filetool__invoker _invoke__diff_strict;
static SG_filetool__invoker _invoke__diff_ignore_eols;
static SG_filetool__invoker _invoke__diff_ignore_whitespace;
static SG_filetool__invoker _invoke__diff_ignore_case;
static SG_filetool__invoker _invoke__diff_ignore_case_and_whitespace;
static SG_filetool__invoker _invoke__diffmerge;

/*
**
** Globals
**
*/

// token keys used by diff tools
static const char* gszToken_FromPath  = "@FROM@";
static const char* gszToken_ToPath    = "@TO@";
static const char* gszToken_FromLabel = "@FROM_LABEL@";
static const char* gszToken_ToLabel   = "@TO_LABEL@";

static const char* gszToken_RightSideIsWritable = "@TO_WRITABLE?@";

/**
 * List of all internal diff tools.
 */
static _internal_tool aInternalTools[] =
{
	{ SG_DIFFTOOL__INTERNAL__SKIP,                            _invoke__skip                           },
	{ SG_DIFFTOOL__INTERNAL__DIFF_STRICT,                     _invoke__diff_strict                    },
	{ SG_DIFFTOOL__INTERNAL__DIFF_IGNORE_EOLS,                _invoke__diff_ignore_eols               },
	{ SG_DIFFTOOL__INTERNAL__DIFF_IGNORE_WHITESPACE,          _invoke__diff_ignore_whitespace         },
	{ SG_DIFFTOOL__INTERNAL__DIFF_IGNORE_CASE,                _invoke__diff_ignore_case               },
	{ SG_DIFFTOOL__INTERNAL__DIFF_IGNORE_CASE_AND_WHITESPACE, _invoke__diff_ignore_case_and_whitespace},
	{ SG_DIFFTOOL__INTERNAL__DIFFMERGE,                       _invoke__diffmerge                      },
};

/**
 * The hard-coded default tool for diff.
 */
static const char * _szDefaultTool = SG_DIFFTOOL__INTERNAL__DIFF_IGNORE_EOLS;


/*
**
** Internal Functions
**
*/

/**
 * Invokes an internal difftool.
 * Basically translates the token values from a filetool call into the arguments for a difftool call.
 */
static void _invoke_built_in_tool(
	SG_context*                pCtx,         //< [in] [out] Error and context info.
	const SG_vhash*            pTokenValues, //< [in] Token/argument values passed to the tool.
	SG_difftool__built_in_tool fInvoker      //< [in] The internal difftool implementation to invoke.
	)
{
	SG_pathname* pFromPath  = NULL;
	SG_string*   pFromLabel = NULL;
	SG_pathname* pToPath    = NULL;
	SG_string*   pToLabel   = NULL;
	const char*  szTemp     = NULL;
	SG_bool bRightSideIsWritable = SG_FALSE;

	SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pTokenValues, gszToken_FromPath, &szTemp)  );
	if(szTemp)
		SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pFromPath, szTemp)  );
	SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pTokenValues, gszToken_FromLabel, &szTemp)  );
	if(szTemp)
		SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &pFromLabel, szTemp)  );

	SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pTokenValues, gszToken_ToPath, &szTemp)  );
	if(szTemp)
		SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pToPath, szTemp)  );
	SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pTokenValues, gszToken_ToLabel, &szTemp)  );
	if(szTemp)
		SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &pToLabel, szTemp)  );

	SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pTokenValues, gszToken_RightSideIsWritable, &szTemp)  );
	if (szTemp)
		bRightSideIsWritable = (szTemp[0] == 'T');

	SG_ERR_CHECK(  fInvoker(pCtx, pFromPath, pToPath, pFromLabel, pToLabel, bRightSideIsWritable, NULL)  );

fail:
	SG_PATHNAME_NULLFREE(pCtx, pFromPath);
	SG_STRING_NULLFREE(pCtx, pFromLabel);
	SG_PATHNAME_NULLFREE(pCtx, pToPath);
	SG_STRING_NULLFREE(pCtx, pToLabel);
	return;
}

/**
 * Invokes the "skip" internal difftool.
 */
static void _invoke__skip(
	SG_context*        pCtx,
	const SG_filetool* pTool,
	const SG_vhash*    pTokenValues,
	SG_int32*          pExitCode
	)
{
	SG_UNUSED(pTool);
	SG_UNUSED(pExitCode);

	SG_ERR_CHECK_RETURN(  _invoke_built_in_tool(pCtx, pTokenValues, SG_difftool__built_in_tool__skip)  );
}

static void _invoke__diff_strict(SG_context* pCtx, const SG_filetool* pTool, const SG_vhash* pTokenValues, SG_int32* pExitCode)
{
	SG_UNUSED(pTool);
	SG_UNUSED(pExitCode);
	SG_ERR_CHECK_RETURN(  _invoke_built_in_tool(pCtx, pTokenValues, SG_difftool__built_in_tool__textfilediff_strict)  );
}
static void _invoke__diff_ignore_eols(SG_context* pCtx, const SG_filetool* pTool, const SG_vhash* pTokenValues, SG_int32* pExitCode)
{
	SG_UNUSED(pTool);
	SG_UNUSED(pExitCode);
	SG_ERR_CHECK_RETURN(  _invoke_built_in_tool(pCtx, pTokenValues, SG_difftool__built_in_tool__textfilediff_ignore_eols)  );
}
static void _invoke__diff_ignore_whitespace(SG_context* pCtx, const SG_filetool* pTool, const SG_vhash* pTokenValues, SG_int32* pExitCode)
{
	SG_UNUSED(pTool);
	SG_UNUSED(pExitCode);
	SG_ERR_CHECK_RETURN(  _invoke_built_in_tool(pCtx, pTokenValues, SG_difftool__built_in_tool__textfilediff_ignore_whitespace)  );
}
static void _invoke__diff_ignore_case(SG_context* pCtx, const SG_filetool* pTool, const SG_vhash* pTokenValues, SG_int32* pExitCode)
{
	SG_UNUSED(pTool);
	SG_UNUSED(pExitCode);
	SG_ERR_CHECK_RETURN(  _invoke_built_in_tool(pCtx, pTokenValues, SG_difftool__built_in_tool__textfilediff_ignore_case)  );
}
static void _invoke__diff_ignore_case_and_whitespace(SG_context* pCtx, const SG_filetool* pTool, const SG_vhash* pTokenValues, SG_int32* pExitCode)
{
	SG_UNUSED(pTool);
	SG_UNUSED(pExitCode);
	SG_ERR_CHECK_RETURN(  _invoke_built_in_tool(pCtx, pTokenValues, SG_difftool__built_in_tool__textfilediff_ignore_case_and_whitespace)  );
}
static void _invoke__diffmerge(SG_context* pCtx, const SG_filetool* pTool, const SG_vhash* pTokenValues, SG_int32* pExitCode)
{
	SG_UNUSED(pTool);
	SG_UNUSED(pExitCode);
	SG_ERR_CHECK_RETURN(  _invoke_built_in_tool(pCtx, pTokenValues, SG_difftool__built_in_tool__exec_diffmerge)  );
}

//////////////////////////////////////////////////////////////////

/**
 * Filetool "plugin" function that locates SG_filetool__invoker
 * functions for the various internal diff tools.
 */
static void _filetool_plugin(
	SG_context*            pCtx,      //< [in] [out] Error and context info.
	const char*            szType,    //< [in] Type of the tool being looked up.
	const char*            szContext, //< [in] The context the tool is being looked up in.
	                                  //<      If NULL, then no specific context is being used.
	const char*            szName,    //< [in] Name of the tool being searched for.
	                                  //<      If NULL, then a custom external tool invoker is being searched for.
	SG_filetool__invoker** ppInvoker  //< [out] Invoker function for the given tool.
	                                  //<       Set to NULL if the tool wasn't found.
	)
{
	SG_uint32             uIndex     = 0u;
	SG_filetool__invoker* pInvoker   = NULL;

	SG_NULLARGCHECK(szType);
	SG_UNUSED(szContext);
	SG_NULLARGCHECK(ppInvoker);

	if (szName != NULL)
	{
		for (uIndex = 0u; uIndex < SG_NrElements(aInternalTools); ++uIndex)
		{
			const _internal_tool* pTool = &(aInternalTools[uIndex]);

			if(strcmp(pTool->szName, szName) == 0)
			{
				pInvoker = pTool->pInvoker;
				break;
			}
		}
	}

	*ppInvoker = pInvoker;
	pInvoker = NULL;

fail:
	return;
}

//////////////////////////////////////////////////////////////////

/**
 * Chooses which tool to use when the user suggested one
 * but we also found a matching one in the settings.
 */
static void _reconcile_suggested_tool(
	SG_context*   pCtx,            //< [in] [out] Error and context info.
	SG_filetool** ppSuggestedTool, //< [in] [out] The tool suggested by the user.
	                               //<            If this tool is output to ppUseTool, this pointer is set to NULL.
	SG_filetool** ppFoundTool,     //< [in] [out] The tool we found in the settings.
	                               //<            If this tool is output to ppUseTool, this pointer is set to NULL.
	SG_filetool** ppUseTool        //< [out] Set to the tool that should be used for the diff.
	)
{
	const char* szSuggestedToolName = NULL;
	const char* szFoundToolName     = NULL;

	SG_NULLARGCHECK(ppSuggestedTool);
	SG_NULLARGCHECK(*ppSuggestedTool);
	SG_NULLARGCHECK(ppFoundTool);
	SG_NULLARGCHECK(*ppFoundTool);
	SG_NULLARGCHECK(ppUseTool);

	// get the names of the tools
	SG_ERR_CHECK(  SG_filetool__get_name(pCtx, *ppSuggestedTool, &szSuggestedToolName)  );
	SG_ERR_CHECK(  SG_filetool__get_name(pCtx, *ppFoundTool, &szFoundToolName)  );

	// compare the tools
	if (strcmp(szSuggestedToolName, szFoundToolName) == 0)
	{
		// both tools are actually the same, so just pick one
		*ppUseTool = *ppSuggestedTool;
	}
	else if (strcmp(szFoundToolName, SG_DIFFTOOL__INTERNAL__SKIP) == 0)
	{
		// for diffs, we'll use the user-suggested tool UNLESS the found tool is SKIP
		// this allows bulk diffs to still skip files that don't make sense to diff
		// files that don't make sense to diff (such as binary files) will be configured to use SKIP in the settings
		*ppUseTool = *ppFoundTool;
	}
	else
	{
		*ppUseTool = *ppSuggestedTool;
	}

	// NULL out the argument that we picked
	if (*ppUseTool == *ppSuggestedTool)
	{
		*ppSuggestedTool = NULL;
	}
	else if (*ppUseTool == *ppFoundTool)
	{
		*ppFoundTool = NULL;
	}
	SG_ASSERT(*ppSuggestedTool == NULL || *ppFoundTool == NULL);

fail:
	return;
}


/*
**
** Public Functions
**
*/

void SG_difftool__exists(
	SG_context* pCtx,
	const char* szName,
	SG_repo*    pRepo,
	SG_bool*    pExists
	)
{
	SG_filetool* pTool = NULL;

	SG_NULLARGCHECK(szName);
	SG_NULLARGCHECK(pExists);

	SG_ERR_CHECK(  SG_filetool__get_tool(pCtx, SG_DIFFTOOL__TYPE, NULL, szName, _filetool_plugin, pRepo, &pTool)  );
	*pExists = (pTool == NULL) ? SG_FALSE : SG_TRUE;

fail:
	SG_FILETOOL_NULLFREE(pCtx, pTool);
	return;
}

void SG_difftool__select(
	SG_context*        pCtx,
	const char*        szPath,
	const SG_pathname* pDiskPath,
	const char*        szSuggestion,
	SG_repo*           pRepo,
	SG_filetool**      ppTool
	)
{
	SG_string*   pClass         = NULL;
	SG_filetool* pSuggestedTool = NULL;
	SG_string*   pFound         = NULL;
	SG_filetool* pFoundTool     = NULL;
	SG_filetool* pFinalTool     = NULL;

	SG_NULLARGCHECK(pDiskPath);
	SG_NULLARGCHECK(ppTool);

	// look up the tool that the user suggested
	// TODO: someday it would be nice to have a mechanism for verifying that the user's suggested
	//       tool actually works with the class of the file that we're working with
	if (szSuggestion != NULL)
	{
		SG_ERR_CHECK(  SG_filetool__get_tool(pCtx, SG_DIFFTOOL__TYPE, NULL, szSuggestion, _filetool_plugin, pRepo, &pSuggestedTool)  );
		if (pSuggestedTool == NULL)
		{
			SG_ERR_THROW2(  SG_ERR_INVALIDARG, (pCtx, "Unknown difftool '%s'.", szSuggestion)  );
		}
	}

	// get the class of the "from" file
	if(pClass==NULL)
		SG_ERR_CHECK(  SG_filetool__get_class(pCtx, szPath, pDiskPath, pRepo, &pClass)  );

	// try to find a tool in the local settings for this class
	SG_ERR_CHECK(  SG_filetool__find_tool(pCtx, SG_DIFFTOOL__TYPE, NULL, SG_string__sz(pClass), pRepo, &pFound)  );
	if (pFound == NULL)
	{
		// couldn't find a tool, use the default one
		SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &pFound, _szDefaultTool)  );
	}
	SG_ASSERT(pFound != NULL);

	// look up the tool we found
	SG_ERR_CHECK(  SG_filetool__get_tool(pCtx, SG_DIFFTOOL__TYPE, NULL, SG_string__sz(pFound), _filetool_plugin, pRepo, &pFoundTool)  );
	SG_ASSERT(pFoundTool != NULL);

	// if the user suggested a tool then we need to decide which one to use
	// otherwise just use the one we found in the settings
	if (pSuggestedTool != NULL)
	{
		SG_ERR_CHECK(  _reconcile_suggested_tool(pCtx, &pSuggestedTool, &pFoundTool, &pFinalTool)  );
	}
	else
	{
		pFinalTool = pFoundTool;
		pFoundTool = NULL;
	}

	// return the tool we found
	*ppTool = pFinalTool;
	pFinalTool = NULL;

fail:
	SG_STRING_NULLFREE(pCtx, pClass);
	SG_FILETOOL_NULLFREE(pCtx, pSuggestedTool);
	SG_STRING_NULLFREE(pCtx, pFound);
	SG_FILETOOL_NULLFREE(pCtx, pFoundTool);
	SG_FILETOOL_NULLFREE(pCtx, pFinalTool);
	return;
}

void SG_difftool__invoke(
	SG_context*        pCtx,
	const SG_filetool* pTool,
	const SG_pathname* pFromPath,
	const SG_string*   pFromLabel,
	const SG_pathname* pToPath,
	const SG_string*   pToLabel,
	SG_bool            bRightSideIsWritable
	)
{
	SG_vhash* pTokenValues = NULL;

	SG_NULLARGCHECK_RETURN(pTool);
	if(pFromPath==NULL && pToPath==NULL)
		SG_ERR_THROW2_RETURN(SG_ERR_INVALIDARG, (pCtx, "pFromPath and pToPath cannot both be NULL"));

	// bundle all of the arguments up into a hash
	SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pTokenValues)  );
	if (pFromPath != NULL)
	{
		SG_ERR_CHECK(  SG_vhash__update__string__sz(pCtx, pTokenValues, gszToken_FromPath,  SG_pathname__sz(pFromPath))  );
		SG_ERR_CHECK(  SG_vhash__update__string__sz(pCtx, pTokenValues, gszToken_FromLabel, SG_pathname__sz(pFromPath))  );
	}
	if (pToPath != NULL)
	{
		SG_ERR_CHECK(  SG_vhash__update__string__sz(pCtx, pTokenValues, gszToken_ToPath,    SG_pathname__sz(pToPath))  );
		SG_ERR_CHECK(  SG_vhash__update__string__sz(pCtx, pTokenValues, gszToken_ToLabel,   SG_pathname__sz(pToPath))  );
	}
	if (pFromLabel != NULL)
	{
		SG_ERR_CHECK(  SG_vhash__update__string__sz(pCtx, pTokenValues, gszToken_FromLabel, SG_string__sz(pFromLabel))  );
	}
	if (pToLabel != NULL)
	{
		SG_ERR_CHECK(  SG_vhash__update__string__sz(pCtx, pTokenValues, gszToken_ToLabel,   SG_string__sz(pToLabel))  );
	}

	SG_ERR_CHECK(  SG_vhash__update__string__sz(pCtx, pTokenValues, gszToken_RightSideIsWritable, ((bRightSideIsWritable)?"T":"F"))  );

	// invoke the tool
	SG_ERR_CHECK(  SG_filetool__invoke(pCtx, pTool, pTokenValues, NULL)  );

fail:
	SG_VHASH_NULLFREE(pCtx, pTokenValues);
	return;
}

void SG_difftool__run(
	SG_context*        pCtx,
	const char*        szPath,
	SG_repo*           pRepo,
	const char*        szSuggestion,
	const SG_pathname* pFromPath,
	const SG_string*   pFromLabel,
	const SG_pathname* pToPath,
	const SG_string*   pToLabel,
	SG_bool            bRightSideIsWritable
	)
{
	SG_filetool* pTool = NULL;
	SG_tempfile * pTempfile = NULL;
	SG_tempfile * pTempfile2 = NULL;

	if(pFromPath==NULL && pToPath==NULL)
		SG_ERR_THROW2_RETURN(SG_ERR_INVALIDARG, (pCtx, "pFromPath and pToPath cannot both be NULL"));

	if(pFromPath!=NULL)
		SG_ERR_CHECK(  SG_difftool__select(pCtx, szPath, pFromPath, szSuggestion, pRepo, &pTool)  );
	else
	{
		//Alloc a temp file to diff against.
		SG_ERR_CHECK(  SG_tempfile__open_new(pCtx, &pTempfile)  );
		SG_ERR_CHECK(  SG_tempfile__close(pCtx, pTempfile)  );
		pFromPath = pTempfile->pPathname;
		SG_ERR_CHECK(  SG_difftool__select(pCtx, szPath, pToPath, szSuggestion, pRepo, &pTool)  );
	}
	if(pToPath==NULL)
	{
		//Alloc a temp file to diff against.
		SG_ERR_CHECK(  SG_tempfile__open_new(pCtx, &pTempfile2)  );
		SG_ERR_CHECK(  SG_tempfile__close(pCtx, pTempfile2)  );
		pToPath = pTempfile2->pPathname;
	}
	SG_ERR_CHECK(  SG_difftool__invoke(pCtx, pTool, pFromPath, pFromLabel, pToPath, pToLabel, bRightSideIsWritable)  );

fail:
	if (pTempfile != NULL)
		SG_tempfile__delete(pCtx, &pTempfile);
	if (pTempfile2 != NULL)
		SG_tempfile__delete(pCtx, &pTempfile2);
	SG_FILETOOL_NULLFREE(pCtx, pTool);
	return;
}

///////////////////////////////////////////////////////////////////////////////

void SG_difftool__built_in_tool__skip(
    SG_context * pCtx,
    const SG_pathname * pPathname0,
    const SG_pathname * pPathname1,
    const SG_string * pLabel0,
    const SG_string * pLabel1,
	SG_bool bRightSideIsWritable,
    SG_string ** ppOutput)
{
	SG_UNUSED(pCtx);
	SG_UNUSED(pPathname0);
	SG_UNUSED(pPathname1);
	SG_UNUSED(pLabel0);
	SG_UNUSED(pLabel1);
	SG_UNUSED( bRightSideIsWritable );
	SG_UNUSED(ppOutput);
}

//////////////////////////////////////////////////////////////////

static void _sg_difftool__perform_textfilediff(
    SG_context * pCtx,
    SG_textfilediff_options diffOptions,
    const SG_pathname * pPathname0,
    const SG_pathname * pPathname1,
    const SG_string * pLabel0,
    const SG_string * pLabel1,
    SG_string **ppOutput)
{
    SG_textfilediff_t * pDiff = NULL;
    SG_string * pOutput = NULL;

    SG_ASSERT(pCtx!=NULL);
    SG_ARGCHECK_RETURN(pPathname0!=NULL || pPathname1!=NULL, pPathname1);

    SG_ERR_CHECK(  SG_textfilediff(pCtx, pPathname0, pPathname1, diffOptions, &pDiff)  );
    SG_ERR_CHECK(  SG_textfilediff__output_unified__string(pCtx, pDiff, pLabel0, pLabel1, 3, &pOutput)  );

    SG_TEXTFILEDIFF_NULLFREE(pCtx, pDiff);

	if(ppOutput!=NULL)
		*ppOutput = pOutput;
	else
	{
		SG_ASSERT(pOutput!=NULL);
		SG_ERR_CHECK(  SG_console(pCtx, SG_CS_STDOUT, "%s", SG_string__sz(pOutput))  );
		SG_STRING_NULLFREE(pCtx, pOutput);
	}

    return;
fail:
    SG_TEXTFILEDIFF_NULLFREE(pCtx, pDiff);
    SG_STRING_NULLFREE(pCtx, pOutput);
}

void SG_difftool__built_in_tool__textfilediff_strict(
    SG_context * pCtx,
    const SG_pathname * pPathname0,
    const SG_pathname * pPathname1,
    const SG_string * pLabel0,
    const SG_string * pLabel1,
	SG_bool bRightSideIsWritable,
    SG_string **ppOutput)
{
	SG_UNUSED( bRightSideIsWritable );

    SG_ERR_CHECK_RETURN(  _sg_difftool__perform_textfilediff(
        pCtx,
        SG_TEXTFILEDIFF_OPTION__STRICT_EOL,
        pPathname0, pPathname1, pLabel0, pLabel1, ppOutput)  );
}

void SG_difftool__built_in_tool__textfilediff_ignore_eols(
    SG_context * pCtx,
    const SG_pathname * pPathname0,
    const SG_pathname * pPathname1,
    const SG_string * pLabel0,
    const SG_string * pLabel1,
	SG_bool bRightSideIsWritable,
    SG_string **ppOutput)
{
	SG_UNUSED( bRightSideIsWritable );

    SG_ERR_CHECK_RETURN(  _sg_difftool__perform_textfilediff(
        pCtx,
        SG_TEXTFILEDIFF_OPTION__REGULAR,
        pPathname0, pPathname1, pLabel0, pLabel1, ppOutput)  );
}

void SG_difftool__built_in_tool__textfilediff_ignore_whitespace(
    SG_context * pCtx,
    const SG_pathname * pPathname0,
    const SG_pathname * pPathname1,
    const SG_string * pLabel0,
    const SG_string * pLabel1,
	SG_bool bRightSideIsWritable,
    SG_string **ppOutput)
{
	SG_UNUSED( bRightSideIsWritable );

    SG_ERR_CHECK_RETURN(  _sg_difftool__perform_textfilediff(
        pCtx,
        SG_TEXTFILEDIFF_OPTION__IGNORE_WHITESPACE,
        pPathname0, pPathname1, pLabel0, pLabel1, ppOutput)  );
}

void SG_difftool__built_in_tool__textfilediff_ignore_case(
    SG_context * pCtx,
    const SG_pathname * pPathname0,
    const SG_pathname * pPathname1,
    const SG_string * pLabel0,
    const SG_string * pLabel1,
	SG_bool bRightSideIsWritable,
    SG_string **ppOutput)
{
	SG_UNUSED( bRightSideIsWritable );

    SG_ERR_CHECK_RETURN(  _sg_difftool__perform_textfilediff(
        pCtx,
        SG_TEXTFILEDIFF_OPTION__STRICT_EOL|SG_TEXTFILEDIFF_OPTION__IGNORE_CASE,
        pPathname0, pPathname1, pLabel0, pLabel1, ppOutput)  );
}

void SG_difftool__built_in_tool__textfilediff_ignore_case_and_whitespace(
    SG_context * pCtx,
    const SG_pathname * pPathname0,
    const SG_pathname * pPathname1,
    const SG_string * pLabel0,
    const SG_string * pLabel1,
	SG_bool bRightSideIsWritable,
    SG_string **ppOutput)
{
	SG_UNUSED( bRightSideIsWritable );

    SG_ERR_CHECK_RETURN(  _sg_difftool__perform_textfilediff(
        pCtx,
        SG_TEXTFILEDIFF_OPTION__IGNORE_CASE|SG_TEXTFILEDIFF_OPTION__IGNORE_WHITESPACE,
        pPathname0, pPathname1, pLabel0, pLabel1, ppOutput)  );
}

//////////////////////////////////////////////////////////////////

void SG_difftool__built_in_tool__exec_diffmerge(
    SG_context * pCtx,
    const SG_pathname * pPathname0,
    const SG_pathname * pPathname1,
    const SG_string * pLabel0,
    const SG_string * pLabel1,
	SG_bool bRightSideIsWritable,
    SG_string ** ppOutput)
{
	SG_pathname * pPathCmd = NULL;
	SG_exec_argvec * pArgVec = NULL;
	SG_file * pNull = NULL;
	SG_exit_status childExitStatus;

	SG_ASSERT(pCtx!=NULL);
	SG_NULLARGCHECK_RETURN(pPathname0);
	SG_NULLARGCHECK_RETURN(pPathname1);

	SG_ERR_CHECK(  SG_filetool__get_builtin_diffmerge_exe(pCtx, &pPathCmd)  );

	SG_ERR_CHECK(  SG_exec_argvec__alloc(pCtx,&pArgVec)  );

	if(pLabel0!=NULL)
	{
		SG_ERR_CHECK(  SG_exec_argvec__append__sz(pCtx, pArgVec, "-t1")  );
		SG_ERR_CHECK(  SG_exec_argvec__append__sz(pCtx, pArgVec, SG_string__sz(pLabel0))  );
	}

	if(pLabel1!=NULL)
	{
		SG_ERR_CHECK(  SG_exec_argvec__append__sz(pCtx, pArgVec, "-t2")  );
		SG_ERR_CHECK(  SG_exec_argvec__append__sz(pCtx, pArgVec, SG_string__sz(pLabel1))  );
	}

	if (!bRightSideIsWritable)
		SG_ERR_CHECK(  SG_exec_argvec__append__sz(pCtx, pArgVec, "-ro2")  );

	SG_ERR_CHECK(  SG_exec_argvec__append__sz(pCtx, pArgVec, SG_pathname__sz(pPathname0))  );
	SG_ERR_CHECK(  SG_exec_argvec__append__sz(pCtx, pArgVec, SG_pathname__sz(pPathname1)  )  );

	// exec diffmerge and wait for exit status.

#if defined(MAC) || defined(LINUX)
	SG_ERR_CHECK(  SG_file__open__null_device(pCtx, SG_FILE_RDONLY, &pNull)  );
#else
	SG_UNUSED(pNull);
#endif
	SG_ERR_CHECK(  SG_exec__exec_sync__files(pCtx, SG_pathname__sz(pPathCmd), pArgVec, NULL, NULL, pNull, &childExitStatus)  );
	SG_FILE_NULLCLOSE(pCtx, pNull);

	if (childExitStatus != 0)
		SG_ERR_THROW2(  SG_ERR_EXTERNAL_TOOL_ERROR,
						(pCtx, "Error returned from '%s'.", SG_pathname__sz(pPathCmd))  );

	if(ppOutput!=NULL)
		SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, ppOutput)  ); // Return an empty string.

fail:
	SG_EXEC_ARGVEC_NULLFREE(pCtx, pArgVec);
	SG_PATHNAME_NULLFREE(pCtx, pPathCmd);
	SG_FILE_NULLCLOSE(pCtx, pNull);
}
