/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg_exec.c
 *
 * @details Routines for shelling-out or otherwise executing a
 * child or subordinate process.
 *
 */

//////////////////////////////////////////////////////////////////

#include <sg.h>
#if defined(MAC) || defined(LINUX)
#include <unistd.h>
#include <sys/wait.h>
#endif

//////////////////////////////////////////////////////////////////

#if defined(DEBUG)
#	define TRACE_EXEC			0
#else
#	define TRACE_EXEC			0
#endif

//////////////////////////////////////////////////////////////////

void _sg_exec__get_last(SG_context* pCtx, const char * sz, const char ** ppResult)
{
	const char * pResult = NULL;
	SG_NULLARGCHECK_RETURN(sz);
	SG_NULLARGCHECK_RETURN(ppResult);
	pResult = sz;
	while(*sz!='\0')
	{
		if('/'==*sz || '\\'==*sz)
			pResult = sz+1;
		sz += 1;
	}
	*ppResult = pResult;
}

void _sg_exec__get_last__string(SG_context* pCtx, const char * sz, SG_string ** ppResult)
{
	SG_ERR_CHECK_RETURN(  _sg_exec__get_last(pCtx, sz, &sz)  );
	SG_ERR_CHECK_RETURN(  SG_STRING__ALLOC__BUF_LEN(pCtx, ppResult, (const SG_byte*)sz, SG_STRLEN(sz))  );
}

#if defined(MAC) || defined(LINUX)
/**
 * Run the child process of synchronous fork/exec.
 * Our caller has already forked.  We are called to
 * do whatever file descriptor table juggling is required
 * and to exec the given command.
 *
 * THIS FUNCTION DOES NOT RETURN.  IT EITHER EXEC() THE
 * GIVEN COMMAND OR IT EXITS WITH AN EXIT STATUS.
 *
 * WE ARE GIVEN A SG_CONTEXT because we need to have one
 * for the routines that we call, but we cannot return a
 * stacktrace to our caller because we do not return.
 */
void _sync__run_child__files(
	SG_context* pCtx,
	const char * pCmd,
	const char ** pArgv,
	SG_file * pFileStdIn,
	SG_file * pFileStdOut,
	SG_file * pFileStdErr)
{
	int status;
	int fd;
	SG_bool bExecAttempted = SG_FALSE;

	if (pFileStdIn)
	{
		SG_ERR_CHECK(  SG_file__get_fd(pCtx,pFileStdIn,&fd)  );
		(void)close(STDIN_FILENO);
		status = dup2(fd,STDIN_FILENO);
		if (status == -1)
			SG_ERR_THROW(  SG_ERR_ERRNO(errno)  );
	}
	else
	{
		// TODO i'm thinking we should let the subordinate process borrow the
		// TODO STDIN of the parent, but i don't have a good reason to say that.
		// TODO
		// TODO Decide if we like this behavior.  Note that if we are in a GUI
		// TODO application, this may be of little use.
	}

	if (pFileStdOut)
	{
		SG_ERR_CHECK(  SG_file__get_fd(pCtx,pFileStdOut,&fd)  );
		(void)close(STDOUT_FILENO);
		status = dup2(fd,STDOUT_FILENO);
		if (status == -1)
			SG_ERR_THROW(  SG_ERR_ERRNO(errno)  );
	}
	else
	{
		// TODO again, i'm going to let the subordinate process borrow the parent's
		// TODO STDOUT.  This might be useful for our command-line client.
	}

	if (pFileStdErr)
	{
		SG_ERR_CHECK(  SG_file__get_fd(pCtx,pFileStdErr,&fd)  );
		(void)close(STDERR_FILENO);
		status = dup2(fd,STDERR_FILENO);
		if (status == -1)
			SG_ERR_THROW(  SG_ERR_ERRNO(errno)  );
	}
	else
	{
		// TODO again, i'm going to let the subordinate process borrow the parent's
		// TODO STDERR.  This might be useful for our command-line client.
	}

	// close all other files (including the other copy of the descriptors
	// that we just dup'd).  this prevents external applications from mucking
	// with any databases that the parent may have open....

	for (fd=STDERR_FILENO+1; fd < (int)FD_SETSIZE; fd++)
		close(fd);

	//////////////////////////////////////////////////////////////////

	execvp( pCmd, (char * const *)pArgv );

	//////////////////////////////////////////////////////////////////

	bExecAttempted = SG_TRUE;	// if we get here, exec() failed.
	SG_ERR_THROW(  SG_ERR_ERRNO(errno)  );

	// fall thru and print something about why the child could not be started.

fail:
#if TRACE_EXEC
	{
		SG_string * pStrErr = NULL;

		SG_context__err_to_string(pCtx, SG_TRUE, &pStrErr);	// IT IS IMPORTANT THAT THIS NOT CALL SG_ERR_IGNORE() because it will hide the info we are reporting.

		// TODO 2010/05/13 We've alredy closed STDERR in the child process, so
		// TODO            this message isn't really going anywhere....

		SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
								   ("execvp() %s:\n"
									"\tCommand: %s\n"
									"\tErrorInfo:\n"
									"%s\n"),
								   ((bExecAttempted)
									? "returned with error"
									: "not attempted because of earlier errors in child"),
								   pCmd,
								   SG_string__sz(pStrErr))  );
		SG_STRING_NULLFREE(pCtx, pStrErr);
		fflush(stderr);
	}
#else
	SG_UNUSED(bExecAttempted);

#endif

	// we we the child.  use _exit() so that none of the atexit() stuff gets run.
	// for example, on Mac/Linux this could cause the debug memory leak report
	// to run and dump a bogus report to stderr.

	_exit(-1);	// an exit of -1 should get mapped to 255 in the parent
}

static void _sync__run_parent__files(SG_context* pCtx,
									 int pid,
									 SG_exit_status * pChildExitStatus,
									 SG_exec_result * per)
{
	int my_errno;
	int statusChild = 0;
	int result;

	//////////////////////////////////////////////////////////////////
	// HACK There is a problem just calling waitpid() on the Mac when
	// HACK in gdb.  We frequently get an interrupted system call.
	// HACK This causes us to abort everything.  I've think I've also
	// HACK seen this from a normal shell.  I think the problem is a
	// HACK race of some kind.
	// HACK
	// HACK The following is an attempt to spin a little until the child
	// HACK gets going enough so that we can actuall wait on it.  I think.
	//
	// TODO consider adding a counter or clock to this so that we don't
	// TODO get stuck here infinitely.  I don't think it'll happen, but
	// TODO it doesn't hurt....
	//////////////////////////////////////////////////////////////////

	while (1)
	{
		result = waitpid(pid,&statusChild,0);
		if (result == pid)
			goto done_waiting;

		my_errno = errno;

#if TRACE_EXEC
		SG_ERR_IGNORE(  SG_console(pCtx,
								   SG_CS_STDERR,
								   "waitpid: [result %d][errno %d]\n",
								   result,my_errno)  );
#endif

		switch (my_errno)
		{
		case 0:
#if 1 || TRACE_EXEC
			if (result > 0)
			{
				// TODO 2011/02/09 if we get no error and a result > 0
				// TODO            it probably means we reaped an async
				// TODO            child.  we should let this loop again
				// TODO            (until we get an ECHILD or the child
				// TODO            that we are expecting), right?
				SG_ERR_IGNORE(  SG_console(pCtx,
										   SG_CS_STDERR,
										   "waitpid: reaper a different child? [result %d][errno %d]\n",
										   result,my_errno)  );
			}
#endif
		default:
		case ECHILD:		// we don't have a child ? ? ?
			*per = SG_EXEC_RESULT__UNKNOWN;
			*pChildExitStatus = SG_EXIT_STATUS_BOGUS;
			return;

		case EDEADLK:		// resource deadlock avoided
		case EAGAIN:		// resource temporarily unavailable
		case EINTR:			// interrupted system call
			break;			// try again
		}
	}

done_waiting:

	if (WIFEXITED(statusChild))
	{
		// child exited properly (with or without error)

		SG_exit_status child_exit_status = WEXITSTATUS(statusChild);

#if TRACE_EXEC
		SG_ERR_IGNORE(  SG_console(pCtx,
								   SG_CS_STDERR,
								   "waitpid: child exited: [pid %d] [0x%x] [exit_status %d]\n",
								   pid, statusChild, child_exit_status)  );
#endif

		if (child_exit_status == 255)		// -1 from child
		{
			// this usually means that the child could not exec the program
			// for whatever reason, such as no access to the exe or the exe
			// did not exist.
			// 
			// we can't tell at this point if it was a EACCESS or a ENOENT.

			*per = SG_EXEC_RESULT__FAILED;
			*pChildExitStatus = SG_EXIT_STATUS_BOGUS;
		}
		else
		{
			*per = SG_EXEC_RESULT__NORMAL;
			*pChildExitStatus = child_exit_status;
		}
	}
	else
	{
		// child crashed or was killed: WTERMSIG() or WCOREDUMP() or ...

#if 1 || TRACE_EXEC
		SG_ERR_IGNORE(  SG_console(pCtx,
							   SG_CS_STDERR,
								   "waitpid: child crashed/killed: [pid %d] [0x%x] [signaled %x][stopped %x] [?term? %d] [?stop? %d]\n",
								   pid,
								   statusChild,
								   WIFSIGNALED(statusChild),WIFSTOPPED(statusChild),
								   ((WIFSIGNALED(statusChild)) ? WTERMSIG(statusChild) : -1),
								   ((WIFSTOPPED(statusChild)) ? WSTOPSIG(statusChild) : -1))  );
#endif

		*per = SG_EXEC_RESULT__SET_ABNORMAL( (WTERMSIG(statusChild)) ); // i only really care about signals
		*pChildExitStatus = SG_EXIT_STATUS_BOGUS;
	}
}

void SG_exec__exec_async__files(
	SG_context* pCtx,
	const char * pCmd,
	const SG_exec_argvec * pArgVec,
	SG_file * pFileStdIn,
	SG_file * pFileStdOut,
	SG_file * pFileStdErr,
	SG_process_id * pProcessID)
{
	// The caller passes a SG_vector of SG_string for ARGV.
	// They DO NOT include the program-to-exec in the vector.

	SG_uint32 nrArgsGiven = 0;
	SG_uint32 nrArgsTotal;
	SG_uint32 kArg;
	const char ** pArgv = NULL;
	const char * pStrCmdEntryname = NULL;
	int pid;

	SG_NULLARGCHECK_RETURN(pCmd);
	SG_NULLARGCHECK_RETURN(pProcessID);

	if (pArgVec)
		SG_ERR_CHECK_RETURN(  SG_exec_argvec__length(pCtx,pArgVec,&nrArgsGiven)  );
	nrArgsTotal = nrArgsGiven + 2;		// +1 for program, +1 for terminating null.

	// We build a platform-appropriate "char * argv[]" to give to execv().

	SG_ERR_CHECK_RETURN(  SG_alloc(pCtx,nrArgsTotal,sizeof(char *),&pArgv)  );

	// On Linux/Mac, we want argv[0] to be the ENTRYNAME of the program.

	// TODO I'm thinking we should convert each arg from utf-8 to an os-buffer
	// TODO aka locale buffer and give that to exec().

	SG_ERR_CHECK(  _sg_exec__get_last(pCtx,pCmd,&pStrCmdEntryname)  );
	pArgv[0] = pStrCmdEntryname;

	for (kArg=0; kArg<nrArgsGiven; kArg++)
	{
		const SG_string * pStr_k;

		SG_ERR_CHECK(  SG_exec_argvec__get(pCtx,pArgVec,kArg,&pStr_k)  );
		if (!pStr_k || (SG_string__length_in_bytes(pStr_k) == 0))
		{
			// allow for blank or empty cell as a "" entry.  (though,
			// i'm not sure why.)

			pArgv[kArg+1] = "";
		}
		else
		{
			// TODO utf-8 --> os-buffer ??
			pArgv[kArg+1] = SG_string__sz(pStr_k);
		}
	}

	// pArgv[nrArgsTotal-1] should already be NULL

#if TRACE_EXEC
	{
		SG_uint32 k;
		SG_ERR_IGNORE(  SG_console(pCtx,
								   SG_CS_STDERR,
								   "Exec [%s]\n",
								   pCmd)  );
		for (k=0; k<nrArgsTotal; k++)
		{
			SG_ERR_IGNORE(  SG_console(pCtx,
									   SG_CS_STDERR,
									   "     Argv[%d] %s\n",
									   k,pArgv[k])  );
		}
	}
#endif

	// just incase we have anything buffered in our output FILE streams
	// (and the child is sharing them), flush them out before the child
	// gets started.

	fflush(stdout);
	fflush(stderr);

	pid = fork();
	if (pid == -1)
	{
		SG_ERR_THROW(  SG_ERR_ERRNO(errno)  );
	}
	else if (pid == 0)
	{
		// we are in the child process.
		//
		// note that we ***CANNOT*** use SG_ERR_CHECK() and the normal
		// goto fail, cleanup, and return -- whether we succeed or fail,
		// we must exit when we are finished.

		_sync__run_child__files(pCtx,pCmd,pArgv,pFileStdIn,pFileStdOut,pFileStdErr);
		_exit(-1);	// not reached
		return;		// not reached
	}

	// otherwise, we are the parent process.
	//For now, do nothing.
	*pProcessID = (SG_process_id)pid;

	// fall thru to common cleanup.

fail:
	SG_NULLFREE(pCtx, pArgv);
}

void SG_exec__exec_sync__files__details(
	SG_context* pCtx,
	const char * pCmd,
	const SG_exec_argvec * pArgVec,
	SG_file * pFileStdIn,
	SG_file * pFileStdOut,
	SG_file * pFileStdErr,
	SG_exit_status * pChildExitStatus,
	SG_exec_result * per,
	SG_process_id * pPid)
{
	// The caller passes a SG_vector of SG_string for ARGV.
	// They DO NOT include the program-to-exec in the vector.

	SG_uint32 nrArgsGiven = 0;
	SG_uint32 nrArgsTotal;
	SG_uint32 kArg;
	const char ** pArgv = NULL;
	const char * pStrCmdEntryname = NULL;
	int pid;
	SG_exit_status exitStatus = 0;
	SG_exec_result execResult = SG_EXEC_RESULT__UNKNOWN;

	SG_NULLARGCHECK_RETURN(pCmd);
	SG_NULLARGCHECK_RETURN(pChildExitStatus);
	SG_NULLARGCHECK_RETURN(per);
	SG_NULLARGCHECK_RETURN(pPid);

	if (pArgVec)
		SG_ERR_CHECK_RETURN(  SG_exec_argvec__length(pCtx,pArgVec,&nrArgsGiven)  );
	nrArgsTotal = nrArgsGiven + 2;		// +1 for program, +1 for terminating null.

	// We build a platform-appropriate "char * argv[]" to give to execv().

	SG_ERR_CHECK_RETURN(  SG_alloc(pCtx,nrArgsTotal,sizeof(char *),&pArgv)  );

	// On Linux/Mac, we want argv[0] to be the ENTRYNAME of the program.

	// TODO I'm thinking we should convert each arg from utf-8 to an os-buffer
	// TODO aka locale buffer and give that to exec().

	SG_ERR_CHECK(  _sg_exec__get_last(pCtx,pCmd,&pStrCmdEntryname)  );
	pArgv[0] = pStrCmdEntryname;

	for (kArg=0; kArg<nrArgsGiven; kArg++)
	{
		const SG_string * pStr_k;

		SG_ERR_CHECK(  SG_exec_argvec__get(pCtx,pArgVec,kArg,&pStr_k)  );
		if (!pStr_k || (SG_string__length_in_bytes(pStr_k) == 0))
		{
			// allow for blank or empty cell as a "" entry.  (though,
			// i'm not sure why.)

			pArgv[kArg+1] = "";
		}
		else
		{
			// TODO utf-8 --> os-buffer ??
			pArgv[kArg+1] = SG_string__sz(pStr_k);
		}
	}

	// pArgv[nrArgsTotal-1] should already be NULL

#if TRACE_EXEC
	{
		SG_uint32 k;
		SG_ERR_IGNORE(  SG_console(pCtx,
								   SG_CS_STDERR,
								   "Exec [%s]\n",
								   pCmd)  );
		for (k=0; k<nrArgsTotal; k++)
		{
			SG_ERR_IGNORE(  SG_console(pCtx,
									   SG_CS_STDERR,
									   "     Argv[%d] %s\n",
									   k,pArgv[k])  );
		}
	}
#endif

	// just incase we have anything buffered in our output FILE streams
	// (and the child is sharing them), flush them out before the child
	// gets started.

	fflush(stdout);
	fflush(stderr);

	pid = fork();
	if (pid == -1)
	{
		SG_ERR_THROW(  SG_ERR_ERRNO(errno)  );
	}
	else if (pid == 0)
	{
		// we are in the child process.
		//
		// note that we ***CANNOT*** use SG_ERR_CHECK() and the normal
		// goto fail, cleanup, and return -- whether we succeed or fail,
		// we must exit when we are finished.

		_sync__run_child__files(pCtx,pCmd,pArgv,pFileStdIn,pFileStdOut,pFileStdErr);
		_exit(-1);	// not reached
		return;		// not reached
	}

	// otherwise, we are the parent process.

	SG_ERR_CHECK(  _sync__run_parent__files(pCtx,pid,&exitStatus,&execResult)  );

	// if have an normal setup/argcheck error and don't get
	// to the point where we try to start the child, we
	// throw just like any other sglib function.  this has
	// already been handled before we did the fork.
	//
	// if we actually try to start the child, we return an exec_result
	// describing what happened to the child.

#if TRACE_EXEC
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "Exec [result %x : %d][status %d] [%s]\n",
							   (execResult & SG_EXEC_RESULT__MASK_TYPE),
							   (execResult & SG_EXEC_RESULT__MASK_VALUE),
							   exitStatus,
							   pCmd)  );
#endif

	*pChildExitStatus = exitStatus;
	*per = execResult;
	*pPid = pid;

fail:
	SG_NULLFREE(pCtx, pArgv);
}

//////////////////////////////////////////////////////////////////

#include "sg_exec__debug_stacktrace.h"

//////////////////////////////////////////////////////////////////

void SG_exec__exec_sync__files(
	SG_context* pCtx,
	const char * pCmd,
	const SG_exec_argvec * pArgVec,
	SG_file * pFileStdIn,
	SG_file * pFileStdOut,
	SG_file * pFileStdErr,
	SG_exit_status * pChildExitStatus)
{
	SG_exit_status childExitStatus;
	SG_exec_result execResult;
	SG_process_id pid;

	SG_ERR_CHECK(  SG_exec__exec_sync__files__details(pCtx, pCmd, pArgVec,
													  pFileStdIn, pFileStdOut, pFileStdErr,
													  &childExitStatus, &execResult, &pid)  );
	
	// since caller didn't want to micro-manage and only
	// wants to have to deal with the exit-status of a
	// regular child that started, ran, and exited
	// properly (with or without an error status), we
	// throw if there was a problem.

	switch (execResult & SG_EXEC_RESULT__MASK_TYPE)
	{
	default:
	case SG_EXEC_RESULT__UNKNOWN:
	case SG_EXEC_RESULT__FAILED:
		SG_ERR_THROW2(  SG_ERR_EXEC_FAILED,
						(pCtx, "%s", pCmd)  );
		break;

	case SG_EXEC_RESULT__ABNORMAL:
#if defined(HAVE_EXEC_DEBUG_STACKTRACE)
		SG_ERR_IGNORE(  SG_exec_debug__get_stacktrace(pCtx, pCmd, pid)  );
#endif
		SG_ERR_THROW2(  SG_ERR_ABNORMAL_TERMINATION,
						(pCtx, "sig(%d): %s",
						 SG_EXEC_RESULT__GET_SUB_VALUE(execResult),
						 pCmd)  );
		break;

	case SG_EXEC_RESULT__NORMAL:
		*pChildExitStatus = childExitStatus;
		break;
	}

fail:
	return;
}
#endif // defined(MAC) || defined(LINUX)

//////////////////////////////////////////////////////////////////

#if defined(WINDOWS)

void _sg_exec__append_arg(
	SG_context * pCtx,
	SG_string * pStrCmdLine,
	const SG_string * pStr_k)
{
	SG_string * pStr = NULL;

	if (!pStr_k || (SG_string__length_in_bytes(pStr_k) == 0))
	{
		SG_ERR_CHECK(  SG_string__append__sz(pCtx,pStrCmdLine,"\"\"")  );
	}
	else
	{
		SG_ERR_CHECK(  SG_string__append__sz(pCtx,pStrCmdLine,"\"")  );

		SG_ERR_CHECK(  SG_STRING__ALLOC__COPY(pCtx, &pStr, pStr_k)  );
		SG_ERR_CHECK(  SG_string__replace_bytes(pCtx, pStr,
			(const SG_byte*)"\"", 1,
			(const SG_byte*)"\\\"", 2,
			SG_UINT32_MAX, SG_TRUE, NULL)  );
		SG_ERR_CHECK(  SG_string__append__string(pCtx,pStrCmdLine,pStr)  );
		SG_STRING_NULLFREE(pCtx, pStr);

		SG_ERR_CHECK(  SG_string__append__sz(pCtx,pStrCmdLine,"\"")  );
	}

	SG_ERR_CHECK(  SG_string__append__sz(pCtx,pStrCmdLine," ")  );

fail:
	SG_STRING_NULLFREE(pCtx, pStr);
}

void SG_exec__exec_async__files(
	SG_context* pCtx,
	const char * pCmd,
	const SG_exec_argvec * pArgVec,
	SG_file * pFileStdIn,
	SG_file * pFileStdOut,
	SG_file * pFileStdErr,
	SG_process_id * pProcessID)
{
	// The caller passes a SG_vector of SG_string for ARGV.
	// They DO NOT include the program-to-exec in the vector.

	SG_string * pStrCmdLine = NULL;
	SG_string * pCmd__string = NULL;
	wchar_t * pWcharCmdLine = NULL;
    PROCESS_INFORMATION pi;
    STARTUPINFO si;
	DWORD dwCreationFlags;
	//DWORD dwChildExitStatus;
	HANDLE hFileStdIn, hFileStdOut, hFileStdErr;

	SG_NULLARGCHECK_RETURN(pCmd);
	SG_NULLARGCHECK_RETURN(pProcessID);

	memset(&pi,0,sizeof(pi));
	memset(&si,0,sizeof(si));

	SG_ERR_CHECK_RETURN(  SG_STRING__ALLOC(pCtx, &pStrCmdLine)  );

	SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &pCmd__string, pCmd)  );
	SG_ERR_CHECK(  _sg_exec__append_arg(pCtx, pStrCmdLine, pCmd__string)  );

	if (pArgVec)
	{
		SG_uint32 kArg, nrArgsGiven;
		SG_ERR_CHECK(  SG_exec_argvec__length(pCtx,pArgVec,&nrArgsGiven)  );
		for (kArg=0; kArg<nrArgsGiven; kArg++)
		{
			const SG_string * pStr_k = NULL;
			SG_ERR_CHECK(  SG_exec_argvec__get(pCtx,pArgVec,kArg,&pStr_k)  );
			SG_ERR_CHECK(  _sg_exec__append_arg(pCtx, pStrCmdLine, pStr_k)  );
		}
	}

#if TRACE_EXEC
	SG_ERR_IGNORE(  SG_console(pCtx,
							   SG_CS_STDERR,
							   "Exec [%s]\n",	// this is close; it won't have the UNC prefix
							   pCmd)  );
	SG_ERR_IGNORE(  SG_console(pCtx,
							   SG_CS_STDERR,
							   "    Args: [%s]\n",
							   SG_string__sz(pStrCmdLine))  );
#endif

	// convert UTF-8 version of the command pathname and the command line to wchar_t.
	SG_ERR_CHECK(  SG_utf8__extern_to_os_buffer__wchar(pCtx,SG_string__sz(pStrCmdLine),&pWcharCmdLine,NULL)  );

    si.cb = sizeof(si);
	si.dwFlags = STARTF_USESTDHANDLES;

	// deal with STDIN/STDOUT/STDERR in the child/subordinate process.
	//
	// if the caller gave us open files, we use the file handles within
	// them; otherwise, we clone our STDIN and friends.
	// we must assume that the files provided by the caller are correctly
	// opened and positioned....

	if (pFileStdIn)
		SG_ERR_CHECK(  SG_file__get_handle(pCtx,pFileStdIn,&hFileStdIn)  );
	else
		hFileStdIn = GetStdHandle(STD_INPUT_HANDLE);

	if (pFileStdOut)
		SG_ERR_CHECK(  SG_file__get_handle(pCtx,pFileStdOut,&hFileStdOut)  );
	else
		hFileStdOut = GetStdHandle(STD_OUTPUT_HANDLE);

	if (pFileStdErr)
		SG_ERR_CHECK(  SG_file__get_handle(pCtx,pFileStdErr,&hFileStdErr)  );
	else
		hFileStdErr = GetStdHandle(STD_ERROR_HANDLE);

	if (!DuplicateHandle(GetCurrentProcess(),hFileStdIn,GetCurrentProcess(),
						 &si.hStdInput,0,TRUE,DUPLICATE_SAME_ACCESS))
	{
		//	SG_ERR_THROW(  SG_ERR_GETLASTERROR(GetLastError())  );
	}

	if (!DuplicateHandle(GetCurrentProcess(),hFileStdOut,GetCurrentProcess(),
						 &si.hStdOutput,0,TRUE,DUPLICATE_SAME_ACCESS))
	{
		//	SG_ERR_THROW(  SG_ERR_GETLASTERROR(GetLastError())  );
	}

	if (!DuplicateHandle(GetCurrentProcess(),hFileStdErr,GetCurrentProcess(),
						 &si.hStdError,0,TRUE,DUPLICATE_SAME_ACCESS))
	{
		//	SG_ERR_THROW(  SG_ERR_GETLASTERROR(GetLastError())  );
	}

#if defined(DEBUG)
	//for debug builds, start in a new, easy-to-kill window
	dwCreationFlags = 0;
#else
	//for release builds, don't bother with a new window.
	dwCreationFlags = CREATE_NO_WINDOW;
#endif

	if (!CreateProcessW(NULL,pWcharCmdLine,
						NULL,NULL,
						TRUE,dwCreationFlags,
						NULL,NULL,
						&si,&pi))
		SG_ERR_THROW2(  SG_ERR_GETLASTERROR(GetLastError()),
						(pCtx,
						 "Cannot execute command [%s] with args [%s]",
						 pCmd,
						 SG_string__sz(pStrCmdLine))  );

	*pProcessID = (SG_process_id) pi.dwProcessId;
	// fall thru to common cleanup.

fail:

	CloseHandle(pi.hProcess);
	CloseHandle(pi.hThread);

	SG_STRING_NULLFREE(pCtx, pStrCmdLine);
	SG_STRING_NULLFREE(pCtx, pCmd__string);
	SG_NULLFREE(pCtx, pWcharCmdLine);

	CloseHandle(si.hStdInput);
	CloseHandle(si.hStdOutput);
	CloseHandle(si.hStdError);
}

void SG_exec__exec_sync__files__details(
	SG_context* pCtx,
	const char * pCmd,
	const SG_exec_argvec * pArgVec,
	SG_file * pFileStdIn,
	SG_file * pFileStdOut,
	SG_file * pFileStdErr,
	SG_exit_status * pChildExitStatus,
	SG_exec_result * per,
	SG_process_id * pPid)
{
	// The caller passes a SG_vector of SG_string for ARGV.
	// They DO NOT include the program-to-exec in the vector.

	SG_string * pStrCmdLine = NULL;
	SG_string * pCmd__string = NULL;
	wchar_t * pWcharCmdLine = NULL;
    PROCESS_INFORMATION pi;
    STARTUPINFO si;
	DWORD dwCreationFlags;
	DWORD dwChildExitStatus;
	HANDLE hFileStdIn, hFileStdOut, hFileStdErr;

	SG_NULLARGCHECK_RETURN(pCmd);
	SG_NULLARGCHECK_RETURN(pChildExitStatus);
	// per is optional
	// pPid is optional

	memset(&pi,0,sizeof(pi));
	memset(&si,0,sizeof(si));

	SG_ERR_CHECK_RETURN(  SG_STRING__ALLOC(pCtx, &pStrCmdLine)  );

	SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &pCmd__string, pCmd)  );
	SG_ERR_CHECK(  _sg_exec__append_arg(pCtx, pStrCmdLine, pCmd__string)  );

	if (pArgVec)
	{
		SG_uint32 kArg, nrArgsGiven;
		SG_ERR_CHECK(  SG_exec_argvec__length(pCtx,pArgVec,&nrArgsGiven)  );
		for (kArg=0; kArg<nrArgsGiven; kArg++)
		{
			const SG_string * pStr_k = NULL;
			SG_ERR_CHECK(  SG_exec_argvec__get(pCtx,pArgVec,kArg,&pStr_k)  );
			SG_ERR_CHECK(  _sg_exec__append_arg(pCtx, pStrCmdLine, pStr_k)  );
		}
	}

#if TRACE_EXEC
	SG_ERR_IGNORE(  SG_console(pCtx,
							   SG_CS_STDERR,
							   "Exec [%s]\n",	// this is close; it won't have the UNC prefix
							   pCmd)  );
	SG_ERR_IGNORE(  SG_console(pCtx,
							   SG_CS_STDERR,
							   "    Args: [%s]\n",
							   SG_string__sz(pStrCmdLine))  );
#endif

	// convert UTF-8 version of the command pathname and the command line to wchar_t.
	SG_ERR_CHECK(  SG_utf8__extern_to_os_buffer__wchar(pCtx,SG_string__sz(pStrCmdLine),&pWcharCmdLine,NULL)  );

    si.cb = sizeof(si);
	si.dwFlags = STARTF_USESTDHANDLES;

	// deal with STDIN/STDOUT/STDERR in the child/subordinate process.
	//
	// if the caller gave us open files, we use the file handles within
	// them; otherwise, we clone our STDIN and friends.
	// we must assume that the files provided by the caller are correctly
	// opened and positioned....

	if (pFileStdIn)
		SG_ERR_CHECK(  SG_file__get_handle(pCtx,pFileStdIn,&hFileStdIn)  );
	else
		hFileStdIn = GetStdHandle(STD_INPUT_HANDLE);

	if (pFileStdOut)
		SG_ERR_CHECK(  SG_file__get_handle(pCtx,pFileStdOut,&hFileStdOut)  );
	else
		hFileStdOut = GetStdHandle(STD_OUTPUT_HANDLE);

	if (pFileStdErr)
		SG_ERR_CHECK(  SG_file__get_handle(pCtx,pFileStdErr,&hFileStdErr)  );
	else
		hFileStdErr = GetStdHandle(STD_ERROR_HANDLE);

	if (!DuplicateHandle(GetCurrentProcess(),hFileStdIn,GetCurrentProcess(),
						 &si.hStdInput,0,TRUE,DUPLICATE_SAME_ACCESS))
		SG_ERR_THROW(  SG_ERR_GETLASTERROR(GetLastError())  );

	if (!DuplicateHandle(GetCurrentProcess(),hFileStdOut,GetCurrentProcess(),
						 &si.hStdOutput,0,TRUE,DUPLICATE_SAME_ACCESS))
		SG_ERR_THROW(  SG_ERR_GETLASTERROR(GetLastError())  );

	if (!DuplicateHandle(GetCurrentProcess(),hFileStdErr,GetCurrentProcess(),
						 &si.hStdError,0,TRUE,DUPLICATE_SAME_ACCESS))
		SG_ERR_THROW(  SG_ERR_GETLASTERROR(GetLastError())  );

	// errors that happen before we try to create the child process
	// throw like normal because they are a problem that the parent
	// had.  just like the code before the fork() in the linux/mac
	// version.

	dwCreationFlags = 0;
	if (!CreateProcessW(NULL,pWcharCmdLine,
						NULL,NULL,
						TRUE,dwCreationFlags,
						NULL,NULL,
						&si,&pi))
	{
		DWORD dwErr = GetLastError();
		SG_error e = SG_ERR_GETLASTERROR(dwErr);

		// I'm going to say that if CreateProcessW() fails that that
		// is equivalent to the execvp() failing after the fork()
		// such that the child process exists and exits it -1 because
		// it can't do anything else.
		//
		// here we just return a normalized failed-to-start.

#if TRACE_EXEC
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "CreateProcessW failed %x for '%s %s'\n",
								   dwErr, pCmd, SG_string__sz(pStrCmdLine))  );
#endif

		if (per)
		{
			*per = SG_EXEC_RESULT__FAILED;
			*pChildExitStatus = SG_EXIT_STATUS_BOGUS;

			if (pPid)
				*pPid = 0;
		}
		else
		{
			// throw a compound error message (with the nice details
			// that the OS gave us, but cross-platform normalized for
			// our callers).

			char buf[SG_ERROR_BUFFER_SIZE];

			SG_error__get_message(e, SG_TRUE, buf, sizeof(buf));
			SG_ERR_THROW2(  SG_ERR_EXEC_FAILED,
							(pCtx, "%s: %s %s", buf, pCmd, SG_string__sz(pStrCmdLine))  );
		}

		goto fail;
	}

	if (WaitForSingleObject(pi.hProcess,INFINITE) != WAIT_OBJECT_0)
	{
		DWORD dwErr = GetLastError();
		SG_error e = SG_ERR_GETLASTERROR(dwErr);

		// I'm going to say that if WaitForSingleObject() fails that
		// that is equivalent to the waitpid() failing with ECHILD.
		// This probably can't happen, but the net-net is that we don't
		// know what happened to the process.
		//
		// normalize this to unknown.

#if TRACE_EXEC
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "WaitForSingleObject failed %x for '%s %s'\n",
								   dwErr, pCmd, SG_string__sz(pStrCmdLine))  );
#endif
		if (per)
		{
			*per = SG_EXEC_RESULT__UNKNOWN;
			*pChildExitStatus = SG_EXIT_STATUS_BOGUS;

			if (pPid)
				*pPid = 0;
		}
		else
		{
			char buf[SG_ERROR_BUFFER_SIZE];

			SG_error__get_message(e, SG_TRUE, buf, sizeof(buf));
			SG_ERR_THROW2(  SG_ERR_EXEC_FAILED,
							(pCtx, "%s: %s %s", buf, pCmd, SG_string__sz(pStrCmdLine))  );
		}

		goto fail;
	}

	if (!GetExitCodeProcess(pi.hProcess,&dwChildExitStatus))
	{
		DWORD dwErr = GetLastError();
		SG_error e = SG_ERR_GETLASTERROR(dwErr);

		// this probably can't happen either, but if it does
		// we still don't know what happened to the process.
		//
		// again, we want to normalize this.

#if TRACE_EXEC
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "GetExitCodeProcess failed %x for '%s %s'\n",
								   dwErr, pCmd, SG_string__sz(pStrCmdLine))  );
#endif

		if (per)
		{
			*per = SG_EXEC_RESULT__UNKNOWN;
			*pChildExitStatus = SG_EXIT_STATUS_BOGUS;

			if (pPid)
				*pPid = 0;
		}
		else
		{
			char buf[SG_ERROR_BUFFER_SIZE];

			SG_error__get_message(e, SG_TRUE, buf, sizeof(buf));
			SG_ERR_THROW2(  SG_ERR_EXEC_FAILED,
							(pCtx, "%s: %s %s", buf, pCmd, SG_string__sz(pStrCmdLine))  );
		}

		goto fail;
	}

	// TODO 2011/02/10 dwChildExitStatus can be either the exit value
	// TODO            if the process exited, an unhandled exception,
	// TODO            or the value sent by another process calling
	// TODO            TerminateProcess.
	// TODO
	// TODO            figure out if there is a way to distinguish
	// TODO            these and return (__ABNORMAL,__STATUS_BOGUS)
	// TODO            for them (and maybe put something in the
	// TODO            value portion).

	if (per)
		*per = SG_EXEC_RESULT__NORMAL;
	if (pPid)
		*pPid = (SG_process_id)pi.dwProcessId;
	*pChildExitStatus = (SG_exit_status)dwChildExitStatus;

	// fall thru to common cleanup.

fail:

	CloseHandle(pi.hProcess);
	CloseHandle(pi.hThread);

	SG_STRING_NULLFREE(pCtx, pStrCmdLine);
	SG_STRING_NULLFREE(pCtx, pCmd__string);
	SG_NULLFREE(pCtx, pWcharCmdLine);

	CloseHandle(si.hStdInput);
	CloseHandle(si.hStdOutput);
	CloseHandle(si.hStdError);
}

void SG_exec__exec_sync__files(
	SG_context* pCtx,
	const char * pCmd,
	const SG_exec_argvec * pArgVec,
	SG_file * pFileStdIn,
	SG_file * pFileStdOut,
	SG_file * pFileStdErr,
	SG_exit_status * pChildExitStatus)
{
	SG_ERR_CHECK_RETURN(  SG_exec__exec_sync__files__details(pCtx, pCmd, pArgVec,
															 pFileStdIn, pFileStdOut, pFileStdErr,
															 pChildExitStatus, NULL, NULL)  );
}
#endif

void SG_exec__system(
	SG_context* pCtx,
	const char* pszCmdLine,
	SG_exit_status* pExitStatus)
{
	int exitStatus = 0;

#if defined(WINDOWS)
	SG_string* pstr = NULL;
	wchar_t* pwszCmdLine = NULL;

	SG_ERR_CHECK(  SG_string__alloc__format(pCtx, &pstr, "\"%s\"", pszCmdLine)  );
	SG_ERR_CHECK(  SG_utf8__extern_to_os_buffer__wchar(pCtx, SG_string__sz(pstr), &pwszCmdLine, NULL)  );

	exitStatus = _wsystem(pwszCmdLine);

	SG_STRING_NULLFREE(pCtx, pstr);
	SG_NULLFREE(pCtx, pwszCmdLine);
#else
	SG_UNUSED(pCtx);
	exitStatus = system(pszCmdLine);
#endif 

	if (pExitStatus)
		*pExitStatus = exitStatus;

	return;

#if defined(WINDOWS)
fail:
	SG_STRING_NULLFREE(pCtx, pstr);
	SG_NULLFREE(pCtx, pwszCmdLine);
#endif
}
