/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include <sg.h>

//////////////////////////////////////////////////////////////////

#if defined(DEBUG)
#define TRACE_DIFFMERGE_LOOKUP	0
#else
#define TRACE_DIFFMERGE_LOOKUP	0
#endif

//////////////////////////////////////////////////////////////////

/*
**
** Types
**
*/

/**
 * Specification of a single file tool.
 */
struct _SG_filetool
{
	SG_string*            pName;       //< The unique name of the tool.
	SG_filetool__invoker* fInvoker;    //< A function that runs the tool's logic.
	SG_string*            pExecutable; //< The tool's executable, or NULL if it's internal.
	                                   //< Might be an absolute path, relative path, or just filename (passed straight to SG_exec).
	SG_varray*            pArguments;  //< List of command line arguments to pass to the tool, or NULL if it's built-in.
	                                   //< These arguments contain tokens to be replaced with actual values when the tool is invoked.
	SG_string*            pStdIn;      //< Token that will resolve to the filename to bind to the tool's STDIN stream.
	                                   //< NULL if the tool's internal, or doesn't use STDIN.
	SG_string*            pStdOut;     //< Token that will resolve to the filename to bind to the tool's STDOUT stream.
	                                   //< NULL if the tool's internal, or doesn't use STDOUT.
	SG_string*            pStdErr;     //< Token that will resolve to the filename to bind to the tool's STDERR stream.
	                                   //< NULL if the tool's internal, or doesn't use STDERR.
};


/*
**
** Globals
**
*/

/**
 * List of built-in tools.
 */
static const struct
{
	const char*           szName;   //< Name of the built-in tool.
	SG_filetool__invoker* fInvoker; //< Function that implements the tool.
}
gpBuiltinFiletools[] =
{
	{ SG_FILETOOL__INTERNAL__NULL, SG_filetool__invoke__null },
};

/**
 * "Magic" numbers used as the first bytes in well-known file formats.
 */
static const SG_byte aMagic_UTF32BE[] = { 0x00, 0x00, 0xFE, 0xFF }; // Unicode UTF-32 Big-Endian
static const SG_byte aMagic_UTF32LE[] = { 0xFF, 0xFE, 0x00, 0x00 }; // Unicode UTF-32 Little-Endian
static const SG_byte aMagic_UTF16BE[] = { 0xFE, 0xFF };             // Unicode UTF-16 Big Endian
static const SG_byte aMagic_UTF16LE[] = { 0xFF, 0xFE };             // Unicode UTF-16 Little Endian
static const SG_byte aMagic_UTF8[]    = { 0xEF, 0xBB, 0xBF };       // Unicode UTF-8

/**
 * List of known magic numbers.
 */
static const struct
{
	const SG_byte* pBytes;  //< The bytes that make up the actual magic number.
	SG_uint32      uSize;   //< The number of bytes pointed to by pBytes.
	const char*    szClass; //< The default built-in file class of files with this magic number.
}
gpFileClassMagicNumbers[] =
{
	{ aMagic_UTF32BE, SG_NrElements(aMagic_UTF32BE), SG_FILETOOL__CLASS__TEXT },
	{ aMagic_UTF32LE, SG_NrElements(aMagic_UTF32LE), SG_FILETOOL__CLASS__TEXT },
	{ aMagic_UTF16BE, SG_NrElements(aMagic_UTF16BE), SG_FILETOOL__CLASS__TEXT },
	{ aMagic_UTF16LE, SG_NrElements(aMagic_UTF16LE), SG_FILETOOL__CLASS__TEXT },
	{ aMagic_UTF8,    SG_NrElements(aMagic_UTF8),    SG_FILETOOL__CLASS__TEXT },
	// TODO: others?
};


/*
**
** Internal Functions
**
*/

/**
 * Allocates a filetool using the vhash data loaded from localsettings.
 */
void _alloc_filetool(
	SG_context*           pCtx,      //< [in] [out] Error and context info.
	const char*           szName,    //< [in] Name of the filetool to allocate.
	SG_filetool__invoker* fInvoker,  //< [in] Function to use to invoke the tool.
	const SG_vhash*       pData,     //< [in] Localsettings data to initialize the filetool with.
	                                 //<      If NULL, then a built-in tool is assumed, and Filename/Arguments are left blank.
	SG_filetool**         ppTool     //< [out] The allocated and initialized tool.
	)
{
	SG_filetool* pTool      = NULL;

	SG_NULLARGCHECK(szName);
	SG_NULLARGCHECK(fInvoker);
	SG_NULLARGCHECK(ppTool);

	SG_alloc1(pCtx, pTool);
	pTool->pName       = NULL;
	pTool->fInvoker    = fInvoker;
	pTool->pExecutable = NULL;
	pTool->pArguments  = NULL;
	pTool->pStdIn      = NULL;
	pTool->pStdOut     = NULL;
	pTool->pStdErr     = NULL;

	SG_STRING__ALLOC__SZ(pCtx, &(pTool->pName), szName);

	if (pData != NULL)
	{
		const char*  szString = NULL;
		SG_varray*   pArray   = NULL;
		SG_bool      bBool    = SG_FALSE;

		// read the path to the tool
		SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pData, SG_FILETOOLCONFIG__TOOLPATH, &szString)  );
		SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &(pTool->pExecutable), szString)  );

		// read the array of arguments
		SG_ERR_CHECK(  SG_vhash__get__varray(pCtx, pData, SG_FILETOOLCONFIG__TOOLARGS, &pArray)  );
		SG_ERR_CHECK(  SG_VARRAY__ALLOC__COPY(pCtx, &(pTool->pArguments), pArray)  );

		// read the STDIN string
		SG_ERR_CHECK(  SG_vhash__has(pCtx, pData, SG_FILETOOLCONFIG__TOOLSTDIN, &bBool)  );
		if (bBool == SG_TRUE)
		{
			SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pData, SG_FILETOOLCONFIG__TOOLSTDIN, &szString)  );
			SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &(pTool->pStdIn), szString)  );
		}

		// read the STDOUT string
		SG_ERR_CHECK(  SG_vhash__has(pCtx, pData, SG_FILETOOLCONFIG__TOOLSTDOUT, &bBool)  );
		if (bBool == SG_TRUE)
		{
			SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pData, SG_FILETOOLCONFIG__TOOLSTDOUT, &szString)  );
			SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &(pTool->pStdOut), szString)  );
		}

		// read the STDERR string
		SG_ERR_CHECK(  SG_vhash__has(pCtx, pData, SG_FILETOOLCONFIG__TOOLSTDERR, &bBool)  );
		if (bBool == SG_TRUE)
		{
			SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pData, SG_FILETOOLCONFIG__TOOLSTDERR, &szString)  );
			SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &(pTool->pStdErr), szString)  );
		}
	}

	*ppTool = pTool;
	pTool = NULL;

fail:
	SG_FILETOOL_NULLFREE(pCtx, pTool);
	return;
}

/**
 * A default plugin for built-in tools.
 */
void _plugin__default(
	SG_context*            pCtx,
	const char*            szType,
	const char*            szContext,
	const char*            szName,
	SG_filetool__invoker** ppInvoker
	)
{
	SG_UNUSED(pCtx);
	SG_UNUSED(szType);
	SG_UNUSED(szContext);

	// check what the caller wants
	if (szName == NULL)
	{
		// return our external tool invoker
		*ppInvoker = SG_filetool__invoke__external;
	}
	else
	{
		SG_uint32 uIndex = 0u;

		// run through the list of built-in files looking for the named internal tool
		for (uIndex = 0u; uIndex < SG_NrElements(gpBuiltinFiletools); ++uIndex)
		{
			// check if this is the named tool
			if (strcmp(gpBuiltinFiletools[uIndex].szName, szName) == 0)
			{
				// this is it, return the invoker
				*ppInvoker = gpBuiltinFiletools[uIndex].fInvoker;
				break;
			}
		}
	}
}

/**
 * Replaces all of the token values in a string
 * using the values given in a token/value hash.
 */
void _replace_token_values(
	SG_context*     pCtx,          //< [in] [out] Error and context info.
	const SG_vhash* pTokenValues,  //< [in] Values for placeholder tokens in the input string.
	                               //<      May be NULL if there are no values to substitute.
	const char*     szInputString, //< [in] String to replace the tokens in.
	SG_string**     ppOutputString //< [out] Input string with tokens replaced.  Caller owns this.
	)
{
	SG_uint32  uTokenCount = 0u;
	SG_uint32  uTokenIndex = 0u;
	SG_string* pResult     = NULL;

	// allocate a result string, starting with the given input
	SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &pResult, szInputString)  );

	// run through each value in the hash of token values
	SG_ERR_CHECK(  SG_vhash__count(pCtx, pTokenValues, &uTokenCount)  );
	for (uTokenIndex = 0u; uTokenIndex < uTokenCount; ++uTokenIndex)
	{
		const char*       szName  = NULL;
		const SG_variant* pValue  = NULL;
		const char*       szValue = NULL;

		// get the current name/value pair
		SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pTokenValues, uTokenIndex, &szName, &pValue)  );
		SG_ERR_CHECK(  SG_variant__get__sz(pCtx, pValue, &szValue)  );

		// do the replacement in the result string
		SG_ERR_CHECK(  SG_string__replace_bytes(pCtx, pResult, (SG_byte*)szName, SG_STRLEN(szName), (SG_byte*)szValue, SG_STRLEN(szValue), SG_UINT32_MAX, SG_TRUE, NULL)  );

		// HACK: also replace "\-" with "-"
		// this is currently necessary because without being able to escape a '-'
		// it isn't possible to set arguments from the command line using 'vv config'
		//
		// for example:
		//   vv config add-to filetools/type/tool/args --something=@VALUE@
		//
		// When vv parses that command line, it thinks that "--something"
		// is a flag for vv, rather than a plain value.
		// So to make it work, we escape it like so:
		//   vv config add-to filetools/type/tool/args \--something=@VALUE@
		// So what we're doing here is removing that escape character.
		//
		// A better fix in the future might be to update the command line parser
		// so that flags in quotes are treated as plain values.
		// Then you could do this instead:
		//   vv config add-to filetools/type/tool/args "--something=@VALUE@"
		// And we wouldn't have a hacky escape character to remove here.
		//
		// TODO 2011/06/23 Figure out a way to let @TO_WRITABLE?@ (or other
		// TODO            boolean values) to do substitutions.  For example,
		// TODO            if it has value 'T' substitute one value and if it
		// TODO            has value 'F' substitute another value.  The main
		// TODO            question is how does the user define a external
		// TODO            with a conditional parameter.

		SG_ERR_CHECK(  SG_string__replace_bytes(pCtx, pResult, (SG_byte*)"\\-", 2u, (SG_byte*)"-", 1u, SG_UINT32_MAX, SG_TRUE, NULL)  );
	}

	// return the result
	*ppOutputString = pResult;
	pResult = NULL;

fail:
	SG_STRING_NULLFREE(pCtx, pResult);
	return;
}

/**
 * Replaces tokens in a tool's argument list with provided values.
 * Generates an SG_exec_argvec from the resulting arguments.
 *
 * The replacement values are provided in the form of an SG_vhash.
 * The keys in the vhash are the placeholder tokens to be replaced.
 * The values in the vhash are what the corresponding token will be replaced with.
 */
void _prepare_arguments(
	SG_context*        pCtx,         //< [in] [out] Error and context info.
	const SG_filetool* pTool,        //< [in] The tool whose arguments to replace tokens in.
	const SG_vhash*    pTokenValues, //< [in] Values for placeholder tokens in the tool's arguments.
	                                 //<      May be NULL if there are no values to substitute.
	SG_exec_argvec**   ppArgVec      //< [out] An argument vector with the resulting command line arguments.
	                                 //<       Caller owns this.
	)
{
	SG_uint32       uArgCount = 0u;
	SG_uint32       uArgIndex = 0u;
	SG_exec_argvec* pArgVec   = NULL;
	SG_string*      pReplaced = NULL;

	SG_NULLARGCHECK(pTool);
	SG_NULLARGCHECK(ppArgVec);

	// allocate an empty argvec
	SG_ERR_CHECK(  SG_exec_argvec__alloc(pCtx, &pArgVec)  );

	// run through each argument in the tool
	SG_ERR_CHECK(  SG_varray__count(pCtx, pTool->pArguments, &uArgCount)  );
	for (uArgIndex = 0u; uArgIndex < uArgCount; ++uArgIndex)
	{
		const char* szArgument = NULL;

		// get the current argument
		SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pTool->pArguments, uArgIndex, &szArgument)  );

		// run the replacement
		SG_ERR_CHECK(  _replace_token_values(pCtx, pTokenValues, szArgument, &pReplaced)  );

		// append the replaced argument to the result argvec
		SG_ERR_CHECK(  SG_exec_argvec__append__sz(pCtx, pArgVec, SG_string__sz(pReplaced))  );

		// free the replaced string, since we'll get a new one next time
		SG_STRING_NULLFREE(pCtx, pReplaced);
	}

	// return the resulting argvec
	*ppArgVec = pArgVec;
	pArgVec = NULL;

fail:
	SG_STRING_NULLFREE(pCtx, pReplaced);
	SG_EXEC_ARGVEC_NULLFREE(pCtx, pArgVec);
	return;
}

/**
 * Prepares/opens a file to be bound to an input/output stream of an external tool.
 */
void _prepare_stream_file(
	SG_context*      pCtx,            //< [in] [out] Error and context info.
	const SG_string* pStreamArgument, //< [in] The stream argument for the stream file to bind.
	                                  //<      One of: pTool->StdIn, pTool->StdOut, or pTool->StdErr
	const SG_vhash*  pTokenValues,    //< [in] Values for placeholder tokens in the tool's stream files.
	                                  //<      May be NULL if there are no values to substitute.
	SG_file_flags    uFlags,          //< [in] Flags to use when opening the file.
	SG_file**        ppFile           //< [out] The file to bind to the tool's stream, or NULL to not bind one.
	)
{
	SG_string*   pReplaced = NULL;
	SG_pathname* pPathname = NULL;
	SG_file*     pFile     = NULL;

	SG_NULLARGCHECK(pTokenValues);
	SG_NULLARGCHECK(ppFile);

	// if the argument is NULL, then we don't need to bind a stream here
	if (pStreamArgument == NULL)
	{
		*ppFile = NULL;
		return;
	}

	// replace token values in the argument
	SG_ERR_CHECK(  _replace_token_values(pCtx, pTokenValues, SG_string__sz(pStreamArgument), &pReplaced)  );

	// allocate a pathname from the replaced value
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPathname, SG_string__sz(pReplaced))  );

	// open the file
	SG_ERR_CHECK(  SG_file__open__pathname(pCtx, pPathname, uFlags, 0600, &pFile)  );

	// return the opened file
	*ppFile = pFile;
	pFile = NULL;

fail:
	SG_STRING_NULLFREE(pCtx, pReplaced);
	SG_PATHNAME_NULLFREE(pCtx, pPathname);
	SG_FILE_NULLCLOSE(pCtx, pFile);
	return;
}

/**
 * Detects which default file class a given file should belong to by sniffing its contents.
 */
void _detect_default_class(
	SG_context*        pCtx,      //< [in] [out] Error and context info.
	const SG_pathname* pPathname, //< [in] Pathname of the file to detect the class of.
	const char**       ppClass    //< [out] The detected class of the file.  One of SG_FILETOOL__CLASS__*.
	)
{
	SG_byte     aBuffer[1024] = { 0, };
	SG_file*    pFile         = NULL;
	SG_uint32   uBytesRead    = 0u;
	SG_uint32   uIndex        = 0u;
	const char* szClass       = NULL;

	// open the file
	SG_ERR_CHECK(  SG_file__open__pathname(pCtx, pPathname, SG_FILE_OPEN_EXISTING | SG_FILE_RDONLY, 0u, &pFile)  );

	// fill our buffer by reading from the file
	SG_file__read(pCtx, pFile, sizeof(aBuffer), aBuffer, &uBytesRead);

	if (SG_context__err_equals(pCtx, SG_ERR_EOF))
    {
		SG_context__err_reset(pCtx);
    }

	SG_ERR_CHECK_CURRENT;
	// first we'll check the initial bytes against our known set of "magic" numbers
	for (uIndex = 0u; uIndex < SG_NrElements(gpFileClassMagicNumbers); ++uIndex)
	{
		SG_uint32 uByte  = 0u;
		SG_bool   bMatch = SG_TRUE;

		// if the magic number is longer than the file, it definitely can't match
		if (gpFileClassMagicNumbers[uIndex].uSize > uBytesRead)
		{
			continue;
		}

		// run through and compare the magic number against the file's first bytes
		for (uByte = 0u; uByte < gpFileClassMagicNumbers[uIndex].uSize; ++uByte)
		{
			if (gpFileClassMagicNumbers[uIndex].pBytes[uByte] != aBuffer[uByte])
			{
				bMatch = SG_FALSE;
				break;
			}
		}

		// if we found a match, use the corresponding file class
		if (bMatch == SG_TRUE)
		{
			szClass = gpFileClassMagicNumbers[uIndex].szClass;
		}
	}

	// TODO: consider checking for #! line at the top of scripts
	//       maybe just add "#!" to the magic number list
	//       or maybe include logic to also look for something resembling a path

	// if we haven't figured it out yet, we'll scan the whole buffer for NUL characters
	// if we find one we'll assume binary, since NUL is generally invalid text
	if (szClass == NULL)
	{
		for (uIndex = 0u; uIndex < uBytesRead; ++uIndex)
		{
			if (aBuffer[uIndex] == '\0')
			{
				szClass = SG_FILETOOL__CLASS__BINARY;
				break;
			}
		}
	}

	// if we haven't figured it out yet, we'll just assume text
	if (szClass == NULL)
	{
		szClass = SG_FILETOOL__CLASS__TEXT;
	}

	// output whatever we found
	*ppClass = szClass;
	szClass = NULL;

fail:
	SG_FILE_NULLCLOSE(pCtx, pFile);
	return;
}


/*
**
** Public Functions
**
*/

void SG_filetool__get_class(
	SG_context*        pCtx,
	const char*        szPath,
	const SG_pathname* pDiskPath,
	SG_repo*           pRepo,
	SG_string**        ppClass
	)
{
	SG_vhash*  pClasses    = NULL;
	SG_uint32  uClassCount = 0u;
	SG_uint32  uClassIndex = 0u;
	SG_string* pClass      = NULL;
	SG_bool    bFound      = SG_FALSE;
	SG_bool	   bAllocedClassesArray = SG_FALSE;
	SG_varray* pPatterns     = NULL;

	SG_NULLARGCHECK(pDiskPath);
	SG_NULLARGCHECK(ppClass);

	// if they didn't supply a path, use the disk path
	if (szPath == NULL)
	{
		szPath = SG_pathname__sz(pDiskPath);
	}

	// allocate our output string
	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pClass)  );

	if (bFound != SG_TRUE)
	{
		// try the file classes defined in localsettings
		SG_ERR_CHECK(  SG_localsettings__get__collapsed_vhash(pCtx, SG_FILETOOLCONFIG__CLASSES, pRepo, &pClasses)  );

		// run through each class
		// note: we're relying on vhash iterating in the order that items were added
		//       because we want types in higher priority scopes to override types in lower priority ones
		SG_ERR_CHECK(  SG_vhash__count(pCtx, pClasses, &uClassCount)  );
		for (uClassIndex = 0u; uClassIndex < uClassCount; ++uClassIndex)
		{
			const char*       szClassName   = NULL;
			const SG_variant* pClassValue   = NULL;
			SG_vhash*         pClassData    = NULL;
			SG_uint32         uPatternCount = 0u;
			SG_uint32         uPatternIndex = 0u;
			SG_bool           bHasPatterns  = SG_FALSE;

			// get the data for this class
			SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pClasses, uClassIndex, &szClassName, &pClassValue)  );
			SG_ERR_CHECK(  SG_variant__get__vhash(pCtx, pClassValue, &pClassData)  );

			// get the list of patterns
			SG_ERR_CHECK(  SG_vhash__has(pCtx, pClassData, SG_FILETOOLCONFIG__CLASSPATTERNS, &bHasPatterns)  );
			if (bHasPatterns == SG_FALSE)
			{
				continue;
			}
		
			SG_vhash__get__varray(pCtx, pClassData, SG_FILETOOLCONFIG__CLASSPATTERNS, &pPatterns);
			bAllocedClassesArray = SG_FALSE;
			if (SG_context__err_equals(pCtx, SG_ERR_VARIANT_INVALIDTYPE) )
			{
				const char * pszValue = NULL;
				SG_context__err_reset(pCtx);
				//Try to load a single string.
				SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pClassData, SG_FILETOOLCONFIG__CLASSPATTERNS, &pszValue)  );
				bAllocedClassesArray = SG_TRUE;
				SG_ERR_CHECK(  SG_VARRAY__ALLOC(pCtx, &pPatterns)  );
				SG_ERR_CHECK(  SG_varray__append__string__sz(pCtx, pPatterns, pszValue)  );
			}
			else
			{
				SG_ERR_CHECK_CURRENT;
			}

			// run through each pattern
			SG_ERR_CHECK(  SG_varray__count(pCtx, pPatterns, &uPatternCount)  );
			for (uPatternIndex = 0u; uPatternIndex < uPatternCount; ++uPatternIndex)
			{
				const char* szPattern = NULL;
				SG_bool     bMatches  = SG_FALSE;

				// get the current pattern
				SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pPatterns, uPatternIndex, &szPattern)  );

				// match this pattern against the filename
				SG_ERR_CHECK(  SG_file_spec__match_pattern(pCtx, szPattern, szPath, SG_FILE_SPEC__MATCH_ANYWHERE, &bMatches)  );

				// if it matched, use the current class name
				if (bMatches == SG_TRUE)
				{
					SG_ERR_CHECK(  SG_string__set__sz(pCtx, pClass, szClassName)  );
					bFound = SG_TRUE;
					break;
				}
			}
			if (bAllocedClassesArray)
				SG_VARRAY_NULLFREE(pCtx, pPatterns);
			
			// if this class matched, stop looking at others
			if (bFound == SG_TRUE)
			{
				break;
			}
		}
	}

	if (bFound != SG_TRUE)
	{
		// try to detect a class
		const char* szDefaultClass = NULL;

		SG_ERR_CHECK(  _detect_default_class(pCtx, pDiskPath, &szDefaultClass)  );
		SG_ERR_CHECK(  SG_string__set__sz(pCtx, pClass, szDefaultClass)  );
		bFound = SG_TRUE;
	}

	if (bFound != SG_TRUE)
	{
		// just default to text
		SG_ERR_CHECK(  SG_string__set__sz(pCtx, pClass, SG_FILETOOL__CLASS__TEXT)  );
		bFound = SG_TRUE;
	}

	// return whatever class we ended up with
	*ppClass = pClass;
	pClass = NULL;

fail:
	if (bAllocedClassesArray)
		SG_VARRAY_NULLFREE(pCtx, pPatterns);
	SG_VHASH_NULLFREE(pCtx, pClasses);
	SG_STRING_NULLFREE(pCtx, pClass);
	return;
}

void SG_filetool__find_tool(
	SG_context* pCtx,
	const char* szType,
	const char* szContext,
	const char* szClass,
	SG_repo*    pRepo,
	SG_string** ppName
	)
{
	SG_string*  pSettingName = NULL;
	SG_variant* pData        = NULL;
	const char* szName       = NULL;

	SG_NULLARGCHECK(szClass);
	SG_NULLARGCHECK(ppName);

	// allocate a string to build setting names in
	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pSettingName)  );

	// check for a tool defined for this type and file class
	if (pData == NULL)
	{
		SG_ERR_CHECK(  SG_string__sprintf(pCtx, pSettingName, "%s/%s/%s", SG_FILETOOLCONFIG__BINDINGS, szType, szClass)  );
		SG_ERR_CHECK(  SG_localsettings__get__variant(pCtx, SG_string__sz(pSettingName), pRepo, &pData, NULL)  );
	}

	// if we found some tool data, parse through it
	if (pData != NULL)
	{
		// check if we're interested in a context-specific tool
		if (szContext != NULL && pData->type == SG_VARIANT_TYPE_VHASH)
		{
			SG_vhash* pContexts = NULL;
			SG_bool   bExists   = SG_FALSE;

			// if there's a tool defined for this context in this class, retrieve its name
			SG_ERR_CHECK(  SG_variant__get__vhash(pCtx, pData, &pContexts)  );
			SG_ERR_CHECK(  SG_vhash__has(pCtx, pContexts, szContext, &bExists)  );
			if (bExists == SG_TRUE)
			{
				SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pContexts, szContext, &szName)  );
			}
		}
		else
		{
			// either we don't care about context, or this class doesn't have context-specific tools
			// just look up the cross-context name
			SG_ERR_CHECK(  SG_variant__get__sz(pCtx, pData, &szName)  );
		}
	}

	// if we found a name, return it
	if (szName != NULL)
	{
		SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, ppName, szName)  );
	}
	else
	{
		*ppName = NULL;
	}

fail:
	SG_STRING_NULLFREE(pCtx, pSettingName);
	SG_VARIANT_NULLFREE(pCtx, pData);
	return;
}

void SG_filetool__get_tool(
	SG_context*         pCtx,
	const char*         szType,
	const char*         szContext,
	const char*         szName,
	SG_filetool__plugin fPlugin,
	SG_repo*            pRepo,
	SG_filetool**       ppTool
	)
{
	SG_string*   pSettingName = NULL;
	SG_vhash*    pData        = NULL;
	SG_filetool* pTool        = NULL;

	SG_NULLARGCHECK(szName);
	SG_NULLARGCHECK(ppTool);

	// look for an external tool in the localsettings
	if (pTool == NULL)
	{
		// build the setting name
		SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pSettingName)  );
		SG_ERR_CHECK(  SG_string__sprintf(pCtx, pSettingName, "%s/%s/%s", SG_FILETOOLCONFIG__TOOLS, szType, szName)  );

		// look up the setting
		SG_ERR_CHECK(  SG_localsettings__get__vhash(pCtx, SG_string__sz(pSettingName), pRepo, &pData, NULL)  );
		if (pData != NULL)
		{
			SG_filetool__invoker* fInvoker = NULL;

			// check if the caller has a custom external tool invoker
			if (fInvoker == NULL && fPlugin != NULL)
			{
				SG_ERR_CHECK(  fPlugin(pCtx, szType, szContext, NULL, &fInvoker)  );
			}

			// check for our own external tool invoker
			if (fInvoker == NULL)
			{
				SG_ERR_CHECK(  _plugin__default(pCtx, szType, szContext, NULL, &fInvoker)  );
			}

			// allocate a tool with the invoker we found
			if (fInvoker != NULL)
			{
				SG_ERR_CHECK(  _alloc_filetool(pCtx, szName, fInvoker, pData, &pTool)  );
			}
		}
	}

	// look for an internal tool
	if (pTool == NULL)
	{
		SG_filetool__invoker* fInvoker = NULL;

		// check if the caller is providing a matching internal tool
		if (fInvoker == NULL && fPlugin != NULL)
		{
			SG_ERR_CHECK(  fPlugin(pCtx, szType, szContext, szName, &fInvoker)  );
		}

		// check for our own matching internal tool
		if (fInvoker == NULL)
		{
			SG_ERR_CHECK(  _plugin__default(pCtx, szType, szContext, szName, &fInvoker)  );
		}

		// allocate a tool with the invoker we found
		if (fInvoker != NULL)
		{
			SG_ERR_CHECK(  _alloc_filetool(pCtx, szName, fInvoker, NULL, &pTool)  );
		}
	}

	// output whatever we found
	*ppTool = pTool;
	pTool = NULL;

fail:
	SG_STRING_NULLFREE(pCtx, pSettingName);
	SG_VHASH_NULLFREE(pCtx, pData);
	SG_FILETOOL_NULLFREE(pCtx, pTool);
	return;
}

void SG_filetool__free(
	SG_context*  pCtx,
	SG_filetool* pTool
	)
{
	if (pTool == NULL)
	{
		return;
	}

	SG_STRING_NULLFREE(pCtx, pTool->pName);
	SG_STRING_NULLFREE(pCtx, pTool->pExecutable);
	SG_VARRAY_NULLFREE(pCtx, pTool->pArguments);
	SG_STRING_NULLFREE(pCtx, pTool->pStdIn);
	SG_STRING_NULLFREE(pCtx, pTool->pStdOut);
	SG_STRING_NULLFREE(pCtx, pTool->pStdErr);

	SG_ERR_CHECK(  SG_free(pCtx, pTool)  );

fail:
	return;
}

void SG_filetool__get_name(
	SG_context*        pCtx,
	const SG_filetool* pTool,
	const char**       ppName
	)
{
	SG_NULLARGCHECK(pTool);
	SG_NULLARGCHECK(ppName);

	*ppName = SG_string__sz(pTool->pName);

fail:
	return;
}

void SG_filetool__get_executable(
	SG_context*        pCtx,
	const SG_filetool* pTool,
	const char**       ppExecutable
	)
{
	SG_NULLARGCHECK(pTool);
	SG_NULLARGCHECK(ppExecutable);

	if (pTool->pExecutable)
		*ppExecutable = SG_string__sz(pTool->pExecutable);
	else
		*ppExecutable = NULL;

fail:
	return;
}

void SG_filetool__invoke(
	SG_context*        pCtx,
	const SG_filetool* pTool,
	const SG_vhash*    pTokenValues,
	SG_int32*          pExitCode
	)
{
	SG_NULLARGCHECK(pTool);

	SG_ERR_CHECK(  (pTool->fInvoker)(pCtx, pTool, pTokenValues, pExitCode)  );

fail:
	return;
}

void SG_filetool__invoke__null(
	SG_context*        pCtx,
	const SG_filetool* pTool,
	const SG_vhash*    pTokenValues,
	SG_int32*          pExitCode
	)
{
	SG_UNUSED(pCtx);
	SG_UNUSED(pTool);
	SG_UNUSED(pTokenValues);

	if (pExitCode != NULL)
	{
		*pExitCode = 0;
	}
}

void SG_filetool__invoke__external(
	SG_context*        pCtx,
	const SG_filetool* pTool,
	const SG_vhash*    pTokenValues,
	SG_int32*          pExitCode
	)
{
	SG_exec_argvec* pArgVec     = NULL;
	SG_file*        pStdIn      = NULL;
	SG_file*        pStdOut     = NULL;
	SG_file*        pStdErr     = NULL;
	SG_exit_status  iExitStatus = 0;

	SG_NULLARGCHECK(pTool);

	// substitute our argument values into the tool's arguments
	SG_ERR_CHECK(  _prepare_arguments(pCtx, pTool, pTokenValues, &pArgVec)  );

	// open any files that we need to bind to streams
	SG_ERR_CHECK(  _prepare_stream_file(pCtx, pTool->pStdIn,  pTokenValues, SG_FILE_RDONLY | SG_FILE_OPEN_EXISTING,                  &pStdIn)  );
	SG_ERR_CHECK(  _prepare_stream_file(pCtx, pTool->pStdOut, pTokenValues, SG_FILE_WRONLY | SG_FILE_OPEN_OR_CREATE | SG_FILE_TRUNC, &pStdOut)  );
	SG_ERR_CHECK(  _prepare_stream_file(pCtx, pTool->pStdErr, pTokenValues, SG_FILE_WRONLY | SG_FILE_OPEN_OR_CREATE | SG_FILE_TRUNC, &pStdErr)  );

	// execute the tool
	SG_ERR_CHECK(  SG_exec__exec_sync__files(pCtx, SG_string__sz(pTool->pExecutable), pArgVec, pStdIn, pStdOut, pStdErr, &iExitStatus)  );

	if (pExitCode != NULL)
	{
		*pExitCode = iExitStatus;
	}

fail:
	SG_FILE_NULLCLOSE(pCtx, pStdIn);
	SG_FILE_NULLCLOSE(pCtx, pStdOut);
	SG_FILE_NULLCLOSE(pCtx, pStdErr);
	SG_EXEC_ARGVEC_NULLFREE(pCtx, pArgVec);
	return;
}

//////////////////////////////////////////////////////////////////

#if defined(WINDOWS)
static void _try__look_for_diffmerge_exe_path_in_registry(SG_context * pCtx, const char * pszTry, SG_pathname ** ppPathExe)
{
	HKEY hKey = NULL;
	char * pszBuf = NULL;
	SG_pathname * pPathExe = NULL;
	DWORD dwType;
	DWORD dwLenBuf;
	SG_bool bExists;
	SG_fsobj_type fsobjType;

	if (RegOpenKeyExA(HKEY_LOCAL_MACHINE, pszTry, 0, KEY_QUERY_VALUE, &hKey) != ERROR_SUCCESS)
		SG_ERR_THROW(  SG_ERR_NOT_FOUND  );

	if (RegQueryValueExA(hKey, "Location", NULL, &dwType, NULL, &dwLenBuf) != ERROR_SUCCESS)
		SG_ERR_THROW(  SG_ERR_NOT_FOUND  );

	if (dwType != REG_SZ)
		SG_ERR_THROW(  SG_ERR_NOT_FOUND  );

	SG_ERR_CHECK(  SG_alloc(pCtx, dwLenBuf+1, sizeof(char), &pszBuf)  );
#if defined(DEBUG)
	_sg_mem__set_caller_data(pszBuf,__FILE__,__LINE__,"RegistryBuffer");
#endif

	if (RegQueryValueExA(hKey, "Location", NULL, &dwType, (LPBYTE)pszBuf, &dwLenBuf) != ERROR_SUCCESS)
		SG_ERR_THROW(  SG_ERR_NOT_FOUND  );

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPathExe, pszBuf)  );
	
	SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pPathExe, &bExists, &fsobjType, NULL)  );
	if (!bExists || (fsobjType != SG_FSOBJ_TYPE__REGULAR))
		SG_ERR_THROW(  SG_ERR_NOT_FOUND  );

#if TRACE_DIFFMERGE_LOOKUP
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "Found DiffMerge path in registry key [HKLM] \\ %s \\ Location = %s\n",
							   pszTry, SG_pathname__sz(pPathExe))  );
#endif

	*ppPathExe = pPathExe;
	pPathExe = NULL;

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathExe);
	SG_NULLFREE(pCtx, pszBuf);
	if (hKey)
		RegCloseKey(hKey);
}

static void _look_for_diffmerge_exe_path_in_registry(SG_context * pCtx, SG_pathname ** ppPathExe)
{
	static const char * aszKeys[] = {
		"SOFTWARE\\SourceGear\\Common\\DiffMerge\\Installer",					// location for MSM-based 3.3.1+
		"SOFTWARE\\SourceGear\\SourceGear DiffMerge\\Installer",				// same-bitness location for 3.3.0 and earlier MSI
		"SOFTWARE\\Wow6432Node\\SourceGear\\SourceGear DiffMerge\\Installer",	// 32-bit 3.3.0 and earlier MSI on 64-bit system 
	};

	SG_uint32 k;

	for (k=0; k<SG_NrElements(aszKeys); k++)
	{
		_try__look_for_diffmerge_exe_path_in_registry(pCtx, aszKeys[k], ppPathExe);
		if (!SG_context__has_err(pCtx))
			return;

		if (!SG_context__err_equals(pCtx, SG_ERR_NOT_FOUND))
			SG_ERR_RETHROW_RETURN;
		SG_context__err_reset(pCtx);
	}

	SG_ERR_THROW_RETURN(  SG_ERR_NOT_FOUND  );
}

static void _try__look_for_diffmerge_exe_path_in_program_files(SG_context * pCtx, const char * pszTry, SG_pathname ** ppPathExe)
{
	SG_pathname * pPathExe = NULL;
	SG_bool bExists;
	SG_fsobj_type fsobjType;

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPathExe, pszTry)  );

	SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pPathExe, &bExists, &fsobjType, NULL)  );
	if (!bExists || (fsobjType != SG_FSOBJ_TYPE__REGULAR))
		SG_ERR_THROW(  SG_ERR_NOT_FOUND  );

#if TRACE_DIFFMERGE_LOOKUP
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "Found DiffMerge path in program files: %s\n",
							   SG_pathname__sz(pPathExe))  );
#endif

	*ppPathExe = pPathExe;
	pPathExe = NULL;

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathExe);
}

static void _look_for_diffmerge_exe_path_in_program_files(SG_context * pCtx, SG_pathname ** ppPathExe)
{
	// 2010/09/30 This is a complete hack.  There, I said it,
	//            but I'm going to do it anyway.
	// 
	//            We want to find SourceGear DiffMerge.  Normally,
	//            I'd just use the entryname and tell the user to
	//            make sure it can be found somewhere in their PATH,
	//            but I just noticed that VS2010 includes a program
	//            called .../Common7/IDE/diffmerge.exe that might
	//            get in the way.  So for now, I'm going to do this.
	// 
	//            This also lets users who have Vault or Fortress
	//            installed (which includes a bundled version of
	//            DiffMerge) to magically work without having to
	//            install the stand-alone version.
	// 
	//            I'm going to list all of the standard install
	//            directories for each product for both 32- and
	//            64-bit machines.

	static const char * aszGuesses[] = {
		"C:\\Program Files\\SourceGear\\Common\\DiffMerge\\sgdm.exe",			// 3.3.1+ MSM-based common
		"C:\\Program Files (x86)\\SourceGear\\Common\\DiffMerge\\sgdm.exe",

		"C:\\Program Files\\SourceGear\\DiffMerge\\DiffMerge.exe",				// 3.3.0 and earlier stand-alone
		"C:\\Program Files (x86)\\SourceGear\\DiffMerge\\DiffMerge.exe",

		"C:\\Program Files\\SourceGear\\Fortress Client\\sgdm.exe",				// bundled versions with other products
		"C:\\Program Files (x86)\\SourceGear\\Fortress Client\\sgdm.exe",
		"C:\\Program Files\\SourceGear\\Vault Client\\sgdm.exe",
		"C:\\Program Files (x86)\\SourceGear\\Vault Client\\sgdm.exe",
		"C:\\Program Files\\SourceGear\\VaultPro Client\\sgdm.exe",
		"C:\\Program Files (x86)\\SourceGear\\VaultPro Client\\sgdm.exe",
	};

	SG_uint32 k;

	for (k=0; k<SG_NrElements(aszGuesses); k++)
	{
		_try__look_for_diffmerge_exe_path_in_program_files(pCtx, aszGuesses[k], ppPathExe);
		if (!SG_context__has_err(pCtx))
			return;

		if (!SG_context__err_equals(pCtx, SG_ERR_NOT_FOUND))
			SG_ERR_RETHROW_RETURN;
		SG_context__err_reset(pCtx);
	}

	SG_ERR_THROW_RETURN(  SG_ERR_NOT_FOUND  );
}
		
static void _try__look_for_diffmerge_exe_path_using_PATH(SG_context * pCtx, const wchar_t * pwszExeName, SG_pathname ** ppPathExe)
{
	SG_uint32 lenAllocated, lenRequired;
	SG_string * pString = NULL;
	wchar_t * pWideBuf = NULL;
	SG_pathname * pPathExe = NULL;

	// The registry lookup failed and we didn't find it in any of the known
	// install directories, so it probably isn't installed.  But before we
	// give up, see if they have a copy in their PATH.  This takes care of
	// the case where they used the DiffMerge ZIP distribution rather than
	// one of the installer packages.
	//
	// I hesitate to call SearchPath() and look
	// for it, but if we just return the entryname and let CreateProcess()
	// try to find it and it gets the MSFT version, the use will see an
	// odd syntax error dialog (because we take different arguments).

	lenAllocated = 4096;
	pWideBuf = (wchar_t *)SG_calloc(lenAllocated,sizeof(wchar_t));
	if (!pWideBuf)
		SG_ERR_THROW( SG_ERR_MALLOCFAILED );

top:
	lenRequired = SearchPath(NULL, pwszExeName, NULL, lenAllocated, pWideBuf, NULL);
	if (lenRequired == 0)				// We did not find any instances anywhere in our PATH.
		SG_ERR_THROW(  SG_ERR_NOT_FOUND  );

	if (lenRequired > lenAllocated)		// pathname too long for our buffer, try again.
	{
		SG_NULLFREE(pCtx, pWideBuf);
		lenAllocated += 4096;
		pWideBuf = (wchar_t *)SG_calloc(lenAllocated,sizeof(wchar_t));
		if (!pWideBuf)
			SG_ERR_THROW( SG_ERR_MALLOCFAILED );
		goto top;
	}

	// We found it somewhere in the PATH, who knows what it is.

	if (wcsstr(pWideBuf, L"Visual Studio"))
		SG_ERR_THROW(  SG_ERR_NOT_FOUND  );

	// If not part of Visual Studio, assume it is ours.
	// We need to convert the wchar pathname into a utf8 psz
	// before we can put it in a SG_pathname.

	SG_ERR_CHECK(  SG_string__alloc(pCtx, &pString)  );
	SG_ERR_CHECK(  SG_utf8__intern_from_os_buffer__wchar(pCtx, pString, pWideBuf)  );
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPathExe, SG_string__sz(pString))  );

#if TRACE_DIFFMERGE_LOOKUP
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "Found DiffMerge path in PATH: %s\n",
							   SG_pathname__sz(pPathExe))  );
#endif

	*ppPathExe = pPathExe;
	pPathExe = NULL;

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathExe);
	SG_STRING_NULLFREE(pCtx, pString);
	SG_NULLFREE(pCtx, pWideBuf);
}

static void _look_for_diffmerge_exe_path_using_PATH(SG_context * pCtx, SG_pathname ** ppPathExe)
{
	static const wchar_t * aszGuesses[] = {
		L"sgdm.exe",
		L"DiffMerge.exe",
	};
	SG_uint32 k;

	for (k=0; k<SG_NrElements(aszGuesses); k++)
	{
		_try__look_for_diffmerge_exe_path_using_PATH(pCtx, aszGuesses[k], ppPathExe);
		if (!SG_context__has_err(pCtx))
			return;

		if (!SG_context__err_equals(pCtx, SG_ERR_NOT_FOUND))
			SG_ERR_RETHROW_RETURN;
		SG_context__err_reset(pCtx);
	}

	SG_ERR_THROW_RETURN(  SG_ERR_NOT_FOUND  );
}

void SG_filetool__get_builtin_diffmerge_exe(
	SG_context * pCtx,          //< [in] [out] Error and context info.
	SG_pathname ** ppPathCmd    //< [out] Path to diffmerge executable. Caller owns it.
	)
{
	// The official DiffMerge MSI/MSM installers recorded the install path in the registry.

	_look_for_diffmerge_exe_path_in_registry(pCtx, ppPathCmd);
	if (!SG_context__has_err(pCtx))
		return;

	if (!SG_context__err_equals(pCtx, SG_ERR_NOT_FOUND))
		SG_ERR_RETHROW_RETURN;
	SG_context__err_reset(pCtx);

	// Search for the EXE in all of the various other products it might be bundled with.

	_look_for_diffmerge_exe_path_in_program_files(pCtx, ppPathCmd);
	if (!SG_context__has_err(pCtx))
		return;

	if (!SG_context__err_equals(pCtx, SG_ERR_NOT_FOUND))
		SG_ERR_RETHROW_RETURN;
	SG_context__err_reset(pCtx);

	// See if it is in their PATH (while avoiding the MSFT program with the same name).

	_look_for_diffmerge_exe_path_using_PATH(pCtx, ppPathCmd);
	if (!SG_context__has_err(pCtx))
		return;

	if (!SG_context__err_equals(pCtx, SG_ERR_NOT_FOUND))
		SG_ERR_RETHROW_RETURN;
	SG_context__err_reset(pCtx);

	// anywhere else we should look ?

	SG_ERR_THROW2_RETURN(  SG_ERR_DIFFMERGE_EXE_NOT_FOUND,
						   (pCtx, "Please install DiffMerge.")  );
}
#endif // end if defined(WINDOWS)

#if defined(LINUX) || defined(MAC)
static void _try__look_for_diffmerge_exe_path_in_known_location(SG_context * pCtx, const char * pszTry, SG_pathname ** ppPathExe)
{
	SG_pathname * pPathExe = NULL;
	SG_bool bExists;
	SG_fsobj_type fsobjType;

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPathExe, pszTry)  );

	SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pPathExe, &bExists, &fsobjType, NULL)  );
	if (!bExists || (fsobjType != SG_FSOBJ_TYPE__REGULAR))
		SG_ERR_THROW(  SG_ERR_NOT_FOUND  );

#if TRACE_DIFFMERGE_LOOKUP
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "Found DiffMerge path in: %s\n",
							   SG_pathname__sz(pPathExe))  );
#endif

	*ppPathExe = pPathExe;
	pPathExe = NULL;

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathExe);
}
#endif

#if defined(LINUX)
static void _look_for_diffmerge_exe_path_in_known_location(SG_context * pCtx, SG_pathname ** ppPathExe)
{
	static const char * aszGuesses[] = {
		"/usr/bin/diffmerge",			// official Ubuntu and Fedora installers put it here
		"/usr/local/bin/diffmerge",
	};

	SG_uint32 k;

	for (k=0; k<SG_NrElements(aszGuesses); k++)
	{
		_try__look_for_diffmerge_exe_path_in_known_location(pCtx, aszGuesses[k], ppPathExe);
		if (!SG_context__has_err(pCtx))
			return;

		if (!SG_context__err_equals(pCtx, SG_ERR_NOT_FOUND))
			SG_ERR_RETHROW_RETURN;
		SG_context__err_reset(pCtx);
	}

	SG_ERR_THROW_RETURN(  SG_ERR_NOT_FOUND  );
}
#endif

#if defined(MAC)
static void _look_for_diffmerge_exe_path_in_known_location(SG_context * pCtx, SG_pathname ** ppPathExe)
{
	static const char * aszGuesses[] = {
		"/usr/bin/diffmerge",			// they installed the shell script I put on the .DMG.
		"/usr/local/bin/diffmerge",
		"/Applications/DiffMerge.app/Contents/Resources/diffmerge.sh",	// use the hidden version of the shell script I put in the .APP.
	};

	SG_uint32 k;

	for (k=0; k<SG_NrElements(aszGuesses); k++)
	{
		_try__look_for_diffmerge_exe_path_in_known_location(pCtx, aszGuesses[k], ppPathExe);
		if (!SG_context__has_err(pCtx))
			return;

		if (!SG_context__err_equals(pCtx, SG_ERR_NOT_FOUND))
			SG_ERR_RETHROW_RETURN;
		SG_context__err_reset(pCtx);
	}

	SG_ERR_THROW_RETURN(  SG_ERR_NOT_FOUND  );
}
#endif

#if defined(LINUX) || defined(MAC)
void SG_filetool__get_builtin_diffmerge_exe(
	SG_context * pCtx,          //< [in] [out] Error and context info.
	SG_pathname ** ppPathCmd    //< [out] Path to diffmerge executable. Caller owns it.
	)
{
	_look_for_diffmerge_exe_path_in_known_location(pCtx, ppPathCmd);
	if (!SG_context__has_err(pCtx))
		return;

	if (!SG_context__err_equals(pCtx, SG_ERR_NOT_FOUND))
		SG_ERR_RETHROW_RETURN;
	SG_context__err_reset(pCtx);

	// anywhere else we should look ?

	SG_ERR_THROW2_RETURN(  SG_ERR_DIFFMERGE_EXE_NOT_FOUND,
						   (pCtx, "Please install DiffMerge.")  );
}
#endif
