/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include <sg.h>

///////////////////////////////////////////////////////////////////////////////


static struct
{
	const char * szApplicationRoot;
	SG_bool readonlyServer;
	SG_bool enableDiagnostics;

	// Flag to create a new cx on every acquire() rather than keeping a pool of them.
	SG_bool ssjsMutable;

	SG_pathname * pPathToDispatchDotJS;
	SG_pathname * pPathToCore;
	SG_pathname * pPathToModules;

	// Members above this comment are not controlled by the mutex.
	// All others are.

	SG_mutex lock;

	// Please take care that all members below this comment be
	// protected by the mutex.

	SG_uint32 numContextsCheckedOut; // Number of SG_jscontexts currently in use by SG_uridispatch.
	SG_jscontext * pFirstAvailableContext; // Pointer to the first SG_jscontext NOT in use by SG_uridispatch.
} * gpJSContextPoolGlobalState = NULL;


static void _sg_jscontextpool_getpaths(SG_context *pCtx)
{
	char * szServerFiles = NULL;

	if (gpJSContextPoolGlobalState->pPathToDispatchDotJS)
		return;

	// Figure out and store relavant paths...
	SG_ERR_CHECK(  SG_localsettings__get__sz(pCtx, SG_LOCALSETTING__SERVER_FILES, NULL, &szServerFiles, NULL)  );

	if (szServerFiles)
	{
		SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &gpJSContextPoolGlobalState->pPathToDispatchDotJS, szServerFiles)  );
		SG_ERR_CHECK(  SG_pathname__append__from_sz(pCtx, gpJSContextPoolGlobalState->pPathToDispatchDotJS, "dispatch.js")  );

		SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &gpJSContextPoolGlobalState->pPathToCore, szServerFiles)  );
		SG_ERR_CHECK(  SG_pathname__append__from_sz(pCtx, gpJSContextPoolGlobalState->pPathToCore, "core")  );

		SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &gpJSContextPoolGlobalState->pPathToModules, szServerFiles)  );
		SG_ERR_CHECK(  SG_pathname__append__from_sz(pCtx, gpJSContextPoolGlobalState->pPathToModules, "modules")  );
	}

fail:
	SG_NULLFREE(pCtx, szServerFiles);
}

void SG_jscontextpool__init(SG_context * pCtx, const char * szApplicationRoot, SG_bool readonlyServer, SG_bool enableDiagnostics, SG_bool skipModules)
{
	char * szSsjsMutable = NULL;

	if(gpJSContextPoolGlobalState != NULL)
		return;

	SG_ERR_CHECK(  SG_alloc1(pCtx, gpJSContextPoolGlobalState)  );

	// Store this for later.
	gpJSContextPoolGlobalState->szApplicationRoot = szApplicationRoot;
	gpJSContextPoolGlobalState->readonlyServer = readonlyServer;
	gpJSContextPoolGlobalState->enableDiagnostics = enableDiagnostics;

	SG_ERR_CHECK(  SG_localsettings__get__sz(pCtx, SG_LOCALSETTING__SERVER_SSJS_MUTABLE, NULL, &szSsjsMutable, NULL)  );
	gpJSContextPoolGlobalState->ssjsMutable = (szSsjsMutable!=NULL && SG_stricmp(szSsjsMutable, "true")==0);
	SG_NULLFREE(pCtx, szSsjsMutable);

	SG_ERR_CHECK(  _sg_jscontextpool_getpaths(pCtx)  );

	// Start up SpiderMonkey.
	SG_ERR_CHECK(  SG_jscore__new_runtime(pCtx, SG_jsglue__context_callback, NULL, skipModules, NULL)  );

	SG_ERR_CHECK(  SG_mutex__init(pCtx, &gpJSContextPoolGlobalState->lock)  );

	return;
fail:
	SG_NULLFREE(pCtx, szSsjsMutable);

	SG_PATHNAME_NULLFREE(pCtx, gpJSContextPoolGlobalState->pPathToDispatchDotJS);
	SG_PATHNAME_NULLFREE(pCtx, gpJSContextPoolGlobalState->pPathToModules);
}


///////////////////////////////////////////////////////////////////////////////

static void _sg_jscontext__create(SG_context * pCtx, SG_jscontext ** ppJs)
{
	SG_jscontext * pJs = NULL;

	SG_ASSERT(pCtx!=NULL);
	SG_NULLARGCHECK_RETURN(ppJs);

	SG_ERR_CHECK(  SG_alloc1(pCtx, pJs)  );

	SG_ERR_CHECK(  SG_jscore__new_context(pCtx, &pJs->cx, &pJs->glob, gpJSContextPoolGlobalState->szApplicationRoot, gpJSContextPoolGlobalState->readonlyServer, gpJSContextPoolGlobalState->enableDiagnostics)  );
	pJs->isInARequest = SG_TRUE;

	*ppJs = pJs;

	return;
fail:
	SG_NULLFREE(pCtx, pJs);
}


///////////////////////////////////////////////////////////////////////////////

void SG_jscontext__acquire(SG_context * pCtx, SG_jscontext ** ppJs)
{
	SG_ASSERT(pCtx!=NULL);
	SG_NULLARGCHECK_RETURN(ppJs);

	if(gpJSContextPoolGlobalState->ssjsMutable)
	{
		_sg_jscontext__create(pCtx, ppJs);
		return;
	}

	SG_ERR_CHECK_RETURN(  SG_mutex__lock(pCtx, &gpJSContextPoolGlobalState->lock)  );

	if(gpJSContextPoolGlobalState->pFirstAvailableContext!=NULL)
	{
		SG_jscontext * pJs = gpJSContextPoolGlobalState->pFirstAvailableContext;

		gpJSContextPoolGlobalState->pFirstAvailableContext = pJs->pNextAvailableContext;
		pJs->pNextAvailableContext = NULL;

		++gpJSContextPoolGlobalState->numContextsCheckedOut;

		SG_ERR_CHECK_RETURN(  SG_mutex__unlock(pCtx, &gpJSContextPoolGlobalState->lock)  );

		(void)JS_SetContextThread(pJs->cx);
		JS_BeginRequest(pJs->cx);
		pJs->isInARequest = SG_TRUE;

		JS_SetContextPrivate(pJs->cx, pCtx);

		*ppJs = pJs;
	}
	else
	{
		++gpJSContextPoolGlobalState->numContextsCheckedOut;
		SG_ERR_CHECK_RETURN(  SG_mutex__unlock(pCtx, &gpJSContextPoolGlobalState->lock)  );

		_sg_jscontext__create(pCtx, ppJs);

		if(SG_context__has_err(pCtx) || *ppJs==NULL)
		{
			/* Use the version of the mutex routines that doesn't touch pCtx,
			because we're already in an error state. */
			SG_mutex__lock__bare(&gpJSContextPoolGlobalState->lock);
			--gpJSContextPoolGlobalState->numContextsCheckedOut;
			SG_mutex__unlock__bare(&gpJSContextPoolGlobalState->lock);
		}
	}
}

void SG_jscontext__release(SG_context * pCtx, SG_jscontext ** ppJs)
{
	if(ppJs==NULL || *ppJs==NULL)
		return;

	if(gpJSContextPoolGlobalState->ssjsMutable)
	{
		JS_EndRequest((*ppJs)->cx);
		JS_DestroyContext((*ppJs)->cx);

		if(SG_context__has_err(pCtx)) // An error was produced during GC...
		{
			SG_log__report_error__current_error(pCtx);
			SG_context__err_reset(pCtx);
		}

		SG_NULLFREE(pCtx, (*ppJs));
	}
	else
	{
		SG_jscontext * pJs = *ppJs;

		JS_MaybeGC(pJs->cx);

		JS_SetContextPrivate(pJs->cx, NULL); // Clear out the old pCtx pointer.

		JS_EndRequest(pJs->cx);
		pJs->isInARequest = SG_FALSE;
		(void)JS_ClearContextThread(pJs->cx);

		SG_ERR_CHECK_RETURN(  SG_mutex__lock(pCtx, &gpJSContextPoolGlobalState->lock)  );

		pJs->pNextAvailableContext = gpJSContextPoolGlobalState->pFirstAvailableContext;
		gpJSContextPoolGlobalState->pFirstAvailableContext = pJs;
		--gpJSContextPoolGlobalState->numContextsCheckedOut;

		SG_ERR_CHECK_RETURN(  SG_mutex__unlock(pCtx, &gpJSContextPoolGlobalState->lock)  );

		*ppJs = NULL;
	}
}

void SG_jscontext__suspend(SG_jscontext * pJs)
{
	if(pJs!=NULL && pJs->isInARequest)
	{
		JS_EndRequest(pJs->cx);
		pJs->isInARequest = SG_FALSE;
	}
}

void SG_jscontext__resume(SG_jscontext * pJs)
{
	if(pJs!=NULL && !pJs->isInARequest)
	{
		JS_BeginRequest(pJs->cx);
		pJs->isInARequest = SG_TRUE;
	}
}


///////////////////////////////////////////////////////////////////////////////

void SG_jscontextpool__teardown(SG_context * pCtx)
{
	if(gpJSContextPoolGlobalState!=NULL)
	{
		SG_ERR_CHECK_RETURN(  SG_mutex__lock(pCtx, &gpJSContextPoolGlobalState->lock)  );

		// Wait until all outstanding SG_jscontexts have been released. Don't try to terminate
		// early, otherwise we have a race condition on our hands: If the outstanding SG_jscontext
		// tries to perform any JavaScript operations before the app terminates, but after we have
		// called JS_Shutdown(), we'll end up crashing on exit. This of course is worse than
		//  making the user hit Ctrl-C again to do a hard shutdown if it's taking too long.
		if(gpJSContextPoolGlobalState->numContextsCheckedOut > 0)
		{
			SG_ERR_IGNORE(  SG_log__report_error(pCtx, "Waiting on %d SG_jscontexts that are still in use.", gpJSContextPoolGlobalState->numContextsCheckedOut)  );
			while(gpJSContextPoolGlobalState->numContextsCheckedOut > 0)
			{
				SG_ERR_CHECK_RETURN(  SG_mutex__unlock(pCtx, &gpJSContextPoolGlobalState->lock)  );
				SG_sleep_ms(10);
				SG_ERR_CHECK_RETURN(  SG_mutex__lock(pCtx, &gpJSContextPoolGlobalState->lock)  );
			}
		}

		SG_ERR_CHECK_RETURN(  SG_mutex__unlock(pCtx, &gpJSContextPoolGlobalState->lock)  );
		SG_mutex__destroy(&gpJSContextPoolGlobalState->lock);

		{
			SG_jscontext * pJs = gpJSContextPoolGlobalState->pFirstAvailableContext;
			while(pJs!=NULL)
			{
				(void)JS_SetContextThread(pJs->cx);
				JS_BeginRequest(pJs->cx);
				JS_SetContextPrivate(pJs->cx, pCtx);
				JS_EndRequest(pJs->cx);

				pJs = pJs->pNextAvailableContext;
			}
		}

		while(gpJSContextPoolGlobalState->pFirstAvailableContext!=NULL)
		{
			SG_jscontext * pJs = gpJSContextPoolGlobalState->pFirstAvailableContext;
			gpJSContextPoolGlobalState->pFirstAvailableContext = pJs->pNextAvailableContext;

			JS_DestroyContext(pJs->cx);

			if(SG_context__has_err(pCtx)) // An error was produced during GC...
			{
				SG_log__report_error__current_error(pCtx);
				SG_context__err_reset(pCtx);
			}

			SG_NULLFREE(pCtx, pJs);
		}

		SG_PATHNAME_NULLFREE(pCtx, gpJSContextPoolGlobalState->pPathToDispatchDotJS);
		SG_PATHNAME_NULLFREE(pCtx, gpJSContextPoolGlobalState->pPathToCore);
		SG_PATHNAME_NULLFREE(pCtx, gpJSContextPoolGlobalState->pPathToModules);
		SG_NULLFREE(pCtx, gpJSContextPoolGlobalState);
	}
}


///////////////////////////////////////////////////////////////////////////////
