/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include <sg.h>
#include <curl/easy.h>

//////////////////////////////////////////////////////////////////////////

/**
 * The structure for which SG_curl is an opaque wrapper.
 */
typedef struct
{
	SG_context* pCtx; // We don't own this and shouldn't attempt to free it. It's the thread's main SG_context.
	CURL* pCurl;
	SG_string* pstrErr; // An error message relayed to us by the server.

	SG_curl__callback* pFnReadRequest;
	void* pReadState; // We don't own this and shouldn't attempt to free it.

	SG_curl__callback* pFnWriteResponse;
	void* pWriteState; // We don't own this and shouldn't attempt to free it.

	SG_curl_progress_callback* pFnProgress;
	void* pProgressState; // We don't own this and shouldn't attempt to free it.
} _sg_curl;

//////////////////////////////////////////////////////////////////////////

#define _SETOPT(val) 	_sg_curl* pMe = (_sg_curl*)pCurl;	\
	CURLcode rc = CURLE_OK;									\
	SG_NULLARGCHECK_RETURN(pCurl);							\
	rc = curl_easy_setopt(pMe->pCurl, option, val);			\
	if (rc)													\
		SG_ERR_THROW_RETURN(SG_ERR_LIBCURL(rc))

static void _setopt__pv(SG_context* pCtx, SG_curl* pCurl, CURLoption option, void* pv)
{
	_SETOPT(pv);
}
static void _setopt__read_cb(SG_context* pCtx, SG_curl* pCurl, CURLoption option, curl_read_callback cb)
{
	_SETOPT(cb);
}
static void _setopt__write_cb(SG_context* pCtx, SG_curl* pCurl, CURLoption option, curl_write_callback cb)
{
	_SETOPT(cb);
}
static void _setopt__progress_cb(SG_context* pCtx, SG_curl* pCurl, CURLoption option, curl_progress_callback cb)
{
	_SETOPT(cb);
}

void SG_curl__setopt__sz(SG_context* pCtx, SG_curl* pCurl, CURLoption option, const char* pszVal)
{
	_SETOPT(pszVal);
}

void SG_curl__setopt__int32(SG_context* pCtx, SG_curl* pCurl, CURLoption option, SG_int32 iVal)
{
	_SETOPT(iVal);
}
void SG_curl__setopt__int64(SG_context* pCtx, SG_curl* pCurl, CURLoption option, SG_int64 iVal)
{
	_SETOPT(iVal);
}

void SG_curl__getinfo__int32(SG_context* pCtx, SG_curl* pCurl, CURLINFO info, SG_int32* piVal)
{
	_sg_curl* p = (_sg_curl*)pCurl;
	CURLcode rc = CURLE_OK;
	SG_int64 val;

	SG_NULLARGCHECK_RETURN(pCurl);

	rc = curl_easy_getinfo(p->pCurl, info, &val);
	if (rc)
		SG_ERR_THROW_RETURN(SG_ERR_LIBCURL(rc));

	*piVal = (SG_int32)val;
}

//////////////////////////////////////////////////////////////////////////

/* libcurl will call these shims for chunked I/O. 
 * They delegate to SG_curl__callback routines, which have sglib-friendly signatures. */

static size_t _callback_shim(char* buffer, size_t size, size_t nmemb, _sg_curl* pMe, SG_curl__callback* pFn, void* pState)
{
	SG_context* pCtx = pMe->pCtx; // so SG_ERR_CHECK works
	SG_uint32 len_handled = 0;

	SG_ERR_CHECK(  pFn(pCtx, (SG_curl*)pMe, buffer, (SG_uint32)(size*nmemb), pState, &len_handled)  );
	return len_handled;

fail:
	return CURL_READFUNC_ABORT;
}

static size_t _read_callback_shim(char* buffer, size_t size, size_t nmemb, void* pVoidState)
{
	_sg_curl* pMe = (_sg_curl*)pVoidState;
	return _callback_shim(buffer, size, nmemb, pMe, pMe->pFnReadRequest, pMe->pReadState);
}

static size_t _write_callback_shim(char* buffer, size_t size, size_t nmemb, void* pVoidState)
{
	_sg_curl* pMe = (_sg_curl*)pVoidState;
	return _callback_shim(buffer, size, nmemb, pMe, pMe->pFnWriteResponse, pMe->pWriteState);
}

static int _progress_callback_shim(void *clientp, double dltotal, double dlnow, double ultotal, double ulnow)
{
	_sg_curl* pMe = (_sg_curl*)clientp;
	SG_context* pCtx = pMe->pCtx; // so SG_ERR_CHECK works
	
	SG_ERR_CHECK(  pMe->pFnProgress(pCtx, dltotal, dlnow, ultotal, ulnow, pMe->pProgressState)  );
	return 0;

fail:
	return CURL_READFUNC_ABORT;
}

//////////////////////////////////////////////////////////////////////////

/* Built-in SG_curl__callbacks for reading/writing SG_strings and SG_files */

// http://curl.haxx.se/libcurl/c/curl_easy_setopt.html#CURLOPTREADFUNCTION

static void _read_file_chunk(SG_context* pCtx, SG_curl* pCurl, char* buffer, SG_uint32 bufLen, void* pVoidState, SG_uint32* pLenHandled)
{
	SG_file* pFile = (SG_file*)pVoidState;

	SG_UNUSED(pCurl);

	SG_ERR_IGNORE(  SG_log__report_verbose(pCtx, "Reading up to %u bytes from fragball", bufLen)  );
	SG_file__read(pCtx, pFile, bufLen, (SG_byte*)buffer, pLenHandled);
	SG_ERR_IGNORE(  SG_log__report_verbose(pCtx, "Read %u bytes from fragball", *pLenHandled)  );
	if (SG_CONTEXT__HAS_ERR(pCtx))
		SG_log__report_verbose__current_error(pCtx);
	SG_ERR_CHECK_RETURN_CURRENT_DISREGARD(SG_ERR_EOF);
}

// http://curl.haxx.se/libcurl/c/curl_easy_setopt.html#CURLOPTWRITEFUNCTION

static void _write_file_chunk(SG_context* pCtx, SG_curl* pCurl, char* buffer, SG_uint32 bufLen, void* pVoidState, SG_uint32* pLenHandled)
{
	SG_file* pFile = (SG_file*)pVoidState;
	_sg_curl* pMe = (_sg_curl*)pCurl;
	SG_int32 responseCode = 0;

	SG_ERR_CHECK(  SG_curl__getinfo__int32(pCtx, pCurl, CURLINFO_RESPONSE_CODE, &responseCode)  );
	if (responseCode == 500)
	{
		if (pMe->pstrErr == NULL)
		{
			SG_int32 lenResponse = 0;

			/* We assume an HTTP 500 response will be small enough to fit into an SG_string. */
			SG_ERR_CHECK(  SG_curl__getinfo__int32(pCtx, pCurl, CURLINFO_CONTENT_LENGTH_DOWNLOAD, &lenResponse)  );
			if (lenResponse > 0)
				SG_ERR_CHECK(  SG_STRING__ALLOC__RESERVE(pCtx, &pMe->pstrErr, (SG_uint32)lenResponse + 1)  );
			else
				SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pMe->pstrErr)  );
		}
		
		SG_ERR_CHECK(  SG_string__append__buf_len(pCtx, pMe->pstrErr, (const SG_byte*)buffer, bufLen)  );
	}

	if (bufLen) // libcurl docs say we may be called with zero length if file is empty
		SG_ERR_CHECK(  SG_file__write(pCtx, pFile, bufLen, (SG_byte*)buffer, pLenHandled)  );

fail:
	;
}

static void _write_string_chunk(SG_context* pCtx, SG_curl* pCurl, char* buffer, SG_uint32 bufLen, void* pVoidState, SG_uint32* pLenHandled)
{
	SG_string* pstrResponse = (SG_string*)pVoidState;
	//_sg_curl* pMe = (_sg_curl*)pCurl;
	*pLenHandled = 0;

	SG_UNUSED(pCurl);

	if (bufLen) // libcurl docs say we may be called with zero length if file is empty
	{
		SG_ERR_CHECK_RETURN(  SG_string__append__buf_len(pCtx, pstrResponse, (const SG_byte*)buffer, bufLen)  );
		*pLenHandled = bufLen;
	}
}

//////////////////////////////////////////////////////////////////////////

/* Set up to use the built-in read/write callbacks. */

void SG_curl__set__read_string(SG_context* pCtx, SG_curl* pCurl, SG_string* pString)
{
	SG_UNUSED(pCtx);
	SG_UNUSED(pCurl);
	SG_UNUSED(pString);

	SG_ERR_THROW_RETURN(SG_ERR_NOTIMPLEMENTED);
}

/**
 * Append the HTTP response to the end of the provided SG_string.
 * The caller retains ownership of the string and should free it.
 */
void SG_curl__set__write_string(SG_context* pCtx, SG_curl* pCurl, SG_string* pString)
{
	_sg_curl* pMe = (_sg_curl*)pCurl;

	SG_NULLARGCHECK_RETURN(pCurl);
	SG_NULLARGCHECK_RETURN(pString);

	pMe->pWriteState = pString;
	pMe->pFnWriteResponse = _write_string_chunk;
	SG_ERR_CHECK_RETURN(  _setopt__pv(pCtx, pCurl, CURLOPT_WRITEDATA, pCurl)  );
	SG_ERR_CHECK_RETURN(  _setopt__write_cb(pCtx, pCurl, CURLOPT_WRITEFUNCTION, _write_callback_shim)  );
}

/**
 * Read the HTTP request from the provided SG_file, which should be open with a readable handle.
 * The caller retains ownership of the file and should free/close/delete it after SG_curl__perform.
 */
void SG_curl__set__read_file(SG_context* pCtx, SG_curl* pCurl, SG_file* pFile)
{
	_sg_curl* pMe = (_sg_curl*)pCurl;

	SG_NULLARGCHECK_RETURN(pCurl);
	SG_NULLARGCHECK_RETURN(pFile);

	pMe->pReadState = pFile;
	pMe->pFnReadRequest = _read_file_chunk;
	SG_ERR_CHECK_RETURN(  _setopt__pv(pCtx, pCurl, CURLOPT_READDATA, pCurl)  );
	SG_ERR_CHECK_RETURN(  _setopt__read_cb(pCtx, pCurl, CURLOPT_READFUNCTION, _read_callback_shim)  );
}

/**
 * Write the HTTP response to the provided SG_file, which should be open with a writeable handle.
 * The caller retains ownership of the file and should free/close/delete it after SG_curl__perform.
 */
void SG_curl__set__write_file(SG_context* pCtx, SG_curl* pCurl, SG_file* pFile)
{
	_sg_curl* pMe = (_sg_curl*)pCurl;

	SG_NULLARGCHECK_RETURN(pCurl);
	SG_NULLARGCHECK_RETURN(pFile);

	pMe->pWriteState = pFile;
	pMe->pFnWriteResponse = _write_file_chunk;
	SG_ERR_CHECK_RETURN(  _setopt__pv(pCtx, pCurl, CURLOPT_WRITEDATA, pCurl)  );
	SG_ERR_CHECK_RETURN(  _setopt__write_cb(pCtx, pCurl, CURLOPT_WRITEFUNCTION, _write_callback_shim)  );
}

/**
 * Use the provided progress callback for this request/response.
 * The caller retains ownership of pState and should free it after SG_curl__perform.
 */
void SG_curl__set__progress_cb(SG_context* pCtx, SG_curl* pCurl, SG_curl_progress_callback* pcb, void* pState)
{
	_sg_curl* pMe = (_sg_curl*)pCurl;

	SG_NULLARGCHECK_RETURN(pCurl);
	SG_NULLARGCHECK_RETURN(pcb);

	pMe->pProgressState = pState;
	pMe->pFnProgress = pcb;

	SG_ERR_CHECK_RETURN(  SG_curl__setopt__int32(pCtx, pCurl, CURLOPT_NOPROGRESS, 0)  );
	SG_ERR_CHECK_RETURN(  _setopt__pv(pCtx, pCurl, CURLOPT_PROGRESSDATA, pCurl)  );
	SG_ERR_CHECK_RETURN(  _setopt__progress_cb(pCtx, pCurl, CURLOPT_PROGRESSFUNCTION, _progress_callback_shim)  );
}

//////////////////////////////////////////////////////////////////////////


void SG_curl__global_init(SG_context* pCtx)
{
	CURLcode rc;

	rc = curl_global_init(CURL_GLOBAL_ALL);

	if (rc)
		SG_ERR_THROW_RETURN(SG_ERR_LIBCURL(rc));
}

void SG_curl__global_cleanup()
{
	(void) curl_global_cleanup();
}

void SG_curl__free(SG_context* pCtx, SG_curl* pCurl)
{
	_sg_curl* p = (_sg_curl*)pCurl;

	if (p)
	{
		SG_STRING_NULLFREE(pCtx, p->pstrErr);
		curl_easy_cleanup(p->pCurl);
		SG_NULLFREE(pCtx, p);
	}
}

void SG_curl__alloc(SG_context* pCtx, SG_curl** ppCurl)
{
	_sg_curl* p = NULL;

	SG_ERR_CHECK(  SG_alloc1(pCtx, p)  );

	p->pCurl = curl_easy_init();
	if (!p->pCurl)
		SG_ERR_THROW_RETURN(SG_ERR_LIBCURL(CURLE_FAILED_INIT));

	p->pCtx = pCtx;

	*ppCurl = (SG_curl*)p;

	return;

fail:
	SG_ERR_IGNORE(  SG_curl__free(pCtx, (SG_curl*)p)  );
}

void SG_curl__reset(SG_context* pCtx, SG_curl* pCurl)
{
	_sg_curl* p = (_sg_curl*)pCurl;

	SG_NULLARGCHECK_RETURN(pCurl);

	SG_STRING_NULLFREE(pCtx, p->pstrErr);
	curl_easy_reset(p->pCurl);

	p->pFnReadRequest = NULL;
	p->pReadState = NULL;
	p->pFnWriteResponse = NULL;
	p->pWriteState = NULL;
	p->pFnProgress = NULL;
	p->pProgressState = NULL;
}

//////////////////////////////////////////////////////////////////////////

void SG_curl__set_one_header(SG_context* pCtx, SG_curl* pCurl, const char* pszHeader, struct curl_slist** ppHeaderList)
{
	//_sg_curl* p = (_sg_curl*)pCurl;
	CURLcode rc = CURLE_OK;
	struct curl_slist* pHeaderList = NULL;

	SG_NULLARGCHECK_RETURN(pCurl);

	SG_NULLARGCHECK_RETURN(ppHeaderList);

	pHeaderList = curl_slist_append(*ppHeaderList, pszHeader);
	if (!pHeaderList)
		SG_ERR_THROW2_RETURN(SG_ERR_UNSPECIFIED, (pCtx, "Failed to add HTTP header."));

	rc = curl_easy_setopt(pCurl, CURLOPT_HTTPHEADER, pHeaderList);
	if (rc)
		SG_ERR_THROW(SG_ERR_LIBCURL(rc));

	SG_RETURN_AND_NULL(pHeaderList, ppHeaderList);

fail:
	if (pHeaderList)
		curl_slist_free_all(pHeaderList);
}

void SG_curl__free_headers(SG_UNUSED_PARAM(SG_context* pCtx), struct curl_slist* pHeaderList)
{
	SG_UNUSED(pCtx);
	if (pHeaderList)
		curl_slist_free_all(pHeaderList);
}

//////////////////////////////////////////////////////////////////////////

void SG_curl__throw_on_non200(SG_context* pCtx, SG_curl* pCurl)
{
	SG_int32 httpResponseCode = 0;
	SG_vhash* pvhErr = NULL;
	_sg_curl* p = (_sg_curl*)pCurl;

	SG_NULLARGCHECK_RETURN(pCurl);

	SG_ERR_CHECK(  SG_curl__getinfo__int32(pCtx, pCurl, CURLINFO_RESPONSE_CODE, &httpResponseCode)  );
	if (httpResponseCode != 200)
	{
		if (httpResponseCode == 500 && p->pstrErr)
		{
			SG_bool bHas = SG_FALSE;
			const char* szMsg = NULL;
			SG_ERR_CHECK(  SG_VHASH__ALLOC__FROM_JSON__SZ(pCtx, &pvhErr, SG_string__sz(p->pstrErr))  );
			SG_ERR_CHECK(  SG_vhash__has(pCtx, pvhErr, "msg", &bHas)  );
			if (bHas)
				SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhErr, "msg", &szMsg)  );
			if (szMsg)
				SG_ERR_THROW2(SG_ERR_EXTENDED_HTTP_500, (pCtx, "%s", szMsg));
			else
				SG_ERR_THROW2(SG_ERR_EXTENDED_HTTP_500, (pCtx, "%s", SG_string__sz(p->pstrErr)));
		}
		else
			SG_ERR_THROW2(SG_ERR_SERVER_HTTP_ERROR, (pCtx, "%d", httpResponseCode));
	}

	/* common cleanup */
fail:
	SG_VHASH_NULLFREE(pCtx, pvhErr);
}

void SG_curl__perform(SG_context* pCtx, SG_curl* pCurl)
{
	CURLcode rc = CURLE_OK;
	_sg_curl* pMe = (_sg_curl*)pCurl;
	
	SG_NULLARGCHECK_RETURN(pCurl);
	
	rc = curl_easy_perform(pMe->pCurl);
	
	// Check for errors in the request and response callbacks.
	SG_ERR_CHECK_RETURN_CURRENT;

	// No callback errors. Make sure curl result code is also kosher.
	if (rc)
		SG_ERR_THROW_RETURN(SG_ERR_LIBCURL(rc));
}
