/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg_mrg__private_file_mrg.h
 *
 * @details
 *
 */

//////////////////////////////////////////////////////////////////

#ifndef H_SG_MRG__PRIVATE_FILE_MRG_H
#define H_SG_MRG__PRIVATE_FILE_MRG_H

BEGIN_EXTERN_C;

//////////////////////////////////////////////////////////////////

struct _amf_data
{
	SG_mrg *				pMrg;
	SG_mrg_cset *			pMrgCSet;
	SG_mrg_cset_stats *		pMrgCSetStats;
};

typedef struct _amf_data _amf_data;

//////////////////////////////////////////////////////////////////

/**
 * Fetch a copy of each version of the file (ancestor and each
 * branch/leaf) into our private temp-dir.
 *
 * These are written to disk to allow an external mergetool to
 * use them and write the merge-result to the same directory.
 *
 * Even if we select the internal the builtin core diff3 engine for
 * the automerge, we still want them written to temp files in case
 * of conflicts and/or so that the user can re-merge them using
 * a GUI tool during a RESOLVE.
 *
 * We do all of this (even the merge-result file) in a private
 * temp directory so that 'vv merge --test' can completely run
 * without trashing/modifying their working directory.
 *
 */
static void _sg_mrg__fetch_file_versions_into_temp_files(SG_context * pCtx,
														 SG_mrg * pMrg,
														 SG_mrg_cset_entry_conflict * pMrgCSetEntryConflict)
{
	SG_rbtree_iterator * pIter = NULL;
	const char * pszKey_k;
	const SG_vector * pVec_k;
	const char * pszFilename = NULL;
	SG_uint32 nrUnique;
	SG_bool bFound;

	SG_ERR_CHECK(  SG_rbtree__count(pCtx, pMrgCSetEntryConflict->prbUnique_File_HidBlob, &nrUnique)  );
	if (nrUnique != 2)
		SG_ERR_THROW2_RETURN(  SG_ERR_NOTIMPLEMENTED,
							   (pCtx, "TODO figure out how to layout temp files when more than 2 versions of the file.")  );

	// create and record pathnames for the 3 inputs and the 1 result file.

	SG_ERR_CHECK(  _sg_mrg__create_pathname_for_conflict_file(pCtx, pMrg, pMrgCSetEntryConflict, "ANCESTOR",
															  &pMrgCSetEntryConflict->pPathTempFile_Ancestor)  );
	SG_ERR_CHECK(  _sg_mrg__create_pathname_for_conflict_file(pCtx, pMrg, pMrgCSetEntryConflict, "BASELINE",
															  &pMrgCSetEntryConflict->pPathTempFile_Baseline)  );
	SG_ERR_CHECK(  _sg_mrg__create_pathname_for_conflict_file(pCtx, pMrg, pMrgCSetEntryConflict, "OTHER",
															  &pMrgCSetEntryConflict->pPathTempFile_Other)  );
	SG_ERR_CHECK(  _sg_mrg__create_pathname_for_conflict_file(pCtx, pMrg, pMrgCSetEntryConflict, "RESULT",
															  &pMrgCSetEntryConflict->pPathTempFile_Result)  );

	// build the gui-friendly labels for each input panel.  BASELINE, ANCESTOR, and OTHER.
	// This is a bit contrived, but also build a RESULT label.
	// I say contrived, because for some GUI tools, the center panel is initially the
	// ancestor version of the file and as they edit and/or perform the merge, the
	// center panel becomes the result file.  So give them tokens for both and they
	// can configure the tool in localsettings either way.

	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pMrgCSetEntryConflict->pStringAncestorLabel)  );
	SG_ERR_CHECK(  SG_pathname__filename(pCtx, pMrgCSetEntryConflict->pPathTempFile_Baseline, &pszFilename)  );
	SG_ERR_CHECK(  SG_string__sprintf(pCtx, pMrgCSetEntryConflict->pStringAncestorLabel, "Ancestor: %s",
									  pszFilename)  );

	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pMrgCSetEntryConflict->pStringBaselineLabel)  );
	SG_ERR_CHECK(  SG_pathname__filename(pCtx, pMrgCSetEntryConflict->pPathTempFile_Baseline, &pszFilename)  );
	SG_ERR_CHECK(  SG_string__sprintf(pCtx, pMrgCSetEntryConflict->pStringBaselineLabel, "Baseline: %s",
									  pszFilename)  );
	
	// fetch the 3 input versions of the file.  (we could use the version
	// in the working directory, but getting a private/fresh copy feels safer.)

	SG_ERR_CHECK(  _sg_mrg__export_to_temp_file(pCtx, pMrg,
												pMrgCSetEntryConflict->pMrgCSetEntry_Ancestor->bufHid_Blob,
												pMrgCSetEntryConflict->pPathTempFile_Ancestor)  );

	SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pIter, pMrgCSetEntryConflict->prbUnique_File_HidBlob, &bFound, &pszKey_k, (void **)&pVec_k)  );
	while (bFound)
	{
		if (strcmp(pszKey_k, pMrgCSetEntryConflict->pMrgCSetEntry_Baseline->bufHid_Blob) == 0)
		{
			SG_ERR_CHECK(  _sg_mrg__export_to_temp_file(pCtx, pMrg,
														pszKey_k,
														pMrgCSetEntryConflict->pPathTempFile_Baseline)  );
		}
		else
		{
			SG_ERR_CHECK(  _sg_mrg__export_to_temp_file(pCtx, pMrg,
														pszKey_k,
														pMrgCSetEntryConflict->pPathTempFile_Other)  );

			SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pMrgCSetEntryConflict->pStringOtherLabel)  );
			SG_ERR_CHECK(  SG_pathname__filename(pCtx, pMrgCSetEntryConflict->pPathTempFile_Other, &pszFilename)  );
			SG_ERR_CHECK(  SG_string__sprintf(pCtx, pMrgCSetEntryConflict->pStringOtherLabel, "Other: %s",
											  pszFilename)  );
		}

		SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pIter, &bFound, &pszKey_k, (void **)&pVec_k)  );
	}

	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pMrgCSetEntryConflict->pStringResultLabel)  );
	SG_ERR_CHECK(  SG_string__sprintf(pCtx, pMrgCSetEntryConflict->pStringResultLabel, "Merge Result")  );

fail:
	SG_RBTREE_ITERATOR_NULLFREE(pCtx, pIter);
}

//////////////////////////////////////////////////////////////////

/**
 * Look at the entryname and/or the file content of the ancestor
 * version of the file and select the best mergetool for it.
 *
 * Update the mergetool name field in the conflict.
 */
static void _sg_mrg__select_mergetool(SG_context * pCtx,
									  SG_mrg * pMrg,
									  SG_mrg_cset_entry_conflict * pMrgCSetEntryConflict)
{
	SG_repo * pRepo;

	SG_ASSERT(  ((pMrgCSetEntryConflict->flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__MASK)
				 == SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__TBD)  );
	SG_ASSERT( (pMrgCSetEntryConflict->pMrgCSetEntry_Composite->tneType == SG_TREENODEENTRY_TYPE_REGULAR_FILE) );

	// use the pathname and contents of the ANCESTOR version of the file
	// to choose the mergetool.
	//
	// this is an arbitrary choice, but it simplifies things.
	// 
	// Most of time, it won't matter, but if there was a rename in one or more
	// of the leaves/branches, then we have to decide which of the new names
	// should be used. and if there is an unresolved-divergent-rename, then
	// we also have to worry about that.  so just use the ancestor name to
	// guide the lookup.
	//
	// likewise we use the ancestor version of the file content (for "#!"
	// (shebang) style matches), if appropriate.
	//
	// Also by using the ancestor entryname, we don't have to worry so much
	// about whether the user has resolved the rename conflict before or
	// after they resolve the file-content conflict.
	//
	// this may return null if there is no match.

	SG_ERR_CHECK(  SG_pendingtree__get_repo(pCtx, pMrg->pPendingTree, &pRepo)  );
	SG_ERR_CHECK(  SG_mergetool__select(pCtx,
										SG_MERGETOOL__CONTEXT__MERGE,
										NULL,
										pMrgCSetEntryConflict->pPathTempFile_Ancestor,
										pMrg->pszMergeToolOverride,
										pRepo,
										&pMrgCSetEntryConflict->pMergeTool)  );

fail:
	return;
}

static void _sg_mrg__invoke_mergetool(SG_context * pCtx,
									  SG_mrg * pMrg,
									  SG_mrg_cset_entry_conflict * pMrgCSetEntryConflict,
									  SG_mrg_cset_stats * pMrgCSetStats)
{
	SG_repo * pRepo;
	SG_varray * pvaExternal = NULL;
	SG_int32 statusMergeTool = SG_MERGETOOL__STATUS__NOT_ATTEMPTED;

	if (pMrgCSetEntryConflict->pMergeTool == NULL)
	{
		// no mergetool defined/appropriate ==> "No Rule"

		pMrgCSetEntryConflict->flags &= ~SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__TBD;
		pMrgCSetEntryConflict->flags |=  SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__NO_RULE;
		pMrgCSetStats->nrAutoMerge_TBD--;
		pMrgCSetStats->nrAutoMerge_NoRule++;
		goto done;
	}

	SG_ERR_CHECK(  SG_pendingtree__get_repo(pCtx, pMrg->pPendingTree, &pRepo)  );
	SG_ERR_CHECK(  SG_mergetool__invoke(pCtx,
										pMrgCSetEntryConflict->pMergeTool,
										pMrgCSetEntryConflict->pPathTempFile_Ancestor, pMrgCSetEntryConflict->pStringAncestorLabel,
										pMrgCSetEntryConflict->pPathTempFile_Baseline, pMrgCSetEntryConflict->pStringBaselineLabel,
										pMrgCSetEntryConflict->pPathTempFile_Other,    pMrgCSetEntryConflict->pStringOtherLabel,
										pMrgCSetEntryConflict->pPathTempFile_Result,   pMrgCSetEntryConflict->pStringResultLabel,
										&statusMergeTool)  );
	switch (statusMergeTool)
	{
	case SG_MERGETOOL__STATUS__NOT_ATTEMPTED:
		// leave as TBD
		goto done;

	case SG_MERGETOOL__STATUS__SUCCESSFUL:
		pMrgCSetEntryConflict->flags &= ~SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__TBD;
		pMrgCSetEntryConflict->flags |=  SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__AUTO_OK;
		pMrgCSetStats->nrAutoMerge_TBD--;
		pMrgCSetStats->nrAutoMerge_OK++;
		goto done;

	case SG_MERGETOOL__STATUS__CONFLICT:
		pMrgCSetEntryConflict->flags &= ~SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__TBD;
		pMrgCSetEntryConflict->flags |=  SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__AUTO_CONFLICT;
		pMrgCSetStats->nrAutoMerge_TBD--;
		pMrgCSetStats->nrAutoMerge_Conflict++;
		goto done;

	default:
		// need to handle all cases to pacify the compiler
		// just treat any status that we don't recognize as an error
	case SG_MERGETOOL__STATUS__ERROR:
		pMrgCSetEntryConflict->flags &= ~SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__TBD;
		pMrgCSetEntryConflict->flags |=  SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__AUTO_ERROR;
		pMrgCSetStats->nrAutoMerge_TBD--;
		pMrgCSetStats->nrAutoMerge_Error++;
		goto done;
	}

done:
	;
fail:
	SG_VARRAY_NULLFREE(pCtx, pvaExternal);
}

//////////////////////////////////////////////////////////////////

/**
 * We get called once for each conflict in SG_mrg_cset.prbConflicts.
 * We may be a structural conflict or a TBD file-content conflict.
 * Ignore the former and try to automerge the contents of the latter.
 */
static SG_rbtree_foreach_callback _try_automerge_files;

static void _try_automerge_files(SG_context * pCtx,
								 const char * pszKey_Gid_Entry,
								 void * pVoidAssocData_MrgCSetEntryConflict,
								 void * pVoid_Data)
{
	SG_mrg_cset_entry_conflict * pMrgCSetEntryConflict = (SG_mrg_cset_entry_conflict *)pVoidAssocData_MrgCSetEntryConflict;
	_amf_data * pAmfData = (_amf_data *)pVoid_Data;

	SG_UNUSED(pszKey_Gid_Entry);

	if ((pMrgCSetEntryConflict->flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__TBD) == 0)
		return;

	SG_ERR_CHECK(  _sg_mrg__fetch_file_versions_into_temp_files(pCtx, pAmfData->pMrg, pMrgCSetEntryConflict)  );

	SG_ERR_CHECK(  _sg_mrg__select_mergetool(pCtx, pAmfData->pMrg, pMrgCSetEntryConflict)  );
	SG_ERR_CHECK(  _sg_mrg__invoke_mergetool(pCtx, pAmfData->pMrg, pMrgCSetEntryConflict, pAmfData->pMrgCSetStats)  );

	// leave the final merge-result and the temp-files in the temp-dir.

fail:
	return;
}

//////////////////////////////////////////////////////////////////

void SG_mrg__automerge_files(SG_context * pCtx, SG_mrg * pMrg, SG_mrg_cset * pMrgCSet, SG_mrg_cset_stats * pMrgCSetStats)
{
	_amf_data amfData;

	SG_NULLARGCHECK_RETURN(pMrg);
	SG_NULLARGCHECK_RETURN(pMrgCSet);

	// if there were no conflicts *anywhere* in the result-cset, we don't have to do anything.

	if (!pMrgCSet->prbConflicts)
		return;

	// visit each entry in the result-cset and do any auto-merges that we can.

	amfData.pMrg = pMrg;
	amfData.pMrgCSet = pMrgCSet;
	amfData.pMrgCSetStats = pMrgCSetStats;

	SG_ERR_CHECK_RETURN(  SG_rbtree__foreach(pCtx,pMrgCSet->prbConflicts,_try_automerge_files,&amfData)  );
}

//////////////////////////////////////////////////////////////////

END_EXTERN_C;

#endif//H_SG_MRG__PRIVATE_FILE_MRG_H
