/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg_mrg__private_master_plan__utils.h
 *
 * @details
 *
 */

//////////////////////////////////////////////////////////////////

#ifndef H_SG_MRG__PRIVATE_MASTER_PLAN__UTILS_H
#define H_SG_MRG__PRIVATE_MASTER_PLAN__UTILS_H

BEGIN_EXTERN_C;

//////////////////////////////////////////////////////////////////

static void _sg_mrg__plan__compute_transient_path__source(SG_context * pCtx,
														  SG_mrg * pMrg,
														  const char * pszGid_Entry,
														  SG_pathname ** ppPath);

//////////////////////////////////////////////////////////////////

/**
 * Compute the transient disk path for the given entry as it
 * WILL APPEAR after we actually execute the plan.
 */
static void _sg_mrg__plan__compute_transient_path__target(SG_context * pCtx,
														  SG_mrg * pMrg,
														  const char * pszGid_Entry,
														  SG_pathname ** ppPath)
{
	SG_pathname * pPath_Allocated = NULL;
	SG_inv_entry * pInvEntry;
	SG_mrg_cset_entry * pMrgCSetEntry;
	SG_bool bFound;

	SG_NULLARGCHECK_RETURN(pMrg);
	SG_NULLARGCHECK_RETURN(pszGid_Entry);

	SG_ERR_CHECK(  SG_inv_dirs__fetch_target_binding(pCtx, pMrg->pInvDirs, pszGid_Entry,
													 &bFound, &pInvEntry, (void **)&pMrgCSetEntry)  );
	// TODO this should probably be an assert, but i want the extra info.
	if (!bFound)
		SG_ERR_THROW2_RETURN(  SG_ERR_NOT_FOUND,
							   (pCtx, "TODO Merge transient target path for [%s]", pszGid_Entry)  );

	// we only worry about the markerValue on the source (baseline) view
	// because the plan is concerned with converting the source view to
	// the target view.  so, for example, we don't park things in the target
	// view.

	// the assumption is that we want the target version of this entry.
	// this includes the target-entryname and the target-parent-directory.
	// *BUT* we don't know what the status of the parent-directory is, so
	// we use the more-general source version.

	if (pMrgCSetEntry->bufGid_Parent[0])
	{
		SG_ERR_CHECK(  _sg_mrg__plan__compute_transient_path__source(pCtx, pMrg,
																	 pMrgCSetEntry->bufGid_Parent,
																	 &pPath_Allocated)  );
		SG_ERR_CHECK(  SG_pathname__append__from_sz(pCtx, pPath_Allocated,
													SG_string__sz(pMrgCSetEntry->pStringEntryname))  );
	}
	else
	{
		SG_ASSERT(  (strcmp(SG_string__sz(pMrgCSetEntry->pStringEntryname),"@") == 0)  );
		SG_ERR_CHECK(  SG_pendingtree__get_working_directory_top(pCtx, pMrg->pPendingTree,
																 &pPath_Allocated)  );
	}

	*ppPath = pPath_Allocated;
	return;

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPath_Allocated);
}

/**
 * Compute the transient disk path for the given entry as
 * it WOULD APPEAR if we were actually executing the plan
 * as we compute it.  That is, compensate for the steps
 * we've already added to the PLAN.  For example, if we have
 * already SCHEDULED for some directories to be parked, a
 * reference to something within that directory needs to use
 * the parked path for the parent directory --- EVEN THOUGH
 * IT IS STILL IN THE WD IN ITS ORIGINAL LOCATION.
 *
 * This is intended for entries in the BASELINE (SOURCE)
 * version of the WD as we juggle it to become the result.
 */
static void _sg_mrg__plan__compute_transient_path__source(SG_context * pCtx,
														  SG_mrg * pMrg,
														  const char * pszGid_Entry,
														  SG_pathname ** ppPath)
{
	SG_pathname * pPath_Allocated = NULL;
	SG_inv_entry * pInvEntry;
	SG_mrg_cset_entry * pMrgCSetEntry;
	SG_bool bFound;

	SG_NULLARGCHECK_RETURN(pMrg);
	SG_NULLARGCHECK_RETURN(pszGid_Entry);

	SG_ERR_CHECK(  SG_inv_dirs__fetch_source_binding(pCtx, pMrg->pInvDirs, pszGid_Entry,
													 &bFound, &pInvEntry, (void **)&pMrgCSetEntry)  );
	if (bFound)
	{
		// the entry is present in the baseline.

		if (pMrgCSetEntry->markerValue & _SG_MRG__PLAN__MARKER__FLAGS__APPLIED)
		{
			// the entry has already been processed.  get the target version.

			SG_ERR_CHECK(  _sg_mrg__plan__compute_transient_path__target(pCtx, pMrg, pszGid_Entry, &pPath_Allocated)  );
		}
		else if (pMrgCSetEntry->markerValue & _SG_MRG__PLAN__MARKER__FLAGS__PARKED)
		{
			// the entry itself was parked.   it is in the parking lot.

			const SG_pathname * pPathParked;

			SG_ERR_CHECK(  SG_inv_entry__get_parked_path__ref(pCtx, pInvEntry, &pPathParked)  );
			SG_ASSERT(  (pPathParked)  );
			SG_ERR_CHECK_RETURN(  SG_PATHNAME__ALLOC__COPY(pCtx, &pPath_Allocated, pPathParked)  );
		}
		else
		{
			// the entry is still where it was in the baseline relative to its parent directory.
			// that is, it is still in the same directory and has the same entyrname.
			// (but its parent directory (or grand*parent) may have been moved or parked.)

			if (pMrgCSetEntry->bufGid_Parent[0])
			{
				SG_ERR_CHECK(  _sg_mrg__plan__compute_transient_path__source(pCtx, pMrg,
																			 pMrgCSetEntry->bufGid_Parent,
																			 &pPath_Allocated)  );
				SG_ERR_CHECK(  SG_pathname__append__from_sz(pCtx, pPath_Allocated,
															SG_string__sz(pMrgCSetEntry->pStringEntryname))  );
			}
			else
			{
				SG_ASSERT(  (strcmp(SG_string__sz(pMrgCSetEntry->pStringEntryname),"@") == 0)  );
				SG_ERR_CHECK(  SG_pendingtree__get_working_directory_top(pCtx, pMrg->pPendingTree,
																		 &pPath_Allocated)  );
			}
		}
	}
	else
	{
		SG_ERR_CHECK(  _sg_mrg__plan__compute_transient_path__target(pCtx, pMrg, pszGid_Entry, &pPath_Allocated)  );
	}

	*ppPath = pPath_Allocated;
	return;

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPath_Allocated);
}

//////////////////////////////////////////////////////////////////
#if 0
/**
 * auto-merge was attempted but not successful.
 * install the conflict file in the WD as the final-result.
 */
static void _sg_mrg__plan__alter_file_from_divergent_edit(SG_context * pCtx,
														  SG_mrg * pMrg,
														  SG_mrg_cset_entry * pMrgCSetEntry_FinalResult, SG_mrg_cset_entry * pMrgCSetEntry_Baseline,
														  const SG_string * pStringRepoPath_FinalResult, const SG_string * pStringRepoPath_Baseline,
														  const SG_pathname * pPath_FinalResult,         const SG_pathname * pPath_Baseline)
{
	SG_ASSERT(  (pMrgCSetEntry_FinalResult)  );
	SG_ASSERT(  (pMrgCSetEntry_FinalResult->bufHid_Blob[0] == 0)  );
	SG_ASSERT(  (pMrgCSetEntry_FinalResult->pMrgCSetEntryConflict)  );
	SG_ASSERT(  (pMrgCSetEntry_FinalResult->pMrgCSetEntryConflict->flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__AUTO_FAILED)  );
	SG_ASSERT(  (pMrgCSetEntry_FinalResult->pMrgCSetEntryConflict->prbUnique_File_HidBlob)  );
	SG_ASSERT(  (pMrgCSetEntry_FinalResult->pMrgCSetEntryConflict->pVec_MrgCSetEntry_Changes)  );
	SG_ASSERT(  (pMrgCSetEntry_FinalResult->pMrgCSetEntryConflict->pVec_MrgCSetEntryNeq_Changes)  );
	SG_ASSERT(  (pMrgCSetEntry_FinalResult->pMrgCSetEntryConflict->pPathTempFile_Result)  );

	// pretend that the auto-merge-result was OK and do a normal __alter_file
	// (so that any move/rename/attrbits/xattr changes get applied and the
	// original file is converted to the auto-merge-result)  This keeps the
	// WD and the pendingtree happy.
	//
	// TODO is it right to set the non-default attrbits/xattrs here -- knowing
	// TODO that whatever tool that they later use to manually merge the content
	// TODO may not respect them.  would it be better to do this during the resolve.

	SG_ERR_CHECK(  SG_wd_plan__alter_file_from_auto_merge_result(pCtx, pMrg->p_wd_plan,
																 pMrgCSetEntry_FinalResult->bufGid_Entry,
																 SG_string__sz(pStringRepoPath_FinalResult), SG_string__sz(pStringRepoPath_Baseline),
																 SG_pathname__sz(pPath_FinalResult),         SG_pathname__sz(pPath_Baseline),
																 SG_pathname__sz(pMrgCSetEntry_FinalResult->pMrgCSetEntryConflict->pPathTempFile_Result),
																 pMrgCSetEntry_Baseline->bufHid_Blob,
																 pMrgCSetEntry_FinalResult->attrBits,        pMrgCSetEntry_Baseline->attrBits,
																 pMrgCSetEntry_FinalResult->bufHid_XAttr,    pMrgCSetEntry_Baseline->bufHid_XAttr,
																 "FILE content of auto-merge conflict.")  );

fail:
	return;
}
#endif

//////////////////////////////////////////////////////////////////

END_EXTERN_C;

#endif//H_SG_MRG__PRIVATE_MASTER_PLAN__UTILS_H
