/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg_mrg__private_temp_dir.h
 *
 * @details
 *
 */

//////////////////////////////////////////////////////////////////

#ifndef H_SG_MRG__PRIVATE_TEMP_DIR_H
#define H_SG_MRG__PRIVATE_TEMP_DIR_H

BEGIN_EXTERN_C;

//////////////////////////////////////////////////////////////////

/**
 * Construct a pathname to a personal TEMP directory within the WD.
 * Use that to allocate the directory on disk.  We may use this directory for
 * holding files as we run external tools, such as diff3, during a merge.
 *
 * We want this temp directory to be local to the WD (so that files can be
 * moved/linked rather than copied -- both for input to diff3 and for the
 * auto-merge result -- IF THAT IS APPROPRIATE.  (And so that we don't have
 * to create a bunch of trash in the user's WD.)
 *
 * We want this TEMP directory to be for the duration of the MERGE/RESOLVE/
 * COMMIT effort.  That is, if MERGE needs to create a set of Ancestor/Mine/
 * Other files for an unresolved file conflict, RESOLVE will know where to
 * find them and COMMIT will clean up the mess.  This allows RESOLVE to have
 * a "redo-merge" option with a different tool, for example.  It also allows
 * us to keep the trash out of the working directory -- both for neatness
 * and to avoid having baggage if there is also a move/rename for the file.
 * 
 * We create a pathname and store it in the SG_mrg so that subsequent calls
 * on this SG_mrg all return the same value.
 *
 * The first call will try to actually create the directory on disk.
 * 
 */
static void _sg_mrg__ensure_temp_dir(SG_context * pCtx, SG_mrg * pMrg)
{
	const SG_pathname * pPathWorkingDirectoryTop;		// we do not own this

	if (pMrg->pPathOurTempDir)
		return;

	// create <sgtemp>/merge_<session>/

	SG_ERR_CHECK_RETURN(  SG_pendingtree__get_working_directory_top__ref(pCtx, pMrg->pPendingTree, &pPathWorkingDirectoryTop)  );
	SG_ERR_CHECK_RETURN(  SG_workingdir__generate_and_create_temp_dir_for_purpose(pCtx,
																				  pPathWorkingDirectoryTop,
																				  "merge",
																				  &pMrg->pPathOurTempDir)  );
}

/**
 * Create a private temp-dir for the various versions of a
 * single file that was modified in multiple branches and
 * may need to be merged using a mergetool.
 *
 * This directory and the files we eventually create may
 * persist after the MERGE is finished in case that RESOLVE
 * needs to refer to them.
 *
 * In any case, this directory should be deleted by COMMIT
 * or a REVERT --all.
 *
 * We create a pathname and store it in the conflict structure
 * so that all versions of this file in this conflict can use it.
 */
static void _sg_mrg__ensure_temp_dir_for_file_conflict(SG_context * pCtx,
													   SG_mrg * pMrg,
													   SG_mrg_cset_entry_conflict * pMrgCSetEntryConflict)
{
	SG_string * pString = NULL;

	if (pMrgCSetEntryConflict->pPathTempDirForFile)
		return;

	SG_ERR_CHECK(  _sg_mrg__ensure_temp_dir(pCtx, pMrg)  );

	// we give a little thought to the name of the per-file temp-dir.
	// we could just make it the GID and go on, but that is a little
	// hard for the user to navigate (but if we've done our job right
	// they shouldn't ever have to, but let's not get personal.... :)
	// if we just use the entryname, we're probably fine unless they
	// have 2 different files with the same name in different parts
	// of the tree and there was an edit conflict in both sets.
	//
	// so, smash the entryname and the gid together and maybe lop
	// off some digits.
	//
	// create "<sgtemp>/merge_<session>/<ancestor_entryname>~<gid_prefix>/"
	//
	// we use the entryname from the file as it appeared in the
	// ancestor version of the tree so we can avoid any pending
	// rename issues.

	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pString)  );
	SG_ERR_CHECK(  SG_string__sprintf(pCtx, pString, "%s~%9.9s",
									  SG_string__sz(pMrgCSetEntryConflict->pMrgCSetEntry_Ancestor->pStringEntryname),
									  pMrgCSetEntryConflict->pMrgCSetEntry_Ancestor->bufGid_Entry)  );

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx, &pMrgCSetEntryConflict->pPathTempDirForFile,
												   pMrg->pPathOurTempDir,
												   SG_string__sz(pString))  );

	SG_ERR_CHECK(  SG_fsobj__mkdir_recursive__pathname(pCtx, pMrgCSetEntryConflict->pPathTempDirForFile)  );

fail:
	SG_STRING_NULLFREE(pCtx, pString);
}

/**
 * Create a pathname in the per-file temp-dir for one
 * version of the file.
 *
 * As above we use the ancestor version of the entryname
 * to avoid issues with pending renames.
 */
static void _sg_mrg__create_pathname_for_conflict_file(SG_context * pCtx,
													   SG_mrg * pMrg,
													   SG_mrg_cset_entry_conflict * pMrgCSetEntryConflict,
													   const char * pszRole,
													   SG_pathname ** ppPathReturned)
{
	SG_pathname * pPath = NULL;
	SG_string * pString = NULL;

	SG_ERR_CHECK(  _sg_mrg__ensure_temp_dir_for_file_conflict(pCtx, pMrg, pMrgCSetEntryConflict)  );

	// create "<sgtemp>/merge_<session>/<ancestor_entryname>~<gid_prefix>/<role_this>~<ancestor_entryname>"
	//
	// for example:
	//     <sgtemp>/merge_20100914_0/foo.c~g123456789/ANCESTOR~foo.c
	//     <sgtemp>/merge_20100914_0/foo.c~g123456789/BASELINE~foo.c

	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pString)  );
	SG_ERR_CHECK(  SG_string__sprintf(pCtx, pString, "%s~%s",
									  pszRole,
									  SG_string__sz(pMrgCSetEntryConflict->pMrgCSetEntry_Ancestor->pStringEntryname))  );
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx, &pPath,
												   pMrgCSetEntryConflict->pPathTempDirForFile,
												   SG_string__sz(pString))  );

	SG_STRING_NULLFREE(pCtx, pString);
	*ppPathReturned = pPath;
	return;

fail:
	SG_STRING_NULLFREE(pCtx, pString);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
}

//////////////////////////////////////////////////////////////////

/**
 * Export the contents of the given file entry (which reflects a specific version)
 * into a temp file so that an external tool can use it.
 *
 * You own the returned pathname and the file on disk.
 */
static void _sg_mrg__export_to_temp_file(SG_context * pCtx, SG_mrg * pMrg,
										 const char * pszHidBlob,
										 const SG_pathname * pPathTempFile)
{
	SG_file * pFile = NULL;
	SG_repo * pRepo;
	SG_bool bExists;

	SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx,pPathTempFile,&bExists,NULL,NULL)  );
	if (bExists)
	{
#if TRACE_MRG
		SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
								   "Skipping export of [blob %s] to [%s]\n",pszHidBlob,SG_pathname__sz(pPathTempFile))  );
#endif
	}
	else
	{
#if TRACE_MRG
		SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
								   "Exporting          [blob %s] to [%s]\n",pszHidBlob,SG_pathname__sz(pPathTempFile))  );
#endif

		// Ideally, when we create this TEMP file it should be read-only.
		// Afterall, it does represent a historical version of the file and
		// it should only be used as INPUT to whatever merge tool the user
		// has configured.  So it should be read-only.  This might allow
		// a GUI merge tool to show locks/whatever and/or prevent accidental
		// editing of these files.
		//
		// However, this can cause an "Access is Denied" error on Windows
		// when we get ready to delete the contents of the TEMP directory.
		// We'll deal with that there rather than here.

		SG_ERR_CHECK(  SG_file__open__pathname(pCtx,pPathTempFile,SG_FILE_WRONLY|SG_FILE_CREATE_NEW,0400,&pFile)  );
		SG_ERR_CHECK(  SG_pendingtree__get_repo(pCtx, pMrg->pPendingTree, &pRepo)  );
		SG_ERR_CHECK(  SG_repo__fetch_blob_into_file(pCtx,pRepo,pszHidBlob,pFile,NULL)  );
		SG_FILE_NULLCLOSE(pCtx,pFile);
	}

	return;

fail:
	if (pFile)			// only if **WE** created the file, do we try to delete it on error.
	{
		SG_FILE_NULLCLOSE(pCtx,pFile);
		SG_ERR_IGNORE(  SG_fsobj__remove__pathname(pCtx,pPathTempFile)  );
	}
}

//////////////////////////////////////////////////////////////////

END_EXTERN_C;

#endif//H_SG_MRG__PRIVATE_TEMP_DIR_H
