/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
 * sg_parents.c
 *
 *  Created on: Apr 19, 2010
 *      Author: jeremy
 */
#include "sg.h"


void SG_parents__list(SG_context * pCtx,
					  SG_repo * pRepo,
					  SG_rev_spec* pRevSpec,
					  const char* pszFilePath,
					  SG_stringarray ** pStringArrayResults)
{
	char *  pszHid = NULL;
	SG_pendingtree * pPendingTree = NULL;
	SG_stringarray * pTempStringArrayResults = NULL;
	SG_history_result* pHistResult = NULL;
	SG_uint32 countRevSpecs = 0;

	/* zero or one rev spec allowed */
	if (pRevSpec)
	{
		SG_ERR_CHECK(  SG_rev_spec__count(pCtx, pRevSpec, &countRevSpecs)  );
		if (countRevSpecs > 1)
			SG_ERR_THROW2_RETURN(SG_ERR_INVALIDARG, (pCtx, "Only one revision specifier is allowed."));
	}

	//If a file argument is given, the revision in which the file was last changed (before the working
	//directory revision or the argument to --rev if given) is printed
	if (pszFilePath != NULL)
	{
		SG_bool bFound = SG_FALSE;

		//This is basically history -max 1 file.
		SG_ERR_CHECK(  SG_vv_verbs__history__workingFolder(pCtx, pszFilePath, 1, &pszFilePath,
			pRevSpec, NULL,  NULL, SG_FALSE, 1, SG_FALSE, 0, SG_INT64_MAX, 
			&bFound, NULL, &pHistResult, NULL)  );
		
		if (bFound)
		{
			SG_ERR_CHECK(  SG_STRINGARRAY__ALLOC(pCtx, &pTempStringArrayResults, 1));

			while (bFound)
			{
				const char * pszParentHid = NULL;
				SG_ERR_CHECK(  SG_history_result__get_cshid(pCtx, pHistResult, &pszParentHid)  );
				SG_ERR_CHECK(  SG_stringarray__add(pCtx, pTempStringArrayResults, (char*)pszParentHid)  );
				SG_ERR_CHECK(  SG_history_result__next(pCtx, pHistResult, &bFound)  );
			}

			SG_HISTORY_RESULT_NULLFREE(pCtx, pHistResult);
		}
	}
	else
	{
		if (countRevSpecs == 0)
		{
			// They want the baseline of the current working copy.

			const SG_varray * pva_wd_parents;			// we do not own this
			SG_uint32 k, nrParents;

			SG_ERR_CHECK(  SG_pendingtree__alloc_from_cwd(pCtx, &pPendingTree)  );

			SG_ERR_CHECK(  SG_STRINGARRAY__ALLOC(pCtx, &pTempStringArrayResults, 1));
			SG_ERR_CHECK(  SG_pendingtree__get_wd_parents__ref(pCtx, pPendingTree, &pva_wd_parents)  );
			SG_ERR_CHECK(  SG_varray__count(pCtx, pva_wd_parents, &nrParents)  );
			for (k=0; k<nrParents; k++)
			{
				const char * psz_hid_parent_k;

				SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pva_wd_parents, k, &psz_hid_parent_k)  );
				SG_ERR_CHECK(  SG_stringarray__add(pCtx, pTempStringArrayResults, psz_hid_parent_k)  );
			}
		}
		else
		{
			//They want the parents of a specific revision.
			SG_dagnode * pCurrentDagnode = NULL;
			SG_rbtree * prb_parents = NULL;
			SG_rbtree_iterator * pit = NULL;
			const char * pszParentID = NULL;
			SG_bool b = SG_FALSE;

			SG_ERR_CHECK(  SG_rev_spec__get_one__repo(pCtx, pRepo, pRevSpec, &pszHid, NULL)  );

			SG_ERR_CHECK(  SG_repo__fetch_dagnode(pCtx, pRepo, SG_DAGNUM__VERSION_CONTROL, pszHid, &pCurrentDagnode)  );
			SG_ERR_CHECK(  SG_dagnode__get_parents__rbtree_ref(pCtx, pCurrentDagnode, &prb_parents)  );
			if (prb_parents)
			{
				SG_ERR_CHECK(  SG_STRINGARRAY__ALLOC(pCtx, &pTempStringArrayResults, 1));
				SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pit, prb_parents, &b, &pszParentID, NULL)  );
				while (b)
				{
					SG_ERR_CHECK(  SG_stringarray__add(pCtx, pTempStringArrayResults, (char*)pszParentID)  );
					SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pit, &b, &pszParentID, NULL)  );
				}
				SG_RBTREE_ITERATOR_NULLFREE(pCtx, pit);
			}
			SG_DAGNODE_NULLFREE(pCtx, pCurrentDagnode);
		}
	}
	*pStringArrayResults =  pTempStringArrayResults;
fail:
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_NULLFREE(pCtx, pszHid);
	SG_HISTORY_RESULT_NULLFREE(pCtx, pHistResult);

}

