/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg_pendingtree__private_remove_safely2.h
 *
 * @details 
 *
 */

//////////////////////////////////////////////////////////////////

#ifndef H_SG_PENDINGTREE__PRIVATE_REMOVE_SAFELY2_H
#define H_SG_PENDINGTREE__PRIVATE_REMOVE_SAFELY2_H

BEGIN_EXTERN_C;

//////////////////////////////////////////////////////////////////

enum _remove_data_task_flags
{
	_TASK__FS__KEEP_PATHNAME	= 0x0001,
	_TASK__FS__BACKUP_PATHNAME	= 0x0002,
	_TASK__FS__MASK__IN_DIR		= 0x000f,

	_TASK__FS__REMOVE_PATHNAME	= 0x0010,
	_TASK__FS__RMDIR_PATHNAME	= 0x0020,
	_TASK__FS__TO_GRAVEYARD		= 0x0040,
	_TASK__FS__DID_NOT_EXIST	= 0x0080,
	_TASK__FS__MASK__NOT_IN_DIR	= 0x00f0,
	
	_TASK__FS__MASK				= _TASK__FS__MASK__IN_DIR | _TASK__FS__MASK__NOT_IN_DIR,

	_TASK__FLUSH_CACHE			= 0x0100,	// remove timestamp cache entry

	_TASK__PTN__MARK_DELETED	= 0x1000,	// set sg_PTNODE_SAVED_FLAG_DELETED
	_TASK__PTN__MARK_UNADDED	= 0x2000,	// set sg_PTNODE_SAVED_FLAG_DELETED and bDeletedAfterAdded
	_TASK__PTN__DELETE_PTNODE	= 0x4000,	// remove ptnode (item will appear as FOUND afterwards)
	_TASK__PTN__MASK			= 0xf000,
};

typedef enum _remove_data_task_flags _remove_data_task_flags;

struct _remove_data_item
{
	struct _remove_data *	pData;				// back ptr, we do not own this
	SG_pathname *			pPath;
	SG_pathname *			pPathDir;
	SG_string *				pStringEntryName;
	SG_string *				pStringRepoPath;
	struct sg_ptnode *		ptnDir;				// we do not own this
	struct sg_ptnode *		ptn;				// we do not own this
	SG_fsobj_stat			st;
	SG_bool					bExistsOnDisk;
	SG_bool					bExistedInBaselineVersion;
	SG_bool					bActive;
	SG_bool					bMergeCreated;		// aka disposable
	_remove_data_task_flags	flags;
};

struct _remove_data_task
{
	SG_pathname *			pPath;
	struct sg_ptnode *		ptn;				// we do not own this
	_remove_data_task_flags	flags;
};
	
struct _remove_data
{
	SG_pendingtree *		pPendingTree;
	SG_repo *				pRepo;
	SG_file_spec *			pFilespec_NoIgnores;
	SG_file_spec *			pFilespec_WithIgnores;
	SG_rbtree *				prbAllItems;		// map[<repo-path> ==> <item>]    we own the items
	SG_rbtree *				prbTopLevelItems;	// map[<repo-path> ==> <top-level-item>]	we do not own the items
	SG_rbtree *				prbTaskList;		// map[<repo-path> ==> <task>]
	SG_bool					bTest;
	SG_bool					bForce;
	SG_bool					bNoBackup;
};

//////////////////////////////////////////////////////////////////

static SG_rbtree_foreach_callback sg_pendingtree__remove_safely__inspect_item;

//////////////////////////////////////////////////////////////////

static void sg_pendingtree__remove_safely__disallow_moved_or_renamed(SG_context * pCtx,
																	 const struct sg_ptnode * ptn,
																	 const char * pszRepoPath)
{
	// For historical reasons we disallow REMOVES of MOVED or RENAMED
	// items.  I'd like to eliminate this restriction, but not today.

	if (ptn->pszCurrentName)
	{
		// we have either the name of an ADDED item -or-
		// the NEW name of a RENAMED item.

		if (ptn->pBaselineEntry)
		{
			// the item was present in the baseline version.
			// the simple presence of both fields *should*
			// be enough to infer that we have a renamed item,
			// but some early code didn't always reduce the
			// ptnode to canonical form, so we'll verify it.

			const char * pszBaselineName = NULL;

			SG_ERR_CHECK(  SG_treenode_entry__get_entry_name(pCtx, ptn->pBaselineEntry, &pszBaselineName)  );
			if (strcmp(ptn->pszCurrentName, pszBaselineName) != 0)
			{
#if TRACE_REMOVE
				SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
										   "RemoveSafely: unsafe -- item renamed: %s\n",
										   pszRepoPath)  );
#endif
				SG_ERR_THROW2(  SG_ERR_CANNOT_REMOVE_SAFELY,
								(pCtx, "Entry was renamed: '%s'",
								 pszRepoPath)  );
			}
		}
	}
	
	if (ptn->pszgidMovedFrom)
	{
#if TRACE_REMOVE
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "RemoveSafely: unsafe -- item moved: %s\n",
								   pszRepoPath)  );
#endif
		SG_ERR_THROW2(  SG_ERR_CANNOT_REMOVE_SAFELY,
						(pCtx, "Entry was moved: '%s'",
						 pszRepoPath)  );
	}

fail:
	;
}

static void sg_pendingtree__remove_safely__disallow_changed_type(SG_context * pCtx,
																 const struct sg_ptnode * ptn,
																 const SG_fsobj_stat * pStat,
																 const char * pszRepoPath)
{
	// Confirm that the thing we observed on the disk
	// is of the same type as what we have in the ptnode.
	// 
	// For example, I'm assuming that if you have a
	// directory 'foo' under version control and /bin/rmdir
	// it and create a file named 'foo' in its place (both
	// without telling us), that if you then do a 'vv remove foo',
	// we should complain.

	if (   ((pStat->type == SG_FSOBJ_TYPE__DIRECTORY) && (ptn->type == SG_TREENODEENTRY_TYPE_DIRECTORY))
		|| ((pStat->type == SG_FSOBJ_TYPE__DIRECTORY) && (ptn->type == SG_TREENODEENTRY_TYPE_SUBMODULE))
		|| ((pStat->type == SG_FSOBJ_TYPE__REGULAR)   && (ptn->type == SG_TREENODEENTRY_TYPE_REGULAR_FILE))
		|| ((pStat->type == SG_FSOBJ_TYPE__SYMLINK)   && (ptn->type == SG_TREENODEENTRY_TYPE_SYMLINK)))
	{
	}
	else
	{
#if TRACE_REMOVE
				SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
										   "RemoveSafely: unsafe -- type changed: %s\n",
										   pszRepoPath)  );
#endif
				SG_ERR_THROW2(  SG_ERR_CANNOT_REMOVE_SAFELY,
								(pCtx, "The type of the entry changed: '%s'",
								 pszRepoPath)  );
	}

fail:
	;
}

static void sg_pendingtree__remove_safely__disallow_certain_inactive(SG_context * pCtx,
																	 const struct sg_ptnode * ptn,
																	 SG_bool bExistsOnDisk,
																	 const char * pszRepoPath,
																	 SG_bool * pbNothingToDo)
{
	SG_bool bNothingToDo = SG_FALSE;

	if (ptn->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY)
	{
		// if we have the phantom of a moved item, then this is
		// not the right ptnode.  (or rather the path they gave
		// us to this ptnode was old.)

#if TRACE_REMOVE
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "RemoveSafely: item mapped to moved-away node: %s\n",
								   pszRepoPath)  );
#endif
		SG_ERR_THROW2(  SG_ERR_NOT_FOUND,
						(pCtx, "'%s'", pszRepoPath)  );
	}

	if (ptn->saved_flags & sg_PTNODE_SAVED_FLAG_DELETED)
	{
		// this node was deleted by a previous 'vv remove' command
		// (but has not yet been committed).  so, in theory, the
		// item should have been deleted from the WD too.
		//
		// also, because we look for active nodes before we look
		// for inactive nodes, if they added a new item with the
		// same name, we would have found the active node rather
		// than this one.

		if (bExistsOnDisk)
		{
			// so if we get here with this inactive node, they
			// created an item on disk, but didn't ADD anything,
			// so we don't really have any claim on this FOUND item.

#if TRACE_REMOVE
			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
									   "RemoveSafely: found item with same name as removed item: %s\n",
									   pszRepoPath)  );
#endif
			SG_ERR_THROW2(  SG_ERR_CANNOT_REMOVE_SAFELY,
							(pCtx, "Unknown entry found: '%s'",
							 pszRepoPath)  );
		}
		else
		{
			// they said 'vv remove foo' twice.  nothing really
			// to do here.
			//
			// TODO 2011/05/04 Should we silently let this pass
			// TODO            or should we complain ?
			//
			// For now, let it slide.

#if TRACE_REMOVE
			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
									   "RemoveSafely: silently allowing remove after remove: %s\n",
									   pszRepoPath)  );
#endif
			bNothingToDo = SG_TRUE;
		}
	}

	*pbNothingToDo = bNothingToDo;

fail:
	;
}

//////////////////////////////////////////////////////////////////

static void sg_pendingtree__remove_safely__free_task(SG_context * pCtx,
													 struct _remove_data_task * pTask)
{
	if (!pTask)
		return;

	SG_PATHNAME_NULLFREE(pCtx, pTask->pPath);
	SG_NULLFREE(pCtx, pTask);
}

static void sg_pendingtree__remove_safely__append_task(SG_context * pCtx,
													   struct _remove_data * pData,
													   const SG_pathname * pPath,
													   const SG_string * pStringRepoPath,
													   struct sg_ptnode * ptn,
													   _remove_data_task_flags flags)
{
	struct _remove_data_task * pTask = NULL;
	const struct _remove_data_task * pTaskFound = NULL;
	SG_bool bFound;

	SG_ERR_CHECK(  SG_rbtree__find(pCtx,
								   pData->prbTaskList,
								   SG_string__sz(pStringRepoPath),
								   &bFound,
								   (void **)&pTaskFound)  );
	if (bFound)
	{
		// This can happen if they say 'vv remove foo/bar foo'
		// where we added a task for bar and then while diving
		// on 'foo', we stumbled upon it again.  in theory, both
		// passes will compute the same set of flags.

#if TRACE_REMOVE
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "RemoveTask: duplicate [flags 0x%x 0x%x]: %s\n",
								   pTaskFound->flags,
								   flags,
								   SG_pathname__sz(pPath))  );
#endif

		if (pTaskFound->flags != flags)
		{
			SG_ERR_THROW2(  SG_ERR_CANNOT_REMOVE_SAFELY,
							(pCtx, "Ambiguous flags [0x%x 0x%x]: %s",
							 pTaskFound->flags,
							 flags,
							 SG_string__sz(pStringRepoPath))  );
		}
		else
		{
			// no need to duplicate the task
		}
	}
	else
	{
		SG_ERR_CHECK(  SG_alloc1(pCtx, pTask)  );
		SG_ERR_CHECK(  SG_PATHNAME__ALLOC__COPY(pCtx, &pTask->pPath, pPath)  );
		pTask->ptn = ptn;
		pTask->flags = flags;

		SG_ERR_CHECK(  SG_rbtree__add__with_assoc(pCtx,
												  pData->prbTaskList,
												  SG_string__sz(pStringRepoPath),
												  pTask)  );
		pTask = NULL;			// rbtree owns it now
	}

fail:
	sg_pendingtree__remove_safely__free_task(pCtx, pTask);
}

static SG_rbtree_foreach_callback sg_pendingtree__remove_safely__apply_task;
static void sg_pendingtree__remove_safely__apply_task(SG_context * pCtx,
													  const char * pszKey_RepoPath,
													  void * pVoid_Task,
													  void * pVoid_Data)
{
	struct _remove_data * pData = (struct _remove_data *)pVoid_Data;
	struct _remove_data_task * pTask = (struct _remove_data_task *)pVoid_Task;
	SG_pathname * pPathBackup = NULL;
	struct sg_ptnode * ptn_owned = NULL;

	// TODO 2011/05/06 Revisit the logging messages here.

#if TRACE_REMOVE
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "Remove: applying task [0x%x]: '%s'...\n",
							   pTask->flags,
							   pszKey_RepoPath)  );
#else
	SG_UNUSED( pszKey_RepoPath );
#endif

	switch (pTask->flags & _TASK__FS__MASK)
	{
	case _TASK__FS__KEEP_PATHNAME:
		break;

	case _TASK__FS__BACKUP_PATHNAME:
		SG_ERR_CHECK(  _pt_log__remove_item(pCtx,
											pTask->ptn->pPendingTree,
											pTask->ptn->pszgid,
											SG_pathname__sz(pTask->pPath),
											"moved_to_backup")  );
		if (!pData->bTest)
		{
			SG_ERR_CHECK(  SG_workingdir__calc_backup_name(pCtx, pTask->pPath, &pPathBackup)  );
			SG_ERR_CHECK(  SG_fsobj__move__pathname_pathname(pCtx, pTask->pPath, pPathBackup)  );
		}
		break;

	case _TASK__FS__REMOVE_PATHNAME:
		SG_ERR_CHECK(  _pt_log__remove_item(pCtx,
											pTask->ptn->pPendingTree,
											pTask->ptn->pszgid,
											SG_pathname__sz(pTask->pPath),
											"removed")  );
		if (!pData->bTest)
			SG_ERR_CHECK(  SG_fsobj__remove__pathname(pCtx, pTask->pPath)  );
		break;

	case _TASK__FS__RMDIR_PATHNAME:
		SG_ERR_CHECK(  _pt_log__rmdir(pCtx,
									  pTask->ptn->pPendingTree,
									  pTask->ptn->pszgid,
									  SG_pathname__sz(pTask->pPath),
									  "removed")  );
		if (!pData->bTest)
			SG_ERR_CHECK(  SG_fsobj__rmdir__pathname(pCtx, pTask->pPath)  );
		break;

	case _TASK__FS__TO_GRAVEYARD:
		SG_ERR_CHECK(  _pt_log__delete_submodule(pCtx,
												 pTask->ptn->pPendingTree,
												 pTask->ptn->pszgid,
												 SG_pathname__sz(pTask->pPath),
												 "moved_to_graveyard")  );
		if (!pData->bTest)
			SG_ERR_CHECK(  SG_workingdir__move_submodule_to_graveyard(pCtx,
																	  pTask->ptn->pPendingTree->pPathWorkingDirectoryTop,
																	  pTask->pPath,
																	  pTask->ptn->pszgid,
																	  NULL)  );
		break;

	case _TASK__FS__DID_NOT_EXIST:
		break;

	default:
		SG_ASSERT(  (0)  );
	}

	if (!pData->bTest)
	{
		if (pTask->flags & _TASK__FLUSH_CACHE)
		{
			SG_timestamp_cache * pTSC = NULL;		// we do not own this

			SG_ERR_CHECK(  SG_pendingtree__get_timestamp_cache(pCtx, pTask->ptn->pPendingTree, &pTSC)  );
			SG_ERR_CHECK(  SG_timestamp_cache__remove(pCtx, pTSC, pTask->ptn->pszgid)  );
		}
	
		switch (pTask->flags & _TASK__PTN__MASK)
		{
		case _TASK__PTN__MARK_DELETED:
			pTask->ptn->saved_flags |= sg_PTNODE_SAVED_FLAG_DELETED;
			pTask->ptn->bDeletedAfterAdded = SG_FALSE;
			pTask->ptn->pszCurrentHid = NULL;	// so it won't look modified after it is deleted
			// TODO 2011/04/28 should we also zero/null attrbits/xattrs ?
			break;

		case _TASK__PTN__MARK_UNADDED:
			pTask->ptn->saved_flags |= sg_PTNODE_SAVED_FLAG_DELETED;
			pTask->ptn->bDeletedAfterAdded = SG_TRUE;
			pTask->ptn->pszCurrentHid = NULL;	// so it won't look modified after it is deleted
			// TODO 2011/04/28 should we also zero/null attrbits/xattrs ?
			break;

		case _TASK__PTN__DELETE_PTNODE:
			SG_ERR_CHECK(  SG_rbtree__remove__with_assoc(pCtx,
														 pTask->ptn->pCurrentParent->prbItems,
														 pTask->ptn->pszgid,
														 (void **)&ptn_owned)  );
			SG_ASSERT(  (ptn_owned == pTask->ptn)  );
			break;

		default:
			SG_ASSERT(  (0)  );
		}
	}

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathBackup);
	SG_PTNODE_NULLFREE(pCtx, ptn_owned);
}

//////////////////////////////////////////////////////////////////

static void sg_pendingtree__remove_safely__free_item(SG_context * pCtx,
													 struct _remove_data_item * pItem)
{
	if (!pItem)
		return;

	SG_PATHNAME_NULLFREE(pCtx, pItem->pPath);
	SG_PATHNAME_NULLFREE(pCtx, pItem->pPathDir);
	SG_STRING_NULLFREE(pCtx, pItem->pStringEntryName);
	SG_STRING_NULLFREE(pCtx, pItem->pStringRepoPath);

	SG_NULLFREE(pCtx, pItem);
}

static void sg_pendingtree__remove_safely__alloc_item(SG_context * pCtx,
													  struct _remove_data * pData,
													  const char * pszItem,
													  struct sg_ptnode * ptnDir,
													  struct sg_ptnode * ptnItem,
													  struct _remove_data_item ** ppItem,
													  SG_bool * pbDuplicate)
{
	struct _remove_data_item * pItem = NULL;
	struct _remove_data_item * pItemFound = NULL;
	SG_pathname * pPath = NULL;
	SG_string * pStringRepoPath = NULL;
	SG_bool bFound;

	SG_NONEMPTYCHECK( pszItem );
	// ptnDir is the ptnode of the parent directory.
	// ptnItem is the ptnode of the item itself.
	// both ptn's are optional.

	// Compute the absolute path and repo-path of the item.

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPath, pszItem)  );
	SG_ERR_CHECK(  SG_workingdir__wdpath_to_repopath(pCtx,
													 pData->pPendingTree->pPathWorkingDirectoryTop,
													 pPath,
													 SG_FALSE,
													 &pStringRepoPath)  );

	// if we have already seen this repo-path, just return the existing item.

	SG_ERR_CHECK(  SG_rbtree__find(pCtx,
								   pData->prbAllItems,
								   SG_string__sz(pStringRepoPath),
								   &bFound,
								   (void **)&pItemFound)  );
	if (bFound)
	{
		SG_STRING_NULLFREE(pCtx, pStringRepoPath);
		SG_PATHNAME_NULLFREE(pCtx, pPath);

		*pbDuplicate = SG_TRUE;
		*ppItem = pItemFound;
		return;
	}
	
	SG_ERR_CHECK(  SG_alloc1(pCtx, pItem)  );
	pItem->pData = pData;
	pItem->pPath = pPath; pPath = NULL;
	pItem->pStringRepoPath = pStringRepoPath; pStringRepoPath = NULL;
	
	SG_ERR_CHECK(  SG_pathname__get_last(pCtx,
										 pItem->pPath,
										 &pItem->pStringEntryName)  );

	// See if it exists on disk.  We want to be able to
	// handle a 'vv remove' *after* they did a /bin/rm,
	// for example.

	SG_ERR_CHECK(  SG_fsobj__exists_stat__pathname(pCtx,
												   pItem->pPath,
												   &pItem->bExistsOnDisk,
												   &pItem->st)  );

	// We want to find the ptnode for the requested item, but because
	// of the quirks in the __find_ api, we need to do this in parts.
	// Compute the pathname of the parent directory.

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__COPY(pCtx, &pItem->pPathDir, pItem->pPath)  );
	SG_ERR_CHECK(  SG_pathname__remove_last(pCtx, pItem->pPathDir)  );

	if (ptnDir)
	{
		pItem->ptnDir = ptnDir;
	}
	else
	{
		// Find the ACTIVE ptnode of the PARENT DIRECTORY, if it exists.
		// (If one of the parent directories are INACTIVE (already deleted or a
		// phantom), we just complain.)

		sg_pendingtree__local_path_to_ptnode(pCtx,
											 pData->pPendingTree,
											 pItem->pPathDir,
											 SG_FALSE,
											 &pItem->ptnDir);
		if (SG_context__has_err(pCtx))
		{
			SG_context__err_reset(pCtx);
			SG_ERR_THROW2(  SG_ERR_DIR_PATH_NOT_FOUND,
							(pCtx, "'%s'", SG_string__sz(pItem->pStringRepoPath))  );
		}
	}

	if (ptnItem)
	{
		pItem->ptn = ptnItem;
		pItem->bActive = (   ((ptnItem->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY) == 0)
		                  && ((ptnItem->saved_flags & sg_PTNODE_SAVED_FLAG_DELETED)    == 0));
	}
	else
	{
		// Third, find the ptnode of this item *within* the parent directory
		// using the item's EntryName.
		// 
		// We allow it to be either ACTIVE or INACTIVE.

		SG_ERR_CHECK(  sg_ptnode__find_entry_by_name_not_recursive__active(pCtx,
																		   pItem->ptnDir,
																		   SG_string__sz(pItem->pStringEntryName),
																		   &pItem->ptn)  );
		if (pItem->ptn)
		{
			pItem->bActive = SG_TRUE;
		}
		else
		{
			// Fourth, look for an INACTIVE node.

			SG_ERR_CHECK(  sg_ptnode__find_entry_by_name_not_recursive__inactive(pCtx,
																				 pItem->ptnDir,
																				 SG_string__sz(pItem->pStringEntryName),
																				 &pItem->ptn)  );
			if (!pItem->ptn)
				SG_ERR_THROW2(  SG_ERR_ITEM_NOT_UNDER_VERSION_CONTROL,
								(pCtx, "'%s'", SG_string__sz(pItem->pStringRepoPath))  );

			pItem->bActive = SG_FALSE;
		}
	}

	// See if this item existed in the baseline.  If not, we
	// have the REMOVE of an ADD.  (Probably an "oops" from
	// the user.)

	pItem->bExistedInBaselineVersion = (pItem->ptn->pBaselineEntry != NULL);

	// See if item was created by MERGE (an ADD from another
	// branch or the override of a DELETE in baseline).

	pItem->bMergeCreated = (pItem->ptn->bDisposableDir || pItem->ptn->pszHidDisposable);

	SG_ERR_CHECK(  SG_rbtree__add__with_assoc(pCtx,
											  pData->prbAllItems,
											  SG_string__sz(pItem->pStringRepoPath),
											  pItem)  );	// rbtree now owns it

	*pbDuplicate = SG_FALSE;
	*ppItem = pItem;
	return;

fail:
	sg_pendingtree__remove_safely__free_item(pCtx, pItem);
	SG_STRING_NULLFREE(pCtx, pStringRepoPath);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
}

/**
 * Pre-qualify a top-level item, gather info for it, and add it
 * to the list of things to do.
 *
 */
static void sg_pendingtree__remove_safely__alloc_top_level_item(SG_context * pCtx,
																struct _remove_data * pData,
																const char * pszItem)
{
	SG_pathname * pPath = NULL;
	SG_string * pStringRepoPath = NULL;
	struct _remove_data_item * pItem = NULL;	// we don't own this
	SG_file_spec__match_result eval;
	SG_bool bDuplicate = SG_FALSE;

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPath, pszItem)  );
	SG_ERR_CHECK(  SG_workingdir__wdpath_to_repopath(pCtx,
													 pData->pPendingTree->pPathWorkingDirectoryTop,
													 pPath,
													 SG_FALSE,
													 &pStringRepoPath)  );

	// pre-qualify the top-level item.  this is more like arg
	// checking to filter out typos and wildcard expansion from
	// the shell.

	sg_pendingtree__remove_safely__alloc_item(pCtx, pData, pszItem,
											  NULL, NULL,
											  &pItem, &bDuplicate);
	if (SG_context__has_err(pCtx))
	{
		if (!SG_context__err_equals(pCtx, SG_ERR_ITEM_NOT_UNDER_VERSION_CONTROL))
			SG_ERR_RETHROW;

		SG_context__err_reset(pCtx);

		// the entry is not under version control. see if it is ignorable.
		// for example, if they say 'vv remove foo*' and there just
		// happens to be an uncontrolled/ignorable "foo~" in the directory,
		// the shell's wildcard expansion will give it to us here.

		SG_ERR_CHECK(  SG_file_spec__should_include(pCtx,
													pData->pFilespec_WithIgnores,
													SG_string__sz(pStringRepoPath),
													SG_FILE_SPEC__CONVERT_REPO_PATH
													| SG_FILE_SPEC__MATCH_TRAILING_SLASH,
													&eval)  );
		switch (eval)
		{
		default:		// quiets compiler
			SG_ASSERT(  (0)  );
			break;
			
		case SG_FILE_SPEC__RESULT__EXCLUDED:
		case SG_FILE_SPEC__RESULT__IGNORED:
		case SG_FILE_SPEC__RESULT__MAYBE:
			goto bottom;

		case SG_FILE_SPEC__RESULT__INCLUDED:
		case SG_FILE_SPEC__RESULT__IMPLIED:
			SG_ERR_THROW2(  SG_ERR_ITEM_NOT_UNDER_VERSION_CONTROL,
							(pCtx, "'%s'", SG_string__sz(pStringRepoPath))  );
			break;
		}
	}
	else if (bDuplicate)
	{
		goto bottom;
	}
	else
	{
		// See if the top-level item should be EXCLUDED.
		// If they say something like "vv remove --exclude='*.h' *"
		// and there happends to be .h files in the current
		// directory, we should exclude them.  (right?)
		//
		// Note that for items that ARE under version control,
		// we DO NOT use the IGNORES.

		SG_ERR_CHECK(  SG_file_spec__should_include(pCtx,
													pData->pFilespec_NoIgnores,
													SG_string__sz(pStringRepoPath),
													SG_FILE_SPEC__CONVERT_REPO_PATH
													| SG_FILE_SPEC__MATCH_TRAILING_SLASH,
													&eval)  );
		switch (eval)
		{
		default:		// queits compiler
			SG_ASSERT(  (0)  );

		case SG_FILE_SPEC__RESULT__EXCLUDED:
		case SG_FILE_SPEC__RESULT__IGNORED:
		case SG_FILE_SPEC__RESULT__MAYBE:
			goto bottom;

		case SG_FILE_SPEC__RESULT__INCLUDED:
		case SG_FILE_SPEC__RESULT__IMPLIED:
			// we want to seriously consider this one. add it to
			// the main queue for inspection.
			SG_ERR_CHECK(  SG_rbtree__add__with_assoc(pCtx,
													  pData->prbTopLevelItems,
													  SG_string__sz(pItem->pStringRepoPath),
													  pItem)  );
			break;
		}
	}

bottom:

#if TRACE_REMOVE
	if (bSkipIt)
	{
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "Remove: skipping: [%s][eval 0x%x][bDup %d]\n",
								   SG_string__sz(pStringRepoPath),
								   eval, bDuplicate)  );
	}
	else
	{
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "Remove: prequalified: [%s]\n",
								   SG_string__sz(pStringRepoPath))  );
	}
#endif

fail:
	SG_STRING_NULLFREE(pCtx, pStringRepoPath);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
}

//////////////////////////////////////////////////////////////////

/**
 * Inspect the existence and/or contents of a file or symlink.
 * (Probably named on the command line.)  Verify that it is
 * can be safely deleted.
 */
static void sg_pendingtree__remove_safely__inspect_file_or_symlink(SG_context * pCtx,
																   struct _remove_data_item * pItem)
{
	_remove_data_task_flags flags = 0;

	SG_ASSERT(  (pItem->bActive)  );	// previously deleted ?
	
	if (pItem->bExistsOnDisk)
	{
		// Use the SCAN_DIR code to make sure that ptn->pszCurrentHid is valid
		// and reflects the current content of the file on disk.

		if (pItem->ptn->type == SG_TREENODEENTRY_TYPE_REGULAR_FILE)
		{
			SG_ERR_CHECK(  _pt_sd__update_ptnode__file_content(pCtx,
															   pItem->ptn->pPendingTree,
															   pItem->ptn,
															   pItem->pPath,
															   NULL,
															   NULL)  );
		}
		else
		{
			SG_ERR_CHECK(  _pt_sd__update_ptnode__symlink_content(pCtx,
																  pItem->ptn->pPendingTree,
																  pItem->ptn,
																  pItem->pPath)  );
		}
	}

	if (pItem->bExistedInBaselineVersion)
	{
		// Since it existed in the baseline, we can use a normal _MARK_DELETED
		// flag causing the ptnode to be kept with deleted status (so 'vv status'
		// can say it was removed).

		if (pItem->bExistsOnDisk)
		{
			if (pItem->ptn->pszCurrentHid == NULL)
			{
				// file content is clean
				flags = _TASK__FS__REMOVE_PATHNAME | _TASK__PTN__MARK_DELETED;
			}
			else if (pItem->bMergeCreated && (strcmp(pItem->ptn->pszCurrentHid,
													 pItem->ptn->pszHidDisposable) == 0))
			{
				// MERGE altered the content and file is unchanged from that
				flags = _TASK__FS__REMOVE_PATHNAME | _TASK__PTN__MARK_DELETED;
			}
			else if (pItem->pData->bForce)
			{
				// file is dirty, but they said --force
				if (pItem->pData->bNoBackup)
					flags = _TASK__FS__REMOVE_PATHNAME | _TASK__PTN__MARK_DELETED;
				else
					flags = _TASK__FS__BACKUP_PATHNAME | _TASK__PTN__MARK_DELETED;
			}
			else
			{
#if TRACE_REMOVE
				SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
										   "RemoveSafely: unsafe file: %s\n", SG_pathname__sz(pItem->pPath))  );
#endif
				SG_ERR_THROW2(  SG_ERR_CANNOT_REMOVE_SAFELY,
								(pCtx, "File has changes: '%s'", SG_string__sz(pItem->pStringRepoPath))  );
			}
		}
		else
		{
			// they did a /bin/rm before they told us about it.
			// silently do the right thing and mark it deleted
			// without making them recreate the file and redo
			// the 'vv remove'.

			flags = _TASK__FS__DID_NOT_EXIST | _TASK__PTN__MARK_DELETED;
		}
	}
	else	// we have a REMOVE of an ADDED item.
	{
		if (pItem->bMergeCreated)
		{
			// MERGE ADDED it.  it was either created in the
			// other branch or MERGE had to override a DELETE-vs-*
			// conflict where it was DELETED in baseline.
			//
			// It's tempting to just delete the ptnode (rather than
			// giving it deleted status), but 'vv status' and 'vv resolve'
			// need it to report on conflicts.  So we have to have a
			// special unadded status (so that treediff doesn't freak out).

			if (pItem->bExistsOnDisk)
			{
				if (strcmp(pItem->ptn->pszCurrentHid,
						   pItem->ptn->pszHidDisposable) == 0)
				{
					flags = _TASK__FS__REMOVE_PATHNAME | _TASK__PTN__MARK_UNADDED;
				}
				else if (pItem->pData->bForce)
				{
					if (pItem->pData->bNoBackup)
						flags = _TASK__FS__REMOVE_PATHNAME | _TASK__PTN__MARK_UNADDED;
					else
						flags = _TASK__FS__BACKUP_PATHNAME | _TASK__PTN__MARK_UNADDED;
				}
				else
				{
#if TRACE_REMOVE
					SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
											   "RemoveSafely: unsafe file: %s\n", SG_pathname__sz(pItem->pPath))  );
#endif
					SG_ERR_THROW2(  SG_ERR_CANNOT_REMOVE_SAFELY,
									(pCtx, "File has changes: '%s'", SG_string__sz(pItem->pStringRepoPath))  );
				}
			}
			else
			{
				flags = _TASK__FS__DID_NOT_EXIST | _TASK__PTN__MARK_UNADDED;
			}
		}
		else
		{
			if (pItem->bExistsOnDisk)
			{
				// User ADDED it and wants to UNDO that.  Granted, a REVERT might
				// be a better thing for them to type, but that's a bit obscure.
				// So we allow it here and effectively do an UNDO -- delete the ptnode
				// and pretend it never happened.
				//
				// Because the actual file WAS NEVER ACTUALLY UNDER VERSION CONTROL
				// (it was never COMMITTED), we DO NOT delete the file.  So it should
				// appear as a FOUND item after we are done.
				//
				// And because it wasn't under version control, we should be able
				// to actually delete the ptnode (or convert it into a FOUND) rather
				// than just marking it with deleted status.
			
				flags = _TASK__FS__KEEP_PATHNAME | _TASK__PTN__DELETE_PTNODE;
			}
			else
			{
				flags = _TASK__FS__DID_NOT_EXIST | _TASK__PTN__DELETE_PTNODE;
			}
		}
	}

	if (pItem->ptn->type == SG_TREENODEENTRY_TYPE_REGULAR_FILE)
		flags |= _TASK__FLUSH_CACHE;

	pItem->flags = flags;

#if TRACE_REMOVE
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "RemoveSafely: flags 0x%x for %s\n",
							   pItem->flags,
							   SG_string__sz(pItem->pStringRepoPath))  );
#endif

	SG_ERR_CHECK(  sg_pendingtree__remove_safely__append_task(pCtx,
															  pItem->pData,
															  pItem->pPath,
															  pItem->pStringRepoPath,
															  pItem->ptn,
															  pItem->flags)  );

fail:
	;
}

//////////////////////////////////////////////////////////////////

static SG_dir_foreach_callback cb_readdir;

static void cb_readdir(SG_context * pCtx,
					   const SG_string * pStringEntryName,
					   SG_fsobj_stat * pfsStat_Unused,
					   void * pVoid_ItemParent)
{
	struct _remove_data_item * pItemParent = (struct _remove_data_item *)pVoid_ItemParent;
	struct _remove_data_item * pItemChild = NULL;	// we do not own this
	SG_pathname * pPath = NULL;
	SG_file_spec__match_result eval;
	SG_bool bDuplicate = SG_FALSE;

	SG_UNUSED( pfsStat_Unused );

	SG_ASSERT(  (pItemParent->bExistsOnDisk)  );

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx, &pPath,
												   pItemParent->pPath,
												   SG_string__sz(pStringEntryName))  );

	sg_pendingtree__remove_safely__alloc_item(pCtx,
											  pItemParent->pData,
											  SG_pathname__sz(pPath),
											  pItemParent->ptn,
											  NULL,
											  &pItemChild,
											  &bDuplicate);
	if (SG_context__has_err(pCtx))
	{
		if (SG_context__err_equals(pCtx, SG_ERR_ITEM_NOT_UNDER_VERSION_CONTROL))
		{
			pItemChild = NULL;
			SG_context__err_reset(pCtx);
		}
		else
		{
			SG_ERR_RETHROW;
		}
	}
	else if (!pItemChild->bActive)
	{
		// the thing we found in the directory on disk has
		// the same name a something that has already been
		// deleted.
		pItemChild = NULL;
	}
	
	if (!pItemChild)
	{
		// this child is not under version control.
		// we don't know anything about it.

		if (pItemParent->flags & _TASK__FS__MASK__NOT_IN_DIR)
		{
			// the child is wants to remain, but we are planning
			// to delete the parent directory.

			SG_ERR_THROW2(  SG_ERR_CANNOT_REMOVE_SAFELY,
							(pCtx, "Uncontrolled entry '%s' prevents deletion of '%s'.",
							 SG_string__sz(pStringEntryName),
							 SG_string__sz(pItemParent->pStringRepoPath))  );
		}
		else
		{
			SG_ASSERT(  (pItemParent->flags & _TASK__FS__MASK__IN_DIR)  );
			
			// parent directory is not going to be deleted (probably
			// an UNDO-style REMOVE of an ADD), so we can leave this
			// entry on the disk.

#if TRACE_REMOVE
			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
									   "Ignoring uncontrolled entry '%s' in '%s'.\n",
									   SG_string__sz(pStringEntryName),
									   SG_string__sz(pItemParent->pStringRepoPath))  );
#endif
		}
	}
	else
	{
		SG_ERR_CHECK(  SG_file_spec__should_include(pCtx,
													pItemParent->pData->pFilespec_NoIgnores,
													SG_string__sz(pItemChild->pStringRepoPath),
													SG_FILE_SPEC__CONVERT_REPO_PATH
													| SG_FILE_SPEC__MATCH_TRAILING_SLASH,
													&eval)  );
		switch (eval)
		{
		default:		// quiets compiler
			SG_ASSERT(  (0)  );

		case SG_FILE_SPEC__RESULT__EXCLUDED:
		case SG_FILE_SPEC__RESULT__IGNORED:
		case SG_FILE_SPEC__RESULT__MAYBE:
			if (pItemParent->flags & _TASK__FS__MASK__NOT_IN_DIR)
			{
				SG_ERR_THROW2(  SG_ERR_CANNOT_REMOVE_SAFELY,
								(pCtx, "Excluded/Ignored entry '%s' prevents deletion of '%s'.",
								 SG_string__sz(pStringEntryName),
								 SG_string__sz(pItemParent->pStringRepoPath))  );
			}
			else
			{
				SG_ASSERT(  (pItemParent->flags & _TASK__FS__MASK__IN_DIR)  );
#if TRACE_REMOVE
				SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
										   "Ignoring excluded/ignored entry '%s' in '%s'.\n",
										   SG_string__sz(pStringEntryName),
										   SG_string__sz(pItemParent->pStringRepoPath))  );
#endif
			}
			break;

		case SG_FILE_SPEC__RESULT__INCLUDED:
		case SG_FILE_SPEC__RESULT__IMPLIED:
			break;
		}
	
		SG_ERR_CHECK(  sg_pendingtree__remove_safely__inspect_item(pCtx,
																   SG_string__sz(pItemChild->pStringRepoPath),
																   pItemChild,
																   pItemChild->pData)  );
		if (pItemParent->flags & _TASK__FS__MASK__NOT_IN_DIR)
		{
			// the parent directory is scheduled for deletion.

			if (pItemChild->flags & _TASK__FS__MASK__NOT_IN_DIR)
			{
				// this child is either scheduled for deletion or doesn't exist.
			}
			else
			{
				// this child currently exists on disk and will remain afterwards
				// (either as an undo-style remove of an add -or- as a backup file).

				SG_ERR_THROW2(  SG_ERR_CANNOT_REMOVE_SAFELY,
								(pCtx, "Entry '%s' prevents deletion of '%s'.",
								 SG_string__sz(pStringEntryName),
								 SG_string__sz(pItemParent->pStringRepoPath))  );
			}
		}
		else
		{
			// the parent directory is an undo-style remove of an add
			// and is not scheduled for deletion (just the ptnode).

			SG_ASSERT(  (pItemParent->flags & _TASK__PTN__DELETE_PTNODE)  );

			if (pItemChild->flags & _TASK__PTN__DELETE_PTNODE)
			{
				// we require that everything within it also be a simple ptnode
				// delete (because we can't have ptnodes-with-deleted status
				// for the children because they would be disconnected when the
				// parent ptnode is deleted).
			}
			else
			{
				SG_ERR_THROW2(  SG_ERR_CANNOT_REMOVE_SAFELY,
								(pCtx, "Entry '%s' prevents deletion of '%s'.",
								 SG_string__sz(pStringEntryName),
								 SG_string__sz(pItemParent->pStringRepoPath))  );
			}
		}
	}

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPath);
}

//////////////////////////////////////////////////////////////////

static SG_rbtree_foreach_callback cb_fake_readdir;

static void cb_fake_readdir(SG_context * pCtx,
							const char * pszKey_gid_child,
							void * pVoid_ptn_child,
							void * pVoid_ItemParent)
{
	struct _remove_data_item * pItemParent = (struct _remove_data_item *)pVoid_ItemParent;
	struct sg_ptnode * ptnChild = (struct sg_ptnode *)pVoid_ptn_child;
	struct _remove_data_item * pItemChild = NULL;	// we do not own this
	const char * pszEntryNameChild = NULL;	// we do not own this
	SG_pathname * pPath = NULL;
	SG_file_spec__match_result eval;
	SG_bool bDuplicate = SG_FALSE;

	SG_UNUSED( pszKey_gid_child );

	SG_ASSERT(  (!pItemParent->bExistsOnDisk)  );
	SG_ASSERT(  (pItemParent->flags & _TASK__FS__DID_NOT_EXIST)  );

	SG_ERR_CHECK(  sg_ptnode__get_name(pCtx, ptnChild, &pszEntryNameChild)  );
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx, &pPath,
												   pItemParent->pPath,
												   pszEntryNameChild)  );

	sg_pendingtree__remove_safely__alloc_item(pCtx,
											  pItemParent->pData,
											  SG_pathname__sz(pPath),
											  pItemParent->ptn,
											  ptnChild,
											  &pItemChild,
											  &bDuplicate);

	if (!pItemChild->bActive)
	{
#if TRACE_REMOVE
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "Ignoring inactive entry '%s' in '%s'.\n",
								   pszEntryNameChild,
								   SG_string__sz(pItemParent->pStringRepoPath))  );
#endif
	}
	else
	{
		SG_ASSERT(  (!pItemChild->bExistsOnDisk)  );

		SG_ERR_CHECK(  SG_file_spec__should_include(pCtx,
													pItemParent->pData->pFilespec_NoIgnores,
													SG_string__sz(pItemChild->pStringRepoPath),
													SG_FILE_SPEC__CONVERT_REPO_PATH
													| SG_FILE_SPEC__MATCH_TRAILING_SLASH,
													&eval)  );
		switch (eval)
		{
		default:		// quiets compiler
			SG_ASSERT(  (0)  );

		case SG_FILE_SPEC__RESULT__EXCLUDED:
		case SG_FILE_SPEC__RESULT__IGNORED:
		case SG_FILE_SPEC__RESULT__MAYBE:
			SG_ERR_THROW2(  SG_ERR_CANNOT_REMOVE_SAFELY,
							(pCtx, "Excluded/Ignored entry '%s' prevents deletion of '%s'.",
							 pszEntryNameChild,
							 SG_string__sz(pItemParent->pStringRepoPath))  );
			break;

		case SG_FILE_SPEC__RESULT__INCLUDED:
		case SG_FILE_SPEC__RESULT__IMPLIED:
			break;
		}

		SG_ERR_CHECK(  sg_pendingtree__remove_safely__inspect_item(pCtx,
																   SG_string__sz(pItemChild->pStringRepoPath),
																   pItemChild,
																   pItemChild->pData)  );
		SG_ASSERT(  (pItemChild->flags & _TASK__FS__DID_NOT_EXIST)  );

		if (pItemParent->flags & _TASK__PTN__DELETE_PTNODE)
		{
			if (pItemChild->flags & _TASK__PTN__DELETE_PTNODE)
			{
				// we require that everything within it also be a simple ptnode
				// delete (because we can't have ptnodes-with-deleted status
				// for the children because they would be disconnected when the
				// parent ptnode is deleted).
			}
			else
			{
				SG_ERR_THROW2(  SG_ERR_CANNOT_REMOVE_SAFELY,
								(pCtx, "Entry '%s' prevents deletion of '%s'.",
								 pszEntryNameChild,
								 SG_string__sz(pItemParent->pStringRepoPath))  );
			}
		}
	}

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPath);
}

static void sg_pendingtree__remove_safely__inspect_directory(SG_context * pCtx,
															 struct _remove_data_item * pItem)
{
	_remove_data_task_flags flags = 0;

	SG_ASSERT(  (pItem->bActive)  );

	// Compute simplified set of estimated task flags.  This is like
	// __inspect_file() but we don't have to worry about backing up
	// the contents of a file.

	if (pItem->bExistedInBaselineVersion)
	{
		if (pItem->bExistsOnDisk)
		{
			flags = _TASK__FS__RMDIR_PATHNAME | _TASK__PTN__MARK_DELETED;
		}
		else
		{
			flags = _TASK__FS__DID_NOT_EXIST | _TASK__PTN__MARK_DELETED;
		}
	}
	else	// we have a REMOVE of an ADDED item.
	{
		if (pItem->bMergeCreated)
		{
			if (pItem->bExistsOnDisk)
			{
				flags = _TASK__FS__RMDIR_PATHNAME | _TASK__PTN__MARK_UNADDED;
			}
			else
			{
				flags = _TASK__FS__DID_NOT_EXIST | _TASK__PTN__MARK_UNADDED;
			}
		}
		else
		{
			if (pItem->bExistsOnDisk)
			{
				flags = _TASK__FS__KEEP_PATHNAME | _TASK__PTN__DELETE_PTNODE;
			}
			else
			{
				flags = _TASK__FS__DID_NOT_EXIST | _TASK__PTN__DELETE_PTNODE;
			}
		}
	}

	pItem->flags = flags;

#if TRACE_REMOVE
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "RemoveSafely: flags 0x%x for directory %s\n",
							   flags,
							   SG_string__sz(pItem->pStringRepoPath))  );
#endif

	SG_ERR_CHECK(  sg_ptnode__load_dir_entries(pCtx, pItem->ptn)  );

	if (pItem->bExistsOnDisk)
	{
		SG_ERR_CHECK(  SG_dir__foreach(pCtx,
									   pItem->pPath,
									   SG_DIR__FOREACH__SKIP_OS,
									   cb_readdir,
									   pItem)  );
	}
	else
	{
		SG_ERR_CHECK(  SG_rbtree__foreach(pCtx,
										  pItem->ptn->prbItems,
										  cb_fake_readdir,
										  pItem)  );
	}

	SG_ERR_CHECK(  sg_pendingtree__remove_safely__append_task(pCtx,
															  pItem->pData,
															  pItem->pPath,
															  pItem->pStringRepoPath,
															  pItem->ptn,
															  pItem->flags)  );
fail:
	;
}

//////////////////////////////////////////////////////////////////

static void sg_pendingtree__remove_safely__inspect_submodule(SG_context * pCtx,
															 struct _remove_data_item * pItem)
{
	_remove_data_task_flags flags = 0;
	SG_bool bDirty_Reference = SG_FALSE;
	SG_bool bDirty_Pending = SG_FALSE;

	SG_ASSERT(  (pItem->bActive)  );	// previously deleted ?
	
	if (pItem->bExistsOnDisk)
	{
		SG_ERR_CHECK(  sg_ptnode__is_submodule_changed_or_dirty(pCtx,
																pItem->ptn,
																&bDirty_Reference,
																&bDirty_Pending)  );
	}

	if (pItem->bExistedInBaselineVersion)
	{
		// Since it existed in the baseline, we can use a normal _MARK_DELETED
		// flag causing the ptnode to be kept with deleted status (so 'vv status'
		// can say it was removed).

		if (pItem->bExistsOnDisk)
		{
			if (!bDirty_Pending)
			{
				flags = _TASK__FS__TO_GRAVEYARD | _TASK__PTN__MARK_DELETED;
			}
			// not sure if (pItem->bMergeCreated) matters here...
			else if (pItem->pData->bForce)
			{
				flags = _TASK__FS__TO_GRAVEYARD | _TASK__PTN__MARK_DELETED;
			}
			else
			{
#if TRACE_REMOVE
				SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
										   "RemoveSafely: unsafe submodule: %s\n", SG_pathname__sz(pItem->pPath))  );
#endif
				SG_ERR_THROW2(  SG_ERR_CANNOT_REMOVE_SAFELY,
								(pCtx, "Submodule has changes: '%s'", SG_string__sz(pItem->pStringRepoPath))  );
			}
		}
		else
		{
			// they did a /bin/rm before they told us about it.
			// silently do the right thing and mark it deleted
			// without making them recreate the file and redo
			// the 'vv remove'.
			//
			// TODO 2011/05/05 We're going to mark it removed without
			// TODO            creating an entry for it in the graveyard.
			// TODO            This might cause 'vv revert' a few problems.

			flags = _TASK__FS__DID_NOT_EXIST | _TASK__PTN__MARK_DELETED;
		}
	}
	else	// we have a REMOVE of an ADDED item.
	{
		if (pItem->bMergeCreated)
		{
			if (pItem->bExistsOnDisk)
			{
				if (!bDirty_Pending)
				{
					flags = _TASK__FS__TO_GRAVEYARD | _TASK__PTN__MARK_UNADDED;
				}
				else if (pItem->pData->bForce)
				{
					flags = _TASK__FS__TO_GRAVEYARD | _TASK__PTN__MARK_UNADDED;
				}
				else
				{
#if TRACE_REMOVE
					SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
											   "RemoveSafely: unsafe submodule: %s\n", SG_pathname__sz(pItem->pPath))  );
#endif
					SG_ERR_THROW2(  SG_ERR_CANNOT_REMOVE_SAFELY,
									(pCtx, "Submodule has changes: '%s'", SG_string__sz(pItem->pStringRepoPath))  );
				}
			}
			else
			{
				flags = _TASK__FS__DID_NOT_EXIST | _TASK__PTN__MARK_UNADDED;
			}
		}
		else
		{
			if (pItem->bExistsOnDisk)
			{
				flags = _TASK__FS__KEEP_PATHNAME | _TASK__PTN__DELETE_PTNODE;
			}
			else
			{
				flags = _TASK__FS__DID_NOT_EXIST | _TASK__PTN__DELETE_PTNODE;
			}
		}
	}

	pItem->flags = flags;

#if TRACE_REMOVE
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "RemoveSafely: flags 0x%x for submodule %s\n",
							   pItem->flags,
							   SG_string__sz(pItem->pStringRepoPath))  );
#endif

	SG_ERR_CHECK(  sg_pendingtree__remove_safely__append_task(pCtx,
															  pItem->pData,
															  pItem->pPath,
															  pItem->pStringRepoPath,
															  pItem->ptn,
															  pItem->flags)  );

fail:
	;
}

//////////////////////////////////////////////////////////////////

static void sg_pendingtree__remove_safely__inspect_item(SG_context * pCtx,
														const char * pszKey_RepoPath,
														void * pVoid_Item,
														void * pVoid_Data)
{
	struct _remove_data_item * pItem = (struct _remove_data_item *)pVoid_Item;

	// struct _remove_data * pData = (struct _remove_data *)pVoid_Data;
	SG_UNUSED( pVoid_Data );

#if TRACE_REMOVE
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "Remove: inspecting item '%s'...\n",
							   pszKey_RepoPath)  );
#else
	SG_UNUSED( pszKey_RepoPath );
#endif

	// For historical reasons we disallow REMOVES of MOVED or RENAMED
	// items.  I'd like to eliminate this restriction, but not today.

	SG_ERR_CHECK(  sg_pendingtree__remove_safely__disallow_moved_or_renamed(pCtx,
																			pItem->ptn,
																			SG_string__sz(pItem->pStringRepoPath))  );
	if (pItem->bExistsOnDisk)
		SG_ERR_CHECK(  sg_pendingtree__remove_safely__disallow_changed_type(pCtx,
																			pItem->ptn,
																			&pItem->st,
																			SG_string__sz(pItem->pStringRepoPath))  );

	if (!pItem->bActive)
	{
		// Qualify an INACTIVE node.  For now, we want to allow
		// *already-deleted* nodes, but not MOVED-AWAY ones.

		SG_bool bNothingToDo = SG_FALSE;

		SG_ERR_CHECK(  sg_pendingtree__remove_safely__disallow_certain_inactive(pCtx,
																				pItem->ptn,
																				pItem->bExistsOnDisk,
																				SG_string__sz(pItem->pStringRepoPath),
																				&bNothingToDo)  );
		if (bNothingToDo)
			goto skip_it;
	}

	switch (pItem->ptn->type)
	{
	default:	// quiets compiler
		SG_ERR_THROW(  SG_ERR_INVALID_PENDINGTREE_OBJECT_TYPE  );

	case SG_TREENODEENTRY_TYPE_DIRECTORY:
		SG_ERR_CHECK(  sg_pendingtree__remove_safely__inspect_directory(pCtx, pItem)  );
		break;

	case SG_TREENODEENTRY_TYPE_REGULAR_FILE:
	case SG_TREENODEENTRY_TYPE_SYMLINK:
		SG_ERR_CHECK(  sg_pendingtree__remove_safely__inspect_file_or_symlink(pCtx, pItem)  );
		break;

	case SG_TREENODEENTRY_TYPE_SUBMODULE:
		SG_ERR_CHECK(  sg_pendingtree__remove_safely__inspect_submodule(pCtx, pItem)  );
		break;
	}

skip_it:
	;
fail:
	;
}

//////////////////////////////////////////////////////////////////

void SG_pendingtree__remove_dont_save_pendingtree(SG_context* pCtx,
												  SG_pendingtree* pPendingTree,
												  SG_uint32 count_items, const char** paszItems,
												  const SG_file_spec* pFilespecGiven,
												  SG_bool bTest, SG_bool bForce, SG_bool bNoBackup)
{
	struct _remove_data data;
	SG_uint32 k;

	memset(&data, 0, sizeof(data));		// must be first

	if (count_items == 0)
		SG_ERR_THROW_RETURN(  SG_ERR_USAGE  );

	SG_NULLARGCHECK( pPendingTree );
	SG_NULLARGCHECK( pFilespecGiven );

	data.bTest = bTest;
	data.bForce = bForce;
	data.bNoBackup = bNoBackup;
	data.pPendingTree = pPendingTree;
	SG_ERR_CHECK(  SG_pendingtree__get_repo(pCtx, data.pPendingTree, &data.pRepo)  );

	SG_ERR_CHECK(  SG_FILE_SPEC__ALLOC__COPY(pCtx, pFilespecGiven, &data.pFilespec_NoIgnores)  );

	SG_ERR_CHECK(  SG_FILE_SPEC__ALLOC__COPY(pCtx, pFilespecGiven, &data.pFilespec_WithIgnores)  );

	// TODO 2011/07/17 Do we actually need the IGNORES for a REMOVE ?
	// TODO            I would expect it for ADD, ADDREMOVE, and SCAN/STATUS
	// TODO            because they look at uncontrolled items.  But REMOVE
	// TODO            should only be dealing with controlled items, right?
	SG_ERR_CHECK(  SG_file_spec__load_ignores(pCtx,
											  data.pFilespec_WithIgnores,
											  data.pRepo,
											  data.pPendingTree->pPathWorkingDirectoryTop,
											  0)  );

	SG_ERR_CHECK(  SG_RBTREE__ALLOC__PARAMS2(pCtx,
											 &data.prbAllItems,
											 0, NULL,
											 SG_rbtree__compare_function__reverse_strcmp)  );
	SG_ERR_CHECK(  SG_RBTREE__ALLOC__PARAMS2(pCtx,
											 &data.prbTopLevelItems,
											 0, NULL,
											 SG_rbtree__compare_function__reverse_strcmp)  );
	SG_ERR_CHECK(  SG_RBTREE__ALLOC__PARAMS2(pCtx,
											 &data.prbTaskList,
											 0, NULL,
											 SG_rbtree__compare_function__reverse_strcmp)  );

	// prequalify each of the top-level named items.
	// this is basically arg-checking for each named item.

	for (k=0; k<count_items; k++)
		SG_ERR_CHECK(  sg_pendingtree__remove_safely__alloc_top_level_item(pCtx, &data, paszItems[k])  );

	// inspect each item (and for directories, the stuff within)
	// to see if it can be safely removed.  we do this in repo-path
	// order (deepest first) in case any of the top-level items are
	// actually contained within one of the other items.  that is,
	// we would like 'vv remove foo foo/bar' to not complain about 'bar'
	// not existing in iteration 2 because the (recursive) remove
	// of 'foo' already took care of it.

	SG_ERR_CHECK(  SG_rbtree__foreach(pCtx,
									  data.prbTopLevelItems,
									  sg_pendingtree__remove_safely__inspect_item,
									  &data)  );
	SG_ERR_CHECK(  SG_rbtree__foreach(pCtx,
									  data.prbTaskList,
									  sg_pendingtree__remove_safely__apply_task,
									  &data)  );

fail:
	SG_FILE_SPEC_NULLFREE(pCtx, data.pFilespec_NoIgnores);
	SG_FILE_SPEC_NULLFREE(pCtx, data.pFilespec_WithIgnores);
	SG_RBTREE_NULLFREE_WITH_ASSOC(pCtx,
								  data.prbTaskList,
								  (SG_free_callback *)sg_pendingtree__remove_safely__free_task);
	SG_RBTREE_NULLFREE(pCtx, data.prbTopLevelItems);
	SG_RBTREE_NULLFREE_WITH_ASSOC(pCtx,
								  data.prbAllItems,
								  (SG_free_callback *)sg_pendingtree__remove_safely__free_item);
}

//////////////////////////////////////////////////////////////////

void SG_pendingtree__remove(SG_context* pCtx,
							SG_pendingtree* pPendingTree,
							SG_uint32 count_items, const char** paszItems,
							const SG_file_spec* pFilespec,
							SG_bool bTest, SG_bool bForce, SG_bool bNoBackup)
{
	SG_ERR_CHECK(  SG_pendingtree__remove_dont_save_pendingtree(pCtx,
																pPendingTree,
																count_items, paszItems,
																pFilespec,
																bTest, bForce, bNoBackup)  );
	if (bTest == SG_FALSE)
		SG_ERR_CHECK(  SG_pendingtree__save(pCtx, pPendingTree)  );

fail:
	return;
}

//////////////////////////////////////////////////////////////////

END_EXTERN_C;

#endif//H_SG_PENDINGTREE__PRIVATE_REMOVE_SAFELY2_H
