/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg_pendingtree__private_resolve_issues.h
 *
 * @details 
 *
 */

//////////////////////////////////////////////////////////////////

#ifndef H_SG_PENDINGTREE__PRIVATE_RESOLVE_ISSUES_H
#define H_SG_PENDINGTREE__PRIVATE_RESOLVE_ISSUES_H

BEGIN_EXTERN_C;

//////////////////////////////////////////////////////////////////

#define H1  ""
#define H2	"\t# "
#define H3	"\t#   "
#define H4	"\t#     "

//////////////////////////////////////////////////////////////////

/**
 * Format the main message for this entry.
 *
 * Append info to pStrOutput.
 *
 * Return the live/current repo-path for this entry (if requested).
 */
static void _pt_resolve__format_header(SG_context * pCtx,
									   SG_pendingtree * pPendingTree,
									   const SG_vhash * pvhIssue,
									   SG_string * pStrOutput,
									   SG_string ** ppStrRepoPath_Returned)
{
	SG_string * pStrRepoPath = NULL;
	const char * pszGid;
	SG_int64 i64;
	SG_pendingtree_wd_issue_status status;
	SG_bool bResolved;

	SG_ERR_CHECK_RETURN(  SG_vhash__get__int64(pCtx, pvhIssue, "status", &i64)  );
	status = (SG_pendingtree_wd_issue_status)i64;
	bResolved = ((status & SG_ISSUE_STATUS__MARKED_RESOLVED) == SG_ISSUE_STATUS__MARKED_RESOLVED);

	SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhIssue, "gid", &pszGid)  );
#if 0 && defined(DEBUG)
	SG_ERR_IGNORE(  SG_vhash_debug__dump_to_console__named(pCtx, pvhIssue, pszGid)  );
	SG_ERR_IGNORE(  SG_pendingtree_debug__dump_existing_to_console(pCtx, pPendingTree)  );
#endif

	// Compute the live/current repo-path for the entry.
	// 
	// WARNING: There currently is a "repopath" field on the conflict, *BUT* it
	//          may be stale.  And I want to say that it is deprecated and going
	//          away.
	// 
	// It was current/valid at the time that the MERGE was performed and before the
	// user was given a chance to modify the WD (either manually or with RESOLVE).
	// That repo-path may still be valid, but then again they may have moved/renamed
	// the item and/or moved/renamed a parent directory.  So try to compute a live
	// repo-path for the item using information in the pendingtree.  This repo-path
	// should match what is currently on disk *and* should match the value printed
	// in the non-conflict portion of the status report.
	//
	// TODO 2010/08/19 We currently have a bug where sometimes __find_repo_path_by_gid()
	// TODO            fails.  This happens when the user has modified the entry in the
	// TODO            WD such that it appears CLEAN relative to the BASELINE (and so
	// TODO            all of the stuff in pendingtree to strip out clean ptnodes bites
	// TODO            us).  What we really need to do is "pin" these ptnodes so that
	// TODO            they won't get stripped out if there's a chance we (STATUS, RESOLVE,
	// TODO            and etc.), might need them.
	// TODO
	// TODO            For now, let this fail, catch it and substitute a pathname with
	// TODO            a warning in it.

	SG_pendingtree__find_repo_path_by_gid(pCtx, pPendingTree, pszGid, &pStrRepoPath);
	if (SG_context__has_err(pCtx))
	{
		const char * pszPath = NULL;

		if (!SG_context__err_equals(pCtx, SG_ERR_NOT_FOUND))
			SG_ERR_RETHROW_RETURN;
			
#if 1 // TODO delete this when we fix the bug -- SEE Y4182.
		SG_context__err_reset(pCtx);

		SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhIssue, "repopath", &pszPath)  );
		SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pStrRepoPath)  );
		SG_ERR_CHECK(  SG_string__sprintf(pCtx, pStrRepoPath, "TODO Fix Path for Display: %s", pszPath)  );
#endif
	}

	SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
											 H1 "%s %s\n",
											 ((bResolved) ? "Resolved:   " : "Unresolved: "),
											 SG_string__sz(pStrRepoPath))  );

	if (ppStrRepoPath_Returned)
	{
		*ppStrRepoPath_Returned = pStrRepoPath;
		pStrRepoPath = NULL;
	}

fail:
	SG_STRING_NULLFREE(pCtx, pStrRepoPath);
}

//////////////////////////////////////////////////////////////////

static void _pt_resolve__format_submodule_divergent_edit_conflict_details(SG_context * pCtx,
																		  SG_pendingtree * pPendingTree,
																		  const SG_vhash * pvhIssue,
																		  SG_string * pStrOutput)
{
	const SG_varray * pva_parents_ref = NULL;
	const char * pszHidOuterBaseline;
	const char * pszHid_chosen_for_wd;
	const SG_vhash * pvhHIDs;
	SG_repo * pRepo;
	SG_varray * pva_referenced_by;
	SG_treenode_submodule * pTSM = NULL;
	char * pszDescription = NULL;
	SG_uint32 kReference, nrReferences;
	SG_uint32 nrHIDs, kHID;
	SG_bool bReferencedByBaseline;

	SG_ERR_CHECK(  SG_pendingtree__get_repo(pCtx, pPendingTree, &pRepo)  );

	// We have a divergent-edit-submodule.  This means that
	// the submodule was pointing to different csets in the
	// 2 or more outer-branches and MERGE doesn't know which
	// cset the submodule should be set to (and we REFUSE to
	// dive into the submodule to investigate).

	// pvhHIDs contains the unique set of submodule hids that
	// are referenced by at least one of the branches in the
	// outer merge.  each vhash-key is the content HID in the
	// TNE. each vhash-value is a varray of the HIDs of the
	// outer csets that referenced that submodule value.
	// 
	// with a 2-way merge, this is a bit overkill, but allows
	// for n-way merges in the future.
	//
	// see sg_mrg__private_build_wd_issues.h
	//
	// "hid_submodule" : "<hid_blob>",
	// "conflict_divergent_hid_submodule" : { "<hid_blob_1>" : [ "<hid_cset_j>", ... ],
	//                                        "<hid_blob_2>" : [ "<hid_cset_k>", ... ],
	//                                        ... },

	SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvhIssue, "conflict_divergent_hid_submodule", (SG_vhash **)&pvhHIDs)  );
	SG_ERR_CHECK(  SG_vhash__count(pCtx, pvhHIDs, &nrHIDs)  );
	SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
											 (H3 "This submodule was changed to reference %d different changesets.\n"
											  H3 "You should inspect the submodule and update/merge it as is\n"
											  H3 "appropriate and then mark it resolved.  MERGE arbitrarily selected\n"
											  H3 "the first of these referenced changesets:\n"),
											 nrHIDs)  );

	// So MERGE arbitrarily chose one of the submodule csets
	// when it populated the outer-WD.  (In the 2-way case,
	// this should always be the cset referenced by the
	// outer-baseline.)  (When we get to n-way merges, this
	// is more arbitrary (because the submodule may not have
	// been present in the baseline and it was changed in 2
	// or more of the other outer branches))

	SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhIssue, "hid_submodule", &pszHid_chosen_for_wd)  );

	// use the chosen-hid and walk the corresponding varray and determine
	// if it was referenced by the baseline (it should in a 2-way).
	// use that to display the details of that submodule version.

	SG_ERR_CHECK(  SG_pendingtree__get_wd_parents__ref(pCtx, pPendingTree, &pva_parents_ref)  );
	SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pva_parents_ref, 0, &pszHidOuterBaseline)  );

	SG_ERR_CHECK(  SG_vhash__get__varray(pCtx, pvhHIDs, pszHid_chosen_for_wd, &pva_referenced_by)  );
	SG_ERR_CHECK(  SG_varray__count(pCtx, pva_referenced_by, &nrReferences)  );
	bReferencedByBaseline = SG_FALSE;
	for (kReference=0; kReference<nrReferences; kReference++)
	{
		const char * pszHidReference;

		SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pva_referenced_by, kReference, &pszHidReference)  );
		if (strcmp(pszHidReference, pszHidOuterBaseline) == 0)
		{
			bReferencedByBaseline = SG_TRUE;
			break;
		}
	}

	SG_ERR_CHECK(  SG_treenode_submodule__load_from_repo(pCtx, pRepo, pszHid_chosen_for_wd, &pTSM)  );
	SG_ERR_CHECK(  SG_treenode_submodule__describe_for_status(pCtx, pTSM, &pszDescription)  );
	SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
											 ((bReferencedByBaseline)
											  ? (H4 "Baseline: %s\n")
											  : (H4 "Selected: %s\n")),
											 pszDescription)  );
	SG_NULLFREE(pCtx, pszDescription);
	SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pTSM);

	// Now display the details of all of the other submodule versions.

	for (kHID=0; kHID<nrHIDs; kHID++)
	{
		const char * pszHid_k;

		SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvhHIDs, kHID, &pszHid_k, NULL)  );
		if (strcmp(pszHid_k, pszHid_chosen_for_wd) == 0)
			continue;

		SG_ERR_CHECK(  SG_treenode_submodule__load_from_repo(pCtx, pRepo, pszHid_k, &pTSM)  );
		SG_ERR_CHECK(  SG_treenode_submodule__describe_for_status(pCtx, pTSM, &pszDescription)  );
		SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
												 (H4 "Other:    %s\n"),
												 pszDescription)  );
		SG_NULLFREE(pCtx, pszDescription);
		SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pTSM);
	}

fail:
	SG_NULLFREE(pCtx, pszDescription);
	SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pTSM);
}

//////////////////////////////////////////////////////////////////

/**
 * Format the sub-messages associated with a conflict.
 *
 */
static void _pt_resolve__format_conflict_flags(SG_context * pCtx,
											   SG_pendingtree * pPendingTree,
											   const SG_vhash * pvhIssue,
											   SG_string * pStrOutput)
{
	SG_string * pStrRepoPathParent = NULL;
	SG_int64 i64;
	SG_mrg_cset_entry_conflict_flags conflict_flags;

	SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvhIssue, "conflict_flags", &i64)  );
	conflict_flags = (SG_mrg_cset_entry_conflict_flags)i64;
	if (conflict_flags == 0)
		return;
	
	SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
											 H2 "Conflicting changes were made to this entry:\n")  );

#if 0 && defined(DEBUG)	// print a debug line with the conflict flags
	{
		SG_bool bHasDebugReasons = SG_FALSE;
		char buf_i64[40];

		SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
												 H3 "DEBUG [%s]",
												 SG_int64_to_sz(i64, buf_i64))  );
		
		SG_ERR_CHECK(  SG_vhash__has(pCtx, pvhIssue, "conflict_reasons", &bHasDebugReasons)  );
		if (bHasDebugReasons)	// if mrg code has conflict_reasons debug enabled, also dump the list of tokens.
		{
			const SG_varray * pvaConflictReasons;
			SG_uint32 k, count;
				
			SG_ERR_CHECK(  SG_vhash__get__varray(pCtx, pvhIssue, "conflict_reasons", (SG_varray **)&pvaConflictReasons)  );
			SG_ERR_CHECK(  SG_string__append__sz(pCtx, pStrOutput, ": [")  );
			SG_ERR_CHECK(  SG_varray__count(pCtx, pvaConflictReasons, &count)  );
			for (k=0; k<count; k++)
			{
				const char * psz_k;

				SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pvaConflictReasons, k, &psz_k)  );
				SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
														 "%s ", psz_k)  );
			}
			SG_ERR_CHECK(  SG_string__append__sz(pCtx, pStrOutput, "]")  );
		}
		SG_ERR_CHECK(  SG_string__append__sz(pCtx, pStrOutput, "\n")  );
	}
#endif

	if (conflict_flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__UNDELETE__MASK)
	{
		const SG_varray * pvaConflictDeletes;
		const char * psz_k;
		SG_uint32 count;

		SG_ERR_CHECK(  SG_vhash__get__varray(pCtx, pvhIssue, "conflict_deletes", (SG_varray **)&pvaConflictDeletes)  );
		SG_ERR_CHECK(  SG_varray__count(pCtx, pvaConflictDeletes, &count)  );
		if (count == 1)
		{
			SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pvaConflictDeletes, 0, &psz_k)  );
			SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
													 H3 "It was deleted in changeset: %s\n",
													 psz_k)  );
		}
		else
		{
			SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
													 H3 "It was deleted in %d changesets:\n",
													 count)  );
		}

		if (conflict_flags & (SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DELETE_VS_MOVE
							  |SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DELETE_VS_RENAME))
			SG_ERR_CHECK(  SG_string__append__sz(pCtx, pStrOutput,
												 H3 "It was moved/renamed in another changeset.\n")  );
		if (conflict_flags & (SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DELETE_VS_ATTRBITS
							  |SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DELETE_VS_XATTR))
			SG_ERR_CHECK(  SG_string__append__sz(pCtx, pStrOutput,
												 H3 "Attributes on the entry were changed in another changeset.\n")  );
		if (conflict_flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DELETE_VS_SYMLINK_EDIT)
			SG_ERR_CHECK(  SG_string__append__sz(pCtx, pStrOutput,
												 H3 "The target of the symbolic link was changed in another changeset.\n")  );
		if (conflict_flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DELETE_VS_SUBMODULE_EDIT)
			SG_ERR_CHECK(  SG_string__append__sz(pCtx, pStrOutput,
												 H3 "The referenced changeset in the submodule was changed in another changeset.\n")  );
		if (conflict_flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DELETE_VS_FILE_EDIT)
			SG_ERR_CHECK(  SG_string__append__sz(pCtx, pStrOutput,
												 H3 "The contents of the file was changed in another changeset.\n")  );
		if (conflict_flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DELETE_CAUSED_ORPHAN)
			SG_ERR_CHECK(  SG_string__append__sz(pCtx, pStrOutput,
												 H3 "Deleting this directory would orphan items created/moved into it in another changeset.\n")  );
	}

	if (conflict_flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__MOVES_CAUSED_PATH_CYCLE)
		SG_ERR_CHECK(  SG_string__append__sz(pCtx, pStrOutput,
											 H3 "The move of this directory caused a pathname cycle when combined with the moves in another changeset.\n")  );

	if (conflict_flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_MOVE)
	{
		const SG_vhash * pvhGIDs;
		SG_uint32 k, count;

		SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvhIssue, "conflict_divergent_moves", (SG_vhash **)&pvhGIDs)  );
		SG_ERR_CHECK(  SG_vhash__count(pCtx, pvhGIDs, &count)  );
		SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
												 H3 "Entry was moved to %d different directories:\n",
												 count)  );
		for (k=0; k<count; k++)
		{
			const char * pszGid_k;

			SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvhGIDs, k, &pszGid_k, NULL)  );
			// TODO 2010/09/02 The following may need the NOT_FOUND stuff too.
			SG_ERR_CHECK(  SG_pendingtree__find_repo_path_by_gid(pCtx, pPendingTree, pszGid_k, &pStrRepoPathParent)  );
			SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
													 H4 "%s\n",
													 SG_string__sz(pStrRepoPathParent))  );
			SG_STRING_NULLFREE(pCtx, pStrRepoPathParent);
		}
	}

	if (conflict_flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_RENAME)
	{
		const SG_vhash * pvhNames;
		SG_uint32 k, count;

		SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvhIssue, "conflict_divergent_renames", (SG_vhash **)&pvhNames)  );
		SG_ERR_CHECK(  SG_vhash__count(pCtx, pvhNames, &count)  );
		SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
												 H3 "Entry was renamed to %d different names:\n",
												 count)  );
		for (k=0; k<count; k++)
		{
			const char * pszName_k;

			SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvhNames, k, &pszName_k, NULL)  );
			SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
													 H4 "%s\n",
													 pszName_k)  );
		}
	}

	if (conflict_flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_ATTRBITS)
	{
		// TODO 2010/09/02 Since we currently only have 1 attrbit (the +x bit),
		// TODO            we can't really get this.  Later (when we have more
		// TODO            bit), we need to decide how to display them here.
		// TODO
		// TODO            For now, just print a message and don't bother
		// TODO            with the actual bits.

		SG_ERR_CHECK(  SG_string__append__sz(pCtx, pStrOutput,
											 H3 "Entry was given diffent attributes.  TODO show details.\n")  );
	}

	if (conflict_flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_XATTR)
	{
		// TODO 2010/09/02 Decide if/how we want to print different sets of XATTRS.

		SG_ERR_CHECK(  SG_string__append__sz(pCtx, pStrOutput,
											 H3 "Entry was given different XAttrs.  TODO show details.\n")  );
	}

	if (conflict_flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_SYMLINK_EDIT)
	{
		const SG_vhash * pvhHIDs;
		SG_uint32 count;

		SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvhIssue, "conflict_divergent_hid_symlink", (SG_vhash **)&pvhHIDs)  );
		SG_ERR_CHECK(  SG_vhash__count(pCtx, pvhHIDs, &count)  );
		SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
												 H3 "This symbolic link was changed to %d different values.  TODO show details.\n",
												 count)  );

		// TODO 2010/09/02 Decide if/how we want to print different symlink targets.
		// TODO            See S00086.
	}

	if (conflict_flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_SUBMODULE_EDIT)
	{
		SG_ERR_CHECK(  _pt_resolve__format_submodule_divergent_edit_conflict_details(pCtx,
																					 pPendingTree,
																					 pvhIssue,
																					 pStrOutput)  );
	}

	if (conflict_flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__MASK)
	{
		// We have 2 types of state:
		// [1] The results computed during the actual MERGE operation are stored in
		//     the conflict_flags for this ISSUE.  This indicates whether auto-merge
		//     was attempted and whether it was successful.
		//
		//     See __DIVERGENT_FILE_EDIT__MASK.
		//
		//     This status is set when during the MERGE and NEVER CHANGED.
		//
		// [2] If auto-merge could not do the job (or if the user decide to redo it
		//     themself), then we have "resolve_mergetool_status" field.  The resolve
		//     status fields will only be present if a previous resolve has been
		//     attempted.
		//
		//     This status field is not added to the ISSUE until a resolve-merge is
		//     atttempted.   It is updated upon subsequent attempts to resolve this
		//     issue.
		//
		// During a RESOLVE we could have an issue that has both an edit conflict
		// and, for example, a divergent rename conflict.  These are handled in 2
		// steps:  [1] invoke their mergetool (and let it update the result-file
		//             in tmp and exit); and
		//         [2] ask them to choose the correct filename.
		// Because the user can interrupt us between the steps in the plan, we
		// allow RESOLVE to update the mergetool status immediately and then
		// deal with [2].

		const char * psz_AutoMerge_ToolName = NULL;
		const char * psz_Resolve_ToolName = NULL;
		SG_int32 statusResolveTool = SG_MERGETOOL__STATUS__NOT_ATTEMPTED;
		SG_bool bHas_AutoMerge_ToolName = SG_FALSE;
		SG_bool bHas_Resolve_ToolName = SG_FALSE;

		SG_ERR_CHECK(  SG_vhash__has(pCtx, pvhIssue, "automerge_mergetool", &bHas_AutoMerge_ToolName)  );
		if (bHas_AutoMerge_ToolName)
		{
			SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhIssue, "automerge_mergetool", &psz_AutoMerge_ToolName)  );
		}
		SG_ERR_CHECK(  SG_vhash__has(pCtx, pvhIssue, "resolve_mergetool", &bHas_Resolve_ToolName)  );
		if (bHas_Resolve_ToolName)
		{
			SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhIssue, "resolve_mergetool", &psz_Resolve_ToolName)  );
			SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvhIssue, "resolve_mergetool_status", &i64)  );
			statusResolveTool = (SG_int32)i64;
		}

		SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
												 H3 "The file was modified in both branches and needs to be merged.\n")  );

		if (conflict_flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__TBD)
		{
			SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
													 H4 "The MERGE command did not attempt to auto-merge them.\n")  );
		}
		else if (conflict_flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__NO_RULE)
		{
			SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
													 H4 "The MERGE command did not know how to auto-merge this type of file.\n")  );
		}
		else if (conflict_flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__AUTO_CONFLICT)
		{
			SG_ASSERT( (bHas_AutoMerge_ToolName) && (*psz_AutoMerge_ToolName) );
			SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
													 H4 "During the MERGE command, the mergetool '%s' found conflicts.\n",
													 psz_AutoMerge_ToolName)  );
		}
		else if (conflict_flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__AUTO_ERROR)
		{
			SG_ASSERT( (bHas_AutoMerge_ToolName) && (*psz_AutoMerge_ToolName) );
			SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
													 H4 "During the MERGE command, the mergetool '%s' had an error.\n",
													 psz_AutoMerge_ToolName)  );
		}
		else if (conflict_flags & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__AUTO_OK)
		{
			SG_ASSERT( (bHas_AutoMerge_ToolName) && (*psz_AutoMerge_ToolName) );
			if (bHas_Resolve_ToolName)
				SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
														 H4 "The MERGE command was able to auto-merge them using the mergetool '%s', but this may have been overwritten by RESOLVE.\n",
														 psz_AutoMerge_ToolName)  );
			else
				SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
														 H4 "The MERGE command was able to auto-merge them using the mergetool '%s'.\n",
														 psz_AutoMerge_ToolName)  );
		}

		if (bHas_Resolve_ToolName)
		{
			switch (statusResolveTool)
			{
			case SG_MERGETOOL__STATUS__SUCCESSFUL:
				SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
														 H4 "The RESOLVE command was able to merge/re-merge them using the mergetool '%s'.\n",
														 psz_Resolve_ToolName)  );
				break;

			case SG_MERGETOOL__STATUS__CONFLICT:
			case SG_MERGETOOL__STATUS__ERROR:
			case SG_MERGETOOL__STATUS__NOT_ATTEMPTED:
			case SG_MERGETOOL__STATUS__EXEC_ERROR:
			case SG_MERGETOOL__STATUS__NO_OUTPUT:
			default:
				// Currently we only update the resolve fields when it is successful,
				// so none of these should ever happen, but I don't want to assert
				// that and blow up.

				SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
														 H4 "The RESOLVE command was not able to merge/re-merge them; mergetool '%s' status %d.\n",
														 psz_Resolve_ToolName, statusResolveTool)  );
				break;
			}
		}
	}

fail:
	SG_STRING_NULLFREE(pCtx, pStrRepoPathParent);
}

/**
 * Format the sub-messages associated with a real collision.
 *
 */
static void _pt_resolve__format_collision(SG_context * pCtx,
										  SG_pendingtree * pPendingTree,
										  const SG_vhash * pvhIssue,
										  SG_string * pStrOutput)
{
	const SG_varray * pvaCollisions;
	const char * pszGid;
	SG_string * pStrRepoPath_k = NULL;
	SG_uint32 k, count;
	SG_bool bCollisions;
	
	SG_ERR_CHECK(  SG_vhash__get__bool(pCtx, pvhIssue, "collision_flags", &bCollisions)  );
	if (!bCollisions)
		return;

	SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhIssue, "gid", &pszGid)  );

	// An exact match entryname collision with one or more other entries.
	// We don't really need to list the other entrynames because they're
	// all the same (and they will each appear in their own issue), but
	// we do so that we can show them the make-unique name that we gave it.

	SG_ERR_CHECK(  SG_vhash__get__varray(pCtx, pvhIssue, "collisions", (SG_varray **)&pvaCollisions)  );
	SG_ERR_CHECK(  SG_varray__count(pCtx, pvaCollisions, &count)  );
	if (count == 1)
		SG_ERR_CHECK(  SG_string__append__sz(pCtx, pStrOutput,
											 H2 "The entry collided with another entry:\n")  );
	else
		SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
												 H2 "The entry collided with %d other entries:\n",
												 count)  );
	for (k=0; k<count; k++)
	{
		const char * pszGid_k;

		SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pvaCollisions, k, &pszGid_k)  );
		if (strcmp(pszGid_k, pszGid) != 0)
		{
			SG_ERR_CHECK(  SG_pendingtree__find_repo_path_by_gid(pCtx, pPendingTree, pszGid_k, &pStrRepoPath_k)  );
			SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
													 H3 "%s\n",
													 SG_string__sz(pStrRepoPath_k))  );
			SG_STRING_NULLFREE(pCtx, pStrRepoPath_k);
		}
	}

fail:
	SG_STRING_NULLFREE(pCtx, pStrRepoPath_k);
}

/**
 * Format the sub-messages associated with a portability warning/issue.
 *
 */
static void _pt_resolve__format_portability_warning(SG_context * pCtx,
													SG_pendingtree * pPendingTree,
													const SG_vhash * pvhIssue,
													SG_string * pStrOutput)
{
	const SG_varray * pvaPortabilityDetails;
	const char * pszGid;
	SG_string * pStrRepoPath_k = NULL;
	SG_int64 i64;
	SG_uint32 k, count;
	SG_portability_flags portability_flags;

	SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvhIssue, "portability_flags", &i64)  );
	portability_flags = (SG_portability_flags)i64;
	if (!portability_flags)
		return;

	SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhIssue, "gid", &pszGid)  );

	// There were POTENTIAL entryname collisions with one or more other entries.

	SG_ERR_CHECK(  SG_vhash__get__varray(pCtx, pvhIssue, "portability_details", (SG_varray **)&pvaPortabilityDetails)  );
	SG_ERR_CHECK(  SG_varray__count(pCtx, pvaPortabilityDetails, &count)  );
	if (count == 1)
		SG_ERR_CHECK(  SG_string__append__sz(pCtx, pStrOutput,
											 H2 "The entry POTENTIALLY collided with another entry:\n")  );
	else
		SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
												 H2 "The entry POTENTIALLY collided with %d other entries:\n",
												 count)  );

	for (k=0; k<count; k++)
	{
		const SG_vhash * pvh_port_detail_k;
		const char * pszGid_k;

		SG_ERR_CHECK(  SG_varray__get__vhash(pCtx, pvaPortabilityDetails, k, (SG_vhash **)&pvh_port_detail_k)  );

		SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh_port_detail_k, "gid1", &pszGid_k)  );
		if (strcmp(pszGid_k, pszGid) != 0)
		{
			SG_ERR_CHECK(  SG_pendingtree__find_repo_path_by_gid(pCtx, pPendingTree, pszGid_k, &pStrRepoPath_k)  );
			SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
													 H3 "%s\n",
													 SG_string__sz(pStrRepoPath_k))  );
			SG_STRING_NULLFREE(pCtx, pStrRepoPath_k);
		}
		SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh_port_detail_k, "gid2", &pszGid_k)  );
		if (strcmp(pszGid_k, pszGid) != 0)
		{
			SG_ERR_CHECK(  SG_pendingtree__find_repo_path_by_gid(pCtx, pPendingTree, pszGid_k, &pStrRepoPath_k)  );
			SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOutput,
													 H3 "%s\n",
													 SG_string__sz(pStrRepoPath_k))  );
			SG_STRING_NULLFREE(pCtx, pStrRepoPath_k);
		}
	}

fail:
	SG_STRING_NULLFREE(pCtx, pStrRepoPath_k);
}

/**
 *
 */
static void _pt_resolve__format_temporary_name_warning(SG_context * pCtx,
													   SG_pendingtree * pPendingTree,
													   const SG_vhash * pvhIssue,
													   SG_string * pStrOutput)
{
	SG_bool bMadeUnique;

	SG_UNUSED(pPendingTree);

	SG_ERR_CHECK(  SG_vhash__get__bool(pCtx, pvhIssue, "made_unique_entryname", &bMadeUnique)  );
	if (!bMadeUnique)
		return;

	// because of a conflicts, collisions, or potential collisions we had to temporarily
	// rename the entry with a unique name.

	SG_ERR_CHECK(  SG_string__append__sz(pCtx, pStrOutput,
										 H2 "The entry was given a temporary name pending resolution.\n")  );

fail:
	return;
}

//////////////////////////////////////////////////////////////////

void SG_pendingtree__format_issue(SG_context * pCtx,
								  SG_pendingtree * pPendingTree,
								  const SG_vhash * pvhIssue,
								  SG_string * pStrOutput,
								  SG_string ** ppStrRepoPath_Returned)
{
	SG_string * pStrRepoPath = NULL;

	SG_NULLARGCHECK_RETURN( pPendingTree );
	SG_NULLARGCHECK_RETURN( pvhIssue );
	SG_NULLARGCHECK_RETURN( pStrOutput );
	// ppStrRepoPath_Returned is optional

	SG_ERR_CHECK(  _pt_resolve__format_header(pCtx, pPendingTree, pvhIssue, pStrOutput, &pStrRepoPath)  );

	SG_ERR_CHECK(  _pt_resolve__format_conflict_flags(pCtx, pPendingTree, pvhIssue, pStrOutput)  );
	SG_ERR_CHECK(  _pt_resolve__format_collision(pCtx, pPendingTree, pvhIssue, pStrOutput)  );
	SG_ERR_CHECK(  _pt_resolve__format_portability_warning(pCtx, pPendingTree, pvhIssue, pStrOutput)  );
	SG_ERR_CHECK(  _pt_resolve__format_temporary_name_warning(pCtx, pPendingTree, pvhIssue, pStrOutput)  );

	if (ppStrRepoPath_Returned)
	{
		*ppStrRepoPath_Returned = pStrRepoPath;
		pStrRepoPath = NULL;
	}

fail:
	SG_STRING_NULLFREE(pCtx, pStrRepoPath);
}

//////////////////////////////////////////////////////////////////

END_EXTERN_C;

#endif//H_SG_PENDINGTREE__PRIVATE_RESOLVE_ISSUES_H
