/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg_pendingtree__private_scan_dir.h
 *
 * @details
 *
 */

//////////////////////////////////////////////////////////////////

#ifndef H_SG_PENDINGTREE__PRIVATE_SCAN_DIR_H
#define H_SG_PENDINGTREE__PRIVATE_SCAN_DIR_H

BEGIN_EXTERN_C;

//////////////////////////////////////////////////////////////////

#if defined(DEBUG)
#	define TRACE_SCAN_DIR		0
#else
#	define TRACE_SCAN_DIR		0
#endif

//////////////////////////////////////////////////////////////////

static void sg_pendingtree__scan_dir(SG_context* pCtx,
									 SG_pendingtree* pPendingTree,
									 SG_pathname* pPathLocalDirectory,
									 const SG_bool bAdd,
									 const SG_bool bRemove,
									 const SG_bool bMarkImplicit,
									 SG_bool bRecursive,
									 struct sg_ptnode* ptn_given,
									 SG_bool bIncludeLostItems,
									 SG_bool bIncludeFoundItems,
									 const SG_file_spec* pFilespec);

//////////////////////////////////////////////////////////////////

struct _pt_sd
{
	SG_pendingtree *		pPendingTree;		// we do not own this
	SG_bool					bAdd;
	SG_bool					bRemove;
	SG_bool					bMarkImplicit;
	SG_bool					bRecursive;
	SG_bool					bIncludeLostItems;
	SG_bool					bIncludeFoundItems;

	const SG_file_spec *	pFilespec;

	struct sg_ptnode *		ptnDirectory;		// we do not own this

	SG_rbtree *				prbEntriesOnDisk;	// map[<entryname> --> _pt_sd_entry *]      we own this
	SG_pathname *			pPathDirectory;		// we own this
};

//////////////////////////////////////////////////////////////////

struct _pt_sd_entry
{
	SG_int64					mtime_ms;
	SG_uint64					size;
	SG_fsobj_perms				perms;		// only SG_FSOBJ_PERMS__MASK bits are defined
	SG_pathname *				pPath;			// we own this
	SG_treenode_submodule *		pTSM;			// we own this -- only set when we see a sub-dir that is a WD.
	SG_treenode_entry_type		tneType;
	SG_bool						bMatched;		// claimed by a ptnode during first pass of scan.
};

void _pt_sd_entry__free(SG_context * pCtx, struct _pt_sd_entry * pEntry)
{
	if (!pEntry)
		return;

	SG_PATHNAME_NULLFREE(pCtx, pEntry->pPath);
	SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pEntry->pTSM);
	SG_NULLFREE(pCtx, pEntry);
}

//////////////////////////////////////////////////////////////////

/**
 * A wrapper for SG_file_spec__should_include().
 *
 * Use the given Ignore/Include/Exclude rules to decide if we should
 * INCLUDE this entry.  This is based strictly upon:
 *      pData->pPathDirectory / pszEntryName
 *
 * The combined absolute pathname MAY OR MAY NOT actually exist on disk.
 */
static void _pt_sd__file_spec__should_include(SG_context * pCtx,
											  struct _pt_sd * pData,
											  const char * pszRepoPath,
											  SG_file_spec__match_result * pEval)
{
	SG_file_spec__match_result eval;

	SG_ERR_CHECK( SG_file_spec__should_include(pCtx, pData->pFilespec, pszRepoPath,
											   SG_FILE_SPEC__CONVERT_REPO_PATH | SG_FILE_SPEC__MATCH_TRAILING_SLASH,
											   &eval));

#if TRACE_SCAN_DIR
	SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
							   "_pt_sd__file_spec__should_include: [result 0x%x] %s\n",
							   (SG_uint32)eval, pszRepoPath)  );
#endif

	*pEval = eval;

fail:
	return;
}

/**
 * A wrapper for SG_file_spec__should_include().  This version should only be given a local path.
 *
 * Use the given Ignore/Include/Exclude rules to decide if we should
 * INCLUDE this entry.  This is based strictly upon:
 *      pData->pPathDirectory / pszEntryName
 *
 */
static void _pt_sd__file_spec__should_include__localpath(SG_context * pCtx,
											  struct _pt_sd * pData,
											  SG_pathname * pLocalPath,
											  SG_file_spec__match_result * pEval)
{
	SG_file_spec__match_result eval;

	SG_ERR_CHECK( SG_file_spec__should_include(pCtx, pData->pFilespec, SG_pathname__sz(pLocalPath),
											   SG_FILE_SPEC__MATCH_TRAILING_SLASH,
											   &eval));

#if TRACE_SCAN_DIR
	SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
							   "_pt_sd__file_spec__should_include__localpath: [result 0x%x] %s\n",
							   (SG_uint32)eval, SG_pathname__sz(pLocalPath))  );
#endif

	*pEval = eval;

fail:
	return;
}

//////////////////////////////////////////////////////////////////

static void _pt_log__add_item(SG_context * pCtx,
						   SG_pendingtree * pPendingTree,
						   const char * pszGid,
						   const char * pszPathDir,
						   const char * pszReason);

//////////////////////////////////////////////////////////////////

/**
 * We get called once for each entry in the directory.
 * We add the entryname and some simple info to an rbtree.
 */
static SG_dir_foreach_callback _pt_sd__read_directory_cb;

static void _pt_sd__read_directory_cb(SG_context * pCtx,
									  const SG_string * pStringEntryName,
									  SG_fsobj_stat * pfsStat,
									  void * pVoidData)
{
	struct _pt_sd * pData = (struct _pt_sd *)pVoidData;
	const char * pszEntryName = SG_string__sz(pStringEntryName);
	struct _pt_sd_entry * pEntry = NULL;

	SG_ERR_CHECK(  SG_alloc1(pCtx,pEntry)  );
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx,&pEntry->pPath,pData->pPathDirectory,pszEntryName)  );
	SG_ERR_CHECK(  sg_pendingtree__fsobjtype_to_treenodeentrytype(pCtx,pfsStat->type,&pEntry->tneType)  );
	if (pEntry->tneType == SG_TREENODEENTRY_TYPE_REGULAR_FILE)
	{
		pEntry->mtime_ms = pfsStat->mtime_ms;
		pEntry->size = pfsStat->size;
		//TODO:  Jeremy needs to know if this should be windows-only.
		pEntry->perms = pfsStat->perms;
	}
	if (pEntry->tneType == SG_TREENODEENTRY_TYPE_DIRECTORY)
	{
		SG_bool bIsWD = SG_FALSE;
		SG_ERR_CHECK(  SG_TREENODE_SUBMODULE__SNIFF_DIRECTORY(pCtx, pEntry->pPath, pData->pPendingTree->pRepo,
															  &bIsWD, &pEntry->pTSM)  );
		if (bIsWD)
			pEntry->tneType = SG_TREENODEENTRY_TYPE_SUBMODULE;
	}
	pEntry->bMatched = SG_FALSE;

#if TRACE_SCAN_DIR
	if (pEntry->tneType == SG_TREENODEENTRY_TYPE_REGULAR_FILE)
	{
		char bufTimestamp[100];

		SG_ERR_IGNORE(  SG_time__format_local__i64(pCtx, pEntry->mtime_ms, bufTimestamp, sizeof(bufTimestamp))  );

		SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
								   "_pt_sd__read_directory_cb: Found [type %d][%s][mtime %s]\n",
								   pEntry->tneType,
								   SG_pathname__sz(pEntry->pPath),
								   bufTimestamp)  );
	}
	else
	{
		SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
								   "_pt_sd__read_directory_cb: Found [type %d][%s]\n",
								   pEntry->tneType,
								   SG_pathname__sz(pEntry->pPath))  );
	}
#endif

	SG_ERR_CHECK(  SG_rbtree__add__with_assoc(pCtx,pData->prbEntriesOnDisk,pszEntryName,(void *)pEntry)  );
	pEntry = NULL;		// pData->prbEntriesOnDisk now owns it

	return;

fail:
	SG_ERR_IGNORE(  _pt_sd_entry__free(pCtx, pEntry)  );
}

/**
 * Build an rbtree containing info for each entry in the directory on disk.
 * The caller must free this.
 */
static void _pt_sd__read_directory(SG_context * pCtx,
								   struct _pt_sd * pData)
{
#if 0
	SG_fsobj_type fsObjType;
	SG_bool bExists;
#endif

#if TRACE_SCAN_DIR
	SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
							   "_pt_sd__read_directory: At top for [%s]\n",
							   SG_pathname__sz(pData->pPathDirectory))  );
#endif

#if 0
	//This is taken out because it's a perf hit.
	//The check should not be done on every folder that we walk through.
	//It should only be done on the top one.
	SG_ERR_CHECK_RETURN(  SG_fsobj__exists__pathname(pCtx,pData->pPathDirectory,&bExists,&fsObjType,NULL)  );
	if (!bExists)
	{
		*pbExists = SG_FALSE;
		*pbIsDirectory = SG_FALSE;
		return;
	}
	if (fsObjType != SG_FSOBJ_TYPE__DIRECTORY)
	{
		*pbExists = SG_TRUE;
		*pbIsDirectory = SG_FALSE;
		return;
	}
#endif

#if TRACE_SCAN_DIR
	SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
							   "_pt_sd__read_directory: Reading [%s]\n",
							   SG_pathname__sz(pData->pPathDirectory))  );
#endif

	SG_ERR_CHECK_RETURN(  SG_dir__foreach(pCtx, pData->pPathDirectory,
										  SG_DIR__FOREACH__SKIP_SPECIAL | SG_DIR__FOREACH__STAT,
										  _pt_sd__read_directory_cb,(void *)pData)  );
#if 0
	*pbExists = SG_TRUE;
	*pbIsDirectory = SG_TRUE;
#endif
}

//////////////////////////////////////////////////////////////////

/**
 * Open the given pathname and compute the HID on the current contents.
 *
 * You own the returned HID.
 */
static void _pt__compute_file_hid(SG_context * pCtx,
								  SG_pendingtree * pPendingTree,
								  const SG_pathname * pPath,
								  const char * pszGid,
								  SG_bool bUpdateTimestampCache,
								  char ** ppszHid)
{
	SG_file * pFile = NULL;
	char * pszHid = NULL;
	SG_fsobj_stat stat;

	SG_ERR_CHECK(  SG_file__open__pathname(pCtx, pPath, SG_FILE_RDONLY | SG_FILE_OPEN_EXISTING, SG_FSOBJ_PERMS__UNUSED, &pFile)  );
	SG_ERR_CHECK(  SG_repo__alloc_compute_hash__from_file(pCtx, pPendingTree->pRepo, pFile, &pszHid)  );
	SG_FILE_NULLCLOSE(pCtx, pFile);

	if (bUpdateTimestampCache)
	{
		SG_timestamp_cache * pTSC = NULL;	// we don't own this

		SG_ERR_CHECK(  SG_fsobj__stat__pathname(pCtx, pPath, &stat)  );
		SG_ERR_CHECK(  SG_pendingtree__get_timestamp_cache(pCtx, pPendingTree, &pTSC)  );
		SG_ERR_CHECK(  SG_timestamp_cache__add(pCtx, pTSC, pszGid, stat.mtime_ms, stat.size, pszHid)  );
	}
	
#if TRACE_SCAN_DIR
	if (bUpdateTimestampCache)
	{
		char buf_i64[40];
		
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "_pt__compute_file_hid: [gid %s] computed [hid %s] [mtime %s] for %s\n",
								   pszGid, pszHid, SG_int64_to_sz(stat.mtime_ms, buf_i64), SG_pathname__sz(pPath))  );
	}
	else
	{
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "_pt__compute_file_hid: [gid %s] computed [hid %s] for %s\n",
								   pszGid, pszHid, SG_pathname__sz(pPath))  );
	}
#endif

	*ppszHid = pszHid;
	return;

fail:
	SG_NULLFREE(pCtx, pszHid);
	SG_FILE_NULLCLOSE(pCtx, pFile);
}

//////////////////////////////////////////////////////////////////

/**
 * Set the current-hid field to the new value if necessary.
 * If the new value matches the baseline, just clear the
 * current value.
 */
static void _pt_sd__set_current_hid(SG_context * pCtx,
									struct sg_ptnode * ptnChild,
									const char * pszHid_new)
{
	const char * pszHid_baseline;

	if (ptnChild->pszCurrentHid)
	{
		if (strcmp(pszHid_new, ptnChild->pszCurrentHid) == 0)
			return;

		ptnChild->pszCurrentHid = NULL;	// string belongs to the pool so we don't need to free it
	}

	if (ptnChild->pBaselineEntry)
	{
		SG_ERR_CHECK(  SG_treenode_entry__get_hid_blob(pCtx, ptnChild->pBaselineEntry, &pszHid_baseline)  );
		if (strcmp(pszHid_new, pszHid_baseline) == 0)
			return;
	}

	SG_ERR_CHECK(  SG_strpool__add__sz(pCtx, ptnChild->pPendingTree->pStrPool, pszHid_new, &ptnChild->pszCurrentHid)  );

fail:
	return;
}

//////////////////////////////////////////////////////////////////

/**
 * Our caller found a sub-directory on the disk that contains a WD root
 * that has the same entryname as a ptnode of type submodule and it
 * wants to ASSUME that they are a match and should be associated with
 * each other.
 *
 * This is a little bit of a sanity check to make sure that they both
 * refer to the same REPO/WD.  That is, if they refer to different CSETs
 * (and hence have a different HID) would it be possible to do some kind
 * of UPDATE operation to get from one CSET to the other?
 *
 * Granted, this is a serious edge case.  For example, if they delete
 * the .sgdrawer and create a .hg (without telling us) our caller should
 * probably throw a fit....  And treat it as a 'veracity-submodule-delete'
 * and a 'hg-submodule-add' rather than a plain 'submodule-update'.
 */
static void _pt_sd__are_submodules_comparable(SG_context * pCtx,
											  SG_pendingtree * pPendingTree,
											  struct sg_ptnode * ptnChild,
											  struct _pt_sd_entry * pEntry,
											  SG_bool * pbSimilar)
{
	const char * pszHidBaseline = NULL;
	SG_treenode_submodule * pTSM_baseline = NULL;

	SG_ASSERT(  (ptnChild->type == SG_TREENODEENTRY_TYPE_SUBMODULE)  );
	SG_ASSERT(  (pEntry->tneType == SG_TREENODEENTRY_TYPE_SUBMODULE)  );
	SG_ASSERT(  (pEntry->pTSM)  );

	SG_ERR_CHECK(  sg_ptnode__get_old_hid(pCtx, ptnChild, &pszHidBaseline)  );
	if (pszHidBaseline)
		SG_ERR_CHECK(  SG_TREENODE_SUBMODULE__LOAD_FROM_REPO(pCtx, pPendingTree->pRepo, pszHidBaseline, &pTSM_baseline)  );

	SG_ERR_CHECK(  SG_treenode_submodule__are_comparable(pCtx, pTSM_baseline, pEntry->pTSM, pbSimilar)  );

fail:
	SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pTSM_baseline);
}

//////////////////////////////////////////////////////////////////

/**
 * Our caller is trying to fill in the "current" fields in the ptnode
 * so that later we can see if any of them are different from the "baseline"
 * fields.  If there are differences, we say the file/symlink/directory is
 * "dirty".
 *
 * Our goal is to see if the file content has changed from the baseline.
 * This requires re-computing the HID on the file content, WHICH IS
 * VERY EXPENSIVE to do on every scan-dir.
 *
 * So, we try to use the file's mtime to see if the content could have
 * changed.
 *
 * When the baseline was established/populated (by GET, COMMIT, and
 * maybe UPDATE), we recorded each file's mtime in the timestamp cache.
 * So we have (baseline-hid, baseline-mtime).
 *
 * In the ptnode, if the file content is clean (or we haven't scanned a
 * change yet), current-hid is null.  If a previous scan-dir determined
 * that the file content changed, it stored the then-current-hid in the
 * ptnode *AND* updated the timestamp cache with the then-current-mtime.
 *
 * So, our task here is to determine if the now-current-mtime is different
 * from the cached mtime.  If so, we compute and store the now-current-hid
 * and update the timestamp cache.
 *
 * TODO 2010/05/27 In the event that the now-current-hid matches the
 * TODO            baseline-hid (say they '/usr/bin/touch' a file), we
 * TODO            have the option of either clearing the current-hid in
 * TODO            the ptnode or setting it to the computed value.  The
 * TODO            pre-timestamp-cache code always set it.  I'm not sure
 * TODO            that it matters one way or the other, but I'm going
 * TODO            clear it so that sg_ptnode__is_file_dirty() is a
 * TODO            little faster.
 *
 * NOTE: It is possible that the timestamp cache is not present or
 * NOTE: doesn't have an entry for this file (because we added it after
 * NOTE: we started dogfooding).  We silently allow this and force the
 * NOTE: HID computation.  And because we always update the cache
 * NOTE: whenever we do the computation, we will back-fill the cache.
 */
static void _pt_sd__update_ptnode__file_content(SG_context * pCtx,
												SG_pendingtree * pPendingTree,
												struct sg_ptnode * ptnChild,
												const SG_pathname * pPath,
												SG_int64 * pmtime_ms,
												SG_uint64 * pSize)
{
	SG_timestamp_cache * pTSC = NULL;			// we don't own this
	const SG_timestamp_data * pTD = NULL;		// we don't own this
	char * pszHid_now_current = NULL;
	SG_fsobj_stat stat_now_current;
	SG_bool bCachedTimeValid = SG_FALSE;
	SG_int64 currentModTime = 0;
	SG_uint64 currentSize = 0;

	SG_ASSERT(  (ptnChild->type == SG_TREENODEENTRY_TYPE_REGULAR_FILE)  );

	if (pmtime_ms == NULL || pSize == NULL)
	{
		SG_ERR_CHECK(  SG_fsobj__stat__pathname(pCtx, pPath, &stat_now_current)  );
		currentSize = stat_now_current.size;
		pmtime_ms = &currentModTime;
		pSize = &currentSize;
	}

	SG_ERR_CHECK(  SG_pendingtree__get_timestamp_cache(pCtx, pPendingTree, &pTSC)  );
	SG_ERR_CHECK(  SG_timestamp_cache__is_valid(pCtx, pTSC, ptnChild->pszgid,
												*pmtime_ms,
												*pSize,
												&bCachedTimeValid,
												&pTD)  );
	if (bCachedTimeValid)
	{
		const char * pszHid_cached;

		// The timestamp_cache thinks that the cache entry is valid for this file.

		SG_ERR_CHECK(  SG_timestamp_data__get_hid__ref(pCtx, pTD, &pszHid_cached)  );

#if 0 && defined(DEBUG)
		// Compute a live hash (defeating the purpose of the cache, I know)
		// and verify that it matches what we think it is.

		SG_ERR_CHECK(  _pt__compute_file_hid(pCtx, pPendingTree, pPath, ptnChild->pszgid,
											 SG_FALSE, &pszHid_now_current)  );
		if (strcmp(pszHid_now_current, pszHid_cached) != 0)
			SG_ERR_THROW2(  SG_ERR_ASSERT,
							(pCtx, "Assumed timestamp_cache was valid, but wasn't for [key %s] [%s] [now %s] [cached %s].",
							 ptnChild->pszgid,
							 SG_pathname__sz(pPath),
							 pszHid_now_current,
							 pszHid_cached)  );

		// since the timestamp_cache gets written more often than the pendingtree,
		// (we don't save the pendingtree on a STATUS, for example).  we shouldn't
		// make any assumptions about whether the cached/computed hid value matches
		// either ptnChild->pszCurrentHid or ptnChild->pBaselineEntry->hid.
#endif

		SG_ERR_CHECK(  _pt_sd__set_current_hid(pCtx, ptnChild, pszHid_cached)  );
	}
	else
	{
		// re-compute the hash (and store it in the timestamp_cache)
		// and update ptnChild->pszCurrentHid.

		SG_ERR_CHECK(  _pt__compute_file_hid(pCtx, pPendingTree, pPath, ptnChild->pszgid,
											 SG_TRUE, &pszHid_now_current)  );
		SG_ERR_CHECK(  _pt_sd__set_current_hid(pCtx, ptnChild, pszHid_now_current)  );
		SG_NULLFREE(pCtx, pszHid_now_current);
	}
	
fail:
	SG_NULLFREE(pCtx, pszHid_now_current);
}

static void _pt_sd__update_ptnode__symlink_content(SG_context * pCtx,
												   SG_pendingtree * pPendingTree,
												   struct sg_ptnode * ptnChild,
												   const SG_pathname * pPath)
{
	char* pszHid = NULL;
	SG_string * pStringLink = NULL;

	SG_ASSERT(  (ptnChild->type == SG_TREENODEENTRY_TYPE_SYMLINK)  );

	// TODO 2010/05/27 Do we need to think about using the timestamp cache for symlinks?
	// TODO            Or does that even make sense?  Or do we have enough symlinks to bother?

	SG_ERR_CHECK(  SG_fsobj__readlink(pCtx, pPath, &pStringLink)  );
	SG_ERR_CHECK(  SG_repo__alloc_compute_hash__from_string(pCtx, pPendingTree->pRepo, pStringLink, &pszHid)  );
	SG_STRING_NULLFREE(pCtx, pStringLink);

	SG_ERR_CHECK(  _pt_sd__set_current_hid(pCtx, ptnChild, pszHid)  );

fail:
	SG_NULLFREE(pCtx, pszHid);
	SG_STRING_NULLFREE(pCtx, pStringLink);
}

static void _pt_sd__update_ptnode__submodule(SG_context * pCtx,
											 SG_pendingtree * pPendingTree,
											 struct sg_ptnode * ptnChild,
											 struct _pt_sd_entry * pEntry)
{
	const char * pszHid = NULL;

	SG_ASSERT(  (ptnChild->type == SG_TREENODEENTRY_TYPE_SUBMODULE)  );
	SG_ASSERT(  (pEntry->pTSM)  );

	SG_ERR_CHECK(  SG_treenode_submodule__get_id_ref(pCtx, pEntry->pTSM, &pszHid)  );
	SG_ERR_CHECK(  SG_strpool__add__sz(pCtx, pPendingTree->pStrPool, pszHid, &ptnChild->pszCurrentHid)  );

	SG_TREENODE_SUBMODULE_NULLFREE(pCtx, ptnChild->pTSM);
	SG_TREENODE_SUBMODULE__ALLOC__COPY(pCtx, &ptnChild->pTSM, pEntry->pTSM);

fail:
	return;
}

/**
 * We have determined that the given ptnode and the given entry on disk
 * should be associated with each other.  Update the various fields in
 * the ptnode from what is currently on disk.
 */
static void _pt_sd__update_ptnode(SG_context * pCtx,
								  struct _pt_sd * pData,
								  struct sg_ptnode * ptnChild,
								  struct _pt_sd_entry * pEntry)
{
#if TRACE_SCAN_DIR
	SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
							   ("_pt_sd__update_ptnode: dir [gid %s][%s]\n"
								"                       sub [gid %s][%s]\n"),
							   pData->ptnDirectory->pszgid,
							   SG_pathname__sz(pData->pPathDirectory),
							   ptnChild->pszgid,
							   SG_pathname__sz(pEntry->pPath))  );
#endif

	SG_ASSERT(  (pEntry->bMatched == SG_FALSE)  );

	SG_ASSERT(  (ptnChild->pCurrentParent == pData->ptnDirectory)  );

	SG_ASSERT(  ((ptnChild->saved_flags & sg_PTNODE_SAVED_FLAG_DELETED) == 0)  );
	SG_ASSERT(  (ptnChild->bDeletedAfterAdded == SG_FALSE)  );
	SG_ASSERT(  ((ptnChild->temp_flags & sg_PTNODE_TEMP_FLAG_IMPLICIT_DELETE) == 0)  );

	// get the XATTRS and ATTRBITS from the disk and update ptnChild.

#ifdef SG_BUILD_FLAG_FEATURE_XATTR
	SG_ERR_CHECK(  sg_ptnode__scan_xattrs(pCtx, ptnChild, pEntry->pPath)  );
#endif
	SG_ERR_CHECK(  sg_ptnode__scan_attrbits(pCtx, ptnChild, pEntry->tneType, pEntry->perms)  );
	

	switch (ptnChild->type)
	{
	case SG_TREENODEENTRY_TYPE_REGULAR_FILE:
		SG_ERR_CHECK(  _pt_sd__update_ptnode__file_content(pCtx, pData->pPendingTree, ptnChild, pEntry->pPath, &pEntry->mtime_ms, &pEntry->size)  );
		break;

	case SG_TREENODEENTRY_TYPE_DIRECTORY:
		if (pData->bRecursive)
		{
			SG_ERR_CHECK(  sg_ptnode__load_dir_entries(pCtx, ptnChild)  );
			SG_ERR_CHECK(  sg_pendingtree__scan_dir(pCtx,
													pData->pPendingTree,
													pEntry->pPath,
													pData->bAdd,
													pData->bRemove,
													pData->bMarkImplicit,
													pData->bRecursive,
													ptnChild,
													pData->bIncludeLostItems,		// bIncludeLostItems
													pData->bIncludeFoundItems,		// bIncludeFoundItems
													pData->pFilespec)  );
		}
		break;

	case SG_TREENODEENTRY_TYPE_SYMLINK:
		SG_ERR_CHECK(  _pt_sd__update_ptnode__symlink_content(pCtx, pData->pPendingTree, ptnChild, pEntry->pPath)  );
		break;

	case SG_TREENODEENTRY_TYPE_SUBMODULE:
		SG_ERR_CHECK(  _pt_sd__update_ptnode__submodule(pCtx, pData->pPendingTree, ptnChild, pEntry)  );
		break;

	default:
		SG_ERR_THROW(  SG_ERR_INVALID_PENDINGTREE_OBJECT_TYPE  );
	}

fail:
	return;
}

//////////////////////////////////////////////////////////////////

/**
 * We get called for each ptnode that should be converted to a DELETE
 * of some kind because we could not find a corresponding entry in the
 * current directory.  THIS IS NOT A VERB FOR DOING A DELETE, but rather
 * a routine to mark various nodes of the tree in memory as deleted
 * because we couldn't find something on the disk.
 *
 * Note that we DO NOT actually delete the ptnode from the parent's
 * list (ptnChild->pCurrentParent->prbItems), we only set some flags.
 *
 * We are possibly in a self-recursive call (deleting the contents of
 * a deleted directory).  We CANNOT respect the Ignore/Include/Exclude
 * rules -- saying don't delete the *.h inside a directory that you
 * ALREADY deleted doesn't make any sense.  Our caller should inspect
 * the Ignore/Include/Exclude stuff before we begin this dive.
 */
static SG_rbtree_foreach_callback _pt_sd__mark_ptnode_as_deleted;

static void _pt_sd__mark_ptnode_as_deleted(SG_context * pCtx,
										   SG_UNUSED_PARAM( const char * pszKey_GidObject ),
										   void * pVoidValue,
										   void * pVoidData)
{
	struct sg_ptnode * ptnChild = (struct sg_ptnode *)pVoidValue;
	struct _pt_sd * pData = (struct _pt_sd *)pVoidData;
	struct _pt_sd pt_sd;
	const char * pszName = NULL;

	SG_UNUSED( pszKey_GidObject );

	SG_ASSERT(  (ptnChild->pCurrentParent == pData->ptnDirectory)  );

	if (ptnChild->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY)		// this is only needed for recursive call
		return;

	SG_ERR_IGNORE(  sg_ptnode__get_name(pCtx,ptnChild,&pszName)  );

#if TRACE_SCAN_DIR
	SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
							   ("_pt_sd__mark_ptnode_as_deleted: dir [gid %s][%s]\n"
								"                                sub [gid %s][%s]\n"),
							   pData->ptnDirectory->pszgid,
							   SG_pathname__sz(pData->pPathDirectory),
							   ptnChild->pszgid,
							   pszName)  );
#endif

	if (ptnChild->saved_flags & sg_PTNODE_SAVED_FLAG_DELETED)
	{
		// if we already know ptnChild is DELETED, we don't update the flags.
		// this keeps us from incorrectly picking up an IMPLICIT flag during
		// a recursive dive for things that we know were EXPLICITLY deleted.
	}
	else
	{
		ptnChild->saved_flags |= sg_PTNODE_SAVED_FLAG_DELETED;
		ptnChild->bDeletedAfterAdded = (ptnChild->pBaselineEntry == NULL);
		if (pData->bMarkImplicit)
			ptnChild->temp_flags |= sg_PTNODE_TEMP_FLAG_IMPLICIT_DELETE;
	}

	if (ptnChild->type != SG_TREENODEENTRY_TYPE_DIRECTORY)
		return;

	if (!pData->bRecursive)
		return;

	SG_ERR_CHECK_RETURN(  sg_ptnode__load_dir_entries(pCtx, ptnChild)  );

	// build a clone of the pData so that we can change the parent directory pathname/ptnode
	// when we hand it to our children.  we don't bother with prbEntriesOnDisk because
	// our directory has been deleted and doesn't exist on disk.
	//
	// TODO i'm passing down the augmented pathname name in a new pt_sd.  currently,
	// TODO i'm only doing this so that the trace can print it.  i don't know yet if
	// TODO the include/exclude stuff will need it to do pathname-wildcard-matching.
	// TODO if not, we could ifdef part of this or pass null for the pathname.

	memset(&pt_sd,0,sizeof(pt_sd));
	pt_sd.pPendingTree  = pData->pPendingTree;
	pt_sd.bAdd			= pData->bAdd;
	pt_sd.bRemove    	= pData->bRemove;
	pt_sd.bMarkImplicit = pData->bMarkImplicit;
	pt_sd.bRecursive    = pData->bRecursive;
	pt_sd.pFilespec     = pData->pFilespec;
	pt_sd.ptnDirectory  = ptnChild;
	//pt_sd.prbEntriesOnDisk is null
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx, &pt_sd.pPathDirectory, pData->pPathDirectory, pszName)  );

	SG_ERR_CHECK(  SG_rbtree__foreach(pCtx, ptnChild->prbItems, _pt_sd__mark_ptnode_as_deleted, (void *)&pt_sd)  );

fail:
	SG_PATHNAME_NULLFREE(pCtx, pt_sd.pPathDirectory);
}

//////////////////////////////////////////////////////////////////

/**
 * We get called for each ptnode in a directory (based on GID).
 * We are also given a list of the entrynames currently found on disk.
 *
 * Our job is to associate an entry from the disk list with this ptnode
 * and "claim" it and update the various fields on the ptnode to reflect
 * any changes on disk.  This includes changes to XATTRS/ATTRBITS, file
 * contents, etc.
 *
 * We only look at the given ptnode and don't worry about siblings
 * which may share the same entryname.
 *
 * We DO NOT modify the ptnChild->pCurrentParent->prbItems because we
 * are in an iteration.
 */
static SG_rbtree_foreach_callback _pt_sd__inspect_child_ptnode_cb;

static void _pt_sd__inspect_child_ptnode_cb(SG_context * pCtx, const char * pszKey_GidObject, void * pVoidValue, void * pVoidData)
{
	struct sg_ptnode * ptnChild = (struct sg_ptnode *)pVoidValue;
	struct _pt_sd * pData = (struct _pt_sd *)pVoidData;
	const char * pszEntryName = NULL;
	struct _pt_sd_entry * pEntry = NULL;
	SG_string * pStringRepoPath = NULL;
	SG_bool bFoundOnDisk;
	SG_file_spec__match_result eval = SG_FILE_SPEC__RESULT__MAYBE;

	SG_ASSERT(  (strcmp(pszKey_GidObject,ptnChild->pszgid) == 0)  );

	if (ptnChild->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY)
	{
		// this ptnode is a phantom.  ignore it.

		return;
	}

	SG_ERR_CHECK(  sg_ptnode__get_name(pCtx,ptnChild,&pszEntryName)  );
	if (!pszEntryName || !*pszEntryName)
	{
#if TRACE_SCAN_DIR
		SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
								   ("_pt_sd__inspect_child_ptnode: [gid %s] has no name\n"),
								   ptnChild->pszgid)  );
#endif
		// TODO 2011/04/07 This should not happen.  If it does, it means
		// TODO            that there's something wrong in wd.json.
		SG_ERR_THROW( SG_ERR_NOTIMPLEMENTED );
	}

	// look up the entryname in the directory on disk.

	SG_ERR_CHECK(  SG_rbtree__find(pCtx,pData->prbEntriesOnDisk,pszEntryName,&bFoundOnDisk,(void **)&pEntry)  );

	if (bFoundOnDisk == SG_FALSE)
	{
		// Compute the repo-path for this ptnode (without regard for whether
		// it currently exists on disk) and see if it matches any of the Ignore/Include/Exclude
		// rules.

		SG_ERR_CHECK(  sg_ptnode__get_repo_path(pCtx, ptnChild, &pStringRepoPath)  );
		SG_ERR_CHECK(  _pt_sd__file_spec__should_include(pCtx,pData,SG_string__sz(pStringRepoPath),&eval)  );
		SG_STRING_NULLFREE(pCtx, pStringRepoPath);
	}
	else
	{
		// Since it's found on disk, we can call should_include with the disk path, which is faster.
		SG_ERR_CHECK(  _pt_sd__file_spec__should_include__localpath(pCtx,pData,pEntry->pPath,&eval)  );
	}

#if TRACE_SCAN_DIR
	SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
							   ("_pt_sd__inspect_child_ptnode: dir [gid %s][%s]\n"
								"                              kid [gid %s][%s][saved_flags %d][temp_flags %d][eval %d]\n"),
							   pData->ptnDirectory->pszgid,
							   SG_pathname__sz(pData->pPathDirectory),
							   ptnChild->pszgid,
							   pszEntryName,
							   ptnChild->saved_flags,
							   ptnChild->temp_flags,
							   eval)  );
#endif

	// look up the entryname in the directory on disk.

	if (ptnChild->saved_flags & sg_PTNODE_SAVED_FLAG_DELETED)
	{
		// this ptnode currently thinks the item has been deleted, but there is an apparent match for it
		// on disk.  we have to assume that it is something new that we may want to add and which should
		// be given a new GID.  we'll do this in a later step.

		if ((bFoundOnDisk) && (pEntry->tneType == ptnChild->type))	// TODO i'm not sure we actually need this.
			return;

		if (!pData->bRemove)
			return;

		// technically, we don't need to do anything for this ptnode (we already think it is deleted).
		// but if it is a directory, we want to dive and delete anything within it that isn't already dead.

		if (ptnChild->type != SG_TREENODEENTRY_TYPE_DIRECTORY)
			return;

		if (!pData->bRecursive)
			return;

		// TODO 2010/06/07 I'm wondering if all 5 cases are actually the same and all
		// TODO            are just "return".

		switch (eval)
		{
		default:
			SG_ASSERT(  (0)  );

		case SG_FILE_SPEC__RESULT__EXCLUDED:
		case SG_FILE_SPEC__RESULT__IGNORED:
			// if this ptnode represents a pathname that should be ignored/excluded, we stop now.
			// this case is a little weird because we really should dive into a deleted directory
			// and make sure that everything within it is deleted too.
			return;

		case SG_FILE_SPEC__RESULT__INCLUDED:
		case SG_FILE_SPEC__RESULT__IMPLIED:
			// TODO 2010/06/07 Confirm that we really need to do this.  We already think it is
			// TODO            deleted.
			if (pData->bIncludeLostItems == SG_TRUE)
				SG_ERR_CHECK(  _pt_sd__mark_ptnode_as_deleted(pCtx,pszKey_GidObject,ptnChild,pData)  );
			return;

		case SG_FILE_SPEC__RESULT__MAYBE:
			// There were INCLUDES, but none matched this directory.
			//
			// We already know that the ptnode thinks it has been deleted and we didn't find a
			// directory on the disk with this name, so in theory all of the ptnodes within
			// ptnChild->prbItems should be deleted too.
			//
			// TODO 2010/06/07 Should we verify this?
			return;
		}
	}
	else							// we currently think the item is present
	{
		// i'm going to assume that we always update the ptnode regardless of
		// the Ignore/Include/Exclude status.  that is, the --include is meant
		// for add/remove and not is-dirty.

		if ((bFoundOnDisk) && (pEntry->tneType == ptnChild->type))
		{
			// this PTNODE has a correspeonding ENTRY on disk with the same ENTRYNAME
			// AND they have the same type.  for files/symlinks/directories we have
			// to ASSUME that they should be associated with each other because we
			// don't have any other information.  for submodules, we do have a little
			// extra info.

			SG_bool bAssumeAssociated = SG_TRUE;

			if (pEntry->tneType == SG_TREENODEENTRY_TYPE_SUBMODULE)
			{
				SG_bool bComparableSubmodules;

				SG_ERR_CHECK(  _pt_sd__are_submodules_comparable(pCtx,
																 pData->pPendingTree,
																 ptnChild,
																 pEntry,
																 &bComparableSubmodules)  );
				if (!bComparableSubmodules)
					bAssumeAssociated = SG_FALSE;
			}
			
			if (bAssumeAssociated)
			{
				SG_ERR_CHECK(  _pt_sd__update_ptnode(pCtx,pData,ptnChild,pEntry)  );
				pEntry->bMatched = SG_TRUE;
				return;
			}
		}

		// either there is no entry with this entryname on disk or the type of the entry on
		// disk is different than what we were expecting or the submodules are incompatable.
		//
		// either way, we have nothing to associate with this ptnode and must consider the
		// ptnode to be a DELETE.  (if it was found on disk, we DO NOT claim it.)

		if (!pData->bRemove)
			return;

		// TODO 2010/06/07 Does it make any sense to treat these 5 cases any differently?
		// TODO            I mean we have established that we have a ptnode for a directory,
		// TODO            we haven't yet marked the ptnode directory as deleted, we know
		// TODO            that the directory does not exist on disk (or is not a directory),
		// TODO            so even if they said --ignore="foo.dir", foo.dir is still gone.
		// TODO            But I guess they could use thot delay us from discovering this?
		// TODO            This needs more thought.

		switch (eval)
		{
		default:
			SG_ASSERT(  (0)  );

		case SG_FILE_SPEC__RESULT__EXCLUDED:
		case SG_FILE_SPEC__RESULT__IGNORED:
			// if this ptnode represents a pathname that should be ignored/excluded, we stop now.
			return;

		case SG_FILE_SPEC__RESULT__INCLUDED:
		case SG_FILE_SPEC__RESULT__IMPLIED:
			if (pData->bIncludeLostItems == SG_TRUE)
				SG_ERR_CHECK(  _pt_sd__mark_ptnode_as_deleted(pCtx,pszKey_GidObject,ptnChild,pData)  );
			return;

		case SG_FILE_SPEC__RESULT__MAYBE:
			// There were INCLUDES, but none matched this directory.
			return;
		}
	}

fail:
	SG_STRING_NULLFREE(pCtx, pStringRepoPath);
	return;
}

//////////////////////////////////////////////////////////////////

static SG_rbtree_foreach_callback _pt_sd__create_new_ptnode_for_unclaimed_entry_cb;

static void _pt_sd__create_new_ptnode_for_unclaimed_entry_cb(SG_context * pCtx, const char * pszKey_EntryName, void * pVoidValue, void * pVoidData)
{
	struct _pt_sd_entry * pEntry = (struct _pt_sd_entry *)pVoidValue;
	struct _pt_sd * pData = (struct _pt_sd *)pVoidData;
	struct sg_ptnode * ptnChild;
	SG_string * pStringRepoPath = NULL;
	SG_rbtree_iterator * pIter = NULL;
	SG_file_spec__match_result eval = SG_FILE_SPEC__RESULT__MAYBE;

	if (pEntry->bMatched)
		return;
	if (pData->bIncludeFoundItems == SG_FALSE)
		return;

	SG_ERR_CHECK(  SG_workingdir__wdpath_to_repopath(pCtx, pData->pPendingTree->pPathWorkingDirectoryTop, pEntry->pPath, SG_TRUE, &pStringRepoPath)  );
	SG_ERR_CHECK(  _pt_sd__file_spec__should_include(pCtx, pData, SG_string__sz(pStringRepoPath), &eval)  );

#if TRACE_SCAN_DIR
	SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
							   "_pt_sd__create_new_ptnode: [eval 0x%x][%s]\n",
							   eval, SG_string__sz(pStringRepoPath))  );
#endif

	switch (eval)
	{
	default:		// quiets compiler
		SG_ASSERT(  (0)  );

	case SG_FILE_SPEC__RESULT__EXCLUDED:
	case SG_FILE_SPEC__RESULT__IGNORED:
		goto done;

	case SG_FILE_SPEC__RESULT__INCLUDED:
	case SG_FILE_SPEC__RESULT__IMPLIED:
		// We have a new (to us) entry on the disk that we want to either
		// list as ADDED or FOUND (this depends upon pData->bMarkImplicit).
		// We respect the passed in setting.
		goto do_create_item;

	case SG_FILE_SPEC__RESULT__MAYBE:
		if (pEntry->tneType != SG_TREENODEENTRY_TYPE_DIRECTORY)
			goto done;

		if (!pData->bRecursive)
			goto done;

		// there were INCLUDES and this directory did not match one of them.
		// so we don't want to ADD this item, but we want to provisionally
		// add a ptnode for it so that we can dive and maybe find some files
		// within it that should be added (which would cause us to be officially
		// added.
		goto do_create_item;
	}

	//////////////////////////////////////////////////////////////////
	// NOT REACHED
	//////////////////////////////////////////////////////////////////

do_create_item:
	SG_ERR_CHECK(  sg_ptnode__create_item(pCtx, pData->ptnDirectory, pszKey_EntryName,
										  pData->bAdd, pData->bMarkImplicit,
										  pEntry->tneType, pEntry->pTSM,
										  &ptnChild)  );
	SG_ERR_CHECK(  _pt_log__add_item(pCtx, pData->pPendingTree, ptnChild->pszgid, SG_pathname__sz(pEntry->pPath), "foundDuringScan" )  );
#if TRACE_SCAN_DIR
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "_pt_sd__create_new_ptnode: created [%s] for %s\n",
							   ptnChild->pszgid,
							   SG_pathname__sz(pEntry->pPath))  );
#endif
#ifdef SG_BUILD_FLAG_FEATURE_XATTR
	SG_ERR_CHECK(  sg_ptnode__scan_xattrs(pCtx, ptnChild, pEntry->pPath)  );
#endif
	SG_ERR_CHECK(  sg_ptnode__scan_attrbits(pCtx, ptnChild, pEntry->tneType, pEntry->perms)  );

	if ((pData->bRecursive) && (ptnChild->type == SG_TREENODEENTRY_TYPE_DIRECTORY))
	{
		SG_ERR_CHECK(  SG_RBTREE__ALLOC(pCtx, &ptnChild->prbItems)  );
		SG_ERR_CHECK(  sg_pendingtree__scan_dir(pCtx,
												pData->pPendingTree,
												pEntry->pPath,
												pData->bAdd,
												pData->bRemove,
												pData->bMarkImplicit,
												pData->bRecursive,
												ptnChild,
												pData->bIncludeLostItems,		// bIncludeLostItems
												pData->bIncludeFoundItems,		// bIncludeFoundItems
												pData->pFilespec)  );

		if ((eval == SG_FILE_SPEC__RESULT__MAYBE) && (!pData->bMarkImplicit) && (ptnChild->temp_flags & sg_PTNODE_TEMP_FLAG_IMPLICIT_ADD))
		{
			// we provisionally created ptnChild for the current directory so that
			// we could scan the contents of the directory.  See if any of the
			// children were actually added and convert our status to actual.
			// it not, leave it a FOUND directory.
			//
			// TODO 2010/06/07 would it be better to remove any of the implicit
			// TODO            items within and then see if we have an empty prbItems?
			// TODO            We could also un-add this directory from our parent.

			SG_bool b;
			struct sg_ptnode * ptn_k;

			SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pIter, ptnChild->prbItems, &b, NULL, (void **)&ptn_k)  );
			while (b)
			{
				if ((ptn_k->temp_flags & sg_PTNODE_TEMP_FLAG_IMPLICIT_ADD) == 0)
				{
					ptnChild->temp_flags &= ~sg_PTNODE_TEMP_FLAG_IMPLICIT_ADD;
					break;
				}
				SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pIter, &b, NULL, (void **)&ptn_k)  );
			}
		}
	}

done:
	;
fail:
	SG_RBTREE_ITERATOR_NULLFREE(pCtx, pIter);
	SG_STRING_NULLFREE(pCtx, pStringRepoPath);
}

//////////////////////////////////////////////////////////////////

/**
 * The job of the "scan" operation is to fix the pending tree to match
 * the working directory.  Flags allow the caller to control whether
 * we want everything to be included or maybe we just want to scan for
 * some of the changes instead of all of them.
 *
 * When used for the purpose of a command like 'sg status', we expect
 * that the resulting pending tree will not be saved to disk.
 */
static void sg_pendingtree__scan_dir(SG_context* pCtx,
									 SG_pendingtree* pPendingTree,
									 SG_pathname* pPathLocalDirectory,
									 const SG_bool bAdd,
									 const SG_bool bRemove,
									 const SG_bool bMarkImplicit,
									 SG_bool bRecursive,
									 struct sg_ptnode* ptn_given,
									 SG_bool bIncludeLostItems,
									 SG_bool bIncludeFoundItems,
									 const SG_file_spec* pFilespec)
{
	struct _pt_sd pt_sd;
	struct sg_ptnode * ptn = NULL;
	
	if (ptn_given)
		ptn = ptn_given;
	else
		SG_ERR_CHECK_RETURN(  sg_pendingtree__local_path_to_ptnode(pCtx, pPendingTree, pPathLocalDirectory, SG_TRUE, &ptn)  );

	//TODO remove this when we figure out how addremove should handle a path not
	//int the pending tree
	if (!ptn)
	{
		SG_ERR_THROW2_RETURN(  SG_ERR_NOTIMPLEMENTED,
							   (pCtx, "TODO sg_pendingtree__scan_dir has null ptn with [%s].", SG_pathname__sz(pPathLocalDirectory))  );
		return;
	}

	SG_ASSERT(	(ptn->type == SG_TREENODEENTRY_TYPE_DIRECTORY)	);

	if (!ptn->prbItems)
		SG_ERR_CHECK_RETURN(  sg_ptnode__load_dir_entries(pCtx, ptn)  );

	// build all the baggage that we need to dive with.

	pt_sd.pPendingTree	   = pPendingTree;
	pt_sd.bAdd			   = bAdd;
	pt_sd.bRemove	   	   = bRemove;
	pt_sd.bMarkImplicit    = bMarkImplicit;
	pt_sd.bRecursive	   = bRecursive;
	pt_sd.pFilespec        = pFilespec;
	pt_sd.ptnDirectory	   = ptn;
	pt_sd.prbEntriesOnDisk = NULL;
	pt_sd.pPathDirectory   = NULL;
	pt_sd.bIncludeLostItems= bIncludeLostItems;
	pt_sd.bIncludeFoundItems= bIncludeFoundItems;
	SG_ERR_CHECK(  SG_RBTREE__ALLOC(pCtx, &pt_sd.prbEntriesOnDisk)	);
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__COPY(pCtx, &pt_sd.pPathDirectory, pPathLocalDirectory)  );

	// Read the contents of the directory as it currently exists on disk (if it exists).


	SG_ERR_CHECK(  _pt_sd__read_directory(pCtx, &pt_sd)  );

	// Iterate over all of the PTNODES (indexed by GID rather than ENTRYNAME) for this directory.
	// Match up the ptnodes with an ENTRY in the prbEntriesOnDisk and do the right thing.  This
	// takes care of entries that were changed or deleted from the disk.

	SG_ERR_CHECK(  SG_rbtree__foreach(pCtx, ptn->prbItems, _pt_sd__inspect_child_ptnode_cb, (void *)&pt_sd)	 );

	// Iterate over all of the ENTRIES that we found in the directory and deal with the "unclaimed" ones.
	// This handled entries that were added

	if (bAdd)
		SG_ERR_CHECK(  SG_rbtree__foreach(pCtx, pt_sd.prbEntriesOnDisk, _pt_sd__create_new_ptnode_for_unclaimed_entry_cb, (void *)&pt_sd)  );

fail:
	SG_PATHNAME_NULLFREE(pCtx, pt_sd.pPathDirectory);
	SG_RBTREE_NULLFREE_WITH_ASSOC(pCtx, pt_sd.prbEntriesOnDisk, (SG_free_callback *)_pt_sd_entry__free);
}

//////////////////////////////////////////////////////////////////

END_EXTERN_C;

#endif//H_SG_PENDINGTREE__PRIVATE_SCAN_DIR_H
