/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include <sg.h>

#define TRACE_PULL 0

/* This exists primarily as a convenience so policy/credential data can be added to it later. 
 * It's only used in this file, there's no opaque wrapper. */
typedef struct
{
	char* pszPullId;
	SG_staging* pStaging;
	SG_repo* pPullIntoRepo;
    SG_vhash* pvh_remote_repo_info;
	SG_bool bAdminOnly; /* True if repo ids don't match but admin ids do. */

	SG_uint32 countRoundtrips;
} sg_pull_instance_data;


/* This is the structure for which SG_pull is an opaque wrapper. */
typedef struct
{
	SG_sync_client* pSyncClient;
	SG_vhash* pvhFragballRequest;
	sg_pull_instance_data* pPullInstance;
} _sg_pull;

static void _free_instance_data(SG_context* pCtx,
								sg_pull_instance_data* pMe)
{
	if (pMe)
	{
		SG_NULLFREE(pCtx, pMe->pszPullId);
		SG_STAGING_NULLFREE(pCtx, pMe->pStaging);
		SG_REPO_NULLFREE(pCtx, pMe->pPullIntoRepo);
        SG_VHASH_NULLFREE(pCtx, pMe->pvh_remote_repo_info);
		SG_NULLFREE(pCtx, pMe);
	}
}
#define _NULLFREE_INSTANCE_DATA(pCtx,p) SG_STATEMENT(  _free_instance_data(pCtx, p); p=NULL;  )

static void _sg_pull__nullfree(SG_context* pCtx, _sg_pull** ppMyPull)
{
	if (ppMyPull && *ppMyPull)
	{
		_sg_pull* pMyPull = *ppMyPull;
		SG_VHASH_NULLFREE(pCtx, pMyPull->pvhFragballRequest);

		if (pMyPull->pPullInstance)
			SG_ERR_IGNORE(  SG_staging__cleanup(pCtx, &pMyPull->pPullInstance->pStaging)  );

		SG_ERR_IGNORE(  _free_instance_data(pCtx, pMyPull->pPullInstance)  );
		SG_SYNC_CLIENT_NULLFREE(pCtx, pMyPull->pSyncClient);
		SG_NULLFREE(pCtx, pMyPull);
		*ppMyPull = NULL;
	}
}


static void _buildPullStats(SG_context* pCtx, _sg_pull* pMyPull, const SG_vhash* pvhStatus, SG_vhash** ppvhStats)
{
	SG_vhash* pvhStats = NULL;
	SG_vhash* pvhRefCounts;

	SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvhStatus, SG_SYNC_STATUS_KEY__COUNTS, &pvhRefCounts)  );

	SG_ERR_CHECK(  SG_VHASH__ALLOC__COPY(pCtx, &pvhStats, pvhRefCounts)  );
	SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvhStats, SG_SYNC_STATUS_KEY__ROUNDTRIPS, pMyPull->pPullInstance->countRoundtrips)  );

	SG_RETURN_AND_NULL(pvhStats, ppvhStats);

	/* common cleanup */
fail:
	SG_VHASH_NULLFREE(pCtx, pvhStats);
}

/**
 * Creates a staging area and returns an initialized instance data structure.
 */
static void _pull_init(SG_context* pCtx, 
					   const char* pszPullIntoRepoDescriptorName,
					   const char* pszRemoteRepoSpec,
					   SG_bool bAdminOnly,
					   _sg_pull** ppMyPull)
{
	char* pszThisRepoId = NULL;
	char* pszThisAdminId = NULL;
	char* pszThisHashMethod = NULL;
	const char* pszRefOtherRepoId = NULL;
	const char* pszRefOtherAdminId = NULL;
	const char* pszRefOtherHashMethod = NULL;

	sg_pull_instance_data* pMe = NULL;
	_sg_pull* pMyPull = NULL;
	
	SG_NULLARGCHECK_RETURN(pszRemoteRepoSpec);
	SG_NULLARGCHECK_RETURN(pszPullIntoRepoDescriptorName);
	SG_NULLARGCHECK_RETURN(ppMyPull);

	SG_ERR_CHECK(  SG_alloc1(pCtx, pMyPull)  );
	SG_ERR_CHECK(  SG_alloc1(pCtx, pMe)  );
	pMyPull->pPullInstance = pMe;

	SG_ERR_CHECK(  SG_sync_client__open(pCtx, pszRemoteRepoSpec, &pMyPull->pSyncClient)  );

	SG_ERR_CHECK(  SG_sync_client__get_repo_info(pCtx, pMyPull->pSyncClient, SG_FALSE, &pMe->pvh_remote_repo_info)  );
	pMe->countRoundtrips++;
    SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pMe->pvh_remote_repo_info, SG_SYNC_REPO_INFO_KEY__REPO_ID, &pszRefOtherRepoId)  );
    SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pMe->pvh_remote_repo_info, SG_SYNC_REPO_INFO_KEY__ADMIN_ID, &pszRefOtherAdminId)  );
    SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pMe->pvh_remote_repo_info, SG_SYNC_REPO_INFO_KEY__HASH_METHOD, &pszRefOtherHashMethod)  );

	SG_ERR_CHECK(  SG_REPO__OPEN_REPO_INSTANCE(pCtx, pszPullIntoRepoDescriptorName, &pMe->pPullIntoRepo)  );

	if (!bAdminOnly)
	{
		SG_ERR_CHECK(  SG_repo__get_repo_id(pCtx, pMe->pPullIntoRepo, &pszThisRepoId)  );
		if (strcmp(pszThisRepoId, pszRefOtherRepoId) != 0)
			SG_ERR_THROW(SG_ERR_REPO_ID_MISMATCH);
	}
	else
		pMyPull->pPullInstance->bAdminOnly = SG_TRUE;

	SG_ERR_CHECK(  SG_repo__get_admin_id(pCtx, pMe->pPullIntoRepo, &pszThisAdminId)  );
	if (strcmp(pszThisAdminId, pszRefOtherAdminId) != 0)
		SG_ERR_THROW(SG_ERR_ADMIN_ID_MISMATCH);

	SG_ERR_CHECK(  SG_repo__get_hash_method(pCtx, pMe->pPullIntoRepo, &pszThisHashMethod)  );
	if (strcmp(pszThisHashMethod, pszRefOtherHashMethod) != 0)
		SG_ERR_THROW(SG_ERR_REPO_HASH_METHOD_MISMATCH);

	// Store pull id (which identifies the staging area) in instance data
	SG_ERR_CHECK(  SG_staging__create(pCtx, pszPullIntoRepoDescriptorName, &pMe->pszPullId, &pMe->pStaging)  );

	SG_RETURN_AND_NULL(pMyPull, ppMyPull);

	/* fall through */
fail:
	SG_NULLFREE(pCtx, pszThisRepoId);
	SG_NULLFREE(pCtx, pszThisAdminId);
	SG_NULLFREE(pCtx, pszThisHashMethod);
	SG_ERR_IGNORE(  _sg_pull__nullfree(pCtx, &pMyPull)  );
}

/**
 * Based on the current contents of our staging area, add dagnodes until the DAGs connect. 
 * May do several roundtrips with the other repo.
 */
static void _add_dagnodes_until_connected(SG_context* pCtx, 
										  SG_vhash** ppvhStagingStatus, 
										  sg_pull_instance_data* pMe, 
										  SG_sync_client* pClient)
{
	SG_bool disconnected = SG_FALSE;
	SG_vhash* pvhRequest = NULL;
	char* pszFragballName = NULL;
	const SG_pathname* pStagingPathname;
	
	SG_rev_spec* pRevSpec = NULL;
	SG_vhash* pvhRevSpec = NULL;
	SG_vhash* pvhRequestDags = NULL;
	SG_vhash* pvhRequestRevs = NULL;
	
	SG_vhash* pvhAllLeaves = NULL;
	SG_vhash* pvhDagLeaves = NULL;
	SG_rbtree* prbLeaves = NULL;
	SG_rbtree_iterator* pit = NULL;

	SG_repo_fetch_dagnodes_handle* pdh = NULL;
	SG_dagnode* pdn = NULL;

	SG_ERR_CHECK(  SG_staging__get_pathname(pCtx, pMe->pStaging, &pStagingPathname)  );

	SG_ERR_CHECK(  SG_vhash__has(pCtx, *ppvhStagingStatus, SG_SYNC_STATUS_KEY__DAGS, &disconnected)  );
	while (disconnected)
	{

#if TRACE_PULL
		SG_ERR_CHECK(  SG_vhash_debug__dump_to_console__named(pCtx, *ppvhStagingStatus, "pull staging status")  );
#endif
		// There's at least one disconnected DAG.  
		
		// Create a fragball request vhash from the staging status vhash.
		SG_ERR_CHECK(  SG_VHASH__ALLOC__PARAMS(pCtx, &pvhRequest, 2, NULL, NULL)  ); // 2 nodes: gens and dags

		{
			SG_vhash* pvhRefStatusDags;
			SG_vhash* pvhRefStatusDag;
			const char* pszRefDagnum;
			SG_uint32 countDags, i;

			SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, *ppvhStagingStatus, SG_SYNC_STATUS_KEY__DAGS, &pvhRefStatusDags)  );
			SG_ERR_CHECK(  SG_vhash__count(pCtx, pvhRefStatusDags, &countDags)  );
			SG_ERR_CHECK(  SG_VHASH__ALLOC__PARAMS(pCtx, &pvhRequestDags, countDags, NULL, NULL)  ); // a node for each dag
			SG_ERR_CHECK(  SG_VHASH__ALLOC__PARAMS(pCtx, &pvhAllLeaves, countDags, NULL, NULL)  ); // a node for each dag

			for (i = 0; i < countDags; i++)
			{
				/* For each disconnected DAG:
				 *  - create a rev spec, serialize the rev spec to a vhash, add that vhash to the request. 
				 *  - add leaves of that dag and their generation, so other repo can build a "best guess" dagfrag. */

				const char* pszRefHid;
				SG_uint32 countHids, j;
				SG_uint64 dagnum;

				SG_ERR_CHECK(  SG_vhash__get_nth_pair__vhash(pCtx, pvhRefStatusDags, i, &pszRefDagnum, &pvhRefStatusDag)  );
				SG_ERR_CHECK(  SG_vhash__count(pCtx, pvhRefStatusDag, &countHids)  );	
				SG_ERR_CHECK(  SG_dagnum__from_sz__hex(pCtx, pszRefDagnum, &dagnum)  );

				SG_ERR_CHECK(  SG_REV_SPEC__ALLOC(pCtx, &pRevSpec)  );
				SG_ERR_CHECK(  SG_rev_spec__set_dagnum(pCtx, pRevSpec, dagnum)  );
				for (j = 0; j < countHids; j++)
				{
					SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvhRefStatusDag, j, &pszRefHid, NULL)  );
					SG_ERR_CHECK(  SG_rev_spec__add_rev(pCtx, pRevSpec, pszRefHid)  );
				}
				SG_ERR_CHECK(  SG_rev_spec__to_vhash(pCtx, pRevSpec, SG_TRUE, &pvhRevSpec)  );
				SG_ERR_CHECK(  SG_vhash__add__vhash(pCtx, pvhRequestDags, pszRefDagnum, &pvhRevSpec )  );
				SG_REV_SPEC_NULLFREE(pCtx, pRevSpec);

				/* Add leaves and gens */
				{
					SG_uint32 countLeaves;
					SG_int32 gen;
					SG_bool b;

					SG_ERR_CHECK(  SG_repo__fetch_dag_leaves(pCtx, pMe->pPullIntoRepo, dagnum, &prbLeaves)  );
					SG_ERR_CHECK(  SG_rbtree__count(pCtx, prbLeaves, &countLeaves)  );
					SG_ERR_CHECK(  SG_VHASH__ALLOC__PARAMS(pCtx, &pvhDagLeaves, countLeaves, NULL, NULL)  );

					SG_ERR_CHECK(  SG_repo__fetch_dagnodes__begin(pCtx, pMe->pPullIntoRepo, dagnum, &pdh)  );
					SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pit, prbLeaves, &b, &pszRefHid, NULL)  );
					while (b)
					{
						SG_ERR_CHECK(  SG_repo__fetch_dagnodes__one(pCtx, pMe->pPullIntoRepo, pdh, pszRefHid, &pdn)  );
						SG_ERR_CHECK(  SG_dagnode__get_generation(pCtx, pdn, &gen)  );
						SG_DAGNODE_NULLFREE(pCtx, pdn);
						SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvhDagLeaves, pszRefHid, gen)  );
						SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pit, &b, &pszRefHid, NULL)  );
					}
					SG_ERR_CHECK(  SG_repo__fetch_dagnodes__end(pCtx, pMe->pPullIntoRepo, &pdh)  );
					SG_RBTREE_NULLFREE(pCtx, prbLeaves);
					SG_RBTREE_ITERATOR_NULLFREE(pCtx, pit);

					SG_ERR_CHECK(  SG_vhash__add__vhash(pCtx, pvhAllLeaves, pszRefDagnum, &pvhDagLeaves)  );
				}
			}

			SG_ERR_CHECK(  SG_vhash__add__vhash(pCtx, pvhRequest, SG_SYNC_STATUS_KEY__DAGS, &pvhRequestDags)  );

			SG_ERR_CHECK(  SG_vhash__add__vhash(pCtx, pvhRequest, SG_SYNC_STATUS_KEY__LEAVES, &pvhAllLeaves)  );
		}

		SG_ERR_CHECK(  SG_sync_client__pull_request_fragball(pCtx, pClient, pvhRequest, 
			SG_FALSE, pStagingPathname, &pszFragballName)  );
		pMe->countRoundtrips++;

		SG_VHASH_NULLFREE(pCtx, pvhRequest);

		SG_ERR_CHECK(  SG_staging__slurp_fragball(pCtx, pMe->pStaging, (const char*)pszFragballName)  );
		SG_NULLFREE(pCtx, pszFragballName);

		SG_VHASH_NULLFREE(pCtx, *ppvhStagingStatus);
		SG_ERR_CHECK(  SG_staging__check_status(pCtx, pMe->pStaging, 
			SG_TRUE, SG_FALSE, SG_FALSE, SG_FALSE, SG_FALSE, 
			ppvhStagingStatus)  );

		SG_ERR_CHECK(  SG_vhash__has(pCtx, *ppvhStagingStatus, SG_SYNC_STATUS_KEY__DAGS, &disconnected)  );

#if TRACE_PULL
		SG_ERR_CHECK(  SG_vhash_debug__dump_to_console__named(pCtx, *ppvhStagingStatus, "pull staging status")  );
#endif
	}

	/* fall through */
fail:
	SG_VHASH_NULLFREE(pCtx, *ppvhStagingStatus);
	SG_VHASH_NULLFREE(pCtx, pvhRequest);
	SG_NULLFREE(pCtx, pszFragballName);
	SG_REV_SPEC_NULLFREE(pCtx, pRevSpec);
	SG_VHASH_NULLFREE(pCtx, pvhRevSpec);
	SG_VHASH_NULLFREE(pCtx, pvhRequestDags);
	SG_VHASH_NULLFREE(pCtx, pvhRequestRevs);
	SG_VHASH_NULLFREE(pCtx, pvhAllLeaves);
	SG_VHASH_NULLFREE(pCtx, pvhDagLeaves);
	SG_RBTREE_NULLFREE(pCtx, prbLeaves);
	SG_RBTREE_ITERATOR_NULLFREE(pCtx, pit);
	if (pdh && pMe && pMe->pPullIntoRepo)
		SG_ERR_IGNORE(  SG_repo__fetch_dagnodes__end(pCtx, pMe->pPullIntoRepo, &pdh)  );
}

/**
 * Based on the current contents of our staging area, request all the blobs we need to complete the pull.
 * May do several roundtrips with the other repo.
 */
static void _add_blobs_until_done(SG_context* pCtx, 
								  _sg_pull* pMyPull) 
{
	SG_staging* pStaging = pMyPull->pPullInstance->pStaging;
	SG_sync_client* pClient = pMyPull->pSyncClient;
	sg_pull_instance_data* pMe = pMyPull->pPullInstance;
	
	SG_bool need_blobs = SG_FALSE;
	SG_vhash* pvhFragballRequest = NULL;
	char* pszFragballName = NULL;
	const SG_pathname* pStagingPathname = NULL;
	SG_vhash* pvhStagingStatus = NULL;

	SG_bool bPopOp = SG_FALSE;
	SG_uint32 i;

	SG_ERR_CHECK(  SG_staging__check_status(pCtx, pStaging, 
		SG_FALSE, SG_FALSE, SG_TRUE, SG_TRUE, SG_FALSE, 
		&pvhStagingStatus)  );

	SG_ERR_CHECK(  SG_vhash__has(pCtx, pvhStagingStatus, SG_SYNC_STATUS_KEY__BLOBS, &need_blobs)  );
	
	if (need_blobs)
		SG_ERR_CHECK(  SG_staging__get_pathname(pCtx, pStaging, &pStagingPathname)  );

	for (i = 0; need_blobs; i++)
	{
		pvhFragballRequest = pvhStagingStatus;
		pvhStagingStatus = NULL;

		if (i)
			SG_ERR_CHECK(  SG_log__push_operation(pCtx, "Downloading user data", SG_LOG__FLAG__NONE)  );
		else
			SG_ERR_CHECK(  SG_log__push_operation(pCtx, "Downloading changesets", SG_LOG__FLAG__NONE)  );
		bPopOp = SG_TRUE;

		SG_ERR_CHECK(  SG_sync_client__pull_request_fragball(pCtx, pClient, pvhFragballRequest, SG_TRUE, pStagingPathname, &pszFragballName)  );
		pMe->countRoundtrips++;
		SG_ERR_CHECK(  SG_log__pop_operation(pCtx)  );
		bPopOp = SG_FALSE;

		SG_VHASH_NULLFREE(pCtx, pvhFragballRequest);

		SG_ERR_CHECK(  SG_staging__slurp_fragball(pCtx, pStaging, (const char*)pszFragballName)  );
		SG_NULLFREE(pCtx, pszFragballName);

		SG_ERR_CHECK(  SG_staging__check_status(pCtx, pStaging, 
			SG_FALSE, SG_FALSE, SG_TRUE, SG_TRUE, SG_FALSE, 
			&pvhStagingStatus)  );

#if TRACE_PULL
		SG_ERR_CHECK(  SG_vhash_debug__dump_to_console__named(pCtx, pvhStagingStatus, "pull staging status")  );
#endif

		SG_ERR_CHECK(  SG_vhash__has(pCtx, pvhStagingStatus, SG_SYNC_STATUS_KEY__BLOBS, &need_blobs)  );
	}

	/* fall through */
fail:
	if (bPopOp)
		SG_log__pop_operation(pCtx);
	SG_VHASH_NULLFREE(pCtx, pvhStagingStatus);
	SG_VHASH_NULLFREE(pCtx, pvhFragballRequest);
	SG_NULLFREE(pCtx, pszFragballName);
}

/**
 * Pulls the requested dagnodes into our staging area.
 * If pMyPull->pvhFragballRequest is NULL, the leaves of every dag in the remote repo are pulled.
 */
static void _add_dagnodes(
	SG_context*        pCtx,
	_sg_pull*		   pMyPull,
	SG_vhash**         ppStatus
	)
{
	SG_staging*        pStaging = pMyPull->pPullInstance->pStaging;
	SG_sync_client*    pClient = pMyPull->pSyncClient;

	const SG_pathname* pStagingPathname = NULL;
	char*              pszFragballName  = NULL;
	SG_vhash*          pvhStatus        = NULL;

	SG_NULLARGCHECK_RETURN(pStaging);
	SG_NULLARGCHECK_RETURN(pClient);
	SG_NULLARGCHECK_RETURN(ppStatus);

	SG_ERR_CHECK(  SG_staging__get_pathname(pCtx, pStaging, &pStagingPathname)  );

	SG_sync_client__pull_request_fragball(pCtx, pClient, pMyPull->pvhFragballRequest, 
		SG_FALSE, pStagingPathname, &pszFragballName);
	if(SG_context__err_equals(pCtx, SG_ERR_BRANCH_HEAD_CHANGESET_NOT_PRESENT))
		SG_ERR_RESET_THROW2(SG_ERR_PULL_FAILED, (pCtx, "The branch cannot be pulled from the requested repository because the repository is missing the latest changeset(s) that it already knows should be in the branch."));
	else
		SG_ERR_CHECK_CURRENT;
	pMyPull->pPullInstance->countRoundtrips++;

	SG_ERR_CHECK(  SG_staging__slurp_fragball(pCtx, pStaging, pszFragballName)  );

	SG_ERR_CHECK(  SG_staging__check_status(pCtx, pStaging, 
		SG_TRUE, SG_FALSE, SG_FALSE, SG_FALSE, SG_FALSE, 
		&pvhStatus)  );
#if TRACE_PULL
	SG_ERR_CHECK(  SG_vhash_debug__dump_to_console__named(pCtx, pvhStatus, "pull staging status")  );
#endif

	*ppStatus = pvhStatus;
	pvhStatus = NULL;

fail:
	SG_VHASH_NULLFREE(pCtx, pvhStatus);
	SG_NULLFREE(pCtx, pszFragballName);
}

/**
 * Pulls the requested dagnodes into our staging area, then roundtrips adding dagnodes until
 * the DAGS connect.
 * A NULL pMyPull->pvhFragballRequest will pull all missing dagnodes from every DAG.
 */
static void _add_dagnodes_and_connect(
	SG_context*            pCtx,
	_sg_pull*			   pMyPull
	)
{
	SG_vhash* pvhStatus = NULL;

	SG_ERR_CHECK(  _add_dagnodes(pCtx, pMyPull, &pvhStatus)  );
	SG_ERR_CHECK(  _add_dagnodes_until_connected(pCtx, &pvhStatus, pMyPull->pPullInstance, pMyPull->pSyncClient)  );

fail:
	SG_VHASH_NULLFREE(pCtx, pvhStatus);
}

/**
 * 
 */
static void _do_pull(
	SG_context* pCtx,
	_sg_pull* pMyPull,
	const char* pszRootOperationDescription,
	SG_vhash** ppvhStats)
{
	SG_vhash* pvhFinalPrecommitStatusForStats = NULL;

	SG_ERR_CHECK(  SG_log__push_operation(pCtx, 
		pszRootOperationDescription ? pszRootOperationDescription : "Pulling", SG_LOG__FLAG__NONE)  );
	SG_ERR_CHECK(  SG_log__set_steps(pCtx, 5u, NULL)  );

	SG_NULLARGCHECK(pMyPull);

	/* Request a fragball containing leaves of every dag */
	/* Then check the status and use it to request more dagnodes until the dags connect */
	SG_ERR_CHECK(  SG_log__set_step(pCtx, "Looking for changes")  );
	SG_ERR_CHECK(  _add_dagnodes_and_connect(pCtx, pMyPull)  );
	SG_ERR_CHECK(  SG_log__finish_step(pCtx)  );

	/* All necessary dagnodes are in the frag at this point.  Now get remaining missing blobs. */
	SG_ERR_CHECK(  SG_log__set_step(pCtx, "Transferring data")  );
	SG_ERR_CHECK(  _add_blobs_until_done(pCtx, pMyPull)  );
	SG_ERR_CHECK(  SG_log__finish_step(pCtx)  );

	if (ppvhStats)
	{
		SG_ERR_CHECK(  SG_staging__check_status(pCtx, pMyPull->pPullInstance->pStaging, 
			SG_FALSE, SG_FALSE, SG_FALSE, SG_FALSE, SG_TRUE, &pvhFinalPrecommitStatusForStats)  );
	}

	/* commit */
	SG_ERR_CHECK(  SG_log__set_step(pCtx, "Committing changes")  );
	SG_ERR_CHECK(  SG_staging__commit(pCtx, pMyPull->pPullInstance->pStaging)  );
	SG_ERR_CHECK(  SG_log__finish_step(pCtx)  );

	/* cleanup */
	SG_ERR_CHECK(  SG_log__set_step(pCtx, "Cleaning up")  );
	SG_ERR_CHECK(  SG_staging__cleanup(pCtx, &pMyPull->pPullInstance->pStaging)  );
	SG_ERR_CHECK(  SG_log__finish_step(pCtx)  );

	/* auto-merge zing dags */
	SG_ERR_CHECK(  SG_log__set_step(pCtx, "Merging databases")  );
	SG_ERR_CHECK(  SG_zing__auto_merge_all_dags(pCtx, pMyPull->pPullInstance->pPullIntoRepo)  );
	SG_ERR_CHECK(  SG_log__finish_step(pCtx)  );

	if (ppvhStats)
		SG_ERR_CHECK(  _buildPullStats(pCtx, pMyPull, pvhFinalPrecommitStatusForStats, ppvhStats)  );

	/* fall through */
fail:
	SG_log__pop_operation(pCtx);
	SG_VHASH_NULLFREE(pCtx, pvhFinalPrecommitStatusForStats);
}

void SG_pull__all(SG_context* pCtx, 
				  const char* pszPullIntoRepoDescriptorName, 
				  const char* pszRemoteRepoSpec,
				  SG_vhash** ppvhStats)
{
	_sg_pull* pMyPull = NULL;

	SG_NULLARGCHECK_RETURN(pszPullIntoRepoDescriptorName);
	SG_NULLARGCHECK_RETURN(pszRemoteRepoSpec);

	SG_ERR_CHECK(  _pull_init(pCtx, pszPullIntoRepoDescriptorName, pszRemoteRepoSpec, SG_FALSE, &pMyPull)  );
	SG_ERR_CHECK(  _do_pull(pCtx, pMyPull, NULL, ppvhStats)  );

	/* fall through */
fail:
	SG_ERR_IGNORE(  _sg_pull__nullfree(pCtx, &pMyPull)  );
}

void SG_pull__begin(
	SG_context* pCtx,
	const char* pszPullIntoRepoDescriptorName, 
	const char* pszRemoteRepoSpec,
	SG_pull** ppPull)
{
	_sg_pull* pMyPull = NULL;

	SG_NULLARGCHECK_RETURN(pszPullIntoRepoDescriptorName);
	SG_NULLARGCHECK_RETURN(pszRemoteRepoSpec);
	SG_NULLARGCHECK_RETURN(ppPull);

	SG_ERR_CHECK(  _pull_init(pCtx, pszPullIntoRepoDescriptorName, pszRemoteRepoSpec, SG_FALSE, &pMyPull)  );

	*ppPull = (SG_pull*)pMyPull;
	pMyPull = NULL;

	/* fall through */
fail:
	SG_ERR_IGNORE(  _sg_pull__nullfree(pCtx, &pMyPull)  );
}

void SG_pull__add(
	SG_context* pCtx,
	SG_pull* pPull,
	SG_rev_spec** ppAddRevSpec)
{
	_sg_pull* pMyPull = NULL;

	SG_uint64 iDagnum;
	char bufDagnum[SG_DAGNUM__BUF_MAX__HEX];
	SG_bool bDagnumRequestExists = SG_FALSE;
	
	/* These need to be freed. */
	SG_vhash* pvhDags = NULL;
	SG_vhash* pvhNewRevSpec = NULL;

	/* These don't. */
	SG_vhash* pvhRefRevSpec = NULL;
	SG_vhash* pvhDagsRef = NULL;
	
	SG_rev_spec* pNewRevSpec = NULL;

	SG_NULLARGCHECK_RETURN(pPull);
	SG_NULL_PP_CHECK_RETURN(ppAddRevSpec);

	pMyPull = (_sg_pull*)pPull;
	if (!pMyPull->pvhFragballRequest)
		SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pMyPull->pvhFragballRequest)  );

	SG_ERR_CHECK(  SG_rev_spec__get_dagnum(pCtx, *ppAddRevSpec, &iDagnum)  );
	SG_ERR_CHECK(  SG_dagnum__to_sz__hex(pCtx, iDagnum, bufDagnum, sizeof(bufDagnum))  );

	/* Get the existing rev spec, if there is one. */
	{
		SG_bool found = SG_FALSE;

		SG_ERR_CHECK(  SG_vhash__has(pCtx, pMyPull->pvhFragballRequest, SG_SYNC_STATUS_KEY__DAGS, &found)  );
		if (found)
			SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pMyPull->pvhFragballRequest, SG_SYNC_STATUS_KEY__DAGS, &pvhDagsRef)  );
		else
		{
			SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvhDags)  );
			pvhDagsRef = pvhDags;
			SG_ERR_CHECK(  SG_vhash__add__vhash(pCtx, pMyPull->pvhFragballRequest, SG_SYNC_STATUS_KEY__DAGS, &pvhDags)  );
		}
		SG_ERR_CHECK(  SG_vhash__has(pCtx, pvhDagsRef, bufDagnum, &bDagnumRequestExists)  );
		if (bDagnumRequestExists)
		{
			SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvhDagsRef, bufDagnum, &pvhRefRevSpec)  );
			SG_ERR_CHECK(  SG_rev_spec__from_vash(pCtx, pvhRefRevSpec, &pNewRevSpec)  );
		}
	}

	/* If there's an existing rev spec, merge the new one into it. */
	if (pNewRevSpec)
	{
		SG_ERR_CHECK(  SG_rev_spec__merge(pCtx, pNewRevSpec, ppAddRevSpec)  );
	}
	else
	{
		/* Otherwise, we'll just add the provided rev spec as-is. */
		pNewRevSpec = *ppAddRevSpec;
		*ppAddRevSpec = NULL;
	}

	/* Put new rev spec into fragball request vhash. Replace the old contents if there were any. */
	if (bDagnumRequestExists)
		SG_ERR_CHECK(  SG_vhash__remove(pCtx, pvhDagsRef, bufDagnum)  );
	SG_ERR_CHECK(  SG_rev_spec__to_vhash(pCtx, pNewRevSpec, SG_TRUE, &pvhNewRevSpec)  );
	SG_ERR_CHECK(  SG_vhash__add__vhash(pCtx, pvhDagsRef, bufDagnum, &pvhNewRevSpec)  );

	/* common cleanup */
fail:
	SG_REV_SPEC_NULLFREE(pCtx, pNewRevSpec);
	SG_VHASH_NULLFREE(pCtx, pvhDags);
	SG_VHASH_NULLFREE(pCtx, pvhNewRevSpec);
}

static void _add_dagnum_to_request(SG_context* pCtx, SG_uint64 iDagnum, SG_vhash** ppvhFragballRequest)
{
	char bufDagnum[SG_DAGNUM__BUF_MAX__HEX];

	/* These need to be freed. */
	SG_vhash* pvhDags = NULL;
	SG_vhash* pvhFragballRequest = NULL;

	/* These don't. */
	SG_vhash* pvhDagsRef = NULL;
	SG_vhash* pvhFragballRequestRef = NULL;

	SG_NULLARGCHECK_RETURN(ppvhFragballRequest);

	if (*ppvhFragballRequest)
		pvhFragballRequestRef = *ppvhFragballRequest;
	else
	{
		SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvhFragballRequest)  );
		pvhFragballRequestRef = pvhFragballRequest;
	}

	SG_ERR_CHECK(  SG_dagnum__to_sz__hex(pCtx, iDagnum, bufDagnum, sizeof(bufDagnum))  );

	/* Get the dags container node into pvhDagsRev. */
	{
		SG_bool found = SG_FALSE;

		SG_ERR_CHECK(  SG_vhash__has(pCtx, pvhFragballRequestRef, SG_SYNC_STATUS_KEY__DAGS, &found)  );
		if (found)
			SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvhFragballRequestRef, SG_SYNC_STATUS_KEY__DAGS, &pvhDagsRef)  );
		else
		{
			SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvhDags)  );
			pvhDagsRef = pvhDags;
			SG_ERR_CHECK(  SG_vhash__add__vhash(pCtx, pvhFragballRequestRef, SG_SYNC_STATUS_KEY__DAGS, &pvhDags)  );
		}
	}

	/* Add a null value for this dagnum, indicating we want it all. */
	SG_ERR_CHECK(  SG_vhash__add__null(pCtx, pvhDagsRef, bufDagnum)  );

	*ppvhFragballRequest = pvhFragballRequestRef;
	pvhFragballRequest = NULL;

	/* common cleanup */
fail:
	SG_VHASH_NULLFREE(pCtx, pvhDags);
	SG_VHASH_NULLFREE(pCtx, pvhFragballRequest);
}

void SG_pull__add__dagnum(SG_context* pCtx, SG_pull* pPull, SG_uint64 iDagnum)
{
	_sg_pull* pMyPull = (_sg_pull*)pPull;

	SG_NULLARGCHECK_RETURN(pPull);
	SG_ARGCHECK_RETURN(iDagnum,iDagnum);

	SG_ERR_CHECK_RETURN(  _add_dagnum_to_request(pCtx, iDagnum, &pMyPull->pvhFragballRequest)  );
}

void SG_pull__commit(
	SG_context* pCtx,
	SG_pull** ppPull,
	SG_vhash** ppvhStats
    )
{
	_sg_pull* pMyPull = NULL;

	SG_NULL_PP_CHECK(ppPull);
	pMyPull = (_sg_pull*)*ppPull;

	SG_ERR_CHECK(  _do_pull(pCtx, pMyPull, NULL, ppvhStats)  );


	/* fall through */
fail:
	SG_ERR_IGNORE(  _sg_pull__nullfree(pCtx, (_sg_pull**)ppPull)  );
}

void SG_pull__abort(
	SG_context* pCtx,
	SG_pull** ppPull)
{
	_sg_pull** ppMyPull = (_sg_pull**)ppPull;

	SG_NULL_PP_CHECK_RETURN(ppPull);

	SG_ERR_CHECK_RETURN(  _sg_pull__nullfree(pCtx, ppMyPull)  );
}

void SG_pull__list_incoming_vc(
	SG_context* pCtx,
	const char* pszPullIntoRepoDescriptorName,
	const char* pszRemoteRepoSpec,
	SG_history_result** ppInfo,
	SG_vhash** ppvhStats)
{
	SG_string* pstrOperation = NULL;
	_sg_pull* pMyPull = NULL;
	SG_vhash* pvhStatus = NULL;

	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pstrOperation)  );
	SG_ERR_CHECK(  SG_string__append__sz(pCtx, pstrOperation, "Incoming from ")  );
	SG_ERR_CHECK(  SG_string__append__sz(pCtx, pstrOperation, pszRemoteRepoSpec)  );
	SG_ERR_CHECK(  SG_log__push_operation(pCtx, SG_string__sz(pstrOperation), SG_LOG__FLAG__NONE)  );
	SG_ERR_CHECK(  SG_log__set_steps(pCtx, 2u, NULL)  );

	SG_NULLARGCHECK(pszPullIntoRepoDescriptorName);
	SG_NULLARGCHECK(pszRemoteRepoSpec);

	SG_ERR_CHECK(  _pull_init(pCtx, pszPullIntoRepoDescriptorName, pszRemoteRepoSpec, SG_FALSE, &pMyPull)  );

	SG_ERR_CHECK(  SG_log__set_step(pCtx, "Looking for changes")  );
	SG_ERR_CHECK(  _add_dagnum_to_request(pCtx, SG_DAGNUM__VERSION_CONTROL, &pMyPull->pvhFragballRequest)  );
	SG_ERR_CHECK(  _add_dagnodes_and_connect(pCtx, pMyPull)  );
	SG_ERR_CHECK(  SG_log__finish_step(pCtx)  );

	SG_ERR_CHECK(  SG_log__set_step(pCtx, "Fetching changeset descriptions")  );
	SG_ERR_CHECK(  SG_staging__check_status(pCtx, pMyPull->pPullInstance->pStaging, 
		SG_FALSE, SG_TRUE, SG_FALSE, SG_FALSE, (ppvhStats != NULL), &pvhStatus)  );
	{
		SG_bool b = SG_FALSE;
		SG_vhash* pvhRequest = NULL;
		SG_ERR_CHECK(  SG_vhash__has(pCtx, pvhStatus, SG_SYNC_STATUS_KEY__NEW_NODES, &b)  );
		if (b)
		{
			/* There are incoming nodes.  Get their info. */
			SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvhStatus, SG_SYNC_STATUS_KEY__NEW_NODES, &pvhRequest)  );
#if TRACE_PULL
			SG_ERR_CHECK(  SG_vhash_debug__dump_to_console__named(pCtx, pvhRequest, "incoming status")  );
#endif
			SG_ERR_CHECK(  SG_sync_client__get_dagnode_info(pCtx, pMyPull->pSyncClient, pvhRequest, ppInfo)  );
		}
	}
	SG_ERR_CHECK(  SG_log__finish_step(pCtx)  );

 	if (ppvhStats)
	{
 		SG_ERR_CHECK(  _buildPullStats(pCtx, pMyPull, pvhStatus, ppvhStats)  );
	}

	/* fall through */
fail:
	SG_ERR_IGNORE(  _sg_pull__nullfree(pCtx, &pMyPull)  );
	SG_VHASH_NULLFREE(pCtx, pvhStatus);
	SG_log__pop_operation(pCtx);
	SG_STRING_NULLFREE(pCtx, pstrOperation); // must be after the operation pop
}

void SG_pull__add__area(
	SG_context* pCtx,
	SG_pull* pPull,
    const char* psz_area_name
    )
{
	_sg_pull* pMe = (_sg_pull*) pPull;
    SG_vhash* pvh_dags = NULL;
    SG_vhash* pvh_areas = NULL;
    SG_uint64 area_id = 0;
    SG_uint32 count_dags = 0;
    SG_uint32 i_dag = 0;
	SG_rev_spec* pRevSpec = NULL;

	SG_NULLARGCHECK_RETURN(pPull);

    SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pMe->pPullInstance->pvh_remote_repo_info, "areas", &pvh_areas)  );
    SG_ERR_CHECK(  SG_vhash__get__uint64(pCtx, pvh_areas, psz_area_name, &area_id)  );
    SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pMe->pPullInstance->pvh_remote_repo_info, "dags", &pvh_dags)  );
    SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_dags, &count_dags)  );
    for (i_dag=0; i_dag<count_dags; i_dag++)
    {
        const char* psz_dagnum = NULL;
        SG_uint64 dagnum = 0;

        SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvh_dags, i_dag, &psz_dagnum, NULL)  );
        SG_ERR_CHECK(  SG_dagnum__from_sz__hex(pCtx, psz_dagnum, &dagnum)  );
        if (SG_DAGNUM__GET_AREA_ID(dagnum) == area_id)
        {
            SG_ERR_CHECK(  SG_pull__add__dagnum(pCtx, pPull, dagnum)  );
        }
    }

fail:
    SG_REV_SPEC_NULLFREE(pCtx, pRevSpec);
}

void SG_pull__add__area__version_control(
	SG_context* pCtx,
	SG_pull* pPull,
	SG_rev_spec** ppRevSpec
    )
{
    SG_uint64 area_id = SG_DAGNUM__GET_AREA_ID(SG_DAGNUM__VERSION_CONTROL);
	_sg_pull* pMe = (_sg_pull*) pPull;
    SG_vhash* pvh_dags = NULL;
    SG_uint32 count_dags = 0;
    SG_uint32 i_dag = 0;

	SG_NULLARGCHECK_RETURN(pPull);
	
	if (ppRevSpec && *ppRevSpec)
	{
		/* If specific version control nodes were requested, add them. */
		SG_uint64 iDagnum;
		SG_ERR_CHECK(  SG_rev_spec__get_dagnum(pCtx, *ppRevSpec, &iDagnum)  );
		if (iDagnum != SG_DAGNUM__VERSION_CONTROL)
			SG_ERR_THROW2(SG_ERR_INVALIDARG, (pCtx, "revision spec's dagnum is not version control"));

		SG_ERR_CHECK(  SG_pull__add(pCtx, pPull, ppRevSpec)  );
	}
	else
	{
		/* Otherwise add the whole dag. */
		SG_ERR_CHECK(  SG_pull__add__dagnum(pCtx, pPull, SG_DAGNUM__VERSION_CONTROL)  );
	}
    

	/* Now add all the other dags that are in the version control area. */
    SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pMe->pPullInstance->pvh_remote_repo_info, "dags", &pvh_dags)  );
    SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_dags, &count_dags)  );
    for (i_dag=0; i_dag<count_dags; i_dag++)
    {
        const char* psz_dagnum = NULL;
        SG_uint64 dagnum = 0;

        SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvh_dags, i_dag, &psz_dagnum, NULL)  );
        SG_ERR_CHECK(  SG_dagnum__from_sz__hex(pCtx, psz_dagnum, &dagnum)  );
        if (SG_DAGNUM__GET_AREA_ID(dagnum) == area_id)
        {
            if (SG_DAGNUM__VERSION_CONTROL != dagnum)
            {
				SG_ERR_CHECK(  SG_pull__add__dagnum(pCtx, pPull, dagnum)  );
            }
        }
    }

fail:
    ;
}

void SG_pull__add__admin(
	SG_context* pCtx,
	SG_pull* pPull
	)
{
	_sg_pull* pMe = (_sg_pull*) pPull;
	SG_vhash* pvh_dags = NULL;
	SG_uint32 count_dags = 0;
	SG_uint32 i_dag = 0;

	SG_NULLARGCHECK_RETURN(pPull);

	/* Add all the remote dags that have their admin bit set. */
	SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pMe->pPullInstance->pvh_remote_repo_info, "dags", &pvh_dags)  );
	SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_dags, &count_dags)  );
	for (i_dag = 0; i_dag < count_dags; i_dag++)
	{
		const char* psz_dagnum = NULL;
		SG_uint64 dagnum = 0;

		SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvh_dags, i_dag, &psz_dagnum, NULL)  );
		SG_ERR_CHECK(  SG_dagnum__from_sz__hex(pCtx, psz_dagnum, &dagnum)  );
		if (SG_DAGNUM__IS_ADMIN(dagnum))
			SG_ERR_CHECK(  SG_pull__add__dagnum(pCtx, pPull, dagnum)  );
	}

fail:
	;
}

void SG_pull__admin(
	SG_context* pCtx, 
	const char* pszPullIntoRepoDescriptorName, 
	const char* pszRemoteRepoSpec,
	SG_vhash** ppvhStats)
{
	_sg_pull* pMyPull = NULL;

	SG_NULLARGCHECK_RETURN(pszPullIntoRepoDescriptorName);
	SG_NULLARGCHECK_RETURN(pszRemoteRepoSpec);

	SG_ERR_CHECK(  _pull_init(pCtx, pszPullIntoRepoDescriptorName, pszRemoteRepoSpec, SG_TRUE, &pMyPull)  );
	SG_ERR_CHECK(  SG_pull__add__admin(pCtx, (SG_pull*)pMyPull)  );
	SG_ERR_CHECK(  _do_pull(pCtx, pMyPull, "Pulling user database", ppvhStats)  );

	/* fall through */
fail:
	SG_ERR_IGNORE(  _sg_pull__nullfree(pCtx, &pMyPull)  );
}