/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

//////////////////////////////////////////////////////////////////

#include <sg.h>

//////////////////////////////////////////////////////////////////

BEGIN_EXTERN_C;

typedef struct __revspec
{
	SG_rev_spec_type specType;
	char*	pszGiven;
} _revspec;

/* The structure for which SG_rev_spec is an opaque wrapper.
 * 
 * Because we use vhashes as a serialization structure for an SG_rev_spec, it would make
 * sense to use a vhash for the core data type here. But we don't, on purpose, because
 * the order in which the rev specs were provided is important in some cases, and a vhash
 * won't preserve that. */
typedef struct __sg_rev_spec
{
	SG_uint64	iDagNum;
	SG_vector*	pvecRevSpecs;
	SG_uint32	iCountRevs;
	SG_uint32	iCountTags;
	SG_uint32	iCountBranches;
} _sg_rev_spec;

static void _add_rev_spec(SG_context* pCtx, _sg_rev_spec* pRevSpec, SG_rev_spec_type specType, const char* pszGiven);

//////////////////////////////////////////////////////////////////

/* TODO: should this routine actually be in SG_vc_branches? 
 *		 I hate poking around inside vhashes created by other modules. */

/* We do branch lookups in these helpers because we look them up in a loop, 
and there's no reason to cleanup or re-retrieve the pile in each iteration. */
static void _ambiguous_branch_lookup(
	SG_context* pCtx, 
	SG_vhash* pvhCleanPileOfBranches, 
	const char* pszBranch, 
	SG_varray** ppvaFullHids)
{
	SG_vhash* pvhRefBranches;
	SG_vhash* pvhRefThisBranch;
	SG_vhash* pvhRefHids;
	SG_varray* pva = NULL;

	SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvhCleanPileOfBranches, "branches", &pvhRefBranches)  );
	
	SG_vhash__get__vhash(pCtx, pvhRefBranches, pszBranch, &pvhRefThisBranch);
	SG_ERR_REPLACE(SG_ERR_VHASH_KEYNOTFOUND, SG_ERR_BRANCH_NOT_FOUND);
	SG_ERR_CHECK_CURRENT;
	
	SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvhRefThisBranch, "values", &pvhRefHids)  );

	SG_ERR_CHECK(  SG_vhash__get_keys(pCtx, pvhRefHids, &pva)  );

	*ppvaFullHids = pva;
	pva = NULL;

	return;

fail:
	SG_VARRAY_NULLFREE(pCtx, pva);
}

/* We do branch lookups in these helpers because we look them up in a loop, 
 * and there's no reason to cleanup or re-retrieve the pile in each iteration. */
static void _unambiguous_branch_lookup(
	SG_context* pCtx, 
	SG_vhash* pvhCleanPileOfBranches, 
	const char* pszBranch, 
	char** ppszFullHid)
{
	SG_varray* pva = NULL;
	SG_uint32 count = 0;

	SG_ERR_CHECK(  _ambiguous_branch_lookup(pCtx, pvhCleanPileOfBranches, pszBranch, &pva)  );

	SG_ERR_CHECK(  SG_varray__count(pCtx, pva, &count)  );
	if (count == 0)
		SG_ERR_THROW(  SG_ERR_BRANCH_NOT_FOUND  );
	else if (count == 1)
	{
		const char* pszRef;
		SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pva, 0, &pszRef)  );
		SG_ERR_CHECK(  SG_STRDUP(pCtx, pszRef, ppszFullHid)  );
	}
	else
		SG_ERR_THROW2(  SG_ERR_BRANCH_NEEDS_MERGE, (pCtx, "%s", pszBranch)  );

	/* common cleanup */
fail:
	SG_VARRAY_NULLFREE(pCtx, pva);
}

//////////////////////////////////////////////////////////////////

void SG_rev_spec__alloc(SG_context* pCtx, SG_rev_spec** ppRevSpec)
{
	_sg_rev_spec* p = NULL;
	
	SG_NULLARGCHECK_RETURN(ppRevSpec);
	
	SG_ERR_CHECK(  SG_alloc1(pCtx, p)  );
	
	p->iDagNum = SG_DAGNUM__VERSION_CONTROL;
	SG_ERR_CHECK(  SG_VECTOR__ALLOC(pCtx, &p->pvecRevSpecs, 5)  );
	
	*ppRevSpec = (SG_rev_spec*)p;

	return;

fail:
	if (p)
	{
		SG_VECTOR_NULLFREE(pCtx, p->pvecRevSpecs);
		SG_NULLFREE(pCtx, p);
	}
}

void SG_rev_spec__alloc__copy(SG_context* pCtx, const SG_rev_spec* pRevSpecSrc, SG_rev_spec** ppRevSpecDest)
{
	SG_rev_spec* pRevSpecDest = NULL;
	_sg_rev_spec* pDest = NULL;
	_sg_rev_spec* pSrc = (_sg_rev_spec*)pRevSpecSrc;
	SG_uint32 i, count;
	_revspec* psSrc;

	SG_NULLARGCHECK_RETURN(pRevSpecSrc);
	SG_NULLARGCHECK_RETURN(ppRevSpecDest);

	SG_ERR_CHECK(  SG_rev_spec__alloc(pCtx, &pRevSpecDest)  );
	pDest = (_sg_rev_spec*)pRevSpecDest;

	SG_ERR_CHECK(  SG_vector__length(pCtx, pSrc->pvecRevSpecs, &count)  );
	for (i = 0; i < count; i++)
	{
		SG_ERR_CHECK(  SG_vector__get(pCtx, pSrc->pvecRevSpecs, i, (void**)&psSrc)  );
		SG_ERR_CHECK(  _add_rev_spec(pCtx, pDest, psSrc->specType, psSrc->pszGiven)  );
	}

	*ppRevSpecDest = pRevSpecDest;
	pRevSpecDest = NULL;

	/* Common cleanup */
fail:
	SG_REV_SPEC_NULLFREE(pCtx, pRevSpecDest);
}

static void _free_revspecs_callback(SG_context* pCtx, void * pVoidData)
{
	if (pVoidData)
	{
		_revspec* prs = (_revspec*)pVoidData;
		SG_NULLFREE(pCtx, prs->pszGiven);
		SG_NULLFREE(pCtx, prs);
	}
}

void SG_rev_spec__free(SG_context* pCtx, SG_rev_spec* pRevSpec)
{
	if (pRevSpec)
	{
		_sg_rev_spec* p = (_sg_rev_spec*)pRevSpec;
		SG_ERR_CHECK_RETURN(  SG_VECTOR_NULLFREE_WITH_ASSOC(pCtx, p->pvecRevSpecs, _free_revspecs_callback)  );
		SG_ERR_CHECK_RETURN(  SG_free(pCtx, p)  );
	}
}

//////////////////////////////////////////////////////////////////

static void _add_rev_spec(SG_context* pCtx, _sg_rev_spec* pRevSpec, SG_rev_spec_type specType, const char* pszGiven)
{
	_revspec* p = NULL;
	
	SG_ERR_CHECK(  SG_alloc1(pCtx, p)  );
	p->specType = specType;
	SG_ERR_CHECK(  SG_STRDUP(pCtx, pszGiven, &p->pszGiven)  );
	SG_ERR_CHECK(  SG_vector__append(pCtx, pRevSpec->pvecRevSpecs, p, NULL)  );

	switch(specType)
	{
		case SPECTYPE_REV:
			pRevSpec->iCountRevs++;
			break;
		case SPECTYPE_TAG:
			pRevSpec->iCountTags++;
			break;
		case SPECTYPE_BRANCH:
			pRevSpec->iCountBranches++;
			break;
		default:
			SG_ERR_THROW2(SG_ERR_INVALID_REV_SPEC, (pCtx, "unknown rev spec type"));
	}

	return;

fail:
	if (p)
	{
		SG_NULLFREE(pCtx, p->pszGiven);
		SG_NULLFREE(pCtx, p);
	}
}

void SG_rev_spec__add_rev(SG_context* pCtx, SG_rev_spec* pRevSpec, const char* pszRev)
{
	_sg_rev_spec* pMe = (_sg_rev_spec*)pRevSpec;

	SG_NULLARGCHECK_RETURN(pRevSpec);

	SG_ERR_CHECK_RETURN(  _add_rev_spec(pCtx, pMe, SPECTYPE_REV, pszRev)  );
}

void SG_rev_spec__add_tag(SG_context* pCtx, SG_rev_spec* pRevSpec, const char* pszTag)
{
	_sg_rev_spec* pMe = (_sg_rev_spec*)pRevSpec;

	SG_NULLARGCHECK_RETURN(pRevSpec);

	SG_ERR_CHECK_RETURN(  _add_rev_spec(pCtx, pMe, SPECTYPE_TAG, pszTag)  );
}

void SG_rev_spec__add_branch(SG_context* pCtx, SG_rev_spec* pRevSpec, const char* pszBranch)
{
	_sg_rev_spec* pMe = (_sg_rev_spec*)pRevSpec;

	SG_NULLARGCHECK_RETURN(pRevSpec);

	SG_ERR_CHECK_RETURN(  _add_rev_spec(pCtx, pMe, SPECTYPE_BRANCH, pszBranch)  );
}

//////////////////////////////////////////////////////////////////

void SG_rev_spec__has_non_vc_specs(SG_context* pCtx, SG_rev_spec* pRevSpec, SG_bool* pbResult)
{
	SG_uint32 count = 0;

	SG_NULLARGCHECK_RETURN(pbResult);

	SG_ERR_CHECK_RETURN(  SG_rev_spec__count_tags(pCtx, pRevSpec, &count)  );
	if (count)
	{
		*pbResult = SG_TRUE;
		return;
	}

	SG_ERR_CHECK_RETURN(  SG_rev_spec__count_branches(pCtx, pRevSpec, &count)  );
	if (count)
	{
		*pbResult = SG_TRUE;
		return;
	}
}

void SG_rev_spec__get_dagnum(SG_context* pCtx, SG_rev_spec* pRevSpec, SG_uint64* piDagnum)
{
	_sg_rev_spec* pMe = (_sg_rev_spec*)pRevSpec;

	SG_NULLARGCHECK_RETURN(pRevSpec);
	SG_NULLARGCHECK_RETURN(piDagnum);

	*piDagnum = pMe->iDagNum;
}

void SG_rev_spec__set_dagnum(SG_context* pCtx, SG_rev_spec* pRevSpec, SG_uint64 iDagnum)
{
	_sg_rev_spec* pMe = (_sg_rev_spec*)pRevSpec;

	SG_NULLARGCHECK_RETURN(pRevSpec);

	pMe->iDagNum = iDagnum;
}

//////////////////////////////////////////////////////////////////

void SG_rev_spec__count_revs(SG_context* pCtx, SG_rev_spec* pRevSpec, SG_uint32* piCount)
{
	_sg_rev_spec* pMe = (_sg_rev_spec*)pRevSpec;
	SG_NULLARGCHECK_RETURN(piCount);

	if (pRevSpec)
		*piCount = pMe->iCountRevs;
	else
		*piCount = 0;
}

void SG_rev_spec__count_tags(SG_context* pCtx, SG_rev_spec* pRevSpec, SG_uint32* piCount)
{
	_sg_rev_spec* pMe = (_sg_rev_spec*)pRevSpec;
	SG_NULLARGCHECK_RETURN(piCount);

	if (pRevSpec)
		*piCount = pMe->iCountTags;
	else
		*piCount = 0;
}

void SG_rev_spec__count_branches(SG_context* pCtx, SG_rev_spec* pRevSpec, SG_uint32* piCount)
{
	_sg_rev_spec* pMe = (_sg_rev_spec*)pRevSpec;
	SG_NULLARGCHECK_RETURN(piCount);

	if (pRevSpec)
		*piCount = pMe->iCountBranches;
	else
		*piCount = 0;
}

void SG_rev_spec__count(SG_context* pCtx, SG_rev_spec* pRevSpec, SG_uint32* piCount)
{
	_sg_rev_spec* pMe = (_sg_rev_spec*)pRevSpec;
	SG_NULLARGCHECK_RETURN(piCount);

	if (pRevSpec)
		SG_ERR_CHECK_RETURN(  SG_vector__length(pCtx, pMe->pvecRevSpecs, piCount)  );
	else
		*piCount = 0;
}
SG_bool SG_rev_spec__is_empty(SG_context* pCtx, SG_rev_spec* pRevSpec)
{
	SG_uint32 iCount=0;
	if(!pRevSpec)
		return SG_TRUE;
	SG_ERR_IGNORE(  SG_rev_spec__count(pCtx, pRevSpec, &iCount)  );
	return iCount==0;
}

//////////////////////////////////////////////////////////////////

static void _to_rbtree(SG_context* pCtx, _sg_rev_spec* pMe, SG_rev_spec_type specType, SG_rbtree** pprb)
{
	SG_rbtree* prb = NULL;
	SG_uint32 count;
	SG_vector* pvecRevSpecs;
	SG_uint32 i = 0;

	SG_NULLARGCHECK_RETURN(pMe);
	pvecRevSpecs = pMe->pvecRevSpecs;
	SG_ERR_CHECK(  SG_vector__length(pCtx, pMe->pvecRevSpecs, &count)  );
	if (count)
	{
		SG_ERR_CHECK(  SG_RBTREE__ALLOC__PARAMS(pCtx, &prb, count, NULL)  );
		for (i = 0; i < count; i++)
		{
			_revspec* p;
			SG_ERR_CHECK(  SG_vector__get(pCtx, pvecRevSpecs, i, (void**)&p)  );
			if (p->specType == specType)
				SG_ERR_CHECK(  SG_rbtree__add(pCtx, prb, p->pszGiven)  );
		}

		SG_ERR_CHECK(  SG_rbtree__count(pCtx, prb, &count)  );
		if (count) /* If there's nothing left after filtering out other types, return null. */
			SG_RETURN_AND_NULL(prb, pprb);
	}

fail:
	SG_RBTREE_NULLFREE(pCtx, prb);
}

void SG_rev_spec__revs_rbt(SG_context* pCtx, SG_rev_spec* pRevSpec, SG_rbtree** pprbRevs)
{
	SG_ERR_CHECK_RETURN(  _to_rbtree(pCtx, (_sg_rev_spec*)pRevSpec, SPECTYPE_REV, pprbRevs)  );
}

void SG_rev_spec__tags_rbt(SG_context* pCtx, SG_rev_spec* pRevSpec, SG_rbtree** pprbTags)
{
	SG_ERR_CHECK_RETURN(  _to_rbtree(pCtx, (_sg_rev_spec*)pRevSpec, SPECTYPE_TAG, pprbTags)  );
}

void SG_rev_spec__branches_rbt(SG_context* pCtx, SG_rev_spec* pRevSpec, SG_rbtree** pprbBranches)
{
	SG_ERR_CHECK_RETURN(  _to_rbtree(pCtx, (_sg_rev_spec*)pRevSpec, SPECTYPE_BRANCH, pprbBranches)  );
}

static void _to_stringarray(SG_context* pCtx, _sg_rev_spec* pMe, SG_rev_spec_type specType, SG_stringarray** ppsa)
{
	SG_stringarray* psa = NULL;
	SG_uint32 count;
	SG_vector* pvecRevSpecs;

	SG_NULLARGCHECK_RETURN(pMe);
	pvecRevSpecs = pMe->pvecRevSpecs;
	SG_ERR_CHECK(  SG_vector__length(pCtx, pMe->pvecRevSpecs, &count)  );
	if (count)
	{
		SG_uint32 i;
		SG_ERR_CHECK(  SG_STRINGARRAY__ALLOC(pCtx, &psa, count)  );
		for (i = 0; i < count; i++)
		{
			_revspec* p;
			SG_ERR_CHECK(  SG_vector__get(pCtx, pvecRevSpecs, i, (void**)&p)  );
			if (p->specType == specType)
				SG_ERR_CHECK(  SG_stringarray__add(pCtx, psa, p->pszGiven)  );
		}
	}

	SG_RETURN_AND_NULL(psa, ppsa);

fail:
	SG_STRINGARRAY_NULLFREE(pCtx, psa);
}

void SG_rev_spec__revs(SG_context* pCtx, SG_rev_spec* pRevSpec, SG_stringarray** ppsaRevs)
{
	SG_ERR_CHECK_RETURN(  _to_stringarray(pCtx, (_sg_rev_spec*)pRevSpec, SPECTYPE_REV, ppsaRevs)  );
}

void SG_rev_spec__tags(SG_context* pCtx, SG_rev_spec* pRevSpec, SG_stringarray** ppsaTags)
{
	SG_ERR_CHECK_RETURN(  _to_stringarray(pCtx, (_sg_rev_spec*)pRevSpec, SPECTYPE_TAG, ppsaTags)  );
}

void SG_rev_spec__branches(SG_context* pCtx, SG_rev_spec* pRevSpec, SG_stringarray** ppsaBranches)
{
	SG_ERR_CHECK_RETURN(  _to_stringarray(pCtx, (_sg_rev_spec*)pRevSpec, SPECTYPE_BRANCH, ppsaBranches)  );
}

//////////////////////////////////////////////////////////////////

void SG_rev_spec__get_all(
	SG_context* pCtx, 
	const char* pszRepoName,
	SG_rev_spec* pRevSpec, 
	SG_bool bAllowAmbiguousBranches,
	SG_stringarray** ppsaFullHids)
{
	SG_repo* pRepo = NULL;

	SG_ERR_CHECK(  SG_REPO__OPEN_REPO_INSTANCE(pCtx, pszRepoName, &pRepo)  );
	SG_ERR_CHECK(  SG_rev_spec__get_all__repo(pCtx, pRepo, pRevSpec, bAllowAmbiguousBranches, ppsaFullHids)  );

fail:
	SG_REPO_NULLFREE(pCtx, pRepo);
}

void SG_rev_spec__get_all__repo(
	SG_context* pCtx, 
	SG_repo* pRepo,
	SG_rev_spec* pRevSpec, 
	SG_bool bAllowAmbiguousBranches,
	SG_stringarray** ppsaFullHids)
{
	_sg_rev_spec* pMe = (_sg_rev_spec*)pRevSpec;
	SG_uint32 count;
	SG_stringarray* psa = NULL;
	char* pszFullHid = NULL;
	SG_vhash* pvhCleanPileOfBranches = NULL;
	SG_varray* pvaBranchHids = NULL;

	SG_NULLARGCHECK_RETURN(pRevSpec);

	/* Don't add a null arg check for ppsaFullHids. It's legit to call us this way,
	 * simply to error check the rev spec. */

	SG_ERR_CHECK(  SG_vector__length(pCtx, pMe->pvecRevSpecs, &count)  );
	if (count)
	{
		SG_uint32 i;

		/* Get a clean pile of branches once. Not in each loop iteration. */
		SG_ERR_CHECK(  SG_vc_branches__cleanup(pCtx, pRepo, &pvhCleanPileOfBranches)  );

		SG_ERR_CHECK(  SG_STRINGARRAY__ALLOC(pCtx, &psa, count)  );

		for (i=0; i < count; i++)
		{
			_revspec* p;
			SG_ERR_CHECK(  SG_vector__get(pCtx, pMe->pvecRevSpecs, i, (void**)&p)  );
			switch (p->specType)
			{
			case SPECTYPE_REV:
				SG_ERR_CHECK(  SG_repo__hidlookup__dagnode(pCtx, pRepo, pMe->iDagNum, p->pszGiven, &pszFullHid)  );
				SG_ERR_CHECK(  SG_stringarray__add(pCtx, psa, pszFullHid)  );
				SG_NULLFREE(pCtx, pszFullHid);
				break;
			case SPECTYPE_TAG:
				if ( pMe->iDagNum != SG_DAGNUM__VERSION_CONTROL )
					SG_ERR_THROW2(SG_ERR_INVALID_REV_SPEC, (pCtx, "tags in a non-version-control query.") );
				SG_ERR_CHECK(  SG_vc_tags__lookup__tag(pCtx, pRepo, p->pszGiven, &pszFullHid)  );
				SG_ERR_CHECK(  SG_stringarray__add(pCtx, psa, pszFullHid)  );
				SG_NULLFREE(pCtx, pszFullHid);
				break;
			case SPECTYPE_BRANCH:
				if ( pMe->iDagNum != SG_DAGNUM__VERSION_CONTROL )
					SG_ERR_THROW2(SG_ERR_INVALID_REV_SPEC, (pCtx, "branches in a non-version-control query.") );
				if (bAllowAmbiguousBranches)
				{
					SG_uint32 j, countBranchHids;

					SG_ERR_CHECK(  _ambiguous_branch_lookup(pCtx, pvhCleanPileOfBranches, p->pszGiven, &pvaBranchHids)  );
					SG_ERR_CHECK(  SG_varray__count(pCtx, pvaBranchHids, &countBranchHids)  );
					for (j = 0; j < countBranchHids; j++)
					{
						const char* pszRefHid;
						SG_bool bExists;
						SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pvaBranchHids, j, &pszRefHid)  );
						SG_ERR_CHECK(  SG_repo__does_blob_exist(pCtx, pRepo, pszRefHid, &bExists)  );
						if (!bExists)
							SG_ERR_CHECK(  SG_vc_branches__throw_changeset_not_present(pCtx, p->pszGiven, pszRefHid)  );
						SG_ERR_CHECK(  SG_stringarray__add(pCtx, psa, pszRefHid)  );
					}
					SG_VARRAY_NULLFREE(pCtx, pvaBranchHids);
				}
				else
				{
					SG_bool bExists;
					SG_ERR_CHECK(  _unambiguous_branch_lookup(pCtx, pvhCleanPileOfBranches, p->pszGiven, &pszFullHid)  );
					SG_ERR_CHECK(  SG_repo__does_blob_exist(pCtx, pRepo, pszFullHid, &bExists)  );
					if (!bExists)
						SG_ERR_CHECK(  SG_vc_branches__throw_changeset_not_present(pCtx, p->pszGiven, pszFullHid)  );
					SG_ERR_CHECK(  SG_stringarray__add(pCtx, psa, pszFullHid)  );
					SG_NULLFREE(pCtx, pszFullHid);
				}
				break;
			default:
				SG_ERR_THROW2(SG_ERR_INVALID_REV_SPEC, (pCtx, "unknown rev spec type"));
			}
		}
	}

	SG_RETURN_AND_NULL(psa, ppsaFullHids);

	/* common cleanup */
fail:
	SG_VHASH_NULLFREE(pCtx, pvhCleanPileOfBranches);
	SG_STRINGARRAY_NULLFREE(pCtx, psa);
	SG_NULLFREE(pCtx, pszFullHid);
	SG_VARRAY_NULLFREE(pCtx, pvaBranchHids);
}

void SG_rev_spec__get_one(
	SG_context* pCtx,
	const char* pszRepoName,
	SG_rev_spec* pRevSpec,
	char** ppszChangesetHid,
	char** ppszBranchName)
{
	SG_repo* pRepo = NULL;

	SG_ERR_CHECK(  SG_REPO__OPEN_REPO_INSTANCE(pCtx, pszRepoName, &pRepo)  );
	SG_ERR_CHECK(  SG_rev_spec__get_one__repo(pCtx, pRepo, pRevSpec, ppszChangesetHid, ppszBranchName)  );

fail:
	SG_REPO_NULLFREE(pCtx, pRepo);
}

void SG_rev_spec__get_one__repo(
	SG_context* pCtx,
	SG_repo* pRepo,
	SG_rev_spec* pRevSpec,
	char** ppszChangesetHid,
	char** ppszBranchName)
{
	_sg_rev_spec* pMe = (_sg_rev_spec*)pRevSpec;
	SG_stringarray* psaAll = NULL;
	SG_uint32 countResults;

	SG_NULLARGCHECK_RETURN(pRevSpec);

	SG_ERR_CHECK(  SG_vector__length(pCtx, pMe->pvecRevSpecs, &countResults)  );
	if (countResults == 1)
	{
		const char* pszRef;

		SG_ERR_CHECK(  SG_rev_spec__get_all__repo(pCtx, pRepo, pRevSpec, SG_FALSE, &psaAll)  );
		
		if (ppszChangesetHid)
		{
			SG_ERR_CHECK(  SG_stringarray__get_nth(pCtx, psaAll, 0, &pszRef)  );
			SG_ERR_CHECK(  SG_STRDUP(pCtx, pszRef, ppszChangesetHid)  );
		}
		if (ppszBranchName)
		{
			SG_uint32 countBranchSpecs = 0;
			SG_ERR_CHECK(  SG_rev_spec__count_branches(pCtx, pRevSpec, &countBranchSpecs)  );
			if (countBranchSpecs == 1)
			{
				_revspec* p;
				SG_ERR_CHECK(  SG_vector__get(pCtx, pMe->pvecRevSpecs, 0, (void**)&p)  );
				SG_ERR_CHECK(  SG_STRDUP(pCtx, p->pszGiven, ppszBranchName)  );
			}
		}
	}
	else if (countResults == 0)
	{
		SG_ERR_THROW2(SG_ERR_NOT_FOUND, (pCtx, "no matching changesets"));
	}
	else
	{
		SG_ERR_THROW2(SG_ERR_LIMIT_EXCEEDED, (pCtx, "more than one match (%u)", countResults));
	}

	/* common cleanup */
fail:
	SG_STRINGARRAY_NULLFREE(pCtx, psaAll);
}

//////////////////////////////////////////////////////////////////

void SG_rev_spec__foreach__repo(SG_context* pCtx, SG_repo* pRepo, SG_rev_spec* pRevSpec, SG_rev_spec_foreach_callback* cb, void* ctx)
{
	_sg_rev_spec* pMe = (_sg_rev_spec*)pRevSpec;
	SG_uint32 i, countBranches;
	char* pszFullHid = NULL;
	SG_vhash* pvhCleanPileOfBranches = NULL;
	SG_varray* pvaFullHids = NULL;

	SG_NULLARGCHECK_RETURN(pRepo);
	SG_NULLARGCHECK_RETURN(pRevSpec);
	SG_NULLARGCHECK_RETURN(cb);

	SG_ERR_CHECK(  SG_vc_branches__cleanup(pCtx, pRepo, &pvhCleanPileOfBranches)  );

	SG_ERR_CHECK(  SG_vector__length(pCtx, pMe->pvecRevSpecs, &countBranches)  );
	for (i = 0; i < countBranches; i++)
	{
		_revspec* p;
		SG_uint32 j;
		SG_uint32 countBranchHids = 0;

		SG_ERR_CHECK(  SG_vector__get(pCtx, pMe->pvecRevSpecs, i, (void**)&p)  );
		switch (p->specType)
		{
		case SPECTYPE_REV:
			SG_ERR_CHECK(  SG_repo__hidlookup__dagnode(pCtx, pRepo, pMe->iDagNum, p->pszGiven, &pszFullHid)  );
			SG_ERR_CHECK(  cb(pCtx, pRepo, SPECTYPE_REV, p->pszGiven, pszFullHid, ctx)  );
			break;
		
		case SPECTYPE_TAG:
			if ( pMe->iDagNum != SG_DAGNUM__VERSION_CONTROL )
				SG_ERR_THROW2(SG_ERR_INVALID_REV_SPEC, (pCtx, "tags in a non-version-control query.") );
			SG_ERR_CHECK(  SG_vc_tags__lookup__tag(pCtx, pRepo, p->pszGiven, &pszFullHid)  );
			SG_ERR_CHECK(  cb(pCtx, pRepo, SPECTYPE_TAG, p->pszGiven, pszFullHid, ctx)  );
			break;
		
		case SPECTYPE_BRANCH:
			if ( pMe->iDagNum != SG_DAGNUM__VERSION_CONTROL )
				SG_ERR_THROW2(SG_ERR_INVALID_REV_SPEC, (pCtx, "branches in a non-version-control query.") );
			SG_ERR_CHECK(  _ambiguous_branch_lookup(pCtx, pvhCleanPileOfBranches, p->pszGiven, &pvaFullHids)  );
			SG_ERR_CHECK(  SG_varray__count(pCtx, pvaFullHids, &countBranchHids)  );
			for (j = 0; j < countBranchHids; j++)
			{
				const char* pszRefHid;
				SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pvaFullHids, j, &pszRefHid)  );
				SG_ERR_CHECK(  cb(pCtx, pRepo, SPECTYPE_BRANCH, p->pszGiven, pszRefHid, ctx)  );
			}
			SG_VARRAY_NULLFREE(pCtx, pvaFullHids);
			break;
		default:
			SG_ERR_THROW2(SG_ERR_INVALID_REV_SPEC, (pCtx, "unknown rev spec type"));
		}
		SG_NULLFREE(pCtx, pszFullHid);
	}

	SG_VHASH_NULLFREE(pCtx, pvhCleanPileOfBranches);

	return;

fail:
	SG_NULLFREE(pCtx, pszFullHid);
	SG_VHASH_NULLFREE(pCtx, pvhCleanPileOfBranches);
	SG_VARRAY_NULLFREE(pCtx, pvaFullHids);
}

//////////////////////////////////////////////////////////////////

struct _sg_rev_spec__contains__baton
{
	const char * szHid;
	SG_bool found;
};

static SG_rev_spec_foreach_callback _sg_rev_spec__contains__cb;
static void _sg_rev_spec__contains__cb(
	SG_context* pCtx, 
	SG_repo* pRepo,
	SG_rev_spec_type specType,
	const char* pszGiven,    /* The value as added to the SG_rev_spec. */
	const char* pszFullHid,	 /* The full HID of the looked-up spec. */
	void* ctx)
{
	struct _sg_rev_spec__contains__baton * baton = (struct _sg_rev_spec__contains__baton *)ctx;
	
	SG_UNUSED(pCtx);
	SG_UNUSED(pRepo);
	SG_UNUSED(specType);
	SG_UNUSED(pszGiven);

	if(baton->found)
		return;

	if(SG_stricmp(pszFullHid, baton->szHid)==0)
		baton->found = SG_TRUE;
}

void SG_rev_spec__contains(
	SG_context* pCtx, 
	SG_repo* pRepo, 
	SG_rev_spec* pRevSpec, 
	const char* szHid, 
	SG_bool* pFound)
{
	struct _sg_rev_spec__contains__baton baton;
	SG_NULLARGCHECK_RETURN(szHid);
	SG_NULLARGCHECK_RETURN(pFound);
	baton.szHid = szHid;
	baton.found = SG_FALSE;
	SG_ERR_CHECK_RETURN(  SG_rev_spec__foreach__repo(pCtx, pRepo, pRevSpec, _sg_rev_spec__contains__cb, &baton)  );
	*pFound = baton.found;
}

//////////////////////////////////////////////////////////////////

#define VHASH_DAGNUM "dagnum"
#define VHASH_REVS "hids"
#define VHASH_TAGS "tags"
#define VHASH_BRANCHES "branches"

/*	A rev spec vhash looks like this:
 *	{
 *		"dagnum" : <dagnum>
 *		"hids" :
 *		{
 *			<full hid> : NULL,
 *			<hid prefix> : NULL,
 *			<revno> : NULL,
 *			...
 *		}
 *		"tags" :
 *		{
 *			<tag name> : NULL,
 *			...
 *		}
 *		"branches" :
 *		{
 *			<branch name> : NULL,
 *			...
 *		}
 *	}
 */

static void _add_to_vhash_throw_nice_or_ignore_dupe(SG_context* pCtx, SG_vhash* pvh, const char* pszKey, SG_bool bIgnoreDupe)
{
	SG_vhash__add__null(pCtx, pvh, pszKey);
	if (bIgnoreDupe)
	{
		SG_ERR_CHECK_RETURN_CURRENT_DISREGARD(SG_ERR_VHASH_DUPLICATEKEY);
	}
	else
	{
		SG_ERR_REPLACE2(SG_ERR_VHASH_DUPLICATEKEY, SG_ERR_INVALID_REV_SPEC, (pCtx, "duplicate spec: %s", pszKey));
		SG_ERR_CHECK_RETURN_CURRENT;
	}
}

void SG_rev_spec__to_vhash(SG_context* pCtx, SG_rev_spec* pRevSpec, SG_bool bIgnoreDupes, SG_vhash** ppvh)
{
	_sg_rev_spec* pMe = (_sg_rev_spec*)pRevSpec;
	SG_vhash* pvhReturn = NULL;
	SG_vhash* pvhHids = NULL;
	SG_vhash* pvhTags = NULL;
	SG_vhash* pvhBranches = NULL;
	SG_uint32 countAllRevSpecs = 0;
	SG_uint32 i;
	char buf_dagnum[SG_DAGNUM__BUF_MAX__HEX];

	SG_NULLARGCHECK_RETURN(pRevSpec);
	SG_NULLARGCHECK_RETURN(ppvh);

	SG_ERR_CHECK(  SG_VHASH__ALLOC__PARAMS(pCtx, &pvhReturn, 4, NULL, NULL)  );

	SG_ERR_CHECK(  SG_dagnum__to_sz__hex(pCtx, pMe->iDagNum, buf_dagnum, sizeof(buf_dagnum))  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhReturn, VHASH_DAGNUM, buf_dagnum)  );

	if (pMe->iCountRevs)
		SG_ERR_CHECK(  SG_VHASH__ALLOC__PARAMS(pCtx, &pvhHids, pMe->iCountRevs, NULL, NULL)  );
	if (pMe->iCountTags)
		SG_ERR_CHECK(  SG_VHASH__ALLOC__PARAMS(pCtx, &pvhTags, pMe->iCountTags, NULL, NULL)  );
	if (pMe->iCountBranches)
		SG_ERR_CHECK(  SG_VHASH__ALLOC__PARAMS(pCtx, &pvhBranches, pMe->iCountBranches, NULL, NULL)  );
	
	SG_ERR_CHECK(  SG_vector__length(pCtx, pMe->pvecRevSpecs, &countAllRevSpecs)  );
	for (i = 0; i < countAllRevSpecs; i++)
	{
		_revspec* p;
		SG_ERR_CHECK(  SG_vector__get(pCtx, pMe->pvecRevSpecs, i, (void**)&p)  );
		switch (p->specType)
		{
		case SPECTYPE_REV:
			SG_ERR_CHECK(  _add_to_vhash_throw_nice_or_ignore_dupe(pCtx, pvhHids, p->pszGiven, bIgnoreDupes)  );
			break;

		case SPECTYPE_TAG:
			if ( pMe->iDagNum != SG_DAGNUM__VERSION_CONTROL )
				SG_ERR_THROW2(SG_ERR_INVALID_REV_SPEC, (pCtx, "tags in a non-version-control query.") );
			SG_ERR_CHECK(  _add_to_vhash_throw_nice_or_ignore_dupe(pCtx, pvhTags, p->pszGiven, bIgnoreDupes)  );
			break;

		case SPECTYPE_BRANCH:
			if ( pMe->iDagNum != SG_DAGNUM__VERSION_CONTROL )
				SG_ERR_THROW2(SG_ERR_INVALID_REV_SPEC, (pCtx, "branches in a non-version-control query.") );
			SG_ERR_CHECK(  _add_to_vhash_throw_nice_or_ignore_dupe(pCtx, pvhBranches, p->pszGiven, bIgnoreDupes)  );
			break;
		default:
			SG_ERR_THROW2(SG_ERR_INVALID_REV_SPEC, (pCtx, "unknown rev spec type"));
		}
	}

	if (pvhHids)
		SG_ERR_CHECK(  SG_vhash__add__vhash(pCtx, pvhReturn, VHASH_REVS, &pvhHids)  );
	if (pvhTags)
		SG_ERR_CHECK(  SG_vhash__add__vhash(pCtx, pvhReturn, VHASH_TAGS, &pvhTags)  );
	if (pvhBranches)
		SG_ERR_CHECK(  SG_vhash__add__vhash(pCtx, pvhReturn, VHASH_BRANCHES, &pvhBranches)  );

	SG_RETURN_AND_NULL(pvhReturn, ppvh);

fail:
	SG_VHASH_NULLFREE(pCtx, pvhReturn);
}

void SG_rev_spec__from_vash(SG_context* pCtx, const SG_vhash* pvh, SG_rev_spec** ppRevSpec)
{
	_sg_rev_spec* pMe = NULL;
	SG_rev_spec* pRevSpec = NULL;
	
	SG_bool bHas;
	SG_vhash* pvhTemp;
	SG_uint32 i, count;
	const char* pszTemp;

	SG_NULLARGCHECK_RETURN(pvh);
	SG_NULLARGCHECK_RETURN(ppRevSpec);

	SG_ERR_CHECK(  SG_REV_SPEC__ALLOC(pCtx, &pRevSpec)  );
	pMe = (_sg_rev_spec*)pRevSpec;

	/* A rev spec without a dagnum is invalid, so there's no need to do a __has lookup here first. */
	SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh, "dagnum", &pszTemp)  );
	SG_ERR_CHECK(  SG_dagnum__from_sz__hex(pCtx, pszTemp, &pMe->iDagNum)  );

	SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh, VHASH_REVS, &bHas)  );
	if (bHas)
	{
		SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvh, VHASH_REVS, &pvhTemp)  );
		SG_ERR_CHECK(  SG_vhash__count(pCtx, pvhTemp, &count)  );
		for (i = 0; i < count; i++)
		{
			SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvhTemp, i, &pszTemp, NULL)  );
			SG_ERR_CHECK(  SG_rev_spec__add_rev(pCtx, pRevSpec, pszTemp)  );
		}
	}

	SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh, VHASH_TAGS, &bHas)  );
	if (bHas)
	{
		SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvh, VHASH_TAGS, &pvhTemp)  );
		SG_ERR_CHECK(  SG_vhash__count(pCtx, pvhTemp, &count)  );
		for (i = 0; i < count; i++)
		{
			SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvhTemp, i, &pszTemp, NULL)  );
			SG_ERR_CHECK(  SG_rev_spec__add_tag(pCtx, pRevSpec, pszTemp)  );
		}
	}

	SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh, VHASH_BRANCHES, &bHas)  );
	if (bHas)
	{
		SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvh, VHASH_BRANCHES, &pvhTemp)  );
		SG_ERR_CHECK(  SG_vhash__count(pCtx, pvhTemp, &count)  );
		for (i = 0; i < count; i++)
		{
			SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvhTemp, i, &pszTemp, NULL)  );
			SG_ERR_CHECK(  SG_rev_spec__add_branch(pCtx, pRevSpec, pszTemp)  );
		}
	}

	SG_RETURN_AND_NULL(pRevSpec, ppRevSpec);

fail:
	SG_REV_SPEC_NULLFREE(pCtx, pRevSpec);
}

//////////////////////////////////////////////////////////////////

void SG_rev_spec__merge(SG_context* pCtx, SG_rev_spec* pRevSpecMergeInto, SG_rev_spec** ppRevSpecWillBeFreed)
{
	_sg_rev_spec* pTarget = (_sg_rev_spec*)pRevSpecMergeInto;
	_sg_rev_spec* pMerge;
	SG_uint32 count, i;

	SG_NULLARGCHECK_RETURN(pRevSpecMergeInto);
	if (!ppRevSpecWillBeFreed || !*ppRevSpecWillBeFreed)
		return;
	pMerge = (_sg_rev_spec*)*ppRevSpecWillBeFreed;

	if (pTarget->iDagNum != pMerge->iDagNum)
		SG_ERR_THROW2(SG_ERR_INVALIDARG, (pCtx, "rev specs have differing dagnums"));

	SG_ERR_CHECK(  SG_vector__length(pCtx, pMerge->pvecRevSpecs, &count)  );

	for (i = 0; i < count; i++)
	{
		_revspec* p;
		SG_ERR_CHECK(  SG_vector__get(pCtx, pMerge->pvecRevSpecs, i, (void**)&p)  );
		SG_ERR_CHECK(  _add_rev_spec(pCtx, pTarget, p->specType, p->pszGiven)  );
	}

	SG_REV_SPEC_NULLFREE(pCtx, *ppRevSpecWillBeFreed);

	return;

fail:
	;
}

END_EXTERN_C;
