/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
 * sg_tag.c
 *
 *  Created on: May 26, 2010
 *      Author: jeremy
 */

#include "sg.h"



void SG_tag__remove(SG_context * pCtx, SG_repo * pRepo, const char* psz_hid_cs, const char** paszTags, SG_uint32 count_tags)
{
    SG_audit q;
	char* psz_hid_assoc_with_tag = NULL;
	SG_uint32 count_valid_tags = 0;
	const char** paszValidTags = NULL;
	SG_uint32 i = 0;

    if (0 == count_tags)
		return;

    SG_ERR_CHECK(  SG_audit__init(pCtx,&q,pRepo,SG_AUDIT__WHEN__NOW,SG_AUDIT__WHO__FROM_SETTINGS)  );

	// SG_vc_tags__remove will throw on the first non-existant tag and stop
	// we don't want to issue errors, just warnings and keep going
	// weed out all the invalid tags here before calling SG_vc_tags__remove
	SG_ERR_CHECK(  SG_alloc(pCtx, count_tags, sizeof(const char*), &paszValidTags)  );
	for (i = 0; i < count_tags; i++)
	{
		SG_vc_tags__lookup__tag(pCtx, pRepo, paszTags[i], &psz_hid_assoc_with_tag);
		if (SG_context__err_equals(pCtx, SG_ERR_TAG_NOT_FOUND)) // tag does not exist
		{
			SG_context__err_reset(pCtx);
			SG_ERR_CHECK(  SG_log__report_warning(pCtx, "Tag not found:  %s\n", paszTags[i])  );
		}
		else
		{
			SG_ERR_CHECK_CURRENT;
			if (psz_hid_cs)
			{
				if (strcmp(psz_hid_cs, psz_hid_assoc_with_tag) == 0) // tag is assoc with given changeset
					paszValidTags[count_valid_tags++] = paszTags[i];
				else // tag not assoc with given changeset, no error, but warn
					SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDOUT, "Tag not associated with given revision:  %s\n", paszTags[i])  );
			}
			else
				paszValidTags[count_valid_tags++] = paszTags[i];
		}
		SG_NULLFREE(pCtx, psz_hid_assoc_with_tag);
	}

	SG_ERR_CHECK(  SG_vc_tags__remove(pCtx, pRepo, &q, count_valid_tags, paszValidTags)  );

fail:
	SG_NULLFREE(pCtx, paszValidTags);
	SG_NULLFREE(pCtx, psz_hid_assoc_with_tag);

}

void SG_tag__add(SG_context * pCtx, SG_repo * pRepo, const char* psz_hid_cs, const char** paszTags, SG_uint32 count_tags, SG_bool bForce)
{
	SG_audit q;
	SG_uint32 i = 0;
	char * psz_current_hid_with_that_tag = NULL;

	SG_NULLARGCHECK_RETURN(pRepo);
	SG_NULLARGCHECK_RETURN(psz_hid_cs);

	SG_ERR_CHECK(  SG_audit__init(pCtx,&q,pRepo,SG_AUDIT__WHEN__NOW,SG_AUDIT__WHO__FROM_SETTINGS)  );

	if (!bForce)
	{
		//Go through and check all tags to make sure that they are not already applied.
		for (i = 0; i < count_tags; i++)
		{
			const char * pszTag = paszTags[i];
			SG_vc_tags__lookup__tag(pCtx, pRepo, pszTag, &psz_current_hid_with_that_tag);
			if(SG_context__err_equals(pCtx, SG_ERR_TAG_NOT_FOUND))
				SG_context__err_reset(pCtx);
			else
			{
				SG_ERR_CHECK_CURRENT;
				if (0 != strcmp(psz_current_hid_with_that_tag, psz_hid_cs)) //The tag has been applied, but not to the given changeset.
					SG_ERR_THROW(SG_ERR_TAG_ALREADY_EXISTS);
			}
			SG_NULLFREE(pCtx, psz_current_hid_with_that_tag);
		}
	}
	for (i = 0; i < count_tags; i++)
	{
		const char * pszTag = paszTags[i];
		SG_vc_tags__lookup__tag(pCtx, pRepo, pszTag, &psz_current_hid_with_that_tag);
		if(SG_context__err_equals(pCtx, SG_ERR_TAG_NOT_FOUND))
		{
			SG_context__err_reset(pCtx);
			psz_current_hid_with_that_tag = NULL;
		}
		else
			SG_ERR_CHECK_CURRENT;
		if (psz_current_hid_with_that_tag == NULL || 0 != strcmp(psz_current_hid_with_that_tag, psz_hid_cs))
		{
			//The tag has not been applied, or it's been applied to a different dagnode.
			if ( psz_current_hid_with_that_tag != NULL && bForce)  //Remove it, if it's already there
					SG_ERR_CHECK(  SG_vc_tags__remove(pCtx, pRepo, &q, 1, &pszTag)  );
			SG_ERR_CHECK(  SG_vc_tags__add(pCtx, pRepo, psz_hid_cs, pszTag, &q)  );
		}
		SG_NULLFREE(pCtx, psz_current_hid_with_that_tag);
	}

fail:
	SG_NULLFREE(pCtx, psz_current_hid_with_that_tag);
}
