/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg_treenode_submodule__private_common.h
 *
 * @details p
 *
 */

//////////////////////////////////////////////////////////////////

#ifndef H_SG_TREENODE_SUBMODULE__PRIVATE_COMMON_H
#define H_SG_TREENODE_SUBMODULE__PRIVATE_COMMON_H

BEGIN_EXTERN_C;

//////////////////////////////////////////////////////////////////

void SG_treenode_submodule__get_version(SG_context * pCtx,
										const SG_treenode_submodule * pTSM,
										SG_treenode_version * pVer)
{
	// This field is present in the VHASH of ALL SUBMODULES.

	SG_int64 v64;

	SG_NULLARGCHECK_RETURN( pTSM );
	SG_NULLARGCHECK_RETURN( pVer );

	*pVer = SG_TN_VERSION__INVALID;

	SG_vhash__get__int64(pCtx, pTSM->pvh, KEY_VERSION, &v64);
	if (SG_context__err_equals(pCtx, SG_ERR_VHASH_KEYNOTFOUND))  // key not found, return __INVALID
	{
		SG_context__err_reset(pCtx);
		return;
	}

	SG_ERR_CHECK_RETURN_CURRENT;

	if (v64 > 0)									// if valid range, return it, otherwise __INVALID.
		*pVer = (SG_treenode_version)v64;
}

void SG_treenode_submodule__get_tech(SG_context * pCtx,
									 const SG_treenode_submodule * pTSM,
									 const char ** ppszTech)
{
	// This field is present in the VHASH of ALL SUBMODULES.

	SG_NULLARGCHECK_RETURN( pTSM );
	SG_NULLARGCHECK_RETURN( ppszTech );

	SG_ERR_CHECK_RETURN(  SG_vhash__get__sz(pCtx, pTSM->pvh, KEY_TECH, ppszTech)  );
}

//////////////////////////////////////////////////////////////////

void SG_treenode_submodule__get_id_ref(SG_context * pCtx,
									   const SG_treenode_submodule * pTSM,
									   const char ** ppszHid)
{
	SG_NULLARGCHECK_RETURN( pTSM );
	SG_NULLARGCHECK_RETURN( ppszHid );
	
	*ppszHid = pTSM->pszHidFrozen;
}

//////////////////////////////////////////////////////////////////

#if defined(DEBUG)
void SG_treenode_submodule_debug__dump_link__by_hid(SG_context * pCtx,
													const char * pszHid,
													SG_repo * pRepo,
													int indent,
													SG_string * pStringResult)
{
	// Dump the raw contents of the submodule link given by the HID.
	// The HID refers to a BLOB containing a JSON serialized VHASH.
	// We don't need to parse the JSON and build the VHASH only to print it,
	// so just splat the JSON directly from the BLOB.

	SG_byte * pBuf = NULL;
	SG_uint64 len = 0;

	SG_NONEMPTYCHECK_RETURN( pszHid );
	SG_NULLARGCHECK_RETURN( pRepo );
	SG_NULLARGCHECK_RETURN( pStringResult );

	SG_ERR_CHECK(  SG_repo__fetch_blob_into_memory(pCtx, pRepo, pszHid, &pBuf, &len)  );
	SG_ERR_CHECK(  SG_string__append__format(pCtx, pStringResult,
											 "%*cTreenodeSubmodule:  [hid %s]: ",
											 indent,' ',
											 pszHid)  );
	SG_ERR_CHECK(  SG_string__append__buf_len(pCtx, pStringResult, pBuf, (SG_uint32)len)  );
	SG_ERR_CHECK(  SG_string__append__sz(pCtx, pStringResult, "\n")  );

fail:
	SG_NULLFREE(pCtx, pBuf);
}

void SG_treenode_submodule_debug__dump_to_console(SG_context * pCtx,
												  const SG_treenode_submodule * pTSM,
												  const char * pszLabel)
{
	SG_string * pStrOut = NULL;

	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pStrOut)  );
	SG_ERR_CHECK(  SG_string__append__format(pCtx, pStrOut,
											 "TreenodeSubmodule: %s [hid %s][impl %s] %s\n",
											 pszLabel,
											 ((pTSM && pTSM->pszHidFrozen && *pTSM->pszHidFrozen) ? pTSM->pszHidFrozen : "(null)"),
											 ((pTSM && pTSM->pImpl) ? pTSM->pImpl->pszTech : "(null)"),
											 ((pTSM && pTSM->pPathWD) ? SG_pathname__sz(pTSM->pPathWD) : "(null)"))  );
	if (pTSM && pTSM->pvh)
	{
		SG_ERR_CHECK(  SG_string__append__sz(pCtx, pStrOut, "\t")  );
		SG_ERR_CHECK(  SG_vhash_debug__dump(pCtx, pTSM->pvh, pStrOut)  );
		SG_ERR_CHECK(  SG_string__append__sz(pCtx, pStrOut, "\n")  );
	}
	
	SG_ERR_CHECK(  SG_console__raw(pCtx, SG_CS_STDERR, SG_string__sz(pStrOut))  );

fail:
	SG_STRING_NULLFREE(pCtx, pStrOut);
}
#endif

//////////////////////////////////////////////////////////////////

static void _sg_treenode_submodule__to_json(SG_context * pCtx, SG_treenode_submodule * pTSM, SG_string ** ppString)
{
	SG_string * pString = NULL;

	// This must be the last step before we convert to JSON.
	// sort the vhash in increasing object gid order.  this is necessary
	// so that when we create the HID from the JSON stream we get a
	// consistent value (independent of the order that fields were added
	// to the vhash).

	SG_ERR_CHECK(  SG_vhash__sort(pCtx, pTSM->pvh, SG_TRUE, SG_vhash_sort_callback__increasing)  );
	
	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pString)  );
	SG_ERR_CHECK(  SG_vhash__to_json(pCtx, pTSM->pvh, pString)  );

	*ppString = pString;
	return;

fail:
	SG_STRING_NULLFREE(pCtx, pString);
}

static void _sg_treenode_submodule__freeze(SG_context * pCtx, SG_treenode_submodule * pTSM, SG_repo * pRepoOuter)
{
	char * pszHid_new = NULL;
	SG_string * pStringJSON = NULL;

	SG_ERR_CHECK(  _sg_treenode_submodule__to_json(pCtx, pTSM, &pStringJSON)  );
	SG_ERR_CHECK(  SG_repo__alloc_compute_hash__from_string(pCtx, pRepoOuter, pStringJSON, &pszHid_new)  );

	if (pTSM->pszHidFrozen)
	{
		SG_ASSERT(  (strcmp(pTSM->pszHidFrozen, pszHid_new) == 0)  );
	}
	else
	{
		pTSM->pszHidFrozen = pszHid_new;
		pszHid_new = NULL;
	}
	
fail:
	SG_STRING_NULLFREE(pCtx, pStringJSON);
	SG_NULLFREE(pCtx, pszHid_new);
}

//////////////////////////////////////////////////////////////////

static void _sg_treenode_submodule__validate(SG_context * pCtx,
											 const SG_treenode_submodule * pTSM)
{
	SG_NULLARGCHECK_RETURN( pTSM );
	SG_NULLARGCHECK_RETURN( pTSM->pImpl );

	SG_ERR_CHECK_RETURN(  (*pTSM->pImpl->pfn_validate)(pCtx, pTSM)  );
}
	
//////////////////////////////////////////////////////////////////

void SG_treenode_submodule__are_comparable(SG_context * pCtx,
										   const SG_treenode_submodule * pTSM1,
										   const SG_treenode_submodule * pTSM2,
										   SG_bool * pbSimilar)
{
	const _sg_treenode_submodule_impl * pImpl1 = ((pTSM1) ? pTSM1->pImpl : NULL);
	const _sg_treenode_submodule_impl * pImpl2 = ((pTSM2) ? pTSM2->pImpl : NULL);
	SG_bool bSimilar;

	// pTSM1 is optional
	// pTSM2 is optional
	SG_NULLARGCHECK_RETURN( pbSimilar );

	// if either or both are null, assume both links belong to the same repo family.
	// this may be a bit contrived, but an ADD operation doesn't store the TSM (so
	// neither ptn->pszCurrentHid nor ptn->pBaselineEntry['KEY_HID_BLOB'] will be
	// set), so a subsequent SCAN/STATUS may not have a reference for one side.

	if ((!pTSM1) || (!pTSM2))
	{
		bSimilar = SG_TRUE;
		goto done;
	}

	if (pImpl1 != pImpl2)
	{
		// if they refer to 2 different DVCS technologies, then we assume that
		// the links refer to 2 different WDs and/or REPOs.  if our caller is
		// STATUS/DIFF, it may want to consider them a DELETE+ADD rather than
		// a MODIFIED, for example.

		bSimilar = SG_FALSE;
		goto done;
	}
	
	SG_ERR_CHECK_RETURN(  (*pImpl1->pfn_comparable)(pCtx, pTSM1, pTSM2, &bSimilar)  );

done:
#if TRACE_SUB
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "TreenodeSubmodule_AreComparable: [result %d] [tech %s %s] [hid1 %s] [hid2 %s]\n",
							   bSimilar,
							   ((pImpl1) ? pImpl1->pszTech : "(null)"),
							   ((pImpl2) ? pImpl2->pszTech : "(null)"),
							   ((pTSM1) ? pTSM1->pszHidFrozen : "(null)"),
							   ((pTSM2) ? pTSM2->pszHidFrozen : "(null)"))  );
#endif

	*pbSimilar = bSimilar;
}

//////////////////////////////////////////////////////////////////

void SG_treenode_submodule__is_live(SG_context * pCtx,
									const SG_treenode_submodule * pTSM,
									SG_bool * pbIsLive)
{
	SG_NULLARGCHECK_RETURN( pTSM );
	SG_NULLARGCHECK_RETURN( pbIsLive );

	// pPathWD is only set when we are associated with an actual live WD.
	// otherwise, we just represent a TSM that was loaded from a blob in the repo.

	*pbIsLive = (pTSM->pPathWD != NULL);
}

//////////////////////////////////////////////////////////////////

void SG_treenode_submodule__are_same_hid(SG_context * pCtx,
										 const SG_treenode_submodule * pTSM1,
										 const SG_treenode_submodule * pTSM2,
										 SG_bool * pbSameHid)
{
	SG_NULLARGCHECK_RETURN( pTSM1 );
	SG_NULLARGCHECK_RETURN( pTSM2 );
	SG_NULLARGCHECK_RETURN( pbSameHid );
	SG_NONEMPTYCHECK_RETURN( pTSM1->pszHidFrozen );
	SG_NONEMPTYCHECK_RETURN( pTSM2->pszHidFrozen );

	*pbSameHid = (strcmp(pTSM1->pszHidFrozen, pTSM2->pszHidFrozen) == 0);
}

//////////////////////////////////////////////////////////////////

void SG_treenode_submodule__is_dirty(SG_context * pCtx,
									 SG_treenode_submodule * pTSM,
									 SG_bool * pbIsDirty)
{
	SG_bool bIsLive = SG_FALSE;

	SG_NULLARGCHECK_RETURN( pTSM );
	SG_NULLARGCHECK_RETURN( pbIsDirty );
	SG_NULLARGCHECK_RETURN( pTSM->pImpl );

	// only if we are live can we represent a dirty WD.

	SG_ERR_CHECK_RETURN(  SG_treenode_submodule__is_live(pCtx, pTSM, &bIsLive)  );
	if (bIsLive)
		SG_ERR_CHECK_RETURN(  (*pTSM->pImpl->pfn_is_dirty)(pCtx, pTSM, pbIsDirty)  );
	else
		*pbIsDirty = SG_FALSE;
}
	
//////////////////////////////////////////////////////////////////

void SG_treenode_submodule__describe_for_status(SG_context * pCtx,
												const SG_treenode_submodule * pTSM,
												char ** ppszDescription)
{
	SG_NULLARGCHECK_RETURN( pTSM );
	SG_NULLARGCHECK_RETURN( pTSM->pImpl );

	SG_ERR_CHECK_RETURN(  (*pTSM->pImpl->pfn_describe_for_status)(pCtx, pTSM, ppszDescription)  );
}

//////////////////////////////////////////////////////////////////

void SG_treenode_submodule__commit(SG_context * pCtx,
								   const SG_treenode_submodule * pTSM,
								   const char * pszComment,
								   SG_bool * pbDidSomething)
{
	SG_bool bDidSomething = SG_FALSE;

	SG_NULLARGCHECK_RETURN( pTSM );
	SG_NULLARGCHECK_RETURN( pTSM->pImpl );
	SG_NULLARGCHECK_RETURN( pTSM->pPathWD );
	// pbDidSomething is optional

// TODO 2010/12/16 convert all of the progress-type messages to use the successor to
// TODO            SG_context__msg__emit__format().

	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "Starting nested commit for submodule '%s'...\n",
							   SG_pathname__sz(pTSM->pPathWD))  );

	SG_ERR_CHECK_RETURN(  (*pTSM->pImpl->pfn_commit)(pCtx, pTSM, pszComment, &bDidSomething)  );

	// Success or fail we *DO NOT* update the TSM (we have a const pointer).
	// If the caller wants an updated TSM, they should create a new one.

	// TODO 2011/04/04 Make these console messages debug only or move to logging.
	if (bDidSomething)
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "Completed nested commit of submodule '%s'.\n",
								   SG_pathname__sz(pTSM->pPathWD))  );
	else
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "Nothing to do for nested commit of submodule '%s'.\n",
								   SG_pathname__sz(pTSM->pPathWD))  );

	if (pbDidSomething)
		*pbDidSomething = bDidSomething;
}

//////////////////////////////////////////////////////////////////

void SG_treenode_submodule__emit_skip_commit_message(SG_context * pCtx,
													 const SG_treenode_submodule * pTSM)
{
	SG_NULLARGCHECK_RETURN( pTSM );
	SG_NULLARGCHECK_RETURN( pTSM->pPathWD );

	// TODO 2011/04/04 Make these console messages debug only or move to logging.
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "Skipping nested commit for submodule '%s'.\n",
							   SG_pathname__sz(pTSM->pPathWD))  );
}

//////////////////////////////////////////////////////////////////

void SG_treenode_submodule__precompute_update(SG_context * pCtx,
											  const SG_pathname * pPathSubmodule,
											  const SG_treenode_submodule * pTSM_pend,
											  const SG_treenode_submodule * pTSM_goal,
											  const SG_vector * pvec_use_submodules,
											  SG_varray * pvaMissingDetails,
											  SG_treenode_submodule_instance_data_update ** ppTSIDU)
{
	SG_uint32 count_before, count_after;

	SG_NULLARGCHECK_RETURN( pTSM_pend );
	SG_NULLARGCHECK_RETURN( pTSM_pend->pImpl );
	// pTSM_pend->pPathWD may or may not be null but we don't rely on it.
	SG_NULLARGCHECK_RETURN( pTSM_goal );
	SG_NULLARGCHECK_RETURN( pTSM_goal->pImpl );
	SG_ASSERT(  (pTSM_goal->pPathWD == NULL)  );
	// pvec_use_submodules is optional
	SG_NULLARGCHECK_RETURN( pvaMissingDetails );
	SG_NULLARGCHECK_RETURN( ppTSIDU );

	// TODO 2011/04/04 Make these console messages debug only or move to logging.
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "Starting scan for nested update of submodule '%s'...\n",
							   SG_pathname__sz(pPathSubmodule))  );

	SG_ERR_CHECK_RETURN(  SG_varray__count(pCtx, pvaMissingDetails, &count_before)  );
	SG_ERR_CHECK_RETURN(  (*pTSM_pend->pImpl->pfn_precompute_update)(pCtx,
																	 pPathSubmodule,
																	 pTSM_pend,
																	 pTSM_goal,
																	 pvec_use_submodules,
																	 pvaMissingDetails,
																	 ppTSIDU)  );
	SG_ERR_CHECK_RETURN(  SG_varray__count(pCtx, pvaMissingDetails, &count_after)  );
	
	if (count_after > count_before)
	{
		SG_ASSERT(  (*ppTSIDU == NULL)  );
		// TODO 2011/04/04 Make these console messages debug only or move to logging.
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "%d issue(s) were found during scan for nested update of submodule '%s'...\n",
								   (count_after - count_before),
								   SG_pathname__sz(pPathSubmodule))  );
	}
	else
	{
		SG_ASSERT(  (*ppTSIDU != NULL)  );
		// TODO 2011/04/04 Make these console messages debug only or move to logging.
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "Completed scan for nested update of submodule '%s'...\n",
								   SG_pathname__sz(pPathSubmodule))  );
	}
}

void SG_treenode_submodule__apply_update(SG_context * pCtx,
										 const SG_pathname * pPathSubmodule,
										 const SG_treenode_submodule * pTSM_pend,
										 const SG_treenode_submodule * pTSM_goal,
										 SG_treenode_submodule_instance_data_update * pTSIDU)
{
	SG_NULLARGCHECK_RETURN( pTSM_pend );
	SG_NULLARGCHECK_RETURN( pTSM_pend->pImpl );
	// pTSM_pend->pPathWD may or may not be null but we don't rely on it.
	SG_NULLARGCHECK_RETURN( pTSM_goal );
	SG_NULLARGCHECK_RETURN( pTSM_goal->pImpl );
	SG_ASSERT(  (pTSM_goal->pPathWD == NULL)  );
	SG_NULLARGCHECK_RETURN( pTSIDU );

	// TODO 2011/04/04 Make these console messages debug only or move to logging.
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "Starting nested update of submodule '%s'...\n",
							   SG_pathname__sz(pPathSubmodule))  );

	SG_ERR_CHECK_RETURN(  (*pTSM_pend->pImpl->pfn_apply_update)(pCtx, pPathSubmodule, pTSM_pend, pTSM_goal, pTSIDU)  );

	// TODO 2011/04/04 Make these console messages debug only or move to logging.
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "Completed nested update of submodule '%s'...\n",
							   SG_pathname__sz(pPathSubmodule))  );
}

//////////////////////////////////////////////////////////////////

void SG_treenode_submodule__precompute_checkout(SG_context * pCtx,
												const SG_pathname * pPathSubmodule,
												const SG_treenode_submodule * pTSM,
												SG_bool bNoSubmodules,
												const SG_vector * pvec_use_submodules,
												SG_varray * pvaMissingDetails,
												SG_vhash ** ppvhCheckoutInstanceData)
{
	SG_uint32 count_before, count_after;

	SG_NULLARGCHECK_RETURN( pTSM );
	SG_NULLARGCHECK_RETURN( pTSM->pImpl );
	SG_ASSERT(  (pTSM->pPathWD == NULL)  );
	// pvec_use_submodules is optional
	SG_NULLARGCHECK_RETURN( pvaMissingDetails );
	SG_NULLARGCHECK_RETURN( ppvhCheckoutInstanceData );

	// TODO 2011/04/04 Make these console messages debug only or move to logging.
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "Starting scan for nested checkout of submodule '%s'...\n",
							   SG_pathname__sz(pPathSubmodule))  );

	SG_ERR_CHECK_RETURN(  SG_varray__count(pCtx, pvaMissingDetails, &count_before)  );
	SG_ERR_CHECK_RETURN(  (*pTSM->pImpl->pfn_precompute_checkout)(pCtx,
																  pPathSubmodule,
																  pTSM,
																  bNoSubmodules,
																  pvec_use_submodules,
																  pvaMissingDetails,
																  ppvhCheckoutInstanceData)  );
	SG_ERR_CHECK_RETURN(  SG_varray__count(pCtx, pvaMissingDetails, &count_after)  );
	
	if (count_after > count_before)
	{
		SG_ASSERT(  (*ppvhCheckoutInstanceData == NULL)  );
		// TODO 2011/04/04 Make these console messages debug only or move to logging.
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "%d issue(s) were found during scan for nested checkout of submodule '%s'...\n",
								   (count_after - count_before),
								   SG_pathname__sz(pPathSubmodule))  );
	}
	else
	{
		SG_ASSERT(  (*ppvhCheckoutInstanceData != NULL)  );
		// TODO 2011/04/04 Make these console messages debug only or move to logging.
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "Completed scan for nested checkout of submodule '%s'...\n",
								   SG_pathname__sz(pPathSubmodule))  );
	}
}

void SG_treenode_submodule__apply_checkout(SG_context * pCtx,
										   const SG_pathname * pPathSubmodule,
										   const SG_treenode_submodule * pTSM,
										   const SG_vhash * pvhCheckoutInstanceData)
{
	SG_NULLARGCHECK_RETURN( pTSM );
	SG_NULLARGCHECK_RETURN( pTSM->pImpl );
	SG_ASSERT(  (pTSM->pPathWD == NULL)  );
	SG_NULLARGCHECK_RETURN( pvhCheckoutInstanceData );

	// WARNING: pPathSubmodule might be a temp dir.

	// TODO 2011/04/04 Make these console messages debug only or move to logging.
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "Starting nested checkout of submodule '%s'...\n",
							   SG_pathname__sz(pPathSubmodule))  );

	SG_ERR_CHECK_RETURN(  (*pTSM->pImpl->pfn_apply_checkout)(pCtx, pPathSubmodule, pvhCheckoutInstanceData)  );

	// TODO 2011/04/04 Make these console messages debug only or move to logging.
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "Completed nested checkout of submodule '%s'...\n",
							   SG_pathname__sz(pPathSubmodule))  );
}

void SG_treenode_submodule__apply_checkout2(SG_context * pCtx,
											const SG_pathname * pPathSubmodule,
											const SG_vhash * pvhCheckoutInstanceData)
{
	// a variant of apply_checkout that does not have a TSM.

	const char * pszTechValue = NULL;
	const _sg_treenode_submodule_impl * pImpl = NULL;

	SG_NULLARGCHECK_RETURN( pPathSubmodule );
	SG_NULLARGCHECK_RETURN( pvhCheckoutInstanceData );

	// WARNING: pPathSubmodule might be a temp dir.

	SG_ERR_CHECK_RETURN(  SG_vhash__get__sz(pCtx, pvhCheckoutInstanceData, SG_CHECKOUT_INSTANCE_DATA__KEY__TECH, &pszTechValue)  );
	SG_ERR_CHECK_RETURN(  _sg_treenode_submodule__find_impl(pCtx, pszTechValue, &pImpl)  );

	// TODO 2011/04/04 Make these console messages debug only or move to logging.
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "Starting nested checkout of submodule '%s'...\n",
							   SG_pathname__sz(pPathSubmodule))  );

	SG_ERR_CHECK_RETURN(  (*pImpl->pfn_apply_checkout)(pCtx, pPathSubmodule, pvhCheckoutInstanceData)  );

	// TODO 2011/04/04 Make these console messages debug only or move to logging.
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "Completed nested checkout of submodule '%s'...\n",
							   SG_pathname__sz(pPathSubmodule))  );
}

//////////////////////////////////////////////////////////////////

void SG_treenode_submodule__precompute_revert(SG_context * pCtx,
											  const SG_pathname * pPathSubmodule,
											  const SG_treenode_submodule * pTSM,	// for vtable only
											  const SG_vector * pvec_use_submodules,
											  SG_bool bNoBackups,
											  SG_varray * pvaMissingDetails)
{
	SG_uint32 count_before, count_after;

	SG_NULLARGCHECK_RETURN( pTSM );
	SG_NULLARGCHECK_RETURN( pTSM->pImpl );
	// pTSM->pPathWD may or may not be null but we don't rely on it.
	// pvec_use_submodules is optional
	SG_NULLARGCHECK_RETURN( pvaMissingDetails );

	// TODO 2011/04/04 Make these console messages debug only or move to logging.
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "Starting scan for nested revert of submodule '%s'...\n",
							   SG_pathname__sz(pPathSubmodule))  );

	SG_ERR_CHECK_RETURN(  SG_varray__count(pCtx, pvaMissingDetails, &count_before)  );
	SG_ERR_CHECK_RETURN(  (*pTSM->pImpl->pfn_precompute_revert)(pCtx,
																pPathSubmodule,
																pvec_use_submodules,
																bNoBackups,
																pvaMissingDetails)  );
	SG_ERR_CHECK_RETURN(  SG_varray__count(pCtx, pvaMissingDetails, &count_after)  );
	
	if (count_after > count_before)
	{
		// TODO 2011/04/04 Make these console messages debug only or move to logging.
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "%d issue(s) were found during scan for nested revert of submodule '%s'...\n",
								   (count_after - count_before),
								   SG_pathname__sz(pPathSubmodule))  );
	}
	else
	{
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "Completed scan for nested revert of submodule '%s'...\n",
								   SG_pathname__sz(pPathSubmodule))  );
	}
}

void SG_treenode_submodule__apply_revert(SG_context * pCtx,
										 const SG_pathname * pPathSubmodule,
										 const SG_treenode_submodule * pTSM,		// for vtable only
										 const SG_vector * pvec_use_submodules,
										 SG_bool bNoBackups)
{
	SG_NULLARGCHECK_RETURN( pTSM );
	SG_NULLARGCHECK_RETURN( pTSM->pImpl );
	// pTSM->pPathWD may or may not be null but we don't rely on it.
	// pvec_use_submodules is optional

	// TODO 2011/04/04 Make these console messages debug only or move to logging.
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "Starting nested revert of submodule '%s'...\n",
							   SG_pathname__sz(pPathSubmodule))  );

	SG_ERR_CHECK_RETURN(  (*pTSM->pImpl->pfn_apply_revert)(pCtx, pPathSubmodule, pvec_use_submodules, bNoBackups)  );

	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "Completed nested revert of submodule '%s'...\n",
							   SG_pathname__sz(pPathSubmodule))  );
}

//////////////////////////////////////////////////////////////////

END_EXTERN_C;

#endif//H_SG_TREENODE_SUBMODULE__PRIVATE_COMMON_H
