/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg_treenode_submodule__private_impl_veracity.h
 *
 * @details 
 *
 */

//////////////////////////////////////////////////////////////////

#ifndef H_SG_TREENODE_SUBMODULE__PRIVATE_IMPL_VERACITY_H
#define H_SG_TREENODE_SUBMODULE__PRIVATE_IMPL_VERACITY_H

BEGIN_EXTERN_C;

//////////////////////////////////////////////////////////////////

// For a Veracity sub-repo, the sub-repo link contains only the
// repo-id and the cset-id.
//
// The idea is that a changeset in the containing/linking repo
// refers to a *specific* changeset in the contained/linked repo.
// WE LINK TO THE EXACT CHANGESET rather than to something like
// the 'head' or to a particular named branch.
//
// Also, WE RECORD ONLY THE GLOBALLY-UNIQUE REPO-ID
// rather than the repo-name (which is itself a per-user, local
// link) that was used to set up the sub's WD because we want
// a CANONICAL LINK.  How this gets mapped to a local repo-instance
// (via a repo-name aka repo-descriptor) is not our problem.

#define KEY_TECH__VERACITY__REPO_ID		"sgrepoid"
#define KEY_TECH__VERACITY__CSET_ID		"sgcsetid"

//////////////////////////////////////////////////////////////////

/**
 * Instance data that we keep in pTSM->pVoidData when necessary.
 */
struct _veracity__data
{
	SG_bool				bIsDirty;

	// TODO status/diff data??
};

typedef struct _veracity__data _veracity__data;

#define _VERACITY__DATA__NULLFREE(pCtx,p) SG_STATEMENT(SG_context__push_level(pCtx); _veracity__data__free(pCtx, p);   SG_ASSERT(!SG_context__has_err(pCtx));SG_context__pop_level(pCtx);p=NULL;)

static void _veracity__data__free(SG_context * pCtx,
								  _veracity__data * pData)
{
	if (!pData)
		return;

	SG_NULLFREE(pCtx, pData);
}

static void _veracity__data__alloc(SG_context * pCtx,
								   _veracity__data ** ppData)
{
	_veracity__data * pData = NULL;

	SG_ERR_CHECK_RETURN(  SG_alloc1(pCtx, pData)  );

	*ppData = pData;
}

//////////////////////////////////////////////////////////////////

static void _veracity__sanity_check_pendingtree(SG_context * pCtx,
												const SG_pathname * pPathCandidate,
												const SG_pendingtree * pPendingTree)
{
	// Our caller was given 'candidate' pathname to what is
	// supposed to be the root of the submodule.
	//
	// The general pendingtree-alloc code will start there and
	// walk back up the tree to find the first directory containing
	// a .sgdrawer.
	//
	// If the user did something wrong, say delete the .sgdrawer
	// directory in a submodule, the pendingtree-alloc will walk
	// back too far.
	//
	// __sniff_and_allocate() tries to catch this by ensuring that
	// the 'candidate' directory is a WD first.
	//
	// But weird things happen sometimes....
	//
	// So this is a sanity check to ensure that the pendingtree
	// that we received matches our expectations.

	const SG_pathname * pPathWorkingDirectoryTop;
	SG_bool bEqual;

	SG_ERR_CHECK(  SG_pendingtree__get_working_directory_top__ref(pCtx, pPendingTree, &pPathWorkingDirectoryTop)  );
	SG_ERR_CHECK(  SG_pathname__equal_ignoring_final_slashes(pCtx, pPathWorkingDirectoryTop, pPathCandidate, &bEqual)  );
	if (!bEqual)
	{
		SG_ERR_THROW2(  SG_ERR_ASSERT,
						(pCtx, "Sanity Check on submodule pathname: [expected %s] [computed %s]",
						 SG_pathname__sz(pPathCandidate),
						 SG_pathname__sz(pPathWorkingDirectoryTop))  );
	}

fail:
	return;
}

//////////////////////////////////////////////////////////////////

static void _veracity__allocate_from_live_wd(SG_context * pCtx,
											 const _sg_treenode_submodule_impl * pImpl,
											 const SG_pathname * pPathCandidate,
											 SG_repo * pRepoOuter,
											 SG_treenode_submodule ** ppTSM)
{
	SG_treenode_submodule * pTSM = NULL;
	SG_pendingtree * pPendingTreeSub = NULL;
	SG_repo * pRepoSub = NULL;					// we don't own this
	const SG_varray * pvaParentsSub = NULL;		// we don't own this
	char * pszRepoIdSub = NULL;
	const char * pszCSetId_0 = NULL;			// we don't own this

	SG_NULLARGCHECK_RETURN( pPathCandidate );
	SG_NULLARGCHECK_RETURN( pRepoOuter );
	SG_NULLARGCHECK_RETURN( ppTSM );

	SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pPathCandidate, &pPendingTreeSub)  );
	SG_ERR_CHECK(  _veracity__sanity_check_pendingtree(pCtx, pPathCandidate, pPendingTreeSub)  );
	SG_ERR_CHECK(  SG_pendingtree__get_repo(pCtx, pPendingTreeSub, &pRepoSub)  );

	SG_ERR_CHECK(  SG_repo__get_repo_id(pCtx, pRepoSub, &pszRepoIdSub)  );

	SG_ERR_CHECK(  SG_pendingtree__get_wd_parents__ref(pCtx, pPendingTreeSub, &pvaParentsSub)  );
	SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pvaParentsSub, 0, &pszCSetId_0)  );

	SG_ERR_CHECK(  SG_alloc1(pCtx, pTSM)  );
	SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pTSM->pvh)  );

	SG_ERR_CHECK(  SG_vhash__update__int64(     pCtx, pTSM->pvh, KEY_VERSION,                  SG_TN_VERSION__CURRENT)  );
	SG_ERR_CHECK(  SG_vhash__update__string__sz(pCtx, pTSM->pvh, KEY_TECH,                     SG_TREENODE_SUBMODULE__TECH__VERACITY)  );
	SG_ERR_CHECK(  SG_vhash__update__string__sz(pCtx, pTSM->pvh, KEY_TECH__VERACITY__REPO_ID,  pszRepoIdSub)  );
	SG_ERR_CHECK(  SG_vhash__update__string__sz(pCtx, pTSM->pvh, KEY_TECH__VERACITY__CSET_ID,  pszCSetId_0)  );

	// freeze the new TSM by computing the HID of the VHASH we just created.
	//
	// This may seem odd, but this TSM is a static snapshot of the
	// contents of the sub-module as it currently exists on disk.  And the HID
	// we compute here can be used to see if the outer WD currently refers to a
	// different CSET than in the outer WD's baseline.
	//
	// WARNING: We are juggling 2 repo pointers.  One for the super- and one for the sub-.
	// WARNING: We use the sub pointer to get details (repo-id) from the sub.  But since
	// WARNING: the treenode that we are creating is a link in the super, we use the super's
	// WARNING: hash algorithm during the freeze to make sure the treenode HID is compatible
	// WARNING: with the super.

	SG_ERR_CHECK(  _sg_treenode_submodule__freeze(pCtx, pTSM, pRepoOuter)  );

	pTSM->pImpl = pImpl;

	// Since this TSM refers to a live WD (that we can later do a STATUS/DIFF on),
	// we record the pathname.

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__COPY(pCtx, &pTSM->pPathWD, pPathCandidate)  );

	// We do not cache the sub's pendingtree in pVoidData because of VFILE locks.
	// Also, we defer diving into the submodule to look for dirt and/or do a
	// status/diff until explicitly requested.

	pTSM->pVoidData = NULL;

	SG_NULLFREE(pCtx, pszRepoIdSub);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTreeSub);

	*ppTSM = pTSM;
	return;

fail:
	SG_NULLFREE(pCtx, pszRepoIdSub);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTreeSub);
	SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pTSM);
}

static void _veracity__sniff(SG_context * pCtx,
							 const SG_pathname * pPathCandidate,
							 SG_bool * pbIsWD)
{
	
	SG_bool bIsWD = SG_FALSE;
	SG_UNUSED(pPathCandidate);

	SG_NULLARGCHECK_RETURN( pbIsWD );
#if 0	
	//TODO:  This should be restored when submodules comes back.
	//It was removed because excessive calls to stat on windows were
	//slowing down status.  
	SG_pathname * pPathMappedTo = NULL;
	SG_string * pStringRepoNameMappedTo = NULL;
	SG_fsobj_type fsType;
	SG_bool bExists;
	SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pPathCandidate, &bExists, &fsType, NULL)  );
	if (bExists  &&  (fsType == SG_FSOBJ_TYPE__DIRECTORY))
	{
		// See if the given candidate-path is the root directory of a working copy.
		// We cannot just try to open a pendingtree on it because pendingtree will silently
		// walk up the tree to containing directories without telling us (and our caller
		// may be holding the vfile lock on this outer pendingtree).
		//
		// So we do it the hard-way and directly see if there is a mapping for it.
		// BUT WE USE A SPECIAL VERSION OF SG_workingdir__find_mapping() that only
		// looks at the given directory and not at any parent directories as is normal.

		SG_workingdir__find_mapping2(pCtx, pPathCandidate,
									 SG_FALSE,	// do not recurse -- only eval the given directory
									 &pPathMappedTo, &pStringRepoNameMappedTo, NULL);
		if (SG_context__has_err(pCtx))
		{
			if (SG_context__err_equals(pCtx, SG_ERR_NOT_A_WORKING_COPY))
			{
				SG_context__err_reset(pCtx);
			}
			else
			{
				SG_ERR_RETHROW;
			}
		}
		else
		{
			bIsWD = SG_TRUE;
		}
	}
#endif
	*pbIsWD = bIsWD;
#if 0
fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathMappedTo);
	SG_STRING_NULLFREE(pCtx, pStringRepoNameMappedTo);
#endif
}
	
static void _impl__veracity__sniff_and_allocate(SG_context * pCtx,
												const _sg_treenode_submodule_impl * pImpl,
												const SG_pathname * pPathCandidate,
												SG_repo * pRepoOuter,
												SG_bool * pbIsWD,
												SG_treenode_submodule ** ppTSM)
{
	SG_NULLARGCHECK_RETURN( pbIsWD );
	// ppTSM is optional

	SG_ERR_CHECK(  _veracity__sniff(pCtx, pPathCandidate, pbIsWD)  );

	if ((*pbIsWD) && (ppTSM))
		SG_ERR_CHECK(  _veracity__allocate_from_live_wd(pCtx, pImpl, pPathCandidate, pRepoOuter, ppTSM)  );

fail:
	return;
}

//////////////////////////////////////////////////////////////////

static void _veracity__validate__v1(SG_context * pCtx,
									const SG_treenode_submodule * pTSM)
{
	const char * pszRepoId;
	const char * pszCSetId;

	SG_vhash__get__sz(pCtx, pTSM->pvh, KEY_TECH__VERACITY__REPO_ID, &pszRepoId);
	if (SG_context__has_err(pCtx))
		SG_ERR_RESET_THROW_RETURN(SG_ERR_TREENODE_VALIDATION_FAILED);

	SG_vhash__get__sz(pCtx, pTSM->pvh, KEY_TECH__VERACITY__CSET_ID, &pszCSetId);
	if (SG_context__has_err(pCtx))
		SG_ERR_RESET_THROW_RETURN(SG_ERR_TREENODE_VALIDATION_FAILED);
}

static void _impl__veracity__validate(SG_context * pCtx,
									  const SG_treenode_submodule * pTSM)
{
	const char * pszTech = NULL;
	SG_treenode_version ver = SG_TN_VERSION__INVALID;
	
	SG_NULLARGCHECK_RETURN( pTSM );
	SG_NULLARGCHECK_RETURN( pTSM->pvh );
	SG_NONEMPTYCHECK_RETURN( pTSM->pszHidFrozen );

	SG_vhash__get__sz(pCtx, pTSM->pvh, KEY_TECH, &pszTech);
	if (SG_context__has_err(pCtx))
		SG_ERR_RESET_THROW_RETURN(SG_ERR_TREENODE_VALIDATION_FAILED);

	if (strcmp(pszTech, SG_TREENODE_SUBMODULE__TECH__VERACITY) != 0)
		SG_ERR_THROW_RETURN(  SG_ERR_TREENODE_VALIDATION_FAILED  );

	SG_treenode_submodule__get_version(pCtx, pTSM, &ver);
	if (SG_context__has_err(pCtx))
		SG_ERR_RESET_THROW_RETURN(SG_ERR_TREENODE_VALIDATION_FAILED);

	switch (ver)
	{
	case SG_TN_VERSION_1:
		_veracity__validate__v1(pCtx, pTSM);
		if (SG_context__has_err(pCtx))
			SG_ERR_RESET_THROW_RETURN(SG_ERR_TREENODE_VALIDATION_FAILED);
		break;

	default:
		SG_ERR_THROW_RETURN(  SG_ERR_TREENODE_VALIDATION_FAILED  );
		break;
	}
}

//////////////////////////////////////////////////////////////////

static void _impl__veracity__dtor(SG_context * pCtx,
								  SG_treenode_submodule * pTSM)
{
	_veracity__data * pData;

	if (!pTSM || !pTSM->pVoidData)
		return;

	pData = (_veracity__data *)pTSM->pVoidData;
	pTSM->pVoidData = NULL;
	
	_VERACITY__DATA__NULLFREE(pCtx, pData);
}

//////////////////////////////////////////////////////////////////

static void _impl__veracity__copy(SG_context * pCtx,
								  const SG_treenode_submodule * pTSM_src,
								  SG_treenode_submodule * pTSM_dest)
{
	if (pTSM_src->pVoidData)
	{
		_veracity__data * pData_src;
		_veracity__data * pData_dest;

		SG_ERR_CHECK_RETURN(  _veracity__data__alloc(pCtx, ((_veracity__data **)&pTSM_dest->pVoidData))  );

		pData_src  = (_veracity__data *)pTSM_src->pVoidData;
		pData_dest = (_veracity__data *)pTSM_dest->pVoidData;

		pData_dest->bIsDirty = pData_src->bIsDirty;
	}
}

//////////////////////////////////////////////////////////////////

static void _impl__veracity__comparable(SG_context * pCtx,
										const SG_treenode_submodule * pTSM1,
										const SG_treenode_submodule * pTSM2,
										SG_bool * pbSimilar)
{
	const char * pszRepoId1 = NULL;
	const char * pszRepoId2 = NULL;
	SG_bool bSimilar = SG_FALSE;

	SG_ERR_CHECK_RETURN(  SG_vhash__get__sz(pCtx, pTSM1->pvh, KEY_TECH__VERACITY__REPO_ID, &pszRepoId1)  );
	SG_ERR_CHECK_RETURN(  SG_vhash__get__sz(pCtx, pTSM2->pvh, KEY_TECH__VERACITY__REPO_ID, &pszRepoId2)  );

	bSimilar = (strcmp(pszRepoId1, pszRepoId2) == 0);

	*pbSimilar = bSimilar;
}

//////////////////////////////////////////////////////////////////

static void _veracity__scan_for_dirt(SG_context * pCtx,
									 SG_treenode_submodule * pTSM)
{
	_veracity__data * pData = (_veracity__data *)pTSM->pVoidData;
	SG_pendingtree * pPendingTree = NULL;
	SG_treediff2 * pTreeDiff = NULL;
	SG_vhash * pVHash = NULL;
	const SG_varray * pva_wd_parents;		// we do not own this
	SG_uint32 nrParents;
	SG_uint32 nrSectionsWithChanges;
	SG_bool bIsDirty = SG_FALSE;

	SG_NULLARGCHECK_RETURN( pTSM->pPathWD );

	// TODO 2010/11/16 Decide how to report the status details. (and also the unresolved issues).
	// TODO            For now, just throw away treediff and everything.

	SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pTSM->pPathWD, &pPendingTree)  );

	// if PendingTree has multiple parents, it has a pending merge and is therefore dirty.

	SG_ERR_CHECK(  SG_pendingtree__get_wd_parents__ref(pCtx, pPendingTree, &pva_wd_parents)  );
	SG_ERR_CHECK(  SG_varray__count(pCtx, pva_wd_parents, &nrParents)  );
	if (nrParents > 1)
	{
		bIsDirty |= SG_TRUE;
	}
	else
	{
		SG_bool bVerbose = SG_FALSE;	// should UNDONE_ entries be included in status?

		SG_ERR_CHECK(  SG_pendingtree__diff_or_status(pCtx, pPendingTree, NULL, NULL, &pTreeDiff)  );

		// We *DO NOT* take a 'SG_file_spec *' or any of the filtering
		// stuff across the submodule boundary.  A STATUS on the outer
		// only reports that submodule is dirty; it does not dive into
		// and list details within.
		//
		// HOWEVER, we do need to filter out the IGNORES so that we don't
		// show false positives when reporting that the submodule is dirty.

		SG_ERR_CHECK(  SG_treediff2__file_spec_filter(pCtx, pTreeDiff, 0, NULL, SG_TRUE, gpBlankFilespec)  );

		// TODO 2010/12/08 After we have filtered the treediff, we ought
		// TODO            to be able to just count the number of items
		// TODO            remaining in the diff, *BUT* this gives false
		// TODO            positives because it still has entries for
		// TODO            MODIFIED DIRECTORIES which we consider noise
		// TODO            in some places and not in others.  Here they
		// TODO            are noise.  (Also there may be other noisy
		// TODO            items -- an undo-added maybe? -- that I don't
		// TODO            want to think about right now.
		// TODO
		// TODO            So, we let treediff build a status vhash
		// TODO            (which will get the same post-filtering-filtering
		// TODO            that the user sees when they type a 'vv status'
		// TODO            command.  This is expensive.
		// TODO
		// TODO            We could speed this up if we had a compatible
		// TODO            post-filtering-filter thing.
		// TODO
		// TODO            Alternatively, we could put the treediff or status vhash
		// TODO            in the pTSM->pVoidData in case it might be useful later.

		SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pVHash)  );
		SG_ERR_CHECK(  SG_treediff2__report_status_to_vhash(pCtx, pTreeDiff, pVHash, bVerbose)  );
		SG_ERR_CHECK(  SG_vhash__count(pCtx, pVHash, &nrSectionsWithChanges)  );
		if (nrSectionsWithChanges > 0)
			bIsDirty |= SG_TRUE;
	}
	
	pData->bIsDirty = bIsDirty;

fail:
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_TREEDIFF2_NULLFREE(pCtx, pTreeDiff);
	SG_VHASH_NULLFREE(pCtx, pVHash);
}

static void _impl__veracity__is_dirty(SG_context * pCtx,
									  SG_treenode_submodule * pTSM,
									  SG_bool * pbIsDirty)
{
	// pPathWD is only set when we are associated with an actual live WD.
	// otherwise, we just represent a TSM that was loaded from a blob in the repo.

	if (!pTSM->pPathWD)
	{
		*pbIsDirty = SG_FALSE;
		return;
	}

	if (pTSM->pVoidData)
	{
		// use cached result
	}
	else
	{
		SG_ERR_CHECK_RETURN(  _veracity__data__alloc(pCtx, ((_veracity__data **)&pTSM->pVoidData))  );
		SG_ERR_CHECK_RETURN(  _veracity__scan_for_dirt(pCtx, pTSM)  );
	}

	*pbIsDirty = ((_veracity__data *)pTSM->pVoidData)->bIsDirty;
}

//////////////////////////////////////////////////////////////////

static void _impl__veracity__describe_for_status(SG_context * pCtx,
												 const SG_treenode_submodule * pTSM,
												 char ** ppszDescription)
{
	const char * pszCSetId;

	SG_ERR_CHECK_RETURN(  SG_vhash__get__sz(pCtx, pTSM->pvh, KEY_TECH__VERACITY__CSET_ID, &pszCSetId)  );
	SG_ERR_CHECK_RETURN(  SG_STRDUP(pCtx, pszCSetId, ppszDescription)  );
}

//////////////////////////////////////////////////////////////////

static void _impl__veracity__commit(SG_context * pCtx,
									const SG_treenode_submodule * pTSM,
									const char * pszComment,
									SG_bool * pbDidSomething)
{
	SG_NULLARGCHECK_RETURN( pTSM->pPathWD );
	SG_NULLARGCHECK_RETURN( pbDidSomething );

	// TODO 2010/12/03 Should we use the changeset comment on the outer WD
	// TODO            exactly as is -or- should we prepend/append something
	// TODO            to their message to indicate that this was a nested
	// TODO            commit?

	// We only do full (non-partial) nested commits.  That is,
	// we do a complete commit within the submodule.  This is
	// mainly because we don't want the ambiguity of deciding
	// whether a set of --includes/--excludes/etc specified on
	// the outer WD should be applied to the submodule.  We also
	// do not pass a files_or_folders, because anything they
	// named on the command line is relative to the outer WD's
	// root directory and the outer treats the submodule as a
	// black box.
	//
	// This also makes it easier to support third-party DVCSs
	// because we don't have to translate include/exclude syntax
	// between them (and that is assuming that they even support
	// the concept).
	//
	// We DO NOT pass a userid across the boundary because the
	// submodule might be in a different admin domain than the
	// outer WD, and so the outer id may not be defined in the
	// inner.
	//
	// We DO NOT pass any associated WIT items to the submodule
	// since work items are defined in the outer WD/repo only.
	//
	// Likewise, we do not pass down the stamps because they
	// are to mark the outer.
	//
	// TODO 2010/12/03 Think about maybe passing down any date
	// TODO            arg that the user passed to the outer.

	SG_vv_verbs__commit(pCtx,
						SG_pathname__sz(pTSM->pPathWD),
						pszComment,
						gpBlankFilespec,	// We *DO NOT* do any filtering across the boundary
						SG_TRUE,			// bRecursive -- We *DO NOT* do any filtering across the boundary
						SG_FALSE,			// bNoSubmodules -- allow nested commit to also do nested commits
						SG_AUDIT__WHO__FROM_SETTINGS,
						SG_AUDIT__WHEN__NOW,
						0, NULL,			// We *DO NOT* pass args (files_or_folders).
						0, NULL,			// We *DO NOT* pass assocs (WITs)
						NULL,				// We *DO NOT* pass stamps.
						NULL, NULL,
						NULL);

	// Success or fail we *DO NOT* update the TSM (we have a const pointer).

	if (!SG_context__has_err(pCtx))
	{
		*pbDidSomething = SG_TRUE;
	}
	else if (SG_context__err_equals(pCtx, SG_ERR_NOTHING_TO_COMMIT))
	{
		// This probably means we only had FOUND-type dirt....
		SG_context__err_reset(pCtx);
		*pbDidSomething = SG_FALSE;
	}
	else
	{
		SG_ERR_RETHROW2_RETURN(  (pCtx, "Error during nested commit of submodule '%s'.", SG_pathname__sz(pTSM->pPathWD))  );
	}
}

//////////////////////////////////////////////////////////////////

static void _veracity__missing_details__update_needs_cset(SG_context * pCtx,
														  SG_varray * pvaMissingDetails,
														  const SG_pathname * pPathSubmodule,
														  SG_pendingtree * pPendingTree,
														  const char * pszHid_requested_cset_id)
{
	SG_vhash * pvhDetails = NULL;	// we never own this
	const SG_pathname * pPathWorkingDirectoryTop;
	SG_string * pStringRepoDescriptorName = NULL;
	SG_bool bEqual;

	SG_ERR_CHECK(  SG_pendingtree__get_working_directory_top__ref(pCtx, pPendingTree, &pPathWorkingDirectoryTop)  );
	SG_ERR_CHECK(  SG_pathname__equal_ignoring_final_slashes(pCtx, pPathWorkingDirectoryTop, pPathSubmodule, &bEqual)  );
	SG_ASSERT(  (bEqual)  );
	SG_ERR_CHECK(  SG_workingdir__find_mapping(pCtx, pPathWorkingDirectoryTop, NULL, &pStringRepoDescriptorName, NULL)  );

	SG_ERR_CHECK(  SG_varray__appendnew__vhash(pCtx, pvaMissingDetails, &pvhDetails)  );

	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "repo_name",      SG_string__sz(pStringRepoDescriptorName))  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "wd_path",        SG_pathname__sz(pPathSubmodule))  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "cset",           pszHid_requested_cset_id)  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "verb",           "update")  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "error",          "cset_not_in_repo")  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "detail_message", "An outer update needs to update the working copy of this submodule, but local repository does not have the required changeset.")  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "action_message", "Do a PULL on this repository and retry.")  );

fail:
	SG_STRING_NULLFREE(pCtx, pStringRepoDescriptorName);
}

static void _veracity__missing_details__checkout_needs_cset(SG_context * pCtx,
															SG_varray * pvaMissingDetails,
															const SG_pathname * pPathSubmodule,
															const char * pszRepoDescriptorName,
															const char * pszHid_requested_cset_id)
{
	SG_vhash * pvhDetails = NULL;	// we never own this

	SG_ERR_CHECK(  SG_varray__appendnew__vhash(pCtx, pvaMissingDetails, &pvhDetails)  );

	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "repo_name",      pszRepoDescriptorName)  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "wd_path",        SG_pathname__sz(pPathSubmodule))  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "cset",           pszHid_requested_cset_id)  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "verb",           "checkout")  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "error",          "cset_not_in_repo")  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "detail_message", "An outer checkout/update needs to do a checkout to create the submodule, but local repository does not have the required changeset.")  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "action_message", "Do a PULL on this repository and retry.")  );

fail:
	return;
}

static void _veracity__missing_details__cannot_open_repo(SG_context * pCtx,
														 SG_varray * pvaMissingDetails,
														 const SG_pathname * pPathSubmodule,
														 const char * pszRepoDescriptorName)
{
	SG_vhash * pvhDetails = NULL;	// we never own this

	SG_ERR_CHECK(  SG_varray__appendnew__vhash(pCtx, pvaMissingDetails, &pvhDetails)  );

	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "repo_name",      pszRepoDescriptorName)  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "wd_path",        SG_pathname__sz(pPathSubmodule))  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "verb",           "checkout")  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "error",          "cannot_open_repo")  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "detail_message", "An outer checkout/update needs to do a checkout to create the submodule, but local repository could not be opened.")  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "action_message", "Verify that this is a valid repository instance.")  );

fail:
	return;
}

static void _veracity__missing_details__checkout_needs_repo_instance(SG_context * pCtx,
																	 SG_varray * pvaMissingDetails,
																	 const SG_pathname * pPathSubmodule,
																	 const char * pszHid_repo_id)
{
	SG_vhash * pvhDetails = NULL;	// we never own this

	SG_ERR_CHECK(  SG_varray__appendnew__vhash(pCtx, pvaMissingDetails, &pvhDetails)  );

	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "repo_id",        pszHid_repo_id)  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "wd_path",        SG_pathname__sz(pPathSubmodule))  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "verb",           "checkout")  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "error",          "no_repo_instance")  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "detail_message", "An outer checkout/update needs to do a checkout to create the submodule, but there are no repository instances on the local system.")  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "action_message", "Do a CLONE to create this repository and retry.")  );

fail:
	return;
}

static void _veracity__missing_details__checkout_multiple_choices(SG_context * pCtx,
																  SG_varray * pvaMissingDetails,
																  const SG_pathname * pPathSubmodule,
																  const char * pszHid_repo_id,
																  SG_vhash ** ppvhListOfChoices)	// we steal the caller's list and null their pointer
{
	SG_vhash * pvhDetails = NULL;	// we never own this

	SG_ERR_CHECK(  SG_varray__appendnew__vhash(pCtx, pvaMissingDetails, &pvhDetails)  );

	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "repo_id",        pszHid_repo_id)  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "wd_path",        SG_pathname__sz(pPathSubmodule))  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "verb",           "checkout")  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "error",          "multiple_repo_instances")  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "detail_message", "An outer checkout/update needs to do a checkout to create the submodule, but there is more than one instance of the repository on the local system.")  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhDetails, "action_message", "TODO Explain how to specify a specific repo-name and retry.")  );
	SG_ERR_CHECK(  SG_vhash__add__vhash(     pCtx, pvhDetails, "choices",        ppvhListOfChoices)  );

fail:
	return;
}


//////////////////////////////////////////////////////////////////

struct _veracity__instance_data_update
{
	SG_pendingtree *		pPendingTree;
};

static void _veracity__precompute_update__dtor(SG_context * pCtx,
											   SG_treenode_submodule_instance_data_update * pTSIDU)
{
	if (!pTSIDU)
		return;

	SG_ASSERT(  (pTSIDU->pfn_dtor == _veracity__precompute_update__dtor)  );

	if (pTSIDU->pVoid)
	{
		struct _veracity__instance_data_update * pidu = (struct _veracity__instance_data_update *)pTSIDU->pVoid;

		SG_PENDINGTREE_NULLFREE(pCtx, pidu->pPendingTree);

		SG_NULLFREE(pCtx, pTSIDU->pVoid);
	}

	SG_NULLFREE(pCtx, pTSIDU);
}

static void _veracity__precompute_update__ctor(SG_context * pCtx,
											   SG_treenode_submodule_instance_data_update ** ppTSIDU,
											   SG_pendingtree * pPendingTree)
{
	SG_treenode_submodule_instance_data_update * pTSIDU = NULL;
	struct _veracity__instance_data_update * pidu = NULL;

	SG_ERR_CHECK(  SG_alloc1(pCtx, pTSIDU)  );
	pTSIDU->pfn_dtor = _veracity__precompute_update__dtor;

	SG_ERR_CHECK(  SG_alloc1(pCtx, pidu)  );
	pTSIDU->pVoid = (void *)pidu;

	pidu->pPendingTree = pPendingTree;
	
	*ppTSIDU = pTSIDU;
	return;

fail:
	SG_ERR_IGNORE(  _veracity__precompute_update__dtor(pCtx, pTSIDU)  );
}

//////////////////////////////////////////////////////////////////

static void _veracity__precompute_checkout__ctor(SG_context * pCtx,
												 SG_vhash ** ppvhCheckoutInstanceData,
												 const char * pszRepoDescriptorName,
												 const char * pszHidCSetId,
												 SG_bool bNoSubmodules,
												 SG_vhash ** ppvhSubs)
{
	SG_vhash * pvh = NULL;

	SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvh)  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvh, SG_CHECKOUT_INSTANCE_DATA__KEY__REPO_NAME, pszRepoDescriptorName)  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvh, SG_CHECKOUT_INSTANCE_DATA__KEY__CSET_ID, pszHidCSetId)  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvh, SG_CHECKOUT_INSTANCE_DATA__KEY__TECH, SG_TREENODE_SUBMODULE__TECH__VERACITY)  );

	SG_ERR_CHECK(  SG_vhash__add__bool(pCtx, pvh, SG_CHECKOUT_INSTANCE_DATA__KEY__NO_SUBS, bNoSubmodules)  );

	SG_ASSERT(  (ppvhSubs != NULL)  );
	if (bNoSubmodules)
	{
		SG_ASSERT(  (*ppvhSubs == NULL)  );
	}
	else
	{
		SG_ASSERT(  (*ppvhSubs != NULL)  );

		SG_ERR_CHECK(  SG_vhash__add__vhash(pCtx, pvh, SG_CHECKOUT_INSTANCE_DATA__KEY__SUBS, ppvhSubs)  );

		*ppvhSubs = NULL;	// we steal caller's vhash of info of stuff *within* this submodule.
	}

	*ppvhCheckoutInstanceData = pvh;
	pvh = NULL;

fail:
	SG_VHASH_NULLFREE(pCtx, pvh);
}

//////////////////////////////////////////////////////////////////

/**
 * Compute whether we could update the existing submodule WD to the
 * requested CSET.
 *
 * [1] does the repo-instance in the closet actually have the requested CSET.
 *
 * [2] go recursive into the submodule and see if the update would be allowed
 * [2a] are there sub-sub modules that need updating
 * [2b] is there dirt in the sub's WD that conflicts
 *
 * We are given an array of missing details.  If we have have issues preventing
 * the update, we append a vhash for each to it.  This will eventually stop the
 * outer update (but let the outer continue to accumulate info for a
 * comprehensive message to the user).
 *
 * If we can do the update, we return instance data that will be used in the apply phase.
 *
 * We try to only throw if we have a hard error.
 */
static void _impl__veracity__precompute_update(SG_context * pCtx,
											   const SG_pathname * pPathSubmodule,
											   const SG_treenode_submodule * pTSM_pend,
											   const SG_treenode_submodule * pTSM_goal,
											   const SG_vector * pvec_use_submodules,
											   SG_varray * pvaMissingDetails,
											   SG_treenode_submodule_instance_data_update ** ppTSIDU)
{
	SG_treenode_submodule_instance_data_update * pTSIDU = NULL;
	SG_pendingtree * pPendingTreeAllocated = NULL;
	SG_changeset * pcs = NULL;
	const char * pszHid_goal_cset_id;
	SG_repo * pRepo;	// we don't own this
	SG_uint32 count_before, count_after;

	SG_NULLARGCHECK_RETURN( pPathSubmodule );
	SG_NULLARGCHECK_RETURN( pTSM_pend );
	SG_NULLARGCHECK_RETURN( pTSM_goal );
	// pvec_use_submodules is optional
	SG_NULLARGCHECK_RETURN( pvaMissingDetails );
	SG_NULLARGCHECK_RETURN( ppTSIDU );

#if defined(DEBUG)
	// pTSM_pend should have come from Ndx_Pend and therefore
	// represent a live TSM.  but there are times when we cloned
	// if from Ndx_Dest in treediff.  Only the Ndx_Pend version
	// has a pathname inside it.
	if (pTSM_pend->pPathWD)
		SG_ARGCHECK_RETURN( (strcmp(SG_pathname__sz(pTSM_pend->pPathWD),
									SG_pathname__sz(pPathSubmodule)) == 0), pPathSubmodule );
#endif

	// [1] see if the backing repo has the requested CSET.

	SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx,
										 pPathSubmodule,
										 &pPendingTreeAllocated)  );
	SG_ERR_CHECK(  SG_pendingtree__get_repo(pCtx, pPendingTreeAllocated, &pRepo)  );

	SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pTSM_goal->pvh, KEY_TECH__VERACITY__CSET_ID, &pszHid_goal_cset_id)  );

	//TODO:  Is there a lighter weight way to check this?  Somethimes changesets can be very large,
	//       so loading them should be avoided if possible.  --jeremy
	SG_changeset__load_from_repo(pCtx, pRepo, pszHid_goal_cset_id, &pcs);
	if (SG_context__has_err(pCtx))
	{
		if (SG_context__err_equals(pCtx, SG_ERR_CHANGESET_BLOB_NOT_FOUND))
		{
			SG_context__err_reset(pCtx);
			SG_ERR_CHECK(  _veracity__missing_details__update_needs_cset(pCtx,
																		 pvaMissingDetails,
																		 pPathSubmodule,
																		 pPendingTreeAllocated,
																		 pszHid_goal_cset_id)  );
			goto done;
		}

		SG_ERR_RETHROW2(  (pCtx, "Error fetching changeset '%s' in submodule '%s'.",
						   pszHid_goal_cset_id,
						   SG_pathname__sz(pPathSubmodule))  );
	}
	
	// the repo/closet knows about the goal changeset.
	//
	// [2] do recursive/nested update precompute to see if
	// there are any conflicts that will cause us problems.

	SG_ERR_CHECK(  SG_varray__count(pCtx, pvaMissingDetails, &count_before)  );
	SG_ERR_CHECK(  SG_pendingtree__update_baseline__precompute(pCtx,
															   pPendingTreeAllocated,
															   pszHid_goal_cset_id,
															   pvec_use_submodules,
															   pvaMissingDetails)  );
	SG_ERR_CHECK(  SG_varray__count(pCtx, pvaMissingDetails, &count_after)  );
	if (count_after > count_before)
		goto done;

	// we can do the update.  return our instance data for the
	// apply-phase.

	SG_ERR_CHECK(  _veracity__precompute_update__ctor(pCtx, &pTSIDU, pPendingTreeAllocated)  );
	pPendingTreeAllocated = NULL;				// instance data now owns our pendingtree

#if 0 && defined(DEBUG)
	{
		const char * pszHid_pend_cset_id;

		SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pTSM_pend->pvh, KEY_TECH__VERACITY__CSET_ID, &pszHid_pend_cset_id)  );
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "Nested update is possible for submodule '%s' from changeset '%s' to '%s'.\n",
								   SG_pathname__sz(pPathSubmodule),
								   pszHid_pend_cset_id,
								   pszHid_goal_cset_id)  );
	}
#endif

done:
	*ppTSIDU = pTSIDU;	pTSIDU = NULL;

fail:
	SG_ERR_IGNORE(  _veracity__precompute_update__dtor(pCtx, pTSIDU)  );
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTreeAllocated);
	SG_CHANGESET_NULLFREE(pCtx, pcs);
}

static void _impl__veracity__apply_update(SG_context * pCtx,
										  const SG_pathname * pPathSubmodule,
										  const SG_treenode_submodule * pTSM_pend,
										  const SG_treenode_submodule * pTSM_goal,
										  SG_treenode_submodule_instance_data_update * pTSIDU)
{
	struct _veracity__instance_data_update * pidu = (struct _veracity__instance_data_update *)pTSIDU->pVoid;
	SG_pendingtree_action_log_enum eActionLog = SG_PT_ACTION__DO_IT;
	const SG_pathname * pPathWorkingDirectoryTop;
	SG_bool bEqual;

	// make sure that they didn't move/rename the submodule (or a parent directory)
	// between the __precompute__ and the __apply__ (because we have cached pathnames
	// in the sub's pendingtree).

	SG_ERR_CHECK(  SG_pendingtree__get_working_directory_top__ref(pCtx, pidu->pPendingTree, &pPathWorkingDirectoryTop)  );
	SG_ERR_CHECK(  SG_pathname__equal_ignoring_final_slashes(pCtx, pPathWorkingDirectoryTop, pPathSubmodule, &bEqual)  );

	if (!bEqual)
	{
		// The path given to precompute should match the path we are given.
		// If not, then the caller perhaps did some moves/renames on the
		// outer between the 2 calls.  This might invalidates the info we cached
		// in the sub-pendingtree.  (or perhaps we shouldn't have cached the
		// submodule's absolute pathname...)

		SG_ERR_THROW2(  SG_ERR_NOTIMPLEMENTED,
							   (pCtx, "Nested submodule path changed [was '%s'] [now '%s'].",
								SG_pathname__sz(pPathWorkingDirectoryTop),
								SG_pathname__sz(pPathSubmodule))  );
	}

	// We don't need the TSMs because the cached pendingtree has all of
	// of the info that we need to do the work.  I'm leaving them in the
	// API because other DVCSs might not be this lucky.

	SG_UNUSED( pTSM_pend );
	SG_UNUSED( pTSM_goal );
	
	// actually do the update.
	//
	// untie the submodule from any named branch
	// because we forced a specific changeset.

	SG_ERR_CHECK(  SG_pendingtree__update_baseline__apply(pCtx,
														  pidu->pPendingTree,
														  eActionLog,
														  NULL)  );

fail:
	return;
}

//////////////////////////////////////////////////////////////////

static void _could_do_checkout__search_tree(SG_context * pCtx,
											const SG_pathname * pPath,
											SG_repo * pRepo,
											const char * pszHidDirectory,
											const SG_vector * pvec_use_submodules,
											SG_varray * pvaMissingDetails,
											SG_vhash * pvhSubs)
{
	SG_treenode * pTreenode = NULL;
	SG_pathname * pPath_k = NULL;
	SG_treenode_submodule * pTSM_k = NULL;
	SG_vhash * pvhSubs_k = NULL;
	SG_uint32 kEntry, nrEntries;

	// The caller should only call us if they wanted a search for submodules.
	SG_bool bNoSubmodules = SG_FALSE;

	SG_ERR_CHECK(  SG_treenode__load_from_repo(pCtx, pRepo, pszHidDirectory, &pTreenode)  );
	SG_ERR_CHECK(  SG_treenode__count(pCtx, pTreenode, &nrEntries)  );
	for (kEntry=0; kEntry < nrEntries; kEntry++)
	{
		const SG_treenode_entry * pTreenodeEntry_k;
		const char * pszGid_k;
		const char * pszName_k;
		const char * pszHidSubDirectory_k;
		const char * pszHidSubmodule_k;
		SG_treenode_entry_type type_k;

		SG_ERR_CHECK(  SG_treenode__get_nth_treenode_entry__ref(pCtx, pTreenode, kEntry,
																&pszGid_k, &pTreenodeEntry_k)  );
		SG_ERR_CHECK(  SG_treenode_entry__get_entry_type(pCtx, pTreenodeEntry_k, &type_k)  );
		switch (type_k)
		{
		default:
			//case SG_TREENODEENTRY_TYPE_REGULAR_FILE:
			//case SG_TREENODEENTRY_TYPE_SYMLINK:
			break;
			
		case SG_TREENODEENTRY_TYPE_DIRECTORY:
			SG_ERR_CHECK(  SG_treenode_entry__get_entry_name(pCtx, pTreenodeEntry_k, &pszName_k)  );
			SG_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx, &pPath_k, pPath, pszName_k)  );
			SG_ERR_CHECK(  SG_treenode_entry__get_hid_blob(pCtx, pTreenodeEntry_k, &pszHidSubDirectory_k)  );
			SG_ERR_CHECK(  _could_do_checkout__search_tree(pCtx, pPath_k, pRepo, pszHidSubDirectory_k, pvec_use_submodules, pvaMissingDetails, pvhSubs)  );
			SG_PATHNAME_NULLFREE(pCtx, pPath_k);
			break;

		case SG_TREENODEENTRY_TYPE_SUBMODULE:
			SG_ERR_CHECK(  SG_treenode_entry__get_entry_name(pCtx, pTreenodeEntry_k, &pszName_k)  );
			SG_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx, &pPath_k, pPath, pszName_k)  );
			SG_ERR_CHECK(  SG_treenode_entry__get_hid_blob(pCtx, pTreenodeEntry_k, &pszHidSubmodule_k)  );
			SG_ERR_CHECK(  SG_treenode_submodule__load_from_repo(pCtx, pRepo, pszHidSubmodule_k, &pTSM_k)  );
			SG_ERR_CHECK(  SG_treenode_submodule__precompute_checkout(pCtx, pPath_k, pTSM_k, bNoSubmodules, pvec_use_submodules, pvaMissingDetails, &pvhSubs_k)  );
            if (pvhSubs_k)
            {
                SG_ERR_CHECK(  SG_vhash__add__vhash(pCtx, pvhSubs, pszGid_k, &pvhSubs_k)  );
                SG_ASSERT(  (pvhSubs_k == NULL)  );
            }
            else
            {
                // code elsewhere depends on the old behavior of add__vhash,
                // which added a null when given a null.
                SG_ERR_CHECK(  SG_vhash__add__null(pCtx, pvhSubs, pszGid_k)  );
            }
			SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pTSM_k);
			SG_PATHNAME_NULLFREE(pCtx, pPath_k);
			break;
		}
	}

fail:
	SG_TREENODE_NULLFREE(pCtx, pTreenode);
	SG_PATHNAME_NULLFREE(pCtx, pPath_k);
	SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pTSM_k);
	SG_VHASH_NULLFREE(pCtx, pvhSubs_k);
}

/**
 * See if we can open the named repo-instance and if it has the desired cset.
 * Walk the treenodes and verify that we *could* do a checkout.  That is,
 * ensure that we could checkout this (repo_descriptor_name,cset_id) *and*
 * also checkout any submodules within.
 *
 * If bNoSubmodules is FALSE, we DO NOT dive into contained submodules.
 * 
 * If not, update pvaMissingDetails with reasons.
 *
 * If so, return vhash of instance data for this submodule.
 *
 * TODO 2011/01/21 This function doesn't really belong in this file.  It
 * TODO            isn't actually doing anything with the caller's submodule.
 */
void SG_treenode_submodule_veracity__could_do_checkout(SG_context * pCtx,
													   const SG_pathname * pPath,	// we only use this for error messages
													   const char * pszRepoDescriptorName,
													   const char * pszHid_cset_id,
													   SG_bool bNoSubmodules,
													   const SG_vector * pvec_use_submodules,
													   SG_varray * pvaMissingDetails,
													   SG_vhash ** ppvhCheckoutInstanceData)
{
	SG_repo * pRepo = NULL;
	SG_changeset * pcs = NULL;
	SG_vhash * pvhSubs = NULL;
	SG_treenode * pTreenodeSuperRoot = NULL;
	const SG_treenode_entry * pTreenodeEntry;
	const char * pszHidSuperRoot;
	const char * pszHidDirectory;
	SG_uint32 count_before, count_after;

	*ppvhCheckoutInstanceData = NULL;

	// try to open the selected repo-instance by name in the closet.

	SG_REPO__OPEN_REPO_INSTANCE(pCtx, pszRepoDescriptorName, &pRepo);
	if (SG_context__has_err(pCtx))
	{
		SG_context__err_reset(pCtx);
		SG_ERR_CHECK(  _veracity__missing_details__cannot_open_repo(pCtx,
																	pvaMissingDetails,
																	pPath,
																	pszRepoDescriptorName)  );
		goto done;
	}

	// try to get the requested cset.
	//TODO:  Is there a lighter weight way to check this?  Somethimes changesets can be very large,
	//       so loading them should be avoided if possible.  --jeremy
	SG_changeset__load_from_repo(pCtx, pRepo, pszHid_cset_id, &pcs);
	if (SG_context__has_err(pCtx))
	{
		if (SG_context__err_equals(pCtx, SG_ERR_CHANGESET_BLOB_NOT_FOUND))
		{
			SG_context__err_reset(pCtx);
			SG_ERR_CHECK(  _veracity__missing_details__checkout_needs_cset(pCtx,
																		   pvaMissingDetails,
																		   pPath,
																		   pszRepoDescriptorName,
																		   pszHid_cset_id)  );
			goto done;
		}

		SG_ERR_RETHROW2(  (pCtx, "Error fetching changeset '%s' in submodule '%s'.",
						   pszHid_cset_id,
						   SG_pathname__sz(pPath))  );
	}

	if (bNoSubmodules)
	{
		// they don't want the contained submodules expanded.
		// this is fine for EXPORT, but questionable for CHECKOUT.
		// the caller should guard this.
		//
		// we build a trivial instance data for just this repo
		// and no submodule info.
	}
	else
	{
		// walk the treenodes within the submodule and look
		// for submoules within it and make sure that they
		// too can be checked out.
		//
		// this will build a vhash[gid_sub ==> vhash_checkout_instance_data]
		// in pvhSubs.

		SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvhSubs)  );
		SG_ERR_CHECK(  SG_varray__count(pCtx, pvaMissingDetails, &count_before)  );
		SG_ERR_CHECK(  SG_changeset__tree__get_root(pCtx, pcs, &pszHidSuperRoot)  );
		SG_ERR_CHECK(  SG_treenode__load_from_repo(pCtx, pRepo, pszHidSuperRoot, &pTreenodeSuperRoot)  );
		SG_ERR_CHECK(  SG_treenode__get_nth_treenode_entry__ref(pCtx, pTreenodeSuperRoot, 0, NULL, &pTreenodeEntry)  );
		SG_ERR_CHECK(  SG_treenode_entry__get_hid_blob(pCtx, pTreenodeEntry, &pszHidDirectory)  );
		SG_ERR_CHECK(  _could_do_checkout__search_tree(pCtx,
													   pPath,
													   pRepo,
													   pszHidDirectory,
													   pvec_use_submodules,
													   pvaMissingDetails,
													   pvhSubs)  );
		SG_ERR_CHECK(  SG_varray__count(pCtx, pvaMissingDetails, &count_after)  );
		if (count_after > count_before)
			goto done;		// don't throw, just return the missing details data
	
		// all went well.  we could do a complete checkout from
		// this repo-instance (and any subs referenced within)
		// into the WD.  return our instance data for the apply
		// phase.

#if 0 && defined(DEBUG)
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "Nested checkout is possible for submodule '%s' at changeset '%s' using repository '%s'.\n",
								   SG_pathname__sz(pPath),
								   pszHid_cset_id,
								   pszRepoDescriptorName)  );
#endif

	}
	
	SG_ERR_CHECK(  _veracity__precompute_checkout__ctor(pCtx,
														ppvhCheckoutInstanceData,
														pszRepoDescriptorName,
														pszHid_cset_id,
														bNoSubmodules,
														&pvhSubs)  );
	SG_ASSERT(  (pvhSubs == NULL)  );


done:	// like fail target but without the throw
	;

fail:
	SG_CHANGESET_NULLFREE(pCtx, pcs);
	SG_REPO_NULLFREE(pCtx, pRepo);
	SG_VHASH_NULLFREE(pCtx, pvhSubs);
	SG_TREENODE_NULLFREE(pCtx, pTreenodeSuperRoot);
}

static void _veracity__precompute_checkout__try_to_reverse_map_repo_id(SG_context * pCtx,
																	   const SG_pathname * pPathSubmodule,
																	   const SG_treenode_submodule * pTSM,
																	   const SG_vector * pvec_use_submodules,
																	   SG_varray * pvaMissingDetails,
																	   char ** ppszRepoDescriptorName)
{
	const char * pszHid_repo_id;
	char * pszRepoDescriptorName = NULL;
	SG_vhash * pvhListOfChoices = NULL;
	SG_uint32 nrChoices;

	// get the repo_id from the submodule-link.
	// this represents the CANONICAL identity of the repo.
	// 
	// note that the submodule-link ***DOES NOT*** have a
	// repo-name or repo-descriptor.
	//
	// if we eventually do do the checkout, the submodule
	// must be bound to a specific (named and local)
	// repo-instance of this repo.

	SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pTSM->pvh, KEY_TECH__VERACITY__REPO_ID,  &pszHid_repo_id )  );

	// use the repo_id and try to reverse map
	// it into a specific repo-instance that is locally
	// available (present in the closet).
	//
	// note that there can be many repo-instances in the
	// closet for any given repo_id.  (for example
	// when you do a local clone of a repo).  so we may get
	// more than one value here.

	SG_ERR_CHECK(  SG_closet__reverse_map_repo_id_to_descriptor_names(pCtx,
																	  pszHid_repo_id,
																	  pvec_use_submodules,
																	  &pvhListOfChoices)  );
	SG_ERR_CHECK(  SG_vhash__count(pCtx, pvhListOfChoices, &nrChoices)  );
	if (nrChoices == 0)
	{
		// the closet does not have any instances of this repo.
		// the user will need to CLONE one from a remote system
		// and only they know where they might find an instance
		// being served.
		//
		// put enough info into the missing-details record that
		// the user can figure out what to do.

		SG_context__err_reset(pCtx);
		SG_ERR_CHECK(  _veracity__missing_details__checkout_needs_repo_instance(pCtx,
																				pvaMissingDetails,
																				pPathSubmodule,
																				pszHid_repo_id)  );
	}
	else if (nrChoices > 1)
	{
		// they should have given us some help using --use-sub....

		SG_context__err_reset(pCtx);
		SG_ERR_CHECK(  _veracity__missing_details__checkout_multiple_choices(pCtx,
																			 pvaMissingDetails,
																			 pPathSubmodule,
																			 pszHid_repo_id,
																			 &pvhListOfChoices)  );
		pvhListOfChoices = NULL;	// the above call steals the list.
	}
	else // (nrChoices == 1)
	{
		// there was exactly 1 repo-instance for this repo_id.
		// *OR* the info in --use-sub was helpful.
		// 
		// we're going to *guess* that the WD for this submodule *should* be bound to
		// this named repo-instance.  most of the time this should be ok.

		const char * pszName_k;

		SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvhListOfChoices, 0, &pszName_k, NULL)  );
		SG_ERR_CHECK(  SG_strdup(pCtx, pszName_k, &pszRepoDescriptorName)  );
	}

	// return name when we identified one; return NULL when we had an issue.

	*ppszRepoDescriptorName = pszRepoDescriptorName;

fail:
	SG_VHASH_NULLFREE(pCtx, pvhListOfChoices);
}

static void _impl__veracity__precompute_checkout(SG_context * pCtx,
												 const SG_pathname * pPathSubmodule,
												 const SG_treenode_submodule * pTSM,
												 SG_bool bNoSubmodules,
												 const SG_vector * pvec_use_submodules,
												 SG_varray * pvaMissingDetails,
												 SG_vhash ** ppvhCheckoutInstanceData)
{
	char * pszRepoDescriptorName = NULL;
	const char * pszHid_cset_id;
	SG_uint32 count_before, count_after;

	SG_NULLARGCHECK_RETURN( pPathSubmodule );
	SG_NULLARGCHECK_RETURN( pTSM );
	SG_NULLARGCHECK_RETURN( pvaMissingDetails );
	SG_NULLARGCHECK_RETURN( ppvhCheckoutInstanceData );

	SG_ARGCHECK_RETURN(  (pTSM->pPathWD == NULL), pTSM  );		// target should not refer to a live WD.

	// reverse map repo_id in the TSM into a unique
	// local repo-descriptor-name, if possible.

	SG_ERR_CHECK(  SG_varray__count(pCtx, pvaMissingDetails, &count_before)  );
	SG_ERR_CHECK(  _veracity__precompute_checkout__try_to_reverse_map_repo_id(pCtx,
																			  pPathSubmodule,
																			  pTSM,
																			  pvec_use_submodules,
																			  pvaMissingDetails,
																			  &pszRepoDescriptorName)  );
	SG_ERR_CHECK(  SG_varray__count(pCtx, pvaMissingDetails, &count_after)  );
	SG_ASSERT(  ((count_after > count_before) == (pszRepoDescriptorName == NULL))  );

	if (count_after > count_before)
		goto done;		// don't throw, just return the missing details data

	// no issues identifying a unique repo-instance for this (repo_id).
	// try to open the selected repo-instance by name in the closet.

	SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pTSM->pvh, KEY_TECH__VERACITY__CSET_ID, &pszHid_cset_id )  );
	SG_ERR_CHECK(  SG_treenode_submodule_veracity__could_do_checkout(pCtx,
																	 pPathSubmodule,
																	 pszRepoDescriptorName,
																	 pszHid_cset_id,
																	 bNoSubmodules,
																	 pvec_use_submodules,
																	 pvaMissingDetails,
																	 ppvhCheckoutInstanceData)  );
	SG_ERR_CHECK(  SG_varray__count(pCtx, pvaMissingDetails, &count_after)  );
	SG_ASSERT(  ((count_after > count_before) == (*ppvhCheckoutInstanceData == NULL))  );

	if (count_after > count_before)
		goto done;		// don't throw, just return the missing details data

done:
	;

fail:
	SG_NULLFREE(pCtx, pszRepoDescriptorName);
}

static void _impl__veracity__apply_checkout(SG_context * pCtx,
											const SG_pathname * pPathSubmodule,
											const SG_vhash * pvhCheckoutInstanceData)
{
	// instead of calling SG_vv_verbs__checkout() we cheat and just do
	// the actual checkout because we've already done the precompute stuff.

	SG_ERR_CHECK(  SG_workingdir__create_and_checkout(pCtx,
													  pPathSubmodule,
													  pvhCheckoutInstanceData,
													  NULL)  );			// no branch -- create it untied

fail:
	return;
}

static void _impl__veracity__precompute_revert(SG_context * pCtx,
											   const SG_pathname * pPathSubmodule,
											   const SG_vector * pvec_use_submodules,
											   SG_bool bNoBackups,
											   SG_varray * pvaMissingDetails)
{
	SG_pendingtree * pPendingTreeAllocated = NULL;

	SG_NULLARGCHECK_RETURN( pPathSubmodule );
	// pvec_use_submodules is optional
	SG_NULLARGCHECK_RETURN( pvaMissingDetails );

	SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx,
										 pPathSubmodule,
										 &pPendingTreeAllocated)  );
	SG_ERR_CHECK(  SG_pendingtree__revert(pCtx,
										  pPendingTreeAllocated,
										  SG_PT_ACTION__DO_IT,		// a lie
										  0, NULL,
										  SG_TRUE,		// bRecursive
										  SG_FALSE,		// bNoSubmodules
										  bNoBackups,
										  pvec_use_submodules,
										  gpBlankFilespec,	// We *DO NOT* do any filtering across the boundary
										  SG_TRUE,		// bPrecomputeOnly
										  pvaMissingDetails)  );
	

fail:
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTreeAllocated);
}

static void _impl__veracity__apply_revert(SG_context * pCtx,
										  const SG_pathname * pPathSubmodule,
										  const SG_vector * pvec_use_submodules,
										  SG_bool bNoBackups)
{
	SG_pendingtree * pPendingTreeAllocated = NULL;
	SG_varray * pvaMissingDetails_XXXXXX = NULL;

	SG_NULLARGCHECK_RETURN( pPathSubmodule );
	// pvec_use_submodules is optional

	// I did not put a missing-details arg in the API to this
	// routine because the caller should have already determined
	// all of this with __precompute_revert, but I need to pass
	// something down because I didn't split SG_pendingtree__revert()
	// into 2 parts.  In theory, it should not return any missing
	// details.

	SG_ERR_CHECK(  SG_VARRAY__ALLOC(pCtx, &pvaMissingDetails_XXXXXX)  );

	SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx,
										 pPathSubmodule,
										 &pPendingTreeAllocated)  );
	SG_ERR_CHECK(  SG_pendingtree__revert(pCtx,
										  pPendingTreeAllocated,
										  SG_PT_ACTION__DO_IT,		// a lie
										  0, NULL,
										  SG_TRUE,		// bRecursive
										  SG_FALSE,		// bNoSubmodules
										  bNoBackups,
										  pvec_use_submodules,
										  gpBlankFilespec,	// We *DO NOT* do any filtering across the boundary
										  SG_FALSE,		// bPrecomputeOnly
										  pvaMissingDetails_XXXXXX)  );
#if defined(DEBUG)
	{
		SG_uint32 nrAfter = 0;

		SG_ERR_CHECK(  SG_varray__count(pCtx, pvaMissingDetails_XXXXXX, &nrAfter)  );
		if (nrAfter > 0)
		{
			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
									   "Nested revert had %d unexpected issues for submodule '%s'.\n",
									   nrAfter, SG_pathname__sz(pPathSubmodule))  );
			// TODO 2011/04/04 Should this throw and/or not be debug-only?
		}
	}
#endif

fail:
	SG_VARRAY_NULLFREE(pCtx, pvaMissingDetails_XXXXXX);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTreeAllocated);
}

//////////////////////////////////////////////////////////////////
// Must be last.
//////////////////////////////////////////////////////////////////

static _sg_treenode_submodule_impl _impl__veracity = {
	SG_TREENODE_SUBMODULE__TECH__VERACITY,
	_impl__veracity__sniff_and_allocate,
	_impl__veracity__validate,
	_impl__veracity__dtor,
	_impl__veracity__copy,
	_impl__veracity__comparable,
	_impl__veracity__is_dirty,
	_impl__veracity__describe_for_status,
	_impl__veracity__commit,
	_impl__veracity__precompute_update,
	_impl__veracity__apply_update,
	_impl__veracity__precompute_checkout,
	_impl__veracity__apply_checkout,
	_impl__veracity__precompute_revert,
	_impl__veracity__apply_revert,
};

//////////////////////////////////////////////////////////////////

END_EXTERN_C;

#endif//H_SG_TREENODE_SUBMODULE__PRIVATE_IMPL_VERACITY_H
