/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg_treenode_submodule__private_typedefs.h
 *
 * @details 
 *
 */

//////////////////////////////////////////////////////////////////

#ifndef H_SG_TREENODE_SUBMODULE__PRIVATE_TYPEDEFS_H
#define H_SG_TREENODE_SUBMODULE__PRIVATE_TYPEDEFS_H

BEGIN_EXTERN_C;

//////////////////////////////////////////////////////////////////

// We define a VERSION field for the vhash for safety, but we
// set it to the current TREENODE version number rather than
// definining another range of version numbers here.  The idea
// being that any changes we make in the layout of a submodule
// vhash (such as supporting new third-party DVCS's) is probably
// won't be done in isolation and we'll want to be able to
// detect it when we open the cset, rather than when we're in
// the middle of doing something and stumble across a newer
// link type.

#define KEY_VERSION						"ver"

// Eventually, we will allow different, third-party DVCS's to
// be referenced as a submodule in addition to a Veracity repo.

#define KEY_TECH						"tech"

//////////////////////////////////////////////////////////////////

typedef struct _sg_treenode_submodule_impl _sg_treenode_submodule_impl;

/**
 * Per-submodule-technology routines.  Think of this as a vtable
 * for a SG_treenode_submodule subclass.
 */
struct _sg_treenode_submodule_impl
{
	const char * pszTech;

	/**
	 * Sniff the given directory and see if it contains a potential
	 * submodule by looking for the various control files that a WD
	 * should have.
	 *
	 * Set bIsWD accordingly.
	 *
	 * Optionally, allocate and return a pTSM with the specifics.
	 * This routine should set pTSM->pPathWD and pTSM->pImpl and
	 * optionally pTSM->pVoidData.
	 */
	void (*pfn_sniff_and_allocate)(SG_context * pCtx,
								   const _sg_treenode_submodule_impl * pImpl,
								   const SG_pathname * pPath,
								   SG_repo * pRepoOuter,
								   SG_bool * pbIsWD,
								   SG_treenode_submodule ** ppTSM);

	/**
	 * Perform basic consistency checks on the given pTSM.
	 * This is mainly to verify that the blob we loaded from the
	 * repo and converted into a TSM was created by a equal or
	 * earlier version of this software.
	 */
	void (*pfn_validate)(SG_context * pCtx,
						const SG_treenode_submodule * pTSM);

	/**
	 * Free any impl-specific instance data stored in pTSM->pVoidData.
	 */
	void (*pfn_dtor)(SG_context * pCtx,
					 SG_treenode_submodule * pTSM);

	/**
	 * Copy any impl-specific instance data stored in pTSM-pVoidData.
	 */
	void (*pfn_copy)(SG_context * pCtx,
					 const SG_treenode_submodule * pTSM_src,
					 SG_treenode_submodule * pTSM_dest);

	/**
	 * Determine if the submodules described in the 2 given TSMs
	 * are comparable.  That is, do they refer to CSETs in the
	 * same repository, such that one could "update" a WD from one
	 * to the other.
	 *
	 * With the thought being that an "update" the submodule directory
	 * should keep the same GID and have status "modified" from one
	 * HID to the other HID.
	 *
	 * If the 2 TSMs are non-comparable, they should (probably) be
	 * treated like a "delete" and an "add" and have 2 different GIDs.
	 * 
	 * This will only be called with TSMs that refer to the same
	 * technology.
	 */
	void (*pfn_comparable)(SG_context * pCtx,
						   const SG_treenode_submodule * pTSM1,
						   const SG_treenode_submodule * pTSM2,
						   SG_bool * pbSimilar);

	/**
	 * See if there is dirt in the WD.  This may cache the results
	 * inside the TSM for a subsequent STATUS/DIFF operation, if
	 * that makes sense.  The purpose of this routine is just to
	 * see if there is dirt.
	 */
	void (*pfn_is_dirty)(SG_context * pCtx,
						 SG_treenode_submodule * pTSM,
						 SG_bool * pbIsDirty);

	/**
	 * Generate a brief string describing the state of the submodule.
	 * Perhaps just the CSID.
	 * 
	 * This should be suitable to appear in flavor text.
	 */
	void (*pfn_describe_for_status)(SG_context * pCtx,
									const SG_treenode_submodule * pTSM,
									char ** ppszDescription);

	/**
	 * Cause a full (non-partial) commit *within* the submodule.
	 * This is conceptually:
	 *     "cd $subdir; vv commit -m $msg" or
	 *     "cd $subdir; hg commit -m $msg" or
	 *     ....
	 *
	 * This interface *DOES NOT* allow/accept any of the filespec
	 * and other args used to do partial commits.  There are 2 reasons:
	 * [1] it is ambiguous when someone says 'vv commit --exclude *.txt'
	 *     whether that exclusion should be applied to only the outer
	 *     WD or to the outer and all nested submodules (however deep).
	 * [2] i don't want to require that third-party DVCS systems
	 *     support such notions (and if they do I don't want to translate
	 *     the syntax).
	 *
	 * SO, if an outer commit causes an automatic submodule commit, we
	 * require the sub to be a full commit.
	 * 
	 * If the sub commit fails (such as for unresolved issues),
	 * this throws (as it does for any other type of error).
	 *
	 * If the sub commit doesn't actually do anything, it should return
	 * false in pbDidSomething rather than throw.
	 *
	 * TODO 2010/12/02 Currently, we pass down the commit COMMENT that
	 * TODO            we were given.  The submodule is free to use it
	 * TODO            as is, or to decorate it with a "nested" message,
	 * TODO            or ignore it.
	 * TODO
	 * TODO            Think about whether we want to normalize how we
	 * TODO            wrap the message (if at all).
	 *
	 * TODO 2010/12/02 Do we want to pass down the DATE (in case the
	 * TODO            outer caller is trying to import a history and
	 * TODO            wanting to set a specific date on the changeset)?
	 *
	 * We DO NOT take a USER (also useful for the import) because the
	 * outer and sub may be in different admin domains.
	 */
	void (*pfn_commit)(SG_context * pCtx,
					   const SG_treenode_submodule * pTSM,
					   const char * pszComment,
					   SG_bool * pbDidSomething);

	/**
	 * Determine *if* the submodule in the live WD *could* be updated
	 * to the goal CSET.
	 *
	 * This may include:
	 * [a] verify that the goal CSET is present in the local copy of repo.
	 * [b] pull it from a remote server, if appropriate/necessary.
	 * [c] verify whatever peer-to-peer/ancestor-to-descendant restrictions
	 *     may be appropraite depending on whether the live WD is dirty.
	 *
	 * We DO NOT change the live WD (because the outer update may still
	 * abort the update if there are other update conflicts).
	 *
	 * This should return instance data from the precompute.  This should
	 * be given to the update-apply step (so that it can avoid re-scanning
	 * re-diffing everything).
	 *
	 * The underlying implementation may or may not PULL if necessary.
	 *
	 * If the implementation cannot do the update (because of missing
	 * information), it should append records to the missing details
	 * (and return a null instance data).
	 * This should have enough info to let the user pull/clone as
	 * necessary and retry.
	 *
	 * We try to only throw for hard errors.
	 */
	void (*pfn_precompute_update)(SG_context * pCtx,
								  const SG_pathname * pPathSubmodule,
								  const SG_treenode_submodule * pTSM_pend,
								  const SG_treenode_submodule * pTSM_goal,
								  const SG_vector * pvec_use_submodules,
								  SG_varray * pvaMissingDetails,
								  SG_treenode_submodule_instance_data_update ** ppTSIDU);

	/**
	 * Actually apply the nested-update using the cached info from the precompute step.
	 */
	void (*pfn_apply_update)(SG_context * pCtx,
							 const SG_pathname * pPathSubmodule,
							 const SG_treenode_submodule * pTSM_pend,
							 const SG_treenode_submodule * pTSM_goal,
							 SG_treenode_submodule_instance_data_update * pTSIDU);

	/**
	 * Determine *if* the submodule described in the given TSM *could*
	 * be checked out into the given directory.
	 *
	 * This IS NOT concerned with the validity of the given pathname.
	 * (It is mainly here so that we can make better error messages.)
	 *
	 * This IS concerned with whether we have a local repo instance
	 * available and whether it contains the desired changeset.
	 *
	 * If the implementation cannot do the checkout (because no
	 * repo-instance can be found or one cannot be uniquely identified,
	 * or the repo-instance does not contain the requested changeset,
	 * or recursively, something within the submodule has such a
	 * problem), it should append records to the missing details
	 * (and return a null instance data).
	 * This should have enough info to let the user clone as necessary
	 * and retry.
	 *
	 * If bNoSubmodules, we do not attempt to visit contained submodules.
	 * This is intended for EXPORT rather than CHECKOUT.  But I guess it
	 * could be used by both.
	 * 
	 * We try to only throw for hard errors.
	 */
	void (*pfn_precompute_checkout)(SG_context * pCtx,
									const SG_pathname * pPathSubmodule,
									const SG_treenode_submodule * pTSM,
									SG_bool bNoSubmodules,
									const SG_vector * pvec_use_submodules,
									SG_varray * pvaMissingDetails,
									SG_vhash ** ppvhCheckoutInstanceData);

	/**
	 * Actually apply the nested-checkout using the cached info from the precompute step.
	 */
	void (*pfn_apply_checkout)(SG_context * pCtx,
							   const SG_pathname * pPathSubmodule,
							   const SG_vhash * pvhCheckoutInstanceData);

	/**
	 * Can the submodule be reverted completely with the info given?
	 * Normally a revert is guaranteed, but if it needs to recreate a
	 * nested submodule within it (a nested CHECKOUT), can it find
	 * the necessary repos.
	 *
	 * If there not, append detail recors to pvaMissingDetails.
	 */
	void (*pfn_precompute_revert)(SG_context * pCtx,
								  const SG_pathname * pPathSubmodule,
								  const SG_vector * pvec_use_submodules,
								  SG_bool bNoBackups,
								  SG_varray * pvaMissingDetails);

	/**
	 *
	 */
	void (*pfn_apply_revert)(SG_context * pCtx,
							 const SG_pathname * pPathSubmodule,
							 const SG_vector * pvec_use_submodules,
							 SG_bool bNoBackups);
	
};

struct SG_treenode_submodule
{
	// Only the pvh gets written to the repo.  The contents of this
	// may vary depending upon the technology used in the submodule.

	SG_vhash *								pvh;

	// The HID of the pvh as usual.

	char *									pszHidFrozen;

	// Binding to routines for the specific submodule technology.

	const _sg_treenode_submodule_impl *		pImpl;

	// The following fields are only set when this TSM was populated
	// from an existing WD.  There are in-memory values only so that
	// we might later ask, for example, if there are uncommitted
	// changes within the submodule.
	//
	// These fields are NOT set when we load a TSM from a blob in the repo.

	SG_pathname *							pPathWD;
	void *									pVoidData;
};

static void _sg_treenode_submodule__find_impl(SG_context * pCtx,
											  const char * pszTech,
											  const _sg_treenode_submodule_impl ** ppImpl);

//////////////////////////////////////////////////////////////////

END_EXTERN_C;

#endif//H_SG_TREENODE_SUBMODULE__PRIVATE_TYPEDEFS_H
