/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg_vv_utils__lookup_wd_info.h
 *
 * @details Routine to take a directory path anywhere within a working directory
 * and return official info for the root of the working directory.  That is,
 * find the directory associated with "@/" from any point within it.  For the
 * root directory, it returns itself.
 *
 */

//////////////////////////////////////////////////////////////////

#ifndef H_SG_VV_UTILS__LOOKUP_WD_INFO_H
#define H_SG_VV_UTILS__LOOKUP_WD_INFO_H

BEGIN_EXTERN_C;

//////////////////////////////////////////////////////////////////

void SG_vv_utils__lookup_wd_info(SG_context * pCtx,
								 const SG_pathname * pPathAnywhereWithinWD,
								 SG_bool * pbInWD, SG_bool * pbUnderVC,
								 SG_pathname ** ppPathRoot,
								 SG_string ** ppStringRepoName,
								 char ** ppszGidEntry)
{
	SG_pathname * pPathRoot = NULL;
	SG_string * pStringRepoName = NULL;
	SG_pendingtree * pPendingTree = NULL;
	char * pszGidEntry = NULL;

	// TODO 2010/10/25 Figure out what ANCHOR is and if we need to do anything with it.

	SG_NULLARGCHECK_RETURN( pPathAnywhereWithinWD );
	SG_NULLARGCHECK_RETURN( pbInWD );
	SG_NULLARGCHECK_RETURN( pbUnderVC );

	// TODO 2010/10/25 We really should fix SG_workingdir__find_mapping to throw _NOT_A_WORKING_COPY
	// TODO            rather than just _NOT_FOUND.

	SG_workingdir__find_mapping(pCtx, pPathAnywhereWithinWD, &pPathRoot, &pStringRepoName, NULL);
	if (SG_context__has_err(pCtx))
	{
		if (!SG_context__err_equals(pCtx, SG_ERR_NOT_A_WORKING_COPY))
			SG_ERR_RETHROW;

		SG_context__err_reset(pCtx);

		*pbInWD = SG_FALSE;
		*pbUnderVC = SG_FALSE;
	}
	else
	{
		// the given pathname is somewhere within a Working Directory.

		*pbInWD = SG_TRUE;

		// see if the entry is under version control (rather than
		// being a found or ignored item).

		SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pPathRoot, &pPendingTree)  );
		SG_pendingtree__get_gid_from_local_path(pCtx, pPendingTree, pPathAnywhereWithinWD, &pszGidEntry);
		if (SG_context__has_err(pCtx))
		{
			if (!SG_context__err_equals(pCtx, SG_ERR_ITEM_NOT_UNDER_VERSION_CONTROL))
				SG_ERR_RETHROW;
			
			SG_context__err_reset(pCtx);

			*pbUnderVC = SG_FALSE;
		}
		else
		{
			*pbUnderVC = SG_TRUE;
		}
	}

	if (ppPathRoot)
	{
		*ppPathRoot = pPathRoot;
		pPathRoot = NULL;
	}
	
	if (ppStringRepoName)
	{
		*ppStringRepoName = pStringRepoName;
		pStringRepoName = NULL;
	}

	if (ppszGidEntry)
	{
		*ppszGidEntry = pszGidEntry;
		pszGidEntry = NULL;
	}

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathRoot);
	SG_STRING_NULLFREE(pCtx, pStringRepoName);
	SG_NULLFREE(pCtx, pszGidEntry);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
}

void SG_vv_utils__throw_NOTAREPOSITORY_for_wd_association(SG_context * pCtx,
                                                          const char * szRepoName,
                                                          SG_pathname * pWdRoot)
{
	SG_ASSERT(pCtx!=NULL);
	SG_ASSERT(szRepoName!=NULL);
	if(pWdRoot)
	{
		SG_ERR_THROW2_RETURN(SG_ERR_NOTAREPOSITORY, (pCtx,
			SG_CR_STR // Special trick to hide the "Not a repository:" prefix.
			"The path '%s' is a working copy for the '%s' repository, which does not exist, or has been deleted and recreated.\n"
			"\n"
			"This error can happen for various reasons, but most likely the repository has been renamed or deleted.\n"
			"\n"
			"To remove the working copy's association with this repository, remove the \".sgdrawer\" folder from the root of the working copy.",
			SG_pathname__sz(pWdRoot), szRepoName
		));
	}
	else
	{
		SG_ERR_THROW2_RETURN(SG_ERR_NOTAREPOSITORY, (pCtx,
			SG_CR_STR // Special trick to hide the "Not a repository:" prefix.
			"This path is in a working copy for the '%s' repository, which does not exist, or has been deleted and recreated.\n"
			"\n"
			"This error can happen for various reasons, but most likely the repository has been renamed or deleted.",
			szRepoName
		));
	}
}

//////////////////////////////////////////////////////////////////

END_EXTERN_C;

#endif//H_SG_VV_UTILS__LOOKUP_WD_INFO_H
