/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg_vv_utils__rev_arg_to_csid.h
 *
 * @details Routine to convert various cset-prefix/rev-nr/tag/whatever
 * into a full CSet HID.
 *
 */

//////////////////////////////////////////////////////////////////

#ifndef H_SG_VV_UTILS__REV_ARG_TO_CSID__NO_WD_H
#define H_SG_VV_UTILS__REV_ARG_TO_CSID__NO_WD_H

BEGIN_EXTERN_C;

//////////////////////////////////////////////////////////////////

/**
 * if given, convert a rev-cset-prefix, a rev-number, or a tag into full
 * and unique CSET HID.  otherwise get the HID of the unique HEAD of the
 * GIVEN BRANCH or the DEFAULT BRANCH or the single unique HEAD.
 *
 * Since we DO NOT assume a working directory, we DO NOT know about any
 * branch ties; that is, only a WD can be "tied" to a branch.
 *
 * TODO 2010/10/22 Eventually we'd like to be able to get rid of
 * TODO            the bRev that we use in all of the vv commands
 * TODO            to determine the rev-vs-tag.
 * TODO
 * TODO            I'd also like to be able to use terms like 'head'
 * TODO            or 'tip-1'.
 *
 * Return both the full HID of the desired CSET *and* (optionally)
 * the name of the branch that it is in (if it is in one).
 */
void SG_vv_utils__rev_spec_to_csid__no_wd(
	SG_context * pCtx,
	const char * pszRepoName,
	SG_rev_spec* pRevSpec,
	char ** ppszHidCSet,
	char ** ppsz_branch_name)
{
	SG_uint32 countRevSpec = 0;
	SG_repo * pRepo = NULL;
	SG_rbtree * prbLeaves = NULL;
	char * pszHidAllocated = NULL;
	char * psz_branch_name = NULL;
    SG_vhash* pvh_pile = NULL;

	SG_NONEMPTYCHECK_RETURN( pszRepoName );
	SG_NULLARGCHECK_RETURN( ppszHidCSet );

	if (pRevSpec)
		SG_ERR_CHECK(  SG_rev_spec__count(pCtx, pRevSpec, &countRevSpec)  );
	if (countRevSpec != 0 && countRevSpec != 1)
		SG_ERR_THROW2_RETURN(  SG_ERR_INVALIDARG,
			(pCtx, "Zero or one revision specifier must be provided")  );
	
	SG_ERR_CHECK(  SG_REPO__OPEN_REPO_INSTANCE(pCtx, pszRepoName, &pRepo)  );

	if (countRevSpec != 0)
		SG_ERR_CHECK(  SG_rev_spec__get_one__repo(pCtx, pRepo, pRevSpec, &pszHidAllocated, &psz_branch_name)  );
	else
	{
		/* There are no revision specifiers. Use the tip of the master branch if it's unambiguous. */
		SG_bool b_exists = SG_FALSE;

		SG_ERR_CHECK(  SG_vc_branches__get_unambiguous(pCtx, pRepo, SG_VC_BRANCHES__DEFAULT, &b_exists, &pszHidAllocated)  );

		if (b_exists && pszHidAllocated)
		{
			SG_ERR_CHECK(  SG_STRDUP(pCtx, SG_VC_BRANCHES__DEFAULT, &psz_branch_name)  );
		}

        if (!pszHidAllocated)
        {
            const char * pszHid_0;
            SG_uint32 count_leaves;
            SG_bool b;

            SG_ERR_CHECK(  SG_repo__fetch_dag_leaves(pCtx, pRepo, SG_DAGNUM__VERSION_CONTROL, &prbLeaves)  );
            SG_ERR_CHECK(  SG_rbtree__count(pCtx, prbLeaves, &count_leaves)  );
            SG_ASSERT(  (count_leaves > 0)  );
            if (count_leaves != 1)
                SG_ERR_THROW(  SG_ERR_MULTIPLE_HEADS_FROM_DAGNODE  );

            SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, NULL, prbLeaves, &b, &pszHid_0, NULL)  );
            SG_ERR_CHECK(  SG_STRDUP(pCtx, pszHid_0, &pszHidAllocated)  );
        }
	}
	
	*ppszHidCSet = pszHidAllocated;
    pszHidAllocated = NULL;

    if (ppsz_branch_name)
    {
        *ppsz_branch_name = psz_branch_name;
        psz_branch_name = NULL;
    }

fail:
	SG_REPO_NULLFREE(pCtx, pRepo);
	SG_RBTREE_NULLFREE(pCtx, prbLeaves);
    SG_VHASH_NULLFREE(pCtx, pvh_pile);
	SG_NULLFREE(pCtx, pszHidAllocated);
	SG_NULLFREE(pCtx, psz_branch_name);
}


//////////////////////////////////////////////////////////////////

END_EXTERN_C;

#endif//H_SG_VV_UTILS__REV_ARG_TO_CSID__NO_WD_H


