/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg_vv_utils__rev_arg_to_csid__wd.h
 *
 * @details 
 *
 */

//////////////////////////////////////////////////////////////////

#ifndef H_SG_VV_UTILS__REV_ARG_TO_CSID__WD_H
#define H_SG_VV_UTILS__REV_ARG_TO_CSID__WD_H

BEGIN_EXTERN_C;

//////////////////////////////////////////////////////////////////

/**
 * if given, convert a rev-set-prefix, a rev-number, or a tag into full
 * and unique CSET HID.  otherwise get the HID of the unique HEAD of the
 * GIVEN BRANCH or the HEAD of the CURRENTLY TIED BRANCH or the single
 * unique DESCENDANT HEAD from the current baseline.
 *
 * TODO 2010/10/22 Eventually we'd like to be able to get rid of
 * TODO            the bRev that we use in all of the vv commands
 * TODO            to determine the rev-vs-tag.
 * TODO
 * TODO            I'd also like to be able to use terms like 'head'
 * TODO            or 'tip-1'.
 *
 * Return both the full HID of the desired CSET *and* (optionally)
 * the name of the branch that it is in (if it is in one).
 *
 * This was written for UPDATE.  Not sure if it is too specific or
 * not for general use.
 */
void SG_vv_utils__rev_spec_to_csid__wd(SG_context * pCtx,
									   const char * pszFolder,
									   SG_rev_spec* pRevSpec,
									   char ** ppszHidCSet,
									   char ** ppsz_branch_name)
{
	SG_pathname * pPath = NULL;
	SG_pendingtree * pPendingTree = NULL;
	SG_repo * pRepo = NULL;		// we do not own this

	SG_uint32 countRevSpecs = 0;
	char * pszTargetChangeset = NULL;
	char * pszTargetBranchName = NULL;

	char * psz_tied_branch_name = NULL;
    char* psz_only_tied_branch_value = NULL;

	// pszFolder is optional
	// pRevSpec is optional
	SG_NULLARGCHECK_RETURN( ppszHidCSet );
	// ppsz_branch_name is optional

	if (pRevSpec)
		SG_ERR_CHECK(  SG_rev_spec__count(pCtx, pRevSpec, &countRevSpecs)  );

	if (countRevSpecs > 1)
		SG_ERR_THROW2_RETURN(  SG_ERR_INVALIDARG,
							   (pCtx, "Only one revision can be specified.")  );

	// use the absoute or relative "folder" name to find the
	// containing working directory (default to CWD if null).
	// then open the associated pendingtree.

	SG_ERR_CHECK(  SG_vv_utils__folder_arg_to_path(pCtx, pszFolder, &pPath)  );
	SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pPath, &pPendingTree)  );
	SG_ERR_CHECK(  SG_pendingtree__get_repo(pCtx, pPendingTree, &pRepo)  );

	if (countRevSpecs)
	{
		SG_ERR_CHECK(  SG_rev_spec__get_one__repo(pCtx, pRepo, pRevSpec, &pszTargetChangeset, &pszTargetBranchName)  );
	}
	else
	{
		const SG_pathname * pPathTop = NULL;		// we do not own this

		SG_ERR_CHECK(  SG_pendingtree__get_working_directory_top__ref(pCtx, pPendingTree, &pPathTop)  );
		SG_ERR_CHECK(  SG_workingdir__get_branch_info(pCtx, pPathTop, &psz_tied_branch_name)  );

		if (psz_tied_branch_name)
		{
			// stay tied to the same branch, update to its head cset (which we haven't lookup yet).
			// update to the most recent version of this branch

			SG_bool b_tied_branch_exists = SG_FALSE;

			SG_ERR_CHECK(  SG_vc_branches__get_unambiguous(pCtx, pRepo, psz_tied_branch_name, &b_tied_branch_exists, &psz_only_tied_branch_value)  );

			if (b_tied_branch_exists)
			{
				pszTargetChangeset = psz_only_tied_branch_value;
				psz_only_tied_branch_value = NULL;
				pszTargetBranchName = psz_tied_branch_name;
				psz_tied_branch_name = NULL;
			}
			else
			{
				// we are tied to a branch that doesn't exist yet
			}
		}
		else
		{
			// we are untied and no branch was given.  see if there is
			// a single unique descendant head from the current baseline.
			// i'm going to leave that code in SG_pendingtree__update_baseline()
			// because it also wants to know dagquery relationship info between
			// the current baseline and the target cset.

			// pass NULL for target changeset and let the UPDATE code find the proper head/tip.

			pszTargetChangeset = NULL;
			pszTargetBranchName = NULL;
		}
	}

	if (ppsz_branch_name)
	{
		if (pszTargetBranchName && *pszTargetBranchName)
			SG_ERR_CHECK(  SG_STRDUP(pCtx, pszTargetBranchName, ppsz_branch_name)  );
		else
			*ppsz_branch_name = NULL;
	}

	*ppszHidCSet = pszTargetChangeset;
	pszTargetChangeset = NULL;

fail:
	SG_NULLFREE(pCtx, pszTargetChangeset);
	SG_NULLFREE(pCtx, pszTargetBranchName);
	SG_NULLFREE(pCtx, psz_tied_branch_name);
	SG_NULLFREE(pCtx, psz_only_tied_branch_value);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
}

//////////////////////////////////////////////////////////////////

END_EXTERN_C;

#endif//H_SG_VV_UTILS__REV_ARG_TO_CSID__WD_H
