/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg_vv_utils__rev_arg_to_csid_for_merge.h
 *
 * @details Convert any given --rev, --tag, --branch, or no arg
 * into the HID of the CSET we should merge with.
 *
 */

//////////////////////////////////////////////////////////////////

#ifndef H_SG_VV_UTILS__REV_ARG_TO_CSID_FOR_MERGE_H
#define H_SG_VV_UTILS__REV_ARG_TO_CSID_FOR_MERGE_H

BEGIN_EXTERN_C;

//////////////////////////////////////////////////////////////////

void SG_vv_utils__rev_spec_to_csid_for_merge(SG_context * pCtx,
											 const char * pszFolder,
											 SG_rev_spec * pRevSpec,
											 char ** ppszHidCSet)
{
	SG_pathname * pPath = NULL;
	SG_pendingtree * pPendingTree = NULL;
	SG_repo * pRepo = NULL;		// we do not own this
	SG_vhash* pvh_pile = NULL;
	const SG_varray * pva_wd_parents;
	char * pszTargetChangeset = NULL;
	char * psz_tied_branch_name = NULL;
	const char * pszHidBaseline;
	SG_bool bIsBaseline;
	SG_bool bGivenRevSpec = SG_FALSE;

	// pszFolder is optional
	// pszWhatever is optional
	SG_NULLARGCHECK_RETURN( ppszHidCSet );

	// use the absoute or relative "folder" name to find the
	// containing working directory (default to CWD if null).
	// then open the associated pendingtree.

	SG_ERR_CHECK(  SG_vv_utils__folder_arg_to_path(pCtx, pszFolder, &pPath)  );
	SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pPath, &pPendingTree)  );
	SG_ERR_CHECK(  SG_pendingtree__get_repo(pCtx, pPendingTree, &pRepo)  );

	SG_ERR_CHECK(  SG_pendingtree__get_wd_parents__ref(pCtx, pPendingTree, &pva_wd_parents)  );
	SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pva_wd_parents, 0, &pszHidBaseline)  );

	if (pRevSpec)
	{
		SG_uint32 count;
		SG_ERR_CHECK(  SG_rev_spec__count(pCtx, pRevSpec, &count)  );
		if (count > 1)
			SG_ERR_THROW2(SG_ERR_INVALIDARG, (pCtx, "Merge can accept (up to) one revision specifier."));
		bGivenRevSpec = count;
	}

	if (bGivenRevSpec)
	{
		SG_ERR_CHECK(  SG_rev_spec__get_one__repo(pCtx, pRepo, pRevSpec, &pszTargetChangeset, NULL)  );

		// make sure the named cset is not the baseline.

		bIsBaseline = (strcmp(pszTargetChangeset, pszHidBaseline) == 0);
		if (bIsBaseline)
			SG_ERR_THROW(SG_ERR_MERGE_BASELINE_IS_IMPLIED);
	}
	else
	{
		const SG_pathname * pPathTop = NULL;		// we do not own this

		SG_ERR_CHECK(  SG_pendingtree__get_working_directory_top__ref(pCtx, pPendingTree, &pPathTop)  );
		SG_ERR_CHECK(  SG_workingdir__get_branch_info(pCtx, pPathTop, &psz_tied_branch_name)  );
        if (psz_tied_branch_name)
        {
            SG_vhash* pvh_branches = NULL;
            SG_bool b_exists = SG_FALSE;
            SG_vhash* pvh_branch_info = NULL;
            SG_vhash* pvh_branch_values = NULL;
            SG_uint32 count_branch_values = 0;

            SG_ERR_CHECK(  SG_vc_branches__cleanup(pCtx, pRepo, &pvh_pile)  );
            SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvh_pile, "branches", &pvh_branches)  );
            SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh_branches, psz_tied_branch_name, &b_exists)  );
            if (!b_exists)
            {
                SG_ERR_THROW(  SG_ERR_BRANCH_DOES_NOT_NEED_MERGE  );
            }

            SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvh_branches, psz_tied_branch_name, &pvh_branch_info)  );

            SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvh_branch_info, "values", &pvh_branch_values)  );
            SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_branch_values, &count_branch_values)  );

            SG_ASSERT(count_branch_values > 0);
            if (2 == count_branch_values)
            {
                const char* psz_val_0 = NULL;
                const char* psz_val_1 = NULL;

                SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvh_branch_values, 0, &psz_val_0, NULL)  );
                SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvh_branch_values, 1, &psz_val_1, NULL)  );

                if (0 == strcmp(pszHidBaseline, psz_val_0))
                {
					SG_ERR_CHECK(  SG_STRDUP(pCtx, psz_val_1, &pszTargetChangeset)  );
                }
                else if (0 == strcmp(pszHidBaseline, psz_val_1))
                {
					SG_ERR_CHECK(  SG_STRDUP(pCtx, psz_val_0, &pszTargetChangeset)  );
                }
                else
                {
                    SG_ERR_THROW(  SG_ERR_TIED_BRANCH_NOT_BASELINE  );
                }
            }
            else if (1 == count_branch_values)
            {
                SG_ERR_THROW(  SG_ERR_BRANCH_DOES_NOT_NEED_MERGE  );
            }
            else
            {
                SG_ERR_THROW(  SG_ERR_TOO_MANY_HEADS  );
            }
        }
        else
        {
            SG_ERR_THROW(  SG_ERR_NOT_TIED  );
        }
	}

	*ppszHidCSet = pszTargetChangeset;
	pszTargetChangeset = NULL;

fail:
	SG_NULLFREE(pCtx, pszTargetChangeset);
	SG_NULLFREE(pCtx, psz_tied_branch_name);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_VHASH_NULLFREE(pCtx, pvh_pile);
}

//////////////////////////////////////////////////////////////////

END_EXTERN_C;

#endif//H_SG_VV_UTILS__REV_ARG_TO_CSID_FOR_MERGE_H
