/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg_vv_verbs__checkout.h
 *
 * @details Routines to perform most of 'vv checkout'.
 *
 */

//////////////////////////////////////////////////////////////////

#ifndef H_SG_VV_VERBS__CHECKOUT_H
#define H_SG_VV_VERBS__CHECKOUT_H

BEGIN_EXTERN_C;

//////////////////////////////////////////////////////////////////

static void _vv_verbs__checkout(SG_context * pCtx,
								const SG_pathname * pPath,
								const char * pszRepoName,
								const char * pszHidCSet,
								const char * psz_branch_name,
								const SG_vector * pvec_use_submodules,
								SG_vhash ** ppvhResult)
{
	SG_repo * pRepo = NULL;
	SG_changeset * pcs = NULL;
	SG_vhash * pvhResult = NULL;
	SG_varray * pvaMissingDetails = NULL;
	SG_vhash * pvhCheckoutInstanceData = NULL;
	SG_uint32 count_after;

	// TODO 2011/02/02 We DO NOT allow partial checkouts where we skip
	// TODO            the contained submodules.  That may be fine for
	// TODO            EXPORT, but for CHECKOUT it wouldn't give a complete
	// TODO            view of the project.
	// TODO
	// TODO            Do we want to soften this?
	SG_bool bNoSubmodules = SG_FALSE;

	// for the top-level checkout, we throw when either the repo-name or cset-id is not valid.
	// this gives the expected error message in response to something the user typed directly.
	// TODO 2011/01/24 Think about letting these return status and details in pvhResult instead.

	SG_ERR_CHECK(  SG_REPO__OPEN_REPO_INSTANCE(pCtx, pszRepoName, &pRepo)  );
	SG_ERR_CHECK(  SG_changeset__load_from_repo(pCtx, pRepo, pszHidCSet, &pcs)  );
    // TODO why do we load this changeset and then immediately free it?
	SG_CHANGESET_NULLFREE(pCtx, pcs);
	SG_REPO_NULLFREE(pCtx, pRepo);

	// before we attempt to do the CHECKOUT, see if there are any submodules
	// within and verify that we have a (repo_instance, cset_id) for each of them.
	// this will either return an array of missing-details *or* instance data for
	// this (repo_instance, cset_id).

	SG_ERR_CHECK(  SG_VARRAY__ALLOC(pCtx, &pvaMissingDetails)  );
	SG_ERR_CHECK(  SG_treenode_submodule_veracity__could_do_checkout(pCtx, pPath,
																	 pszRepoName, pszHidCSet,
																	 bNoSubmodules,
																	 pvec_use_submodules,
																	 pvaMissingDetails,
																	 &pvhCheckoutInstanceData)  );
	SG_ERR_CHECK(  SG_varray__count(pCtx, pvaMissingDetails, &count_after)  );

	if (count_after == 0)
	{
		// everything *should* work, so go ahead and populate the WD.

#if 0 && defined(DEBUG)
		SG_ERR_IGNORE(  SG_vhash_debug__dump_to_console__named(pCtx,
															   pvhCheckoutInstanceData,
															   "vv_verbs__checkout: CheckoutInstanceData")  );
#endif

		SG_ERR_CHECK(  SG_workingdir__create_and_checkout(pCtx,
														  pPath,
														  pvhCheckoutInstanceData,
														  psz_branch_name)  );
		if (ppvhResult)
		{
			SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvhResult)  );
			SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvhResult, "status", SG_VV_VERBS__STATUS__OK)  );
		}
	}
	else /* (count_after > 0) */
	{
		// there is something in the tree that we *will not* be able to populate.

		if (ppvhResult)
		{
			// the caller wants the details, so we don't throw.
			// rather we put the details and status in the vhash.

			SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvhResult)  );

			SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvhResult, "status", SG_VV_VERBS__STATUS__SUBMODULE_PROBLEM)  );
			SG_ERR_CHECK(  SG_vhash__add__varray(pCtx, pvhResult, "sub", &pvaMissingDetails)  );	// this steals our varray
			SG_ASSERT(  (pvaMissingDetails == NULL)  );
		}
		else
		{
			SG_ERR_THROW(  SG_ERR_SUBMODULE_NESTED_UPDATE_CHECKOUT_PROBLEM  );
		}
	}

	if (ppvhResult)
	{
		*ppvhResult = pvhResult;
		pvhResult = NULL;
	}

fail:
	SG_VHASH_NULLFREE(pCtx, pvhResult);
	SG_VARRAY_NULLFREE(pCtx, pvaMissingDetails);
	SG_CHANGESET_NULLFREE(pCtx, pcs);
	SG_REPO_NULLFREE(pCtx, pRepo);
	SG_VHASH_NULLFREE(pCtx, pvhCheckoutInstanceData);
}

void SG_vv_verbs__checkout(SG_context * pCtx,
						   const char * pszRepoName,
						   const char * pszHidCSet,		//< [in] Full HID of CSET (caller should use SG_vv_utils__lookup_cset()).
						   const char * pszFolder,		//< [in] arbitrary path (abs or rel); defaults to CWD if blank or null.
						   SG_bool bRequireNewDirOrEmpty,
						   SG_bool bSubmodule,
                           const char* psz_branch_name,
						   const SG_vector * pvec_use_submodules,
						   SG_vhash ** ppvhResult)
{
	SG_pathname * pPathAllocated = NULL;
	SG_pathname * pPathOuterWD = NULL;
	SG_string * pStringOuterRepoName = NULL;
	char * pszGidEntry = NULL;
	SG_bool bAlreadyInWD;
	SG_bool bUnderVC;

	SG_NONEMPTYCHECK_RETURN( pszRepoName );
	SG_NONEMPTYCHECK_RETURN( pszHidCSet );
	// pszFolder is optional and defaults to CWD.

	SG_ERR_CHECK(  SG_vv_utils__folder_arg_to_path(pCtx, pszFolder, &pPathAllocated)  );
	
	// see if the desired directory is already within an
	// outer WD and possibly already under version control.

	SG_ERR_CHECK(  SG_vv_utils__lookup_wd_info(pCtx,
											   pPathAllocated,
											   &bAlreadyInWD,
											   &bUnderVC,
											   &pPathOuterWD,
											   &pStringOuterRepoName,
											   &pszGidEntry)  );
#if 0 && defined(DEBUG)
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   ("vv_verbs__checkout: wd info\n"
								"           [path %s]\n"
								"           [in wd %d]\n"
								"           [under vc %d]\n"
								"           [wd root %s]\n"
								"           [repo name %s]\n"
								"           [gid entry %s]\n"),
							   ((pPathAllocated) ? SG_pathname__sz(pPathAllocated) : "(null)"),
							   bAlreadyInWD,
							   bUnderVC,
							   ((pPathOuterWD) ? SG_pathname__sz(pPathOuterWD) : "(null)"),
							   ((pStringOuterRepoName) ? SG_string__sz(pStringOuterRepoName) : "(null)"),
							   pszGidEntry)  );
#endif

	if (bUnderVC)
	{
		// the pathname is already under version control in
		// an outer WD.  if we were to create a WD/submodule
		// here, we would be stealing the directory from the
		// outer WD.  the outer would see this (and everything
		// within it) as LOST.
		//
		// we THROW here rather than creating a row in pvhResult.
		// i'm not sure it matters one way or another, but this
		// feels more like a hard error than something that the
		// user could fix by cloning/pulling.  (This will cause
		// a hard error in JS.)

		SG_ERR_THROW2(  SG_ERR_ENTRY_ALREADY_UNDER_VERSION_CONTROL,
						(pCtx, "The path '%s' is already under version control in the working copy for repository '%s' rooted at '%s'.",
						 SG_pathname__sz(pPathAllocated),
						 SG_string__sz(pStringOuterRepoName),
						 SG_pathname__sz(pPathOuterWD))  );
	}

	if (bRequireNewDirOrEmpty)
	{
		// when checking-out an arbitrary version we require either a new or
		// empty target directory so that we are not trashing something.

		SG_ERR_CHECK(  SG_vv_utils__verify_dir_empty(pCtx, pPathAllocated)  );
	}
	else
	{
		// when initializing a new WD we allow there to be stuff in the
		// directory (stuff that a subsequent addremove would deal with).
		// but we do require that if the path exists on disk that it be
		// a directory.

		SG_fsobj_type fsType = SG_FSOBJ_TYPE__UNSPECIFIED;
		SG_bool bExists = SG_FALSE;

		SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pPathAllocated, &bExists, &fsType, NULL)  );
		if (bExists && (fsType != SG_FSOBJ_TYPE__DIRECTORY))
		{
			SG_ERR_THROW2(  SG_ERR_NOT_A_DIRECTORY,
							(pCtx, "Cannot create a WD here; the path '%s' already exists and is not a directory.",
							 SG_pathname__sz(pPathAllocated))  );
		}
	}

	if (bAlreadyInWD)
	{
		if (bSubmodule)
		{
			SG_ERR_CHECK(  _vv_verbs__checkout(pCtx, pPathAllocated, pszRepoName, pszHidCSet, psz_branch_name, pvec_use_submodules, ppvhResult)  );

			// TODO 2010/11/08 Should be auto-add this submodule directory to the containing WD?
		}
		else
		{
			// The desired WD is already within an outer WD and they
			// didn't indicate that they were creating a submodule.

			SG_ERR_THROW2(  SG_ERR_WORKING_DIRECTORY_IN_USE,
							(pCtx, "The directory '%s' is already part of the working copy for repository '%s' rooted at '%s'; use 'submodule' option.",
							 SG_pathname__sz(pPathAllocated),
							 SG_string__sz(pStringOuterRepoName),
							 SG_pathname__sz(pPathOuterWD))  );
		}
	}
	else
	{
		if (bSubmodule)
		{
			// They indicated that they wanted to create this WD as
			// a submodule of an existing/outer WD.

			SG_ERR_THROW2(  SG_ERR_NOT_A_WORKING_COPY,
							(pCtx, "Option 'submodule' invalid; the directory '%s' is not within an existing working copy.",
							 SG_pathname__sz(pPathAllocated))  );
		}
		else
		{
			SG_ERR_CHECK(  _vv_verbs__checkout(pCtx, pPathAllocated, pszRepoName, pszHidCSet, psz_branch_name, pvec_use_submodules, ppvhResult)  );
		}
	}

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathAllocated);
	SG_PATHNAME_NULLFREE(pCtx, pPathOuterWD);
	SG_STRING_NULLFREE(pCtx, pStringOuterRepoName);
	SG_NULLFREE(pCtx, pszGidEntry);
}

//////////////////////////////////////////////////////////////////

END_EXTERN_C;

#endif//H_SG_VV_VERBS__CHECKOUT_H
