/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg_vv_verbs__commit.h
 *
 * @details Routines to perform most of 'vv commit'.
 *
 */

//////////////////////////////////////////////////////////////////

#ifndef H_SG_VV_VERBS__COMMIT_H
#define H_SG_VV_VERBS__COMMIT_H

BEGIN_EXTERN_C;

//////////////////////////////////////////////////////////////////


void SG_vv_verbs__commit(SG_context * pCtx,
						 const char * pszFolder,			//<[in] arbitrary path (abs or rel) to a WD or a location within a WD; defaults to CWD if blank or null.
						 const char * pszComment,			//<[in] the checkin comment (optional); can be added later using 'vv comment'.
						 const SG_file_spec * pFilespec,	//<[in] the set of --includes/--excludes/etc.
						 SG_bool bRecursive,
						 SG_bool bNoSubmodules,				//<[in] do not do nested commits for dirty submodules.
						 const char * pszUser,				//<[in] credit this user (useful for import mainly); defaults to the current user.
						 SG_int64 i64Date,					//<[in] date to assign to checkin (useful for import mainly); see SG_AUDIT__WHEN__NOW.
						 SG_uint32 count_args, const char ** paszArgs,	//<[in] (optional) list of files_or_folders (relative to pszFolder).
						 SG_uint32 count_assocs, const char * const * ppszAssocs,
						 const SG_stringarray * psaStamps,
						 void (*promptfn)(SG_context *, SG_string *,char **, const SG_varray *),
						 SG_string *boilerplate,
						 char ** ppsz_hid_cset)				//<[out] (optional) the HID of the newly created CSET.
{
	SG_pathname * pPathGiven = NULL;
	SG_pendingtree * pPendingTree = NULL;
    SG_dagnode * pdn = NULL;
    SG_repo * pRepo = NULL;				// we don't own this
	SG_treediff2 * pTreeDiff = NULL;
	SG_treediff2_stats tdStats;
	SG_audit q;
	char *localComment = NULL;
	char *trimmedComment = NULL;
	SG_varray *descs = NULL;

	// convert the given folder into an absolute directory.
	// (this may be a WD root or any subdirectory within it.)
	// default to CWD if not given.

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC(pCtx, &pPathGiven)  );
	if (pszFolder && *pszFolder)
		SG_ERR_CHECK(  SG_pathname__set__from_sz(pCtx, pPathGiven, pszFolder)  );
	else
		SG_ERR_CHECK(  SG_pathname__set__from_cwd(pCtx, pPathGiven)  );
	
    // Do a COMPLETE scan first.  This ensures that the pendingtree knows
	// about everything that is dirty in the working directory WITHOUT REGARD
	// for what they actually want to commit.  This scans everything and
	// writes it out to wd.json.

	SG_ERR_CHECK(  SG_pendingtree__alloc(pCtx, pPathGiven, &pPendingTree)  );
	SG_ERR_CHECK(  SG_pendingtree__scan(pCtx,
										pPendingTree,
										SG_TRUE,	// Assume fully-recursive for the scan, so that we know how much dirt there is.
										SG_FALSE,	// Commit does not include Lost or Found items in the scan, since it can't do anything about them.
										SG_FALSE,	// Commit does not include Lost or Found items in the scan, since it can't do anything about them.
										gpBlankFilespec)  );
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);

	// We've decided that we want COMMIT to throw an error if there
	// are any LOST items (on a full COMMIT or any LOST items chosen
	// on a partial COMMIT).
	//
	// Problem: the _mark_notes_for_commit stuff only knows about
	// the *potential* set of of things that should be committed
	// (based upon argv and/or include/exclude/ignore filters); it
	// doesn't know about what is dirty -- that gets computed later
	// as dirty stuff is identified and blobs are created.  That's
	// too late to stop the music.  So, we build a full treediff and
	// apply the filters.  The net result should be a subset of the
	// *potential* set that contains just the *dirty* stuff that
	// needs to be committed.
	//
	// We have to to this on a fresh pendingtree because we use
	// different settings to compute a diff (and which should not
	// be saved).

	SG_ERR_CHECK(  SG_pendingtree__alloc(pCtx, pPathGiven, &pPendingTree)  );
	SG_ERR_CHECK(  SG_pendingtree__diff_or_status(pCtx, pPendingTree, NULL, NULL, &pTreeDiff)  );
	SG_ERR_CHECK(  SG_treediff2__file_spec_filter(pCtx, pTreeDiff,
												  count_args, paszArgs, bRecursive,
												  pFilespec)  );
	memset(&tdStats, 0, sizeof(tdStats));
	SG_ERR_CHECK(  SG_treediff2__compute_stats(pCtx, pTreeDiff, &tdStats)  );
	if (tdStats.nrLost > 0)
	{
		SG_ERR_THROW(  SG_ERR_CANNOT_COMMIT_LOST_ENTRY  );
	}
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);

	// Reload the pendingtree (because the above save/save was destructive to
	// the in-memory copy).
	//
	// TODO 2010/12/03 It'd be so much faster if we fixed __scan() and
	// TODO            the __save() to not trash the in-memory pendingtree
	// TODO            so that the following step could just *USE* the
	// TODO            tree we just built....
	// TODO
	// TODO 2011/03/04 I just split SG_pendingtree__scan() into 2 parts
	// TODO            so we could use the first part here and speed things
	// TODO            up.  See SG_pendingtree__scan__no_unload__no_save().

	SG_ERR_CHECK(  SG_pendingtree__alloc(pCtx, pPathGiven, &pPendingTree)  );

	// build an audit record for this commit.

    SG_ERR_CHECK(  SG_pendingtree__get_repo(pCtx, pPendingTree, &pRepo)  );
	SG_ERR_CHECK(  SG_audit__init(pCtx, &q, pRepo,
								  i64Date,
								  ((pszUser && *pszUser) ? pszUser : SG_AUDIT__WHO__FROM_SETTINGS))  );

	SG_ERR_CHECK(  SG_pendingtree__precommit_check(pCtx, pPendingTree, &q,
										  count_args, paszArgs,
										  bRecursive, pFilespec,
										  ppszAssocs, count_assocs,
												   &descs)  );

	if (promptfn)
	{
		SG_bool bNoComment = SG_FALSE;

		SG_NULLARGCHECK( boilerplate );

		SG_ERR_CHECK(  (*promptfn)(pCtx, boilerplate, &localComment, descs )  );

		if (localComment)
		{
			SG_ERR_CHECK(  SG_sz__is_all_whitespace(pCtx, localComment, &bNoComment)  );
		}
		else
			bNoComment = SG_TRUE;

		if(bNoComment)
		{
			SG_ERR_THROW(SG_ERR_EMPTYCOMMITMESSAGE);
		}

		SG_ERR_CHECK(  SG_sz__trim(pCtx, localComment, NULL, &trimmedComment));
		pszComment = trimmedComment;
	}
	else if(pszComment)
	{
		SG_ERR_CHECK(  SG_sz__trim(pCtx, pszComment, NULL, &trimmedComment));
		pszComment = trimmedComment;
	}

	// do the guts of the commit.  this takes care of creating the
	// new CSET, cleaning up temp dirs & dirt, setting the pending
	// tree to the post-commit dirt, adding an audit record, logging
	// the checkin comment, adding any WIT associations, and applying
	// stamps.

	SG_ERR_CHECK(  SG_pendingtree__commit(pCtx, pPendingTree, &q,
										  count_args, paszArgs,
										  bRecursive, pFilespec,
										  ppszAssocs, count_assocs,
										  psaStamps,
										  bNoSubmodules,
										  pszComment,
										  ((ppsz_hid_cset) ? &pdn : NULL))  );
	if (ppsz_hid_cset)
		SG_ERR_CHECK(  SG_dagnode__get_id(pCtx, pdn, ppsz_hid_cset)  );

fail:
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
    SG_DAGNODE_NULLFREE(pCtx, pdn);
	SG_PATHNAME_NULLFREE(pCtx, pPathGiven);
	SG_TREEDIFF2_NULLFREE(pCtx, pTreeDiff);
	SG_NULLFREE(pCtx, localComment);
	SG_NULLFREE(pCtx, trimmedComment);
	SG_VARRAY_NULLFREE(pCtx, descs);
}

//////////////////////////////////////////////////////////////////


END_EXTERN_C;

#endif//H_SG_VV_VERBS__COMMIT_H
