/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg_vv_verbs__export.h
 *
 * @details Routiens to export the contents of the version
 * control tree to a directory -- like a checkout but without
 * the baggage.
 *
 */

//////////////////////////////////////////////////////////////////

#ifndef H_SG_VV_VERBS__EXPORT_H
#define H_SG_VV_VERBS__EXPORT_H

BEGIN_EXTERN_C;

//////////////////////////////////////////////////////////////////

static void _vv_verbs__export(SG_context * pCtx,
							  const SG_pathname * pPath,
							  const char * pszRepoName,
							  const char * pszHidCSet,
							  SG_bool bNoSubmodules,
							  const SG_vector * pvec_use_submodules,
							  SG_vhash ** ppvhResult)
{
	SG_repo * pRepo = NULL;
	SG_changeset * pcs = NULL;
	SG_vhash * pvhResult = NULL;
	SG_varray * pvaMissingDetails = NULL;
	SG_vhash * pvhCheckoutInstanceData = NULL;
	SG_uint32 count_after = 0;

	// for the top-level EXPORT, we throw when either the repo-name or cset-id is not valid.
	// this gives the expected error message in response to something the user typed directly.
	// TODO 2011/02/01 Think about letting these return status and details in pvhResult instead.

	SG_ERR_CHECK(  SG_REPO__OPEN_REPO_INSTANCE(pCtx, pszRepoName, &pRepo)  );
	SG_ERR_CHECK(  SG_changeset__load_from_repo(pCtx, pRepo, pszHidCSet, &pcs)  );
	//TODO:  why do we load a changeset and then immediately free it.  Should be have a 
	//"does this changeset exist" call instead?
	SG_CHANGESET_NULLFREE(pCtx, pcs);
	SG_REPO_NULLFREE(pCtx, pRepo);

	// before we attempt to do the EXPORT, build instance data for the checkout/export.
	// bNoSubmodules controls how much data is constructed and verify that if there
	// are any submodules that we have a (repo_instance, cset_id) for each of them.
	// this will either return an array of missing-details *or* instance data for
	// this (repo_instance, cset_id).
	//
	// call __could_do_checkout() since we assume that both checkout and export
	// are identical in what they need on the local system (repo, cset) for the
	// all of the nested submodules within.

	SG_ERR_CHECK(  SG_VARRAY__ALLOC(pCtx, &pvaMissingDetails)  );
	SG_ERR_CHECK(  SG_treenode_submodule_veracity__could_do_checkout(pCtx, pPath,
																	 pszRepoName, pszHidCSet,
																	 bNoSubmodules,
																	 pvec_use_submodules,
																	 pvaMissingDetails,
																	 &pvhCheckoutInstanceData)  );
	SG_ERR_CHECK(  SG_varray__count(pCtx, pvaMissingDetails, &count_after)  );
	if (count_after > 0)
	{
		// there is something in the tree that we *will not* be able to populate.

		SG_ASSERT(  (pvhCheckoutInstanceData == NULL)  );

		if (ppvhResult)
		{
			// the caller wants the details, so we don't throw.
			// rather we put the details and status in the vhash.

			SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvhResult)  );

			SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvhResult, "status", SG_VV_VERBS__STATUS__SUBMODULE_PROBLEM)  );
			SG_ERR_CHECK(  SG_vhash__add__varray(pCtx, pvhResult, "sub", &pvaMissingDetails)  );	// this steals our varray
			SG_ASSERT(  (pvaMissingDetails == NULL)  );

			*ppvhResult = pvhResult;
			pvhResult = NULL;
			goto done;
		}
		else
		{
			// TODO 2011/02/01 Should we have a __NESTED_EXPORT_PROBLEM code?
			SG_ERR_THROW(  SG_ERR_SUBMODULE_NESTED_UPDATE_CHECKOUT_PROBLEM  );
		}
	}

	// any required submodules should be able to be exported too.

#if 0 && defined(DEBUG)
	SG_ERR_IGNORE(  SG_vhash_debug__dump_to_console__named(pCtx,
														   pvhCheckoutInstanceData,
														   "vv_verbs__export: CheckoutInstanceData")  );
#endif
	
	SG_ERR_CHECK(  SG_workingdir__do_export(pCtx, pPath, pvhCheckoutInstanceData)  );

	if (ppvhResult)
	{
		SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvhResult)  );
		SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvhResult, "status", SG_VV_VERBS__STATUS__OK)  );
		*ppvhResult = pvhResult;
		pvhResult = NULL;
	}

done:
	;
fail:
	SG_VHASH_NULLFREE(pCtx, pvhResult);
	SG_VARRAY_NULLFREE(pCtx, pvaMissingDetails);
	SG_CHANGESET_NULLFREE(pCtx, pcs);
	SG_REPO_NULLFREE(pCtx, pRepo);
	SG_VHASH_NULLFREE(pCtx, pvhCheckoutInstanceData);
}

//////////////////////////////////////////////////////////////////

void SG_vv_verbs__export(SG_context * pCtx,
						 const char * pszRepoName,
						 const char * pszHidCSet,		//< [in] Full HID of CSET (caller should use SG_vv_utils__lookup_cset()).
						 const char * pszFolder,		//< [in] arbitrary path (abs or rel); defaults to CWD if blank or null.
						 SG_bool bNoSubmodules,			//< [in] do not export contained submodules
						 const SG_vector * pvec_use_submodules,	//< [in] optional list of hints for repo-names of referenced submodules
						 SG_vhash ** ppvhResult)		//< [out] list of submodule problems
{
	SG_pathname * pPathAllocated = NULL;

	SG_NONEMPTYCHECK_RETURN( pszRepoName );
	SG_NONEMPTYCHECK_RETURN( pszHidCSet );
	// pszFolder is optional and defaults to CWD.

	SG_ERR_CHECK(  SG_vv_utils__folder_arg_to_path(pCtx, pszFolder, &pPathAllocated)  );

	// when exporting an arbitrary version we require either a new or
	// empty target directory so that we are not trashing something.

	SG_ERR_CHECK(  SG_vv_utils__verify_dir_empty(pCtx, pPathAllocated)  );

	SG_ERR_CHECK(  _vv_verbs__export(pCtx, pPathAllocated, pszRepoName, pszHidCSet, bNoSubmodules, pvec_use_submodules, ppvhResult)  );
	
fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathAllocated);
}

//////////////////////////////////////////////////////////////////

END_EXTERN_C;

#endif//H_SG_VV_VERBS__EXPORT_H
