/*
 Copyright 2010 SourceGear, LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 */

/**
 *
 * @file sg_vv_verbs__history.h
 *
 * @details Routines to perform most of 'vv history'.
 *
 */

//////////////////////////////////////////////////////////////////

#ifndef H_SG_VV_VERBS__HISTORY_H
#define H_SG_VV_VERBS__HISTORY_H
#include "sg.h"
BEGIN_EXTERN_C;
//////////////////////////////////////////////////////////////////

void _sg_vv_verbs__history_core(SG_context * pCtx, SG_repo * pRepo,
		SG_stringarray * pStringArrayGIDs,
		SG_stringarray * pStringArrayStartingChangesets, const char* pszUser,
		const char* pszStamp, SG_uint32 nResultLimit, SG_bool bLeaves,
		SG_bool bHideObjectMerges, SG_int64 nFromDate, SG_int64 nToDate,
		SG_bool bRecommendDagWalk, SG_bool* pbHasResult,
		SG_history_result ** ppResult, SG_history_token ** ppHistoryToken) {
	SG_uint32 nGIDCount = 0;
	
	if (pStringArrayGIDs) {
		SG_ERR_CHECK( SG_stringarray__count(pCtx, pStringArrayGIDs, &nGIDCount) );
	}
	//if (nGIDCount > 0)
	//	bRecommendDagWalk = SG_TRUE;
	//They are not querying by 
	if (bLeaves && bRecommendDagWalk == SG_FALSE) {
		// They aren't asking for anything complicated, just use Ian's rev numbers
		// to iterate over all changesets.  This is easier, because no dag walk is necessary.
		SG_ERR_CHECK( SG_history__list(pCtx, pRepo, nResultLimit, SG_UINT32_MAX, pStringArrayGIDs, pszUser, pszStamp, nFromDate, nToDate, bHideObjectMerges, pbHasResult, ppResult, ppHistoryToken) );
	} else {
		SG_ERR_CHECK( SG_history__query(pCtx, pRepo, pStringArrayGIDs, pStringArrayStartingChangesets, pszUser, pszStamp, nResultLimit, bHideObjectMerges, nFromDate, nToDate, bRecommendDagWalk, pbHasResult, ppResult, ppHistoryToken) );
	}

	/* fall through */
	fail: ;
}

void _sg_vv_verbs__history__get_starting_changesets(
		SG_context * pCtx,
		SG_repo * pRepo, 
		SG_rev_spec* pRevSpec,
		SG_stringarray ** ppsaHidChangesets, 
		SG_bool * pbRecommendDagWalk,
		SG_bool * pbLeaves) 
{
	SG_uint32 countRevSpecs = 0;
	SG_stringarray* psaHidChangesets = NULL;
	SG_rbtree* prbHidChangesets = NULL;

	*pbLeaves = SG_FALSE;

	if (pRevSpec)
		SG_ERR_CHECK(  SG_rev_spec__count(pCtx, pRevSpec, &countRevSpecs)  );
	if (countRevSpecs)
	{
		SG_uint32 countBranches = 0;
		SG_ERR_CHECK(  SG_rev_spec__get_all__repo(pCtx, pRepo, pRevSpec, SG_TRUE, &psaHidChangesets)  );
		SG_ERR_CHECK(  SG_rev_spec__count_branches(pCtx, pRevSpec, &countBranches)  );
		if (countBranches)
			*pbRecommendDagWalk = SG_TRUE;
	}
	else //Assume that they want all history.  Fill the leaves.
	{
		SG_ERR_CHECK(  SG_repo__fetch_dag_leaves(pCtx, pRepo, SG_DAGNUM__VERSION_CONTROL, &prbHidChangesets)  );
		SG_ERR_CHECK(  SG_rbtree__to_stringarray__keys_only(pCtx, prbHidChangesets, &psaHidChangesets)  );
		*pbRecommendDagWalk = SG_FALSE; //If we are looking at all leaves, we can use the faster
										//list method, instead of querying.
		*pbLeaves = SG_TRUE;
	}

	SG_RETURN_AND_NULL(psaHidChangesets, ppsaHidChangesets);
	
	/* common cleanup */
fail:
	SG_RBTREE_NULLFREE(pCtx, prbHidChangesets);
	SG_STRINGARRAY_NULLFREE(pCtx, psaHidChangesets);
}

void _sg_vv_verbs__history__lookup_gids_by_repopaths(SG_context * pCtx,
		SG_repo * pRepo, const char * psz_dagnode_hid, SG_uint32 count_args,
		const char ** paszArgs, //The arguments can be repo paths
		SG_stringarray ** ppStringArrayGIDs) {
	SG_stringarray * pStringArrayGIDs = NULL;
	char * pszHidForRoot = NULL;
	SG_uint32 i = 0;

	char * pszSearchItemGID = NULL;
	SG_treenode * pTreeNodeRoot = NULL;
	SG_treenode_entry * ptne = NULL;

	SG_ERR_CHECK( SG_STRINGARRAY__ALLOC(pCtx, &pStringArrayGIDs, count_args) );
	SG_ERR_CHECK( SG_changeset__tree__find_root_hid(pCtx, pRepo, psz_dagnode_hid, &pszHidForRoot) );
	SG_ERR_CHECK( SG_treenode__load_from_repo(pCtx, pRepo, pszHidForRoot, &pTreeNodeRoot) );
	for (i = 0; i < count_args; i++) {
		SG_ERR_CHECK( SG_treenode__find_treenodeentry_by_path(pCtx, pRepo, pTreeNodeRoot, paszArgs[i], &pszSearchItemGID, &ptne) );
		if (pszSearchItemGID == NULL) {
			SG_TREENODE_NULLFREE(pCtx, pTreeNodeRoot);
			SG_NULLFREE(pCtx, pszHidForRoot);
			SG_ERR_THROW( SG_ERR_NOT_FOUND );
		}
		SG_ERR_CHECK( SG_stringarray__add(pCtx, pStringArrayGIDs, pszSearchItemGID) );
		SG_NULLFREE(pCtx, pszSearchItemGID);
		SG_TREENODE_ENTRY_NULLFREE(pCtx, ptne);
	}
	SG_RETURN_AND_NULL(pStringArrayGIDs, ppStringArrayGIDs);
	fail:
	SG_NULLFREE(pCtx, pszHidForRoot);
	SG_TREENODE_NULLFREE(pCtx, pTreeNodeRoot);
	SG_STRINGARRAY_NULLFREE(pCtx, pStringArrayGIDs);
	SG_NULLFREE(pCtx, pszSearchItemGID);
	SG_TREENODE_ENTRY_NULLFREE(pCtx, ptne);
}

//////////////////////////////////////////////////////////////////
void SG_vv_verbs__history__repo2(
		SG_context * pCtx,
		SG_repo * pRepo, 
		SG_uint32 count_args,
		const char ** paszArgs, //The arguments can be repo paths
		SG_rev_spec* pRevSpec,
		const char* pszUser, const char* pszStamp,
		SG_uint32 nResultLimit, SG_bool bHideObjectMerges, SG_int64 nFromDate,
		SG_int64 nToDate, SG_bool* pbHasResult, SG_history_result ** ppResult,
		SG_history_token ** ppHistoryToken) {

	SG_stringarray * pStringArrayChangesets = NULL;
	SG_bool bRecommendDagWalk = SG_FALSE;
	SG_bool bLeaves = SG_FALSE;
	const char * pszCurrentDagnodeID = NULL;
	SG_stringarray * pStringArrayGIDs = NULL;
	SG_uint32 i = 0, nChangesetCount = 0;

	//Determine the starting changeset IDs.  strBranch and bLeaves control this.
	//We do this step here, so that repo paths can be looked up before we call into history__core.
	SG_ERR_CHECK( _sg_vv_verbs__history__get_starting_changesets(pCtx, pRepo, pRevSpec, &pStringArrayChangesets, &bRecommendDagWalk, &bLeaves) );

	SG_ERR_CHECK(  SG_stringarray__count(pCtx, pStringArrayChangesets, &nChangesetCount)  );
	//Look up the GIDs for all of the arguments.
	//Try every changeset, until we get one that has the GID in question.
	for (i = 0; i < nChangesetCount; i++)
	{
		SG_ERR_CHECK( SG_stringarray__get_nth(pCtx, pStringArrayChangesets, i, &pszCurrentDagnodeID) );

		//This might be used if you have --leaves, or if there are multiple parents
		//since they specified a changeset, we need to use the full repo path @/blah/blah to look up the objects
		_sg_vv_verbs__history__lookup_gids_by_repopaths(pCtx, pRepo, pszCurrentDagnodeID, count_args, paszArgs, &pStringArrayGIDs);
		if (SG_context__has_err(pCtx))
		{
			if (i == (nChangesetCount - 1) || ! SG_context__err_equals(pCtx, SG_ERR_NOT_FOUND) )
			{
				SG_STRINGARRAY_NULLFREE(pCtx, pStringArrayGIDs);
				SG_ERR_RETHROW;
			}
			else
			{
				SG_STRINGARRAY_NULLFREE(pCtx, pStringArrayGIDs);
				SG_context__err_reset(pCtx);
			}
		}
		else
			break;
	}

	//Call history core with the GIDs
	SG_ERR_CHECK( _sg_vv_verbs__history_core(pCtx, pRepo, pStringArrayGIDs,
					pStringArrayChangesets,
					pszUser, pszStamp, nResultLimit, bLeaves, bHideObjectMerges,
					nFromDate, nToDate, bRecommendDagWalk, pbHasResult, ppResult, ppHistoryToken) );

fail:
	SG_STRINGARRAY_NULLFREE(pCtx, pStringArrayGIDs);
	SG_STRINGARRAY_NULLFREE(pCtx, pStringArrayChangesets);
}


//////////////////////////////////////////////////////////////////
void SG_vv_verbs__history__repo(
		SG_context * pCtx,
		const char * pszRepoDescriptorName, //The local descriptor name.
		SG_uint32 count_args,
		const char ** paszArgs, //The arguments can be repo paths
		SG_rev_spec* pRevSpec,
		const char* pszUser, const char* pszStamp,
		SG_uint32 nResultLimit, SG_bool bHideObjectMerges, SG_int64 nFromDate,
		SG_int64 nToDate, SG_bool* pbHasResult, 
		SG_vhash** ppvhBranchPile,
		SG_history_result ** ppResult,
		SG_history_token ** ppHistoryToken) 
{
	SG_repo * pRepo = NULL;
    
	//Open the repo.
	SG_ERR_CHECK( SG_REPO__OPEN_REPO_INSTANCE(pCtx, pszRepoDescriptorName, &pRepo) );
    SG_ERR_CHECK( SG_vv_verbs__history__repo2(pCtx, pRepo, count_args,paszArgs,pRevSpec,
        pszUser, pszStamp, nResultLimit, bHideObjectMerges, nFromDate, nToDate, pbHasResult, ppResult, ppHistoryToken));

	/* This is kind of a hack. History callers often need branch data to format ouput.
	 * But we open the repo down here. I didn't want to open/close it again. So we do this. */
	if (ppvhBranchPile)
		SG_ERR_CHECK(  SG_vc_branches__cleanup(pCtx, pRepo, ppvhBranchPile)  );
	
fail:
	SG_REPO_NULLFREE(pCtx, pRepo);
	
}



///////////////////////////////////////////////////////////////////

void SG_vv_verbs__history__workingFolder(
		SG_context * pCtx,
		const char * pszPendingTreePath, //The local path that will be used to lookup the pending tree.  If this is null, the pendingtree will be looked up by the current working folder.
		SG_uint32 count_args,
		const char ** paszArgs, //The arguments can be either local paths or repo paths
		SG_rev_spec* pRevSpec,
		const char* pszUser, const char* pszStamp,
		SG_bool bDetectCurrentBranch, SG_uint32 nResultLimit,
		SG_bool bHideObjectMerges, SG_int64 nFromDate, SG_int64 nToDate,
		SG_bool* pbHasResult, 
		SG_vhash** ppvhBranchPile,
		SG_history_result ** ppResult,
		SG_history_token ** ppHistoryToken)
{
	SG_pendingtree * pPendingTree = NULL;
	SG_pathname * pPathname = NULL;
	SG_pathname * pPathnameTop = NULL;
	SG_repo * pRepo = NULL;
	SG_stringarray * pStringArrayGIDs = NULL;
	SG_bool bHistoryOnRoot = SG_FALSE;
	SG_stringarray * pStringArrayChangesets = NULL;
	SG_bool bRecommendDagWalk = SG_FALSE;
	SG_bool bLeaves = SG_FALSE;
	char * pszBranchName = NULL;
	SG_vhash* pvhBranchPile = NULL;
	SG_varray* pvaParents = NULL;
	SG_bool bMyBranchWalkRecommendation = SG_FALSE;
	
	SG_rev_spec* pMyRevSpec = pRevSpec;
	SG_rev_spec* pFreeRevSpec = NULL;
	
	if (pszPendingTreePath == NULL) 
	{
		SG_ERR_CHECK( SG_PENDINGTREE__ALLOC_FROM_CWD(pCtx, &pPendingTree) );
	} else {
		SG_ERR_CHECK( SG_PATHNAME__ALLOC__SZ(pCtx, &pPathname, pszPendingTreePath) );
		SG_ERR_CHECK( SG_PENDINGTREE__ALLOC(pCtx, pPathname, &pPendingTree) );
	}
	SG_ERR_CHECK( SG_STRINGARRAY__ALLOC(pCtx, &pStringArrayGIDs, count_args) );
	SG_ERR_CHECK( SG_pendingtree__get_gids_for_paths(pCtx, pPendingTree, count_args, paszArgs, pStringArrayGIDs, &bHistoryOnRoot) );

	SG_ERR_CHECK( SG_pendingtree__get_repo(pCtx, pPendingTree, &pRepo) );

	/* If no revisions were specified, and the caller wants us to use the current branch, create a revision spec with the current branch. */
	{
		SG_uint32 countRevsSpecified = 0;
		if (pMyRevSpec)
			SG_ERR_CHECK(  SG_rev_spec__count(pCtx, pMyRevSpec, &countRevsSpecified)  );
		else
		{
			SG_ERR_CHECK(  SG_REV_SPEC__ALLOC(pCtx, &pFreeRevSpec)  );
			pMyRevSpec = pFreeRevSpec;
		}
		if (bDetectCurrentBranch && countRevsSpecified == 0) 
		{
			SG_ERR_CHECK(  SG_pendingtree__get_working_directory_top(pCtx, pPendingTree, &pPathnameTop)  );
			SG_ERR_CHECK(  SG_workingdir__get_branch_info(pCtx, pPathnameTop, &pszBranchName)  );
			if (pszBranchName)
			{
				/* The working folder is attached to a branch. Does it exist? */
				SG_bool bHasBranches = SG_FALSE;
				SG_bool bBranchExists = SG_FALSE;

				SG_ERR_CHECK(  SG_vc_branches__cleanup(pCtx, pRepo, &pvhBranchPile)  );
				SG_ERR_CHECK(  SG_vhash__has(pCtx, pvhBranchPile, "branches", &bHasBranches)  );
				if (bHasBranches)
				{
					SG_vhash* pvhRefBranches;
					SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvhBranchPile, "branches", &pvhRefBranches)  );
					SG_ERR_CHECK(  SG_vhash__has(pCtx, pvhRefBranches, pszBranchName, &bBranchExists)  );
				}
				
				if (bBranchExists)
				{
					SG_uint32 numParents, i;
					const char* pszRefParent;

					/* If that branch exists, just add to our rev spec. */
					SG_ERR_CHECK(  SG_rev_spec__add_branch(pCtx, pMyRevSpec, pszBranchName)  );

					/* Plus, if the working folder's parents are not in the branch (yet), add them as well
					 * (they'll be in it after the user commits something...). */
					SG_ERR_CHECK(  SG_pendingtree__get_wd_parents(pCtx, pPendingTree, &pvaParents)  );
					SG_ERR_CHECK(  SG_varray__count(pCtx, pvaParents, &numParents)  );
					for (i = 0; i < numParents; i++)
					{
						SG_bool already_in_rev_spec = SG_FALSE;
						SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pvaParents, i, &pszRefParent)  );
						SG_ERR_CHECK(  SG_rev_spec__contains(pCtx, pRepo, pRevSpec, pszRefParent, &already_in_rev_spec)  );
						if(!already_in_rev_spec)
							SG_ERR_CHECK(  SG_rev_spec__add_rev(pCtx, pRevSpec, pszRefParent)  );
					}
				}
				else
				{
					/* If the branch doesn't exist, add the working folder's baseline(s) to the rev spec
					 * and force a dag walk. */
					SG_uint32 numParents, i;
					const char* pszRefParent;

					SG_ERR_CHECK(  SG_pendingtree__get_wd_parents(pCtx, pPendingTree, &pvaParents)  );
					SG_ERR_CHECK(  SG_varray__count(pCtx, pvaParents, &numParents)  );
					for (i = 0; i < numParents; i++)
					{
						SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pvaParents, i, &pszRefParent)  );
						SG_ERR_CHECK(  SG_rev_spec__add_rev(pCtx, pRevSpec, pszRefParent)  );
					}
					bMyBranchWalkRecommendation = SG_TRUE;
				}

			}
		}
	}

	// Determine the starting changeset IDs.  strBranch and bLeaves control this.
	// We do this step here, so that repo paths can be looked up before we call into history__core.
	SG_ERR_CHECK( _sg_vv_verbs__history__get_starting_changesets(pCtx, pRepo, pMyRevSpec, &pStringArrayChangesets, &bRecommendDagWalk, &bLeaves) );

	bRecommendDagWalk = bRecommendDagWalk || bMyBranchWalkRecommendation;
	
	//This hack is here to  detect when we're being asked for the parent of a certain
	//object from the sg_parents code.  parents always wants the dag walk.
	//The better solution would be to allow users to pass in a flag about their dagwalk
	//preferences
	if (count_args == 1 && nResultLimit == 1)
		bRecommendDagWalk = SG_TRUE;

	SG_ERR_CHECK( _sg_vv_verbs__history_core(pCtx, pRepo, pStringArrayGIDs,
					pStringArrayChangesets,
					pszUser, pszStamp, nResultLimit, bLeaves, bHideObjectMerges,
					nFromDate, nToDate, bRecommendDagWalk, pbHasResult, ppResult, ppHistoryToken) );

	/* This is kind of a hack. History callers often need branch data to format ouput.
	 * But we open the repo down here. I didn't want to open/close it again. And there's logic
	 * in here about which repo to open. So instead, we do this. */
	if (ppvhBranchPile)
	{
		if (pvhBranchPile)
		{
			*ppvhBranchPile = pvhBranchPile;
			pvhBranchPile = NULL;
		}
		else
			SG_ERR_CHECK(  SG_vc_branches__cleanup(pCtx, pRepo, ppvhBranchPile)  );
	}

	/* common cleanup */
fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathnameTop);
	SG_REV_SPEC_NULLFREE(pCtx, pFreeRevSpec);
	SG_NULLFREE(pCtx, pszBranchName);
	SG_PATHNAME_NULLFREE(pCtx, pPathname);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_STRINGARRAY_NULLFREE(pCtx, pStringArrayChangesets);
	SG_STRINGARRAY_NULLFREE(pCtx, pStringArrayGIDs);
	SG_VHASH_NULLFREE(pCtx, pvhBranchPile);
	SG_VARRAY_NULLFREE(pCtx, pvaParents);
}

void SG_vv_verbs__history__fetch_more__repo(
		SG_context * pCtx,
		const char * pszRepoDescriptorName, //The local descriptor name.
		SG_history_token * pHistoryToken, SG_uint32 nResultLimit,
		SG_history_result ** ppHistoryResults, SG_history_token **ppNewToken) {
	SG_repo * pRepo = NULL;

	//Open the repo.
	SG_ERR_CHECK( SG_REPO__OPEN_REPO_INSTANCE(pCtx, pszRepoDescriptorName, &pRepo) );

	SG_ERR_CHECK( SG_history__fetch_more(pCtx, pRepo, pHistoryToken, nResultLimit, ppHistoryResults, ppNewToken) );
	fail:
	SG_REPO_NULLFREE(pCtx, pRepo);

}

void SG_vv_verbs__history__fetch_more__repo2(
		SG_context * pCtx,
		SG_repo * pRepo, //The local descriptor name.
		SG_history_token * pHistoryToken, SG_uint32 nResultLimit,
		SG_history_result ** ppHistoryResults, SG_history_token **ppNewToken) {


	SG_ERR_CHECK_RETURN( SG_history__fetch_more(pCtx, pRepo, pHistoryToken, nResultLimit, ppHistoryResults, ppNewToken) );

}

void SG_vv_verbs__history__fetch_more__working_folder(
		SG_context * pCtx,
		const char * pszPendingTreePath, //The local path that will be used to lookup the pending tree.  If this is null, the pendingtree will be looked up by the current working folder.
		SG_history_token * pHistoryToken, SG_uint32 nResultLimit,
		SG_history_result ** ppHistoryResults, SG_history_token **ppNewToken) {
	SG_repo * pRepo = NULL;
	SG_pendingtree * pPendingTree = NULL;
	SG_pathname * pPathname = NULL;

	if (pszPendingTreePath == NULL) {
		SG_ERR_CHECK( SG_PENDINGTREE__ALLOC_FROM_CWD(pCtx, &pPendingTree) );
	} else {
		SG_ERR_CHECK( SG_PATHNAME__ALLOC__SZ(pCtx, &pPathname, pszPendingTreePath) );
		SG_ERR_CHECK( SG_PENDINGTREE__ALLOC(pCtx, pPathname, &pPendingTree) );
	}

	SG_ERR_CHECK( SG_pendingtree__get_repo(pCtx, pPendingTree, &pRepo) );
	SG_ERR_CHECK( SG_history__fetch_more(pCtx, pRepo, pHistoryToken, nResultLimit, ppHistoryResults, ppNewToken) );
	fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathname);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);

}
END_EXTERN_C;

#endif//H_SG_VV_VERBS__HISTORY_H
