/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg_vv_verbs__merge.h
 *
 * @details Routines to perform most of 'vv merge'.
 *
 */

//////////////////////////////////////////////////////////////////

#ifndef H_SG_VV_VERBS__MERGE_H
#define H_SG_VV_VERBS__MERGE_H

BEGIN_EXTERN_C;

//////////////////////////////////////////////////////////////////

void SG_vv_verbs__merge(SG_context * pCtx,
						const char * pszFolder,
						const char * psz_other_csetid,
						const char * pszToolOverride,
						SG_bool bTest,
						const SG_vector * pvec_use_submodules,
						SG_vhash ** ppvhResults,
						SG_wd_plan ** ppPlan)
{
	SG_pathname * pPath = NULL;
	SG_pendingtree * pPendingTree = NULL;
	SG_varray * pva_wd_parents_copy = NULL;
	SG_varray * pvaMissingDetails = NULL;
	SG_varray * pvaWarnings = NULL;
	SG_varray * pva_wd_issues_copy = NULL;
	SG_vhash * pvhResults = NULL;
	SG_mrg * pMrg = NULL;
	const SG_varray * pva_wd_issues_ref = NULL;
	SG_uint32 nrParents;
	SG_uint32 nrUnresolvedIssues;
	SG_bool bHaveIssues;

	// TODO 2010/07/06 Decide if we want to warn them if the baseline is not a leaf/head
	// TODO            or make them give us a --force or something?
	// TODO
	// TODO            We have to allow them to do the merge using the baseline, but we
	// TODO            might want to warn them (and maybe nudge them) that they probably
	// TODO            should be starting from a leaf rather than something with children.
	// TODO
	// TODO            See sprawl-964.

	SG_bool bComplainIfBaselineNotLeaf = SG_FALSE;

	// pszFolder is optional
	SG_NONEMPTYCHECK_RETURN(psz_other_csetid);
	// pszToolOverride is optional
	// pvec_use_submodules is optional
	SG_NULLARGCHECK_RETURN(ppvhResults);
	SG_NULLARGCHECK_RETURN(ppPlan);

	SG_ERR_CHECK(  SG_vv_utils__folder_arg_to_path(pCtx, pszFolder, &pPath)  );

	SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pPath, &pPendingTree)  );

	// we do not allow them to start merge if the WD has an
	// uncommitted merge.  get our own copy of the parent array.

	SG_ERR_CHECK(  SG_pendingtree__get_wd_parents(pCtx, pPendingTree, &pva_wd_parents_copy)  );
	SG_ERR_CHECK(  SG_varray__count(pCtx, pva_wd_parents_copy, &nrParents)  );
	if (nrParents > 1)
	{
		SG_ERR_THROW2(  SG_ERR_CANNOT_DO_WHILE_UNCOMMITTED_MERGE,
						(pCtx, "Working copy already contains an uncommitted merge."));
	}

	SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvhResults)  );
	SG_ERR_CHECK(  SG_VARRAY__ALLOC(pCtx, &pvaMissingDetails)  );

	SG_ERR_CHECK(  SG_MRG__ALLOC(pCtx, pPendingTree, bComplainIfBaselineNotLeaf, pszToolOverride, &pMrg)  );

	SG_mrg__compute_merge(pCtx, pMrg,
						  psz_other_csetid, pvec_use_submodules,
						  pvaMissingDetails, &nrUnresolvedIssues);
	if (SG_context__has_err(pCtx))
	{
		if (SG_context__err_equals(pCtx, SG_ERR_PORTABILITY_WARNINGS))
		{
			SG_context__err_reset(pCtx);
			SG_ERR_CHECK(  SG_pendingtree__steal_warnings(pCtx, pPendingTree, &pvaWarnings)  );
			SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvhResults, "status", SG_VV_VERBS__STATUS__PORTABILITY_WARNING)  );
			SG_ERR_CHECK(  SG_vhash__add__varray(pCtx, pvhResults, "port", &pvaWarnings)  );
			goto done;
		}
		else if (SG_context__err_equals(pCtx, SG_ERR_SUBMODULE_NESTED_UPDATE_CHECKOUT_PROBLEM))
		{
			SG_context__err_reset(pCtx);
			SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvhResults, "status", SG_VV_VERBS__STATUS__SUBMODULE_PROBLEM)  );
			SG_ERR_CHECK(  SG_vhash__add__varray(pCtx, pvhResults, "sub", &pvaMissingDetails)  );
			goto done;
		}

		SG_ERR_RETHROW;
	}

	if (!bTest)
	{
		SG_ERR_CHECK(  SG_mrg__apply_merge(pCtx, pMrg)  );

		// The issues are computed during __apply_merge and the whole
		// pendingtree data (pending and issues) are deleted during __save(),
		// so if we want the issues we have to get them now.  (Or we could
		// reload the pendingtree after the save.)  Yet another instance of
		// that single operation quirk.

		SG_ERR_CHECK(  SG_pendingtree__get_wd_issues__ref(pCtx, pPendingTree, &bHaveIssues, &pva_wd_issues_ref)  );
		if (bHaveIssues)
			SG_ERR_CHECK(  SG_varray__alloc__copy(pCtx, &pva_wd_issues_copy, pva_wd_issues_ref)  );

		SG_ERR_CHECK(  SG_pendingtree__save(pCtx, pPendingTree)  );
	}

	// if we actually did the merge, the live parents array would now have 2 entries.
	// if we were given --test, the live array will still have only 1 entry, but we
	// want to report what the parents *would* be if the merge were actually performed.
	// so we fake up the copy we made at the top.
	SG_ERR_CHECK(  SG_varray__append__string__sz(pCtx, pva_wd_parents_copy, psz_other_csetid)  );

	SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvhResults, "status", SG_VV_VERBS__STATUS__OK)  );
	SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvhResults, "nr_unresolved_issues", nrUnresolvedIssues)  );
	SG_ERR_CHECK(  SG_vhash__add__varray(pCtx, pvhResults, "parents", &pva_wd_parents_copy)  );

	if (pva_wd_issues_copy)
		SG_ERR_CHECK(  SG_vhash__add__varray(pCtx, pvhResults, "issues", &pva_wd_issues_copy)  );

	// TODO 2011/02/22 I'd like to put the plan in the pvhResults
	// TODO            but I'm not sure that converting a plan to a multi-level
	// TODO            vhash is worth the bother right now).  So for now, just
	// TODO            return it directly.
	// TODO
	// TODO            Perhaps we could put the plan-stats into pvhResults.
	SG_ERR_CHECK(  SG_mrg__steal_wd_plan(pCtx, pMrg, ppPlan)  );

done:
	*ppvhResults = pvhResults;
	pvhResults = NULL;

fail:
	// leave the MERGE TEMP directory so that RESOLVE can use it for any
	// manual file content merges.

	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_VARRAY_NULLFREE(pCtx, pvaMissingDetails);
	SG_VARRAY_NULLFREE(pCtx, pvaWarnings);
	SG_MRG_NULLFREE(pCtx, pMrg);
	SG_VHASH_NULLFREE(pCtx, pvhResults);
	SG_VARRAY_NULLFREE(pCtx, pva_wd_parents_copy);
	SG_VARRAY_NULLFREE(pCtx, pva_wd_issues_copy);
}

//////////////////////////////////////////////////////////////////

END_EXTERN_C;

#endif//H_SG_VV_VERBS__MERGE_H
