/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg_vv_verbs__revert.h
 *
 * @details Routines to perform most of 'vv revert'.
 *
 */

//////////////////////////////////////////////////////////////////

#ifndef H_SG_VV_VERBS__REVERT_H
#define H_SG_VV_VERBS__REVERT_H

BEGIN_EXTERN_C;

//////////////////////////////////////////////////////////////////

void SG_vv_verbs__revert(SG_context * pCtx,
						 const char * pszFolder,
						 SG_bool bRecursive,
						 SG_bool bNoSubmodules,
						 SG_bool bVerbose,
						 SG_bool bTest,
						 SG_bool bNoBackups,
						 const SG_vector * pvec_use_submodules,
						 const SG_file_spec * pFilespec,
						 SG_uint32 count_items, const char ** paszItems,
						 SG_vhash ** ppvhResult)
{
	SG_pathname * pPath = NULL;
	SG_pendingtree * pPendingTree = NULL;
	SG_pendingtree_action_log_enum eActionLog = 0;
	SG_varray * pvaMissingDetails = NULL;
	SG_varray * pvaActionLog = NULL;
	SG_varray * pvaWarnings = NULL;
	SG_vhash * pvhResult = NULL;
	SG_uint32 count_details;

	SG_NULLARGCHECK_RETURN( ppvhResult );

	if (bTest)
	{
		eActionLog = SG_PT_ACTION__LOG_IT;
	}
	else
	{
		eActionLog = SG_PT_ACTION__DO_IT;
		if (bVerbose)
			eActionLog |= SG_PT_ACTION__LOG_IT;
	}

	SG_ERR_CHECK(  SG_VARRAY__ALLOC(pCtx, &pvaMissingDetails)  );

	SG_ERR_CHECK(  SG_vv_utils__folder_arg_to_path(pCtx, pszFolder, &pPath)  );
	SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pPath, &pPendingTree)  );

	SG_pendingtree__revert(pCtx,
						   pPendingTree,
						   eActionLog,
						   count_items, paszItems,
						   bRecursive,
						   bNoSubmodules,
						   bNoBackups,
						   pvec_use_submodules,
						   pFilespec,
						   SG_FALSE,	// bPrecomputeOnly
						   pvaMissingDetails);
	if (SG_context__has_err(pCtx))
	{
		if (SG_context__err_equals(pCtx, SG_ERR_PORTABILITY_WARNINGS))
		{
			SG_context__err_reset(pCtx);
			SG_ERR_CHECK(  SG_pendingtree__steal_warnings(pCtx, pPendingTree, &pvaWarnings)  );

			SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvhResult)  );
			SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvhResult, "status", SG_VV_VERBS__STATUS__PORTABILITY_WARNING)  );
			SG_ERR_CHECK(  SG_vhash__add__varray(pCtx, pvhResult, "port", &pvaWarnings)  );
			goto done;
		}

		SG_ERR_RETHROW;
	}

	SG_ERR_CHECK(  SG_varray__count(pCtx, pvaMissingDetails, &count_details)  );
	if (count_details > 0)
	{
		// if there were any submodules that could not be reverted (such as
		// when the nested revert requires we re-create a deleted (but not
		// committed) submodule and we can't uniquely figure out what to
		// checkout), then we stopped everything early and did not do
		// the full precompute (but hey, we didn't throw).

		SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvhResult)  );
		SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvhResult, "status", SG_VV_VERBS__STATUS__SUBMODULE_PROBLEM)  );
		SG_ERR_CHECK(  SG_vhash__add__varray(pCtx, pvhResult, "sub", &pvaMissingDetails)  );
		goto done;
	}

	SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvhResult)  );
	SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvhResult, "status", SG_VV_VERBS__STATUS__OK)  );

	SG_ERR_CHECK(  SG_pendingtree__steal_action_log(pCtx, pPendingTree, &pvaActionLog)  );
	if (pvaActionLog)	// should only have been allocated if --test or --verbose
		SG_ERR_CHECK(  SG_vhash__add__varray(pCtx, pvhResult, "log", &pvaActionLog)  );
	else
		SG_ERR_CHECK(  SG_vhash__add__null(pCtx, pvhResult, "log")  );

done:
	*ppvhResult = pvhResult;
	pvhResult = NULL;

fail:
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_VARRAY_NULLFREE(pCtx, pvaActionLog);
	SG_VARRAY_NULLFREE(pCtx, pvaWarnings);
	SG_VARRAY_NULLFREE(pCtx, pvaMissingDetails);
	SG_VHASH_NULLFREE(pCtx, pvhResult);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
}

//////////////////////////////////////////////////////////////////

END_EXTERN_C;

#endif//H_SG_VV_VERBS__REVERT_H
