/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg_vv_verbs__update.h
 *
 * @details Routines to perform most of 'vv update'.
 *
 */

//////////////////////////////////////////////////////////////////

#ifndef H_SG_VV_VERBS__UPDATE_H
#define H_SG_VV_VERBS__UPDATE_H

BEGIN_EXTERN_C;

//////////////////////////////////////////////////////////////////

void SG_vv_verbs__update(SG_context * pCtx,
						 const char * pszFolder,
						 const char * pszTargetChangeset,
						 const char * pszTargetBranchName,
						 SG_bool bVerbose,
						 SG_bool bTest,
						 const SG_vector * pvec_use_submodules,
						 SG_varray ** ppvaActionLog,
						 SG_varray ** ppvaWarnings,
						 SG_varray ** ppvaMissingDetails)
{
	SG_pathname * pPath = NULL;
	SG_pendingtree * pPendingTree = NULL;
	SG_pendingtree_action_log_enum eActionLog = 0;
	SG_varray * pvaMissingDetails = NULL;
	SG_uint32 count_details;

	// pszFolder is optional
	// pszTargetChangeset is optional
	// pszTargetBranchName is optional
	// pvec_use_submodules is optional
	// ppvaActionLog is optional
	// ppvaWarnings is optional
	// ppvaMissingDetails is optional

	SG_ERR_CHECK(  SG_vv_utils__folder_arg_to_path(pCtx, pszFolder, &pPath)  );

	SG_ERR_CHECK(  SG_VARRAY__ALLOC(pCtx, &pvaMissingDetails)  );

	SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pPath, &pPendingTree)  );
	SG_pendingtree__update_baseline__precompute(pCtx,
												pPendingTree,
												pszTargetChangeset,
												pvec_use_submodules,
												pvaMissingDetails);
	if (SG_CONTEXT__HAS_ERR(pCtx))
	{
		if (SG_context__err_equals(pCtx, SG_ERR_CHANGESET_BLOB_NOT_FOUND) && pszTargetBranchName && *pszTargetBranchName)
		{
			SG_ERR_REPLACE(SG_ERR_CHANGESET_BLOB_NOT_FOUND, SG_ERR_BRANCH_HEAD_CHANGESET_NOT_PRESENT);
			/* Deliberately not using an SG_ERR_CHECK macro. Leave it that way. */
			SG_vc_branches__throw_changeset_not_present(pCtx, pszTargetBranchName, pszTargetChangeset);
		}
		else
			SG_ERR_RETHROW;
	}

	SG_ERR_CHECK(  SG_varray__count(pCtx, pvaMissingDetails, &count_details)  );
	if (count_details > 0)
	{
		// if there were any submodules that could not be updated (such as
		// when the nested update requires a cset that the sub repo doesn't
		// have locally), then we stopped everything early and did not do
		// the full precompute (but hey, we didn't throw).

		if (ppvaMissingDetails)
		{
			*ppvaMissingDetails = pvaMissingDetails;
			pvaMissingDetails = NULL;
			goto done;
		}
		else
		{
			SG_ERR_THROW2(  SG_ERR_SUBMODULE_NESTED_UPDATE_CHECKOUT_PROBLEM,
							(pCtx, "%s", SG_pathname__sz(pPath))  );
		}
	}

	// no submodule issues so the full precompute ran and we can continue
	// with the apply phase.

	if (bTest)
		eActionLog |= SG_PT_ACTION__LOG_IT;
	else
		eActionLog |= SG_PT_ACTION__DO_IT;

	if (bVerbose)
		eActionLog |= SG_PT_ACTION__LOG_IT;

	SG_ERR_CHECK(  SG_pendingtree__update_baseline__apply(pCtx,
														  pPendingTree,
														  eActionLog,														  
														  pszTargetBranchName)  );

	// steal the logs so that we can destroy the pendingtree

	if (ppvaActionLog)
		SG_ERR_CHECK(  SG_pendingtree__steal_action_log(pCtx, pPendingTree, ppvaActionLog)  );

	if (ppvaMissingDetails)
		*ppvaMissingDetails = NULL;

fail:
	if (ppvaWarnings && SG_context__err_equals(pCtx,SG_ERR_PORTABILITY_WARNINGS))
		SG_ERR_IGNORE(  SG_pendingtree__steal_warnings(pCtx, pPendingTree, ppvaWarnings)  );

done:
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_VARRAY_NULLFREE(pCtx, pvaMissingDetails);
}

//////////////////////////////////////////////////////////////////

void SG_vv_verbs__update2(SG_context * pCtx,
						  const char * pszFolder,
						  const char * pszTargetChangeset,
						  const char * pszTargetBranchName,
						  SG_bool bVerbose,
						  SG_bool bTest,
						  const SG_vector * pvec_use_submodules,
						  SG_vhash ** ppvhResult)
{
	SG_vhash * pvhResult = NULL;
	SG_varray * pvaActionLog = NULL;
	SG_varray * pvaWarnings = NULL;
	SG_varray * pvaMissingDetails = NULL;

	// pvec_use_submodules is optional

	SG_ERR_CHECK_RETURN(  SG_VHASH__ALLOC(pCtx, &pvhResult)  );

	// TODO 2010/01/20 for historical reasons, the api to this routine
	// TODO            is a little convoluted.  part of this comes from
	// TODO            the throw-model we have.
	//
	// if there are missing submodule issues, this RETURNS (OK, pvaMissingDetails).
	// else if there are portability warnings, this SETS pvaWarnings and THROWS (non standard).
	// else it tries to do the UPDATE and RETURNS the log if requested.
	// else it THROWS any other error.

	SG_vv_verbs__update(pCtx,
						pszFolder, pszTargetChangeset, pszTargetBranchName,
						bVerbose, bTest,
						pvec_use_submodules,
						&pvaActionLog,
						&pvaWarnings,
						&pvaMissingDetails);
	if (SG_context__has_err(pCtx))
	{
		SG_ASSERT(  (pvaActionLog == NULL)  );
		SG_ASSERT(  (pvaMissingDetails == NULL)  );

		if (SG_context__err_equals(pCtx,SG_ERR_PORTABILITY_WARNINGS))
		{
			SG_ASSERT(  (pvaWarnings != NULL)  );

			SG_context__err_reset(pCtx);
			SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvhResult, "status", SG_VV_VERBS__STATUS__PORTABILITY_WARNING)  );
			SG_ERR_CHECK(  SG_vhash__add__varray(pCtx, pvhResult, "port", &pvaWarnings)  );
		}
		else
		{
			SG_ASSERT(  (pvaWarnings == NULL)  );

			SG_ERR_RETHROW;
		}
	}
	else if (pvaMissingDetails)
	{
		SG_ASSERT(  (pvaActionLog == NULL)  );
		SG_ASSERT(  (pvaWarnings == NULL)  );

		SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvhResult, "status", SG_VV_VERBS__STATUS__SUBMODULE_PROBLEM)  );
		SG_ERR_CHECK(  SG_vhash__add__varray(pCtx, pvhResult, "sub", &pvaMissingDetails)  );
	}
	else
	{
		SG_ASSERT(  (pvaMissingDetails == NULL)  );
		SG_ASSERT(  (pvaWarnings == NULL)  );

		SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvhResult, "status", SG_VV_VERBS__STATUS__OK)  );
		if (pvaActionLog)		// should only be set if --test or --verbose
		{
			SG_ERR_CHECK(  SG_vhash__add__varray(pCtx, pvhResult, "log", &pvaActionLog)  );
		}
	}

	*ppvhResult = pvhResult;
	return;

fail:
	SG_VARRAY_NULLFREE(pCtx, pvaActionLog);
	SG_VARRAY_NULLFREE(pCtx, pvaWarnings);
	SG_VARRAY_NULLFREE(pCtx, pvaMissingDetails);
	SG_VHASH_NULLFREE(pCtx, pvhResult);
}

//////////////////////////////////////////////////////////////////

END_EXTERN_C;

#endif//H_SG_VV_VERBS__UPDATE_H
