/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

// Config vars
var sessionLifetime = 60 * 60;
var sessionVersion = 1;

/**
 * vvSession - Session State methods
 *
 * This library provides a simple file based mechanism for persisting data between
 * HTTP requests.  
 *
 * = Usage =
 * Sessions are not automatically created on each request.  In order to request
 * that a session object be created, you must set enableSessions to true inside 
 * your route definition:
 * 
 * <pre>
 *	"/test/cookies/session": {
 *		"GET": {
 *			enableSessions: true,
 *			onDispatch: function (request)
 *			{
 *				var session = vvSession.get();
 *				return OK;
 *			}
 *		}
 *	},
 * </pre>
 *
 * Once you have a session you can store data in the session by adding it like 
 * you would for any other JavaScript object.
 *
 * <pre>
 * 		var session = vvSession.get();
 * 		var data = { key: "value", foo: "bar" };
 * 		session["store_me"] = data;
 * 		session.bar = "some value";
 * 		delete session.bar;
 * </pre>
 *
 * The session will serialize itself to disk at the end of the request.
 **/

var vvSession = (function() {
	var _inited = false;
	var options = {}
	var rootInstance = null;
	
	var init = function()
	{
		var sessionDir = sg.local_settings().machine.server.session_storage;
	
		if (sessionDir)
		{
			if (! sessionDir.match(/[\\\/]$/))
				sessionDir += "/";
		}
		else
		{
			sessionDir = sg.server_state_dir();

			if (! sessionDir.match(/[\\\/]$/))
				sessionDir += "/";

			sessionDir += "vvsession/";
		}
	
		if (! sg.fs.exists(sessionDir))
			sg.fs.mkdir(sessionDir);
		
		options["sessionDir"] = sessionDir;
	}
	
	var start = function(cookies)
	{
		if (!_inited)
			init();
		
		if (!rootInstance)
			rootInstance = new vvSessionInstance(cookies);
			
		return rootInstance;
	};
	
	var get = function()
	{
		if (!rootInstance)
			return null;
			
		return rootInstance;
	}
	
	var setCookies = function(response, instance)
	{
		if (!response.headers)
			response.headers = {};
		
		var session_cookie = {
			name: "vvSession",
			value: instance.hash,
			max_age: sessionLifetime,
			version: sessionVersion,
			path: "/"
		};
			
		if (!response.headers["Set-Cookie"])
		{
			response.headers["Set-Cookie"] = [];
		}
		else
		{
			if (typeof response.headers["Set-Cookie"] != "array")
			{
				var existing = response.headers["Set-Cookie"];
				response.headers["Set-Cookie"] = [ existing ];
			}
		}
		
		response.headers["Set-Cookie"].push(session_cookie);
	}
	
	var save = function(instance)
	{
		var json = sg.to_json(instance);
		var path = sessionPath(instance);
		
		removeOldSessions();
		
		if (!json)
			return;
			
		sg.file.write(path, json);
	}
	
	var reset = function()
	{
		rootInstance = null;
	}
	
	var removeOldSessions = function()
	{
		if (! sg.fs.exists(options.sessionDir))
			return;
		
		var now = new Date();
		var ts_now = now.getTime();
		
		try
		{
			var files = sg.fs.readdir(options.sessionDir);
			
			for ( var i = 0; i < files.length; ++i )
			{
				var file = files[i];
				
				// Remove session files that are older than the session lifetime
				if ((ts_now - (sessionLifetime * 1000)) > file.modtime.getTime())
				{
					var path = options.sessionDir + "/" + file.name;
					try
					{
						if (sg.fs.exists(path))
							sg.fs.remove(path);
					}
					catch(e)
					{
						vv.log("Error removing session file: " + path);
					}
				}
			}
		}
		catch(e)
		{
			// ignore occasional errors trying to stat a file-in-progress
		}
	}
	
	var sessionPath = function(instance)
	{
		return options.sessionDir + sessionFileName(instance);
	}
	
	var sessionFileName = function(instance)
	{
		return instance.hash;
	}
	
	var vvSessionInstance = function(request)
	{
		var _hash = null;
		
		var init = function(instance)
		{
			if (request.cookies['vvSession'])
			{
				_hash = request.cookies['vvSession'];
				instance.hash = _hash;
				load(instance);
			} 
			else
			{
				_hash = sessionHash();
				instance.hash = _hash;
			}
			
			return instance;
		}
		
		var load = function(instance)
		{
			var path = sessionPath(instance);
			
			if (!sg.fs.exists( path ))
			{
				return;
			}
			
			var json = sg.file.read( path );
			
			if (!json)
				return;
				
			var obj = sg.from_json(json);
			
			for(var key in obj)
			{
				if (key == "hash")
					continue;
					
				instance[key] = obj[key];
			}
		};
		
		var sessionHash = function()
		{
			if (_hash)
				return _hash;
				
			var t = (new Date).getTime();
			var r = Math.random();
			var key = "" + t + r;
			var hash = sg.hash(key);
			return "vv_" + hash;
		}
		
		return init(this);
		
		
	};
	
	// export the public API
	return {
		start: start,
		get: get,
		setCookies: setCookies,
		save: save,
		reset: reset
	}
})();
