/*
Copyright 20102011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 * Create and fully initialize a response object based on a template
 * file. All include processing and variable expansion happens here.
 * 
 * @param request request object describing our context
 * @param packageName the package within which this request is being handled
 * @param title the page title to be used
 * @param replacer (optional) keyword replacement function
 *                 overrides the default replacements. The function takes
 *                 two parameters : keyword and request object. It should
 *                 return nothing if it has no match for the keyword; the
 *                 default replacement fills in the gap.
 */
function vvUiTemplateResponse(request, packageName, title, replacer, templateFilename)
{
	this._request = request;
	this._packageName = packageName;
	this._replacer = replacer || false;
	this._title = title;

	this._serverFiles = sg.local_settings().machine.server.files;

	this._coreTemplateDir = this._serverFiles + "/core/templates";

	if (packageName == "core")
		this._templateDir = this._coreTemplateDir;
	else
		this._templateDir = this._serverFiles + "/modules/" + packageName + "/templates";

	this._fn = templateFilename || false;
	this.statusCode = STATUS_CODE__NOT_FOUND;
	this._str = null;

	if (!this._fn)
	{
		var uri = trimSlashes(request.uri);
		var matches = uri.match(/([^\/]+\.html)$/);
		if(matches)
			this._fn = matches[1];
		else
		{
			matches = uri.match(/([^\/]+)$/);
			if(matches)
				this._fn = matches[1]+".html";
		}
	}

	if(this._fn)
	{
		if (this._fn.match(/^([a-z]:)?\//i))
			this._fullPath = this._fn;
		else
			this._fullPath = this._templateDir + "/" + this._fn;

		if (sg.fs.exists(this._fullPath))
		{
			var text = sg.file.read(this._fullPath);

			if (!! text)
			{
				this.statusCode = STATUS_CODE__OK;
				this._str = this.fillTemplate(text);
			}
		}
	}

	if(this._str===null)
		throw "The template file was not found.";

	this._str = sg.to_sz(this._str);
	this.headers = {
		"Content-Length" : this._str.length,
		"Content-Type" : htmlContentTypeForUserAgent(request.headers["User-Agent"])
	};

	return(this);
}

vvUiTemplateResponse.prototype.onChunk = function () {
	var ret = this._str;
	this._str = null;
	return ret;
};
vvUiTemplateResponse.prototype.finalize = function () {
	this._str.fin();
	this._str = null;
};


vvUiTemplateResponse.prototype.expandTemplateFile = function(fn) {
	var tpath = this._templateDir + "/" + fn;
	
	if (! sg.fs.exists(tpath))
		tpath = this._coreTemplateDir + "/" + fn;

	var text = sg.file.read(tpath);
	if (!! text)
		return(this.fillTemplate(text));
	else
		return("");
};


vvUiTemplateResponse.prototype.fillTemplate = function(templateText) {
	var result = "";
	var matches;
	var len = templateText.length;
	var remainder = 0;
	var skipping = 0;

	while (remainder < len)
	{
		var p = templateText.indexOf("{{{", remainder);

		if (p < 0)
		{
			result += templateText.substring(remainder);
			remainder = len;
		}
		else
		{
			if(!skipping)
				result += templateText.substring(remainder, p);

			var endp = templateText.indexOf("}}}", p);

			if (endp < 0)
			{
				result += templateText.substring(p);
				remainder = len;
			}
			else
			{
				var key = templateText.substring(p + 3, endp);
				remainder = endp + 3;

				if ((key.length > 0) && (key.charAt(0) == '<'))
				{
					result += this.expandTemplateFile(key.substring(1));
				}
				else if ((key.length >= 3) && (key[0]=='I') && (key[1]=='F') && ((key[2]==' ')||(key[2]=='(')))
				{
					if(skipping>0)
						skipping = skipping+1;
					else
					{
						var if_statement_result = false;
						try
						{
							if_statement_result = eval(key.slice(2));
						}
						catch(ex)
						{
						}

						if(!if_statement_result)
							skipping = 1;
					}
				}
				else if(key=="ENDIF")
				{
					if(skipping>0)
						skipping = skipping - 1;
				}
				else if(!skipping)
				{
					var replacement = undefined;
					var jsEscape = false;

					matches = key.match(/^js:(.+)/);

					if (matches)
					{
						key = matches[1];
						jsEscape = true;
					}

					if (this._replacer)
						replacement = this._replacer(key, this._request);

					if (replacement===undefined)
						replacement = this.defaultReplacer(key, this._request, jsEscape);
					else
						replacement = this.escape(replacement, jsEscape);

					result += replacement;
				}
			}
		}
	}

	return(result);
};

vvUiTemplateResponse.prototype.escape = function(text, jsEscape)
{
	var esc = text || '';

	if (jsEscape)
	{
		esc = esc.
			replace(/\\/g, "\\\\").
			replace(/"/g, '\\"').
			replace(/'/g, "\\'").
			replace(/\n/g, "\\n").
			replace(/\r/g, "\\r").
			replace(/\t/g, "\\t").
			replace(/\v/g, "\\v");
	}
	else
	{
		esc = esc.
			replace(/\&/g, '&amp;').
			replace(/"/g, '&quot;').
			replace(/'/g, "&#39;").
			replace(/</g, '&lt;').
			replace(/>/g, '&gt;');
	}
	
	return( esc );
};


vvUiTemplateResponse.prototype.scriptTags = function(scriptList, linkRoot)
{
	var base = linkRoot + "/ui/";
	var scripts = scriptList.split(/[ \t\r\n,]+/);
	var results = [];

	if (enableDiagnostics)
	{
		for ( var i = 0; i < scripts.length; ++i )
		{
			results.push("<script type='text/javascript' src='" + base + scripts[i] + "'></script>");
		}

		return(results.join("\r\n"));
	}
	else
	{
		return("<script type='text/javascript' src='"+linkRoot+"/vvscripts.js?s=" + scripts.join(",") + "'></script>");
	}
};

vvUiTemplateResponse.prototype.defaultReplacer = function(key, request, jsEscape) {
	var replacement = "";

	var matches = key.match(/^SCRIPTS[ \t]+(.+?)[ \t]*$/);

	if (matches)
		return(this.scriptTags(matches[1], request.linkRoot));

	switch (key)
	{
		case "REPONAME":
			replacement = this.escape(request.repoName || getDefaultRepoName(request), jsEscape);
			break;

		case "LINKROOT":
			replacement = this.escape(request.linkRoot, jsEscape);
			break;

		case "STATICLINKROOT":
			replacement = this.escape(request.linkRoot, jsEscape) + "/ui";
			break;

		case "USERNAME":
		case "EMAIL":
			replacement = this.escape(request.headers.From, jsEscape);
			break;

		case "LOCALREPO":
			replacement = this.escape(getLocalRepoName(request), jsEscape);
			break;

		case "TITLE":
			replacement = this.escape(this._title, jsEscape);
			break;

		case "MODULEMENUS":
			var mods = [];
			var files = [];

			if (typeof(vvModules) != 'undefined')
				mods = vvModules;

			var base = sg.local_settings().machine.server.files;

			if (base)
			{
				if (! base.match(/[\\\/]$/))
					base += "/";

				for ( var i = 0; i < mods.length; ++i )
				{
					var area = mods[i].area;
					var tfn = base + "ui/modules/" + area + "/menu.js";

					if (sg.fs.exists(tfn))
						files.push("modules/" + area + "/menu.js");
				}

				if (files.length > 0)
				{
					replacement = "<script src='/vvscripts.js?s=" +
						this.escape(files.join(",")) + "'></script>\n";
				}
			}

			break;

		case "CURRENTSPRINT":
		    replacement = "";
		
			if (request.repo)
			{
				var cfg = new Configuration(request.repo);
				var current = cfg.get('current_sprint');

				if (current)
					replacement = this.escape(current.value, jsEscape);
			}
			break;
	}

	return(replacement);
};
