/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/


//
// Call init() from C code in order to configure settings that we wouldn't
// otherwise be able to determine from within JavaScript.
//

var applicationRoot = "";
var enableDiagnostics = false;
var readonlyServer = false;

var init = function(applicationRootParam, readonlyServerParam, enableDiagnosticsParam){
	applicationRoot = applicationRootParam;
	readonlyServer = readonlyServerParam;
	enableDiagnostics = enableDiagnosticsParam;
}


//
// Misc helpers
//

String.prototype.trim = function(chars) {
	var str = this.rtrim(chars);
	str = str.ltrim(chars);
	return str;

}
String.prototype.ltrim = function(chars) {
	chars = chars || "\\s";
	return this.replace(new RegExp("^[" + chars + "]+", "g"), "");
}

String.prototype.rtrim = function(chars) {
	chars = chars || "\\s";
	return this.replace(new RegExp("[" + chars + "]+$", "g"), "");
}

Array.isArray = Array.isArray || function(o) { return Object.prototype.toString.call(o) === '[object Array]'; };

var trimSlashes = function(str) {
	return str.replace(new RegExp("^\\/+", "g"), "").replace(new RegExp("\\/+$", "g"), "");
}

var firstKeyIn = function (obj) {
	for (var key in obj)
		return key;
}


var STATUS_CODE__OK = "200 OK";
var STATUS_CODE__MOVED_PERMANENTLY = "301 Moved Permanently";
var STATUS_CODE__NOT_MODIFIED = "304 Not Modified";
var STATUS_CODE__BAD_REQUEST = "400 Bad Request";
var STATUS_CODE__UNAUTHORIZED = "401 Unauthorized";
var STATUS_CODE__NOT_FOUND = "404 Not Found";
var STATUS_CODE__METHOD_NOT_ALLOWED = "405 Method Not Allowed";
var STATUS_CODE__PRECONDITION_FAILED = "412 Precondition Failed";
var STATUS_CODE__INTERNAL_SERVER_ERROR = "500 Internal Server Error";

var CONTENT_TYPE__ATOM = "application/atom+xml;charset=UTF-8";
var CONTENT_TYPE__FRAGBALL = "application/fragball";
var CONTENT_TYPE__JSON = "application/json;charset=UTF-8";
var CONTENT_TYPE__TEXT = "text/plain;charset=UTF-8";
var CONTENT_TYPE__JAVASCRIPT = "application/javascript;charset=UTF-8";

var CONTENT_TYPE__HTML = "text/html;charset=UTF-8";
var CONTENT_TYPE__XHTML = "application/xhtml+xml;charset=UTF-8";
var htmlContentTypeForUserAgent = function(ua) {
	if(!ua)
		return CONTENT_TYPE__XHTML;

	var m = ua.match(/MSIE\W+\d+/);
	if( m && (m[0].slice(5) < 9))
		return CONTENT_TYPE__HTML;

	return CONTENT_TYPE__XHTML;
}

var chunk_sz = function () {
	var ret = this.sz;
	this.sz = null;
	return ret;
}
var free_sz = function () {
	this.sz.fin();
	this.sz = null;
}

var jsonResponse = function(object, statusCode, headers) {
	var response = {};
	response.onChunk = chunk_sz;
	response.finalize = free_sz;
	response.sz = sg.to_json__sz(object);
	response.statusCode = statusCode || STATUS_CODE__OK;

	response.headers = headers || {};
	response.headers["Content-Length"] = response.sz.length;
	response.headers["Content-Type"] = CONTENT_TYPE__JSON;

	return response;
}

var OK = {
	statusCode: STATUS_CODE__OK,
	headers: {
		"Content-Length": 0
	}
};

var textResponse = function(text) {
	var response = {};
	response.onChunk = chunk_sz;
	response.finalize = free_sz;
	response.sz = sg.to_sz(text);
	response.statusCode = STATUS_CODE__OK;
	response.headers = ({
		"Content-Length": response.sz.length,
		"Content-Type": CONTENT_TYPE__TEXT
	});
	return response;
};

var fileResponse = function(file, contentType, download, filename) {
	var response = {
		statusCode: STATUS_CODE__OK,
		headers: {
			"Content-Length": file.total_length
		},
		onChunk: function () { return file.next_chunk(); },
		finalize: function () { file.abort();  },
		onFinish: function () { file.abort();  }
	};

	if (contentType)
		response.headers["Content-Type"] = contentType;

	if (download) {
		if (filename)
			response.headers["Content-Disposition"] = "attachment; filename=" + filename;
		else
			response.headers["Content-Disposition"] = "attachment";
	}

	return response;
};

// Parameters:
//   statusCode: Http status code (for example "400 Bad Request").
//   request:    The request associated with this error response. Note: This parameter is considered
//               optional, but if it's not provided we'll be forced to return a plaintext response.
//               If provided, we'll choose between plaintext and html based on what was requested.
//   message:    Plaintext string that will be included in the response. Optional.
var errorResponse = function (statusCode, request, message, dontEscapeHtml) {
	statusCode = statusCode || STATUS_CODE__INTERNAL_SERVER_ERROR;
	message = message || statusCode;

	var response;

	if( request
		&& request.requestMethod === "GET"
		&& ( request.uri.slice(request.uri.length-5)===".html" || request.uri.lastIndexOf('.')<=request.uri.lastIndexOf('/') )
	)
	{
		// Html response.
		sg.log('debug', "Constructing html page for \""+statusCode+"\" response to GET of "+request.uri);
		if(!dontEscapeHtml)
			message = "<pre><![CDATA["+message+"]]></pre>";
		response = new vvUiTemplateResponse(
			request,
			"core",
			"veracity",
			function(key){if(key=="ERRORMESSAGE")return message;},
			"error.html");
		response.statusCode = statusCode;
	}
	else {
		// Plaintext response.
		sg.log('debug', "Constructing plaintext \""+statusCode+"\" response.");
		response = {};
		response.statusCode = statusCode;
		response.sz = sg.to_sz(message);
		response.onChunk  = chunk_sz;
		response.finalize = free_sz;
		response.headers  = ({
			"Content-Length": response.sz.length,
			"Content-Type": CONTENT_TYPE__TEXT
		});
	}

	return response;
}

var json500Response = function(request, message)
{
	var statusCode = STATUS_CODE__INTERNAL_SERVER_ERROR;
	sg.log('debug', "Constructing JSON \""+statusCode+"\" response for " + message);

	/* When jsglue gives proper exception objects for sglib errors, this regex hack won't be necessary. */

	/* Error 140 (sglib): There is no working copy here: The path 'C:/Users/iolsen/' is not inside a working copy.
	         [1] [2]      [3]*/
	var reMsg = /Error\s([0-9]{1,7})\s\((\w+)\):\s(.+)/; //
	var match = reMsg.exec(message);

	var objMsg = {};
	objMsg.num = match[1];
	objMsg.src = match[2];
	objMsg.msg = match[3];

	var szMsg = sg.to_json(objMsg);

	sg.log('debug', "JSON \""+statusCode+"\" response: " + szMsg);

	// JSON response
	var response = {};
	response.statusCode = statusCode;
	response.sz = sg.to_sz(szMsg);
	response.onChunk  = chunk_sz;
	response.finalize = free_sz;
	response.headers  = ({
							 "Content-Length": response.sz.length,
							 "Content-Type": CONTENT_TYPE__JSON
						 });

	return response;
};

var badRequestResponse = function (request, message) {
	return errorResponse(STATUS_CODE__BAD_REQUEST, request, message || "Bad request.");
};
var preconditionFailedResponse = function (request, message) {
	return errorResponse(STATUS_CODE__PRECONDITION_FAILED, request, message || "Precondition failed.");
};
var methodNotAllowedResponse = function (request, message, allow) {
	response = errorResponse(STATUS_CODE__METHOD_NOT_ALLOWED, request, message || STATUS_CODE__METHOD_NOT_ALLOWED);
	if (allow)
		response.headers["Allow"] = allow;
	return response;
}
var notFoundResponse = function (request, message) {
	return errorResponse(STATUS_CODE__NOT_FOUND, request, message || "Not found.");
}
var unauthorizedResponse = function (request, message) {
	return errorResponse(STATUS_CODE__UNAUTHORIZED, request, message || "Unauthorized.");
}

var getLocalRepoName = function(request) {
	if (! request.localRepoName)
	{
		var localRepo = false;

		try
		{
			localRepo = sg.open_local_repo();
			if (!! localRepo)
				request.localRepoName = localRepo.name;
			else
				request.localRepoName = "";
		}
		catch (e)
		{
			request.localRepoName = "";
		}

		if (localRepo)
			localRepo.close();
	}

	return(request.localRepoName);
};

var getDefaultRepoName = function (request) {
	var defaultRepoName = getLocalRepoName(request);

	if(defaultRepoName) return defaultRepoName;

	try {
		defaultRepoName = sg.local_settings().machine.server.repos['default'];
	}
	catch (e) {
	}

	if(defaultRepoName) return defaultRepoName;

	defaultRepoName = firstKeyIn(sg.list_descriptors());

	return defaultRepoName;
};


// "routes" is a structure that stores an aggregation of all registered
// routes. It is a hierarchical map of url "parts" (ie substrings when split
// on the slashes) with the empty string, a file extension, and an http verb
// included on the end as the final three "parts". For example, all of the
// following might exist in routes.
//
//     routes.repos."".html.GET                      // to get /repos.html
//     routes.repos."".json.GET                      // to get /repos.json
//     routes.repos."*"."".html.GET                  // to get /repos/myrepo.html
//     routes.repos."*"."".json.GET                  // to get /repos/myrepo.json
//     routes.repos."*".activity."".atom.GET         // to get /repos/myrepo/activity.atom
//     routes.repos."*".history."".html.GET          // to get /repos/myrepo/history.html
//     routes.repos."*".sync."".(undefined).POST     // to post to /repos/myrepo/sync
//     routes.repos."*".sync."*"."".(undefined).GET  // to get /repos/myrepo/sync/syncid
//     routes.repos."*".sync."*"."".(undefined).POST // to post to /repos/myrepo/sync/syncid
//     routes.repos."*".users."".json.GET            // to get /repos/myrepo/users.json
//     ... etc...
//
// Each of these would evaluate to an object containing callback functions to
// handle a request, plus an optional flag or two. Possible members include:
//
//     availableOnReadonlyServer: bool - Can be used to enable or disable a
//         given route for when the server is in readonly mode. Defaults to
//         true for GET routes and false for all others.
//     onDispatch: function(request) - Called after the http request headers
//         have been received.
//     onJsonReceived: function(request, data) - Use this function when you
//         are expecting incoming json. The 'data' parameter will be the
//         object, already interpreted from the json.
//     onFileReceived: function(request, file_path) - Use this function
//         if you want the incoming message body to be stored in a file and
//         then handed to you. The file will appear in the "user temp
//         directory". If the file still exists at the original location when
//         your function returns, it will be deleted.
//     onMimeFileReceived: function(request, path, filename, mimetype) - Similar
//         to onFileReceived().
//     onRequestAborted: function(request) - Called if we are going to cancel
//         the request before it is finished being recieved. Will only be called
//         *instead of* one of the above onReceived functions. It will not be
//         called if an error is produce during or after one of them.
//
// Notice that the request object is passed to each of these functions. If you
// need to store state, to be used by subsequent calls, go ahead and store it
// in the request object.
//
// Any of the functions (except for onRequestAborted), can return a response
// object. If a response object is returned, no further calls will be made to
// any of the request handling functions (even if the request body wasn't
// finished being received).
//
// A response object can contain the following information that the framework
// will use to send the response:
//
//     statusCode: string **REQUIRED**
//     headers: object **REQUIRED** This is a JS object that acts as a key-value
//                                  hash where each key should be a HTTP header
//                                  field name. "Content-Length" is a required
//                                  key in this hash.  The value for a key can be
//                                  the string or integer to send for that header
//                                  or an array of values.  In that case, the
//                                  header will be sent multiple times to the client.
//
//     onChunk: function() **REQUIRED (unless Content-Length is 0)**
//                         Function must return a "cbuffer".
//     onFinish: function() - Called after the last chunk has been sent. If this
//                            function throws an exception or tries to return a
//                            value of any kind it will be ignored.
//     finalize: function() - Called if the chunking of the response is aborted for
//                            any reason. If this function throws an exception or
//                            tries to return a value of any kind it will be ignored.
//                            (Note: an object's finalize function is automatically
//                            called by SpiderMonkey on GC of the object. The reason
//                            it is NOT called if the response is sent successfully
//                            is because SG_uridispatch.c sets the response object's
//                            finalize member to null when it calls onFinish. It
//                            sets it to null right *before* calling onFinish, so if
//                            you wanted the finalize function to be called even
//                            after onFinish, it would technically still be possible
//                            by setting it again in your onFinish function.)
//
//     ...
//     And add your own members for state information needed by subsequent callbacks.

var routes = ({});


//
// registerRoutes() is what registers mappings that match a url pattern to a set of
// handlers for the request.
//
// If a route already exists for a given url, the old route will be overridden.
//

var registerRoutes = function(new_routes) {
	for(var path in new_routes) {
		var route = new_routes[path];
		route.pathParts = trimSlashes(path).split("/");

		var pRoute = routes;
		var suffix = undefined;
		for(var i=0; i<route.pathParts.length; ++i) {
			var lastPart = (i===route.pathParts.length-1);

			var part = route.pathParts[i];

			if(lastPart){
				var dotPos = part.lastIndexOf(".");
				if(dotPos!==-1) {
					suffix = part.slice(dotPos+1);
					part = part.slice(0, dotPos);
				}
			}

			if(part[0]==="<") {
				route.pathParts[i] = part.slice(1,-1);
				part = "*";
			}

			if(!pRoute[part])
				pRoute[part] = ({});
			pRoute = pRoute[part];
		}

		if(!pRoute[""])
			pRoute[""] = ({});
		pRoute = pRoute[""];

		pRoute[suffix] = route;
	}
}


//
// dispatch() is the function that is called for every incoming request to the server
//

var dispatch = function (request) {

	//
	// First add some more members to the request object that the C layer didn't cover.
	//

	// Make sure headers member exists
	request.headers = request.headers || {};

	// Link root
	// (Note: If the server is behind a proxy we need to use the X-Forwarded-Host
	// header to determine the correct linkRoot).
	request.scheme = request.scheme || sg.local_settings().machine.server.scheme || "http:";
	request.hostname = request.headers["X-Forwarded-Host"] || request.headers["Host"] || sg.local_settings().machine.server.hostname

	if(request.hostname)
		request.linkRoot = request.scheme + "//" + request.hostname + applicationRoot;
	else
		request.linkRoot = applicationRoot;

	// HTTP Cookie object
	var cookieObj = {};
	if (request.headers.Cookie) {
		var pairs = request.headers.Cookie.split(";");
		for (var i = 0; i < pairs.length; i++) {
			var parts = pairs[i].split("=");
			if (parts.length == 2) {
				var key = parts[0].trim();
				var val = parts[1].trim();
				cookieObj[key] = val;
			}
		}
	}
	request.cookies = cookieObj;

	// Destructor
	request.close = function() {
		if(this.repo)
			this.repo.close();
	}

	//
	// Next we'll find a route that matches the uri.
	//

	var pathParts = trimSlashes(request.uri).split("/");
	var suffix = undefined;

	var lastPart = pathParts[pathParts.length-1];
	var dotPos = lastPart.lastIndexOf(".");
	if(dotPos!==-1) {
		suffix = lastPart.slice(dotPos+1)
		pathParts[pathParts.length-1] = lastPart.slice(0, dotPos);
	}

	var substitutions = [];
	var pRoute = routes;
	for(var i=0; i<pathParts.length; ++i) {
		if(pRoute[pathParts[i]])
			pRoute = pRoute[pathParts[i]];
		else if(pRoute["*"]) {
			substitutions.push(i);
			pRoute = pRoute["*"];
		}
		else
			return notFoundResponse(request);
	}

	if(pRoute[""])
		pRoute = pRoute[""];
	else
		return notFoundResponse(request);

	var pNext = pRoute[suffix];
	if (!pNext) {
		if (pRoute["$"]) {
			// A ".$" extension means that the route wants the file extension to be included in the last "pathPart" token.
			if (suffix !== undefined)
				pathParts[pathParts.length - 1] = pathParts[pathParts.length - 1] + "." + suffix;
			pNext = pRoute["$"];
		}
		else {
			pNext = pRoute["html"] || pRoute["json"] || pRoute[firstKeyIn(pRoute)];
		}
	}
	pRoute = pNext;

	if(!pRoute)
		return notFoundResponse(request);

	var pathPartNames = pRoute.pathParts;

	if(!pRoute[request.requestMethod])
	{
		var allow = "";
		for(var allowedMethod in pRoute)
		{
			if(allowedMethod!=="pathParts")
				allow += allowedMethod + " ";
			if(allowedMethod==="GET")
				allow += "HEAD ";
		}
		return methodNotAllowedResponse(request, null, allow.rtrim(" "));
	}
	else
	{
	    pRoute = pRoute[request.requestMethod];

        // We now have the route. See if we need to reject the request it based on readonly server.
		if(readonlyServer)
		{
			if(request.requestMethod==="GET" && pRoute.availableOnReadonlyServer===false)
				return notFoundResponse(request, "Resource not available when server is in readonly mode.");
			if(request.requestMethod!=="GET" && pRoute.availableOnReadonlyServer!==true)
				return methodNotAllowedResponse(request, "Operation not allowed when server is in readonly mode.");
		}

        // Now see if we need to reject the request based on they needed a From: header for POST/PUT.
	}

	// Create a member in the request for each fill-in-the-blank from the path that we matched it to
	for (var i = 0; i < substitutions.length; ++i) {
		request[pathPartNames[substitutions[i]]] = pathParts[substitutions[i]];
		if(pathPartNames[substitutions[i]]=="repoName")
		{
			try
			{
				var uid = request.headers.From || null;

				request["repo"] = sg.open_repo(request["repoName"], uid);
			}
			catch(ex)
			{
				if(ex.message.indexOf("Error 10 (sglib)")>=0) // SG_ERR_NOTAREPOSITORY
				{
					return notFoundResponse(request, "The repository '"+request["repoName"]+"' does not exist on this server or is not available.");
				}
				else
				{
					return response_for_exception_caught_by_dispatcher(ex, "sg.open_repo", request);
				}
			}
		}
	}

	if (pRoute.enableSessions)
		vvSession.start(request); // Initialize vvSession for this client.

	//
	// We've mapped the uri (+ request method) to a route, which should provide
	// us with all the callback functions we need to construct a repsonse.
	//
	// First thing we'll do is call the onDispatch function, if they provided one,
	// and see if it can produce a response based on the http headers alone.
	//

	if(pRoute.onDispatch!=undefined)
	{
		try
		{
			var response = pRoute.onDispatch(request);

			if(response!==undefined && response!==null)
			{
				verify_response_object(response);
				return response;
			}
		}
		catch (ex)
		{
			return response_for_exception_caught_by_dispatcher(ex, "onDispatch callback", request);
		}
	}

	//
	// A response has not yet been generated based on the request headers.
	// We now construct a request-handler object with the route's callback
	// functions to handle the rest of the request. Then we'll be able to
	// form a response after receiving the message body.
	//

	var requestHandler = {request: request};
	if(pRoute.onJsonReceived) {
		requestHandler.onJsonReceived_real = pRoute.onJsonReceived;
		requestHandler.onJsonReceived = function(request, data) {
			try
			{
				var response = this.onJsonReceived_real(request, data);
				verify_response_object(response);
				return response;
			}
			catch (ex)
			{
				return response_for_exception_caught_by_dispatcher(ex, "onJsonReceived callback", request);
			}
		};
		requestHandler.onJsonReceived_ = true;
	}
	if(pRoute.onFileReceived) {
		requestHandler.onFileReceived_real = pRoute.onFileReceived;
		requestHandler.onFileReceived = function(request, file_path) {
			try
			{
				var response = this.onFileReceived_real(request, file_path);
				verify_response_object(response);
				return response;
			}
			catch (ex)
			{
				return response_for_exception_caught_by_dispatcher(ex, "onFileReceived callback", request);
			}
		};
		requestHandler.onFileReceived_ = true;
	}
	if(pRoute.onMimeFileReceived) {
		requestHandler.onMimeFileReceived_real = pRoute.onMimeFileReceived;
		requestHandler.onMimeFileReceived = function(request, file_path, filename, mimetype) {
			try
			{
				var response = this.onMimeFileReceived_real(request, file_path, filename, mimetype);
				verify_response_object(response);
				return response;
			}
			catch (ex)
			{
				return response_for_exception_caught_by_dispatcher(ex, "onMimeFileReceived callback", request);
			}
		};
		requestHandler.onMimeFileReceived_ = true;
	}
	return requestHandler;
}

// Helper function for dispatch(). Simply throws and exception if an object is not a valid "repsonse object".
var verify_response_object = function(obj) {
	// undefined and null are considered valid:
	// - When returned by an onDispatch() callback, this means the actual response will be generated later.
	// - When returned by an onReceived() callback, this implies the request was processed with no errors, and
	//   the C layer can just return a "200 OK" response with no message body.
	if(obj!==undefined && obj!==null)
	{
		if(!obj.statusCode)
			throw "Callback returned an object without a status code";
		if(!(obj.headers["Content-Length"]>=0))
			throw "Callback returned an object with missing or invalid Content-Length.";
	}
}

// Helper function for dispatch(). Creates an error response based on an exception that was caught.
var response_for_exception_caught_by_dispatcher = function(ex, functionName, request) {
	var is_response = false;
	try
	{
		verify_response_object(ex);
		is_response = true;
	}
	catch(ex2)
	{
	}

	if(is_response)
	{
		// An actual response object was thrown. Just return it as THE response.
		return ex;
	}
	else
	{
		// Otherwise construct a response for the exception.
		sg.log('urgent', functionName + " threw exception during request to "+ request.uri +":\n" + ex.toString());
		if(enableDiagnostics)
			return errorResponse(STATUS_CODE__INTERNAL_SERVER_ERROR, request, ex.toString());
		else
			return errorResponse(STATUS_CODE__INTERNAL_SERVER_ERROR, request, "Internal Server Error. Details about this error have been logged to the server's log file.");
	}
}

// This function is called from C code when the response is all ready to be sent.
// All it does is save the current session state, and insert the correct headers
// into the response for the session.
var onBeforeSendResponse = function(response) {
	// (The C code ignores all output and errors generated by this function,
	// so we'll wrap in a try/catch and log any exception.)
	var session = null;
	try
	{
		if( response && (response.statusCode==STATUS_CODE__OK) )
		{
			var session = vvSession.get();
			if (session)
			{
				vvSession.setCookies(response, session);
				vvSession.save(session);
			}
			formatCookies(response);
		}
	}
	catch(ex)
	{
		sg.log('urgent', "Exception caught in onBeforeSendResponse: " + ex.toString());
	}
	finally
	{
		if (session)
			vvSession.reset();
	}
}

function formatCookies(response)
{
	// format cookies
	if (response.headers && response.headers['Set-Cookie'])
	{
		var toFormat = response.headers['Set-Cookie'];

		if (Array.isArray(toFormat))
		{
			var cookies = [];
			for (var i = 0; i < toFormat.length; i++)
			{
				var cookie = toFormat[i];
				if (typeof cookie == "string")
				{
					cookies.push(cookie);
				}
				else
				{
					var formatted = formatCookieObject(cookie);
					if (formatted != "")
						cookies.push(formatted);
				}
			}

			response.headers['Set-Cookie'] = cookies;
		}
		else if (typeof toFormat == "object")
		{
			response.headers['Set-Cookie'] = formatCookieObject(toFormat);
		}
	}
}

/**
 * Take a cookie object and convert it to a string that is valid to be used in a
 * Set-Cookie HTTP header
 *
 * The cookie object can be and JS object with following keys:
 *
 * name (required) -	The name of the cookie
 *
 * value (required) -	The value to be set in the cookie.  This my be a  string.
 * 						If it is not a string it will be saved in the cookie as
 * 						whatever value toString() returns for that value.
 *
 * expires (optional) -	The date at which the cookie should expire.  This can
 * 						be a Javascript Date object or a string that represents
 * 						a valid expiration date for a HTTP cookie.
 *
 * max_age (optional) -	The number of seconds in the future the cookie should
 * 						expire.
 *
 * path (optional) -	The server path that this cookie should be returned for.
 *
 * domain (optional) -	The domain this cookie should be valid for.
 *
 * version (optional) - The version number for the cookie.
 *
 * secure (optional) - 	A boolean that tells the client to only send to cookie
 * 						back to the server when using a secure connection.
 *
 * httponly (optional) - A boolean that tells the client that the cookie should
 * 						 only be used over http connections and should not be
 * 						 visible to client side scripts.
 *
 * = Cookie Lifespan =
 * It should be noted that if both expires and max_age are excluded the cookie
 * will be considered a session cookie and will expire when the user closes their
 * browser window.
 *
 **/
function formatCookieObject(obj)
{
	// If the cookie object is a string, assume the caller has already formatted
	// the cookie string properly
	if (typeof obj == "string")
		return obj;

	//  Name and Value are required
	if (!obj.name || !obj.value)
		return "";

	var parts = [];
	var base = obj.name + "=" + obj.value;
	parts.push(base);

	if (obj.expires)
	{
		if (obj.expires.getDate)
		{
			parts.push("Expires=" + obj.expires.strftime("%a, %d-%b-%Y %T %Z"));
		}
		else
		{
			parts.push("Expires=" + obj.expires);

		}
	}

	if (obj.max_age)
	{
		parts.push("Max-Age=" + obj.max_age);
	}

	if (obj.path)
	{
		parts.push("path=" + obj.path);
	}

	if (obj.domain)
	{
		parts.push("domain=" + obj.domain);
	}

	if (obj.version)
	{
		parts.push("Version=" + obj.version);
	}

	if (obj.secure)
	{
		parts.push("Secure");
	}

	if (obj.httponly)
	{
		parts.push("httponly");
	}

	return parts.join("; ");
}
