/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include "precompiled.h"
#include "vvCommitCommand.h"

#include "vvContext.h"
#include "vvCommitDialog.h"
#include "vvVerbs.h"


/*
**
** Globals
**
*/

/**
 * Global instance of vvCommitCommand.
 */
static const vvCommitCommand gcCommitCommand;

/*
 * Command specification.
 * These need to be char* instead of wxString, because they're used by a static constructor.
 * And we can't be sure the wxStrings will be constructed before the static instance that uses them.
 */
static const char*        gszCommandName        = "commit";
static const unsigned int guCommandFlags        = 0u;
static const char*        gszCommandDescription = "Commit a working folder to its repository.";


/*
**
** Public Functions
**
*/

vvCommitCommand::vvCommitCommand()
	: vvCommand(gszCommandName, guCommandFlags, gszCommandDescription)
	, mwDialog(NULL)
{
}

bool vvCommitCommand::ParseArguments(
	const wxArrayString& cArguments
	)
{
	// make sure we got at least one path
	if (cArguments.Count() < 1u)
	{
		wxLogError("No path(s) specified to commit from.");
		return false;
	}

	// store the given paths
	for (wxArrayString::const_iterator it = cArguments.begin(); it != cArguments.end(); ++it)
	{
		this->mcAbsolutePaths.Add(*it);
	}

	// success
	return true;
}

bool vvCommitCommand::Init()
{
	vvVerbs*   pVerbs   = this->GetVerbs();
	vvContext* pContext = this->GetContext();

	wxASSERT(pVerbs != NULL);
	wxASSERT(pContext != NULL);

	// find the working folder containing our paths
	wxString sFolder = wxEmptyString;
	if (!pVerbs->FindWorkingFolder(*pContext, this->mcAbsolutePaths, &sFolder, NULL))
	{
		pContext->Error_ResetReport("Cannot find single working folder for the given set of paths.");
		return false;
	}
	wxASSERT(!sFolder.IsEmpty());
	this->msFolder = sFolder;

	// convert our absolute paths to repo paths
	if (!pVerbs->GetRepoPaths(*pContext, this->mcAbsolutePaths, this->msFolder, this->mcRepoPaths))
	{
		pContext->Error_ResetReport("Failed to translate the given paths into repo paths relative to working folder: %s", this->msFolder);
		return false;
	}
	wxASSERT(this->mcAbsolutePaths.size() == this->mcRepoPaths.size());

	return true;
}

vvDialog* vvCommitCommand::CreateDialog(
	wxWindow* wParent
	)
{
	vvVerbs*   pVerbs   = this->GetVerbs();
	vvContext* pContext = this->GetContext();

	wxASSERT(pVerbs != NULL);
	wxASSERT(pContext != NULL);

	this->mwDialog = new vvCommitDialog();
	if (!this->mwDialog->Create(wParent, this->msFolder, *pVerbs, *pContext, &this->mcRepoPaths))
	{
		delete this->mwDialog;
		return NULL;
	}
	return this->mwDialog;
}

bool vvCommitCommand::Execute()
{
	// get data from the dialog
	bool                 bAll   = false;
	const wxArrayString& cFiles = this->mwDialog->GetSelectedFiles(&bAll);
	wxString             sUser  = this->mwDialog->GetSelectedUser();

	// build appropriate flags
	vvFlags cFlags;
	if (!this->mwDialog->GetCommitSubmodules())
	{
		cFlags.AddFlags(vvVerbs::COMMIT_NO_SUBMODULES);
	}

	// run the commit
	wxString sHid;
	return this->GetVerbs()->Commit(
		*this->GetContext(),
		this->msFolder,
		this->mwDialog->GetComment(),
		!bAll ? &cFiles : NULL,
		&this->mwDialog->GetSelectedStamps(),
		!sUser.IsEmpty() ? &sUser : NULL,
		cFlags,
		&sHid
		);
}
