/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include "precompiled.h"
#include "vvCommitDialog.h"

#include "vvContext.h"
#include "vvDialogHeaderControl.h"
#include "vvStampsControl.h"
#include "vvStatusControl.h"
#include "vvValidatorEmptyConstraint.h"
#include "vvVerbs.h"


/*
**
** Globals
**
*/

// style for the status control
const unsigned int guStatusControlStyle = vvStatusControl::STYLE_SELECT;

// style for the stamps control
const unsigned int guStampsControlStyle = vvStampsControl::STYLE_CHECKABLE;

/**
 * Suffixes for tab names.
 */
static const wxString gsTabSuffix_Clean  = "";
static const wxString gsTabSuffix_Dirty  = " (*)";
static const wxString gsTabSuffix_Number = " (%d)";
static const wxString gsTabSuffix_Subset = " (%d/%d)";

/*
 * Hints for various text box controls.
 */
static const wxString gsHint_Comment    = "Enter a comment to associate with the new changeset.";
static const wxString gsHint_Stamps_New = "New stamp name";

/**
 * String displayed at the top of the validation error message box.
 */
static const wxString gsValidationErrorMessage = "There was an error in the information specified.  Please correct the following problems and try again.";

/*
 * Strings displayed as field labels when validation errors occur.
 */
static const wxString gsValidationField_Commit_Comment = "Comment";
static const wxString gsValidationField_Files          = "Files";
static const wxString gsValidationField_User           = "User";


/*
**
** Internal Functions
**
*/

/**
 * Checks if a path list contains a single entry that indicates the repo root.
 * Returns true if this is the case, or false otherwise.
 */
bool _ListContainsOnlyRepoRoot(
	const wxArrayString& cPaths //< [in] The list of paths to check.
	)
{
	// if we don't have exactly one entry, no match
	if (cPaths.size() != 1u)
	{
		return false;
	}

	// if the one entry exactly equals the repo root, match
	const wxString& cPath = cPaths.front();
	if (cPath == vvVerbs::sszRepoRoot)
	{
		return true;
	}

	// if the entry doesn't have one more character than the repo root, no match
	size_t uLength = strlen(vvVerbs::sszRepoRoot);
	if (cPath.Length() != uLength + 1u)
	{
		return false;
	}

	// if the entry's extra character isn't a patch separator, no match
	if (wxFileName::GetPathSeparators().Index(cPath[uLength]) == 0)
	{
		return false;
	}

	// must be one entry with a trailing path separator, match
	return true;
}


/*
**
** Public Functions
**
*/

vvCommitDialog::vvCommitDialog()
	: msDirectory(wxEmptyString)
	, mbSubmodules(true)
	, mbAllFiles(true)
	, mwComment(NULL)
	, mwFiles(NULL)
	, mwStamps(NULL)
	, mwStamps_New(NULL)
	, mwStamps_Add(NULL)
	, mwUser(NULL)
	, mwSubmodules(NULL)
{
}

vvCommitDialog::~vvCommitDialog()
{
	if (this->mwFiles != NULL)
	{
		delete this->mwFiles;
	}
}

bool vvCommitDialog::Create(
	wxWindow*            wParent,
	const wxString&      sDirectory,
	vvVerbs&             cVerbs,
	vvContext&           cContext,
	const wxArrayString* pFiles,
	const wxArrayString* pStamps,
	bool                 bSubmodules,
	const wxString&      sUser,
	const wxString&      sComment
	)
{
	// create the dialog window
	if (!vvDialog::Create(wParent, &cVerbs, &cContext, this->GetClassInfo()->GetClassName()))
	{
		return false;
	}

	// look up the repo associated with the given working directory
	if (!cVerbs.FindWorkingFolder(cContext, sDirectory, NULL, &this->msRepo))
	{
		cContext.Error_ResetReport("Cannot find the working folder for path: %s", sDirectory);
		return false;
	}

	// store the given initialization data
	this->msDirectory  = sDirectory;
	this->msComment    = sComment;
	this->msUser       = sUser;
	this->mbSubmodules = bSubmodules;
	if (pStamps != NULL)
	{
		this->mcStamps.assign(pStamps->begin(), pStamps->end());
	}
	if (pFiles == NULL || _ListContainsOnlyRepoRoot(*pFiles))
	{
		this->mbAllFiles = true;
	}
	else
	{
		this->mbAllFiles = false;
		this->mcFiles.assign(pFiles->begin(), pFiles->end());
	}

	// get all of our widgets
	this->mwComment       = XRCCTRL(*this, "wComment",       wxTextCtrl);
	this->mwNotebook      = XRCCTRL(*this, "wNotebook",      wxNotebook);
	this->mwStamps_New    = XRCCTRL(*this, "wStamps_New",    wxTextCtrl);
	this->mwStamps_Add    = XRCCTRL(*this, "wStamps_Add",    wxButton);
	this->mwUser          = XRCCTRL(*this, "wUser",          wxChoice);
	this->mwUser_Override = XRCCTRL(*this, "wUser_Override", wxCheckBox);
	this->mwSubmodules    = XRCCTRL(*this, "wSubmodules",    wxCheckBox);

	// create our header control
	wxXmlResource::Get()->AttachUnknownControl("wHeader", new vvDialogHeaderControl(this, wxID_ANY, this->msDirectory, cVerbs, cContext));

	// create our status control
	this->mwFiles = new vvStatusControl(this, wxID_ANY, wxDefaultPosition, wxDefaultSize, guStatusControlStyle);
	this->mwFiles->SetDisplay(vvStatusControl::MODE_COLLAPSED, vvStatusControl::ITEM_DEFAULT, vvStatusControl::STATUS_DEFAULT);
	wxXmlResource::Get()->AttachUnknownControl("wFiles", this->mwFiles);

	// create our stamps control
	this->mwStamps = new vvStampsControl(this, wxID_ANY, wxDefaultPosition, wxDefaultSize, guStampsControlStyle);
	wxXmlResource::Get()->AttachUnknownControl("wStamps", this->mwStamps);

	// configure controls
	this->mwComment->SetHint(gsHint_Comment);
	this->mwStamps_New->SetHint(gsHint_Stamps_New);

	// get the names of the various tabs
	wxASSERT(this->mwNotebook->GetPageCount() == TAB_COUNT);
	mcTabNames.resize(TAB_COUNT);
	mcTabNames[TAB_FILES]      = this->mwNotebook->GetPageText(TAB_FILES);
	mcTabNames[TAB_STAMPS]     = this->mwNotebook->GetPageText(TAB_STAMPS);
	mcTabNames[TAB_WORK_ITEMS] = this->mwNotebook->GetPageText(TAB_WORK_ITEMS);
	mcTabNames[TAB_ADVANCED]   = this->mwNotebook->GetPageText(TAB_ADVANCED);

	// setup validators
	this->mcValidatorReporter.SetHelpMessage(gsValidationErrorMessage);
	this->mwComment->SetValidator(vvValidator(gsValidationField_Commit_Comment)
		.SetReporter(&this->mcValidatorReporter)
		.AddConstraint(vvValidatorEmptyConstraint(false))
	);
	this->mwFiles->SetValidator(vvValidator(gsValidationField_Files)
		.SetReporter(&this->mcValidatorReporter)
		.AddConstraint(vvValidatorEmptyConstraint(false))
	);
	this->mwUser->SetValidator(vvValidator(gsValidationField_User)
		.SetReporter(&this->mcValidatorReporter)
		.AddConstraint(vvValidatorEmptyConstraint(false))
	);

	// HACK: update the splitter's sash position to force it to re-layout the notebook tabs
	//       otherwise for some reason the status control isn't expanded to fill its tab
	wxSplitterWindow* wSplitter = XRCCTRL(*this, "wSplitter", wxSplitterWindow);
	int iOldPosition = wSplitter->GetSashPosition();
	wSplitter->SetSashPosition(iOldPosition + 1);
	wSplitter->SetSashPosition(iOldPosition);

	// focus on the comment control
	this->mwComment->SetFocus();

	// success
	return true;
}

bool vvCommitDialog::TransferDataToWindow()
{
	// ----- COMMENT -----

	// set the comment
	this->mwComment->SetValue(this->msComment);

	// retrieve status data and populate the file list
	if (!this->mwFiles->FillData(*this->GetVerbs(), *this->GetContext(), this->msDirectory))
	{
		this->GetContext()->Error_ResetReport("Unable to populate status data.");
	}

	// ----- FILES -----

	// expand all of the files
	this->mwFiles->ExpandAll();

	// select the appropriate files
	if (this->mbAllFiles)
	{
		this->mwFiles->SetItemSelectionAll(true);
	}
	else if (this->mwFiles->SetItemSelectionByPath(this->mcFiles, true) == 0u)
	{
		this->mwFiles->SetItemSelectionAll(true);
	}

	// ----- STAMPS -----

	// populate the stamps control
	this->mwStamps->AddItems(*this->GetVerbs(), *this->GetContext(), this->msRepo);

	// pre-check any items in our list
	for (wxArrayString::const_iterator it = this->mcStamps.begin(); it != this->mcStamps.end(); ++it)
	{
		wxDataViewItem cItem = this->mwStamps->GetItemByName(*it);
		if (cItem.IsOk())
		{
			this->mwStamps->CheckItem(cItem);
		}
	}

	// ----- USER -----

	// decide if we're overriding the current user
	if (this->msUser.IsEmpty())
	{
		this->mwUser_Override->SetValue(false);

		// temporarily store the current user in this->msUser, for pre-selection purposes
		// we'll blank it out again after we populate the choice control
		if (!this->GetVerbs()->GetCurrentUser(*this->GetContext(), this->msRepo, this->msUser))
		{
			this->GetContext()->Error_ResetReport("Unable to retrieve current user.");
		}
	}
	else
	{
		this->mwUser_Override->SetValue(true);
	}

	// get the list of users and populate the choice control
	vvVerbs::stlUserDataList cUsers;
	if (!this->GetVerbs()->GetUsers(*this->GetContext(), this->msRepo, cUsers))
	{
		this->GetContext()->Error_ResetReport("Unable to retrieve user list.");
	}
	this->mwUser->Clear();
	for (vvVerbs::stlUserDataList::const_iterator it = cUsers.begin(); it != cUsers.end(); ++it)
	{
		int iIndex = this->mwUser->Append(it->sName);

		// select this user, if it's the one we currently have stored
		if (it->sName == this->msUser)
		{
			this->mwUser->SetSelection(iIndex);
		}
	}
	if (cUsers.empty())
	{
		this->GetContext()->Log_ReportError("No users exist in the repository, you won't be able to commit.");
	}

	// blank out the current user if we only set it temporarily
	if (!this->mwUser_Override->GetValue())
	{
		this->msUser.Clear();
	}

	// ----- FLAGS -----

	// set the state of various options
	this->mwSubmodules->SetValue(this->mbSubmodules);

	// ----- END -----

	// update our display to reflect any widget changes
	this->UpdateDisplay();

	// success
	return true;
}

bool vvCommitDialog::TransferDataFromWindow()
{
	// get the comment
	this->msComment = this->mwComment->GetValue();

	// get the selected files
	vvStatusControl::stlItemDataList cStatusItems;
	this->mwFiles->GetItemData(&cStatusItems, vvStatusControl::COLUMN_FILENAME);
	this->mcFiles.clear();
	this->mbAllFiles = true;
	for (vvStatusControl::stlItemDataList::const_iterator it = cStatusItems.begin(); it != cStatusItems.end(); ++it)
	{
		if (it->bSelected)
		{
			this->mcFiles.Add(it->sRepoPath);
		}
		else
		{
			this->mbAllFiles = false;
		}
	}

	// get the checked stamps
	this->mcStamps.clear();
	vvStampsControl::stlItemDataList cStampItems;
	this->mwStamps->GetItemData(&cStampItems, true);
	for (vvStampsControl::stlItemDataList::const_iterator it = cStampItems.begin(); it != cStampItems.end(); ++it)
	{
		// we asked for only checked items, so they'd better all be checked
		wxASSERT(it->bChecked);
		this->mcStamps.Add(it->sName);
	}

	// get the selected user
	if (this->mwUser_Override->GetValue())
	{
		this->msUser = this->mwUser->GetStringSelection();
	}
	else
	{
		this->msUser = wxEmptyString;
	}

	// get the various option values
	this->mbSubmodules = this->mwSubmodules->GetValue();

	return true;
}

wxString vvCommitDialog::GetComment() const
{
	return this->msComment;
}

wxString vvCommitDialog::GetSelectedUser() const
{
	return this->msUser;
}

const wxArrayString& vvCommitDialog::GetSelectedStamps() const
{
	return this->mcStamps;
}

bool vvCommitDialog::GetCommitSubmodules() const
{
	return this->mbSubmodules;
}

const wxArrayString& vvCommitDialog::GetSelectedFiles(
	bool* pAll
	) const
{
	if (pAll != NULL)
	{
		*pAll = this->mbAllFiles;
	}
	return this->mcFiles;
}

void vvCommitDialog::UpdateDisplay()
{
	// enable/disable the user selection control appropriately
	this->mwUser->Enable(this->mwUser_Override->GetValue());

	// update the Files tab's suffix
	unsigned int uFileCount         = this->mwFiles->GetItemData(NULL, 0u);
	unsigned int uSelectedFileCount = this->mwFiles->GetItemData(NULL, 0u, true);
	this->SetTabSuffix(TAB_FILES, wxString::Format(gsTabSuffix_Subset, uSelectedFileCount, uFileCount));

	// update the Stamps tab's suffix
	unsigned int uStampCount = this->mwStamps->GetItemData(NULL, true);
	this->SetTabSuffix(TAB_STAMPS, wxString::Format(gsTabSuffix_Number, uStampCount));

	// update the Work Item tab's suffix
	// TODO: depends how work items end up being implemented
	this->SetTabSuffix(TAB_WORK_ITEMS, wxString::Format(gsTabSuffix_Number, 0u));

	// update the Advanced tab's suffix
	bool bAdvancedDirty = this->mwUser_Override->GetValue() || !this->mwSubmodules->GetValue();
	this->SetTabSuffix(TAB_ADVANCED, bAdvancedDirty ? gsTabSuffix_Dirty : gsTabSuffix_Clean);

	// relayout our controls and fit the window in case child sizes change
	this->Layout();
	this->Fit();
}

void vvCommitDialog::OnUpdateEvent(
	wxDataViewEvent& WXUNUSED(cEvent)
	)
{
	this->UpdateDisplay();
}

void vvCommitDialog::OnUpdateEvent(
	wxCommandEvent& WXUNUSED(cEvent)
	)
{
	this->UpdateDisplay();
}

void vvCommitDialog::OnAddStamp(
	wxCommandEvent& WXUNUSED(cEvent)
	)
{
	// if they didn't type in a stamp name, don't bother
	wxString sStamp = this->mwStamps_New->GetValue();
	if (sStamp.IsEmpty())
	{
		return;
	}

	// check if the item is in the list already
	wxDataViewItem cItem = this->mwStamps->GetItemByName(sStamp);
	if (cItem.IsOk())
	{
		// in the list already, check it
		this->mwStamps->CheckItem(cItem);
	}
	else
	{
		// not in the list yet, add it
		cItem = this->mwStamps->AddItem(sStamp, 0u, true);
	}

	// select and scroll to the item
	this->mwStamps->EnsureVisible(cItem);
	this->mwStamps->Select(cItem);

	// clear the new stamp text control
	this->mwStamps_New->SetValue(wxEmptyString);

	// update the display accordingly
	this->UpdateDisplay();
}

void vvCommitDialog::SetTabSuffix(
	NotebookTab     eTab,
	const wxString& sSuffix
	)
{
	this->mwNotebook->SetPageText(eTab, this->mcTabNames[eTab] + sSuffix);
}

IMPLEMENT_DYNAMIC_CLASS(vvCommitDialog, vvDialog);

BEGIN_EVENT_TABLE(vvCommitDialog, vvDialog)
	EVT_DATAVIEW_ITEM_VALUE_CHANGED(XRCID("wFiles"),         OnUpdateEvent)
	EVT_DATAVIEW_ITEM_VALUE_CHANGED(XRCID("wStamps"),        OnUpdateEvent)
	EVT_BUTTON(                     XRCID("wStamps_Add"),    OnAddStamp)
	EVT_TEXT_ENTER(                 XRCID("wStamps_New"),    OnAddStamp)
	EVT_CHECKBOX(                   XRCID("wUser_Override"), OnUpdateEvent)
	EVT_CHECKBOX(                   XRCID("wSubmodules"),    OnUpdateEvent)
END_EVENT_TABLE()
