/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include "precompiled.h"
#include "vvCreateCommand.h"

#include "vvContext.h"
#include "vvCreateDialog.h"
#include "vvVerbs.h"


/*
**
** Globals
**
*/

/**
 * Global instance of vvCreateCommand.
 */
static const vvCreateCommand gcCreateCommand;

/*
 * Command specification.
 * These need to be char* instead of wxString, because they're used by a static constructor.
 * And we can't be sure the wxStrings will be constructed before the static instance that uses them.
 */
static const char*        gszCommandName        = "create";
static const unsigned int guCommandFlags        = 0;
static const char*        gszCommandDescription = "Create a working folder from a new or existing repo.";


/*
**
** Public Functions
**
*/

vvCreateCommand::vvCreateCommand()
	: vvCommand(gszCommandName, guCommandFlags, gszCommandDescription)
	, mwDialog(NULL)
{
}

bool vvCreateCommand::ParseArguments(
	const wxArrayString& cArguments
	)
{
	// get the directory name from the first argument
	if (cArguments.Count() < 1u)
	{
		wxLogError("No directory specified to create a working folder in.");
		return false;
	}
	wxString sDirectory = cArguments[0];

	// make sure this is actually an existing directory
	if (!wxDirExists(sDirectory))
	{
		wxLogError("Specified directory doesn't exist: %s", sDirectory);
		return false;
	}

	// make sure the directory is empty
	wxDir cDirectory(sDirectory);
	if (!cDirectory.IsOpened())
	{
		wxLogError("Cannot open directory for traversal: %s", sDirectory);
		return false;
	}
	if (cDirectory.HasFiles() || cDirectory.HasSubDirs())
	{
		wxLogError("The specified directory is not empty: %s", sDirectory);
		return false;
	}

	// success
	this->msDirectory = sDirectory;
	return true;
}

vvDialog* vvCreateCommand::CreateDialog(
	wxWindow* wParent
	)
{
	vvVerbs*   pVerbs   = this->GetVerbs();
	vvContext* pContext = this->GetContext();

	wxASSERT(pVerbs != NULL);
	wxASSERT(pContext != NULL);

	this->mwDialog = new vvCreateDialog();
	if (!this->mwDialog->Create(wParent, this->msDirectory, *pVerbs, *pContext))
	{
		delete this->mwDialog;
		return NULL;
	}
	return this->mwDialog;
}

bool vvCreateCommand::Execute()
{
	vvVerbs*   pVerbs   = this->GetVerbs();
	vvContext* pContext = this->GetContext();

	wxASSERT(pVerbs != NULL);
	wxASSERT(pContext != NULL);

	pContext->Log_PushOperation("Creating working folder", vvContext::LOG_FLAG_NONE);
	pContext->Log_SetSteps(2u);
	wxScopeGuard cPopOperation = wxMakeGuard(&vvContext::Log_PopOperation_Static, pContext);
	wxUnusedVar(cPopOperation);

	switch (this->mwDialog->GetCloneAction())
	{
	case vvCreateDialog::CLONE_EMPTY:
		if (!pVerbs->New(*pContext, this->mwDialog->GetRepo(), this->msDirectory))
		{
			pContext->Error_ResetReport("Unable to initialize new repository.");
			return false;
		}
		break;
	case vvCreateDialog::CLONE_LOCAL:
	case vvCreateDialog::CLONE_REMOTE:
		if (!pVerbs->Clone(*pContext, this->mwDialog->GetCloneSource(), this->mwDialog->GetRepo()))
		{
			pContext->Error_ResetReport("Unable to clone repository: %s", this->mwDialog->GetCloneSource());
			return false;
		}
		break;
	}
	pContext->Log_FinishStep();

	if (this->mwDialog->GetCloneAction() != vvCreateDialog::CLONE_EMPTY)
	{
		switch (this->mwDialog->GetCheckoutType())
		{
		case vvCreateDialog::CHECKOUT_HEAD:
			if (!pVerbs->Checkout(*pContext, this->mwDialog->GetRepo(), this->msDirectory, vvVerbs::CHECKOUT_HEAD, wxEmptyString))
			{
				pContext->Error_ResetReport("Unable to checkout head of repository.");
				return false;
			}
			break;
		case vvCreateDialog::CHECKOUT_BRANCH:
			if (!pVerbs->Checkout(*pContext, this->mwDialog->GetRepo(), this->msDirectory, vvVerbs::CHECKOUT_BRANCH, this->mwDialog->GetCheckoutTarget()))
			{
				pContext->Error_ResetReport("Unable to checkout branch: %s", this->mwDialog->GetCheckoutTarget());
				return false;
			}
			break;
		case vvCreateDialog::CHECKOUT_CHANGESET:
			if (!pVerbs->Checkout(*pContext, this->mwDialog->GetRepo(), this->msDirectory, vvVerbs::CHECKOUT_CHANGESET, this->mwDialog->GetCheckoutTarget()))
			{
				pContext->Error_ResetReport("Unable to checkout changeset: %s", this->mwDialog->GetCheckoutTarget());
				return false;
			}
			break;
		case vvCreateDialog::CHECKOUT_TAG:
			if (!pVerbs->Checkout(*pContext, this->mwDialog->GetRepo(), this->msDirectory, vvVerbs::CHECKOUT_TAG, this->mwDialog->GetCheckoutTarget()))
			{
				pContext->Error_ResetReport("Unable to checkout tag: %s", this->mwDialog->GetCheckoutTarget());
				return false;
			}
			break;
		}
	}
	pContext->Log_FinishStep();

	return true;
}
