/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include "precompiled.h"
#include "vvCreateDialog.h"

#include "vvDialogHeaderControl.h"
#include "vvVerbs.h"
#include "vvValidatorEmptyConstraint.h"
#include "vvValidatorRepoExistsConstraint.h"


/*
**
** Types
**
*/

/**
 * IDs of controls in the dialog.
 */
enum Control
{
	CONTROL_REPO, //< The repo drop-down.
};


/*
**
** Globals
**
*/

/**
 * String displayed in the Repo list to indicate the creation of a new repo.
 * Must match the string used in the XRC file.
 */
static const wxString gsNewRepoString = "<New Repository>";

/*
 * Hints for various text box controls.
 */
static const wxString gsHint_New_Name                 = "Local name for the new repository";
static const wxString gsHint_New_Remote_Value         = "URL of the remote repository to clone";
static const wxString gsHint_Revision_Branch_Value    = "Name of the branch to checkout the head revision from";
static const wxString gsHint_Revision_Tag_Value       = "Name of the tag to checkout";
static const wxString gsHint_Revision_Changeset_Value = "ID of the changeset to checkout";

/**
 * String displayed at the top of the validation error message box.
 */
static const wxString gsValidationErrorMessage = "There was an error in the information specified.  Please correct the following problems and try again.";

/*
 * Strings displayed as field labels when validation errors occur.
 */
static const wxString gsValidationField_Repo                     = "Repository";
static const wxString gsValidationField_New_Name                 = "New Repository Name";
static const wxString gsValidationField_New_Remote_Value         = "Remote Repository";
static const wxString gsValidationField_New_Local_Value          = "Local Repository";
static const wxString gsValidationField_Revision_Branch_Value    = "Checkout Branch";
static const wxString gsValidationField_Revision_Tag_Value       = "Checkout Tag";
static const wxString gsValidationField_Revision_Changeset_Value = "Checkout Changeset";

/*
 * Strings used by the Select Branch dialog.
 */
static const wxString gsSelectBranch_Caption = "Select Branch";
static const wxString gsSelectBranch_Message = "Select the branch to checkout.";

/*
 * Strings used by the No Branches dialog.
 */
static const wxString gsNoBranches_Caption = "No Branches";
static const wxString gsNoBranches_Message = "The selected repository has no branches.";
static const long     giNoBranches_Style   = wxOK | wxICON_INFORMATION;


/*
**
** Public Functions
**
*/

bool vvCreateDialog::Create(
	wxWindow*            pParent,
	const wxString&      sFolder,
	vvVerbs&             cVerbs,
	vvContext&           cContext,
	wxString             sRepoName,
	CloneAction          eCloneAction,
	wxString             sCloneSource,
	CheckoutType         eCheckoutType,
	wxString             sCheckoutTarget
	)
{
	// create the dialog window
	if (!vvDialog::Create(pParent, &cVerbs, &cContext, this->GetClassInfo()->GetClassName()))
	{
		return false;
	}

	// store the given initialization data
	this->msRepoName       = sRepoName;
	this->meCloneAction    = eCloneAction;
	this->msCloneSource    = sCloneSource;
	this->meCheckoutType   = eCheckoutType;
	this->msCheckoutTarget = sCheckoutTarget;

	// retrieve pointers to all our widgets
	this->mwNew_Name                  = XRCCTRL(*this, "wNew_Name",                  wxTextCtrl);
	this->mwNew_Remote_Select         = XRCCTRL(*this, "wNew_Remote_Select",         wxRadioButton);
	this->mwNew_Remote_Value          = XRCCTRL(*this, "wNew_Remote_Value",          wxComboBox);
	this->mwNew_Local_Select          = XRCCTRL(*this, "wNew_Local_Select",          wxRadioButton);
	this->mwNew_Local_Value           = XRCCTRL(*this, "wNew_Local_Value",           wxChoice);
	this->mwNew_Empty_Select          = XRCCTRL(*this, "wNew_Empty_Select",          wxRadioButton);
	this->mwNew_Empty_Value           = XRCCTRL(*this, "wNew_Empty_Value",           wxStaticText);
	this->mwRevision_Head_Select      = XRCCTRL(*this, "wRevision_Head_Select",      wxRadioButton);
	this->mwRevision_Head_Value       = XRCCTRL(*this, "wRevision_Head_Value",       wxStaticText);
	this->mwRevision_Branch_Select    = XRCCTRL(*this, "wRevision_Branch_Select",    wxRadioButton);
	this->mwRevision_Branch_Value     = XRCCTRL(*this, "wRevision_Branch_Value",     wxTextCtrl);
	this->mwRevision_Branch_Browse    = XRCCTRL(*this, "wRevision_Branch_Browse",    wxButton);
	this->mwRevision_Tag_Select       = XRCCTRL(*this, "wRevision_Tag_Select",       wxRadioButton);
	this->mwRevision_Tag_Value        = XRCCTRL(*this, "wRevision_Tag_Value",        wxTextCtrl);
	this->mwRevision_Tag_Browse       = XRCCTRL(*this, "wRevision_Tag_Browse",       wxButton);
	this->mwRevision_Changeset_Select = XRCCTRL(*this, "wRevision_Changeset_Select", wxRadioButton);
	this->mwRevision_Changeset_Value  = XRCCTRL(*this, "wRevision_Changeset_Value",  wxTextCtrl);
	this->mwRevision_Changeset_Browse = XRCCTRL(*this, "wRevision_Changeset_Browse", wxButton);

	// create our header control
	vvDialogHeaderControl* wHeader = new vvDialogHeaderControl(this);
	this->mwRepo = new wxChoice(wHeader, CONTROL_REPO);
	wHeader->AddFolderValue(sFolder, cVerbs, cContext);
	wHeader->AddControlValue(vvDialogHeaderControl::sszLabel_Repo, this->mwRepo, false);
	wxXmlResource::Get()->AttachUnknownControl("wHeader", wHeader);

	// set the height of the blank static text boxes that are replacing actual text controls
	// we want the height to be the same as a text control
	// that way all of the associated radio buttons will be evenly spaced
	this->mwNew_Empty_Value    ->SetMinSize(this->mwNew_Remote_Value  ->GetSize());
	this->mwRevision_Head_Value->SetMinSize(this->mwRevision_Tag_Value->GetSize());

	// setup validators
	this->mcValidatorReporter.SetHelpMessage(gsValidationErrorMessage);
	this->mwRepo->SetValidator(vvValidator(gsValidationField_Repo)
		.SetValidatorFlags(vvValidator::VALIDATE_SUCCEED_IF_DISABLED | vvValidator::VALIDATE_SUCCEED_IF_HIDDEN)
		.SetReporter(&this->mcValidatorReporter)
		.AddConstraint(vvValidatorEmptyConstraint(false))
	);
	this->mwNew_Name->SetValidator(vvValidator(gsValidationField_New_Name)
		.SetValidatorFlags(vvValidator::VALIDATE_SUCCEED_IF_DISABLED | vvValidator::VALIDATE_SUCCEED_IF_HIDDEN)
		.SetReporter(&this->mcValidatorReporter)
		.AddConstraint(vvValidatorEmptyConstraint(false))
		.AddConstraint(vvValidatorRepoExistsConstraint(*this->GetVerbs(), *this->GetContext(), false))
		);
	this->mwNew_Remote_Value->SetValidator(vvValidator(gsValidationField_New_Remote_Value)
		.SetValidatorFlags(vvValidator::VALIDATE_SUCCEED_IF_DISABLED | vvValidator::VALIDATE_SUCCEED_IF_HIDDEN)
		.SetReporter(&this->mcValidatorReporter)
		.AddConstraint(vvValidatorEmptyConstraint(false))
	);
	this->mwNew_Local_Value->SetValidator(vvValidator(gsValidationField_New_Local_Value)
		.SetValidatorFlags(vvValidator::VALIDATE_SUCCEED_IF_DISABLED | vvValidator::VALIDATE_SUCCEED_IF_HIDDEN)
		.SetReporter(&this->mcValidatorReporter)
		.AddConstraint(vvValidatorEmptyConstraint(false))
	);
	this->mwRevision_Branch_Value->SetValidator(vvValidator(gsValidationField_Revision_Branch_Value)
		.SetValidatorFlags(vvValidator::VALIDATE_SUCCEED_IF_DISABLED | vvValidator::VALIDATE_SUCCEED_IF_HIDDEN)
		.SetReporter(&this->mcValidatorReporter)
		.AddConstraint(vvValidatorEmptyConstraint(false))
	);
	this->mwRevision_Tag_Value->SetValidator(vvValidator(gsValidationField_Revision_Tag_Value)
		.SetValidatorFlags(vvValidator::VALIDATE_SUCCEED_IF_DISABLED | vvValidator::VALIDATE_SUCCEED_IF_HIDDEN)
		.SetReporter(&this->mcValidatorReporter)
		.AddConstraint(vvValidatorEmptyConstraint(false))
	);
	this->mwRevision_Changeset_Value->SetValidator(vvValidator(gsValidationField_Revision_Changeset_Value)
		.SetValidatorFlags(vvValidator::VALIDATE_SUCCEED_IF_DISABLED | vvValidator::VALIDATE_SUCCEED_IF_HIDDEN)
		.SetReporter(&this->mcValidatorReporter)
		.AddConstraint(vvValidatorEmptyConstraint(false))
	);

	// set text field hints
	// can't currently do this in the XRC file
	this->mwNew_Name                ->SetHint(gsHint_New_Name);
	this->mwNew_Remote_Value        ->SetHint(gsHint_New_Remote_Value);
	this->mwRevision_Branch_Value   ->SetHint(gsHint_Revision_Branch_Value);
	this->mwRevision_Tag_Value      ->SetHint(gsHint_Revision_Tag_Value);
	this->mwRevision_Changeset_Value->SetHint(gsHint_Revision_Changeset_Value);

	// focus on the repo select list
	this->mwRepo->SetFocus();

	// success
	return true;
}

bool vvCreateDialog::TransferDataToWindow()
{
	// populate the lists of local repos
	this->mwRepo->Clear();
	this->mwNew_Local_Value->Clear();
	if (this->GetVerbs() != NULL)
	{
		wxArrayString cLocalRepos;
		this->GetVerbs()->GetLocalRepos(*this->GetContext(), cLocalRepos);
		for (wxArrayString::const_iterator itRepo = cLocalRepos.begin(); itRepo != cLocalRepos.end(); ++itRepo)
		{
			this->mwRepo->Append(*itRepo);
			this->mwNew_Local_Value->Append(*itRepo);
		}
	}
	this->mwRepo->Append(gsNewRepoString);

	// populate the list of recently used remote URLs
	this->mwNew_Remote_Value->Clear();
	if (this->GetVerbs() != NULL)
	{
		wxArrayString cRemoteRepos;
		wxArrayString cRemoteRepoAliases;
		wxString sRepoName;
		this->GetVerbs()->GetRemoteRepos(*this->GetContext(), sRepoName, cRemoteRepos, cRemoteRepoAliases);
		for (wxArrayString::const_iterator itRepo = cRemoteRepos.begin(); itRepo != cRemoteRepos.end(); ++itRepo)
		{
			this->mwNew_Remote_Value->Append(*itRepo);
		}
	}

	// check which repo to pre-select
	if (this->meCloneAction != CLONE_NOTHING || this->mwRepo->GetCount() == 1u)
	{
		this->mwRepo->SetSelection(0);
		this->mwNew_Name->SetValue(this->msRepoName);
	}
	else if (this->msRepoName.IsEmpty())
	{
		this->mwRepo->SetSelection(wxNOT_FOUND);
	}
	else
	{
		if (!this->mwRepo->SetStringSelection(this->msRepoName))
		{
			wxLogWarning("Couldn't pre-select repository '%s' because it's not in the local repository list.", this->msRepoName);
		}
	}

	// pre-fill in the appropriate clone data
	switch (this->meCloneAction)
	{
	case CLONE_EMPTY:
		this->mwNew_Empty_Select->SetValue(true);
		break;
	case CLONE_LOCAL:
		this->mwNew_Local_Select->SetValue(true);
		if (this->msCloneSource.IsEmpty())
		{
			this->mwNew_Local_Value->SetSelection(wxNOT_FOUND);
		}
		else
		{
			if (!this->mwNew_Local_Value->SetStringSelection(this->msCloneSource))
			{
				wxLogWarning("Couldn't pre-select local clone source '%s' because it's not in the local repository list.", this->msCloneSource);
			}
		}
		break;
	case CLONE_REMOTE:
		this->mwNew_Remote_Select->SetValue(true);
		this->mwNew_Remote_Value->SetValue(this->msCloneSource);
		break;
	}

	// pre-fill in the appropriate checkout data
	switch (this->meCheckoutType)
	{
	case CHECKOUT_HEAD:
		this->mwRevision_Head_Select->SetValue(true);
		break;
	case CHECKOUT_BRANCH:
		this->mwRevision_Branch_Select->SetValue(true);
		this->mwRevision_Branch_Value->SetValue(this->msCheckoutTarget);
		break;
	case CHECKOUT_CHANGESET:
		this->mwRevision_Changeset_Select->SetValue(true);
		this->mwRevision_Changeset_Value->SetValue(this->msCheckoutTarget);
		break;
	case CHECKOUT_TAG:
		this->mwRevision_Tag_Select->SetValue(true);
		this->mwRevision_Tag_Value->SetValue(this->msCheckoutTarget);
		break;
	}

	// update our display to reflect any widget changes
	this->UpdateDisplay();

	// success
	return true;
}

bool vvCreateDialog::TransferDataFromWindow()
{
	// check if the user wants to create a new repo
	if (this->mwRepo->GetStringSelection() != gsNewRepoString)
	{
		this->msRepoName    = this->mwRepo->GetStringSelection();
		this->meCloneAction = CLONE_NOTHING;
		this->msCloneSource = wxEmptyString;
	}
	else
	{
		this->msRepoName = this->mwNew_Name->GetValue();
		if (this->mwNew_Empty_Select->GetValue())
		{
			this->meCloneAction = CLONE_EMPTY;
			this->msCloneSource = wxEmptyString;
		}
		else if (this->mwNew_Local_Select->GetValue())
		{
			this->meCloneAction = CLONE_LOCAL;
			this->msCloneSource = this->mwNew_Local_Value->GetStringSelection();
		}
		else if (this->mwNew_Remote_Select->GetValue())
		{
			this->meCloneAction = CLONE_REMOTE;
			this->msCloneSource = this->mwNew_Remote_Value->GetValue();
		}
		else
		{
			wxFAIL_MSG("vvCreateDialog somehow ended up with no clone action selected.");
			return false;
		}
	}

	// check which revision the user wants to checkout
	if (this->mwRevision_Head_Select->GetValue())
	{
		this->meCheckoutType   = CHECKOUT_HEAD;
		this->msCheckoutTarget = wxEmptyString;
	}
	else if (this->mwRevision_Branch_Select->GetValue())
	{
		this->meCheckoutType   = CHECKOUT_BRANCH;
		this->msCheckoutTarget = this->mwRevision_Branch_Value->GetValue();
	}
	else if (this->mwRevision_Changeset_Select->GetValue())
	{
		this->meCheckoutType   = CHECKOUT_CHANGESET;
		this->msCheckoutTarget = this->mwRevision_Changeset_Value->GetValue();
	}
	else if (this->mwRevision_Tag_Select->GetValue())
	{
		this->meCheckoutType   = CHECKOUT_TAG;
		this->msCheckoutTarget = this->mwRevision_Tag_Value->GetValue();
	}
	else
	{
		wxFAIL_MSG("vvCreateDialog somehow ended up with no checkout type selected.");
		return false;
	}

	// success
	return true;
}

wxString vvCreateDialog::GetRepo() const
{
	return this->msRepoName;
}

vvCreateDialog::CloneAction vvCreateDialog::GetCloneAction() const
{
	return this->meCloneAction;
}

wxString vvCreateDialog::GetCloneSource() const
{
	return this->msCloneSource;
}

vvCreateDialog::CheckoutType vvCreateDialog::GetCheckoutType() const
{
	return this->meCheckoutType;
}

wxString vvCreateDialog::GetCheckoutTarget() const
{
	return this->msCheckoutTarget;
}

void vvCreateDialog::UpdateDisplay()
{
	// check a few things up front
	bool bRepoSelected    = this->mwRepo->GetSelection() != wxNOT_FOUND;
	bool bNewSelected     = this->mwRepo->GetStringSelection() == gsNewRepoString;
	bool bEmptySelected   = this->mwNew_Empty_Select->GetValue();
	bool bUseExistingRepo = bRepoSelected && !(bNewSelected && bEmptySelected);

	// show the New Repo block if they've selected to create a new repo
	vvXRCSIZERITEM(*this, "cNew")->Show(bNewSelected);

	// if the local list of repos to clone is empty, disable and hide it
	if (this->mwNew_Local_Value->GetCount() == 0u)
	{
		this->mwNew_Local_Select->SetValue(false);
		this->mwNew_Local_Select->Enable(false);
		vvXRCSIZERITEM(*this, "cNew_Local_Select")->Show(false);
		vvXRCSIZERITEM(*this, "cNew_Local_Value") ->Show(false);
	}

	// show the Checkout Revision block if:
	// - an existing repo is selected
	// - they're not creating a new empty repo
	vvXRCSIZERITEM(*this, "cRevision")->Show(bUseExistingRepo);

	// enable the text controls according to their corresponding to radio buttons
	this->mwNew_Remote_Value        ->Enable(this->mwNew_Remote_Select        ->GetValue());
	this->mwNew_Local_Value         ->Enable(this->mwNew_Local_Select         ->GetValue());
	this->mwRevision_Branch_Value   ->Enable(this->mwRevision_Branch_Select   ->GetValue());
	this->mwRevision_Tag_Value      ->Enable(this->mwRevision_Tag_Select      ->GetValue());
	this->mwRevision_Changeset_Value->Enable(this->mwRevision_Changeset_Select->GetValue());

	// if they want to use an existing repo, check if the one they've entered/chosen validates
	// if so, they can browse it for branches/tags/changesets
	bool bBrowsableRepo = false;
	if (bUseExistingRepo)
	{
		if (bRepoSelected && !bNewSelected)
		{
			// they're making a working folder from a repo we've already got
			bBrowsableRepo = true;
		}
		else if (
			   this->mwNew_Remote_Select->GetValue()
			&& this->mwNew_Remote_Value->GetValidator() != NULL
			&& this->mwNew_Remote_Value->GetValidator()->Validate(this)
			)
		{
			// they're making a working folder from a cloned remote repo
			bBrowsableRepo = true;
		}
		else if (
			   this->mwNew_Local_Select->GetValue()
			&& this->mwNew_Local_Value->GetValidator() != NULL
			&& this->mwNew_Local_Value->GetValidator()->Validate(this)
			)
		{
			// they're making a working folder from a cloned local repo
			bBrowsableRepo = true;
		}
	}
	this->mwRevision_Branch_Browse   ->Enable(this->mwRevision_Branch_Select   ->GetValue() && bBrowsableRepo);
	this->mwRevision_Tag_Browse      ->Enable(this->mwRevision_Tag_Select      ->GetValue() && bBrowsableRepo);
	this->mwRevision_Changeset_Browse->Enable(this->mwRevision_Changeset_Select->GetValue() && bBrowsableRepo);

	// our minimum height might have changed if we changed which blocks are visible
	this->SetMinSize(wxSize(this->GetSize().GetWidth(), this->GetSizer()->GetMinSize().GetHeight()));

	// relayout our controls and fit the window to their new size
	this->Layout();
	this->Fit();
}

void vvCreateDialog::OnControlValueChange(
	wxCommandEvent& WXUNUSED(cEvent)
	)
{
	this->UpdateDisplay();
}

void vvCreateDialog::OnBrowseBranches(
	wxCommandEvent& WXUNUSED(cEvent)
	)
{
	// figure out which repo we need to get the list from
	wxString sRepo = this->mwRepo->GetStringSelection();
	if (sRepo == gsNewRepoString)
	{
		if (this->mwNew_Local_Select->GetValue())
		{
			sRepo = this->mwNew_Local_Value->GetStringSelection();
		}
		else
		{
			wxMessageBox("Browsing for branches in remote repositories is not yet supported.", "Not Implemented", wxOK | wxICON_INFORMATION, this);
			return;
		}
	}

	// get the list of branches in the repo
	wxArrayString cBranches;
	{
		wxBusyCursor cBusyCursor;

		if (!this->GetVerbs()->GetBranches(*this->GetContext(), sRepo, true, cBranches))
		{
			wxLogError("Error retrieving branch list from repository: %s", sRepo);
			return;
		}
		if (cBranches.size() == 0u)
		{
			wxMessageBox(gsNoBranches_Message, gsNoBranches_Caption, giNoBranches_Style, this);
			return;
		}
	}

	// create a dialog to allow the user to choose one of them
	wxSingleChoiceDialog cDialog;
	{
		wxBusyCursor cBusyCursor;

		if (!cDialog.Create(this, gsSelectBranch_Message, gsSelectBranch_Caption, cBranches))
		{
			wxLogError("Error creating branch selection dialog.");
			return;
		}
	}

	// show the dialog
	if (cDialog.ShowModal() != wxID_OK)
	{
		return;
	}

	// set the value to the branch they chose
	this->mwRevision_Branch_Value->SetValue(cDialog.GetStringSelection());
}

void vvCreateDialog::OnBrowseTags(
	wxCommandEvent& WXUNUSED(cEvent)
	)
{
	wxMessageBox("Browsing for tags in the selected repository is not implemented yet.", "Not Implemented", wxOK | wxICON_INFORMATION, this);
}

void vvCreateDialog::OnBrowseChangesets(
	wxCommandEvent& WXUNUSED(cEvent)
	)
{
	wxMessageBox("Browsing for changesets in the selected repository is not implemented yet.", "Not Implemented", wxOK | wxICON_INFORMATION, this);
}

IMPLEMENT_DYNAMIC_CLASS(vvCreateDialog, vvDialog);

BEGIN_EVENT_TABLE(vvCreateDialog, vvDialog)
	// when a value changes, we need to update the display to match
	EVT_CHOICE(     CONTROL_REPO,                        vvCreateDialog::OnControlValueChange)
	EVT_RADIOBUTTON(XRCID("wNew_Remote_Select"),         vvCreateDialog::OnControlValueChange)
	EVT_TEXT(       XRCID("wNew_Remote_Value"),          vvCreateDialog::OnControlValueChange)
	EVT_RADIOBUTTON(XRCID("wNew_Local_Select"),          vvCreateDialog::OnControlValueChange)
	EVT_CHOICE(     XRCID("wNew_Local_Value"),           vvCreateDialog::OnControlValueChange)
	EVT_RADIOBUTTON(XRCID("wNew_Empty_Select"),          vvCreateDialog::OnControlValueChange)
	EVT_RADIOBUTTON(XRCID("wRevision_Head_Select"),      vvCreateDialog::OnControlValueChange)
	EVT_RADIOBUTTON(XRCID("wRevision_Branch_Select"),    vvCreateDialog::OnControlValueChange)
	EVT_RADIOBUTTON(XRCID("wRevision_Tag_Select"),       vvCreateDialog::OnControlValueChange)
	EVT_RADIOBUTTON(XRCID("wRevision_Changeset_Select"), vvCreateDialog::OnControlValueChange)

	// when a browse button is clicked, call up the appropriate browse dialog
	EVT_BUTTON(XRCID("wRevision_Branch_Browse"),    vvCreateDialog::OnBrowseBranches)
	EVT_BUTTON(XRCID("wRevision_Tag_Browse"),       vvCreateDialog::OnBrowseTags)
	EVT_BUTTON(XRCID("wRevision_Changeset_Browse"), vvCreateDialog::OnBrowseChangesets)
END_EVENT_TABLE()
