/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#ifndef VV_CREATE_DIALOG_H
#define VV_CREATE_DIALOG_H

#include "precompiled.h"
#include "vvDialog.h"
#include "vvValidatorMessageBoxReporter.h"


/**
 * A dialog box used to create a working folder in a directory.
 * Allows the user to attach the working folder to:
 * (1) an existing repo (checkout)
 * (2) a brand new repo (init)
 * (3) a freshly cloned repo (clone, checkout)
 */
class vvCreateDialog : public vvDialog
{
// types
public:
	/**
	 * Actions the user might select to create a freshly cloned repo to create the working folder from.
	 */
	enum CloneAction
	{
		CLONE_NOTHING, //< User chose not to create a new repo.  Clone source is not valid.
		CLONE_EMPTY,   //< User chose to create a new blank repo.  Clone source is not valid.
		CLONE_LOCAL,   //< User chose to clone a local repo.  The clone source will be the name.
		CLONE_REMOTE,  //< User chose to clone a remote repo.  The clone source will be a URL.
	};

	/**
	 * Methods that the user can use to select which revision to checkout into the new working folder.
	 */
	enum CheckoutType
	{
		CHECKOUT_HEAD,      //< User chose to checkout the head revision.  Checkout target is not valid.
		CHECKOUT_BRANCH,    //< User chose to checkout the head revision of a branch.  Checkout target is the branch name.
		CHECKOUT_CHANGESET, //< User chose to checkout a specific changeset.  Checkout target will be the changeset ID.
		CHECKOUT_TAG,       //< User chose to checkout a tagged revision.  Checkout target will be the tag name.
	};

// functionality
public:
	/**
	 * Creates the actual dialog window.
	 */
	virtual bool Create(
		wxWindow*            pParent,                         //< [in] The new dialog's parent window, or NULL if it won't have one.
		const wxString&      sFolder,                         //< [in] The folder to create a working folder in.
		class vvVerbs&       cVerbs,                          //< [in] The vvVerbs implementation to retrieve local repo and config info from.
		class vvContext&     cContext,                        //< [in] Context to use with pVerbs.
		wxString             sRepoName       = wxEmptyString, //< [in] The name of the repo to pre-select.
		CloneAction          eCloneAction    = CLONE_NOTHING, //< [in] the clone action to pre-select.
		wxString             sCloneSource    = wxEmptyString, //< [in] The clone source to pre-select.  Where it's placed depends on eCloneAction.
		CheckoutType         eCheckoutType   = CHECKOUT_HEAD, //< [in] The checkout type to pre-select.
		wxString             sCheckoutTarget = wxEmptyString  //< [in] The checkout target to pre-select.  Where it's placed depends on eCheckoutType.
		);

	/**
	 * Transfers our internal data to our widgets.
	 */
	virtual bool TransferDataToWindow();

	/**
	 * Stores data from our widgets to our internal data.
	 */
	virtual bool TransferDataFromWindow();

	/**
	 * Retrieves the name of the repo chosen by the user to create the working folder from.
	 * Might be the name of a repo that doesn't yet exist, check GetCloneAction.
	 */
	wxString GetRepo() const;

	/**
	 * Gets the action chosen by the user for attaching to a freshly cloned repo.
	 */
	CloneAction GetCloneAction() const;

	/**
	 * The source that the user wants to clone a repo from prior to attaching to it.
	 * Check GetCloneAction to determine whether it's a remote URL or another local repo name.
	 */
	wxString GetCloneSource() const;

	/**
	 * Gets the type of checkout the user wants to perform to populate the new working folder.
	 * If they chose to checkout a particular revision, use GetCheckoutTarget to get the appropriate target.
	 */
	CheckoutType GetCheckoutType() const;

	/**
	 * The target that the user is checking out from the repo.
	 * Might be the name of a tag, branch, changeset, etc.  Use GetCheckoutType to tell.
	 */
	wxString GetCheckoutTarget() const;

// internal functionality
protected:
	/**
	 * Update the visible/enabled state of all controls, based on the current values.
	 */
	void UpdateDisplay();

// event handlers
protected:
	void OnControlValueChange(wxCommandEvent& cEvent);
	void OnBrowseBranches(wxCommandEvent& cEvent);
	void OnBrowseTags(wxCommandEvent& cEvent);
	void OnBrowseChangesets(wxCommandEvent& cEvent);

// private data
private:
	// internal data
	wxString     msRepoName;       //< The name of the repo to create the working folder from.
	CloneAction  meCloneAction;    //< The action to take to create a freshly cloned repo.
	wxString     msCloneSource;    //< The source to create a freshly cloned repo from.
	CheckoutType meCheckoutType;   //< The method to use to checkout a revision from the repo.
	wxString     msCheckoutTarget; //< The target revision to checkout from the repo.

	// widgets
	wxChoice*      mwRepo;
	wxTextCtrl*    mwNew_Name;
	wxRadioButton* mwNew_Remote_Select;
	wxComboBox*    mwNew_Remote_Value;
	wxRadioButton* mwNew_Local_Select;
	wxChoice*      mwNew_Local_Value;
	wxRadioButton* mwNew_Empty_Select;
	wxStaticText*  mwNew_Empty_Value;
	wxRadioButton* mwRevision_Head_Select;
	wxStaticText*  mwRevision_Head_Value;
	wxRadioButton* mwRevision_Branch_Select;
	wxTextCtrl*    mwRevision_Branch_Value;
	wxButton*      mwRevision_Branch_Browse;
	wxRadioButton* mwRevision_Tag_Select;
	wxTextCtrl*    mwRevision_Tag_Value;
	wxButton*      mwRevision_Tag_Browse;
	wxRadioButton* mwRevision_Changeset_Select;
	wxTextCtrl*    mwRevision_Changeset_Value;
	wxButton*      mwRevision_Changeset_Browse;

	// misc
	vvValidatorMessageBoxReporter mcValidatorReporter; //< Reporter that will display validation errors.

// macro declarations
private:
	DECLARE_DYNAMIC_CLASS(vvCreateDialog);
	DECLARE_EVENT_TABLE();
};


#endif
