/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include "precompiled.h"
#include "vvDialogHeaderControl.h"

#include "vvContext.h"
#include "vvFlags.h"
#include "vvVerbs.h"


/*
**
** Globals
**
*/

const wxArtID vvDialogHeaderControl::scDefaultBitmapName = "SourceGear_64";

const char* vvDialogHeaderControl::sszLabel_Folder = "Folder:";
const char* vvDialogHeaderControl::sszLabel_Repo   = "Repository:";
const char* vvDialogHeaderControl::sszLabel_Branch = "Branch:";


/*
**
** Public Functions
**
*/

vvDialogHeaderControl::vvDialogHeaderControl()
	: mpSizer(NULL)
	, mwBitmap(NULL)
{
}

vvDialogHeaderControl::vvDialogHeaderControl(
	wxWindow*      pParent,
	wxWindowID     cId,
	const wxArtID& cBitmap,
	const wxPoint& cPosition,
	const wxSize&  cSize,
	long           iStyle
	)
	: mpSizer(NULL)
	, mwBitmap(NULL)
{
	this->Create(pParent, cId, cBitmap, cPosition, cSize, iStyle);
}

vvDialogHeaderControl::vvDialogHeaderControl(
	wxWindow*       pParent,
	wxWindowID      cId,
	const wxString& sFolder,
	vvVerbs&        cVerbs,
	vvContext&      cContext,
	const wxArtID&  cBitmap,
	const wxPoint&  cPosition,
	const wxSize&   cSize,
	long            iStyle
	)
	: mpSizer(NULL)
	, mwBitmap(NULL)
{
	this->Create(pParent, cId, sFolder, cVerbs, cContext, cBitmap, cPosition, cSize, iStyle);
}

bool vvDialogHeaderControl::Create(
	wxWindow*      pParent,
	wxWindowID     cId,
	const wxArtID& cBitmap,
	const wxPoint& cPosition,
	const wxSize&  cSize,
	long           iStyle
	)
{
	// create the base control
	if (!wxPanel::Create(pParent, cId, cPosition, cSize, iStyle))
	{
		return false;
	}

	// create the bitmap control
	this->mwBitmap = new wxStaticBitmap(this, wxID_ANY, wxArtProvider::GetBitmap(cBitmap));

	// create a flex grid sizer to layout any added controls
	this->mpSizer = new wxFlexGridSizer(2, wxSize(8,8));
	this->mpSizer->AddGrowableCol(1u, 1);

	// create a horizontal box sizer to layout the grid and bitmap
	wxBoxSizer* pBoxSizer = new wxBoxSizer(wxHORIZONTAL);
	pBoxSizer->Add(this->mpSizer, wxSizerFlags().Center().Proportion(1));
	pBoxSizer->AddSpacer(8);
	pBoxSizer->Add(this->mwBitmap, wxSizerFlags().Proportion(0));

	// set the horizontal sizer to this window
	this->SetSizer(pBoxSizer);

	// done
	return true;
}

bool vvDialogHeaderControl::Create(
	wxWindow*       pParent,
	wxWindowID      cId,
	const wxString& sFolder,
	vvVerbs&        cVerbs,
	vvContext&      cContext,
	const wxArtID&  cBitmap,
	const wxPoint&  cPosition,
	const wxSize&   cSize,
	long            iStyle
	)
{
	return (
		   this->Create(pParent, cId, cBitmap, cPosition, cSize, iStyle)
		&& this->AddDefaultValues(sFolder, cVerbs, cContext)
		);
}

bool vvDialogHeaderControl::AddControlValue(
	const wxString& sLabel,
	wxControl*      wControl,
	bool            bExpand
	)
{
	wxASSERT(wControl->GetParent() == this);

	// create a label and add it
	wxStaticText* wLabel = new wxStaticText(this, wxID_ANY, sLabel);
	this->mpSizer->Add(wLabel, wxSizerFlags().Align(wxALIGN_CENTER_VERTICAL));

	// add the control
	wxSizerFlags cFlags;
	cFlags.Align(wxALIGN_CENTER_VERTICAL);
	if (bExpand)
	{
		cFlags.Expand();
	}
	this->mpSizer->Add(wControl, cFlags);

	// re-layout the control
	return this->Layout();
}

bool vvDialogHeaderControl::AddTextValue(
	const wxString& sLabel,
	const wxString& sValue
	)
{
	// Note: It may seem odd to do this in two steps, but it seems to be necessary.
	//       1) Create a text control with no text value and a height of zero.
	//       2) Assign a text value to the control.
	//
	//       If you pass the value directly to the constructor, you get an exception
	//       that is usually caught and apparently handled correctly in the Win32 code.
	//       However, aside from it looking ominous in a debugger output window,
	//       I have also occasionally seen an unhandled exception occur that seems to
	//       be strongly correlated to the ones that seem to be handled fine.
	//       (It happens deep in the text control creation, while the wx code is
	//       handling a command message from the control, possible a CREATE-like one).
	//
	//       Passing a height of zero is necessary to get the control to adjust its
	//       height to exactly match the text you give it, which is required for it
	//       to properly align vertically with its associated label.
	wxTextCtrl* pValue = new wxTextCtrl(this, wxID_ANY, wxEmptyString, wxDefaultPosition, wxSize(wxDefaultCoord,0), wxBORDER_NONE | wxTE_READONLY);
	pValue->SetValue(sValue);
	return this->AddControlValue(sLabel, pValue, true);
}

bool vvDialogHeaderControl::AddFolderValue(
	const wxString& sFolder,
	vvVerbs&        cVerbs,
	vvContext&      cContext
	)
{
	wxString sWorkingFolder;
	if (cVerbs.FindWorkingFolder(cContext, sFolder, &sWorkingFolder, NULL))
	{
		return this->AddTextValue(sszLabel_Folder, sWorkingFolder);
	}
	else
	{
		cContext.Error_Reset();
		return this->AddTextValue(sszLabel_Folder, sFolder);
	}
}

bool vvDialogHeaderControl::AddRepoValue(
	const wxString& sFolder,
	vvVerbs&        cVerbs,
	vvContext&      cContext
	)
{
	wxString sRepo;
	return (
		   cVerbs.FindWorkingFolder(cContext, sFolder, NULL, &sRepo)
		&& this->AddTextValue(sszLabel_Repo, sRepo)
		);
}

bool vvDialogHeaderControl::AddFolderAndRepoValues(
	const wxString& sFolder,
	vvVerbs&        cVerbs,
	vvContext&      cContext
	)
{
	wxString sWorkingFolder;
	wxString sRepo;
	return (
		   cVerbs.FindWorkingFolder(cContext, sFolder, &sWorkingFolder, &sRepo)
		&& this->AddTextValue(sszLabel_Folder, sWorkingFolder)
		&& this->AddTextValue(sszLabel_Repo, sRepo)
		);
}

bool vvDialogHeaderControl::AddBranchValue(
	const wxString& sFolder,
	vvVerbs&        cVerbs,
	vvContext&      cContext
	)
{
	wxString sBranch;
	return (
		   cVerbs.GetBranch(cContext, sFolder, sBranch)
		&& this->AddTextValue(sszLabel_Branch, sBranch)
		);
}

bool vvDialogHeaderControl::AddDefaultValues(
	const wxString& sFolder,
	vvVerbs&        cVerbs,
	vvContext&      cContext
	)
{
	return (
		   this->AddFolderAndRepoValues(sFolder, cVerbs, cContext)
		&& this->AddBranchValue(sFolder, cVerbs, cContext)
		);
}

IMPLEMENT_DYNAMIC_CLASS(vvDialogHeaderControl, wxControl);

BEGIN_EVENT_TABLE(vvDialogHeaderControl, wxControl)
END_EVENT_TABLE()
