/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#ifndef VV_FLAGS_H
#define VV_FLAGS_H

#include "precompiled.h"


/**
 * A mix-in base class for classes that contain a set of flags.
 */
class vvHasFlags
{
// construction
public:
	/**
	 * Constructs a new set of flags.
	 */
	vvHasFlags(
		unsigned int uFlags = 0u //< [in] Initial value for the flags.
		)
		: muFlags(uFlags)
	{
	}

// functionality
public:
	/**
	 * Gets the raw value of the flags.
	 */
	unsigned int GetFlags() const
	{
		return this->muFlags;
	}

	/**
	 * Checks if the flags are empty.
	 */
	bool FlagsEmpty() const
	{
		return this->muFlags == 0u;
	}

	/**
	 * Checks if the flags contain exactly one flag.
	 */
	bool HasOneFlag() const
	{
		// This amounts to checking if the flags are a power of 2.
		// http://graphics.stanford.edu/~seander/bithacks.html#DetermineIfPowerOf2
		return !this->FlagsEmpty() && ((this->muFlags & (this->muFlags - 1u)) == 0u);
	}

	/**
	 * Checks if the flags contain more than one flag.
	 */
	bool HasMultipleFlags() const
	{
		return !this->FlagsEmpty() && !this->HasOneFlag();
	}

	/**
	 * Checks if the flags contain any of the given flags.
	 */
	bool HasAnyFlag(
		unsigned int uFlags //< [in] Flags to check for.
		) const
	{
		return ((this->muFlags & uFlags) != 0);
	}

	/**
	 * Checks if the flags contain all of the given flags.
	 */
	bool HasAllFlags(
		unsigned int uFlags //< [in] Flags to check for.
		) const
	{
		return ((this->muFlags & uFlags) == uFlags);
	}

	/**
	 * Checks if the flags contain any flags that are NOT among the given flags.
	 * Equivalent to HasAnyFlag(~uFlags).
	 */
	bool HasAnyFlagExcept(
		unsigned int uFlags //< [in] Flags to check for.
		) const
	{
		return this->HasAnyFlag(~uFlags);
	}

	/**
	 * Checks if the flags contain all flags that are NOT among the given flags.
	 * Equivalent to HasAllFlags(~uFlags);
	 */
	bool HasAllFlagsExcept(
		unsigned int uFlags //< [in] Flags to check for.
		) const
	{
		return this->HasAllFlags(~uFlags);
	}

	/**
	 * Sets the flags to an entirely new value.
	 * Returns the new value.
	 */
	unsigned int SetFlags(
		unsigned int uFlags //< [in] New value for the flags.
		)
	{
		this->muFlags = uFlags;
		return this->muFlags;
	}

	/**
	 * Adds one or more flags.
	 * Returns the new value.
	 */
	unsigned int AddFlags(
		unsigned int uFlags //< [in] Flags to add.
		)
	{
		this->muFlags |= uFlags;
		return this->muFlags;
	}

	/**
	 * Removes one or more flags.
	 * Returns the new value.
	 */
	unsigned int RemoveFlags(
		unsigned int uFlags //< [in] Flags to remove.
		)
	{
		this->muFlags &= ~uFlags;
		return this->muFlags;
	}

	/**
	 * Simultaneously adds and removes flags.
	 * Returns the new value.
	 */
	unsigned int ModifyFlags(
		unsigned int uAddFlags,   //< [in] Flags to add.
		unsigned int uRemoveFlags //< [in] Flags to remove.
		)
	{
		this->AddFlags(uAddFlags);
		this->RemoveFlags(uRemoveFlags);
		return this->muFlags;
	}

// private data
private:
	unsigned int muFlags; //< Current flags value.
};


/**
 * A class that wraps an unsigned integer and provides
 * functionality for treating it like a set of bit flags.
 */
class vvFlags : public vvHasFlags
{
// constants
public:
	static const unsigned int ALL     = ~0u;          //< Mask with all possible flags enabled.
	static const unsigned int LOWEST  = 1u;           //< The lowest possible flag.
	static const unsigned int HIGHEST = ~((~0u)>>1u); //< The highest possible flag.

// types
public:
	/**
	 * A class for iterating through each flag in a vvFlags instance, lowest to highest.
	 * Optionally only iterates flags that also match a given mask.
	 */
	class iterator
	{
	// construction
	public:
		/**
		 * Constructor.
		 * The iterator is initialized to "one past the end" state.
		 * Use operator++ to advance it to point at the first flag.
		 */
		iterator(
			const vvFlags& cFlags,     //< [in] The flags instance to iterate through.
			unsigned int   uMask = ALL //< [in] The mask to match against when iterating.
			)
			: mpFlags(&cFlags)
			, muMask(uMask)
			, muValue(0u)
		{
		}

	// functionality
	public:
		/**
		 * Equality operator.
		 * Crashes if comparing iterators that point to different flag values.
		 */
		bool operator==(
			const iterator& that
			)
		{
			wxASSERT(this->mpFlags == that.mpFlags);
			return (this->muValue == that.muValue && this->muMask == that.muMask);
		}

		/**
		 * Inequality operator.
		 * Opposite of equality operator.
		 */
		bool operator!=(
			const iterator& that
			)
		{
			return !(*this == that);
		}

		/**
		 * Dereference operator.
		 * Crashes if the iterator has an invalid value.
		 */
		vvFlags operator*()
		{
			wxASSERT(this->muValue != 0u);
			return this->muValue;
		}

		/**
		 * Member dereference operator.
		 * Crashes if the iterator has an invalid value.
		 */
		vvFlags operator->()
		{
			wxASSERT(this->muValue != 0u);
			return this->muValue;
		}

		/**
		 * Pre-increment operator.
		 */
		iterator& operator++()
		{
			wxASSERT(this->muValue == 0u || vvFlags(this->muValue).HasOneFlag());
			for (unsigned int uCurrent = this->muValue == 0u ? vvFlags::LOWEST : this->muValue << 1u; uCurrent != 0u; uCurrent <<= 1u)
			{
				if (this->mpFlags->HasAnyFlag(uCurrent) && (this->muMask & uCurrent) != 0u)
				{
					this->muValue = uCurrent;
					return *this;
				}
			}
			this->muValue = 0u;
			return *this;
		}

		/**
		 * Post-increment operator.
		 */
		iterator operator++(
			int
			)
		{
			iterator cResult = *this;
			++(*this);
			return cResult;
		}

		/**
		 * Pre-decrement operator.
		 */
		iterator& operator--()
		{
			// Note: It's very important that the index be unsigned!
			//       This ensures that the right-shift is logical.
			//       Right-shifts on signed numbers are compiler-dependent, but usually arithmetic.
			//       http://stackoverflow.com/questions/7622/shift-operator-in-c
			//       If we end up with an arithmetic shift, then this loop might go infinite.
			wxASSERT(this->muValue == 0u || vvFlags(this->muValue).HasOneFlag());
			for (unsigned int uCurrent = this->muValue == 0u ? vvFlags::HIGHEST : this->muValue >> 1u; uCurrent != 0u; uCurrent >>= 1u)
			{
				if (this->mpFlags->HasAnyFlag(uCurrent) && (this->muMask & uCurrent) != 0u)
				{
					this->muValue = uCurrent;
					return *this;
				}
			}
			this->muValue = 0u;
			return *this;
		}

		/**
		 * Post-decrement operator.
		 */
		iterator operator--(
			int
			)
		{
			iterator cResult = *this;
			--(*this);
			return cResult;
		}

	// private data
	private:
		const vvFlags* mpFlags; //< The flags that are being iterated.
		unsigned int   muMask;  //< The mask of flags to iterate.
		unsigned int   muValue; //< The iterator's current value (zero indicates null/invalid iterator).
	};

// construction
public:
	/**
	 * Constructs a new set of flags.
	 */
	vvFlags(
		unsigned int uValue = 0u //< [in] Initial value for the flags.
		)
		: vvHasFlags(uValue)
	{
	}

// operators
public:
	/**
	 * Assignment from an unsigned int.
	 */
	vvFlags& operator=(
		unsigned int uValue
		)
	{
		this->SetValue(uValue);
		return *this;
	}

	/**
	 * Implicit conversion back to an unsigned int.
	 */
	operator unsigned int() const
	{
		return this->GetFlags();
	}

	/**
	 * Bitwise NOT operator.
	 */
	vvFlags operator~() const
	{
		return vvFlags(~(this->GetFlags()));
	}

	/**
	 * Bitwise AND operator.
	 */
	vvFlags operator&(const vvFlags& that) const
	{
		return vvFlags(this->GetFlags() & that.GetFlags());
	}

	/**
	 * Bitwise OR operator.
	 */
	vvFlags operator|(const vvFlags& that) const
	{
		return vvFlags(this->GetFlags() | that.GetFlags());
	}

	/**
	 * Bitwise XOR operator.
	 */
	vvFlags operator^(const vvFlags& that) const
	{
		return vvFlags(this->GetFlags() ^ that.GetFlags());
	}

// functionality
public:
	/**
	 * Convenience wrapper around base GetFlags().
	 */
	unsigned int GetValue() const
	{
		return this->GetFlags();
	}

	/**
	 * Convenience wrapper around base SetFlags().
	 */
	unsigned int SetValue(
		unsigned int uValue
		)
	{
		return this->SetFlags(uValue);
	}

	/**
	 * Gets an iterator to the first (lowest) flag in the flags.
	 */
	iterator begin() const
	{
		iterator it(*this);
		++it;
		return it;
	}

	/**
	 * Gets an iterator to the first (lowest) flag in the flags
	 * that only iterates flags in the given mask.
	 */
	iterator begin(
		vvFlags cMask //< [in] The returned iterator will only iterate through flags in this mask.
		) const
	{
		iterator it(*this, cMask);
		++it;
		return it;
	}

	/**
	 * Gets an iterator to one past the last (highest) flag in the flags.
	 */
	iterator end() const
	{
		return iterator(*this, ALL);
	}

	/**
	 * Gets an iterator to one past the last (highest) flag in the flags
	 * that only iterates flags in the given mask.
	 */
	iterator end(
		vvFlags cMask //< [in] The returned iterator will only iterate through flags in this mask.
		) const
	{
		return iterator(*this, cMask);
	}
};


/**
 * Bitwise AND operator for unsigned int.
 */
inline vvFlags operator&(unsigned int uLeft, const vvFlags& cRight)
{
	return vvFlags(uLeft & cRight.GetFlags());
}

/**
 * Bitwise OR operator for unsigned int.
 */
inline vvFlags operator|(unsigned int uLeft, const vvFlags& cRight)
{
	return vvFlags(uLeft | cRight.GetFlags());
}

/**
 * Bitwise XOR operator for unsigned int.
 */
inline vvFlags operator^(unsigned int uLeft, const vvFlags& cRight)
{
	return vvFlags(uLeft ^ cRight.GetFlags());
}


#endif
