#include "precompiled.h"
#include <wx/dataview.h>
#include <wx/thread.h>
#include "vvHistoryListControl.h"
#include "vvVerbs.h"
#include "vvHistoryObjects.h"
#include "vvContext.h"
#include "vvSgHelpers.h"
#include "vvDataViewToggleIconTextRenderer.h"


wxDECLARE_EVENT(wxEVT_COMMAND_MODEL_FETCH_MORE__COMPLETE, wxThreadEvent);
/**
 * Models data about the history in a repository.
 */
class vvHistoryModel : public wxDataViewVirtualListModel
{
// types
public:
	/**
	 * Enumeration of columns used by the model.
	 *
	 * The B/S/U prefixes indicate the value type of the column (Bool, String, Unsigned).
	 * The prefix codes are always listed alphabetically, if it helps you remember the names.
	 *
	 * B and S values use variant types "bool" and "string" respectively.
	 * U values use variant type "long", but are actually cast unsigned ints, and should be cast back to unsigned int before use.
	 */
	enum Column
	{
		//COLUMN_B_CHECKED, //< Item's checked state.
		COLUMN_S_REVISION,
		COLUMN_S_COMMENT,
		COLUMN_S_BRANCH,
		COLUMN_S_STAMP,
		COLUMN_S_TAG,
		COLUMN_U_DATE,
		COLUMN_S_USER,
		COLUMN_COUNT,     //< Number of entries in this enumeration, for iteration.
	};

// construction
public:
	/**
	 * Default constructor.
	 */
	vvHistoryModel(vvVerbs& cVerbs,	vvContext& cContext, vvHistoryListControl * parent, wxString sRepoName);

// functionality
public:
	
	/**
	 * Removes all items from the data model.
	 */
	void Clear();

// implementation
public:
	void LoadHistory();
	void GetMoreData();
	void SetData(vvVerbs::HistoryData data);
 	virtual unsigned int GetColumnCount() const;
	virtual unsigned int vvHistoryModel::GetCount() const;
	virtual wxString GetColumnType(
		unsigned int uColumn //< [in] The column to get the type of.
		) const;

	void GetValueByRow(
		wxVariant&            cValue,
		unsigned int 		  row,
		unsigned int 		  uColumn
		) const;

	bool SetValueByRow(const wxVariant &variant,
	                               unsigned row, unsigned col);

	void SetFilter(vvHistoryFilterOptions vvHistoryFilter);
	vvVerbs::HistoryEntry GetHistoryDataAt(int row)
	{
		wxCriticalSectionLocker lock(m_dataAccessLock);
		return this->mcDataItems[row];
	}
private:
	vvVerbs * mpVerbs;
	vvContext * mpContext;
	
	void * pHistoryToken;
	unsigned int nEndOfPage;
	wxCriticalSection m_lock;
	wxCriticalSection m_fetchCheckCriticalSection;
	wxCriticalSection m_dataAccessLock;
	bool m_bAlreadyFetching;
	wxString m_sRepoName;
	vvVerbs::HistoryData mcDataItems; //< The data items that we're displaying.
	vvHistoryFilterOptions m_historyFilter;
	wxCriticalSection m_historyFilterLock;
// macro declarations
private:
	vvHistoryListControl * mpListControl;

	vvDISABLE_COPY(vvHistoryModel);
};

vvHistoryModel::vvHistoryModel(vvVerbs& cVerbs,	vvContext& cContext, vvHistoryListControl * parent, wxString sRepoName)
{
	mpVerbs = &cVerbs;
	mpContext = &cContext;
	pHistoryToken = NULL;
	nEndOfPage = 0;
	mpListControl = parent;
	m_sRepoName = sRepoName;
	m_bAlreadyFetching = false;
}

void vvHistoryModel::Clear()
{
	{
		wxCriticalSectionLocker lock(m_dataAccessLock);
		this->mcDataItems.clear();
		nEndOfPage = wxUINT32_MAX;
	}
	this->Reset(0);
}

void vvHistoryModel::SetData(vvVerbs::HistoryData data)
{
	Clear();
	wxCriticalSectionLocker lock(this->m_dataAccessLock);
	this->mcDataItems = data;
	this->pHistoryToken = NULL;
	this->Reset(mcDataItems.size());
}

unsigned int vvHistoryModel::GetCount() const
{
	vvHistoryModel* nonConst = (vvHistoryModel*)this;
	wxCriticalSectionLocker lock(nonConst->m_dataAccessLock);
	return this->mcDataItems.size();
}

void vvHistoryModel::SetFilter(vvHistoryFilterOptions vvHistoryFilter)
{
	wxCriticalSectionLocker lock(m_historyFilterLock);
	m_historyFilter = vvHistoryFilter;
}
void vvHistoryModel::LoadHistory()
{
	
	vvVerbs::HistoryData newData;
	{
		wxCriticalSectionLocker lock(m_historyFilterLock);
		mpListControl->SendStatusMessage("Loading...", false);
		mpVerbs->History__Free_Token(*mpContext, this->pHistoryToken);
		this->pHistoryToken = NULL;
		if (mpVerbs->History(*mpContext, m_sRepoName, m_historyFilter, newData, &pHistoryToken))
		{
			wxCriticalSectionLocker lock(m_dataAccessLock);
			mcDataItems = newData;
			this->Reset(newData.size());
		}
		int count = mcDataItems.size();
		if (count != 1)
			mpListControl->SendStatusMessage(wxString::Format("%d results", count), pHistoryToken != NULL);
		else
			mpListControl->SendStatusMessage(wxString::Format("%d result", count), pHistoryToken != NULL);
	}
	if (pHistoryToken != NULL)
		nEndOfPage = newData.size() - 10;
	else
		nEndOfPage = wxUINT32_MAX;
	
	//m_bAlreadyFetching = false;

}

void vvHistoryModel::GetMoreData()
{
	void * pNewHistoryToken = NULL;
	vvVerbs::HistoryData cDataItemsNew;
	//wxCriticalSectionLocker lock(m_lock);
	{
		//wxCriticalSectionLocker lock(m_fetchCheckCriticalSection);
		//if (m_bAlreadyFetching)
			//return;
		//else
//			m_bAlreadyFetching = true;
	}

	if (this->pHistoryToken)
	{
		mpListControl->SendStatusMessage("Loading More...", false);
		if (mpVerbs->History__Fetch_More(*mpContext, m_sRepoName, 200, this->pHistoryToken, cDataItemsNew, &pNewHistoryToken))
		{
			wxCriticalSectionLocker lock(m_dataAccessLock);
			mcDataItems.insert(mcDataItems.end(), cDataItemsNew.begin(), cDataItemsNew.end());
			this->Reset(mcDataItems.size());
			this->pHistoryToken = pNewHistoryToken;
			this->nEndOfPage = mcDataItems.size() - 10;
		}
		int count = mcDataItems.size();
		if (count != 1)
			mpListControl->SendStatusMessage(wxString::Format("%d results", count), pHistoryToken != NULL);
		else
			mpListControl->SendStatusMessage(wxString::Format("%d result", count), pHistoryToken != NULL);
	}
	{
		//wxCriticalSectionLocker lock(m_fetchCheckCriticalSection);
		//m_bAlreadyFetching = false;
	}

}

//This seems to never be called.
unsigned int vvHistoryModel::GetColumnCount() const
{
	return vvHistoryModel::COLUMN_COUNT;
}

//This seems to never be called.
wxString vvHistoryModel::GetColumnType(
	unsigned int uColumn
	) const
{
	switch (static_cast<Column>(uColumn))
	{
	//case COLUMN_B_CHECKED:
	//	return "bool";

	case COLUMN_S_REVISION:
	case COLUMN_S_COMMENT:
	case COLUMN_S_BRANCH:
	case COLUMN_S_TAG:
	case COLUMN_S_STAMP:
	case COLUMN_S_USER:
		return "string";

	case COLUMN_U_DATE:
		return "date";

	default:
		wxLogError("Unknown StampsModel column: %u", uColumn);
		return wxEmptyString;
	}
}

bool vvHistoryModel::SetValueByRow(const wxVariant &variant,
                               unsigned row, unsigned col)
{
	row;
	col;
	variant;
	return true;
}

void vvHistoryModel::GetValueByRow(
	wxVariant&            cValue,
	unsigned int 		  row,
	unsigned int 		  uColumn
	) const
{
	row;
	vvVerbs::HistoryEntry entry;
	//if (static_cast<Column>(uColumn) == COLUMN_B_CHECKED)
	//	cValue = "false";
	//if (m_bAlreadyFetching) //We're fetching data right now.
	//	return;
	{
		vvHistoryModel* nonConst = (vvHistoryModel*)this;
		wxCriticalSectionLocker lock(nonConst->m_dataAccessLock);
		if (row >= this->mcDataItems.size())
			return;
		 entry = this->mcDataItems[row];
	}
	if (row >= this->nEndOfPage)
	{
		bool bFetchingCopy = false;
		{
			//wxCriticalSectionLocker lock(m_fetchCheckCriticalSection);
			bFetchingCopy = m_bAlreadyFetching ;
		}
		if (bFetchingCopy == false)
			mpListControl->StartGetMoreData();
	}
	wxString rev;
	switch (static_cast<Column>(uColumn))
	{
	//case COLUMN_B_CHECKED:
		//Todo: Jeremy should fix this.
		//cValue = true;
		//return;

	case COLUMN_S_REVISION:
		rev << entry.nRevno;
		rev << ":";
		rev << entry.sChangesetHID;
		cValue = rev;
		return;

	case COLUMN_S_COMMENT:
		if (entry.cComments.size() > 0)
		{
			wxString comment = entry.cComments[0].sCommentText;
			comment.Replace("\r", " ", true);
			comment.Replace("\n", " ", true);
			cValue = comment;
		}
		return;

	case COLUMN_S_BRANCH:
		for (unsigned int i = 0; i < entry.cBranches.size(); i++)
		{
			if (i > 0)
				rev << ", ";
			rev << entry.cBranches[i];
		}
		cValue = rev;
		return;

	case COLUMN_S_STAMP:
		for (unsigned int i = 0; i < entry.cStamps.size(); i++)
		{
			if (i > 0)
				rev << ", ";
			rev << entry.cStamps[i];
		}
		cValue = rev;
		return;

	case COLUMN_S_TAG:
		for (unsigned int i = 0; i < entry.cTags.size(); i++)
		{
			if (i > 0)
				rev << ", ";
			rev << entry.cTags[i];
		}
		cValue = rev;
		return;

	case COLUMN_U_DATE:
		if (entry.cAudits.size() > 0)
			cValue = entry.cAudits[0].nWhen;
		return;

	case COLUMN_S_USER:
			if (entry.cAudits.size() > 0)
				cValue = entry.cAudits[0].sWho;
			return;
	default:
		wxLogError("Unknown StampsModel column: %u", uColumn);
		cValue.MakeNull();
		return;
	}
}


vvHistoryListControl::vvHistoryListControl()
{
}

vvHistoryListControl::vvHistoryListControl(
	vvVerbs& cVerbs,
	vvContext& cContext,
	wxString& sRepoName,
	wxWindow*          pParent,
	wxWindowID         cId,
	const wxPoint&     cPosition,
	const wxSize&      cSize,
	long               iStyle,
	const wxValidator& cValidator
	) : wxPanel(pParent, cId,cPosition, cSize, iStyle, wxT("History List Control") )
{
	cValidator;
	cVerbs;
	&cContext;

	m_list = NULL;
	wxBoxSizer * sizer = new wxBoxSizer(wxHORIZONTAL);
	m_list = new wxDataViewListCtrl(this,-1,wxDefaultPosition,cSize,
			wxDV_HORIZ_RULES | wxDV_SINGLE);
	sizer->Add(m_list, 1, wxEXPAND);
	sizer->SetSizeHints(this);
	wxPanel::SetSizer(sizer);

	m_pVerbs = &cVerbs;
	m_pContext = &cContext;
	m_sRepoName = sRepoName;
}

bool vvHistoryListControl::Create(
	wxWindow*          pParent,
	wxWindowID         cId,
	const wxPoint&     cPosition,
	const wxSize&      cSize,
	long               iStyle,
	const wxValidator&  cValidator
	)
{
	cValidator;
	// create our parent
	if (!wxPanel::Create(pParent, cId, cPosition, cSize, iStyle))
	{
		return false;
	}

	// done
	return true;
}

void vvHistoryListControl::LoadHistory(
	vvVerbs::HistoryData data
	)
{
	((vvHistoryModel*)m_list->GetModel())->SetData(data);
}

void vvHistoryListControl::LoadHistory(
		vvHistoryFilterOptions& vvHistoryFilter
		)
{
	{
		
		m_historyFilter = vvHistoryFilter;
		((vvHistoryModel*)m_list->GetModel())->SetFilter(vvHistoryFilter);
		((vvHistoryModel*)m_list->GetModel())->Clear();
	}
	StartGetMoreData();
}

void vvHistoryListControl::SendStatusMessage(wxString sStatusString, bool bHasMoreData)
{
	wxCommandEvent ev(newEVT_HISTORY_STATUS, GetId() );

	ev.SetString(sStatusString);
	ev.SetInt(bHasMoreData);
	ev.SetEventObject(this);
	this->ProcessEvent(ev);
}

void vvHistoryListControl::StartGetMoreData()
{
	// we want to start a long task, but we don't want our GUI to block
    // while it's executed, so we use a thread to do it.
	if (GetThread() == NULL || !GetThread()->IsAlive())
		if ( CreateThread(wxTHREAD_JOINABLE) != wxTHREAD_NO_ERROR)
		{
			wxLogError("Could not create the worker thread!");
			return;
		}


    // go!
	if (!GetThread()->IsRunning())
		GetThread()->Run();
}

wxThread::ExitCode vvHistoryListControl::Entry()
{
    // IMPORTANT:
    // this function gets executed in the secondary thread context!
	try
	{
		if (((vvHistoryModel*)m_list->GetModel())->GetCount() == 0)
		{
			((vvHistoryModel*)m_list->GetModel())->LoadHistory();
		}
		else
			((vvHistoryModel*)m_list->GetModel())->GetMoreData();
	}
	catch (std::exception e)
	{
		int tmp = 4;
		tmp++;
	}


    return (wxThread::ExitCode)0;
}

wxArrayString vvHistoryListControl::GetSelection()
{
	wxArrayString returnVal;
	wxDataViewItem sel = m_list->GetSelection();
	if (sel.IsOk())
	{
		int row = ((vvHistoryModel*)m_list->GetModel())->GetRow(sel);
		returnVal.Add(((vvHistoryModel*)m_list->GetModel())->GetHistoryDataAt(row).sChangesetHID);
	}
  
	return returnVal;
}

void SetColumnWidth(wxDataViewColumn * column, int width)
{
	if (width > 0)
		column->SetWidth(width);
}

void vvHistoryListControl::ApplyColumnPreferences(vvHistoryColumnPreferences& columnPrefs)
{
	if (m_list->GetColumnCount() == 0)
	{
		const int columnCount = 7;
		unsigned int model_column = 0;
		wxDataViewColumn* columnArray[columnCount];
	
		//columnArray[model_column++] = m_toggleColumn = new wxDataViewColumn( "", new wxDataViewToggleRenderer( wxT("bool"), wxDATAVIEW_CELL_ACTIVATABLE), model_column, 20);
		columnArray[model_column++] = m_revisionColumn = new wxDataViewColumn( "Revision", new wxDataViewTextRenderer( wxT("string"), wxDATAVIEW_CELL_ACTIVATABLE), model_column, 80, wxALIGN_LEFT, wxCOL_REORDERABLE | wxCOL_RESIZABLE);
		columnArray[model_column++] = m_commentColumn = new wxDataViewColumn( "Comment", new wxDataViewTextRenderer( wxT("string"), wxDATAVIEW_CELL_ACTIVATABLE), model_column, 80, wxALIGN_LEFT, wxCOL_REORDERABLE | wxCOL_RESIZABLE);
		columnArray[model_column++] = m_branchColumn = new wxDataViewColumn( "Branch", new wxDataViewTextRenderer( wxT("string"), wxDATAVIEW_CELL_ACTIVATABLE), model_column, 60, wxALIGN_LEFT, wxCOL_REORDERABLE | wxCOL_RESIZABLE);
		columnArray[model_column++] = m_stampColumn = new wxDataViewColumn( "Stamp", new wxDataViewTextRenderer( wxT("string"), wxDATAVIEW_CELL_ACTIVATABLE), model_column, 60, wxALIGN_LEFT, wxCOL_REORDERABLE | wxCOL_RESIZABLE);
		columnArray[model_column++] = m_tagColumn = new wxDataViewColumn( "Tag", new wxDataViewTextRenderer( wxT("string"), wxDATAVIEW_CELL_ACTIVATABLE), model_column, 60, wxALIGN_LEFT, wxCOL_REORDERABLE | wxCOL_RESIZABLE);
		columnArray[model_column++] = m_dateColumn = new wxDataViewColumn( "Date", new wxDataViewTextRenderer( wxT("datetime"), wxDATAVIEW_CELL_ACTIVATABLE), model_column, 60, wxALIGN_LEFT, wxCOL_REORDERABLE | wxCOL_RESIZABLE);
		columnArray[model_column++] = m_userColumn = new wxDataViewColumn( "User", new wxDataViewTextRenderer( wxT("string"), wxDATAVIEW_CELL_ACTIVATABLE), model_column, 80, wxALIGN_LEFT, wxCOL_REORDERABLE | wxCOL_RESIZABLE);

		if (columnPrefs.columnOrder.size() != columnCount)
		{
			for (int i=0; i < columnCount; i++)
				m_list->AppendColumn(columnArray[i]);
		}
		else
		{
			//The column order array contains numbers, which correspond to "where should I put the 
			//column which would normally go at the Nth position?"  For example, the array "0, 3, 2 1" means
			//that the revision column (usually column 1), should go at position 3 instead.
			//the comment column (usually column 2), should go at position 2 as normal.
			//the branch column (usually column 3), should go at position 1 instead.
			for (int i=0; i < columnCount; i++)
				for (int j=0; j < columnCount; j++)
				{
					if (columnPrefs.columnOrder[j] == i)
						m_list->AppendColumn(columnArray[j]);
				}
		}
		
		// create our model
		this->mpDataModel = new vvHistoryModel(*m_pVerbs, *m_pContext, this, m_sRepoName);
		m_list->AssociateModel(this->mpDataModel.get());

		SetColumnWidth(m_revisionColumn, columnPrefs.WidthRevisionColumn);
		SetColumnWidth(m_commentColumn, columnPrefs.WidthCommentColumn);
		SetColumnWidth(m_branchColumn, columnPrefs.WidthBranchColumn);
		SetColumnWidth(m_stampColumn, columnPrefs.WidthStampColumn);
		SetColumnWidth(m_tagColumn, columnPrefs.WidthTagColumn);
		SetColumnWidth(m_dateColumn, columnPrefs.WidthDateColumn);
		SetColumnWidth(m_userColumn, columnPrefs.WidthUserColumn);
	}

	//Just do the hidden/shown settings 
	m_revisionColumn->SetHidden(!columnPrefs.ShowRevisionColumn);
	m_commentColumn->SetHidden(!columnPrefs.ShowCommentColumn);
	m_branchColumn->SetHidden(!columnPrefs.ShowBranchColumn);
	m_stampColumn->SetHidden(!columnPrefs.ShowStampColumn);
	m_tagColumn->SetHidden(!columnPrefs.ShowTagColumn);
	m_dateColumn->SetHidden(!columnPrefs.ShowDateColumn);
	m_userColumn->SetHidden(!columnPrefs.ShowUserColumn);
}

void vvHistoryListControl::GetColumnSettings(vvHistoryColumnPreferences& columnPrefs)
{
	columnPrefs.WidthRevisionColumn = m_revisionColumn->GetWidth();
	columnPrefs.WidthCommentColumn	= m_commentColumn->GetWidth();
	columnPrefs.WidthBranchColumn	= m_branchColumn->GetWidth();
	columnPrefs.WidthStampColumn	= m_stampColumn->GetWidth();
	columnPrefs.WidthTagColumn		= m_tagColumn->GetWidth();
	columnPrefs.WidthDateColumn		= m_dateColumn->GetWidth();
	columnPrefs.WidthUserColumn		= m_userColumn->GetWidth();
	columnPrefs.columnOrder.clear();
	//columnPrefs.columnOrder.Add(m_list->GetColumnPosition(m_toggleColumn));
	columnPrefs.columnOrder.Add(m_list->GetColumnPosition(m_revisionColumn));
	columnPrefs.columnOrder.Add(m_list->GetColumnPosition(m_commentColumn));
	columnPrefs.columnOrder.Add(m_list->GetColumnPosition(m_branchColumn));
	columnPrefs.columnOrder.Add(m_list->GetColumnPosition(m_stampColumn));
	columnPrefs.columnOrder.Add(m_list->GetColumnPosition(m_tagColumn));
	columnPrefs.columnOrder.Add(m_list->GetColumnPosition(m_dateColumn));
	columnPrefs.columnOrder.Add(m_list->GetColumnPosition(m_userColumn));
	
}

wxDEFINE_EVENT(wxEVT_COMMAND_MODEL_FETCH_MORE__COMPLETE, wxThreadEvent);
IMPLEMENT_DYNAMIC_CLASS(vvHistoryListControl, wxDataViewCtrl);
BEGIN_EVENT_TABLE(vvHistoryListControl, wxPanel)
END_EVENT_TABLE()
