/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include "precompiled.h"
#include "vvProgressExecutor.h"

#include "vvCommand.h"
#include "vvContext.h"
#include "vvProgressDialog.h"


/*
**
** Public Functions
**
*/

vvProgressExecutor::vvProgressExecutor(
	vvCommand& cCommand
	)
	: mpCommand(&cCommand)
	, mcMessages()
	, mwProgress(NULL)
{
	// create our background thread and start it
	wxThreadError eThreadError = this->CreateThread(wxTHREAD_JOINABLE);
	if (eThreadError != wxTHREAD_NO_ERROR)
	{
		wxLogError("Error creating background command execution thread: %d", eThreadError);
		return;
	}
	eThreadError = this->GetThread()->Run();
	if (eThreadError != wxTHREAD_NO_ERROR)
	{
		wxLogError("Error starting background command execution thread: %d", eThreadError);
		return;
	}
}

vvProgressExecutor::~vvProgressExecutor()
{
	// clear the message queue
	this->mcMessages.Clear();

	// send an exit message to the background thread
	this->mcMessages.Post(MESSAGE_EXIT);

	// wait until the thread exits
	this->GetThread()->Wait();
}

bool vvProgressExecutor::IsOk() const
{
	wxThread* pThread = this->GetThread();
	return pThread != NULL && pThread->IsAlive() && pThread->IsRunning();
}

bool vvProgressExecutor::Execute(
	wxWindow* wParent
	)
{
	wxASSERT(wxThread::IsMain());
	wxASSERT(this->IsOk());

	// create a progress dialog to display the execution's progress to the user
	vvProgressDialog cDialog(wParent);

	// store the progress dialog to a member variable
	// so that we'll have access to it from the other thread
	// this being non-NULL also signifies that execution is in progress
	this->mwProgress = &cDialog;

	// send a message to the background thread to execute the command
	this->mcMessages.Post(MESSAGE_EXECUTE);

	// display the progress dialog
	// it's modal so that the user can't interact with anything while the command executes
	// this will be closed either by the background thread or by the user clicking OK/Cancel
	int iResult = cDialog.ShowModal();

	// blank out the progress dialog member
	// no execution in progress anymore
	this->mwProgress = NULL;

	// return the result
	return (iResult == wxID_OK);
}

wxThread::ExitCode vvProgressExecutor::Entry()
{
	ThreadMessage eMessage;
	vvContext     cContext;

	// wait for a message from the main thread
	while (this->mcMessages.Receive(eMessage) == wxMSGQUEUE_NO_ERROR)
	{
		// check for status changes in the thread
		if (this->GetThread()->TestDestroy())
		{
			break;
		}

		// if this is an exit message, then exit
		if (eMessage == MESSAGE_EXIT)
		{
			break;
		}

		// otherwise, assume we have an execute message
		// since no other message types exist right now
		wxASSERT(eMessage == MESSAGE_EXECUTE);

		// switch the command's context to the one for this thread
		// we'll restore its old one after we're finished
		vvContext* pOldContext = this->mpCommand->GetContext();
		this->mpCommand->SetContext(&cContext);

		// register the current progress dialog to receive sglib progress
		wxASSERT(this->mwProgress != NULL);
		if (!this->mwProgress->RegisterSgLogHandler(cContext))
		{
			this->mwProgress->SetReturnCode(wxID_CANCEL);
			wxLogError("Error registering progress dialog as sglib logger: %s", cContext.Error_GetMessage());
			cContext.Error_Reset();
			continue;
		}

		// execute the command
		bool bSuccess = this->mpCommand->Execute();
		wxASSERT_MSG(!(bSuccess && cContext.Error_Check()), "Command execution succeeded, but left the context in an error state.");
		if (!bSuccess)
		{
			this->mwProgress->SetReturnCode(wxID_CANCEL);
			if (cContext.Error_Check())
			{
				wxLogError(cContext.Error_GetMessage());
				cContext.Error_Reset();
			}
			else
			{
				wxLogError("Command execution failed.  There may be reported and/or logged errors with more information.");
			}
		}

		// done receiving progress for the dialog
		if (!this->mwProgress->UnregisterSgLogHandler(cContext))
		{
			cContext.Error_Reset();
		}

		// restore the command's old context
		this->mpCommand->SetContext(pOldContext);

		// finished successfully
		this->mwProgress->SetReturnCode(wxID_OK);
	}

	// return success
	return (wxThread::ExitCode)0;
}
