/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include "precompiled.h"
#include "vvSgHelpers.h"

#include "sg.h"
#include "vvContext.h"


/*
**
** Public Functions
**
*/

bool vvSgHelpers::Convert_sgBool_cppBool(
	int bValue
	)
{
	wxCOMPILE_TIME_ASSERT(sizeof(bValue) == sizeof(SG_bool), SgBoolSizeMismatch);
	wxASSERT(bValue == SG_TRUE || bValue == SG_FALSE);
	return bValue == SG_FALSE ? false : true;
}

int vvSgHelpers::Convert_cppBool_sgBool(
	bool bValue
	)
{
	wxCOMPILE_TIME_ASSERT(sizeof(int) == sizeof(SG_bool), SgBoolSizeMismatch);
	return bValue ? SG_TRUE : SG_FALSE;
}

wxString vvSgHelpers::Convert_sgString_wxString(
	const SG_string* sValue
	)
{
	return vvSgHelpers::Convert_cString_wxString(SG_string__sz(sValue));
}

wxString vvSgHelpers::Convert_cString_wxString(
	const char* szValue
	)
{
	wxString sValue(szValue);

	// wx seems to expect strings in memory to use a platform-neutral newline.
	// It is converted to a platform-specific one as necessary (like when written to disk).
	sValue.Replace(SG_PLATFORM_NATIVE_EOL_STR, "\n", true);

	return sValue;
}

wxString vvSgHelpers::Convert_sgPathname_wxString(
	const SG_pathname* sValue
	)
{
	// run it through wxFileName to normalize the path to the current platform
	return wxFileName(vvSgHelpers::Convert_cString_wxString(SG_pathname__sz(sValue))).GetFullPath();
}

SG_pathname* vvSgHelpers::Convert_wxString_sgPathname(
	vvContext&      pCtx,
	const wxString& sValue
	)
{
	SG_pathname* pValue = NULL;
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pValue, sValue.c_str())  );

fail:
	return pValue;
}

void vvSgHelpers::Convert_wxArrayString_cStringArrayCount(
	vvContext&           pCtx,
	const wxArrayString& cStrings,
	const char***        pppStrings,
	wxUint32*            pCount
	)
{
	const char** ppStrings = NULL;
	
	SG_ERR_CHECK(  SG_alloc(pCtx, sizeof(const char*), cStrings.size(), &ppStrings)  );

	for (unsigned int uIndex = 0u; uIndex < cStrings.size(); ++uIndex)
	{
		ppStrings[uIndex] = cStrings[uIndex].c_str();
	}

	if (pppStrings != NULL)
	{
		*pppStrings = ppStrings;
		ppStrings = NULL;
	}

	if (pCount != NULL)
	{
		*pCount = cStrings.size();
	}

fail:
	SG_NULLFREE(pCtx, ppStrings);
	return;
}


void vvSgHelpers::Convert_listStampData_wxArrayString(
	vvContext& pCtx,
	vvVerbs::stlStampDataList stampData,
	wxArrayString& cStrings
	)
{
	pCtx;
	for (std::list<vvVerbs::StampData>::const_iterator it = stampData.begin(); it != stampData.end(); ++it)
	{
		cStrings.Add((*it).sName);
	}
	return;
}

void vvSgHelpers::Convert_listTagData_wxArrayString(
	vvContext& pCtx,
	vvVerbs::stlTagDataList tagData,
	wxArrayString& cStrings
	)
{
	pCtx;
	for (std::list<vvVerbs::TagData>::const_iterator it = tagData.begin(); it != tagData.end(); ++it)
	{
		cStrings.Add((*it).sName);
	}
	return;
}

void vvSgHelpers::Convert_wxArrayString_sgStringArray(
	vvContext& pCtx,
	const wxArrayString& cStrings,
	SG_stringarray**     ppStrings
	)
{
	SG_stringarray* pStrings = NULL;

	SG_ERR_CHECK(  SG_STRINGARRAY__ALLOC(pCtx, &pStrings, cStrings.size())  );
	for (wxArrayString::const_iterator it = cStrings.begin(); it != cStrings.end(); ++it)
	{
		SG_ERR_CHECK(  SG_stringarray__add(pCtx, pStrings, it->c_str())  );
	}

	if (ppStrings != NULL)
	{
		*ppStrings = pStrings;
		pStrings = NULL;
	}

fail:
	SG_STRINGARRAY_NULLFREE(pCtx, pStrings);
	return;
}

void vvSgHelpers::Convert_wxArrayString_sgRBTree(
	vvContext& pCtx,
	const wxArrayString& cStrings,
	SG_rbtree**     pprbtree,
	bool bNullIsOK
	)
{
	SG_rbtree* prb = NULL;

	if (cStrings.Count() == 0 && bNullIsOK)
		return;
	SG_ERR_CHECK(  SG_RBTREE__ALLOC(pCtx, &prb)  );
	for (wxArrayString::const_iterator it = cStrings.begin(); it != cStrings.end(); ++it)
	{
		SG_ERR_CHECK(  SG_rbtree__add__with_assoc(pCtx, prb, it->c_str(), NULL)  );
	}

	if (pprbtree != NULL)
	{
		*pprbtree = prb;
		prb = NULL;
	}

fail:
	SG_RBTREE_NULLFREE(pCtx, prb);
	return;
}

void vvSgHelpers::Convert_wxArrayString_sgVhash(
	vvContext& pCtx,
	const wxArrayString& cStrings,
	SG_vhash**     ppvhash,
	bool bNullIsOK
	)
{
	SG_vhash* pvh = NULL;

	if (cStrings.Count() == 0 && bNullIsOK)
		return;
	SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvh)  );
	for (wxArrayString::const_iterator it = cStrings.begin(); it != cStrings.end(); ++it)
	{
		SG_ERR_CHECK(  SG_vhash__update__null(pCtx, pvh, it->c_str())  );
	}

	if (ppvhash != NULL)
	{
		*ppvhash = pvh;
		pvh = NULL;
	}

fail:
	SG_VHASH_NULLFREE(pCtx, pvh);
	return;
}
void vvSgHelpers::Convert_sgVarray_wxArrayInt(
	vvContext& pCtx,
	const void*     pvarray,
	wxArrayInt& cInts
	)
{
	wxArrayInt returnVal;
	SG_uint32 varrayCount = 0;
	SG_int64 value = 0;
	if (pvarray)
	{
		SG_ERR_CHECK(  SG_varray__count(pCtx, (const SG_varray *)pvarray, &varrayCount)  );
		for (SG_uint32 i = 0; i < varrayCount; i++)
		{
			SG_ERR_CHECK(  SG_varray__get__int64(pCtx, (const SG_varray *)pvarray, i, &value)  );
			returnVal.Add(value);
		}
	}
fail:
	cInts = returnVal;
}

void vvSgHelpers::Convert_wxArrayInt_sgVarray(
	vvContext& pCtx,
	const wxArrayInt& cInts,
	void**     ppvarray,
	bool bNullIsOK
	)
{
	SG_varray* pva = NULL;

	if (cInts.Count() == 0 && bNullIsOK)
		return;
	SG_ERR_CHECK(  SG_VARRAY__ALLOC(pCtx, &pva)  );
	for (wxArrayInt::const_iterator it = cInts.begin(); it != cInts.end(); ++it)
	{
		SG_ERR_CHECK(  SG_varray__append__int64(pCtx, pva, *it)  );
	}

	if (ppvarray != NULL)
	{
		*ppvarray = (void*)pva;
		pva = NULL;
	}

fail:
	SG_VARRAY_NULLFREE(pCtx, pva);
	return;
}
wxDateTime vvSgHelpers::Convert_sgTime_wxDateTime(
	wxInt64 iValue
	)
{
	wxCOMPILE_TIME_ASSERT(sizeof(iValue) == sizeof(SG_int64), SgTimeSizeMismatch);
	wxDateTime cWxTime;
	cWxTime.Set(iValue / 1000u);
	cWxTime.SetMillisecond(iValue % 1000u);
	wxASSERT(cWxTime.IsValid());
	return cWxTime;
}



vvVerbs::HistoryData vvSgHelpers::Convert_sg_history_result_HistoryData(
	class vvContext& pCtx,
	SG_history_result * phr_result
	)
{
	vvVerbs::HistoryData cHistoryData;
	//print the information for each
	SG_bool bFound = (phr_result != NULL);
	const char* currentInfoItem = NULL;
	SG_uint32 revno;
	SG_uint32 nCount = 0;
	SG_uint32 nIndex = 0;
	const char * pszTag = NULL;
	const char * pszComment = NULL;
	const char * pszStamp = NULL;
	const char * pszParent = NULL;
	SG_uint32 nResultCount = 0;
    SG_vhash* pvh_values = NULL;

	/*
    if (pvh_pile)
    {
        SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvh_pile, "values", &pvh_values)  );
    } */

	if (phr_result == NULL)
		return cHistoryData;
	SG_ERR_CHECK(  SG_history_result__count(pCtx, phr_result, &nResultCount)  );
	while (nResultCount != 0 && bFound)
	{
		vvVerbs::HistoryEntry he;
        SG_vhash* pvh_branch_names = NULL;

		SG_ERR_CHECK(  SG_history_result__get_revno(pCtx, phr_result, &revno)  );
		SG_ERR_CHECK(  SG_history_result__get_cshid(pCtx, phr_result, &currentInfoItem)  );
		he.nRevno = revno;
		he.sChangesetHID = currentInfoItem;

        if (pvh_values)
        {
            SG_bool b_has = SG_FALSE;

            SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh_values, currentInfoItem, &b_has)  );
            if (b_has)
            {
                SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvh_values, currentInfoItem, &pvh_branch_names)  );
            }
        }

        if (pvh_branch_names)
        {
            SG_uint32 count_branch_names = 0;
            SG_uint32 i;

            SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_branch_names, &count_branch_names)  );
            for (i=0; i<count_branch_names; i++)
            {
                const char* psz_branch_name = NULL;

                SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvh_branch_names, i, &psz_branch_name, NULL)  );
				he.cBranches.Add(psz_branch_name);
            }
        }

		SG_ERR_CHECK(  SG_history_result__get_audit__count(pCtx, phr_result, &nCount)  );
		for (nIndex = 0; nIndex < nCount; nIndex++)
		{
			SG_int64 itime = -1;
			const char * pszUser = NULL;
			vvVerbs::AuditEntry ae;

			SG_ERR_CHECK(  SG_history_result__get_audit__who(pCtx, phr_result, nIndex, &pszUser)  );
			SG_ERR_CHECK(  SG_history_result__get_audit__when(pCtx, phr_result, nIndex, &itime)  );
			ae.sWho = pszUser;
			ae.nWhen = wxDateTime((time_t)(itime / 1000));
			he.cAudits.push_back(ae);
		}

		SG_ERR_CHECK(  SG_history_result__get_tag__count(pCtx, phr_result, &nCount)  );
		for (nIndex = 0; nIndex < nCount; nIndex++)
		{
			SG_ERR_CHECK(  SG_history_result__get_tag__text(pCtx, phr_result, nIndex, &pszTag)  );
			he.cTags.Add(pszTag);
		}

		SG_ERR_CHECK(  SG_history_result__get_comment__count(pCtx, phr_result, &nCount)  );
		for (nIndex = 0; nIndex < nCount; nIndex++)
		{
			SG_int64 itime = -1;
			const char * pszUser = NULL;
			vvVerbs::AuditEntry ae;
			vvVerbs::CommentEntry ce;
			SG_uint32 nHistoryCount = 0;
			SG_uint32 nCommentHistoryIndex = 0;
			SG_uint32 nCommentAuditCount = 0;
			SG_uint32 nCommentAuditIndex = 0;
			SG_varray * pvaCommentHistory = NULL;
			SG_vhash * pvhThisCommentHistoryItem = NULL;
			SG_varray * pvaCommentAudits = NULL;
			SG_vhash * pvhThisCommentHistoryAudit = NULL;

			SG_ERR_CHECK(  SG_history_result__get_comment__text(pCtx, phr_result, nIndex, &pszComment)  );
			SG_ERR_CHECK(  SG_history_result__get_comment__history__ref(pCtx, phr_result, nIndex, &pvaCommentHistory)  );

			SG_ERR_CHECK(  SG_varray__count(pCtx, pvaCommentHistory, &nHistoryCount)  );
			for (nCommentHistoryIndex = 0; nCommentHistoryIndex < nHistoryCount; nCommentHistoryIndex++)
			{
				SG_ERR_CHECK(  SG_varray__get__vhash(pCtx, pvaCommentHistory, nCommentHistoryIndex, &pvhThisCommentHistoryItem)  );
				SG_ERR_CHECK(  SG_vhash__get__varray(pCtx, pvhThisCommentHistoryItem, "audits", &pvaCommentAudits)  );
				SG_ERR_CHECK(  SG_varray__count(pCtx, pvaCommentAudits, &nCommentAuditCount)  );
				for (nCommentAuditIndex = 0; nCommentAuditIndex < nCommentAuditCount; nCommentAuditIndex++)
				{
					SG_ERR_CHECK(  SG_varray__get__vhash(pCtx, pvaCommentAudits, nCommentAuditIndex, &pvhThisCommentHistoryAudit)  );
					SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhThisCommentHistoryAudit, "username", &pszUser)  );
					SG_ERR_CHECK(  SG_vhash__get__int64_or_double(pCtx, pvhThisCommentHistoryAudit, "timestamp", &itime)  );
					ae.sWho = pszUser;
					ae.nWhen = wxDateTime((time_t)(itime / 1000));
					ce.cAudits.push_back(ae);
				}
			}
			ce.sCommentText = pszComment;
			he.cComments.push_back(ce);
		}

		SG_ERR_CHECK(  SG_history_result__get_stamp__count(pCtx, phr_result, &nCount)  );
		for (nIndex = 0; nIndex < nCount; nIndex++)
		{
			SG_ERR_CHECK(  SG_history_result__get_stamp__text(pCtx, phr_result, nIndex, &pszStamp)  );
			he.cStamps.Add(pszStamp);
		}

		SG_ERR_CHECK(  SG_history_result__get_parent__count(pCtx, phr_result, &nCount)  );
		for (nIndex = 0; nIndex < nCount; nIndex++)
		{
			vvVerbs::RevisionEntry pe;
			SG_ERR_CHECK(  SG_history_result__get_parent(pCtx, phr_result, nIndex, &pszParent, &revno)  );
			pe.sChangesetHID = pszParent;
			pe.nRevno = revno;
			he.cParents.push_back(pe);
		}
		cHistoryData.push_back(he);
		SG_ERR_CHECK(  SG_history_result__next(pCtx, phr_result, &bFound)  );
	}
fail:
	return cHistoryData;
}

void vvSgHelpers::Add_int__to_vhash_if_int_is_positive(vvContext& pCtx, SG_vhash * pvh, const char* setting, int number)
{
	if (number > 0)
		SG_ERR_CHECK(  SG_vhash__update__int64(pCtx, pvh, setting, number)  ); 
fail:
	;
}

void vvSgHelpers::Convert_wxArrayStrings_sgRevSpec(
	vvContext& pCtx,
	const wxArrayString& cRevs,
	const wxArrayString& cTags,
	const wxArrayString& cBranches,
	SG_rev_spec** ppRevSpec)
{
	SG_rev_spec* pRevSpec = NULL;

	SG_ERR_CHECK(  SG_REV_SPEC__ALLOC(pCtx, &pRevSpec)  );

	for (wxArrayString::const_iterator it = cRevs.begin(); it != cRevs.end(); ++it)
		SG_ERR_CHECK(  SG_rev_spec__add_rev(pCtx, pRevSpec, it->c_str())  );
	for (wxArrayString::const_iterator it = cTags.begin(); it != cTags.end(); ++it)
		SG_ERR_CHECK(  SG_rev_spec__add_tag(pCtx, pRevSpec, it->c_str())  );
	for (wxArrayString::const_iterator it = cBranches.begin(); it != cBranches.end(); ++it)
		SG_ERR_CHECK(  SG_rev_spec__add_branch(pCtx, pRevSpec, it->c_str())  );

	if (ppRevSpec)
	{
		*ppRevSpec = pRevSpec;
		pRevSpec = NULL;
	}

fail:
	SG_REV_SPEC_NULLFREE(pCtx, pRevSpec);
}

#if defined(DEBUG)

wxString vvSgHelpers::Dump_sgVhash(
	vvContext&      pCtx,
	const SG_vhash* pValue
	)
{
	wxString   sValue  = wxEmptyString;
	SG_string* pString = NULL;

	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pString)  );
	SG_ERR_CHECK(  SG_vhash_debug__dump(pCtx, pValue, pString)  );
	sValue = vvSgHelpers::Convert_sgString_wxString(pString);

fail:
	SG_STRING_NULLFREE(pCtx, pString);
	return sValue;
}

#endif