/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#ifndef VV_SG_HELPERS_H
#define VV_SG_HELPERS_H

#include "precompiled.h"
#include "vvVerbs.h"
#include "sg.h"

/*
 * This is a collection of general helper functions for working with sglib.
 */
class vvSgHelpers
{
// static functionality
public:
	/**
	 * Converts from an SG_bool to a C++ bool.
	 */
	static bool Convert_sgBool_cppBool(
		int bValue //< [in] The SG_bool value to convert.
		);

	/**
	 * Converts from a C++ bool to an SG_bool.
	 */
	static int Convert_cppBool_sgBool(
		bool bValue //< [in] The bool value to convert.
		);

	/**
	 * Converts from an SG_string to a wxString.
	 */
	static wxString Convert_sgString_wxString(
		const struct _SG_string* sValue //< [in] The SG_string value to convert.
		);

	/**
	 * Converts from an sglib-generated const char* to a wxString.
	 */
	static wxString Convert_cString_wxString(
		const char* szValue //< [in] The const char* value to convert.
		);

	/**
	 * Converts from an SG_pathname to a wxString.
	 */
	static wxString Convert_sgPathname_wxString(
		const struct _SG_pathname* sValue //< [in] The SG_pathname value to convert.
		);

	/**
	 * Converts from a wxString to an SG_pathname.
	 * Caller owns the returned pointer.
	 */
	static struct _SG_pathname* Convert_wxString_sgPathname(
		class vvContext& cContext, //< [in] [out] Error and context info.
		const wxString&  sValue    //< [in] The wxString value to convert.
		);

	/**
	 *Converts a list of stampdata to a wxArrayString.
	 */
	static void vvSgHelpers::Convert_listStampData_wxArrayString(
		vvContext& pCtx,
		vvVerbs::stlStampDataList stampData,
		wxArrayString& cStrings
		);
	
	/**
	 *Converts a list of tagdata to a wxArrayString.
	 */
	static void vvSgHelpers::Convert_listTagData_wxArrayString(
		vvContext& pCtx,
		vvVerbs::stlTagDataList tagData,
		wxArrayString& cStrings
		);
	/**
	 * Converts from a wxArrayString to a const char** and size/count.
	 */
	static void Convert_wxArrayString_cStringArrayCount(
		class vvContext&     cContext,   //< [in] [out] Error and context info.
		const wxArrayString& cStrings,   //< [in] The string array to convert.
		const char***        pppStrings, //< [out] The converted string array.
		                                 //<       Caller owns the array pointer (which should be freed with SG_NULLFREE),
		                                 //<       but the individual string pointers are owned by the given wxArrayString.
		wxUint32*            pCount      //< [out] Size of the returned array (SG_int32).
		);

	/**
	 * Converts from a wxArrayString to an SG_stringarray.
	 */
	static void Convert_wxArrayString_sgStringArray(
		class vvContext&         cContext, //< [in] [out] Error and context info.
		const wxArrayString&     cStrings, //< [in] The string array to convert.
		struct _sg_stringarray** ppStrings //< [out] The converted string array.
		                                   //<       Caller owns this pointer.
		);

		/**
	 * Converts from a wxArrayString to an SG_stringarray.
	 */
	static void Convert_wxArrayString_sgRBTree(
		class vvContext&         cContext, //< [in] [out] Error and context info.
		const wxArrayString&     cStrings, //< [in] The string array to convert.
		struct _sg_rbtree** pprbTree,		//< [out] The converted rbtree
											//<       Caller owns this pointer.
		bool bNullIsOK = false
		                                   
		);


			/**
	 * Converts from a wxArrayString to an SG_stringarray.
	 */
	static void Convert_wxArrayString_sgVhash(
		class vvContext&         cContext, //< [in] [out] Error and context info.
		const wxArrayString&     cStrings, //< [in] The string array to convert.
		struct _SG_vhash** ppvhash, //< [out] The converted vhash.
		                                   //<       Caller owns this pointer.
		bool bNullIsOK = false
		);

	static void Convert_wxArrayInt_sgVarray(
		vvContext& pCtx,
		const wxArrayInt& cInts,
		void**     ppvarray,
		bool bNullIsOK
		);

	static void Convert_sgVarray_wxArrayInt(
		vvContext& pCtx,
		const void*     pvarray,
		wxArrayInt& cInts
		);
	/**
	 * Converts from an sglib time to a wxDateTime.
	 * The sglib time is expected to be milliseconds since 1/1/1970 00:00:00.000 UTC.
	 */
	static wxDateTime Convert_sgTime_wxDateTime(
		wxInt64 iValue //< The SG_int64 time value to convert.
		);
	/**
	 * Converts from SG_history_result to HistoryData
	 **/
	static vvVerbs::HistoryData Convert_sg_history_result_HistoryData(
		class vvContext& pCtx,
		struct _opaque_history_result * phr_result
		);
	/**
	 * Does what it says.
	 */
	static void Add_int__to_vhash_if_int_is_positive(vvContext& pCtx, _SG_vhash * pvh, const char* setting, int number);

	/**
	 * Until there's a wrapper class for SG_rev_spec, this routine converts arrays of revs, tags, and branches to an SG_rev_spec.
	 */
	static void Convert_wxArrayStrings_sgRevSpec(
		vvContext& pCtx,
		const wxArrayString& cRevs,
		const wxArrayString& cTags,
		const wxArrayString& cBranches,
		SG_rev_spec** ppRevSpec
		);

#if defined(DEBUG)

	/**
	 * Dumps a vhash to a string.
	 */
	static wxString Dump_sgVhash(
		class vvContext&        cContext,
		const struct _SG_vhash* pValue
		);

#endif
};


#endif
