/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include "precompiled.h"
#include "vvSgLogHandler.h"

#include "sg.h"
#include "vvContext.h"
#include "vvSgHelpers.h"


/*
**
** Internal Functions
**
*/

/**
 * SG_log_text__writer implementation used by vvSgTextLogHandler.
 */
void _TextWriter(
	SG_context*      pCtx,        //< [in] [out] Error and context information.
	void*            pWriterData, //< [in] Data specific to the writer.
	const SG_string* pText,       //< [in] The text to write.
	                              //<      If NULL, then the writer is being initialized or destroyed (check bError to determine which).
	SG_bool          bError       //< [in] Whether or not the text should be considered an error.
	                              //<      This is for things like choosing STDERR vs STDOUT.
	                              //<      If szText is NULL, then this will be true for initialization or false for destruction.
	)
{
	vvContext cContext(pCtx);

	if (pText == NULL)
	{
		return;
	}

	vvSgTextLogHandler* pThis = static_cast<vvSgTextLogHandler*>(pWriterData);
	pThis->DoWrite(cContext, vvSgHelpers::Convert_sgString_wxString(pText), vvSgHelpers::Convert_sgBool_cppBool(bError));
}


/*
**
** Public Functions
**
*/

// ----- vvSgLogHandler -----

vvSgLogHandler::vvSgLogHandler()
	: mbRegistered(false)
{
}

vvSgLogHandler::~vvSgLogHandler()
{
	wxASSERT(!this->mbRegistered);
}

bool vvSgLogHandler::RegisterSgLogHandler(
	vvContext& cContext
	)
{
	if (this->mbRegistered)
	{
		return true;
	}

	SG_log__global_init(cContext);
	if (cContext.Error_Check())
	{
		return false;
	}

	if (!this->DoRegister(cContext))
	{
		return false;
	}

	this->mbRegistered = true;
	return true;
}

bool vvSgLogHandler::UnregisterSgLogHandler(
	vvContext& cContext
	)
{
	if (!this->mbRegistered)
	{
		return true;
	}

	if (!this->DoUnregister(cContext))
	{
		return false;
	}

	this->mbRegistered = false;
	return true;
}

// ----- vvSgGeneralLogHandler -----

vvSgGeneralLogHandler::vvSgGeneralLogHandler(
	sgLogHandler& cHandler,
	void*         pUserData
	)
	: mpHandler(&cHandler)
	, mpUserData(pUserData)
{
}

bool vvSgGeneralLogHandler::DoRegister(
	vvContext& cContext
	)
{
	SG_log__register_handler(cContext, this->mpHandler, this->mpUserData, SG_LOG__FLAG__HANDLER_TYPE__ALL);
	return !cContext.Error_Check();
}

bool vvSgGeneralLogHandler::DoUnregister(
	vvContext& cContext
	)
{
	SG_log__unregister_handler(cContext, this->mpHandler, this->mpUserData);
	return !cContext.Error_Check();
}

// ----- vvSgTextLogHandler -----

vvSgTextLogHandler::vvSgTextLogHandler(
	vvContext& cContext
	)
{
	this->mpHandler = new SG_log_text__data;

	SG_log_text__set_defaults(cContext, this->mpHandler);
	if (cContext.Error_Check())
	{
		delete this->mpHandler;
		this->mpHandler = NULL;
		return;
	}

	this->mpHandler->fWriter     = _TextWriter;
	this->mpHandler->pWriterData = this;
}

vvSgTextLogHandler::~vvSgTextLogHandler()
{
	if (this->mpHandler != NULL)
	{
		delete this->mpHandler;
	}
}

bool vvSgTextLogHandler::IsOk() const
{
	return (this->mpHandler != NULL);
}

#define IMPLEMENT_PROPERTY__BOOL(FunctionName, PropertyName)         \
	void vvSgTextLogHandler::FunctionName(                           \
		bool bValue                                                  \
		)                                                            \
	{                                                                \
		this->mpHandler->PropertyName = bValue ? SG_TRUE : SG_FALSE; \
	}

#define IMPLEMENT_PROPERTY__STRING(FunctionName, PropertyName)            \
void vvSgTextLogHandler::FunctionName(                                    \
	const wxString& sValue                                                \
	)                                                                     \
{                                                                         \
	size_t uIndex = this->mcStrings.Add(sValue);                          \
	this->mpHandler->PropertyName = this->mcStrings.Item(uIndex).c_str(); \
}

IMPLEMENT_PROPERTY__BOOL(SetLogVerboseOperations, bLogVerboseOperations);
IMPLEMENT_PROPERTY__BOOL(SetLogVerboseValues,     bLogVerboseValues);

IMPLEMENT_PROPERTY__STRING(SetRegisterMessage,     szRegisterMessage);
IMPLEMENT_PROPERTY__STRING(SetUnregisterMessage,   szUnregisterMessage);
IMPLEMENT_PROPERTY__STRING(SetDateTimeFormat,      szDateTimeFormat);
IMPLEMENT_PROPERTY__STRING(SetProcessThreadFormat, szProcessThreadFormat);
IMPLEMENT_PROPERTY__STRING(SetIndent,              szIndent);
IMPLEMENT_PROPERTY__STRING(SetElapsedTimeFormat,   szElapsedTimeFormat);
IMPLEMENT_PROPERTY__STRING(SetCompletionFormat,    szCompletionFormat);
IMPLEMENT_PROPERTY__STRING(SetOperationFormat,     szOperationFormat);
IMPLEMENT_PROPERTY__STRING(SetValueFormat,         szValueFormat);
IMPLEMENT_PROPERTY__STRING(SetStepFormat,          szStepFormat);
IMPLEMENT_PROPERTY__STRING(SetStepMessage,         szStepMessage);
IMPLEMENT_PROPERTY__STRING(SetResultFormat,        szResultFormat);
IMPLEMENT_PROPERTY__STRING(SetVerboseFormat,       szVerboseFormat);
IMPLEMENT_PROPERTY__STRING(SetInfoFormat,          szInfoFormat);
IMPLEMENT_PROPERTY__STRING(SetWarningFormat,       szWarningFormat);
IMPLEMENT_PROPERTY__STRING(SetErrorFormat,         szErrorFormat);

#undef IMPLEMENT_PROPERTY__BOOL
#undef IMPLEMENT_PROPERTY__STRING

bool vvSgTextLogHandler::DoRegister(
	vvContext& cContext
	)
{
	SG_log_text__register(cContext, this->mpHandler, SG_LOG__FLAG__HANDLER_TYPE__ALL);
	return !cContext.Error_Check();
}

bool vvSgTextLogHandler::DoUnregister(
	vvContext& cContext
	)
{
	SG_log_text__unregister(cContext, this->mpHandler);
	return !cContext.Error_Check();
}
