/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include "precompiled.h"
#include "vvStatusDialog.h"

#include "vvContext.h"
#include "vvDialogHeaderControl.h"
#include "vvStatusControl.h"


/*
**
** Globals
**
*/

const unsigned int guStatusControlStyle = 0u;


/*
**
** Public Functions
**
*/

vvStatusDialog::vvStatusDialog()
	: msDirectory(wxEmptyString)
	, mwStatus(NULL)
{
}

vvStatusDialog::~vvStatusDialog()
{
	if (this->mwStatus != NULL)
	{
		delete this->mwStatus;
	}
}

bool vvStatusDialog::Create(
	wxWindow*       pParent,
	const wxString& sDirectory,
	vvVerbs&        cVerbs,
	vvContext&      cContext
	)
{
	// create the dialog window
	if (!vvDialog::Create(pParent, &cVerbs, &cContext, this->GetClassInfo()->GetClassName()))
	{
		return false;
	}

	// store the given initialization data
	this->msDirectory = sDirectory;

	// create our header control
	wxXmlResource::Get()->AttachUnknownControl("wHeader", new vvDialogHeaderControl(this, wxID_ANY, this->msDirectory, cVerbs, cContext));

	// create our status control
	this->mwStatus = new vvStatusControl(this, wxID_ANY, wxDefaultPosition, wxDefaultSize, guStatusControlStyle);
	wxXmlResource::Get()->AttachUnknownControl("wStatus", this->mwStatus);

	// get all of our other widgets
	this->mcStatusFilters[vvStatusControl::STATUS_ADDED]         = XRCCTRL(*this, "wOptions_StatusFilter_Added",    wxCheckBox);
	this->mcStatusFilters[vvStatusControl::STATUS_CONFLICT]      = XRCCTRL(*this, "wOptions_StatusFilter_Conflict", wxCheckBox);
	this->mcStatusFilters[vvStatusControl::STATUS_DELETED]       = XRCCTRL(*this, "wOptions_StatusFilter_Deleted",  wxCheckBox);
	this->mcStatusFilters[vvStatusControl::STATUS_FOUND]         = XRCCTRL(*this, "wOptions_StatusFilter_Found",    wxCheckBox);
	this->mcStatusFilters[vvStatusControl::STATUS_IGNORED]       = XRCCTRL(*this, "wOptions_StatusFilter_Ignored",  wxCheckBox);
	this->mcStatusFilters[vvStatusControl::STATUS_LOST]          = XRCCTRL(*this, "wOptions_StatusFilter_Lost",     wxCheckBox);
	this->mcStatusFilters[vvStatusControl::STATUS_MODIFIED]      = XRCCTRL(*this, "wOptions_StatusFilter_Modified", wxCheckBox);
	this->mcStatusFilters[vvStatusControl::STATUS_MULTIPLE]      = XRCCTRL(*this, "wOptions_StatusFilter_Multiple", wxCheckBox);
	this->mcStatusFilters[vvStatusControl::STATUS_MOVED]         = XRCCTRL(*this, "wOptions_StatusFilter_Moved",    wxCheckBox);
	this->mcStatusFilters[vvStatusControl::STATUS_RENAMED]       = XRCCTRL(*this, "wOptions_StatusFilter_Renamed",  wxCheckBox);
	this->mcTypeFilters[vvStatusControl::ITEM_FILE]              = XRCCTRL(*this, "wOptions_TypeFilter_File",       wxCheckBox);
	this->mcTypeFilters[vvStatusControl::ITEM_FOLDER]            = XRCCTRL(*this, "wOptions_TypeFilter_Folder",     wxCheckBox);
	this->mcTypeFilters[vvStatusControl::ITEM_SYMLINK]           = XRCCTRL(*this, "wOptions_TypeFilter_Symlink",    wxCheckBox);
	this->mcTypeFilters[vvStatusControl::ITEM_SUBMODULE]         = XRCCTRL(*this, "wOptions_TypeFilter_Submodule",  wxCheckBox);
	this->mcColumnFilters[vvStatusControl::COLUMN_FILENAME]      = XRCCTRL(*this, "wOptions_Column_Filename",       wxCheckBox);
	this->mcColumnFilters[vvStatusControl::COLUMN_STATUS]        = XRCCTRL(*this, "wOptions_Column_Status",         wxCheckBox);
	this->mcColumnFilters[vvStatusControl::COLUMN_FROM]          = XRCCTRL(*this, "wOptions_Column_From",           wxCheckBox);
	this->mcColumnFilters[vvStatusControl::COLUMN_ATTRIBUTES]    = XRCCTRL(*this, "wOptions_Column_Attributes",     wxCheckBox);
	this->mcColumnFilters[vvStatusControl::COLUMN_EX_ATTRIBUTES] = XRCCTRL(*this, "wOptions_Column_ExAttributes",   wxCheckBox);
	this->mcColumnFilters[vvStatusControl::COLUMN_GID]           = XRCCTRL(*this, "wOptions_Column_Gid",            wxCheckBox);
	this->mcColumnFilters[vvStatusControl::COLUMN_HID_NEW]       = XRCCTRL(*this, "wOptions_Column_NewHid",         wxCheckBox);
	this->mcColumnFilters[vvStatusControl::COLUMN_HID_OLD]       = XRCCTRL(*this, "wOptions_Column_OldHid",         wxCheckBox);
	this->mwOptions_Misc_CollapseMode                            = XRCCTRL(*this, "wOptions_Misc_CollapseMode",     wxCheckBox);

	// set our checkbox labels according to the status control data
	for (stlValueCheckboxMap::const_iterator it = this->mcStatusFilters.begin(); it != this->mcStatusFilters.end(); ++it)
	{
		if (it->second->GetLabelText().IsEmpty())
		{
			it->second->SetLabel(vvStatusControl::GetStatusData(static_cast<vvStatusControl::Status>(it->first))->szName);
		}
	}
	for (stlValueCheckboxMap::const_iterator it = this->mcTypeFilters.begin(); it != this->mcTypeFilters.end(); ++it)
	{
		if (it->second->GetLabelText().IsEmpty())
		{
			it->second->SetLabel(vvStatusControl::GetItemTypeData(static_cast<vvStatusControl::ItemType>(it->first))->szName);
		}
	}
	for (stlValueCheckboxMap::const_iterator it = this->mcColumnFilters.begin(); it != this->mcColumnFilters.end(); ++it)
	{
		if (it->second->GetLabelText().IsEmpty())
		{
			it->second->SetLabel(vvStatusControl::GetColumnData(static_cast<vvStatusControl::Column>(it->first))->szName);
		}
	}
	// BUG: The Layout/Fit calls don't seem to correctly handle checkbox labels that got longer.
	//      Layout/Fit are called in UpdateDisplay, which is called by TransferDataToWindow, which is called after this function,
	//      though it doesn't seem to make a difference to call them here, either.
	//      If the label was lengthened, it gets clipped by its parent static box sizer.
	//      I haven't been able to figure that out yet.
	//      In the meantime, I went back to hard-coding the labels in the XRC file.

	// success
	return true;
}

bool vvStatusDialog::TransferDataToWindow()
{
	// retrieve status data and populate widgets
	if (!this->mwStatus->FillData(*this->GetVerbs(), *this->GetContext(), this->msDirectory))
	{
		this->GetContext()->Error_ResetReport("Unable to populate state data.");
		return false;
	}

	// update our display to reflect any widget changes
	this->UpdateDisplay();

	// success
	return true;
}

void vvStatusDialog::UpdateDisplay()
{
	// build the status filter from the checked boxes
	vvFlags cStatusFilter = 0u;
	for (stlValueCheckboxMap::const_iterator it = this->mcStatusFilters.begin(); it != this->mcStatusFilters.end(); ++it)
	{
		if (it->second->IsChecked())
		{
			cStatusFilter.AddFlags(it->first);
		}
	}

	// build the types filter from the checked boxes
	vvFlags cTypesFilter = 0u;
	for (stlValueCheckboxMap::const_iterator it = this->mcTypeFilters.begin(); it != this->mcTypeFilters.end(); ++it)
	{
		if (it->second->IsChecked())
		{
			cTypesFilter.AddFlags(it->first);
		}
	}

	// decide which mode to use
	vvStatusControl::Mode eMode = this->mwStatus->GetDisplayMode();
	if (eMode != vvStatusControl::MODE_FLAT)
	{
		eMode = this->mwOptions_Misc_CollapseMode->IsChecked() ? vvStatusControl::MODE_COLLAPSED : vvStatusControl::MODE_FOLDER;
	}

	// build the columns filter from the checked boxes
	vvFlags cColumnsFilter = 0u;
	for (stlValueCheckboxMap::const_iterator it = this->mcColumnFilters.begin(); it != this->mcColumnFilters.end(); ++it)
	{
		if (it->second->IsChecked())
		{
			cColumnsFilter.AddFlags(it->first);
		}
	}

	// set the status display according to the options
	this->mwStatus->SetDisplay(eMode, cTypesFilter.GetValue(), cStatusFilter.GetValue());

	// set the columns according to the options
	this->mwStatus->SetColumns(cColumnsFilter.GetValue());

	// expand all the items
	this->mwStatus->ExpandAll();

	// relayout our controls and fit the window in case child sizes change
	this->Layout();
	this->Fit();
}

void vvStatusDialog::OnNotebookPageChanged(
	wxBookCtrlEvent& cEvent
	)
{
	// if we haven't even been created yet, ignore this
	if (this->mwStatus == NULL)
	{
		return;
	}

	// if we're changing to the Status tab, then update based on our options
	if (cEvent.GetSelection() == 0u)
	{
		this->UpdateDisplay();
	}
}

IMPLEMENT_DYNAMIC_CLASS(vvStatusDialog, vvDialog);

BEGIN_EVENT_TABLE(vvStatusDialog, vvDialog)
	EVT_NOTEBOOK_PAGE_CHANGED(wxID_ANY, OnNotebookPageChanged)
END_EVENT_TABLE()
