/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include "precompiled.h"
#include "vvSyncCommand.h"

#include "vvContext.h"
#include "vvSyncDialog.h"
#include "vvVerbs.h"


/*
**
** Globals
**
*/

/**
 * Global instance of vvSyncCommand.
 */
static const vvSyncCommand gcSyncCommand;

/*
 * Command specification.
 * These need to be char* instead of wxString, because they're used by a static constructor.
 * And we can't be sure the wxStrings will be constructed before the static instance that uses them.
 */
static const char*        gszCommandName        = "sync";
static const unsigned int guCommandFlags        = 0;
static const char*        gszCommandDescription = "Synchronize changes between repository instances.";


/*
**
** Public Functions
**
*/

vvSyncCommand::vvSyncCommand()
	: vvCommand(gszCommandName, guCommandFlags, gszCommandDescription)
	, mwDialog(NULL)
{
}

bool vvSyncCommand::ParseArguments(
	const wxArrayString& cArguments
	)
{
	// get the directory name from the first argument
	if (cArguments.Count() < 1u)
	{
		wxLogError("No argument was passed in to sync.");
		return false;
	}
	wxString sItem = cArguments[0];

	// make sure this is actually an existing directory
	if ( ! (wxFileExists(sItem) || wxDirExists(sItem)) )
	{
		wxLogError("Specified item doesn't exist: %s", sItem);
		return false;
	}

	// success
	this->msDirectory = sItem;
	return true;
}

vvDialog* vvSyncCommand::CreateDialog(
	wxWindow* wParent
	)
{
	vvVerbs*   pVerbs   = this->GetVerbs();
	vvContext* pContext = this->GetContext();

	wxASSERT(pVerbs != NULL);
	wxASSERT(pContext != NULL);
	wParent;
	this->mwDialog = new vvSyncDialog();
	if (!this->mwDialog->Create(wParent, this->msDirectory, *pVerbs, *pContext))
	{
		delete this->mwDialog;
		return NULL;
	}
	return this->mwDialog;
}

bool vvSyncCommand::Execute()
{
	vvVerbs*   pVerbs   = this->GetVerbs();
	vvContext* pContext = this->GetContext();

	wxASSERT(pVerbs != NULL);
	wxASSERT(pContext != NULL);

	vvSyncDialog::SyncAction selectedAction = this->mwDialog->GetSyncAction();

	if (selectedAction == vvSyncDialog::SYNC_PULL || selectedAction == vvSyncDialog::SYNC_PULL_THEN_PUSH)
	{
		pContext->Log_PushOperation("Pulling", vvContext::LOG_FLAG_NONE);
		pContext->Log_SetSteps(2u);
//		wxScopeGuard cPopOperation = wxMakeGuard(&vvContext::Log_PopOperation_Static, pContext);
//		wxUnusedVar(cPopOperation);

		pVerbs->Pull(*pContext, this->mwDialog->GetOtherRepo(), this->mwDialog->GetMyRepoName(), 
			this->mwDialog->GetAreas(), wxArrayString(), this->mwDialog->GetChangesets(), wxArrayString());

		pContext->Log_FinishStep();
		pContext->Log_PopOperation();
	}
	
	if (!pContext->Error_Check() && (selectedAction == vvSyncDialog::SYNC_PUSH || selectedAction == vvSyncDialog::SYNC_PULL_THEN_PUSH))
	{
		pContext->Log_PushOperation("Pushing", vvContext::LOG_FLAG_NONE);
		pContext->Log_SetSteps(2u);
		wxScopeGuard cPopOperation = wxMakeGuard(&vvContext::Log_PopOperation_Static, pContext);
		wxUnusedVar(cPopOperation);

		pVerbs->Push(*pContext, this->mwDialog->GetOtherRepo(), this->mwDialog->GetMyRepoName(), 
			this->mwDialog->GetAreas(), wxArrayString(), this->mwDialog->GetChangesets(), wxArrayString());

		pContext->Log_FinishStep();
	}
	return !pContext->Error_Check();
}
