/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include "precompiled.h"
#include "vvSyncDialog.h"

#include "vvVerbs.h"
#include "vvValidatorEmptyConstraint.h"
#include "vvValidatorRepoExistsConstraint.h"
#include "vvProgressExecutor.h"
#include "vvInOutCommand.h"
#include "vvDialogHeaderControl.h"
#include "vvHistoryListControl.h"


/*
**
** Globals
**
*/

static const wxString gsValidationField_Repo                     = "Other Repository";
static const wxString gsValidationField_Revision_Branch_Value    = "Branches";
static const wxString gsValidationField_Revision_Tag_Value       = "Tags";
static const wxString gsValidationField_Revision_Changeset_Value = "Changesets";

static const wxString gsHint_Repos                 = "Local repository name, or remote repository URL";
static const wxString gsHint_Dags			       = "All areas will be transferred";
static const wxString gsHint_Revision_Branch_Value    = "Name of the branches to transfer";
static const wxString gsHint_Revision_Tag_Value       = "Name of the tags to transfer";
static const wxString gsHint_Revision_Changeset_Value = "ID of the changesets to transfer";

/*
**
** Public Functions
**
*/

bool vvSyncDialog::Create(
	wxWindow*            pParent,
	const wxString&      sInitialValue,
	vvVerbs&             cVerbs,
	vvContext&           cContext
	)
{
	// create the dialog window
	if (!vvDialog::Create(pParent, &cVerbs, &cContext, this->GetClassInfo()->GetClassName()))
	{
		return false;
	}
	// create our header control
	wxXmlResource::Get()->AttachUnknownControl("wHeader", new vvDialogHeaderControl(this, wxID_ANY, sInitialValue, cVerbs, cContext));

	this->mwComboBox						  = XRCCTRL(*this, "wxOperation",                 wxComboBox);
	this->mwOKButton						  = XRCCTRL(*this, "wxID_OK",                 wxButton);
	this->mwDagBrowseButton					  = XRCCTRL(*this, "wDagBrowse",                 wxButton);
	this->mwDags							  = XRCCTRL(*this, "wDags",                 wxTextCtrl);
	this->mwCheckBoxUpdate					  = XRCCTRL(*this, "wCheckBox_Update",                 wxCheckBox);
	this->mwComboBoxRepos					  = XRCCTRL(*this, "wxRepos",                 wxComboCtrl);
	this->msPathProvided = sInitialValue;
	this->mwListViewComboPopup = new vv_wxListViewComboPopup();

	this->mwComboBoxRepos->UseAltPopupWindow();
	this->mwComboBoxRepos->EnablePopupAnimation(false);
    this->mwComboBoxRepos->SetPopupMinWidth(200);
	this->mwComboBoxRepos->SetPopupControl(this->mwListViewComboPopup);
	bSyncingVersionControl = true;

	this->mwComboBoxRepos			->SetHint(gsHint_Repos);
	this->mwDags					->SetHint(gsHint_Dags);
	
	this->mwComboBoxRepos->SetValidator(vvValidator(gsValidationField_Repo)
		.SetValidatorFlags(vvValidator::VALIDATE_SUCCEED_IF_DISABLED | vvValidator::VALIDATE_SUCCEED_IF_HIDDEN)
		.SetReporter(&this->mcValidatorReporter)
		.AddConstraint(vvValidatorEmptyConstraint(false))
	);
	wxString psRepoDescriptorName;
	cVerbs.FindWorkingFolder(cContext, msPathProvided, NULL, &psRepoDescriptorName);
	m_sRepoName = wxString(psRepoDescriptorName);
	m_revisionsControl = new vvHistoryListControl(cVerbs, cContext, m_sRepoName, this, wxID_ANY);
	wxXmlResource::Get()->AttachUnknownControl("wRevisions", m_revisionsControl, this);

	Layout();
	Fit();
	return true;
}

bool vvSyncDialog::TransferDataToWindow()
{ 
	vvContext&     pCtx = *this->GetContext();
	vvVerbs*   pVerbs   = this->GetVerbs();
	wxArrayString cRemoteRepos;
	wxArrayString cRemoteRepoAliases;
	
	
	pVerbs->GetRemoteRepos(pCtx, m_sRepoName, cRemoteRepos, cRemoteRepoAliases);
	
	this->mwListViewComboPopup->GetListView()->InsertColumn(0, "URL or repo name");
	this->mwListViewComboPopup->GetListView()->InsertColumn(1, "Alias", wxLIST_FORMAT_RIGHT);
	for (unsigned int uIndex = 0u; uIndex < cRemoteRepos.size(); ++uIndex)
	{
		wxListItem item;

		this->mwListViewComboPopup->GetListView()->InsertItem(uIndex, "test");
		//this->mwListViewComboPopup->SetItem(this->mwListViewComboPopup->GetItemCount(), 0, cRemoteRepos[uIndex] );
		this->mwListViewComboPopup->GetListView()->SetItem(uIndex, 0, cRemoteRepos[uIndex] );
		this->mwListViewComboPopup->GetListView()->SetItem(uIndex, 1, cRemoteRepoAliases[uIndex] );
		if (uIndex == 0)
			this->mwComboBoxRepos->SetValue(cRemoteRepos[0]);
	}
	vvHistoryColumnPreferences cColumnPrefs;
	pVerbs->GetColumnPreferences(pCtx, cColumnPrefs);
	this->m_revisionsControl->ApplyColumnPreferences(cColumnPrefs);
	UpdateDisplay();
	//Set the column widths.
	int comboWidth = (int)this->mwComboBoxRepos->GetSize().GetWidth();
	int firstColumnWidth =  (int)((double)comboWidth * 0.75);
	int secondColumnWidth =  (int)((double)comboWidth * 0.25);
	this->mwListViewComboPopup->GetListView()->SetColumnWidth(0, firstColumnWidth);
	this->mwListViewComboPopup->GetListView()->SetColumnWidth(1, secondColumnWidth);

	mcAreas_all.Add("Version Control");
	mcAreas_all.Add("Work Item Tracking");
	mcAreas_all.Add("Scrum");
	return true;
}

bool vvSyncDialog::TransferDataFromWindow()
{
	// success
	this->msOtherRepoSpec = this->mwComboBoxRepos->GetValue();
	this->mSelectedOperation = gui_GetSyncAction();
	if (!bSyncingVersionControl)
	{
		mRevisionMode = REV_MODE_ALL;
	}
	else
	{
		if (XRCCTRL(*this, "wRevision_Select_One", wxRadioButton)	->GetValue())
		{
			mcChangesets_selected = m_revisionsControl->GetSelection();
		}
		mRevisionMode = REV_MODE_CHANGESETS;
		/*if (XRCCTRL(*this, "wRevision_Branch_Select", wxRadioButton)	->GetValue())
			mRevisionMode = REV_MODE_BRANCHES;
		else if (XRCCTRL(*this, "wRevision_Tag_Select", wxRadioButton)	->GetValue())
			mRevisionMode = REV_MODE_TAGS;
		else if (XRCCTRL(*this, "wRevision_Changeset_Select", wxRadioButton)	->GetValue())
			mRevisionMode = REV_MODE_CHANGESETS;
			*/
	}
	return true;
}

vvSyncDialog::SyncAction vvSyncDialog::GetSyncAction()
{
	return mSelectedOperation;
}

wxString vvSyncDialog::GetOtherRepo()
{
	return msOtherRepoSpec;
}

wxString vvSyncDialog::GetMyRepoName()
{
	return m_sRepoName;
}

vvSyncDialog::SyncAction vvSyncDialog::gui_GetSyncAction()
{
	return (SyncAction)this->mwComboBox->GetCurrentSelection();
}

//////////////////////////////Events//////////////////
void vvSyncDialog::OnComboBoxSelectionChanged(
	wxCommandEvent& cEvent
	)
{
	if (cEvent.GetEventObject() == this->mwComboBox)
		this->mwOKButton->SetLabel(this->mwComboBox->GetValue());
	ShowOrHideUpdateControl();
	LoadRevisionsIfNecessary();
}

void vvSyncDialog::OnComboBoxDropdown(wxCommandEvent& cEvent)
{
	cEvent;
	this->mwListViewComboPopup->SetFocus();
}

void vvSyncDialog::OnComboBoxText(wxCommandEvent& cEvent)
{
	cEvent;
}

void vvSyncDialog::OnComboBoxResized(
	wxSizeEvent& cEvent
	)
{
	cEvent;
	/*if (this->mwListViewComboPopup != NULL)
	{
		this->mwListViewComboPopup->SetSize(this->mwComboBoxRepos->GetSize().GetWidth(), 300);
	}*/
	cEvent.Skip();
}

void vvSyncDialog::ShowOrHideUpdateControl()
{
	if (gui_GetSyncAction() != SYNC_PULL)
	{
		//They aren't doing a pull.  Always hide
		vvXRCSIZERITEM(*this, "cUpdate")->Show(false);
	}
	else
	{
		//They may not be syncing the version control area, so this will sometimes be hiding, and sometimes showing.
		vvXRCSIZERITEM(*this, "cUpdate")->Show(bSyncingVersionControl);
	}
	if (gui_GetSyncAction() == SYNC_PULL_THEN_PUSH)
		vvXRCSIZERITEM(*this, "cRevision")->Show(false);
	else
		//They may not be syncing the version control area, so this will sometimes be hiding, and sometimes showing.
		vvXRCSIZERITEM(*this, "cRevision")->Show(bSyncingVersionControl);

	this->SetMinSize(wxSize(this->GetSize().GetWidth(), this->GetSizer()->GetMinSize().GetHeight()));
	this->Layout();
	this->Fit();
	this->Refresh();
}

void vvSyncDialog::LoadRevisionsIfNecessary()
{
	if (m_revisionsControl->IsEnabled())
	{
		m_bDoPull = gui_GetSyncAction() == SYNC_PULL;
		//Reset the list to empty.
		vvVerbs::HistoryData emptyList;
		m_revisionsControl->LoadHistory(emptyList);
		// we want to start a long task, but we don't want our GUI to block
		// while it's executed, so we use a thread to do it.
		if (GetThread() == NULL || !GetThread()->IsAlive())
			if ( CreateThread(wxTHREAD_JOINABLE) != wxTHREAD_NO_ERROR)
			{
				wxLogError("Could not create the worker thread!");
				return;
			}


		// go!
		if (!GetThread()->IsRunning())
			GetThread()->Run();
	}
}

void vvSyncDialog::UpdateDisplay()
{
	if (XRCCTRL(*this, "wRevision_Select_One", wxRadioButton)->GetValue())
	{
		m_revisionsControl->Enable(true);
	}
	else
		m_revisionsControl->Enable(false);
	LoadRevisionsIfNecessary();
	/*
	this->mwBranches   ->Enable(XRCCTRL(*this, "wRevision_Branch_Select", wxRadioButton)	->GetValue());
	this->mwTags      ->Enable(XRCCTRL(*this, "wRevision_Tag_Select", wxRadioButton)     ->GetValue());
	this->mwChangesets->Enable(XRCCTRL(*this, "wRevision_Changeset_Select", wxRadioButton)		->GetValue());
	this->mwButtonBrowseBranches->Enable(XRCCTRL(*this, "wRevision_Branch_Select", wxRadioButton)	->GetValue());
	this->mwButtonBrowseTags	->Enable(XRCCTRL(*this, "wRevision_Tag_Select", wxRadioButton)     ->GetValue());
	this->mwButtonBrowseChangesets->Enable(XRCCTRL(*this, "wRevision_Changeset_Select", wxRadioButton)		->GetValue());
	*/
}
void vvSyncDialog::OnControlValueChange(
	wxCommandEvent& cEvent
	)
{
	cEvent;
	UpdateDisplay();
}
void vvSyncDialog::OnBrowseButtonPressed(
	wxCommandEvent& cEvent
	)
{
	if (cEvent.GetEventObject() == this->mwDagBrowseButton)
	{
		{
			
			// create a dialog to allow the user to choose one of them
			wxMultiChoiceDialog cDialog;
			{
				if (!cDialog.Create(this, "Select the areas that you wish to sync.", "Select Areas", mcAreas_all))
				{
					wxLogError("Error creating area selection dialog.");
					return;
				}
				wxArrayInt selectionsIn;
				for (size_t index = 0; index < mcAreas_all.Count(); index++)
				{
					selectionsIn.Add(index);
				}
				cDialog.SetSelections(selectionsIn);
				mcAreas_selected.Empty();

				// show the dialog
				if (cDialog.ShowModal() != wxID_OK)
				{
					return;
				}

				if (cDialog.GetSelections()[0] != 0)
				{
					//They did not select Version control.  Hide the version control stuff.
					bSyncingVersionControl = false;
					ShowOrHideUpdateControl();
					vvXRCSIZERITEM(*this, "cRevision")->Show(false);
				}
				else
				{
					bSyncingVersionControl = true;
					ShowOrHideUpdateControl();
					vvXRCSIZERITEM(*this, "cRevision")->Show(true);
				}
				this->Layout();
				this->Fit();
				this->Refresh();
				// set the value to the branch they chose
				this->mwDags->Clear();
				wxArrayInt selections = cDialog.GetSelections();
				if (selections.Count() == mcAreas_all.Count())
				{
					return;
				}
				bool addedOneSoFar = false;
				mcAreas_selected.Empty();
				for (size_t index = 0; index < selections.Count(); index++)
				{
					if (addedOneSoFar == true)
						this->mwDags->AppendText(", ");
					else
						addedOneSoFar = true;
					this->mwDags->AppendText(mcAreas_all[cDialog.GetSelections()[index]]);
					mcAreas_selected.Add(mcAreas_all[cDialog.GetSelections()[index]]);
				}
				
			}
		}
	}
	this->SetMinSize(wxSize(this->GetSize().GetWidth(), this->GetSizer()->GetMinSize().GetHeight()));
	this->Layout();
	this->Fit();
	this->Refresh();
}

wxArrayString vvSyncDialog::GetAreas()
{
	return mcAreas_selected;	
}

vvSyncDialog::RevisionMode vvSyncDialog::GetRevisionMode()
{
	return mRevisionMode;
}

wxArrayString vvSyncDialog::GetChangesets()
{
	if (GetRevisionMode() == REV_MODE_CHANGESETS)
		return mcChangesets_selected;
	else
	{
		wxArrayString emptyArray;
		return emptyArray;
	}
}

wxThread::ExitCode vvSyncDialog::Entry()
{
    // IMPORTANT:
    // this function gets executed in the secondary thread context!
	try
	{
		vvVerbs::HistoryData cHistoryResults;
		if (gui_GetSyncAction() == SYNC_PULL)
		{
			this->GetVerbs()->Incoming(*this->GetContext(), this->mwComboBoxRepos->GetValue(), m_sRepoName, cHistoryResults);
		}
		else
		{
			this->GetVerbs()->Outgoing(*this->GetContext(), this->mwComboBoxRepos->GetValue(), m_sRepoName, cHistoryResults);
		}
		m_revisionsControl->LoadHistory(cHistoryResults);
	}
	catch (std::exception e)
	{
		int tmp = 4;
		tmp++;
	}


    return (wxThread::ExitCode)0;
}

IMPLEMENT_DYNAMIC_CLASS(vvSyncDialog, vvDialog);
BEGIN_EVENT_TABLE(vvSyncDialog, vvDialog)
	EVT_COMBOBOX(wxID_ANY, OnComboBoxSelectionChanged)
	EVT_TEXT(XRCID("wxRepos"), OnComboBoxText)
	EVT_COMBOBOX_DROPDOWN(XRCID("wxRepos"), OnComboBoxDropdown)
	EVT_BUTTON(XRCID("wDagBrowse"),         vvSyncDialog::OnBrowseButtonPressed)

	EVT_RADIOBUTTON(XRCID("wRevision_All_Select"),      vvSyncDialog::OnControlValueChange)
	EVT_RADIOBUTTON(XRCID("wRevision_Select_One"),    vvSyncDialog::OnControlValueChange)
	EVT_SIZE(vvSyncDialog::OnComboBoxResized)
END_EVENT_TABLE()
