/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#ifndef VV_VERBS_H
#define VV_VERBS_H

#include "precompiled.h"
#include "vvFlags.h"
#include "vvHistoryObjects.h"


/*
 * This is a wrapper around the actual sglib verb operations used by vvDialog.
 * It is higher level and takes care of a lot of sglib details so dialogs don't have to.
 * It also hides sglib-specific types and translates them to wxWidgets and Windows equivalents.
 * All functions that return a bool are indicating success/failure.
 *
 * This is a base class interface, which allows for several implementations.
 * This is helpful so that a debug or mock implementation can be easily swapped in for debugging/testing purposes.
 */
class vvVerbs
{
// types
public:
	/**
	 * Data about a user in a repository.
	 */
	struct UserData
	{
		wxString sGid;    //< Unique ID of the user.
		wxString sName;   //< Name of the user.
		wxString sPrefix; //< Prefix used to help generate unique IDs from this user.
	};

	/**
	 * A list of user data.
	 */
	typedef std::list<UserData> stlUserDataList;

	/**
	 * Data about a stamp in a repository.
	 */
	struct StampData
	{
		wxString     sName;  //< Name of the stamp.
		unsigned int uCount; //< Number of times the stamp is used in the repo.
	};

	/**
	 * A list of stamp data.
	 */
	typedef std::list<StampData> stlStampDataList;

	
	/**
	 * Data about a tag in a repository.
	 */
	struct TagData
	{
		wxString     sName;  //< Name of the tag.
		wxString     sCSID;  //< revision that the tag is applied to.
	};

	/**
	 * A list of stamp data.
	 */
	typedef std::list<TagData> stlTagDataList;

	/**
	 * List of possible checkout types.
	 */
	enum CheckoutType
	{
		CHECKOUT_HEAD,      //< Checkout the head revision.
		CHECKOUT_BRANCH,    //< Checkout the head revision in a branch.
		CHECKOUT_CHANGESET, //< Checkout an explicit changeset ID.
		CHECKOUT_TAG,       //< Checkout by tag name.
		CHECKOUT_COUNT,     //< Number of values, for iteration.
	};

	/**
	 * Data about changes made to a single version-controlled item.
	 */
	struct DiffEntry
	{
		/**
		 * List of possible types of items in a diff.
		 *
		 * Roughly corresponds to SG_treenode_entry_type in
		 * sg_treenode_entry_typedefs.h, but is not necessarily intended to mirror
		 * them.  They are intended to provide a potentially simpler UI-centric
		 * view of the sglib internal data.
		 */
		enum ItemType
		{
			ITEM_FILE,      //< Diff entry is for a file.
			ITEM_FOLDER,    //< Diff entry is for a folder.
			ITEM_SYMLINK,   //< Diff entry is for a symbolic link.
			ITEM_SUBMODULE, //< Diff entry is for a submodule.
			ITEM_COUNT,     //< Number of values, for iteration.
		};

		/**
		 * List of possible changes to an item in a diff.
		 *
		 * Roughly corresponds to SG_DIFFSTATUS_FLAGS__* in sg_treediff2_typedefs.h,
		 * but it is not necessarily intended to mirror them.  These flags are
		 * intended to provide a potentially simpler UI-centric view of the sglib
		 * internal data.
		 */
		enum ChangeFlags
		{
			CHANGE_NONE              = 0,
			CHANGE_ADDED             = 1 <<  0, //< Item was added to source control.
			CHANGE_DELETED           = 1 <<  1, //< Item was removed from source control and disk.
			CHANGE_MODIFIED          = 1 <<  2, //< Controlled item was changed.
			CHANGE_LOST              = 1 <<  3, //< Item is controlled but not on disk.
			CHANGE_FOUND             = 1 <<  4, //< Item is on disk but uncontrolled.
			CHANGE_RENAMED           = 1 <<  5, //< Controlled item was renamed within a directory.
			CHANGE_MOVED             = 1 <<  6, //< Controlled item was moved to a new directory.
			CHANGE_XATTRS            = 1 <<  7, //< Controlled item's extended attributes were changed.
			CHANGE_ATTRS             = 1 <<  8, //< Controlled item's basic attributes were changed.
			CHANGE_LAST              = 1 <<  9, //< Last item in the enumeration, for iteration purposes.

			// bits in this mask are mutually exclusive
			CHANGE_MASK_EXCLUSIVE    = CHANGE_ADDED
			                         | CHANGE_DELETED
			                         | CHANGE_MODIFIED
			                         | CHANGE_LOST
			                         | CHANGE_FOUND,

			// bits in this mask may be combined with each other and other flags
			CHANGE_MASK_MODIFIER     = CHANGE_RENAMED
			                         | CHANGE_MOVED
			                         | CHANGE_XATTRS
			                         | CHANGE_ATTRS,
		};

		// data
		ItemType eItemType;           //< Type of item the entry is about.
		wxString sItemGid;            //< GID of the item that's different.
		wxString sItemPath;           //< Repo path of the item that's different.
		vvFlags  cItemChanges;        //< ChangeFlags describing what about the item changed.
		bool     bItemIgnored;        //< Whether or not the item is ignored by the include/exclude/ignore settings.
		wxString sContentHid;         //< HID of the item's content.
		wxString sOldContentHid;      //< HID of the item's old content.
		wxString sMoved_From;         //< Old path, for MOVED diffs.
		wxString sRenamed_From;       //< Old filename, for RENAMED diffs.
		wxString sSubmodule_Current;  //< Current version token, for SUBMODULE items.
		wxString sSubmodule_Original; //< Original/baseline version token, for SUBMODULE items.
		bool     bSubmodule_Dirty;    //< Whether or not a submodule's folder is dirty, for SUBMODULE items.
	};

	/**
	 * List of item diffs returned by Diff/Status.
	 */
	typedef std::list<DiffEntry> DiffData;

	/**
	 * Data about a conflict on a single version-controlled item.
	 *
	 * Note: There is a lot more data to add to this once conflicts
	 *       and resolve functionality are more well defined.
	 */
	struct ConflictEntry
	{
		// basic data
		wxString sItemGid;  //< GID of the item that's in conflict.
		bool     bResolved; //< Whether or not the conflict has been resolved yet.
		                    //< Corresponds to SG_ISSUE_STATUS__* in sg_pendingtree_typedefs.h.
	};

	/**
	 * List of conflicts returned by Status.
	 */
	typedef std::list<ConflictEntry> ConflictData;

	/**
	 * List of flags used with the Commit function.
	 */
	enum CommitFlag
	{
		COMMIT_NO_SUBMODULES, //< Ignore submodules when committing.
		COMMIT_NO_RECURSE,    //< Old --non-recursive flag, I don't think this is used anymore (replaced by SG_file_spec).
	};

	enum RepoItemType
	{
		INVALID=-1,					/**< Invalid Value, must be -1. */

		REGULAR_FILE=1,				/**< A regular file. */
		DIRECTORY=2,					/**< A sub-directory. */
		SYMLINK=3,					/**< A symlink. */

		SUBMODULE=4					/**< A submodule directory */
	};
	
	/**
	 * Data about a single item inside the repo (like "@/blah/blah/blah.txt")
	 */
	struct RepoItem
	{
		RepoItemType eItemType;
		wxString sItemName;
	};

	/**
	 * List of conflicts returned by Get_Children.
	 */
	typedef std::vector<RepoItem> RepoItemList;

	/**
	 * Data about a single parent of a changeset
	 */
	struct RevisionEntry
	{
		unsigned int nRevno;            
		wxString sChangesetHID;           
	};

	typedef std::vector<RevisionEntry> RevisionList;

	/**
	 * Data about a single audit on a changeset or comment
	 */
	struct AuditEntry
	{
		wxString sWho;            
		wxDateTime nWhen;           
	};

	/**
	 * Data about a single comment on a changeset
	 */
	struct CommentEntry
	{
		std::vector<AuditEntry> cAudits;
		wxString sCommentText;           
	};
	
	/**
	 * Data about a single changeset
	 */
	struct HistoryEntry
	{
		// data
		unsigned int nRevno;            //< GID of the item that's different.
		wxString sChangesetHID;           //< Repo path of the item that's different.
		RevisionList cParents;
		wxArrayString cTags;
		wxArrayString cStamps;
		wxArrayString cBranches;
		std::vector<AuditEntry> cAudits;
		std::vector<CommentEntry> cComments;
	};
	
	/**
	 * List of item diffs returned by Diff/Status.
	 */
	typedef std::vector<HistoryEntry> HistoryData;


// constants
public:
	/**
	 * The name of the directory that symbolizes a repository root: @
	 * Does not include a trailing slash.
	 */
	static const char* sszRepoRoot;

// construction
public:
	virtual ~vvVerbs();

// functionality
public:
	/**
	 * Initializes sglib globally.
	 * Most other functions will assert until this one is called successfully.
	 */
	virtual bool Initialize(
		class vvContext& cContext //< [in] [out] Error and context info.
		) = 0;

	/**
	 * Shuts down sglib globally.
	 * Most other functions will assert after this one is called.
	 */
	virtual void Shutdown(
		class vvContext& cContext //< [in] [out] Error and context info.
		) = 0;

	/**
	 * Gets a list of repos on the local machine.
	 */
	virtual bool GetLocalRepos(
		class vvContext& cContext,   //< [in] [out] Error and context info.
		wxArrayString&   cLocalRepos //< [out] Array to add local repos to.
		) = 0;

	/**
	 * Checks to see if a local repo with the given name exists.
	 * Returns true if the repo exists, false otherwise.
	 */
	virtual bool LocalRepoExists(
		class vvContext& cContext, //< [in] [out] Error and context info.
		const wxString&  sName     //< [in] The local repo name to check for.
		) = 0;

	/**
	 * Gets a list of remote repos that are known by the local machine.
	 */
	virtual bool GetRemoteRepos(
		class vvContext& cContext,    //< [in] [out] Error and context info.
		wxString& sRepoDescriptorName,//< [in] The repo descriptor name.  Leave this null or empty to get all paths for all repos.
		wxArrayString&   cRemoteRepos,//< [out] Array to add remote repos to.
		wxArrayString&   cRemoteRepoAliases //< [out] Array of aliases.  It will be the same length as cRemoteRepos.  
		) = 0;

	/**
	 * Finds the working folder that contains a given path.
	 * Returns true if the given path is in a working folder, or false if it isn't.
	 */
	virtual bool FindWorkingFolder(
		class vvContext& cContext, //< [in] [out] Error and context info.
		const wxString&  sPath,    //< [in] The path to find a working folder from.
		wxString*        pFolder,  //< [out] The root of the found working folder.
		                           //<       Not modified if no working folder is found.
		                           //<       If set, this is suitable for passing to other vvVerbs functions.
		wxString*        pRepo,    //< [out] The local name of the repo that the found working folder is from.
		                           //<       Not modified if no working folder is found.
		                           //<       If set, this is suitable for passing to other vvVerbs functions.
		wxArrayString*        pBaselineHIDs = NULL     //< [out] The baseline HID of the working folder.
		                           //<       Not modified if no working folder is found.
		                           //<       If set, this is suitable for passing to other vvVerbs functions.
		) = 0;

	/**
	 * Finds the single working folder that contains a given set of paths.
	 * Returns true if the given paths are in a single working folder, or false if they aren't.
	 */
	virtual bool FindWorkingFolder(
		class vvContext&     cContext, //< [in] [out] Error and context info.
		const wxArrayString& cPaths,   //< [in] The path to find a working folder from.
		wxString*            pFolder,  //< [out] The root of the found working folder.
		                               //<       Not modified if no working folder is found.
		                               //<       If set, this is suitable for passing to other vvVerbs functions.
		wxString*            pRepo,    //< [out] The local name of the repo that the found working folder is from.
		                               //<       Not modified if no working folder is found.
		                               //<       If set, this is suitable for passing to other vvVerbs functions.
		wxArrayString*        pBaselineHIDs   = NULL   //< [out] The baseline HID of the working folder.
		                           //<       Not modified if no working folder is found.
		                           //<       If set, this is suitable for passing to other vvVerbs functions.
		);

	/**
	 * Converts an absolute path into a repo path in a given working folder.
	 */
	virtual bool GetRepoPath(
		class vvContext& cContext,       //< [in] [out] Error and context info.
		const wxString&  sAbsolutePath,  //< [in] The absolute paths to convert.
		const wxString&  sWorkingFolder, //< [in] The working folder to convert relative to.
		wxString&        sRepoPath       //< [out] The converted repo path.
		) = 0;

	/**
	 * Converts a set of absolute paths into a set of repo paths in a given working folder.
	 */
	virtual bool GetRepoPaths(
		class vvContext&     cContext,       //< [in] [out] Error and context info.
		const wxArrayString& cAbsolutePaths, //< [in] The absolute paths to convert.
		const wxString&      sWorkingFolder, //< [in] The working folder to convert relative to.
		wxArrayString&       cRepoPaths      //< [in] The converted repo paths.
		);

	/**
	 * Converts a repo path into a path that is relative to its repository's root folder.
	 */
	virtual bool GetRelativePath(
		class vvContext& cContext,     //< [in] [out] Error and context info.
		const wxString&  sRepoPath,    //< [in] The repo path to convert.
		wxString&        sRelativePath //< [out] The converted relative path.
		);

	/**
	 * Converts a set of repo paths into paths that are relative to their repository's root folders.
	 */
	virtual bool GetRelativePaths(
		class vvContext&     cContext,      //< [in] [out] Error and context info.
		const wxArrayString& cRepoPaths,    //< [in] The repo paths to convert.
		wxArrayString&       cRelativePaths //< [out] The converted relative paths.
		);
	
	/**
	 * Converts a set of repo paths into paths that are relative to their repository's root folders.
	 */
	virtual bool GetDiskPaths(
		class vvContext&     cContext,      //< [in] [out] Error and context info.
		const wxString&		 cRootDiskFolder, //<[in] root of working folder.
		const wxArrayString& cRepoPaths,    //< [in] The repo paths to convert.
		wxArrayString&       cRelativePaths //< [out] The converted relative paths.
		);

	/**
	 * Gets a list of branches in a repo.
	 */
	virtual bool GetBranches(
		class vvContext& cContext, //< [in] [out] Error and context info.
		const wxString&  sRepo,    //< [in] The name of the repo to get a list of branches from.
		bool bIncludeClosed,	   //< [in] Include branches that have been marked closed.
		wxArrayString&   cBranches //< [out] Array to add branches to.
		) = 0;

	/**
	 * Gets a list of users in a repo.
	 */
	virtual bool GetUsers(
		class vvContext& cContext, //< [in] [out] Error and context info.
		const wxString&  sRepo,    //< [in] The name of the repo to get a list of users from.
		stlUserDataList& cUsers    //< [out] List to add users to.
		) = 0;

	/**
	 * Gets the e-mail of the current user in a repo.
	 */
	virtual bool GetCurrentUser(
		class vvContext& cContext, //< [in] [out] Error and context info.
		const wxString&  sRepo,    //< [in] The name of the repo to get the current user from.
		wxString&        sUser     //< [out] The current user in the given repo, or wxEmptyString if none is set.
		) = 0;

	/**
	 * Gets a list of stamps in a repo.
	 */
	virtual bool GetStamps(
		class vvContext&  cContext, //< [in] [out] Error and context info.
		const wxString&   sRepo,    //< [in] The name of the repo to get a list of stamps from.
		stlStampDataList& cStamps   //< [out] List to add stamps to.
		) = 0;
	
	/**
	 * Gets a list of stamps in a repo.
	 */
	virtual bool GetTags(
		class vvContext&  cContext, //< [in] [out] Error and context info.
		const wxString&   sRepo,    //< [in] The name of the repo to get a list of tags from.
		stlTagDataList& cTags   //< [out] List to add tags to.
		) = 0;
	/**
	 * Gets the branch that a given working folder is tied to.
	 */
	virtual bool GetBranch(
		class vvContext& cContext, //< [in] [out] Error and context ifno.
		const wxString&  sFolder,  //< [in] The working folder to get the tied branch from.
		wxString&        sBranch   //< [out] The branch that the given working folder is tied to.
		                           //<       Set to an empty string if the folder is not tied.
		) = 0;

	virtual bool GetColumnPreferences(
		vvContext&     pCtx,						//< [in] [out] Error and context ifno.
		vvHistoryColumnPreferences& cColumnPrefs	//< [out] The column preferences that are stored in localsettings.
		) = 0;

	virtual bool GetHistoryFilterDefaults(
		vvContext&     pCtx,						//< [in] [out] Error and context ifno.
		wxString& sRepoName,						//< [in] The repo to load defaults for.  These are repo instance specific.
		vvHistoryFilterOptions& cHistoryDefaults	//< [out] The history filter defaults that are stored in localsettings.
		) = 0;

	virtual bool SaveColumnPreferences(
		vvContext&     pCtx,						//< [in] [out] Error and context ifno.
		vvHistoryColumnPreferences& cColumnPrefs	//< [out] The column preferences to be stored in localsettings.
		) = 0;
	/**
	 * Creates a new local repository.
	 */
	virtual bool New(
		class vvContext& cContext,                       //< [in] [out] Error and context info.
		const wxString&  sName,                          //< [in] The unique (among local repos) name for the new repo.
		const wxString&  sFolder        = wxEmptyString, //< [in] A folder to checkout the new repo into.
		                                                 //<      Empty string indicates to not checkout the new repo.
		const wxString&  sStorageDriver = wxEmptyString, //< [in] The storage driver to use for the new repo.
		                                                 //<      Empty string to use the default storage driver.
		const wxString&  sHashMethod    = wxEmptyString  //< [in] The hash method to use in the new repo.
		                                                //<      Empty string to use the default hash method.
		) = 0;

	/**
	 * Creates a new local repo that is a clone of an existing repo.
	 */
	virtual bool Clone(
		class vvContext& cContext,      //< [in] [out] Error and context info.
		const wxString&  sExistingRepo, //< [in] The existing repo to clone.  Might be a local repo name or remote repo URL.
		const wxString&  sNewRepoName   //< [in] The unique (among local repos) name to use for the new repo.
		) = 0;

	/**
	 * Update a working folder to the given branch or revision.
	 */
	virtual bool Update(
		vvContext&      pCtx,		//< [in] [out] Error and context info.
		const wxString& sFolder,	//< [in] the disk path to the working folder
		CheckoutType    eType,		//< [in] the type of update.
		const wxString& sTarget		//< [in] the name of the branch, or the revision id.
		) = 0;
		
	/**
	 * Creates a working folder from a local repo.
	 * sRevision and sTag cannot both be specified.
	 * If neither is specified, the HEAD revision will be checked out.
	 */
	virtual bool Checkout(
		class vvContext& cContext, //< [in] [out] Error and context info.
		const wxString&  sName,    //< [in] The name of the local repo to checkout.
		const wxString&  sFolder,  //< [in] Absolute path of the folder to check the repo out into.
		CheckoutType     eType,    //< [in] The type of item to checkout from.
		const wxString&  sTarget   //< [in] The target to checkout (possible values depend on eType).
		) = 0;

	/**
	 * Computes the difference between a working folder and a changeset.
	 */
	virtual bool Status(
		class vvContext& cContext,  //< [in] [out] Error and context info.
		const wxString&  sFolder,   //< [in] The working folder to run the diff against.
		const wxString&  sRevision, //< [in] The revision to diff the working folder against.
		                            //<      If empty, the first parent of the working folder is used.
		DiffData&        cDiff      //< [out] Data about the differences between the given folder and revision.
		) = 0;

	/**
	 * Gets the list of conflicts in a working folder.
	 */
	virtual bool Conflicts(
		class vvContext& cContext,  //< [in] [out] Error and context info.
		const wxString&  sFolder,   //< [in] The working folder to get conflicts from.
		ConflictData&    cConflicts //< [out] Data about the conflicts in the given folder.
		) = 0;

	/**
	 * Rename an object on disk.
	 */
	virtual bool Rename(
		class vvContext&      pCtx,					// [in] [out] Error and context info.
		const wxString& sFullPath_objectToRename,	// [in] The full disk path of the object to rename.  The object must exist on disk.
		const wxString& sNewName					// [in] The new name to give to the item.
		) = 0;

	/**
	 * Commits a working directory to a repo as a new changeset.
	 */
	virtual bool Commit(
		class vvContext&     cContext,         //< [in] [out] Error and context info.
		const wxString&      sFolder,          //< [in] The working folder to commit.
		const wxString&      sComment,         //< [in] Comment to add to the new changeset.
		const wxArrayString* pPaths    = NULL, //< [in] Repo paths of individual dirty items to commit.
		                                       //<      If NULL, then all dirty items will be committed.
		const wxArrayString* pStamps   = NULL, //< [in] List of stamps to apply to the new changeset.
		                                       //<      Passing NULL the same as passing an empty list.
		const wxString*      pUser     = NULL, //< [in] The username to commit as.
		                                       //<      If NULL, then the current user is used.
		vvFlags              cFlags    = 0u,   //< [in] COMMIT_* flags to customize behavior.
		wxString*            pHid      = NULL  //< [out] The HID of the new changeset.
		) = 0;

	/**
	 * Add a comment to an existing changeset.
	 */
	virtual bool AddComment(
		class vvContext& cContext, //< [in] [out] Error and context info.
		const wxString&  sRepoName,    //< [in] The local repo name for the changeset.
		const wxString&  sChangesetHID,  //< [in] The changeset ID
		const wxString&  sNewComment	//< [in] The text of the new comment
		) = 0;

	virtual bool AddStamp(
		class vvContext& cContext, //< [in] [out] Error and context info.
		const wxString&  sRepoName,    //< [in] The local repo name for the changeset.
		const wxString&  sChangesetHID,  //< [in] The changeset ID
		const wxString&  sNewStamp	//< [in] The new stamp
		) = 0;

	virtual bool AddTag(
		class vvContext& cContext, //< [in] [out] Error and context info.
		const wxString&  sRepoName,    //< [in] The local repo name for the changeset.
		const wxString&  sChangesetHID,  //< [in] The changeset ID
		const wxString&  sNewTag	//< [in] The new tag
		) = 0;

	virtual bool DeleteStamp(
		class vvContext& pCtx,			//< [in] [out] Error and context info.
		const wxString&  sRepoName,		//< [in] The local repo name for the changeset.
		const wxString&  sChangesetHID,	//< [in] The changeset ID
		const wxString&  sRemoveMe		//< [in] The stamp to remove
		) = 0;

	virtual bool DeleteTag(
		class vvContext& pCtx,			//< [in] [out] Error and context info.
		const wxString&  sRepoName,		//< [in] The local repo name for the changeset.
		const wxString&  sRemoveMe		//< [in] The tag to remove
		) = 0;
	/**
	 * Move an object on disk.
	 */
	virtual bool Move(
		class vvContext&      pCtx,					// [in] [out] Error and context info.
		const wxArrayString& sFullPath_objectToMove,		// [in] The full disk path of the object to move.  The object must exist on disk.
		const wxString& sNewParentPath				// [in] The new parent path.
		) = 0;
	
	/**
	 * Add an object on disk.
	 */
	virtual bool Add(
		class vvContext&      pCtx,						// [in] [out] Error and context info.
		const wxArrayString& sFullPath_objectToAdd		// [in] The full disk path of the object to add.  The object must exist on disk.
		) = 0;

	/**
	 * Remove an object from the repository.
	 */
	virtual bool Remove(
		class vvContext&      pCtx,						// [in] [out] Error and context info.
		const wxArrayString& sFullPath_objectToRemove	// [in] The full disk path of the object to remove.  The object must exist on disk.
		) = 0;

	/**
	 * Pull changes from a remote server or from a local repo.
	 */
	virtual bool Pull(
		class vvContext& cContext,						//< [in] [out] Error and context info.
		const wxString&  sSrcLocalRepoNameOrRemoteURL,	//< [in] The local repo name, or remote repo URL.  Changes will be pulled from that repo.
		const wxString&  sDestLocalRepoName,			//< [in] The name of the local repo to pull changes into.
		const wxArrayString& saAreas,					//< [in] The area names of areas to update.
		const wxArrayString& saTags,					//< [in] The tags that apply to dagnodes to pull
		const wxArrayString& saDagnodes,				//< [in] The dagnode ids that will be pulled.
		const wxArrayString& saBranchNames				//< [in] The names of branches that will be pulled.
		) = 0;

	/**
	 * Pushes changes from a local repo to another repo.
	 */
	virtual bool Push(
		class vvContext& cContext,						//< [in] [out] Error and context info.
		const wxString&  sDestLocalRepoNameOrRemoteURL,	//< [in] The local repo name, or remote repo URL.  Changes will be pulled from that repo.
		const wxString&  sSrcLocalRepoName,				//< [in] The name of the local repo to push changes from.
		const wxArrayString& saAreas,					//< [in] The area names of areas to update
		const wxArrayString& saTags,					//< [in] The tags that apply to dagnodes to push
		const wxArrayString& saDagnodes,				//< [in] The dagnode ids that will be pushed.
		const wxArrayString& saBranchNames				//< [in] The names of branches that will be pushed.
		) = 0;
		

	/**
	 * Load the changesets which are different bettween a local and remote repos.
	 */
	virtual bool Incoming(
		class vvContext& cContext,						//< [in] [out] Error and context info.
		const wxString&  sSrcLocalRepoNameOrRemoteURL,	//< [in] The local repo name, or remote repo URL.  Changes will be pulled from that repo.
		const wxString&  sDestLocalRepoName,			//< [in] The name of the local repo to pull changes into.
		HistoryData& cHistoryData
		) = 0;

	/**
	 * Load the changesets which are different bettween a local and remote repos.
	 */
	virtual bool Outgoing(
		class vvContext& cContext,						//< [in] [out] Error and context info.
		const wxString&  sDestLocalRepoNameOrRemoteURL,	//< [in] The local repo name, or remote repo URL.  Changes will be pulled from that repo.
		const wxString&  sSrcLocalRepoName,			//< [in] The name of the local repo to pull changes into.
		HistoryData& cHistoryData
		) = 0;

	virtual bool History(
		class vvContext& cContext,						//< [in] [out] Error and context info.
		const wxString&  sLocalRepoName,				//< [in] The name of the local repo to pull changes into.
		const vvHistoryFilterOptions filterOptions,		//< [in] History Options.
		HistoryData& cHistoryData,						//< [out] History Results
		void ** ppNewHistoryToken						//< [out] A token that can be used to fetch more data.
		) = 0;

	virtual bool History__Fetch_More(
		class vvContext& pCtx,							//< [in] [out] Error and context info.
		const wxString&  sLocalRepoName,				//< [in] The name of the local repo to pull changes into.
		unsigned int nNumResultsDesired,				//< [in] The number of additional results desired.
		void * pHistoryToken,							//< [in]  The token that was returned by a previous call to history.  This will be freed if the fetch is successful.
		HistoryData& cHistoryData,						//< [out] The new data.
		void ** ppNewHistoryToken						//< [out] A new token that can be used to fetch more data.
		) = 0;

	virtual bool History__Free_Token(
		class vvContext& pCtx,							//< [in] [out] Error and context info.
		void * pHistoryToken) = 0;							//< [in] The token to free.

	virtual bool Repo_Item_Get_Children(
		class vvContext& pCtx,							//< [in] [out] Error and context info.
		wxString sRepoName,								//< [in] The name of the local repo 
		wxString sChangesetHID,							//< [in] The changeset to browse
		wxString sItemPath,								//< [in] The full path of the item to look up.  This must be a repo path.
		RepoItemList& cChildren) = 0;					//< [out]  The list of children.

	virtual bool Get_Revision_Details(
		class vvContext& pCtx,							//< [in] [out] Error and context info.
		const wxString&  sLocalRepoName,				//< [in] The name of the local repo to pull changes into.
		const wxString& sRevID,							//< [in] The revision to load
		HistoryEntry& cHistoryEntry,						//< [out] The details.
		vvVerbs::RevisionList& children						//< [out] A vector of the children of the revision.
		) = 0;

	virtual bool Export_Revision(
		class vvContext& pCtx,							//< [in] [out] Error and context info.
		const wxString&  sLocalRepoName,				//< [in] The name of the local repo to pull changes into.
		const wxString& sRevID,							//< [in] The revision to load
		wxString sLocalOutputPath						//< [in] The local path to output to.
		) = 0;
};


#endif
