/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include "precompiled.h"
#include "vvVerbsImpl.h"

extern "C" {
#include "sg.h"
}
#include "vvContext.h"
#include "vvSgHelpers.h"
#include "vvHistoryObjects.h"
#include "wx/datetime.h"
/*
**
** Internal Functions
**
*/

/**
 * Data required by _TranslateDiffObject as user data.
 */
struct _TranslateDiffObject_Data
{
	vvVerbs::DiffData* pDiffData; //< The diff data to add translated diff objects to.
	SG_file_spec*      pFilespec; //< The file spec to use to check if an object is ignored.
};

/**
 * Gets the item type associated with a diff object.
 */
vvVerbs::DiffEntry::ItemType _TranslateDiffObject_GetItemType(
	SG_context*                      pCtx,                    //< [in] Error and context info.
	SG_treediff2*                    WXUNUSED(pTreeDiff),     //< [in] The tree diff that owns the object.
	const SG_treediff2_ObjectData*   pObject,                 //< [in] The object to get the data from.
	const _TranslateDiffObject_Data* WXUNUSED(pTranslateData) //< [in] Options to use when translating the object.
	)
{
	SG_treenode_entry_type eType;
	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_tneType(pCtx, pObject, &eType)  );
	switch (eType)
	{
	case SG_TREENODEENTRY_TYPE_REGULAR_FILE:
		return vvVerbs::DiffEntry::ITEM_FILE;
	case SG_TREENODEENTRY_TYPE_DIRECTORY:
		return vvVerbs::DiffEntry::ITEM_FOLDER;
	case SG_TREENODEENTRY_TYPE_SYMLINK:
		return vvVerbs::DiffEntry::ITEM_SYMLINK;
	case SG_TREENODEENTRY_TYPE_SUBMODULE:
		return vvVerbs::DiffEntry::ITEM_SUBMODULE;
	default:
		SG_ERR_THROW2(SG_ERR_INVALIDARG, (pCtx, "Unknown SG_TREENODEENTRY_TYPE: %d", eType));
	}

fail:
	return vvVerbs::DiffEntry::ITEM_COUNT;
}

wxString _TranslateDiffObject_GetItemPath(
	SG_context*                      pCtx,                    //< [in] Error and context info.
	SG_treediff2*                    WXUNUSED(pTreeDiff),     //< [in] The tree diff that owns the object.
	const SG_treediff2_ObjectData*   pObject,                 //< [in] The object to get the data from.
	const _TranslateDiffObject_Data* WXUNUSED(pTranslateData) //< [in] Options to use when translating the object.
	)
{
	const char* szValue = NULL;
	wxString    sResult = wxEmptyString;
	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_repo_path(pCtx, pObject, &szValue)  );
	sResult = vvSgHelpers::Convert_cString_wxString(szValue);

fail:
	return sResult;
}

/**
 * Gets the change flags associated with a diff object.
 */
vvFlags _TranslateDiffObject_GetChangeFlags(
	SG_context*                      pCtx,                    //< [in] Error and context info.
	SG_treediff2*                    WXUNUSED(pTreeDiff),     //< [in] The tree diff that owns the object.
	const SG_treediff2_ObjectData*   pObject,                 //< [in] The object to get the data from.
	const _TranslateDiffObject_Data* WXUNUSED(pTranslateData) //< [in] Options to use when translating the object.
	)
{
	unsigned int uValue         = 0u;
	vvFlags      cIncomingFlags = uValue;
	vvFlags      cOutgoingFlags = 0u;
	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_dsFlags(pCtx, pObject, &uValue)  );
	cIncomingFlags = uValue;
	#define CHECK_FLAG(sgFlag, vvFlag)                           \
		if (cIncomingFlags.HasAnyFlag(sgFlag))                   \
		{                                                        \
			cOutgoingFlags.AddFlags(vvVerbs::DiffEntry::vvFlag); \
		}
	CHECK_FLAG(SG_DIFFSTATUS_FLAGS__ADDED,                   CHANGE_ADDED);
	CHECK_FLAG(SG_DIFFSTATUS_FLAGS__DELETED,                 CHANGE_DELETED);
	CHECK_FLAG(SG_DIFFSTATUS_FLAGS__MODIFIED,                CHANGE_MODIFIED);
	CHECK_FLAG(SG_DIFFSTATUS_FLAGS__LOST,                    CHANGE_LOST);
	CHECK_FLAG(SG_DIFFSTATUS_FLAGS__FOUND,                   CHANGE_FOUND);
	CHECK_FLAG(SG_DIFFSTATUS_FLAGS__RENAMED,                 CHANGE_RENAMED);
	CHECK_FLAG(SG_DIFFSTATUS_FLAGS__MOVED,                   CHANGE_MOVED);
	CHECK_FLAG(SG_DIFFSTATUS_FLAGS__CHANGED_XATTRS,          CHANGE_XATTRS);
	CHECK_FLAG(SG_DIFFSTATUS_FLAGS__CHANGED_ATTRBITS,        CHANGE_ATTRS);
	#undef CHECK_FLAG
	SG_ASSERT(cOutgoingFlags.GetValue() != 0u); // pretty sure we should have at least one flag

fail:
	return cOutgoingFlags;
}

/**
 * Gets whether or not the diff object should be considered ignored.
 */
bool _TranslateDiffObject_GetIgnored(
	SG_context*                      pCtx,                //< [in] Error and context info.
	SG_treediff2*                    WXUNUSED(pTreeDiff), //< [in] The tree diff that owns the object.
	const SG_treediff2_ObjectData*   pObject,             //< [in] The object to get the data from.
	const _TranslateDiffObject_Data* pTranslateData       //< [in] Options to use when translating the object.
	)
{
	bool         bResult = false;
	unsigned int uFlags  = 0u;

	// check the item's status flags
	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_dsFlags(pCtx, pObject, &uFlags)  );
	if (vvFlags(uFlags).HasAnyFlag(SG_DIFFSTATUS_FLAGS__FOUND))
	{
		// FOUND items need to be matched against our filespec to see if they match
		const char* szRepoPath = NULL;
		SG_bool     bMatches   = SG_FALSE;

		SG_ERR_CHECK(  SG_treediff2__ObjectData__get_repo_path(pCtx, pObject, &szRepoPath)  );
		SG_ERR_CHECK(  SG_file_spec__match_path(pCtx, pTranslateData->pFilespec, szRepoPath, SG_FILE_SPEC__MATCH_TRAILING_SLASH, &bMatches)  );
		bResult = !vvSgHelpers::Convert_sgBool_cppBool(bMatches);
	}
	else
	{
		// any other status is never ignored
		bResult = false;
	}

fail:
	return bResult;
}

/**
 * Gets the content HID associated with a diff object.
 */
wxString _TranslateDiffObject_GetContentHid(
	SG_context*                      pCtx,                    //< [in] Error and context info.
	SG_treediff2*                    WXUNUSED(pTreeDiff),     //< [in] The tree diff that owns the object.
	const SG_treediff2_ObjectData*   pObject,                 //< [in] The object to get the data from.
	const _TranslateDiffObject_Data* WXUNUSED(pTranslateData) //< [in] Options to use when translating the object.
	)
{
	const char* szValue = NULL;
	wxString    sResult = wxEmptyString;
	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_content_hid(pCtx, pObject, &szValue)  );
	sResult = vvSgHelpers::Convert_cString_wxString(szValue);

fail:
	return sResult;
}

/**
 * Gets the old content HID associated with a diff object.
 */
wxString _TranslateDiffObject_GetOldContentHid(
	SG_context*                      pCtx,                    //< [in] Error and context info.
	SG_treediff2*                    WXUNUSED(pTreeDiff),     //< [in] The tree diff that owns the object.
	const SG_treediff2_ObjectData*   pObject,                 //< [in] The object to get the data from.
	const _TranslateDiffObject_Data* WXUNUSED(pTranslateData) //< [in] Options to use when translating the object.
	)
{
	const char* szValue = NULL;
	wxString    sResult = wxEmptyString;
	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_old_content_hid(pCtx, pObject, &szValue)  );
	if (szValue != NULL)
	{
		sResult = vvSgHelpers::Convert_cString_wxString(szValue);
	}

fail:
	return sResult;
}

/**
 * Gets the old repo path associated with a diff object.
 */
wxString _TranslateDiffObject_GetMoved_From(
	SG_context*                      pCtx,                    //< [in] Error and context info.
	SG_treediff2*                    pTreeDiff,               //< [in] The tree diff that owns the object.
	const SG_treediff2_ObjectData*   pObject,                 //< [in] The object to get the data from.
	const _TranslateDiffObject_Data* WXUNUSED(pTranslateData) //< [in] Options to use when translating the object.
	)
{
	const char* szValue = NULL;
	wxString    sResult = wxEmptyString;
	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_moved_from_repo_path(pCtx, pTreeDiff, pObject, &szValue)  );
	sResult = vvSgHelpers::Convert_cString_wxString(szValue);

fail:
	return sResult;
}

/**
 * Gets the old name associated with a diff object.
 */
wxString _TranslateDiffObject_GetRenamed_From(
	SG_context*                      pCtx,                    //< [in] Error and context info.
	SG_treediff2*                    WXUNUSED(pTreeDiff),     //< [in] The tree diff that owns the object.
	const SG_treediff2_ObjectData*   pObject,                 //< [in] The object to get the data from.
	const _TranslateDiffObject_Data* WXUNUSED(pTranslateData) //< [in] Options to use when translating the object.
	)
{
	const char* szValue = NULL;
	wxString    sResult = wxEmptyString;
	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_old_name(pCtx, pObject, &szValue)  );
	sResult = vvSgHelpers::Convert_cString_wxString(szValue);

fail:
	return sResult;
}

/**
 * Gets the original submodule description associated with a diff object.
 */
wxString _TranslateDiffObject_GetSubmodule_Original(
	SG_context*                      pCtx,                    //< [in] Error and context info.
	SG_treediff2*                    WXUNUSED(pTreeDiff),     //< [in] The tree diff that owns the object.
	const SG_treediff2_ObjectData*   pObject,                 //< [in] The object to get the data from.
	const _TranslateDiffObject_Data* WXUNUSED(pTranslateData) //< [in] Options to use when translating the object.
	)
{
	SG_treenode_submodule* pSubmodule = NULL;
	char*                  szValue    = NULL;
	wxString               sResult    = wxEmptyString;
	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_old_tsm(pCtx, pObject, &pSubmodule)  );
	SG_ERR_CHECK(  SG_treenode_submodule__describe_for_status(pCtx, pSubmodule, &szValue)  );
	sResult = vvSgHelpers::Convert_cString_wxString(szValue);

fail:
	SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pSubmodule);
	SG_NULLFREE(pCtx, szValue);
	return sResult;
}

/**
 * Gets the current submodule description associated with a diff object.
 */
wxString _TranslateDiffObject_GetSubmodule_Current(
	SG_context*                      pCtx,                    //< [in] Error and context info.
	SG_treediff2*                    WXUNUSED(pTreeDiff),     //< [in] The tree diff that owns the object.
	const SG_treediff2_ObjectData*   pObject,                 //< [in] The object to get the data from.
	const _TranslateDiffObject_Data* WXUNUSED(pTranslateData) //< [in] Options to use when translating the object.
	)
{
	SG_treenode_submodule* pSubmodule = NULL;
	char*                  szValue    = NULL;
	wxString               sResult    = wxEmptyString;
	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_tsm(pCtx, pObject, &pSubmodule)  );
	SG_ERR_CHECK(  SG_treenode_submodule__describe_for_status(pCtx, pSubmodule, &szValue)  );
	sResult = vvSgHelpers::Convert_cString_wxString(szValue);

fail:
	SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pSubmodule);
	SG_NULLFREE(pCtx, szValue);
	return sResult;
}

/**
 * Gets the dirty flag from the current submodule associated with a diff object.
 */
bool _TranslateDiffObject_GetSubmodule_Dirty(
	SG_context*                      pCtx,                    //< [in] Error and context info.
	SG_treediff2*                    WXUNUSED(pTreeDiff),     //< [in] The tree diff that owns the object.
	const SG_treediff2_ObjectData*   pObject,                 //< [in] The object to get the data from.
	const _TranslateDiffObject_Data* WXUNUSED(pTranslateData) //< [in] Options to use when translating the object.
	)
{
	SG_treenode_submodule* pSubmodule = NULL;
	SG_bool                bValue     = SG_FALSE;
	bool                   bResult    = false;
	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_tsm(pCtx, pObject, &pSubmodule)  );
	SG_ERR_CHECK(  SG_treenode_submodule__is_dirty(pCtx, pSubmodule, &bValue)  );
	bResult = vvSgHelpers::Convert_sgBool_cppBool(bValue);

fail:
	SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pSubmodule);
	return bResult;
}

/**
 * Translates an SG_treediff2_ObjectData into a vvVerbs::DiffEntry and adds it to a vvVerbs::DiffData.
 * Intended for use as an SG_treediff2_foreach_callback.
 */
void _TranslateDiffObject(
	SG_context*                    pCtx,      //< [in] [out] Error and context info.
	SG_treediff2*                  pTreeDiff, //< [in] The treediff that is being iterated.
	const char*                    szGid,     //< [in] The GID of the current object.
	const SG_treediff2_ObjectData* pObject,   //< [in] The current object.
	void*                          pUserData  //< [in] A populated instance of _TranslateDiffObject_Data.
	)
{
	_TranslateDiffObject_Data* pData = NULL;
	vvVerbs::DiffEntry         cEntry;

	SG_NULLARGCHECK(pUserData);

	// get the caller-specified data
	pData = static_cast<_TranslateDiffObject_Data*>(pUserData);

	// the logic for when to retrieve which fields is pulled from
	// _sg_report_status_for_section_vhash_cb in sg_treediff2.c
	cEntry.sItemGid     = szGid;
	cEntry.eItemType    = _TranslateDiffObject_GetItemType(pCtx, pTreeDiff, pObject, pData);
	cEntry.sItemPath    = _TranslateDiffObject_GetItemPath(pCtx, pTreeDiff, pObject, pData);
	cEntry.cItemChanges = _TranslateDiffObject_GetChangeFlags(pCtx, pTreeDiff, pObject, pData);
	cEntry.bItemIgnored = _TranslateDiffObject_GetIgnored(pCtx, pTreeDiff, pObject, pData);
	cEntry.sContentHid  = _TranslateDiffObject_GetContentHid(pCtx, pTreeDiff, pObject, pData);
	if (cEntry.eItemType != vvVerbs::DiffEntry::ITEM_FOLDER && cEntry.cItemChanges.HasAnyFlag(vvVerbs::DiffEntry::CHANGE_MODIFIED | vvVerbs::DiffEntry::CHANGE_ATTRS | vvVerbs::DiffEntry::CHANGE_XATTRS))
	{
		cEntry.sOldContentHid = _TranslateDiffObject_GetOldContentHid(pCtx, pTreeDiff, pObject, pData);
	}
	if (cEntry.cItemChanges.HasAnyFlag(vvVerbs::DiffEntry::CHANGE_MOVED))
	{
		cEntry.sMoved_From = _TranslateDiffObject_GetMoved_From(pCtx, pTreeDiff, pObject, pData);
	}
	if (cEntry.cItemChanges.HasAnyFlag(vvVerbs::DiffEntry::CHANGE_RENAMED))
	{
		cEntry.sRenamed_From = _TranslateDiffObject_GetRenamed_From(pCtx, pTreeDiff, pObject, pData);
	}
	if (cEntry.eItemType == vvVerbs::DiffEntry::ITEM_SUBMODULE)
	{
		if (cEntry.cItemChanges.HasAnyFlag(vvVerbs::DiffEntry::CHANGE_MODIFIED | vvVerbs::DiffEntry::CHANGE_ADDED | vvVerbs::DiffEntry::CHANGE_FOUND))
		{
			cEntry.sSubmodule_Current = _TranslateDiffObject_GetSubmodule_Current(pCtx, pTreeDiff, pObject, pData);
			cEntry.bSubmodule_Dirty   = _TranslateDiffObject_GetSubmodule_Dirty(pCtx, pTreeDiff, pObject, pData);
		}
		if (cEntry.cItemChanges.HasAnyFlag(vvVerbs::DiffEntry::CHANGE_MODIFIED | vvVerbs::DiffEntry::CHANGE_DELETED | vvVerbs::DiffEntry::CHANGE_LOST))
		{
			cEntry.sSubmodule_Original = _TranslateDiffObject_GetSubmodule_Original(pCtx, pTreeDiff, pObject, pData);
		}
	}

	// add the entry to the given diff data
	pData->pDiffData->push_back(cEntry);

fail:
	return;
}

/**
 * Translates a single conflict entry from an sglib vhash into a vvVerbs::ConflictEntry.
 */
void _TranslateConflictEntry(
	vvContext&              pCtx,      //< [in] [out] Error and context info.
	SG_vhash*               pIncoming, //< [in] The sglib conflict entry.
	vvVerbs::ConflictEntry& cOutgoing  //< [out] The translated entry.
	)
{
	// see the "issues" section of wd.json (in a repo with conflicts) for the source of these
	static const char* szKey_ItemGid          = "gid";
	static const char* szKey_Status           = "status";

	const char* szValue = NULL;
	SG_string*  sValue  = NULL;
	SG_int64    iValue  = 0;
	vvFlags     cFlags;

	// sItemGid
	SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pIncoming, szKey_ItemGid, &szValue)  );
	cOutgoing.sItemGid = szValue;

	// bResolved
	SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pIncoming, szKey_Status, &iValue)  );
	cFlags = iValue;
	if (cFlags.HasAnyFlag(SG_ISSUE_STATUS__MARKED_RESOLVED))
	{
		cOutgoing.bResolved = true;
	}
	else
	{
		if (!cFlags.FlagsEmpty())
		{
			// If this trips, check sg_pendingtree_typedefs.h for new SG_ISSUE_STATUS__* flags.
			// If more get added, bResolved might have to be changed from a bool to flags.
			wxLogError("Unknown flag in conflict status: %u", cFlags.GetValue());
		}
		cOutgoing.bResolved = false;
	}

fail:
	SG_STRING_NULLFREE(pCtx, sValue);
	return;
}

/**
 * Translates a set of conflict entries from an sglib varray into a vvVerbs::ConflictData.
 * The incoming varray is as returned by SG_pendingtree__get_wd_issues__ref.
 */
void _TranslateConflictData(
	vvContext&             pCtx,      //< [in] [out] Error and context info.
	const SG_varray*       pIncoming, //< [in] The sglib conflict data.
	vvVerbs::ConflictData& cOutgoing  //< [out] The translated data.
	)
{
	// check if we even have an array of conflicts to walk
	if (pIncoming == NULL)
	{
		return;
	}

	// run through each entry in the array
	SG_uint32 uCount = 0u;
	SG_ERR_CHECK(  SG_varray__count(pCtx, pIncoming, &uCount)  );
	for (SG_uint32 uIndex = 0u; uIndex < uCount; ++uIndex)
	{
		SG_vhash*              pEntry = NULL;
		vvVerbs::ConflictEntry cEntry;

		// get and translate the current entry
		SG_ERR_CHECK(  SG_varray__get__vhash(pCtx, pIncoming, uIndex, &pEntry)  );
		SG_ERR_CHECK(  _TranslateConflictEntry(pCtx, pEntry, cEntry)  );

		// add the new entry to the list
		cOutgoing.push_back(cEntry);
	}

fail:
	return;
}


/*
**
** Public Functions
**
*/

bool vvVerbsImpl::Initialize(
	vvContext& pCtx
	)
{
	SG_lib__global_initialize(pCtx);
	return !pCtx.Error_Check();
}

void vvVerbsImpl::Shutdown(
	vvContext& pCtx
	)
{
	SG_lib__global_cleanup(pCtx);
}

bool vvVerbsImpl::GetLocalRepos(
	vvContext&     pCtx,
	wxArrayString& cLocalRepos
	)
{
	// regular expression to filter out test repos
	static wxRegEx cTestNamePattern("^t[0-9A-Fa-f]{32}$|^vscript test repo for repo[0-9]{13}$");

	SG_vhash*  pRepoHash  = NULL;
	SG_varray* pRepoList  = NULL;
	SG_uint32  uRepoCount = NULL;

	// setup
	pCtx.Log_PushOperation("Retrieving local repositories", vvContext::LOG_FLAG_VERBOSE);
	SG_ASSERT(cTestNamePattern.IsValid());

	// get the list of repos from the closet
	SG_ERR_CHECK(  SG_closet__descriptors__list(pCtx, &pRepoHash)  );
	SG_ERR_CHECK(  SG_vhash__get_keys(pCtx, pRepoHash, &pRepoList)  );

	// run through the list
	SG_ERR_CHECK(  SG_varray__count(pCtx, pRepoList, &uRepoCount)  );
	pCtx.Log_SetSteps(uRepoCount);
	for (SG_uint32 uIndex = 0u; uIndex < uRepoCount; ++uIndex)
	{
		const char* szRepoName = NULL;

		// get the current repo name
		SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pRepoList, uIndex, &szRepoName)  );
		pCtx.Log_SetStep(szRepoName);

		// check if this is a test repo
		if (cTestNamePattern.Matches(szRepoName))
		{
			// skip the test repo
			pCtx.Log_ReportVerbose("Skipping test repository: %s", szRepoName);
		}
		else
		{
			// add the real repo to the list
			pCtx.Log_ReportVerbose("Found repository: %s", szRepoName);
			cLocalRepos.Add(szRepoName);
		}

		// next
		pCtx.Log_FinishStep();
	}

fail:
	SG_VARRAY_NULLFREE(pCtx, pRepoList);
	SG_VHASH_NULLFREE(pCtx, pRepoHash);
	pCtx.Log_PopOperation();
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::LocalRepoExists(
	vvContext&      pCtx,
	const wxString& sName
	)
{
	SG_vhash*  pRepoHash  = NULL;
	SG_varray* pRepoList  = NULL;
	SG_uint32  uRepoCount = NULL;
	bool       bFound     = false;

	pCtx.Log_PushOperation("Checking if repository exists", vvContext::LOG_FLAG_VERBOSE);
	pCtx.Log_SetValue("Repository", sName, vvContext::LOG_FLAG_INPUT);

	// get the list of repos from the closet
	SG_ERR_CHECK(  SG_closet__descriptors__list(pCtx, &pRepoHash)  );
	SG_ERR_CHECK(  SG_vhash__get_keys(pCtx, pRepoHash, &pRepoList)  );

	// run through the list
	SG_ERR_CHECK(  SG_varray__count(pCtx, pRepoList, &uRepoCount)  );
	pCtx.Log_SetSteps(uRepoCount);
	for (SG_uint32 uIndex = 0u; uIndex < uRepoCount && !bFound; ++uIndex)
	{
		const char* szRepoName = NULL;

		// get the current repo name
		SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pRepoList, uIndex, &szRepoName)  );
		pCtx.Log_SetStep(szRepoName);

		// check if this is the name we're looking for
		if (sName == szRepoName) {
			bFound = true;
		}
		pCtx.Log_FinishStep();
	}

	// mention our result
	pCtx.Log_SetValue("Exists", bFound, vvContext::LOG_FLAG_OUTPUT);

fail:
	SG_VARRAY_NULLFREE(pCtx, pRepoList);
	SG_VHASH_NULLFREE(pCtx, pRepoHash);
	pCtx.Log_PopOperation();
	return bFound;
}

bool vvVerbsImpl::GetRemoteRepos(
	vvContext&     pCtx,
	wxString& sRepoDescriptorName,
	wxArrayString& cRemoteRepos,
	wxArrayString& cRemoteRepoAliases
	)
{
	SG_string * pString = NULL;
	SG_vhash * pvhBindings = NULL;
	SG_uint32 nCountBindings = 0;
	SG_uint32 i = 0;
	const SG_variant * pVariant = NULL;
	const char * pszKeyName = NULL;
	const char * pszRepoString = NULL;
	SG_varray * pva_targets = NULL;

	//SG_context__alloc(&pCtx);
	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pString)  );
	SG_ERR_CHECK(  SG_string__sprintf(pCtx, pString, "%s/%s/%s",
									  SG_LOCALSETTING__SCOPE__INSTANCE,
									  (const char *)sRepoDescriptorName.c_str(),
									  SG_LOCALSETTING__PATHS)  );
	

	SG_ERR_IGNORE(  SG_localsettings__get__vhash(pCtx,
												 SG_string__sz(pString),
												 NULL,		// null repo because we only want /instance/ settings.
												 &pvhBindings,
												 NULL)  );
	if (pvhBindings != NULL)
	{
		SG_ERR_CHECK(  SG_vhash__count(pCtx, pvhBindings, &nCountBindings)  );
		for (i = 0; i < nCountBindings; i++)
		{
			SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvhBindings, i, &pszKeyName, &pVariant)  );
			SG_ERR_CHECK(  SG_variant__get__sz(pCtx, pVariant, &pszRepoString)  );
			cRemoteRepoAliases.Add(pszKeyName);
			cRemoteRepos.Add(pszRepoString);
		}
	}

	//Now get the "previous sync targets" that we save in localsettings
	SG_ERR_CHECK(  SG_string__sprintf(pCtx, pString, "%s/%s/%s",
									SG_LOCALSETTING__SCOPE__INSTANCE,
									(const char *)sRepoDescriptorName.c_str(),
									SG_LOCALSETTING__SYNC_TARGETS)  );
	SG_ERR_CHECK(  SG_localsettings__get__varray(pCtx, SG_string__sz(pString), NULL, &pva_targets, NULL)  );
	if (pva_targets)
	{
		SG_uint32 nCountTargets = 0;
		SG_uint32 nIndex = 0;
		SG_ERR_CHECK(  SG_varray__count(pCtx, pva_targets, &nCountTargets)  );
		for (nIndex = 0; nIndex < nCountTargets; nIndex++)
		{
			SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pva_targets, nIndex, &pszRepoString)  );
			if (cRemoteRepos.Index(pszRepoString, false) < 0)
			{
				//We haven't added this path yet, add it (with an empty alias)
				cRemoteRepoAliases.Add("");
				cRemoteRepos.Add(pszRepoString);
			}
		}
	}
	SG_VARRAY_NULLFREE(pCtx, pva_targets);
	SG_STRING_NULLFREE(pCtx, pString);
	
	SG_VHASH_NULLFREE(pCtx, pvhBindings);
	// success
	return true;
fail:
	SG_VARRAY_NULLFREE(pCtx, pva_targets);
	SG_STRING_NULLFREE(pCtx, pString);
	SG_VHASH_NULLFREE(pCtx, pvhBindings);
	return false;
}

bool vvVerbsImpl::FindWorkingFolder(
	class vvContext& pCtx,
	const wxString&  sPath,
	wxString*        pFolder,
	wxString*        pRepo,
	wxArrayString*        pBaselineHIDs
	)
{
	SG_pathname* pInputPath  = NULL;
	SG_pathname* pOutputPath = NULL;
	SG_string*   pOutputRepo = NULL;
	SG_pendingtree * pPendingTree = NULL;

	pCtx.Log_PushOperation("Finding working folder", vvContext::LOG_FLAG_VERBOSE);
	pCtx.Log_SetValue("Path", sPath, vvContext::LOG_FLAG_INPUT);

	SG_PATHNAME__ALLOC__SZ(pCtx, &pInputPath, sPath.c_str());
	SG_ERR_CHECK(  SG_workingdir__find_mapping2(pCtx, pInputPath, SG_TRUE, &pOutputPath, &pOutputRepo, NULL)  );

	if (pFolder != NULL)
	{
		pFolder->assign(vvSgHelpers::Convert_sgPathname_wxString(pOutputPath));
		pCtx.Log_SetValue("Working folder", *pFolder, vvContext::LOG_FLAG_OUTPUT);
	}
	if (pRepo != NULL)
	{
		pRepo->assign(vvSgHelpers::Convert_sgString_wxString(pOutputRepo));
		pCtx.Log_SetValue("Repository", *pRepo, vvContext::LOG_FLAG_OUTPUT);
	}
	if (pBaselineHIDs != NULL)
	{
		const SG_varray * pvaParents = NULL;
		SG_uint32 k, nrParents;
		const char * psz_hid_parent_k = NULL;
		SG_ERR_CHECK( SG_PENDINGTREE__ALLOC(pCtx, pOutputPath, &pPendingTree)  );
		SG_ERR_CHECK(  SG_pendingtree__get_wd_parents__ref(pCtx, pPendingTree, &pvaParents)  );
		SG_ERR_CHECK(  SG_varray__count(pCtx, pvaParents, &nrParents)  );
		for (k=0; k<nrParents; k++)
		{
			SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pvaParents, k, &psz_hid_parent_k)  );
			pBaselineHIDs->Add(psz_hid_parent_k);
		}
	}

fail:
	SG_PATHNAME_NULLFREE(pCtx, pInputPath);
	SG_PATHNAME_NULLFREE(pCtx, pOutputPath);
	SG_STRING_NULLFREE(pCtx, pOutputRepo);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	pCtx.Log_PopOperation();
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::GetRepoPath(
	vvContext&      pCtx,
	const wxString& sAbsolutePath,
	const wxString& sWorkingFolder,
	wxString&       sRepoPath
	)
{
	SG_pathname* pAbsolutePath  = NULL;
	SG_pathname* pWorkingFolder = NULL;
	SG_string*   pRepoPath      = NULL;

	// setup
	pCtx.Log_PushOperation("Converting to repo path", vvContext::LOG_FLAG_VERBOSE);
	pCtx.Log_SetValue("Absolute Path", sAbsolutePath, vvContext::LOG_FLAG_INPUT);
	pCtx.Log_SetValue("Working Folder", sWorkingFolder, vvContext::LOG_FLAG_INPUT);

	// do the actual conversion
	pAbsolutePath  = vvSgHelpers::Convert_wxString_sgPathname(pCtx, sAbsolutePath);
	pWorkingFolder = vvSgHelpers::Convert_wxString_sgPathname(pCtx, sWorkingFolder);
	SG_ERR_CHECK(  SG_workingdir__wdpath_to_repopath(pCtx, pWorkingFolder, pAbsolutePath, SG_FALSE, &pRepoPath)  );

	// return the result
	sRepoPath.assign(vvSgHelpers::Convert_sgString_wxString(pRepoPath));
	pCtx.Log_SetValue("Repository Path", sRepoPath, vvContext::LOG_FLAG_OUTPUT);

fail:
	SG_PATHNAME_NULLFREE(pCtx, pAbsolutePath);
	SG_PATHNAME_NULLFREE(pCtx, pWorkingFolder);
	SG_STRING_NULLFREE(pCtx, pRepoPath);
	pCtx.Log_PopOperation();
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::GetBranches(
	vvContext&      pCtx,
	const wxString& sRepo,
	bool bIncludeClosed,
	wxArrayString&  cBranches
	)
{
	static const char* szKey_Branches = "branches";
	static const char* szKey_Closed = "closed";
	static const char* szKey_Values   = "values";

	SG_repo*  pRepo        = NULL;
	SG_vhash* pBranchPile  = NULL;
	SG_vhash* pBranches    = NULL;
	SG_vhash* pvh_pile_closed = NULL;
	SG_uint32 uBranchCount = 0u;

	// setup
	pCtx.Log_PushOperation("Retrieving branches", vvContext::LOG_FLAG_VERBOSE);
	pCtx.Log_SetValue("Repository", sRepo, vvContext::LOG_FLAG_INPUT);

	// open the repo and get a list of branches from it
	SG_ERR_CHECK(  SG_REPO__OPEN_REPO_INSTANCE(pCtx, sRepo.c_str(), &pRepo)  );
	SG_ERR_CHECK(  SG_vc_branches__cleanup(pCtx, pRepo, &pBranchPile)  );
	SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pBranchPile, szKey_Branches, &pBranches)  );
	SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pBranchPile, szKey_Closed, &pvh_pile_closed)  );

	// run through the list
	SG_ERR_CHECK(  SG_vhash__count(pCtx, pBranches, &uBranchCount)  );
	pCtx.Log_SetSteps(uBranchCount);
	for (SG_uint32 uIndex = 0u; uIndex < uBranchCount; ++uIndex)
	{
		const char* szName = NULL;
		SG_bool b_include_this_one = SG_FALSE;
		// get the current branch name and add it to the list
		SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pBranches, uIndex, &szName, NULL)  );
		
		if (bIncludeClosed)
        {
            b_include_this_one = SG_TRUE;
        }
        else
        {
            SG_bool b_closed = SG_FALSE;

            SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh_pile_closed, szName, &b_closed)  );
            if (b_closed)
            {
                b_include_this_one = SG_FALSE;
            }
            else
            {
                b_include_this_one = SG_TRUE;
            }
        }

		if (b_include_this_one == SG_FALSE)
		{
			pCtx.Log_ReportVerbose("Skipping closed branch: %s", szName);
		}
		else
		{
			pCtx.Log_ReportVerbose("Found branch: %s", szName);
			cBranches.Add(szName);
		}
		pCtx.Log_FinishStep();
	}

fail:
	SG_VHASH_NULLFREE(pCtx, pBranchPile);
	SG_REPO_NULLFREE(pCtx, pRepo);
	pCtx.Log_PopOperation();
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::GetUsers(
	vvContext&       pCtx,
	const wxString&  sRepo,
	stlUserDataList& cUsers
	)
{
	static const char* szKey_Gid    = "recid";
	static const char* szKey_Name   = "name";
	static const char* szKey_Prefix = "prefix";

	SG_repo*   pRepo  = NULL;
	SG_varray* pUsers = NULL;
	SG_uint32  uCount = 0u;

	// setup
	pCtx.Log_PushOperation("Retrieving users", vvContext::LOG_FLAG_VERBOSE);
	pCtx.Log_SetValue("Repository", sRepo, vvContext::LOG_FLAG_INPUT);

	// open the repo and get a list of users from it
	SG_ERR_CHECK(  SG_REPO__OPEN_REPO_INSTANCE(pCtx, sRepo.c_str(), &pRepo)  );
	SG_ERR_CHECK(  SG_user__list_all(pCtx, pRepo, &pUsers)  );

	// run through the list
	SG_ERR_CHECK(  SG_varray__count(pCtx, pUsers, &uCount)  );
	pCtx.Log_SetSteps(uCount);
	for (SG_uint32 uIndex = 0u; uIndex < uCount; ++uIndex)
	{
		SG_vhash*   pUser   = NULL;
		const char* szValue = NULL;
		UserData    cUser;

		// get the current user data and add it to the list
		SG_ERR_CHECK(  SG_varray__get__vhash(pCtx, pUsers, uIndex, &pUser)  );
		SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pUser, szKey_Gid, &szValue)  );
		cUser.sGid = szValue;
		SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pUser, szKey_Name, &szValue)  );
		cUser.sName = szValue;
		SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pUser, szKey_Prefix, &szValue)  );
		cUser.sPrefix = szValue;
		pCtx.Log_ReportVerbose("Found user: %s", cUser.sName);
		cUsers.push_back(cUser);

		// done with this one
		pCtx.Log_FinishStep();
	}

fail:
	SG_VARRAY_NULLFREE(pCtx, pUsers);
	SG_REPO_NULLFREE(pCtx, pRepo);
	pCtx.Log_PopOperation();
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::GetCurrentUser(
	class vvContext& pCtx,
	const wxString&  sRepo,
	wxString&        sUser
	)
{
	SG_repo* pRepo  = NULL;
	char*    szUser = NULL;

	// setup
	pCtx.Log_PushOperation("Retrieving current user", vvContext::LOG_FLAG_VERBOSE);
	pCtx.Log_SetValue("Repository", sRepo, vvContext::LOG_FLAG_INPUT);

	// open the repo
	SG_ERR_CHECK(  SG_REPO__OPEN_REPO_INSTANCE(pCtx, sRepo.c_str(), &pRepo)  );

	// get the current username
	SG_user__get_username_for_repo(pCtx, pRepo, &szUser);
	if (SG_context__err_equals(pCtx, SG_ERR_NO_CURRENT_WHOAMI) == SG_TRUE)
	{
		sUser = wxEmptyString;
		pCtx.Error_Reset();
	}
	else if (SG_context__has_err(pCtx) == SG_TRUE)
	{
		SG_ERR_CHECK_CURRENT;
	}
	else
	{
		sUser = szUser;
	}

fail:
	SG_NULLFREE(pCtx, szUser);
	SG_REPO_NULLFREE(pCtx, pRepo);
	pCtx.Log_PopOperation();
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::GetStamps(
	vvContext&        pCtx,
	const wxString&   sRepo,
	stlStampDataList& cStamps
	)
{
	static const char* szKey_Name  = "stamp";
	static const char* szKey_Count = "count";

	SG_repo*   pRepo   = NULL;
	SG_varray* pStamps = NULL;
	SG_uint32  uCount  = 0u;

	// setup
	pCtx.Log_PushOperation("Retrieving stamps", vvContext::LOG_FLAG_VERBOSE);
	pCtx.Log_SetValue("Repository", sRepo, vvContext::LOG_FLAG_INPUT);

	// open the repo and get a list of stamps from it
	SG_ERR_CHECK(  SG_REPO__OPEN_REPO_INSTANCE(pCtx, sRepo.c_str(), &pRepo)  );
	SG_ERR_CHECK(  SG_vc_stamps__list_all_stamps(pCtx, pRepo, &pStamps)  );

	// run through the list
	SG_ERR_CHECK(  SG_varray__count(pCtx, pStamps, &uCount)  );
	pCtx.Log_SetSteps(uCount);
	for (SG_uint32 uIndex = 0u; uIndex < uCount; ++uIndex)
	{
		SG_vhash*   pStamp  = NULL;
		const char* szValue = NULL;
		SG_int64    iValue  = 0u;
		StampData   cStamp;

		// get the current stamp data and add it to the list
		SG_ERR_CHECK(  SG_varray__get__vhash(pCtx, pStamps, uIndex, &pStamp)  );
		SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pStamp, szKey_Name, &szValue)  );
		cStamp.sName = szValue;
		SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pStamp, szKey_Count, &iValue)  );
		cStamp.uCount = static_cast<unsigned int>(iValue);
		pCtx.Log_ReportVerbose("Found stamp: %s", cStamp.sName);
		cStamps.push_back(cStamp);

		// done with this one
		pCtx.Log_FinishStep();
	}

fail:
	SG_VARRAY_NULLFREE(pCtx, pStamps);
	SG_REPO_NULLFREE(pCtx, pRepo);
	pCtx.Log_PopOperation();
	return !pCtx.Error_Check();
}


bool vvVerbsImpl::GetTags(
	vvContext&        pCtx,
	const wxString&   sRepo,
	vvVerbs::stlTagDataList& cTags
	)
{
	static const char* szKey_Name  = "tag";
	static const char* szKey_CSID = "csid";

	SG_repo*   pRepo   = NULL;
	SG_varray* pTags = NULL;
	SG_uint32  uCount  = 0u;

	// setup
	pCtx.Log_PushOperation("Retrieving tags", vvContext::LOG_FLAG_VERBOSE);
	pCtx.Log_SetValue("Repository", sRepo, vvContext::LOG_FLAG_INPUT);

	// open the repo and get a list of stamps from it
	SG_ERR_CHECK(  SG_REPO__OPEN_REPO_INSTANCE(pCtx, sRepo.c_str(), &pRepo)  );
	SG_ERR_CHECK(  SG_vc_tags__list_all(pCtx, pRepo, &pTags)  );

	// run through the list
	SG_ERR_CHECK(  SG_varray__count(pCtx, pTags, &uCount)  );
	pCtx.Log_SetSteps(uCount);
	for (SG_uint32 uIndex = 0u; uIndex < uCount; ++uIndex)
	{
		SG_vhash*   pTag  = NULL;
		const char* szValue = NULL;
		vvVerbs::TagData cTag;

		// get the current stamp data and add it to the list
		SG_ERR_CHECK(  SG_varray__get__vhash(pCtx, pTags, uIndex, &pTag)  );
		SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pTag, szKey_Name, &szValue)  );
		cTag.sName = szValue;
		SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pTag, szKey_CSID, &szValue)  );
		cTag.sCSID = szValue;
		pCtx.Log_ReportVerbose("Found tag: %s", cTag.sName);
		cTags.push_back(cTag);

		// done with this one
		pCtx.Log_FinishStep();
	}

fail:
	SG_VARRAY_NULLFREE(pCtx, pTags);
	SG_REPO_NULLFREE(pCtx, pRepo);
	pCtx.Log_PopOperation();
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::GetBranch(
	vvContext&      pCtx,
	const wxString& sFolder,
	wxString&       sBranch
	)
{
	char*        szBranch = NULL;
	SG_pathname* pPath    = NULL;

	pCtx.Log_PushOperation("Retrieving tied branch", vvContext::LOG_FLAG_VERBOSE);
	pCtx.Log_SetValue("Folder", sFolder, vvContext::LOG_FLAG_INPUT);

	pPath = vvSgHelpers::Convert_wxString_sgPathname(pCtx, sFolder);
	SG_ERR_CHECK(  SG_workingdir__get_branch_info(pCtx, pPath, &szBranch)  );

	if (szBranch == NULL)
	{
		sBranch.Clear();
	}
	else
	{
		sBranch = szBranch;
	}
	pCtx.Log_SetValue("Branch", sBranch, vvContext::LOG_FLAG_OUTPUT);

fail:
	SG_NULLFREE(pCtx, szBranch);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	pCtx.Log_PopOperation();
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::New(
	vvContext&      pCtx,
	const wxString& sName,
	const wxString& sFolder,
	const wxString& sStorageDriver,
	const wxString& sHashMethod
	)
{
	const char* szName          = NULL;
	const char* szFolder        = NULL;
	SG_bool     bNoWD           = SG_TRUE;
	const char* szStorageDriver = NULL;
	const char* szHashMethod    = NULL;

	pCtx.Log_PushOperation("Creating new repository", vvContext::LOG_FLAG_NONE);
	pCtx.Log_SetValue("Repository", sName, vvContext::LOG_FLAG_INPUT);

	SG_ARGCHECK(!sName.IsEmpty(), sName);

	szName = sName.c_str();
	if (!sFolder.IsEmpty())
	{
		szFolder = sFolder.c_str();
		bNoWD    = SG_FALSE;
		pCtx.Log_SetValue("Checkout folder", sFolder, vvContext::LOG_FLAG_INPUT);
	}
	if (!sStorageDriver.IsEmpty())
	{
		szStorageDriver = sStorageDriver.c_str();
		pCtx.Log_SetValue("Storage driver", sStorageDriver, vvContext::LOG_FLAG_INPUT);
	}
	if (!sHashMethod.IsEmpty())
	{
		szHashMethod = sHashMethod.c_str();
		pCtx.Log_SetValue("Hash method", sHashMethod, vvContext::LOG_FLAG_INPUT);
	}

	SG_ERR_CHECK(  SG_vv_verbs__init_new_repo(pCtx, szName, szFolder, szStorageDriver, szHashMethod, bNoWD, SG_FALSE, NULL, NULL, NULL)  );

fail:
	pCtx.Log_PopOperation();
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::Clone(
	vvContext&      pCtx,
	const wxString& sExistingRepo,
	const wxString& sNewRepoName
	)
{
	const char* szExistingRepo = NULL;
	const char* szNewRepoName  = NULL;

	pCtx.Log_PushOperation("Cloning repository", vvContext::LOG_FLAG_NONE);
	pCtx.Log_SetValue("Source", sExistingRepo, vvContext::LOG_FLAG_INPUT);
	pCtx.Log_SetValue("Target", sNewRepoName, vvContext::LOG_FLAG_INPUT);

	SG_ARGCHECK(!sExistingRepo.IsEmpty(), sExistingRepo);
	SG_ARGCHECK(!sNewRepoName.IsEmpty(), sNewRepoName);

	szExistingRepo = sExistingRepo.c_str();
	szNewRepoName  = sNewRepoName.c_str();

	SG_ERR_CHECK(  SG_clone__exact(pCtx, szExistingRepo, szNewRepoName)  );
	SG_ERR_CHECK(  SG_localsettings__descriptor__update__sz(pCtx, szNewRepoName, SG_LOCALSETTING__PATHS_DEFAULT, szExistingRepo)  );

fail:
	pCtx.Log_PopOperation();
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::Update(
	vvContext&      pCtx,
	const wxString& sFolder,
	CheckoutType    eType,
	const wxString& sTarget
	)
{
	SG_vhash * pvhResult = NULL;
	SG_rev_spec* pRevSpec         = NULL;
	char*        szResolvedTarget = NULL;
	char*        szResolvedBranch = NULL;

	pCtx.Log_PushOperation("Updating Working Folder", vvContext::LOG_FLAG_NONE);
	pCtx.Log_SetValue("Target Folder", sFolder, vvContext::LOG_FLAG_INPUT);

	SG_ERR_CHECK(  SG_rev_spec__alloc(pCtx, &pRevSpec)  );

	switch (eType)
	{
	case CHECKOUT_HEAD:
		break;
	case CHECKOUT_BRANCH:
		SG_ERR_CHECK(  SG_rev_spec__add_branch(pCtx, pRevSpec, sTarget.c_str())  );
		pCtx.Log_SetValue("Target Branch", sTarget, vvContext::LOG_FLAG_INPUT);
		break;
	case CHECKOUT_CHANGESET:
		SG_ERR_CHECK(  SG_rev_spec__add_rev(pCtx, pRevSpec, sTarget.c_str())  );
		pCtx.Log_SetValue("Target Changeset", sTarget, vvContext::LOG_FLAG_INPUT);
		break;
	case CHECKOUT_TAG:
		SG_ERR_CHECK(  SG_rev_spec__add_tag(pCtx, pRevSpec, sTarget.c_str())  );
		pCtx.Log_SetValue("Target Tag", sTarget, vvContext::LOG_FLAG_INPUT);
		break;
	default:
		SG_ERR_THROW(SG_ERR_INVALIDARG);
	}

	SG_ERR_CHECK(  SG_vv_utils__rev_spec_to_csid__wd(pCtx, sFolder.c_str(), pRevSpec, &szResolvedTarget, &szResolvedBranch)  );
	
	SG_ERR_CHECK(  SG_vv_verbs__update2(pCtx, sFolder.c_str(), szResolvedTarget, szResolvedBranch ? szResolvedBranch : sTarget.c_str(), SG_TRUE, SG_FALSE, NULL, &pvhResult)  );

fail:
	SG_NULLFREE(pCtx, szResolvedTarget);
	SG_NULLFREE(pCtx, szResolvedBranch);
	SG_REV_SPEC_NULLFREE(pCtx, pRevSpec);
	SG_VHASH_NULLFREE(pCtx, pvhResult);
	pCtx.Log_PopOperation();
	return !pCtx.Error_Check();
}
bool vvVerbsImpl::Checkout(
	vvContext&      pCtx,
	const wxString& sName,
	const wxString& sFolder,
	CheckoutType    eType,
	const wxString& sTarget
	)
{
	const char*  szName           = NULL;
	const char*  szFolder         = NULL;
	char*        szResolvedTarget = NULL;
	char*        szResolvedBranch = NULL;
	SG_rev_spec* pRevSpec         = NULL;

	pCtx.Log_PushOperation("Checking out repository", vvContext::LOG_FLAG_NONE);
	pCtx.Log_SetValue("Repository", sName, vvContext::LOG_FLAG_INPUT);
	pCtx.Log_SetValue("Target Folder", sFolder, vvContext::LOG_FLAG_INPUT);

	SG_ARGCHECK(!sName.IsEmpty(), sName);
	SG_ARGCHECK(!sFolder.IsEmpty(), sFolder);
	SG_ARGCHECK(0 <= eType && eType < CHECKOUT_COUNT, eType);

	SG_ERR_CHECK(  SG_rev_spec__alloc(pCtx, &pRevSpec)  );

	szName   = sName.c_str();
	szFolder = sFolder.c_str();
	switch (eType)
	{
	case CHECKOUT_HEAD:
		break;
	case CHECKOUT_BRANCH:
		SG_ERR_CHECK(  SG_rev_spec__add_branch(pCtx, pRevSpec, sTarget.c_str())  );
		pCtx.Log_SetValue("Target Branch", sTarget, vvContext::LOG_FLAG_INPUT);
		break;
	case CHECKOUT_CHANGESET:
		SG_ERR_CHECK(  SG_rev_spec__add_rev(pCtx, pRevSpec, sTarget.c_str())  );
		pCtx.Log_SetValue("Target Changeset", sTarget, vvContext::LOG_FLAG_INPUT);
		break;
	case CHECKOUT_TAG:
		SG_ERR_CHECK(  SG_rev_spec__add_tag(pCtx, pRevSpec, sTarget.c_str())  );
		pCtx.Log_SetValue("Target Tag", sTarget, vvContext::LOG_FLAG_INPUT);
		break;
	default:
		SG_ERR_THROW(SG_ERR_INVALIDARG);
	}

	SG_ERR_CHECK(  SG_vv_utils__rev_spec_to_csid__no_wd(pCtx, szName, pRevSpec, &szResolvedTarget, &szResolvedBranch)  );
	if (szResolvedTarget != NULL)
	{
		pCtx.Log_SetValue("Resolved Changeset", szResolvedTarget, vvContext::LOG_FLAG_INTERMEDIATE);
	}
	if (szResolvedBranch != NULL)
	{
		pCtx.Log_SetValue("Resolved Branch", szResolvedBranch, vvContext::LOG_FLAG_INTERMEDIATE);
	}

	// TODO 2011/01/25 Andy, I added an extra arg to SG_vv_verbs__checkout() that
	// TODO            returns detailed info when we can't do a checkout because
	// TODO            we don't have one of the required submodules or one of the
	// TODO            required submodule csets.  I don't know if you have the
	// TODO            ability (or want) to show the user this (any maybe help them
	// TODO            clone and/or pull as necessary and try again).  For now, I'm
	// TODO            passing NULL which will cause the lower layer to throw.
	//
	// TODO 2011/01/27 Andy, I also added a vector of "hints" for the repo-names
	// TODO            of nested submodules.  For now, I'm just going to set this
	// TODO            to null.  I'm not sure you could use it anyway at this point.
	SG_ERR_CHECK(  SG_vv_verbs__checkout(pCtx, szName, szResolvedTarget, szFolder, SG_TRUE, SG_FALSE, szResolvedBranch, NULL, NULL)  );

fail:
	SG_NULLFREE(pCtx, szResolvedTarget);
	SG_NULLFREE(pCtx, szResolvedBranch);
	SG_REV_SPEC_NULLFREE(pCtx, pRevSpec);
	pCtx.Log_PopOperation();
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::Status(
	vvContext&      pCtx,
	const wxString& sFolder,
	const wxString& sRevision,
	DiffData&       cDiff
	)
{
	SG_pathname*    pPath        = NULL;
	SG_pendingtree* pPendingTree = NULL;
	SG_repo*        pRepo        = NULL;
	char*           szHid        = NULL;
	SG_treediff2*   pDiff        = NULL;
	SG_pathname*    pPathWorkingDirTop = NULL;

	_TranslateDiffObject_Data cTranslateData;
	cTranslateData.pDiffData = &cDiff;
	cTranslateData.pFilespec = NULL;

	pCtx.Log_PushOperation("Getting status", vvContext::LOG_FLAG_VERBOSE);
	pCtx.Log_SetValue("Folder", sFolder, vvContext::LOG_FLAG_INPUT);
	if (sRevision != wxEmptyString)
	{
		pCtx.Log_SetValue("Revision", sRevision, vvContext::LOG_FLAG_INPUT);
	}

	// allocate a pendingtree for the specified folder
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPath, sFolder.c_str())  );
	SG_ERR_CHECK(  SG_pendingtree__alloc(pCtx, pPath, &pPendingTree)  );
	SG_ERR_CHECK(  SG_pendingtree__get_working_directory_top(pCtx, pPendingTree, &pPathWorkingDirTop)  );

	// look up the specified revision, if any
	if (sRevision != wxEmptyString)
	{
		SG_ERR_CHECK(  SG_pendingtree__get_repo(pCtx, pPendingTree, &pRepo)  );
		SG_ERR_CHECK(  SG_repo__hidlookup__dagnode(pCtx, pRepo, SG_DAGNUM__VERSION_CONTROL, sRevision.c_str(), &szHid)  );
	}

	// run the status
	SG_ERR_CHECK(  SG_pendingtree__diff_or_status(pCtx, pPendingTree, szHid, NULL, &pDiff)  );

	// translate the status into new diff entries on the given diff data
	SG_ERR_CHECK(  SG_FILE_SPEC__ALLOC(pCtx, &cTranslateData.pFilespec)  );
	SG_ERR_CHECK(  SG_file_spec__load_ignores(pCtx, cTranslateData.pFilespec, pRepo, pPathWorkingDirTop, 0u)  );
	SG_ERR_CHECK(  SG_treediff2__foreach(pCtx, pDiff, _TranslateDiffObject, &cTranslateData)  );

fail:
	SG_FILE_SPEC_NULLFREE(pCtx, cTranslateData.pFilespec);
	SG_NULLFREE(pCtx, szHid);
	SG_TREEDIFF2_NULLFREE(pCtx, pDiff);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_PATHNAME_NULLFREE(pCtx, pPathWorkingDirTop);
	pCtx.Log_PopOperation();
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::Conflicts(
	vvContext&      pCtx,
	const wxString& sFolder,
	ConflictData&   cConflicts
	)
{
	SG_pathname*     pPath        = NULL;
	SG_pendingtree*  pPendingTree = NULL;
	SG_bool          bConflicts   = SG_FALSE;
	const SG_varray* pConflicts   = NULL;

	pCtx.Log_PushOperation("Getting conflicts", vvContext::LOG_FLAG_VERBOSE);
	pCtx.Log_SetValue("Folder", sFolder, vvContext::LOG_FLAG_INPUT);

	// allocate a pendingtree for the specified folder
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPath, sFolder.c_str())  );
	SG_ERR_CHECK(  SG_pendingtree__alloc(pCtx, pPath, &pPendingTree)  );

	// get the conflicts
	SG_ERR_CHECK(  SG_pendingtree__get_wd_issues__ref(pCtx, pPendingTree, &bConflicts, &pConflicts)  );

	// translate the vhash data into our ConflictData
	SG_ERR_CHECK(  _TranslateConflictData(pCtx, pConflicts, cConflicts)  );

fail:
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	pCtx.Log_PopOperation();
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::Commit(
	vvContext&           pCtx,
	const wxString&      sFolder,
	const wxString&      sComment,
	const wxArrayString* pPaths,
	const wxArrayString* pStamps,
	const wxString*      pUser,
	vvFlags              cFlags,
	wxString*            pHid
	)
{
	bool            bResult          = true;
	wxArrayString   cDiskPaths;
	SG_uint32       uConvertedPaths  = 0u;
	const char**    ppConvertedPaths = NULL;
	SG_stringarray* pConvertedStamps = NULL;
	const char*     szUser           = SG_AUDIT__WHO__FROM_SETTINGS;
	char*           szHid            = NULL;
	wxString        sHid;

	pCtx.Log_PushOperation("Committing", 0u);
	pCtx.Log_SetValue("Folder", sFolder, vvContext::LOG_FLAG_INPUT);
	pCtx.Log_SetValue("Comment", sComment, vvContext::LOG_FLAG_INPUT);
	pCtx.Log_SetValue("Flags", cFlags, vvContext::LOG_FLAG_INPUT | vvContext::LOG_FLAG_VERBOSE);
	if (pPaths != NULL)
	{
		if (!this->GetDiskPaths(pCtx, sFolder, *pPaths, cDiskPaths))
		{
			bResult = false;
			goto fail;
		}
		pCtx.Log_SetValue("Items", wxJoin(cDiskPaths, ','), vvContext::LOG_FLAG_INPUT);
	}
	if (pStamps != NULL)
	{
		pCtx.Log_SetValue("Stamps", wxJoin(*pStamps, ','), vvContext::LOG_FLAG_INPUT);
	}
	if (pUser != NULL)
	{
		pCtx.Log_SetValue("User", *pUser, vvContext::LOG_FLAG_INPUT);
	}

	if (pPaths != NULL)
	{
		SG_ERR_CHECK(  vvSgHelpers::Convert_wxArrayString_cStringArrayCount(pCtx, cDiskPaths, &ppConvertedPaths, &uConvertedPaths)  );
	}
	if (pStamps != NULL)
	{
		SG_ERR_CHECK(  vvSgHelpers::Convert_wxArrayString_sgStringArray(pCtx, *pStamps, &pConvertedStamps)  );
	}
	if (pUser != NULL)
	{
		szUser = pUser->c_str();
	}

	SG_ERR_CHECK(  SG_vv_verbs__commit(
		pCtx,
		sFolder.c_str(),
		sComment.c_str(),
		gpBlankFilespec,
		vvSgHelpers::Convert_cppBool_sgBool(!cFlags.HasAnyFlag(COMMIT_NO_RECURSE)),
		vvSgHelpers::Convert_cppBool_sgBool(cFlags.HasAnyFlag(COMMIT_NO_SUBMODULES)),
		szUser,
		SG_AUDIT__WHEN__NOW,
		uConvertedPaths,
		ppConvertedPaths,
		0u,
		NULL,
		pConvertedStamps,
		NULL, NULL,
		&szHid
	)  );

	sHid = vvSgHelpers::Convert_cString_wxString(szHid);
	pCtx.Log_SetValue("Hid", sHid, vvContext::LOG_FLAG_OUTPUT);
	if (pHid != NULL)
	{
		*pHid = sHid;
	}

fail:
	SG_NULLFREE(pCtx, ppConvertedPaths);
	SG_STRINGARRAY_NULLFREE(pCtx, pConvertedStamps);
	SG_NULLFREE(pCtx, szHid);
	pCtx.Log_PopOperation();
	return bResult && !pCtx.Error_Check();
}

bool vvVerbsImpl::Rename(
	vvContext&      pCtx,
	const wxString& sFullPath_objectToRename,
	const wxString& sNewName
	)
{
	SG_pathname*    pPath        = NULL;
	SG_pendingtree* pPendingTree = NULL;

	pCtx.Log_PushOperation("Renaming item", vvContext::LOG_FLAG_VERBOSE);
	pCtx.Log_SetValue("Item", sFullPath_objectToRename, vvContext::LOG_FLAG_INPUT);

	// allocate a pendingtree for the specified folder
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPath, sFullPath_objectToRename.c_str())  );
	SG_ERR_CHECK(  SG_pendingtree__alloc(pCtx, pPath, &pPendingTree)  );

	// rename the object
	SG_ERR_CHECK(  SG_pendingtree__rename(pCtx, pPendingTree, pPath, sNewName.c_str(), SG_FALSE)  );

fail:
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	pCtx.Log_PopOperation();
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::AddComment(
		class vvContext& pCtx, 
		const wxString&  sRepoName,    
		const wxString&  sChangesetHID,  
		const wxString&  sNewComment	
		)
{
	SG_repo * pRepo = NULL;
	SG_audit pAudit;
	pCtx;
	sRepoName;
	sChangesetHID;
	sNewComment;
	SG_ERR_CHECK(  SG_REPO__OPEN_REPO_INSTANCE(pCtx, sRepoName.c_str(), &pRepo)  );
	SG_ERR_CHECK(  SG_audit__init(pCtx, &pAudit, pRepo,
								  SG_AUDIT__WHEN__NOW,
								  SG_AUDIT__WHO__FROM_SETTINGS)  );
	SG_ERR_CHECK(  SG_vc_comments__add(pCtx, pRepo, sChangesetHID.c_str(), sNewComment.c_str(), &pAudit)  );

fail:
	SG_REPO_NULLFREE(pCtx, pRepo);
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::AddStamp(
		class vvContext& pCtx, 
		const wxString&  sRepoName,    
		const wxString&  sChangesetHID,  
		const wxString&  sNewStamp
		)
{
	SG_repo * pRepo = NULL;
	SG_audit pAudit;

	SG_ERR_CHECK(  SG_REPO__OPEN_REPO_INSTANCE(pCtx, sRepoName.c_str(), &pRepo)  );
	SG_ERR_CHECK(  SG_audit__init(pCtx, &pAudit, pRepo,
								  SG_AUDIT__WHEN__NOW,
								  SG_AUDIT__WHO__FROM_SETTINGS)  );
	SG_ERR_CHECK(  SG_vc_stamps__add(pCtx, pRepo, sChangesetHID.c_str(), sNewStamp.c_str(), &pAudit, NULL)  );

fail:
	SG_REPO_NULLFREE(pCtx, pRepo);
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::AddTag(
		class vvContext& pCtx, 
		const wxString&  sRepoName,    
		const wxString&  sChangesetHID,  
		const wxString&  sNewTag
		)
{
	SG_repo * pRepo = NULL;
	SG_audit pAudit;
	
	SG_ERR_CHECK(  SG_REPO__OPEN_REPO_INSTANCE(pCtx, sRepoName.c_str(), &pRepo)  );
	SG_ERR_CHECK(  SG_audit__init(pCtx, &pAudit, pRepo,
								  SG_AUDIT__WHEN__NOW,
								  SG_AUDIT__WHO__FROM_SETTINGS)  );
	SG_ERR_CHECK(  SG_vc_tags__add(pCtx, pRepo, sChangesetHID.c_str(), sNewTag.c_str(), &pAudit)  );

fail:
	SG_REPO_NULLFREE(pCtx, pRepo);
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::DeleteStamp(
		class vvContext& pCtx, 
		const wxString&  sRepoName,    
		const wxString&  sChangesetHID,  
		const wxString&  sRemoveMe	
		)
{
	SG_repo * pRepo = NULL;
	SG_audit pAudit;
	const char * szRemoveMe = sRemoveMe.c_str();
	SG_ERR_CHECK(  SG_REPO__OPEN_REPO_INSTANCE(pCtx, sRepoName.c_str(), &pRepo)  );
	SG_ERR_CHECK(  SG_audit__init(pCtx, &pAudit, pRepo,
								  SG_AUDIT__WHEN__NOW,
								  SG_AUDIT__WHO__FROM_SETTINGS)  );
	SG_ERR_CHECK(  SG_vc_stamps__remove(pCtx, pRepo, &pAudit, sChangesetHID.c_str(), 1, &szRemoveMe)  );

fail:
	SG_REPO_NULLFREE(pCtx, pRepo);
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::DeleteTag(
		class vvContext& pCtx, 
		const wxString&  sRepoName,    
		const wxString&  sRemoveMe	
		)
{
	SG_repo * pRepo = NULL;
	SG_audit pAudit;
	const char * szRemoveMe = sRemoveMe.c_str();
	SG_ERR_CHECK(  SG_REPO__OPEN_REPO_INSTANCE(pCtx, sRepoName.c_str(), &pRepo)  );
	SG_ERR_CHECK(  SG_audit__init(pCtx, &pAudit, pRepo,
								  SG_AUDIT__WHEN__NOW,
								  SG_AUDIT__WHO__FROM_SETTINGS)  );
	SG_ERR_CHECK(  SG_vc_tags__remove(pCtx, pRepo, &pAudit, 1, &szRemoveMe)  );

fail:
	SG_REPO_NULLFREE(pCtx, pRepo);
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::Move(
	vvContext&      pCtx,
	const wxArrayString& sFullPath_objectsToMove,
	const wxString& sNewParentPath
	)
{
	SG_pathname*    pPath        = NULL;
	SG_pendingtree* pPendingTree = NULL;
	sNewParentPath;

	pCtx.Log_PushOperation("Moving item", vvContext::LOG_FLAG_VERBOSE);
	pCtx.Log_SetValue("Item", sFullPath_objectsToMove[0], vvContext::LOG_FLAG_INPUT);

	// allocate a pendingtree for the specified folder
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPath, sFullPath_objectsToMove[0].c_str())  );
	SG_ERR_CHECK(  SG_pendingtree__alloc(pCtx, pPath, &pPendingTree)  );

	//Convert the wxArrayString to a char** array.
	SG_pathname ** pasz_pathsToMove = NULL;
	SG_pathname * pPathnameTemp = NULL;
	SG_alloc(pCtx, sFullPath_objectsToMove.GetCount(), sizeof(SG_pathname*), &pasz_pathsToMove);
	for (SG_uint32 i =0; i < sFullPath_objectsToMove.GetCount(); i++)
	{
		SG_ERR_CHECK( SG_PATHNAME__ALLOC__SZ(pCtx, &pPathnameTemp, sFullPath_objectsToMove[i].c_str())  );
		pasz_pathsToMove[i] = pPathnameTemp;
	}
	SG_pathname * pPathnameDest = NULL;
	SG_ERR_CHECK( SG_PATHNAME__ALLOC__SZ(pCtx, &pPathnameDest, sNewParentPath.c_str())  );
	// move the objects
	SG_ERR_CHECK(  SG_pendingtree__move(pCtx, pPendingTree, (const SG_pathname**)pasz_pathsToMove, sFullPath_objectsToMove.GetCount(), pPathnameDest)  );

fail:
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	pCtx.Log_PopOperation();
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::Add(
	vvContext&      pCtx,
	const wxArrayString& sFullPath_objectsToAdd
	)
{
	SG_pathname*    pPath        = NULL;
	SG_pendingtree* pPendingTree = NULL;
	SG_pathname* pPathCwd = NULL;
	SG_file_spec * pFileSpec = NULL;
	SG_repo *		pRepo = NULL;
	char ** pasz_pathsToAdd = NULL;
	SG_pathname*    pPathWorkingDirTop = NULL;

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC(pCtx, &pPathCwd)  );
	SG_ERR_CHECK(  SG_pathname__set__from_cwd(pCtx, pPathCwd)  );

	pCtx.Log_PushOperation("Adding item", vvContext::LOG_FLAG_VERBOSE);
	pCtx.Log_SetValue("Item", sFullPath_objectsToAdd[0], vvContext::LOG_FLAG_INPUT);

	// allocate a pendingtree for the specified folder
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPath, sFullPath_objectsToAdd[0].c_str())  );
	SG_ERR_CHECK(  SG_pendingtree__alloc(pCtx, pPath, &pPendingTree)  );
	SG_ERR_CHECK(  SG_pendingtree__get_repo(pCtx, pPendingTree, &pRepo)  );
	SG_ERR_CHECK(  SG_pendingtree__get_working_directory_top(pCtx, pPendingTree, &pPathWorkingDirTop)  );

	//Convert the wxArrayString to a char** array.
	SG_pathname * pPathnameTemp = NULL;
	SG_alloc(pCtx, sFullPath_objectsToAdd.GetCount(), sizeof(const char *), &pasz_pathsToAdd);
	for (SG_uint32 i =0; i < sFullPath_objectsToAdd.GetCount(); i++)
	{
		SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPathnameTemp, sFullPath_objectsToAdd[i].c_str())  );
		SG_ERR_CHECK(  SG_STRDUP(pCtx, SG_pathname__sz(pPathnameTemp), &(pasz_pathsToAdd[i]))  );
	}
	SG_ERR_CHECK(  SG_FILE_SPEC__ALLOC(pCtx, &pFileSpec)  );
	SG_ERR_CHECK(  SG_file_spec__load_ignores(pCtx, pFileSpec, pRepo, pPathWorkingDirTop, 0)  );
	// add the objects
	SG_ERR_CHECK(  SG_pendingtree__add(pCtx, pPendingTree, sFullPath_objectsToAdd.GetCount(), (const char **)pasz_pathsToAdd, SG_TRUE, pFileSpec, SG_FALSE)  );

fail:
	if (pasz_pathsToAdd)
	{
		for (SG_uint32 i =0; i < sFullPath_objectsToAdd.GetCount(); i++)
		{
			SG_NULLFREE(pCtx, pasz_pathsToAdd[i]);
		}
	}
	SG_NULLFREE(pCtx, pasz_pathsToAdd);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_PATHNAME_NULLFREE(pCtx, pPathWorkingDirTop);
	pCtx.Log_PopOperation();
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::Remove(
	vvContext&      pCtx,
	const wxArrayString& sFullPath_objectsToRemove
	)
{
	SG_pathname*    pPath        = NULL;
	SG_pendingtree* pPendingTree = NULL;
	SG_pathname* pPathCwd = NULL;
	SG_file_spec * pFileSpec = NULL;
	SG_repo *		pRepo = NULL;
	char ** pasz_pathsToRemove = NULL;
	SG_pathname*    pPathWorkingDirTop = NULL;

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC(pCtx, &pPathCwd)  );
	SG_ERR_CHECK(  SG_pathname__set__from_cwd(pCtx, pPathCwd)  );

	pCtx.Log_PushOperation("Removing item", vvContext::LOG_FLAG_VERBOSE);
	pCtx.Log_SetValue("Item", sFullPath_objectsToRemove[0], vvContext::LOG_FLAG_INPUT);

	// allocate a pendingtree for the specified folder
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPath, sFullPath_objectsToRemove[0].c_str())  );
	SG_ERR_CHECK(  SG_pendingtree__alloc(pCtx, pPath, &pPendingTree)  );
	SG_ERR_CHECK(  SG_pendingtree__get_repo(pCtx, pPendingTree, &pRepo)  );
	SG_ERR_CHECK(  SG_pendingtree__get_working_directory_top(pCtx, pPendingTree, &pPathWorkingDirTop)  );

	//Convert the wxArrayString to a char** array.
	SG_pathname * pPathnameTemp = NULL;
	SG_alloc(pCtx, sFullPath_objectsToRemove.GetCount(), sizeof(const char *), &pasz_pathsToRemove);
	for (SG_uint32 i =0; i < sFullPath_objectsToRemove.GetCount(); i++)
	{
		SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPathnameTemp, sFullPath_objectsToRemove[i].c_str())  );
		SG_ERR_CHECK(  SG_STRDUP(pCtx, SG_pathname__sz(pPathnameTemp), &(pasz_pathsToRemove[i]))  );
	}
	SG_ERR_CHECK(  SG_FILE_SPEC__ALLOC(pCtx, &pFileSpec)  );
	SG_ERR_CHECK(  SG_file_spec__load_ignores(pCtx, pFileSpec, pRepo, pPathWorkingDirTop, 0)  );
	// add the objects
	SG_ERR_CHECK(  SG_pendingtree__remove(pCtx, pPendingTree, sFullPath_objectsToRemove.GetCount(), (const char **)pasz_pathsToRemove, pFileSpec, SG_FALSE, SG_FALSE, SG_FALSE)  );

fail:
	if (pasz_pathsToRemove)
	{
		for (SG_uint32 i =0; i < sFullPath_objectsToRemove.GetCount(); i++)
		{
			SG_NULLFREE(pCtx, pasz_pathsToRemove[i]);
		}
	}
	SG_NULLFREE(pCtx, pasz_pathsToRemove);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_PATHNAME_NULLFREE(pCtx, pPathWorkingDirTop);
	pCtx.Log_PopOperation();
	return !pCtx.Error_Check();
}


bool vvVerbsImpl::Pull(
		class vvContext& pCtx,						
		const wxString&  sSrcLocalRepoNameOrRemoteURL,		
		const wxString&  sDestLocalRepoName,				
		const wxArrayString& saAreas,					
		const wxArrayString& saTags,					
		const wxArrayString& saDagnodes,
		const wxArrayString& saBranchNames
		)
{
	SG_vhash * pvhAreas = NULL;
	SG_rev_spec* pRevSpec = NULL;

	SG_ERR_CHECK(  vvSgHelpers::Convert_wxArrayString_sgVhash(pCtx, saAreas, &pvhAreas, true)  );
	SG_ERR_CHECK(  vvSgHelpers::Convert_wxArrayStrings_sgRevSpec(pCtx, saDagnodes, saTags, saBranchNames, &pRevSpec)  );	

	SG_ERR_CHECK(  SG_vv_verbs__pull(pCtx, sSrcLocalRepoNameOrRemoteURL.c_str(), sDestLocalRepoName.c_str(), pvhAreas, &pRevSpec, NULL)  );
fail:
	SG_VHASH_NULLFREE(pCtx, pvhAreas);
	SG_REV_SPEC_NULLFREE(pCtx, pRevSpec);
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::Push(
		class vvContext& pCtx,						
		const wxString&  sDestLocalRepoNameOrRemoteURL,	
		const wxString&  sSrcLocalRepoName,				
		const wxArrayString& saAreas,					
		const wxArrayString& saTags,					
		const wxArrayString& saDagnodes,
		const wxArrayString& saBranchNames
		)
{
	SG_vhash * pvhAreas = NULL;
	SG_rev_spec* pRevSpec = NULL;
	
	SG_ERR_CHECK(  vvSgHelpers::Convert_wxArrayString_sgVhash(pCtx, saAreas, &pvhAreas, true)  );
	SG_ERR_CHECK(  vvSgHelpers::Convert_wxArrayStrings_sgRevSpec(pCtx, saDagnodes, saTags, saBranchNames, &pRevSpec)  );

    // TODO the SG_FALSE in the line below is a force flag
	SG_ERR_CHECK(  SG_vv_verbs__push(pCtx, sDestLocalRepoNameOrRemoteURL.c_str(), sSrcLocalRepoName.c_str(), pvhAreas, pRevSpec, SG_FALSE, NULL)  );
	
fail:
	SG_VHASH_NULLFREE(pCtx, pvhAreas);
	SG_REV_SPEC_NULLFREE(pCtx, pRevSpec);
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::Incoming(
		class vvContext& pCtx,							//< [in] [out] Error and context info.
		const wxString&  sSrcLocalRepoNameOrRemoteURL,	//< [in] The local repo name, or remote repo URL.  Changes will be pulled from that repo.
		const wxString&  sDestLocalRepoName,			//< [in] The name of the local repo to pull changes into.
		HistoryData& cHistoryData
		)
{
	SG_history_result* pInfo = NULL;
	SG_ERR_CHECK(  SG_vv_verbs__incoming(pCtx, sSrcLocalRepoNameOrRemoteURL.c_str(), sDestLocalRepoName.c_str(), &pInfo, NULL)  );
	cHistoryData = vvSgHelpers::Convert_sg_history_result_HistoryData(pCtx, pInfo);
	
fail:
	//SG_HISTORY_RESULT_NULLFREE(pCtx, pInfo);
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::Outgoing(
		class vvContext& pCtx,						//< [in] [out] Error and context info.
		const wxString&  sDestLocalRepoNameOrRemoteURL,	//< [in] The local repo name, or remote repo URL.  Changes will be pulled from that repo.
		const wxString&  sSrcLocalRepoName,			//< [in] The name of the local repo to pull changes into.
		HistoryData& cHistoryData
		)
{
	SG_history_result* pInfo = NULL;
	SG_ERR_CHECK(  SG_vv_verbs__outgoing(pCtx, sDestLocalRepoNameOrRemoteURL.c_str(), sSrcLocalRepoName.c_str(), &pInfo, NULL)  );
	cHistoryData = vvSgHelpers::Convert_sg_history_result_HistoryData(pCtx, pInfo);
	
fail:
	//SG_HISTORY_RESULT_NULLFREE(pCtx, pInfo);
	return !pCtx.Error_Check();
}

#define _REVISION_SETTING "revision"
#define _COMMENT_SETTING "comment"
#define _BRANCH_SETTING "branch"
#define _STAMP_SETTING "stamp"
#define _TAG_SETTING "tag"
#define _DATE_SETTING "date"
#define _USER_SETTING "user"
#define _REVISION_WIDTH_SETTING "revision_width"
#define _COMMENT_WIDTH_SETTING "comment_width"
#define _BRANCH_WIDTH_SETTING "branch_width"
#define _STAMP_WIDTH_SETTING "stamp_width"
#define _TAG_WIDTH_SETTING "tag_width"
#define _DATE_WIDTH_SETTING "date_width"
#define _USER_WIDTH_SETTING "user_width"
#define _COLUMN_ORDER_SETTING "column_order"
bool vvVerbsImpl::GetColumnPreferences(
	vvContext&     pCtx,
	vvHistoryColumnPreferences& cColumnPrefs
	)
{
	SG_string * pString = NULL;
	SG_vhash * pvhSetting = NULL;
	SG_uint32 nCountBindings = 0;
	SG_uint32 i = 0;
	const SG_variant * pVariant = NULL;
	const char * pszKeyName = NULL;
	SG_bool bValue = SG_FALSE;

	//SG_context__alloc(&pCtx);
	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pString)  );
	SG_ERR_CHECK(  SG_string__sprintf(pCtx, pString, "%s/%s",
									  SG_LOCALSETTING__SCOPE__MACHINE,
									  SG_LOCALSETTING__TORTOISE_HISTORY_COLUMNS)  );
	

	SG_ERR_IGNORE(  SG_localsettings__get__vhash(pCtx,
												 SG_string__sz(pString),
												 NULL,		// null repo because we only want /machine/ settings.
												 &pvhSetting,
												 NULL)  );
	if (pvhSetting != NULL)
	{
		SG_ERR_CHECK(  SG_vhash__count(pCtx, pvhSetting, &nCountBindings)  );
		for (i = 0; i < nCountBindings; i++)
		{
			SG_int64 intValue = -1;
			SG_varray * pvarrayValue = NULL;
			SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvhSetting, i, &pszKeyName, &pVariant)  );
			if (pVariant->type == SG_VARIANT_TYPE_BOOL)
				SG_ERR_CHECK(  SG_variant__get__bool(pCtx, pVariant, &bValue)  );
			else if (pVariant->type == SG_VARIANT_TYPE_INT64)
				SG_ERR_CHECK(  SG_variant__get__int64(pCtx, pVariant, &intValue)  );
			else if (pVariant->type == SG_VARIANT_TYPE_VARRAY)
				SG_ERR_CHECK(  SG_variant__get__varray(pCtx, pVariant, &pvarrayValue)  );
			bool boolValue = (bValue == SG_TRUE);
			
			if (strcmp(pszKeyName, _REVISION_SETTING) == 0)
				cColumnPrefs.ShowRevisionColumn = boolValue;
			else if (strcmp(pszKeyName, _COMMENT_SETTING) == 0)
				cColumnPrefs.ShowCommentColumn = boolValue;
			else if (strcmp(pszKeyName, _BRANCH_SETTING) == 0)
				cColumnPrefs.ShowBranchColumn = boolValue;
			else if (strcmp(pszKeyName, _STAMP_SETTING) == 0)
				cColumnPrefs.ShowStampColumn = boolValue;
			else if (strcmp(pszKeyName, _TAG_SETTING) == 0)
				cColumnPrefs.ShowTagColumn = boolValue;
			else if (strcmp(pszKeyName, _DATE_SETTING) == 0)
				cColumnPrefs.ShowDateColumn = boolValue;
			else if (strcmp(pszKeyName, _USER_SETTING) == 0)
				cColumnPrefs.ShowUserColumn = boolValue;
			if (strcmp(pszKeyName, _REVISION_WIDTH_SETTING) == 0)
				cColumnPrefs.WidthRevisionColumn = intValue;
			else if (strcmp(pszKeyName, _COMMENT_WIDTH_SETTING) == 0)
				cColumnPrefs.WidthCommentColumn = intValue;
			else if (strcmp(pszKeyName, _BRANCH_WIDTH_SETTING) == 0)
				cColumnPrefs.WidthBranchColumn = intValue;
			else if (strcmp(pszKeyName, _STAMP_WIDTH_SETTING) == 0)
				cColumnPrefs.WidthStampColumn = intValue;
			else if (strcmp(pszKeyName, _TAG_WIDTH_SETTING) == 0)
				cColumnPrefs.WidthTagColumn = intValue;
			else if (strcmp(pszKeyName, _DATE_WIDTH_SETTING) == 0)
				cColumnPrefs.WidthDateColumn = intValue;
			else if (strcmp(pszKeyName, _USER_WIDTH_SETTING) == 0)
				cColumnPrefs.WidthUserColumn = intValue;
			else if (strcmp(pszKeyName, _COLUMN_ORDER_SETTING) == 0)
			{
				vvSgHelpers::Convert_sgVarray_wxArrayInt(pCtx, pvarrayValue, cColumnPrefs.columnOrder);
			}
		}
	}
fail:
	SG_STRING_NULLFREE(pCtx, pString);
	SG_VHASH_NULLFREE(pCtx, pvhSetting);
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::SaveColumnPreferences(
	vvContext&     pCtx,
	vvHistoryColumnPreferences& cColumnPrefs
	)
{
	SG_string * pString = NULL;
	SG_vhash * pvhSetting = NULL;
	SG_varray * pva_columnOrder = NULL;
	//SG_varray * pvaSetting = NULL;
	

	//SG_context__alloc(&pCtx);
	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pString)  );
	SG_ERR_CHECK(  SG_string__sprintf(pCtx, pString, "%s/%s",
									  SG_LOCALSETTING__SCOPE__MACHINE,
									  SG_LOCALSETTING__TORTOISE_HISTORY_COLUMNS)  );
	

	SG_ERR_IGNORE(  SG_localsettings__get__vhash(pCtx,
												 SG_string__sz(pString),
												 NULL,		// null repo because we only want /machine/ settings.
												 &pvhSetting,
												 NULL)  );
	if (pvhSetting == NULL)
	{
		SG_VHASH__ALLOC(pCtx, &pvhSetting);
	}

	SG_ERR_CHECK(  SG_vhash__update__bool(pCtx, pvhSetting, _REVISION_SETTING, cColumnPrefs.ShowRevisionColumn)  ); 
	SG_ERR_CHECK(  SG_vhash__update__bool(pCtx, pvhSetting, _COMMENT_SETTING, cColumnPrefs.ShowCommentColumn)  ); 
	SG_ERR_CHECK(  SG_vhash__update__bool(pCtx, pvhSetting, _BRANCH_SETTING, cColumnPrefs.ShowBranchColumn)  ); 
	SG_ERR_CHECK(  SG_vhash__update__bool(pCtx, pvhSetting, _TAG_SETTING, cColumnPrefs.ShowTagColumn)  ); 
	SG_ERR_CHECK(  SG_vhash__update__bool(pCtx, pvhSetting, _STAMP_SETTING, cColumnPrefs.ShowStampColumn)  ); 
	SG_ERR_CHECK(  SG_vhash__update__bool(pCtx, pvhSetting, _DATE_SETTING, cColumnPrefs.ShowDateColumn)  ); 
	SG_ERR_CHECK(  SG_vhash__update__bool(pCtx, pvhSetting, _USER_SETTING, cColumnPrefs.ShowUserColumn)  ); 

	SG_ERR_CHECK(  vvSgHelpers::Add_int__to_vhash_if_int_is_positive(pCtx, pvhSetting, _REVISION_WIDTH_SETTING, cColumnPrefs.WidthRevisionColumn)  ); 
	SG_ERR_CHECK(  vvSgHelpers::Add_int__to_vhash_if_int_is_positive(pCtx, pvhSetting, _COMMENT_WIDTH_SETTING, cColumnPrefs.WidthCommentColumn)  ); 
	SG_ERR_CHECK(  vvSgHelpers::Add_int__to_vhash_if_int_is_positive(pCtx, pvhSetting, _BRANCH_WIDTH_SETTING, cColumnPrefs.WidthBranchColumn)  ); 
	SG_ERR_CHECK(  vvSgHelpers::Add_int__to_vhash_if_int_is_positive(pCtx, pvhSetting, _TAG_WIDTH_SETTING, cColumnPrefs.WidthTagColumn)  ); 
	SG_ERR_CHECK(  vvSgHelpers::Add_int__to_vhash_if_int_is_positive(pCtx, pvhSetting, _STAMP_WIDTH_SETTING, cColumnPrefs.WidthStampColumn)  ); 
	SG_ERR_CHECK(  vvSgHelpers::Add_int__to_vhash_if_int_is_positive(pCtx, pvhSetting, _DATE_WIDTH_SETTING, cColumnPrefs.WidthDateColumn)  ); 
	SG_ERR_CHECK(  vvSgHelpers::Add_int__to_vhash_if_int_is_positive(pCtx, pvhSetting, _USER_WIDTH_SETTING, cColumnPrefs.WidthUserColumn)  ); 
	SG_ERR_CHECK(  vvSgHelpers::Convert_wxArrayInt_sgVarray(pCtx, cColumnPrefs.columnOrder, (void**)&pva_columnOrder, false)  );
	SG_ERR_CHECK(  SG_vhash__update__varray(pCtx, pvhSetting, _COLUMN_ORDER_SETTING, &pva_columnOrder)  );

	SG_ERR_CHECK(  SG_localsettings__update__vhash(pCtx, SG_string__sz(pString), (const SG_vhash *)pvhSetting)  );
fail:
	SG_STRING_NULLFREE(pCtx, pString);
	SG_VHASH_NULLFREE(pCtx, pvhSetting);
	return !pCtx.Error_Check();
}
		
bool vvVerbsImpl::GetHistoryFilterDefaults(
	vvContext&     pCtx,
	wxString& sRepoName,
	vvHistoryFilterOptions& cFilterOptions
	)
{
	SG_string * pString = NULL;
	SG_vhash * pvhSetting = NULL;
	SG_bool bValue = SG_FALSE;

	//SG_context__alloc(&pCtx);
	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pString)  );
	SG_ERR_CHECK(  SG_string__sprintf(pCtx, pString, "%s/%s/%s",
									  SG_LOCALSETTING__SCOPE__INSTANCE,
									  (const char *)sRepoName.c_str(),
									  SG_LOCALSETTING__TORTOISE_HISTORY_FILTER_DEFAULTS)  );
	

	SG_ERR_IGNORE(  SG_localsettings__get__vhash(pCtx,
												 SG_string__sz(pString),
												 NULL,		// null repo because we only want /machine/ settings.
												 &pvhSetting,
												 NULL)  );
	if (pvhSetting != NULL)
	{
		SG_bool bFound = SG_FALSE;
		const char * pszValue = NULL;
		SG_ERR_CHECK(  SG_vhash__has(pCtx, pvhSetting, "hideMerges", &bFound)  );
		if (bFound)
		{
			SG_ERR_CHECK(  SG_vhash__get__bool(pCtx, pvhSetting, "hideMerges", &bValue)  );
			cFilterOptions.bHideMerges = bValue == SG_TRUE;
		}
		SG_ERR_CHECK(  SG_vhash__has(pCtx, pvhSetting, "user", &bFound)  );
		if (bFound)
		{
			SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhSetting, "user", &pszValue)  );
			cFilterOptions.sUser = pszValue;
		}
		SG_ERR_CHECK(  SG_vhash__has(pCtx, pvhSetting, "branch", &bFound)  );
		if (bFound)
		{
			SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhSetting, "branch", &pszValue)  );
			cFilterOptions.sBranch = pszValue;
		}
		SG_ERR_CHECK(  SG_vhash__has(pCtx, pvhSetting, "tag", &bFound)  );
		if (bFound)
		{
			SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhSetting, "tag", &pszValue)  );
			cFilterOptions.sTag = pszValue;
		}
		SG_ERR_CHECK(  SG_vhash__has(pCtx, pvhSetting, "stamp", &bFound)  );
		if (bFound)
		{
			SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhSetting, "stamp", &pszValue)  );
			cFilterOptions.sStamp = pszValue;
		}
		SG_ERR_CHECK(  SG_vhash__has(pCtx, pvhSetting, "fileOrFolder", &bFound)  );
		if (bFound)
		{
			SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhSetting, "fileOrFolder", &pszValue)  );
			cFilterOptions.sFileOrFolderPath = pszValue;
		}
		SG_ERR_CHECK(  SG_vhash__has(pCtx, pvhSetting, "beginDate", &bFound)  );
		if (bFound)
		{
			SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhSetting, "beginDate", &pszValue)  );
			cFilterOptions.dBeginDate.ParseDate(pszValue);
		}
		SG_ERR_CHECK(  SG_vhash__has(pCtx, pvhSetting, "endDate", &bFound)  );
		if (bFound)
		{
			SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhSetting, "endDate", &pszValue)  );
			cFilterOptions.dEndDate.ParseDate(pszValue);
		}
	}
fail:
	SG_STRING_NULLFREE(pCtx, pString);
	SG_VHASH_NULLFREE(pCtx, pvhSetting);
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::History(
		class vvContext& pCtx,							//< [in] [out] Error and context info.
		const wxString&  sLocalRepoName,				//< [in] The name of the local repo to pull changes into.
		const vvHistoryFilterOptions filterOptions,		//< [in] History Options.
		HistoryData& cHistoryData,
		void ** ppHistoryToken_void
		)
{
	filterOptions;
	sLocalRepoName;
	cHistoryData;
	SG_uint32 nCountArgs = 0;
	const char * psz_arg = NULL;
	const char * psz_stamp = NULL;
	const char * psz_user = NULL;
	SG_history_result * pHistoryResults = NULL;
	SG_history_token * pHistoryToken = NULL;
	SG_rev_spec* pRevSpec = NULL;

	SG_ERR_CHECK(  SG_rev_spec__alloc(pCtx, &pRevSpec)  );

	if (filterOptions.sFileOrFolderPath != wxEmptyString)
	{
		nCountArgs = 1;
		psz_arg = filterOptions.sFileOrFolderPath.c_str();
	}
	if (filterOptions.sTag != wxEmptyString)
	{
		SG_ERR_CHECK(  SG_rev_spec__add_tag(pCtx, pRevSpec, filterOptions.sTag.c_str())  );
	}
	if (filterOptions.sBranch != wxEmptyString)
	{
		SG_ERR_CHECK(  SG_rev_spec__add_branch(pCtx, pRevSpec, filterOptions.sBranch.c_str())  );
	}
	if (filterOptions.sStamp != wxEmptyString)
	{
		psz_stamp = filterOptions.sStamp.c_str();
	}
	if (filterOptions.sUser != wxEmptyString)
	{
		psz_user = filterOptions.sUser.c_str();
	}

	SG_ERR_CHECK(  SG_vv_verbs__history__repo(pCtx, sLocalRepoName.c_str(), nCountArgs, &psz_arg, 
		pRevSpec, psz_user, psz_stamp, 
		filterOptions.nResultsAtOneTime, filterOptions.bHideMerges, 
		filterOptions.dBeginDate.IsValid() ? filterOptions.dBeginDate.GetTicks() * 1000 : 0, 
		filterOptions.dEndDate.IsValid() ? filterOptions.dEndDate.GetTicks() * 1000: SG_INT64_MAX,
		NULL, NULL, &pHistoryResults, &pHistoryToken)  );
	if (pHistoryToken != NULL && ppHistoryToken_void != NULL)
	{
		*ppHistoryToken_void = (void *)pHistoryToken;
		pHistoryToken = NULL;
	}
	cHistoryData = vvSgHelpers::Convert_sg_history_result_HistoryData(pCtx, pHistoryResults);
fail:
	SG_HISTORY_TOKEN_NULLFREE(pCtx, pHistoryToken);
	SG_REV_SPEC_NULLFREE(pCtx, pRevSpec);
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::History__Fetch_More(
		class vvContext& pCtx,							//< [in] [out] Error and context info.
		const wxString&  sLocalRepoName,				//< [in] The name of the local repo to pull changes into.
		unsigned int nNumResultsDesired,				//< [in] The number of additional results desired.
		void * pHistoryToken_void,
		HistoryData& cHistoryData,
		void ** ppNewHistoryToken_void
		)
{
	sLocalRepoName;
	cHistoryData;
	SG_history_result * pHistoryResults = NULL;
	SG_history_token * pHistoryToken = (SG_history_token *)pHistoryToken_void;
	SG_history_token ** ppNewHistoryToken = (SG_history_token **) ppNewHistoryToken_void;

	SG_ERR_CHECK(  SG_vv_verbs__history__fetch_more__repo(pCtx, sLocalRepoName.c_str(), pHistoryToken, nNumResultsDesired,
		&pHistoryResults, ppNewHistoryToken)  );

	cHistoryData = vvSgHelpers::Convert_sg_history_result_HistoryData(pCtx, pHistoryResults);
	//Free the old history token.
	SG_HISTORY_TOKEN_NULLFREE(pCtx, pHistoryToken);
	return !pCtx.Error_Check();
fail:
	SG_HISTORY_TOKEN_NULLFREE(pCtx, pHistoryToken);
	if (ppNewHistoryToken != NULL && *ppNewHistoryToken != NULL)
		SG_HISTORY_TOKEN_NULLFREE(pCtx, *ppNewHistoryToken);
	return !pCtx.Error_Check();
}


bool vvVerbsImpl::History__Free_Token(
		class vvContext& pCtx,							//< [in] [out] Error and context info.
		void * pHistoryToken_void)
{
	SG_history_token * pHistoryToken = (SG_history_token *)pHistoryToken_void;
	SG_HISTORY_TOKEN_NULLFREE(pCtx, pHistoryToken);
	return true;
}
bool sortRepoItems (vvVerbs::RepoItem i,vvVerbs::RepoItem j) 
{ 
	if (i.eItemType == j.eItemType)
	{
		return i.sItemName.CmpNoCase(j.sItemName) < 0;
	}
	else
	{
		return i.eItemType > j.eItemType;
	}
}
bool vvVerbsImpl::Repo_Item_Get_Children(
	class vvContext& pCtx,
	wxString sRepoName,
	wxString sChangesetHID,
	wxString sItemPath,
	vvVerbs::RepoItemList& cChildren)
{
	SG_repo * pRepo = NULL;
	SG_treenode * pnodeRoot = NULL;
	SG_treenode * pnodeFolder = NULL;
	char * pCS_root_HID = NULL;
	SG_treenode_entry * ptne = NULL;
	const SG_treenode_entry * pTreenodeEntrySub = NULL;
	const char * pFolderHID = NULL;
	SG_treenode_entry_type tneType;
	SG_treenode_entry_type tneTypeSub;
	SG_uint32 kLimit, k;
	const char * szEntryname = NULL;

	cChildren.clear();

	SG_ERR_CHECK(  SG_REPO__OPEN_REPO_INSTANCE(pCtx, sRepoName.c_str(), &pRepo)  );
	SG_ERR_CHECK(  SG_changeset__tree__find_root_hid(pCtx, pRepo, sChangesetHID.c_str(), &pCS_root_HID)  );
	SG_ERR_CHECK(  SG_treenode__load_from_repo(pCtx, pRepo, pCS_root_HID, &pnodeRoot)  );
	SG_ERR_CHECK(  SG_treenode__find_treenodeentry_by_path(pCtx, pRepo, pnodeRoot, sItemPath.c_str(), NULL, &ptne)  );

	SG_ERR_CHECK(  SG_treenode_entry__get_hid_blob(pCtx, ptne, &pFolderHID)  );
	SG_ERR_CHECK(  SG_treenode_entry__get_entry_type(pCtx, ptne, &tneType)  );
	if (tneType == SG_TREENODEENTRY_TYPE_DIRECTORY)
	{
		//We found the one we were looking for, now load it and loop over its children.
		SG_ERR_CHECK(  SG_treenode__load_from_repo(pCtx, pRepo, pFolderHID, &pnodeFolder)  );
		SG_ERR_CHECK(  SG_treenode__count(pCtx,pnodeFolder,&kLimit)  );
		for (k=0; k < kLimit; k++)
		{
			SG_treenode__get_nth_treenode_entry__ref(pCtx,pnodeFolder,k,NULL,&pTreenodeEntrySub);
			vvVerbs::RepoItem ri;
			SG_ERR_CHECK(  SG_treenode_entry__get_entry_type(pCtx,pTreenodeEntrySub,&tneTypeSub)  );
			ri.eItemType = (vvVerbs::RepoItemType)tneTypeSub;
			
			//SG_ERR_CHECK(  SG_treenode_entry__get_hid_blob(pCtx,pTreenodeEntrySub,&szHidBlob)  );
			//SG_ERR_CHECK(  SG_treenode_entry__get_hid_xattrs(pCtx,pTreenodeEntrySub,&szHidXAttrs)  );
			//SG_ERR_CHECK(  SG_treenode_entry__get_attribute_bits(pCtx,pTreenodeEntrySub,&attrs)  );
			SG_ERR_CHECK(  SG_treenode_entry__get_entry_name(pCtx,pTreenodeEntrySub,&szEntryname)  );
			ri.sItemName = szEntryname;
			cChildren.push_back(ri);
		}
	}
	sort (cChildren.begin(), cChildren.end(), sortRepoItems);
fail:
	SG_TREENODE_NULLFREE(pCtx, pnodeRoot);
	SG_TREENODE_NULLFREE(pCtx, pnodeFolder);

	SG_NULLFREE(pCtx, pCS_root_HID);
	SG_REPO_NULLFREE(pCtx, pRepo);
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::Get_Revision_Details(
	class vvContext& pCtx,							//< [in] [out] Error and context info.
	const wxString&  sLocalRepoName,				//< [in] The name of the local repo to pull changes into.
	const wxString& sRevID,							//< [in] The revision to load
	vvVerbs::HistoryEntry& cHistoryEntry,						//< [out] The details.
	vvVerbs::RevisionList& children						//< [out] A vector of the children of the revision.
)
{
	pCtx;
	sLocalRepoName;
	sRevID;
	cHistoryEntry;
	children;
	SG_repo * pRepo = NULL;
	vvVerbs::HistoryData cHistoryData;

	SG_bool bHasResult = SG_FALSE;
	SG_history_result * pHistoryResults = NULL;
	SG_varray * pvaChildren = NULL;
	SG_rev_spec * prev_spec;

	SG_ERR_CHECK(  SG_REPO__OPEN_REPO_INSTANCE(pCtx, sLocalRepoName.c_str(), &pRepo)  );
	SG_ERR_CHECK(  SG_REV_SPEC__ALLOC(pCtx, &prev_spec)  );
	SG_ERR_CHECK(  SG_rev_spec__add_rev(pCtx, prev_spec, sRevID.c_str())  );
	SG_ERR_CHECK(  SG_vv_verbs__history__repo2(pCtx, pRepo, 0, NULL, prev_spec, NULL, NULL, 1, SG_FALSE, 0, SG_INT64_MAX,  &bHasResult, &pHistoryResults, NULL) );
	//SG_ERR_CHECK(  SG_history__get_changeset_description(pCtx, pRepo, pszHid, SG_TRUE, &pvhChangesetDescription)  );
	
	cHistoryData = vvSgHelpers::Convert_sg_history_result_HistoryData(pCtx, pHistoryResults);
	
	if (cHistoryData.size() > 0)
	{
		SG_ERR_CHECK(  SG_history__get_children_description(pCtx, pRepo, cHistoryData[0].sChangesetHID, &pvaChildren)  );
		SG_uint32 nChildrenLen = 0;
		SG_ERR_CHECK(  SG_varray__count(pCtx, pvaChildren, &nChildrenLen)  );
		
		SG_vhash * pvh_currentChild = NULL;
		SG_int64 nRevNo = 0;
		const char * pszChildHid = NULL;
		
		for (SG_uint32 i = 0; i < nChildrenLen; i++)
		{
			vvVerbs::RevisionEntry childRev;
			SG_ERR_CHECK(  SG_varray__get__vhash(pCtx, pvaChildren, i, &pvh_currentChild)  );
			SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvh_currentChild, "revno", &nRevNo)  );
			SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh_currentChild, "changeset_id", &pszChildHid)  );
			childRev.nRevno = nRevNo;
			childRev.sChangesetHID = pszChildHid;
			children.push_back(childRev);

		}
	}
	
	if (cHistoryData.size() > 0)
		cHistoryEntry = cHistoryData[0];

fail:
	SG_REPO_NULLFREE(pCtx, pRepo);
	return !pCtx.Error_Check();
}

bool vvVerbsImpl::Export_Revision(
	class vvContext& pCtx,							//< [in] [out] Error and context info.
	const wxString&  sLocalRepoName,				//< [in] The name of the local repo to pull changes into.
	const wxString& sRevID,							//< [in] The revision to load
	wxString sLocalOutputPath						//< [in] The local path to output to.
)
{

	SG_ERR_CHECK(  SG_vv_verbs__export(pCtx,
		sLocalRepoName.c_str(), sRevID.c_str(), sLocalOutputPath.c_str(),
									   SG_FALSE,
									   NULL,
									   NULL)  );

fail:
	return !pCtx.Error_Check();
}