/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/
#include "stdafx.h"
#include "vvWinSync.h"

#define HEARTBEAT_EVERY_N_SECONDS 60

static void ReportError(SG_context*, HWND);
static void LogHearbeatFailure(SG_context*);
static void GetConfiguration(SG_context*, HWND, const char*, char**, char**, SG_int64*, BOOL*);
static void DoHeartbeat(SG_context*, HWND, const char*, const char*, BOOL*, BOOL*);
static void DoSync(SG_context*, HWND, const char*, const char*, BOOL*);

void SetTooltipWithLastSyncMessage(SG_context* pCtx, DWORD lSinceLastSync, const char* szSrc, const char* prefix) 
{
	SG_string* pstrTemp = NULL;
	LPWSTR pwszTemp = NULL;

	SG_STRING__ALLOC__RESERVE(pCtx, &pstrTemp, MAX_TOOLTIP);
	
	if (prefix)
		SG_ERR_CHECK(  SG_string__append__sz(pCtx, pstrTemp, prefix)  );

	if (lSinceLastSync < 60000)
	{
		SG_ERR_CHECK(  SG_string__append__format(pCtx, pstrTemp, "Synced less than a minute ago with %s", szSrc)  );
	}
	else
	{
		SG_uint32 niceMinutes = lSinceLastSync / 60000;
		if (niceMinutes == 1)
			SG_ERR_CHECK(  SG_string__append__format(pCtx, pstrTemp, "Synced 1 minute ago with %s", szSrc)  );
		else
			SG_ERR_CHECK(  SG_string__append__format(pCtx, pstrTemp, "Synced %u minutes ago with %s", niceMinutes, szSrc)  );
	}

	SG_ERR_CHECK(  SG_utf8__extern_to_os_buffer__wchar(pCtx, SG_string__sz(pstrTemp), &pwszTemp, NULL)  );
	SetToolTip(pwszTemp);

	/* common cleanup */
fail:
	SG_NULLFREE(pCtx, pwszTemp);
	SG_STRING_NULLFREE(pCtx, pstrTemp);
}


void WorkerThread(void* pVoid)
{
	volatile WorkerThreadParms* pParms = (WorkerThreadParms*)pVoid;
	SG_context* pCtx = NULL;

	char* pszDest = NULL;
	char* pszSrc = NULL;

	SG_context__alloc(&pCtx);
	if (!pCtx)
	{
		PostQuitMessage(-2);
		return;
	}

	BOOL bHeartbeatSuccess = FALSE;
	BOOL bSyncSucceededOrDidNotRun = FALSE;
	char* pszLastSyncSrc = NULL;
	
	while (TRUE)
	{
		BOOL bGetConfigSuccess = FALSE;
		BOOL bHeartbeatOwnsTooltip = FALSE;

		DWORD lNow = GetTickCount();
		DWORD lSinceLastSync = (lNow - pParms->lLastSync);
		DWORD lSinceLastHeartbeat = (lNow - pParms->lLastHearbeat);

		SG_int64 lSyncIntervalMinutes;

		SG_ERR_CHECK(  GetConfiguration(pCtx, pParms->hWnd, pParms->szProfile, &pszDest, &pszSrc, &lSyncIntervalMinutes, &bGetConfigSuccess)  );
		if (!bGetConfigSuccess)
			goto wait;

		BOOL bTimeForHeartbeat = lSinceLastHeartbeat >= (HEARTBEAT_EVERY_N_SECONDS * 1000);
		BOOL bTimeForSync = (pParms->lLastSync == 0) || (lSinceLastSync >= (lSyncIntervalMinutes * 60000));

		/* Do heartbeat if it's time for heartbeat or for sync. */
		if ( bTimeForHeartbeat || bTimeForSync )
		{
			SG_ERR_CHECK(  DoHeartbeat(pCtx, pParms->hWnd, pszSrc, pszDest, &bHeartbeatSuccess, &bHeartbeatOwnsTooltip)  );

			/* Update "last heartbeat" timestamps whether it succeeded or not. */
			pParms->lLastHearbeat = lNow;
			lSinceLastHeartbeat = 0;
		}

		/* Do sync if it's time and we're online */
		if ( bTimeForSync && bHeartbeatSuccess )
		{
			SG_ERR_CHECK(  DoSync(pCtx, pParms->hWnd, pszSrc, pszDest, &bSyncSucceededOrDidNotRun)  );
			if (bSyncSucceededOrDidNotRun)
			{
				/* Update "last sync" timestamps if it was successful. */
				pParms->lLastSync = lNow;
				lSinceLastSync = 0;

				/* If the source repo gets changed somewhere other than our
				   config dialog, pszSrc has a repo we haven't synced with, and 
				   the "Last synced..." tooltip would be wrong. So we keep a copy
				   of the source repo that we actually synced with. */
				SG_NULLFREE(pCtx, pszLastSyncSrc);
				SG_ERR_CHECK(  SG_STRDUP(pCtx, pszSrc, &pszLastSyncSrc)  );
			}
		}
		else
		{
			bSyncSucceededOrDidNotRun = TRUE; // Didn't run because it wasn't time or heartbeat failed.
		}

		/* If the sync failed, or the heartbeat said so, it set the tooltip with an error message. Don't overwrite it. */
		if (bSyncSucceededOrDidNotRun && !bHeartbeatOwnsTooltip)
		{
			/* The tooltip shows time since the last sync, so we intentionally update it 
			   in every loop iteration where we have a known last sync time. */
			if (pParms->lLastSync)
			{
				if (bHeartbeatSuccess)
				{
					SG_ERR_CHECK(  SetTooltipWithLastSyncMessage(pCtx, lSinceLastSync, pszLastSyncSrc, NULL)  );
					if ( SendMessage(pParms->hWnd, WM_APP_READY, 0, 0) )
						SG_ERR_THROW(SG_ERR_GETLASTERROR(GetLastError()));
				}
				else
				{
					SG_ERR_CHECK(  SetTooltipWithLastSyncMessage(pCtx, lSinceLastSync, pszLastSyncSrc, "Offline\n")  );
					if ( SendMessage(pParms->hWnd, WM_APP_OFFLINE, 0, 0) )
						SG_ERR_THROW(SG_ERR_GETLASTERROR(GetLastError()));
				}
			}
			else
			{
				/* We don't know when we last synced, but we should set status/tooltip info anyway. */
				if (bHeartbeatSuccess)
				{
					SetToolTip(L"Online\nFirst sync pending.");
					if ( SendMessage(pParms->hWnd, WM_APP_READY, 0, 0) )
						SG_ERR_THROW(SG_ERR_GETLASTERROR(GetLastError()));
				}
				else
				{
					SetToolTip(L"Offline");
					if ( SendMessage(pParms->hWnd, WM_APP_OFFLINE, 0, 0) )
						SG_ERR_THROW(SG_ERR_GETLASTERROR(GetLastError()));
				}
			}
		}


		SG_NULLFREE(pCtx, pszDest);
		SG_NULLFREE(pCtx, pszSrc);

wait:
		DWORD waitResult = WaitForSingleObject(pParms->hEvent, INFINITE);
		if (waitResult == WAIT_FAILED)
			SG_ERR_THROW(SG_ERR_GETLASTERROR(GetLastError()));

		if (pParms->bShutdown)
			break;
	}

	SG_ERR_IGNORE(  SG_log__report_message(pCtx, SG_LOG__MESSAGE__VERBOSE, "Worker thread shut down cleanly.")  );

	/* This will quit the application (it shouldn't be in the fail block). */
	if ( SendMessage(pParms->hWnd, WM_APP_WORKER_DONE, 0, 0) )
		SG_ERR_THROW_RETURN(SG_ERR_GETLASTERROR(GetLastError()));

fail:
	if (SG_CONTEXT__HAS_ERR(pCtx) && !pParms->bShutdown)
		ReportError(pCtx, pParms->hWnd);

	SG_NULLFREE(pCtx, pszDest);
	SG_NULLFREE(pCtx, pszSrc);
	SG_NULLFREE(pCtx, pszLastSyncSrc);
	
	SG_CONTEXT_NULLFREE(pCtx);

	/* If we did terminate abnormally by jumping to fail, let the app
	know we're not running and it can shut down whenever it's ready. */
	pParms->bShutdown = TRUE;
}

static void DoHeartbeat(
	SG_context* pCtx, 
	HWND hWnd,
	const char* pszSrc, 
	const char* pszDest,
	BOOL* pbSuccess,
	BOOL* pbLeaveTooltip) 
{
	SG_sync_client* pSyncClient = NULL;
	SG_vhash* pvhSrcRepoInfo = NULL;
	LPWSTR pwszErr = NULL;
	SG_repo* pRepoDest = NULL;
	char* pszDestRepoId = NULL;

	*pbSuccess = FALSE;
	*pbLeaveTooltip = FALSE;

	SG_ERR_CHECK(  SG_log__report_verbose(pCtx, "Checking heartbeat of %s.", pszSrc)  );
	SG_ERR_CHECK(  SG_sync_client__open(pCtx, pszSrc, &pSyncClient)  );
	SG_sync_client__heartbeat(pCtx, pSyncClient, &pvhSrcRepoInfo);
	if (SG_CONTEXT__HAS_ERR(pCtx))
	{
		LogHearbeatFailure(pCtx);
		SG_ERR_DISCARD;
		SG_ERR_CHECK(  SG_log__report_verbose(pCtx, "Hearbeat: offline.")  );
	}
	else
	{
		/* Verify that repo IDs match */
		const char* pszSrcRepoIdRef = NULL;
		SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhSrcRepoInfo, "RepoID", &pszSrcRepoIdRef)  );
		SG_REPO__OPEN_REPO_INSTANCE(pCtx, pszDest, &pRepoDest);
		if (SG_CONTEXT__HAS_ERR(pCtx))
		{
			if (SG_context__err_equals(pCtx, SG_ERR_SG_LIBRARY(SG_ERR_NOTAREPOSITORY)))
			{
				SG_ERR_DISCARD;

				if ( FAILED(SetToolTip(L"Veracity Sync Offline: The local destination repository does not exist.", pwszErr)) )
					SG_ERR_THROW(SG_ERR_GETLASTERROR(GetLastError()));

				SendMessage(hWnd, WM_APP_CONFIG_ERR, 0, 0);

				/* More icky string copies for no good reason that could be cleaned up. */
				static const LPWSTR ERR_NO_REPO = L"The local destination repository does not exist.";
				size_t len_in_chars = 0;
				if (FAILED(  StringCchLength(ERR_NO_REPO, STRSAFE_MAX_CCH, &len_in_chars)  ))
					SG_ERR_THROW(SG_ERR_GETLASTERROR(GetLastError()));
				pwszErr = (LPWSTR)malloc(sizeof(wchar_t) * (len_in_chars + 1));
				if (FAILED(  StringCchCopy(pwszErr, len_in_chars + 1, ERR_NO_REPO)  ))
					SG_ERR_THROW(SG_ERR_GETLASTERROR(GetLastError()));

				SendMessage(hWnd, WM_APP_SHOW_BALLOON, (WPARAM)L"Veracity Sync Offline", (LPARAM)pwszErr);
				pwszErr = NULL;
				*pbLeaveTooltip = TRUE;
			}
			else
				SG_ERR_RETHROW;
		}
		else
		{
			/* Opened local destination repo successfully. */
			SG_ERR_CHECK(  SG_repo__get_repo_id(pCtx, pRepoDest, &pszDestRepoId)  );
			int repoIdsMatch = strcmp(pszSrcRepoIdRef, pszDestRepoId);

			if (repoIdsMatch != 0)
			{
				if ( FAILED(SetToolTip(L"Veracity Sync Offline: The repositories must be clones.", pwszErr)) )
					SG_ERR_THROW(SG_ERR_GETLASTERROR(GetLastError()));

				SendMessage(hWnd, WM_APP_CONFIG_ERR, 0, 0);

				/* More icky string copies for no good reason that could be cleaned up. */
				static const LPWSTR ERR_NOT_CLONES = L"The repositories must be clones.";
				size_t len_in_chars = 0;
				if (FAILED(  StringCchLength(ERR_NOT_CLONES, STRSAFE_MAX_CCH, &len_in_chars)  ))
					SG_ERR_THROW(SG_ERR_GETLASTERROR(GetLastError()));
				pwszErr = (LPWSTR)malloc(sizeof(wchar_t) * (len_in_chars + 1));
				if (FAILED(  StringCchCopy(pwszErr, len_in_chars + 1, ERR_NOT_CLONES)  ))
					SG_ERR_THROW(SG_ERR_GETLASTERROR(GetLastError()));

				SendMessage(hWnd, WM_APP_SHOW_BALLOON, (WPARAM)L"Veracity Sync Offline", (LPARAM)pwszErr);
				pwszErr = NULL;
				*pbLeaveTooltip = TRUE;
			}
			else
			{
				*pbSuccess = TRUE;
				SG_ERR_CHECK(  SG_log__report_verbose(pCtx, "Hearbeat: online.")  );
			}
		}
	}

	/* common cleanup */
fail:
	if (pwszErr)
		free(pwszErr);
	SG_ERR_IGNORE(  SG_sync_client__close_free(pCtx, pSyncClient)  );
	SG_VHASH_NULLFREE(pCtx, pvhSrcRepoInfo);
	SG_REPO_NULLFREE(pCtx, pRepoDest);
	SG_NULLFREE(pCtx, pszDestRepoId);
}

static void DoSync(
	SG_context* pCtx, 
	HWND hWnd, 
	const char* pszSrc, 
	const char* pszDest,
	BOOL* pbSuccess) 
{
	*pbSuccess = FALSE;

	if ( FAILED(SetToolTip(L"Syncing now...")) )
		SG_ERR_THROW_RETURN(SG_ERR_GETLASTERROR(GetLastError()));
	if ( SendMessage(hWnd, WM_APP_BUSY, 0, 0) )
		SG_ERR_THROW_RETURN(SG_ERR_GETLASTERROR(GetLastError()));

	/* Pull everything */
	SG_ERR_CHECK_RETURN(  SG_log__report_verbose(pCtx, "Syncing from %s.", pszSrc)  );
	SG_pull__all(pCtx, pszDest, pszSrc, NULL);
	if (SG_CONTEXT__HAS_ERR(pCtx))
	{
		ReportError(pCtx, hWnd);
		SG_ERR_DISCARD;
	}
	else
	{
		*pbSuccess = TRUE;
	}
}

/**
 * Does not propagate errors via pCtx, despite what you might guess, looking at the function signature.
 *
 * The "heartbeat" uses this. It expects to fail, so logs failures at the verbose level.
 */
static void LogHearbeatFailure(SG_context* pCtx)
{
	SG_string* pstrMessage = NULL;

	SG_error err = SG_context__err_to_string(pCtx, SG_TRUE, &pstrMessage);
	if (SG_IS_OK(err))
	{
		SG_ERR_IGNORE(  SG_string__insert__sz(pCtx, pstrMessage, 0, "Heartbeat failure: ")  );
		SG_ERR_IGNORE(  SG_log__report_verbose(pCtx, SG_string__sz(pstrMessage))  );
	}

	SG_STRING_NULLFREE(pCtx, pstrMessage);
}

static void GetConfiguration(
	SG_context* pCtx, 
	HWND hWnd, 
	const char* szProfile,
	char** ppszDest, 
	char** ppszSrc, 
	SG_int64* plSyncIntervalMinutes,
	BOOL* pbSuccess) 
{
	char* pszDest = NULL;
	char* pszSrc = NULL;
	SG_int64 lSyncIntervalMinutes = 0;
	LPWSTR pwszErr = NULL;

	*pbSuccess = FALSE;

	// We deliberately look these up every time through the loop so we pick up changes while we're running.
	SG_ERR_CHECK(  GetProfileSettings(pCtx, szProfile, &pszDest, &pszSrc, &lSyncIntervalMinutes, &pwszErr)  );
	if (pwszErr)
	{
		if ( FAILED(SetToolTip(L"Veracity Sync Offline: %s", pwszErr)) )
			SG_ERR_THROW(SG_ERR_GETLASTERROR(GetLastError()));

		SendMessage(hWnd, WM_APP_CONFIG_ERR, 0, 0);
		/* This is kind of a hack: WndProc knows more about this than it should. 
		It always frees lParam and never frees wParam, for example. 
		If WM_APP_SHOW_BALLOON gets used again, this should be better generalized. */
		SendMessage(hWnd, WM_APP_SHOW_BALLOON, (WPARAM)L"Veracity Sync Offline", (LPARAM)pwszErr);
		pwszErr = NULL;
	}
	else
	{
		SG_RETURN_AND_NULL(pszDest, ppszDest);
		SG_RETURN_AND_NULL(pszSrc, ppszSrc);
		if (plSyncIntervalMinutes)
			*plSyncIntervalMinutes = lSyncIntervalMinutes;
		*pbSuccess = TRUE;
	}

	/* common cleanup */
fail:
	if (pwszErr)
		free(pwszErr);
	SG_NULLFREE(pCtx, pszDest);
	SG_NULLFREE(pCtx, pszSrc);
}

static void ReportError(SG_context* pCtx, HWND hWnd) 
{
	SG_string* pstrErr = NULL;
	LPTSTR szErr = NULL;
	SG_log__report_error__current_error(pCtx);

	/* Show the error message in a balloon notification. */
	SG_error err = SG_context__err_to_string(pCtx, SG_FALSE, &pstrErr);
	if (SG_IS_OK(err))
	{
		SG_ERR_IGNORE(  SG_utf8__extern_to_os_buffer__wchar(pCtx, SG_string__sz(pstrErr), &szErr, NULL)  );
		SetToolTip(szErr);
		/* This is kind of a hack: WndProc knows more about this than it should. 
		It always frees lParam and never frees wParam, for example. 
		If WM_APP_SHOW_BALLOON gets used again, this should be better generalized. */
		SendMessage(hWnd, WM_APP_SHOW_BALLOON, (WPARAM)L"Veracity Sync Error", (LPARAM)szErr);
	}

	SendMessage(hWnd, WM_APP_OFFLINE, 0, 0);

	SG_STRING_NULLFREE(pCtx, pstrErr);
}
