/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include <sg.h>
#include "unittests.h"


/*
**
** get_class
**
*/

/**
 * A single test case for detecting the class of a file.
 */
typedef struct
{
	const char* szFilename;
	const char* szClass;
}
u0106__get_class__case;

/**
 * List of test files and expected results.
 */
static const u0106__get_class__case u0106__get_class__cases[] =
{
	// mostly nameless/extensionless files
	// no classes should exist that match these files by their filename
	// so it will have to fall back on the binary file detection algorithm
	{ "bmp",            SG_FILETOOL__CLASS__BINARY },
	{ "doc",            SG_FILETOOL__CLASS__BINARY },
	{ "docx",           SG_FILETOOL__CLASS__BINARY },
	{ "gif",            SG_FILETOOL__CLASS__BINARY },
	{ "jpg",            SG_FILETOOL__CLASS__BINARY },
	{ "odp",            SG_FILETOOL__CLASS__BINARY },
	{ "ods",            SG_FILETOOL__CLASS__BINARY },
	{ "odt",            SG_FILETOOL__CLASS__BINARY },
	{ "png",            SG_FILETOOL__CLASS__BINARY },
	{ "ppt",            SG_FILETOOL__CLASS__BINARY },
	{ "pptx",           SG_FILETOOL__CLASS__BINARY },
	{ "tga",            SG_FILETOOL__CLASS__BINARY },
	{ "tif",            SG_FILETOOL__CLASS__BINARY },
	{ "xls",            SG_FILETOOL__CLASS__BINARY },
	{ "xlsx",           SG_FILETOOL__CLASS__BINARY },
	{ "txt",            SG_FILETOOL__CLASS__TEXT   },
	{ "utf8",           SG_FILETOOL__CLASS__TEXT   },
	{ "utf8_nul",       SG_FILETOOL__CLASS__TEXT   },
	{ "utf8_nobom",     SG_FILETOOL__CLASS__TEXT   },
	{ "utf8_nobom_nul", SG_FILETOOL__CLASS__BINARY }, // this one will fool the binary detection
	{ "utf16be",        SG_FILETOOL__CLASS__TEXT   },
	{ "utf16be_nul",    SG_FILETOOL__CLASS__TEXT   },
	{ "utf16le",        SG_FILETOOL__CLASS__TEXT   },
	{ "utf16le_nul",    SG_FILETOOL__CLASS__TEXT   },
	{ "utf32be",        SG_FILETOOL__CLASS__TEXT   },
	{ "utf32be_nul",    SG_FILETOOL__CLASS__TEXT   },
	{ "utf32le",        SG_FILETOOL__CLASS__TEXT   },
	{ "utf32le_nul",    SG_FILETOOL__CLASS__TEXT   },
};

static void u0106__get_class__run_test_cases(SG_context* pCtx, const SG_pathname* pDataDir)
{
	SG_pathname* pDataPath = NULL;
	SG_bool      bExists   = SG_FALSE;
	SG_uint32    uIndex    = 0u;
	SG_pathname* pPathname = NULL;
	SG_string*   pClass    = NULL;

	// find the data path with all of our files
	VERIFY_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx, &pDataPath, pDataDir, "u0106_filetool_data")  );
	VERIFY_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pDataPath, &bExists, NULL, NULL)  );
	VERIFYP_COND("Data folder doesn't exist.", bExists == SG_TRUE, ("Expected data path: %s", SG_pathname__sz(pDataPath)));

	// iterate through all of our test cases
	for (uIndex = 0u; uIndex < SG_NrElements(u0106__get_class__cases); ++uIndex)
	{
		const u0106__get_class__case* pCase = u0106__get_class__cases + uIndex;

		// if you need to debug through a specific case
		// then update this if statement to check for the index you care about and stick a breakpoint inside
		// failure messages include the index of the failing case for easy reference
		if (uIndex == SG_UINT32_MAX)
		{
			pClass = NULL;
		}

		// build the path to the current file
		VERIFY_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx, &pPathname, pDataPath, pCase->szFilename)  );
		VERIFY_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pPathname, &bExists, NULL, NULL)  );
		VERIFYP_COND("Test file doesn't exist.", bExists == SG_TRUE, ("Expected test file: %s", SG_pathname__sz(pPathname)));

		// run the test
		if (bExists == SG_TRUE)
		{
			// get the class for this file
			VERIFY_ERR_CHECK(  SG_filetool__get_class(pCtx, NULL, pPathname, NULL, &pClass)  );

			// make sure we got the expected result
			VERIFYP_COND("Filename had NULL class.", pClass != NULL, ("index(%d) file(%s) class(NULL), expected(%s)", uIndex, pCase->szFilename, pCase->szClass));
			if (pClass != NULL)
			{
				VERIFYP_COND("Filename had unexpected class.", strcmp(SG_string__sz(pClass), pCase->szClass) == 0, ("index(%d) file(%s) class(%s) expected(%s)", uIndex, pCase->szFilename, SG_string__sz(pClass), pCase->szClass));
			}
		}

		// free the temporary data
		SG_PATHNAME_NULLFREE(pCtx, pPathname);
		SG_STRING_NULLFREE(pCtx, pClass);
	}

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathname);
	SG_STRING_NULLFREE(pCtx, pClass);
	SG_PATHNAME_NULLFREE(pCtx, pDataPath);
	return;
}

/*
**
** MAIN
**
*/

TEST_MAIN(u0106_filetool)
{
	TEMPLATE_MAIN_START;

	BEGIN_TEST(  u0106__get_class__run_test_cases(pCtx, pDataDir)  );

	TEMPLATE_MAIN_END;
}
