#!/bin/sh
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2010-2011 SourceGear, LLC
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

###############################################################################
## Script to compile third-party packages that we use and that
## are not typically installed on the system.
##
## This script will download and compile each package and
## "install" them into a "vv-thirdparty" directory.
###############################################################################
## Usage:
##     ./build_mac.sh [universal | i386]
##
## Build either universal or 32-bit versions of these libraries.
## Set CMAKE_OSX_ARCHITECTURES in the main Veracity CMake
## script to whatever value you use here.
##
## If you omit the argument, we will make a guess based on the value of 
## "hw.cpu64bit_capable" in sysctl.  If the platform is 64bit capable, we will
## build universal.  On machines that are not 64bit capable, we will build 32bit.
##
###############################################################################
IS64BIT=`sysctl -n hw.cpu64bit_capable`

if [ $# -eq 0 ]; then
	if [ ${IS64BIT} == "1" ]; then
		ARCH=universal
	else
		ARCH=i386
	fi
elif [ $1 == universal ]; then
	ARCH=universal
elif [ $1 == i386 ]; then
	ARCH=i386
else
   echo "Usage: $0 [universal | i386]"
   exit 1;
fi

PWD=`pwd`
PARENT_DIR=../
if [ $# -eq 2 ]; then
	# PARENT_DIR is an undocumented extra parameter that you can use to build
	# the "vv-thirdparty" directory in a non-default location. Most likely this
	# is because your source directory and your build directory are not peers
	# (and the fact that you will more likely want "vv-thirdparty" to be a peer
	# of your build directory than a peer of your of your source directory).
	# Simply provide this parameter, giving it the location of the the
	# directory that you want to be the parent containing "vv-thirdparty".
	PARENT_DIR="$2"
fi
# Stop on any error and report failure.
set -e
trap "echo ; echo An error has occurred! Veracity thirdparty libraries did NOT finish building!" ERR

# Get the path to this script
SCRIPT_PATH=`dirname $0`
THIRDPARTY_DIR=`cd $SCRIPT_PATH; pwd`
cd $PWD

# Some supporting files are in veracity/thirdparty/patches
PATCHES=$THIRDPARTY_DIR/patches

cd "$PARENT_DIR"
VVTHIRDPATH_PATH=`pwd`/vv-thirdparty
OUTPUT64="$VVTHIRDPATH_PATH"/x86_64
OUTPUT32="$VVTHIRDPATH_PATH"/i386
if [ ${ARCH} == "universal" ]; then
	OUTPUT="$VVTHIRDPATH_PATH"/universal
else
	OUTPUT="$OUTPUT32"
fi

echo "Installing veracity thirdparty dependencies to $VVTHIRDPATH_PATH"
rm -rf "$OUTPUT32"
rm -rf "$OUTPUT64"
rm -rf "$OUTPUT"
mkdir -p "$OUTPUT32"/include "$OUTPUT64"/include "$OUTPUT"/include
mkdir -p "$OUTPUT32"/lib "$OUTPUT64"/lib "$OUTPUT"/lib
mkdir -p "$OUTPUT32"/bin "$OUTPUT64"/bin "$OUTPUT"/bin

# Intermediate folders. Can be deleted after a successful build, if desired.
DOWNLOADS=`pwd`/vv-thirdparty/downloads
mkdir -p "$DOWNLOADS"
BUILDS=`pwd`/vv-thirdparty/builds
mkdir -p "$BUILDS"

# Helper funciton to download a file if it hasn't already been downloaded.
download () {
	echo "Downloading..."
	cd "$DOWNLOADS"
	if [ -f $2 ]; then
		echo "(Already downloaded.)"
	else
		curl -L $1 -o $2
	fi
}

###############################################################################
## ICU (International Components for Unicode)

echo
echo "ICU (icu4c-4_6)"
echo "================="
if [ -f "$BUILDS"/icu4c-4_6-build_finished ]; then
	echo "(Using build output from previous build.)"
else
	download http://download.icu-project.org/files/icu4c/4.6/icu4c-4_6-src.tgz icu4c-4_6-src.tgz
	
	echo "Unpacking..."
	cd "$BUILDS"
	rm -rf icu icu4c-4_6 # Clear files from previous build.
	tar xzf "$DOWNLOADS"/icu4c-4_6-src.tgz
	chmod +x icu/source/runConfigureICU icu/source/configure icu/source/install-sh
	cp -r icu icu4c-4_6-32
	mv icu icu4c-4_6-64
	
	echo "Building 32bit..."
	cd "$BUILDS"/icu4c-4_6-32/source
	LOGFILE="$BUILDS"/icu4c-4_6-32/build_log.txt
	echo "(Build output logged at $LOGFILE)"
	
	./runConfigureICU MacOSX --prefix="$OUTPUT32" --disable-icuio --disable-layout --with-library-bits=32 --with-iostream=none > "$LOGFILE" 2>&1
	make >> "$LOGFILE" 2>&1
	
	if [ ${ARCH} == "universal" ]; then
		echo "Building 64bit..."
		cd "$BUILDS"/icu4c-4_6-64/source
		LOGFILE="$BUILDS"/icu4c-4_6-64/build_log.txt
		echo "(Build output logged at $LOGFILE)"
	
		./runConfigureICU MacOSX --prefix="$OUTPUT64" --disable-icuio --disable-layout --with-library-bits=64 --with-iostream=none > "$LOGFILE" 2>&1
		make >> "$LOGFILE" 2>&1
	fi
	touch "$BUILDS"/icu4c-4_6-build_finished
fi

echo "Copying 32bit files..."
cd "$BUILDS"/icu4c-4_6-32/source
LOGFILE="$BUILDS"/icu4c-4_6-32/install_log.txt
echo "(Output logged at $LOGFILE)"
make install > "$LOGFILE"

if [ ${ARCH} == "universal" ]; then
	echo "Copying 64bit files..."
	cd "$BUILDS"/icu4c-4_6-64/source
	LOGFILE="$BUILDS"/icu4c-4_6-64/install_log.txt
	echo "(Output logged at $LOGFILE)"
	make install > "$LOGFILE"

	cd "$OUTPUT"/lib

	echo "Creating Universal Binaries"
	lipo -create "$OUTPUT32"/lib/libicudata.46.0.dylib "$OUTPUT64"/lib/libicudata.46.0.dylib -output libicudata.46.0.dylib
	lipo -create "$OUTPUT32"/lib/libicui18n.46.0.dylib "$OUTPUT64"/lib/libicui18n.46.0.dylib -output libicui18n.46.0.dylib
	lipo -create "$OUTPUT32"/lib/libicutest.46.0.dylib "$OUTPUT64"/lib/libicutest.46.0.dylib -output libicutest.46.0.dylib
	lipo -create "$OUTPUT32"/lib/libicutu.46.0.dylib "$OUTPUT64"/lib/libicutu.46.0.dylib -output libicutu.46.0.dylib
	lipo -create "$OUTPUT32"/lib/libicuuc.46.0.dylib "$OUTPUT64"/lib/libicuuc.46.0.dylib -output libicuuc.46.0.dylib

	echo "Creating library symlinks"
	ln -s libicudata.46.0.dylib libicudata.46.dylib
	ln -s libicudata.46.0.dylib libicudata.dylib
	ln -s libicui18n.46.0.dylib libicui18n.46.dylib
	ln -s libicui18n.46.0.dylib libicui18n.dylib
	ln -s libicutest.46.0.dylib libicutest.46.dylib
	ln -s libicutest.46.0.dylib libicutest.dylib
	ln -s libicutu.46.0.dylib libicutu.46.dylib
	ln -s libicutu.46.0.dylib libicutu.dylib
	ln -s libicuuc.46.0.dylib libicuuc.46.dylib
	ln -s libicuuc.46.0.dylib libicuuc.dylib
	
	echo "Copying Includes"
	cp -r "$OUTPUT64"/include/unicode "$OUTPUT"/include/unicode
fi 

echo "Performing quick sanity check on the bitness of the ICU libraries..."
file "$OUTPUT"/lib/libicuuc.dylib
echo "Ok."

echo "Done."

###############################################################################
## NSPR (Netscape Portable Runtime)
##     This library is needed in order to build SpiderMonkey with the
##     JS_THREADSAFE option.

echo
echo "NSPR (nspr-4.8.6)"
echo "================="
if [ -f "$BUILDS"/nspr-4.8.6/build_finished ]; then
	echo "(Using build output from previous build.)"
else
	download http://ftp.mozilla.org/pub/mozilla.org/nspr/releases/v4.8.6/src/nspr-4.8.6.tar.gz nspr-4.8.6.tar.gz
	
	echo "Unpacking..."
	cd "$BUILDS"
	rm -rf nspr-4.8.6 # Clear files from previous build.
	tar xzf "$DOWNLOADS"/nspr-4.8.6.tar.gz
	
	echo "Building Universal Binary..."
	cd "$BUILDS"/nspr-4.8.6/mozilla/nsprpub
	LOGFILE="$BUILDS"/nspr-4.8.6/build_log.txt
	echo "(Build output logged at $LOGFILE)"
	if [ ${ARCH} == "universal" ]; then
		env CFLAGS="-arch i386 -arch x86_64" LDFLAGS="-arch i386 -arch x86_64" ./configure --disable-debug --enable-optimize > "$LOGFILE" 2>&1
	else
		./configure --disable-debug --enable-optimize > "$LOGFILE" 2>&1
	fi
	
	make >> "$LOGFILE" 2>&1
	
	touch "$BUILDS"/nspr-4.8.6/build_finished
fi

echo "Copying files..."
cp "$BUILDS"/nspr-4.8.6/mozilla/nsprpub/dist/lib/* "$OUTPUT"/lib/

###############################################################################
## SpiderMonkey - JavaScript interpreter

echo
echo "Spidermonkey (js-1.8.0-rc1)"
echo "==========================="
if [ -f "$BUILDS"/js-1.8.0-rc1-build_finished ]; then
	echo "(Using build output from previous build.)"
else
	download http://ftp.mozilla.org/pub/mozilla.org/js/js-1.8.0-rc1.tar.gz js-1.8.0-rc1.tar.gz
	
	echo "Unpacking..."
	cd "$BUILDS"
	rm -rf js js-1.8.0-rc1 # Clear files from previous build.
	tar xzf "$DOWNLOADS"/js-1.8.0-rc1.tar.gz
	if [ ${ARCH} == "universal" ]; then
		cp -r js js-1.8.0-rc1-64
	fi
	
	mv js js-1.8.0-rc1-32
	
	## There is a syntax error in JS 1.8.0-rc1 (jsprf.c:644) that shows up
	## with various versions of GCC.  Here is a patch for it from macports.
	if [ ${ARCH} == "universal" ]; then
		patch "$BUILDS"/js-1.8.0-rc1-64/src/jsprf.c "$PATCHES"/mac__js-1.8.0-rc1__jsprf.c.patch
		patch "$BUILDS"/js-1.8.0-rc1-64/src/jsapi.c "$PATCHES"/js-1.8.0-rc1__jsapi.c.patch
		patch "$BUILDS"/js-1.8.0-rc1-64/src/jscntxt.c "$PATCHES"/js-1.8.0-rc1__jscntxt.c.patch
		patch "$BUILDS"/js-1.8.0-rc1-64/src/jscntxt.h "$PATCHES"/js-1.8.0-rc1__jscntxt.h.patch
	fi
	
	patch "$BUILDS"/js-1.8.0-rc1-32/src/jsprf.c "$PATCHES"/mac__js-1.8.0-rc1__jsprf.c.patch
	patch "$BUILDS"/js-1.8.0-rc1-32/src/jsapi.c "$PATCHES"/js-1.8.0-rc1__jsapi.c.patch
	patch "$BUILDS"/js-1.8.0-rc1-32/src/jscntxt.c "$PATCHES"/js-1.8.0-rc1__jscntxt.c.patch
	patch "$BUILDS"/js-1.8.0-rc1-32/src/jscntxt.h "$PATCHES"/js-1.8.0-rc1__jscntxt.h.patch
	
	echo "Building 32bit..."
	## Add bitness arg to GCC command line.  I tried adding it to the
	## XCFLAGS variable that was mentioned in the makefile, but that
	## didn't work (some of the loads didn't see it).
	MY_CC="/usr/bin/gcc -arch i386"
	cd "$BUILDS"/js-1.8.0-rc1-32/src
	LOGFILE="$BUILDS"/js-1.8.0-rc1-32/build_log.txt
	echo "(Build output logged at $LOGFILE)"
	make CC="$MY_CC" JS_DIST="$BUILDS"/nspr-4.8.6/mozilla/nsprpub/dist JS_THREADSAFE=1 BUILD_OPT=1 -f Makefile.ref > "$LOGFILE" 2>&1

	if [ ${ARCH} == "universal" ]; then
		echo "Building 64bit..."
		## Add bitness arg to GCC command line.  I tried adding it to the
		## XCFLAGS variable that was mentioned in the makefile, but that
		## didn't work (some of the loads didn't see it).
		MY_CC="/usr/bin/gcc -arch x86_64"
		cd "$BUILDS"/js-1.8.0-rc1-64/src
		LOGFILE="$BUILDS"/js-1.8.0-rc1-64/build_log.txt
		echo "(Build output logged at $LOGFILE)"
		make CC="$MY_CC" JS_DIST="$BUILDS"/nspr-4.8.6/mozilla/nsprpub/dist JS_THREADSAFE=1 BUILD_OPT=1 -f Makefile.ref > "$LOGFILE" 2>&1
	fi
	
	echo "Performing quick sanity check that the js executable is the right bitness..."
	
	touch "$BUILDS"/js-1.8.0-rc1-build_finished
fi

echo "Copying files..."
if [ ${ARCH} == "universal" ]; then
	mkdir -p "$OUTPUT"/include/spidermonkey/i386
	mkdir -p "$OUTPUT"/include/spidermonkey/x86_64
	cp "$BUILDS"/js-1.8.0-rc1-64/src/Darwin_OPT.OBJ/*.h "$OUTPUT"/include/spidermonkey/x86_64
	cp "$BUILDS"/js-1.8.0-rc1-64/src/*.h                "$OUTPUT"/include/spidermonkey/
	cp "$BUILDS"/js-1.8.0-rc1-64/src/*.msg              "$OUTPUT"/include/spidermonkey/
	cp "$BUILDS"/js-1.8.0-rc1-64/src/*.tbl              "$OUTPUT"/include/spidermonkey/
	cp "$BUILDS"/js-1.8.0-rc1-32/src/Darwin_OPT.OBJ/*.h "$OUTPUT"/include/spidermonkey/i386

	patch "$OUTPUT"/include/spidermonkey/jstypes.h "$PATCHES"/mac__js-1.8.0-rc1__jstypes.h.patch

	lipo -create "$BUILDS"/js-1.8.0-rc1-32/src/Darwin_OPT.OBJ/libjs.a  \
		"$BUILDS"/js-1.8.0-rc1-64/src/Darwin_OPT.OBJ/libjs.a -output "$OUTPUT"/lib/libspidermonkey.a
else
	mkdir -p "$OUTPUT"/include/spidermonkey/
	cp "$BUILDS"/js-1.8.0-rc1-32/src/*.h                "$OUTPUT"/include/spidermonkey/
	cp "$BUILDS"/js-1.8.0-rc1-32/src/*.msg              "$OUTPUT"/include/spidermonkey/
	cp "$BUILDS"/js-1.8.0-rc1-32/src/*.tbl              "$OUTPUT"/include/spidermonkey/
	cp "$BUILDS"/js-1.8.0-rc1-32/src/Darwin_OPT.OBJ/*.h "$OUTPUT"/include/spidermonkey/
	cp "$BUILDS"/js-1.8.0-rc1-32/src/Darwin_OPT.OBJ/libjs.a "$OUTPUT"/lib/libspidermonkey.a
fi

## (Veracity only links with the libspidermonkey.a, but /usr/bin/file
## doesn't tell us anything usefull for .a's and the JS makefile
## builds both.)
file "$OUTPUT"/lib/libspidermonkey.a

###############################################################################
## Done.

if [ ${ARCH} == "universal" ]; then
	echo "Cleaning up..."
	rm -r "$OUTPUT32"
	rm -r "$OUTPUT64"
fi

echo
echo
echo "Finished building vv-thirdparty!"
echo
echo "It is safe to delete the \"downloads\" and \"builds\" directories."
