/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include <sg.h>
#include <sghash.h>
#include <spidermonkey/jsapi.h>
#include <spidermonkey/jsdate.h>
#include "sg_js_safeptr__private.h"

// Function in uridispatch that is not a part of its regular public interface.
void sg_uridispatch__query_repo_availability(SG_context * pCtx, const char * szRepoName, SG_bool * pbIsAvailable);

/* TODO make sure we tell JS that strings are utf8 */

/*
 * Any object that requires a pointer (like an SG_repo)
 * is handled as a special class with private data
 * for storing the pointer (as an SG_safeptr).  See
 * sg_jsglue__get_object_private().
 *
 * Any object that can be serialized to a vhash is
 * simply handled as a plain JS object.
 */

/* Whenever we put a 64 bit integer into JSON, or into
 * a JSObject, if it won't fit into a 32 bit signed
 * int, we actually store it as a string instead of
 * as an int. */

//////////////////////////////////////////////////////////////////


SG_context * SG_jsglue__get_clean_sg_context(JSContext * cx)
{
	// TODO Do we want to use a SG_safeptr to wrap this?
	// TODO We have a bit of a chicken-n-egg problem.
	// TODO
	// TODO I don't think it is as much of an issue because
	// TODO we only have one JSContext.  Which is not like
	// TODO a JSObject where we have many.
	//
	// I'm going to say no for now.
	SG_context * ptr = NULL;
	ptr = (SG_context *)JS_GetContextPrivate(cx);
	SG_context__err_reset(ptr);
	return ptr;
}

SG_context * SG_jsglue__get_sg_context(JSContext * cx)
{
	// TODO Do we want to use a SG_safeptr to wrap this?
	// TODO We have a bit of a chicken-n-egg problem.
	// TODO
	// TODO I don't think it is as much of an issue because
	// TODO we only have one JSContext.  Which is not like
	// TODO a JSObject where we have many.
	//
	// I'm going to say no for now.
	SG_context * ptr = NULL;
	ptr = (SG_context *)JS_GetContextPrivate(cx);
	return ptr;
}
void SG_jsglue__set_sg_context(SG_context * pCtx, JSContext * cx)
{
	// remember pCtx in the cx so that it can be referenced
	// by methods that we export to JavaScript.
	JS_SetContextPrivate(cx, pCtx);
}

//////////////////////////////////////////////////////////////////

/**
 * This is a JSContextCallback.
 *
 * This gets called whenever a CX is created or destroyed.
 *
 * https://developer.mozilla.org/en/SpiderMonkey/JSAPI_Reference/JS_SetContextCallback
 */
JSBool SG_jsglue__context_callback(JSContext * cx, uintN contextOp)
{
	switch (contextOp)
	{
	case JSCONTEXT_NEW:
		JS_SetErrorReporter(cx,SG_jsglue__error_reporter);
		JS_SetVersion(cx, JSVERSION_LATEST);
		// TODO see SetContextOptions() in vscript.c for other goodies to set.
		return JS_TRUE;

	default:
	//case JSCONTEXT_DESTROY:
		return JS_TRUE;
	}
}

/**
 * This is a JSErrorReporter.
 *
 * This gets called whenever an error happens on this CX (either inside JavaScript proper
 * or within our glue code when SG_jsglue__report_sg_error() is called).
 */
void SG_jsglue__error_reporter(JSContext * cx, const char * pszMessage, JSErrorReport * pJSErrorReport)
{
	SG_context * pCtx = SG_jsglue__get_sg_context(cx);

	if (SG_context__has_err(pCtx))
	{
		// we had an error within the jsglue code in sglib (which has pushed
		// one or more sglib-stackframes) before calling SG_jsglue__report_sg_error()
		// at the end of the METHOD/PROPERY.
		//
		// So, the information in the JSErrorReport may or may not be of that
		// much value.
		//
		// We ignore the supplied pszMessage because it is just a copy of
		// the description that was provided when the original sglib-error
		// was thrown.
		//
		// TODO experiment with this and see if the JSErrorReport gives us any
		// TODO new information (like the location in the script) that we'll want.
		// TODO
		// TODO for now, just append a stackframe with the info and see if we
		// TODO get a duplicate entry....

		SG_context__err_stackframe_add(pCtx,
									   ((pJSErrorReport->filename) ? pJSErrorReport->filename : "<no_filename>"),
									   pJSErrorReport->lineno ? pJSErrorReport->lineno : 1);
	}
	else
	{
		// something within JavaScript threw an error (and sglib was not involved).
		// start a new error in pCtx using all of the information in JSErrorReport.

		SG_context__err(pCtx, SG_ERR_JS,
						((pJSErrorReport->filename) ? pJSErrorReport->filename : "<no_filename>"),
						pJSErrorReport->lineno ? pJSErrorReport->lineno : 1,
						pszMessage);
	}
}

/**
 * This should be called as the last step in the fail-block of ***all***
 * METHOD/PROPERTY functions.
 *
 * This is necessary so that JavaScript try/exception handling works
 * as expected.
 *
 * IT IS IMPORTANT THAT ALL CALLS TO SG_jsglue__report_sg_error() ***NOT*** BE
 * WRAPPED IN SG_ERR_IGNORE() (because it would hide the error from us).
 */
void SG_jsglue__report_sg_error(/*const*/ SG_context * pCtx, JSContext * cx)
{
	SG_error err;
	char szErr[SG_ERROR_BUFFER_SIZE];
	SG_string * pStrContextDump = NULL;

	SG_ASSERT(  SG_context__has_err(pCtx)  );

	SG_context__err_to_string(pCtx, SG_TRUE, &pStrContextDump);
	if (!pStrContextDump)
	{
		// We're in an error-on-error state.
		// SG_context's error routines won't give us the description because it doesn't want to allocate memory.
		SG_context__get_err(pCtx, &err);
		SG_error__get_message(err, SG_TRUE, szErr, sizeof(szErr));
		JS_ReportError(cx, "%s", szErr);
	}
	else
	{
		JS_ReportError(cx, "%s", SG_string__sz(pStrContextDump));
		SG_STRING_NULLFREE(pCtx, pStrContextDump);
	}

	SG_context__err_reset(pCtx);
}

//////////////////////////////////////////////////////////////////

/**
 * A simple wrapper around SG_free() because it is a macro
 * rather than a function pointer and I can reference it from
 * SG_vector__free__with_assoc() call below.
 */
static SG_free_callback _my_sg_free;
static void _my_sg_free(SG_context * pCtx, void * pVoid)
{
	SG_free(pCtx, pVoid);
}

//////////////////////////////////////////////////////////////////

/**
 * We cannot control the prototypes of methods and properties;
 * these are defined by SpiderMonkey.
 *
 * Note that they do not take a SG_Context.  Methods and properties
 * will have to extract it from the js-context-private data.
 */
#define SG_JSGLUE_METHOD_NAME(class, name) sg_jsglue__method__##class##__##name
#define SG_JSGLUE_METHOD_PROTOTYPE(class, name) static JSBool SG_JSGLUE_METHOD_NAME(class, name)(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)

#define SG_JSGLUE_PROPERTY_NAME(class, name) sg_jsglue__property__##class##__##name
#define SG_JSGLUE_PROPERTY_PROTOTYPE(class, name) static JSBool SG_JSGLUE_PROPERTY_NAME(class, name)(JSContext *cx, JSObject *obj, jsval idval, jsval *vp)

void sg_jsglue__set_variant_as_property(SG_context * pCtx, JSContext* cx, JSObject* obj, const char* psz_name, const SG_variant* pv);
void sg_jsglue__set_variant_as_element(SG_context * pCtx, JSContext* cx, JSObject* obj, SG_uint32 i_ndx, const SG_variant* pv);

/* ---------------------------------------------------------------- */
/* JSClass structures.  The InitClass calls are at the bottom of
 * the file */
/* ---------------------------------------------------------------- */

static JSClass sg_audit_class = {
    "sg_audit",
    JSCLASS_HAS_PRIVATE,
    JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub,
    JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub,
    JSCLASS_NO_OPTIONAL_MEMBERS
};

static JSClass sg_cbuffer_class = {
    "sg_cbuffer",
    JSCLASS_HAS_PRIVATE,
    JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub,
    JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub,
    JSCLASS_NO_OPTIONAL_MEMBERS
};

static JSClass sg_blobset_class = {
    "sg_blobset",
    JSCLASS_HAS_PRIVATE,
    JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub,
    JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub,
    JSCLASS_NO_OPTIONAL_MEMBERS
};

static JSClass sg_repo_class = {
    "sg_repo",
    JSCLASS_HAS_PRIVATE,
    JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub,
    JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub,
    JSCLASS_NO_OPTIONAL_MEMBERS
};

static JSClass sg_fetchblobhandle_class = {
    "sg_fetchblobhandle",
    JSCLASS_HAS_PRIVATE,
    JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub,
    JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub,
    JSCLASS_NO_OPTIONAL_MEMBERS
};

static JSClass sg_fetchfilehandle_class = {
    "sg_fetchfilehandle",
    JSCLASS_HAS_PRIVATE,
    JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub,
    JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub,
    JSCLASS_NO_OPTIONAL_MEMBERS
};

static JSClass sg_class = {
    "sg",
    0,
    JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub,
    JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub,
    JSCLASS_NO_OPTIONAL_MEMBERS
};

static JSClass fs_class = {
    "sg_fs",
    0,
    JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub,
    JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub,
    JSCLASS_NO_OPTIONAL_MEMBERS
};

static JSClass server_class = {
    "sg_server",
    0,
    JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub,
    JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub,
    JSCLASS_NO_OPTIONAL_MEMBERS
};

static JSClass file_class = {
    "sg_file",
    0,
    JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub,
    JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub,
    JSCLASS_NO_OPTIONAL_MEMBERS
};

static JSClass fsobj_properties_class = {
    "fsobj_properties",
    0,
    JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub,
    JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub,
    JSCLASS_NO_OPTIONAL_MEMBERS
};

/* TODO need a class for sg_file */

static JSClass pending_tree_class = {
    "sg_pending_tree",
    0,
    JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub,
    JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub,
    JSCLASS_NO_OPTIONAL_MEMBERS
};

/* ---------------------------------------------------------------- */
/* Utility functions */
/* ---------------------------------------------------------------- */


SG_safeptr* sg_jsglue__get_object_private(JSContext *cx, JSObject *obj)
{
    SG_safeptr* psp = (SG_safeptr*) JS_GetPrivate(cx, obj);
    return psp;
}

void sg_jsglue__copy_stringarray_into_jsobject(SG_context * pCtx, JSContext* cx, SG_stringarray* psa, JSObject* obj)
{
    SG_uint32 count = 0;
    SG_uint32 i;

    SG_ERR_CHECK(  SG_stringarray__count(pCtx, psa, &count)  );

    for (i=0; i<count; i++)
    {
        const char * pszv = NULL;
        jsval jv;

        SG_ERR_CHECK(  SG_stringarray__get_nth(pCtx, psa, i, &pszv)  );
        JSVAL_FROM_SZ(jv, pszv);

        SG_JS_BOOL_CHECK(  JS_SetElement(cx, obj, i, &jv)  );
    }

	return;

fail:
    return;
}

void sg_jsglue__copy_varray_into_jsobject(SG_context * pCtx, JSContext* cx, SG_varray* pva, JSObject* obj)
{
    SG_uint32 count = 0;
    SG_uint32 i;

    SG_ARGCHECK(JS_IsArrayObject(cx, obj), obj);

    SG_ERR_CHECK(  SG_varray__count(pCtx, pva, &count)  );

    for (i=0; i<count; i++)
    {
        const SG_variant* pv = NULL;

        SG_ERR_CHECK(  SG_varray__get__variant(pCtx, pva, i, &pv)  );
        SG_ERR_CHECK(  sg_jsglue__set_variant_as_element(pCtx, cx, obj, i, pv)  );
    }

	return;

fail:
    /* TODO free obj */
    return;
}

void sg_jsglue__copy_rbtree_keys_into_jsobject(SG_context * pCtx, JSContext* cx, SG_rbtree* prb, JSObject* obj)
{
	SG_rbtree_iterator* pit = NULL;
	const char* szKey;
	SG_bool bFound;
	SG_uint32 i;

	SG_ARGCHECK(JS_IsArrayObject(cx, obj), obj);

	SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pit, prb, &bFound, &szKey, NULL)  );
	for (i = 0; bFound; i++)
	{
		jsval jv;

		JSVAL_FROM_SZ(jv, szKey);

		SG_JS_BOOL_CHECK(  JS_SetElement(cx, obj, i, &jv)  );

		SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pit, &bFound, &szKey, NULL)  );
	}

	/* fall through */

fail:
	SG_RBTREE_ITERATOR_NULLFREE(pCtx, pit);
	
}

void sg_jsglue__int64_to_jsval(SG_context * pCtx, JSContext* cx, SG_int64 i, jsval* pjv)
{
    jsval jv = JSVAL_VOID;

    if (
            SG_int64__fits_in_int32(i)
            && INT_FITS_IN_JSVAL((SG_int32) i)
            )
    {
        jv = INT_TO_JSVAL((SG_int32) i);
    }
    else if (SG_int64__fits_in_double(i))
    {
        double d = (double) i;

        SG_JS_BOOL_CHECK(  JS_NewNumberValue(cx, d, &jv)  );
    }
    else
    {
        SG_ERR_THROW(  SG_ERR_INTEGER_OVERFLOW  );
    }

    *pjv = jv;

    return;

fail:
    return;
}

void sg_jsglue__set_variant_as_element(SG_context * pCtx, JSContext* cx, JSObject* obj, SG_uint32 i_ndx, const SG_variant* pv)
{
    jsval jv = JSVAL_VOID;
    JSString* pjs = NULL;
    JSObject* jso = NULL;

    SG_ARGCHECK(JS_IsArrayObject(cx, obj), obj);

	switch (pv->type)
	{
	case SG_VARIANT_TYPE_DOUBLE:
        {
            double d;

            SG_ERR_CHECK(  SG_variant__get__double(pCtx, pv, &d)  );
            SG_JS_BOOL_CHECK(  JS_NewNumberValue(cx, d, &jv)  );
            SG_JS_BOOL_CHECK(  JS_SetElement(cx, obj, i_ndx, &jv)  );
            break;
        }

	case SG_VARIANT_TYPE_INT64:
        {
            SG_int64 i = 0;

            SG_ERR_CHECK(  SG_variant__get__int64(pCtx, pv, &i)  );

            SG_ERR_CHECK(  sg_jsglue__int64_to_jsval(pCtx, cx, i, &jv)  );
            SG_JS_BOOL_CHECK(  JS_SetElement(cx, obj, i_ndx, &jv)  );

            break;
        }

	case SG_VARIANT_TYPE_BOOL:
        {
            SG_bool b;
            int v;

            SG_ERR_CHECK(  SG_variant__get__bool(pCtx, pv, &b)  );

            /* BOOLEAN_TO_JSVAL is documented as accepting only 0 or 1, so
             * we make sure */
            v = b ? 1 : 0;
            jv = BOOLEAN_TO_JSVAL(v);
            SG_JS_BOOL_CHECK(  JS_SetElement(cx, obj, i_ndx, &jv)  );
            break;
        }

	case SG_VARIANT_TYPE_NULL:
        {
            jv = JSVAL_NULL;
            SG_JS_BOOL_CHECK(  JS_SetElement(cx, obj, i_ndx, &jv)  );
            break;
        }

	case SG_VARIANT_TYPE_SZ:
        {
            const char* psz = NULL;

            SG_ERR_CHECK(  SG_variant__get__sz(pCtx, pv, &psz)  );
            SG_JS_NULL_CHECK(  (pjs = JS_NewStringCopyZ(cx, psz))  );
            jv = STRING_TO_JSVAL(pjs);
            SG_JS_BOOL_CHECK(  JS_SetElement(cx, obj, i_ndx, &jv)  );
            pjs = NULL;
            break;
        }

	case SG_VARIANT_TYPE_VHASH:
        {
            SG_vhash* pvh = NULL;

            SG_ERR_CHECK(  SG_variant__get__vhash(pCtx, pv, &pvh)  );
            SG_JS_NULL_CHECK(  (jso = JS_NewObject(cx, NULL, NULL, NULL))  );
            jv = OBJECT_TO_JSVAL(jso);
            SG_JS_BOOL_CHECK(  JS_SetElement(cx, obj, i_ndx, &jv)  );

            SG_ERR_CHECK(  sg_jsglue__copy_vhash_into_jsobject(pCtx, cx, pvh, jso)  );
            jso = NULL;
            break;
        }

	case SG_VARIANT_TYPE_VARRAY:
        {
            SG_varray* pva = NULL;

            SG_ERR_CHECK(  SG_variant__get__varray(pCtx, pv, &pva)  );
            SG_JS_NULL_CHECK(  (jso = JS_NewArrayObject(cx, 0, NULL))  );
            jv = OBJECT_TO_JSVAL(jso);
            SG_JS_BOOL_CHECK(  JS_SetElement(cx, obj, i_ndx, &jv)  );

            SG_ERR_CHECK(  sg_jsglue__copy_varray_into_jsobject(pCtx, cx, pva, jso)  );
            jso = NULL;
            break;
        }
	}

    return;

fail:
    /* TODO free pjs */
    /* TODO free jso */
    return;
}

void sg_jsglue__set_variant_as_property(SG_context * pCtx, JSContext* cx, JSObject* obj, const char* psz_name, const SG_variant* pv)
{
    jsval jv = JSVAL_VOID;
    JSString* pjs = NULL;
    JSObject* jso = NULL;

	switch (pv->type)
	{
	case SG_VARIANT_TYPE_DOUBLE:
        {
            double d;

            SG_ERR_CHECK(  SG_variant__get__double(pCtx, pv, &d)  );
            SG_JS_BOOL_CHECK(  JS_NewNumberValue(cx, d, &jv)  );
            SG_JS_BOOL_CHECK(  JS_SetProperty(cx, obj, psz_name, &jv)  );
            break;
        }

	case SG_VARIANT_TYPE_INT64:
        {
            SG_int64 i = 0;

            SG_ERR_CHECK(  SG_variant__get__int64(pCtx, pv, &i)  );

            SG_ERR_CHECK(  sg_jsglue__int64_to_jsval(pCtx, cx, i, &jv)  );
            SG_JS_BOOL_CHECK(  JS_SetProperty(cx, obj, psz_name, &jv)  );

            break;
        }

	case SG_VARIANT_TYPE_BOOL:
        {
            SG_bool b;
            int v;

            SG_ERR_CHECK(  SG_variant__get__bool(pCtx, pv, &b)  );

            /* BOOLEAN_TO_JSVAL is documented as accepting only 0 or 1, so
             * we make sure */
            v = b ? 1 : 0;
            jv = BOOLEAN_TO_JSVAL(v);
            SG_JS_BOOL_CHECK(  JS_SetProperty(cx, obj, psz_name, &jv)  );
            break;
        }

	case SG_VARIANT_TYPE_NULL:
        {
            jv = JSVAL_NULL;
            SG_JS_BOOL_CHECK(  JS_SetProperty(cx, obj, psz_name, &jv)  );
            break;
        }

	case SG_VARIANT_TYPE_SZ:
        {
            const char* psz = NULL;

            SG_ERR_CHECK(  SG_variant__get__sz(pCtx, pv, &psz)  );
            SG_JS_NULL_CHECK(  (pjs = JS_NewStringCopyZ(cx, psz))  );
            jv = STRING_TO_JSVAL(pjs);
            SG_JS_BOOL_CHECK(  JS_SetProperty(cx, obj, psz_name, &jv)  );
            pjs = NULL;
            break;
        }

	case SG_VARIANT_TYPE_VHASH:
        {
            SG_vhash* pvh = NULL;

            SG_ERR_CHECK(  SG_variant__get__vhash(pCtx, pv, &pvh)  );
            SG_JS_NULL_CHECK(  (jso = JS_NewObject(cx, NULL, NULL, NULL))  );
            jv = OBJECT_TO_JSVAL(jso);
            SG_JS_BOOL_CHECK(  JS_SetProperty(cx, obj, psz_name, &jv)  );

            SG_ERR_CHECK(  sg_jsglue__copy_vhash_into_jsobject(pCtx, cx, pvh, jso)  );
            jso = NULL;
            break;
        }

	case SG_VARIANT_TYPE_VARRAY:
        {
            SG_varray* pva = NULL;

            SG_ERR_CHECK(  SG_variant__get__varray(pCtx, pv, &pva)  );
            SG_JS_NULL_CHECK(  (jso = JS_NewArrayObject(cx, 0, NULL))  );
            jv = OBJECT_TO_JSVAL(jso);
            SG_JS_BOOL_CHECK(  JS_SetProperty(cx, obj, psz_name, &jv)  );
            SG_ERR_CHECK(  sg_jsglue__copy_varray_into_jsobject(pCtx, cx, pva, jso)  );
            jso = NULL;
            break;
        }
	}

    return;

fail:
    /* TODO free pjs */
    /* TODO free jso */
    return;
}

void sg_jsglue__copy_vhash_into_jsobject(SG_context * pCtx, JSContext* cx, SG_vhash* pvh, JSObject* obj)
{
    SG_uint32 count = 0;
    SG_uint32 i;

    SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh, &count)  );

    for (i=0; i<count; i++)
    {
        const char* psz_name = NULL;
        const SG_variant* pv = NULL;

        SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvh, i, &psz_name, &pv)  );
        SG_ERR_CHECK(  sg_jsglue__set_variant_as_property(pCtx, cx, obj, psz_name, pv)  );
    }

fail:
    /* TODO free obj */
    return;
}

void sg_jsglue__jsobject_to_varray(SG_context * pCtx, JSContext* cx, JSObject* obj, SG_varray** ppva)
{
    jsuint len = 0;
    SG_uint32 i = 0;
    SG_varray* pva = NULL;
    SG_varray* pva_sub = NULL;
    SG_vhash* pvh_sub = NULL;

    SG_JS_BOOL_CHECK(  JS_GetArrayLength(cx, obj, &len)  );

    SG_ERR_CHECK(  SG_VARRAY__ALLOC(pCtx, &pva)  );

    for (i=0; i<len; i++)
    {
        jsval jv_value;

        SG_JS_BOOL_CHECK(  JS_GetElement(cx, obj, i, &jv_value)  );

        if (JSVAL_IS_STRING(jv_value))
        {
            SG_ERR_CHECK(  SG_varray__append__string__sz(pCtx, pva, JS_GetStringBytes(JSVAL_TO_STRING(jv_value)))  );
        }
        else if (JSVAL_IS_INT(jv_value))
        {
            SG_ERR_CHECK(  SG_varray__append__int64(pCtx, pva, (SG_int64) JSVAL_TO_INT(jv_value))  );
        }
        else if (
                JSVAL_IS_NUMBER(jv_value)
                || JSVAL_IS_DOUBLE(jv_value)
                )
        {
            jsdouble d = 0;
            SG_JS_BOOL_CHECK(  JS_ValueToNumber(cx, jv_value, &d)  );
            SG_ERR_CHECK(  SG_varray__append__double(pCtx, pva, (double) d)  );
        }
        else if (JSVAL_IS_BOOLEAN(jv_value))
        {
            SG_ERR_CHECK(  SG_varray__append__bool(pCtx, pva, JSVAL_TO_BOOLEAN(jv_value))  );
        }
        else if (JSVAL_IS_NULL(jv_value))
        {
            SG_ERR_CHECK(  SG_varray__append__null(pCtx, pva)  );
        }
        else if (JSVAL_IS_OBJECT(jv_value))
        {
            JSObject* po = JSVAL_TO_OBJECT(jv_value);

            if (JS_IsArrayObject(cx, po))
            {
                SG_ERR_CHECK(  sg_jsglue__jsobject_to_varray(pCtx, cx, po, &pva_sub)  );

                SG_ERR_CHECK(  SG_varray__append__varray(pCtx, pva, &pva_sub)  );
            }
            else
            {
                SG_ERR_CHECK(  sg_jsglue__jsobject_to_vhash(pCtx, cx, po, &pvh_sub)  );

                SG_ERR_CHECK(  SG_varray__append__vhash(pCtx, pva, &pvh_sub)  );
            }
        }
        else if (JSVAL_IS_VOID(jv_value))
        {
			SG_ERR_THROW( SG_ERR_JS );
        }
        else
        {
            SG_ERR_THROW( SG_ERR_JS );
        }
    }

    *ppva = pva;
    pva = NULL;

	return;

fail:
    SG_VHASH_NULLFREE(pCtx, pvh_sub);
    SG_VARRAY_NULLFREE(pCtx, pva_sub);
	SG_VARRAY_NULLFREE(pCtx, pva);
}

void sg_jsglue__jsval_to_json(SG_context * pCtx, JSContext * cx, jsval obj, SG_bool bPretty, SG_string ** ppstr)
{
	SG_vhash* pvh = NULL;
    SG_varray* pva = NULL;
	SG_string * pstr = NULL;
	char * psz_foreign = NULL;
	JSObject * jso = NULL;
	JSString * pjs = NULL;
	
	
	char buf[256];
	
	if (JSVAL_IS_NULL(obj))
	{
		SG_ERR_CHECK( SG_STRING__ALLOC__RESERVE(pCtx, &pstr, 4) );
		SG_ERR_CHECK( SG_string__set__sz(pCtx, pstr, "null") );
	}
	else if (JSVAL_IS_BOOLEAN(obj))
	{
		SG_bool bValue = JSVAL_TO_BOOLEAN(obj);

		if (bValue)
		{
			SG_ERR_CHECK( SG_STRING__ALLOC__RESERVE(pCtx, &pstr, 4));
			SG_ERR_CHECK( SG_string__set__sz(pCtx, pstr, "true") );
		}
		else
		{
			SG_ERR_CHECK( SG_STRING__ALLOC__RESERVE(pCtx, &pstr, 5) );
			SG_ERR_CHECK( SG_string__set__sz(pCtx, pstr, "false") );
		}
	}
	else if(JSVAL_IS_OBJECT(obj))
	{
		jso = JSVAL_TO_OBJECT(obj);

		if (JS_IsArrayObject(cx, jso))
		{
			SG_ERR_CHECK(  sg_jsglue__jsobject_to_varray(pCtx, cx, jso, &pva)  );
			SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pstr)  );
			if (bPretty)
				SG_ERR_CHECK(  SG_varray__to_json__pretty_print_NOT_for_storage(pCtx, pva, pstr)  );
			else
				SG_ERR_CHECK(  SG_varray__to_json(pCtx, pva, pstr)  );
				
			SG_VARRAY_NULLFREE(pCtx, pva);
		}
		else
		{
			SG_ERR_CHECK(  sg_jsglue__jsobject_to_vhash(pCtx, cx, jso, &pvh)  );
			SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pstr)  );
			if (bPretty)
				SG_ERR_CHECK(  SG_vhash__to_json__pretty_print_NOT_for_storage(pCtx, pvh, pstr)  );
			else
				SG_ERR_CHECK(  SG_vhash__to_json(pCtx, pvh, pstr)  );
				
			SG_VHASH_NULLFREE(pCtx, pvh);
		}
	}
	else if(JSVAL_IS_STRING(obj))
	{
		pjs = JSVAL_TO_STRING(obj);
		psz_foreign = SG_ALLOC_FOREIGN( JS_EncodeString(cx, pjs) );
		SG_ERR_CHECK(  SG_sz__to_json(pCtx, psz_foreign, &pstr)  );
	}
	else if(JSVAL_IS_INT(obj))
	{
		SG_int_to_string_buffer tmp;
		SG_int64 i = JSVAL_TO_INT(obj);
		
		SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pstr)  );
		SG_ERR_CHECK(  SG_string__set__sz(pCtx, pstr, SG_int64_to_sz(i, tmp)) );
	}
	else if(JSVAL_IS_NUMBER(obj) || JSVAL_IS_DOUBLE(obj))
	{
		jsdouble d = 0;
		SG_JS_BOOL_CHECK(  JS_ValueToNumber(cx, obj, &d)  );
		SG_ERR_CHECK(  SG_sprintf(pCtx, buf, 255, "%f", (double) d)  );
		SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pstr)  );
		SG_ERR_CHECK(  SG_string__append__sz(pCtx, pstr, buf) );
	}
	else
	{
		SG_ERR_THROW(SG_ERR_VALUE_CANNOT_BE_SERIALIZED_TO_JSON);
	}
	
	JS_free(cx, SG_FREE_FOREIGN(psz_foreign));

	*ppstr = pstr;
	pstr = NULL;
	return;
	
fail:
	SG_VARRAY_NULLFREE(pCtx, pva);
	SG_VHASH_NULLFREE(pCtx, pvh);
	SG_STRING_NULLFREE(pCtx, pstr);
	JS_free(cx, SG_FREE_FOREIGN(psz_foreign));
}

void sg_jsglue__jsobject_to_vhash(SG_context * pCtx, JSContext* cx, JSObject* obj, SG_vhash** ppvh)
{
    JSIdArray* properties = NULL;
    SG_vhash* pvh = NULL;
    SG_int32 i;
    SG_vhash* pvh_sub = NULL;
    SG_varray* pva_sub = NULL;

    SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvh)  );

    SG_JS_NULL_CHECK(  properties = JS_Enumerate(cx, obj)  );

    for (i=0; i<properties->length; i++)
    {
        jsval jv_name;
        jsval jv_value;
        char* str_name = NULL;
        SG_int_to_string_buffer str_name_from_int;

        SG_JS_BOOL_CHECK(  JS_IdToValue(cx, properties->vector[i], &jv_name)  );
        if(JSVAL_IS_INT(jv_name))
        {
            // Keys may have been converted to type integer if the string was an integer
            // (Even with objects made by sg_jsglue__copy_vhash_into_jsobject(), which
            // inserts values using JS_SetProperty(), which takes a char * for the key!).
            SG_int64_to_sz(JSVAL_TO_INT(jv_name), str_name_from_int);
            str_name = str_name_from_int;
        }
        else
        {
            SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(jv_name)  );
            str_name = JS_GetStringBytes(JSVAL_TO_STRING(jv_name));
        }

        SG_JS_BOOL_CHECK(  JS_GetProperty(cx, obj, str_name, &jv_value)  );

        if (JSVAL_IS_STRING(jv_value))
        {
            SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvh, str_name, JS_GetStringBytes(JSVAL_TO_STRING(jv_value)))  );
        }
        else if (JSVAL_IS_INT(jv_value))
        {
            SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh, str_name, (SG_int64) JSVAL_TO_INT(jv_value))  );
        }
        else if (
                JSVAL_IS_NUMBER(jv_value)
                || JSVAL_IS_DOUBLE(jv_value)
                )
        {
            jsdouble d = 0;
            SG_JS_BOOL_CHECK(  JS_ValueToNumber(cx, jv_value, &d)  );
            SG_ERR_CHECK(  SG_vhash__add__double(pCtx, pvh, str_name, (double) d)  );
        }
        else if (JSVAL_IS_BOOLEAN(jv_value))
        {
            SG_ERR_CHECK(  SG_vhash__add__bool(pCtx, pvh, str_name, JSVAL_TO_BOOLEAN(jv_value))  );
        }
        else if (JSVAL_IS_NULL(jv_value))
        {
            SG_ERR_CHECK(  SG_vhash__add__null(pCtx, pvh, str_name)  );
        }
        else if (JSVAL_IS_OBJECT(jv_value))
        {
            JSObject* po = JSVAL_TO_OBJECT(jv_value);

            if (JS_IsArrayObject(cx, po))
            {
                SG_ERR_CHECK(  sg_jsglue__jsobject_to_varray(pCtx, cx, po, &pva_sub)  );
                SG_ERR_CHECK(  SG_vhash__add__varray(pCtx, pvh, str_name, &pva_sub)  );

            }
            else
            {
                SG_ERR_CHECK(  sg_jsglue__jsobject_to_vhash(pCtx, cx, po, &pvh_sub)  );

                SG_ERR_CHECK(  SG_vhash__add__vhash(pCtx, pvh, str_name, &pvh_sub)  );
            }
        }
        else if (JSVAL_IS_VOID(jv_value))
        {
            SG_ERR_THROW2(  SG_ERR_JS,
                           (pCtx, "%s cannot be undefined/void", str_name )  );
        }
        else
        {
            SG_ERR_THROW( SG_ERR_JS );
        }
    }

    JS_DestroyIdArray(cx, properties);

    *ppvh = pvh;
    pvh = NULL;

	return;

fail:
    JS_DestroyIdArray(cx, properties);
    SG_VHASH_NULLFREE(pCtx, pvh);
    SG_VHASH_NULLFREE(pCtx, pvh_sub);
    SG_VARRAY_NULLFREE(pCtx, pva_sub);
}

SG_bool sg_jsglue_get_equals_flag(const char* psz_flag, const char* psz_match, const char** psz_value)
{
    int len = SG_STRLEN(psz_match);
    if (
            (0 == memcmp(psz_flag, psz_match, len))
            && ('=' == psz_flag[len])
        )
    {
        *psz_value = psz_flag + len + 1;
        return SG_TRUE;
    }

    return SG_FALSE;
}


void sg_jsglue_get_pattern_array(SG_context * pCtx, const char* psz_patterns, SG_stringarray** psaPatterns)
{
	char** pppResults = NULL;
	SG_uint32 i;
	SG_uint32 ncount = 0;


	// TODO 2010/06/16 From what I can tell by skimming, this code is intended to
	// TODO            allow the caller to receive "exclude=a,b,c" and we cut these
	// TODO            apart (breaking on the comma) into individual strings and
	// TODO            append them to the array (which we allocate if necessary).
	// TODO
	// TODO            Do we REALLY want to do this?
	// TODO
	// TODO            The callers are set up to allow an arg array containing
	// TODO            something like [ "exclude=a,b,c", "exclude=*.h", ...],
	// TODO            so they could get the same effect with the multiple n/v pairs.
	// TODO
	// TODO            Commas are valid characters in filenames and patterns, so
	// TODO            this prohibits us from, for example, excluding **,v" files.

	SG_ERR_CHECK (  SG_string__split__sz_asciichar(pCtx, psz_patterns, ',', SG_STRLEN(psz_patterns) + 1, &pppResults, &ncount) );
	if (!(*psaPatterns))
	{
		SG_ERR_CHECK (  SG_STRINGARRAY__ALLOC(pCtx, psaPatterns, ncount) );
	}
	for(i = 0; i < ncount; i++)
	{
		SG_ERR_CHECK (  SG_stringarray__add(pCtx, *psaPatterns, pppResults[i])  );
	}


fail:
	if (pppResults)
	{
		for(i = 0; i < ncount; i++)
			SG_NULLFREE(pCtx, pppResults[i]);
		SG_NULLFREE(pCtx, pppResults);
	}
}

static void sg_jsglue__get_repo_from_jsval(SG_context* pCtx, JSContext* cx, jsval arg, SG_repo** ppRepo, const char *uid)
{
	SG_vhash* pvh_descriptor = NULL;
	SG_repo* pRepo = NULL;
	SG_vhash* pUser = NULL;

	if (JSVAL_IS_STRING(arg))
	{
		const char * szRepoName = NULL;
		SG_bool isAvailable = SG_TRUE;

		szRepoName = JS_GetStringBytes(JSVAL_TO_STRING(arg));
		SG_ERR_CHECK(  sg_uridispatch__query_repo_availability(pCtx, szRepoName, &isAvailable)  );
		if(!isAvailable)
			SG_ERR_THROW(SG_ERR_NOTAREPOSITORY);

		SG_ERR_CHECK(  SG_jscore__check_module_dags(pCtx, cx, JS_GetGlobalObject(cx), szRepoName, uid, "", SG_FALSE, SG_FALSE)  );

		SG_ERR_CHECK(  SG_REPO__OPEN_REPO_INSTANCE(pCtx, szRepoName, &pRepo)  );

		if (uid)
		{
			SG_ERR_CHECK(  SG_user__lookup_by_userid(pCtx, pRepo, uid, &pUser)  );

			if (! pUser)
				SG_ERR_THROW2( SG_ERR_USERNOTFOUND, (pCtx, "uid %s not found in repo %s", uid, szRepoName)  );
		}
	}
	else if (!JSVAL_IS_NULL(arg) && JSVAL_IS_OBJECT(arg))
	{
		JSObject* po = JSVAL_TO_OBJECT(arg);
		SG_ERR_CHECK(  sg_jsglue__jsobject_to_vhash(pCtx, cx, po, &pvh_descriptor)  );
		SG_ERR_CHECK(  SG_repo__open_repo_instance__unnamed(pCtx, &pvh_descriptor, &pRepo)  );
	}
	else
	{
		// TODO what should the error code be?  all i found here was a goto fail.
		SG_ERR_THROW( SG_ERR_JS );
	}

	SG_RETURN_AND_NULL(pRepo, ppRepo);

	/* fall through */
fail:
	SG_VHASH_NULLFREE(pCtx, pvh_descriptor);
	SG_REPO_NULLFREE(pCtx, pRepo);
	SG_VHASH_NULLFREE(pCtx, pUser);
}

static void sg_jsglue__get_repo_from_cwd(SG_context* pCtx, SG_repo** ppRepo, SG_pathname** ppPathCwd, SG_pathname** ppPathWorkingDirectoryTop)
{
	SG_pathname * pPathCwd = NULL;
	SG_repo * pRepo = NULL;
	SG_pathname * pPathWorkingDirectoryTop = NULL;
	SG_string * pstrRepoDescriptorName = NULL;

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC(pCtx, &pPathCwd)  );
	SG_ERR_CHECK(  SG_pathname__set__from_cwd(pCtx, pPathCwd)  );
	SG_ERR_CHECK(  SG_workingdir__find_mapping(pCtx, pPathCwd, &pPathWorkingDirectoryTop, &pstrRepoDescriptorName, NULL)  );
	SG_ERR_CHECK(  SG_REPO__OPEN_REPO_INSTANCE(pCtx, SG_string__sz(pstrRepoDescriptorName), &pRepo)  );

	SG_RETURN_AND_NULL(pPathCwd, ppPathCwd);
	SG_RETURN_AND_NULL(pPathWorkingDirectoryTop, ppPathWorkingDirectoryTop);
	SG_RETURN_AND_NULL(pRepo, ppRepo);

	/* fall through */
fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathCwd);
	SG_PATHNAME_NULLFREE(pCtx, pPathWorkingDirectoryTop);
	SG_STRING_NULLFREE(pCtx, pstrRepoDescriptorName);
}

/**
 * Checks if a JS argument is a filespec argument.
 * include, exclude, include-from, exclude-from, no-ignores
 */
static SG_bool sg_jsglue__is_filespec_arg(
	const char* szOption //< [in] Option to check.
	)
{
	/* See: Y2908 
	const char* szValue = NULL;
	
	if (sg_jsglue_get_equals_flag (szOption, "include", &szValue))
	{
		return SG_TRUE;
	}
	else if (sg_jsglue_get_equals_flag (szOption, "exclude", &szValue))
	{
		return SG_TRUE;
	}
	else if (sg_jsglue_get_equals_flag (szOption, "include-from", &szValue))
	{
		return SG_TRUE;
	}
	else if (sg_jsglue_get_equals_flag (szOption, "exclude-from", &szValue))
	{
		return SG_TRUE;
	}
	else */
	if (strcmp(szOption, "no-ignores") == 0)
	{
		return SG_TRUE;
	}
	else
	{
		return SG_FALSE;
	}
}

/**
 * Processes JS arguments identified as filespec arguments.
 * Manipulates the given filespec object according to the argument's value.
 */
static void sg_jsglue__process_filespec_arg(
	SG_context*   pCtx,     //< [in] [out] Error and context info.
	const char*   szOption, //< [in] The argument that was identified as a filespec argument using sg_jsglue__is_filespec_arg.
	SG_file_spec* pFilespec //< [in] [out] Filespec to manipulate according to the value.
	)
{
	const char*     szValue = NULL;
	SG_stringarray* pValues = NULL;
	SG_pathname * pPathname = NULL;

	SG_NULLARGCHECK(szOption);
	SG_NULLARGCHECK(pFilespec);

	if (sg_jsglue_get_equals_flag (szOption, "include", &szValue))
	{
		// TODO: read the patterns directly from JS into the filespec
		//       rather than using an intermediate stringarray
		SG_ERR_CHECK(  SG_STRINGARRAY__ALLOC(pCtx, &pValues, 1)  );
		SG_ERR_CHECK(  sg_jsglue_get_pattern_array(pCtx, szValue, &pValues)  );
		SG_ERR_CHECK(  SG_file_spec__add_patterns__stringarray(pCtx, pFilespec, SG_FILE_SPEC__PATTERN__INCLUDE, pValues, 0u)  );
	}
	else if (sg_jsglue_get_equals_flag (szOption, "exclude", &szValue))
	{
		// TODO: read the patterns directly from JS into the filespec
		//       rather than using an intermediate stringarray
		SG_ERR_CHECK(  SG_STRINGARRAY__ALLOC(pCtx, &pValues, 1)  );
		SG_ERR_CHECK(  sg_jsglue_get_pattern_array(pCtx, szValue, &pValues)  );
		SG_ERR_CHECK(  SG_file_spec__add_patterns__stringarray(pCtx, pFilespec, SG_FILE_SPEC__PATTERN__EXCLUDE, pValues, 0u)  );
	}
	else if (sg_jsglue_get_equals_flag (szOption, "include-from", &szValue))
	{
		SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPathname, szValue)  );
		SG_ERR_CHECK(  SG_file_spec__add_patterns__file(pCtx, pFilespec, SG_FILE_SPEC__PATTERN__INCLUDE, pPathname, 0u)  );
	}
	else if (sg_jsglue_get_equals_flag (szOption, "exclude-from", &szValue))
	{
		SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPathname, szValue)  );
		SG_ERR_CHECK(  SG_file_spec__add_patterns__file(pCtx, pFilespec, SG_FILE_SPEC__PATTERN__EXCLUDE, pPathname, 0u)  );
	}
	else if (strcmp(szOption, "no-ignores") == 0)
	{
		SG_ERR_CHECK(  SG_file_spec__modify_flags(pCtx, pFilespec, SG_FILE_SPEC__NO_IGNORES, 0u, NULL, NULL)  );
	}

fail:
	SG_STRINGARRAY_NULLFREE(pCtx, pValues);
	SG_PATHNAME_NULLFREE(pCtx, pPathname);
}

static char* sg_jsglue__getObjectType(SG_treenode_entry_type type)
{
	if (type == SG_TREENODEENTRY_TYPE_REGULAR_FILE)
		return "File";
	else if (type == SG_TREENODEENTRY_TYPE_DIRECTORY)
			return "Directory";
	else if (type == SG_TREENODEENTRY_TYPE_SYMLINK)
			return "Symlink";
	else
		return "Invalid";
}

static void sg_jsglue__treenode_entry_hash( SG_context * pCtx,
	 SG_repo* pRepo,
	 const char* pszChangeSetID,
	 SG_treenode_entry* pTreenodeEntry,
	 const char* pszPath,
	 const char* pszGid,
	 SG_vhash** ppvhResults )
{
	SG_string* content = NULL;
	SG_uint32 count;
	SG_uint32 k;
	const char* pszHid = NULL;
	const char* pszName = NULL;
	SG_treenode_entry_type ptneType;
	SG_vhash* pvhResults = NULL;
	SG_vhash* pvhChild = NULL;
	SG_varray* pvaChildren = NULL;
	SG_treenode* pTreenode = NULL;
	SG_string* pStrPath = NULL;

	SG_ERR_CHECK(  SG_treenode_entry__get_entry_type(pCtx, pTreenodeEntry, &ptneType)  );
	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &content)  );
	SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvhResults)  );
	SG_ERR_CHECK(  SG_VARRAY__ALLOC(pCtx, &pvaChildren)  );

	SG_ERR_CHECK(  SG_treenode_entry__get_hid_blob(pCtx, pTreenodeEntry, &pszHid)  );
	SG_ERR_CHECK(  SG_treenode_entry__get_entry_name(pCtx, pTreenodeEntry, &pszName)  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhResults, "name", pszName) );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhResults, "changeset_id", pszChangeSetID) );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhResults, "path", pszPath) );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhResults, "gid", pszGid)  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhResults, "type", sg_jsglue__getObjectType(ptneType))  );
	SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhResults, "hid", pszHid)  );


	if (ptneType == SG_TREENODEENTRY_TYPE_DIRECTORY)
	{
		SG_ERR_CHECK(  SG_treenode__load_from_repo(pCtx, pRepo, pszHid, &pTreenode)  );

		SG_ERR_CHECK(  SG_treenode__count(pCtx, pTreenode, &count)  );

		for (k=0; k < count; k++)
		{
			const char* szGidObject = NULL;
			const char* pszCurHid = NULL;
			const char* pszName = NULL;
			SG_treenode_entry_type ptneCurType;
			const SG_treenode_entry* ptneChild;

			SG_VHASH__ALLOC(pCtx, &pvhChild);
			SG_ERR_CHECK(  SG_treenode__get_nth_treenode_entry__ref(pCtx, pTreenode, k, &szGidObject, &ptneChild)  );
			SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhChild, "gid", szGidObject)  );

			SG_treenode_entry__get_hid_blob(pCtx, ptneChild, &pszCurHid);
			SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhChild, "hid", pszCurHid)  );

			SG_treenode_entry__get_entry_type(pCtx, ptneChild, &ptneCurType);
			SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhChild, "type", sg_jsglue__getObjectType(ptneCurType))  );

			SG_treenode_entry__get_entry_name(pCtx, ptneChild, &pszName);
			SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhChild, "name", pszName)  );

			SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &pStrPath, pszPath)  );
			SG_ERR_CHECK(  SG_repopath__append_entryname(pCtx, pStrPath, pszName, SG_FALSE) );
			SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhChild, "path", SG_string__sz(pStrPath))  );

			SG_varray__append__vhash(pCtx, pvaChildren, &pvhChild);

			SG_VHASH_NULLFREE(pCtx, pvhChild);
			SG_STRING_NULLFREE(pCtx, pStrPath);
		}
	}
	SG_ERR_CHECK(  SG_vhash__add__varray(pCtx, pvhResults, "contents", &pvaChildren) );
	pvaChildren = NULL;
    *ppvhResults = pvhResults;
	pvhResults = NULL;
fail:
	SG_VHASH_NULLFREE(pCtx, pvhResults);
	SG_STRING_NULLFREE(pCtx, content);
	SG_TREENODE_NULLFREE(pCtx, pTreenode);
    SG_VARRAY_NULLFREE(pCtx, pvaChildren);
	SG_VHASH_NULLFREE(pCtx, pvhChild);
	SG_STRING_NULLFREE(pCtx, pStrPath);

}


static void sg_jsglue__generic_file_response__context__alloc(
	SG_context* pCtx,
	SG_file* pFileFragball,
	SG_pathname* pPathFile,
	SG_bool bDeleteWhenDone,
	SG_generic_file_response_context** ppState)
{
	SG_generic_file_response_context* pState = NULL;

	SG_ERR_CHECK(  SG_alloc1(pCtx, pState)  );
	pState->pFileFragball = pFileFragball;
	pState->pPathFile = pPathFile;
	pState->bDeleteWhenDone = bDeleteWhenDone;

	SG_RETURN_AND_NULL(pState, ppState);
fail:
	SG_NULLFREE(pCtx, pState);
}

static void sg_jsglue__generic_file_response__context__free(SG_context* pCtx, SG_generic_file_response_context* pState)
{
	if (pState)
	{
		if (pState->bDeleteWhenDone && pState->pPathFile)
		{
			SG_ERR_IGNORE(  SG_file__close(pCtx, &pState->pFileFragball)  );
			SG_ERR_IGNORE(  SG_fsobj__remove__pathname(pCtx, pState->pPathFile)  );
		}

		SG_FILE_NULLCLOSE(pCtx, pState->pFileFragball);
		SG_PATHNAME_NULLFREE(pCtx, pState->pPathFile);
		SG_NULLFREE(pCtx, pState);
	}
}
#define _GENERIC_FILE_RESPONSE__CONTEXT__NULLFREE(pCtx,p) SG_STATEMENT(SG_context__push_level(pCtx); sg_jsglue__generic_file_response__context__free(pCtx,p); SG_context__pop_level(pCtx); p=NULL;)
/* ---------------------------------------------------------------- */
/* JSNative method definitions */
/* ---------------------------------------------------------------- */

SG_JSGLUE_METHOD_PROTOTYPE(fs, length)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    const char* psz_arg = NULL;
    SG_pathname* pPath = NULL;
    SG_uint64 i;
    SG_fsobj_type t;
    jsval jv;

    SG_UNUSED(obj);

    SG_JS_BOOL_CHECK(  1 == argc  );
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    psz_arg = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));

    SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPath, psz_arg)  );

    SG_ERR_CHECK(  SG_fsobj__length__pathname(pCtx, pPath, &i, &t)  );

    SG_ERR_CHECK(  sg_jsglue__int64_to_jsval(pCtx, cx, i, &jv)  );

    /* TODO verify t */

    SG_PATHNAME_NULLFREE(pCtx, pPath);

    *rval = jv;

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    SG_PATHNAME_NULLFREE(pCtx, pPath);
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(fs, mkdir)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    const char* psz_arg = NULL;
    SG_pathname* pPath = NULL;
    SG_bool b = SG_FALSE;

    SG_UNUSED(obj);

    SG_JS_BOOL_CHECK(  1 == argc  );
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    psz_arg = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));

    SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPath, psz_arg)  );

	//TODO:  This currently throws an error (stopping javascript execution)
    //if there's something there.  Should it just return false?

    SG_ERR_CHECK(  SG_fsobj__mkdir__pathname(pCtx, pPath)  );
	b = SG_TRUE;
    SG_PATHNAME_NULLFREE(pCtx, pPath);

    *rval = BOOLEAN_TO_JSVAL(b);

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    SG_PATHNAME_NULLFREE(pCtx, pPath);
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(fs, rmdir)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    const char* psz_arg = NULL;
    SG_pathname* pPath = NULL;
    SG_bool b = SG_FALSE;

    SG_UNUSED(obj);

    SG_JS_BOOL_CHECK(  1 == argc  );
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    psz_arg = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));

    SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPath, psz_arg)  );

	//TODO:  This currently throws an error (stopping javascript execution)
    //if there is not something there.  Should it just return false?

    SG_ERR_CHECK(  SG_fsobj__rmdir__pathname(pCtx, pPath)  );

	b = SG_TRUE;
    SG_PATHNAME_NULLFREE(pCtx, pPath);

    *rval = BOOLEAN_TO_JSVAL(b);

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    SG_PATHNAME_NULLFREE(pCtx, pPath);
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(fs, mkdir_recursive)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    const char* psz_arg = NULL;
    SG_pathname* pPath = NULL;
    SG_bool b = SG_FALSE;

    SG_UNUSED(obj);

    SG_JS_BOOL_CHECK(  1 == argc  );
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    psz_arg = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));

    SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPath, psz_arg)  );

    //TODO:  This currently throws an error (stopping javascript execution)
    //if there's something there.  Should it just return false?

    SG_ERR_CHECK(  SG_fsobj__mkdir_recursive__pathname(pCtx, pPath)  );

    b = SG_TRUE;
    SG_PATHNAME_NULLFREE(pCtx, pPath);

    *rval = BOOLEAN_TO_JSVAL(b);

    return JS_TRUE;

fail:
    SG_jsglue__report_sg_error(pCtx, cx);
    SG_PATHNAME_NULLFREE(pCtx, pPath);
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(fs, rmdir_recursive)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    const char* psz_arg = NULL;
    SG_pathname* pPath = NULL;
    SG_bool b = SG_FALSE;

    SG_UNUSED(obj);

    SG_JS_BOOL_CHECK(  1 == argc  );
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    psz_arg = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));

    SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPath, psz_arg)  );
    //TODO:  This currently throws an error (stopping javascript execution)
    //if there is not something there.  Should it just return false?
    SG_ERR_CHECK(  SG_fsobj__rmdir_recursive__pathname(pCtx, pPath)  );
    b = SG_TRUE;
    SG_PATHNAME_NULLFREE(pCtx, pPath);

    *rval = BOOLEAN_TO_JSVAL(b);

    return JS_TRUE;

fail:
    SG_jsglue__report_sg_error(pCtx, cx);
    SG_PATHNAME_NULLFREE(pCtx, pPath);
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(fs, exists)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    const char* psz_arg = NULL;
    SG_pathname* pPath = NULL;
    SG_bool b = SG_FALSE;

    SG_UNUSED(obj);

    SG_JS_BOOL_CHECK(  1 == argc  );
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    psz_arg = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));
    if (strlen(psz_arg) != 0)
		SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPath, psz_arg)  );

    SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pPath, &b, NULL, NULL)  );

    SG_PATHNAME_NULLFREE(pCtx, pPath);

    *rval = BOOLEAN_TO_JSVAL(b);

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    SG_PATHNAME_NULLFREE(pCtx, pPath);
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(fs, remove)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    const char* psz_arg = NULL;
    SG_pathname* pPath = NULL;

    SG_UNUSED(obj);

    SG_JS_BOOL_CHECK(  1 == argc  );
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    psz_arg = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));

    SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPath, psz_arg)  );

    SG_ERR_CHECK(  SG_fsobj__remove__pathname(pCtx, pPath)  );

    SG_PATHNAME_NULLFREE(pCtx, pPath);

    *rval = JSVAL_VOID;

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    SG_PATHNAME_NULLFREE(pCtx, pPath);
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(fs, move)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    const char* psz_arg1 = NULL;
    const char* psz_arg2 = NULL;
    SG_pathname* pPath = NULL;
    SG_pathname* pNewPath = NULL;

    SG_UNUSED(obj);

    SG_JS_BOOL_CHECK(  2 == argc  );
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  );
    psz_arg1 = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));
    psz_arg2 = JS_GetStringBytes(JSVAL_TO_STRING(argv[1]));

    SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPath, psz_arg1)  );
    SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pNewPath, psz_arg2)  );

    SG_ERR_CHECK(  SG_fsobj__move__pathname_pathname(pCtx, pPath, pNewPath)  );

    SG_PATHNAME_NULLFREE(pCtx, pPath);
    SG_PATHNAME_NULLFREE(pCtx, pNewPath);

    *rval = JSVAL_TRUE;

    return JS_TRUE;

fail:
    *rval = JSVAL_FALSE;
    SG_jsglue__report_sg_error(pCtx, cx);
    SG_PATHNAME_NULLFREE(pCtx, pPath);
    SG_PATHNAME_NULLFREE(pCtx, pNewPath);
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(fs, getcwd)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_string* pPath = NULL;
    JSString* pjs = NULL;

    SG_UNUSED(argv);
    SG_UNUSED(obj);

    SG_JS_BOOL_CHECK(  0 == argc  );

    SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pPath)  );
    SG_ERR_CHECK(  SG_fsobj__getcwd(pCtx, pPath)  );

    SG_JS_NULL_CHECK(  (pjs = JS_NewStringCopyZ(cx, SG_string__sz(pPath)))  );
    SG_STRING_NULLFREE(pCtx, pPath);

    *rval = STRING_TO_JSVAL(pjs);

    return JS_TRUE;

fail:
    SG_jsglue__report_sg_error(pCtx, cx);
    SG_STRING_NULLFREE(pCtx, pPath);
    *rval = JSVAL_VOID;
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(fs, getcwd_top)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_pathname* pPathCwd = NULL;
    SG_pathname* pPathRepoTop = NULL;
    SG_string* pstrRepo = NULL;
    char* pszgid = NULL;
    JSString* pjs = NULL;

    SG_UNUSED(argv);
    SG_UNUSED(obj);

    SG_JS_BOOL_CHECK(  0 == argc  );
   
    SG_ERR_CHECK(  SG_PATHNAME__ALLOC(pCtx, &pPathCwd)  );    
    SG_ERR_CHECK(  SG_pathname__set__from_cwd(pCtx, pPathCwd)  );  	
    SG_ERR_CHECK(  SG_workingdir__find_mapping(pCtx, pPathCwd, &pPathRepoTop, &pstrRepo, &pszgid));
	
    SG_PATHNAME_NULLFREE(pCtx, pPathCwd);    
    SG_NULLFREE(pCtx, pszgid);  
    SG_STRING_NULLFREE(pCtx, pstrRepo);

    SG_JS_NULL_CHECK(  (pjs = JS_NewStringCopyZ(cx, SG_pathname__sz(pPathRepoTop)))  );
  
    SG_PATHNAME_NULLFREE(pCtx, pPathRepoTop);
    *rval = STRING_TO_JSVAL(pjs);

    return JS_TRUE;

fail:
    SG_jsglue__report_sg_error(pCtx, cx);
    SG_STRING_NULLFREE(pCtx, pstrRepo);
    SG_PATHNAME_NULLFREE(pCtx, pPathCwd);
    SG_PATHNAME_NULLFREE(pCtx, pPathRepoTop);
    SG_NULLFREE(pCtx, pszgid); 
    *rval = JSVAL_VOID;
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(fs, tmpdir)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    JSString* pjs = NULL;
    SG_pathname *tmppath = NULL;

    SG_UNUSED(argv);
    SG_UNUSED(obj);

    SG_JS_BOOL_CHECK(  0 == argc  );

    SG_ERR_CHECK(  SG_PATHNAME__ALLOC__USER_TEMP_DIRECTORY(pCtx, &tmppath)  );

    SG_JS_NULL_CHECK(  (pjs = JS_NewStringCopyZ(cx, SG_pathname__sz(tmppath)))  );
    SG_PATHNAME_NULLFREE(pCtx, tmppath);

    *rval = STRING_TO_JSVAL(pjs);

    return JS_TRUE;

fail:
    SG_jsglue__report_sg_error(pCtx, cx);
    SG_PATHNAME_NULLFREE(pCtx, tmppath);
    *rval = JSVAL_VOID;
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(fs, cd)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_pathname* pPath = NULL;
    JSString* pjs = NULL;
    const char* psz_arg = NULL;

    SG_UNUSED(obj);

    SG_JS_BOOL_CHECK(  1 == argc  );
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );

	pjs = JSVAL_TO_STRING(argv[0]);

    psz_arg = JS_GetStringBytes(pjs);

    SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPath, psz_arg)  );
    SG_ERR_CHECK(  SG_fsobj__cd__pathname(pCtx, pPath)   );
    SG_PATHNAME_NULLFREE(pCtx, pPath);

    *rval = STRING_TO_JSVAL(pjs);

    return JS_TRUE;

fail:
    SG_jsglue__report_sg_error(pCtx, cx);
    SG_PATHNAME_NULLFREE(pCtx, pPath);
    *rval = JSVAL_VOID;
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(fs, chmod)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    const char* psz_arg = NULL;
    SG_pathname* pPath = NULL;
    SG_bool b = SG_FALSE;
    SG_fsobj_perms perms = 0;

    SG_UNUSED(obj);

    SG_JS_BOOL_CHECK(  2 == argc  );
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    SG_JS_BOOL_CHECK(  JSVAL_IS_INT(argv[1])  );
    psz_arg = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));
    perms = JSVAL_TO_INT(argv[1]);

    SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPath, psz_arg)  );

    //TODO:  This currently throws an error (stopping javascript execution)
    //if there's something there.  Should it just return false?

    SG_ERR_CHECK(  SG_fsobj__chmod__pathname(pCtx, pPath, perms )  );

    b = SG_TRUE;
    SG_PATHNAME_NULLFREE(pCtx, pPath);

    *rval = BOOLEAN_TO_JSVAL(b);

    return JS_TRUE;

fail:
    SG_jsglue__report_sg_error(pCtx, cx);
    SG_PATHNAME_NULLFREE(pCtx, pPath);
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(fs, stat)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    const char* psz_arg = NULL;
    SG_pathname* pPath = NULL;
    SG_fsobj_stat FsObjStat;
    int iIsFile;
    int iIsDir;
    int iIsSymlink;
    jsval perms;
    jsval modtime;
    jsval size;
    jsval dateval;
    jsval isFile;
    jsval isDirectory;
    jsval isSymlink;
    JSObject* object = NULL;
    JSObject* dateObject = NULL;
    jsval privaterval;

    SG_UNUSED(obj);

    SG_JS_BOOL_CHECK(  1 == argc  );
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    psz_arg = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));

    SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPath, psz_arg)  );

    SG_ERR_CHECK(  SG_fsobj__stat__pathname(pCtx, pPath, &FsObjStat)  );

    SG_JS_NULL_CHECK(  object = JS_NewObject(cx, &fsobj_properties_class, NULL, NULL)  );
    *rval = OBJECT_TO_JSVAL(object);

    //Permissions are in an int.
    perms = INT_TO_JSVAL(FsObjStat.perms);
    SG_JS_BOOL_CHECK(  JS_SetProperty(cx, object, "permissions", &perms)  );

    //Create a Javascript Date object for the modtime.
    SG_JS_NULL_CHECK(  dateObject = JS_NewObject(cx, &js_DateClass, NULL, NULL)  );
    dateval = OBJECT_TO_JSVAL(dateObject);
    SG_ERR_CHECK(  sg_jsglue__int64_to_jsval(pCtx, cx, FsObjStat.mtime_ms, &modtime)   );
    SG_JS_BOOL_CHECK( JS_CallFunctionName(cx, dateObject, "setTime", 1, &modtime, &privaterval) );
	SG_JS_BOOL_CHECK(  JS_SetProperty(cx, object, "modtime", &dateval )  );

	SG_ERR_CHECK(  sg_jsglue__int64_to_jsval(pCtx, cx, FsObjStat.size, &size)   );
	SG_JS_BOOL_CHECK(  JS_SetProperty(cx, object, "size", &size)  );

	iIsFile = FsObjStat.type == SG_FSOBJ_TYPE__REGULAR;
	iIsDir = FsObjStat.type == SG_FSOBJ_TYPE__DIRECTORY;
	iIsSymlink = FsObjStat.type == SG_FSOBJ_TYPE__SYMLINK;
	isFile = BOOLEAN_TO_JSVAL(iIsFile);
	isDirectory = BOOLEAN_TO_JSVAL(iIsDir);
	isSymlink = BOOLEAN_TO_JSVAL(iIsSymlink);

	SG_JS_BOOL_CHECK(  JS_SetProperty(cx, object, "isFile", &isFile)  );
	SG_JS_BOOL_CHECK(  JS_SetProperty(cx, object, "isDirectory", &isDirectory)  );
	SG_JS_BOOL_CHECK(  JS_SetProperty(cx, object, "isSymlink", &isSymlink)  );

    SG_PATHNAME_NULLFREE(pCtx, pPath);

    return JS_TRUE;

fail:
    SG_jsglue__report_sg_error(pCtx, cx);
    SG_PATHNAME_NULLFREE(pCtx, pPath);
    return JS_FALSE;
}

typedef struct
{
	JSObject *listObject;
	JSContext *cx;
} _readdirEachContext;

static void _readdirEach(
	SG_context *      pCtx,             //< [in] [out] Error and context info.
	const SG_string * pStringEntryName, //< [in] Name of the current entry.
	SG_fsobj_stat *   pfsStat,          //< [in] Stat data for the current entry (only if SG_DIR__FOREACH__STAT is used).
	void *            pVoidData)        //< [in] Data provided by the caller of SG_dir__foreach.
{
	_readdirEachContext *ctx = (_readdirEachContext *)pVoidData;
	JSObject *listObject = ctx->listObject;
	JSContext *cx = ctx->cx;
	JSObject *object = NULL;
    int iIsFile;
    int iIsDir;
    int iIsSymlink;
    jsval perms;
    jsval modtime;
    jsval size;
    jsval dateval;
    jsval isFile;
    jsval isDirectory;
	jsval isSymlink;
	JSObject* dateObject = NULL;
	jsuint length;
    jsval privaterval;
	jsval v;
	jsval name;
	JSString *pjstr;

    SG_JS_NULL_CHECK(  object = JS_NewObject(cx, &fsobj_properties_class, NULL, NULL)  );
	v = OBJECT_TO_JSVAL(object);

	// Root new object now to avoid GC.
	SG_JS_BOOL_CHECK(  JS_GetArrayLength(cx, listObject, &length)  );
	SG_JS_BOOL_CHECK(  JS_SetElement(cx, listObject, length, &v)  );

    //Permissions are in an int.
    perms = INT_TO_JSVAL(pfsStat->perms);
    SG_JS_BOOL_CHECK(  JS_SetProperty(cx, object, "permissions", &perms)  );

    //Create a Javascript Date object for the modtime.
    SG_JS_NULL_CHECK(  dateObject = JS_NewObject(cx, &js_DateClass, NULL, NULL)  );
    dateval = OBJECT_TO_JSVAL(dateObject);
    SG_JS_BOOL_CHECK(  JS_SetProperty(cx, object, "modtime", &dateval )  ); // Root now to avoid GC.
    SG_ERR_CHECK(  sg_jsglue__int64_to_jsval(pCtx, cx, pfsStat->mtime_ms, &modtime)   );
    SG_JS_BOOL_CHECK( JS_CallFunctionName(cx, dateObject, "setTime", 1, &modtime, &privaterval) );

	SG_ERR_CHECK(  sg_jsglue__int64_to_jsval(pCtx, cx, pfsStat->size, &size)   );
	SG_JS_BOOL_CHECK(  JS_SetProperty(cx, object, "size", &size)  );

	iIsFile = pfsStat->type == SG_FSOBJ_TYPE__REGULAR;
	iIsDir = pfsStat->type == SG_FSOBJ_TYPE__DIRECTORY;
	iIsSymlink = pfsStat->type == SG_FSOBJ_TYPE__SYMLINK;
	isFile = BOOLEAN_TO_JSVAL(iIsFile);
	isDirectory = BOOLEAN_TO_JSVAL(iIsDir);
	isSymlink = BOOLEAN_TO_JSVAL(iIsSymlink);

    SG_JS_NULL_CHECK(  (pjstr = JS_NewStringCopyZ(cx, SG_string__sz(pStringEntryName)))  );

	name = STRING_TO_JSVAL(pjstr);

	SG_JS_BOOL_CHECK(  JS_SetProperty(cx, object, "isFile", &isFile)  );
	SG_JS_BOOL_CHECK(  JS_SetProperty(cx, object, "isDirectory", &isDirectory)  );
	SG_JS_BOOL_CHECK(  JS_SetProperty(cx, object, "isSymlink", &isSymlink)  );
	SG_JS_BOOL_CHECK(  JS_SetProperty(cx, object, "name", &name)  );

 fail:
	;
}


SG_JSGLUE_METHOD_PROTOTYPE(fs, readdir)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    const char* psz_arg = NULL;
    SG_pathname* pPath = NULL;
	JSObject* listObject = NULL;
	_readdirEachContext ctx;

    SG_UNUSED(obj);

    SG_JS_BOOL_CHECK(  argc == 1 );
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    psz_arg = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPath, psz_arg)  );

	SG_JS_NULL_CHECK(  listObject = JS_NewArrayObject(cx, 0, NULL)  );
	// immediately cause the new object to be referenced
	// so that the GC doesn't kill it while we populate
	// the fields within it.  (yes, this is subtle.)
    *rval = OBJECT_TO_JSVAL(listObject);

	ctx.listObject = listObject;
	ctx.cx = cx;

	SG_ERR_CHECK(  SG_dir__foreach(pCtx,
								   pPath,
								   SG_DIR__FOREACH__STAT
								   | SG_DIR__FOREACH__SKIP_SELF
								   | SG_DIR__FOREACH__SKIP_PARENT
								   | SG_DIR__FOREACH__SKIP_SGDRAWER,
								   _readdirEach,
								   &ctx)  );

    SG_PATHNAME_NULLFREE(pCtx, pPath);

    return JS_TRUE;

fail:
    SG_jsglue__report_sg_error(pCtx, cx);
    SG_PATHNAME_NULLFREE(pCtx, pPath);
    return JS_FALSE;
}

///sg.fs.fetch_file(path, chunksize, deletewhendone) 
SG_JSGLUE_METHOD_PROTOTYPE(fs, fetch_file)
{
    SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_safeptr* psp_ffh = NULL;
    JSObject* jso = NULL;

    SG_int_to_string_buffer buf_uint64;
    SG_generic_file_response_context* pResponseCtx = NULL;
    SG_uint64 length = 0;
    SG_file* pFile = NULL;
    const char* pszPathFile = NULL;
    SG_pathname* pPathFile = NULL;
    SG_bool bDeleteOnSuccessfulFinish;
    jsval jv;
        
    SG_UNUSED(obj);

    SG_JS_BOOL_CHECK(  argc == 3 );
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    SG_JS_BOOL_CHECK(  JSVAL_IS_INT(argv[1])  );
    SG_JS_BOOL_CHECK(  JSVAL_IS_BOOLEAN(argv[2])  );

    pszPathFile = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));
    bDeleteOnSuccessfulFinish = JSVAL_TO_BOOLEAN(argv[2]);

    SG_ERR_CHECK(  SG_pathname__alloc__sz(pCtx, &pPathFile, pszPathFile)  );

    SG_ERR_CHECK(  SG_file__open__pathname(pCtx, pPathFile, SG_FILE_RDONLY | SG_FILE_OPEN_EXISTING, SG_FSOBJ_PERMS__UNUSED, &pFile)  );
    SG_ERR_CHECK(  SG_file__seek_end(pCtx, pFile, &length)  );

    SG_ERR_CHECK(  sg_jsglue__generic_file_response__context__alloc(pCtx, pFile, pPathFile, bDeleteOnSuccessfulFinish, &pResponseCtx)  );

    SG_JS_NULL_CHECK(  (jso = JS_NewObject(cx, &sg_fetchfilehandle_class, NULL, NULL))  );  
    *rval = OBJECT_TO_JSVAL(jso);
    SG_ERR_CHECK(  SG_safeptr__wrap__fetchfilehandle(pCtx, pResponseCtx, &psp_ffh)  );
    JS_SetPrivate(cx, jso, psp_ffh);
        
    jv = INT_TO_JSVAL(argv[1]);
    SG_JS_BOOL_CHECK(  JS_SetProperty(cx, jso, "chunk_length", &jv)  );

    JSVAL_FROM_SZ(jv, "0");
    SG_JS_BOOL_CHECK(  JS_SetProperty(cx, jso, "processed_length", &jv)  );

    JSVAL_FROM_SZ(jv, SG_uint64_to_sz(length, buf_uint64));
    SG_JS_BOOL_CHECK(  JS_SetProperty(cx, jso, "total_length", &jv)  );

    pFile = NULL;
    pPathFile = NULL;
    pResponseCtx = NULL;
    
    return JS_TRUE;
 fail:
    SG_jsglue__report_sg_error(pCtx, cx);     
    SG_FILE_NULLCLOSE(pCtx, pFile);
    _GENERIC_FILE_RESPONSE__CONTEXT__NULLFREE(pCtx, pResponseCtx);
    SG_SAFEPTR_NULLFREE(pCtx, psp_ffh); 
    SG_PATHNAME_NULLFREE(pCtx, pPathFile);
    return JS_FALSE;
}



/**
 * fetchfilehandle.next_chunk() returns a cbuffer.
 * A cbuffer is essentially an object that is meant to be returned back down to
 * (or passed back up into) C code, without having to be converted back and
 * forth between a JavaScript native type/object. See sg_cbuffer_typedefs.h.
 * 
 * If *don't* pass the returned cbuffer back to C code, you can free it
 * directly from javascript by calling its ".fin()" member function.
 */
SG_JSGLUE_METHOD_PROTOTYPE(fetchfilehandle, next_chunk)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_safeptr* psp_ffh = NULL;
    SG_generic_file_response_context* pContext = NULL;
	
	SG_safeptr* psp_cbuffer = NULL;
	SG_cbuffer * pCbuffer = NULL;
	
	jsval val;   
    jsval jv;
	SG_uint32 chunk_length = 0;
	JSObject * jso = NULL;
    SG_uint64 processed_length = 0;
    SG_uint64 total_length = 0;
    const char * szVal = NULL;
    SG_int_to_string_buffer buf_uint64;
    
    SG_UNUSED(argc);
    SG_UNUSED(argv);

	psp_ffh = sg_jsglue__get_object_private(cx, obj);
	SG_safeptr__unwrap__fetchfilehandle(pCtx, psp_ffh, &pContext);
	if(SG_context__err_equals(pCtx, SG_ERR_SAFEPTR_NULL))
	{
		SG_context__err_reset(pCtx);
		*rval = JSVAL_NULL; // Return null after the last chunk has been sent.
		return JS_TRUE;
	}
	else
		SG_ERR_CHECK_CURRENT;    	

	SG_JS_BOOL_CHECK(  JS_GetProperty(cx, obj, "chunk_length", &val)  );
	chunk_length = JSVAL_TO_INT(val);

    SG_JS_BOOL_CHECK(  JS_GetProperty(cx, obj, "total_length", &val)  );
    szVal = JS_GetStringBytes(JSVAL_TO_STRING(val));
    SG_ERR_CHECK(  SG_uint64__parse__strict(pCtx, &total_length, szVal)  );

    SG_JS_BOOL_CHECK(  JS_GetProperty(cx, obj, "processed_length", &val)  );
    szVal = JS_GetStringBytes(JSVAL_TO_STRING(val));
    SG_ERR_CHECK(  SG_uint64__parse__strict(pCtx, &processed_length, szVal)  );

	if(total_length-processed_length<(SG_uint64)chunk_length)
		chunk_length = (SG_uint32)(total_length-processed_length);
	SG_ERR_CHECK(  SG_cbuffer__alloc__new(pCtx, &pCbuffer, chunk_length)  );

    SG_ERR_CHECK(  SG_file__seek(pCtx, pContext->pFileFragball, processed_length)  );

	SG_ERR_CHECK(  SG_file__read(pCtx, pContext->pFileFragball, pCbuffer->len, pCbuffer->pBuf, NULL)  );
	processed_length+=chunk_length;
	if (processed_length >= total_length)
	{
		SG_SAFEPTR_NULLFREE(pCtx, psp_ffh);
		_GENERIC_FILE_RESPONSE__CONTEXT__NULLFREE(pCtx, pContext);
		JS_SetPrivate(cx, obj, NULL);
	}
	JSVAL_FROM_SZ(jv, SG_uint64_to_sz(processed_length, buf_uint64));
	SG_JS_BOOL_CHECK(  JS_SetProperty(cx, obj, "processed_length", &jv)  );

	SG_JS_NULL_CHECK(  jso = JS_NewObject(cx, &sg_cbuffer_class, NULL, NULL)  );
	*rval = OBJECT_TO_JSVAL(jso);
	SG_ERR_CHECK(  SG_safeptr__wrap__cbuffer(pCtx, pCbuffer, &psp_cbuffer)  );
	JS_SetPrivate(cx, jso, psp_cbuffer);       

	return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx); // DO NOT SG_ERR_IGNORE() THIS
	SG_SAFEPTR_NULLFREE(pCtx, psp_cbuffer);
     _GENERIC_FILE_RESPONSE__CONTEXT__NULLFREE(pCtx, pContext);
	SG_cbuffer__nullfree(&pCbuffer, cx);
	return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(fetchfilehandle, abort)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_safeptr* psp_ffh = NULL;
	SG_generic_file_response_context * pHandle = NULL;

	SG_UNUSED(argv);
    SG_UNUSED(argc);
	
	psp_ffh = sg_jsglue__get_object_private(cx, obj);
	SG_safeptr__unwrap__fetchfilehandle(pCtx, psp_ffh, &pHandle);
	if(SG_context__err_equals(pCtx, SG_ERR_SAFEPTR_NULL))
	{
		SG_context__err_reset(pCtx);
		*rval = JSVAL_NULL; // We seem to have already aborted or finished.
		return JS_TRUE;
	}
	else
		SG_ERR_CHECK_CURRENT;

	_GENERIC_FILE_RESPONSE__CONTEXT__NULLFREE(pCtx, pHandle);

	SG_SAFEPTR_NULLFREE(pCtx, psp_ffh);
	JS_SetPrivate(cx, obj, NULL);

	*rval = JSVAL_VOID;
	return JS_TRUE;

fail:
    _GENERIC_FILE_RESPONSE__CONTEXT__NULLFREE(pCtx, pHandle);
    SG_SAFEPTR_NULLFREE(pCtx, psp_ffh);
	SG_jsglue__report_sg_error(pCtx,cx); // DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}


extern SG_bool _sg_uridispatch__debug_remote_shutdown;

SG_JSGLUE_METHOD_PROTOTYPE(server, debug_shutdown)
{
	SG_UNUSED(cx);
	SG_UNUSED(obj);
	SG_UNUSED(argc);
	SG_UNUSED(argv);
	SG_UNUSED(rval);
	
	_sg_uridispatch__debug_remote_shutdown = SG_TRUE;
	return JS_TRUE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, from_json)
{
    SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    const char* psz_arg = NULL;
    SG_vhash* pvh = NULL;
    SG_varray* pva = NULL;
    JSObject* jso = NULL;

    SG_UNUSED(obj);

    SG_JS_BOOL_CHECK(  1 == argc  );
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    psz_arg = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));

    SG_ERR_CHECK(  SG_veither__parse_json__buflen(pCtx, psz_arg, SG_STRLEN(psz_arg), &pvh, &pva)  );

    SG_ASSERT(pva || pvh);

    if (pva)
    {
        SG_JS_NULL_CHECK(  (jso = JS_NewArrayObject(cx, 0, NULL))  );
        *rval = OBJECT_TO_JSVAL(jso);
        SG_ERR_CHECK(  sg_jsglue__copy_varray_into_jsobject(pCtx, cx, pva, jso)  );
        SG_VARRAY_NULLFREE(pCtx, pva);
    }
    else if (pvh)
    {
        SG_JS_NULL_CHECK(  (jso = JS_NewObject(cx, NULL, NULL, NULL))  );
        *rval = OBJECT_TO_JSVAL(jso);
        SG_ERR_CHECK(  sg_jsglue__copy_vhash_into_jsobject(pCtx, cx, pvh, jso)  );
        SG_VHASH_NULLFREE(pCtx, pvh);
    }

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    SG_VHASH_NULLFREE(pCtx, pvh);
    SG_VARRAY_NULLFREE(pCtx, pva);
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, hash)
{
    SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_varray *strlist = NULL;
    char leafhash[SG_HID_MAX_BUFFER_LENGTH];
    SGHASH_handle *hashHandle = NULL;
    SG_uint32 i, count;
    JSString* pjs = NULL;
    JSObject* jso = NULL;
	const char *hashalgo = "SHA1/160";
	char * psz_foreign = NULL;
	char * hashalgo_foreign = NULL;

    SG_UNUSED(obj);

    SG_JS_BOOL_CHECK(  (argc >= 1) && (argc <= 2)  );
	
	if (JSVAL_IS_NULL(argv[0]))
	{
		// TODO how should null be hashed
		 SG_ERR_THROW(SG_ERR_NOTIMPLEMENTED);
	}
	else if(JSVAL_IS_OBJECT(argv[0]))
    {
        jso = JSVAL_TO_OBJECT(argv[0]);

        if (JS_IsArrayObject(cx, jso))
        {
            SG_ERR_CHECK(  sg_jsglue__jsobject_to_varray(pCtx, cx, jso, &strlist)  );
        }
        else
        {
            SG_ERR_THROW(SG_ERR_NOTIMPLEMENTED);//todo: numbers, bool, null
        }
    }
    else if(JSVAL_IS_STRING(argv[0]))
    {
        pjs = JSVAL_TO_STRING(argv[0]);
		SG_ERR_CHECK(  SG_varray__alloc(pCtx, &strlist)  );
		psz_foreign = SG_ALLOC_FOREIGN( JS_EncodeString(cx, pjs) );
		SG_ERR_CHECK(  SG_varray__append__string__sz(pCtx, strlist, psz_foreign)  );
    }
    else
    {
        SG_ERR_THROW(SG_ERR_NOTIMPLEMENTED);//todo: numbers, bool, null
    }

	if (argc == 2)
	{
		if (JSVAL_IS_STRING(argv[1]))
		{
			pjs = JSVAL_TO_STRING(argv[1]);
			hashalgo_foreign = SG_ALLOC_FOREIGN( JS_EncodeString(cx, pjs) );
			hashalgo = hashalgo_foreign;
		}
		else
		{
			SG_ERR_THROW(SG_ERR_NOTIMPLEMENTED);//todo: numbers, bool, null
		}
	}

    SGHASH_init(hashalgo, &hashHandle);

    SG_ERR_CHECK(  SG_varray__count(pCtx, strlist, &count)  );

    for ( i = 0; i < count; ++i )
    {
	const char *st;
	
	SG_ERR_CHECK(  SG_varray__get__sz(pCtx, strlist, i, &st)  );
	SGHASH_update(hashHandle, (const SG_byte *)st, (SG_uint32)(strlen(st) * sizeof(char)));
    }

    SGHASH_final(&hashHandle, leafhash, sizeof(leafhash));

    SG_JS_NULL_CHECK(  (pjs = JS_NewStringCopyZ(cx, leafhash))  );
    SG_VARRAY_NULLFREE(pCtx, strlist);

	JS_free(cx, SG_FREE_FOREIGN(psz_foreign));
	JS_free(cx, SG_FREE_FOREIGN(hashalgo_foreign));

    *rval = STRING_TO_JSVAL(pjs);
    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    SG_VARRAY_NULLFREE(pCtx, strlist);
	JS_free(cx, SG_FREE_FOREIGN(psz_foreign));
	JS_free(cx, SG_FREE_FOREIGN(hashalgo_foreign));
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, to_json)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_string* pstr = NULL;
    JSString* pjs = NULL;

    SG_UNUSED(obj);

    SG_JS_BOOL_CHECK(  1 == argc  );
	
	SG_ERR_CHECK(  sg_jsglue__jsval_to_json(pCtx, cx, argv[0], SG_FALSE, &pstr)  );

    SG_JS_NULL_CHECK(  (pjs = JS_NewStringCopyZ(cx, SG_string__sz(pstr)))  );
    SG_STRING_NULLFREE(pCtx, pstr);

    *rval = STRING_TO_JSVAL(pjs);
    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    SG_STRING_NULLFREE(pCtx, pstr);
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, to_json__pretty_print)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_string* pstr = NULL;
    JSString* pjs = NULL;

    SG_UNUSED(obj);

    SG_JS_BOOL_CHECK(  1 == argc  );

	SG_ERR_CHECK(  sg_jsglue__jsval_to_json(pCtx, cx, argv[0], SG_TRUE, &pstr)  );

    SG_JS_NULL_CHECK(  (pjs = JS_NewStringCopyZ(cx, SG_string__sz(pstr)))  );
    SG_STRING_NULLFREE(pCtx, pstr);

    *rval = STRING_TO_JSVAL(pjs);
    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    SG_STRING_NULLFREE(pCtx, pstr);
    return JS_FALSE;
}

/**
 * sg.to_json__sz() returns a cbuffer.
 * A cbuffer is essentially an object that is meant to be returned back down to
 * (or passed back up into) C code, without having to be converted back and
 * forth between a JavaScript native type/object. See sg_cbuffer_typedefs.h.
 * 
 * If *don't* pass the returned cbuffer back to C code, you can free it
 * directly from javascript by calling its ".fin()" member function.
 */
SG_JSGLUE_METHOD_PROTOTYPE(sg, to_json__sz)
{
    SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_string* pstr = NULL;
    SG_cbuffer* pCbuffer = NULL;
    SG_safeptr* psp = NULL;
    JSObject* jso = NULL;

    SG_UNUSED(obj);

    SG_JS_BOOL_CHECK(  1 == argc  );
	
	SG_ERR_CHECK(  sg_jsglue__jsval_to_json(pCtx, cx, argv[0], SG_FALSE, &pstr) );

    SG_ERR_CHECK(  SG_cbuffer__alloc__string(pCtx, &pCbuffer, &pstr)  );
	pstr = NULL;	// cbuffer owns it now

    SG_JS_NULL_CHECK(  jso = JS_NewObject(cx, &sg_cbuffer_class, NULL, NULL)  );
    *rval = OBJECT_TO_JSVAL(jso);
    SG_ERR_CHECK(  SG_safeptr__wrap__cbuffer(pCtx, pCbuffer, &psp)  );
    JS_SetPrivate(cx, jso, psp);

    return JS_TRUE;

fail:
    SG_jsglue__report_sg_error(pCtx,cx);    // DO NOT SG_ERR_IGNORE() THIS
    SG_STRING_NULLFREE(pCtx, pstr);
    SG_cbuffer__nullfree(&pCbuffer, cx);
    SG_SAFEPTR_NULLFREE(pCtx, psp);
    return JS_FALSE;
}

/**
 * sg.to_sz() returns a cbuffer.
 * A cbuffer is essentially an object that is meant to be returned back down to
 * (or passed back up into) C code, without having to be converted back and
 * forth between a JavaScript native type/object. See sg_cbuffer_typedefs.h.
 * 
 * If *don't* pass the returned cbuffer back to C code, you can free it
 * directly from javascript by calling its ".fin()" member function.
 */
SG_JSGLUE_METHOD_PROTOTYPE(sg, to_sz)
{
    SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    char * sz = NULL;
    SG_cbuffer* pCbuffer = NULL;
    SG_safeptr* psp = NULL;
    JSObject* jso = NULL;

    SG_UNUSED(obj);

    SG_JS_BOOL_CHECK(argc==1);
    SG_JS_BOOL_CHECK(JSVAL_IS_STRING(argv[0]));

    sz = SG_ALLOC_FOREIGN( JS_EncodeString(cx, JSVAL_TO_STRING(argv[0])) );

    SG_ERR_CHECK(  SG_cbuffer__alloc__js_sz(pCtx, &pCbuffer, &sz)  );

    SG_JS_NULL_CHECK(  jso = JS_NewObject(cx, &sg_cbuffer_class, NULL, NULL)  );
    *rval = OBJECT_TO_JSVAL(jso);
    SG_ERR_CHECK(  SG_safeptr__wrap__cbuffer(pCtx, pCbuffer, &psp)  );
    JS_SetPrivate(cx, jso, psp);

    return JS_TRUE;

fail:
    SG_jsglue__report_sg_error(pCtx,cx);    // DO NOT SG_ERR_IGNORE() THIS
    JS_free(cx, SG_FREE_FOREIGN(sz));
    SG_cbuffer__nullfree(&pCbuffer, cx);
    SG_SAFEPTR_NULLFREE(pCtx, psp);
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, repoid_to_descriptors)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    char * pszRepoId_foreign = NULL;
	JSObject* jso = NULL;
	SG_vhash * pvhList = NULL;

	SG_UNUSED(obj);

    SG_JS_BOOL_CHECK(argc==1);
    SG_JS_BOOL_CHECK(JSVAL_IS_STRING(argv[0]));

    pszRepoId_foreign = SG_ALLOC_FOREIGN( JS_EncodeString(cx, JSVAL_TO_STRING(argv[0])) );

	// TODO 2011/01/27 allow pvec_use_submodules hints to be passed in.
	SG_ERR_CHECK(  SG_closet__reverse_map_repo_id_to_descriptor_names(pCtx, pszRepoId_foreign, NULL, &pvhList)  );

	SG_JS_NULL_CHECK(  (jso = JS_NewObject(cx, NULL, NULL, NULL))  );
    *rval = OBJECT_TO_JSVAL(jso);
    SG_ERR_CHECK(  sg_jsglue__copy_vhash_into_jsobject(pCtx, cx, pvhList, jso)  );

	SG_VHASH_NULLFREE(pCtx, pvhList);
	JS_free(cx, SG_FREE_FOREIGN(pszRepoId_foreign));
	return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	SG_VHASH_NULLFREE(pCtx, pvhList);
	JS_free(cx, SG_FREE_FOREIGN(pszRepoId_foreign));
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, exec)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	const char * pszPathToExec = NULL;
	SG_exec_argvec * pArgVec = NULL;
	SG_exit_status childExitStatus = 0;
	SG_tempfile * pTempStdOut;
	SG_tempfile * pTempStdErr;
	SG_string * pStringErr = NULL;
	SG_string * pStringOut = NULL;
	uint i = 0;
	jsval jsChildExitStatus = JSVAL_VOID;
	jsval jsExecResult = JSVAL_VOID;
	jsval jsstderr = JSVAL_VOID;
	jsval jsstdout = JSVAL_VOID;
	JSObject* object = NULL;
	SG_exec_result execResult = SG_EXEC_RESULT__UNKNOWN;
	SG_process_id pid;

    SG_UNUSED(obj);

	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
	pszPathToExec = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));

	SG_ERR_CHECK(  SG_exec_argvec__alloc(pCtx, &pArgVec)   );
	//allocate the new string array that we will send down.
	for (i = 1; i < argc; i++)
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[i])  );
		SG_ERR_CHECK(  SG_exec_argvec__append__sz(pCtx, pArgVec, JS_GetStringBytes(JSVAL_TO_STRING(argv[i]))) );
	}

	//We need to create temp files to handle the stdout and stderr output.
	SG_ERR_CHECK( SG_tempfile__open_new(pCtx, &pTempStdOut)  );
	SG_ERR_CHECK( SG_tempfile__open_new(pCtx, &pTempStdErr)  );

	// Run the requested command.  If we run into a setup-type error,
	// this will throw as usual.  But if the child has problems, it
	// returns normally (with both EXEC_RESULT and EXIT_STATUS).
	
	SG_ERR_CHECK(  SG_exec__exec_sync__files__details(pCtx,
													  pszPathToExec, pArgVec,
													  NULL, pTempStdOut->pFile, pTempStdErr->pFile,
													  &childExitStatus,
													  &execResult,
													  &pid)  );

	SG_ERR_CHECK( SG_file__fsync(pCtx, pTempStdOut->pFile)   );
	SG_ERR_CHECK( SG_file__fsync(pCtx, pTempStdErr->pFile)   );

	SG_ERR_CHECK( SG_file__seek(pCtx, pTempStdOut->pFile, 0)   );
	SG_ERR_CHECK( SG_file__seek(pCtx, pTempStdErr->pFile, 0)   );

	SG_ERR_CHECK( SG_STRING__ALLOC(pCtx, &pStringErr) );
	SG_ERR_CHECK( SG_STRING__ALLOC(pCtx, &pStringOut) );
	SG_ERR_IGNORE(  SG_file__read_utf8_file_into_string(pCtx, pTempStdErr->pFile, pStringErr)  );	// allow this to fail incase the output is huge
	SG_ERR_IGNORE(  SG_file__read_utf8_file_into_string(pCtx, pTempStdOut->pFile, pStringOut)  );	// allow this to fail incase the output is huge

	SG_ERR_CHECK( SG_tempfile__close_and_delete(pCtx, &pTempStdErr) );
	SG_ERR_CHECK( SG_tempfile__close_and_delete(pCtx, &pTempStdOut) );

	//Bundle a Javascript object to return.

	SG_JS_NULL_CHECK(  (object = JS_NewObject(cx, &fsobj_properties_class, NULL, NULL))  );
	*rval = OBJECT_TO_JSVAL(object);

	jsExecResult = INT_TO_JSVAL(execResult);
	SG_JS_BOOL_CHECK( JS_SetProperty(cx, object, "exec_result", &jsExecResult)  );

	jsChildExitStatus = INT_TO_JSVAL(childExitStatus);
	SG_JS_BOOL_CHECK( JS_SetProperty(cx, object, "exit_status", &jsChildExitStatus) );

	JSVAL_FROM_SZ(jsstderr, SG_string__sz(pStringErr));
	SG_JS_BOOL_CHECK( JS_SetProperty(cx, object, "stderr", &jsstderr) );

	JSVAL_FROM_SZ(jsstdout, SG_string__sz(pStringOut));
	SG_JS_BOOL_CHECK( JS_SetProperty(cx, object, "stdout", &jsstdout) );

#if defined(HAVE_EXEC_DEBUG_STACKTRACE)
	if (execResult & SG_EXEC_RESULT__ABNORMAL)
	{
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   ("Child process crashed with stderr:\n"
									"==================================\n"
									"%s\n"
									"==================================\n"),
								   SG_string__sz(pStringErr))  );

		SG_ERR_IGNORE(  SG_exec_debug__get_stacktrace(pCtx, pszPathToExec, pid)  );
	}
#endif

	SG_STRING_NULLFREE(pCtx, pStringOut);
	SG_STRING_NULLFREE(pCtx, pStringErr);
	SG_EXEC_ARGVEC_NULLFREE(pCtx, pArgVec);
	return JS_TRUE;
fail:
	SG_ERR_IGNORE( SG_tempfile__close_and_delete(pCtx, &pTempStdErr) );
	SG_ERR_IGNORE( SG_tempfile__close_and_delete(pCtx, &pTempStdOut) );
	SG_STRING_NULLFREE(pCtx, pStringOut);
	SG_STRING_NULLFREE(pCtx, pStringErr);
	SG_EXEC_ARGVEC_NULLFREE(pCtx, pArgVec);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, exec_nowait)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	const char * pszPathToExec = NULL;
	SG_exec_argvec * pArgVec = NULL;
	SG_process_id processID = 0;
	SG_pathname * pPathNameForOutput = NULL;
	SG_pathname * pPathNameForError = NULL;
	SG_file * pFileForOutput = NULL;
	SG_file * pFileForError = NULL;
	uint i = 0;
	SG_uint32 firstArg = 0;
	JSObject* jso = NULL;
	SG_varray* pva_options = NULL;
	const char * pszVal = NULL;

    SG_UNUSED(obj);
    SG_UNUSED(rval);

	if (argc > 1)
	{
		jsval arg0 = argv[0];

		jso = JSVAL_TO_OBJECT(arg0);

		if (JS_IsArrayObject(cx, jso))
		{
			 SG_uint32 count = 0;

			 SG_uint32 j = 0;
			 const char* pszOption;

			 SG_ERR_CHECK(  sg_jsglue__jsobject_to_varray(pCtx, cx, jso, &pva_options)  );

			 SG_ERR_CHECK(  SG_varray__count(pCtx, pva_options, &count)  );

			 for (j = 0; j < count; j++)
			 {

				 SG_ERR_CHECK( SG_varray__get__sz(pCtx, pva_options, j, &pszOption) );

				 if ( sg_jsglue_get_equals_flag (pszOption, "output", &pszVal) )
				 {
					SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPathNameForOutput, pszVal)  );
				 }
				 else if ( sg_jsglue_get_equals_flag (pszOption, "error", &pszVal) )
				 {
					 SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPathNameForError, pszVal)  );
				 }
			 }
			 firstArg++;
		}
	}

	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[firstArg])  );
	pszPathToExec = JS_GetStringBytes(JSVAL_TO_STRING(argv[firstArg]));

	SG_ERR_CHECK(  SG_exec_argvec__alloc(pCtx, &pArgVec)   );
	//allocate the new string array that we will send down.
	for (i = firstArg + 1; i < argc; i++)
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[i])  );
		SG_ERR_CHECK(  SG_exec_argvec__append__sz(pCtx, pArgVec, JS_GetStringBytes(JSVAL_TO_STRING(argv[i]))) );
	}

	//We need to create temp files to handle the stdout and stderr output.
	if (pPathNameForOutput != NULL)
	{
		SG_ERR_CHECK( SG_file__open__pathname(pCtx, pPathNameForOutput, SG_FILE_OPEN_OR_CREATE|SG_FILE_WRONLY,  0644, &pFileForOutput)   );
	}
	if (pPathNameForError != NULL)
	{
		SG_ERR_CHECK( SG_file__open__pathname(pCtx, pPathNameForError, SG_FILE_OPEN_OR_CREATE|SG_FILE_WRONLY,  0644, &pFileForError)   );
	}
	SG_ERR_CHECK(   SG_exec__exec_async__files(pCtx, pszPathToExec, pArgVec, NULL, pFileForOutput, pFileForError, &processID)  );
	SG_EXEC_ARGVEC_NULLFREE(pCtx, pArgVec);
	SG_VARRAY_NULLFREE(pCtx, pva_options);
	SG_PATHNAME_NULLFREE(pCtx, pPathNameForOutput);
	SG_PATHNAME_NULLFREE(pCtx, pPathNameForError);
	//It's ok to close these, because the child process has its own open file descriptor
	SG_FILE_NULLCLOSE(pCtx, pFileForError);
	SG_FILE_NULLCLOSE(pCtx, pFileForOutput);
    *rval = INT_TO_JSVAL(processID);
	return JS_TRUE;
fail:
	//It's ok to close these, because the child process has its own open file descriptor
	SG_FILE_NULLCLOSE(pCtx, pFileForError);
	SG_FILE_NULLCLOSE(pCtx, pFileForOutput);
	SG_EXEC_ARGVEC_NULLFREE(pCtx, pArgVec);
	SG_VARRAY_NULLFREE(pCtx, pva_options);
	SG_PATHNAME_NULLFREE(pCtx, pPathNameForOutput);
	SG_PATHNAME_NULLFREE(pCtx, pPathNameForError);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, time)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_int64 t;

    SG_UNUSED(argc);
    SG_UNUSED(obj);
    SG_UNUSED(argv);

    // This function is here partly because JS new Date().getTime() is
    // supposed to do the exact same thing, but on Windows, it kind
    // of doesn't.  See
    //
    // https://bugzilla.mozilla.org/show_bug.cgi?id=363258

    SG_time__get_milliseconds_since_1970_utc(pCtx, &t);

    if (SG_int64__fits_in_double(t))
    {
        double d = (double) t;
        jsval jv;

        SG_JS_BOOL_CHECK(  JS_NewNumberValue(cx, d, &jv)  );
        *rval = jv;
    }
    else
    {
        SG_ERR_THROW(  SG_ERR_INTEGER_OVERFLOW  );
    }

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, sleep_ms)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_UNUSED(obj);
	SG_JS_BOOL_CHECK(argc==1);
	if(JSVAL_IS_INT(argv[0]))
		SG_sleep_ms((SG_int64)JSVAL_TO_INT(argv[0]));
	else if(JSVAL_IS_STRING(argv[0]))
	{
		SG_uint32 i;
		SG_ERR_CHECK(SG_uint32__parse(pCtx, &i, JS_GetStringBytes(JSVAL_TO_STRING(argv[0]))));
		SG_sleep_ms(i);
	}
	else
		SG_ERR_THROW(SG_ERR_INVALIDARG);

	*rval=JSVAL_VOID;
	return JS_TRUE;
fail:
	SG_jsglue__report_sg_error(pCtx, cx); // DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, platform)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);

	SG_UNUSED(argc);
	SG_UNUSED(argv);
	SG_UNUSED(obj);

#if defined(LINUX)
	JSVAL_FROM_SZ(*rval, "LINUX");
	return JS_TRUE;
#elif defined(MAC)
	JSVAL_FROM_SZ(*rval, "MAC");
	return JS_TRUE;
#elif defined(WINDOWS)
	JSVAL_FROM_SZ(*rval, "WINDOWS");
	return JS_TRUE;
#else
	return JS_FALSE;
#endif
fail:
	return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, rand)
{
	SG_context * pCtx = SG_jsglue__get_sg_context(cx);
    SG_uint32 v = 0;

    SG_UNUSED(obj);

	SG_JS_BOOL_CHECK(  2 == argc  );
	SG_JS_BOOL_CHECK(  JSVAL_IS_INT(argv[0])  );
	SG_JS_BOOL_CHECK(  JSVAL_IS_INT(argv[1])  );

    v = SG_random_uint32__2(JSVAL_TO_INT(argv[0]), JSVAL_TO_INT(argv[1]));

    *rval = INT_TO_JSVAL(v);

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, gid)
{
	SG_context * pCtx = SG_jsglue__get_sg_context(cx);
    char buf[SG_GID_BUFFER_LENGTH];

    SG_UNUSED(obj);
    SG_UNUSED(argc);
    SG_UNUSED(argv);

    SG_ERR_CHECK(  SG_gid__generate(pCtx, buf, sizeof(buf))  );

    JSVAL_FROM_SZ(*rval, buf);

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, set_local_setting)
{
	SG_context * pCtx = SG_jsglue__get_sg_context(cx);

	SG_UNUSED(obj);

	SG_JS_BOOL_CHECK(  2 == argc  );
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  );

    SG_ERR_CHECK(  SG_localsettings__update__sz(
                pCtx,
                JS_GetStringBytes(JSVAL_TO_STRING(argv[0])),
                JS_GetStringBytes(JSVAL_TO_STRING(argv[1]))
                ) );

    // TODO do we want to return something here?
	*rval = JSVAL_NULL;

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, set_local_setting__system_wide)
{
	SG_context * pCtx = SG_jsglue__get_sg_context(cx);

	SG_UNUSED(obj);

	SG_JS_BOOL_CHECK(  2 == argc  );
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  );

	SG_ERR_CHECK(  SG_localsettings__system_wide__update_sz(
		pCtx,
		JS_GetStringBytes(JSVAL_TO_STRING(argv[0])),
		JS_GetStringBytes(JSVAL_TO_STRING(argv[1]))
		) );

	// TODO do we want to return something here?
	*rval = JSVAL_NULL;

	return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, local_settings)
{
	SG_context * pCtx = SG_jsglue__get_sg_context(cx);
	SG_vhash * pvh = NULL;
	JSObject* jso = NULL;
	const char * argOne = NULL;
	const char * argTwo = NULL;
	SG_uint32 firstArg = 0;
	SG_bool bReset = SG_FALSE;
	SG_varray* pva_options = NULL;

	SG_UNUSED(obj);

    // TODO rework this.  it's a very un-javascript-like API
    // see set_local_setting above for a partial replacement

	if (argc > 0)
	{
		jsval arg0 = argv[0];

		jso = JSVAL_TO_OBJECT(arg0);

		if (JS_IsArrayObject(cx, jso))
		{
			 SG_uint32 count = 0;

			 SG_uint32 j = 0;
			 const char* pszOption;

			 SG_ERR_CHECK(  sg_jsglue__jsobject_to_varray(pCtx, cx, jso, &pva_options)  );

			 SG_ERR_CHECK(  SG_varray__count(pCtx, pva_options, &count)  );

			 for (j = 0; j < count; j++)
			 {

				 SG_ERR_CHECK( SG_varray__get__sz(pCtx, pva_options, j, &pszOption) );

				 if ( strcmp(pszOption, "reset")==0)
				 {
					bReset = SG_TRUE;
				 }
			 }
			 firstArg++;
		}
		if (argc >= 2)
		{
			argOne = JS_GetStringBytes(JSVAL_TO_STRING(argv[firstArg]));
			if (argc > (firstArg + 1))
			{
				if (JSVAL_IS_STRING(argv[firstArg + 1]))
				{
					argTwo = JS_GetStringBytes(JSVAL_TO_STRING(argv[firstArg + 1]));
					SG_ERR_CHECK(  SG_localsettings__update__sz(pCtx, argOne, argTwo) );
				}
				else if (JSVAL_IS_INT(argv[firstArg + 1]))
                {
                    SG_int_to_string_buffer sz_i;

                    SG_int64_to_sz(JSVAL_TO_INT(argv[firstArg + 1]), sz_i);
					SG_ERR_CHECK(  SG_localsettings__update__sz(pCtx, argOne, sz_i)  );
                }
			}
			else if (bReset)
			{
				SG_ERR_CHECK(  SG_localsettings__reset(pCtx, argOne)  );
			}
		}
	}

    SG_ERR_CHECK(  SG_localsettings__list__vhash(pCtx, &pvh)  );

    SG_JS_NULL_CHECK(  (jso = JS_NewObject(cx, NULL, NULL, NULL))  );
    *rval = OBJECT_TO_JSVAL(jso);
    SG_ERR_CHECK(  sg_jsglue__copy_vhash_into_jsobject(pCtx, cx, pvh, jso)  );
    SG_VHASH_NULLFREE(pCtx, pvh);
	SG_VARRAY_NULLFREE(pCtx, pva_options);

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	SG_VHASH_NULLFREE(pCtx, pvh);
	SG_VARRAY_NULLFREE(pCtx, pva_options);
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, jsmin)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	const char *psz_arg = NULL;
	SG_string *inJs = NULL;
	SG_string *outJs = NULL;
    JSString* pjs = NULL;

	SG_UNUSED(obj);

    SG_JS_BOOL_CHECK(  1 == argc  );
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    psz_arg = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));
	SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &inJs, psz_arg)  );

	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &outJs)  );
	SG_ERR_CHECK(  SG_jsmin(pCtx, inJs, outJs)  );
	SG_STRING_NULLFREE(pCtx, inJs);

    SG_JS_NULL_CHECK(  (pjs = JS_NewStringCopyZ(cx, SG_string__sz(outJs)))  );
    SG_STRING_NULLFREE(pCtx, outJs);
	*rval = STRING_TO_JSVAL(pjs);
 
	return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS

	SG_STRING_NULLFREE(pCtx, inJs);
	SG_STRING_NULLFREE(pCtx, outJs);

    return JS_FALSE;

}

SG_JSGLUE_METHOD_PROTOTYPE(sg, list_descriptors)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_vhash* pvh = NULL;
	JSObject* jso = NULL;
	SG_uint32 i = 0;
	SG_uint32 numDescriptors = 0;

	SG_UNUSED(obj);
	SG_UNUSED(argv);

	SG_JS_BOOL_CHECK(argc==0);

	SG_ERR_CHECK(  SG_closet__descriptors__list(pCtx, &pvh)  );

	SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh, &numDescriptors)  );
	while(i<numDescriptors)
	{
		const char * szRepoName = NULL;
		SG_bool isAvailable = SG_TRUE;

		SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvh, i, &szRepoName, NULL)  );
		SG_ERR_CHECK(  sg_uridispatch__query_repo_availability(pCtx, szRepoName, &isAvailable)  );
		if(!isAvailable)
		{
			SG_ERR_CHECK(  SG_vhash__remove(pCtx, pvh, szRepoName)  );
			--numDescriptors;
		}
		else
			++i;
	}

    SG_JS_NULL_CHECK(  (jso = JS_NewObject(cx, NULL, NULL, NULL))  );
    *rval = OBJECT_TO_JSVAL(jso);
    SG_ERR_CHECK(  sg_jsglue__copy_vhash_into_jsobject(pCtx, cx, pvh, jso)  );
    SG_VHASH_NULLFREE(pCtx, pvh);

	return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx); // DO NOT SG_ERR_IGNORE() THIS
	SG_VHASH_NULLFREE(pCtx, pvh);
	return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, zip)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	const char* psz_repo = NULL;
	const char* psz_csid = NULL;
	const char* psz_path = NULL;
	SG_repo* pRepo = NULL;

    SG_UNUSED(obj);
    SG_UNUSED(argc);
    
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
	psz_repo = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  );
	psz_csid = JS_GetStringBytes(JSVAL_TO_STRING(argv[1]));
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[2])  );
	psz_path = JS_GetStringBytes(JSVAL_TO_STRING(argv[2]));

	SG_ERR_CHECK(  SG_REPO__OPEN_REPO_INSTANCE(pCtx, psz_repo, &pRepo)  );
    SG_ERR_CHECK(  SG_repo__zip(pCtx, pRepo, psz_csid, psz_path)  );
	SG_REPO_NULLFREE(pCtx, pRepo);

    *rval = JSVAL_VOID;
    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	SG_REPO_NULLFREE(pCtx, pRepo);
    return JS_FALSE;
}


SG_JSGLUE_METHOD_PROTOTYPE(sg, server_state_dir)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    JSString* pjs = NULL;
    SG_pathname *tmppath = NULL;

    SG_UNUSED(argv);
    SG_UNUSED(obj);

    SG_JS_BOOL_CHECK(  0 == argc  );

	SG_ERR_CHECK(  SG_closet__get_server_state_path(pCtx, &tmppath)  );

    SG_JS_NULL_CHECK(  (pjs = JS_NewStringCopyZ(cx, SG_pathname__sz(tmppath)))  );
    SG_PATHNAME_NULLFREE(pCtx, tmppath);

    *rval = STRING_TO_JSVAL(pjs);

    return JS_TRUE;

fail:
    SG_jsglue__report_sg_error(pCtx, cx);
    SG_PATHNAME_NULLFREE(pCtx, tmppath);
    *rval = JSVAL_VOID;
    return JS_FALSE;
}

void SG_jsglue__np__required__uint32(
    SG_context* pCtx,
    SG_vhash* pvh_args,
    SG_vhash* pvh_got,
    const char* psz_key,
    SG_uint32* pi
    )
{
    SG_bool b_has = SG_FALSE;

    SG_ERR_CHECK_RETURN(  SG_vhash__has(pCtx, pvh_args, psz_key, &b_has)  );
    if (b_has)
    {
        SG_ERR_CHECK_RETURN(  SG_vhash__get__uint32(pCtx, pvh_args, psz_key, pi)  );
        SG_ERR_CHECK_RETURN(  SG_vhash__add__null(pCtx, pvh_got, psz_key)  );
    }
    else
    {
        SG_ERR_THROW2_RETURN(SG_ERR_JS_NAMED_PARAM_REQUIRED, (pCtx, "%s", psz_key));
    }
}

void SG_jsglue__np__required__varray(
    SG_context* pCtx,
    SG_vhash* pvh_args,
    SG_vhash* pvh_got,
    const char* psz_key,
    SG_varray** ppva
    )
{
    SG_varray* pva = NULL;

    SG_ERR_CHECK_RETURN(  SG_vhash__check__varray(pCtx, pvh_args, psz_key, &pva)  );
    if (pva)
    {
        SG_ERR_CHECK_RETURN(  SG_vhash__add__null(pCtx, pvh_got, psz_key)  );
        *ppva = pva;
    }
    else
    {
        SG_ERR_THROW2_RETURN(SG_ERR_JS_NAMED_PARAM_REQUIRED, (pCtx, "%s", psz_key));
    }
}

void SG_jsglue__np__required__vhash(
    SG_context* pCtx,
    SG_vhash* pvh_args,
    SG_vhash* pvh_got,
    const char* psz_key,
    SG_vhash** ppvh
    )
{
    SG_vhash* pvh = NULL;

    SG_ERR_CHECK_RETURN(  SG_vhash__check__vhash(pCtx, pvh_args, psz_key, &pvh)  );
    if (pvh)
    {
        SG_ERR_CHECK_RETURN(  SG_vhash__add__null(pCtx, pvh_got, psz_key)  );
        *ppvh = pvh;
    }
    else
    {
        SG_ERR_THROW2_RETURN(SG_ERR_JS_NAMED_PARAM_REQUIRED, (pCtx, "%s", psz_key));
    }
}

void SG_jsglue__np__required__sz(
    SG_context* pCtx,
    SG_vhash* pvh_args,
    SG_vhash* pvh_got,
    const char* psz_key,
    const char** ppsz
    )
{
    const char* p = NULL;

    SG_ERR_CHECK_RETURN(  SG_vhash__check__sz(pCtx, pvh_args, psz_key, &p)  );
    if (p)
    {
        SG_ERR_CHECK_RETURN(  SG_vhash__add__null(pCtx, pvh_got, psz_key)  );
        *ppsz = p;
    }
    else
    {
        SG_ERR_THROW2_RETURN(SG_ERR_JS_NAMED_PARAM_REQUIRED, (pCtx, "%s", psz_key));
    }
}

static void SG_jsglue__convert_arg__bool(
    SG_context* pCtx,
    SG_vhash* pvh_args,
    const char* psz_key,
    SG_bool* p
    )
{
    SG_uint16 t = 0; 

    SG_ERR_CHECK_RETURN(  SG_vhash__typeof(pCtx, pvh_args, psz_key, &t)  ); 
    if (SG_VARIANT_TYPE_BOOL == t) 
    {
        SG_ERR_CHECK_RETURN(  SG_vhash__get__bool(pCtx, pvh_args, psz_key, p)  );
    }
    else if (SG_VARIANT_TYPE_INT64 == t) 
    {
        SG_int64 i64 = 0;

        SG_ERR_CHECK_RETURN(  SG_vhash__get__int64(pCtx, pvh_args, psz_key, &i64)  );

        if (i64)
        {
            *p = SG_TRUE;
        }
        else
        {
            *p = SG_FALSE;
        }
    }
    else if (SG_VARIANT_TYPE_SZ == t)
    {
        const char* psz = NULL;

        SG_ERR_CHECK_RETURN(  SG_vhash__get__sz(pCtx, pvh_args, psz_key, &psz)  );

        if (*psz)
        {
            *p = SG_TRUE;
        }
        else
        {
            *p = SG_FALSE;
        }
    }
    else if (SG_VARIANT_TYPE_SZ == t)
    {
        *p = SG_FALSE;
    }
    else if (SG_VARIANT_TYPE_VHASH == t)
    {
        *p = SG_TRUE;
    }
    else if (SG_VARIANT_TYPE_VARRAY == t)
    {
        *p = SG_TRUE;
    }
    else if (SG_VARIANT_TYPE_DOUBLE == t)
    {
        double d = 0;

        SG_ERR_CHECK_RETURN(  SG_vhash__get__double(pCtx, pvh_args, psz_key, &d)  );

        if (0 == ( (int) d ))
        {
            *p = SG_FALSE;
        }
        else
        {
            *p = SG_TRUE;
        }
    }
    else
    {
        SG_ERR_THROW_RETURN(  SG_ERR_NOTIMPLEMENTED  );
    }
}

void SG_jsglue__np__required__bool(
    SG_context* pCtx,
    SG_vhash* pvh_args,
    SG_vhash* pvh_got,
    const char* psz_key,
    SG_bool* p
    )
{
    SG_bool b_has = SG_FALSE; 

    SG_ERR_CHECK_RETURN(  SG_vhash__has(pCtx, pvh_args, psz_key, &b_has)  ); 
    if (!b_has) 
    { 
        SG_ERR_THROW2_RETURN(SG_ERR_JS_NAMED_PARAM_REQUIRED, (pCtx, "%s", psz_key));
    } 
    else 
    { 
        SG_ERR_CHECK_RETURN(  SG_jsglue__convert_arg__bool(pCtx, pvh_args, psz_key, p)  );
        SG_ERR_CHECK_RETURN(  SG_vhash__add__null(pCtx, pvh_got, psz_key)  );
    }
}

void SG_jsglue__np__optional__sz(
    SG_context* pCtx,
    SG_vhash* pvh_args,
    SG_vhash* pvh_got,
    const char* psz_key,
    const char* psz_def,
    const char** ppsz
    )
{
    const char* p = NULL;

    SG_ERR_CHECK_RETURN(  SG_vhash__check__sz(pCtx, pvh_args, psz_key, &p)  );
    if (p)
    {
        SG_ERR_CHECK_RETURN(  SG_vhash__add__null(pCtx, pvh_got, psz_key)  );
        *ppsz = p;
    }
    else
    {
        *ppsz = psz_def;
    }
}

void SG_jsglue__np__optional__bool(
    SG_context* pCtx,
    SG_vhash* pvh_args,
    SG_vhash* pvh_got,
    const char* psz_key,
    SG_bool def,
    SG_bool* p
    )
{
    SG_bool b_has = SG_FALSE; 

    SG_ERR_CHECK_RETURN(  SG_vhash__has(pCtx, pvh_args, psz_key, &b_has)  ); 
    if (!b_has) 
    { 
        *p = def; 
    } 
    else 
    { 
        SG_ERR_CHECK_RETURN(  SG_jsglue__convert_arg__bool(pCtx, pvh_args, psz_key, p)  );
        SG_ERR_CHECK_RETURN(  SG_vhash__add__null(pCtx, pvh_got, psz_key)  );
    }
}

static void SG_jsglue__convert_arg__uint32(
    SG_context* pCtx,
    SG_vhash* pvh_args,
    const char* psz_key,
    SG_uint32* p
    )
{
    SG_uint16 t = 0; 

    SG_ERR_CHECK_RETURN(  SG_vhash__typeof(pCtx, pvh_args, psz_key, &t)  ); 

    if (SG_VARIANT_TYPE_INT64 == t) 
    { 
        SG_int64 i64 = 0; 

        SG_ERR_CHECK_RETURN(  SG_vhash__get__int64(pCtx, pvh_args, psz_key, &i64)  ); 
        if (SG_int64__fits_in_uint32(i64)) 
        { 
            *p = (SG_uint32) i64; 
        } 
        else 
        { 
            SG_ERR_THROW_RETURN(  SG_ERR_INTEGER_OVERFLOW  ); 
        } 
    } 
    else if (SG_VARIANT_TYPE_SZ == t) 
    { 
        const char* psz = NULL; 
        SG_uint32 ui32 = 0; 

        SG_ERR_CHECK_RETURN(  SG_vhash__get__sz(pCtx, pvh_args, psz_key, &psz)  ); 
        SG_ERR_CHECK_RETURN(  SG_uint32__parse__strict(pCtx, &ui32, psz)  ); 
        *p = ui32; 
    } 
    else
    {
        SG_ERR_THROW_RETURN(  SG_ERR_JS_NAMED_PARAM_TYPE_CONVERSION  ); 
    }
}

static void SG_jsglue__convert_arg__int32(
    SG_context* pCtx,
    SG_vhash* pvh_args,
    const char* psz_key,
    SG_int32* p
    )
{
    SG_uint16 t = 0; 

    SG_ERR_CHECK_RETURN(  SG_vhash__typeof(pCtx, pvh_args, psz_key, &t)  ); 

    if (SG_VARIANT_TYPE_INT64 == t) 
    { 
        SG_int64 i64 = 0; 

        SG_ERR_CHECK_RETURN(  SG_vhash__get__int64(pCtx, pvh_args, psz_key, &i64)  ); 
        if (SG_int64__fits_in_int32(i64)) 
        { 
            *p = (SG_int32) i64; 
        } 
        else 
        { 
            SG_ERR_THROW_RETURN(  SG_ERR_INTEGER_OVERFLOW  ); 
        } 
    } 
    else if (SG_VARIANT_TYPE_SZ == t) 
    { 
        const char* psz = NULL; 
        SG_int32 i32 = 0; 

        SG_ERR_CHECK_RETURN(  SG_vhash__get__sz(pCtx, pvh_args, psz_key, &psz)  ); 
        SG_ERR_CHECK_RETURN(  SG_int32__parse__strict(pCtx, &i32, psz)  ); 
        *p = i32; 
    } 
    else
    {
        SG_ERR_THROW_RETURN(  SG_ERR_JS_NAMED_PARAM_TYPE_CONVERSION  ); 
    }
}

static void SG_jsglue__convert_arg__uint64(
    SG_context* pCtx,
    SG_vhash* pvh_args,
    const char* psz_key,
    SG_uint64* p
    )
{
    SG_uint16 t = 0; 

    SG_ERR_CHECK_RETURN(  SG_vhash__typeof(pCtx, pvh_args, psz_key, &t)  ); 

    if (SG_VARIANT_TYPE_INT64 == t) 
    { 
        SG_int64 i64 = 0; 

        SG_ERR_CHECK_RETURN(  SG_vhash__get__int64(pCtx, pvh_args, psz_key, &i64)  ); 
        if (SG_int64__fits_in_uint64(i64)) 
        { 
            *p = i64; 
        } 
        else 
        { 
            SG_ERR_THROW_RETURN(  SG_ERR_INTEGER_OVERFLOW  ); 
        } 
    } 
    else if (SG_VARIANT_TYPE_SZ == t) 
    { 
        const char* psz = NULL; 
        SG_uint64 ui64 = 0; 

        SG_ERR_CHECK_RETURN(  SG_vhash__get__sz(pCtx, pvh_args, psz_key, &psz)  ); 
        SG_ERR_CHECK_RETURN(  SG_uint64__parse__strict(pCtx, &ui64, psz)  ); 
        *p = ui64; 
    } 
    else
    {
        SG_ERR_THROW_RETURN(  SG_ERR_JS_NAMED_PARAM_TYPE_CONVERSION  ); 
    }
}

void SG_jsglue__np__optional__uint64(
    SG_context* pCtx,
    SG_vhash* pvh_args,
    SG_vhash* pvh_got,
    const char* psz_key,
    SG_uint64 def,
    SG_uint64* p
    )
{
    SG_bool b_has = SG_FALSE; 

    SG_ERR_CHECK_RETURN(  SG_vhash__has(pCtx, pvh_args, psz_key, &b_has)  ); 
    if (!b_has) 
    { 
        *p = def; 
    } 
    else 
    { 
        SG_ERR_CHECK_RETURN(  SG_jsglue__convert_arg__uint64(pCtx, pvh_args, psz_key, p)  );
        SG_ERR_CHECK_RETURN(  SG_vhash__add__null(pCtx, pvh_got, psz_key)  );
    }
}

void SG_jsglue__np__optional__uint32(
    SG_context* pCtx,
    SG_vhash* pvh_args,
    SG_vhash* pvh_got,
    const char* psz_key,
    SG_uint32 def,
    SG_uint32* p
    )
{
    SG_bool b_has = SG_FALSE; 

    SG_ERR_CHECK_RETURN(  SG_vhash__has(pCtx, pvh_args, psz_key, &b_has)  ); 
    if (!b_has) 
    { 
        *p = def; 
    } 
    else 
    { 
        SG_ERR_CHECK_RETURN(  SG_jsglue__convert_arg__uint32(pCtx, pvh_args, psz_key, p)  );
        SG_ERR_CHECK_RETURN(  SG_vhash__add__null(pCtx, pvh_got, psz_key)  );
    }
}

void SG_jsglue__np__optional__int32(
    SG_context* pCtx,
    SG_vhash* pvh_args,
    SG_vhash* pvh_got,
    const char* psz_key,
    SG_int32 def,
    SG_int32* p
    )
{
    SG_bool b_has = SG_FALSE; 

    SG_ERR_CHECK_RETURN(  SG_vhash__has(pCtx, pvh_args, psz_key, &b_has)  ); 
    if (!b_has) 
    { 
        *p = def; 
    } 
    else 
    { 
        SG_ERR_CHECK_RETURN(  SG_jsglue__convert_arg__int32(pCtx, pvh_args, psz_key, p)  );
        SG_ERR_CHECK_RETURN(  SG_vhash__add__null(pCtx, pvh_got, psz_key)  );
    }
}

void SG_jsglue__np__anything_not_got_is_invalid(
        SG_context* pCtx,
        const char* psz_func,
        SG_vhash* pvh_args,
        SG_vhash* pvh_got
        )
{
    SG_string* pstr = NULL;
    SG_uint32 count = 0;
    SG_uint32 errs = 0;

    if (pvh_args)
    {
        SG_ERR_CHECK(  SG_vhash__count(pCtx, pvh_args, &count)  );
        if (count > 0)
        {
            SG_uint32 i = 0;

            SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pstr)  );
            for (i=0; i<count; i++)
            {
                const char* psz_param_name = NULL;
                SG_bool b_got = SG_FALSE;

                SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvh_args, i, &psz_param_name, NULL)  );
                SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh_got, psz_param_name, &b_got)  );

                if (!b_got)
                {
                    if (errs)
                    {
                        SG_ERR_CHECK(  SG_string__append__sz(pCtx, pstr, ", ")  );
                    }
                    SG_ERR_CHECK(  SG_string__append__sz(pCtx, pstr, psz_param_name)  );
                    errs++;
                }
            }
            if (errs)
            {
                SG_ERR_THROW2(  SG_ERR_JS_NAMED_PARAM_NOT_ALLOWED,
                        (pCtx, "%s: %s", psz_func, SG_string__sz(pstr))
                        );
            }
        }
    }

fail:
    SG_STRING_NULLFREE(pCtx, pstr);
}

static JSBool do_clone_all_to_something(
    JSContext *cx, 
    JSObject *obj, 
    uintN argc, 
    jsval *argv, 
    jsval *rval,
    const char* psz_func,
    SG_blob_encoding encoding
    )
{
    SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    const char* psz_existing = NULL;
    const char* psz_new = NULL;
    SG_vhash* pvh_args = NULL;
    SG_vhash* pvh_got = NULL;
    SG_bool b_override_alwaysfull = SG_FALSE;
    struct SG_clone__params__all_to_something ats;
    struct SG_clone__demands demands;

    SG_UNUSED(obj);

    memset(&ats, 0, sizeof(ats));
    ats.new_encoding = encoding;

    SG_ERR_CHECK(  SG_clone__init_demands(pCtx, &demands)  );

	SG_JS_BOOL_CHECK(1 == argc);
    SG_JS_BOOL_CHECK(  !JSVAL_IS_NULL(argv[0]) && JSVAL_IS_OBJECT(argv[0])  );
    SG_ERR_CHECK(  sg_jsglue__jsobject_to_vhash(pCtx, cx, JSVAL_TO_OBJECT(argv[0]), &pvh_args)  );
    SG_ERR_CHECK(  SG_vhash__alloc(pCtx, &pvh_got)  );

    SG_ERR_CHECK(  SG_jsglue__np__required__sz(pCtx, pvh_args, pvh_got, "existing_repo", &psz_existing)  );
    SG_ERR_CHECK(  SG_jsglue__np__required__sz(pCtx, pvh_args, pvh_got, "new_repo", &psz_new)  );

    SG_ERR_CHECK(  SG_jsglue__np__optional__uint64(pCtx, pvh_args, pvh_got, "low_pass", 0, &ats.low_pass)  );
    SG_ERR_CHECK(  SG_jsglue__np__optional__uint64(pCtx, pvh_args, pvh_got, "high_pass", 0, &ats.high_pass)  );
    SG_ERR_CHECK(  SG_jsglue__np__optional__bool(pCtx, pvh_args, pvh_got, "override_alwaysfull", SG_FALSE, &b_override_alwaysfull)  );
    if (SG_IS_BLOBENCODING_ZLIB(encoding))
    {
        SG_ERR_CHECK(  SG_jsglue__np__optional__int32(pCtx, pvh_args, pvh_got, SG_CLONE_CHANGES__DEMAND_ZLIB_SAVINGS_OVER_FULL, -1, &demands.zlib_savings_over_full)  );
    }

    SG_ERR_CHECK(  SG_jsglue__np__anything_not_got_is_invalid(pCtx, psz_func, pvh_args, pvh_got)  );

    if (b_override_alwaysfull)
    {
        ats.flags |= SG_CLONE_FLAG__OVERRIDE_ALWAYSFULL;
    }

    SG_ERR_CHECK(  SG_clone__all_to_something(
                pCtx, 
                psz_existing, 
                psz_new, 
                &ats,
                &demands
                )  );

	// Save default push/pull location for the new repository instance.
	SG_ERR_CHECK(  SG_localsettings__descriptor__update__sz(pCtx, psz_new, SG_LOCALSETTING__PATHS_DEFAULT, psz_existing)  );

    SG_VHASH_NULLFREE(pCtx, pvh_args);
    SG_VHASH_NULLFREE(pCtx, pvh_got);

    *rval = JSVAL_VOID;
    return JS_TRUE;

fail:
    SG_VHASH_NULLFREE(pCtx, pvh_args);
    SG_VHASH_NULLFREE(pCtx, pvh_got);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, clone__all_full)
{
    return do_clone_all_to_something(cx, obj, argc, argv, rval, "clone__all_full", SG_BLOBENCODING__KEEPFULLFORNOW);
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, clone__all_zlib)
{
    return do_clone_all_to_something(cx, obj, argc, argv, rval, "clone__all_zlib", SG_BLOBENCODING__ZLIB);
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, pull_branch_stuff_only)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	const char* psz_from = NULL;
	const char* psz_to = NULL;
    SG_vhash* pvh_args = NULL;
    SG_vhash* pvh_got = NULL;
    char* psz_repo_upstream = NULL;

    SG_UNUSED(obj);

	SG_JS_BOOL_CHECK(1 == argc);
    SG_JS_BOOL_CHECK(  JSVAL_IS_OBJECT(argv[0])  );
    SG_ERR_CHECK(  sg_jsglue__jsobject_to_vhash(pCtx, cx, JSVAL_TO_OBJECT(argv[0]), &pvh_args)  );
    SG_ERR_CHECK(  SG_vhash__alloc(pCtx, &pvh_got)  );

    SG_ERR_CHECK(  SG_jsglue__np__optional__sz(pCtx, pvh_args, pvh_got, "from", NULL, &psz_from)  );
    SG_ERR_CHECK(  SG_jsglue__np__required__sz(pCtx, pvh_args, pvh_got, "to", &psz_to)  );

    SG_ERR_CHECK(  SG_jsglue__np__anything_not_got_is_invalid(pCtx, "pull_branch_stuff_only", pvh_args, pvh_got)  );

    if (!psz_from)
    {
        // we need the repo spec for the other side
        SG_ERR_CHECK(  SG_localsettings__descriptor__get__sz(pCtx, psz_to, "paths/default", &psz_repo_upstream)  );
        psz_from = psz_repo_upstream;
    }

    SG_ERR_CHECK(  SG_vc_locks__pull_locks_and_branches(pCtx, psz_from, psz_to)  );

    SG_NULLFREE(pCtx, psz_repo_upstream);
    SG_VHASH_NULLFREE(pCtx, pvh_args);
    SG_VHASH_NULLFREE(pCtx, pvh_got);

    *rval = JSVAL_VOID;
    return JS_TRUE;

fail:
    SG_NULLFREE(pCtx, psz_repo_upstream);
    SG_VHASH_NULLFREE(pCtx, pvh_args);
    SG_VHASH_NULLFREE(pCtx, pvh_got);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, clone__pack)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	const char* psz_existing = NULL;
	const char* psz_new = NULL;
	SG_int32 keyframe_density;
	SG_int32 revisions_to_leave_undeltified;
	SG_int32 min_revisions;
    SG_uint64 low_pass = 0;
    SG_uint64 high_pass = 0;
    struct SG_clone__demands demands;
    SG_vhash* pvh_args = NULL;
    SG_vhash* pvh_got = NULL;

    SG_UNUSED(obj);

    SG_ERR_CHECK(  SG_clone__init_demands(pCtx, &demands)  );

	SG_JS_BOOL_CHECK(1 == argc);
    SG_JS_BOOL_CHECK(  JSVAL_IS_OBJECT(argv[0])  );
    SG_ERR_CHECK(  sg_jsglue__jsobject_to_vhash(pCtx, cx, JSVAL_TO_OBJECT(argv[0]), &pvh_args)  );
    SG_ERR_CHECK(  SG_vhash__alloc(pCtx, &pvh_got)  );

    SG_ERR_CHECK(  SG_jsglue__np__required__sz(pCtx, pvh_args, pvh_got, "existing_repo", &psz_existing)  );
    SG_ERR_CHECK(  SG_jsglue__np__required__sz(pCtx, pvh_args, pvh_got, "new_repo", &psz_new)  );

    SG_ERR_CHECK(  SG_jsglue__np__optional__uint64(pCtx, pvh_args, pvh_got, "low_pass", 0, &low_pass)  );
    SG_ERR_CHECK(  SG_jsglue__np__optional__uint64(pCtx, pvh_args, pvh_got, "high_pass", 0, &high_pass)  );
    SG_ERR_CHECK(  SG_jsglue__np__optional__int32(pCtx, pvh_args, pvh_got, SG_CLONE_CHANGES__DEMAND_ZLIB_SAVINGS_OVER_FULL, -1, &demands.zlib_savings_over_full)  );
    SG_ERR_CHECK(  SG_jsglue__np__optional__int32(pCtx, pvh_args, pvh_got, SG_CLONE_CHANGES__DEMAND_VCDIFF_SAVINGS_OVER_FULL, -1, &demands.vcdiff_savings_over_full)  );
    SG_ERR_CHECK(  SG_jsglue__np__optional__int32(pCtx, pvh_args, pvh_got, SG_CLONE_CHANGES__DEMAND_VCDIFF_SAVINGS_OVER_ZLIB, -1, &demands.vcdiff_savings_over_zlib)  );
    SG_ERR_CHECK(  SG_jsglue__np__optional__int32(pCtx, pvh_args, pvh_got, "keyframe_density", SG_PACK_DEFAULT__KEYFRAMEDENSITY, &keyframe_density)  );
    SG_ERR_CHECK(  SG_jsglue__np__optional__int32(pCtx, pvh_args, pvh_got, "leave_full", SG_PACK_DEFAULT__REVISIONSTOLEAVEFULL, &revisions_to_leave_undeltified)  );
    SG_ERR_CHECK(  SG_jsglue__np__optional__int32(pCtx, pvh_args, pvh_got, "min_revisions", SG_PACK_DEFAULT__MINREVISIONS, &min_revisions)  );

    SG_ERR_CHECK(  SG_jsglue__np__anything_not_got_is_invalid(pCtx, "clone__pack", pvh_args, pvh_got)  );

    SG_ERR_CHECK(  SG_clone__pack(
                pCtx, 
                psz_existing, 
                psz_new, 
                keyframe_density, 
                revisions_to_leave_undeltified, 
                min_revisions,
                low_pass,
                high_pass,
                &demands
                )  );

	// Save default push/pull location for the new repository instance.
	SG_ERR_CHECK(  SG_localsettings__descriptor__update__sz(pCtx, psz_new, SG_LOCALSETTING__PATHS_DEFAULT, psz_existing)  );

    SG_VHASH_NULLFREE(pCtx, pvh_args);
    SG_VHASH_NULLFREE(pCtx, pvh_got);

    *rval = JSVAL_VOID;
    return JS_TRUE;

fail:
    SG_VHASH_NULLFREE(pCtx, pvh_args);
    SG_VHASH_NULLFREE(pCtx, pvh_got);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, clone__exact)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	const char* psz_existing = NULL;
	const char* psz_new = NULL;

    SG_UNUSED(obj);
    
	SG_JS_BOOL_CHECK(2 == argc);

	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
	psz_existing = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));

	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  );
	psz_new = JS_GetStringBytes(JSVAL_TO_STRING(argv[1]));

    SG_ERR_CHECK(  SG_clone__exact(pCtx, psz_existing, psz_new)  );

	// Save default push/pull location for the new repository instance.
	SG_ERR_CHECK(  SG_localsettings__descriptor__update__sz(pCtx, psz_new, SG_LOCALSETTING__PATHS_DEFAULT, psz_existing)  );

    *rval = JSVAL_VOID;
    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, clone__specific_changes)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	const char* psz_existing = NULL;
	const char* psz_new = NULL;
    SG_vhash* pvh_specific = NULL;
    SG_vhash* pvh_args = NULL;
    SG_vhash* pvh_got = NULL;
    SG_clone* pClone = NULL;
    struct SG_clone__demands demands;

    SG_UNUSED(obj);
    
    SG_ERR_CHECK(  SG_clone__init_demands(pCtx, &demands)  );

	SG_JS_BOOL_CHECK(1 == argc);
    SG_JS_BOOL_CHECK(  JSVAL_IS_OBJECT(argv[0])  );
    SG_ERR_CHECK(  sg_jsglue__jsobject_to_vhash(pCtx, cx, JSVAL_TO_OBJECT(argv[0]), &pvh_args)  );
    SG_ERR_CHECK(  SG_vhash__alloc(pCtx, &pvh_got)  );

    SG_ERR_CHECK(  SG_jsglue__np__required__sz(pCtx, pvh_args, pvh_got, "existing_repo", &psz_existing)  );
    SG_ERR_CHECK(  SG_jsglue__np__required__sz(pCtx, pvh_args, pvh_got, "new_repo", &psz_new)  );

    SG_ERR_CHECK(  SG_jsglue__np__required__vhash(pCtx, pvh_args, pvh_got, "specific", &pvh_specific)  );

    SG_ERR_CHECK(  SG_jsglue__np__optional__int32(pCtx, pvh_args, pvh_got, SG_CLONE_CHANGES__DEMAND_ZLIB_SAVINGS_OVER_FULL, -1, &demands.zlib_savings_over_full)  );
    SG_ERR_CHECK(  SG_jsglue__np__optional__int32(pCtx, pvh_args, pvh_got, SG_CLONE_CHANGES__DEMAND_VCDIFF_SAVINGS_OVER_FULL, -1, &demands.vcdiff_savings_over_full)  );
    SG_ERR_CHECK(  SG_jsglue__np__optional__int32(pCtx, pvh_args, pvh_got, SG_CLONE_CHANGES__DEMAND_VCDIFF_SAVINGS_OVER_ZLIB, -1, &demands.vcdiff_savings_over_zlib)  );

    SG_ERR_CHECK(  SG_jsglue__np__anything_not_got_is_invalid(pCtx, "clone__specific_changes", pvh_args, pvh_got)  );

    SG_ERR_CHECK(  SG_clone__create(pCtx, psz_existing, &pClone)  );

	SG_ERR_CHECK(  SG_clone__commit(pCtx, pClone, psz_new, pvh_specific, NULL, &demands)  );

    SG_ERR_CHECK(  SG_clone__cleanup(pCtx, &pClone)  );

	// Save default push/pull location for the new repository instance.
	SG_ERR_CHECK(  SG_localsettings__descriptor__update__sz(pCtx, psz_new, SG_LOCALSETTING__PATHS_DEFAULT, psz_existing)  );

    SG_VHASH_NULLFREE(pCtx, pvh_args);
    SG_VHASH_NULLFREE(pCtx, pvh_got);

    *rval = JSVAL_VOID;
    return JS_TRUE;

fail:
    SG_ERR_IGNORE(  SG_clone__cleanup(pCtx, &pClone)  );
    SG_VHASH_NULLFREE(pCtx, pvh_args);
    SG_VHASH_NULLFREE(pCtx, pvh_got);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}

/**
* void sg.push_all(string src_repo_descriptor, string dest_repo_spec, [bool force=false])
*/	
SG_JSGLUE_METHOD_PROTOTYPE(sg, push_all)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	const char* pszSrc = NULL;
	const char* pszDest = NULL;
	SG_bool bForce = SG_FALSE;

	SG_UNUSED(obj);

	SG_JS_BOOL_CHECK(2 == argc || 3 == argc);

	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
	pszSrc = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));

	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  );
	pszDest = JS_GetStringBytes(JSVAL_TO_STRING(argv[1]));

	if (3 == argc)
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_BOOLEAN(argv[2])  );
		bForce = JSVAL_TO_BOOLEAN(argv[2]);
	}

	SG_ERR_CHECK(  SG_vv_verbs__push(pCtx, pszDest, pszSrc, NULL, NULL, bForce, NULL)  );

	*rval = JSVAL_VOID;
	return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}

/**
* void sg.push_branch(string src_repo_descriptor, string dest_repo_spec, string branch_name, [bool force=false])
*/	
SG_JSGLUE_METHOD_PROTOTYPE(sg, push_branch)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	const char* pszSrc = NULL;
	const char* pszDest = NULL;
	const char* pszBranch = NULL;
	SG_rev_spec* pRevSpec = NULL;
	SG_bool bForce = SG_FALSE;

	SG_UNUSED(obj);

	SG_JS_BOOL_CHECK(3 == argc || 4 == argc);

	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
	pszSrc = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));

	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  );
	pszDest = JS_GetStringBytes(JSVAL_TO_STRING(argv[1]));

	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[2])  );
	pszBranch = JS_GetStringBytes(JSVAL_TO_STRING(argv[2]));

	if (4 == argc)
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_BOOLEAN(argv[3])  );
		bForce = JSVAL_TO_BOOLEAN(argv[3]);
	}

	SG_ERR_CHECK(  SG_REV_SPEC__ALLOC(pCtx, &pRevSpec)  );
	SG_ERR_CHECK(  SG_rev_spec__add_branch(pCtx, pRevSpec, pszBranch)  );
	SG_ERR_CHECK(  SG_vv_verbs__push(pCtx, pszDest, pszSrc, NULL, pRevSpec, bForce, NULL)  );
	SG_REV_SPEC_NULLFREE(pCtx, pRevSpec);

	*rval = JSVAL_VOID;
	return JS_TRUE;

fail:
	SG_REV_SPEC_NULLFREE(pCtx, pRevSpec);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}

/**
* void sg.pull_all(string src_repo_spec, string dest_repo_descriptor)
*/	
SG_JSGLUE_METHOD_PROTOTYPE(sg, pull_all)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	const char* pszSrc = NULL;
	const char* pszDest = NULL;

	SG_UNUSED(obj);

	SG_JS_BOOL_CHECK(2 == argc);

	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
	pszSrc = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));

	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  );
	pszDest = JS_GetStringBytes(JSVAL_TO_STRING(argv[1]));

	SG_ERR_CHECK(  SG_vv_verbs__pull(pCtx, pszSrc, pszDest, NULL, NULL, NULL)  );

	*rval = JSVAL_VOID;
	return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, open_repo)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    JSObject* jso = NULL;
    SG_repo* pRepo = NULL;
    SG_safeptr* psp = NULL;
	const char *uid = NULL;

    SG_UNUSED(obj);

	SG_JS_BOOL_CHECK( (argc == 1) || (argc == 2) );

	if ((argc == 2) && ! JSVAL_IS_NULL(argv[1]))
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  );
		uid = JS_GetStringBytes(JSVAL_TO_STRING(argv[1]));
	}

    SG_JS_NULL_CHECK(  jso = JS_NewObject(cx, &sg_repo_class, NULL, NULL)  );
    *rval = OBJECT_TO_JSVAL(jso);

	SG_ERR_CHECK(  sg_jsglue__get_repo_from_jsval(pCtx, cx, argv[0], &pRepo, uid)  );

    SG_ERR_CHECK(  SG_safeptr__wrap__repo(pCtx, pRepo, &psp)  );
    JS_SetPrivate(cx, jso, psp);
    psp = NULL;

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	// TODO do we need to free anything?
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, open_local_repo)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    JSObject* jso = NULL;
    SG_repo* pRepo = NULL;
    SG_safeptr* psp = NULL;

    SG_UNUSED(obj);
    SG_UNUSED(argc);
	SG_UNUSED(argv);

    SG_JS_NULL_CHECK(  jso = JS_NewObject(cx, &sg_repo_class, NULL, NULL)  );
    *rval = OBJECT_TO_JSVAL(jso);

	SG_ERR_CHECK(  sg_jsglue__get_repo_from_cwd(pCtx, &pRepo, NULL, NULL)  );

    SG_ERR_CHECK(  SG_safeptr__wrap__repo(pCtx, pRepo, &psp)  );
    JS_SetPrivate(cx, jso, psp);
    psp = NULL;

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	// TODO do we need to free anything?
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, console)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    const char * msg;

	SG_UNUSED(obj);
	SG_UNUSED(rval);

	SG_JS_BOOL_CHECK(0 < argc);

	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    msg = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));
    
	SG_console(pCtx, SG_CS_STDOUT, "%s\n", msg );     
	
	return JS_TRUE;
fail:
	return JS_FALSE;
}
SG_JSGLUE_METHOD_PROTOTYPE(sg, log)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	const char * level;
	const char * msg;

	SG_UNUSED(obj);
	SG_UNUSED(rval);

	SG_JS_BOOL_CHECK(0 < argc);

	if (argc == 1)
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
		msg = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));

		SG_log__report_info(pCtx, "%s", msg);
	}
	else
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
		level = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));
		SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  );
		msg = JS_GetStringBytes(JSVAL_TO_STRING(argv[1]));

		if (0 == strcmp(level, "debug"))
		{
			SG_log__report_verbose(pCtx, "%s", msg);
		}
		else if (0 == strcmp(level, "urgent"))
		{
			SG_log__report_error(pCtx, "%s", msg);
		}
		else if (0 == strcmp(level, "normal"))
		{
			SG_log__report_info(pCtx, "%s", msg);
		}
	}
	return JS_TRUE;
fail:
	return JS_FALSE;
}

//SG_JSGLUE_METHOD_PROTOTYPE(sg, open_pending_tree)
//{
//	SG_context * pCtx = SG_jsglue__get_sg_context(cx);
//    JSObject* jso = NULL;
//    SG_safeptr* psp = NULL;
//    SG_pendingtree* pPendingTree = NULL;
//    SG_pathname* pPath = NULL;
//
//    SG_JS_NULL_CHECK(  jso = JS_NewObject(cx, &pending_tree_class, NULL, NULL)  );
//    *rval = OBJECT_TO_JSVAL(jso);
//
//    SG_ERR_CHECK(  SG_PATHNAME__ALLOC(pCtx, &pPath)  );
//	SG_ERR_CHECK(  SG_pathname__set__from_cwd(pCtx, pPath)  );
//    SG_ERR_CHECK(  SG_pendingtree__alloc(pCtx, pPath, &pPendingTree)  );
//
//    SG_PATHNAME_NULLFREE(pCtx, pPath);
//    SG_ERR_CHECK(  SG_safeptr__wrap__pending_tree(pCtx, pPendingTree, &psp)  );
//    JS_SetPrivate(cx, jso, psp);
//    psp = NULL;
//
//    return JS_TRUE;
//
//fail:
//	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
//	// TODO do we need to free anything?
//    return JS_FALSE;
//}

//SG_JSGLUE_METHOD_PROTOTYPE(pending_tree, close)
//{
//	SG_context * pCtx = SG_jsglue__get_sg_context(cx);
//    SG_pendingtree* pPendingTree = NULL;
//    SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
//
//    SG_ERR_CHECK(  SG_safeptr__unwrap__pending_tree(pCtx, psp, &pPendingTree)  );
//    SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
//    SG_SAFEPTR_NULLFREE(pCtx, psp);
//    JS_SetPrivate(cx, obj, NULL);
//
//    *rval = JSVAL_VOID;
//    return JS_TRUE;
//
//fail:
//	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
//    /* TODO */
//    return JS_FALSE;
//}

/*  usage: sg.add([options],[files/folders])
 *
 *  possible options: nonrecursive, include=pattern, exclude=pattern, test
 */
SG_JSGLUE_METHOD_PROTOTYPE(pending_tree, add)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_pendingtree* pPendingTree = NULL;
	SG_pathname* pPath = NULL;
	const char** newArgv = NULL;
	uint i;
	uint n = 0;
	JSObject* jso = NULL;
	jsval arg0 = argv[0];
	SG_bool bRecursive = SG_TRUE;
	SG_bool bTestOnly = SG_FALSE;
	SG_uint32 nCountItems = 0;
	SG_varray* pva_options = NULL;
	SG_file_spec* pFilespec = NULL;

	SG_UNUSED(obj);

	jso = JSVAL_TO_OBJECT(arg0);

	SG_ERR_CHECK (  SG_FILE_SPEC__ALLOC(pCtx, &pFilespec)  );

    if (argc > 0 && JS_IsArrayObject(cx, jso))
    {
		 SG_uint32 count = 0;

		 SG_uint32 j = 0;
		 const char* pszOption;

         SG_ERR_CHECK(  sg_jsglue__jsobject_to_varray(pCtx, cx, jso, &pva_options)  );

		 SG_ERR_CHECK(  SG_varray__count(pCtx, pva_options, &count)  );

		 for (j = 0; j < count; j++)
		 {
			 SG_ERR_CHECK( SG_varray__get__sz(pCtx, pva_options, j, &pszOption) );

			 if ( strcmp(pszOption, "nonrecursive")==0)
			 {
				bRecursive = SG_FALSE;
			 }
			 else if ( strcmp(pszOption, "test")==0)
			 {
				bTestOnly = SG_TRUE;
			 }
			 else if ( sg_jsglue__is_filespec_arg(pszOption) )
			 {
				 SG_ERR_CHECK(  sg_jsglue__process_filespec_arg(pCtx, pszOption, pFilespec)  );
			 }
		 }

		 n++;
	}
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC(pCtx, &pPath)  );
	SG_ERR_CHECK(  SG_pathname__set__from_cwd(pCtx, pPath)  );

	SG_ERR_CHECK(  SG_pendingtree__alloc(pCtx, pPath, &pPendingTree)  );

	//allocate the new string array that we will send down.
	SG_ERR_CHECK(  SG_alloc(pCtx, argc, sizeof(char*), &newArgv)  );

	for (i = n; i < argc; i++)
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[i])  );
		newArgv[nCountItems] = JS_GetStringBytes(JSVAL_TO_STRING(argv[i]));
		nCountItems++;
	}

	SG_ERR_CHECK(  SG_pendingtree__add(
		pCtx, pPendingTree,
		nCountItems,
		newArgv,
		bRecursive,
		pFilespec,
		bTestOnly)  );

	SG_FILE_SPEC_NULLFREE(pCtx, pFilespec);
	SG_NULLFREE(pCtx, newArgv);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_VARRAY_NULLFREE(pCtx, pva_options);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	*rval = JSVAL_VOID;
	return JS_TRUE;

fail:
	SG_FILE_SPEC_NULLFREE(pCtx, pFilespec);
	SG_NULLFREE(pCtx, newArgv);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_VARRAY_NULLFREE(pCtx, pva_options);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    /* TODO */
    return JS_FALSE;
}


/*  usage: sg.remove([options],[files/folders])
 *
 *  possible options: include=pattern, exclude=pattern, test
 */
SG_JSGLUE_METHOD_PROTOTYPE(pending_tree, remove)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_pendingtree* pPendingTree = NULL;
	SG_pathname* pPath = NULL;
	const char** newArgv = NULL;
	uint i;
	uint n = 0;
	JSObject* jso = NULL;
	jsval arg0 = argv[0];
	SG_bool bTestOnly = SG_FALSE;
	SG_bool bForce = SG_FALSE;
	SG_bool bNoBackups = SG_FALSE;
	SG_uint32 nCountItems = 0;
	SG_varray* pva_options = NULL;
	SG_file_spec* pFilespec = NULL;

	SG_UNUSED(obj);

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC(pCtx, &pPath)  );
	SG_ERR_CHECK(  SG_pathname__set__from_cwd(pCtx, pPath)  );

	SG_ERR_CHECK(  SG_pendingtree__alloc(pCtx, pPath, &pPendingTree)  );

	SG_ERR_CHECK(  SG_FILE_SPEC__ALLOC(pCtx, &pFilespec)  );

	jso = JSVAL_TO_OBJECT(arg0);

    if (argc > 0 && JS_IsArrayObject(cx, jso))
    {
		 SG_uint32 j = 0;
		 const char* pszOption;
		 SG_uint32 count;

         SG_ERR_CHECK(  sg_jsglue__jsobject_to_varray(pCtx, cx, jso, &pva_options)  );

		 SG_ERR_CHECK(  SG_varray__count(pCtx, pva_options, &count)  );

		 for (j = 0; j < count; j++)
		 {
			 SG_ERR_CHECK( SG_varray__get__sz(pCtx, pva_options, j, &pszOption) );

			 if ( strcmp(pszOption, "test")==0)
			 {
				bTestOnly = SG_TRUE;
			 }
			 else if (strcmp(pszOption, "force") == 0)
			 {
				 bForce = SG_TRUE;
			 }
			 else if (strcmp(pszOption, "no-backups") == 0)	// only makes sense when used with "force"
			 {
				 bNoBackups = SG_TRUE;
			 }
			 else if ( sg_jsglue__is_filespec_arg(pszOption) )
			 {
				SG_ERR_CHECK(  sg_jsglue__process_filespec_arg(pCtx, pszOption, pFilespec)  );
			 }
			 else
			 {
				 
				 SG_ERR_THROW2(  SG_ERR_INVALIDARG,
								 (pCtx, "sg.pending_tree.remove(['%s'...)", pszOption)  );
			 }
		 }
		 n++;

	}

	//allocate the new string array that we will send down.
	SG_ERR_CHECK(  SG_alloc(pCtx, argc, sizeof(char*), &newArgv)  );
	for (i = n; i < argc; i++)
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[i])  );
		newArgv[nCountItems] = JS_GetStringBytes(JSVAL_TO_STRING(argv[i]));
		nCountItems++;
	}
	if (nCountItems > 0)
		SG_ERR_CHECK(  SG_pendingtree__remove(pCtx,
										  pPendingTree,
										  nCountItems, newArgv,
										  pFilespec,
										  bTestOnly, bForce, bNoBackups)  );

	SG_NULLFREE(pCtx, newArgv);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_VARRAY_NULLFREE(pCtx, pva_options);
	SG_FILE_SPEC_NULLFREE(pCtx, pFilespec);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	*rval = JSVAL_VOID;
	return JS_TRUE;

fail:
	SG_NULLFREE(pCtx, newArgv);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_VARRAY_NULLFREE(pCtx, pva_options);
	SG_FILE_SPEC_NULLFREE(pCtx, pFilespec);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    /* TODO */
    return JS_FALSE;
}

/*  usage: sg.rename([options], file_or_folder, new_name)
 */
SG_JSGLUE_METHOD_PROTOTYPE(pending_tree, rename)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_pendingtree* pPendingTree = NULL;
	SG_pathname* pPath = NULL;
	SG_pathname* pPathToRename = NULL;
	const char* psz_arg1 = NULL;
	const char* psz_arg2 = NULL;
	uint n = 0;
	JSObject* jso = NULL;
	jsval arg0 = argv[0];
	SG_varray* pva_options = NULL;

	SG_UNUSED(obj);

	jso = JSVAL_TO_OBJECT(arg0);

    if (argc > 0 && JS_IsArrayObject(cx, jso))
    {
		 SG_uint32 count = 0;
		 SG_uint32 j = 0;
		 const char* pszOption;

         SG_ERR_CHECK(  sg_jsglue__jsobject_to_varray(pCtx, cx, jso, &pva_options)  );

		 SG_ERR_CHECK(  SG_varray__count(pCtx, pva_options, &count)  );

		 for (j = 0; j < count; j++)
		 {
			 SG_ERR_CHECK( SG_varray__get__sz(pCtx, pva_options, j, &pszOption) );

			 // TODO 2011/05/09 Parse any options....

			 if (pszOption && *pszOption)	// this test is dumb, but it makes MSVC happy
				 SG_ERR_THROW2(  SG_ERR_INVALIDARG,
								 (pCtx, "sg.pending_tree.rename(['%s'...)", pszOption)  );
		 }
		 n++;
	}

	SG_JS_BOOL_CHECK(  2 == (argc - n)  );
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[n])  );
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[n+1])  );
	psz_arg1 = JS_GetStringBytes(JSVAL_TO_STRING(argv[n]));
	psz_arg2 = JS_GetStringBytes(JSVAL_TO_STRING(argv[n+1]));

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPathToRename, psz_arg1)  );

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC(pCtx, &pPath)  );
	SG_ERR_CHECK(  SG_pathname__set__from_cwd(pCtx, pPath)  );

	SG_ERR_CHECK(  SG_pendingtree__alloc(pCtx, pPath, &pPendingTree)  );

	SG_ERR_CHECK(  SG_pendingtree__rename(pCtx, pPendingTree, pPathToRename, psz_arg2, SG_FALSE)  );

	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_PATHNAME_NULLFREE(pCtx, pPathToRename);
	SG_VARRAY_NULLFREE(pCtx, pva_options);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	*rval = JSVAL_VOID;
	return JS_TRUE;

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_PATHNAME_NULLFREE(pCtx, pPathToRename);
	SG_VARRAY_NULLFREE(pCtx, pva_options);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    /* TODO */
    return JS_FALSE;
}

/*  usage: sg.move([options], file_or_folder, new_destination)
 */
SG_JSGLUE_METHOD_PROTOTYPE(pending_tree, move)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_pendingtree* pPendingTree = NULL;
	SG_pathname* pPath = NULL;
	SG_pathname* pPathToMove = NULL;
	SG_pathname* pPathMoveToLocation = NULL;
	SG_pathname ** newArgv = NULL;
	SG_pathname * pPathnameTemp = NULL;
	const char* psz_arg2 = NULL;
	uint n = 0;
	uint i = 0;
	SG_uint32 nCountItems = 0;
	JSObject* jso = NULL;
	jsval arg0 = argv[0];
	SG_varray* pva_options = NULL;
	SG_bool bExists = SG_FALSE;
	SG_fsobj_type t;
	SG_uint32 errorCount = 0;
	SG_error nLastError = 0;

	SG_UNUSED(obj);

	jso = JSVAL_TO_OBJECT(arg0);

    if (argc > 0 && JS_IsArrayObject(cx, jso))
    {
		 SG_uint32 count = 0;
		 SG_uint32 j = 0;
		 const char* pszOption;

         SG_ERR_CHECK(  sg_jsglue__jsobject_to_varray(pCtx, cx, jso, &pva_options)  );

		 SG_ERR_CHECK(  SG_varray__count(pCtx, pva_options, &count)  );

		 for (j = 0; j < count; j++)
		 {
			 SG_ERR_CHECK( SG_varray__get__sz(pCtx, pva_options, j, &pszOption) );

			 // TODO 2011/05/09 Parse any options....

			 if (pszOption && *pszOption)	// this test is dumb, but it makes MSVC happy
				 SG_ERR_THROW2(  SG_ERR_INVALIDARG,
								 (pCtx, "sg.pending_tree.move(['%s'...)", pszOption)  );
		}
		n++;
	}

	SG_JS_BOOL_CHECK( (argc - n) >= 2  );


	//allocate the new string array that we will send down.
	SG_ERR_CHECK(  SG_alloc(pCtx, argc, sizeof(SG_pathname*), &newArgv)  );

	for (i = n; i < argc - 1; i++)
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[i])  );
		SG_ERR_CHECK( SG_PATHNAME__ALLOC__SZ(pCtx, &pPathnameTemp, JS_GetStringBytes(JSVAL_TO_STRING(argv[i])))  );
		newArgv[i] = pPathnameTemp;
		nCountItems++;
	}

	psz_arg2 = JS_GetStringBytes(JSVAL_TO_STRING(argv[argc - 1]));
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPathMoveToLocation, psz_arg2)  );

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC(pCtx, &pPath)  );
	SG_ERR_CHECK(  SG_pathname__set__from_cwd(pCtx, pPath)  );

	/* First make sure the paths exist */
	SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pPathMoveToLocation, &bExists, &t, NULL)  );

	if (!bExists || t != SG_FSOBJ_TYPE__DIRECTORY)
	{
		SG_ERR_THROW(  SG_ERR_NOT_A_DIRECTORY  );
	}
	for (i = 0; i < nCountItems; i ++)
	{
		SG_ERR_CHECK(  SG_pendingtree__alloc(pCtx, pPath, &pPendingTree)  );
		SG_pendingtree__move(pCtx, pPendingTree, (const SG_pathname **)&newArgv[i], 1, pPathMoveToLocation);
		if (SG_context__has_err(pCtx))
		{
			if (i == nCountItems - 1)
				SG_ERR_CHECK_CURRENT;  //We're on the last item to move, just throw the error.
			errorCount++;
			SG_context__get_err(pCtx, &nLastError);
			SG_context__err_to_console(pCtx, SG_CS_STDERR);
			SG_context__err_reset(pCtx);
		}
		SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	}

	if (errorCount == 1)
		SG_ERR_THROW2( nLastError, (pCtx, "1 error encountered." ) );
	if (errorCount > 1)
		SG_ERR_THROW2( nLastError, (pCtx, "%d errors encountered.", errorCount) );
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_PATHNAME_NULLFREE(pCtx, pPathToMove);
	SG_PATHNAME_NULLFREE(pCtx, pPathMoveToLocation);
	SG_VARRAY_NULLFREE(pCtx, pva_options);

	for (i = 0; i < nCountItems; i++)
	{
		SG_PATHNAME_NULLFREE(pCtx, newArgv[i]);
	}
	SG_NULLFREE(pCtx, newArgv);
	*rval = JSVAL_VOID;
	return JS_TRUE;

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_PATHNAME_NULLFREE(pCtx, pPathToMove);
	SG_PATHNAME_NULLFREE(pCtx, pPathMoveToLocation);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_VARRAY_NULLFREE(pCtx, pva_options);
	for (i = 0; i < nCountItems; i++)
	{
		SG_PATHNAME_NULLFREE(pCtx, newArgv[i]);
	}
	SG_NULLFREE(pCtx, newArgv);

	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    /* TODO */
    return JS_FALSE;
}

/*  usage: ???
 *
 *  possible options: ????
 */
SG_JSGLUE_METHOD_PROTOTYPE(pending_tree, scan)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_pendingtree* pPendingTree = NULL;
	SG_pathname* pPath = NULL;
	JSObject* jso = NULL;
	SG_varray* pva_options = NULL;
	SG_file_spec* pFilespec = NULL;
	SG_timestamp_cache * pTSC = NULL;			// we do not own this

	SG_UNUSED(obj);

	SG_ERR_CHECK(  SG_FILE_SPEC__ALLOC(pCtx, &pFilespec)  );

	if (argc > 0)
	{
		jsval arg0 = argv[0];

		jso = JSVAL_TO_OBJECT(arg0);

		if (JS_IsArrayObject(cx, jso))
		{
			 SG_uint32 count = 0;
			 SG_uint32 j = 0;
			 const char* pszOption;

			 SG_ERR_CHECK(  sg_jsglue__jsobject_to_varray(pCtx, cx, jso, &pva_options)  );

			 SG_ERR_CHECK(  SG_varray__count(pCtx, pva_options, &count)  );

			 for (j = 0; j < count; j++)
			 {
				 SG_ERR_CHECK( SG_varray__get__sz(pCtx, pva_options, j, &pszOption) );
				//TODO scan options
				if ( sg_jsglue__is_filespec_arg(pszOption) )
				{
					 SG_ERR_CHECK (  sg_jsglue__process_filespec_arg(pCtx, pszOption, pFilespec)  );
				}
			 }
		}
	}
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC(pCtx, &pPath)  );
	SG_ERR_CHECK(  SG_pathname__set__from_cwd(pCtx, pPath)  );

	SG_ERR_CHECK(  SG_pendingtree__alloc(pCtx, pPath, &pPendingTree)  );

	// start scan with an empty timestamp_cache and rebuild it.
	//
	// TODO 2010/08/11 should we only flush the cache when given a --force, like vv.exe?

	SG_ERR_CHECK(  SG_pendingtree__get_timestamp_cache(pCtx, pPendingTree, &pTSC)  );
	SG_ERR_CHECK(  SG_timestamp_cache__remove_all(pCtx, pTSC)  );

	SG_ERR_CHECK(  SG_pendingtree__scan(pCtx, pPendingTree, SG_TRUE, SG_TRUE, SG_TRUE, pFilespec)  );

	SG_FILE_SPEC_NULLFREE(pCtx, pFilespec);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_VARRAY_NULLFREE(pCtx, pva_options);

	*rval = JSVAL_VOID;
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);

	return JS_TRUE;

fail:
	SG_FILE_SPEC_NULLFREE(pCtx, pFilespec);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_VARRAY_NULLFREE(pCtx, pva_options);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    /* TODO */

    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(pending_tree, detach)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_pathname* pPath = NULL;
    SG_pathname* pPath_top = NULL;

	SG_UNUSED(obj);
	SG_UNUSED(argc);
	SG_UNUSED(argv);

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC(pCtx, &pPath)  );
	SG_ERR_CHECK(  SG_pathname__set__from_cwd(pCtx, pPath)  );
	SG_ERR_CHECK(  SG_workingdir__find_mapping(pCtx, pPath, &pPath_top, NULL, NULL)  );
    SG_ERR_CHECK(  SG_workingdir__remove_branch_info(pCtx, pPath_top)  );

	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_PATHNAME_NULLFREE(pCtx, pPath_top);

    *rval = JSVAL_VOID;

	return JS_TRUE;

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_PATHNAME_NULLFREE(pCtx, pPath_top);
	return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(pending_tree, attach)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_bool bNewBranch = SG_FALSE;
	const char* pszBranchName;
	SG_repo* pRepo = NULL;
	SG_vhash* pvh_pile = NULL;
	SG_pathname* pPath = NULL;
	SG_pathname* pPath_top = NULL;


	SG_UNUSED(obj);

	SG_JS_BOOL_CHECK(  1 == argc || 2 == argc  );
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
	if ( 2 == argc )
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_BOOLEAN(argv[1])  );
		bNewBranch = JSVAL_TO_BOOLEAN(argv[1]);
	}


	pszBranchName = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));

	SG_ERR_CHECK(  sg_jsglue__get_repo_from_cwd(pCtx, &pRepo, NULL, NULL)  );
	SG_ERR_CHECK(  SG_vc_branches__cleanup(pCtx, pRepo, &pvh_pile)  );
	{
		SG_bool bHas;
		SG_vhash* pvhRefBranches;
		SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvh_pile, "branches", &pvhRefBranches)  );
		SG_ERR_CHECK(  SG_vhash__has(pCtx, pvhRefBranches, pszBranchName, &bHas)  );
		if ( !bNewBranch && !bHas)
			SG_ERR_THROW2(  SG_ERR_BRANCH_NOT_FOUND, (pCtx, "Branch %s does not exist.", pszBranchName)  );
	}

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC(pCtx, &pPath)  );
	SG_ERR_CHECK(  SG_pathname__set__from_cwd(pCtx, pPath)  );
	SG_ERR_CHECK(  SG_workingdir__find_mapping(pCtx, pPath, &pPath_top, NULL, NULL)  );
	SG_ERR_CHECK(  SG_workingdir__set_branch_info(pCtx, pPath_top, pszBranchName)  );

	SG_REPO_NULLFREE(pCtx, pRepo);
	SG_VHASH_NULLFREE(pCtx, pvh_pile);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_PATHNAME_NULLFREE(pCtx, pPath_top);

	*rval = JSVAL_VOID;

	return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	SG_REPO_NULLFREE(pCtx, pRepo);
	SG_VHASH_NULLFREE(pCtx, pvh_pile);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_PATHNAME_NULLFREE(pCtx, pPath_top);
	return JS_FALSE;
}

/*  usage: sg.stamp([options])
 *
 *  possible options: stamp=STAMPTOAPPLY, rev=CSID, tag=TAG, remove
 */
SG_JSGLUE_METHOD_PROTOTYPE(pending_tree, stamp)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_stringarray * pasz_stamps = NULL;
	SG_stringarray * pasz_revisions = NULL;
	SG_bool bRemove = SG_FALSE;
	JSObject* jso = NULL;
	SG_varray* pva_options = NULL;

	SG_repo * pRepo = NULL;

	SG_uint32 nRevisionCount = 0;
	SG_uint32 nStampsCount = 0;
	SG_uint32 nStampsIndex = 0;
	SG_uint32 nRevisionIndex = 0;
	SG_audit q;
	const char * psz_revision = NULL;
	const char * pszCurrentStamp = NULL;

	SG_UNUSED(obj);
	SG_UNUSED(rval);

	SG_ERR_CHECK(  sg_jsglue__get_repo_from_cwd(pCtx, &pRepo, NULL, NULL)  );

	SG_ERR_CHECK(  SG_STRINGARRAY__ALLOC(pCtx, &pasz_revisions, 1)  );
	SG_ERR_CHECK(  SG_STRINGARRAY__ALLOC(pCtx, &pasz_stamps, 1)  );
	if (argc > 0)
	{
		jsval arg0 = argv[0];
		jso = JSVAL_TO_OBJECT(arg0);

		if (JS_IsArrayObject(cx, jso))
		{
			 SG_uint32 count = 0;
			 SG_uint32 j = 0;
			 const char* pszOption;

			 SG_ERR_CHECK(  sg_jsglue__jsobject_to_varray(pCtx, cx, jso, &pva_options)  );

			 SG_ERR_CHECK(  SG_varray__count(pCtx, pva_options, &count)  );

			 for (j = 0; j < count; j++)
			 {
				 const char* pszVal = NULL;
				 SG_ERR_CHECK( SG_varray__get__sz(pCtx, pva_options, j, &pszOption) );

				 if ( strcmp(pszOption, "remove")==0)
				 {
					 bRemove = SG_TRUE;
				 }
				 else if ( sg_jsglue_get_equals_flag (pszOption, "stamp", &pszVal)  )
				 {
					 SG_ERR_CHECK(  SG_stringarray__add(pCtx, pasz_stamps, pszVal)  );
				 }
				 else if ( sg_jsglue_get_equals_flag (pszOption, "rev", &pszVal) )
				{
					 char * psz_rev = NULL;
					 SG_ERR_CHECK(  SG_repo__hidlookup__dagnode(pCtx, pRepo, SG_DAGNUM__VERSION_CONTROL, pszVal, &psz_rev)  );
					if (psz_rev != NULL)
					{

						SG_ERR_CHECK(  SG_stringarray__add(pCtx, pasz_revisions, psz_rev)  );
					}
					SG_NULLFREE(pCtx, psz_rev);
				}
				else if ( sg_jsglue_get_equals_flag (pszOption, "tag", &pszVal) )
				{
					char * psz_rev = NULL;
					SG_ERR_CHECK(  SG_vc_tags__lookup__tag(pCtx, pRepo, pszVal, &psz_rev)  );
					SG_ERR_CHECK(  SG_stringarray__add(pCtx, pasz_revisions, psz_rev)  );
					SG_NULLFREE(pCtx, psz_rev);
				}
			 }
		}
	}

	SG_ERR_CHECK(  SG_stringarray__count(pCtx, pasz_revisions, &nRevisionCount)  );
	SG_ERR_CHECK(  SG_stringarray__count(pCtx, pasz_stamps, &nStampsCount)  );
	if (nStampsCount == 0 && !bRemove)
				SG_ERR_THROW2(  SG_ERR_INVALIDARG,
				                           (pCtx, "Must include at least one stamp.")  );

	if (nRevisionCount == 0)
		SG_ERR_THROW2(  SG_ERR_INVALIDARG,
								   (pCtx, "Must include at least one changeset using either rev or tag")  );

	SG_ERR_CHECK(  SG_audit__init(pCtx,&q,pRepo,SG_AUDIT__WHEN__NOW,SG_AUDIT__WHO__FROM_SETTINGS)  );

	for (nRevisionIndex = 0; nRevisionIndex < nRevisionCount; nRevisionIndex++)
	{
		SG_ERR_CHECK(  SG_stringarray__get_nth(pCtx, pasz_revisions, nRevisionIndex, &psz_revision)  );

		if (psz_revision != NULL && !bRemove)
		{
			for (nStampsIndex = 0; nStampsIndex < nStampsCount; nStampsIndex++)
			{
				SG_ERR_CHECK(  SG_stringarray__get_nth(pCtx, pasz_stamps, nStampsIndex, &pszCurrentStamp)  );
				SG_ERR_CHECK(  SG_vc_stamps__add(pCtx, pRepo, psz_revision, pszCurrentStamp, &q, NULL)  );
			}
		}
		else if (psz_revision != NULL && bRemove)
		{
			const char * const * pasz_char_stamps = NULL;
			if (nStampsCount > 0)
				SG_ERR_CHECK(  SG_stringarray__sz_array(pCtx, pasz_stamps, &pasz_char_stamps)  );
			//Remove all stamps from this revision.
			SG_ERR_CHECK(  SG_vc_stamps__remove(pCtx, pRepo, &q, (const char *)psz_revision, nStampsCount, pasz_char_stamps)  );
		}
	}

	SG_VARRAY_NULLFREE(pCtx, pva_options);
	SG_STRINGARRAY_NULLFREE(pCtx, pasz_revisions);
	SG_STRINGARRAY_NULLFREE(pCtx, pasz_stamps);
	SG_REPO_NULLFREE(pCtx, pRepo);
	return JS_TRUE;

fail:
	SG_VARRAY_NULLFREE(pCtx, pva_options);
	SG_STRINGARRAY_NULLFREE(pCtx, pasz_revisions);
	SG_STRINGARRAY_NULLFREE(pCtx, pasz_stamps);
	SG_REPO_NULLFREE(pCtx, pRepo);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    /* TODO */
    return JS_FALSE;
}

/*  usage: sg.stamps()
 *
 */
SG_JSGLUE_METHOD_PROTOTYPE(pending_tree, stamps)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_repo * pRepo = NULL;
	SG_varray* pVArray = NULL;
	JSObject* jso = NULL;

	SG_UNUSED(obj);
	SG_UNUSED(argc);
	SG_UNUSED(argv);

	SG_ERR_CHECK(  sg_jsglue__get_repo_from_cwd(pCtx, &pRepo, NULL, NULL)  );

	SG_ERR_CHECK(  SG_vc_stamps__list_all_stamps(pCtx, pRepo, &pVArray)  );

    SG_JS_NULL_CHECK(  (jso = JS_NewArrayObject(cx, 0, NULL))  );
    *rval = OBJECT_TO_JSVAL(jso);
    SG_ERR_CHECK(  sg_jsglue__copy_varray_into_jsobject(pCtx, cx, pVArray, jso)  );
	SG_VARRAY_NULLFREE(pCtx, pVArray);

	SG_REPO_NULLFREE(pCtx, pRepo);
	return JS_TRUE;

fail:
	SG_VARRAY_NULLFREE(pCtx, pVArray);
	SG_REPO_NULLFREE(pCtx, pRepo);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	/* TODO */
	return JS_FALSE;
}

/*  usage: sg.tag([options])
 *
 *  possible options: tag=EXISTINGTAG, rev=CSID, tag=TAG, remove, force
 */
SG_JSGLUE_METHOD_PROTOTYPE(pending_tree, tag)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_bool bRemove = SG_FALSE;
	SG_bool bForce = SG_FALSE;
	JSObject* jso = NULL;
	SG_varray* pva_options = NULL;
	SG_pendingtree * pPendingTree = NULL;
	SG_rev_spec* pRevSpec = NULL;
	SG_repo * pRepo = NULL;
	uint n = 0;
	char * pszSelectedCSID = NULL;
	uint i = 0;
	SG_uint32 nCountItems = 0;
	const char** newArgv = NULL;
	SG_pathname* pPathCwd = NULL;
	SG_uint32 countRevSpecs;
	
	SG_UNUSED(obj);
	SG_UNUSED(rval);

	SG_ERR_CHECK(  sg_jsglue__get_repo_from_cwd(pCtx, &pRepo, &pPathCwd, NULL)  );
	SG_ERR_CHECK(  SG_REV_SPEC__ALLOC(pCtx, &pRevSpec)  );

	if (argc > 0)
	{
		jsval arg0 = argv[0];
		jso = JSVAL_TO_OBJECT(arg0);

		if (JS_IsArrayObject(cx, jso))
		{
			SG_uint32 count = 0;
			SG_uint32 j = 0;
			const char* pszOption;

			SG_ERR_CHECK(  sg_jsglue__jsobject_to_varray(pCtx, cx, jso, &pva_options)  );

			SG_ERR_CHECK(  SG_varray__count(pCtx, pva_options, &count)  );

			for (j = 0; j < count; j++)
			{
				const char* pszVal = NULL;
				SG_ERR_CHECK( SG_varray__get__sz(pCtx, pva_options, j, &pszOption) );

				if ( strcmp(pszOption, "remove")==0)
				{
					bRemove = SG_TRUE;
				}
				else if ( strcmp(pszOption, "force")==0)
				{
					bForce = SG_TRUE;
				}
				else if ( sg_jsglue_get_equals_flag (pszOption, "rev", &pszVal) )
				{
					SG_ERR_CHECK(  SG_rev_spec__add_rev(pCtx, pRevSpec, pszVal)  );
				}
				else if ( sg_jsglue_get_equals_flag (pszOption, "tag", &pszVal) )
				{
					SG_ERR_CHECK(  SG_rev_spec__add_tag(pCtx, pRevSpec, pszVal)  );
				}
			}
			n++;
		}
	}

	SG_ERR_CHECK(  SG_rev_spec__count(pCtx, pRevSpec, &countRevSpecs)  );
	if (countRevSpecs > 1)
		SG_ERR_THROW2(SG_ERR_USAGE, (pCtx, "A tag can only be applied to one changeset.")  );
	if (countRevSpecs == 1)
		SG_ERR_CHECK(  SG_rev_spec__get_one__repo(pCtx, pRepo, pRevSpec, &pszSelectedCSID, NULL)  );
	else
	{
		const SG_varray * pvaRefParents;
		SG_uint32 countParents;
		const char * pszRefHidParent;
		SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx, pPathCwd, &pPendingTree)  );
		SG_ERR_CHECK(  SG_pendingtree__get_wd_parents__ref(pCtx, pPendingTree, &pvaRefParents)  );
		SG_ERR_CHECK(  SG_varray__count(pCtx, pvaRefParents, &countParents)  );
		if (countParents > 1)
			SG_ERR_THROW(  SG_ERR_CANNOT_DO_WHILE_UNCOMMITTED_MERGE  );
		SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pvaRefParents, 0, &pszRefHidParent)  );
		SG_ERR_CHECK(  SG_STRDUP(pCtx, pszRefHidParent, &pszSelectedCSID)  );
	}

	//allocate the new string array that we will send down.
	SG_ERR_CHECK(  SG_alloc(pCtx, argc, sizeof(char*), &newArgv)  );

	for (i = n; i < argc; i++)
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[i])  );
		newArgv[nCountItems] = JS_GetStringBytes(JSVAL_TO_STRING(argv[i]));
		nCountItems++;
	}

	if (bRemove)
	{
		SG_ERR_CHECK(  SG_tag__remove(pCtx, pRepo, countRevSpecs ? pszSelectedCSID : NULL, newArgv, nCountItems)  );
	}
	else
	{
		SG_ERR_CHECK(  SG_tag__add(pCtx, pRepo, pszSelectedCSID, newArgv, nCountItems, bForce)  );
	}

	SG_NULLFREE(pCtx, newArgv);
	SG_REV_SPEC_NULLFREE(pCtx, pRevSpec);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_VARRAY_NULLFREE(pCtx, pva_options);
	SG_REPO_NULLFREE(pCtx, pRepo);
	SG_NULLFREE(pCtx, pszSelectedCSID);
	SG_REV_SPEC_NULLFREE(pCtx, pRevSpec);
	SG_PATHNAME_NULLFREE(pCtx, pPathCwd);
	return JS_TRUE;

fail:
	SG_NULLFREE(pCtx, newArgv);
	SG_REV_SPEC_NULLFREE(pCtx, pRevSpec);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_VARRAY_NULLFREE(pCtx, pva_options);
	SG_REPO_NULLFREE(pCtx, pRepo);
	SG_NULLFREE(pCtx, pszSelectedCSID);
	SG_REV_SPEC_NULLFREE(pCtx, pRevSpec);
	SG_PATHNAME_NULLFREE(pCtx, pPathCwd);
	SG_jsglue__report_sg_error(pCtx,cx);    // DO NOT SG_ERR_IGNORE() THIS
	/* TODO */
	return JS_FALSE;
}


/*  usage: sg.tags()
 *
 */
SG_JSGLUE_METHOD_PROTOTYPE(pending_tree, tags)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_repo * pRepo = NULL;
	SG_varray* pVArray = NULL;
	JSObject* jso = NULL;
	const char * psz_tag = NULL;
	const char * psz_csid = NULL;
	SG_rbtree * prbtree = NULL;
	SG_rbtree_iterator * pIter = NULL;
	SG_bool b = SG_FALSE;
	SG_vhash * pvh_current = NULL;

	SG_UNUSED(obj);
	SG_UNUSED(argc);
	SG_UNUSED(argv);

	SG_ERR_CHECK(  sg_jsglue__get_repo_from_cwd(pCtx, &pRepo, NULL, NULL)  );

	SG_ERR_CHECK(  SG_vc_tags__list(pCtx, pRepo, &prbtree)  );
	SG_ERR_CHECK(  SG_VARRAY__ALLOC(pCtx, &pVArray)  );

	if (prbtree)
	{
		SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pIter, prbtree, &b, &psz_tag, (void**)&psz_csid)  );

		while (b)
		{
			SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvh_current)  );
			SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvh_current, "tag", psz_tag) );
			SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvh_current, "csid", psz_csid) );
			SG_ERR_CHECK(  SG_varray__append__vhash(pCtx, pVArray, &pvh_current)  );
			SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pIter, &b, &psz_tag, (void**)&psz_csid)  );
		}
		SG_rbtree__iterator__free(pCtx, pIter);
		pIter = NULL;
		SG_RBTREE_NULLFREE(pCtx, prbtree);
	}

    SG_JS_NULL_CHECK(  (jso = JS_NewArrayObject(cx, 0, NULL))  );
    *rval = OBJECT_TO_JSVAL(jso);
    SG_ERR_CHECK(  sg_jsglue__copy_varray_into_jsobject(pCtx, cx, pVArray, jso)  );
    SG_VARRAY_NULLFREE(pCtx, pVArray);

	SG_RBTREE_NULLFREE(pCtx, prbtree);
	SG_REPO_NULLFREE(pCtx, pRepo);
	return JS_TRUE;

fail:
	SG_RBTREE_NULLFREE(pCtx, prbtree);
	SG_VARRAY_NULLFREE(pCtx, pVArray);
	SG_REPO_NULLFREE(pCtx, pRepo);
	SG_VHASH_NULLFREE(pCtx, pvh_current);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	/* TODO */
	return JS_FALSE;
}

/*  usage: sg.status([options], [files,folders])
 *
 *  possible options: nonrecursive, include=pattern, exclude=pattern
 */
SG_JSGLUE_METHOD_PROTOTYPE(pending_tree, status)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_pendingtree* pPendingTree = NULL;
	SG_pathname* pPath = NULL;
	SG_treediff2* pTreeDiff = NULL;
	SG_vhash* pVHash = NULL;
	const char** newArgv = NULL;
	uint i;
	JSObject* jso = NULL;
	uint n = 0;

	SG_bool bRecursive = SG_TRUE;
	SG_uint32 nCountItems = 0;
	SG_varray* pva_options = NULL;
	SG_file_spec* pFilespec = NULL;
	char * psz_cset_1 = NULL;
	char * psz_cset_2 = NULL;

	SG_UNUSED(obj);

	SG_ERR_CHECK(  SG_FILE_SPEC__ALLOC(pCtx, &pFilespec)  );

	if (argc > 0)
	{
		jsval arg0 = argv[0];
		jso = JSVAL_TO_OBJECT(arg0);

		if (JS_IsArrayObject(cx, jso))
		{
			SG_uint32 count = 0;
			SG_uint32 j = 0;
			const char* pszOption;

			SG_ERR_CHECK(  sg_jsglue__jsobject_to_varray(pCtx, cx, jso, &pva_options)  );

			SG_ERR_CHECK(  SG_varray__count(pCtx, pva_options, &count)  );

			 for (j = 0; j < count; j++)
			 {
				const char* pszVal = NULL;

				SG_ERR_CHECK( SG_varray__get__sz(pCtx, pva_options, j, &pszOption) );

				if ( strcmp(pszOption, "nonrecursive")==0)
				{
					bRecursive = SG_FALSE;
				}
				else if ( sg_jsglue__is_filespec_arg(pszOption) )
				{
					SG_ERR_CHECK (  sg_jsglue__process_filespec_arg(pCtx, pszOption, pFilespec)  );
				}
				else if ( sg_jsglue_get_equals_flag(pszOption, "rev", &pszVal) )
				{
					// TODO 2010/06/03 Should we allow rev prefixes like we do in the command line client?
					if (psz_cset_1 == NULL)
						SG_ERR_CHECK(  SG_STRDUP(pCtx, pszVal, &psz_cset_1)  );
					else if (psz_cset_2 == NULL)
						SG_ERR_CHECK(  SG_STRDUP(pCtx, pszVal, &psz_cset_2)  );
					else
					{
						// ignore 3rd and beyond...
					}
				}

				// TODO 2010/06/03 Should we allow "tag" like we do in the command line client?
			}
			n++;
		}
	}
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC(pCtx, &pPath)  );
	SG_ERR_CHECK(  SG_pathname__set__from_cwd(pCtx, pPath)  );

	SG_ERR_CHECK(  SG_pendingtree__alloc(pCtx, pPath, &pPendingTree)  );
	SG_ERR_CHECK(  SG_pendingtree__diff_or_status(pCtx, pPendingTree, psz_cset_1, psz_cset_2, &pTreeDiff)  );

	//allocate the new string array that we will send down.
	SG_ERR_CHECK(  SG_alloc(pCtx, argc, sizeof(char*), &newArgv)  );
	for (i = n; i < argc; i++)
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[i])  );
		newArgv[nCountItems] = JS_GetStringBytes(JSVAL_TO_STRING(argv[i]));
		nCountItems++;
	}

    //  if ((nCountItems > 0) || (count_includes > 0) || (count_excludes > 0))
		SG_ERR_CHECK(  SG_treediff2__file_spec_filter(pCtx, pTreeDiff,
													    nCountItems, newArgv, bRecursive,
														pFilespec)  );

	SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pVHash)   );
	SG_ERR_CHECK(  SG_treediff2__report_status_to_vhash(pCtx, pTreeDiff, pVHash, SG_TRUE )   );

    SG_JS_NULL_CHECK(  (jso = JS_NewObject(cx, NULL, NULL, NULL))  );
    *rval = OBJECT_TO_JSVAL(jso);
    SG_ERR_CHECK(  sg_jsglue__copy_vhash_into_jsobject(pCtx, cx, pVHash, jso)  );
    SG_VHASH_NULLFREE(pCtx, pVHash);

	SG_FILE_SPEC_NULLFREE(pCtx, pFilespec);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_TREEDIFF2_NULLFREE(pCtx, pTreeDiff);
	SG_VARRAY_NULLFREE(pCtx, pva_options);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_NULLFREE(pCtx, psz_cset_1);
	SG_NULLFREE(pCtx, psz_cset_2);
	SG_NULLFREE(pCtx, newArgv);
	return JS_TRUE;

fail:
	SG_FILE_SPEC_NULLFREE(pCtx, pFilespec);
	SG_VHASH_NULLFREE(pCtx, pVHash);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_TREEDIFF2_NULLFREE(pCtx, pTreeDiff);
	SG_VARRAY_NULLFREE(pCtx, pva_options);
	SG_NULLFREE(pCtx, newArgv);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_NULLFREE(pCtx, psz_cset_1);
	SG_NULLFREE(pCtx, psz_cset_2);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    /* TODO */
    return JS_FALSE;
}

/*  usage: sg.status([options], [files,folders])
 *
 *  possible options: nonrecursive, include=pattern, exclude=pattern
 *
 *  TODO 2010/06/03 this usage list is out of date.
 */
SG_JSGLUE_METHOD_PROTOTYPE(pending_tree, status_string)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_pendingtree* pPendingTree = NULL;
	SG_pathname* pPath = NULL;
	SG_treediff2* pTreeDiff = NULL;
	SG_string* pStrReport = NULL;
	const char** newArgv = NULL;
	uint i;
	uint n = 0;
	JSObject* jso = NULL;

	SG_bool bRecursive = SG_TRUE;
	SG_uint32 nCountItems = 0;
	SG_varray* pva_options = NULL;
	SG_file_spec* pFilespec = NULL;
	char * psz_cset_1 = NULL;
	char * psz_cset_2 = NULL;

	SG_bool bHaveIssuesIn = SG_FALSE;
	const SG_varray * pvaIssuesIn = NULL;
	SG_vhash * pVhash = NULL;

	SG_UNUSED(obj);

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC(pCtx, &pPath)  );
	SG_ERR_CHECK(  SG_pathname__set__from_cwd(pCtx, pPath)  );

	SG_ERR_CHECK(  SG_FILE_SPEC__ALLOC(pCtx, &pFilespec)  );

	if (argc > 0)
	{
		jsval arg0 = argv[0];
		jso = JSVAL_TO_OBJECT(arg0);

		if (JS_IsArrayObject(cx, jso))
		{
			SG_uint32 count = 0;
			SG_uint32 j = 0;
			const char* pszOption;

			SG_ERR_CHECK(  sg_jsglue__jsobject_to_varray(pCtx, cx, jso, &pva_options)  );

			SG_ERR_CHECK(  SG_varray__count(pCtx, pva_options, &count)  );

			for (j = 0; j < count; j++)
			{
				const char* pszVal = NULL;
				SG_ERR_CHECK( SG_varray__get__sz(pCtx, pva_options, j, &pszOption) );

				if ( strcmp(pszOption, "nonrecursive")==0)
				{
					bRecursive = SG_FALSE;
				}
				else if ( sg_jsglue__is_filespec_arg(pszOption) )
				{
					SG_ERR_CHECK (  sg_jsglue__process_filespec_arg(pCtx, pszOption, pFilespec)  );
				}
				else if ( sg_jsglue_get_equals_flag(pszOption, "rev", &pszVal) )
				{
					// TODO 2010/06/03 Should these be prefix-lookups like we do for the command line client?
					if (psz_cset_1 == NULL)
						SG_ERR_CHECK(  SG_STRDUP(pCtx, pszVal, &psz_cset_1)  );
					else if (psz_cset_2 == NULL)
						SG_ERR_CHECK(  SG_STRDUP(pCtx, pszVal, &psz_cset_2)  );
					else
					{
						// ignore 3rd and beyond...
					}
				}

				// TODO 2010/06/03 Should we allow "tag" options like we do for the command line client?
			}
			n++;
		}
	}

	SG_ERR_CHECK(  SG_pendingtree__alloc(pCtx, pPath, &pPendingTree)  );
	SG_ERR_CHECK(  SG_pendingtree__diff_or_status(pCtx, pPendingTree, psz_cset_1, psz_cset_2, &pTreeDiff)  );

	//allocate the new string array that we will send down.
	SG_ERR_CHECK(  SG_alloc(pCtx, argc, sizeof(char*), &newArgv)  );
	for (i = n; i < argc; i++)
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[i])  );
		newArgv[i] = JS_GetStringBytes(JSVAL_TO_STRING(argv[i]));
		nCountItems++;
	}

	SG_ERR_CHECK(  SG_treediff2__file_spec_filter(pCtx, pTreeDiff,
													nCountItems, newArgv, bRecursive,
													pFilespec)  );

	// __diff_or_status() freed pPendingTree->pSuperRoot as part of that
	// one operation per pendingtree issue.  reload it here so that we
	// can _status_formatter__ can compute live repo-paths for things.

	SG_ERR_CHECK(  SG_pendingtree__reload_pendingtree_from_vfile(pCtx, pPendingTree)  );

	if (psz_cset_1 == NULL && psz_cset_2 == NULL)
		SG_ERR_CHECK(  SG_pendingtree__get_wd_issues__ref(pCtx, pPendingTree, &bHaveIssuesIn, &pvaIssuesIn)  );

	SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pVhash)  );
	SG_ERR_CHECK(  SG_treediff2__report_status_to_vhash(pCtx, pTreeDiff, pVhash, SG_TRUE)  );
	SG_ERR_CHECK(  SG_status_formatter__vhash_to_string(pCtx, pVhash, pvaIssuesIn, pPendingTree, NULL, NULL, &pStrReport)  );
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);

	JSVAL_FROM_SZ(*rval, SG_string__sz(pStrReport));

	SG_FILE_SPEC_NULLFREE(pCtx, pFilespec);
	SG_VHASH_NULLFREE(pCtx, pVhash);
	SG_STRING_NULLFREE(pCtx, pStrReport);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_TREEDIFF2_NULLFREE(pCtx, pTreeDiff);
	SG_VARRAY_NULLFREE(pCtx, pva_options);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_NULLFREE(pCtx, psz_cset_1);
	SG_NULLFREE(pCtx, psz_cset_2);
	SG_NULLFREE(pCtx, newArgv);
	return JS_TRUE;

fail:
	SG_FILE_SPEC_NULLFREE(pCtx, pFilespec);
	SG_VHASH_NULLFREE(pCtx, pVhash);
	SG_STRING_NULLFREE(pCtx, pStrReport);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_TREEDIFF2_NULLFREE(pCtx, pTreeDiff);
	SG_NULLFREE(pCtx, newArgv);
	SG_VARRAY_NULLFREE(pCtx, pva_options);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_NULLFREE(pCtx, psz_cset_1);
	SG_NULLFREE(pCtx, psz_cset_2);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    /* TODO */
    return JS_FALSE;
}


/*  usage: sg.history([options], [files,folders])
 *
 *  possible options: leaves, rev=CHANGESET, user=USERNAME, max=MAXRESULTS, from=MINIMUMDATE, to=MAXIMUMDATE
 */
SG_JSGLUE_METHOD_PROTOTYPE(pending_tree, history)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_treediff2* pTreeDiff = NULL;
	SG_history_result* pHistResult = NULL;
	const char** newArgv = NULL;
	uint i;
	JSObject* jso = NULL;
	JSObject* jsoReturn = NULL;
	uint n = 0;

	SG_bool bLeaves = SG_FALSE;
	SG_uint32 nCountItems = 0;
	SG_varray* pva_options = NULL;
	SG_rev_spec* pRevSpec = NULL;

	SG_int64 maxDate = SG_INT64_MAX;
	SG_int64 minDate = 0;
	SG_uint32 maxResults = SG_UINT32_MAX;
	const char * pszUser = NULL;
	const char * pszStamp = NULL;
	SG_pathname * pCwd = NULL;
	SG_bool bHideMerges = SG_FALSE;

	SG_UNUSED(obj);

	SG_ERR_CHECK(  SG_REV_SPEC__ALLOC(pCtx, &pRevSpec)  );

	if (argc > 0)
	{
		jsval arg0 = argv[0];
		jso = JSVAL_TO_OBJECT(arg0);

		if (JS_IsArrayObject(cx, jso))
		{
			SG_uint32 count = 0;
			SG_uint32 j = 0;
			const char* pszOption;
			const char* pszVal;

			SG_ERR_CHECK(  sg_jsglue__jsobject_to_varray(pCtx, cx, jso, &pva_options)  );

			SG_ERR_CHECK(  SG_varray__count(pCtx, pva_options, &count)  );

			 for (j = 0; j < count; j++)
			 {
				SG_ERR_CHECK( SG_varray__get__sz(pCtx, pva_options, j, &pszOption) );
				if ( strcmp(pszOption, "leaves")==0)
				{
					bLeaves = SG_TRUE;
				}
				else if ( sg_jsglue_get_equals_flag (pszOption, "from", &pszVal) )
				{
					SG_ERR_CHECK(  SG_int64__parse__strict(pCtx, &minDate, pszVal)  );
				}
				else if ( sg_jsglue_get_equals_flag (pszOption, "to", &pszVal) )
				{
					SG_ERR_CHECK(  SG_int64__parse__strict(pCtx, &maxDate, pszVal)  );
				}
				else if ( sg_jsglue_get_equals_flag (pszOption, "rev", &pszVal) )
				{
					SG_ERR_CHECK(  SG_rev_spec__add_rev(pCtx, pRevSpec, pszVal)  );
				}
				else if ( sg_jsglue_get_equals_flag (pszOption, "tag", &pszVal) )
				{
					SG_ERR_CHECK(  SG_rev_spec__add_tag(pCtx, pRevSpec, pszVal)  );
				}
				else if ( sg_jsglue_get_equals_flag (pszOption, "max", &pszVal) )
				{
					SG_ERR_CHECK(  SG_uint32__parse(pCtx, &maxResults, pszVal)  );
				}
				else if ( sg_jsglue_get_equals_flag (pszOption, "user", &pszVal) )
				{
					pszUser = pszVal;
				}
				else if ( sg_jsglue_get_equals_flag (pszOption, "branch", &pszVal) )
				{
					SG_ERR_CHECK(  SG_rev_spec__add_branch(pCtx, pRevSpec, pszVal)  );
				}
				else if (strcmp(pszOption, "hide-merges")==0 )
				{
					bHideMerges = SG_TRUE;
				}
			}
			n++;
		}
	}

	SG_JS_BOOL_CHECK(  1 >= (argc - n)  );

	SG_ERR_CHECK(  SG_alloc(pCtx, argc, sizeof(char*), &newArgv)  );
	//allocate the new string array that we will send down.
	for (i = n; i < argc; i++)
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[i])  );
		newArgv[nCountItems] = JS_GetStringBytes(JSVAL_TO_STRING(argv[i]));
		nCountItems++;
	}
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC(pCtx, &pCwd)  );
	SG_ERR_CHECK(  SG_fsobj__getcwd__pathname(pCtx, pCwd)  );

	SG_ERR_CHECK(  SG_vv_verbs__history__workingFolder(pCtx, SG_pathname__sz(pCwd), nCountItems, newArgv, 
				pRevSpec, 
				pszUser, pszStamp, !bLeaves, maxResults, bHideMerges, minDate,
				maxDate, NULL, NULL, &pHistResult, NULL)  );

	if (pHistResult)
	{
		SG_varray* pvaRef;

#if 1 && defined(DEBUG)
		SG_ERR_IGNORE(  SG_history_debug__dump_result_to_console(pCtx, pHistResult)  );
#endif
		SG_ERR_CHECK(  SG_history_result__get_root(pCtx, pHistResult, &pvaRef)  );

        SG_JS_NULL_CHECK(  (jsoReturn = JS_NewArrayObject(cx, 0, NULL))  );
        *rval = OBJECT_TO_JSVAL(jsoReturn);
        SG_ERR_CHECK(  sg_jsglue__copy_varray_into_jsobject(pCtx, cx, pvaRef, jsoReturn)  );
	}
	else
    {
		*rval = JSVAL_VOID;
    }

	
	SG_REV_SPEC_NULLFREE(pCtx, pRevSpec);
	SG_HISTORY_RESULT_NULLFREE(pCtx, pHistResult);
	SG_TREEDIFF2_NULLFREE(pCtx, pTreeDiff);
	SG_VARRAY_NULLFREE(pCtx, pva_options);
	SG_NULLFREE(pCtx, newArgv);
	SG_PATHNAME_NULLFREE(pCtx, pCwd);
	return JS_TRUE;

fail:
	SG_REV_SPEC_NULLFREE(pCtx, pRevSpec);
	SG_HISTORY_RESULT_NULLFREE(pCtx, pHistResult);
	SG_TREEDIFF2_NULLFREE(pCtx, pTreeDiff);
	SG_VARRAY_NULLFREE(pCtx, pva_options);
	SG_NULLFREE(pCtx, newArgv);
	SG_PATHNAME_NULLFREE(pCtx, pCwd);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    /* TODO */
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(pending_tree, manually_set_parents)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_pendingtree* pPendingTree = NULL;
	SG_varray * pVArrayParents = NULL;
	SG_uint32 i = 0;

	SG_UNUSED(obj);
	SG_UNUSED(rval);

	SG_JS_BOOL_CHECK( argc >= 1);
	SG_ERR_CHECK(  SG_VARRAY__ALLOC(pCtx, &pVArrayParents)  );
	for (i = 0; i < argc; i++)
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[i])  );
		SG_ERR_CHECK(  SG_varray__append__string__sz(pCtx, pVArrayParents, JS_GetStringBytes(JSVAL_TO_STRING(argv[i])))  );
	}
	SG_ERR_CHECK(  SG_pendingtree__alloc_from_cwd(pCtx, &pPendingTree)  );
	SG_ERR_CHECK(  SG_pendingtree__set_wd_parents(pCtx,  pPendingTree, &pVArrayParents)  );
	pVArrayParents = NULL;
	SG_ERR_CHECK(  SG_pendingtree__save(pCtx, pPendingTree)  );
	SG_VARRAY_NULLFREE(pCtx, pVArrayParents);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	return JS_TRUE;
fail:
	SG_VARRAY_NULLFREE(pCtx, pVArrayParents);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(pending_tree, parents)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	JSObject* jso = NULL;
	SG_rev_spec* pRevSpec = NULL;
	SG_varray * pva_options = NULL;
	SG_uint32 n = 0;
	SG_stringarray * pStringArrayResults = NULL;
	const char * pszFilePath = NULL;
	JSObject* jsoReturn = NULL;
	SG_repo * pRepo = NULL;

	SG_UNUSED(obj);

	SG_ERR_CHECK(  sg_jsglue__get_repo_from_cwd(pCtx, &pRepo, NULL, NULL)  );

	if (argc > 0)
	{
		jsval arg0 = argv[0];
		jso = JSVAL_TO_OBJECT(arg0);

		if (JS_IsArrayObject(cx, jso))
		{
			SG_uint32 count = 0;
			SG_uint32 j = 0;
			const char* pszOption;
			const char* pszVal;

			SG_ERR_CHECK(  sg_jsglue__jsobject_to_varray(pCtx, cx, jso, &pva_options)  );

			SG_ERR_CHECK(  SG_varray__count(pCtx, pva_options, &count)  );

			SG_JS_BOOL_CHECK(  count <= 1  );

			SG_ERR_CHECK(  SG_REV_SPEC__ALLOC(pCtx, &pRevSpec)  );

			for (j = 0; j < count; j++)
			{
				SG_ERR_CHECK( SG_varray__get__sz(pCtx, pva_options, j, &pszOption) );

				if ( sg_jsglue_get_equals_flag (pszOption, "rev", &pszVal) )
				{
					SG_ERR_CHECK(  SG_rev_spec__add_rev(pCtx, pRevSpec, pszVal)  );
				}
				else if ( sg_jsglue_get_equals_flag (pszOption, "tag", &pszVal) )
				{
					SG_ERR_CHECK(  SG_rev_spec__add_tag(pCtx, pRevSpec, pszVal)  );
				}
			}
			n++;
		}
		if (argc > n)
		{
			SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[n])  );
			pszFilePath = JS_GetStringBytes(JSVAL_TO_STRING(argv[n]));
		}
	}

	SG_ERR_CHECK( SG_parents__list(pCtx, pRepo, pRevSpec, pszFilePath, &pStringArrayResults)  );

	if (pStringArrayResults != NULL)
	{
        SG_JS_NULL_CHECK(  (jsoReturn = JS_NewArrayObject(cx, 0, NULL))  );
		*rval = OBJECT_TO_JSVAL(jsoReturn);
		SG_ERR_CHECK(  sg_jsglue__copy_stringarray_into_jsobject(pCtx, cx, pStringArrayResults, jsoReturn)  );
	}
	SG_STRINGARRAY_NULLFREE(pCtx, pStringArrayResults);
	SG_REV_SPEC_NULLFREE(pCtx, pRevSpec);
	SG_REPO_NULLFREE(pCtx, pRepo);
	SG_VARRAY_NULLFREE(pCtx, pva_options);
	return JS_TRUE;
fail:
	SG_REV_SPEC_NULLFREE(pCtx, pRevSpec);
	SG_STRINGARRAY_NULLFREE(pCtx, pStringArrayResults);
	SG_REPO_NULLFREE(pCtx, pRepo);
	SG_VARRAY_NULLFREE(pCtx, pva_options);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    /* TODO */
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(pending_tree, leaves)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_varray* pVArrayResults = NULL;
	JSObject* jsoReturn = NULL;

	SG_rbtree * prbLeaves = NULL;
	SG_repo * pRepo = NULL;

	SG_UNUSED(obj);
	SG_UNUSED(argc);
	SG_UNUSED(argv);

	SG_ERR_CHECK(  sg_jsglue__get_repo_from_cwd(pCtx, &pRepo, NULL, NULL)  );

	SG_ERR_CHECK(  SG_repo__fetch_dag_leaves(pCtx,pRepo,SG_DAGNUM__VERSION_CONTROL,&prbLeaves)  );

	SG_ERR_CHECK(  SG_rbtree__to_varray__keys_only(pCtx, prbLeaves, &pVArrayResults)  );

	if (pVArrayResults != NULL)
	{
        SG_JS_NULL_CHECK(  (jsoReturn = JS_NewArrayObject(cx, 0, NULL))  );
        *rval = OBJECT_TO_JSVAL(jsoReturn);
        SG_ERR_CHECK(  sg_jsglue__copy_varray_into_jsobject(pCtx, cx, pVArrayResults, jsoReturn)  );
        SG_VARRAY_NULLFREE(pCtx, pVArrayResults);
	}
	else
    {
		*rval = JSVAL_VOID;
    }
	SG_RBTREE_NULLFREE(pCtx, prbLeaves);
	SG_REPO_NULLFREE(pCtx, pRepo);
	return JS_TRUE;

fail:
	SG_VARRAY_NULLFREE(pCtx, pVArrayResults);
	SG_RBTREE_NULLFREE(pCtx, prbLeaves);
	SG_REPO_NULLFREE(pCtx, pRepo);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    /* TODO */
    return JS_FALSE;
}

/*  usage: sg.addremove([options], path)
 *
 *  possible options: nonrecursive, test, include=pattern, exclude=pattern
 */
SG_JSGLUE_METHOD_PROTOTYPE(pending_tree, addremove)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_pendingtree* pPendingTree = NULL;
	SG_pathname* pPath = NULL;
	uint n = 0;
	JSObject* jso = NULL;
	const char** newArgv = NULL;
	SG_bool bRecursive = SG_TRUE;
	SG_bool bTestOnly = SG_FALSE;
	SG_varray* pva_options = NULL;
	SG_file_spec* pFilespec = NULL;
	SG_uint32 i;
	SG_uint32 nCountItems = 0;

	SG_UNUSED(obj);

	SG_ERR_CHECK(  SG_FILE_SPEC__ALLOC(pCtx, &pFilespec)  );

	if (argc > 0)
	{
		jsval arg0 = argv[0];
		jso = JSVAL_TO_OBJECT(arg0);

		if (JS_IsArrayObject(cx, jso))
		{
			 SG_uint32 count = 0;
			 SG_uint32 j = 0;
			 const char* pszOption;

			 SG_ERR_CHECK(  sg_jsglue__jsobject_to_varray(pCtx, cx, jso, &pva_options)  );

			 SG_ERR_CHECK(  SG_varray__count(pCtx, pva_options, &count)  );

			 for (j = 0; j < count; j++)
			 {
				 SG_ERR_CHECK( SG_varray__get__sz(pCtx, pva_options, j, &pszOption) );

				 if ( strcmp(pszOption, "nonrecursive")==0)
				 {
					bRecursive = SG_FALSE;
				 }
				 else if ( strcmp(pszOption, "test")==0)
				 {
					bTestOnly = SG_TRUE;
				 }
				 else if ( sg_jsglue__is_filespec_arg (pszOption) )
				 {
					 SG_ERR_CHECK (  sg_jsglue__process_filespec_arg(pCtx, pszOption, pFilespec)  );
				 }
			}
			n++;
		}
	}

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC(pCtx, &pPath)  );
	SG_ERR_CHECK(  SG_pathname__set__from_cwd(pCtx, pPath)  );

	//allocate the new string array that we will send down.
	SG_ERR_CHECK(  SG_alloc(pCtx, argc, sizeof(char*), &newArgv)  );

	for (i = n; i < argc; i++)
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[i])  );
		newArgv[nCountItems] = JS_GetStringBytes(JSVAL_TO_STRING(argv[i]));
		nCountItems++;
	}

	SG_ERR_CHECK(  SG_pendingtree__alloc(pCtx, pPath, &pPendingTree)  );

	SG_ERR_CHECK(  SG_pendingtree__addremove(pCtx, pPendingTree, nCountItems,  newArgv, bRecursive, pFilespec, bTestOnly)  );

	SG_FILE_SPEC_NULLFREE(pCtx, pFilespec);
	SG_VARRAY_NULLFREE(pCtx, pva_options);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	*rval = JSVAL_VOID;
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_NULLFREE(pCtx, newArgv);

	return JS_TRUE;

fail:
	SG_FILE_SPEC_NULLFREE(pCtx, pFilespec);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_VARRAY_NULLFREE(pCtx, pva_options);
	SG_PENDINGTREE_NULLFREE(pCtx, pPendingTree);
	SG_NULLFREE(pCtx, newArgv);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    /* TODO */
    return JS_FALSE;
}

/*  usage: sg.pending_tree.commit([options], [files/folders])
 *
 *  TODO 2010/12/03 [] we should not support 'no-ignores', because we are not
 *  TODO               doing an add-remove.  (it gets silently allowed by the
 *  TODO               sg_jsglue__is_filespec_arg().)

 *  TODO 2010/11/30 in jsglue, this routine is in "sg.pending_tree" as
 *  TODO            "sg.pending_tree.commit()", but it does not take a pendingtree
 *  TODO            pointer; rather it just uses the CWD.  should it take a pathname
 *  TODO            for the WD?  the new call to SG_vv_verbs__commit() would allows
 *  TODO            it, but we don't have it here.
 */
SG_JSGLUE_METHOD_PROTOTYPE(pending_tree, commit)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	const char** newArgv = NULL;
	uint i;
	char* psz_hid_cset = NULL;
	char* psz_comment = NULL;
	uint n = 0;
	JSObject* jso = NULL;
	SG_bool bRecursive = SG_TRUE;
	SG_uint32 nCountItems = 0;
	SG_varray* pva_options = NULL;
	SG_file_spec* pFilespec = NULL;
	SG_stringarray* psa_stamps = NULL;
	SG_int64 nCommitDate = SG_AUDIT__WHEN__NOW;
	SG_bool bNoSubmodules = SG_FALSE;
	char * psz_user = SG_AUDIT__WHO__FROM_SETTINGS;

	SG_UNUSED(obj);

	SG_ERR_CHECK (  SG_FILE_SPEC__ALLOC(pCtx, &pFilespec) );

	if (argc > 0)
	{
		jsval arg0 = argv[0];

		jso = JSVAL_TO_OBJECT(arg0);

		if (JS_IsArrayObject(cx, jso))
		{
			 SG_uint32 count = 0;
			 SG_uint32 j = 0;
			 const char* pszOption;

			 SG_ERR_CHECK(  sg_jsglue__jsobject_to_varray(pCtx, cx, jso, &pva_options)  );

			 SG_ERR_CHECK(  SG_varray__count(pCtx, pva_options, &count)  );

			 for (j = 0; j < count; j++)
			 {
				 const char* pszVal = NULL;

				 SG_ERR_CHECK( SG_varray__get__sz(pCtx, pva_options, j, &pszOption) );

				 if ( strcmp(pszOption, "nonrecursive")==0)
				 {
					bRecursive = SG_FALSE;
				 }
				 else if ( sg_jsglue__is_filespec_arg (pszOption) )
				 {
					 SG_ERR_CHECK (  sg_jsglue__process_filespec_arg(pCtx, pszOption, pFilespec)  );
				 }
				 else if (  sg_jsglue_get_equals_flag (pszOption, "message", &pszVal) )
				 {
					SG_ERR_CHECK(  SG_STRDUP(pCtx, pszVal, &psz_comment)  );
				 }
				 else if ( sg_jsglue_get_equals_flag (pszOption, "stamp", &pszVal) )
				 {
					if (psa_stamps == NULL)
						SG_ERR_CHECK(  SG_STRINGARRAY__ALLOC(pCtx, &psa_stamps, 1)  );
					SG_ERR_CHECK(  SG_stringarray__add(pCtx, psa_stamps, pszVal)  );
				 }
				 else if ( sg_jsglue_get_equals_flag (pszOption, "date", &pszVal) )
				 {
					SG_ERR_CHECK(  SG_int64__parse__strict(pCtx, &nCommitDate, pszVal)  );
				 }
				 /* See: W6233 else if (strcmp(pszOption, "no-submodules") == 0)
				 {
					 bNoSubmodules = SG_TRUE;
				 }*/
				 else if ( sg_jsglue_get_equals_flag (pszOption, "user", &pszVal) )
				 {
					 SG_ERR_CHECK(  SG_STRDUP(pCtx, pszVal, &psz_user)  );
				 }

				 // TODO 2010/12/03 Should we allow JS to pass an array of associated WITs?
			 }

			 n++;
		}
	}

	// package up the remaining file_or_folders args into an (argc,argv)

	SG_ERR_CHECK(  SG_alloc(pCtx, argc, sizeof(char*), &newArgv)  );
	for (i = n; i < argc; i++)
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[i])  );
		newArgv[nCountItems] = JS_GetStringBytes(JSVAL_TO_STRING(argv[i]));
		nCountItems++;
	}

	SG_ERR_CHECK(  SG_vv_verbs__commit(pCtx,
									   NULL,				// assume CWD
									   psz_comment,
									   pFilespec,
									   bRecursive,
									   bNoSubmodules,
									   psz_user,
									   nCommitDate,
									   nCountItems, newArgv,
									   0, NULL,				// TODO 2010/12/03 associations?
									   psa_stamps,
									   NULL, NULL,
									   &psz_hid_cset)  );

	JSVAL_FROM_SZ(*rval, psz_hid_cset);

	SG_FILE_SPEC_NULLFREE(pCtx, pFilespec);
	SG_NULLFREE(pCtx, psz_comment);
	SG_NULLFREE(pCtx, psz_hid_cset);
	SG_NULLFREE(pCtx, newArgv);
	SG_VARRAY_NULLFREE(pCtx, pva_options);
	SG_STRINGARRAY_NULLFREE(pCtx, psa_stamps);


	return JS_TRUE;

fail:
	SG_FILE_SPEC_NULLFREE(pCtx, pFilespec);
	SG_NULLFREE(pCtx, psz_comment);
	SG_NULLFREE(pCtx, psz_hid_cset);
	SG_NULLFREE(pCtx, newArgv);

	SG_VARRAY_NULLFREE(pCtx, pva_options);
	SG_STRINGARRAY_NULLFREE(pCtx, psa_stamps);

	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    /* TODO */
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(file, write)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_pathname* pPath = NULL;
	SG_file* pFile = NULL;
	const char* psz_arg1 = NULL;
	const char* psz_arg2 = NULL;

	SG_UNUSED(obj);

	SG_JS_BOOL_CHECK(  2 == argc  );
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  );
	psz_arg1 = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));
	psz_arg2 = JS_GetStringBytes(JSVAL_TO_STRING(argv[1]));

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPath, psz_arg1)  );

	SG_ERR_CHECK(  SG_file__open__pathname(pCtx, pPath, SG_FILE_OPEN_OR_CREATE|SG_FILE_WRONLY, 0644, &pFile)   );

	SG_ERR_CHECK(  SG_file__truncate(pCtx, pFile)   );
	SG_ERR_CHECK(  SG_file__write__sz(pCtx, pFile, psz_arg2)  );
	SG_ERR_CHECK(  SG_file__close(pCtx, &pFile)   );

	SG_PATHNAME_NULLFREE(pCtx, pPath);

	*rval = JSVAL_VOID;
	return JS_TRUE;

fail:
	SG_FILE_NULLCLOSE(pCtx, pFile);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    /* TODO */
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(file, append)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_pathname* pPath = NULL;
	SG_file* pFile = NULL;
	const char* psz_arg1 = NULL;
	const char* psz_arg2 = NULL;

	SG_UNUSED(obj);

	SG_JS_BOOL_CHECK(  2 == argc  );
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  );
	psz_arg1 = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));
	psz_arg2 = JS_GetStringBytes(JSVAL_TO_STRING(argv[1]));

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPath, psz_arg1)  );

	SG_ERR_CHECK(  SG_file__open__pathname(pCtx, pPath, SG_FILE_OPEN_OR_CREATE|SG_FILE_WRONLY,  0644, &pFile)   );
	SG_ERR_CHECK(  SG_file__seek_end(pCtx, pFile, NULL)   );
	SG_ERR_CHECK(  SG_file__write__sz(pCtx, pFile, psz_arg2)   );
	SG_ERR_CHECK(  SG_file__close(pCtx, &pFile)   );

	SG_PATHNAME_NULLFREE(pCtx, pPath);

	*rval = JSVAL_VOID;
	return JS_TRUE;

fail:
	SG_FILE_NULLCLOSE(pCtx, pFile);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	/* TODO */
	return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(file, read)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_pathname* pPath = NULL;
	SG_file* pFile = NULL;
	const char* psz_arg1 = NULL;
	SG_string * pString = NULL;

	SG_UNUSED(obj);

	SG_JS_BOOL_CHECK(  1 == argc  );
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
	psz_arg1 = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPath, psz_arg1)  );

	SG_ERR_CHECK(  SG_file__open__pathname(pCtx, pPath, SG_FILE_RDONLY|SG_FILE_OPEN_EXISTING, SG_FSOBJ_PERMS__UNUSED, &pFile)   );

	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pString)   );
	SG_ERR_CHECK(  SG_file__read_utf8_file_into_string(pCtx, pFile, pString)  );
	SG_ERR_CHECK(  SG_file__close(pCtx, &pFile)   );

	SG_PATHNAME_NULLFREE(pCtx, pPath);

	JSVAL_FROM_SZ(*rval, SG_string__sz(pString));
	SG_STRING_NULLFREE(pCtx, pString);
	return JS_TRUE;

fail:
	SG_FILE_NULLCLOSE(pCtx, pFile);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_STRING_NULLFREE(pCtx, pString);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	/* TODO */
	return JS_FALSE;
}

SG_JSGLUE_PROPERTY_PROTOTYPE(cbuffer, length)
{
    SG_context* pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_safeptr* psp = NULL;
    SG_cbuffer* pCbuffer = NULL;

    SG_UNUSED(idval);

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__cbuffer(pCtx, psp, &pCbuffer)  );

    *vp = INT_TO_JSVAL(pCbuffer->len);
    return JS_TRUE;

fail:
    SG_jsglue__report_sg_error(pCtx,cx);    // DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(cbuffer, fin)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_cbuffer* pCbuffer = NULL;
	SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);

	SG_UNUSED(argc);
	SG_UNUSED(argv);

	SG_ERR_CHECK(  SG_safeptr__unwrap__cbuffer(pCtx, psp, &pCbuffer)  );
	SG_cbuffer__nullfree(&pCbuffer, cx);
	SG_SAFEPTR_NULLFREE(pCtx, psp);
	JS_SetPrivate(cx, obj, NULL);

	*rval = JSVAL_VOID;
	return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx); // DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, install_vc_hook)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_repo* pRepo = NULL;
    SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
    SG_vhash* pvh_args = NULL;
    SG_vhash* pvh_got = NULL;
    const char* psz_interface = NULL;
    const char* psz_js = NULL;
	const char *module = NULL;
	SG_uint32 version = 0;
	SG_bool replace = SG_FALSE;
    SG_audit q;
	const char *uid = NULL;

	SG_JS_BOOL_CHECK(1 == argc);
    SG_JS_BOOL_CHECK(  JSVAL_IS_OBJECT(argv[0])  );
    SG_ERR_CHECK(  sg_jsglue__jsobject_to_vhash(pCtx, cx, JSVAL_TO_OBJECT(argv[0]), &pvh_args)  );
    SG_ERR_CHECK(  SG_vhash__alloc(pCtx, &pvh_got)  );

    SG_ERR_CHECK(  SG_jsglue__np__required__sz(pCtx, pvh_args, pvh_got, "interface", &psz_interface)  );
    SG_ERR_CHECK(  SG_jsglue__np__required__sz(pCtx, pvh_args, pvh_got, "js", &psz_js)  );
	SG_ERR_CHECK(  SG_jsglue__np__optional__bool(pCtx, pvh_args, pvh_got, "replace", SG_FALSE, &replace)  );
	SG_ERR_CHECK(  SG_jsglue__np__optional__sz(pCtx, pvh_args, pvh_got, "module", NULL, &module)  );
	SG_ERR_CHECK(  SG_jsglue__np__optional__sz(pCtx, pvh_args, pvh_got, "userid", NULL, &uid)  );
	SG_ERR_CHECK(  SG_jsglue__np__optional__uint32(pCtx, pvh_args, pvh_got, "version", 0, &version)  );
    SG_ERR_CHECK(  SG_jsglue__np__anything_not_got_is_invalid(pCtx, "install_vc_hook", pvh_args, pvh_got)  );

    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

	if (! uid)
		uid = SG_AUDIT__WHO__FROM_SETTINGS;

    SG_ERR_CHECK(  SG_audit__init(pCtx,&q,pRepo,SG_AUDIT__WHEN__NOW,uid)  );
    SG_ERR_CHECK(  SG_vc_hooks__install(
                pCtx, 
                pRepo, 
                psz_interface,
                psz_js,
				module,
				version,
				replace,
                &q
                )  );

    *rval = JSVAL_VOID;
    SG_VHASH_NULLFREE(pCtx, pvh_args);
    SG_VHASH_NULLFREE(pCtx, pvh_got);
    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    SG_VHASH_NULLFREE(pCtx, pvh_args);
    SG_VHASH_NULLFREE(pCtx, pvh_got);
    /* TODO */
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, vc_hooks_installed)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_repo* pRepo = NULL;
    SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
    const char* psz_interface = NULL;
	SG_varray *recs = NULL;
	JSObject *jso = NULL;

	SG_JS_BOOL_CHECK(1 == argc);
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
	psz_interface = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));

    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );
	SG_ERR_CHECK(  SG_vc_hooks__lookup_by_interface(pCtx, pRepo, psz_interface, &recs)  );

	SG_JS_NULL_CHECK(  (jso = JS_NewArrayObject(cx, 0, NULL))  );
	*rval = OBJECT_TO_JSVAL(jso);
	SG_ERR_CHECK(  sg_jsglue__copy_varray_into_jsobject(pCtx, cx, recs, jso)  );

	SG_VARRAY_NULLFREE(pCtx, recs);

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS

	SG_VARRAY_NULLFREE(pCtx, recs);

    /* TODO */
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, get_status_of_version_control_locks)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_repo* pRepo = NULL;
    SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
    SG_vhash* pvh_result = NULL;
    JSObject* jso = NULL;
    SG_vhash* pvh_args = NULL;
    SG_vhash* pvh_got = NULL;
    SG_vhash* pvh_pile = NULL;
    SG_vhash* pvh_locks_by_branch = NULL;
    SG_vhash* pvh_branches_needing_merge = NULL;
    SG_vhash* pvh_duplicate_locks = NULL;
    SG_vhash* pvh_violated_locks = NULL;

	SG_JS_BOOL_CHECK(1 == argc);
    SG_JS_BOOL_CHECK(  JSVAL_IS_OBJECT(argv[0])  );
    SG_ERR_CHECK(  sg_jsglue__jsobject_to_vhash(pCtx, cx, JSVAL_TO_OBJECT(argv[0]), &pvh_args)  );
    SG_ERR_CHECK(  SG_vhash__alloc(pCtx, &pvh_got)  );

    SG_ERR_CHECK(  SG_jsglue__np__anything_not_got_is_invalid(pCtx, "get_status_of_version_control_locks", pvh_args, pvh_got)  );

    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

	SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvh_result)  );

    SG_ERR_CHECK(  SG_vc_locks__check__okay_to_push(
                pCtx, 
                pRepo, 
                NULL,
                &pvh_locks_by_branch,
                &pvh_duplicate_locks,
                &pvh_violated_locks
                )  );

    if (pvh_locks_by_branch)
    {
        SG_ERR_CHECK(  SG_vhash__add__vhash(pCtx, pvh_result, "locks", &pvh_locks_by_branch)  );
    }

    if (pvh_duplicate_locks)
    {
        SG_ERR_CHECK(  SG_vhash__add__vhash(pCtx, pvh_result, "duplicate", &pvh_duplicate_locks)  );
    }

    if (pvh_violated_locks)
    {
        SG_ERR_CHECK(  SG_vhash__add__vhash(pCtx, pvh_result, "violated", &pvh_violated_locks)  );
    }

    SG_JS_NULL_CHECK(  (jso = JS_NewObject(cx, NULL, NULL, NULL))  );
    *rval = OBJECT_TO_JSVAL(jso);
    SG_ERR_CHECK(  sg_jsglue__copy_vhash_into_jsobject(pCtx, cx, pvh_result, jso)  );
    SG_VHASH_NULLFREE(pCtx, pvh_result);
    SG_VHASH_NULLFREE(pCtx, pvh_args);
    SG_VHASH_NULLFREE(pCtx, pvh_got);
    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    SG_VHASH_NULLFREE(pCtx, pvh_result);
    SG_VHASH_NULLFREE(pCtx, pvh_pile);
    SG_VHASH_NULLFREE(pCtx, pvh_locks_by_branch);
    SG_VHASH_NULLFREE(pCtx, pvh_branches_needing_merge);
    SG_VHASH_NULLFREE(pCtx, pvh_violated_locks);
    SG_VHASH_NULLFREE(pCtx, pvh_duplicate_locks);
    SG_VHASH_NULLFREE(pCtx, pvh_args);
    SG_VHASH_NULLFREE(pCtx, pvh_got);
    /* TODO */
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, close)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_repo* pRepo = NULL;
    SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);

    SG_UNUSED(argc);
    SG_UNUSED(argv);

    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );
    SG_REPO_NULLFREE(pCtx, pRepo);
    SG_SAFEPTR_NULLFREE(pCtx, psp);
    JS_SetPrivate(cx, obj, NULL);

    *rval = JSVAL_VOID;
    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    /* TODO */
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, set_user)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_repo* pRepo = NULL;
    SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
    const char* psz_username = NULL;

    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

	SG_JS_BOOL_CHECK(  1 == argc  );
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
	psz_username = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));

    SG_ERR_CHECK(  SG_user__set_user__repo(pCtx, pRepo, psz_username)  );

    *rval = JSVAL_VOID;
    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    /* TODO */
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, fetch_group)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_repo* pRepo = NULL;
    SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
    const char* psz_name = NULL;
    const char* psz_recid = NULL;
    SG_vhash* pvh = NULL;
    JSObject* jso = NULL;
    SG_uint32 count_given = 0;
    SG_vhash* pvh_args = NULL;
    SG_vhash* pvh_got = NULL;

    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

	SG_JS_BOOL_CHECK(1 == argc);
    SG_JS_BOOL_CHECK(  JSVAL_IS_OBJECT(argv[0])  );

    SG_ERR_CHECK(  sg_jsglue__jsobject_to_vhash(pCtx, cx, JSVAL_TO_OBJECT(argv[0]), &pvh_args)  );
    SG_ERR_CHECK(  SG_vhash__alloc(pCtx, &pvh_got)  );

    SG_ERR_CHECK(  SG_jsglue__np__optional__sz(pCtx, pvh_args, pvh_got, "name", NULL, &psz_name)  );
    SG_ERR_CHECK(  SG_jsglue__np__optional__sz(pCtx, pvh_args, pvh_got, "recid", NULL, &psz_recid)  );
    if (psz_name)
    {
        count_given++;
    }
    if (psz_recid)
    {
        count_given++;
    }
    SG_ERR_CHECK(  SG_jsglue__np__anything_not_got_is_invalid(pCtx, "fetch_group", pvh_args, pvh_got)  );

    if (1 != count_given)
    {
        SG_ERR_THROW2(  SG_ERR_JS_NAMED_PARAM_REQUIRED,
                (pCtx, "Must specific exactly one of name and recid")
                );
    }

    if (psz_name)
    {
        SG_ERR_CHECK(  SG_group__get_by_name(pCtx, pRepo, psz_name, &pvh)  );
    }
    else
    {
        SG_ERR_CHECK(  SG_group__get_by_recid(pCtx, pRepo, psz_recid, &pvh)  );
    }


    SG_JS_NULL_CHECK(  (jso = JS_NewObject(cx, NULL, NULL, NULL))  );
    *rval = OBJECT_TO_JSVAL(jso);
    SG_ERR_CHECK(  sg_jsglue__copy_vhash_into_jsobject(pCtx, cx, pvh, jso)  );
    SG_VHASH_NULLFREE(pCtx, pvh);
    SG_VHASH_NULLFREE(pCtx, pvh_args);
    SG_VHASH_NULLFREE(pCtx, pvh_got);

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    SG_VHASH_NULLFREE(pCtx, pvh);
    SG_VHASH_NULLFREE(pCtx, pvh_args);
    SG_VHASH_NULLFREE(pCtx, pvh_got);
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, create_group)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_repo* pRepo = NULL;
    SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
    const char* psz_name = NULL;
    SG_vhash* pvh_args = NULL;
    SG_vhash* pvh_got = NULL;

    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

	SG_JS_BOOL_CHECK(1 == argc);
    SG_JS_BOOL_CHECK(  JSVAL_IS_OBJECT(argv[0])  );

    SG_ERR_CHECK(  sg_jsglue__jsobject_to_vhash(pCtx, cx, JSVAL_TO_OBJECT(argv[0]), &pvh_args)  );
    SG_ERR_CHECK(  SG_vhash__alloc(pCtx, &pvh_got)  );

    SG_ERR_CHECK(  SG_jsglue__np__required__sz(pCtx, pvh_args, pvh_got, "name", &psz_name)  );
    SG_ERR_CHECK(  SG_jsglue__np__anything_not_got_is_invalid(pCtx, "create_group", pvh_args, pvh_got)  );

    SG_ERR_CHECK(  SG_group__create(pCtx, pRepo, psz_name)  );

    SG_VHASH_NULLFREE(pCtx, pvh_args);
    SG_VHASH_NULLFREE(pCtx, pvh_got);

    *rval = JSVAL_VOID;
    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    SG_VHASH_NULLFREE(pCtx, pvh_args);
    SG_VHASH_NULLFREE(pCtx, pvh_got);
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, create_user)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_repo* pRepo = NULL;
    SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
    const char* psz_username = NULL;
    char* psz_userid = NULL;

    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

	SG_JS_BOOL_CHECK(  1 == argc  );
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
	psz_username = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));

    SG_ERR_CHECK(  SG_user__create(pCtx, pRepo, psz_username, &psz_userid)  );

    JSVAL_FROM_SZ(*rval, psz_userid);
    SG_NULLFREE(pCtx, psz_userid);
    return JS_TRUE;

fail:
    SG_NULLFREE(pCtx, psz_userid);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    /* TODO */
    return JS_FALSE;
}

// usage:  history(max_returned, user, stamp, date_from, date_to, branch, file_folder_path, hidemerges);
SG_JSGLUE_METHOD_PROTOTYPE(repo, history)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    JSObject* jso = NULL;
    SG_repo* pRepo = NULL;
	SG_uint32 max = 25; // default
	
	SG_history_result* pHistResult = NULL;
    SG_vhash* pvh_result = NULL;
    SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
    SG_varray *pvaRef = NULL;

	SG_history_token * pToken = NULL;

	const char *pszUser = NULL;
	const char *pszStamp = NULL;
    const char *pszPath = NULL;


    SG_bool bHideMerges = SG_FALSE;
    SG_bool bResult = SG_FALSE;
	SG_int64 nDateFrom = 0;
	SG_int64 nDateTo = SG_INT64_MAX;
    SG_int32 nCountArgs = 0;

	SG_rev_spec* pRevSpec = NULL;
   
	SG_JS_BOOL_CHECK( argc <= 8 );

    //max
	if ((argc >= 1) && ! JSVAL_IS_NULL(argv[0]))
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_INT(argv[0])  );
		max = (SG_uint32)JSVAL_TO_INT(argv[0]);

		if (max == 0)
			max = SG_INT32_MAX;
	}

    //user
	if ((argc >= 2) && ! JSVAL_IS_NULL(argv[1]))
	{
        SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  );
		pszUser = JS_GetStringBytes(JSVAL_TO_STRING(argv[1]));	
    }    	

    //stamp
	if ((argc >= 3) && ! JSVAL_IS_NULL(argv[2]))
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[2])  );
		pszStamp = JS_GetStringBytes(JSVAL_TO_STRING(argv[2])); 
	}

    //date from
	if ((argc >= 4) && ! JSVAL_IS_NULL(argv[3]))
	{
		const char *timestr;
		SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[3])  );
		timestr = JS_GetStringBytes( JSVAL_TO_STRING(argv[3]) );

		SG_ERR_CHECK(  SG_time__parse__informal__local(pCtx, timestr, &nDateFrom, SG_FALSE)  );
	}

    //date to
	if ((argc >= 5) && ! JSVAL_IS_NULL(argv[4]))
	{
		const char *timestr;
		SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[4])  );
		timestr = JS_GetStringBytes( JSVAL_TO_STRING(argv[4]) );

		SG_ERR_CHECK(  SG_time__parse__informal__local(pCtx, timestr, &nDateTo, SG_TRUE)  );
	}

    //branch
    if ((argc >= 6) && ! JSVAL_IS_NULL(argv[5]))
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[5])  );
		SG_ERR_CHECK(  SG_REV_SPEC__ALLOC(pCtx, &pRevSpec)  );
		SG_ERR_CHECK(  SG_rev_spec__add_branch(pCtx, pRevSpec, JS_GetStringBytes(JSVAL_TO_STRING(argv[5])))  );
	}

    //file or folder path
    if ((argc >= 7) && ! JSVAL_IS_NULL(argv[6]))
	{
        nCountArgs = 1;
		SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[6])  );
		pszPath = JS_GetStringBytes(JSVAL_TO_STRING(argv[6])); 
	}

    //max
	if ((argc >= 8) && ! JSVAL_IS_NULL(argv[7]))
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_BOOLEAN(argv[7])  );
		bHideMerges = (SG_uint32)JSVAL_TO_BOOLEAN(argv[7]);	
	}
    
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

    SG_ERR_CHECK(  SG_vv_verbs__history__repo2(pCtx, pRepo, nCountArgs, &pszPath, pRevSpec, pszUser, pszStamp, max, bHideMerges, nDateFrom, nDateTo, &bResult, &pHistResult, &pToken ) );
    SG_REV_SPEC_NULLFREE(pCtx, pRevSpec);

    if (bResult)
    {           
        SG_varray* pvaRefNew = NULL;
        SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvh_result)  );
        SG_JS_NULL_CHECK(  (jso = JS_NewObject(cx, NULL, NULL, NULL))  );
        *rval = OBJECT_TO_JSVAL(jso);
        
        SG_ERR_CHECK(  SG_history_result__get_root(pCtx, pHistResult, &pvaRef)  );
        SG_ERR_CHECK(  SG_varray__alloc__copy(pCtx, &pvaRefNew, pvaRef)  );
        SG_ERR_CHECK(  SG_vhash__add__varray(pCtx, pvh_result, "items", &pvaRefNew)  ); 
        SG_VARRAY_NULLFREE(pCtx, pvaRefNew);
        if (pToken)
        {         
            SG_ERR_CHECK(  SG_vhash__add__vhash(pCtx, pvh_result, "next_token", (SG_vhash**)&pToken)  );  
          
        }
        SG_ERR_CHECK(  sg_jsglue__copy_vhash_into_jsobject(pCtx, cx, pvh_result, jso)  );   
            
        SG_VHASH_NULLFREE(pCtx, pvh_result);   
        SG_HISTORY_RESULT_NULLFREE(pCtx, pHistResult); 
        SG_HISTORY_TOKEN_NULLFREE(pCtx, pToken);
     
    }
    else
    {
        *rval = JSVAL_NULL;
    }
    
    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    SG_VHASH_NULLFREE(pCtx, pvh_result); 
    SG_REV_SPEC_NULLFREE(pCtx, pRevSpec);

 	SG_HISTORY_RESULT_NULLFREE(pCtx, pHistResult);
    SG_HISTORY_TOKEN_NULLFREE(pCtx, pToken);
    
    return JS_FALSE;
}

//usage:  history_fetch_more(reponame, max_returned);
SG_JSGLUE_METHOD_PROTOTYPE(repo, history_fetch_more)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    JSObject* jso = NULL;
    SG_repo* pRepo = NULL;

	SG_uint32 max = 25; // default
	
	SG_history_result* pHistResult = NULL;
    SG_vhash* pvh_result = NULL;
    SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
  

    SG_varray *pvaRef = NULL;

	SG_vhash* pToken = NULL;
    SG_history_token * pNewToken = NULL;

   
	SG_JS_BOOL_CHECK( argc >= 1 );
    SG_JS_BOOL_CHECK(  !JSVAL_IS_NULL(argv[0]) && JSVAL_IS_OBJECT(argv[0])  );
    SG_ERR_CHECK(  sg_jsglue__jsobject_to_vhash(pCtx, cx, JSVAL_TO_OBJECT(argv[0]), &pToken)  );

	if ((argc >= 2) && ! JSVAL_IS_NULL(argv[1]))
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_INT(argv[1])  );
		max = (SG_uint32)JSVAL_TO_INT(argv[1]);

		if (max == 0)
			max = SG_INT32_MAX;
	}
   
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );
    SG_ERR_CHECK(  SG_vv_verbs__history__fetch_more__repo2(pCtx, pRepo, (SG_history_token*)pToken, max, &pHistResult, &pNewToken)  ); 
 	
    if (pHistResult)
    {           
        SG_varray* pvaRefNew = NULL;

        SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvh_result)  );
        SG_JS_NULL_CHECK(  (jso = JS_NewObject(cx, NULL, NULL, NULL))  );
        *rval = OBJECT_TO_JSVAL(jso);
        
        SG_ERR_CHECK(  SG_history_result__get_root(pCtx, pHistResult, &pvaRef)  );
        SG_ERR_CHECK(  SG_varray__alloc__copy(pCtx, &pvaRefNew, pvaRef)  );
        SG_ERR_CHECK(  SG_vhash__add__varray(pCtx, pvh_result, "items", &pvaRefNew)  ); 
        SG_VARRAY_NULLFREE(pCtx, pvaRefNew);
        if (pNewToken)
        {         
            SG_ERR_CHECK(  SG_vhash__add__vhash(pCtx, pvh_result, "next_token", (SG_vhash**)&pNewToken)  );             
        }
        SG_ERR_CHECK(  sg_jsglue__copy_vhash_into_jsobject(pCtx, cx, pvh_result, jso)  );   
        
        
        SG_VHASH_NULLFREE(pCtx, pvh_result);    
        SG_HISTORY_RESULT_NULLFREE(pCtx, pHistResult);
        SG_VHASH_NULLFREE(pCtx, pToken);
        SG_HISTORY_TOKEN_NULLFREE(pCtx, pNewToken);

    }
    else
    {
        *rval = JSVAL_NULL;
    }
    
    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
 	
   
    SG_VHASH_NULLFREE(pCtx, pvh_result); 
    SG_VHASH_NULLFREE(pCtx, pToken);
    SG_HISTORY_RESULT_NULLFREE(pCtx, pHistResult);
    SG_HISTORY_TOKEN_NULLFREE(pCtx, pNewToken);
    return JS_FALSE;
}



/*SG_JSGLUE_METHOD_PROTOTYPE(repo, history_old)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    JSObject* jso = NULL;
    SG_repo* pRepo = NULL;
	SG_uint32 max = 25; // default
	SG_varray *pvaRef = NULL;
	SG_history_result* pHistResult = NULL;
    SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
	SG_uint32 startWith = 0;
	const char *historyUser = NULL;
	const char *stamp = NULL;
	SG_int64 nDateFrom = 0;
	SG_int64 nDateTo = SG_INT64_MAX;

    SG_UNUSED(obj);
    SG_UNUSED(argc);

	SG_JS_BOOL_CHECK( argc <= 6 );

	if ((argc >= 1) && ! JSVAL_IS_NULL(argv[0]))
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_INT(argv[0])  );
		max = (SG_uint32)JSVAL_TO_INT(argv[0]);

		if (max == 0)
			max = SG_INT32_MAX;
	}

	if ((argc >= 2) && ! JSVAL_IS_NULL(argv[1]))
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_INT(argv[1])  );
		startWith = (SG_uint32)JSVAL_TO_INT(argv[1]);
    }

	if ((argc >= 3) && ! JSVAL_IS_NULL(argv[2]))
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[2])  );
		historyUser = JS_GetStringBytes(JSVAL_TO_STRING(argv[2])); 
	}

	if ((argc >= 4) && ! JSVAL_IS_NULL(argv[3]))
	{
		SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[3])  );
		stamp = JS_GetStringBytes(JSVAL_TO_STRING(argv[3])); 
	}

	if ((argc >= 5) && ! JSVAL_IS_NULL(argv[4]))
	{
		const char *timestr;
		SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[4])  );
		timestr = JS_GetStringBytes( JSVAL_TO_STRING(argv[4]) );

		SG_ERR_CHECK(  SG_time__parse__informal__local(pCtx, timestr, &nDateFrom, SG_FALSE)  );
	}

	if ((argc >= 6) && ! JSVAL_IS_NULL(argv[5]))
	{
		const char *timestr;
		SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[5])  );
		timestr = JS_GetStringBytes( JSVAL_TO_STRING(argv[5]) );

		SG_ERR_CHECK(  SG_time__parse__informal__local(pCtx, timestr, &nDateTo, SG_TRUE)  );
	}

	

    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

    SG_ERR_CHECK(  SG_history__list(pCtx, pRepo, max, startWith, historyUser, stamp, nDateFrom, nDateTo, NULL, &pHistResult, NULL)  );

	SG_ERR_CHECK(  SG_history_result__get_root(pCtx, pHistResult, &pvaRef)  );

    SG_JS_NULL_CHECK(  (jso = JS_NewArrayObject(cx, 0, NULL))  );
    *rval = OBJECT_TO_JSVAL(jso);
    SG_ERR_CHECK(  sg_jsglue__copy_varray_into_jsobject(pCtx, cx, pvaRef, jso)  );

 	SG_HISTORY_RESULT_NULLFREE(pCtx, pHistResult);

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
 	SG_HISTORY_RESULT_NULLFREE(pCtx, pHistResult);

    return JS_FALSE;
}*/



SG_JSGLUE_METHOD_PROTOTYPE(repo, fetch_blob_into_tempfile)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_repo* pRepo = NULL;
    SG_safeptr* psp = NULL;
    JSObject* jso = NULL;
    SG_tempfile* ptf = NULL;
    char* psz_objectid = NULL;
    SG_int64 i_objectid_timestamp = -1;
    SG_blob_encoding encoding = 0;
    char* psz_vcdiff_reference = NULL;
    SG_uint64 len_encoded = 0;
    SG_uint64 len_full = 0;
    jsval jv;
    const char* psz_blobid = NULL;
    SG_bool b_convert_to_full = SG_FALSE;

    SG_UNUSED(argc);

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

    SG_JS_BOOL_CHECK(  2 == argc  );

    // blobid
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    psz_blobid = JS_GetStringBytes(JSVAL_TO_STRING(argv[0])); 

    // convert_to_full
    SG_JS_BOOL_CHECK(  JSVAL_IS_BOOLEAN(argv[1])  );
    b_convert_to_full = JSVAL_TO_BOOLEAN(argv[1]);

    if (b_convert_to_full)
    {
        SG_ERR_CHECK( SG_tempfile__open_new(pCtx, &ptf)  );
        SG_ERR_CHECK(  
                SG_repo__fetch_blob_into_file(
                    pCtx, 
                    pRepo, 
                    psz_blobid,
                    ptf->pFile,
                    &len_full
                    )  );
        SG_ERR_CHECK( SG_tempfile__close(pCtx, ptf) );
        SG_ASSERT(ptf->pPathname);
        SG_ASSERT(!ptf->pFile);
    }
    else
    {
        SG_ERR_CHECK( SG_tempfile__open_new(pCtx, &ptf)  );
        SG_ERR_CHECK(  
                SG_repo__fetch_blob_into_file__encoded(
                    pCtx, 
                    pRepo, 
                    psz_blobid,
                    ptf->pFile,
                    &psz_objectid,
                    &i_objectid_timestamp,
                    &encoding,
                    &psz_vcdiff_reference,
                    &len_encoded,
                    &len_full
                    )  );
        SG_ERR_CHECK( SG_tempfile__close(pCtx, ptf) );
        SG_ASSERT(ptf->pPathname);
        SG_ASSERT(!ptf->pFile);
    }

	SG_JS_NULL_CHECK(  (jso = JS_NewObject(cx, NULL, NULL, NULL))  );
    *rval = OBJECT_TO_JSVAL(jso);

    JSVAL_FROM_SZ(jv, SG_pathname__sz(ptf->pPathname));
    SG_JS_BOOL_CHECK(  JS_SetProperty(cx, jso, "path", &jv)  );

    JSVAL_FROM_SZ(jv, psz_blobid);
    SG_JS_BOOL_CHECK(  JS_SetProperty(cx, jso, "blobid", &jv)  );

    {
        SG_int_to_string_buffer buf_uint64;

        JSVAL_FROM_SZ(jv, SG_uint64_to_sz(len_full, buf_uint64));
        SG_JS_BOOL_CHECK(  JS_SetProperty(cx, jso, "len_full", &jv)  );
    }

    if (!b_convert_to_full)
    {
        if (psz_vcdiff_reference)
        {
            JSVAL_FROM_SZ(jv, psz_vcdiff_reference);
        }
        else
        {
            jv = JSVAL_NULL;
        }
        SG_JS_BOOL_CHECK(  JS_SetProperty(cx, jso, "vcdiff_reference", &jv)  );

        if (psz_objectid)
        {
            JSVAL_FROM_SZ(jv, psz_objectid);
        }
        else
        {
            jv = JSVAL_NULL;
        }
        SG_JS_BOOL_CHECK(  JS_SetProperty(cx, jso, "objectid", &jv)  );

        {
            SG_int_to_string_buffer buf_int64;

            JSVAL_FROM_SZ(jv, SG_int64_to_sz(i_objectid_timestamp, buf_int64));
            SG_JS_BOOL_CHECK(  JS_SetProperty(cx, jso, "objectid_timestamp", &jv)  );
        }

        {
            char buf_encoding[2];

            SG_ASSERT(  (((unsigned char) encoding) >= 'a')  );
            SG_ASSERT(  (((unsigned char) encoding) <= 'z')  );

            buf_encoding[0] = encoding;
            buf_encoding[1] = 0;
            JSVAL_FROM_SZ(jv, buf_encoding);
            SG_JS_BOOL_CHECK(  JS_SetProperty(cx, jso, "encoding", &jv)  );
        }

        {
            SG_int_to_string_buffer buf_uint64;

            JSVAL_FROM_SZ(jv, SG_uint64_to_sz(len_encoded, buf_uint64));
            SG_JS_BOOL_CHECK(  JS_SetProperty(cx, jso, "len_encoded", &jv)  );
        }
    }

    SG_PATHNAME_NULLFREE(pCtx, ptf->pPathname);
    SG_NULLFREE(pCtx, ptf);

    SG_NULLFREE(pCtx, psz_objectid);
    SG_NULLFREE(pCtx, psz_vcdiff_reference);

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    /* TODO */
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, fetch_blob)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_repo* pRepo = NULL;
	SG_safeptr* psp_repo = NULL;
	SG_safeptr* psp_fbh = NULL;
	JSObject* jso = NULL;
	SG_uint64 len_full = 0;
	const char* psz_blobid = NULL;
	SG_repo_fetch_blob_handle * pHandle = NULL;

	psp_repo = sg_jsglue__get_object_private(cx, obj);
	SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp_repo, &pRepo)  );

	SG_JS_BOOL_CHECK(argc==1 || argc==2);

	// blobid
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
	psz_blobid = JS_GetStringBytes(JSVAL_TO_STRING(argv[0])); 

	// chunk length
	SG_JS_BOOL_CHECK(argc<2 || JSVAL_IS_INT(argv[1]));

	if(argc<2 || JSVAL_TO_INT(argv[1])<0)
	{
		SG_ERR_THROW2(SG_ERR_NOTIMPLEMENTED, (pCtx, "todo: return the entire blob as a cbuffer."));
	}
	else
	{
		jsval jv;
		SG_int_to_string_buffer buf_uint64;

		SG_ERR_CHECK(  SG_repo__fetch_blob__begin(pCtx,  pRepo, psz_blobid, SG_TRUE, NULL, NULL, NULL, NULL, NULL, &len_full, &pHandle)  );

		SG_JS_NULL_CHECK(  (jso = JS_NewObject(cx, &sg_fetchblobhandle_class, NULL, NULL))  );
		*rval = OBJECT_TO_JSVAL(jso);
		SG_ERR_CHECK(  SG_safeptr__wrap__fetchblobhandle(pCtx, pHandle, &psp_fbh)  );
		JS_SetPrivate(cx, jso, psp_fbh);

		jv = INT_TO_JSVAL(argv[1]);
		SG_JS_BOOL_CHECK(  JS_SetProperty(cx, jso, "chunk_length", &jv)  );

		JSVAL_FROM_SZ(jv, SG_uint64_to_sz(len_full, buf_uint64));
		SG_JS_BOOL_CHECK(  JS_SetProperty(cx, jso, "length", &jv)  );
	}

	return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx); // DO NOT SG_ERR_IGNORE() THIS
	SG_SAFEPTR_NULLFREE(pCtx, psp_fbh);
	SG_ERR_IGNORE(  SG_repo__fetch_blob__abort(pCtx, pRepo, &pHandle)  );
	return JS_FALSE;
}

/**
 * fetchblobhandle.next_chunk() returns a cbuffer.
 * A cbuffer is essentially an object that is meant to be returned back down to
 * (or passed back up into) C code, without having to be converted back and
 * forth between a JavaScript native type/object. See sg_cbuffer_typedefs.h.
 * 
 * If *don't* pass the returned cbuffer back to C code, you can free it
 * directly from javascript by calling its ".fin()" member function.
 */
SG_JSGLUE_METHOD_PROTOTYPE(fetchblobhandle, next_chunk)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_safeptr* psp_fbh = NULL;
	SG_repo_fetch_blob_handle * pHandle = NULL;
	SG_safeptr* psp_repo = NULL;
	SG_repo* pRepo = NULL;
	SG_safeptr* psp_cbuffer = NULL;
	SG_cbuffer * pCbuffer = NULL;
	SG_bool done = SG_FALSE;
	jsval val;
	SG_uint32 chunk_length = 0;
	JSObject * jso = NULL;

	psp_fbh = sg_jsglue__get_object_private(cx, obj);
	SG_safeptr__unwrap__fetchblobhandle(pCtx, psp_fbh, &pHandle);
	if(SG_context__err_equals(pCtx, SG_ERR_SAFEPTR_NULL))
	{
		SG_context__err_reset(pCtx);
		*rval = JSVAL_NULL; // Return null after the last chunk has been sent.
		return JS_TRUE;
	}
	else
		SG_ERR_CHECK_CURRENT;

	SG_JS_BOOL_CHECK(argc==1);

	SG_JS_BOOL_CHECK(JSVAL_IS_OBJECT(argv[0]));
	psp_repo = sg_jsglue__get_object_private(cx, JSVAL_TO_OBJECT(argv[0]));
	SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp_repo, &pRepo)  );

	SG_JS_BOOL_CHECK(  JS_GetProperty(cx, obj, "chunk_length", &val)  );
	chunk_length = JSVAL_TO_INT(val);

	SG_ERR_CHECK(  SG_cbuffer__alloc__new(pCtx, &pCbuffer, chunk_length)  );
	SG_ERR_CHECK(  SG_repo__fetch_blob__chunk(pCtx,  pRepo, pHandle, pCbuffer->len, pCbuffer->pBuf, &pCbuffer->len, &done)  );

	if(done)
	{
		SG_ERR_CHECK(  SG_repo__fetch_blob__end(pCtx, pRepo, &pHandle)  );
		SG_SAFEPTR_NULLFREE(pCtx, psp_fbh);
		JS_SetPrivate(cx, obj, NULL);
	}

	SG_JS_NULL_CHECK(  jso = JS_NewObject(cx, &sg_cbuffer_class, NULL, NULL)  );
	*rval = OBJECT_TO_JSVAL(jso);
	SG_ERR_CHECK(  SG_safeptr__wrap__cbuffer(pCtx, pCbuffer, &psp_cbuffer)  );
	JS_SetPrivate(cx, jso, psp_cbuffer);

	return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx); // DO NOT SG_ERR_IGNORE() THIS
	SG_SAFEPTR_NULLFREE(pCtx, psp_cbuffer);
	SG_cbuffer__nullfree(&pCbuffer, cx);
	return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(fetchblobhandle, abort)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_safeptr* psp_fbh = NULL;
	SG_repo_fetch_blob_handle * pHandle = NULL;
	SG_safeptr* psp_repo = NULL;
	SG_repo* pRepo = NULL;

	psp_fbh = sg_jsglue__get_object_private(cx, obj);
	SG_safeptr__unwrap__fetchblobhandle(pCtx, psp_fbh, &pHandle);
	if(SG_context__err_equals(pCtx, SG_ERR_SAFEPTR_NULL))
	{
		SG_context__err_reset(pCtx);
		*rval = JSVAL_NULL; // We seem to have already aborted or finished.
		return JS_TRUE;
	}
	else
		SG_ERR_CHECK_CURRENT;

	SG_JS_BOOL_CHECK(argc==1);

	SG_JS_BOOL_CHECK(JSVAL_IS_OBJECT(argv[0]));
	psp_repo = sg_jsglue__get_object_private(cx, JSVAL_TO_OBJECT(argv[0]));
	SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp_repo, &pRepo)  );

	SG_ERR_CHECK(  SG_repo__fetch_blob__abort(pCtx,  pRepo, &pHandle)  );
	SG_SAFEPTR_NULLFREE(pCtx, psp_fbh);
	JS_SetPrivate(cx, obj, NULL);

	*rval = JSVAL_VOID;
	return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx); // DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, file_classes)
{
	SG_context * pCtx = SG_jsglue__get_sg_context(cx);    
    SG_vhash* pvhClasses = NULL;
    SG_repo* pRepo = NULL;    
    JSObject* jso = NULL;
     SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);

    SG_UNUSED(argv);
    SG_UNUSED(argc);

    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

    // try the file classes defined in localsettings
    SG_ERR_CHECK(  SG_localsettings__get__collapsed_vhash(pCtx, SG_FILETOOLCONFIG__CLASSES, pRepo, &pvhClasses)  );
    
	*rval = JSVAL_NULL;
  	if (pvhClasses)
    {
        SG_JS_NULL_CHECK(  (jso = JS_NewObject(cx, NULL, NULL, NULL))  );
        *rval = OBJECT_TO_JSVAL(jso);

        SG_ERR_CHECK(  sg_jsglue__copy_vhash_into_jsobject(pCtx, cx, pvhClasses, jso)  );
        SG_VHASH_NULLFREE(pCtx, pvhClasses);	 
    }
    
    return JS_TRUE;

fail:
    SG_VHASH_NULLFREE(pCtx, pvhClasses);	
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}


static void get_time_value(
        SG_context* pCtx,
        JSContext* cx,
        jsval jv,
        SG_int64* pi_time
        )
{
    SG_int64 i = -1;

    if (JSVAL_IS_INT(jv))
    {
        i = JSVAL_TO_INT(jv);
    }
    else if (JSVAL_IS_NUMBER(jv))
    {
        jsdouble d;

        d = *JSVAL_TO_DOUBLE(jv);
        i = (SG_int64)d;
    }
    else if (JSVAL_IS_STRING(jv))
    {
        SG_ERR_CHECK(  SG_int64__parse__strict(pCtx, &i, JS_GetStringBytes(JSVAL_TO_STRING(jv)))  );
    }
    else if (JSVAL_IS_OBJECT(jv))
    {
        JSObject* jsotmp = NULL;

        jsotmp = JSVAL_TO_OBJECT(jv);
        if (JS_InstanceOf(cx, jsotmp, &js_DateClass, NULL))
        {
            jsval tmpval;
            jsdouble d;

            SG_JS_BOOL_CHECK( JS_CallFunctionName(cx, jsotmp, "getTime", 0, NULL, &tmpval) );

            d = *JSVAL_TO_DOUBLE(jv);
            i = (SG_int64)d;
        }
        else
        {
            SG_ERR_THROW(  SG_ERR_ZING_TYPE_MISMATCH  );
        }
    }
    else
    {
        SG_ERR_THROW(  SG_ERR_ZING_TYPE_MISMATCH  );
    }

    *pi_time = i;

fail:
    ;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, rebuild_indexes)
{
    SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_repo* pRepo = NULL;
    SG_safeptr* psp = NULL;

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

    SG_JS_BOOL_CHECK(0==argc);
    SG_UNUSED(argv);

    SG_ERR_CHECK(  SG_repo__rebuild_indexes(pCtx, pRepo)  );

    *rval = JSVAL_VOID;
    return JS_TRUE;

fail:
    SG_jsglue__report_sg_error(pCtx,cx); // DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, list_blobs)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    JSObject* jso = NULL;
    SG_repo* pRepo = NULL;
    SG_safeptr* psp_repo = NULL;
    SG_safeptr* psp_blobset = NULL;
    SG_blobset* pbs = NULL;

    psp_repo = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp_repo, &pRepo)  );

    SG_UNUSED(obj);
    SG_UNUSED(argc);
    SG_UNUSED(argv);

    // TODO allow args here for encoding, limit, offset
#if 0
    if (1 == argc)
    {
        int32 jsi = 0;
        SG_JS_BOOL_CHECK(  JSVAL_IS_INT(argv[0])  );
        JS_ValueToInt32(cx, argv[0], &jsi);
        limit = (SG_uint32) jsi;
    }
#endif

    SG_ERR_CHECK(  SG_repo__list_blobs(pCtx, pRepo, 0, 0, 0, &pbs)  );

    SG_JS_NULL_CHECK(  jso = JS_NewObject(cx, &sg_blobset_class, NULL, NULL)  );
    *rval = OBJECT_TO_JSVAL(jso);
    SG_ERR_CHECK(  SG_safeptr__wrap__blobset(pCtx, pbs, &psp_blobset)  );
    JS_SetPrivate(cx, jso, psp_blobset);
    psp_blobset = NULL;

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	// TODO do we need to free anything?
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, list_areas)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_repo* pRepo = NULL;
    SG_safeptr* psp_repo = NULL;
	JSObject *jso = NULL;
	SG_vhash *vhAreas = NULL;

	SG_JS_BOOL_CHECK(argc == 0);
	SG_UNUSED(argv);

    psp_repo = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp_repo, &pRepo)  );

	SG_ERR_CHECK(  SG_area__list(pCtx, pRepo, &vhAreas)  );

	SG_JS_NULL_CHECK(  (jso = JS_NewObject(cx, NULL, NULL, NULL))  );
	*rval = OBJECT_TO_JSVAL(jso);
	SG_ERR_CHECK(  sg_jsglue__copy_vhash_into_jsobject(pCtx, cx, vhAreas, jso)  );
	SG_VHASH_NULLFREE(pCtx, vhAreas);

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS

	SG_VHASH_NULLFREE(pCtx, vhAreas);
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, area_add)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_repo* pRepo = NULL;
    SG_safeptr* psp_repo = NULL;
	SG_audit q;
	SG_uint64 area_id = 0;
	const char *area_name = NULL;
	const char *uid = SG_AUDIT__WHO__FROM_SETTINGS;

	SG_JS_BOOL_CHECK( (argc == 2) || (argc == 3) );
	SG_JS_BOOL_CHECK( JSVAL_IS_STRING( argv[0] ) );
	SG_JS_BOOL_CHECK( JSVAL_IS_INT( argv[1] ) );

	if ((argc == 3) && ! JSVAL_IS_NULL(argv[2]))
	{
		SG_JS_BOOL_CHECK( JSVAL_IS_STRING( argv[2] ) );
		uid = JS_GetStringBytes(JSVAL_TO_STRING(argv[2]));
	}

	area_name = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));
	area_id = (SG_uint64)JSVAL_TO_INT(argv[1]);

    psp_repo = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp_repo, &pRepo)  );

	SG_ERR_CHECK(  SG_audit__init(pCtx, &q, pRepo, SG_AUDIT__WHEN__NOW, uid)  );

	SG_ERR_CHECK(  SG_area__add(pCtx, pRepo, area_name, area_id, &q)  );

	*rval = JSVAL_VOID;

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS

    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, list_dags)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    JSObject* jso = NULL;
    SG_repo* pRepo = NULL;
    SG_safeptr* psp_repo = NULL;
	SG_uint32 nDags = 0;
	SG_uint64 *paDagnums = NULL;
	SG_varray *results = NULL;
	SG_uint32 i = 0;

    psp_repo = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp_repo, &pRepo)  );

    SG_UNUSED(argc);
    SG_UNUSED(argv);

	SG_ERR_CHECK(  SG_repo__list_dags(pCtx, pRepo, &nDags, &paDagnums)  );
	SG_ERR_CHECK(  SG_varray__alloc(pCtx, &results)  );

	for ( i = 0; i < nDags; ++i )
	{
		SG_int64 dagnum = paDagnums[i];
		char buf_dagnum[SG_DAGNUM__BUF_MAX__HEX];

		SG_ERR_CHECK(  SG_dagnum__to_sz__hex(pCtx, dagnum, buf_dagnum, sizeof(buf_dagnum))  );
		
		SG_ERR_CHECK(  SG_varray__append__string__sz(pCtx, results, buf_dagnum)  );
	}

	SG_NULLFREE(pCtx, paDagnums);

	SG_JS_NULL_CHECK(  (jso = JS_NewArrayObject(cx, 0, NULL))  );
	*rval = OBJECT_TO_JSVAL(jso);
	SG_ERR_CHECK(  sg_jsglue__copy_varray_into_jsobject(pCtx, cx, results, jso)  );
	SG_VARRAY_NULLFREE(pCtx, results);

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	SG_VARRAY_NULLFREE(pCtx, results);
	SG_NULLFREE(pCtx, paDagnums);									  

    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(blobset, cleanup)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_blobset* pbs = NULL;
    SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);

    SG_UNUSED(argc);
    SG_UNUSED(argv);

    SG_ERR_CHECK(  SG_safeptr__unwrap__blobset(pCtx, psp, &pbs)  );
    SG_BLOBSET_NULLFREE(pCtx, pbs);
    SG_SAFEPTR_NULLFREE(pCtx, psp);
    JS_SetPrivate(cx, obj, NULL);

    *rval = JSVAL_VOID;
    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    /* TODO */
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(blobset, get_stats)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_blobset* pbs = NULL;
    SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
    SG_vhash* pvh_args = NULL;
    SG_vhash* pvh_got = NULL;
    SG_vhash* pvh_result = NULL;
    JSObject* jso = NULL;
    SG_uint32 count = 0;
    SG_uint64 len_encoded = 0;
    SG_uint64 len_full = 0;
    SG_uint64 max_len_encoded = 0;
    SG_uint64 max_len_full = 0;
    SG_blob_encoding encoding;
    SG_uint32 i32 = 0;

	SG_JS_BOOL_CHECK(1 == argc);
    SG_JS_BOOL_CHECK(  JSVAL_IS_OBJECT(argv[0])  );

    SG_ERR_CHECK(  sg_jsglue__jsobject_to_vhash(pCtx, cx, JSVAL_TO_OBJECT(argv[0]), &pvh_args)  );
    SG_ERR_CHECK(  SG_vhash__alloc(pCtx, &pvh_got)  );

    SG_ERR_CHECK(  SG_jsglue__np__optional__uint32(pCtx, pvh_args, pvh_got, "encoding", 0, &i32)  );
    encoding = (SG_blob_encoding) i32;

    SG_ERR_CHECK(  SG_safeptr__unwrap__blobset(pCtx, psp, &pbs)  );

    SG_ERR_CHECK(  SG_blobset__get_stats(
                pCtx,
                pbs,
                encoding,
                &count,
                &len_encoded,
                &len_full,
                &max_len_encoded,
                &max_len_full
                )  );

    SG_ERR_CHECK(  SG_vhash__alloc(pCtx, &pvh_result)  );
    SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh_result, "count", count)  );
    SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh_result, "len_encoded", len_encoded)  );
    SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh_result, "len_full", len_full)  );
    SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh_result, "max_len_encoded", max_len_encoded)  );
    SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh_result, "max_len_full", max_len_full)  );

    SG_JS_NULL_CHECK(  (jso = JS_NewObject(cx, NULL, NULL, NULL))  );
    *rval = OBJECT_TO_JSVAL(jso);
    SG_ERR_CHECK(  sg_jsglue__copy_vhash_into_jsobject(pCtx, cx, pvh_result, jso)  );
    SG_VHASH_NULLFREE(pCtx, pvh_result);

    SG_VHASH_NULLFREE(pCtx, pvh_args);
    SG_VHASH_NULLFREE(pCtx, pvh_got);

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    SG_VHASH_NULLFREE(pCtx, pvh_result);
    SG_VHASH_NULLFREE(pCtx, pvh_args);
    SG_VHASH_NULLFREE(pCtx, pvh_got);
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(blobset, get)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_blobset* pbs = NULL;
    SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
    SG_vhash* pvh_args = NULL;
    SG_vhash* pvh_got = NULL;
    SG_vhash* pvh = NULL;
    JSObject* jso = NULL;
    SG_uint32 limit = 0;
    SG_uint32 skip = 0;

	SG_JS_BOOL_CHECK(1 == argc);
    SG_JS_BOOL_CHECK(  JSVAL_IS_OBJECT(argv[0])  );

    SG_ERR_CHECK(  sg_jsglue__jsobject_to_vhash(pCtx, cx, JSVAL_TO_OBJECT(argv[0]), &pvh_args)  );
    SG_ERR_CHECK(  SG_vhash__alloc(pCtx, &pvh_got)  );

    SG_ERR_CHECK(  SG_jsglue__np__required__uint32(pCtx, pvh_args, pvh_got, "limit", &limit)  );
    SG_ERR_CHECK(  SG_jsglue__np__required__uint32(pCtx, pvh_args, pvh_got, "skip", &skip)  );
    SG_ERR_CHECK(  SG_jsglue__np__anything_not_got_is_invalid(pCtx, "get", pvh_args, pvh_got)  );

    SG_ERR_CHECK(  SG_safeptr__unwrap__blobset(pCtx, psp, &pbs)  );

    SG_ERR_CHECK(  SG_blobset__get(
                pCtx, 
                pbs, 
                limit,
                skip,
                &pvh
                )  );
    SG_JS_NULL_CHECK(  (jso = JS_NewObject(cx, NULL, NULL, NULL))  );
    *rval = OBJECT_TO_JSVAL(jso);
    SG_ERR_CHECK(  sg_jsglue__copy_vhash_into_jsobject(pCtx, cx, pvh, jso)  );
    SG_VHASH_NULLFREE(pCtx, pvh);

    SG_VHASH_NULLFREE(pCtx, pvh_args);
    SG_VHASH_NULLFREE(pCtx, pvh_got);

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    SG_VHASH_NULLFREE(pCtx, pvh);
    SG_VHASH_NULLFREE(pCtx, pvh_args);
    SG_VHASH_NULLFREE(pCtx, pvh_got);
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(blobset, lookup)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_blobset* pbs = NULL;
    SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
    SG_vhash* pvh_args = NULL;
    SG_vhash* pvh_got = NULL;
    SG_vhash* pvh_found = NULL;
    JSObject* jso = NULL;
    const char* psz_hid = NULL;

	SG_JS_BOOL_CHECK(1 == argc);
    SG_JS_BOOL_CHECK(  JSVAL_IS_OBJECT(argv[0])  );

    SG_ERR_CHECK(  sg_jsglue__jsobject_to_vhash(pCtx, cx, JSVAL_TO_OBJECT(argv[0]), &pvh_args)  );
    SG_ERR_CHECK(  SG_vhash__alloc(pCtx, &pvh_got)  );

    SG_ERR_CHECK(  SG_jsglue__np__required__sz(pCtx, pvh_args, pvh_got, "hid", &psz_hid)  );
    SG_ERR_CHECK(  SG_jsglue__np__anything_not_got_is_invalid(pCtx, "lookup", pvh_args, pvh_got)  );

    SG_ERR_CHECK(  SG_safeptr__unwrap__blobset(pCtx, psp, &pbs)  );

    SG_ERR_CHECK(  SG_blobset__lookup__vhash(
                pCtx, 
                pbs, 
                psz_hid, 
                &pvh_found
                )  );
    if (pvh_found)
    {
        SG_JS_NULL_CHECK(  (jso = JS_NewObject(cx, NULL, NULL, NULL))  );
        *rval = OBJECT_TO_JSVAL(jso);
        SG_ERR_CHECK(  sg_jsglue__copy_vhash_into_jsobject(pCtx, cx, pvh_found, jso)  );
        SG_VHASH_NULLFREE(pCtx, pvh_found);
    }
    else
    {
        *rval = JSVAL_NULL;
    }

    SG_VHASH_NULLFREE(pCtx, pvh_args);
    SG_VHASH_NULLFREE(pCtx, pvh_got);

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    SG_VHASH_NULLFREE(pCtx, pvh_found);
    SG_VHASH_NULLFREE(pCtx, pvh_args);
    SG_VHASH_NULLFREE(pCtx, pvh_got);
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, base_to_audit)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    const char* psz_dagnum = NULL;
    SG_uint64 base_dagnum = 0;
    SG_uint64 audit_dagnum = 0;
	char buf_dagnum[SG_DAGNUM__BUF_MAX__HEX];

    SG_UNUSED(obj);

	SG_JS_BOOL_CHECK(1 == argc);
    SG_JS_BOOL_CHECK(JSVAL_IS_STRING(argv[0]));

    psz_dagnum = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));
    SG_ERR_CHECK(  SG_dagnum__from_sz__hex(pCtx, psz_dagnum, &base_dagnum)  );

    audit_dagnum = SG_DAGNUM__BASE_TO_AUDIT(base_dagnum);

	SG_ERR_CHECK(  SG_dagnum__to_sz__hex(pCtx, audit_dagnum, buf_dagnum, sizeof(buf_dagnum))  );

    JSVAL_FROM_SZ(*rval, buf_dagnum);
    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, make_dagnum)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_vhash* pvh_args = NULL;
    SG_vhash* pvh_got = NULL;
    SG_bool b_audit = SG_FALSE;
    SG_bool b_admin = SG_FALSE;
    SG_bool b_trivial = SG_FALSE;
    SG_bool b_hardwired_template = SG_FALSE;
    SG_bool b_single_rectype = SG_FALSE;
    SG_uint64 dagnum = 0;
	char buf_dagnum[SG_DAGNUM__BUF_MAX__HEX];
    const char* psz_type = NULL;
    SG_uint32 id = 0;
    SG_uint32 vendor = 0;
    SG_uint32 grouping = 0;

    SG_UNUSED(obj);
    
	SG_JS_BOOL_CHECK(1 == argc);
    SG_JS_BOOL_CHECK(  JSVAL_IS_OBJECT(argv[0])  );
    SG_ERR_CHECK(  sg_jsglue__jsobject_to_vhash(pCtx, cx, JSVAL_TO_OBJECT(argv[0]), &pvh_args)  );
    SG_ERR_CHECK(  SG_vhash__alloc(pCtx, &pvh_got)  );

    SG_ERR_CHECK(  SG_jsglue__np__required__uint32(pCtx, pvh_args, pvh_got, "id", &id)  );
    SG_ERR_CHECK(  SG_jsglue__np__required__uint32(pCtx, pvh_args, pvh_got, "vendor", &vendor)  );
    SG_ERR_CHECK(  SG_jsglue__np__required__uint32(pCtx, pvh_args, pvh_got, "grouping", &grouping)  );
    SG_ERR_CHECK(  SG_jsglue__np__required__sz(pCtx, pvh_args, pvh_got, "type", &psz_type)  );
    SG_ERR_CHECK(  SG_jsglue__np__optional__bool(pCtx, pvh_args, pvh_got, "trivial", SG_FALSE, &b_trivial)  );
    SG_ERR_CHECK(  SG_jsglue__np__optional__bool(pCtx, pvh_args, pvh_got, "hardwired_template", SG_FALSE, &b_hardwired_template)  );
    SG_ERR_CHECK(  SG_jsglue__np__optional__bool(pCtx, pvh_args, pvh_got, "audit", SG_FALSE, &b_audit)  );
    SG_ERR_CHECK(  SG_jsglue__np__optional__bool(pCtx, pvh_args, pvh_got, "admin", SG_FALSE, &b_admin)  );
    SG_ERR_CHECK(  SG_jsglue__np__optional__bool(pCtx, pvh_args, pvh_got, "single_rectype", SG_FALSE, &b_single_rectype)  );

    SG_ERR_CHECK(  SG_jsglue__np__anything_not_got_is_invalid(pCtx, "make_dagnum", pvh_args, pvh_got)  );

    dagnum = SG_DAGNUM__SET_DAGID(id, 0);
    dagnum = SG_DAGNUM__SET_VENDOR(vendor, dagnum);
    dagnum = SG_DAGNUM__SET_GROUPING(grouping, dagnum);

    if (0 == strcmp(psz_type, "db"))
    {
        dagnum = SG_DAGNUM__SET_TYPE(sg_DAGNUM__TYPE__DB, dagnum);
    }
    else if (0 == strcmp(psz_type, "tree"))
    {
        dagnum = SG_DAGNUM__SET_TYPE(sg_DAGNUM__TYPE__TREE, dagnum);
    }
    else
    {
        SG_ERR_THROW(SG_ERR_NOTIMPLEMENTED);//todo: numbers, bool, null
    }

    if (b_trivial)
    {
        dagnum |= SG_DAGNUM__FLAG__TRIVIAL_MERGE;
    }
    if (b_hardwired_template)
    {
        dagnum |= SG_DAGNUM__FLAG__HARDWIRED_TEMPLATE;
    }
    if (b_audit)
    {
        dagnum |= SG_DAGNUM__FLAG__AUDIT;
    }
    if (b_admin)
    {
        dagnum |= SG_DAGNUM__FLAG__ADMIN;
    }
    if (b_single_rectype)
    {
        dagnum |= SG_DAGNUM__FLAG__SINGLE_RECTYPE;
    }

	SG_ERR_CHECK(  SG_dagnum__to_sz__hex(pCtx, dagnum, buf_dagnum, sizeof(buf_dagnum))  );

    SG_VHASH_NULLFREE(pCtx, pvh_args);
    SG_VHASH_NULLFREE(pCtx, pvh_got);

    JSVAL_FROM_SZ(*rval, buf_dagnum);
    return JS_TRUE;

fail:
    SG_VHASH_NULLFREE(pCtx, pvh_args);
    SG_VHASH_NULLFREE(pCtx, pvh_got);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(sg, get_area_id)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_uint64 dagnum = 0;
	SG_uint64 area = 0;
	const char *dagstr = NULL;

    SG_UNUSED(obj);
    
	SG_JS_BOOL_CHECK(1 == argc);
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
	dagstr = JS_GetStringBytes( JSVAL_TO_STRING(argv[0]) );

	SG_ERR_CHECK(  SG_dagnum__from_sz__hex(pCtx, dagstr, &dagnum)  );

	area = SG_DAGNUM__GET_AREA_ID(dagnum);

	*rval = INT_TO_JSVAL(area);

	return JS_TRUE;

fail:
    SG_jsglue__report_sg_error(pCtx, cx);
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, low_level_delete_record)
{
    SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_repo* pRepo = NULL;
    SG_uint64 dagnum = 0;
    SG_safeptr* psp = NULL;
    JSObject* jso = NULL;
    const char* psz_hid_rec = NULL;
    const char* psz_csid_parent = NULL;
    const char* psz_hid_template = NULL;
    SG_pendingdb* ptx = NULL;
    SG_changeset* pcs = NULL;
    SG_dagnode* pdn = NULL;
    SG_vhash* pvh = NULL;
    const char* psz_dagnum = NULL;

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

    SG_JS_BOOL_CHECK(3==argc);

    // dagnum
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    psz_dagnum = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));
    SG_ERR_CHECK(  SG_dagnum__from_sz__hex(pCtx, psz_dagnum, &dagnum)  );

    // csid
    SG_JS_BOOL_CHECK(JSVAL_IS_STRING(argv[1]));
    psz_csid_parent = JS_GetStringBytes(JSVAL_TO_STRING(argv[1]));

    // hidrec
    SG_JS_BOOL_CHECK(JSVAL_IS_STRING(argv[2]));
    psz_hid_rec = JS_GetStringBytes(JSVAL_TO_STRING(argv[2]));

    // fetch the template
    if (!SG_DAGNUM__HAS_HARDWIRED_TEMPLATE(dagnum))
    {
        SG_ERR_CHECK(  SG_zing__get_cached_hid_for_template__csid(pCtx, pRepo, dagnum, psz_csid_parent, &psz_hid_template)  );
    }

    // do the tx
    SG_ERR_CHECK(  SG_pendingdb__alloc(pCtx, pRepo, dagnum, psz_csid_parent, psz_hid_template, &ptx)  );
    SG_ERR_CHECK(  SG_pendingdb__add_parent(pCtx, ptx, psz_csid_parent)  );

    SG_ERR_CHECK(  SG_pendingdb__remove(pCtx, ptx, psz_hid_rec)  );

    // and COMMIT
    if (SG_DAGNUM__IS_AUDIT(dagnum))
    {
        SG_ERR_CHECK(  SG_pendingdb__commit(pCtx, ptx, NULL, &pcs, &pdn)  );
    }
    else
    {
        SG_audit q;

        SG_ERR_CHECK(  SG_audit__init(pCtx,&q,pRepo,SG_AUDIT__WHEN__NOW,SG_AUDIT__WHO__FROM_SETTINGS)  );
        SG_ERR_CHECK(  SG_pendingdb__commit(pCtx, ptx, &q, &pcs, &pdn)  );
    }

    SG_ERR_CHECK(  SG_vhash__alloc(pCtx, &pvh)  );

    if (pdn)
	{
        SG_int32 generation = -1;
        const char* psz_hid_cs0 = NULL;

		SG_ERR_CHECK(  SG_dagnode__get_id_ref(pCtx, pdn, &psz_hid_cs0)  );
        SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvh, "csid", psz_hid_cs0)  );

		SG_ERR_CHECK(  SG_dagnode__get_generation(pCtx, pdn, &generation)  );
        SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh, "generation", (SG_int64) generation)  );
	}

    SG_DAGNODE_NULLFREE(pCtx, pdn);
    SG_CHANGESET_NULLFREE(pCtx, pcs);
    SG_PENDINGDB_NULLFREE(pCtx, ptx);

    SG_JS_NULL_CHECK(  (jso = JS_NewObject(cx, NULL, NULL, NULL))  );
    *rval = OBJECT_TO_JSVAL(jso);
    SG_ERR_CHECK(  sg_jsglue__copy_vhash_into_jsobject(pCtx, cx, pvh, jso)  );
    SG_VHASH_NULLFREE(pCtx, pvh);

    return JS_TRUE;

fail:
    SG_VHASH_NULLFREE(pCtx, pvh);
    SG_jsglue__report_sg_error(pCtx,cx); // DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, add_audit)
{
    SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_repo* pRepo = NULL;
    SG_uint64 i_dagnum = 0;
    SG_safeptr* psp = NULL;
    SG_int64 i_when = -1;
    SG_audit q;
    const char* psz_dagnum = NULL;

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

    SG_JS_BOOL_CHECK(4==argc);

    // dagnum
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    psz_dagnum = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));
    SG_ERR_CHECK(  SG_dagnum__from_sz__hex(pCtx, psz_dagnum, &i_dagnum)  );

    // csid
    SG_JS_BOOL_CHECK(JSVAL_IS_STRING(argv[1]));

    // userid
    SG_JS_BOOL_CHECK(JSVAL_IS_STRING(argv[2]));

    // when
    SG_ERR_CHECK(  get_time_value(pCtx, cx, argv[3], &i_when)  );

    // add the audit
    SG_ERR_CHECK(  SG_audit__init(pCtx, &q, pRepo, i_when, JS_GetStringBytes(JSVAL_TO_STRING(argv[2])))  );

    SG_ERR_CHECK(  SG_audit__add(pCtx, pRepo, JS_GetStringBytes(JSVAL_TO_STRING(argv[1])), i_dagnum, &q)  );

    *rval = JSVAL_VOID;
    return JS_TRUE;

fail:
    SG_jsglue__report_sg_error(pCtx,cx); // DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, create_audit)
{
    SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_repo* pRepo = NULL;
    JSObject* jso = NULL;
    SG_audit * pq = NULL;
    SG_safeptr* psp = NULL;

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

    SG_JS_BOOL_CHECK(1==argc);
    SG_JS_BOOL_CHECK(JSVAL_IS_STRING(argv[0]));

    SG_ERR_CHECK(  SG_alloc1(pCtx, pq)  );
    SG_ERR_CHECK(  SG_audit__init(pCtx, pq, pRepo, SG_AUDIT__WHEN__NOW, JS_GetStringBytes(JSVAL_TO_STRING(argv[0])))  );

    SG_JS_NULL_CHECK(  jso = JS_NewObject(cx, &sg_audit_class, NULL, NULL)  );
    *rval = OBJECT_TO_JSVAL(jso);
    SG_ERR_CHECK(  SG_safeptr__wrap__audit(pCtx, pq, &psp)  );
    JS_SetPrivate(cx, jso, psp);
    psp = NULL;

    return JS_TRUE;

fail:
    SG_jsglue__report_sg_error(pCtx,cx);    // DO NOT SG_ERR_IGNORE() THIS
    
    SG_NULLFREE(pCtx, pq);

    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(audit, fin)
{
    SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_audit* pq = NULL;
    SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);

    SG_UNUSED(argc);
    SG_UNUSED(argv);

    SG_ERR_CHECK(  SG_safeptr__unwrap__audit(pCtx, psp, &pq)  );
    SG_NULLFREE(pCtx, pq);
    SG_SAFEPTR_NULLFREE(pCtx, psp);
    JS_SetPrivate(cx, obj, NULL);

    *rval = JSVAL_VOID;
    return JS_TRUE;

fail:
    SG_jsglue__report_sg_error(pCtx,cx); // DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, add_vc_comment)
{
    SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_repo* pRepo = NULL;
    SG_audit * pq;
    SG_safeptr* psp = NULL;
    const char* psz_csid = NULL;
    const char* psz_text = NULL;

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

    SG_JS_BOOL_CHECK(argc==3);

    // csid
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    psz_csid = JS_GetStringBytes(JSVAL_TO_STRING(argv[0])); 

    // text
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  );
    psz_text = JS_GetStringBytes(JSVAL_TO_STRING(argv[1])); 

    // audit
    psp = sg_jsglue__get_object_private(cx, JSVAL_TO_OBJECT(argv[2]));
    SG_ERR_CHECK(  SG_safeptr__unwrap__audit(pCtx, psp, &pq)  );

    SG_ERR_CHECK(  SG_vc_comments__add(pCtx, pRepo, psz_csid, psz_text, pq)  );

    *rval = JSVAL_VOID;
    return JS_TRUE;

fail:
    SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    /* TODO */
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, fetch_vc_comments)
{
    SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_repo* pRepo = NULL;
    SG_safeptr* psp = NULL;
    const char * pChangesetHid;
    SG_varray* pvaComments = NULL;
    JSObject* jso = NULL;

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

    SG_JS_BOOL_CHECK(argc==1);
    SG_JS_BOOL_CHECK(JSVAL_IS_STRING(argv[0]));
    pChangesetHid = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));

    SG_ERR_CHECK(   SG_history__get_changeset_comments(pCtx, pRepo, pChangesetHid, &pvaComments)  );

    SG_JS_NULL_CHECK(  (jso = JS_NewArrayObject(cx, 0, NULL))  );
    *rval = OBJECT_TO_JSVAL(jso);
    SG_ERR_CHECK(  sg_jsglue__copy_varray_into_jsobject(pCtx, cx, pvaComments, jso)  );
    SG_VARRAY_NULLFREE(pCtx, pvaComments);

    return JS_TRUE;

fail:
    SG_jsglue__report_sg_error(pCtx,cx);    // DO NOT SG_ERR_IGNORE() THIS
    SG_VARRAY_NULLFREE(pCtx, pvaComments);
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, lookup_branch_names)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_repo* pRepo = NULL;
    SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
    SG_varray* pva_changesets = NULL;
    SG_varray* pva_copy = NULL;
    SG_vhash* pvh_result = NULL;
    SG_vhash* pvh_args = NULL;
    SG_vhash* pvh_got = NULL;
    JSObject* jso = NULL;

    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

	SG_JS_BOOL_CHECK(1 == argc);
    SG_JS_BOOL_CHECK(  JSVAL_IS_OBJECT(argv[0])  );

    SG_ERR_CHECK(  sg_jsglue__jsobject_to_vhash(pCtx, cx, JSVAL_TO_OBJECT(argv[0]), &pvh_args)  );
    SG_ERR_CHECK(  SG_vhash__alloc(pCtx, &pvh_got)  );

    SG_ERR_CHECK(  SG_jsglue__np__required__varray(pCtx, pvh_args, pvh_got, "changesets", &pva_changesets)  );
    SG_ERR_CHECK(  SG_jsglue__np__anything_not_got_is_invalid(pCtx, "lookup_branch_names", pvh_args, pvh_got)  );

    SG_ERR_CHECK(  SG_varray__alloc__copy(pCtx, &pva_copy, pva_changesets)  );
    SG_ERR_CHECK(  SG_vc_branches__get_branch_names_for_given_changesets(pCtx, pRepo, &pva_copy, &pvh_result)  );

	SG_JS_NULL_CHECK(  (jso = JS_NewObject(cx, NULL, NULL, NULL))  );
	*rval = OBJECT_TO_JSVAL(jso);

	SG_ERR_CHECK(  sg_jsglue__copy_vhash_into_jsobject(pCtx, cx, pvh_result, jso)  );
	jso = NULL;

    SG_VHASH_NULLFREE(pCtx, pvh_result);
    SG_VHASH_NULLFREE(pCtx, pvh_args);
    SG_VHASH_NULLFREE(pCtx, pvh_got);

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    SG_VHASH_NULLFREE(pCtx, pvh_args);
    SG_VHASH_NULLFREE(pCtx, pvh_got);
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, named_branches)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_repo* pRepo = NULL;
	SG_safeptr* psp = NULL;
	SG_vhash* pvhReturn = NULL;
	SG_vhash* pvhRefTemp = NULL;
	SG_vhash* pvhBranchPile = NULL;
	SG_bool bHas = SG_FALSE;
	JSObject* jso = NULL;

    SG_UNUSED(argv);

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

    SG_JS_BOOL_CHECK(argc==0);

	SG_JS_NULL_CHECK(  (jso = JS_NewObject(cx, NULL, NULL, NULL))  );
	*rval = OBJECT_TO_JSVAL(jso);
	SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvhReturn)  );

	SG_ERR_CHECK(  SG_vc_branches__cleanup(pCtx, pRepo, &pvhBranchPile)  );

	SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvhBranchPile, "branches", &pvhRefTemp)  );
	SG_ERR_CHECK(  SG_vhash__addcopy__vhash(pCtx, pvhReturn, "branches", pvhRefTemp)  );   

	SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvhBranchPile, "values", &pvhRefTemp)  );
	SG_ERR_CHECK(  SG_vhash__addcopy__vhash(pCtx, pvhReturn, "values", pvhRefTemp)  );  
	
	SG_ERR_CHECK(  SG_vhash__has(pCtx, pvhBranchPile, "closed", &bHas)  );
	if (bHas)
	{
		SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pvhBranchPile, "closed", &pvhRefTemp)  );
		SG_ERR_CHECK(  SG_vhash__addcopy__vhash(pCtx, pvhReturn, "closed", pvhRefTemp)  );  
	}

	SG_ERR_CHECK(  sg_jsglue__copy_vhash_into_jsobject(pCtx, cx, pvhReturn, jso)  );
	
	jso = NULL;
	SG_VHASH_NULLFREE(pCtx, pvhReturn);
	SG_VHASH_NULLFREE(pCtx, pvhBranchPile);

    return JS_TRUE;

fail:
    SG_jsglue__report_sg_error(pCtx,cx);    // DO NOT SG_ERR_IGNORE() THIS
    SG_VHASH_NULLFREE(pCtx, pvhReturn);  
    SG_VHASH_NULLFREE(pCtx, pvhBranchPile);
    
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, fetch_json)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_repo* pRepo = NULL;
    SG_safeptr* psp = NULL;
    JSObject* jso = NULL;
    const char* psz_hid = NULL;

	SG_byte* buf = NULL;
	SG_uint64 len = 0;
	SG_vhash* pvh = NULL;
	SG_varray* pva = NULL;

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

    SG_JS_BOOL_CHECK(  1 == argc  );
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );

    psz_hid = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));

	SG_ERR_CHECK(  SG_repo__fetch_blob_into_memory(pCtx, pRepo, psz_hid, &buf, &len)  );
	if (len > SG_UINT32_MAX)
		SG_ERR_THROW(SG_ERR_LIMIT_EXCEEDED);
	SG_ERR_CHECK(  SG_veither__parse_json__buflen(pCtx, (const char*)buf, (SG_uint32)len, &pvh, &pva)  );
	SG_ASSERT (pvh != NULL || pva != NULL);

	if (pva == NULL)
	{
		SG_JS_NULL_CHECK(  (jso = JS_NewObject(cx, NULL, NULL, NULL))  );
		*rval = OBJECT_TO_JSVAL(jso);
		SG_ERR_CHECK(  sg_jsglue__copy_vhash_into_jsobject(pCtx, cx, pvh, jso)  );
		SG_VHASH_NULLFREE(pCtx, pvh);
	}
	else
	{
		SG_uint32 count = 0;
		SG_ERR_CHECK(  SG_varray__count(pCtx, pva, &count)  );
		SG_JS_NULL_CHECK(  (jso = JS_NewArrayObject(cx, count, NULL))  );
		*rval = OBJECT_TO_JSVAL(jso);
		SG_ERR_CHECK(  sg_jsglue__copy_varray_into_jsobject(pCtx, cx, pva, jso)  );
		SG_VARRAY_NULLFREE(pCtx, pva);
	}

    jso = NULL;
	SG_NULLFREE(pCtx, buf);

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	SG_NULLFREE(pCtx, buf);
	SG_VARRAY_NULLFREE(pCtx, pva);
	SG_VHASH_NULLFREE(pCtx, pvh);
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, fetch_string)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_repo* pRepo = NULL;
    SG_safeptr* psp = NULL;
	SG_byte * pbuf = NULL;
    const char* psz_hid = NULL;
    JSString* pjs = NULL;
    SG_uint64 len = 0;

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

    SG_JS_BOOL_CHECK(  1 == argc  );
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );

    psz_hid = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));

	SG_ERR_CHECK(  SG_repo__fetch_blob_into_memory(pCtx,pRepo,psz_hid,&pbuf,&len)  );

    pjs = JS_NewStringCopyN(cx, (char*) pbuf, (size_t) len);
    *rval = STRING_TO_JSVAL(pjs);
	SG_NULLFREE(pCtx, pbuf);

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    /* TODO */
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, fetch_dagnode)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_uint64 dagnum = 0;
    const char* psz_dagnum = NULL;
    SG_repo* pRepo = NULL;
    SG_safeptr* psp = NULL;
    SG_dagnode* pdn = NULL;
    SG_vhash* pvh = NULL;
    JSObject* jso = NULL;

    SG_UNUSED(argc);

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    psz_dagnum = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));
    SG_ERR_CHECK(  SG_dagnum__from_sz__hex(pCtx, psz_dagnum, &dagnum)  );

	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  );

    SG_ERR_CHECK(  SG_repo__fetch_dagnode(pCtx, pRepo, dagnum, JS_GetStringBytes(JSVAL_TO_STRING(argv[1])), &pdn)  );

    SG_ERR_CHECK(  SG_dagnode__to_vhash__shared(pCtx, pdn, NULL, &pvh)  );
    SG_DAGNODE_NULLFREE(pCtx, pdn);

    SG_JS_NULL_CHECK(  (jso = JS_NewObject(cx, NULL, NULL, NULL))  );
    *rval = OBJECT_TO_JSVAL(jso);
    SG_ERR_CHECK(  sg_jsglue__copy_vhash_into_jsobject(pCtx, cx, pvh, jso)  );
    SG_VHASH_NULLFREE(pCtx, pvh);

    return JS_TRUE;

fail:
    SG_VHASH_NULLFREE(pCtx, pvh);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    /* TODO */
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, fetch_dag_leaves)
{
   SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_repo* pRepo = NULL;
    SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
    SG_rbtree* prb_leaves = NULL;
    SG_uint64 dagnum;
    SG_rbtree_iterator* pit = NULL;
    SG_bool b = SG_FALSE;
    const char* psz_id = NULL;
    SG_uint32 i;
    JSObject* jso;
    const char* psz_dagnum = NULL;

    SG_UNUSED(argc);

    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    psz_dagnum = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));
    SG_ERR_CHECK(  SG_dagnum__from_sz__hex(pCtx, psz_dagnum, &dagnum)  );

    SG_ERR_CHECK(  SG_repo__fetch_dag_leaves(pCtx, pRepo, dagnum, &prb_leaves)  );

    SG_JS_NULL_CHECK(  jso = JS_NewArrayObject(cx, 0, NULL)  );
    *rval = OBJECT_TO_JSVAL(jso);

    SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pit, prb_leaves, &b, &psz_id, NULL)  );
    i = 0;
    while (b)
    {
        jsval jv;

        JSVAL_FROM_SZ(jv, psz_id);
        SG_JS_BOOL_CHECK(  JS_SetElement(cx, jso, i++, &jv)  );

        SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pit, &b, &psz_id, NULL)  );
    }
    SG_rbtree__iterator__free(pCtx, pit);
    SG_RBTREE_NULLFREE(pCtx, prb_leaves);

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	// TODO do we need to free anything?
    return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, compare)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_safeptr* psp = NULL;

	SG_repo* pRepo = NULL;
	SG_repo* pOtherRepo = NULL;
	SG_bool bIdentical;

	psp = sg_jsglue__get_object_private(cx, obj);
	SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

	SG_JS_BOOL_CHECK(  1 == argc  );
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
	SG_ERR_CHECK(  SG_REPO__OPEN_REPO_INSTANCE(pCtx, JS_GetStringBytes(JSVAL_TO_STRING(argv[0])), &pOtherRepo)  );

	SG_ERR_CHECK(  SG_sync_debug__compare_repo_dags(pCtx, pRepo, pOtherRepo, &bIdentical)  );

	if (bIdentical)
    {
		SG_ERR_CHECK(  SG_sync__compare_repo_blobs(pCtx, pRepo, pOtherRepo, &bIdentical)  );
    }

	SG_REPO_NULLFREE(pCtx, pOtherRepo);

	*rval = BOOLEAN_TO_JSVAL(bIdentical);
	return JS_TRUE;

fail:
	SG_REPO_NULLFREE(pCtx, pOtherRepo);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}


SG_JSGLUE_METHOD_PROTOTYPE(repo, vc_lookup_hids)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_safeptr* psp = NULL;
	SG_repo* pRepo = NULL;
    JSObject* pjo = NULL;

	SG_rbtree* prbHids = NULL;
	
	psp = sg_jsglue__get_object_private(cx, obj);
	SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );
   
	SG_JS_BOOL_CHECK(  1 == argc  );
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );

	SG_ERR_CHECK(  SG_repo__find_dagnodes_by_prefix(pCtx, pRepo, SG_DAGNUM__VERSION_CONTROL, 
					JS_GetStringBytes(JSVAL_TO_STRING(argv[0])), &prbHids)  );

    SG_JS_NULL_CHECK(  (pjo = JS_NewArrayObject(cx, 0, NULL))  );
    *rval = OBJECT_TO_JSVAL(pjo);
	SG_ERR_CHECK(  sg_jsglue__copy_rbtree_keys_into_jsobject(pCtx, cx, prbHids, pjo)  ); 

    SG_RBTREE_NULLFREE(pCtx, prbHids);
	return JS_TRUE;

fail:
	SG_RBTREE_NULLFREE(pCtx, prbHids);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}


SG_JSGLUE_METHOD_PROTOTYPE(repo, lookup_hid)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_safeptr* psp = NULL;
	SG_repo* pRepo = NULL;

	char* pszTargetChangeset = NULL;

	psp = sg_jsglue__get_object_private(cx, obj);
	SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

	SG_JS_BOOL_CHECK(  1 == argc  );
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );

	SG_ERR_CHECK(  SG_repo__hidlookup__dagnode(pCtx, pRepo, SG_DAGNUM__VERSION_CONTROL, 
		JS_GetStringBytes(JSVAL_TO_STRING(argv[0])), &pszTargetChangeset)  );

	JSVAL_FROM_SZ(*rval, pszTargetChangeset);
	SG_NULLFREE(pCtx, pszTargetChangeset);

	return JS_TRUE;

fail:
	SG_NULLFREE(pCtx, pszTargetChangeset);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}

/* string repo.store_blobs_from_files(string filePath, string filePath, ...) */
SG_JSGLUE_METHOD_PROTOTYPE(repo, store_blobs_from_files)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
	SG_repo* pRepo = NULL;

	SG_pathname* pPath = NULL;
	SG_file* pFile = NULL;
	SG_repo_tx_handle* pRepoTx = NULL;

	char* pszNewHid = NULL;

	SG_vhash* pvhResults = NULL;
	JSObject* pjo = NULL;

	SG_uint32 i;

	SG_NULLARGCHECK(psp);

	for (i = 0; i < argc; i++)
		SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[i])  );

	SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

	SG_ERR_CHECK(  SG_repo__begin_tx(pCtx, pRepo, &pRepoTx)  );
	SG_ERR_CHECK(  SG_VHASH__ALLOC__PARAMS(pCtx, &pvhResults, argc, NULL, NULL)  );

	for (i = 0; i < argc; i++)
	{
		const char* pszPath = JS_GetStringBytes(JSVAL_TO_STRING(argv[i]));
		SG_ERR_CHECK(  SG_pathname__alloc__sz(pCtx, &pPath, pszPath)  );
		SG_ERR_CHECK(  SG_file__open__pathname(pCtx, pPath, SG_FILE_OPEN_EXISTING|SG_FILE_RDONLY, SG_FSOBJ_PERMS__UNUSED, &pFile)  );

		SG_ERR_CHECK(  SG_repo__store_blob_from_file(pCtx, pRepo, pRepoTx, NULL, 0, SG_FALSE, pFile, &pszNewHid, NULL)  );

		SG_ERR_CHECK(  SG_file__close(pCtx, &pFile)  );
		SG_PATHNAME_NULLFREE(pCtx, pPath);

		SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhResults, pszNewHid, pszPath)  );

		SG_NULLFREE(pCtx, pszNewHid);
	}

	SG_ERR_CHECK(  SG_repo__commit_tx(pCtx, pRepo, &pRepoTx)  );

    SG_JS_NULL_CHECK(  (pjo = JS_NewObject(cx, NULL, NULL, NULL))  );
    *rval = OBJECT_TO_JSVAL(pjo);
    SG_ERR_CHECK(  sg_jsglue__copy_vhash_into_jsobject(pCtx, cx, pvhResults, pjo)  );
    SG_VHASH_NULLFREE(pCtx, pvhResults);

	return JS_TRUE;

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_NULLFREE(pCtx, pszNewHid);
	SG_VHASH_NULLFREE(pCtx, pvhResults);
	SG_ERR_IGNORE(  SG_file__close(pCtx, &pFile)  );
	if (pRepo && pRepoTx)
		SG_ERR_IGNORE(  SG_repo__abort_tx(pCtx, pRepo, &pRepoTx)  );

	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}

/* string repo.new_since(string startChangesetHid, string endChangesetHid) */
SG_JSGLUE_METHOD_PROTOTYPE(repo, new_since)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
	SG_repo* pRepo = NULL;
	SG_rbtree* prbNewNodes = NULL;
	JSObject* pjo = NULL;

	SG_NULLARGCHECK(psp);
	SG_JS_BOOL_CHECK(  2 == argc  );
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  );

	SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

	SG_ERR_CHECK(  SG_dagquery__new_since(pCtx, pRepo, SG_DAGNUM__VERSION_CONTROL, 
		JS_GetStringBytes(JSVAL_TO_STRING(argv[0])), JS_GetStringBytes(JSVAL_TO_STRING(argv[1])), &prbNewNodes)  );

	SG_JS_NULL_CHECK(  (pjo = JS_NewArrayObject(cx, 0, NULL))  );
	*rval = OBJECT_TO_JSVAL(pjo);
	SG_ERR_CHECK(  sg_jsglue__copy_rbtree_keys_into_jsobject(pCtx, cx, prbNewNodes, pjo)  );

	SG_RBTREE_NULLFREE(pCtx, prbNewNodes);
	
	return JS_TRUE;

fail:
	SG_RBTREE_NULLFREE(pCtx, prbNewNodes);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}

/* object repo.diff_changesets(string changesetHid1, string changesetHid2) */
SG_JSGLUE_METHOD_PROTOTYPE(repo, diff_changesets)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
	SG_repo* pRepo = NULL;
    SG_vhash* pvhDiffHash = NULL;
	JSObject* pjoReturn = NULL;
    SG_treediff2* pTreeDiff = NULL;

	SG_NULLARGCHECK(psp);
	SG_JS_BOOL_CHECK(  2 == argc  );
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  );

	SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );    
   
    SG_ERR_CHECK(  SG_treediff2__alloc(pCtx, pRepo, &pTreeDiff)  );
	SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvhDiffHash)  );
	SG_ERR_CHECK(  SG_treediff2__compare_cset_vs_cset(pCtx, pTreeDiff,  JS_GetStringBytes(JSVAL_TO_STRING(argv[0])), JS_GetStringBytes(JSVAL_TO_STRING(argv[1])))  );
	SG_ERR_CHECK(  SG_treediff2__report_status_to_vhash(pCtx, pTreeDiff, pvhDiffHash, SG_FALSE) );	
	
	SG_TREEDIFF2_NULLFREE(pCtx, pTreeDiff);	
		    
    SG_JS_NULL_CHECK(  (pjoReturn = JS_NewObject(cx, NULL, NULL, NULL))  );
    *rval = OBJECT_TO_JSVAL(pjoReturn);
    SG_ERR_CHECK(  sg_jsglue__copy_vhash_into_jsobject(pCtx, cx, pvhDiffHash, pjoReturn)  );
    SG_VHASH_NULLFREE(pCtx, pvhDiffHash);

    return JS_TRUE;
fail:
    SG_VHASH_NULLFREE(pCtx, pvhDiffHash);
	SG_TREEDIFF2_NULLFREE(pCtx, pTreeDiff);		
	
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}

/* object repo.get_changeset_description(string changesetHid) */
SG_JSGLUE_METHOD_PROTOTYPE(repo, get_changeset_description)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
	SG_repo* pRepo = NULL;
	SG_vhash* pvhCS = NULL;
	JSObject* pjoReturn = NULL;

	SG_NULLARGCHECK(psp);
	SG_JS_BOOL_CHECK(  1 == argc  );
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );

	SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

	SG_ERR_CHECK(  SG_history__get_changeset_description(pCtx, pRepo, JS_GetStringBytes(JSVAL_TO_STRING(argv[0])), SG_TRUE, &pvhCS)  );

    SG_JS_NULL_CHECK(  (pjoReturn = JS_NewObject(cx, NULL, NULL, NULL))  );
    *rval = OBJECT_TO_JSVAL(pjoReturn);
    SG_ERR_CHECK(  sg_jsglue__copy_vhash_into_jsobject(pCtx, cx, pvhCS, pjoReturn)  );
    SG_VHASH_NULLFREE(pCtx, pvhCS);

	return JS_TRUE;

fail:
	SG_VHASH_NULLFREE(pCtx, pvhCS);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}

/* object repo.get_vc_tags( <csid optional>) */
SG_JSGLUE_METHOD_PROTOTYPE(repo, fetch_vc_tags)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
	JSObject* jso = NULL;   
    SG_repo* pRepo = NULL;   
    SG_varray* pva_tags = NULL;

    if (argc > 0)
    {
	    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    }

	SG_NULLARGCHECK(psp);

    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );		

    if (argc > 0)
    {
        SG_vc_tags__lookup(pCtx, pRepo, JS_GetStringBytes(JSVAL_TO_STRING(argv[0])), &pva_tags);
    }
    else
	    SG_ERR_CHECK(  SG_vc_tags__list_all(pCtx, pRepo, &pva_tags)  );   

    SG_JS_NULL_CHECK(  (jso = JS_NewArrayObject(cx, 0, NULL))  );
	*rval = OBJECT_TO_JSVAL(jso);
    SG_ERR_CHECK(  sg_jsglue__copy_varray_into_jsobject(pCtx, cx, pva_tags, jso)  );
    SG_VARRAY_NULLFREE(pCtx, pva_tags);
	return JS_TRUE;

fail:
    SG_VARRAY_NULLFREE(pCtx, pva_tags);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}

/* object repo.add_vc_tag(csid, tag, audit) */
SG_JSGLUE_METHOD_PROTOTYPE(repo, add_vc_tag)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_repo* pRepo = NULL;
    SG_audit * pq;
    SG_safeptr* psp = NULL;
    const char* psz_csid = NULL;
    const char* psz_text = NULL;

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

    SG_JS_BOOL_CHECK(argc==3);

    // csid
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    psz_csid = JS_GetStringBytes(JSVAL_TO_STRING(argv[0])); 

    // text
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  );
    psz_text = JS_GetStringBytes(JSVAL_TO_STRING(argv[1])); 

    // audit
    psp = sg_jsglue__get_object_private(cx, JSVAL_TO_OBJECT(argv[2]));
    SG_ERR_CHECK(  SG_safeptr__unwrap__audit(pCtx, psp, &pq)  );

    SG_ERR_CHECK(  SG_vc_tags__add(pCtx, pRepo, psz_csid, psz_text, pq)  );

    *rval = JSVAL_VOID;
    return JS_TRUE;

fail:
    SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    /* TODO */
    return JS_FALSE;
}

/* object repo.delete_vc_tag(tag, audit) */
SG_JSGLUE_METHOD_PROTOTYPE(repo, delete_vc_tag)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_repo* pRepo = NULL;
    SG_audit * pq;
    SG_safeptr* psp = NULL;
    const char* psz_text = NULL;

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

    SG_JS_BOOL_CHECK(argc==3);

    // text
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  );
    psz_text = JS_GetStringBytes(JSVAL_TO_STRING(argv[1])); 

    // audit
    psp = sg_jsglue__get_object_private(cx, JSVAL_TO_OBJECT(argv[2]));
    SG_ERR_CHECK(  SG_safeptr__unwrap__audit(pCtx, psp, &pq)  );

    SG_ERR_CHECK( SG_vc_tags__remove(pCtx, pRepo, pq, 1, &psz_text) );

    *rval = JSVAL_VOID;
    return JS_TRUE;

fail:
    SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    /* TODO */
    return JS_FALSE;
}

/* object repo.add_vc_stamp(csid, stamp, audit) */
SG_JSGLUE_METHOD_PROTOTYPE(repo, add_vc_stamp)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_repo* pRepo = NULL;
    SG_audit * pq;
    SG_safeptr* psp = NULL;
    const char* psz_csid = NULL;
    const char* psz_text = NULL;

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

    SG_JS_BOOL_CHECK(argc==3);

    // csid
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    psz_csid = JS_GetStringBytes(JSVAL_TO_STRING(argv[0])); 

    // text
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  );
    psz_text = JS_GetStringBytes(JSVAL_TO_STRING(argv[1])); 

    // audit
    psp = sg_jsglue__get_object_private(cx, JSVAL_TO_OBJECT(argv[2]));
    SG_ERR_CHECK(  SG_safeptr__unwrap__audit(pCtx, psp, &pq)  );

    SG_ERR_CHECK(  SG_vc_stamps__add(pCtx, pRepo, psz_csid, psz_text, pq, NULL)  );

    *rval = JSVAL_VOID;
    return JS_TRUE;

fail:
    SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    /* TODO */
    return JS_FALSE;
}

/* object repo.delete_vc_stamp(csid, stamp, audit) */
SG_JSGLUE_METHOD_PROTOTYPE(repo, delete_vc_stamp)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_repo* pRepo = NULL;
    SG_audit * pq;
    SG_safeptr* psp = NULL;
    const char* psz_csid = NULL;
    const char* psz_text = NULL;

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

    SG_JS_BOOL_CHECK(argc==3);

    // csid
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    psz_csid = JS_GetStringBytes(JSVAL_TO_STRING(argv[0])); 

    // text
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  );
    psz_text = JS_GetStringBytes(JSVAL_TO_STRING(argv[1])); 

    // audit
    psp = sg_jsglue__get_object_private(cx, JSVAL_TO_OBJECT(argv[2]));
    SG_ERR_CHECK(  SG_safeptr__unwrap__audit(pCtx, psp, &pq)  );

    SG_ERR_CHECK( SG_vc_stamps__remove(pCtx, pRepo, pq, psz_csid, 1, (const char* const*) &psz_text) );

    *rval = JSVAL_VOID;
    return JS_TRUE;

fail:
    SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    /* TODO */
    return JS_FALSE;
}

/* object repo.get_vc_stamps(<csid optional>) */
SG_JSGLUE_METHOD_PROTOTYPE(repo, fetch_vc_stamps)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
	JSObject* jso = NULL;   
    SG_repo* pRepo = NULL;   
    SG_varray* pva_stamps = NULL;

   
    if (argc > 0)
    {
	    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    }

	SG_NULLARGCHECK(psp);

    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );		

    if (argc > 0)
    {
        SG_ERR_CHECK(  SG_vc_stamps__lookup(pCtx, pRepo, JS_GetStringBytes(JSVAL_TO_STRING(argv[0])), &pva_stamps)  );	  
    }
    else
        SG_ERR_CHECK(  SG_vc_stamps__list_all_stamps(pCtx, pRepo, &pva_stamps)  );  

    SG_JS_NULL_CHECK(  (jso = JS_NewArrayObject(cx, 0, NULL))  );
	*rval = OBJECT_TO_JSVAL(jso);
    SG_ERR_CHECK(  sg_jsglue__copy_varray_into_jsobject(pCtx, cx, pva_stamps, jso)  );
    SG_VARRAY_NULLFREE(pCtx, pva_stamps);
	return JS_TRUE;

fail:
    SG_VARRAY_NULLFREE(pCtx, pva_stamps);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}

/* object repo.get_treenode_entry_by_gid(changeset_id, gid) */
SG_JSGLUE_METHOD_PROTOTYPE(repo, get_treenode_info_by_gid)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
    JSObject* jso = NULL;   
    SG_repo* pRepo = NULL;   

    SG_vhash* pvhResults = NULL;
    SG_treenode_entry* tne = NULL;
    SG_treenode* ptn = NULL;
    const char* pChangesetHid = NULL;
    const char* pFolderGid = NULL;
    SG_string* pstr_path = NULL;  
    SG_changeset* pChangeset = NULL;

	SG_NULLARGCHECK(psp);

    SG_JS_BOOL_CHECK(  2 == argc  );
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  );

    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );	

    pChangesetHid = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));
	pFolderGid = JS_GetStringBytes(JSVAL_TO_STRING(argv[1]));
	
	SG_ERR_CHECK(  SG_changeset__load_from_repo(pCtx, pRepo, pChangesetHid, &pChangeset)  );

	SG_ERR_CHECK(  SG_repo__treendx__get_path_in_dagnode(pCtx, pRepo, SG_DAGNUM__VERSION_CONTROL, pFolderGid, pChangesetHid, &pstr_path, &tne)  );
	
    if (tne)
    {
        SG_ERR_CHECK(  sg_jsglue__treenode_entry_hash(pCtx, pRepo, pChangesetHid, tne, SG_string__sz(pstr_path), pFolderGid, &pvhResults)  );
	    SG_JS_NULL_CHECK(  (jso = JS_NewObject(cx, NULL, NULL, NULL))  );
	    *rval = OBJECT_TO_JSVAL(jso);
	    SG_ERR_CHECK(  sg_jsglue__copy_vhash_into_jsobject(pCtx, cx, pvhResults, jso)  );
    }
    SG_VHASH_NULLFREE(pCtx, pvhResults);
    SG_STRING_NULLFREE(pCtx, pstr_path);
	SG_TREENODE_ENTRY_NULLFREE(pCtx, tne);
	SG_CHANGESET_NULLFREE(pCtx, pChangeset);
	SG_TREENODE_NULLFREE(pCtx, ptn);

	return JS_TRUE;

fail:
    SG_VHASH_NULLFREE(pCtx, pvhResults);
    SG_STRING_NULLFREE(pCtx, pstr_path);
	SG_TREENODE_ENTRY_NULLFREE(pCtx, tne);
	SG_CHANGESET_NULLFREE(pCtx, pChangeset);
	SG_TREENODE_NULLFREE(pCtx, ptn);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}

/* object repo.get_treenode_entry_by_path(changeset_id, path) */
SG_JSGLUE_METHOD_PROTOTYPE(repo, get_treenode_info_by_path)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
    JSObject* jso = NULL;   
    SG_repo* pRepo = NULL;   

    SG_vhash* pvhResults = NULL;
    SG_treenode_entry* tne = NULL;
	char* pszHidRoot = NULL;
    const char* pszPath = NULL;
    const char* pszChangesetID;
	SG_treenode* ptn = NULL;
	char* pszGid = NULL;		

	SG_NULLARGCHECK(psp);

    SG_JS_BOOL_CHECK(  2 == argc  );
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  );

    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );	

    pszChangesetID = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));
	pszPath = JS_GetStringBytes(JSVAL_TO_STRING(argv[1]));

    SG_ERR_CHECK(  SG_changeset__tree__find_root_hid(pCtx, pRepo, pszChangesetID, &pszHidRoot)  );
	SG_ERR_CHECK(  SG_treenode__load_from_repo(pCtx, pRepo, pszHidRoot, &ptn)  );	
	
	SG_ERR_CHECK(  SG_treenode__find_treenodeentry_by_path(pCtx, pRepo, ptn, pszPath, &pszGid, &tne)  );

    if (tne)
    {
        SG_ERR_CHECK(  sg_jsglue__treenode_entry_hash(pCtx, pRepo, pszChangesetID, tne, pszPath, pszGid, &pvhResults)  );
	    SG_JS_NULL_CHECK(  (jso = JS_NewObject(cx, NULL, NULL, NULL))  );
	    *rval = OBJECT_TO_JSVAL(jso);
	    SG_ERR_CHECK(  sg_jsglue__copy_vhash_into_jsobject(pCtx, cx, pvhResults, jso)  );
    }
    SG_VHASH_NULLFREE(pCtx, pvhResults);
    SG_TREENODE_ENTRY_NULLFREE(pCtx, tne);
	SG_NULLFREE(pCtx, pszHidRoot);
	SG_TREENODE_NULLFREE(pCtx, ptn);
	SG_NULLFREE(pCtx, pszGid);
	
	return JS_TRUE;

fail:
    SG_TREENODE_ENTRY_NULLFREE(pCtx, tne);
	SG_NULLFREE(pCtx, pszHidRoot);
    SG_VHASH_NULLFREE(pCtx, pvhResults);
	SG_TREENODE_NULLFREE(pCtx, ptn);
	SG_NULLFREE(pCtx, pszGid);	
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}


SG_JSGLUE_PROPERTY_PROTOTYPE(repo, admin_id)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_safeptr* psp = NULL;
    SG_repo* pRepo = NULL;
    char* psz_id = NULL;

    SG_UNUSED(idval);

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

    SG_ERR_CHECK(  SG_repo__get_admin_id(pCtx, pRepo, &psz_id)  );

    JSVAL_FROM_SZ(*vp, psz_id);

    SG_NULLFREE(pCtx, psz_id);

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}

SG_JSGLUE_PROPERTY_PROTOTYPE(repo, name)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_safeptr* psp = NULL;
    SG_repo* pRepo = NULL;
    const char* psz_name = NULL;

    SG_UNUSED(idval);

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

	SG_ERR_CHECK(  SG_repo__get_descriptor_name(pCtx, pRepo, &psz_name)  );

    JSVAL_FROM_SZ(*vp, psz_name);

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}

SG_JSGLUE_PROPERTY_PROTOTYPE(repo, repo_id)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_safeptr* psp = NULL;
    SG_repo* pRepo = NULL;
    char* psz_id = NULL;

    SG_UNUSED(idval);

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

    SG_ERR_CHECK(  SG_repo__get_repo_id(pCtx, pRepo, &psz_id)  );

    JSVAL_FROM_SZ(*vp, psz_id);

    SG_NULLFREE(pCtx, psz_id);

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}

SG_JSGLUE_PROPERTY_PROTOTYPE(blobset, count)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_safeptr* psp = NULL;
    SG_blobset* pRepo = NULL;
    SG_uint32 count = 0;

    SG_UNUSED(idval);

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__blobset(pCtx, psp, &pRepo)  );

    SG_ERR_CHECK(  SG_blobset__count(pCtx, pRepo, &count)  );

    *vp = INT_TO_JSVAL(count);

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}

SG_JSGLUE_PROPERTY_PROTOTYPE(repo, hash_method)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_safeptr* psp = NULL;
    SG_repo* pRepo = NULL;
    char* psz_method = NULL;

    SG_UNUSED(idval);

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

    SG_ERR_CHECK(  SG_repo__get_hash_method(pCtx, pRepo, &psz_method)  );

    JSVAL_FROM_SZ(*vp, psz_method);

    SG_NULLFREE(pCtx, psz_method);

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}

SG_JSGLUE_PROPERTY_PROTOTYPE(repo, supports_live_with_working_copy)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_repo* pRepo = NULL;
	SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
	SG_bool bAnswer = SG_FALSE;

	SG_UNUSED(idval);

	SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

	SG_ERR_CHECK(  SG_repo__query_implementation(pCtx, pRepo, SG_REPO__QUESTION__BOOL__SUPPORTS_LIVE_WITH_WORKING_DIRECTORY, &bAnswer, NULL, NULL, 0, NULL)  );
	*vp = BOOLEAN_TO_JSVAL(bAnswer);

	return JS_TRUE;
fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	// TODO do we need to free anything?
	return JS_FALSE;
}

SG_JSGLUE_PROPERTY_PROTOTYPE(repo, supports_zlib)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_repo* pRepo = NULL;
	SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
	SG_bool bAnswer = SG_FALSE;

	SG_UNUSED(idval);

	SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

	SG_ERR_CHECK(  SG_repo__query_implementation(pCtx, pRepo, SG_REPO__QUESTION__BOOL__SUPPORTS_ZLIB, &bAnswer, NULL, NULL, 0, NULL)  );
	*vp = BOOLEAN_TO_JSVAL(bAnswer);

	return JS_TRUE;
fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	// TODO do we need to free anything?
	return JS_FALSE;
}

SG_JSGLUE_PROPERTY_PROTOTYPE(repo, supports_vcdiff)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_repo* pRepo = NULL;
	SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
	SG_bool bAnswer = SG_FALSE;

	SG_UNUSED(idval);

	SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

	SG_ERR_CHECK(  SG_repo__query_implementation(pCtx, pRepo, SG_REPO__QUESTION__BOOL__SUPPORTS_VCDIFF, &bAnswer, NULL, NULL, 0, NULL)  );
	*vp = BOOLEAN_TO_JSVAL(bAnswer);

	return JS_TRUE;
fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	// TODO do we need to free anything?
	return JS_FALSE;
}
SG_JSGLUE_PROPERTY_PROTOTYPE(repo, supports_dbndx)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_repo* pRepo = NULL;
	SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
	SG_bool bAnswer = SG_FALSE;

	SG_UNUSED(idval);

	SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

	SG_ERR_CHECK(  SG_repo__query_implementation(pCtx, pRepo, SG_REPO__QUESTION__BOOL__SUPPORTS_DBNDX, &bAnswer, NULL, NULL, 0, NULL)  );
	*vp = BOOLEAN_TO_JSVAL(bAnswer);

	return JS_TRUE;
fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	// TODO do we need to free anything?
	return JS_FALSE;
}
SG_JSGLUE_PROPERTY_PROTOTYPE(repo, supports_treendx)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_repo* pRepo = NULL;
	SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
	SG_bool bAnswer = SG_FALSE;

	SG_UNUSED(idval);

	SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

	SG_ERR_CHECK(  SG_repo__query_implementation(pCtx, pRepo, SG_REPO__QUESTION__BOOL__SUPPORTS_TREEDX, &bAnswer, NULL, NULL, 0, NULL)  );
	*vp = BOOLEAN_TO_JSVAL(bAnswer);

	return JS_TRUE;
fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	// TODO do we need to free anything?
	return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, create_nobody)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
    SG_repo* pRepo = NULL;

	SG_UNUSED(argv);
	SG_NULLARGCHECK(psp);

    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

	SG_JS_BOOL_CHECK(  (0 == argc)  );

    SG_ERR_CHECK(  SG_user__create_nobody(pCtx, pRepo)  );

	*rval = JSVAL_VOID;
	return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}

SG_JSGLUE_METHOD_PROTOTYPE(repo, lookup_audits)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
	JSObject* jso = NULL;
    const char* psz_csid = NULL;
    SG_repo* pRepo = NULL;
    SG_uint64 iDagNum = 0;
    SG_varray* pva_audits = NULL;
    const char* psz_dagnum = NULL;

	SG_NULLARGCHECK(psp);

    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );

	SG_JS_BOOL_CHECK(  (2 == argc)  );

    // dagnum
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    psz_dagnum = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));
    SG_ERR_CHECK(  SG_dagnum__from_sz__hex(pCtx, psz_dagnum, &iDagNum)  );

    // csid
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  );
    psz_csid = JS_GetStringBytes(JSVAL_TO_STRING(argv[1]));

    if (SG_DAGNUM__IS_DB(iDagNum))
    {
        SG_ERR_CHECK(  SG_repo__dbndx__lookup_audits(pCtx, pRepo, iDagNum, psz_csid, &pva_audits)  );
    }
    else
    {
        SG_ERR_CHECK(  SG_audit__lookup(pCtx, pRepo, psz_csid, iDagNum, &pva_audits)  );
    }

    SG_JS_NULL_CHECK(  (jso = JS_NewArrayObject(cx, 0, NULL))  );
	*rval = OBJECT_TO_JSVAL(jso);
    SG_ERR_CHECK(  sg_jsglue__copy_varray_into_jsobject(pCtx, cx, pva_audits, jso)  );
    SG_VARRAY_NULLFREE(pCtx, pva_audits);
	return JS_TRUE;

fail:
    SG_VARRAY_NULLFREE(pCtx, pva_audits);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}

//sg.repo.file_diff(hid1, hid2, filename)
SG_JSGLUE_METHOD_PROTOTYPE(repo, diff_file)
{
    SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);	 
    SG_repo* pRepo = NULL;
    JSString* pjs = NULL;

	SG_tempfile* pTempFileA = NULL;
	SG_tempfile* pTempFileB = NULL;

	SG_string* pDiffText = NULL;
	SG_string* pStringLabel_a = NULL;
	SG_string* pStringLabel_b = NULL;

    
	//TODO get this from a config file of some sort. For now just hardcode it	
    SG_difftool__built_in_tool * pfnCompare = SG_difftool__built_in_tool__textfilediff_ignore_eols;	
    
    SG_NULLARGCHECK(psp);
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );	
   
	SG_JS_BOOL_CHECK(  3 == argc  );
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  );
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[2])  );
   
	
    SG_ERR_CHECK(  SG_tempfile__open_new(pCtx, &pTempFileA)  );
	SG_ERR_CHECK(  SG_repo__fetch_blob_into_file(pCtx,pRepo, JS_GetStringBytes(JSVAL_TO_STRING(argv[0])),pTempFileA->pFile,NULL)  );
	SG_ERR_CHECK(  SG_tempfile__close(pCtx, pTempFileA)  );
	SG_ERR_CHECK(  SG_diff_utils__make_label(pCtx,JS_GetStringBytes(JSVAL_TO_STRING(argv[2])),JS_GetStringBytes(JSVAL_TO_STRING(argv[0])),NULL,&pStringLabel_a)  );
    
    SG_ERR_CHECK(  SG_tempfile__open_new(pCtx, &pTempFileB)  );
	SG_ERR_CHECK(  SG_repo__fetch_blob_into_file(pCtx,pRepo, JS_GetStringBytes(JSVAL_TO_STRING(argv[1])),pTempFileB->pFile,NULL)  );
	SG_ERR_CHECK(  SG_tempfile__close(pCtx, pTempFileB)  );
	SG_ERR_CHECK(  SG_diff_utils__make_label(pCtx,JS_GetStringBytes(JSVAL_TO_STRING(argv[2])),JS_GetStringBytes(JSVAL_TO_STRING(argv[1])),NULL,&pStringLabel_b)  );
	

	SG_ERR_CHECK(  (pfnCompare)(pCtx, pTempFileA->pPathname, pTempFileB->pPathname,
										   pStringLabel_a,pStringLabel_b,
										   SG_FALSE,
										   &pDiffText)  );
    if (pDiffText)
    {
        SG_JS_NULL_CHECK(  (pjs = JS_NewStringCopyZ(cx, SG_string__sz(pDiffText)))  );    

        *rval = STRING_TO_JSVAL(pjs);
    }
    else
    {
        *rval = JSVAL_VOID;
    }
	
    SG_STRING_NULLFREE(pCtx, pStringLabel_a);
	SG_STRING_NULLFREE(pCtx, pStringLabel_b);
	SG_ERR_IGNORE(  SG_tempfile__close_and_delete(pCtx, &pTempFileA)  );
	SG_ERR_IGNORE(  SG_tempfile__close_and_delete(pCtx, &pTempFileB)  );
	SG_STRING_NULLFREE(pCtx, pDiffText);

    return JS_TRUE;

fail:
	SG_STRING_NULLFREE(pCtx, pStringLabel_a);
	SG_STRING_NULLFREE(pCtx, pStringLabel_b);
	SG_ERR_IGNORE(  SG_tempfile__close_and_delete(pCtx, &pTempFileA)  );
	SG_ERR_IGNORE(  SG_tempfile__close_and_delete(pCtx, &pTempFileB)  );
	SG_STRING_NULLFREE(pCtx, pDiffText);
    SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}

//diff a file at a certain version with the working copy file
//sg.repo.diff_file_local(hid1, repopath)
SG_JSGLUE_METHOD_PROTOTYPE(repo, diff_file_local)
{
    SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);	 
    SG_repo* pRepo = NULL;
    JSString* pjs = NULL;

	SG_tempfile* pTempFileA = NULL;

	SG_string* pDiffText = NULL;  
    SG_string* pstrRepo = NULL;
    char* pszgid = NULL;
	SG_string* pStringLabel_a = NULL;
	SG_string* pStringLabel_b = NULL;
    SG_pathname* pPathRepoTop = NULL;
    SG_pathname* pPathLocal = NULL;
    SG_pathname* pPathCwd = NULL;	
    
	//TODO get this from a config file of some sort. For now just hardcode it	
    SG_difftool__built_in_tool * pfnCompare = SG_difftool__built_in_tool__textfilediff_ignore_eols;	

    SG_NULLARGCHECK(psp);
    SG_ERR_CHECK(  SG_safeptr__unwrap__repo(pCtx, psp, &pRepo)  );	

	SG_JS_BOOL_CHECK(  2 == argc  );
	SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  );
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  );
    
    SG_ERR_CHECK(  SG_PATHNAME__ALLOC(pCtx, &pPathCwd)  );
	SG_ERR_CHECK(  SG_pathname__set__from_cwd(pCtx, pPathCwd)  );
  	
    SG_ERR_CHECK(  SG_workingdir__find_mapping(pCtx, pPathCwd, &pPathRepoTop, &pstrRepo, &pszgid));
	SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path(pCtx, pPathRepoTop, JS_GetStringBytes(JSVAL_TO_STRING(argv[1])), &pPathLocal) );
    SG_PATHNAME_NULLFREE(pCtx, pPathCwd);
    SG_PATHNAME_NULLFREE(pCtx, pPathRepoTop);
    SG_NULLFREE(pCtx, pszgid);
    SG_STRING_NULLFREE(pCtx, pstrRepo);
	
	//fetch the blob into a temp file
	SG_ERR_CHECK(  SG_tempfile__open_new(pCtx, &pTempFileA)  );
	SG_ERR_CHECK(  SG_repo__fetch_blob_into_file(pCtx,pRepo, JS_GetStringBytes(JSVAL_TO_STRING(argv[0])),pTempFileA->pFile,NULL)  );
	SG_ERR_CHECK(  SG_tempfile__close(pCtx, pTempFileA)  );
	SG_ERR_CHECK(  SG_diff_utils__make_label(pCtx, JS_GetStringBytes(JSVAL_TO_STRING(argv[1])),JS_GetStringBytes(JSVAL_TO_STRING(argv[0])),NULL,&pStringLabel_a)  );
	SG_ERR_CHECK(  SG_diff_utils__make_label(pCtx, SG_pathname__sz(pPathLocal), NULL, NULL, &pStringLabel_b)  );

	SG_ERR_CHECK(  (pfnCompare)(pCtx, pTempFileA->pPathname, pPathLocal,
										   pStringLabel_a,pStringLabel_b,
										   SG_FALSE,
										   &pDiffText)  );
    if (pDiffText)
    {
        SG_JS_NULL_CHECK(  (pjs = JS_NewStringCopyZ(cx, SG_string__sz(pDiffText)))  );   

        *rval = STRING_TO_JSVAL(pjs);
    }
    else
    {
        *rval = JSVAL_VOID;
    }
	
    SG_STRING_NULLFREE(pCtx, pStringLabel_a);
	SG_STRING_NULLFREE(pCtx, pStringLabel_b);
	SG_ERR_IGNORE(  SG_tempfile__close_and_delete(pCtx, &pTempFileA)  );
	SG_STRING_NULLFREE(pCtx, pDiffText); 
	SG_PATHNAME_NULLFREE(pCtx, pPathLocal);

    return JS_TRUE;

fail:
    SG_STRING_NULLFREE(pCtx, pStringLabel_a);
	SG_STRING_NULLFREE(pCtx, pStringLabel_b);
    SG_STRING_NULLFREE(pCtx, pstrRepo);
	SG_PATHNAME_NULLFREE(pCtx, pPathLocal);
	SG_STRING_NULLFREE(pCtx, pDiffText);
    SG_PATHNAME_NULLFREE(pCtx, pPathCwd);
    SG_PATHNAME_NULLFREE(pCtx, pPathRepoTop);
    SG_NULLFREE(pCtx, pszgid);

	SG_ERR_IGNORE(  SG_tempfile__close_and_delete(pCtx, &pTempFileA)  );	

    SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}

/* ---------------------------------------------------------------- */
/* xmlwriter class */
/* ---------------------------------------------------------------- */

static void xmlwriter_finalize(JSContext *cx, JSObject *obj);
static void xmlwriter__free(SG_context *pCtx, sg_xmlwriter *pWriter);

static JSClass xmlwriter_class = {
    "xmlwriter",
    JSCLASS_HAS_PRIVATE,
    JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub,
    JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, xmlwriter_finalize,
    JSCLASS_NO_OPTIONAL_MEMBERS
};


struct sg_xmlwriter
{
	SG_xmlwriter *writer;
	SG_string *dest;
};


static void xmlwriter__alloc(
	SG_context *pCtx,
	sg_xmlwriter **ppWriter)
{
	sg_xmlwriter *pWriter = NULL;

	SG_NULLARGCHECK(ppWriter);

	*ppWriter = NULL;

	SG_ERR_CHECK(  SG_alloc1(pCtx, pWriter)  );

	SG_ERR_CHECK(  SG_string__alloc(pCtx, &(pWriter->dest))  );
	SG_ERR_CHECK(  SG_xmlwriter__alloc(pCtx, &(pWriter->writer), pWriter->dest, SG_TRUE)  );

	*ppWriter = pWriter;
	pWriter = NULL;

fail:
	if (pWriter)
	{
		xmlwriter__free(pCtx, pWriter);
		pWriter = NULL;
	}
}


static void xmlwriter__free(
	SG_context *pCtx,
	sg_xmlwriter *pWriter)
{
	if (pWriter)
	{
		if (pWriter->writer)
			SG_XMLWRITER_NULLFREE(pCtx, pWriter->writer);
		if (pWriter->dest)
			SG_STRING_NULLFREE(pCtx, pWriter->dest);

		SG_NULLFREE(pCtx, pWriter);
	}
}


JSBool xmlwriter_constructor(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
	JSObject* jso = NULL;
	SG_safeptr* psp = NULL;
	sg_xmlwriter *writer = NULL;

    SG_UNUSED(obj);
	SG_UNUSED(argv);

	SG_JS_BOOL_CHECK(  0 == argc  );

	SG_JS_NULL_CHECK(  jso = JS_NewObject(cx, &xmlwriter_class, NULL, NULL)  );
	*rval = OBJECT_TO_JSVAL(jso);

	SG_ERR_CHECK(  xmlwriter__alloc(pCtx, &writer)  );

	SG_ERR_CHECK(  SG_xmlwriter__write_start_document(pCtx, writer->writer)  );

	SG_ERR_CHECK(  SG_safeptr__wrap__xmlwriter(pCtx, writer, &psp)  );
    JS_SetPrivate(cx, jso, psp);

	return JS_TRUE;

fail:
    xmlwriter__free(pCtx, writer);
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
	return JS_FALSE;
}

static void xmlwriter_finalize(JSContext *cx, JSObject *obj)
{
	SG_context * pCtx = NULL;
	SG_safeptr* psp = sg_jsglue__get_object_private(cx, obj);
	sg_xmlwriter* writer = NULL;

	SG_context__alloc__temporary(&pCtx);
	//err = SG_context__alloc__temporary(&pCtx);
    // TODO check err

	SG_NULLARGCHECK(psp);

	SG_ERR_CHECK(  SG_safeptr__unwrap__xmlwriter(pCtx, psp, &writer)  );
	xmlwriter__free(pCtx, writer);
	SG_SAFEPTR_NULLFREE(pCtx, psp);
	SG_CONTEXT_NULLFREE(pCtx);
	return;

fail:
	xmlwriter__free(pCtx, writer);
	SG_SAFEPTR_NULLFREE(pCtx, psp);
	SG_CONTEXT_NULLFREE(pCtx);

}



/* .start_element(name) */
SG_JSGLUE_METHOD_PROTOTYPE(xmlwriter, start_element)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_safeptr* psp = NULL;
	const char *tagname = NULL;
	sg_xmlwriter *writer = NULL;

	SG_UNUSED(rval);

    SG_JS_BOOL_CHECK(  argc == 1 );

    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  ); 
    tagname = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__xmlwriter(pCtx, psp, &writer)  );

	SG_ERR_CHECK(  SG_xmlwriter__write_start_element__sz(pCtx, writer->writer, tagname)  );

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}


/* .attribute(name, value) */
SG_JSGLUE_METHOD_PROTOTYPE(xmlwriter, attribute)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_safeptr* psp = NULL;
	const char *attname = NULL;
	const char *attval = NULL;
	sg_xmlwriter *writer = NULL;

	SG_UNUSED(rval);

    SG_JS_BOOL_CHECK(  argc == 2 );

    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  ); 
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  ); 

    attname = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));
    attval = JS_GetStringBytes(JSVAL_TO_STRING(argv[1]));

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__xmlwriter(pCtx, psp, &writer)  );

	SG_ERR_CHECK(  SG_xmlwriter__write_attribute__sz(pCtx, writer->writer, attname, attval)  );

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}


/* .content(text) */
SG_JSGLUE_METHOD_PROTOTYPE(xmlwriter, content)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_safeptr* psp = NULL;
	const char *text = NULL;
	sg_xmlwriter *writer = NULL;

	SG_UNUSED(rval);

    SG_JS_BOOL_CHECK(  argc == 1 );

    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  ); 
    text = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__xmlwriter(pCtx, psp, &writer)  );

	SG_ERR_CHECK(  SG_xmlwriter__write_content__sz(pCtx, writer->writer, text)  );

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}


/* .end_element() */
SG_JSGLUE_METHOD_PROTOTYPE(xmlwriter, end_element)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_safeptr* psp = NULL;
	sg_xmlwriter *writer = NULL;

	SG_UNUSED(rval);
	SG_UNUSED(argv);

    SG_JS_BOOL_CHECK(  argc == 0 );

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__xmlwriter(pCtx, psp, &writer)  );

	SG_ERR_CHECK(  SG_xmlwriter__write_end_element(pCtx, writer->writer)  );

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}


/* .element(name, content) */
SG_JSGLUE_METHOD_PROTOTYPE(xmlwriter, element)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_safeptr* psp = NULL;
	const char *tagname = NULL;
	const char *text = NULL;
	sg_xmlwriter *writer = NULL;
	
	SG_UNUSED(rval);

    SG_JS_BOOL_CHECK(  argc == 2 );

    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[0])  ); 
    SG_JS_BOOL_CHECK(  JSVAL_IS_STRING(argv[1])  ); 

	tagname = JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));
    text = JS_GetStringBytes(JSVAL_TO_STRING(argv[1]));

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__xmlwriter(pCtx, psp, &writer)  );

	SG_ERR_CHECK(  SG_xmlwriter__write_element__sz(pCtx, writer->writer, tagname, text)  );

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}

/* .finish() */
/* end doc, return text */
SG_JSGLUE_METHOD_PROTOTYPE(xmlwriter, finish)
{
	SG_context * pCtx = SG_jsglue__get_clean_sg_context(cx);
    SG_safeptr* psp = NULL;
	sg_xmlwriter *writer = NULL;
    JSString* pjs = NULL;

	SG_UNUSED(argv);

    SG_JS_BOOL_CHECK(  argc == 0 );

    psp = sg_jsglue__get_object_private(cx, obj);
    SG_ERR_CHECK(  SG_safeptr__unwrap__xmlwriter(pCtx, psp, &writer)  );

	SG_ERR_CHECK(  SG_xmlwriter__write_end_document(pCtx, writer->writer)  );

    SG_JS_NULL_CHECK(  (pjs = JS_NewStringCopyZ(cx, SG_string__sz(writer->dest)))  );
	*rval = STRING_TO_JSVAL(pjs);

    return JS_TRUE;

fail:
	SG_jsglue__report_sg_error(pCtx,cx);	// DO NOT SG_ERR_IGNORE() THIS
    return JS_FALSE;
}

// todo: how/when do we free?



/* ---------------------------------------------------------------- */
/* Now the method tables */
/* ---------------------------------------------------------------- */

static JSPropertySpec sg_audit_properties[] =
{
    //{"who", ...},
    //{"when", ...},
    {NULL,0,0,NULL,NULL}
};
static JSFunctionSpec sg_audit_methods[] =
{
    {"fin", SG_JSGLUE_METHOD_NAME(audit,fin), 0,0,0},
    {NULL,NULL,0,0,0}
};

static JSPropertySpec sg_cbuffer_properties[] =
{
    {"length", 0, JSPROP_READONLY, SG_JSGLUE_PROPERTY_NAME(cbuffer, length), NULL},
    {NULL,0,0,NULL,NULL}
};
static JSFunctionSpec sg_cbuffer_methods[] =
{
    {"fin", SG_JSGLUE_METHOD_NAME(cbuffer,fin), 0,0,0},
    {NULL,NULL,0,0,0}
};

static JSPropertySpec sg_blobset_properties[] =
{
    {"count",           0,      JSPROP_READONLY,       SG_JSGLUE_PROPERTY_NAME(blobset, count), NULL},
    {NULL,0,0,NULL,NULL}
};

static JSFunctionSpec sg_blobset_methods[] = 
{
    {"cleanup", 			  	SG_JSGLUE_METHOD_NAME(blobset, cleanup),0,0,0},
    {"get_stats", 			  	SG_JSGLUE_METHOD_NAME(blobset, get_stats),1,0,0},
    {"lookup", 			  	SG_JSGLUE_METHOD_NAME(blobset, lookup),1,0,0},
    {"get", 			  	SG_JSGLUE_METHOD_NAME(blobset, get),1,0,0},
	JS_FS_END
};

/*
 * These properties & methods are available on a repo class.  To create a repo
 * use the javascript "var repo = sg.open_repo("repo name")
 */
static JSPropertySpec sg_repo_properties[] = {
    {"hash_method",           0,      JSPROP_READONLY,       SG_JSGLUE_PROPERTY_NAME(repo, hash_method), NULL},
    {"repo_id",           0,      JSPROP_READONLY,       SG_JSGLUE_PROPERTY_NAME(repo, repo_id), NULL},
    {"admin_id",          0,     JSPROP_READONLY,       SG_JSGLUE_PROPERTY_NAME(repo, admin_id), NULL},
	{"name",	0, JSPROP_READONLY, SG_JSGLUE_PROPERTY_NAME(repo, name), NULL},
    {"supports_live_with_working_copy",          0,     JSPROP_READONLY,       SG_JSGLUE_PROPERTY_NAME(repo, supports_live_with_working_copy), NULL},
    {"supports_zlib",          0,     JSPROP_READONLY,       SG_JSGLUE_PROPERTY_NAME(repo, supports_zlib), NULL},
    {"supports_vcdiff",          0,     JSPROP_READONLY,       SG_JSGLUE_PROPERTY_NAME(repo, supports_vcdiff), NULL},
    {"supports_dbndx",          0,     JSPROP_READONLY,       SG_JSGLUE_PROPERTY_NAME(repo, supports_dbndx), NULL},
    {"supports_treendx",          0,     JSPROP_READONLY,       SG_JSGLUE_PROPERTY_NAME(repo, supports_treendx), NULL},
    {NULL,0,0,NULL,NULL}
};
static JSFunctionSpec sg_repo_methods[] = {
    {"lookup_audits", 			SG_JSGLUE_METHOD_NAME(repo,lookup_audits),0,0,0},
    {"create_nobody", 			SG_JSGLUE_METHOD_NAME(repo,create_nobody),0,0,0},
    {"fetch_dag_leaves", SG_JSGLUE_METHOD_NAME(repo,fetch_dag_leaves),1,0,0},
    {"close", SG_JSGLUE_METHOD_NAME(repo,close),0,0,0},
    {"install_vc_hook", SG_JSGLUE_METHOD_NAME(repo,install_vc_hook),1,0,0},
    {"vc_hooks_installed", SG_JSGLUE_METHOD_NAME(repo,vc_hooks_installed),1,0,0},
    {"get_status_of_version_control_locks", SG_JSGLUE_METHOD_NAME(repo,get_status_of_version_control_locks),1,0,0},
    {"set_user", SG_JSGLUE_METHOD_NAME(repo,set_user),1,0,0},
    {"create_user", SG_JSGLUE_METHOD_NAME(repo,create_user),1,0,0},
    {"create_group", SG_JSGLUE_METHOD_NAME(repo,create_group),1,0,0},
    {"fetch_group", SG_JSGLUE_METHOD_NAME(repo,fetch_group),1,0,0},
    {"fetch_vc_tags", SG_JSGLUE_METHOD_NAME(repo,fetch_vc_tags),0,0,0},
    {"delete_vc_tag", SG_JSGLUE_METHOD_NAME(repo,delete_vc_tag),3,0,0},
    {"add_vc_tag", SG_JSGLUE_METHOD_NAME(repo,add_vc_tag),3,0,0},
    {"fetch_vc_stamps", SG_JSGLUE_METHOD_NAME(repo,fetch_vc_stamps),0,0,0},
    {"delete_vc_stamp", SG_JSGLUE_METHOD_NAME(repo,delete_vc_stamp),3,0,0},
    {"add_vc_stamp", SG_JSGLUE_METHOD_NAME(repo,add_vc_stamp),3,0,0},
    {"fetch_dagnode", SG_JSGLUE_METHOD_NAME(repo,fetch_dagnode),1,0,0},
    {"fetch_string", SG_JSGLUE_METHOD_NAME(repo,fetch_string),1,0,0},
    {"fetch_json", SG_JSGLUE_METHOD_NAME(repo,fetch_json),1,0,0},
    {"fetch_blob_into_tempfile", SG_JSGLUE_METHOD_NAME(repo,fetch_blob_into_tempfile),1,0,0},
    {"fetch_blob", SG_JSGLUE_METHOD_NAME(repo,fetch_blob), 2,0,0},
    {"diff_file", SG_JSGLUE_METHOD_NAME(repo,diff_file), 3,0,0},
    {"diff_file_local", SG_JSGLUE_METHOD_NAME(repo, diff_file_local), 2,0,0},
    {"get_treenode_info_by_gid", SG_JSGLUE_METHOD_NAME(repo, get_treenode_info_by_gid), 2,0,0},
    {"get_treenode_info_by_path", SG_JSGLUE_METHOD_NAME(repo, get_treenode_info_by_path), 2,0,0},
    {"history", SG_JSGLUE_METHOD_NAME(repo,history), 6, 0, 0 },

	{"area_add", SG_JSGLUE_METHOD_NAME(repo,area_add), 2, 0, 0},
	{"list_areas", SG_JSGLUE_METHOD_NAME(repo,list_areas), 0, 0, 0},

	{"history_fetch_more", SG_JSGLUE_METHOD_NAME(repo,history_fetch_more), 2, 0, 0 },

    {"list_blobs", SG_JSGLUE_METHOD_NAME(repo,list_blobs), 3,0,0},
    {"low_level_delete_record", SG_JSGLUE_METHOD_NAME(repo,low_level_delete_record), 3,0,0},
    {"add_audit", SG_JSGLUE_METHOD_NAME(repo,add_audit), 4,0,0},
    {"rebuild_indexes", SG_JSGLUE_METHOD_NAME(repo,rebuild_indexes), 0,0,0},
    {"create_audit", SG_JSGLUE_METHOD_NAME(repo,create_audit), 2,0,0},
    {"add_vc_comment", SG_JSGLUE_METHOD_NAME(repo,add_vc_comment), 3,0,0},
    {"fetch_vc_comments", SG_JSGLUE_METHOD_NAME(repo,fetch_vc_comments), 1,0,0},
    {"named_branches", SG_JSGLUE_METHOD_NAME(repo,named_branches), 1,0,0},
    {"lookup_branch_names", SG_JSGLUE_METHOD_NAME(repo,lookup_branch_names), 1,0,0},
	{"list_dags", SG_JSGLUE_METHOD_NAME(repo,list_dags), 0, 0, 0},
    {"file_classes", SG_JSGLUE_METHOD_NAME(repo,file_classes), 0, 0, 0},


	/**
	 * boolean repo.compare(string descriptorName)
	 * 
	 * Compares repo to the repository corresponding with descriptorName.
	 * Returns true if all dags and blobs are identical, false if not.
	 */	
	{"compare", SG_JSGLUE_METHOD_NAME(repo,compare),1,0,0},

	/**
	 * string repo.lookup_hid(string prefixOrRevisionID)
	 * 
	 * Returns the full VC changeset HID if the provided prefix 
	 * or revision ID could be found, otherwise throws.
	 */	
	{"lookup_hid", SG_JSGLUE_METHOD_NAME(repo,lookup_hid),1,0,0},

    /**
	 * object repo.lookup_hid(string prefixOrRevisionID)
	 * 
	 * Returns the list VC changeset HIDs if the provided prefix is found
	 * or if revision IDs has matches, otherwise throws.
	 */	
	{"vc_lookup_hids", SG_JSGLUE_METHOD_NAME(repo,vc_lookup_hids),1,0,0},

	/**
	 * string repo.store_blobs_from_files(string filePath1, string filePath2, ...)
	 * 
	 * Stores the contents of the listed files as repo blobs.
	 * Returns an object with the files' HIDs as properties 
	 * and the correcsponding file path as their value.
	 */	
	{"store_blobs_from_files", SG_JSGLUE_METHOD_NAME(repo,store_blobs_from_files),1,0,0},

	/**
	 * string repo.new_since(string startChangesetHid, string endChangesetHid)
	 * 
	 * Returns an array of changeset hids representing nodes that are new between
	 * startChangesetHid and endChangesetHid. In other words, returns the nodes
	 * that are ancestors of endChangesetHid that are NOT ancestors of 
	 * startChangesetHid.
	 */	
	{"new_since", SG_JSGLUE_METHOD_NAME(repo, new_since), 2,0,0},

	/**
	 * object repo.get_changeset_description(string changesetHid)
	 */	
	{"get_changeset_description", SG_JSGLUE_METHOD_NAME(repo, get_changeset_description), 1,0,0},

    /**
	 * object repo.diff_changesets(string changesetHid1, string changesetHid2)
	 */	
	{"diff_changesets", SG_JSGLUE_METHOD_NAME(repo, diff_changesets), 2,0,0},

	JS_FS_END
};

/*
 * properties and methods of a fetchblobhandle
 */
static JSPropertySpec sg_fetchblobhandle_properties[] = {
	{NULL,0,0,NULL,NULL}
};
static JSFunctionSpec sg_fetchblobhandle_methods[] = {
    {"next_chunk", SG_JSGLUE_METHOD_NAME(fetchblobhandle,next_chunk),0,0,0},
    {"abort", SG_JSGLUE_METHOD_NAME(fetchblobhandle,abort),0,0,0},
    JS_FS_END
};

/*
 * properties and methods of a fetchfilehandle
 */
static JSPropertySpec sg_fetchfilehandle_properties[] = {
	{NULL,0,0,NULL,NULL}
};
static JSFunctionSpec sg_fetchfilehandle_methods[] = {
    {"next_chunk", SG_JSGLUE_METHOD_NAME(fetchfilehandle, next_chunk),0,0,0},
    {"abort", SG_JSGLUE_METHOD_NAME(fetchfilehandle, abort),0,0,0},
    JS_FS_END
};

/*
 * These methods are available on the global static "sg" object.
 * example of usage:
 * sg.time()
 * sg.gid()
 */
static JSFunctionSpec sg_methods[] = {
    {"open_repo",            SG_JSGLUE_METHOD_NAME(sg,open_repo),       1,0,0},
	{"open_local_repo",	   SG_JSGLUE_METHOD_NAME(sg,open_local_repo), 0, 0, 0},
//  {"open_pending_tree",  SG_JSGLUE_METHOD_NAME(sg,open_pending_tree),       1,0,0},

    {"clone__specific_changes",             SG_JSGLUE_METHOD_NAME(sg,clone__specific_changes), 2,0,0},
    {"clone__exact",             SG_JSGLUE_METHOD_NAME(sg,clone__exact), 2,0,0},
    {"clone__pack",             SG_JSGLUE_METHOD_NAME(sg,clone__pack), 2,0,0},
    {"clone__all_full",             SG_JSGLUE_METHOD_NAME(sg,clone__all_full), 1,0,0},
    {"clone__all_zlib",             SG_JSGLUE_METHOD_NAME(sg,clone__all_zlib), 1,0,0},
    {"pull_branch_stuff_only",             SG_JSGLUE_METHOD_NAME(sg,pull_branch_stuff_only), 1,0,0},

	/**
	 * void sg.push_all(string src_repo_descriptor, string dest_repo_spec, [bool force=false])
	 */	
	{"push_all",               SG_JSGLUE_METHOD_NAME(sg, push_all), 2,0,0},

	/**
	* void sg.push_branch(string src_repo_descriptor, string dest_repo_spec, string branch_name, [bool force=false])
	*/	
	{"push_branch",               SG_JSGLUE_METHOD_NAME(sg, push_branch), 3,0,0},

	/**
	 * void sg.pull_all(string src_repo_spec, string dest_repo_descriptor)
	 */	
	{"pull_all",               SG_JSGLUE_METHOD_NAME(sg, pull_all), 2,0,0},

    {"zip",               SG_JSGLUE_METHOD_NAME(sg,zip), 3,0,0},

	{"server_state_dir", SG_JSGLUE_METHOD_NAME(sg,server_state_dir), 0,0,0},
    
	{"time",              SG_JSGLUE_METHOD_NAME(sg,time),     0,0,0},
    {"sleep_ms",          SG_JSGLUE_METHOD_NAME(sg,sleep_ms), 1,0,0},
    {"platform",          SG_JSGLUE_METHOD_NAME(sg,platform), 0,0,0},
    {"rand",              SG_JSGLUE_METHOD_NAME(sg,rand),     0,0,0},
    {"make_dagnum",               SG_JSGLUE_METHOD_NAME(sg,make_dagnum),      1,0,0},
    {"base_to_audit",               SG_JSGLUE_METHOD_NAME(sg,base_to_audit),      1,0,0},
    {"gid",               SG_JSGLUE_METHOD_NAME(sg,gid),      0,0,0},
    {"set_local_setting", SG_JSGLUE_METHOD_NAME(sg,set_local_setting),      2,0,0},
	{"set_local_setting__system_wide", SG_JSGLUE_METHOD_NAME(sg,set_local_setting__system_wide),      2,0,0},
    {"local_settings",    SG_JSGLUE_METHOD_NAME(sg,local_settings),      0,0,0},
	{"get_area_id", SG_JSGLUE_METHOD_NAME(sg,get_area_id), 1, 0, 0},

// TODO Now that we allow REPOs to select a Hash Algorithm, HIDs are
// TODO essentially variable length.  We need to have a REPO variable
// TODO on hand before we can request a HID be computed.  I'm commenting
// TODO out this method because I don't think it is being used.  I think
// TODO it was only added for debugging when we first got started.
//    {"hid",             SG_JSGLUE_METHOD_NAME(sg,hid),      1,0,0},

	{"jsmin",				   SG_JSGLUE_METHOD_NAME(sg,jsmin), 1, 0, 0},

    {"exec",            SG_JSGLUE_METHOD_NAME(sg,exec),       1,0,0},
    {"exec_nowait",            SG_JSGLUE_METHOD_NAME(sg,exec_nowait),       1,0,0},
    {"list_descriptors",            SG_JSGLUE_METHOD_NAME(sg,list_descriptors),       0,0,0},
    {"to_json",               SG_JSGLUE_METHOD_NAME(sg,to_json),               1,0,0},
    {"to_json__pretty_print", SG_JSGLUE_METHOD_NAME(sg,to_json__pretty_print), 1,0,0},
    {"to_json__sz",           SG_JSGLUE_METHOD_NAME(sg,to_json__sz),           1,0,0},
    {"to_sz",                 SG_JSGLUE_METHOD_NAME(sg,to_sz),                 1,0,0},
    {"from_json",             SG_JSGLUE_METHOD_NAME(sg,from_json),             1,0,0},
    {"log",                   SG_JSGLUE_METHOD_NAME(sg,log),                   1,0,0},
    {"console",                SG_JSGLUE_METHOD_NAME(sg,console),           1,0,0},
	{"hash",                  SG_JSGLUE_METHOD_NAME(sg,hash),                  2,0,0},
	{"repoid_to_descriptors", SG_JSGLUE_METHOD_NAME(sg,repoid_to_descriptors), 1,0,0},	// should this be debug only
    {NULL,NULL,0,0,0}
};

/*
 * xmlwriter is not a global object; we create them as needed
 */

static JSPropertySpec xmlwriter_properties[] =
{
    {NULL,0,0,NULL,NULL}
};

static JSFunctionSpec xmlwriter_methods[] = {
	/* .start_element(name) */
	{"start_element", SG_JSGLUE_METHOD_NAME(xmlwriter, start_element), 1,0,0},

	/* .attribute(name, value) */
	{"attribute", SG_JSGLUE_METHOD_NAME(xmlwriter, attribute), 2,0,0},

	/* .content(text) */
	{"content", SG_JSGLUE_METHOD_NAME(xmlwriter, content), 1,0,0},

	/* .end_element() */
	{"end_element", SG_JSGLUE_METHOD_NAME(xmlwriter, end_element), 0,0,0},

	/* .element(name, content) */
	{"element", SG_JSGLUE_METHOD_NAME(xmlwriter, element), 2,0,0},

	/* .finish() */
	{"finish", SG_JSGLUE_METHOD_NAME(xmlwriter, finish), 0,0,0},

	{NULL, NULL, 0, 0, 0}
};

/*
 * These methods are available on the global static "sg.fs" object.
 * examples of usage: sg.fs.length("/etc/hosts")
 * sg.fs.exists("/etc/hotss")
 * sg.fs.getcwd()
 */
static JSFunctionSpec fs_methods[] = {
    {"length",            SG_JSGLUE_METHOD_NAME(fs,length),       1,0,0},
    {"remove",            SG_JSGLUE_METHOD_NAME(fs,remove),       1,0,0},
    {"exists",            SG_JSGLUE_METHOD_NAME(fs,exists),       1,0,0},
    {"mkdir",            SG_JSGLUE_METHOD_NAME(fs,mkdir),       1,0,0},
    {"rmdir",            SG_JSGLUE_METHOD_NAME(fs,rmdir),       1,0,0},
    {"mkdir_recursive",   SG_JSGLUE_METHOD_NAME(fs,mkdir_recursive),       1,0,0},
    {"rmdir_recursive",   SG_JSGLUE_METHOD_NAME(fs,rmdir_recursive),       1,0,0},
    {"move",  	      SG_JSGLUE_METHOD_NAME(fs,move),       2,0,0},
    //Gets the Current Working Directory
    {"getcwd",  	  	      SG_JSGLUE_METHOD_NAME(fs,getcwd),       0,0,0},
    {"getcwd_top",  	  	      SG_JSGLUE_METHOD_NAME(fs,getcwd_top),       0,0,0},
    {"chmod",  	  	      SG_JSGLUE_METHOD_NAME(fs,chmod),       2,0,0},
	//This returns a javascript object that contains information about the timestamp and permissions, and other stuff
    {"stat",  	  	      SG_JSGLUE_METHOD_NAME(fs,stat),       1,0,0},
	{"readdir",			SG_JSGLUE_METHOD_NAME(fs,readdir),	1,0,0},
    //Change the Current Working Directory
    {"cd",  	  	      SG_JSGLUE_METHOD_NAME(fs,cd),       1,0,0},
    {"tmpdir", SG_JSGLUE_METHOD_NAME(fs,tmpdir), 0,0,0},
    {"fetch_file",            SG_JSGLUE_METHOD_NAME(fs,fetch_file),       3,0,0},
    // fsobj_type?
    // perms?
    // time?
    {NULL,NULL,0,0,0}
};

/*
 * These methods are available on the global static "sg.server" object.
 * example usage: sg.server.debug_shutdown()
 */
static JSFunctionSpec server_methods[] = {
	{"debug_shutdown", SG_JSGLUE_METHOD_NAME(server,debug_shutdown), 0,0,0},
	{NULL,NULL,0,0,0}
};

/*
 * These methods are available on the global static "sg.pending_tree" object.
 * examples of usage: sg.pending_tree.add("relative/path/to/object")
 * sg.pending_tree.status()
 */
static JSFunctionSpec pending_tree_methods[] = {
    //{"close",            SG_JSGLUE_METHOD_NAME(pending_tree,close),       0,0,0},
    {"add",            SG_JSGLUE_METHOD_NAME(pending_tree,add),       1,0,0},
    {"commit",            SG_JSGLUE_METHOD_NAME(pending_tree,commit),       0,0,0},
    {"remove",            SG_JSGLUE_METHOD_NAME(pending_tree,remove),       1,0,0},
    {"rename",            SG_JSGLUE_METHOD_NAME(pending_tree,rename),       2,0,0},
    {"move",            SG_JSGLUE_METHOD_NAME(pending_tree,move),       2,0,0},
    {"scan",            SG_JSGLUE_METHOD_NAME(pending_tree,scan),       0,0,0},
    {"addremove",            SG_JSGLUE_METHOD_NAME(pending_tree,addremove),       0,0,0},
    //This returns the history as an array.
    {"history",            SG_JSGLUE_METHOD_NAME(pending_tree,history),       0,0,0},
	{"stamp",            SG_JSGLUE_METHOD_NAME(pending_tree,stamp),       0,0,0},
	{"stamps",            SG_JSGLUE_METHOD_NAME(pending_tree,stamps),       0,0,0},
    //This returns the status as an array.
    {"status",            SG_JSGLUE_METHOD_NAME(pending_tree,status),       0,0,0},
    //This returns the status as a string, formatted exactly like the CLC does it.
    {"status_string",            SG_JSGLUE_METHOD_NAME(pending_tree,status_string),       0,0,0},
    {"tag",            SG_JSGLUE_METHOD_NAME(pending_tree,tag),       0,0,0},
    {"tags",            SG_JSGLUE_METHOD_NAME(pending_tree,tags),       0,0,0},
    {"detach",            SG_JSGLUE_METHOD_NAME(pending_tree,detach),       0,0,0},
	{"attach",            SG_JSGLUE_METHOD_NAME(pending_tree,attach),       1,0,0},
    {"parents",            SG_JSGLUE_METHOD_NAME(pending_tree,parents),       0,0,0},
    {"leaves",             SG_JSGLUE_METHOD_NAME(pending_tree,leaves),        0,0,0},
    {"manually_set_parents",            SG_JSGLUE_METHOD_NAME(pending_tree,manually_set_parents),       0,0,0},
    {NULL,NULL,0,0,0}
};
/*
 * These methods are available on the global static "sg.file" object.
 * examples of usage: sg.file.write("relative/path/to/object", "file contents")
 * sg.file.append("file.txt", "new line to append")
 */
static JSFunctionSpec file_methods[] = {
    {"write",            SG_JSGLUE_METHOD_NAME(file,write),       2,0,0},
    {"append",            SG_JSGLUE_METHOD_NAME(file,append),       2,0,0},
    {"read",            SG_JSGLUE_METHOD_NAME(file,read),       1,0,0},
    {NULL,NULL,0,0,0}
};

// TODO need class/methods to do SG_pathname operations

//////////////////////////////////////////////////////////////////

#include "sg_jsglue__private_setprop.h"
#include "sg_jsglue__private_sg_config.h"
#include "sg_jsglue__private_sg_vv.h"
#include "sg_jsglue__private_sync_remote.h"

//////////////////////////////////////////////////////////////////

/* ---------------------------------------------------------------- */
/* This is the function which sets everything up */
/* ---------------------------------------------------------------- */

void SG_jsglue__install_scripting_api(SG_context * pCtx,
									  JSContext *cx, JSObject *glob)
{
	const char * szVersion = NULL;
    JSObject* pjsobj_sg = NULL;
	JSObject* pjsobj_vv = NULL;
    JSObject* pjsobj_fs = NULL;
	JSObject* pjsobj_sync_remote = NULL;
    JSObject* pjsobj_server = NULL;
    JSObject* pjsobj_pending_tree = NULL;
    JSObject* pjsobj_file = NULL;
    JSObject* pjsobj_dagnums = NULL;
    JSObject* pjsobj_blobencodings = NULL;
	JSObject* pjsobj_vendors = NULL;
    jsval jv;
	char buf_dagnum[SG_DAGNUM__BUF_MAX__HEX];

	SG_JS_NULL_CHECK(  pjsobj_sg = JS_DefineObject(cx, glob, "sg", &sg_class, NULL, 0)  );
	SG_JS_BOOL_CHECK(  JS_DefineFunctions(cx, pjsobj_sg, sg_methods)  );

	// TODO 2010/11/30 This defines "sg.version".  But we also have "sg.config.version.*"
	// TODO            (defined in sg_jsglue__private_sg_config.h).  Shouldn't we combine these?

	SG_ERR_CHECK(  SG_lib__version(pCtx, &szVersion)  );
	JSVAL_FROM_SZ(jv, szVersion);
	SG_JS_BOOL_CHECK(  JS_SetProperty(cx, pjsobj_sg, "version", &jv)  );

	SG_JS_NULL_CHECK(  pjsobj_vv = JS_DefineObject(cx, pjsobj_sg, "vv", &sg_vv__class, NULL, 0)  );
	SG_JS_BOOL_CHECK(  JS_DefineFunctions(cx, pjsobj_vv, sg_vv__methods)  );

	SG_JS_NULL_CHECK(  pjsobj_fs = JS_DefineObject(cx, pjsobj_sg, "fs", &fs_class, NULL, 0)  );
	SG_JS_BOOL_CHECK(  JS_DefineFunctions(cx, pjsobj_fs, fs_methods)  );

	SG_JS_NULL_CHECK(  pjsobj_sync_remote = JS_DefineObject(cx, pjsobj_sg, "sync_remote", &sg_sync_remote__class, NULL, 0)  );
	SG_JS_BOOL_CHECK(  JS_DefineFunctions(cx, pjsobj_sync_remote, sg_sync_remote__methods)  );

	SG_JS_NULL_CHECK(  pjsobj_server = JS_DefineObject(cx, pjsobj_sg, "server", &server_class, NULL, 0)  );
	SG_JS_BOOL_CHECK(  JS_DefineFunctions(cx, pjsobj_server, server_methods)  );

	SG_JS_NULL_CHECK(  pjsobj_pending_tree = JS_DefineObject(cx, pjsobj_sg, "pending_tree", &pending_tree_class, NULL, 0)  );
	SG_JS_BOOL_CHECK(  JS_DefineFunctions(cx, pjsobj_pending_tree, pending_tree_methods)  );

    SG_JS_NULL_CHECK(  pjsobj_file = JS_DefineObject(cx, pjsobj_sg, "file", &file_class, NULL, 0)  );
	SG_JS_BOOL_CHECK(  JS_DefineFunctions(cx, pjsobj_file, file_methods)  );

	// Define global sg.config. array with a list of known values.
	SG_ERR_CHECK(  sg_jsglue__install_sg_config(pCtx, cx, pjsobj_sg)  );

    SG_JS_NULL_CHECK(  pjsobj_blobencodings = JS_DefineObject(cx, pjsobj_sg, "blobencoding", NULL, NULL, 0)  );

    jv = INT_TO_JSVAL(SG_BLOBENCODING__FULL);
    SG_JS_BOOL_CHECK(  JS_SetProperty(cx, pjsobj_blobencodings, "FULL", &jv)  );

    jv = INT_TO_JSVAL(SG_BLOBENCODING__KEEPFULLFORNOW);
    SG_JS_BOOL_CHECK(  JS_SetProperty(cx, pjsobj_blobencodings, "KEEPFULLFORNOW", &jv)  );

    jv = INT_TO_JSVAL(SG_BLOBENCODING__ALWAYSFULL);
    SG_JS_BOOL_CHECK(  JS_SetProperty(cx, pjsobj_blobencodings, "ALWAYSFULL", &jv)  );

    jv = INT_TO_JSVAL(SG_BLOBENCODING__ZLIB);
    SG_JS_BOOL_CHECK(  JS_SetProperty(cx, pjsobj_blobencodings, "ZLIB", &jv)  );

    jv = INT_TO_JSVAL(SG_BLOBENCODING__VCDIFF);
    SG_JS_BOOL_CHECK(  JS_SetProperty(cx, pjsobj_blobencodings, "VCDIFF", &jv)  );

	SG_JS_NULL_CHECK(  pjsobj_vendors = JS_DefineObject(cx, pjsobj_sg, "vendor", NULL, NULL, 0)  );

	jv = INT_TO_JSVAL(SG_VENDOR__SOURCEGEAR);
	SG_JS_BOOL_CHECK(  JS_SetProperty(cx, pjsobj_vendors, "SOURCEGEAR", &jv)  );

    SG_JS_NULL_CHECK(  pjsobj_dagnums = JS_DefineObject(cx, pjsobj_sg, "dagnum", NULL, NULL, 0)  );

	SG_ERR_CHECK(  SG_dagnum__to_sz__hex(pCtx, SG_DAGNUM__VERSION_CONTROL, buf_dagnum, sizeof(buf_dagnum))  );
    JSVAL_FROM_SZ(jv, buf_dagnum);
    SG_JS_BOOL_CHECK(  JS_SetProperty(cx, pjsobj_dagnums, "VERSION_CONTROL", &jv)  );

#if 0
	SG_ERR_CHECK(  SG_dagnum__to_sz__hex(pCtx, SG_DAGNUM__WORK_ITEMS, buf_dagnum, sizeof(buf_dagnum))  );
    JSVAL_FROM_SZ(jv, buf_dagnum);
    SG_JS_BOOL_CHECK(  JS_SetProperty(cx, pjsobj_dagnums, "WORK_ITEMS", &jv)  );

	SG_ERR_CHECK(  SG_dagnum__to_sz__hex(pCtx, SG_DAGNUM__BUILDS, buf_dagnum, sizeof(buf_dagnum))  );
    jv = STRING_TO_JSVAL(JS_NewStringCopyZ(cx, buf_dagnum));
	SG_JS_BOOL_CHECK(  JS_SetProperty(cx, pjsobj_dagnums, "BUILDS", &jv)  );
#endif

	SG_ERR_CHECK(  SG_dagnum__to_sz__hex(pCtx, SG_DAGNUM__USERS, buf_dagnum, sizeof(buf_dagnum))  );
    JSVAL_FROM_SZ(jv, buf_dagnum);
    SG_JS_BOOL_CHECK(  JS_SetProperty(cx, pjsobj_dagnums, "USERS", &jv)  );

	SG_ERR_CHECK(  SG_dagnum__to_sz__hex(pCtx, SG_DAGNUM__AREAS, buf_dagnum, sizeof(buf_dagnum))  );
    JSVAL_FROM_SZ(jv, buf_dagnum);
    SG_JS_BOOL_CHECK(  JS_SetProperty(cx, pjsobj_dagnums, "AREAS", &jv)  );

	SG_ERR_CHECK(  SG_dagnum__to_sz__hex(pCtx, SG_DAGNUM__VC_COMMENTS, buf_dagnum, sizeof(buf_dagnum))  );
    JSVAL_FROM_SZ(jv, buf_dagnum);
    SG_JS_BOOL_CHECK(  JS_SetProperty(cx, pjsobj_dagnums, "VC_COMMENTS", &jv)  );

	SG_ERR_CHECK(  SG_dagnum__to_sz__hex(pCtx, SG_DAGNUM__VC_TAGS, buf_dagnum, sizeof(buf_dagnum))  );
    JSVAL_FROM_SZ(jv, buf_dagnum);
    SG_JS_BOOL_CHECK(  JS_SetProperty(cx, pjsobj_dagnums, "VC_TAGS", &jv)  );

	SG_ERR_CHECK(  SG_dagnum__to_sz__hex(pCtx, SG_DAGNUM__VC_BRANCHES, buf_dagnum, sizeof(buf_dagnum))  );
    JSVAL_FROM_SZ(jv, buf_dagnum);
    SG_JS_BOOL_CHECK(  JS_SetProperty(cx, pjsobj_dagnums, "VC_BRANCHES", &jv)  );

	SG_ERR_CHECK(  SG_dagnum__to_sz__hex(pCtx, SG_DAGNUM__VC_STAMPS, buf_dagnum, sizeof(buf_dagnum))  );
    JSVAL_FROM_SZ(jv, buf_dagnum);
    SG_JS_BOOL_CHECK(  JS_SetProperty(cx, pjsobj_dagnums, "VC_STAMPS", &jv)  );

	SG_ERR_CHECK(  SG_dagnum__to_sz__hex(pCtx, SG_DAGNUM__VC_HOOKS, buf_dagnum, sizeof(buf_dagnum))  );
    jv = STRING_TO_JSVAL(JS_NewStringCopyZ(cx, buf_dagnum));
    SG_JS_BOOL_CHECK(  JS_SetProperty(cx, pjsobj_dagnums, "VC_HOOKS", &jv)  );

// the following are only used in the test suite
#define SG_DAGNUM__TESTING__DB   \
        SG_DAGNUM__SET_VENDOR(SG_VENDOR__SOURCEGEAR, \
        SG_DAGNUM__SET_GROUPING(SG_GROUPING__SOURCEGEAR__TESTING, \
        SG_DAGNUM__SET_DAGID(1, \
        SG_DAGNUM__SET_TYPE(sg_DAGNUM__TYPE__DB, \
        ((SG_uint64) 0)))))

#define SG_DAGNUM__TESTING2__DB   \
        SG_DAGNUM__SET_VENDOR(SG_VENDOR__SOURCEGEAR, \
        SG_DAGNUM__SET_GROUPING(SG_GROUPING__SOURCEGEAR__TESTING, \
        SG_DAGNUM__SET_DAGID(2, \
        SG_DAGNUM__SET_TYPE(sg_DAGNUM__TYPE__DB, \
        ((SG_uint64) 0)))))

	SG_ERR_CHECK(  SG_dagnum__to_sz__hex(pCtx, SG_DAGNUM__TESTING__DB, buf_dagnum, sizeof(buf_dagnum))  );
    JSVAL_FROM_SZ(jv, buf_dagnum);
    SG_JS_BOOL_CHECK(  JS_SetProperty(cx, pjsobj_dagnums, "TESTING_DB", &jv)  );

	SG_ERR_CHECK(  SG_dagnum__to_sz__hex(pCtx, SG_DAGNUM__TESTING2__DB, buf_dagnum, sizeof(buf_dagnum))  );
    JSVAL_FROM_SZ(jv, buf_dagnum);
    SG_JS_BOOL_CHECK(  JS_SetProperty(cx, pjsobj_dagnums, "TESTING2_DB", &jv)  );

    SG_JS_NULL_CHECK(  JS_InitClass(
            cx,
            glob,
            NULL, /* parent proto */
            &sg_audit_class,
            NULL, /* no constructor */
            0, /* nargs */
            sg_audit_properties,
            sg_audit_methods,
            NULL,  /* static properties */
            NULL   /* static methods */
            )  );

    SG_JS_NULL_CHECK(  JS_InitClass(
            cx,
            glob,
            NULL, /* parent proto */
            &sg_cbuffer_class,
            NULL, /* no constructor */
            0, /* nargs */
            sg_cbuffer_properties,
            sg_cbuffer_methods,
            NULL,  /* static properties */
            NULL   /* static methods */
            )  );

    SG_JS_NULL_CHECK(  JS_InitClass(
            cx,
            glob,
            NULL, /* parent proto */
            &sg_repo_class,
            NULL, /* no constructor */
            0, /* nargs */
            sg_repo_properties,
            sg_repo_methods,
            NULL,  /* static properties */
            NULL   /* static methods */
            )  );

    SG_JS_NULL_CHECK(  JS_InitClass(
            cx,
            glob,
            NULL, /* parent proto */
            &sg_blobset_class,
            NULL, /* no constructor */
            0, /* nargs */
            sg_blobset_properties,
            sg_blobset_methods,
            NULL,  /* static properties */
            NULL   /* static methods */
            )  );

    SG_JS_NULL_CHECK(  JS_InitClass(
            cx,
            glob,
            NULL, /* parent proto */
            &sg_fetchblobhandle_class,
            NULL, /* no constructor */
            0, /* nargs */
            sg_fetchblobhandle_properties,
            sg_fetchblobhandle_methods,
            NULL,  /* static properties */
            NULL   /* static methods */
            )  );

    SG_JS_NULL_CHECK(  JS_InitClass(
            cx,
            glob,
            NULL, /* parent proto */
            &sg_fetchfilehandle_class,
            NULL, /* no constructor */
            0, /* nargs */
            sg_fetchfilehandle_properties,
            sg_fetchfilehandle_methods,
            NULL,  /* static properties */
            NULL   /* static methods */
            )  );
      

	SG_JS_NULL_CHECK(  JS_InitClass(
            cx,
            glob,
            NULL, /* parent proto */
            &xmlwriter_class,
            xmlwriter_constructor,
            0, /* nargs */
            xmlwriter_properties,
            xmlwriter_methods,
            NULL,  /* static properties */
            NULL   /* static methods */
            )  );

    return;

fail:
    // TODO we had an error installing stuff.
    // TODO do we need to tear down anything that we just created?
    return;
}

#if 0
/*
TODO 3/31/10 Things Jeff wants in the JS glue:
[] Need way to tell which platform we are on so that we can know
   about AttrBits and XAttrs.
[] Need interfaces to AttrBits and XAttrs as appropriate.
[] Need fsobj.symlink.
[] Need way to do readdir() and iterate over contents of directory.
[] Need way to read treenode (maybe?)
[] Need to do SG_wdrepo changes.
[] When pendingtree operation throw a portability warning, we need
   to get the varray/vhash of problems back to inspect/print.
[] pendingtree.rename(path,entryname) and pendingtree.move(path,directory), right?
[] the "_" in "pending_tree" class is hard to type after typing SG_pendingtree.
[] several routines (like pending_tree.move) allocate the includes/excludes
   arrays before they know whether they need them.  we should defer those allocs
   until we actually see a --include/--exclude since the SG_pendingtree__ code
   properly handles null pointers for them.
*/
#endif
