/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include <sg.h>

//////////////////////////////////////////////////////////////////

#if defined(DEBUG)
#	define TRACE_PENDINGTREE		0
#	define TRACE_COMMIT				0
#	define TRACE_ADD				0
#	define TRACE_ADDREMOVE			0
#	define TRACE_REVERT				0
#	define TRACE_REMOVE				0
#	define TRACE_CHECK_PORT			0
#	define TRACE_RENAME				0
#	define TRACE_MOVE				0
#else
#	define TRACE_PENDINGTREE		0
#	define TRACE_COMMIT				0
#	define TRACE_ADD				0
#	define TRACE_ADDREMOVE			0
#	define TRACE_REVERT				0
#	define TRACE_REMOVE				0
#	define TRACE_CHECK_PORT			0
#	define TRACE_RENAME				0
#	define TRACE_MOVE				0
#endif

//////////////////////////////////////////////////////////////////

/* TODO many of the SG_pendingtree__ functions save the tree when they
 * exit.  Not sure this is a good idea, because we can't do two
 * operations in a row.
 *
 * TODO [X] this is potentially bad because the chunk of code that says:
 *         // we don't change the saved pending changeset here.
 *         sg_ptnode__free(pPendingTree->pSuperRoot);
 *         pPendingTree->pSuperRoot = NULL;
 * only gets called when the main function is successful.  If something
 * low level takes an error, the free will get called by the containing
 * function and it may write out the data.
 *
 * We should probably change this.
 *
 */

#define sg_PTNODE_TEMP_FLAG_CHECKMYNAME    2

/* sg_PTNODE_TEMP_FLAG_COMMITTING means that the item is in the process
 * of being committed to the repo.  When a commit operation begins,
 * first we go through the tree and set this COMMITTING flag on
 * every item which is supposed to be included in the commit.  If
 * nothing has been specified, we assume we are supposed to commit all
 * changes and we just set the flag on every node.  For every file
 * which has been specified, we set the flag and bubble it up.  For
 * every directory that has been specified, we set the flag on it and
 * all its descendants and also bubble it up. */

#define sg_PTNODE_TEMP_FLAG_COMMITTING    4

#define sg_PTNODE_SAVED_FLAG_DELETED    8

#define sg_PTNODE_SAVED_FLAG_MOVED_AWAY    16

#define sg_PTNODE_TEMP_FLAG_IMPLICIT_DELETE 32

#define sg_PTNODE_TEMP_FLAG_REVERTING    64

#define sg_PTNODE_TEMP_FLAG_IMPLICIT_ADD 256

#define sg_PTNODE_TEMP_FLAG_PARKED_FOR_CYCLE 512

#define sg_PTNODE_TEMP_FLAG_NESTED_SUBMODULE_COMMIT 1024

/* Some items in the ptnode are CURRENT.  They tell us information
 * which is reflected in the pending tree, in the working copy.
 * Other items are BASELINE.  They tell us information about how
 * things were in the baseline version where we started. */
struct sg_ptnode
{
	SG_pendingtree* pPendingTree; /* never null */

	struct sg_ptnode* pCurrentParent; /* null for the root treenode
									   * only.  This is the CURRENT
									   * parent. */
	
	const char* pszgidMovedFrom; /* non-null iff this item was moved here
								  * from somewhere else. used to check
								  * for split moves */

	SG_treenode_entry* pBaselineEntry; /* NULL if this item wasn't in
										* the baseline.  Also NULL for
										* the very top node in the
										* tree.  If non-NULL, this is
										* a copy which needs to be
										* freed. */

	SG_treenode_entry* pNewEntry; /* During commit, each item which
								   * appears in the committed tree
								   * gets a new entry, which is stored
								   * here.  Post-commit this entry is
								   * moved to pBaselineEntry. */

	const char* pszgid; /* This is here for convenience.  If
						 * pBaselineEntry is present, its gid will
						 * always be the same as this one.  If
						 * pCurrentParent is non-NULL, then this gid
						 * is the same gid which is used for the key
						 * in pCurrentParent->prbItems. Note that this
						 * member is not identified as current or
						 * baseline, because it cannot change. */
	
	const char* pszCurrentName; /* If NULL, use the name in the entry. */
	
	const char* pszCurrentHid;
    const char* pszHidXattrs;
	const char * pszHidDisposable;
	SG_bool bDisposableDir;		// TODO 2011/03/07 should this have just been another _SAVED_FLAG_ ?
	SG_bool bDeletedAfterAdded;	// Set when we have an ADDED+DELETED and can't just remove the ptnode (after merge)
	SG_bool bStickyDir;         // somebody needs this ptnode to always be available even when clean (see merge/resolve)

    SG_int64 iCurrentAttributeBits;

	SG_treenode_entry_type type;
	SG_uint16 saved_flags;
	SG_uint16 temp_flags;
	
	/* Stuff that's valid only for type == DIRECTORY */
	SG_rbtree* prbItems;  /* NULL until it's needed. */

	// Stuff that is only valid for a submodule.  This is only set
	// when somebody has examined the submodule and wants to cache
	// the TSM.  We can clone the TSM from the entry in scan-dir and
	// avoid re-computing it during commits, for example.
	// This is a TEMPORARY field; we DO NOT store this in pending.json
	// because (like the contents of a file) the user can make changes
	// to the submodule (like they can the contents of a file) without
	// needing to tell us.
	SG_treenode_submodule * pTSM;	// NULL until it is needed.

	struct sg_ptnode * ptnPhantomMe; /* transient. only set for MOVES during REVERT.
									  * this helps us undo moves when there are path
									  * cycles (and the tree is temporarily broken
									  * (either disconnected, cyclic, or both)). */

	/* Note that there is no local path here.  Here's why: If there
	 * were, then a rename operation on a directory node would have to
	 * go through all its child nodes and fix up the path.  Not so
	 * bad, but then a revert of that rename operation would have to
	 * go through and fix them all.  Better to just not have a path
	 * stored here at all.  Instead, when the path is needed (during
	 * commit, for adding a file blob), we simply calculate it during
	 * commit. */
};

struct _SG_pendingtree
{
	SG_string* pstrRepoDescriptorName;
	SG_pathname* pPathWorkingDirectoryTop;
	SG_strpool* pStrPool;
	SG_repo* pRepo;
	struct sg_ptnode* pSuperRoot;

	// 4/21/10 We hold the ".sgdrawer/wd.json" VFILE open and locked for the
	// duration of an operation.  We also keep the VHASH in memory.  These
	// contain was we used to keep in "repo.json" and "pending.json".
	SG_vfile * pvf_wd;
	SG_vhash * pvh_wd;

#if defined(DEBUG)
	SG_bool b_pvh_wd__parents__dirty;		// someone called __set_wd_parents() but not __save()
	SG_bool b_pvh_wd__issues__dirty;		// someone called __set_wd_issues() but not __save()
	SG_bool b_pvh_wd__choices__dirty;		// someone called __set_wd_choices() but not __save()
#endif

	SG_utf8_converter *			pConverter;
    SG_varray *					pvaWarnings;
	SG_portability_flags		portMask;

	SG_pendingtree_action_log_enum		eActionLog;
	SG_varray *							pvaActionLog;	// varray[<vhash *>] of details

	// We now optionally keep a handle to the timestamp_cache
	// rather than holding the individual rbtree and rbtreedb
	// fields.  We only load this when we need it.
	//
	// TODO 2010/08/11 Comment on when you should save the cache
	// TODO            relative to saving the pendingtree.  And
	// TODO            about saving the cache even when you don't
	// TODO            save the pendingtree.

	SG_timestamp_cache *				pTSC;

	/* TODO ndx ? */

	SG_pendingtree_update_data *	pud;	// NULL until needed by __precompute_update and __apply_update

	SG_bool bDebugExport;	// true if we are in __export_to_vhash and want extra info

	SG_rbtree * prbQueuedPortabilityChecks;		// map[gidDir --> ptnDir]

};

//////////////////////////////////////////////////////////////////

static void sg_ptnode__free(SG_context* pCtx, struct sg_ptnode* ptn);

#define SG_PTNODE_NULLFREE(pCtx, p)    SG_STATEMENT( SG_context__push_level(pCtx); sg_ptnode__free(pCtx, p); SG_ASSERT(!SG_context__has_err(pCtx)); SG_context__pop_level(pCtx);  p=NULL; )

//////////////////////////////////////////////////////////////////

static void sg_ptnode__get_name(SG_context* pCtx, 
	const struct sg_ptnode* ptn,
	const char** ppszName
	);

static void sg_ptnode__set_disposable_hid(SG_context * pCtx,
										  struct sg_ptnode * ptn,
										  const char * pszHid);
static void sg_ptnode__set_disposable_dir(SG_context * pCtx,
										  struct sg_ptnode * ptn);

static void sg_ptnode__set_sticky_dir(SG_context * pCtx, struct sg_ptnode * ptn);

//////////////////////////////////////////////////////////////////

static void _sg_pendingtree__do_diff__unload_clean_objects(SG_context* pCtx, SG_pendingtree * pPendingTree,
														   const char * szHidArbitraryCSet,		/* optional */
														   SG_treediff2 * pTreeDiff_cache_owner,	/* optional */
														   SG_treediff2 ** ppTreeDiff);

static void sg_pendingtree__hack__unadd_dont_save__by_ptn(SG_context * pCtx,
												   struct sg_ptnode * ptn);

//////////////////////////////////////////////////////////////////

/**
 * Allocate a utf-8-->? converter for testing for portability problems.
 * Since we internally store all entrynames in UTF-8, we *may* have a
 * problem expressing them on a platform that requires us to convert them
 * to an OS/native/locale/whatever (non-unicode) encoding.
 *
 * We allocate the proper converter and stuff it in the pendingtree, so
 * subsequent calls will just reference this value; only the first call
 * will actually allocate it.
 */
static void sg_pendingtree__configure_portability_converter(SG_context * pCtx, SG_pendingtree * pPendingTree)
{
	if (!pPendingTree->pConverter)
	{
		const char * pszCharSetName = NULL;

		// TODO determine if the REPO has a specific LOCALE/CHARSET defined and fetch it.
		// TODO for now, we assume UTF-8.

		SG_ERR_CHECK_RETURN(  SG_utf8_converter__alloc(pCtx,pszCharSetName,&pPendingTree->pConverter)  );
	}
}

//////////////////////////////////////////////////////////////////

static void sg_ptnode__alter_name(SG_context * pCtx,
								  SG_pendingtree * pPendingTree,
								  struct sg_ptnode * ptnEntry,
								  const char * pszNewName)
{
	if (pszNewName && *pszNewName)
		SG_ERR_CHECK_RETURN(  SG_strpool__add__sz(pCtx, pPendingTree->pStrPool, pszNewName, &ptnEntry->pszCurrentName)  );
	else
		ptnEntry->pszCurrentName = NULL;

    ptnEntry->temp_flags |= sg_PTNODE_TEMP_FLAG_CHECKMYNAME;
}

//////////////////////////////////////////////////////////////////

static void _do_portability_check(SG_context * pCtx,
								  struct sg_ptnode * ptnDir,
								  SG_portability_flags * pPortWarningsLogged)
{
	SG_bool b = SG_FALSE;
	const char* pszGid = NULL;
	struct sg_ptnode* ptnSub = NULL;
	SG_rbtree_iterator* pIterator = NULL;
	SG_portability_dir* pPortabilityDir = NULL;

	// TODO jeff says: I think we need to revisit the NULL for relative-path in the call
	// TODO below.  If we don't have the repo-path for the parent directory, we can't do
	// TODO some of the pathname-length checks.  Granted we won't get them all because
	// TODO the "@/" is just a placeholder for whereever they create their working copy,
	// TODO but we can catch the cases where the part of the repo-path after the "@/" is
	// TODO really long and there's a long entryname and the result is > 256 and might
	// TODO cause problems on other platforms.

	SG_ERR_CHECK(  SG_portability_dir__alloc(pCtx,
											 ptnDir->pPendingTree->portMask,
											 NULL, /* no relative path needed, I think */
											 ptnDir->pPendingTree->pConverter,
											 ptnDir->pPendingTree->pvaWarnings,
											 &pPortabilityDir)  );

    SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pIterator, ptnDir->prbItems, &b, &pszGid, (void**) &ptnSub)  );
    while (b)
    {
		if (
			(ptnSub->saved_flags & sg_PTNODE_SAVED_FLAG_DELETED)
			|| (ptnSub->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY)
			|| (ptnSub->temp_flags & sg_PTNODE_TEMP_FLAG_IMPLICIT_ADD)
			)
		{
		}
		else
		{
            const char* pszName = NULL;
			SG_bool bIsDuplicate = SG_FALSE;

			SG_ERR_CHECK(  sg_ptnode__get_name(pCtx, ptnSub, &pszName)  );

            //////////////////////////////////////////////////////////////////
            // compute the portability and/or potential portability problems with this entryname.
            // this includes filenames that are/may be invalid on some platform and *possible*
            // collisions on some platforms due to things like case-folding.
            // 
            // we only log a complaint about an item if it is a new/renamed/moved item.
            // we always compute the info because we need the collision data.

            SG_ERR_CHECK(  SG_portability_dir__add_item(pCtx, 
														pPortabilityDir,
														NULL,
														pszName,
														((ptnSub->temp_flags & sg_PTNODE_TEMP_FLAG_CHECKMYNAME)?SG_TRUE:SG_FALSE),
														((ptnSub->temp_flags & sg_PTNODE_TEMP_FLAG_CHECKMYNAME)?SG_TRUE:SG_FALSE),
														&bIsDuplicate)  );

			// TODO the original SG_portability_dir__add_item() threw SG_ERR_RBTREE_DUPLICATEKEY when
            // TODO you have 2 identical strings in this directory.  THIS **CAN** HAPPEN IN THINGS
            // TODO LIKE "vv rename" BECAUSE WE DO THE PORTABILITY TEST BEFORE WE MODIFY THE DIRECTORY.
			// TODO
			// TODO I changed the portability code to return a flag for that case rather than throwing an error.
            // TODO
            // TODO I need to revisit the callers of this function and see if they'd rather have the error or
            // TODO a flag (or a different error code).  But for now, I throw the original error code

			if (bIsDuplicate)
				SG_ERR_THROW2(  SG_ERR_RBTREE_DUPLICATEKEY,
								(pCtx, "%s", pszName)  );
        }
        
        SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pIterator, &b, &pszGid, (void**) &ptnSub)  );
    }

	SG_ERR_CHECK(  SG_portability_dir__get_result_flags(pCtx, pPortabilityDir, NULL, NULL, pPortWarningsLogged)  );

fail:
	SG_PORTABILITY_DIR_NULLFREE(pCtx, pPortabilityDir);
    SG_RBTREE_ITERATOR_NULLFREE(pCtx, pIterator);
}

static void sg_ptnode__immediate_portability_check(SG_context * pCtx, struct sg_ptnode * ptnDir)
{
	SG_portability_flags portWarningsLogged = SG_PORT_FLAGS__NONE;

	SG_NULLARGCHECK_RETURN(ptnDir);

#if TRACE_CHECK_PORT
	{
		SG_uint32 count = 0;

		if (ptnDir->prbItems)
			SG_ERR_CHECK(  SG_rbtree__count(pCtx, ptnDir->prbItems, &count)  );
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "sg_ptnode__immediate_portability_check: [gid %s][isDir %d][nrItems %d]\n",
								   ptnDir->pszgid,
								   (ptnDir->type == SG_TREENODEENTRY_TYPE_DIRECTORY),
								   count)  );
	}
#endif

	if (ptnDir->type != SG_TREENODEENTRY_TYPE_DIRECTORY)
        return;

    if (!ptnDir->prbItems)
        return;

	// this can still throw _RBTREE_DUPLICATEKEY
	SG_ERR_CHECK(  _do_portability_check(pCtx, ptnDir, &portWarningsLogged)  );

	if (portWarningsLogged != SG_PORT_FLAGS__NONE)
		SG_ERR_THROW(  SG_ERR_PORTABILITY_WARNINGS  );

fail:
	return;
}

static void sg_ptnode__queue_portability_check(SG_context * pCtx, struct sg_ptnode * ptnDir)
{
	SG_NULLARGCHECK_RETURN(ptnDir);

#if TRACE_CHECK_PORT
	{
		SG_uint32 count = 0;

		if (ptnDir->prbItems)
			SG_ERR_CHECK(  SG_rbtree__count(pCtx, ptnDir->prbItems, &count)  );
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "sg_ptnode__queue_portability_check: queueing [gid %s][isDir %d][nrItems %d]\n",
								   ptnDir->pszgid,
								   (ptnDir->type == SG_TREENODEENTRY_TYPE_DIRECTORY),
								   count)  );
	}
#endif

	if (ptnDir->type != SG_TREENODEENTRY_TYPE_DIRECTORY)
        return;

    if (!ptnDir->prbItems)
        return;

	if (!ptnDir->pPendingTree->prbQueuedPortabilityChecks)
		SG_ERR_CHECK(  SG_RBTREE__ALLOC(pCtx, &ptnDir->pPendingTree->prbQueuedPortabilityChecks)  );

	SG_ERR_CHECK(  SG_rbtree__update__with_assoc(pCtx,
												 ptnDir->pPendingTree->prbQueuedPortabilityChecks,
												 ptnDir->pszgid,
												 (void *)ptnDir,
												 NULL)  );

fail:
	return;
}

static void sg_ptnode__do_queued_portability_checks(SG_context * pCtx,
													SG_pendingtree * pPendingTree)
{
	SG_rbtree_iterator * pIter = NULL;
	SG_portability_flags portWarningsLogged_sum = SG_PORT_FLAGS__NONE;
	SG_rbtree * prb_owned = NULL;

#if TRACE_CHECK_PORT
	{
		SG_uint32 count = 0;

		if (pPendingTree->prbQueuedPortabilityChecks)
			SG_ERR_CHECK(  SG_rbtree__count(pCtx, pPendingTree->prbQueuedPortabilityChecks, &count)  );

		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "sg_ptnode__do_queued_portability_checks: [nrDirs %d]\n",
								   count)  );
	}
#endif

	if (pPendingTree->prbQueuedPortabilityChecks)
	{
		const char * pszgid_k;
		struct sg_ptnode * ptnDir_k;
		SG_bool b;

		prb_owned = pPendingTree->prbQueuedPortabilityChecks;	// steal the rbtree from pendingtree
		pPendingTree->prbQueuedPortabilityChecks = NULL;

		SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pIter, prb_owned,
												  &b, &pszgid_k, (void **)&ptnDir_k)  );
		while (b)
		{
			SG_portability_flags portWarningsLogged_k = SG_PORT_FLAGS__NONE;

			// accumulate any portability warnings.
			// this can still throw _RBTREE_DUPLICATEKEY.
			SG_ERR_CHECK(  _do_portability_check(pCtx, ptnDir_k, &portWarningsLogged_k)  );

			portWarningsLogged_sum |= portWarningsLogged_k;

			SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pIter,
													 &b, &pszgid_k, (void**)&ptnDir_k)  );
		}

		if (portWarningsLogged_sum != SG_PORT_FLAGS__NONE)
			SG_ERR_THROW(  SG_ERR_PORTABILITY_WARNINGS  );
	}

fail:
    SG_RBTREE_ITERATOR_NULLFREE(pCtx, pIter);
	SG_RBTREE_NULLFREE(pCtx, prb_owned);
}

//////////////////////////////////////////////////////////////////

static void sg_ptnode__clone_for_move(
	SG_context* pCtx, 
	const struct sg_ptnode* ptnOrig,
	struct sg_ptnode** ppCopy
	)
{
	struct sg_ptnode* ptn;

	// Note: 2011/03/07 This routine doesn't seem to set
	//                  ptn->saved_flags and ptn->temp_flags.

	SG_ERR_CHECK(  SG_alloc1(pCtx, ptn)  );

	ptn->pPendingTree = ptnOrig->pPendingTree;
	SG_ERR_CHECK(  SG_TREENODE_ENTRY__ALLOC__COPY(pCtx, &ptn->pBaselineEntry, ptnOrig->pBaselineEntry)  );
	ptn->pszgid = ptnOrig->pszgid;
	ptn->pszCurrentName = ptnOrig->pszCurrentName;
	ptn->type = ptnOrig->type;
	ptn->pszCurrentHid = ptnOrig->pszCurrentHid;
	ptn->pszHidDisposable = ptnOrig->pszHidDisposable;
	ptn->bDisposableDir = ptnOrig->bDisposableDir;
	ptn->bStickyDir = ptnOrig->bStickyDir;
	ptn->bDeletedAfterAdded = ptnOrig->bDeletedAfterAdded;
    ptn->pszHidXattrs = ptnOrig->pszHidXattrs;
    ptn->iCurrentAttributeBits = ptnOrig->iCurrentAttributeBits;

	*ppCopy = ptn;
	
	return;

fail:
	if (ptn != NULL && ptn->pBaselineEntry != NULL)
		SG_TREENODE_ENTRY_NULLFREE(pCtx, ptn->pBaselineEntry);
	SG_PTNODE_NULLFREE(pCtx, ptn);
}

static void sg_ptnode__alloc__top(
	SG_context* pCtx, 
	SG_pendingtree* pPendingTree,
	const char* pszHid
	)
{
	struct sg_ptnode* ptn;
	
    SG_ASSERT(!pPendingTree->pSuperRoot);

	SG_ERR_CHECK(  SG_alloc1(pCtx, ptn)  );

	ptn->pPendingTree = pPendingTree;
	ptn->type = SG_TREENODEENTRY_TYPE_DIRECTORY;
	ptn->prbItems = NULL;
	ptn->pszgid = NULL;
	SG_ERR_CHECK(  SG_strpool__add__sz(pCtx, pPendingTree->pStrPool, pszHid, &ptn->pszCurrentHid)  );

	pPendingTree->pSuperRoot = ptn;

	return;

fail:
	SG_PTNODE_NULLFREE(pCtx, ptn);
}

static void sg_ptnode__alloc__load(
	SG_context* pCtx, 
	struct sg_ptnode** ppNew,
	SG_pendingtree* pPendingTree,
	const SG_treenode_entry* pEntry,
	const char* pszCurrentName,
	SG_treenode_entry_type type,
	const char* pszGid,
	const char* pszHid,
	const char* pszHidXattrs,
    SG_int64 iAttributeBits,
	const char* pszgidMovedFrom,
	const char * pszHidDisposable,
	SG_bool bDisposableDir,
	SG_bool bDeletedAfterAdded,
	SG_uint16 saved_flags,
	SG_bool bStickyDir
	)
{
	struct sg_ptnode* ptn;
	
	SG_ERR_CHECK(  SG_alloc1(pCtx, ptn)  );

	ptn->pPendingTree = pPendingTree;
	ptn->type = type;
	ptn->saved_flags = saved_flags;

	if (pEntry)
	{
		SG_ERR_CHECK(  SG_TREENODE_ENTRY__ALLOC__COPY(pCtx, &ptn->pBaselineEntry, pEntry)  );
	}
	if (pszCurrentName)
	{
		SG_ERR_CHECK(  SG_strpool__add__sz(pCtx, pPendingTree->pStrPool, pszCurrentName, &ptn->pszCurrentName)  );
	}
	if (pszGid)
	{
		SG_ERR_CHECK(  SG_strpool__add__sz(pCtx, pPendingTree->pStrPool, pszGid, &ptn->pszgid)  );
	}
	if (pszHid)
	{
		SG_ERR_CHECK(  SG_strpool__add__sz(pCtx, pPendingTree->pStrPool, pszHid, &ptn->pszCurrentHid)  );
	}
	if (pszHidXattrs)
	{
		SG_ERR_CHECK(  SG_strpool__add__sz(pCtx, pPendingTree->pStrPool, pszHidXattrs, &ptn->pszHidXattrs)  );
    }
    ptn->iCurrentAttributeBits = iAttributeBits;
	if (pszgidMovedFrom)
	{
		SG_ERR_CHECK(  SG_strpool__add__sz(pCtx, pPendingTree->pStrPool, pszgidMovedFrom, &ptn->pszgidMovedFrom)  );
	}

	if (pszHidDisposable && *pszHidDisposable)
		SG_ERR_CHECK(  sg_ptnode__set_disposable_hid(pCtx, ptn, pszHidDisposable)  );
	if (bDisposableDir)
		SG_ERR_CHECK(  sg_ptnode__set_disposable_dir(pCtx, ptn)  );
	if (bStickyDir)
		SG_ERR_CHECK(  sg_ptnode__set_sticky_dir(pCtx, ptn) );

	ptn->bDeletedAfterAdded = bDeletedAfterAdded;

	*ppNew = ptn;
	
	return;

fail:
	SG_PTNODE_NULLFREE(pCtx, ptn);
}

static void sg_ptnode__alloc__entry(
	SG_context* pCtx, 
	struct sg_ptnode** ppNew,
	SG_pendingtree* pPendingTree,
	const SG_treenode_entry* pEntry,
	const char* pszgid
	)
{
	struct sg_ptnode* ptn;

	SG_NULLARGCHECK_RETURN(ppNew);
	SG_NULLARGCHECK_RETURN(pPendingTree);
	SG_NULLARGCHECK_RETURN(pEntry);

	SG_ERR_CHECK(  SG_alloc1(pCtx, ptn)  );

	ptn->pPendingTree = pPendingTree;
	SG_ERR_CHECK(  SG_treenode_entry__get_entry_type(pCtx, pEntry, &ptn->type)  );
	SG_ERR_CHECK(  SG_TREENODE_ENTRY__ALLOC__COPY(pCtx, &ptn->pBaselineEntry, pEntry)  );
	SG_ERR_CHECK(  SG_strpool__add__sz(pCtx, ptn->pPendingTree->pStrPool, pszgid, &ptn->pszgid)  );
    SG_ERR_CHECK(  SG_treenode_entry__get_attribute_bits(pCtx, pEntry, &ptn->iCurrentAttributeBits)  );
	ptn->prbItems = NULL;

	*ppNew = ptn;

	return;
	
fail:
	SG_PTNODE_NULLFREE(pCtx, ptn);
}

static void sg_ptnode__alloc__new__with_gid(
	SG_context* pCtx, 
	struct sg_ptnode** ppNew,
	SG_pendingtree* pPendingTree,
	const char* pszName,
	SG_treenode_entry_type type,
	const char * pszGid
	)
{
	struct sg_ptnode* ptn;
	
	SG_NULLARGCHECK_RETURN(ppNew);
	SG_NULLARGCHECK_RETURN(pPendingTree);
	SG_NULLARGCHECK_RETURN(pszName);

	SG_ARGCHECK_RETURN(  ((type == SG_TREENODEENTRY_TYPE_DIRECTORY)
						  || (type == SG_TREENODEENTRY_TYPE_REGULAR_FILE)
						  || (type == SG_TREENODEENTRY_TYPE_SYMLINK)
						  || (type == SG_TREENODEENTRY_TYPE_SUBMODULE)),
						 type);
	
	SG_ERR_CHECK(  SG_alloc1(pCtx, ptn)  );

	ptn->pPendingTree = pPendingTree;
	ptn->type = type;
	ptn->prbItems = NULL;
	SG_ERR_CHECK(  SG_strpool__add__sz(pCtx, pPendingTree->pStrPool, pszName, &ptn->pszCurrentName)  );
	SG_ERR_CHECK(  SG_strpool__add__sz(pCtx, pPendingTree->pStrPool, pszGid, &ptn->pszgid)  );
	
	*ppNew = ptn;

	return;

fail:
	SG_PTNODE_NULLFREE(pCtx, ptn);
}

static void sg_ptnode__alloc__new(
	SG_context* pCtx, 
	struct sg_ptnode** ppNew,
	SG_pendingtree* pPendingTree,
	const char* pszName,
	SG_treenode_entry_type type
	)
{
	char bufGid[SG_GID_BUFFER_LENGTH];

	SG_ERR_CHECK_RETURN(  SG_gid__generate(pCtx, bufGid, sizeof(bufGid))  );
	SG_ERR_CHECK_RETURN(  sg_ptnode__alloc__new__with_gid(pCtx, ppNew, pPendingTree, pszName, type, bufGid)  );
}

//static SG_free_callback sg_ptnode__free;

static void sg_ptnode__free(SG_context* pCtx, struct sg_ptnode* ptn)
{
	if (!ptn)
	{
		return;
	}

#if TRACE_PENDINGTREE
	SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
							   "sg_ptnode__free: [gid %s][saved %d][temp %d]\n",
							   ptn->pszgid,ptn->saved_flags,ptn->temp_flags)  );
#endif

	if (ptn->prbItems)
	{
		SG_RBTREE_NULLFREE_WITH_ASSOC(pCtx, ptn->prbItems, (SG_free_callback *)sg_ptnode__free);
	}

	SG_TREENODE_SUBMODULE_NULLFREE(pCtx, ptn->pTSM);

	SG_TREENODE_ENTRY_NULLFREE(pCtx, ptn->pBaselineEntry);
	SG_TREENODE_ENTRY_NULLFREE(pCtx, ptn->pNewEntry);

	SG_NULLFREE(pCtx, ptn);
}

/*
 * If a dir node is not dirty, then we don't need its info loaded,
 * so unload it.
 */
static void sg_ptnode__unload_dir_entries(SG_context* pCtx, struct sg_ptnode* ptn)
{
	SG_NULLARGCHECK_RETURN(ptn);

	SG_ARGCHECK_RETURN(ptn->type == SG_TREENODEENTRY_TYPE_DIRECTORY, ptn->type);

#if TRACE_PENDINGTREE
	SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
							   "sg_ptnode__unload_dir_entries: [gid %s][saved %d][temp %d]\n",
							   ptn->pszgid,ptn->saved_flags,ptn->temp_flags)  );
#endif

	if (ptn->prbItems)
	{
		SG_RBTREE_NULLFREE_WITH_ASSOC(pCtx, ptn->prbItems, (SG_free_callback *)sg_ptnode__free);
	}

}

static void sg_ptnode__is_file_dirty(
	SG_context* pCtx, 
	const struct sg_ptnode* ptn,
	SG_bool* pbDirty
	)
{
	const char* pszhid = NULL;
	
	if (!ptn->pBaselineEntry)
	{
		*pbDirty = SG_TRUE;
		return;
	}

	if (!ptn->pszCurrentHid)
	{
		*pbDirty = SG_FALSE;
		return;
	}
	
	SG_ERR_CHECK(  SG_treenode_entry__get_hid_blob(pCtx, ptn->pBaselineEntry, &pszhid)  );

	if (0 == strcmp(pszhid, ptn->pszCurrentHid))
	{
		*pbDirty = SG_FALSE;
		return;
	}

	*pbDirty = SG_TRUE;
	return;

fail:
	return;
}

static void sg_ptnode__is_symlink_dirty(
	SG_context* pCtx, 
	const struct sg_ptnode* ptn,
	SG_bool* pbDirty
	)
{
	// sg_ptnode__is_file_dirty() happens to work for both files and symlinks
	// because all it does is check the blob HID for changes to the
	// content *within* the entry.	we do not care about changes *on*
	// the entry (such as the entryname) that are a property of the
	// containing directory.

	SG_ERR_CHECK_RETURN(  sg_ptnode__is_file_dirty(pCtx, ptn, pbDirty)  );
}

static void sg_ptnode__force_load_tsm(SG_context * pCtx, struct sg_ptnode * ptn,
									  SG_bool * pb_dir_not_found,
									  SG_bool * pb_dir_not_a_wd);

static void sg_ptnode__is_submodule_changed_or_dirty(
	SG_context * pCtx,
	struct sg_ptnode * ptn,
	SG_bool * pbDirty_Reference,		//< Has the cset that we reference changed?
	SG_bool * pbDirty_Pending			//< Is there pending dirt within the submodule?
	)
{
	// Is the submodule dirty?  There are 2 ways to
	// answer that question.  Did the blob HID of the
	// entry change (just like we do for a file or
	// symlink).  This is an indication of whether the
	// submodule has changed baselines.  The other way
	// is to look at both the blob HID **and** whether
	// there is dirt within the submodule's WD; that is,
	// is there dirty stuff in the sub's pendingtree?
	// 
	// I'm currently not sure if there is a single
	// right answer, so let the caller decide.

	// For the reference, we rely on the blob HID of
	// the SG_treenode_submodule (just like we do for
	// files and symlinks), so we don't need to actually
	// ask the TSM code to compare their contents.

	if (pbDirty_Reference)
		SG_ERR_CHECK_RETURN(  sg_ptnode__is_file_dirty(pCtx, ptn, pbDirty_Reference)  );

	if (pbDirty_Pending)
	{
		if (!ptn->pTSM)
		{
			SG_bool b_dir_not_found = SG_FALSE;
			SG_bool b_dir_not_a_wd = SG_FALSE;

			SG_ERR_CHECK_RETURN(  sg_ptnode__force_load_tsm(pCtx, ptn,
															&b_dir_not_found,
															&b_dir_not_a_wd)  );
			if (b_dir_not_found || b_dir_not_a_wd)
			{
				// in either case, something is wrong -- the submodule
				// is not in the WD like we think it should be.  so we
				// declare it to be dirty (even though we have no details).

				*pbDirty_Pending = SG_TRUE;
			}
			else
			{
				SG_ASSERT(  (ptn->pTSM)  );
			}
		}

		if (ptn->pTSM)
			SG_ERR_CHECK_RETURN(  SG_treenode_submodule__is_dirty(pCtx, ptn->pTSM, pbDirty_Pending)  );
	}
}

static void sg_ptnode__check_parent_dirty_non_recursive(
	SG_context* pCtx, 
	struct sg_ptnode* ptn,
	SG_bool* pbParentDirty
	);

static void sg_ptnode__unload_dir_entries_if_not_dirty(SG_context* pCtx, struct sg_ptnode* ptn, SG_bool* pbParentDirty)
{
	SG_bool b = SG_FALSE;
	const char* pszKey = NULL;
	struct sg_ptnode* ptnSub = NULL;
	SG_rbtree_iterator* pIterator = NULL;
	SG_bool bSelfDirty = SG_FALSE;

	switch (ptn->type)
	{
	case SG_TREENODEENTRY_TYPE_DIRECTORY:

		if (ptn->prbItems)
		{
			SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pIterator, ptn->prbItems, &b, &pszKey, (void**) &ptnSub)  );
			while (b)
			{
				SG_ERR_CHECK(  sg_ptnode__unload_dir_entries_if_not_dirty(pCtx, ptnSub, &bSelfDirty)  );
				
				SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pIterator, &b, &pszKey, (void**) &ptnSub)  );
			}
			SG_rbtree__iterator__free(pCtx, pIterator);
			pIterator = NULL;

			if (ptn->bDisposableDir)
			{
				// If MERGE gave us a "disposable" bit, then we need to
				// effectively PIN the ptnode in the pendingtree because
				// it isn't stored anywhere else.
				bSelfDirty = SG_TRUE;
			}
			else if (ptn->bDeletedAfterAdded)
			{
				bSelfDirty = SG_TRUE;
			}
			else if (ptn->bStickyDir)
			{
				bSelfDirty = SG_TRUE;
			}
			
			if (!bSelfDirty)
			{
				SG_ERR_CHECK(  sg_ptnode__unload_dir_entries(pCtx, ptn)  );
			}
		}
		break;

	case SG_TREENODEENTRY_TYPE_REGULAR_FILE:
		if (ptn->pszHidDisposable && *ptn->pszHidDisposable)
		{
			// If MERGE gave us "disposable-if" HID info, we need to
			// effectively PIN the ptnode in the pendingtree because
			// it isn't stored anywhere else.  that is, if after a MERGE
			// and they edit the file and make it match their original
			// baseline, we would have unloaded this ptnode (because it
			// isn't dirty anymore) and would have lost the disposable
			// info.
			bSelfDirty = SG_TRUE;
		}
		else if (ptn->bDeletedAfterAdded)
		{
			bSelfDirty = SG_TRUE;
		}
		else
		{
			SG_ERR_CHECK(  sg_ptnode__is_file_dirty(pCtx, ptn, &bSelfDirty)  );
		}
		break;

	case SG_TREENODEENTRY_TYPE_SYMLINK:
		if (ptn->pszHidDisposable && *ptn->pszHidDisposable)
		{
			bSelfDirty = SG_TRUE;	// PIN the ptnode in the pendingtree.
		}
		else if (ptn->bDeletedAfterAdded)
		{
			bSelfDirty = SG_TRUE;
		}
		else
		{
			SG_ERR_CHECK(  sg_ptnode__is_symlink_dirty(pCtx, ptn, &bSelfDirty)  );
		}
		break;

	case SG_TREENODEENTRY_TYPE_SUBMODULE:
		if (ptn->pszHidDisposable && *ptn->pszHidDisposable)
		{
			bSelfDirty = SG_TRUE;	// PIN the ptnode in the pendingtree.
		}
		else if (ptn->bDeletedAfterAdded)
		{
			bSelfDirty = SG_TRUE;
		}
		else if (ptn->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY)
		{
			// TODO 2010/11/17 I did this here for submodules because the bDirty_Pending
			// TODO            flag makes us try to hit the disk and in the case of a
			// TODO            phantom/moved-away node, there isn't anything on the disk
			// TODO            to examine.
			// TODO
			// TODO            But that raises a larger question of why we don't check
			// TODO            for moved-away nodes at the top of this function.  Afterall
			// TODO            a phantom is by-definition dirt.

			bSelfDirty = SG_TRUE;
		}
		else if (ptn->saved_flags & sg_PTNODE_SAVED_FLAG_DELETED)
		{
			// a deleted or lost item is, by definition, dirty

			bSelfDirty = SG_TRUE;
		}
		else
		{
			SG_bool bDirty_Reference;
			SG_bool bDirty_Pending;

			SG_ERR_CHECK(  sg_ptnode__is_submodule_changed_or_dirty(pCtx, ptn, &bDirty_Reference, &bDirty_Pending)  );
			bSelfDirty = (bDirty_Reference | bDirty_Pending);
		}
		break;

	default:
		SG_ERR_THROW_RETURN( SG_ERR_INVALID_PENDINGTREE_OBJECT_TYPE );
		break;
	}

	if (bSelfDirty)
	{
		*pbParentDirty = SG_TRUE;
	}
	else
	{
		SG_ERR_CHECK(  sg_ptnode__check_parent_dirty_non_recursive(pCtx, ptn, pbParentDirty)  );
	}
	
	return;

fail:
    SG_RBTREE_ITERATOR_NULLFREE(pCtx, pIterator);
}

static void sg_ptnode__undo_implicit(SG_context* pCtx, struct sg_ptnode* ptn)
{
	SG_bool b = SG_FALSE;
	const char* pszKey = NULL;
	struct sg_ptnode* ptnSub = NULL;
	SG_rbtree_iterator* pIterator = NULL;
	SG_rbtree* prbRemove = NULL;

	if (ptn->temp_flags & sg_PTNODE_TEMP_FLAG_IMPLICIT_DELETE)
	{
		ptn->saved_flags &= ~sg_PTNODE_SAVED_FLAG_DELETED;
		ptn->temp_flags &= ~sg_PTNODE_TEMP_FLAG_IMPLICIT_DELETE;
	}
	
	if (ptn->type == SG_TREENODEENTRY_TYPE_DIRECTORY)
	{
		if (ptn->prbItems)
		{
			SG_ERR_CHECK(  SG_RBTREE__ALLOC(pCtx, &prbRemove)  );
			
			SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pIterator, ptn->prbItems, &b, &pszKey, (void**) &ptnSub)  );
			while (b)
			{
				if (ptnSub->temp_flags & sg_PTNODE_TEMP_FLAG_IMPLICIT_ADD)
				{
					SG_ERR_CHECK(  SG_rbtree__add(pCtx, prbRemove, pszKey)  );
				}
				
				SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pIterator, &b, &pszKey, (void**) &ptnSub)  );
			}
			SG_rbtree__iterator__free(pCtx, pIterator);
			pIterator = NULL;

			SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pIterator, prbRemove, &b, &pszKey, NULL)  );
			while (b)
			{
				SG_ERR_CHECK(  SG_rbtree__remove__with_assoc(pCtx, ptn->prbItems, pszKey, (void**) &ptnSub)  );
				SG_PTNODE_NULLFREE(pCtx, ptnSub);
				
				SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pIterator, &b, &pszKey, NULL)  );
			}
			SG_rbtree__iterator__free(pCtx, pIterator);
			pIterator = NULL;

			SG_RBTREE_NULLFREE(pCtx, prbRemove);

			SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pIterator, ptn->prbItems, &b, &pszKey, (void**) &ptnSub)  );
			while (b)
			{
				SG_ERR_CHECK(  sg_ptnode__undo_implicit(pCtx, ptnSub)  );
				
				SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pIterator, &b, &pszKey, (void**) &ptnSub)  );
			}
			SG_rbtree__iterator__free(pCtx, pIterator);
			pIterator = NULL;
		}
	}

	return;

fail:
    SG_RBTREE_ITERATOR_NULLFREE(pCtx, pIterator);
}

static void sg_ptnode__get_name(
	SG_context* pCtx, 
	const struct sg_ptnode* ptn,
	const char** ppszName
	)
{
	if (ptn->pszCurrentName)
	{
		*ppszName = ptn->pszCurrentName;
	}
	else if (ptn->pBaselineEntry)
	{
		SG_ERR_CHECK_RETURN(  SG_treenode_entry__get_entry_name(pCtx, ptn->pBaselineEntry, ppszName)  );
	}
	else
	{
		*ppszName = NULL;
	}
}

/**
 * Like sg_ptnode__get_name(), but we get the original version of the name,
 * ignoring any renames.
 */
static void sg_ptnode__get_old_name(
	SG_context* pCtx, 
	const struct sg_ptnode* ptn,
	const char** ppszName
	)
{
	if (ptn->pBaselineEntry)
	{
		SG_ERR_CHECK_RETURN(  SG_treenode_entry__get_entry_name(pCtx, ptn->pBaselineEntry, ppszName)  );
	}
	else
	{
		*ppszName = NULL;
	}
}

static void sg_ptnode__get_hid(
	SG_context* pCtx, 
	const struct sg_ptnode* ptn,
	const char** ppszhid
	)
{
	
	if (ptn->pszCurrentHid)
	{
		*ppszhid = ptn->pszCurrentHid;
	}
	else if (ptn->pBaselineEntry)
	{
		SG_ERR_CHECK_RETURN(  SG_treenode_entry__get_hid_blob(pCtx, ptn->pBaselineEntry, ppszhid)  );
	}
	else
	{
		*ppszhid = NULL;
	}
}

static void sg_ptnode__get_old_hid(
	SG_context* pCtx, 
	const struct sg_ptnode* ptn,
	const char** ppszhid
	)
{
	
	if (ptn->pBaselineEntry)
	{
		SG_ERR_CHECK_RETURN(  SG_treenode_entry__get_hid_blob(pCtx, ptn->pBaselineEntry, ppszhid)  );
	}
	else
	{
		*ppszhid = NULL;
	}
}

static void sg_ptnode__get_xattrs_hid(
	SG_context* pCtx, 
	const struct sg_ptnode* ptn,
	const char** ppszhid
	)
{
	SG_NULLARGCHECK_RETURN(ptn);
	SG_NULLARGCHECK_RETURN(ppszhid);

    *ppszhid = ptn->pszHidXattrs;
}

static void sg_ptnode__get_old_xattrs_hid(
	SG_context* pCtx, 
	const struct sg_ptnode* ptn,
	const char** ppszhid
	)
{
	if (ptn->pBaselineEntry)
	{
		SG_ERR_CHECK_RETURN(  SG_treenode_entry__get_hid_xattrs(pCtx, ptn->pBaselineEntry, ppszhid)  );
	}
	else
	{
		*ppszhid = NULL;
	}
}

static void sg_ptnode__find_entry_by_name_not_recursive__active(SG_context* pCtx, const struct sg_ptnode* ptnDir, const char* pszName, struct sg_ptnode** ppResult)
{
	SG_bool b = SG_FALSE;
	const char* pszKey = NULL;
	struct sg_ptnode* ptnSub = NULL;
	SG_rbtree_iterator* pIterator = NULL;
	const char* pszEntryName = NULL;
	struct sg_ptnode* ptnResult = NULL;

	SG_NULLARGCHECK_RETURN(ptnDir);
	SG_NULLARGCHECK_RETURN(ptnDir->prbItems);

	SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pIterator, ptnDir->prbItems, &b, &pszKey, (void**) &ptnSub)  );
	while (b)
	{
		if (
			(ptnSub->saved_flags & sg_PTNODE_SAVED_FLAG_DELETED)
			|| (ptnSub->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY)
			)
		{
		}
		else
		{
			SG_ERR_CHECK(  sg_ptnode__get_name(pCtx, ptnSub, &pszEntryName)  );
			
			if (0 == strcmp(pszName, pszEntryName))
			{
				ptnResult = ptnSub;
				break;
			}
		}
		
		SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pIterator, &b, &pszKey, (void**) &ptnSub)  );
	}
	SG_rbtree__iterator__free(pCtx, pIterator);
	pIterator = NULL;

	*ppResult = ptnResult;

	return;

fail:
	SG_RBTREE_ITERATOR_NULLFREE(pCtx, pIterator);
}

static void sg_ptnode__find_entry_by_name_not_recursive__inactive(SG_context* pCtx, const struct sg_ptnode* ptnDir, const char* pszName, struct sg_ptnode** ppResult)
{
	SG_bool b = SG_FALSE;
	const char* pszKey = NULL;
	struct sg_ptnode* ptnSub = NULL;
	SG_rbtree_iterator* pIterator = NULL;
	const char* pszEntryName = NULL;
	struct sg_ptnode* ptnResult = NULL;
	
	SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pIterator, ptnDir->prbItems, &b, &pszKey, (void**) &ptnSub)  );
	while (b)
	{
		if (
			(ptnSub->saved_flags & sg_PTNODE_SAVED_FLAG_DELETED)
			|| (ptnSub->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY)
			)
		{
			SG_ERR_CHECK(  sg_ptnode__get_name(pCtx, ptnSub, &pszEntryName)  );
			
			if (0 == strcmp(pszName, pszEntryName))
			{
				ptnResult = ptnSub;
				break;
			}
		}
		else
		{
		}
		
		SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pIterator, &b, &pszKey, (void**) &ptnSub)  );
	}
	SG_rbtree__iterator__free(pCtx, pIterator);
	pIterator = NULL;

	*ppResult = ptnResult;

	return;

fail:
    SG_RBTREE_ITERATOR_NULLFREE(pCtx, pIterator);
}

void sg_ptnode__add_entry(SG_context* pCtx, struct sg_ptnode* ptnParent, struct sg_ptnode* ptnEntry)
{
#if TRACE_PENDINGTREE
	SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,"sg_ptnode__add_entry: [parent %s][child %s][child saved %d][child temp %d]\n",
							   ((ptnParent->pszgid) ? ptnParent->pszgid : "(null)"),
							   ((ptnEntry->pszgid) ? ptnEntry->pszgid : "(null)"),
							   ptnEntry->saved_flags,
							   ptnEntry->temp_flags)  );
#endif

	SG_ERR_CHECK(  SG_rbtree__add__with_assoc(pCtx, ptnParent->prbItems, ptnEntry->pszgid, ptnEntry)  );

	ptnEntry->pCurrentParent = ptnParent;

fail:
	return;
}

void sg_ptnode__load_dir_entries(SG_context* pCtx, struct sg_ptnode* ptn)
{
	SG_uint32 count = 0;
	SG_uint32 i;
	struct sg_ptnode* pNewNode = NULL;
	const char* pszidGid = NULL;
	SG_treenode* pTreenode = NULL;
	const char* pszhid = NULL;
	
	SG_NULLARGCHECK_RETURN(ptn);
	SG_ARGCHECK_RETURN(ptn->type == SG_TREENODEENTRY_TYPE_DIRECTORY, ptn->type);

	if (ptn->prbItems)
		return;

	SG_ERR_CHECK(  SG_RBTREE__ALLOC(pCtx, &ptn->prbItems)  );

	SG_ERR_CHECK(  sg_ptnode__get_hid(pCtx, ptn, &pszhid)  );
	
	if (pszhid)
	{
		SG_ERR_CHECK(  SG_treenode__load_from_repo(pCtx, ptn->pPendingTree->pRepo, pszhid, &pTreenode)  );
		
		SG_ERR_CHECK(  SG_treenode__count(pCtx, pTreenode, &count)  );
		for (i=0; i<count; i++)
		{
			const SG_treenode_entry* pEntry = NULL;
			
			SG_ERR_CHECK(  SG_treenode__get_nth_treenode_entry__ref(pCtx, pTreenode, i, &pszidGid, &pEntry)  );
			SG_ERR_CHECK(  sg_ptnode__alloc__entry(pCtx, &pNewNode, ptn->pPendingTree, pEntry, pszidGid)  );
			SG_ERR_CHECK(  sg_ptnode__add_entry(pCtx, ptn, pNewNode)  );
			pNewNode = NULL;
		}
		
		SG_TREENODE_NULLFREE(pCtx, pTreenode);
	}

	return;

fail:
	SG_TREENODE_NULLFREE(pCtx, pTreenode);
	SG_PTNODE_NULLFREE(pCtx, pNewNode);
}

//////////////////////////////////////////////////////////////////

void sg_ptnode__find_directory(
	SG_context* pCtx, 
	struct sg_ptnode* pNode,
	SG_bool bSynthesizeMissingParents,
	const char** ppszPathParts,
	SG_uint32 count_parts,
	SG_uint32 cur_part,
	struct sg_ptnode** ppResult
	)
{
	struct sg_ptnode* ptn = NULL;
	struct sg_ptnode* pNewNode = NULL;
	
	SG_NULLARGCHECK_RETURN(pNode);
	SG_NULLARGCHECK_RETURN(ppszPathParts);
	SG_NULLARGCHECK_RETURN(ppResult);
	SG_ARGCHECK_RETURN((count_parts > 0), count_parts);
	SG_ARGCHECK_RETURN((SG_TREENODEENTRY_TYPE_DIRECTORY == pNode->type), nNode->type);

	SG_ERR_CHECK(  sg_ptnode__load_dir_entries(pCtx, pNode)  );

	SG_ERR_CHECK(  sg_ptnode__find_entry_by_name_not_recursive__active(pCtx, pNode, ppszPathParts[cur_part], &ptn)  );
	
	if (!ptn)
	{
		if (!bSynthesizeMissingParents)
			SG_ERR_THROW(  SG_ERR_PENDINGTREE_PARENT_DIR_MISSING_OR_INACTIVE  );

#if TRACE_PENDINGTREE
		SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
								   "sg_ptnode__find_directory: synthesizing directory node for [%s]\n",
								   ppszPathParts[cur_part])  );
#endif
		SG_ERR_CHECK(  sg_ptnode__alloc__new(pCtx, &pNewNode, pNode->pPendingTree, ppszPathParts[cur_part], SG_TREENODEENTRY_TYPE_DIRECTORY)  );
		SG_ERR_CHECK(  sg_ptnode__add_entry(pCtx, pNode, pNewNode)  );
		ptn = pNewNode;
		pNewNode = NULL;
	}
	
	if ((cur_part +1) < count_parts)
	{
		SG_ERR_CHECK_RETURN(  sg_ptnode__find_directory(pCtx, ptn, bSynthesizeMissingParents, ppszPathParts, count_parts, cur_part+1, ppResult)  );
		return;
	}
	else
	{
		SG_ARGCHECK_RETURN((SG_TREENODEENTRY_TYPE_DIRECTORY == ptn->type), ptn->type);

		SG_ERR_CHECK(  sg_ptnode__load_dir_entries(pCtx, ptn)  );
		
		*ppResult = ptn;
		return;
	}
	
fail:
	SG_PTNODE_NULLFREE(pCtx, pNewNode);
}

void sg_ptnode__find_directory__allow_deleted(
	SG_context* pCtx, 
	struct sg_ptnode* pNode,
	const char** ppszPathParts,
	SG_uint32 count_parts,
	SG_uint32 cur_part,
	struct sg_ptnode** ppResult
	)
{
	struct sg_ptnode* ptn = NULL;
	struct sg_ptnode* pNewNode = NULL;
	
	SG_NULLARGCHECK_RETURN(pNode);
	SG_NULLARGCHECK_RETURN(ppszPathParts);
	SG_NULLARGCHECK_RETURN(ppResult);
	SG_ARGCHECK_RETURN((count_parts > 0), count_parts);
	SG_ARGCHECK_RETURN((SG_TREENODEENTRY_TYPE_DIRECTORY == pNode->type), nNode->type);

	SG_ERR_CHECK(  sg_ptnode__load_dir_entries(pCtx, pNode)  );

	SG_ERR_CHECK(  sg_ptnode__find_entry_by_name_not_recursive__active(pCtx, pNode, ppszPathParts[cur_part], &ptn)  );
	
	if (!ptn)
	{
		SG_ERR_CHECK(  sg_ptnode__find_entry_by_name_not_recursive__inactive(pCtx, pNode, ppszPathParts[cur_part], &ptn)  );
		if (ptn && (ptn->saved_flags & sg_PTNODE_SAVED_FLAG_DELETED))
		{
#if TRACE_PENDINGTREE
			SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
									   "sg_ptnode__find_directory__allow_deleted: found deleted directory node for [%s]\n",
									   ppszPathParts[cur_part])  );
#endif
		}
		else
		{
#if TRACE_PENDINGTREE
			SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
									   "sg_ptnode__find_directory__allow_deleted: not found or disallow node [%s]\n",
									   ppszPathParts[cur_part])  );
#endif
			SG_ERR_THROW(  SG_ERR_PENDINGTREE_PARENT_DIR_MISSING_OR_INACTIVE  );
		}
	}
	
	if ((cur_part +1) < count_parts)
	{
		// TODO 2011/05/11 we could speed up the recursion if we passed in
		// TODO            a flag to indicate that we found a deleted/inactive
		// TODO            and let the recursive call avoid the __active() lookup.
		SG_ERR_CHECK_RETURN(  sg_ptnode__find_directory__allow_deleted(pCtx, ptn, ppszPathParts, count_parts, cur_part+1, ppResult)  );
		return;
	}
	else
	{
		SG_ARGCHECK_RETURN((SG_TREENODEENTRY_TYPE_DIRECTORY == ptn->type), ptn->type);

		SG_ERR_CHECK(  sg_ptnode__load_dir_entries(pCtx, ptn)  );
		
		*ppResult = ptn;
		return;
	}
	
fail:
	SG_PTNODE_NULLFREE(pCtx, pNewNode);
}

//////////////////////////////////////////////////////////////////

static void _sg_pendingtree__load_cached_or_baseline(SG_context* pCtx, SG_pendingtree* pPendingTree);

/**
 * core __alloc routine for pendingtree.
 * this ***DOES NOT*** call __load().
 */
static void _sg_pendingtree__alloc(
	SG_context* pCtx, 
	const SG_pathname* pPathWorkingDirectoryTop,   /* TODO as implemented below, this doesn't REALLY have to be the top. */
	SG_pendingtree** ppNew
	)
{
	SG_pendingtree* ptree = NULL;
	SG_pathname* pPathActualTop = NULL;
	char* pszidGidAnchorDirectory = NULL;
	
	SG_ERR_CHECK(  SG_alloc1(pCtx, ptree)  );

	// don't load/create timestamp cache until we know what caller wants to do.
	ptree->pTSC = NULL;

	SG_ERR_CHECK(  SG_VARRAY__ALLOC(pCtx, &ptree->pvaWarnings)  );
    ptree->portMask = SG_PORT_FLAGS__ALL;  /* TODO get this from the repo config */
	SG_ERR_CHECK(  sg_pendingtree__configure_portability_converter(pCtx, ptree)  );

	SG_ERR_CHECK(  SG_STRPOOL__ALLOC(pCtx, &ptree->pStrPool, 32768)  ); /* TODO constant */
	
	SG_ERR_CHECK(  SG_workingdir__find_mapping(pCtx, pPathWorkingDirectoryTop, &pPathActualTop, &ptree->pstrRepoDescriptorName, &pszidGidAnchorDirectory)  );

	ptree->pPathWorkingDirectoryTop = pPathActualTop;
	pPathActualTop = NULL;

	SG_NULLFREE(pCtx, pszidGidAnchorDirectory);    /* TODO do something with pszidGidAnchorDirectory */

	SG_REPO__OPEN_REPO_INSTANCE(pCtx, SG_string__sz(ptree->pstrRepoDescriptorName), &ptree->pRepo);
	if(SG_context__err_equals(pCtx, SG_ERR_NOTAREPOSITORY))
	{
		SG_context__err_reset(pCtx);
		SG_ERR_CHECK(  SG_vv_utils__throw_NOTAREPOSITORY_for_wd_association(pCtx, SG_string__sz(ptree->pstrRepoDescriptorName), ptree->pPathWorkingDirectoryTop)  );
	}
	else
		SG_ERR_CHECK_CURRENT;

	*ppNew = ptree;
	return;

fail:
	SG_PENDINGTREE_NULLFREE(pCtx, ptree);
	SG_NULLFREE(pCtx, pszidGidAnchorDirectory);
	SG_PATHNAME_NULLFREE(pCtx, pPathActualTop);
}

void SG_pendingtree__alloc_from_cwd(
	SG_context* pCtx,
	SG_pendingtree** ppNew
	)
{
	SG_pathname * pPathCwd = NULL;
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC(pCtx, &pPathCwd)  );
	SG_ERR_CHECK(  SG_pathname__set__from_cwd(pCtx, pPathCwd)  );
	SG_ERR_CHECK(  SG_pendingtree__alloc(pCtx, pPathCwd, ppNew)  );
fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathCwd);
}

void SG_pendingtree__alloc(
	SG_context* pCtx, 
	const SG_pathname* pPathWorkingDirectoryTop,   /* TODO as implemented below, this doesn't REALLY have to be the top. */
	SG_pendingtree** ppNew
	)
{
	SG_pendingtree * pNew = NULL;

#if TRACE_PENDINGTREE
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "SG_pendingtree__alloc() on %s\n",
							   ((pPathWorkingDirectoryTop) ? SG_pathname__sz(pPathWorkingDirectoryTop) : "(null)"))  );
#endif

	SG_ERR_CHECK(  _sg_pendingtree__alloc(pCtx, pPathWorkingDirectoryTop,
										  &pNew)  );
	SG_ERR_CHECK(  _sg_pendingtree__load_cached_or_baseline(pCtx, pNew)  );

	*ppNew = pNew;
	return;

fail:
	SG_PENDINGTREE_NULLFREE(pCtx, pNew);
}

//////////////////////////////////////////////////////////////////

void sg_ptnode__to_vhash(
	SG_context* pCtx, 
	const struct sg_ptnode* ptn,
	SG_vhash* pParent,
	SG_vhash** ppResult
	)
{
	SG_vhash* pvh = NULL;
	SG_vhash* pvhSub = NULL;
	SG_vhash* pvhEntries = NULL;
	SG_treenode_entry* pEntry = NULL;
	SG_rbtree_iterator* pIterator = NULL;
	
	SG_NULLARGCHECK_RETURN(ptn);

	/* TODO use #defines for these key strings */

	/* TODO add a version number */
	
	SG_ERR_CHECK(  SG_VHASH__ALLOC__SHARED(pCtx, &pvh, 10, pParent)  );

	if (ptn->pPendingTree->bDebugExport)
	{
		SG_ERR_CHECK(  SG_vhash__add__int64(pCtx,pvh,"temp_flags",(SG_int64)ptn->temp_flags)  );
	}
	
	SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh, "flags", (SG_int64) ptn->saved_flags)  );

	if (ptn->pBaselineEntry)
	{
		SG_ERR_CHECK(  SG_TREENODE_ENTRY__ALLOC__COPY(pCtx, &pEntry, ptn->pBaselineEntry)  );

		/* TODO this cast of a treenode entry to a vhash is in very
		 * poor taste.  The fact that a treenode entry is really a
		 * vhash is something that this code should not know. */
		SG_ERR_CHECK(  SG_vhash__add__vhash(pCtx, pvh, "entry", (SG_vhash**)&pEntry)  );
	}
	
	SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh, "type", ptn->type)  );

    if (ptn->pszCurrentName)
    {
        SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvh, "name", ptn->pszCurrentName)  );
    }
    if (ptn->pszCurrentHid)
    {
        SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvh, "hid", ptn->pszCurrentHid)  );
    }
    if (ptn->pszHidXattrs)
    {
        SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvh, "xattrs", ptn->pszHidXattrs)  );
    }
	SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pvh, "attrbits", ptn->iCurrentAttributeBits)  );
    if (ptn->pszgid)
    {
        SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvh, "gid", ptn->pszgid)  );
    }

    if (ptn->pszgidMovedFrom)
    {
        SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvh, "moved_from", ptn->pszgidMovedFrom)  );
    }

	if (ptn->prbItems)
	{
		SG_uint32 count;
		SG_bool b = SG_FALSE;
		const char* pszKey = NULL;
		struct sg_ptnode* ptnSub = NULL;

		SG_ERR_CHECK(  SG_rbtree__count(pCtx, ptn->prbItems, &count)  );
		
		SG_ERR_CHECK(  SG_VHASH__ALLOC__SHARED(pCtx, &pvhEntries, count, pParent)  );
		
		SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pIterator, ptn->prbItems, &b, &pszKey, (void**) &ptnSub)  );
		while (b)
		{
			SG_ASSERT(  (ptnSub)  );
			SG_ERR_CHECK(  sg_ptnode__to_vhash(pCtx, ptnSub, pParent, &pvhSub)  );
			
			SG_ERR_CHECK(  SG_vhash__add__vhash(pCtx, pvhEntries, pszKey, &pvhSub)  );

			SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pIterator, &b, &pszKey, (void**) &ptnSub)  );
		}
		SG_rbtree__iterator__free(pCtx, pIterator);
		pIterator = NULL;

		SG_ERR_CHECK(  SG_vhash__add__vhash(pCtx, pvh, "entries", &pvhEntries)  );
	}

	if (ptn->pszHidDisposable && *ptn->pszHidDisposable)
		SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvh, "hid_disposable", ptn->pszHidDisposable)  );
	if (ptn->bDisposableDir)
		SG_ERR_CHECK(  SG_vhash__add__bool(pCtx, pvh, "dir_disposable", SG_TRUE)  );
	if (ptn->bStickyDir)
		SG_ERR_CHECK(  SG_vhash__add__bool(pCtx, pvh, "dir_sticky", SG_TRUE)  );

	if (ptn->bDeletedAfterAdded)
		SG_ERR_CHECK(  SG_vhash__add__bool(pCtx, pvh, "deleted_after_added", SG_TRUE)  );

	*ppResult = pvh;

	return;

fail:
    SG_RBTREE_ITERATOR_NULLFREE(pCtx, pIterator);
    SG_TREENODE_ENTRY_NULLFREE(pCtx, pEntry);
    SG_VHASH_NULLFREE(pCtx, pvhSub);
    SG_VHASH_NULLFREE(pCtx, pvhEntries);
}

static void sg_pendingtree__to_vhash(
	SG_context* pCtx, 
	const SG_pendingtree* pPendingTree,
	SG_vhash** ppResult
	)
{
	SG_vhash* pvh = NULL;
	
	SG_NULLARGCHECK_RETURN(pPendingTree);
	SG_NULLARGCHECK_RETURN(ppResult);

	SG_ERR_CHECK(  sg_ptnode__to_vhash(pCtx, pPendingTree->pSuperRoot, NULL, &pvh)  );

	*ppResult = pvh;

	return;

fail:
	SG_VHASH_NULLFREE(pCtx, pvh);
}

void sg_pendingtree__get_save_path(
	SG_context* pCtx, 
	SG_pendingtree* pPendingTree,
	SG_pathname** ppPath)
{
	SG_pathname* pDrawerPath = NULL;
	SG_pathname* pPendingTreePath = NULL;

	SG_ERR_CHECK(  SG_workingdir__get_drawer_path(pCtx, pPendingTree->pPathWorkingDirectoryTop, &pDrawerPath)  );
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx, &pPendingTreePath, pDrawerPath, "wd.json")  );

	SG_PATHNAME_NULLFREE(pCtx, pDrawerPath);
	
	*ppPath = pPendingTreePath;

	return;

fail:
	SG_PATHNAME_NULLFREE(pCtx, pDrawerPath);
}
	
void SG_pendingtree__get_working_directory_top(
	SG_context* pCtx,
	const SG_pendingtree* pPendingTree,
	SG_pathname** ppPath)
{
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, ppPath, SG_pathname__sz(pPendingTree->pPathWorkingDirectoryTop))   );
fail:
	return;
}

void SG_pendingtree__get_working_directory_top__ref(
	SG_context * pCtx,
	const SG_pendingtree * pPendingTree,
	const SG_pathname ** ppPath)
{
	SG_NULLARGCHECK_RETURN(  pPendingTree  );
	SG_NULLARGCHECK_RETURN(  ppPath  );

	*ppPath = pPendingTree->pPathWorkingDirectoryTop;
}

void SG_pendingtree__save(
	SG_context* pCtx, 
	SG_pendingtree* pPendingTree
	)
{
	SG_vhash * pvh_pending = NULL;

	// Before we write the pendingtree (and release the lock),
	// see if we have a dirty timestamp_cache and write it out.

	if (pPendingTree->pTSC)
		SG_ERR_CHECK(  SG_timestamp_cache__save(pCtx, pPendingTree->pTSC)  );

#if TRACE_PENDINGTREE
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "SG_pendingtree__save: [pending non-empty %d][parents dirty %d][issues dirty %d][choices dirty %d]\n",
							   (pPendingTree->pSuperRoot != NULL),
							   pPendingTree->b_pvh_wd__parents__dirty,
							   pPendingTree->b_pvh_wd__issues__dirty,
							   pPendingTree->b_pvh_wd__choices__dirty)  );
#endif

	// update the "pending" sub-vhash based upon our in-memory pending tree.

	if (pPendingTree->pSuperRoot)
	{
		SG_ERR_CHECK(  sg_pendingtree__to_vhash(pCtx, pPendingTree, &pvh_pending)  );
		SG_ERR_CHECK(  SG_vhash__update__vhash(pCtx, pPendingTree->pvh_wd, "pending", &pvh_pending)  );
	}
	else
	{
		SG_vhash__remove(pCtx, pPendingTree->pvh_wd, "pending");
		SG_ERR_CHECK_CURRENT_DISREGARD(  SG_ERR_VHASH_KEYNOTFOUND  );
	}

	// NOTE: I'm assuming that any changes to the "parents" were made directly to
	// NOTE: sub-vhash in pPendingTree->pvh_wd, so that we don't have to do anything here
	// NOTE: other than write them out with everything else.

	// NOTE: I'm assuming that any changes to the "issues" were made directly to
	// NOTE: sub-vhash in pPendingTree->pvh_wd, so that we don't have to do anything here
	// NOTE: other than write them out with everything else.

	// When we write out the VFILE, we release the lock on it and forget that we had it open.
	SG_ERR_CHECK(  SG_vfile__end(pCtx, &pPendingTree->pvf_wd, pPendingTree->pvh_wd)  );
	SG_VHASH_NULLFREE(pCtx, pPendingTree->pvh_wd);
#if defined(DEBUG)
	pPendingTree->b_pvh_wd__parents__dirty = SG_FALSE;
	pPendingTree->b_pvh_wd__issues__dirty = SG_FALSE;
	pPendingTree->b_pvh_wd__choices__dirty = SG_FALSE;
#endif

	return;
	
fail:
	SG_VHASH_NULLFREE(pCtx, pvh_pending);
}

void sg_ptnode__from_vhash(SG_context* pCtx, struct sg_ptnode** ppResult, SG_pendingtree* pPendingTree, const SG_vhash* pvh)
{
	struct sg_ptnode* ptn = NULL;
	const char* pszGid = NULL;
	const char* pszHid = NULL;
	const char* pszHidXattrs = NULL;
	SG_int64 type = 0;
    SG_int64 iAttributeBits = 0;
	SG_int64 flags = 0;
	SG_vhash* pvhOriginalEntry = NULL;
	SG_vhash* pvhEntries = NULL;
	SG_uint32 count;
	SG_uint32 i;
	const char* pszGidMovedFrom = NULL;
	const char* pszCurrentName = NULL;
	const char * pszHidDisposable = NULL;
	SG_bool bDisposableDir = SG_FALSE;
	SG_bool bStickyDir = SG_FALSE;
	SG_bool bDeletedAfterAdded = SG_FALSE;

	SG_ERR_CHECK(  SG_vhash__check__vhash(pCtx, pvh, "entry", &pvhOriginalEntry)  );
	SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pvh, "name", &pszCurrentName)  );
	SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvh, "type", &type)  );
	SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pvh, "hid", &pszHid)  );
	SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pvh, "xattrs", &pszHidXattrs)  );
	SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvh, "attrbits", &iAttributeBits)  );
	SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pvh, "gid", &pszGid)  );
	SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pvh, "moved_from", &pszGidMovedFrom)  );
	SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pvh, "flags", &flags)  );
	SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pvh, "hid_disposable", &pszHidDisposable)  );

	// since we only add the "dir_disposable" key with value SG_TRUE, we
	// don't actually need to look at the value.
	SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh, "dir_disposable", &bDisposableDir)  );

	SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh, "dir_sticky", &bStickyDir)  );

	SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh, "deleted_after_added", &bDeletedAfterAdded)  );

	/* TODO this cast of a treenode entry to a vhash is in very
	 * poor taste.  The fact that a treenode entry is really a
	 * vhash is something that this code should not know. */
	SG_ERR_CHECK(  sg_ptnode__alloc__load(pCtx,
										  &ptn,
										  pPendingTree,
										  (SG_treenode_entry*) pvhOriginalEntry,
										  pszCurrentName,
										  type,
										  pszGid,
										  pszHid,
										  pszHidXattrs,
										  iAttributeBits,
										  pszGidMovedFrom,
										  pszHidDisposable,
										  bDisposableDir,
										  bDeletedAfterAdded,
										  (SG_uint16) flags,
										  bStickyDir)  );
	
	SG_ERR_CHECK(  SG_vhash__check__vhash(pCtx, pvh, "entries", &pvhEntries)  );
	if (pvhEntries)
	{
		SG_ERR_CHECK(  SG_vhash__count(pCtx, pvhEntries, &count)  );

		SG_ERR_CHECK(  SG_RBTREE__ALLOC__PARAMS(pCtx, &ptn->prbItems, count, NULL)  );
		
		for (i=0; i<count; i++)
		{
			const char* pszName = NULL;
			const SG_variant* pV = NULL;
			SG_vhash* pvhSub = NULL;
			struct sg_ptnode* ptnSub = NULL;
			
			SG_ERR_CHECK(  SG_vhash__get_nth_pair(pCtx, pvhEntries, i, &pszName, &pV)  );
			SG_ERR_CHECK(  SG_variant__get__vhash(pCtx, pV, &pvhSub)  );
			SG_ERR_CHECK(  sg_ptnode__from_vhash(pCtx, &ptnSub, pPendingTree, pvhSub)  );
			SG_ERR_CHECK(  sg_ptnode__add_entry(pCtx, ptn, ptnSub)  );
		}
	}

	*ppResult = ptn;
	
	return;

fail:
	SG_PTNODE_NULLFREE(pCtx, ptn);
}

void SG_pendingtree__from_vhash(
	SG_context* pCtx, 
	SG_pendingtree* pPendingTree,
	const SG_vhash* pvh
	)
{
    SG_ASSERT(!pPendingTree->pSuperRoot);

	SG_ERR_CHECK_RETURN(  sg_ptnode__from_vhash(pCtx, &pPendingTree->pSuperRoot, pPendingTree, pvh)  );
}

//////////////////////////////////////////////////////////////////

static void _sg_pendingtree__intern_vfile_contents(SG_context * pCtx,
												   SG_pendingtree * pPendingTree)
{
	char* pszidUserSuperRoot = NULL;
	if (pPendingTree->pvh_wd)
	{
		// the VFile has a valid VHash in it.

		const SG_varray * pva_wd_parents = NULL;		// we don't own this
		SG_uint32 nrParents = 0;
		SG_bool bFound = SG_FALSE;

		SG_ERR_CHECK(  SG_pendingtree__get_wd_parents__ref(pCtx, pPendingTree, &pva_wd_parents)  );
		SG_ERR_CHECK(  SG_varray__count(pCtx, pva_wd_parents, &nrParents)  );
		SG_ASSERT(  (nrParents >= 1)  );

		// we only have a "pending" sub-vhash when there is dirt in the WD.
		// if the WD is dirty, we re-load the dirt into the in-memory pendingtree.
		// if the WD is clean, we load minimal info from the changeset's superroot.
		SG_ERR_CHECK(  SG_vhash__has(pCtx, pPendingTree->pvh_wd, "pending", &bFound)  );
		if (bFound)
		{
			SG_vhash * pvh_pending;		// we don't own this
			SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pPendingTree->pvh_wd, "pending", &pvh_pending)  );
#if 0
			SG_ERR_CHECK(  SG_vhash_debug__dump_to_console(pCtx, pvh_pending)  );
#endif
			SG_ERR_CHECK(  SG_pendingtree__from_vhash(pCtx, pPendingTree, pvh_pending)  );
		}
		else
		{
			const char * psz_hid_parent_0;	// we don't own this

			// I want to assert that (nrParents == 1), but I suppose that it is possible
			// for a merge to produce an empty pendingtree if both sides made the exact same changes.
			// (we'd still need the merge to resolve the parents and collapse the dag, but we might
			// not have any actual changes to the WD.
			// either way, we use parent[0] to populate the pendingtree.
			SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pva_wd_parents, 0, &psz_hid_parent_0)  );

			SG_ERR_CHECK(  SG_changeset__tree__find_root_hid(pCtx, pPendingTree->pRepo, psz_hid_parent_0, &pszidUserSuperRoot)  );
			SG_ERR_CHECK(  sg_ptnode__alloc__top(pCtx, pPendingTree, pszidUserSuperRoot)  );
		}
	}
	else
	{
		// When a WD is first created (probably by SG_workingdir__create_and_checkout(),
		// we don't have a "wd.json" yet.  VFile fakes it and creates and locks the
		// file, but just returns NULL for the VHash.  We continue the illusion and
		// create one here.
		//
		// NOTE: We don't know what the baseline/parents should be, so we don't set
		// NOTE: anything for it.

		SG_ASSERT(  (pPendingTree->pvf_wd)  );
		SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pPendingTree->pvh_wd)  );
	}

fail:
	SG_NULLFREE(pCtx, pszidUserSuperRoot);
}

static void _sg_pendingtree__load_cached_or_baseline(
	SG_context* pCtx, 
	SG_pendingtree* pPendingTree
	)
{
	SG_pathname* pPath = NULL;
	
	SG_ERR_CHECK(  sg_pendingtree__get_save_path(pCtx, pPendingTree, &pPath)  );
	
#if TRACE_PENDINGTREE
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "_sg_pendingtree__load_cached_or_baseline() on %s\n",
							   SG_pathname__sz(pPath))  );
#endif

	SG_ASSERT(  (pPendingTree->pvf_wd == NULL)  );	// we should only be called once
	SG_ASSERT(  (pPendingTree->pvh_wd == NULL)  );

	// load the contents of the VFILE into a new VHASH.
	// then convert the VHASH (if present) into PTNODES and etc.

	SG_ERR_CHECK(  SG_vfile__begin(pCtx, pPath, SG_FILE_RDWR | SG_FILE_OPEN_OR_CREATE, &pPendingTree->pvh_wd, &pPendingTree->pvf_wd)  );
	SG_ERR_CHECK(  _sg_pendingtree__intern_vfile_contents(pCtx, pPendingTree)  );
	
fail:
	SG_PATHNAME_NULLFREE(pCtx, pPath);
}

//////////////////////////////////////////////////////////////////

void SG_pendingtree__reload_pendingtree_from_vfile(SG_context * pCtx,
												   SG_pendingtree * pPendingTree)
{
	SG_ASSERT(  (pPendingTree->pvf_wd)  );
	
#if TRACE_PENDINGTREE
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "_sg_pendingtree__reload_pendingtree_from_vfile()\n")  );
#endif

#if defined(DEBUG)
	pPendingTree->b_pvh_wd__parents__dirty = SG_FALSE;
	pPendingTree->b_pvh_wd__issues__dirty = SG_FALSE;
	pPendingTree->b_pvh_wd__choices__dirty = SG_FALSE;
#endif
	SG_VHASH_NULLFREE(pCtx, pPendingTree->pvh_wd);
	SG_PTNODE_NULLFREE(pCtx, pPendingTree->pSuperRoot);

	// reload the contents of the VFILE into a new VHASH.
	// We ASSUME that we are still holding the file lock.
	// then convert the VHASH (if present) into PTNODES and etc.

	SG_ERR_CHECK(  SG_vfile__reread(pCtx, pPendingTree->pvf_wd, &pPendingTree->pvh_wd)  );
	SG_ERR_CHECK(  _sg_pendingtree__intern_vfile_contents(pCtx, pPendingTree)  );

fail:
	return;
}

//////////////////////////////////////////////////////////////////

static void sg_pendingtree__fsobjtype_to_treenodeentrytype(
    SG_context* pCtx,
    SG_fsobj_type t_fsobj,
    SG_treenode_entry_type* p_t_tne
    )
{
    SG_treenode_entry_type t = SG_TREENODEENTRY_TYPE__INVALID;

	if (t_fsobj == SG_FSOBJ_TYPE__DIRECTORY)
	{
		// This layer just converts a FSOBJ_TYPE__DIRECTORY to
		// a TREENODE_TYPE_DIRECTORY.  It does not sniff for
		// submodules.  The layer above takes care of that.

        t = SG_TREENODEENTRY_TYPE_DIRECTORY;
	}
	else if (t_fsobj == SG_FSOBJ_TYPE__REGULAR)
	{
        t = SG_TREENODEENTRY_TYPE_REGULAR_FILE;
	}
	else if (t_fsobj == SG_FSOBJ_TYPE__SYMLINK)
	{
        t = SG_TREENODEENTRY_TYPE_SYMLINK;
	}
	else
	{
		SG_ERR_THROW_RETURN(  SG_ERR_INVALID_PENDINGTREE_OBJECT_TYPE  );
	}

    *p_t_tne = t;
}

static void sg_ptnode__get_repo_path(
	SG_context* pCtx, 
	struct sg_ptnode* ptn,
	SG_string** ppstr
	);

static void sg_ptnode__add_symlink_portability_warning(
    SG_context* pCtx, 
    struct sg_ptnode* ptn
    )
{
    SG_vhash* pvhItem = NULL;
    SG_string* pstrRepoPath = NULL;
    SG_pathname* pPath = NULL;

    SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pvhItem)  );

    SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhItem, "type", "symlink")  );

	SG_ERR_CHECK(  sg_ptnode__get_repo_path(pCtx, ptn, &pstrRepoPath)  );
	SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path(pCtx,
																		 ptn->pPendingTree->pPathWorkingDirectoryTop,
																		 SG_string__sz(pstrRepoPath),
																		 &pPath)  );
    SG_STRING_NULLFREE(pCtx, pstrRepoPath);
    SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pvhItem, "path", SG_pathname__sz(pPath)) );
    SG_PATHNAME_NULLFREE(pCtx, pPath);

	if (!ptn->pPendingTree->pvaWarnings)
		SG_ERR_CHECK(  SG_VARRAY__ALLOC(pCtx, &ptn->pPendingTree->pvaWarnings)  );

    SG_ERR_CHECK(  SG_varray__append__vhash(pCtx, ptn->pPendingTree->pvaWarnings, &pvhItem)  );

    return;

fail:
    SG_PATHNAME_NULLFREE(pCtx, pPath);
    SG_STRING_NULLFREE(pCtx, pstrRepoPath);
    SG_VHASH_NULLFREE(pCtx, pvhItem);
}

static void sg_ptnode__create_item__with_gid(
	SG_context* pCtx, 
	struct sg_ptnode* ptn, 
	const char* pszName,
	SG_bool bAdd,
	SG_bool bMarkImplicitAdd, 
	SG_treenode_entry_type tneType,
	const SG_treenode_submodule * pTSM,
	const char * pszGid,
	struct sg_ptnode** ppNewNode
	)
{
	struct sg_ptnode* pNewNode_Allocated = NULL;
	struct sg_ptnode* pNewNode;						// we do not own this
	struct sg_ptnode* ptnFoundExisting = NULL;		// we do not own this
	SG_bool bFoundExisting = SG_FALSE;

	// Normally, we don't expect to find an existing ptnode with the requested GID
	// already in the directory, but there are some times (after a MERGE or RESOLVE
	// for example) when a ADD is really an UNDELETE.

	SG_ERR_CHECK(  SG_rbtree__find(pCtx, ptn->prbItems, pszGid, &bFoundExisting, (void**) &ptnFoundExisting)  );
	if (bFoundExisting)
	{
		SG_ASSERT(  (ptnFoundExisting)  );
		if (ptnFoundExisting->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY)
		{
			// This is a coding error.  This directory has the PHANTOM moved-away
			// node (the item used to be in this directory, but was moved to somewhere
			// else in the tree).  The caller should call the MOVE code and let it
			// "unmove" the item back to here.
			SG_ERR_THROW2(  SG_ERR_NOTIMPLEMENTED,
							(pCtx, "Caller should use MOVE rather than ADD to un-move '%s'.", pszName)  );
		}
		else if (ptnFoundExisting->saved_flags & sg_PTNODE_SAVED_FLAG_DELETED)
		{
			// treat this ADD as an UNDELETE.

			ptnFoundExisting->saved_flags &= ~sg_PTNODE_SAVED_FLAG_DELETED;
			ptnFoundExisting->bDeletedAfterAdded = SG_FALSE;

			if (bMarkImplicitAdd)
				ptnFoundExisting->temp_flags |= sg_PTNODE_TEMP_FLAG_IMPLICIT_ADD;
			if (bAdd)
				ptnFoundExisting->temp_flags |= sg_PTNODE_TEMP_FLAG_CHECKMYNAME;
	
			pNewNode = ptnFoundExisting;
		}
		else
		{
			SG_ERR_THROW2(  SG_ERR_PENDINGTREE_OBJECT_ALREADY_EXISTS,
							(pCtx, "%s", pszName)  );
		}
	}
	else
	{
		SG_ERR_CHECK(  sg_ptnode__alloc__new__with_gid(pCtx, &pNewNode_Allocated, ptn->pPendingTree, pszName, tneType, pszGid)  );
		if (bMarkImplicitAdd)
			pNewNode_Allocated->temp_flags |= sg_PTNODE_TEMP_FLAG_IMPLICIT_ADD;
		if (bAdd)
			pNewNode_Allocated->temp_flags |= sg_PTNODE_TEMP_FLAG_CHECKMYNAME;
	
		SG_ERR_CHECK(  sg_ptnode__add_entry(pCtx, ptn, pNewNode_Allocated)  );
		pNewNode = pNewNode_Allocated;
		pNewNode_Allocated = NULL;			// we don't own it any more.
	}

	// We just added a new item to the 'ptn' directory.
	// Schedule a portability_check for the contents
	// of this directory after our callers are done
	// adding stuff.

    SG_ERR_CHECK(  sg_ptnode__queue_portability_check(pCtx, ptn)  );

    if (SG_TREENODEENTRY_TYPE_SYMLINK == tneType)
    {
        /* TODO we need some sort of a mask mechanism to turn this warning off */
        SG_ERR_CHECK(  sg_ptnode__add_symlink_portability_warning(pCtx, pNewNode)  );
    }

	if (tneType == SG_TREENODEENTRY_TYPE_SUBMODULE)
	{
		if (pTSM)
			SG_ERR_CHECK(  SG_TREENODE_SUBMODULE__ALLOC__COPY(pCtx, &pNewNode->pTSM, pTSM)  );
#if 1 && defined(DEBUG)
		else
			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "ptnode__create_item: submodule with no tsm\n")  );
#endif
	}
	
	*ppNewNode = pNewNode;

	return;

fail:
    SG_PTNODE_NULLFREE(pCtx, pNewNode_Allocated);
}

static void sg_ptnode__create_item(
	SG_context* pCtx, 
	struct sg_ptnode* ptn, 
	const char* pszName,
	SG_bool bAdd,
	SG_bool bMarkImplicitAdd, 
	SG_treenode_entry_type tneType,
	const SG_treenode_submodule * pTSM,
	struct sg_ptnode** ppNewNode
	)
{
	char bufGid[SG_GID_BUFFER_LENGTH];

	SG_ERR_CHECK_RETURN(  SG_gid__generate(pCtx, bufGid, sizeof(bufGid))  );
	SG_ERR_CHECK_RETURN(  sg_ptnode__create_item__with_gid(pCtx, ptn, pszName,
														   bAdd, bMarkImplicitAdd,
														   tneType, pTSM,
														   bufGid,
														   ppNewNode)  );
}

static void sg_ptnode__force_load_tsm(SG_context * pCtx, struct sg_ptnode * ptn,
									  SG_bool * pb_dir_not_found,
									  SG_bool * pb_dir_not_a_wd)
{
	// TODO 2010/11/18 Should this verify that ptn is not a phantom?

	SG_string * pStrRepoPath = NULL;
	SG_pathname * pPath = NULL;
	SG_bool bIsWD;
	SG_bool bExists;
	SG_fsobj_type fsobjType;

	SG_ASSERT(  (ptn->pTSM == NULL)  );

	SG_ERR_CHECK(  sg_ptnode__get_repo_path(pCtx, ptn, &pStrRepoPath)  );
	SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path(pCtx,
																		 ptn->pPendingTree->pPathWorkingDirectoryTop,
																		 SG_string__sz(pStrRepoPath),
																		 &pPath)  );
#if TRACE_PENDINGTREE
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "Force loading tsm for '%s'\n", SG_pathname__sz(pPath))  );
#endif

	SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pPath, &bExists, &fsobjType, NULL)  );
	if (!bExists)
	{
		*pb_dir_not_found = SG_TRUE;
	}
	else if (fsobjType != SG_FSOBJ_TYPE__DIRECTORY)
	{
		*pb_dir_not_found = SG_FALSE;
		*pb_dir_not_a_wd = SG_TRUE;
	}
	else
	{
		*pb_dir_not_found = SG_FALSE;
		SG_ERR_CHECK(  SG_TREENODE_SUBMODULE__SNIFF_DIRECTORY(pCtx, pPath, ptn->pPendingTree->pRepo, &bIsWD, &ptn->pTSM)  );
		*pb_dir_not_a_wd = (!bIsWD);
	}

#if TRACE_PENDINGTREE
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "After force loading TSM for '%s': [b_dir_not_found %d] [b_dir_not_a_wd %d]\n",
							   SG_pathname__sz(pPath),
							   *pb_dir_not_found,
							   *pb_dir_not_a_wd)  );
#endif

fail:
    SG_STRING_NULLFREE(pCtx, pStrRepoPath);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
}

static void sg_pendingtree__local_path_to_ptnode(
	SG_context* pCtx, 
	SG_pendingtree* pPendingTree,
	const SG_pathname* pPathLocalDirectory,
	SG_bool bSynthesizeMissingParents,
	struct sg_ptnode** ppNode
	)
{
	SG_string* pStrRepoPath = NULL;
	const char** apszParts = NULL;
	SG_uint32 count_parts = 0;
	SG_strpool* pPool = NULL;
	struct sg_ptnode* ptn = NULL;

	SG_ERR_CHECK(  SG_workingdir__wdpath_to_repopath(pCtx, pPendingTree->pPathWorkingDirectoryTop, pPathLocalDirectory, SG_TRUE, &pStrRepoPath)  );
	SG_ERR_CHECK(  SG_repopath__ensure_final_slash(pCtx, pStrRepoPath)  );
	SG_ERR_CHECK(  SG_STRPOOL__ALLOC(pCtx, &pPool, 10)  );
	
	SG_ERR_CHECK(  SG_repopath__split(pCtx, SG_string__sz(pStrRepoPath), pPool, &apszParts, &count_parts)  );

	SG_ERR_CHECK(  sg_ptnode__find_directory(pCtx, pPendingTree->pSuperRoot, bSynthesizeMissingParents, apszParts, count_parts, 0, &ptn)  );
	SG_NULLFREE(pCtx, apszParts);
	SG_STRING_NULLFREE(pCtx, pStrRepoPath);

	SG_STRPOOL_NULLFREE(pCtx, pPool);

	*ppNode = ptn;

	return;

fail:
	SG_STRING_NULLFREE(pCtx, pStrRepoPath);
	SG_NULLFREE(pCtx, apszParts);
	SG_STRPOOL_NULLFREE(pCtx, pPool);
}

static void sg_pendingtree__local_path_to_ptnode__allow_deleted(
	SG_context* pCtx, 
	SG_pendingtree* pPendingTree,
	const SG_pathname* pPathLocalDirectory,
	struct sg_ptnode** ppNode
	)
{
	SG_string* pStrRepoPath = NULL;
	const char** apszParts = NULL;
	SG_uint32 count_parts = 0;
	SG_strpool* pPool = NULL;
	struct sg_ptnode* ptn = NULL;

	SG_ERR_CHECK(  SG_workingdir__wdpath_to_repopath(pCtx, pPendingTree->pPathWorkingDirectoryTop, pPathLocalDirectory, SG_TRUE, &pStrRepoPath)  );
	SG_ERR_CHECK(  SG_repopath__ensure_final_slash(pCtx, pStrRepoPath)  );
	SG_ERR_CHECK(  SG_STRPOOL__ALLOC(pCtx, &pPool, 10)  );
	
	SG_ERR_CHECK(  SG_repopath__split(pCtx, SG_string__sz(pStrRepoPath), pPool, &apszParts, &count_parts)  );

	SG_ERR_CHECK(  sg_ptnode__find_directory__allow_deleted(pCtx, pPendingTree->pSuperRoot, apszParts, count_parts, 0, &ptn)  );
	SG_NULLFREE(pCtx, apszParts);
	SG_STRING_NULLFREE(pCtx, pStrRepoPath);

	SG_STRPOOL_NULLFREE(pCtx, pPool);

	*ppNode = ptn;

	return;

fail:
	SG_STRING_NULLFREE(pCtx, pStrRepoPath);
	SG_NULLFREE(pCtx, apszParts);
	SG_STRPOOL_NULLFREE(pCtx, pPool);
}

//////////////////////////////////////////////////////////////////

	static void sg_ptnode__scan_attrbits(SG_context* pCtx, struct sg_ptnode* ptn, SG_treenode_entry_type type, SG_fsobj_perms perms)
{
    SG_int64 iBaselineAttributeBits = 0;
    SG_int64 iCurrentAttributeBits = 0;

    if (ptn->pBaselineEntry)
    {
        SG_ERR_CHECK(  SG_treenode_entry__get_attribute_bits(pCtx, ptn->pBaselineEntry, &iBaselineAttributeBits)  );
    }

	SG_ERR_CHECK(  SG_attributes__bits__from_perms(pCtx, type == SG_TREENODEENTRY_TYPE_REGULAR_FILE ? SG_FSOBJ_TYPE__REGULAR : SG_FSOBJ_TYPE__UNSPECIFIED, perms, iBaselineAttributeBits, &iCurrentAttributeBits)  );

    ptn->iCurrentAttributeBits = iCurrentAttributeBits;

fail:
    return;
}

#ifdef SG_BUILD_FLAG_FEATURE_XATTR

static void sg_ptnode__get_xattrs(SG_context* pCtx, struct sg_ptnode* ptn, const SG_pathname* pPathLocal, char** ppszCurrentAttributesHid, SG_vhash** ppvh, SG_rbtree** pprb, SG_string** ppstr)
{
    SG_vhash* pvhAttrs = NULL;
    SG_vhash* pvhBaselineAttrs = NULL;
    const char* pszBaselineAttributesHid = NULL;
    SG_rbtree* prb = NULL;
    SG_string* pstr = NULL;
    char* pszCurrentAttributesHid = NULL;
    SG_uint32 count = 0;

    /* Read the xattrs from the local file */
    SG_ERR_CHECK(  SG_attributes__xattrs__read(pCtx, ptn->pPendingTree->pRepo, pPathLocal, &pvhAttrs, &prb)  );

    /* Compute the HID */
    SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pstr)  );
    SG_ERR_CHECK(  SG_vhash__to_json(pCtx, pvhAttrs,pstr)  );
    SG_ERR_CHECK(  SG_repo__alloc_compute_hash__from_string(pCtx, ptn->pPendingTree->pRepo, pstr, &pszCurrentAttributesHid)  );

    if (ptn->pBaselineEntry)
    {
		SG_ERR_CHECK(  SG_treenode_entry__get_hid_xattrs(pCtx, ptn->pBaselineEntry, &pszBaselineAttributesHid)  );
        if (pszBaselineAttributesHid)
        {
            if (0 != strcmp(pszCurrentAttributesHid, pszBaselineAttributesHid))
            {
                SG_ERR_CHECK(  SG_repo__fetch_vhash(pCtx, ptn->pPendingTree->pRepo,pszBaselineAttributesHid,&pvhBaselineAttrs)  );
                SG_ERR_CHECK(  SG_attributes__xattrs__update_with_baseline(pCtx, pvhBaselineAttrs, pvhAttrs)  );

                SG_ERR_CHECK(  SG_string__clear(pCtx, pstr)  );
                SG_ERR_CHECK(  SG_vhash__to_json(pCtx, pvhAttrs,pstr)  );
                SG_NULLFREE(pCtx, pszCurrentAttributesHid);
                SG_ERR_CHECK(  SG_repo__alloc_compute_hash__from_string(pCtx, ptn->pPendingTree->pRepo, pstr, &pszCurrentAttributesHid)  );
            }
        }
    }

    SG_ERR_CHECK(  SG_vhash__count(pCtx, pvhAttrs, &count)  );
    if (count == 0)
    {
        SG_NULLFREE(pCtx, pszCurrentAttributesHid);
        SG_VHASH_NULLFREE(pCtx, pvhAttrs);
        SG_RBTREE_NULLFREE(pCtx, prb);
        SG_STRING_NULLFREE(pCtx, pstr);
    }

    SG_VHASH_NULLFREE(pCtx, pvhBaselineAttrs);

    *ppstr = pstr;
    *pprb = prb;
    *ppvh = pvhAttrs;
    *ppszCurrentAttributesHid = pszCurrentAttributesHid;

    return;

fail:
    SG_VHASH_NULLFREE(pCtx, pvhBaselineAttrs);
    SG_VHASH_NULLFREE(pCtx, pvhAttrs);
}

static void sg_ptnode__scan_xattrs(SG_context* pCtx, struct sg_ptnode* ptn, SG_pathname* pPathLocal)
{
    SG_vhash* pvhAttrs = NULL;
    SG_rbtree* prb = NULL;
    char* pszCurrentAttributesHid = NULL;
    SG_string* pstr = NULL;

    SG_ERR_CHECK(  sg_ptnode__get_xattrs(pCtx, ptn, pPathLocal, &pszCurrentAttributesHid, &pvhAttrs, &prb, &pstr)  );
    if (pszCurrentAttributesHid)
    {
        SG_ERR_CHECK(  SG_strpool__add__sz(pCtx, ptn->pPendingTree->pStrPool, pszCurrentAttributesHid, &ptn->pszHidXattrs)  );
    }
    else
    {
        ptn->pszHidXattrs = NULL;
    }

    SG_NULLFREE(pCtx, pszCurrentAttributesHid);
    SG_STRING_NULLFREE(pCtx, pstr);
    SG_RBTREE_NULLFREE(pCtx, prb);
    SG_VHASH_NULLFREE(pCtx, pvhAttrs);

    return;

fail:
    SG_NULLFREE(pCtx, pszCurrentAttributesHid);
    SG_STRING_NULLFREE(pCtx, pstr);
    SG_RBTREE_NULLFREE(pCtx, prb);
    SG_VHASH_NULLFREE(pCtx, pvhAttrs);
}

static void sg_ptnode__commit_xattrs(
	SG_context* pCtx, 
	struct sg_ptnode* ptn,
	const SG_pathname* pPathLocal,
	SG_committing* ptx
	)
{
    SG_vhash* pvhAttrs = NULL;
    SG_rbtree* prb = NULL;
    char* pszCurrentAttributesHid = NULL;
	SG_bool b = SG_FALSE;
	const char* pszHid = NULL;
	SG_rbtree_iterator* pIterator = NULL;
    SG_byte* pPacked = NULL;
    char* pszActualHid = NULL;
    SG_string* pstr = NULL;

    SG_ERR_CHECK(  sg_ptnode__get_xattrs(pCtx, ptn, pPathLocal, &pszCurrentAttributesHid, &pvhAttrs, &prb, &pstr)  );
    SG_VHASH_NULLFREE(pCtx, pvhAttrs);
    if (!pszCurrentAttributesHid)
    {
        ptn->pszHidXattrs = NULL;
    }
    else
    {
        SG_ERR_CHECK(  SG_committing__add_bytes__string(
                    pCtx, 
                    ptx, 
                    NULL, // TODO objectid
                    -1, // TODO objectid_timestamp
                    pstr, 
                    &pszActualHid
                    )  );
        SG_STRING_NULLFREE(pCtx, pstr);

        /* TODO should we check to make sure actual hid is what it should be ? */
        SG_ERR_CHECK(  SG_strpool__add__sz(pCtx, ptn->pPendingTree->pStrPool, pszActualHid, &ptn->pszHidXattrs)  );
        SG_NULLFREE(pCtx, pszActualHid);

        SG_NULLFREE(pCtx, pszCurrentAttributesHid);

        /* And some of the attributes may require side blobs, so those need to get
         * saved as well.  */
        SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pIterator, prb, &b, &pszHid, (void**) &pPacked)  );
        while (b)
        {
            SG_uint32 len;
            const SG_byte* pData = NULL;

            SG_ERR_CHECK(  SG_memoryblob__unpack(pCtx, pPacked, &pData, &len)  );
            SG_ERR_CHECK(  SG_committing__add_bytes__buflen(pCtx, ptx, 
                    NULL, // TODO objectid
                    -1, // TODO objectid_timestamp
                        pData, len, &pszActualHid)  );
            /* TODO should we check to make sure actual hid is what it should be ? */
            SG_NULLFREE(pCtx, pszActualHid);

            SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pIterator, &b, &pszHid, (void**) &pPacked)  );
        }
        SG_RBTREE_ITERATOR_NULLFREE(pCtx, pIterator);
    }

    SG_RBTREE_NULLFREE(pCtx, prb);

    return;

fail:
    SG_RBTREE_NULLFREE(pCtx, prb);
    SG_NULLFREE(pCtx, pszCurrentAttributesHid);
    SG_STRING_NULLFREE(pCtx, pstr);
    SG_NULLFREE(pCtx, pszActualHid);
    SG_VHASH_NULLFREE(pCtx, pvhAttrs);
    SG_RBTREE_ITERATOR_NULLFREE(pCtx, pIterator);
}
#endif

//////////////////////////////////////////////////////////////////

#include "sg_pendingtree__private_scan_dir.h"

//////////////////////////////////////////////////////////////////

void SG_pendingtree__scan__no_unload__no_save(SG_context* pCtx,
											  SG_pendingtree* pPendingTree,
											  SG_bool bRecursive,
											  SG_bool bIncludeLostItems,
											  SG_bool bIncludeFoundItems,
											  const SG_file_spec* pFilespec)
{
	SG_file_spec* pInternalFilespec = NULL;

	SG_ERR_CHECK(  SG_FILE_SPEC__ALLOC__COPY(pCtx, pFilespec, &pInternalFilespec)  );
	SG_ERR_CHECK(  SG_file_spec__load_ignores(pCtx,
											  pInternalFilespec,
											  pPendingTree->pRepo,
											  pPendingTree->pPathWorkingDirectoryTop,
											  0u)  );

	SG_ERR_CHECK(  sg_pendingtree__scan_dir(pCtx,
											pPendingTree,
											pPendingTree->pPathWorkingDirectoryTop,
											SG_FALSE,		// bAdd
											SG_FALSE,		// bRemove
											SG_FALSE,		// bMarkImplicit
											bRecursive,
											NULL,			// ptnode
											bIncludeLostItems,		// bIncludeLostItems
											bIncludeFoundItems,		// bIncludeFoundItems
											pInternalFilespec)  );

fail:
	SG_FILE_SPEC_NULLFREE(pCtx, pInternalFilespec);
}

void SG_pendingtree__scan(SG_context* pCtx,
						  SG_pendingtree* pPendingTree,
						  SG_bool bRecursive,
						  SG_bool bIncludeLostItems,
						  SG_bool bIncludeFoundItems,
						  const SG_file_spec* pFilespec)
{
	SG_bool bDirty = SG_FALSE;

	SG_ERR_CHECK(  SG_pendingtree__scan__no_unload__no_save(pCtx,
															pPendingTree,
															bRecursive,
															bIncludeLostItems,
															bIncludeFoundItems,
															pFilespec)  );

	// if any items were actually added do a portability check
	// on each affected directory.
	SG_ERR_CHECK(  sg_ptnode__do_queued_portability_checks(pCtx, pPendingTree)  );

	SG_ERR_CHECK(  sg_ptnode__unload_dir_entries_if_not_dirty(pCtx, pPendingTree->pSuperRoot, &bDirty)  );

	if (!(pPendingTree->pSuperRoot->prbItems))
	{
		SG_PTNODE_NULLFREE(pCtx, pPendingTree->pSuperRoot);
	}

	SG_ERR_CHECK(  SG_pendingtree__save(pCtx, pPendingTree)  );

fail:
	return;
}

//////////////////////////////////////////////////////////////////

/**
 * Do ADDREMOVE with no ITEMS were given in the ITEMLIST.
 * This implicitly assumes the current working copy within the WD.
 */
static void _sg_pendingtree__addremove__when_no_items(SG_context * pCtx,
													  SG_pendingtree * pPendingTree,
													  SG_bool bRecursive,
													  const SG_file_spec* pFilespec)
{
	SG_pathname * pPathCwd = NULL;
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC(pCtx, &pPathCwd)  );
	SG_ERR_CHECK(  SG_pathname__set__from_cwd(pCtx, pPathCwd)  );

#if TRACE_ADDREMOVE
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "AddRemove: Scanning from where... CWD [%s]\n",
							   SG_pathname__sz(pPathCwd))  );
#endif

	SG_ERR_CHECK(  sg_pendingtree__scan_dir(pCtx,
											pPendingTree,
											pPathCwd,
											SG_TRUE,		// bAdd
											SG_TRUE,		// bRemove
											SG_FALSE,		// bMarkImplicit
											bRecursive,
											NULL,			// ptnode
											SG_TRUE,		// bIncludeLostItems
											SG_TRUE,		// bIncludeFoundItems
											pFilespec)  );

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathCwd);
}

/**
 * Do ADDREMOVE on one of the ITEMS from the ITEMLIST.
 * This item forms the top of the sub-tree that we look at.
 * Some of the arguments only apply to things that we respect
 * at this top-level.
 */
static void _sg_pendingtree__addremove__one_item(SG_context * pCtx,
												 SG_pendingtree * pPendingTree,
												 const char * pszItem_k,
												 SG_bool bRecursive,
												 const SG_file_spec* pFilespec)
{
	SG_string* pstrName = NULL;
	SG_pathname* pPathLocal = NULL;
	SG_pathname* pPathDir = NULL;
	struct sg_ptnode* ptnParent;
	struct sg_ptnode* pNewNode = NULL;
	struct sg_ptnode* ptnExistingNode = NULL;
	SG_string * pStringRepoPath = NULL;
	SG_fsobj_stat st;
	SG_file_spec__match_result eval;
	SG_bool bExists;
	SG_int32 compareResult = 0;
	SG_treenode_submodule * pTSM = NULL;
	SG_treenode_entry_type tneType = SG_TREENODEENTRY_TYPE__INVALID;

	SG_ERR_CHECK(  SG_pathname__alloc__sz(pCtx, &pPathLocal, pszItem_k) );
	SG_ERR_CHECK(  SG_workingdir__wdpath_to_repopath(pCtx, pPendingTree->pPathWorkingDirectoryTop, pPathLocal, SG_FALSE, &pStringRepoPath)  );
	SG_ERR_CHECK(  SG_file_spec__should_include(pCtx, pFilespec, SG_string__sz(pStringRepoPath), SG_FILE_SPEC__CONVERT_REPO_PATH | SG_FILE_SPEC__MATCH_TRAILING_SLASH, &eval)  );
#if TRACE_ADDREMOVE
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "AddRemove: Considering item [eval 0x%x][%s]\n", eval, SG_string__sz(pStringRepoPath))  );
#endif

	switch (eval)
	{
	default:	// quiets compiler
		SG_ASSERT(  (0)  );
		goto fail;

	case SG_FILE_SPEC__RESULT__RESERVED:
		SG_ERR_THROW2(  SG_ERR_NOT_ALLOWED,
						(pCtx, "Veracity system folders/files cannot be added to version control: %s",
						 SG_string__sz(pStringRepoPath))  );

	case SG_FILE_SPEC__RESULT__EXCLUDED:
#if TRACE_ADDREMOVE
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "AddRemove: item excluded [%s]\n", SG_string__sz(pStringRepoPath))  );
#endif
		goto done;

	case SG_FILE_SPEC__RESULT__IGNORED:
#if TRACE_ADDREMOVE
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "AddRemove: item ignored [%s]\n", SG_string__sz(pStringRepoPath))  );
#endif
		goto done;

	case SG_FILE_SPEC__RESULT__INCLUDED:
	case SG_FILE_SPEC__RESULT__IMPLIED:
		SG_ERR_CHECK(  SG_fsobj__exists__pathname__exact(pCtx, pPendingTree->pPathWorkingDirectoryTop, pPathLocal, &bExists)  );
		if (!bExists)
			goto do_maybe_lost;
		
		SG_ERR_CHECK(  SG_fsobj__exists_stat__pathname(pCtx, pPathLocal, &bExists, &st)  );
		if (!bExists)
			goto do_maybe_lost;
		SG_ERR_CHECK(  sg_pendingtree__fsobjtype_to_treenodeentrytype(pCtx, st.type, &tneType)  );

		// hard-add a ptnode for this entry regardless of whether it is a
		// file or folder.  if it is a folder, we'll deal with the contents
		// in a minute.

#if TRACE_ADDREMOVE
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "AddRemove: including item [%s]\n", SG_string__sz(pStringRepoPath))  );
#endif
		SG_ERR_CHECK(  SG_repopath__compare(pCtx, SG_string__sz(pStringRepoPath), "@/", &compareResult)  );
		if ( compareResult == 0)
		{
			SG_ERR_CHECK(  sg_pendingtree__local_path_to_ptnode(pCtx, pPendingTree, pPathLocal, SG_TRUE, &pNewNode)  );
			//Don't output any warning messages, just recurse.
		}
		else
		{
			SG_ERR_CHECK(  SG_PATHNAME__ALLOC__COPY(pCtx, &pPathDir, pPathLocal)  );
			SG_ERR_CHECK(  SG_pathname__remove_last(pCtx, pPathDir)  );
			SG_ERR_CHECK(  sg_pendingtree__local_path_to_ptnode(pCtx, pPendingTree, pPathDir, SG_TRUE, &ptnParent)  );
			SG_ERR_CHECK(  SG_pathname__get_last(pCtx, pPathLocal, &pstrName)  );
			SG_ERR_CHECK(  sg_ptnode__find_entry_by_name_not_recursive__active(pCtx, ptnParent, SG_string__sz(pstrName), &ptnExistingNode)  );
			if (ptnExistingNode != NULL)
			{
				SG_ERR_CHECK(  SG_log__report_warning(pCtx, "The item %s is already controlled.", SG_pathname__sz(pPathLocal))  );
				pNewNode = ptnExistingNode;
			}
			else
			{
				if (tneType == SG_TREENODEENTRY_TYPE_DIRECTORY)
				{
					SG_bool bIsWD;
					SG_ERR_CHECK(  SG_TREENODE_SUBMODULE__SNIFF_DIRECTORY(pCtx, pPathLocal, pPendingTree->pRepo, &bIsWD, &pTSM)  );
					if (bIsWD)
						tneType = SG_TREENODEENTRY_TYPE_SUBMODULE;
				}
				SG_ERR_CHECK(  sg_ptnode__create_item(pCtx, ptnParent, SG_string__sz(pstrName),
													  SG_TRUE,		// bAddRemove (actually just check-my-name)
													  SG_FALSE,		// bMarkImplicit
													  tneType,
													  pTSM,
													  &pNewNode)  );
				SG_ERR_CHECK( _pt_log__add_item(pCtx, pPendingTree, pNewNode->pszgid, SG_pathname__sz(pPathLocal), "explicitlyAdded")  );
			}
		}
		if (tneType == SG_TREENODEENTRY_TYPE_DIRECTORY)
			goto do_add_dir;
		else
			goto done;
			
	case SG_FILE_SPEC__RESULT__MAYBE:
		SG_ERR_CHECK(  SG_fsobj__exists__pathname__exact(pCtx, pPendingTree->pPathWorkingDirectoryTop, pPathLocal, &bExists)  );
		if (!bExists)
			goto do_maybe_lost;

		SG_ERR_CHECK(  SG_fsobj__exists_stat__pathname(pCtx, pPathLocal, &bExists, &st)  );
		if (!bExists)
			goto do_maybe_lost;
		SG_ERR_CHECK(  sg_pendingtree__fsobjtype_to_treenodeentrytype(pCtx, st.type, &tneType)  );
		
		if (tneType == SG_TREENODEENTRY_TYPE_DIRECTORY)
		{
			SG_bool bIsWD;
			SG_ERR_CHECK(  SG_treenode_submodule__sniff_directory(pCtx, pPathLocal, pPendingTree->pRepo, &bIsWD, NULL)  );
			if (bIsWD)
			{
				tneType = SG_TREENODEENTRY_TYPE_SUBMODULE;
#if TRACE_ADDREMOVE
				SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "AddRemove: omitting submodule-maybe [%s]\n", SG_string__sz(pStringRepoPath))  );
#endif
			}
			else
			{
#if TRACE_ADDREMOVE
				SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "AddRemove: maybe including directory [%s]\n", SG_string__sz(pStringRepoPath))  );
#endif
				goto do_add_dir;
			}
		}
		else
		{
#if TRACE_ADDREMOVE
			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "AddRemove: omitting non-directory-maybe [%s]\n", SG_string__sz(pStringRepoPath))  );
#endif
			goto done;
		}
	}

	//////////////////////////////////////////////////////////////////
	// NOTREACHED
	//////////////////////////////////////////////////////////////////

do_add_dir:
	// do an ADDREMOVE starting with this sub-directory within the tree.
	// we may or may not have already done a hard-add on the sub-directory.
	// here we scan the contents of the sub-dir and add/remove anything
	// we find within.  if a child does get added, it will back-fill us.

	if (bRecursive)
	{
#if TRACE_ADDREMOVE
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "AddRemove: Scanning directory [%s]\n", SG_string__sz(pStringRepoPath))  );
#endif
		SG_ERR_CHECK(  sg_pendingtree__scan_dir(pCtx,
												pPendingTree,
												pPathLocal,
												SG_TRUE,	    // bAdd
												SG_TRUE,		// bRemove
												SG_FALSE,
												bRecursive,
												NULL,
												SG_TRUE,		// bIncludeLostItems
												SG_TRUE,		// bIncludeFoundItems
												pFilespec)  );
	}
	else
	{
#if TRACE_ADDREMOVE
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "AddRemove: Skipping contents of directory because not recursive [%s]\n", SG_string__sz(pStringRepoPath))  );
#endif
	}
	goto done;

	//////////////////////////////////////////////////////////////////
	// NOTREACHED
	//////////////////////////////////////////////////////////////////

do_maybe_lost:	
	{
		struct sg_ptnode * ptnExistingNode = SG_FALSE;

		// They named something that isn't in the WD.  It is either bogus or
		// possibly refers to something that is currently LOST.
		// 
		// This should probably look in the ptnode and see if
		// this is a LOST entry that should be REMOVED.

#if TRACE_ADDREMOVE
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "ADDREMOVE given pathname to non-existing item [eval %d] [%s]\n",
								   eval, SG_string__sz(pStringRepoPath))  );
#endif
		SG_ERR_CHECK(  SG_PATHNAME__ALLOC__COPY(pCtx, &pPathDir, pPathLocal)  );
		SG_ERR_CHECK(  SG_pathname__remove_last(pCtx, pPathDir)  );
		// find the item's parent directory -- BUT DO NOT SYNTHESIZE -- intermediate ptnodes;
		sg_pendingtree__local_path_to_ptnode(pCtx, pPendingTree, pPathDir, SG_FALSE, &ptnParent);
		if (SG_context__has_err(pCtx))
		{
			if (SG_context__err_equals(pCtx, SG_ERR_PENDINGTREE_PARENT_DIR_MISSING_OR_INACTIVE))
			{
				SG_context__err_reset(pCtx);
				SG_ERR_THROW2(  SG_ERR_NOT_FOUND,
								(pCtx, "ADDREMOVE could not find a parent directory of: '%s'",
								 SG_pathname__sz(pPathLocal))  );
			}
			else
			{
				SG_ERR_RETHROW;
			}
		}
		SG_ERR_CHECK(  SG_pathname__get_last(pCtx, pPathLocal, &pstrName)  );
		SG_ERR_CHECK(  sg_ptnode__find_entry_by_name_not_recursive__active(pCtx, ptnParent, SG_string__sz(pstrName), &ptnExistingNode)  );
		if (!ptnExistingNode)
			SG_ERR_THROW2(  SG_ERR_ITEM_NOT_UNDER_VERSION_CONTROL,
							(pCtx, "%s", SG_pathname__sz(pPathLocal))  );

		// we found the ptnode for the given pathname.
		// we know the file/dir is not on disk, so we
		// can let REMOVE take care of it (and we don't
		// have to worry about the 2-pass thing throwing
		// an error).

		{
			const char * pszPath = SG_pathname__sz(pPathLocal);
			SG_ERR_CHECK(  SG_pendingtree__remove_dont_save_pendingtree(pCtx,
																		pPendingTree,
																		1, &pszPath,
																		gpBlankFilespec,
																		SG_FALSE, SG_FALSE, SG_FALSE)  );
		}

		goto done;
	}

	//////////////////////////////////////////////////////////////////
	// NOTREACHED
	//////////////////////////////////////////////////////////////////

done:
	;
fail:
	SG_STRING_NULLFREE(pCtx, pstrName);
	SG_STRING_NULLFREE(pCtx, pStringRepoPath);
	SG_PATHNAME_NULLFREE(pCtx, pPathDir);
	SG_PATHNAME_NULLFREE(pCtx, pPathLocal);
	SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pTSM);
}

void SG_pendingtree__addremove(SG_context* pCtx, 
							   SG_pendingtree* pPendingTree,
							   SG_uint32 count_items, const char** paszItems,
							   SG_bool bRecursive,
							   const SG_file_spec* pFilespec,
							   SG_bool bTest)
{	
	SG_bool bDirty = SG_FALSE;
	SG_file_spec* pInternalFilespec = NULL;

	SG_ERR_CHECK(  SG_FILE_SPEC__ALLOC__COPY(pCtx, pFilespec, &pInternalFilespec)  );
	SG_ERR_CHECK(  SG_file_spec__load_ignores(pCtx,
											  pInternalFilespec,
											  pPendingTree->pRepo,
											  pPendingTree->pPathWorkingDirectoryTop,
											  0u)  );
	SG_ERR_CHECK(  SG_file_spec__load_reserved(pCtx, pInternalFilespec)  );

	if (count_items == 0)
	{
		SG_ERR_CHECK(  _sg_pendingtree__addremove__when_no_items(pCtx,
																 pPendingTree,
																 bRecursive,
																 pInternalFilespec)  );
	}
	else
	{
		SG_uint32 i;

		for (i=0; i<count_items; i++)
		{
			SG_ERR_CHECK(  _sg_pendingtree__addremove__one_item(pCtx,
																pPendingTree,
																paszItems[i],
																bRecursive,
																pInternalFilespec)  );
		}
	}

	// if any items were actually added do a portability check
	// on each affected directory.
	SG_ERR_CHECK(  sg_ptnode__do_queued_portability_checks(pCtx, pPendingTree)  );

	/* TODO - this may cause problem if there are paszItems, not sure
	 *
	 * TODO 2010/06/05 The above TODO is very old.  Think about whether it is still relevant. 
	 *
	 */
	SG_ERR_CHECK(  sg_ptnode__unload_dir_entries_if_not_dirty(pCtx, pPendingTree->pSuperRoot, &bDirty)  );

	if (!(pPendingTree->pSuperRoot->prbItems))
	{
		SG_PTNODE_NULLFREE(pCtx, pPendingTree->pSuperRoot);
	}

	if (bTest == SG_FALSE)
		SG_ERR_CHECK(  SG_pendingtree__save(pCtx, pPendingTree)  );

fail:
	SG_FILE_SPEC_NULLFREE(pCtx, pInternalFilespec);
	return;
}

//////////////////////////////////////////////////////////////////

static void sg_ptnode__get_repo_path(
	SG_context* pCtx, 
	struct sg_ptnode* ptn,
	SG_string** ppstr
	)
{
	const char* pszName = NULL;
	
	if (ptn->pCurrentParent)
	{
		SG_ERR_CHECK(  sg_ptnode__get_repo_path(pCtx, ptn->pCurrentParent, ppstr)  );
	}
	else
	{
		SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, ppstr)  );
	}

	SG_ERR_CHECK(  sg_ptnode__get_name(pCtx, ptn, &pszName)  );
	if (pszName)
	{
#if 0		
		if (0 != SG_string__length_in_bytes(*ppstr))
		{
			SG_ERR_CHECK(  SG_repopath__ensure_final_slash(pCtx, *ppstr)  );
		}
#endif
		SG_ERR_CHECK(  SG_string__append__sz(pCtx, *ppstr, pszName)  );
		if (SG_TREENODEENTRY_TYPE_DIRECTORY == ptn->type)
		{
			SG_ERR_CHECK(  SG_repopath__ensure_final_slash(pCtx, *ppstr)  );
		}
	}

fail:
	return;
}

/**
 * Like sg_ptnode__get_repo_path(), but we get the original version of
 * the path (before any pending changes (moves/renames) in a parent.
 */
/* This isn't used, which makes gcc 4.3.3 unhappy.
static SG_error sg_ptnode__get_old_repo_path(
	struct sg_ptnode* ptn,
	SG_string** ppstr
	)
{
	SG_error err;
	const char* pszName = NULL;
	
	if (ptn->pCurrentParent)
	{
		SG_ERR_CHECK(  sg_ptnode__get_old_repo_path(ptn->pCurrentParent, ppstr)  );
	}
	else
	{
		SG_ERR_CHECK(  SG_STRING__ALLOC(ppstr)  );
	}

	SG_ERR_CHECK(  sg_ptnode__get_old_name(ptn, &pszName)  );
	if (pszName)
	{
#if 0		
		if (0 != SG_string__length_in_bytes(*ppstr))
		{
			SG_ERR_CHECK(  SG_repopath__ensure_final_slash(*ppstr)  );
		}
#endif
		SG_ERR_CHECK(  SG_string__append__sz(*ppstr, pszName)  );
		if (SG_TREENODEENTRY_TYPE_DIRECTORY == ptn->type)
		{
			SG_ERR_CHECK(  SG_repopath__ensure_final_slash(*ppstr)  );
		}
	}
	
	return SG_ERR_OK;

fail:
	return err;
}
*/

static void sg_pendingtree__find_by_gid(
	SG_context* pCtx, 
	SG_pendingtree* pPendingTree,
	const char* pszGid,
	struct sg_ptnode** ppResult
	);

void SG_pendingtree__get_repo_path(
	SG_context* pCtx, 
	SG_pendingtree* pPendingTree,
	const char* pszgid,
	SG_string** ppResult
	)
{
	struct sg_ptnode* ptn = NULL;
	SG_string* pstr = NULL;

	SG_ERR_CHECK(  SG_gid__argcheck(pCtx, pszgid)  );

	SG_ERR_CHECK(  sg_pendingtree__find_by_gid(pCtx, pPendingTree, pszgid, &ptn)  );
	if (!ptn)
	{
		SG_ERR_THROW(  SG_ERR_NOT_FOUND  );
	}

	SG_ERR_CHECK(  sg_ptnode__get_repo_path(pCtx, ptn, &pstr)  );

	*ppResult = pstr;

fail:
	return;
}

//////////////////////////////////////////////////////////////////

void SG_pendingtree__add_dont_save_pendingtree(SG_context* pCtx, 
											   SG_pendingtree* pPendingTree,
											   SG_uint32 count_items, const char** paszItems,
											   SG_bool bRecursive,
											   const SG_file_spec* pFilespec)
{	
	SG_uint32 i;
	struct sg_ptnode* ptnParent;
	SG_string* pstrName = NULL;
	SG_pathname* pPathDir = NULL;
	SG_pathname* pPathLocal = NULL;
	struct sg_ptnode* pNewNode = NULL;
	struct sg_ptnode* ptnExistingNode = NULL;
	SG_string * pStringRepoPath = NULL;
	SG_file_spec__match_result eval;
	SG_fsobj_stat st;
	SG_treenode_entry_type tneType = SG_TREENODEENTRY_TYPE__INVALID;
	SG_int32 compareResult = 0;
	SG_file_spec* pInternalFilespec = NULL;
	SG_pathname * pPathClosestRoot = NULL;
	SG_pathname * pPathPendingTreeRoot = NULL;
	SG_string * pStringDeltaRoots = NULL;
	SG_treenode_submodule * pTSM = NULL;
	SG_bool bExists = SG_FALSE;
	SG_string * pRelativePathFromClosestRoot = NULL;

	SG_ARGCHECK_RETURN(  (count_items > 0), count_items  );

	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__COPY(pCtx, &pPathPendingTreeRoot, pPendingTree->pPathWorkingDirectoryTop)  );
	SG_ERR_IGNORE(  SG_pathname__remove_final_slash(pCtx, pPathPendingTreeRoot, NULL)  );
	
	SG_ERR_CHECK(  SG_FILE_SPEC__ALLOC__COPY(pCtx, pFilespec, &pInternalFilespec)  );
	SG_ERR_CHECK(  SG_file_spec__load_ignores(pCtx,
											  pInternalFilespec,
											  pPendingTree->pRepo,
											  pPendingTree->pPathWorkingDirectoryTop,
											  0u)  );
	SG_ERR_CHECK(  SG_file_spec__load_reserved(pCtx, pInternalFilespec)  );

	for (i=0; i<count_items; i++)
	{
		bExists = SG_FALSE;
		SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPathLocal, paszItems[i])  );
		SG_ERR_IGNORE(  SG_pathname__remove_final_slash(pCtx, pPathLocal, NULL)  );

		// With submodules present, we cannot assume that they didn't give us an explicit
		// pathname that crosses into a sub-dir that is a submodule.  So see if the absolute
		// path of the candidate should properly be contained within this WD/pendingtree or
		// within a submodule.

		SG_ERR_CHECK(  SG_workingdir__find_mapping(pCtx, pPathLocal, &pPathClosestRoot, NULL, NULL)  );
		SG_ERR_IGNORE(  SG_pathname__remove_final_slash(pCtx, pPathClosestRoot, NULL)  );

		SG_ERR_CHECK(  SG_pathname__make_relative(pCtx,
												  pPathPendingTreeRoot,
												  pPathClosestRoot,
												  &pStringDeltaRoots)  );
#if TRACE_ADD
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   ("Add: test\n"
									"     [path       %s]\n"
									"     [closest WD %s]\n"
									"     [pending WD %s]\n"
									"     [delta   WD %s]\n"),
								   SG_pathname__sz(pPathLocal),
								   SG_pathname__sz(pPathClosestRoot),
								   SG_pathname__sz(pPathPendingTreeRoot),
								   SG_string__sz(pStringDeltaRoots))  );
#endif
		if (SG_string__length_in_bytes(pStringDeltaRoots) == 0)
		{
			// both root pathnames point to the same WD.  this is the normal case
			// when no submodules are involved.  allow the ADD to proceed.
		}
		else
		{
			if (strcmp(SG_pathname__sz(pPathClosestRoot),
					   SG_pathname__sz(pPathLocal)) == 0)
			{
				// The given path is the root of a submodule.  Allow the ADD to proceed
				// under the assumption that we'll be adding the entry as a submodule
				// rather than a regular directory.
			}
			else
			{
				// if (strncmp(SG_string__sz(pStringDeltaRoots), "../", 3) == 0)
				//     the closest containing WD of the given path is *NOT* within
				//     the tree rooted at the WD root associated with this pendingtree.
				// else
				//     the closest containing WD of the given path is somewhere nested
				//     inside the pendingtree's WD.
				// either way, the candidate entry properly belongs to another WD not us.

				SG_ERR_THROW2(  SG_ERR_PATH_IN_ANOTHER_SUBMODULE,
								(pCtx, "The path '%s' should belong to the working copy/submodule rooted at '%s' rather than the one at '%s'.",
								 SG_pathname__sz(pPathLocal),
								 SG_pathname__sz(pPathClosestRoot),
								 SG_pathname__sz(pPathPendingTreeRoot))  );
			}
		}

		SG_ERR_CHECK(  SG_workingdir__wdpath_to_repopath(pCtx, pPendingTree->pPathWorkingDirectoryTop, pPathLocal, SG_TRUE, &pStringRepoPath)  );

		SG_ERR_CHECK(  SG_fsobj__exists_stat__pathname(pCtx, pPathLocal, &bExists, &st)  );
		if (!bExists)
			SG_ERR_THROW2(  SG_ERR_NOT_FOUND,
							(pCtx, "Could not find '%s'", SG_pathname__sz(pPathLocal))  );
		SG_ERR_CHECK(  SG_fsobj__exists__pathname__exact(pCtx, pPathClosestRoot, pPathLocal, &bExists)  );
		if (!bExists)
			SG_ERR_THROW2(  SG_ERR_NOT_FOUND,
							(pCtx, "Could not find '%s' (check case)", SG_pathname__sz(pPathLocal))  );

		SG_ERR_CHECK(  sg_pendingtree__fsobjtype_to_treenodeentrytype(pCtx, st.type, &tneType)  );

#if TRACE_ADD
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "Add: calling __should_include [%s]\n", SG_string__sz(pStringRepoPath))  );
#endif
		SG_ERR_CHECK(  SG_file_spec__should_include(pCtx,
													pInternalFilespec,
													SG_string__sz(pStringRepoPath),
													SG_FILE_SPEC__CONVERT_REPO_PATH | SG_FILE_SPEC__MATCH_TRAILING_SLASH,
													&eval)  );
#if TRACE_ADD
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "Add: considering [eval 0x%x][%s]\n", eval, SG_string__sz(pStringRepoPath))  );
#endif

		switch (eval)
		{
		default:	// quiets compiler
			SG_ASSERT(  (0)  );
			goto done_with_this_one;

		case SG_FILE_SPEC__RESULT__RESERVED:
			SG_ERR_THROW2(  SG_ERR_NOT_ALLOWED,
							(pCtx, "Veracity system folders/files cannot be added to version control: %s",
							 SG_string__sz(pStringRepoPath))  );

		case SG_FILE_SPEC__RESULT__EXCLUDED:
#if TRACE_ADD
			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "Add: item excluded [%s]\n", SG_string__sz(pStringRepoPath))  );
#endif
			goto done_with_this_one;

		case SG_FILE_SPEC__RESULT__IGNORED:
#if TRACE_ADD
			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "Add: item ignored [%s]\n", SG_string__sz(pStringRepoPath))  );
#endif
			goto done_with_this_one;

		case SG_FILE_SPEC__RESULT__INCLUDED:
		case SG_FILE_SPEC__RESULT__IMPLIED:
#if TRACE_ADD
			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "Add: Adding [%s]\n", SG_string__sz(pStringRepoPath))  );
#endif
			SG_ERR_CHECK(  SG_repopath__compare(pCtx, SG_string__sz(pStringRepoPath), "@/", &compareResult)  );
			if ( compareResult == 0)
			{
				SG_ERR_CHECK(  sg_pendingtree__local_path_to_ptnode(pCtx, pPendingTree, pPathLocal, SG_TRUE, &pNewNode)  );
				//Don't output any warning messages, just recurse.
			}
			else
			{
				SG_ERR_CHECK(  SG_PATHNAME__ALLOC__COPY(pCtx, &pPathDir, pPathLocal)  );
				SG_ERR_CHECK(  SG_pathname__remove_last(pCtx, pPathDir)  );
				SG_ERR_CHECK(  sg_pendingtree__local_path_to_ptnode(pCtx, pPendingTree, pPathDir, SG_TRUE, &ptnParent)  );
				SG_ERR_CHECK(  SG_pathname__get_last(pCtx, pPathLocal, &pstrName)  );
				SG_ERR_CHECK(  sg_ptnode__find_entry_by_name_not_recursive__active(pCtx, ptnParent, SG_string__sz(pstrName), &ptnExistingNode)  );
				if (ptnExistingNode != NULL)
				{
					SG_ERR_CHECK(  SG_log__report_warning(pCtx, "The item %s is already controlled.", SG_pathname__sz(pPathLocal))  );
					pNewNode = ptnExistingNode;
				}
				else
				{
					// before we add this item see if might be a submodule.  for directories
					// that have DVCS control files we convert the tneType to be a submodule-link
					// and treat it as a black box rather than a directory.

					SG_ASSERT(  (pTSM == NULL)  );		// this assert is only to help find leaks.

					// TODO 2010/10/26 Add arg to this function to see if they *want* to consider submodules.

					if (tneType == SG_TREENODEENTRY_TYPE_DIRECTORY)
					{
						SG_bool bIsWD;
						SG_ERR_CHECK(  SG_TREENODE_SUBMODULE__SNIFF_DIRECTORY(pCtx, pPathLocal, pPendingTree->pRepo, &bIsWD, &pTSM)  );
						if (bIsWD)
						{
							// we mark it a submodule by updating the tneType only.
							// we **DO NOT** set the 'pszCurrentHid' or store the TSM
							// in the ptnode because we don't have a place for it and
							// we don't want to have to write blob to the repo for the
							// TSM until an actual commit.  also, since the submodule
							// is live, we need to recompute the HID (like we do on
							// the content of files) during a scan-dir.

							tneType = SG_TREENODEENTRY_TYPE_SUBMODULE;
						}
					}

					SG_ERR_CHECK(  sg_ptnode__create_item(pCtx, ptnParent, SG_string__sz(pstrName),
														  SG_TRUE,		// bAddRemove (actually just check-my-name)
														  SG_FALSE,		// bMarkImplicit
														  tneType,
														  pTSM,
														  &pNewNode)  );
					SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pTSM);
					SG_ERR_CHECK( _pt_log__add_item(pCtx, pPendingTree, pNewNode->pszgid, SG_pathname__sz(pPathLocal), "explicitlyAdded")  );
				}
			}
			if (tneType == SG_TREENODEENTRY_TYPE_DIRECTORY)
				goto do_add_dir;
			else
				goto done_with_this_one;

		case SG_FILE_SPEC__RESULT__MAYBE:
			if (tneType == SG_TREENODEENTRY_TYPE_DIRECTORY)
			{
				// TODO 2010/10/26 Add arg to this function to see if they *want* to consider submodules.

				SG_bool bIsWD;
				SG_ERR_CHECK(  SG_treenode_submodule__sniff_directory(pCtx, pPathLocal, pPendingTree->pRepo, &bIsWD, NULL)  );
				if (bIsWD)
				{
					tneType = SG_TREENODEENTRY_TYPE_SUBMODULE;
#if TRACE_ADD
					SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "Add: omitting submodule-maybe [%s]\n", SG_string__sz(pStringRepoPath))  );
#endif
				}
				else
				{
#if TRACE_ADD
					SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "Add: maybe including directory [%s]\n", SG_string__sz(pStringRepoPath))  );
#endif
					goto do_add_dir;
				}
			}
			else
			{
#if TRACE_ADD
				SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "Add: omitting not-directory-maybe [%s]\n", SG_string__sz(pStringRepoPath))  );
#endif
				goto done_with_this_one;
			}
		}
		
		//////////////////////////////////////////////////////////////////
		// NOT REACHED
		//////////////////////////////////////////////////////////////////

do_add_dir:
		if (bRecursive)
		{
#if TRACE_ADD
			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "Add: Scanning directory [%s]\n", SG_string__sz(pStringRepoPath))  );
#endif
			SG_ERR_CHECK(  sg_pendingtree__scan_dir(pCtx,
													pPendingTree,
													pPathLocal,
													SG_TRUE,	// bAdd
													SG_FALSE,	// bRemove
													SG_FALSE,	// bMarkImplicit
													bRecursive,
													pNewNode,
													SG_TRUE,		// bIncludeLostItems
													SG_TRUE,		// bIncludeFoundItems
													pInternalFilespec)  );
		}
		else
		{
#if TRACE_ADD
			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "Add: Skipping contents of directory because not recursive [%s]\n", SG_string__sz(pStringRepoPath))  );
#endif
		}

done_with_this_one:
		SG_STRING_NULLFREE(pCtx, pstrName);
		SG_PATHNAME_NULLFREE(pCtx, pPathDir);
		SG_PATHNAME_NULLFREE(pCtx, pPathLocal);
		SG_STRING_NULLFREE(pCtx, pStringRepoPath);
		SG_STRING_NULLFREE(pCtx, pStringDeltaRoots);
		SG_PATHNAME_NULLFREE(pCtx, pPathClosestRoot);
		SG_STRING_NULLFREE(pCtx, pRelativePathFromClosestRoot);
		SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pTSM);
	}

	SG_PATHNAME_NULLFREE(pCtx, pPathPendingTreeRoot);
	SG_FILE_SPEC_NULLFREE(pCtx, pInternalFilespec);

	// when new ptnodes were added, we queued a portability check
	// on the parent directory.  do all of them now.
	SG_ERR_CHECK(  sg_ptnode__do_queued_portability_checks(pCtx, pPendingTree)  );

	return;

fail:
	SG_FILE_SPEC_NULLFREE(pCtx, pInternalFilespec);
	SG_STRING_NULLFREE(pCtx, pstrName);
	SG_PATHNAME_NULLFREE(pCtx, pPathDir);
	SG_PATHNAME_NULLFREE(pCtx, pPathLocal);
	SG_STRING_NULLFREE(pCtx, pStringRepoPath);
	SG_STRING_NULLFREE(pCtx, pStringDeltaRoots);
	SG_PATHNAME_NULLFREE(pCtx, pPathClosestRoot);
	SG_PATHNAME_NULLFREE(pCtx, pPathPendingTreeRoot);
	SG_STRING_NULLFREE(pCtx, pRelativePathFromClosestRoot);
	SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pTSM);
}

void SG_pendingtree__add(SG_context* pCtx, 
						 SG_pendingtree* pPendingTree,
						 SG_uint32 count_items, const char** paszItems,
						 SG_bool bRecursive,
						 const SG_file_spec* pFilespec,
						 SG_bool bTest)
{
	if (count_items == 0)
		SG_ERR_THROW_RETURN(  SG_ERR_USAGE  );

	SG_ERR_CHECK(  SG_pendingtree__add_dont_save_pendingtree(pCtx, pPendingTree,
															 count_items, paszItems,
															 bRecursive,
															 pFilespec)  );

	if (bTest == SG_FALSE)
		SG_ERR_CHECK(  SG_pendingtree__save(pCtx, pPendingTree)  );

fail:
	return;
}

void SG_pendingtree__add_dont_save_pendingtree__with_gid(SG_context * pCtx,
														 SG_pendingtree * pPendingTree,
														 const char * pszGid,
														 const SG_pathname * pPath)
{
	// We DO NOT take a bRecursive option to tell us to scan the disk
	// when the item is a diretory.  The assumption is that if the caller
	// already has a GID for this item, it is not scanning and is in complete
	// control of what is being added (such as when doing a MERGE or RESOLVE).
	// 
	// Likewise, we DO NOT look at the INCLUDES/EXCLUDES/IGNORES list for this.
	//
	// And we DO NOT synthesize missing parent directories.

	SG_string * pStringEntryname = NULL;
	SG_pathname * pPathDir = NULL;
	struct sg_ptnode * ptn;
	struct sg_ptnode * pNewNode;
	SG_fsobj_stat st;
	SG_treenode_entry_type tneType = SG_TREENODEENTRY_TYPE__INVALID;
	SG_treenode_submodule * pTSM = NULL;

	// split full pathname into <dir> and <entryname>

	SG_ERR_CHECK(  SG_pathname__get_last(pCtx, pPath, &pStringEntryname)  );
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__COPY(pCtx, &pPathDir, pPath)  );
	SG_ERR_CHECK(  SG_pathname__remove_last(pCtx, pPathDir)  );

	// find ptnode for <dir>.

	SG_ERR_CHECK(  sg_pendingtree__local_path_to_ptnode(pCtx, pPendingTree, pPathDir, SG_FALSE, &ptn)  );

	// make sure the entry is present on the disk.

	SG_ERR_CHECK(  SG_fsobj__stat__pathname(pCtx, pPath, &st)  );
	SG_ERR_CHECK(  sg_pendingtree__fsobjtype_to_treenodeentrytype(pCtx, st.type, &tneType)  );
	if (tneType == SG_TREENODEENTRY_TYPE_DIRECTORY)
	{
		SG_bool bIsWD;
		SG_ERR_CHECK(  SG_TREENODE_SUBMODULE__SNIFF_DIRECTORY(pCtx, pPath, pPendingTree->pRepo, &bIsWD, &pTSM)  );
		if (bIsWD)
			tneType = SG_TREENODEENTRY_TYPE_SUBMODULE;
	}
	SG_ERR_CHECK(  sg_ptnode__create_item__with_gid(pCtx, ptn, SG_string__sz(pStringEntryname),
													SG_TRUE,		// bAddRemove
													SG_FALSE,		// bMarkImplicit
													tneType,
													pTSM,
													pszGid,
													&pNewNode)  );

	// when new ptnodes were added, we queued a portability check
	// on the parent directory.  do all of them now.
	
	SG_ERR_CHECK(  sg_ptnode__do_queued_portability_checks(pCtx, pPendingTree)  );

fail:
	SG_STRING_NULLFREE(pCtx, pStringEntryname);
	SG_PATHNAME_NULLFREE(pCtx, pPathDir);
	SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pTSM);
}

void SG_pendingtree__move_dont_save_pendingtree(
	SG_context* pCtx,
	SG_pendingtree* pPendingTree,
	const SG_pathname* pPathMoveMe,
	const SG_pathname* pPathMoveTo,
	SG_bool bAllowDestinationDirInDrawer
	)
{
	struct sg_ptnode* ptnParentDirOfItemBeingMoved = NULL;
	struct sg_ptnode* ptnMoveTo = NULL;
	struct sg_ptnode* ptnCheckForParentMove = NULL;
	struct sg_ptnode* ptnCloneOfItemBeingMoved = NULL;
	SG_string* pstrName = NULL;
	struct sg_ptnode* ptnItemBeingMoved = NULL;
	SG_pathname* pPathParentDirOfItemBeingMoved = NULL;
	SG_pathname* pPathMoved = NULL;
	SG_bool bExists = SG_FALSE;
    SG_fsobj_type t;
	SG_file_spec * pFilespec = NULL;
	SG_string * pStringRepoPathDestinationDir = NULL;
	SG_file_spec__match_result eval;

	SG_NULLARGCHECK_RETURN( pPendingTree );
	SG_NULLARGCHECK_RETURN( pPathMoveMe );
	SG_NULLARGCHECK_RETURN( pPathMoveTo );

#if TRACE_MOVE
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "MOVE %s ==> %s\n",
							   SG_pathname__sz(pPathMoveMe),
							   SG_pathname__sz(pPathMoveTo))  );
#endif

	// Make sure destination directory exists (and with correct case).

	SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pPathMoveTo, &bExists, &t, NULL)  );
    if (!bExists || t != SG_FSOBJ_TYPE__DIRECTORY)
        SG_ERR_THROW2(  SG_ERR_NOT_A_DIRECTORY,
						(pCtx, "Move destination must be a directory: %s", SG_pathname__sz(pPathMoveTo))  );
	SG_ERR_CHECK(  SG_fsobj__exists__pathname__exact(pCtx,
													 pPendingTree->pPathWorkingDirectoryTop,
													 pPathMoveTo,
													 &bExists)  );
	if (!bExists)
		SG_ERR_THROW2(  SG_ERR_NOT_FOUND,
						(pCtx, "Could not find destination directory '%s' (check case).",
						 SG_pathname__sz(pPathMoveTo))  );

	if (bAllowDestinationDirInDrawer)
	{
		// MERGE moves (parks) some things in the parking lot
		// as it juggles the working directory.  so .sgdrawer
		// may get TEMPORARILY added to the pendingtree (to
		// keep from partitioning the tree).
	}
	else
	{
		// normally we DO NOT allow anything within .sgdrawer
		// to be added (DIRECTLY by ADD).  here we want to keep
		// .sgdrawer from being INDIRECTLY added by MOVE.

		SG_ERR_CHECK(  SG_FILE_SPEC__ALLOC(pCtx, &pFilespec)  );
		// no includes/excludes/ignores.
		SG_ERR_CHECK(  SG_file_spec__load_reserved(pCtx, pFilespec)  );

		SG_ERR_CHECK(  SG_workingdir__wdpath_to_repopath(pCtx, pPendingTree->pPathWorkingDirectoryTop,
														 pPathMoveTo, SG_FALSE, &pStringRepoPathDestinationDir)  );
		SG_ERR_CHECK(  SG_file_spec__should_include(pCtx, pFilespec, SG_string__sz(pStringRepoPathDestinationDir),
													SG_FILE_SPEC__CONVERT_REPO_PATH |
													SG_FILE_SPEC__NO_INCLUDES |
													SG_FILE_SPEC__NO_EXCLUDES |
													SG_FILE_SPEC__NO_IGNORES,
													&eval)  );
#if TRACE_MOVE
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "MOVE: considering destination [eval 0x%x][%s]\n", eval, SG_string__sz(pStringRepoPathDestinationDir))  );
#endif
		switch (eval)
		{
		default:
			break;

		case SG_FILE_SPEC__RESULT__RESERVED:
			SG_ERR_THROW2(  SG_ERR_NOT_ALLOWED,
							(pCtx, "Cannot move item '%s' into Veracity system folder '%s'",
							 SG_pathname__sz(pPathMoveMe),
							 SG_pathname__sz(pPathMoveTo))  );
		}
	}
		
	SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pPathMoveMe, &bExists, NULL, NULL)  );
    if (!bExists)
    {
    	SG_ERR_THROW2( SG_ERR_NOT_FOUND, (pCtx,"%s",SG_pathname__sz(pPathMoveMe) )  );
    }

    /* We want the parent directory of this item.  So copy the path and remove the last component */
    SG_ERR_CHECK(  SG_PATHNAME__ALLOC__COPY(pCtx, &pPathParentDirOfItemBeingMoved, pPathMoveMe)  );
    SG_ERR_CHECK(  SG_pathname__remove_last(pCtx, pPathParentDirOfItemBeingMoved)  );

    /* Find the ptnodes */
    SG_ERR_CHECK(  sg_pendingtree__local_path_to_ptnode(pCtx, pPendingTree, pPathParentDirOfItemBeingMoved, SG_FALSE, &ptnParentDirOfItemBeingMoved)  );
    if (!ptnParentDirOfItemBeingMoved)
    {
        SG_ERR_THROW2(  SG_ERR_ITEM_NOT_UNDER_VERSION_CONTROL,
						(pCtx, "Parent directory of '%s'.", SG_pathname__sz(pPathMoveMe))  );
    }
    SG_ERR_CHECK(  sg_pendingtree__local_path_to_ptnode(pCtx, pPendingTree, pPathMoveTo, SG_TRUE, &ptnMoveTo)  );
    if (!ptnMoveTo)
    {
        SG_ERR_THROW2(  SG_ERR_ITEM_NOT_UNDER_VERSION_CONTROL,
						(pCtx, "Destination directory '%s'.", SG_pathname__sz(pPathMoveTo))  );
    }

    if (ptnParentDirOfItemBeingMoved == ptnMoveTo)
    {
        SG_ERR_THROW2(  SG_ERR_CANNOT_MOVE_INTO_CURRENT_PARENT,
						(pCtx, "Parent directory of item '%s' is the same as the destination directory '%s'.",
						 SG_pathname__sz(pPathMoveMe),
						 SG_pathname__sz(pPathMoveTo))  );
    }

    /* Get the base name from the original path */
    SG_ERR_CHECK(  SG_pathname__get_last(pCtx, pPathMoveMe, &pstrName)  );

    /* leave the item in both places, one as a phantom */

    SG_ERR_CHECK(  sg_ptnode__find_entry_by_name_not_recursive__active(pCtx, ptnParentDirOfItemBeingMoved, SG_string__sz(pstrName), &ptnItemBeingMoved)  );
    if (!ptnItemBeingMoved)
    {
        SG_ERR_THROW2(  SG_ERR_ITEM_NOT_UNDER_VERSION_CONTROL,
						(pCtx, "%s", SG_pathname__sz(pPathMoveMe))  );
    }

    if (ptnItemBeingMoved == ptnMoveTo)
    {
        SG_ERR_THROW2(  SG_ERR_CANNOT_MOVE_INTO_SUBFOLDER,
						(pCtx, "Cannot move item '%s' into itself.", SG_pathname__sz(pPathMoveMe))  );
    }
    ptnCheckForParentMove = ptnMoveTo;
    while (ptnCheckForParentMove != NULL)
    {
        if (ptnCheckForParentMove == ptnItemBeingMoved)
            SG_ERR_THROW2(  SG_ERR_CANNOT_MOVE_INTO_SUBFOLDER,
							(pCtx, "Cannot move item '%s' into a sub-directory of itself.", SG_pathname__sz(pPathMoveMe))  );
        else
        {
            if (ptnCheckForParentMove->pCurrentParent != NULL)
                ptnCheckForParentMove = ptnCheckForParentMove->pCurrentParent;
            else
                break;
        }
    }

    if (ptnItemBeingMoved->pszgidMovedFrom)
    {
        struct sg_ptnode* ptn2 = NULL;

        /* this item has already been moved.  So it already has
         * a phantom in its original location.  We don't need to
         * create a new one.  We just need to move this ptnode
         * into the new location.  Leave pszgidMovedFrom alone. */

        SG_ERR_CHECK(  SG_rbtree__remove__with_assoc(pCtx, ptnParentDirOfItemBeingMoved->prbItems, ptnItemBeingMoved->pszgid, (void**) &ptn2)  );
        SG_ASSERT(ptn2 == ptnItemBeingMoved);
        ptnItemBeingMoved->pCurrentParent = NULL;
        ptn2 = NULL;

        if (0 == strcmp(ptnMoveTo->pszgid, ptnItemBeingMoved->pszgidMovedFrom))
        {
            /* item is being moved back to where it came from.  we need
             * to get rid of the old phantom node. */

            SG_ERR_CHECK(  SG_rbtree__remove__with_assoc(pCtx, ptnMoveTo->prbItems, ptnItemBeingMoved->pszgid, (void**) &ptn2)  );
            SG_ASSERT(ptn2->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY);

            SG_PTNODE_NULLFREE(pCtx, ptn2);

            ptnItemBeingMoved->pszgidMovedFrom = NULL;
        }

        SG_ERR_CHECK(  sg_ptnode__add_entry(pCtx, ptnMoveTo, ptnItemBeingMoved)  );
        ptnItemBeingMoved->temp_flags |= sg_PTNODE_TEMP_FLAG_CHECKMYNAME;

        SG_ERR_CHECK(  sg_ptnode__immediate_portability_check(pCtx, ptnMoveTo)  );
    }
    else
    {
        if (ptnItemBeingMoved->pBaselineEntry)
        {
            SG_ERR_CHECK(  sg_ptnode__clone_for_move(pCtx, ptnItemBeingMoved, &ptnCloneOfItemBeingMoved)  );

            SG_ERR_CHECK(  sg_ptnode__add_entry(pCtx, ptnMoveTo, ptnCloneOfItemBeingMoved)  );
            ptnCloneOfItemBeingMoved->temp_flags |= sg_PTNODE_TEMP_FLAG_CHECKMYNAME;

            SG_ERR_CHECK(  SG_strpool__add__sz(pCtx, pPendingTree->pStrPool, ptnParentDirOfItemBeingMoved->pszgid, &ptnCloneOfItemBeingMoved->pszgidMovedFrom)  );

            ptnItemBeingMoved->saved_flags |= sg_PTNODE_SAVED_FLAG_MOVED_AWAY;
            ptnCloneOfItemBeingMoved->prbItems = ptnItemBeingMoved->prbItems;
            ptnItemBeingMoved->prbItems = NULL;
            ptnCloneOfItemBeingMoved = NULL;
            SG_ERR_CHECK(  sg_ptnode__immediate_portability_check(pCtx, ptnMoveTo)  );
        }
        else
        {
            /* this item is newly added (and being moved before it was ever committed)
			 * (just move it to the new directory without leaving a phantom). */
            struct sg_ptnode* ptn2 = NULL;

            SG_ERR_CHECK(  SG_rbtree__remove__with_assoc(pCtx, ptnParentDirOfItemBeingMoved->prbItems, ptnItemBeingMoved->pszgid, (void**) &ptn2)  );
            SG_ASSERT(ptn2 == ptnItemBeingMoved);
            ptnItemBeingMoved->pCurrentParent = NULL;
            ptn2 = NULL;

            SG_ERR_CHECK(  sg_ptnode__add_entry(pCtx, ptnMoveTo, ptnItemBeingMoved)  );
            ptnItemBeingMoved->temp_flags |= sg_PTNODE_TEMP_FLAG_CHECKMYNAME;

            SG_ERR_CHECK(  sg_ptnode__immediate_portability_check(pCtx, ptnMoveTo)  );
        }
    }

    /* Do the local rename */
    SG_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx, &pPathMoved, pPathMoveTo, SG_string__sz(pstrName))  );
    SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pPathMoved, &bExists, NULL, NULL)  );
    if (bExists)
    {
        SG_ERR_THROW(  SG_ERR_PENDINGTREE_OBJECT_ALREADY_EXISTS  );
    }
    SG_ERR_CHECK(  SG_fsobj__move__pathname_pathname(pCtx, pPathMoveMe, pPathMoved)  );

fail:
    SG_PTNODE_NULLFREE(pCtx, ptnCloneOfItemBeingMoved);
	SG_PATHNAME_NULLFREE(pCtx, pPathParentDirOfItemBeingMoved);
	SG_PATHNAME_NULLFREE(pCtx, pPathMoved);
	SG_STRING_NULLFREE(pCtx, pstrName);
	SG_FILE_SPEC_NULLFREE(pCtx, pFilespec);
	SG_STRING_NULLFREE(pCtx, pStringRepoPathDestinationDir);
}

void SG_pendingtree__move(
	SG_context* pCtx,
	SG_pendingtree* pPendingTree,
	const SG_pathname** pArrayPathsToMove,
	SG_uint32 nCountMoves,
	const SG_pathname* pPathMoveTo
	)
{
	SG_bool bAllowDestinationDirInDrawer = SG_FALSE;	// only MERGE gets to do this for parked stuff
	SG_uint32 nPathIndex = 0;

	for (nPathIndex = 0; nPathIndex < nCountMoves; nPathIndex++)
	{
		SG_ERR_CHECK_RETURN(  SG_pendingtree__move_dont_save_pendingtree(pCtx,
																		 pPendingTree,
																		 pArrayPathsToMove[nPathIndex],
																		 pPathMoveTo,
																		 bAllowDestinationDirInDrawer)  );
	}
	SG_ERR_CHECK_RETURN(  SG_pendingtree__save(pCtx, pPendingTree)  );
}

//////////////////////////////////////////////////////////////////

/**
 * We want to rename "src" ==> "dest", but we have been told that
 * both paths already exist on disk and aren't sure if:
 * [1] they are ***actually*** 2 different items on disk;
 * [2] they are the ***same*** item because of case-folding.
 *
 * The first case should fail.
 *
 * The second we want to succeed using a temp pathname.
 *
 */
static void _rename_try_2step(SG_context * pCtx,
							  const SG_pathname * pPathSrc,
							  const SG_pathname * pPathDest)
{
	SG_pathname * pPathTemp = NULL;
	SG_bool bExists_dest;

	// start the 2-step move (src --> temp)
	SG_ERR_CHECK(  SG_workingdir__calc_backup_name(pCtx, pPathSrc, &pPathTemp)  );
	SG_ERR_CHECK(  SG_fsobj__move__pathname_pathname(pCtx, pPathSrc, pPathTemp)  );

	SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pPathDest, &bExists_dest, NULL, NULL)  );
	if (bExists_dest)
	{
		// dest still exists, so we must have [1].
		// undo the move of src --> temp
		// and throw

		SG_ERR_CHECK(  SG_fsobj__move__pathname_pathname(pCtx, pPathTemp, pPathSrc)  );

		// TODO 2011/07/05 We just found something on disk that isn't
		// TODO            under version control.  Should we have a
		// TODO            different error code here?  This error code
		// TODO            (was intended to) imply an attempt to create
		// TODO            a duplicate item under version control.

		SG_ERR_THROW2(  SG_ERR_PENDINGTREE_OBJECT_ALREADY_EXISTS,
						(pCtx, "%s", SG_pathname__sz(pPathDest))  );
	}
	else
	{
		SG_ERR_CHECK(  SG_fsobj__move__pathname_pathname(pCtx, pPathTemp, pPathDest)  );
#if TRACE_RENAME
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "RENAME used 2-step to rename: '%s' --> '%s'\n",
								   SG_pathname__sz(pPathSrc), SG_pathname__sz(pPathDest))  );
#endif
	}

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathTemp);
}

void SG_pendingtree__rename_dont_save_pendingtree(
	SG_context* pCtx,
	SG_pendingtree* pPendingTree,
	const SG_pathname* pPathLocal,
	const char* pszNewName,
	SG_bool bIgnoreErrorNotFoundOnDisk
	)
{
	struct sg_ptnode* ptnDir = NULL;
	SG_string* pstrName = NULL;
	struct sg_ptnode* ptnRenameMe = NULL;
	SG_pathname* pPathDir = NULL;
	SG_pathname* pPathRenamed = NULL;
	SG_portability_dir * pPort = NULL;
	SG_portability_flags portFlags_destination;
	SG_bool bExists_source;
	SG_bool bExists_destination;
	SG_bool bIsDuplicate;
	SG_bool bCollidesWithSelf;

	// Cut the source pathname into parts and try to find the
	// parent directory and entry ptnodes.

	SG_ERR_CHECK(  SG_pathname__get_last(pCtx, pPathLocal, &pstrName)  );
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__COPY(pCtx, &pPathDir, pPathLocal)  );
	SG_ERR_CHECK(  SG_pathname__remove_last(pCtx, pPathDir)  );
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx, &pPathRenamed, pPathDir, pszNewName)  );

	SG_ERR_CHECK(  sg_pendingtree__local_path_to_ptnode(pCtx, pPendingTree, pPathDir, SG_TRUE, &ptnDir)  );
	SG_ERR_CHECK(  sg_ptnode__find_entry_by_name_not_recursive__active(pCtx, ptnDir, SG_string__sz(pstrName), &ptnRenameMe)  );
    if (!ptnRenameMe)
    {
        SG_ERR_THROW2(  SG_ERR_ITEM_NOT_UNDER_VERSION_CONTROL,
						(pCtx, "%s", SG_pathname__sz(pPathLocal))  );
    }

	// Use the portability code to check first the new entryname
	// by itself (for non-portable names) and then to see if it
	// will actually or potentially collide with something already
	// under version control.

	SG_ERR_CHECK(  SG_portability_dir__alloc(pCtx,
											 pPendingTree->portMask,
											 NULL,
											 pPendingTree->pConverter,
											 pPendingTree->pvaWarnings,
											 &pPort)  );
	SG_ERR_CHECK(  SG_portability_dir__add_item(pCtx, pPort, NULL, pszNewName, SG_TRUE, SG_TRUE, &bIsDuplicate)  );
	SG_ASSERT(  (bIsDuplicate == SG_FALSE)  );
	SG_ERR_CHECK(  SG_portability_dir__get_item_result_flags(pCtx, pPort, pszNewName,
															 NULL, &portFlags_destination)  );
	if (portFlags_destination & SG_PORT_FLAGS__MASK_INDIV)
	{
		if (portFlags_destination & SG_PORT_FLAGS__INDIV__INVALID_00_7F)
		{
			if (strchr(pszNewName, '/')  != NULL
				|| strchr(pszNewName, '\\')  != NULL)
			{
				// the destination of a RENAME must be an ENTRYNAME only (we do
				// not support MOVE+RENAME).
				// 
				// special case this one because we want a nicer error message
				// for a common error.
				//
				// we leave the data in pPendingTree->pvaWarningLog because it
				// doesn't matter since we're throwing a different error.

				SG_ERR_THROW2(  SG_ERR_INVALIDARG,
								(pCtx, "The destination of a RENAME must be a simple entryname, not a pathname: '%s'", pszNewName)  );
			}
		}

		// since a RENAME creates a *new* entryname (for an existing item), the
		// new entryname must be qualified just like we do in ADD.  for example,
		// if they say something like "vv rename FOO COM1" it'll cause problems
		// for Windows users.
		//
		// so for any __INDIV__ issue, just throw a portability warning
		// and let the outer caller fish an error message from the pvaWarningLog.

		SG_ERR_THROW2(  SG_ERR_PORTABILITY_WARNINGS,
						(pCtx, "The destination entryname is not portable: '%s'", pszNewName)  );
	}
	
	// we accept the new entryname as not (potentially) causing problems
	// all by itself.
	//
	// see if the new entryname actually or potentially collides with the
	// original entryname for this item.  we do this step before looking at
	// the full directory so that we can catch things like "vv rename FOO foo"
	// and silently let it work if there aren't any other issues.  (On a MAC
	// or Windows system, this causes us problems because both exists() tests
	// will succeed but are actually looking a the same single item and a
	// naive rename() would throw an already exists error.

	SG_ERR_CHECK(  SG_portability_dir__add_item(pCtx, pPort, NULL, SG_string__sz(pstrName),
											  SG_FALSE, SG_TRUE, &bIsDuplicate)  );
	if (bIsDuplicate)
	{
		SG_ERR_THROW2(  SG_ERR_INVALIDARG,
					   (pCtx, "The new entryname '%s' is exactly the same as the current name: '%s'",
					   pszNewName, SG_pathname__sz(pPathLocal))  );
	}
	SG_ERR_CHECK(  SG_portability_dir__get_item_result_flags(pCtx, pPort, pszNewName,
															 NULL, &portFlags_destination)  );
	bCollidesWithSelf = ((portFlags_destination & SG_PORT_FLAGS__MASK_COLLISION) != 0);
	SG_PORTABILITY_DIR_NULLFREE(pCtx, pPort);

	// Update the PTNODE with the new name and see if it will collide
	// with OTHER items in the directory.

	SG_ERR_CHECK(  sg_ptnode__alter_name(pCtx, pPendingTree, ptnRenameMe, pszNewName)  );
	sg_ptnode__immediate_portability_check(pCtx, ptnDir);
	if (SG_context__has_err(pCtx))
	{
		if (SG_context__err_equals(pCtx, SG_ERR_RBTREE_DUPLICATEKEY))
		{
			// we get a _RBTREE_DUPLICATEKEY error when the entryname is an
			// *exact* match for an existing entry in the directory. 

			SG_context__err_reset(pCtx);
			SG_ERR_THROW2(  SG_ERR_PENDINGTREE_OBJECT_ALREADY_EXISTS,
							(pCtx, "%s", SG_pathname__sz(pPathRenamed))  );
		}
		else if (SG_context__err_equals(pCtx, SG_ERR_PORTABILITY_WARNINGS))
		{
			SG_ERR_RETHROW2(  (pCtx, "'%s' potentially collides with another entry in '%s'.",
							   pszNewName, SG_pathname__sz(pPathDir))  );
		}
		else
		{
			SG_ERR_RETHROW;
		}
	}

	// we now know that there are no POTENTIAL FINAL COLLISIONS with OTHER
	// items.  but we may or may not have a TRANSIENT COLLISION with the
	// original entryname of the item.

	// see if the source exists on disk.  in theory, the user is supposed to
	// use "vv rename FOO BAR" to do a rename, but sometimes they forget and
	// do a plain "/bin/mv FOO BAR" and then when they see "vv status" report
	// LOST and FOUND items, the try to do the "vv rename" (after the fact).
	// we want to allow this to succeed if we can (and not be pedantic and
	// make them "mv" it back and let us "vv rename" it).
	//
	// note that the system can lie to us.  if i say "vv rename FOO foo", the
	// existance tests on both names may succeed on Mac/Windows and we'll
	// happily continue.

	SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pPathLocal,   &bExists_source,      NULL, NULL)  );
	SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pPathRenamed, &bExists_destination, NULL, NULL)  );
	if (!bExists_source)
	{
		if (bIgnoreErrorNotFoundOnDisk)
		{
			if (bExists_destination)
			{
#if TRACE_RENAME
				SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "Allow/Assume after-the-fact RENAME: '%s' --> '%s'\n",
										   SG_string__sz(pstrName), pszNewName)  );
#endif
				goto done;
			}
		}
		
		SG_ERR_THROW2(  SG_ERR_NOT_FOUND,
						(pCtx, "%s", SG_pathname__sz(pPathLocal))  );
    }
	else
	{
		if (bExists_destination)
		{
			// when the destination pathname exists, it can either mean that:
			// [1] we are on an ambiguous filesystem (such as Mac/Windows) and
			//     were told that both "FOO" and "foo" exist in "vv rename FOO foo";
			// [2] that "BAR" really does exist when given "vv rename FOO BAR" and
			//     that "BAR" is not under version control;
			// [3] that "foo" really does exist independently of "FOO" on Linux
			//     and "foo" is not under version control.

			if (bCollidesWithSelf)     // [1] or [3]
			{
				SG_ERR_CHECK(  _rename_try_2step(pCtx, pPathLocal, pPathRenamed)  );
				goto done;
			}
			else
			{
				// TODO 2011/07/05 We just found something on disk that isn't
				// TODO            under version control.  Should we have a
				// TODO            different error code here?  This error code
				// TODO            (was intended to) imply an attempt to create
				// TODO            a duplicate item under version control.
				SG_ERR_THROW2(  SG_ERR_PENDINGTREE_OBJECT_ALREADY_EXISTS,
								(pCtx, "%s", SG_pathname__sz(pPathRenamed))  );
			}
		}
		else
		{
			// finally, the normal, non-screw case....
			SG_ERR_CHECK(  SG_fsobj__move__pathname_pathname(pCtx, pPathLocal, pPathRenamed)  );
		}
	}

done:
	;
fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathDir);
	SG_PATHNAME_NULLFREE(pCtx, pPathRenamed);
	SG_STRING_NULLFREE(pCtx, pstrName);
	SG_PORTABILITY_DIR_NULLFREE(pCtx, pPort);
}

void SG_pendingtree__rename(
	SG_context* pCtx,
	SG_pendingtree* pPendingTree,
	const SG_pathname* pPathLocal,
	const char* pszNewName,
	SG_bool bIgnoreErrorNotFoundOnDisk
	)
{
	SG_ERR_CHECK_RETURN(  SG_pendingtree__rename_dont_save_pendingtree(pCtx, pPendingTree,
																	   pPathLocal, pszNewName,
																	   bIgnoreErrorNotFoundOnDisk)  );
	SG_ERR_CHECK_RETURN(  SG_pendingtree__save(pCtx, pPendingTree)  );
}

static void sg_ptnode__create_entry_for_commit(
	SG_context* pCtx,
	struct sg_ptnode* ptnSub
	)
{
	SG_treenode_entry* pEntry = NULL;

	if (!(ptnSub->temp_flags & sg_PTNODE_TEMP_FLAG_COMMITTING))
	{
		/* The original entry may not be present.  This can happen if
		 * the item is newly added but has been specifically excluded
		 * for this commit operation. */
		if (ptnSub->pBaselineEntry)
		{
			SG_ERR_CHECK(  SG_TREENODE_ENTRY__ALLOC__COPY(pCtx, &pEntry, ptnSub->pBaselineEntry)  );
		}
		else
		{
			/* this item is not supposed to be committed, and it
			 * has no original entry, so it is newly added but
			 * should be ignored now. */
		}
	}
	else
	{
		if (
			(ptnSub->saved_flags & sg_PTNODE_SAVED_FLAG_DELETED)
			|| (ptnSub->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY)
			)
		{
			/* ignore this then */
		}
		else
		{
			const char* pszName = NULL;
			const char* pszHid = NULL;
			const char* pszAttributesHid = NULL;

			SG_ERR_CHECK(  sg_ptnode__get_name(pCtx, ptnSub, &pszName)  );
			SG_ERR_CHECK(  sg_ptnode__get_hid(pCtx, ptnSub, &pszHid)  );
			SG_ERR_CHECK(  sg_ptnode__get_xattrs_hid(pCtx, ptnSub, &pszAttributesHid)  );

			if (!pszHid || !*pszHid)
				SG_ERR_THROW2(  SG_ERR_TREENODE_ENTRY_VALIDATION_FAILED,
								(pCtx, "HID not set when creating TNE for %s '%s'",
								 ptnSub->pszgid, ((pszName) ? pszName : "(null)"))  );

			SG_ERR_CHECK(  SG_TREENODE_ENTRY__ALLOC(pCtx, &pEntry)  );
			SG_ERR_CHECK(  SG_treenode_entry__set_entry_name(pCtx, pEntry, pszName)  );
			SG_ERR_CHECK(  SG_treenode_entry__set_entry_type(pCtx, pEntry, ptnSub->type)  );
			SG_ERR_CHECK(  SG_treenode_entry__set_hid_blob(pCtx, pEntry, pszHid)  );
			SG_ERR_CHECK(  SG_treenode_entry__set_hid_xattrs(pCtx, pEntry, pszAttributesHid)  );
			SG_ERR_CHECK(  SG_treenode_entry__set_attribute_bits(pCtx, pEntry, ptnSub->iCurrentAttributeBits)  );
		}
	}

	ptnSub->pNewEntry = pEntry;

	return;

fail:
	SG_TREENODE_ENTRY_NULLFREE(pCtx, pEntry);
}

static void sg_ptnode__commit(
	SG_context* pCtx,
	struct sg_ptnode* ptn,
	const SG_string* pRepoPath,
	SG_committing* ptx,
	SG_bool* pbParentDirty,
	const char * psz_comment_for_nested_submodule_commit,
    SG_vhash* pvh_locks
	)
{
	SG_rbtree_iterator* pIterator = NULL;
	SG_string* pMyRepoPath = NULL;
	const char* pszgid = NULL;
	const char* pszName = NULL;
	SG_bool b;
	SG_treenode* pTreenode = NULL;
	SG_file* pFile = NULL;
	char* pszidHid = NULL;
	SG_pathname* pPathLocal = NULL;
	SG_treenode_entry* pEntry = NULL;
	SG_string* pstrLink = NULL;
	SG_timestamp_cache * pTSC = NULL;		// we do not own this
	SG_int64 tim = -1;
    SG_string* pstr_repo_path = NULL;
    SG_vhash* pvh_user = NULL;

	SG_NULLARGCHECK_RETURN(ptn);
	SG_NULLARGCHECK_RETURN(ptx);

#if TRACE_COMMIT
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "ptnode__commit: entering %s\n",
							   ((pRepoPath) ? SG_string__sz(pRepoPath) : "(null)"))  );
#endif

	if (
		!(ptn->temp_flags & sg_PTNODE_TEMP_FLAG_COMMITTING)
		)
	{
		/* We do NOT clear *pbParentDirty here.  We simply don't set it. */
		goto done;
	}

	if (ptn->bDeletedAfterAdded)
	{
		// This item was ADDED (by MERGE) and then DELETED (by the user).
		// In theory, we do not change the dirtyness of the parent directory
		// (it wasn't in it in the baseline and it isn't in it now).
		// BUT I think I want to force this (in case the parent was also
		// just added and has never been committed without this item).
		*pbParentDirty = SG_TRUE;
		goto done;
	}

	if (
		(ptn->saved_flags & sg_PTNODE_SAVED_FLAG_DELETED)
		|| (ptn->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY)
		)
	{
		/* ignore this then */
		*pbParentDirty = SG_TRUE;
		goto done;
	}

    if (pRepoPath)
    {
        SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path(pCtx,
																			 ptn->pPendingTree->pPathWorkingDirectoryTop,
																			 SG_string__sz(pRepoPath),
																			 &pPathLocal)  );
#ifdef SG_BUILD_FLAG_FEATURE_XATTR
        SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pPathLocal, &bFoundOnDisk, NULL, NULL)  );
        if (bFoundOnDisk)
        	SG_ERR_CHECK(  sg_ptnode__commit_xattrs(pCtx, ptn, pPathLocal, ptx)  );
#endif
    }

	if (SG_TREENODEENTRY_TYPE_DIRECTORY == ptn->type)
	{
		SG_bool bNeedToWriteMyTreenode = SG_FALSE;

		SG_ERR_CHECK(  SG_treenode__alloc(pCtx, &pTreenode)  );
		SG_ERR_CHECK(  SG_treenode__set_version(pCtx, pTreenode, SG_TN_VERSION__CURRENT)  );

		if (ptn->prbItems)
		{
			struct sg_ptnode* ptnSub = NULL;

			SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pIterator, ptn->prbItems, &b, &pszgid, (void**) &ptnSub)  );

			while (b)
			{
				SG_ERR_CHECK(  sg_ptnode__get_name(pCtx, ptnSub, &pszName)  );

				if (pRepoPath)
				{
					SG_ERR_CHECK(  SG_repopath__combine__sz(pCtx, SG_string__sz(pRepoPath), pszName, (SG_TREENODEENTRY_TYPE_DIRECTORY == ptnSub->type), &pMyRepoPath)  );
				}
				else
				{
					SG_ASSERT('@' == pszName[0]);
					SG_ASSERT(0 == pszName[1]);

					SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &pMyRepoPath, "@/")  );
				}

				SG_ERR_CHECK(  sg_ptnode__commit(pCtx, ptnSub, pMyRepoPath, ptx, &bNeedToWriteMyTreenode, psz_comment_for_nested_submodule_commit, pvh_locks)  );

				SG_ERR_CHECK(  sg_ptnode__create_entry_for_commit(pCtx, ptnSub)  );

				if (ptnSub->pNewEntry)
				{
					SG_ERR_CHECK(  SG_TREENODE_ENTRY__ALLOC__COPY(pCtx, &pEntry, ptnSub->pNewEntry)  );
					SG_ERR_CHECK(  SG_treenode__add_entry(pCtx, pTreenode, ptnSub->pszgid, &pEntry)  );
				}

				SG_STRING_NULLFREE(pCtx, pMyRepoPath);

				SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pIterator, &b, &pszgid, (void**) &ptnSub)  );
			}
			SG_RBTREE_ITERATOR_NULLFREE(pCtx, pIterator);
		}

		if (
			(!ptn->pBaselineEntry)
			&& (ptn->pCurrentParent)
			)
		{
			bNeedToWriteMyTreenode = SG_TRUE;
		}

		if (bNeedToWriteMyTreenode)
		{

            SG_ERR_CHECK(  SG_time__get_milliseconds_since_1970_utc(pCtx, &tim)  );
			SG_ERR_CHECK(  SG_committing__tree__add_treenode(pCtx, ptx, &pTreenode, ptn->pszgid, tim, &pszidHid)  );
		}

		SG_TREENODE_NULLFREE(pCtx, pTreenode);
	}
	else if (SG_TREENODEENTRY_TYPE_REGULAR_FILE == ptn->type)
	{
		SG_fsobj_stat stat;
		SG_bool bFileIsDirty = SG_FALSE;

		SG_ERR_CHECK(  sg_ptnode__is_file_dirty(pCtx, ptn, &bFileIsDirty)  );

		if (bFileIsDirty)
		{
            if (pvh_locks)
            {
                const char* psz_lock_userid = NULL;

                SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pvh_locks, ptn->pszgid, &psz_lock_userid)  );
                if (psz_lock_userid)
                {
                    const char* psz_username = NULL;

                    SG_ERR_CHECK(  SG_user__lookup_by_userid(pCtx, ptn->pPendingTree->pRepo, psz_lock_userid, &pvh_user)  );
                    SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvh_user, "name", &psz_username)  );
                    SG_ERR_CHECK(  sg_ptnode__get_repo_path(pCtx, ptn, &pstr_repo_path)  );
                    SG_ERR_THROW2(  SG_ERR_VC_LOCK_VIOLATION,
                        (pCtx, "%s is locked by %s", SG_string__sz(pstr_repo_path), psz_username)
                        );
                }
            }

            SG_ERR_CHECK(  SG_time__get_milliseconds_since_1970_utc(pCtx, &tim)  );

			SG_ERR_CHECK(  SG_file__open__pathname(pCtx, pPathLocal, SG_FILE_LOCK | SG_FILE_RDONLY | SG_FILE_OPEN_EXISTING, SG_FSOBJ_PERMS__UNUSED, &pFile)  );

            /* TODO fix the b_dont_bother (SG_BLOBENCODING__*) flag below according to the filename of this item */
			SG_ERR_CHECK(  SG_committing__add_file(pCtx, 
                        ptx, 
                        ptn->pszgid, 
                        tim,
                        SG_FALSE, 
                        pFile, 
                        &pszidHid
                        )  );

			SG_FILE_NULLCLOSE(pCtx, pFile);

			*pbParentDirty = SG_TRUE;

			SG_ERR_CHECK(  SG_fsobj__stat__pathname(pCtx, pPathLocal, &stat)  );
			SG_ERR_CHECK(  SG_pendingtree__get_timestamp_cache(pCtx, ptn->pPendingTree, &pTSC)  );
			SG_ERR_CHECK(  SG_timestamp_cache__add(pCtx, pTSC, ptn->pszgid, stat.mtime_ms, stat.size, pszidHid)  );
		}
		else
		{
//			// TODO 2010/08/10 This may not be necessary.  If the file isn't dirty, what are we doing here.
//			// TODO            And we probably don't need to update the timestamp_cache for it.
//
//			const char * pszHid_baseline;
//
//			SG_ERR_CHECK(  sg_ptnode__get_hid(pCtx, ptn, &pszHid_baseline)  );
//
//			SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pPathLocal, &bFoundOnDisk, NULL, NULL)  );
//			SG_ASSERT(  (bFoundOnDisk)  );
//
//			SG_ERR_CHECK(  SG_fsobj__stat__pathname(pCtx, pPathLocal, &stat)  );
//
//			SG_ERR_CHECK(  SG_pendingtree__get_timestamp_cache(pCtx, ptn->pPendingTree, &pTSC)  );
//			SG_ERR_CHECK(  SG_timestamp_cache__add(pCtx, pTSC, ptn->pszgid, stat.mtime_ms, pszHid_baseline)  );
		}
	}
	else if (SG_TREENODEENTRY_TYPE_SYMLINK == ptn->type)
	{
		SG_bool bIsLinkDirty = SG_FALSE;

        /* TODO the logic here differs from the logic for a file, above.
         * should it?  for a file, we trust the hid which was calculated during
         * the last scan.  for a symlink, we just read it and check the hid
         * right now, since it's always going to be so short anyway. */

		SG_ERR_CHECK(  SG_fsobj__readlink(pCtx, pPathLocal, &pstrLink)  );
		SG_ERR_CHECK(  SG_repo__alloc_compute_hash__from_string(pCtx, ptn->pPendingTree->pRepo, pstrLink, &pszidHid)  );

		if (!ptn->pBaselineEntry)
		{
			bIsLinkDirty = SG_TRUE;
		}
		else
		{
			const char* pszBaselineHid = NULL;

			SG_ERR_CHECK(  SG_treenode_entry__get_hid_blob(pCtx, ptn->pBaselineEntry, &pszBaselineHid)  );

			if (0 != strcmp(pszBaselineHid, pszidHid))
			{
				bIsLinkDirty = SG_TRUE;
			}
		}

		SG_NULLFREE(pCtx, pszidHid);

		if (bIsLinkDirty)
		{
			SG_ERR_CHECK(  SG_time__get_milliseconds_since_1970_utc(pCtx, &tim)  );
			SG_ERR_CHECK(  SG_committing__add_bytes__string(
                        pCtx, 
                        ptx, 
                        NULL,
                        tim,
                        pstrLink, 
                        SG_TRUE, // ALWAYSFULL.  symlinks are too small to bother with compression
                        &pszidHid
                        )  );
			*pbParentDirty = SG_TRUE;
		}
		SG_STRING_NULLFREE(pCtx, pstrLink);
	}
	else if (SG_TREENODEENTRY_TYPE_SUBMODULE == ptn->type)
	{
		const char * pszHid = NULL;
		SG_bool bDirty_Reference, bDirty_Pending;
		SG_bool bIsWD = SG_FALSE;

		if (!ptn->pTSM)		// this should only be necessary if scan-dir or __deal_with_possibly_dirty_submodule() didn't run.  we may not need it.
		{
			// warning: __sniff_directory() allocates a new TSM and populates it with data
			// warning: from the live WD directory.  BUT IT DOES NOT update the current HID
			// warning: in the ptnode.  we have to do that when we want it (because the
			// warning: underlying submodule can change just like the contents of a file can
			// warning: change, we only set the current hid in the ptnode when we need it).
			// warning: and we have to set it before we ask __is_submodule_changed_or_dirty()
			// warning: to compute bDirty_Reference (because it just compares the 2 HIDs).
			SG_ERR_CHECK(  SG_TREENODE_SUBMODULE__SNIFF_DIRECTORY(pCtx, pPathLocal, ptn->pPendingTree->pRepo,
																  &bIsWD, &ptn->pTSM)  );
			if (!bIsWD)
				SG_ERR_THROW2(  SG_ERR_NOTIMPLEMENTED,
								(pCtx, "TODO path '%s' was a submodule but does not appear to be one now.",
								 SG_pathname__sz(pPathLocal))  );
			SG_ERR_CHECK(  SG_treenode_submodule__get_id_ref(pCtx, ptn->pTSM, &pszHid)  );
			SG_ERR_CHECK(  SG_strpool__add__sz(pCtx, ptn->pPendingTree->pStrPool, pszHid, &ptn->pszCurrentHid)  );
		}

		SG_ERR_CHECK(  sg_ptnode__is_submodule_changed_or_dirty(pCtx, ptn, &bDirty_Reference, &bDirty_Pending)  );
		if (bDirty_Pending)
		{
			if (ptn->temp_flags & sg_PTNODE_TEMP_FLAG_NESTED_SUBMODULE_COMMIT)
			{
				// dive into the dirty submodule and do a commit.

				SG_ERR_CHECK(  SG_treenode_submodule__commit(pCtx, ptn->pTSM, psz_comment_for_nested_submodule_commit, NULL)  );
				
				// after the nested commit we delete the ptn->pTSM and re-scan the directory
				// to get fresh values for current-hid and the various flags.

				SG_TREENODE_SUBMODULE_NULLFREE(pCtx, ptn->pTSM);
				SG_ERR_CHECK(  SG_TREENODE_SUBMODULE__SNIFF_DIRECTORY(pCtx, pPathLocal, ptn->pPendingTree->pRepo,
																	  &bIsWD, &ptn->pTSM)  );
				if (!bIsWD)
					SG_ERR_THROW2(  SG_ERR_NOTIMPLEMENTED,
									(pCtx, "TODO path '%s' was a submodule but does not appear to be one now.",
									 SG_pathname__sz(pPathLocal))  );
				SG_ERR_CHECK(  SG_treenode_submodule__get_id_ref(pCtx, ptn->pTSM, &pszHid)  );
				SG_ERR_CHECK(  SG_strpool__add__sz(pCtx, ptn->pPendingTree->pStrPool, pszHid, &ptn->pszCurrentHid)  );
				SG_ERR_CHECK(  sg_ptnode__is_submodule_changed_or_dirty(pCtx, ptn, &bDirty_Reference, &bDirty_Pending)  );
			}
			else
			{
				SG_ERR_CHECK(  SG_treenode_submodule__emit_skip_commit_message(pCtx, ptn->pTSM)  );
			}
		}

		if (bDirty_Reference)
		{
			SG_ERR_CHECK(  SG_time__get_milliseconds_since_1970_utc(pCtx, &tim)  );
			SG_ERR_CHECK(  SG_committing__tree__add_treenode_submodule(pCtx, ptx, ptn->pTSM, ptn->pszgid, tim, &pszidHid)  );

			*pbParentDirty = SG_TRUE;
		}
	}
	else
	{
		SG_ERR_THROW( SG_ERR_INVALID_PENDINGTREE_OBJECT_TYPE );
	}

	if (pszidHid)
	{
#if TRACE_COMMIT
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "ptnode__commit: computed hid %s\n",
								   pszidHid)  );
#endif
		/* TODO is pszidHid really different? If not, we could just
		 * leave pszCurrentHid null and not set *pbParentDirty */
		SG_ERR_CHECK(  SG_strpool__add__sz(pCtx, ptn->pPendingTree->pStrPool, pszidHid, &ptn->pszCurrentHid)  );
		SG_NULLFREE(pCtx, pszidHid);
		*pbParentDirty = SG_TRUE;
	}

#if 1
	/* TODO the following extra check feels hack-ish, like we're
	 * trying to catch cases we missed above.  Shouldn't we have
	 * figured this out by now? */
	if (!*pbParentDirty)
	{
		SG_ERR_CHECK(  sg_ptnode__check_parent_dirty_non_recursive(pCtx, ptn, pbParentDirty)  );
	}
#endif

	SG_PATHNAME_NULLFREE(pCtx, pPathLocal);

done:
#if TRACE_COMMIT
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "ptnode__commit: leaving  %s\n",
							   ((pRepoPath) ? SG_string__sz(pRepoPath) : "(null)"))  );
#endif
	return;

fail:
	SG_VHASH_NULLFREE(pCtx, pvh_user);
	SG_STRING_NULLFREE(pCtx, pstr_repo_path);
	SG_FILE_NULLCLOSE(pCtx, pFile);
	SG_STRING_NULLFREE(pCtx, pstrLink);
	SG_NULLFREE(pCtx, pszidHid);
	SG_PATHNAME_NULLFREE(pCtx, pPathLocal);
	SG_STRING_NULLFREE(pCtx, pMyRepoPath);
	SG_RBTREE_ITERATOR_NULLFREE(pCtx, pIterator);
    SG_TREENODE_ENTRY_NULLFREE(pCtx, pEntry);
	SG_TREENODE_NULLFREE(pCtx, pTreenode);
}

static void STUPID_HACK_TODO_sg_ptnode__find_by_gid(
	SG_context* pCtx, 
	struct sg_ptnode* ptn,
	const char* pszGid,
	SG_uint32 depth,
	struct sg_ptnode** ppResult
	)
{
	SG_rbtree_iterator* pIterator = NULL;
	struct sg_ptnode* ptnSub = NULL;

#if TRACE_PENDINGTREE
	SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,"%*c looking at:  [gid %s][saved %d][temp %d][prb %d]\n",
							   depth*4,' ',
							   ((ptn->pszgid) ? ptn->pszgid : "(null)"),
							   ptn->saved_flags,
							   ptn->temp_flags,
							   (ptn->prbItems != NULL))  );
#endif

	/* to ensure that this function is unambiguous, we should probably
	 * not return results which are MOVED_AWAY.  That happens to be
	 * one case where a gid appears twice in a pending tree. */

	if (
		!(ptn->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY)
		&& (ptn->pszgid)
		&& (0 == strcmp(pszGid, ptn->pszgid))
		)
	{
		*ppResult = ptn;
		return;
	}

	if (ptn->prbItems)
	{
        SG_bool b = SG_FALSE;

		SG_ERR_CHECK(  SG_rbtree__find(pCtx, ptn->prbItems, pszGid, &b, (void**) &ptnSub)  );

		if (b && (!(ptnSub->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY)) )
        {
            *ppResult = ptnSub;
            return;
        }
        else
		{
			const char* pszKey = NULL;

			/* Recurse downward into every entry */
			SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pIterator, ptn->prbItems, &b, &pszKey, (void**) &ptnSub)  );
			while (b)
			{
				SG_ERR_CHECK(  STUPID_HACK_TODO_sg_ptnode__find_by_gid(pCtx, ptnSub, pszGid, depth+1, ppResult)  );

				if (*ppResult)
				{
					break;
				}

				SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pIterator, &b, &pszKey, (void**) &ptnSub)  );
			}
			SG_RBTREE_ITERATOR_NULLFREE(pCtx, pIterator);
		}
	}

	return;

fail:
	SG_RBTREE_ITERATOR_NULLFREE(pCtx, pIterator);
}

static void sg_pendingtree__find_by_gid(
	SG_context* pCtx, 
	SG_pendingtree* pPendingTree,
	const char* pszGid,
	struct sg_ptnode** ppResult
	)
{
	struct sg_ptnode* ptn = NULL;

#if TRACE_PENDINGTREE
	SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
							   "sg_pendingtree__find_by_gid: looking for [gid %s]\n",
							   pszGid)  );
#endif

	SG_ERR_CHECK(  STUPID_HACK_TODO_sg_ptnode__find_by_gid(pCtx, pPendingTree->pSuperRoot, pszGid, 1, &ptn)  );

	if (!ptn)
	{
#if TRACE_PENDINGTREE || defined(DEBUG)
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "PendingTreeFindByGid: About to throw NOT_FOUND for '%s'\n", pszGid)  );
		SG_ERR_IGNORE(  SG_pendingtree_debug__dump_existing_to_console(pCtx, pPendingTree)  );
#endif
		SG_ERR_THROW2(  SG_ERR_NOT_FOUND,
						(pCtx,"PendingTreeFindByGid '%s'",pszGid)  );
	}

#if TRACE_PENDINGTREE
	SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
							   "sg_pendingtree__find_by_gid: found ptn [parent %s][child %s][child saved %d][child temp %d]\n",
							   ((ptn->pCurrentParent->pszgid) ? ptn->pCurrentParent->pszgid : "(null)"),
							   ptn->pszgid,ptn->saved_flags,ptn->temp_flags)  );
#endif

	*ppResult = ptn;

fail:
	return;
}

void SG_pendingtree__find_repo_path_by_gid(SG_context * pCtx,
										   SG_pendingtree * pPendingTree,
										   const char * pszGid,
										   SG_string ** ppStrRepoPath)
{
	struct sg_ptnode * ptn = NULL;

	SG_ERR_CHECK_RETURN(  sg_pendingtree__find_by_gid(pCtx, pPendingTree, pszGid, &ptn)  );
	SG_ERR_CHECK_RETURN(  sg_ptnode__get_repo_path(pCtx, ptn, ppStrRepoPath)  );
}

/**
 * This function finds places where we are committing the destination
 * end of a move operation before the source end.  This is not
 * allowed, because the resulting tree will have the item appearing in
 * both places.
 *
 * We do, however, allow committing the source end without the
 * destination end.  The result is that the moved item is not in the
 * tree at all.  Not sure why anybody would want to do this, but it
 * doesn't violate a fundamental law, so we'll allow it.
 *
 * A corollary is that we allow committing the destination end without
 * the source end /if/ the source end was already committed in some
 * previous operation.
 *
 * TODO think about the revert operations and other stuff to make sure
 * this is all okay.  If any problems come up, we'll just require that
 * both ends always get committed at the same time.  But that will
 * require more bookkeeping in the ptnode structure.
 *
 */
static void sg_ptnode__fail_if_any_split_moves(
	SG_context* pCtx, 
	struct sg_ptnode* ptn,
	SG_uint16 flag
	)
{
	SG_rbtree_iterator* pIterator = NULL;
	struct sg_ptnode* ptnOriginalParent = NULL;

	if (!(ptn->temp_flags & flag))
	{
		return;
	}

	if (ptn->pszgidMovedFrom)
	{
		SG_ERR_CHECK(  sg_pendingtree__find_by_gid(pCtx, ptn->pPendingTree, ptn->pszgidMovedFrom, &ptnOriginalParent)  );

		if (
			!(ptnOriginalParent->temp_flags & flag)
			)
		{
			SG_ERR_THROW(  SG_ERR_SPLIT_MOVE  );
		}
	}

	if (ptn->prbItems)
	{
		SG_bool b = SG_FALSE;
		const char* pszKey = NULL;
		struct sg_ptnode* ptnSub = NULL;

		/* Recurse downward into every entry */
		SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pIterator, ptn->prbItems, &b, &pszKey, (void**) &ptnSub)  );
		while (b)
		{
			SG_ERR_CHECK(  sg_ptnode__fail_if_any_split_moves(pCtx, ptnSub, flag)  );

			SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pIterator, &b, &pszKey, (void**) &ptnSub)  );
		}
		SG_RBTREE_ITERATOR_NULLFREE(pCtx, pIterator);
	}

	return;

fail:
	SG_RBTREE_ITERATOR_NULLFREE(pCtx, pIterator);
}

//////////////////////////////////////////////////////////////////

/**
 * Set the COMMITTING flag on this ptnode and propagate to parent directories.
 */
static void _my_set_commit_flag(SG_context * pCtx, struct sg_ptnode * ptn)
{
	struct sg_ptnode * ptn_temp;

#if 0 && defined(DEBUG)
	{
		const char * pszEntryname = NULL;

		SG_ERR_IGNORE(  sg_ptnode__get_name(pCtx, ptn, &pszEntryname)  );
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "my_set_commit_flag: [saved flags %d][temp flags %d] [gid %s] %s\n",
								   ptn->saved_flags,
								   ptn->temp_flags,
								   ptn->pszgid,
								   pszEntryname)  );
	}
#else
	SG_UNUSED(pCtx);
#endif

	ptn->temp_flags |= sg_PTNODE_TEMP_FLAG_COMMITTING;

	// the commit flag must be pushed back up to the parent directory so that
	// the commit of a file causes the parent directory to be commited too (so
	// that the directory HIDs change on the treenodes.
	//
	// TODO 2010/06/07 We may need to split the COMMITING BIT into 2 (one for
	// TODO            actually committing this node and one for there is dirt
	// TODO            within this directory to be committed, so keep diving).
	// TODO            Because if they said "vv commit --include='**.h' ."
	// TODO            they probably just want to commit the *.h changes and
	// TODO            not any moves/renames/xattr/attrbit changes on any of
	// TODO            the parent/ancestor directories that happen to contain
	// TODO            a .h file.  Right?  See sprawl-819.
	//
	// we set the flag on this node.  in theory, if we are in a straight
	// dive from the root, our parent's node should also have its flag set.
	// but if we got started from an ITEM-LIST, our parent may not have been
	// visited.  so we set the bits on our ancestor directories so that when
	// the bubble up happens they will be committed too.

	ptn_temp = ptn->pCurrentParent;
	while (ptn_temp && ((ptn_temp->temp_flags & sg_PTNODE_TEMP_FLAG_COMMITTING) == 0))
	{
		ptn_temp->temp_flags |= sg_PTNODE_TEMP_FLAG_COMMITTING;
		ptn_temp = ptn_temp->pCurrentParent;
	}
}

static void sg_ptnode__maybe_mark_ptnode_for_commit__recursive(SG_context * pCtx,
															   struct sg_ptnode * ptn,
															   SG_bool bRecursive,
															   const SG_file_spec* pFilespec,
															   SG_bool bPartialCommit_2, SG_bool bNoSubmodules);

/**
 * Do the actual dirty work of walking the contents of the directory
 * and marking the contents.
 *
 * WE DO NOT TAKE A bRecursive FLAG because it only applied to the
 * top-level items named in SG_pendingtree__commit() and as discussed in
 * sg_ptnode__maybe_mark_ptnode_for_commit__recursive() we don't want to
 * propagate that.
 */
static void _my_commit_include_dir(SG_context * pCtx,
								   struct sg_ptnode * ptn,
								   SG_string * pStringRepoPath,
								   const SG_file_spec * pFilespec,
								   SG_bool bPartialCommit_2, SG_bool bNoSubmodules)
{
	SG_rbtree_iterator * pIterator = NULL;
	const char * pszKey;
	struct sg_ptnode * ptnSub;
	SG_bool bFound = SG_FALSE;

#if (TRACE_COMMIT == 0)
	SG_UNUSED(pStringRepoPath);
#endif

	SG_ASSERT(  (ptn->type == SG_TREENODEENTRY_TYPE_DIRECTORY)  );

	if (ptn->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY)
	{
		// the phantom moved-from ptnode should not have any children
		// because they are in the node that is in the new location.

		SG_ASSERT(  (ptn->prbItems == NULL)  );
		
#if TRACE_COMMIT
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "MarkPtnodesForCommit: Ignore empty contents of phantom moved-away node for [%s][gid %s][saved %x][temp %x].\n",
								   ((pStringRepoPath) ? SG_string__sz(pStringRepoPath) : "(null)"),
								   ptn->pszgid, ptn->saved_flags, ptn->temp_flags)  );
#endif
		return;
	}

	// if there are no items it could be that the directory is empty *or*
	// that we just didn't load it (like there was nothing dirty in it).
	// this bothers me for some reason, but i think we are doing it right.

	if (!ptn->prbItems)
	{
#if TRACE_COMMIT
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "MarkPtnodesForCommit: TODO determine if the directory is really empty or just not loaded [%s][gid %s][saved %x][temp %x].\n",
								   ((pStringRepoPath) ? SG_string__sz(pStringRepoPath) : "(null)"),
								   ptn->pszgid, ptn->saved_flags, ptn->temp_flags)  );
#endif
		return;
	}

#if TRACE_COMMIT
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "MarkPtnodesForCommit: Considering contents of directory [%s] [gid %s][saved %x][temp %x]\n",
							   ((pStringRepoPath) ? SG_string__sz(pStringRepoPath) : "(null)"),
							   ptn->pszgid, ptn->saved_flags, ptn->temp_flags)  );
#endif

	SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pIterator, ptn->prbItems, &bFound, &pszKey, (void**) &ptnSub)  );
	while (bFound)
	{
#if TRACE_COMMIT
		{
			const char * pszEntryNameSub = NULL;
			if (ptnSub->pszCurrentName)
				pszEntryNameSub = ptnSub->pszCurrentName;
			else if (ptnSub->pBaselineEntry)
				SG_ERR_IGNORE(  SG_treenode_entry__get_entry_name(pCtx, ptnSub->pBaselineEntry, &pszEntryNameSub)  );

			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
									   "MarkPtnodesForCommit: ptnSub [%s][gid %s][saved %x][temp %x]\n",
									   ((*pszEntryNameSub) ? pszEntryNameSub : "(null)"),
									   ptnSub->pszgid, (SG_uint32)ptnSub->saved_flags, (SG_uint32)ptnSub->temp_flags)  );
		}
#endif

		SG_ERR_CHECK(  sg_ptnode__maybe_mark_ptnode_for_commit__recursive(pCtx,
																		  ptnSub,
																		  SG_TRUE,		// do not propagate bRecursive
																		  pFilespec,
																		  bPartialCommit_2, bNoSubmodules)  );

		SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pIterator, &bFound, &pszKey, (void**) &ptnSub)  );
	}

fail:
	SG_RBTREE_ITERATOR_NULLFREE(pCtx, pIterator);
}

/**
 * A directory was EXPLICITLY INCLUDED.  (there were INCLUDES and we matched one of them.)
 *
 * TODO 2010/06/07 So we matched the directory, what does this imply for
 * TODO            the contents of the directory?  I'm going to assume for
 * TODO            the moment that it means that we add the directory and
 * TODO            then (optinally) dive into it WITHOUT any INCLUDES so
 * TODO            that we get the full contents.
 * TODO
 * TODO            That is, we treat:
 * TODO                vv add --include='**.dir' --include='**.c' .
 * TODO
 * TODO            like the combination of 2 commands:
 * TODO                vv add --include='**.c' .
 * TODO                find . -name="*.dir" | xargs vv add
 * TODO
 * TODO            I don't think it makes any sense to treat it like:
 * TODO                vv add --include='**.c' .
 * TODO                find . -name="*.dir" | xargs vv add --include='**.c'
 * TODO
 * TODO            See sprawl-818.
 *
 * We do, however, carry along the EXCLUDES.
 */
static void _my_commit_explicit_include_dir(SG_context * pCtx,
											struct sg_ptnode * ptn,
											SG_string * pStringRepoPath,
											SG_bool bRecursive,
											const SG_file_spec * pFilespec,
											SG_bool bPartialCommit_2, SG_bool bNoSubmodules)
{
	SG_file_spec* pFilespecNoIncludes = NULL;

	if (!bRecursive)
	{
#if TRACE_COMMIT
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "MarkPtnodesForCommit: Skipping dive on explicitly-included directory [%s] because not-recursive\n",
								   ((pStringRepoPath) ? SG_string__sz(pStringRepoPath) : "(null)"))  );
#endif
		return;
	}

	SG_ERR_CHECK(  SG_FILE_SPEC__ALLOC__COPY(pCtx, pFilespec, &pFilespecNoIncludes)  );
	SG_ERR_CHECK(  SG_file_spec__modify_flags(pCtx, pFilespecNoIncludes, SG_FILE_SPEC__NO_INCLUDES, 0u, NULL, NULL)  );
	SG_ERR_CHECK(  _my_commit_include_dir(pCtx, ptn, pStringRepoPath, pFilespecNoIncludes, bPartialCommit_2, bNoSubmodules)  );

fail:
	SG_FILE_SPEC_NULLFREE(pCtx, pFilespecNoIncludes);
}

/**
 * There were no INCLUDES, so we are including everything we can find.
 */
static void _my_commit_implicit_include_dir(SG_context * pCtx,
											struct sg_ptnode * ptn,
											SG_string * pStringRepoPath,
											SG_bool bRecursive,
											const SG_file_spec * pFilespec,
											SG_bool bPartialCommit_2, SG_bool bNoSubmodules)
{
	if (!bRecursive)
	{
#if TRACE_COMMIT
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "MarkPtnodesForCommit: Skipping dive on implicitly-included directory [%s] because not-recursive\n",
								   ((pStringRepoPath) ? SG_string__sz(pStringRepoPath) : "(null)"))  );
#endif
		return;
	}

	SG_ERR_CHECK(  _my_commit_include_dir(pCtx, ptn, pStringRepoPath, pFilespec, bPartialCommit_2, bNoSubmodules)  );

fail:
	return;
}

/**
 * A provisional dir.  We had INCLUDES, but didn't exactly match this directory,
 * allow children to be inspected.
 */
static void _my_commit_provisional_include_dir(SG_context * pCtx,
											   struct sg_ptnode * ptn,
											   SG_string * pStringRepoPath,
											   SG_bool bRecursive,
											   const SG_file_spec * pFilespec,
											   SG_bool bPartialCommit_2, SG_bool bNoSubmodules)
{
	if (!bRecursive)
	{
#if TRACE_COMMIT
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "MarkPtnodesForCommit: Skipping provisional dive on directory [%s] because not-recursive\n",
								   ((pStringRepoPath) ? SG_string__sz(pStringRepoPath) : "(null)"))  );
#endif
		return;
	}

	SG_ERR_CHECK_RETURN(  _my_commit_include_dir(pCtx, ptn, pStringRepoPath, pFilespec, bPartialCommit_2, bNoSubmodules)  );
}

/**
 * When we mark the ptnode of a submodule for commit, that can
 * mean 2 things:
 * [1] dive/cd into the submodule and request that it do
 *     a commit, if it is dirty.
 * [2] whether it is (currently) dirty or not, it still may
 *     refer to a different cset/baseline, so mark the ptnode
 *     for commit to get the new link target.
 *
 * There are times we want to automatically/silently
 * do nested commits.  And there are times that we don't.
 *
 * if bNoSubmodules is set, we do not go into the submodule
 * (think of this like -xdev on find(1)).
 *
 * if the outer is doing a partial commit and the sub is dirty,
 * we throw an error because we don't want the ambiguity.  for
 * example, should the sub get the list of --excludes too or
 * should the sub always do a full commit?
 *
 */
static void _my_commit__deal_with_possibly_dirty_submodule(SG_context * pCtx,
														   struct sg_ptnode * ptn,
														   SG_bool bPartialCommit_2, SG_bool bNoSubmodules)
{
	SG_string * pStringRepoPath = NULL;
	SG_bool bDirty_Pending = SG_FALSE;		// is there uncommitted dirt?

	if (ptn->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY)
	{
		// TODO 2011/01/27 I'm not sure why the caller is trying
		// TODO            to step into the phantom.
#if TRACE_COMMIT
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "DealWithPossiblyDirtySubmodule: given moved away ptnode [gid %s]\n",
								   ptn->pszgid)  );
#endif
		return;
	}
	
	if (bNoSubmodules)
	{
		// do not try to dive into the submodule and force a commit
		// (without regard to whether it is clean/dirty and/or whether
		// this is a full/partial commit).

		return;
	}

	if (ptn->saved_flags & sg_PTNODE_SAVED_FLAG_DELETED)
	{
		// the submodule directory has already been deleted from the WD
		// so we can't sniff it.
		return;
	}

	SG_ERR_CHECK(  sg_ptnode__is_submodule_changed_or_dirty(pCtx, ptn, NULL, &bDirty_Pending)  );
	if (!bDirty_Pending)
	{
		// the submodule is clean.  (it may or may not reference a
		// different baseline, but we don't care about that here;
		// the point is that we don't need to dive into it and force
		// a commit.)

		return;
	}

	if (bPartialCommit_2)
	{
		SG_ERR_CHECK(  sg_ptnode__get_repo_path(pCtx, ptn, &pStringRepoPath)  );

		SG_ERR_THROW2(  SG_ERR_CANNOT_PARTIAL_COMMIT_WITH_DIRTY_SUBMODULE,
						(pCtx, "The submodule '%s' is dirty.", SG_string__sz(pStringRepoPath))  );
	}
	else
	{
		// tell the actual sg_ptnode__commit() code to dive into the submodule
		// and commit it before it deals with the ptnode.

		ptn->temp_flags |= sg_PTNODE_TEMP_FLAG_NESTED_SUBMODULE_COMMIT;

#if TRACE_COMMIT
		SG_ERR_CHECK(  sg_ptnode__get_repo_path(pCtx, ptn, &pStringRepoPath)  );
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "MarkPtnodesForCommit: scheduled nested commit for submodule '%s'.\n",
								   SG_string__sz(pStringRepoPath))  );
#endif
	}
	
fail:
	SG_STRING_NULLFREE(pCtx, pStringRepoPath);
}


/**
 * Evaluate the given PTNODE and see if it should be marked
 * for COMMITING based upon the INCLUDE/EXCLUDE args.
 *
 * WE DO NOT TAKE AN IGNORES LIST because we are not scanning
 * the WD; rather, we are walking the ptnodes representing things
 * already under version control.
 *
 * WE TAKE BUT DO NOT PROPAGATE the bRecursive FLAG because that
 * flag only applies to the list of ITEMS that are passed in to the
 * top-level of SG_pendingtree__commit() for things they explicitly
 * named; once we are diving inside of one of them, we have to keep
 * diving.
 *
 * If the PTNODE is a directory, dive into it and try to
 * mark all of the children.
 *
 * For our purposes here a submodule looks just like a file/symlink:
 * a treenode entry with a HID for the content, which in the case of
 * a submodule is the hash of the unique repo-id/cset-id/etc fields.
 */
static void sg_ptnode__maybe_mark_ptnode_for_commit__recursive(SG_context * pCtx,
															   struct sg_ptnode * ptn,
															   SG_bool bRecursive,
															   const SG_file_spec* pFilespec,
															   SG_bool bPartialCommit_2, SG_bool bNoSubmodules)
{
	SG_string * pStringRepoPath = NULL;
	SG_file_spec__match_result eval = SG_FILE_SPEC__RESULT__IMPLIED;

	// we allow this routine to process phantom moved-away nodes
	// so that the dirty bit gets set on their parent nodes.

	if (ptn->pCurrentParent)
	{
		SG_bool bHasIncludeExclude = SG_FALSE;
		SG_ERR_CHECK(  SG_file_spec__has_pattern(pCtx, pFilespec, SG_FILE_SPEC__PATTERN__INCLUDE | SG_FILE_SPEC__PATTERN__EXCLUDE, &bHasIncludeExclude)  );
		if (bHasIncludeExclude == SG_TRUE)
		{
			SG_ERR_CHECK(  sg_ptnode__get_repo_path(pCtx, ptn, &pStringRepoPath)  );
			SG_ASSERT(  (pStringRepoPath)  );
			SG_ERR_CHECK(  SG_file_spec__should_include(pCtx,
														pFilespec,
														SG_string__sz(pStringRepoPath),
														SG_FILE_SPEC__CONVERT_REPO_PATH | SG_FILE_SPEC__MATCH_TRAILING_SLASH | SG_FILE_SPEC__NO_IGNORES,
														&eval)  );
		}
	}
	else
	{
		// Don't bother trying to filter the super-root since it doesn't
		// have a repo-path.
	}

	switch (eval)
	{
	default:				// quiets compiler
		SG_ASSERT( (0) );

		// TODO 2011/07/13 should we check for __RESULT__RESERVED and throw ?

	case SG_FILE_SPEC__RESULT__EXCLUDED:
	case SG_FILE_SPEC__RESULT__IGNORED:
		// An EXCLUDE/IGNORE pattern can match either a file or folder.  Either way,
		// if we match an EXCLUDE/IGNORE pattern, we can just stop.  This means that
		// when we match a directory, we implicilty EXCLUDE everything within it.
#if TRACE_COMMIT
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "MarkPtnodesForCommit: Excluding/Ignoring [%s]\n",
								   ((pStringRepoPath) ? SG_string__sz(pStringRepoPath) : "(null)"))  );
#endif
		break;

	case SG_FILE_SPEC__RESULT__INCLUDED:
		SG_ASSERT(  (bPartialCommit_2)  );	// by definition since we had an --include to match
		if (ptn->type == SG_TREENODEENTRY_TYPE_SUBMODULE)
			SG_ERR_CHECK(  _my_commit__deal_with_possibly_dirty_submodule(pCtx, ptn, bPartialCommit_2, bNoSubmodules)  );

		SG_ERR_CHECK(  _my_set_commit_flag(pCtx, ptn)  );
		if (ptn->type == SG_TREENODEENTRY_TYPE_DIRECTORY)
			SG_ERR_CHECK(  _my_commit_explicit_include_dir(pCtx, ptn, pStringRepoPath,
														   bRecursive,
														   pFilespec,
														   bPartialCommit_2, bNoSubmodules)  );
		break;

	case SG_FILE_SPEC__RESULT__IMPLIED:
		if (ptn->type == SG_TREENODEENTRY_TYPE_SUBMODULE)
			SG_ERR_CHECK(  _my_commit__deal_with_possibly_dirty_submodule(pCtx, ptn, bPartialCommit_2, bNoSubmodules)  );

		SG_ERR_CHECK(  _my_set_commit_flag(pCtx, ptn)  );
		if (ptn->type == SG_TREENODEENTRY_TYPE_DIRECTORY)
			SG_ERR_CHECK(  _my_commit_implicit_include_dir(pCtx, ptn, pStringRepoPath,
														   bRecursive,
														   pFilespec,
														   bPartialCommit_2, bNoSubmodules)  );
		break;

	case SG_FILE_SPEC__RESULT__MAYBE:
		// a maybe means there were actual includes, but we didn't match one of them.
		SG_ASSERT(  (bPartialCommit_2)  );	// by definition since we had an --include to match
		if (ptn->type == SG_TREENODEENTRY_TYPE_DIRECTORY)
		{
			// for a directory, a "maybe" means we should dive and see if there are
			// children that match.
			SG_ERR_CHECK(  _my_commit_provisional_include_dir(pCtx, ptn, pStringRepoPath,
															  bRecursive,
															  pFilespec,
															  bPartialCommit_2, bNoSubmodules)  );
		}
		else if (ptn->type == SG_TREENODEENTRY_TYPE_SUBMODULE)
		{
			// a "maybe" on a submodule directory just means that file_spec couldn't
			// match an --include for it.  All in all we want to treat this like an
			// exclude/ignore:
			// [a] we DO NOT mark this node (don't want bDirty_Reference changes).
			// [b] we DO NOT dive into it and force a nested commit (don't want bDirty_Pending changes).
			// 
			// because of the ambiguity partial-commits crossing the submodule boundary,
			// we also let this throw an error if the sub is dirty -- BECAUSE WE TOUCHED
			// THE PTNODE.
			SG_ERR_CHECK(  _my_commit__deal_with_possibly_dirty_submodule(pCtx, ptn, bPartialCommit_2, bNoSubmodules)  );

#if TRACE_COMMIT
			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
									   "MarkPtnodesForCommit: submodule-maybe ==> Excluding/Ignoring  [%s]\n",
									   ((pStringRepoPath) ? SG_string__sz(pStringRepoPath) : "(null)"))  );
#endif
		}
		else
		{
			// for a non-directory, a "maybe" should be treated like an exclude/ignore.
#if TRACE_COMMIT
			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
									   "MarkPtnodesForCommit: non-directory-maybe ==> Excluding/Ignoring  [%s]\n",
									   ((pStringRepoPath) ? SG_string__sz(pStringRepoPath) : "(null)"))  );
#endif
		}
	}

fail:
	SG_STRING_NULLFREE(pCtx, pStringRepoPath);
}

/**
 * We are given a partial pathname and a list of entries/items within it.
 * Build the complete pathname, lookup the associated PTNODE and see if
 * it should be marked for COMMITTING based upon the INCLUDE/EXCLUDE args.
 *
 * WE DO NOT TAKE AN IGNORES LIST because we are not scanning
 * the WD; rather, we are walking the ptnodes representing things
 * already under version control.
 *
 * If the PTNODE is a directory, we treat it as the root of a sub-tree
 * and walk all of the stuff within it.
 */
static void sg_ptnode__maybe_mark_items_for_commit(SG_context * pCtx,
												   SG_pendingtree * pPendingTree,
												   SG_uint32 count_items, const char** paszItems,
												   SG_bool bRecursive,
												   const SG_file_spec* pFilespec,
												   SG_bool bPartialCommit_2, SG_bool bNoSubmodules)
{
	SG_pathname * pPathItem_k = NULL;
	SG_pathname * pPathParent_k = NULL;
	SG_string * pStringEntryname_k = NULL;
	SG_uint32 k;

	SG_ASSERT(  (count_items > 0)  );

	for (k=0; k<count_items; k++)
	{
		struct sg_ptnode * ptn_k = NULL;		// we don't own this
		struct sg_ptnode * ptnParent_k = NULL;	// we don't own this

		SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPathItem_k, paszItems[k])  );

		SG_ERR_CHECK(  SG_PATHNAME__ALLOC__COPY(pCtx, &pPathParent_k, pPathItem_k)  );
		SG_ERR_CHECK(  SG_pathname__remove_last(pCtx, pPathParent_k)  );
		SG_ERR_CHECK(  sg_pendingtree__local_path_to_ptnode__allow_deleted(pCtx, pPendingTree, pPathParent_k, &ptnParent_k)  );

		SG_ERR_CHECK(  SG_pathname__get_last(pCtx, pPathItem_k, &pStringEntryname_k)  );
		SG_ERR_CHECK(  sg_ptnode__find_entry_by_name_not_recursive__active(pCtx, ptnParent_k, SG_string__sz(pStringEntryname_k), &ptn_k)  );
		if (!ptn_k)
		{
			SG_ERR_CHECK(  sg_ptnode__find_entry_by_name_not_recursive__inactive(pCtx, ptnParent_k, SG_string__sz(pStringEntryname_k), &ptn_k)  );
			if (!ptn_k)
				SG_ERR_THROW2(  SG_ERR_NOT_FOUND,
								(pCtx, "'%s'", SG_pathname__sz(pPathItem_k))  );
		}

#if TRACE_COMMIT
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "MarkItemsForCommit: Starting at [%s]\n",
								   SG_pathname__sz(pPathItem_k))  );
#endif

		SG_ERR_CHECK(  sg_ptnode__maybe_mark_ptnode_for_commit__recursive(pCtx,
																		  ptn_k,
																		  bRecursive,
																		  pFilespec,
																		  bPartialCommit_2, bNoSubmodules)  );

		SG_PATHNAME_NULLFREE(pCtx, pPathItem_k);
		SG_PATHNAME_NULLFREE(pCtx, pPathParent_k);
		SG_STRING_NULLFREE(pCtx, pStringEntryname_k);
	}

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathItem_k);
	SG_PATHNAME_NULLFREE(pCtx, pPathParent_k);
	SG_STRING_NULLFREE(pCtx, pStringEntryname_k);
}

//////////////////////////////////////////////////////////////////

/**
 * Set the REVERTING flag on this ptnode.
 */
static void _my_set_revert_flag(SG_context * pCtx, struct sg_ptnode * ptn)
{
	SG_UNUSED(pCtx);

	ptn->temp_flags |= sg_PTNODE_TEMP_FLAG_REVERTING;

	// WE DEVIATE HERE FROM THE COMMIT VERSION.
	//
	// I don't think we want to force set the REVERTING flag in the ptnodes
	// of the parent directories.  In the case of a COMMIT, when we commit
	// changes to a file, we alter the treenode of the parent directory (and
	// change its HID) which alters its parent directory and so on.  However,
	// in the case of REVERT, the user has only said that they want to revert
	// the given ptnode -- that is completely independent of any changes
	// made to any of the parent directories.
	//
	// For example, if I have:
	//     @/dir1/file1
	// in the tree and do:
	//     vv rename dir1/file1 file2
	//     vv rename dir1 dir2
	// and say:
	//     vv revert dir2/file2
	// we should have just:
	//     @/dir2/file1
	// because there's no reason to assume that the rename of dir2 should be
	// reverted too.
}

static void sg_ptnode__maybe_mark_ptnode_for_revert__recursive(SG_context * pCtx,
															   struct sg_ptnode * ptn,
															   SG_bool bRecursive,
															   const SG_file_spec* pFilespec);

/**
 * Do the actual dirty work of walking the contents of the directory
 * and marking the contents.
 *
 * WE DO NOT TAKE A bRecursive FLAG because it only applied to the
 * top-level items named in SG_pendingtree__revert() and as discussed in
 * sg_ptnode__maybe_mark_ptnode_for_revert__recursive() we don't want to
 * propagate that.
 */
static void _my_revert_include_dir(SG_context * pCtx,
								   struct sg_ptnode * ptn,
								   SG_string * pStringRepoPath,
								   const SG_file_spec* pFilespec)
{
	SG_rbtree_iterator * pIterator = NULL;
	const char * pszKey;
	struct sg_ptnode * ptnSub;
	SG_bool bFound = SG_FALSE;

	SG_ASSERT(  (ptn->type == SG_TREENODEENTRY_TYPE_DIRECTORY)  );

	if (!ptn->prbItems)
	{
#if 0 && defined(DEBUG)
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "MarkPtnodesForRevert: TODO determine if the directory [%s] is really empty or just not loaded.\n",
								   ((pStringRepoPath) ? SG_string__sz(pStringRepoPath) : "(null)"))  );
#else
        SG_UNUSED(pStringRepoPath);
#endif
		return;
	}

#if TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "MarkPtnodesForRevert: Considering contents of directory [%s]\n",
							   ((pStringRepoPath) ? SG_string__sz(pStringRepoPath) : "(null)"))  );
#endif

	SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pIterator, ptn->prbItems, &bFound, &pszKey, (void**) &ptnSub)  );
	while (bFound)
	{
		SG_ERR_CHECK(  sg_ptnode__maybe_mark_ptnode_for_revert__recursive(pCtx,
																		  ptnSub,
																		  SG_TRUE,		// do not propagate bRecursive
																		  pFilespec)  );

		SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pIterator, &bFound, &pszKey, (void**) &ptnSub)  );
	}

fail:
	SG_RBTREE_ITERATOR_NULLFREE(pCtx, pIterator);
}

/**
 * A directory was EXPLICITLY INCLUDED.  (there were INCLUDES and we matched one of them.)
 *
 * TODO 2010/06/07 So we matched the directory, what does this imply for
 * TODO            the contents of the directory?  I'm going to assume for
 * TODO            the moment that it means that we add the directory and
 * TODO            then (optinally) dive into it WITHOUT any INCLUDES so
 * TODO            that we get the full contents.
 * TODO
 * TODO            That is, we treat:
 * TODO                vv add --include='**.dir' --include='**.c' .
 * TODO
 * TODO            like the combination of 2 commands:
 * TODO                vv add --include='**.c' .
 * TODO                find . -name="*.dir" | xargs vv add
 * TODO
 * TODO            I don't think it makes any sense to treat it like:
 * TODO                vv add --include='**.c' .
 * TODO                find . -name="*.dir" | xargs vv add --include='**.c'
 * TODO
 * TODO            See sprawl-818.
 *
 * We do, however, carry along the EXCLUDES.
 */
static void _my_revert_explicit_include_dir(SG_context * pCtx,
											struct sg_ptnode * ptn,
											SG_string * pStringRepoPath,
											SG_bool bRecursive,
											const SG_file_spec * pFilespec)
{
	SG_file_spec* pFilespecNoIncludes = NULL;

	if (!bRecursive)
	{
#if TRACE_REVERT
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "MarkPtnodesForRevert: Skipping dive on explicitly-included directory [%s] because not-recursive\n",
								   ((pStringRepoPath) ? SG_string__sz(pStringRepoPath) : "(null)"))  );
#endif
		return;
	}

	SG_ERR_CHECK(  SG_FILE_SPEC__ALLOC__COPY(pCtx, pFilespec, &pFilespecNoIncludes)  );
	SG_ERR_CHECK(  SG_file_spec__modify_flags(pCtx, pFilespecNoIncludes, SG_FILE_SPEC__NO_INCLUDES, 0u, NULL, NULL)  );
	SG_ERR_CHECK(  _my_revert_include_dir(pCtx, ptn, pStringRepoPath, pFilespecNoIncludes)  );

fail:
	SG_FILE_SPEC_NULLFREE(pCtx, pFilespecNoIncludes);
}

/**
 * There were no INCLUDES, so we are including everything we can find.
 */
static void _my_revert_implicit_include_dir(SG_context * pCtx,
											struct sg_ptnode * ptn,
											SG_string * pStringRepoPath,
											SG_bool bRecursive,
											const SG_file_spec * pFilespec)
{
	if (!bRecursive)
	{
#if TRACE_REVERT
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "MarkPtnodesForRevert: Skipping dive on implicitly-included directory [%s] because not-recursive\n",
								   ((pStringRepoPath) ? SG_string__sz(pStringRepoPath) : "(null)"))  );
#endif
		return;
	}

	SG_ERR_CHECK(  _my_revert_include_dir(pCtx, ptn, pStringRepoPath, pFilespec)  );

fail:
	return;
}

/**
 * A provisional dir.  We had INCLUDES, but didn't exactly match this directory,
 * allow children to be inspected.
 */
static void _my_revert_provisional_include_dir(SG_context * pCtx,
											   struct sg_ptnode * ptn,
											   SG_string * pStringRepoPath,
											   SG_bool bRecursive,
											   const SG_file_spec * pFilespec)
{
	if (!bRecursive)
	{
#if TRACE_REVERT
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "MarkPtnodesForRevert: Skipping provisional dive on directory [%s] because not-recursive\n",
								   ((pStringRepoPath) ? SG_string__sz(pStringRepoPath) : "(null)"))  );
#endif
		return;
	}

	SG_ERR_CHECK_RETURN(  _my_revert_include_dir(pCtx, ptn, pStringRepoPath, pFilespec)  );
}

/**
 * Evaluate the given PTNODE and see if it should be marked
 * for REVERTING based upon the INCLUDE/EXCLUDE args.
 *
 * WE DO NOT TAKE AN IGNORES LIST because we are not scanning
 * the WD; rather, we are walking the ptnodes representing things
 * already under version control.
 *
 * WE TAKE BUT DO NOT PROPAGATE the bRecursive FLAG because that
 * flag only applies to the list of ITEMS that are passed in to the
 * top-level of SG_pendingtree__revert() for things they explicitly
 * named; once we are diving inside of one of them, we have to keep
 * diving.
 *
 * If the PTNODE is a directory, dive into it and try to
 * mark all of the children.
 */
static void sg_ptnode__maybe_mark_ptnode_for_revert__recursive(SG_context * pCtx,
															   struct sg_ptnode * ptn,
															   SG_bool bRecursive,
															   const SG_file_spec* pFilespec)
{
	SG_string * pStringRepoPath = NULL;
	SG_file_spec__match_result eval = SG_FILE_SPEC__RESULT__IMPLIED;

	// Don't bother trying to filter the super-root since it doesn't
	// have a repo-path.

	if (ptn->pCurrentParent)
	{
		SG_bool bHasIncludeExclude = SG_FALSE;
		SG_ERR_CHECK(  SG_file_spec__has_pattern(pCtx, pFilespec, SG_FILE_SPEC__PATTERN__INCLUDE | SG_FILE_SPEC__PATTERN__EXCLUDE, &bHasIncludeExclude)  );
		if (bHasIncludeExclude == SG_TRUE)
		{
			SG_ERR_CHECK(  sg_ptnode__get_repo_path(pCtx, ptn, &pStringRepoPath)  );
			SG_ASSERT(  (pStringRepoPath)  );
			SG_ERR_CHECK(  SG_file_spec__should_include(pCtx,
														pFilespec,
														SG_string__sz(pStringRepoPath),
														SG_FILE_SPEC__CONVERT_REPO_PATH | SG_FILE_SPEC__MATCH_TRAILING_SLASH | SG_FILE_SPEC__NO_IGNORES,
														&eval)  );
		}
	}

	switch (eval)
	{
	default:			// quiets compiler
		SG_ASSERT( (0) );

	case SG_FILE_SPEC__RESULT__EXCLUDED:
	case SG_FILE_SPEC__RESULT__IGNORED:
		// An EXCLUDE/IGNORE pattern can match either a file or folder.  Either way,
		// if we match an EXCLUDE/IGNORE pattern, we can just stop.  This means that
		// when we match a directory, we implicilty EXCLUDE everything within it.
#if TRACE_REVERT
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "MarkPtnodesForRevert: Excluding/Ignoring [%s]\n",
								   ((pStringRepoPath) ? SG_string__sz(pStringRepoPath) : "(null)"))  );
#endif
		break;

	case SG_FILE_SPEC__RESULT__INCLUDED:
		SG_ERR_CHECK(  _my_set_revert_flag(pCtx, ptn)  );
		if (ptn->type == SG_TREENODEENTRY_TYPE_DIRECTORY)
			SG_ERR_CHECK(  _my_revert_explicit_include_dir(pCtx, ptn, pStringRepoPath,
														   bRecursive,
														   pFilespec)  );
		break;

	case SG_FILE_SPEC__RESULT__IMPLIED:
		SG_ERR_CHECK(  _my_set_revert_flag(pCtx, ptn)  );
		if (ptn->type == SG_TREENODEENTRY_TYPE_DIRECTORY)
			SG_ERR_CHECK(  _my_revert_implicit_include_dir(pCtx, ptn, pStringRepoPath,
														   bRecursive,
														   pFilespec)  );
		break;

	case SG_FILE_SPEC__RESULT__MAYBE:
		// a maybe means there were actual includes, but we didn't match one of them.
		if (ptn->type == SG_TREENODEENTRY_TYPE_DIRECTORY)
		{
			// for a directory, a "maybe" means we should dive and see if there are
			// children that match.
			SG_ERR_CHECK(  _my_revert_provisional_include_dir(pCtx, ptn, pStringRepoPath,
															  bRecursive,
															  pFilespec)  );
		}
		else
		{
			// for a non-directory, a "maybe" should be treated like an exclude/ignore.
#if TRACE_REVERT
			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
									   "MarkPtnodesForRevert: non-directory-maybe ==> Excluding/Ignoring  [%s]\n",
									   ((pStringRepoPath) ? SG_string__sz(pStringRepoPath) : "(null)"))  );
#endif
		}
	}

fail:
	SG_STRING_NULLFREE(pCtx, pStringRepoPath);
}

/**
 * We are given a partial pathname and a list of entries/items within it.
 * Build the complete pathname, lookup the associated PTNODE and see if
 * it should be marked for REVERTING based upon the INCLUDE/EXCLUDE args.
 *
 * WE DO NOT TAKE AN IGNORES LIST because we are not scanning
 * the WD; rather, we are walking the ptnodes representing things
 * already under version control.
 *
 * If the PTNODE is a directory, we treat it as the root of a sub-tree
 * and walk all of the stuff within it.
 */
static void sg_ptnode__maybe_mark_items_for_revert(SG_context * pCtx,
												   SG_pendingtree * pPendingTree,
												   SG_uint32 count_items, const char** paszItems,
												   SG_bool bRecursive,
												   const SG_file_spec* pFilespec)
{
	SG_pathname * pPathItem_k = NULL;
	SG_pathname * pPathParent_k = NULL;
	SG_string * pStringEntryname_k = NULL;
	struct sg_ptnode * ptn_k;
	struct sg_ptnode * ptnParent_k;
	SG_uint32 k;

	for (k=0; k<count_items; k++)
	{
		SG_ERR_CHECK(  SG_PATHNAME__ALLOC__SZ(pCtx, &pPathItem_k, paszItems[k])  );

		SG_ERR_CHECK(  SG_PATHNAME__ALLOC__COPY(pCtx, &pPathParent_k, pPathItem_k)  );
		SG_ERR_CHECK(  SG_pathname__remove_last(pCtx, pPathParent_k)  );
		SG_ERR_CHECK(  sg_pendingtree__local_path_to_ptnode(pCtx, pPendingTree, pPathParent_k, SG_TRUE, &ptnParent_k)  );

		SG_ERR_CHECK(  SG_pathname__get_last(pCtx, pPathItem_k, &pStringEntryname_k)  );
		SG_ERR_CHECK(  sg_ptnode__find_entry_by_name_not_recursive__active(pCtx, ptnParent_k, SG_string__sz(pStringEntryname_k), &ptn_k)  );
		if (!ptn_k)
			SG_ERR_CHECK(  sg_ptnode__find_entry_by_name_not_recursive__inactive(pCtx, ptnParent_k, SG_string__sz(pStringEntryname_k), &ptn_k)  );

		// TODO 2010/06/08 Sprawl-825.  If we are trying to revert something that was
		// TODO            deleted, we need to do an __inactive() lookup to get its ptnode.
		// TODO            We also need to verify that the parent is active.
		// TODO            That is if they deleted the parent (or really any ancestor)
		// TODO            directory, we cannot un-delete this entry without first doing
		// TODO            the parents.
		// TODO
		// TODO            We should throw
		// TODO            SG_ERR_THROW(  SG_ERR_PENDINGTREE_PARENT_DIR_MISSING_OR_INACTIVE  );

		if (!ptn_k)
			SG_ERR_THROW2(  SG_ERR_NOT_FOUND,
							(pCtx, "Item [%s] not found in pendingtree.", SG_pathname__sz(pPathItem_k))  );

#if TRACE_REVERT
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "MarkItemsForRevert: Starting at [%s]\n",
								   SG_pathname__sz(pPathItem_k))  );
#endif

		SG_ERR_CHECK(  sg_ptnode__maybe_mark_ptnode_for_revert__recursive(pCtx, ptn_k, bRecursive, pFilespec)  );

		SG_PATHNAME_NULLFREE(pCtx, pPathItem_k);
		SG_PATHNAME_NULLFREE(pCtx, pPathParent_k);
		SG_STRING_NULLFREE(pCtx, pStringEntryname_k);
	}

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathItem_k);
	SG_PATHNAME_NULLFREE(pCtx, pPathParent_k);
	SG_STRING_NULLFREE(pCtx, pStringEntryname_k);
}


static void sg_ptnode__check_parent_dirty_non_recursive(
	SG_context* pCtx,
	struct sg_ptnode* ptn,
	SG_bool* pbParentDirty
	)
{
	/* This function never clears the flag, it merely sets it when
	 * needed.  So if it is already set, we're done. */
	if (*pbParentDirty)
	{
		return;
	}

	if (
		(ptn->saved_flags & sg_PTNODE_SAVED_FLAG_DELETED)
		|| (ptn->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY)
		|| (ptn->pszgidMovedFrom)
		|| (
			(!ptn->pBaselineEntry)
			&& ptn->pCurrentParent
			)
		)
	{
		*pbParentDirty = SG_TRUE;
		return;
	}

	if (ptn->pszCurrentName)
	{
		const char* pszBaselineName = NULL;

		SG_ERR_CHECK(  SG_treenode_entry__get_entry_name(pCtx, ptn->pBaselineEntry, &pszBaselineName)  );

		if (0 != strcmp(ptn->pszCurrentName, pszBaselineName))
		{
			*pbParentDirty = SG_TRUE;
			return;
		}
	}

    if (ptn->pBaselineEntry)
    {
        SG_int64 iBaselineAttributeBits = 0;
        const char* pszBaselineHidAttributes = NULL;

        SG_ERR_CHECK(  SG_treenode_entry__get_attribute_bits(pCtx, ptn->pBaselineEntry, &iBaselineAttributeBits)  );
        if (iBaselineAttributeBits != ptn->iCurrentAttributeBits)
        {
            *pbParentDirty = SG_TRUE;
            return;
        }

		SG_ERR_CHECK(  SG_treenode_entry__get_hid_xattrs(pCtx, ptn->pBaselineEntry, &pszBaselineHidAttributes)  );
        if (ptn->pszHidXattrs && pszBaselineHidAttributes)
        {
            if (0 != strcmp(ptn->pszHidXattrs, pszBaselineHidAttributes))
            {
                *pbParentDirty = SG_TRUE;
                return;
            }
        }
        else if (ptn->pszHidXattrs || pszBaselineHidAttributes)
        {
            *pbParentDirty = SG_TRUE;
            return;
        }
    }

fail:
	return;
}

static void sg_ptnode__post_commit_fixup(
	SG_context* pCtx,
	struct sg_ptnode* ptn,
	SG_bool* pbRemoveMe
	)
{
	SG_rbtree_iterator* pIterator = NULL;
	SG_rbtree* prbRemove = NULL;

	if (!(ptn->temp_flags & sg_PTNODE_TEMP_FLAG_COMMITTING))
	{
		*pbRemoveMe = SG_FALSE;
		return;
	}

	ptn->temp_flags &= ~sg_PTNODE_TEMP_FLAG_COMMITTING;
	if (
		(ptn->saved_flags & sg_PTNODE_SAVED_FLAG_DELETED)
		|| (ptn->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY)
		)
	{
		*pbRemoveMe = SG_TRUE;
		return;
	}

	if (ptn->pCurrentParent)
	{
		/*
		 * As long as this is not the top of the tree, we expect this
		 * node to have a new entry, because all of the cases where
		 * it does not have already been checked above.
		 *
		 * So, we take the new entry and make it the baseline entry,
		 * since the commit succeeded.  This means that later if this
		 * node is checked for dirty, it will claim that it is not
		 * dirty.
		 */

		SG_ASSERT(ptn->pNewEntry);
		SG_treenode_entry__free(pCtx, ptn->pBaselineEntry);
		ptn->pBaselineEntry = ptn->pNewEntry;
		ptn->pNewEntry = NULL;

		ptn->pszCurrentName = NULL;
		ptn->pszCurrentHid = NULL;
		ptn->pszHidDisposable = NULL;
		ptn->bDisposableDir = SG_FALSE;
		ptn->bStickyDir = SG_FALSE;
		ptn->bDeletedAfterAdded = SG_FALSE;
        ptn->pszHidXattrs = NULL;
        SG_ERR_CHECK(  SG_treenode_entry__get_hid_xattrs(pCtx, ptn->pBaselineEntry, &ptn->pszHidXattrs)  );
		ptn->pszgidMovedFrom = NULL;
        SG_ERR_CHECK(  SG_treenode_entry__get_attribute_bits(pCtx, ptn->pBaselineEntry, &ptn->iCurrentAttributeBits)  );
	}

	if (SG_TREENODEENTRY_TYPE_DIRECTORY == ptn->type)
	{
		SG_bool b = SG_FALSE;
		const char* pszKey = NULL;
		struct sg_ptnode* ptnSub = NULL;
		SG_uint32 count = 0;

		if (ptn->prbItems)
		{
			SG_ERR_CHECK(  SG_rbtree__count(pCtx, ptn->prbItems, &count)  );
		}

		if (count)
		{
			SG_ERR_CHECK(  SG_RBTREE__ALLOC__PARAMS(pCtx, &prbRemove, count, NULL)  );

			SG_ASSERT(ptn->prbItems);

			/* Recurse downward into every entry */
			SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pIterator, ptn->prbItems, &b, &pszKey, (void**) &ptnSub)  );
			while (b)
			{
				SG_bool bRemoveThis = SG_FALSE;

				SG_ERR_CHECK(  sg_ptnode__post_commit_fixup(pCtx, ptnSub, &bRemoveThis)  );
				if (bRemoveThis)
				{
					SG_ERR_CHECK(  SG_rbtree__add(pCtx, prbRemove, pszKey)  );
				}

				SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pIterator, &b, &pszKey, (void**) &ptnSub)  );
			}
			SG_rbtree__iterator__free(pCtx, pIterator);
			pIterator = NULL;

			/* Now remove anything which asked to be removed */
			SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pIterator, prbRemove, &b, &pszKey, NULL)  );
			while (b)
			{
				SG_ERR_CHECK(  SG_rbtree__remove__with_assoc(pCtx, ptn->prbItems, pszKey, (void**) &ptnSub)  );

				SG_PTNODE_NULLFREE(pCtx, ptnSub);

				SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pIterator, &b, &pszKey, (void**) &ptnSub)  );
			}
			SG_RBTREE_ITERATOR_NULLFREE(pCtx, pIterator);

			SG_RBTREE_NULLFREE(pCtx, prbRemove);
		}
	}

	return;

fail:
	SG_RBTREE_ITERATOR_NULLFREE(pCtx, pIterator);
}

void sg_pendingtree__post_commit_fixup(
	SG_context* pCtx,
	SG_pendingtree* pPendingTree
	)
{
	SG_bool bRemoveTop = SG_FALSE;
	SG_bool bDirty = SG_FALSE;

	SG_ERR_CHECK(  sg_ptnode__post_commit_fixup(pCtx, pPendingTree->pSuperRoot, &bRemoveTop)  );
	SG_ERR_CHECK(  sg_ptnode__unload_dir_entries_if_not_dirty(pCtx, pPendingTree->pSuperRoot, &bDirty)  );

	if (!(pPendingTree->pSuperRoot->prbItems))
	{
		SG_PTNODE_NULLFREE(pCtx, pPendingTree->pSuperRoot);
	}

fail:
	return;
}

static void _sg_pendingtree__precommit_check(SG_context* pCtx,
							SG_pendingtree* pPendingTree,
							const SG_audit* pq,
							SG_uint32 count_items, const char** paszItems,
							SG_bool bRecursive,
							const SG_file_spec* pFilespec,
							const char* const* paszAssocs, SG_uint32 count_assocs,
											 SG_varray **pdescs,
									 SG_bool *pPartial1, SG_bool *pPartial2, 
									 SG_file_spec **ppInternalFilespec,
											 SG_varray **ppva_wd_parents, SG_uint32 *pnrParents
	)

{
	SG_uint32 nrUnresolved = 0;
	SG_varray * pva_wd_parents = NULL;
	SG_uint32 nrParents = 0;
	SG_bool bPartialCommit_1 = SG_FALSE, bPartialCommit_2 = SG_FALSE;
	SG_file_spec* pInternalFilespec = NULL;
	SG_bool bHasIncludeExclude = SG_FALSE;
	SG_varray *descs = NULL;

	SG_NULLARGCHECK_RETURN(pPendingTree);
	SG_NULLARGCHECK_RETURN(pq);

	SG_ARGCHECK_RETURN(  ((   count_items==0)   || (   paszItems != NULL)), count_items     );
	SG_ARGCHECK_RETURN(  ((bRecursive==SG_TRUE) || (count_items > 0)),      bRecursive      );	// if not recursive, require at least one item.

	// verify that any/all merge issues have been resolved.

	SG_ERR_CHECK(  SG_pendingtree__count_unresolved_wd_issues(pCtx, pPendingTree, &nrUnresolved)  );
	if (nrUnresolved > 0)
		SG_ERR_THROW(  SG_ERR_CANNOT_COMMIT_WITH_UNRESOLVED_ISSUES  );

	// this will THROW if the vfile didn't know what the parents are.

	SG_ERR_CHECK(  SG_pendingtree__get_wd_parents(pCtx, pPendingTree, &pva_wd_parents)  );
	SG_ERR_CHECK(  SG_varray__count(pCtx, pva_wd_parents, &nrParents)  );
	SG_ASSERT(  (nrParents >= 1)  );

	// There are 2 ways to have to have a partial-commit:
	// [1] a list of items;
	// [2] --include/--exclude/--nonrecursive.
	//
	// Disallow partial-commits where we have a merge.
	//
	// We also want to disallow [2] partial-commits when we bump
	// into dirty submodules.  For that we have to pass the info
	// to the __mark_ptnode_for_commit code.

	bPartialCommit_1 = (count_items > 0);
	if (bPartialCommit_1 && (nrParents > 1))
		SG_ERR_THROW(  SG_ERR_CANNOT_PARTIAL_COMMIT_AFTER_MERGE  );

	SG_ERR_CHECK(  SG_FILE_SPEC__ALLOC__COPY(pCtx, pFilespec, &pInternalFilespec)  );

	SG_ERR_CHECK(  SG_file_spec__has_pattern(pCtx, pInternalFilespec, SG_FILE_SPEC__PATTERN__INCLUDE | SG_FILE_SPEC__PATTERN__EXCLUDE, &bHasIncludeExclude)  );
	bPartialCommit_2 = ((bHasIncludeExclude == SG_TRUE) || (!bRecursive));
	if (bPartialCommit_2 && (nrParents > 1))
		SG_ERR_THROW(  SG_ERR_CANNOT_PARTIAL_COMMIT_AFTER_MERGE  );

	if (paszAssocs!=NULL && count_assocs>0)
    {
		SG_ERR_CHECK(  SG_varray__alloc(pCtx, &descs)  );
		SG_ERR_CHECK(  SG_vc_hooks__ASK__WIT__VALIDATE_ASSOCIATIONS( pCtx, pPendingTree->pRepo, paszAssocs, count_assocs, descs )  );

		if (pdescs)
		{
			*pdescs = descs;
			descs = NULL;
		}
	}

	if (pPartial1)
		*pPartial1 = bPartialCommit_1;
	if (pPartial2)
		*pPartial2 = bPartialCommit_2;
	if (pnrParents)
		*pnrParents = nrParents;
	if (ppInternalFilespec)
	{
		*ppInternalFilespec = pInternalFilespec;
		pInternalFilespec = NULL;
	}
	if (ppva_wd_parents)
	{
		*ppva_wd_parents = pva_wd_parents;
		pva_wd_parents = NULL;
	}

fail:
    SG_VARRAY_NULLFREE(pCtx, pva_wd_parents);
	SG_FILE_SPEC_NULLFREE(pCtx, pInternalFilespec);
	SG_VARRAY_NULLFREE(pCtx, descs);
}

void SG_pendingtree__precommit_check(SG_context* pCtx,
							SG_pendingtree* pPendingTree,
							const SG_audit* pq,
							SG_uint32 count_items, const char** paszItems,
							SG_bool bRecursive,
							const SG_file_spec* pFilespec,
							const char* const* paszAssocs, SG_uint32 count_assocs,
											 SG_varray **descs)
{
	_sg_pendingtree__precommit_check(pCtx, pPendingTree, pq, count_items, paszItems,
									 bRecursive, pFilespec, paszAssocs, count_assocs, descs,
									 NULL, NULL, NULL, NULL, NULL);
}


/* TODO why doesn't "sg commit ." work ?
 *
 * TODO 2010/05/18 Is the above comment still valid?
 */
void SG_pendingtree__commit(SG_context* pCtx,
							SG_pendingtree* pPendingTree,
							const SG_audit* pq,
							SG_uint32 count_items, const char** paszItems,
							SG_bool bRecursive,
							const SG_file_spec* pFilespec,
							const char* const* paszAssocs, SG_uint32 count_assocs,
							const SG_stringarray * psaStamps,
							SG_bool bNoSubmodules,
							const char * psz_comment,	// optional
							SG_dagnode** ppdn)
{
	// We DO NOT need/use the IGNORES list because we are not scanning.

	SG_committing* ptx = NULL;
	SG_changeset * pChangesetOriginal = NULL;
	SG_changeset * pcs_branches_dag = NULL;
	SG_dagnode * pDagnodeOriginal = NULL;
	SG_dagnode * pdn_branches_dag = NULL;
	const char* psz_hid_parent_k = NULL;
	SG_bool bSomethingCommitted = SG_FALSE;
	SG_uint32 k, nrParents;
	char * psz_hid_new_changeset = NULL;
	SG_varray * pva_wd_parents = NULL;

	SG_file_spec* pInternalFilespec = NULL;
	SG_pathname * pPathMergeTempDir = NULL;
	SG_bool bHasMergeTempDir = SG_FALSE;
    char* psz_tied_branch_name = NULL;
    SG_vhash* pvh_locks = NULL;
    SG_zingtx* pztx = NULL;
	SG_bool bPartialCommit_1 = SG_FALSE;
	SG_bool bPartialCommit_2 = SG_FALSE;

	SG_ERR_CHECK(  _sg_pendingtree__precommit_check(
					   pCtx, pPendingTree, pq, count_items, paszItems, bRecursive, pFilespec,
					   paszAssocs, count_assocs, 
					   NULL, &bPartialCommit_1, &bPartialCommit_2, &pInternalFilespec, &pva_wd_parents, &nrParents)  );

    SG_ERR_CHECK(  SG_workingdir__get_branch_info(pCtx, pPendingTree->pPathWorkingDirectoryTop, &psz_tied_branch_name)  );
    if (psz_tied_branch_name)
    {
        SG_ERR_CHECK(  SG_vc_locks__pre_commit(pCtx, pPendingTree->pRepo, pq->who_szUserId, psz_tied_branch_name, &pvh_locks)  );
    }

	//////////////////////////////////////////////////////////////////
	// mark all of the ptnodes that should participate in the commit.

	if (count_items > 0)
	{
		SG_ERR_CHECK(  sg_ptnode__maybe_mark_items_for_commit(pCtx,
															  pPendingTree,
															  count_items, paszItems,
															  bRecursive,
															  pInternalFilespec,
															  bPartialCommit_2, bNoSubmodules)  );

		/* check the tree to see if the user is trying to commit only
		 * the target half of a move operation.  If so, fail.  We
		 * cannot allow this case because it can result in a situation
		 * where the same object (identified by its gid) exists in two
		 * places in the tree at the same time. */
		SG_ERR_CHECK(  sg_ptnode__fail_if_any_split_moves(pCtx, pPendingTree->pSuperRoot, sg_PTNODE_TEMP_FLAG_COMMITTING)  );
	}
	else
	{
		SG_ERR_CHECK(  sg_ptnode__maybe_mark_ptnode_for_commit__recursive(pCtx,
																		  pPendingTree->pSuperRoot,
																		  SG_TRUE,	// we've already asserted that bRecursive is true
																		  pInternalFilespec,
																		  bPartialCommit_2, bNoSubmodules)  );
	}

	//////////////////////////////////////////////////////////////////
	// begin creating the cset.

	SG_ERR_CHECK(  SG_committing__alloc(pCtx, &ptx, pPendingTree->pRepo, SG_DAGNUM__VERSION_CONTROL, pq, SG_CSET_VERSION__CURRENT)  );

	for (k=0; k<nrParents; k++)
	{
		SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pva_wd_parents, k, &psz_hid_parent_k)  );
		SG_ERR_CHECK(  SG_committing__add_parent(pCtx, ptx, psz_hid_parent_k)  );
	}

	// write all dirty blobs (both user files and treenodes) to the repo.
	// this sets pPendingTree->pSuperRoot->pszCurrentHid.  this should
	// silently allow any/all of the blobs to already exist in the repo.
	//
	// bSomethingCommitted means that there was something dirty that needed
	// to be written; if this is false, it does not mean that we can stop
	// working because it could be that we are doing a merge where there
	// were no actual tree changes but we are bringing together the leaves.
	SG_ERR_CHECK(  sg_ptnode__commit(pCtx, pPendingTree->pSuperRoot, NULL, ptx, &bSomethingCommitted, psz_comment, pvh_locks)  );

	if (!bSomethingCommitted && (nrParents == 1))
	{
		// don't allow trivial changeset where the only change is
		// the parent.
		//
		// WARNING: we've probably trashed the pendingtree in memory.

		SG_ERR_THROW(  SG_ERR_NOTHING_TO_COMMIT  );
	}

	SG_ERR_CHECK(  SG_committing__tree__set_root(pCtx, ptx, pPendingTree->pSuperRoot->pszCurrentHid)  );

	// write the actual changeset and dagnode.  this computes the resulting
	// changeset's HID.  this eats any duplicate changeset/dagnode warning
	// silently continues if the changeset/dagnode already exists in the
	// repo and returns handles to the existing ones.
	SG_ERR_CHECK(  SG_committing__end(pCtx, ptx, &pChangesetOriginal, &pDagnodeOriginal)  );
	ptx = NULL;

	// The changeset-proper is now completely recorded.
	//////////////////////////////////////////////////////////////////
	// refurbish the pendingtree after the commit.  this should strip out
	// any of the dirt, clear the unresolved issues, and update the parents
	// array to reflect that we have a new baseline.

	SG_ERR_CHECK(  sg_pendingtree__post_commit_fixup(pCtx, pPendingTree)  );

	// after a commit, we have a single new "baseline".  stuff it into the pendingtree.
	SG_ERR_CHECK(  SG_dagnode__get_id(pCtx, pDagnodeOriginal, &psz_hid_new_changeset)  );
	SG_ERR_CHECK(  SG_pendingtree__set_single_wd_parent(pCtx, pPendingTree, psz_hid_new_changeset)  );

	// If we had a merge, we need to clear "issues" section in wd.json.
	// Our goal here is to forget all of them (without regard to whether
	// they were resolved or unresolved) because after the commit and we
	// update to the new baseline, these issues are not longer appropriate.
	SG_ERR_CHECK(  SG_pendingtree__clear_wd_issues(pCtx, pPendingTree)  );
	SG_ERR_CHECK(  SG_pendingtree__set_wd_choices(pCtx, pPendingTree, NULL)  );

	// Likewise, clear the MERGE-TEMP-DIR associated with the pending merge.
	SG_ERR_CHECK(  SG_pendingtree__get_mrg_temp_dir(pCtx, pPendingTree, &bHasMergeTempDir, &pPathMergeTempDir)  );
	if (bHasMergeTempDir)
		SG_ERR_CHECK(  SG_pendingtree__set_mrg_temp_dir(pCtx, pPendingTree, NULL)  );

	SG_ERR_CHECK(  SG_pendingtree__save(pCtx, pPendingTree)  );

	// The new pendingtree on disk is now ready for future work.
	//////////////////////////////////////////////////////////////////
	// Record any auxillary data associated with the changeset.

	if (ppdn)
    {
		*ppdn = pDagnodeOriginal;
        pDagnodeOriginal = NULL;
    }

	if (bHasMergeTempDir)
	{
		// After we get the commit completely finalized (and don't care
		// so much about an error), we'll try to delete the merge-temp-dir
		// from the disk.  this is pretty much trash now.  if it fails,
		// it is no big deal.

		SG_ERR_IGNORE(  SG_fsobj__rmdir_recursive__pathname2(pCtx, pPathMergeTempDir, SG_TRUE)  );
		SG_PATHNAME_NULLFREE(pCtx, pPathMergeTempDir);
	}

    // update the branch if this wd is tied

    if (psz_tied_branch_name)
    {
        SG_ERR_CHECK(  SG_vc_branches__add_head(pCtx, pPendingTree->pRepo, psz_tied_branch_name, psz_hid_new_changeset, pva_wd_parents, SG_FALSE, pq)  );
    }
    else
    {
        // not tied.  so do nothing.
    }

	if (psz_comment && *psz_comment)
		SG_ERR_CHECK(  SG_vc_comments__add(pCtx, pPendingTree->pRepo, psz_hid_new_changeset, psz_comment, pq)  );

	// add any stamps.

	if (psaStamps)
	{
		SG_uint32 iCount = 0;
		SG_uint32 iCountStamps = 0;

		SG_ERR_CHECK(  SG_stringarray__count(pCtx, psaStamps, &iCountStamps )  );
		while (iCount < iCountStamps)
		{
			const char* pszstamp;

			SG_ERR_CHECK(  SG_stringarray__get_nth(pCtx, psaStamps, iCount, &pszstamp)  );
			SG_ERR_CHECK(  SG_vc_stamps__add(pCtx, pPendingTree->pRepo, psz_hid_new_changeset, pszstamp, pq, NULL)  );
			iCount++;
		}
	}

	// Update any work-item-tracking issues.

	if (paszAssocs!=NULL && count_assocs>0)
	{
        SG_ERR_CHECK(  SG_vc_hooks__ASK__WIT__ADD_ASSOCIATIONS(
                    pCtx,
                    pPendingTree->pRepo,
                    pChangesetOriginal,
                    psz_tied_branch_name,
                    pq,
                    psz_comment,
                    paszAssocs,
                    count_assocs,
                    psaStamps
                    )  );
	}

    SG_ERR_CHECK(  SG_vc_hooks__BROADCAST__AFTER_COMMIT(
                pCtx,
                pPendingTree->pRepo,
                pChangesetOriginal,
                psz_tied_branch_name,
                pq,
                psz_comment,
                paszAssocs,
                count_assocs,
                psaStamps
                )  );


fail:
    if (ptx)
    {
        SG_ERR_IGNORE(  SG_committing__abort(pCtx, ptx)  );
    }
    if (pztx)
    {
        SG_ERR_IGNORE(  SG_zing__abort_tx(pCtx, &pztx)  );
    }
    SG_NULLFREE(pCtx, psz_hid_new_changeset);
    SG_NULLFREE(pCtx, psz_tied_branch_name);
	SG_DAGNODE_NULLFREE(pCtx, pDagnodeOriginal);
	SG_CHANGESET_NULLFREE(pCtx, pChangesetOriginal);
	SG_FILE_SPEC_NULLFREE(pCtx, pInternalFilespec);
	SG_PATHNAME_NULLFREE(pCtx, pPathMergeTempDir);
    SG_CHANGESET_NULLFREE(pCtx, pcs_branches_dag);
    SG_DAGNODE_NULLFREE(pCtx, pdn_branches_dag);
    SG_VARRAY_NULLFREE(pCtx, pva_wd_parents);
    SG_VHASH_NULLFREE(pCtx, pvh_locks);
}

//////////////////////////////////////////////////////////////////

#include "sg_pendingtree__private_log.h"
#include "sg_pendingtree__private_revert.h"
#include "sg_pendingtree__private_resolve_issues.h"
#include "sg_pendingtree__private_update_baseline.h"

//////////////////////////////////////////////////////////////////

void SG_pendingtree__free(SG_context* pCtx, SG_pendingtree* pThis)
{
	if (!pThis)
	{
		return;
	}
	
#if TRACE_PENDINGTREE
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "SG_pendingtree__free()\n")  );
#endif

#if 0 && defined(DEBUG)
	// A crude check to see if a caller did a __set_wd_parents()
	// without a __save().  However, it is possible for us to get
	// here as part of a fail-block cleanup/throw, so we don't assert.
	if (pThis->b_pvh_wd__parents__dirty)
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "SG_pendingtree__free: with b_pvh_wd__parents__dirty\n")  );

	// A crude check to see if a caller did a __set_wd_issues()
	// without a __save().  However, it is possible for us to get
	// here as part of a fail-block cleanup/throw, so we don't assert.
	if (pThis->b_pvh_wd__issues__dirty)
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "SG_pendingtree__free: with b_pvh_wd__issues__dirty\n")  );

	// A crude check to see if a caller did a __set_wd_choices()
	// without a __save().  However, it is possible for us to get
	// here as part of a fail-block cleanup/throw, so we don't assert.
	if (pThis->b_pvh_wd__choices__dirty)
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "SG_pendingtree__free: with b_pvh_wd__choices__dirty\n")  );
#endif

	if (pThis->pvf_wd)
	{
		SG_ERR_IGNORE(  SG_vfile__abort(pCtx, &pThis->pvf_wd)  );
	}
	SG_VHASH_NULLFREE(pCtx, pThis->pvh_wd);

#if 1
	if (pThis->prbQueuedPortabilityChecks)
	{
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "TODO someone didn't process queued portability checks...\n")  );
		SG_ASSERT(  (0)  );
	}
#endif
	SG_RBTREE_NULLFREE(pCtx, pThis->prbQueuedPortabilityChecks);

    SG_UTF8_CONVERTER_NULLFREE(pCtx, pThis->pConverter);
	SG_PTNODE_NULLFREE(pCtx, pThis->pSuperRoot);

    SG_STRING_NULLFREE(pCtx, pThis->pstrRepoDescriptorName);
    SG_VARRAY_NULLFREE(pCtx, pThis->pvaWarnings);
	SG_VARRAY_NULLFREE(pCtx, pThis->pvaActionLog);
	SG_PATHNAME_NULLFREE(pCtx, pThis->pPathWorkingDirectoryTop);
	SG_REPO_NULLFREE(pCtx, pThis->pRepo);
	SG_STRPOOL_NULLFREE(pCtx, pThis->pStrPool);

	SG_TIMESTAMP_CACHE_NULLFREE(pCtx, pThis->pTSC);

	SG_PENDINGTREE_UPDATE_DATA_NULLFREE(pCtx, pThis->pud);
	
	SG_NULLFREE(pCtx, pThis);
}

void SG_pendingtree__get_repo(SG_UNUSED_PARAM(SG_context* pCtx),
        SG_pendingtree* pPendingTree,
        SG_repo** ppRepo
        )
{
	SG_UNUSED(  pCtx  );
    *ppRepo = pPendingTree->pRepo;
}

//////////////////////////////////////////////////////////////////

/**
 * Update the information in the tree-diff for a single tree-node-entry
 * using the state in the pending-tree.
 *
 * This is similar to sg_ptnode__status() in spirit.
 */
static void _sg_ptnode__annotate_treediff2(SG_context* pCtx, struct sg_ptnode * ptn,
											   SG_treediff2 * pTreeDiff,
											   SG_bool * pbDirty)
{
	SG_rbtree_iterator * pIterator = NULL;
	SG_diffstatus_flags dsFlags = SG_DIFFSTATUS_FLAGS__ZERO;
	const char * szParentGid_base;
	const char * szParentGid_pend;
	const char * szHidContent_base;
	const char * szHidContent_pend;
	const char * szHidXAttrs_base;
	const char * szHidXAttrs_pend = NULL;
	const char * szEntryName_base;
	const char * szEntryName_pend;
	SG_int64 attrBits_base, attrBits_pend;
	SG_bool bInvalidateFolderContentHid = SG_FALSE;

	SG_NULLARGCHECK_RETURN(ptn);

	// get the baseline and current values of these fields.
	// if they haven't changed, both will point to the same string.
	// these may return a null string if we don't have a value for
	// them (for example when we have an __ADDED entry).

	SG_ERR_CHECK(  sg_ptnode__get_old_name(pCtx, ptn,&szEntryName_base)  );
	SG_ERR_CHECK(  sg_ptnode__get_name(pCtx, ptn,&szEntryName_pend)  );

	// the content HIDs may or may not be set.  in some cases
	// we are lazy and don't compute the HID because we don't
	// need it.  in other case (such as a folder) we *can't*
	// compute it because is sg_treenode's job and we don't
	// necessarily have all of the information needed to do it.

	SG_ERR_CHECK(  sg_ptnode__get_old_hid(pCtx, ptn,&szHidContent_base)  );
	SG_ERR_CHECK(  sg_ptnode__get_hid(pCtx, ptn,&szHidContent_pend)  );

	// the XAttr HIDs may or may not be null because the entry
	// may or may not have XAttrs on disk (and or on this platform).

	SG_ERR_CHECK(  sg_ptnode__get_old_xattrs_hid(pCtx, ptn,&szHidXAttrs_base)  );
	SG_ERR_CHECK(  sg_ptnode__get_xattrs_hid(pCtx, ptn,&szHidXAttrs_pend)  );

	attrBits_pend = ptn->iCurrentAttributeBits;
	if (ptn->pBaselineEntry)
		SG_ERR_CHECK(  SG_treenode_entry__get_attribute_bits(pCtx, ptn->pBaselineEntry,&attrBits_base)  );
	else
		attrBits_base = attrBits_pend;
	
	szParentGid_pend = ((ptn->pCurrentParent) ? ptn->pCurrentParent->pszgid : NULL);
	szParentGid_base = ((ptn->pszgidMovedFrom) ? ptn->pszgidMovedFrom : szParentGid_pend);


	if (ptn->saved_flags & sg_PTNODE_SAVED_FLAG_DELETED)
	{
		if (ptn->temp_flags & sg_PTNODE_TEMP_FLAG_IMPLICIT_DELETE)
			dsFlags |= SG_DIFFSTATUS_FLAGS__LOST;
		else
		{
			dsFlags |= SG_DIFFSTATUS_FLAGS__DELETED;
			if (ptn->bDeletedAfterAdded)
				dsFlags |= SG_DIFFSTATUS_FLAGS__HACK__DELETED_ADDED;
		}
	}

	else if (ptn->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY)
	{
		// this is the ghost of the move.  the node for the entry
		// in the from-directory.  it is not in the tree proper, but this
		// is a placeholder for where it was.
		//
		// we do *not* set a dsFlag for this because we don't want the
		// ghost in the tree-diff data structures.  it will get the version
		// of the entry in the to-directory (and __MOVED will be set).

#if 0
		// TODO convert this to a trace message
		SG_console(SG_CS_STDERR,"MOVED_AWAY: [gid %s][name_base %s][name_pend %s][cur %s]\n",
				   ptn->pszgid,
				   ((szEntryName_base) ? szEntryName_base : "(NULL)"),
				   ((szEntryName_pend) ? szEntryName_pend : "(NULL)"),
				   ((ptn->pszCurrentName) ? ptn->pszCurrentName : "(NULL)"));

#endif
		dsFlags = SG_DIFFSTATUS_FLAGS__ZERO;
		goto skip_all;
	}

	if (   !(ptn->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY)
	    && ptn->pszCurrentName
		&& ptn->pBaselineEntry)
	{
#if 0
		// TODO convert this to a trace message
		SG_console(SG_CS_STDERR,"XXXXX: [name_base %s][name_pend %s][cur %s]\n",
				   ((szEntryName_base) ? szEntryName_base : "(NULL)"),
				   ((szEntryName_pend) ? szEntryName_pend : "(NULL)"),
				   ((ptn->pszCurrentName) ? ptn->pszCurrentName : "(NULL)"));
#endif
		if (strcmp( ((szEntryName_base) ? szEntryName_base : ""),
					((szEntryName_pend) ? szEntryName_pend : "")) != 0)
			dsFlags |= SG_DIFFSTATUS_FLAGS__RENAMED;
	}

	if (ptn->pCurrentParent && !ptn->pBaselineEntry)
	{
		// !pBaselineEntry means that the item did not
		// appear in the baseline and must be an ADDED
		// or FOUND.

		if (dsFlags & SG_DIFFSTATUS_FLAGS__LOST)
		{
			// if the user does:
			//     vv add foo
			//     /bin/rm foo
			// we have something that looks like ADDED+LOST
			// but we don't want that (both are bits in the
			// mutually-exclusive section for one thing) and
			// it kinda looks silly for another.  i'm not
			// sure that there is a *right* answer, so for
			// now, let's let LOST win.

			dsFlags |= SG_DIFFSTATUS_FLAGS__HACK__LOST_ADDED;

#if TRACE_PENDINGTREE
			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
									   "AnnotateTreediff: BOGUS ADDED+LOST or FOUND+LOST: [gid %s] %s\n",
									   ptn->pszgid,
									   ((szEntryName_pend) ? szEntryName_pend : "(null)"))  );
#endif
		}
		else if (dsFlags & SG_DIFFSTATUS_FLAGS__DELETED)
		{
			// if MERGE added the item and the user then DELETED it,
			// we mark it.  we can't just delete the ptnode because
			// STATUS and RESOLVE need it.

			SG_ASSERT(  (dsFlags & SG_DIFFSTATUS_FLAGS__HACK__DELETED_ADDED)  );
		}
		else
		{
			if (ptn->temp_flags & sg_PTNODE_TEMP_FLAG_IMPLICIT_ADD)
				dsFlags |= SG_DIFFSTATUS_FLAGS__FOUND;
			else
				dsFlags |= SG_DIFFSTATUS_FLAGS__ADDED;
		}
	}

	if (ptn->pszgidMovedFrom)
	{
		dsFlags |= SG_DIFFSTATUS_FLAGS__MOVED;
	}

	if (ptn->pBaselineEntry &&  !(ptn->saved_flags & sg_PTNODE_SAVED_FLAG_DELETED))
	{
		if (strcmp( ((szHidXAttrs_base) ? szHidXAttrs_base : ""),
					((szHidXAttrs_pend) ? szHidXAttrs_pend : "")) != 0)
			dsFlags |= SG_DIFFSTATUS_FLAGS__CHANGED_XATTRS;

		if (attrBits_base != attrBits_pend)
			dsFlags |= SG_DIFFSTATUS_FLAGS__CHANGED_ATTRBITS;
	}

	if (ptn->type == SG_TREENODEENTRY_TYPE_DIRECTORY)
	{
		if (ptn->prbItems)
		{
            SG_bool bChildrenDirty = SG_FALSE;
			struct sg_ptnode * ptnChild;
			const char * pszgidChild;
			SG_bool bChildDirty;
			SG_bool b;

			// dive into the files/sub-directories within this directory and let them
			// annotate the tree-diff.  because we dive first and wait to see if any
			// of them were dirty before we annotate the entry for this directory, the
			// tree-diff will get our children before it gets our data.  this may cause
			// tree-diff a little problem because it normally gets data completely
			// top-down.
			//
			// the reason we defer the annotation of this directory is that we only
			// want to annotate dirty nodes -- we still want to try to keep the
			// tree-diff code from loading the entire tree if it doesn't have to
			// (we want that equal-content-hid-short-circuit to keep working).

			SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pIterator, ptn->prbItems, &b, &pszgidChild, (void**) &ptnChild)  );
			while (b)
			{
				SG_ERR_CHECK(  _sg_ptnode__annotate_treediff2(pCtx, ptnChild,pTreeDiff,&bChildDirty)  );
				bChildrenDirty |= bChildDirty;

				// if this child is a move-ghost, we also need to mark the folder dirty.
				// we don't get this automatically from the recursive call because ghosts
				// don't set any flags.

				bChildrenDirty |= ((ptnChild->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY) != 0);

				SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pIterator, &b, &pszgidChild, (void**) &ptnChild)  );
			}
			SG_RBTREE_ITERATOR_NULLFREE(pCtx, pIterator);

			// if one or more children changed, the Content-HID of this directory will change.
			// if none of the other MASK_ME bits are set, force the _MODIFIED bit.  this is a
			// bit contrived -- because the status/treediff code doesn't usually look at a _MODIFIED
			// state on a folder -- but it ensures that they get the message that it is dirty.

			if (bChildrenDirty && ((dsFlags & SG_DIFFSTATUS_FLAGS__MASK_MUTUALLY_EXCLUSIVE) == 0))
			{
				dsFlags |= SG_DIFFSTATUS_FLAGS__MODIFIED;
				bInvalidateFolderContentHid = SG_TRUE;
			}
		}
	}
	else if (ptn->type == SG_TREENODEENTRY_TYPE_REGULAR_FILE)
	{
		if (ptn->pBaselineEntry && !(ptn->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY))
		{
			SG_bool bFileIsDirty = SG_FALSE;

			SG_ERR_CHECK(  sg_ptnode__is_file_dirty(pCtx, ptn, &bFileIsDirty)  );
			if (bFileIsDirty)
				dsFlags |= SG_DIFFSTATUS_FLAGS__MODIFIED;
		}
	}
	else if (ptn->type == SG_TREENODEENTRY_TYPE_SYMLINK)
	{
		if (ptn->pBaselineEntry && !(ptn->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY))
		{
			SG_bool bSymlinkIsDirty = SG_FALSE;

			SG_ERR_CHECK(  sg_ptnode__is_symlink_dirty(pCtx, ptn, &bSymlinkIsDirty)  );
			if (bSymlinkIsDirty)
				dsFlags |= SG_DIFFSTATUS_FLAGS__MODIFIED;
		}
	}
	else if (ptn->type == SG_TREENODEENTRY_TYPE_SUBMODULE)
	{
		// TODO 2010/12/20 i added OR _DELETED to keep us from trying to see if the sub
		// TODO            is dirty and not bother setting the __MODIFIED flag.  but i
		// TODO            wonder if we shouldn't also do that for the file/symlink cases too.
		if (ptn->pBaselineEntry && !(ptn->saved_flags & (sg_PTNODE_SAVED_FLAG_MOVED_AWAY|sg_PTNODE_SAVED_FLAG_DELETED)))
		{
			SG_bool bDirty_Reference;
			SG_bool bDirty_Pending;

			SG_ERR_CHECK(  sg_ptnode__is_submodule_changed_or_dirty(pCtx, ptn, &bDirty_Reference, &bDirty_Pending)  );
			if (bDirty_Reference || bDirty_Pending)
				dsFlags |= SG_DIFFSTATUS_FLAGS__MODIFIED;
		}
	}
	else
	{
		SG_ERR_THROW_RETURN( SG_ERR_INVALID_PENDINGTREE_OBJECT_TYPE );
	}

	// if this node is dirty, tell the tree-diff about it.

	if (dsFlags)
	{
		// because we get called with the super-root (above "@/"),
		// we need to skip that one.

		if (ptn->pszgid)
		{
			// TODO 2011/03/21 Add debug code to verify that only one bit in
			// TODO            the mutually-exclusive section is set.

			// if we are a folder and anything changed within it, we invalidate the
			// content HID because we cannot compute it here.  the treediff code
			// would like to have it, but it can cope.

			if (bInvalidateFolderContentHid)
				szHidContent_pend = NULL;

			// TODO for files/symlinks compute szHidContent_pend if not set.
			// TODO 2010/11/16 Is the above comment for files/symlinks still valid?
			// TODO 2010/11/16 Do we also want to compute/not-compute it for submodules?

			SG_ERR_CHECK(  SG_treediff2__pendingtree_annotate_cb(pCtx, pTreeDiff,dsFlags,
																 ptn->pszgid,
																 szParentGid_base,szParentGid_pend,
																 ptn->type,
																 szEntryName_base,szEntryName_pend,
																 szHidContent_base,szHidContent_pend,
																 szHidXAttrs_base,szHidXAttrs_pend,
																 attrBits_base,attrBits_pend,
																 ptn->pTSM)  );
		}
	}


skip_all:
	*pbDirty = (dsFlags != SG_DIFFSTATUS_FLAGS__ZERO);
	SG_ASSERT( !SG_context__has_err(pCtx) );
	return;

fail:
	SG_RBTREE_ITERATOR_NULLFREE(pCtx, pIterator);
}

//////////////////////////////////////////////////////////////////

void SG_pendingtree__get_wd_parents__ref(SG_context * pCtx,
										 SG_pendingtree * pPendingTree,
										 const SG_varray ** ppva_wd_parents)
{
	SG_varray * pvaSrc;

	SG_NULLARGCHECK_RETURN(pPendingTree );
	SG_NULLARGCHECK_RETURN(ppva_wd_parents);

	SG_ASSERT(  (pPendingTree->pvh_wd)  );		// __load_cached_or_baseline() should have been called

	// this will THROW if the vfile didn't know what the parents are.
	SG_ERR_CHECK_RETURN(  SG_vhash__get__varray(pCtx, pPendingTree->pvh_wd, "parents", &pvaSrc)  );

	*ppva_wd_parents = pvaSrc;					// you can borrow our copy, but don't change it.
}

void SG_pendingtree__get_wd_parents(SG_context * pCtx,
									SG_pendingtree * pPendingTree,
									SG_varray ** ppva_wd_parents)
{
	SG_varray * pvaAllocated = NULL;
	const SG_varray * pvaSrc = NULL;
	const char * pszHid_k = NULL;
	SG_uint32 k, nr;

	SG_NULLARGCHECK_RETURN(pPendingTree );
	SG_NULLARGCHECK_RETURN(ppva_wd_parents);

	SG_VARRAY__ALLOC(pCtx, &pvaAllocated);

	SG_ERR_CHECK(  SG_pendingtree__get_wd_parents__ref(pCtx, pPendingTree, &pvaSrc)  );
	SG_ERR_CHECK(  SG_varray__count(pCtx, pvaSrc, &nr)  );
	for (k=0; k<nr; k++)
	{
		SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pvaSrc, k, &pszHid_k)  );
		SG_ERR_CHECK(  SG_varray__append__string__sz(pCtx, pvaAllocated, pszHid_k)  );
	}

	*ppva_wd_parents = pvaAllocated;
	return;

fail:
	SG_VARRAY_NULLFREE(pCtx, pvaAllocated);
}

void SG_pendingtree__set_wd_parents(SG_context * pCtx,
									SG_pendingtree * pPendingTree,
									SG_varray ** ppva_new_wd_parents)
{
	SG_NULLARGCHECK_RETURN(pPendingTree);
	SG_NULL_PP_CHECK_RETURN(ppva_new_wd_parents);

	SG_ASSERT(  (pPendingTree->pvh_wd)  );

#if TRACE_PENDINGTREE
	{
		SG_uint32 k, nrParents;
		const char * psz_hid_parent_k;

		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "SG_pendingtree__set_wd_parents:\n")  );
		SG_ERR_IGNORE(  SG_varray__count(pCtx, *ppva_new_wd_parents, &nrParents)  );
		for (k=0; k<nrParents; k++)
		{
			SG_ERR_IGNORE(  SG_varray__get__sz(pCtx, *ppva_new_wd_parents, k, &psz_hid_parent_k)  );
			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "        [%d]: %s\n", k, psz_hid_parent_k)  );
		}
	}
#endif
	// steal the callers VArray and stuff it into the VHash.

	SG_ERR_CHECK_RETURN(  SG_vhash__update__varray(pCtx, pPendingTree->pvh_wd, "parents", ppva_new_wd_parents)  );

#if defined(DEBUG)
	pPendingTree->b_pvh_wd__parents__dirty = SG_TRUE;
#endif
}

void SG_pendingtree__set_single_wd_parent(SG_context * pCtx,
										  SG_pendingtree * pPendingTree,
										  const char * psz_hid_parent)
{
	SG_varray * pva = NULL;

	SG_NULLARGCHECK_RETURN(pPendingTree);
	SG_NONEMPTYCHECK_RETURN(psz_hid_parent);

	SG_ERR_CHECK(  SG_VARRAY__ALLOC(pCtx, &pva)  );
	SG_ERR_CHECK(  SG_varray__append__string__sz(pCtx, pva, psz_hid_parent)  );
	SG_ERR_CHECK(  SG_pendingtree__set_wd_parents(pCtx, pPendingTree, &pva)  );		// this steals our varray

	return;

fail:
	SG_VARRAY_NULLFREE(pCtx, pva);
}

//////////////////////////////////////////////////////////////////

void SG_pendingtree__clear_wd_issues(SG_context * pCtx,
									 SG_pendingtree * pPendingTree)
{
	// if the "issues" section is present, remove it.

	SG_vhash__remove(pCtx, pPendingTree->pvh_wd, "issues");
	if (SG_context__has_err(pCtx))
	{
		if (SG_context__err_equals(pCtx, SG_ERR_VHASH_KEYNOTFOUND))
			SG_context__err_reset(pCtx);
		else
			SG_ERR_RETHROW_RETURN;
	}

#if defined(DEBUG)
	pPendingTree->b_pvh_wd__issues__dirty = SG_TRUE;
#endif
}

void SG_pendingtree__set_wd_issues(SG_context * pCtx,
								   SG_pendingtree * pPendingTree,
								   SG_varray ** ppva_new_wd_issues)
{
	SG_uint32 count;

	SG_NULLARGCHECK_RETURN(pPendingTree);
	// ppva_new_wd_issues can be null or refer to a null vhash.

	SG_ASSERT(  (pPendingTree->pvh_wd)  );

	SG_ERR_CHECK_RETURN(  SG_pendingtree__clear_wd_issues(pCtx, pPendingTree)  );

	if (!ppva_new_wd_issues)
		return;

	if (!*ppva_new_wd_issues)
		return;

	SG_ERR_CHECK_RETURN(  SG_varray__count(pCtx, (*ppva_new_wd_issues), &count)  );
	if (count > 0)
		SG_ERR_CHECK_RETURN(  SG_vhash__update__varray(pCtx, pPendingTree->pvh_wd, "issues", ppva_new_wd_issues)  );	// this steals our caller's varray
	else
		SG_VARRAY_NULLFREE(pCtx, (*ppva_new_wd_issues));	// this steals our caller's varray

#if defined(DEBUG)
	pPendingTree->b_pvh_wd__issues__dirty = SG_TRUE;
#endif
}

void SG_pendingtree__get_wd_issues__ref(SG_context * pCtx,
										SG_pendingtree * pPendingTree,
										SG_bool * pbHaveIssues,
										const SG_varray ** ppva_wd_issues)
{
	SG_bool bHaveIssues = SG_FALSE;
	SG_varray * pva = NULL;			// we do not own this

	SG_NULLARGCHECK_RETURN(pPendingTree);
	SG_NULLARGCHECK_RETURN(pbHaveIssues);
	SG_NULLARGCHECK_RETURN(ppva_wd_issues);

	SG_ASSERT(  (pPendingTree->pvh_wd)  );

	// return a read-only pointer to the in-memory version of the "issues"
	// section in wd.json (if it exists).
	//
	// if there is no "issues" section or if it points to a null vhash,
	// assume that we have no issues to report.

	SG_vhash__get__varray(pCtx, pPendingTree->pvh_wd, "issues", &pva);
	if (SG_context__has_err(pCtx))
	{
		if (SG_context__err_equals(pCtx, SG_ERR_VHASH_KEYNOTFOUND))
			SG_context__err_reset(pCtx);
		else
			SG_ERR_RETHROW_RETURN;
	}
	else
	{
		if (pva)
		{
			SG_uint32 count;

			SG_ERR_CHECK_RETURN(  SG_varray__count(pCtx, pva, &count)  );
			bHaveIssues = (count > 0);
		}
	}

	*pbHaveIssues = bHaveIssues;
	*ppva_wd_issues = pva;
}

void SG_pendingtree__set_wd_choices(
	SG_context*     pCtx,
	SG_pendingtree* pPendingTree,
	SG_vhash**      ppChoices
	)
{
	SG_NULLARGCHECK(pPendingTree);
	SG_ASSERT(  (pPendingTree->pvh_wd)  );

	// if the "choices" section is present in our wd data, remove it.
	SG_vhash__remove(pCtx, pPendingTree->pvh_wd, "choices");
	SG_ERR_CHECK_CURRENT_DISREGARD(SG_ERR_VHASH_KEYNOTFOUND);

	// if we have new choices to save, do that
	if (ppChoices != NULL && *ppChoices != NULL)
	{
		SG_uint32 uCount;

		// check if the hash is empty
		SG_ERR_CHECK(  SG_vhash__count(pCtx, *ppChoices, &uCount)  );
		if (uCount > 0u)
		{
			// this steals our caller's vhash
			SG_ERR_CHECK(  SG_vhash__update__vhash(pCtx, pPendingTree->pvh_wd, "choices", ppChoices)  );
		}
		else
		{
			// we're not going to give the caller's vhash to our own
			// but we still became the owner of it, so just free it
			SG_VHASH_NULLFREE(pCtx, *ppChoices);
		}
	}

#if defined(DEBUG)
	// mark that the choices are dirty
	// this allows us to verify that they're clean (saved) before we're freed
	pPendingTree->b_pvh_wd__choices__dirty = SG_TRUE;
#endif

fail:
	return;
}

void SG_pendingtree__get_wd_choices__ref(
	SG_context*      pCtx,
	SG_pendingtree*  pPendingTree,
	SG_bool*         pHaveChoices,
	const SG_vhash** ppChoices
	)
{
	SG_bool   bHaveChoices = SG_FALSE;
	SG_vhash* pChoices     = NULL;

	SG_NULLARGCHECK(pPendingTree);
	SG_ASSERT(  (pPendingTree->pvh_wd)  );

	SG_vhash__get__vhash(pCtx, pPendingTree->pvh_wd, "choices", &pChoices);
	SG_ERR_CHECK_CURRENT_DISREGARD(SG_ERR_VHASH_KEYNOTFOUND);

	if (pChoices != NULL)
	{
		SG_uint32 uCount;

		SG_ERR_CHECK_RETURN(  SG_vhash__count(pCtx, pChoices, &uCount)  );
		if (uCount > 0u)
		{
			bHaveChoices = SG_TRUE;
		}
		else
		{
			bHaveChoices = SG_FALSE;
		}
	}

	if (pHaveChoices != NULL)
	{
		*pHaveChoices = bHaveChoices;
	}
	if (ppChoices != NULL)
	{
		*ppChoices = pChoices;
	}

fail:
	return;
}

void SG_pendingtree__set_wd_issue_status__dont_save_pendingtree(SG_context * pCtx,
																SG_pendingtree * pPendingTree,
																const SG_vhash * pvhIssue,
																SG_pendingtree_wd_issue_status status)
{
	// update the pendingtree.issues[nr].status directly (without copying the issues varray).
	//
	// only reason we need this rather than letting the caller just update the pvhIssue
	// directly is so that we get the dirty bit set here.

	SG_NULLARGCHECK_RETURN(pPendingTree);
	SG_NULLARGCHECK_RETURN(pvhIssue);

	SG_ASSERT(  (pPendingTree->pvh_wd)  );

	SG_ERR_CHECK_RETURN(  SG_vhash__update__int64(pCtx, (SG_vhash *)pvhIssue, "status", (SG_int64)status)  );

#if defined(DEBUG)
	pPendingTree->b_pvh_wd__issues__dirty = SG_TRUE;
#endif
}

void SG_pendingtree__set_wd_issue_status(SG_context * pCtx,
										 SG_pendingtree * pPendingTree,
										 const SG_vhash * pvhIssue,
										 SG_pendingtree_wd_issue_status status)
{
	SG_ERR_CHECK_RETURN(  SG_pendingtree__set_wd_issue_status__dont_save_pendingtree(pCtx, pPendingTree, pvhIssue, status)  );
	SG_ERR_CHECK_RETURN(  SG_pendingtree__save(pCtx, pPendingTree)  );
}

void SG_pendingtree__set_wd_issue_resolve_mergetool_status(SG_context * pCtx,
														   SG_pendingtree * pPendingTree,
														   const SG_vhash * pvhIssue,
														   const char * pszResolveMergeToolName,
														   SG_int32 statusResolveMergeTool)
{
	// update pendingtree.issues[nr].resolve_mergetool
	// and    pendingtree.issues[nr].resolve_mergetool_status

	SG_NULLARGCHECK_RETURN(pPendingTree);
	SG_NULLARGCHECK_RETURN(pvhIssue);

	SG_ASSERT(  (pPendingTree->pvh_wd)  );

	SG_ERR_CHECK_RETURN(  SG_vhash__update__string__sz(pCtx, (SG_vhash *)pvhIssue, "resolve_mergetool", pszResolveMergeToolName)  );
	SG_ERR_CHECK_RETURN(  SG_vhash__update__int64(pCtx, (SG_vhash *)pvhIssue, "resolve_mergetool_status", (SG_int64)statusResolveMergeTool)  );

#if defined(DEBUG)
	pPendingTree->b_pvh_wd__issues__dirty = SG_TRUE;
#endif

	SG_ERR_CHECK_RETURN(  SG_pendingtree__save(pCtx, pPendingTree)  );
}

void SG_pendingtree__set_wd_issue_resolve_structural_status(SG_context * pCtx,
															SG_pendingtree * pPendingTree,
															const SG_vhash * pvhIssue,
															SG_bool bValue)
{
	// update pendingtree.issues[nr].resolve_structural_fixed

	SG_NULLARGCHECK_RETURN(pPendingTree);
	SG_NULLARGCHECK_RETURN(pvhIssue);

	SG_ASSERT(  (pPendingTree->pvh_wd)  );

	SG_ERR_CHECK_RETURN(  SG_vhash__update__bool(pCtx, (SG_vhash *)pvhIssue, "resolve_structural_fixed", bValue)  );

#if defined(DEBUG)
	pPendingTree->b_pvh_wd__issues__dirty = SG_TRUE;
#endif

	SG_ERR_CHECK_RETURN(  SG_pendingtree__save(pCtx, pPendingTree)  );
}

void SG_pendingtree__count_unresolved_wd_issues(SG_context * pCtx,
												SG_pendingtree * pPendingTree,
												SG_uint32 * pNrUnresolved)
{
	SG_varray * pvaIssues;
	SG_uint32 nrUnresolved = 0;
	SG_bool bHasIssues;

	SG_NULLARGCHECK_RETURN(pPendingTree);
	SG_NULLARGCHECK_RETURN(pNrUnresolved);

	SG_ASSERT(  (pPendingTree->pvh_wd)  );

	SG_ERR_CHECK_RETURN(  SG_vhash__has(pCtx, pPendingTree->pvh_wd, "issues", &bHasIssues)  );
	if (bHasIssues)
	{
		SG_vhash * pvhIssue;
		SG_uint32 k, count;

		SG_ERR_CHECK_RETURN(  SG_vhash__get__varray(pCtx, pPendingTree->pvh_wd, "issues", &pvaIssues)  );
		SG_ERR_CHECK_RETURN(  SG_varray__count(pCtx, pvaIssues, &count)  );
		if (count > 0)
		{
			for (k=0; k<count; k++)
			{
				SG_int64 s;
				SG_pendingtree_wd_issue_status status;

				SG_ERR_CHECK_RETURN(  SG_varray__get__vhash(pCtx, pvaIssues, k, &pvhIssue)  );
				SG_ERR_CHECK_RETURN(  SG_vhash__get__int64(pCtx, pvhIssue, "status", &s)  );
				status = (SG_pendingtree_wd_issue_status)s;

				if ((status & SG_ISSUE_STATUS__MARKED_RESOLVED) == 0)
					nrUnresolved++;
			}
		}
	}

	*pNrUnresolved = nrUnresolved;
}

void SG_pendingtree__find_wd_issue_by_gid(SG_context * pCtx,
										  SG_pendingtree * pPendingTree,
										  const char * pszGidToFind,
										  SG_bool * pbFound,
										  const SG_vhash ** ppvhIssue)
{
	SG_bool bHaveIssues = SG_FALSE;
	const SG_varray * pvaIssues;

	SG_NULLARGCHECK_RETURN(pPendingTree);
	SG_NONEMPTYCHECK_RETURN(pszGidToFind);
	SG_NULLARGCHECK_RETURN(pbFound);
	// ppvhIssue is optional

	SG_ERR_CHECK(  SG_pendingtree__get_wd_issues__ref(pCtx, pPendingTree, &bHaveIssues, &pvaIssues)  );
	if (bHaveIssues)
	{
		SG_uint32 k, nrIssues;

		SG_ERR_CHECK(  SG_varray__count(pCtx, pvaIssues, &nrIssues)  );
		for (k=0; k<nrIssues; k++)
		{
			const SG_vhash * pvhIssue_k;
			const char * pszGid_k;

			SG_ERR_CHECK(  SG_varray__get__vhash(pCtx, pvaIssues, k, (SG_vhash **)&pvhIssue_k)  );
			SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pvhIssue_k, "gid", &pszGid_k)  );
			if (strcmp(pszGid_k, pszGidToFind) == 0)
			{
				*pbFound = SG_TRUE;
				if (ppvhIssue)
					*ppvhIssue = pvhIssue_k;
				return;
			}
		}
	}

	*pbFound = SG_FALSE;

fail:
	return;
}

//////////////////////////////////////////////////////////////////

void SG_pendingtree__set_mrg_temp_dir(SG_context * pCtx,
									  SG_pendingtree * pPendingTree,
									  const SG_pathname * pPathMergeTempDir)
{
	SG_string * pStringRepoPath = NULL;

	SG_NULLARGCHECK_RETURN(pPendingTree);
	// pPathMergeTempDir is optional; if null we unset it in WD.JSON.

	SG_ASSERT(  (pPendingTree->pvh_wd)  );

	if (pPathMergeTempDir)
	{
		SG_ERR_CHECK(  SG_workingdir__wdpath_to_repopath(pCtx,
														 pPendingTree->pPathWorkingDirectoryTop,
														 pPathMergeTempDir,
														 SG_FALSE,
														 &pStringRepoPath)  );
		SG_ERR_CHECK(  SG_vhash__update__string__sz(pCtx,
													pPendingTree->pvh_wd,
													"merge_temp_dir",
													SG_string__sz(pStringRepoPath))  );
	}
	else
	{
		// delete the key/value if present.

		SG_ERR_IGNORE(  SG_vhash__remove(pCtx,
										 pPendingTree->pvh_wd,
										 "merge_temp_dir")  );
	}

#if defined(DEBUG)
	// I don't think we need a b_pvh_wd_*_dirty flag for this field.
#endif

fail:
	SG_STRING_NULLFREE(pCtx, pStringRepoPath);
}

void SG_pendingtree__get_mrg_temp_dir(SG_context * pCtx,
									  SG_pendingtree * pPendingTree,
									  SG_bool * pbHasOne,
									  SG_pathname ** ppPathMergeTempDir)
{
	SG_pathname * pPath = NULL;
	const char * pszRepoPath;
	SG_bool bHasOne = SG_FALSE;

	SG_NULLARGCHECK_RETURN(pPendingTree);
	SG_ASSERT(  (pPendingTree->pvh_wd)  );

	SG_ERR_CHECK(  SG_vhash__has(pCtx, pPendingTree->pvh_wd, "merge_temp_dir", &bHasOne)  );
	if (bHasOne && ppPathMergeTempDir)
	{
		SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pPendingTree->pvh_wd, "merge_temp_dir", &pszRepoPath)  );
		SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path(pCtx,
																			 pPendingTree->pPathWorkingDirectoryTop,
																			 pszRepoPath,
																			 &pPath)  );
	}

	if (pbHasOne)
		*pbHasOne = bHasOne;
	if (ppPathMergeTempDir)
		*ppPathMergeTempDir = pPath;
	return;

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPath);
}

//////////////////////////////////////////////////////////////////

/**
 * do the actual work of computing a tree-diff using the pending-tree.
 *
 * you own the returned tree-diff.
 *
 * WARNING: we call scan_dir() but DO NOT use any of the --include/--exclude
 * file-spec stuff.  Diffs should be filtered afterwards.
 */
static void _sg_pendingtree__do_diff(SG_context* pCtx, SG_pendingtree * pPendingTree,
									 SG_bool bUnloadCleanObjects,
									 const char * szHidArbitraryCSet,		/* optional */
									 SG_treediff2 * pTreeDiff_cache_owner,	/* optional */
									 SG_treediff2 ** ppTreeDiff)
{
	SG_treediff2 * pTreeDiff = NULL;
	const SG_varray * pva_wd_parents = NULL;
	const char * psz_hid_baseline = NULL;
	SG_uint32 nrParents;
	SG_bool bDirty = SG_FALSE;
	SG_bool bDirty2 = SG_FALSE;

	SG_bool bAdd = SG_TRUE;
	SG_bool bRemove = SG_TRUE;
	SG_bool bMarkImplicit = SG_TRUE;
	SG_bool bRecursive = SG_TRUE;
	SG_file_spec* pFilespec = NULL;

	// read the complete tree into memory (read all of the treenodes) into ptnodes and walk the
	// working-directory and make any modifications that we detect.

	SG_ERR_CHECK(  SG_FILE_SPEC__ALLOC(pCtx, &pFilespec)  );
	SG_ERR_CHECK(  sg_pendingtree__scan_dir(pCtx,
											pPendingTree,
											pPendingTree->pPathWorkingDirectoryTop,
											bAdd,
											bRemove,
											bMarkImplicit,
											bRecursive,
											NULL,
											SG_TRUE,		// bIncludeLostItems
											SG_TRUE,		// bIncludeFoundItems
											pFilespec)  );
	SG_FILE_SPEC_NULLFREE(pCtx, pFilespec);

	// if any items were actually added do a portability check
	// on each affected directory.
	SG_ERR_CHECK(  sg_ptnode__do_queued_portability_checks(pCtx, pPendingTree)  );

	if (bUnloadCleanObjects)
	{
		// strip out the clean objects from the ptnodes.

		SG_ERR_CHECK(  sg_ptnode__unload_dir_entries_if_not_dirty(pCtx, pPendingTree->pSuperRoot, &bDirty)  );
	}

	// get the HID of the baseline.

	SG_ERR_CHECK(  SG_pendingtree__get_wd_parents__ref(pCtx, pPendingTree, &pva_wd_parents)  );
	SG_ERR_CHECK(  SG_varray__count(pCtx, pva_wd_parents, &nrParents)  );
	SG_ASSERT(  (nrParents >= 1)  );
	SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pva_wd_parents, 0, &psz_hid_baseline)  );

	// compute tree-diff as "baseline vs working-directory" or "arbitrary-cset vs working-directory".

	SG_ERR_CHECK(  SG_treediff2__alloc__shared(pCtx, pPendingTree->pRepo,pTreeDiff_cache_owner,&pTreeDiff)  );
	SG_ERR_CHECK(  SG_treediff2__begin_pendingtree_diff(pCtx, pTreeDiff,
														szHidArbitraryCSet,
														psz_hid_baseline,
														pPendingTree->pPathWorkingDirectoryTop)  );
	SG_ERR_CHECK(  _sg_ptnode__annotate_treediff2(pCtx, pPendingTree->pSuperRoot,pTreeDiff,&bDirty2)  );
	SG_ERR_CHECK(  SG_treediff2__finish_pendingtree_diff(pCtx, pTreeDiff)  );

	*ppTreeDiff = pTreeDiff;

	return;

fail:
	SG_TREEDIFF2_NULLFREE(pCtx, pTreeDiff);
	SG_FILE_SPEC_NULLFREE(pCtx, pFilespec);
}

//////////////////////////////////////////////////////////////////

/**
 * TODO 2010/08/16 Think about getting rid of the treediff-cache-sharing
 * TODO            stuff.  I'm not sure it's worth the bother here.  At
 * TODO            the time I wrote it, MERGE was going to have multiple
 * TODO            diffs (LCA-vs-each-leaf) and then compare the outputs
 * TODO            so it made sense to share some stuff.  But MERGE went
 * TODO            a different direction, so this doesn't matter as much.
 */
void SG_pendingtree__diff__shared(SG_context* pCtx, SG_pendingtree * pPendingTree,
								  const char * szHidArbitraryCSet,		/* optional */
								  SG_treediff2 * pTreeDiff_cache_owner,	/* optional */
								  SG_treediff2 ** ppTreeDiff)
{
	SG_treediff2 * pTreeDiff = NULL;

	SG_NULLARGCHECK_RETURN(  ppTreeDiff  );

	SG_ERR_CHECK(  _sg_pendingtree__do_diff__unload_clean_objects(pCtx, pPendingTree, szHidArbitraryCSet, pTreeDiff_cache_owner, &pTreeDiff)  );

	/* we don't change the saved pending changeset here. */
	SG_PTNODE_NULLFREE(pCtx, pPendingTree->pSuperRoot);

	*ppTreeDiff = pTreeDiff;

	return;

fail:
	// TODO see note [X] at top of file
	SG_TREEDIFF2_NULLFREE(pCtx, pTreeDiff);
}

/**
 * Compute a tree-diff on the values in the pendingtree.
 *
 * szHidArbitraryCSet, if set, is the HID of an arbitrary CSET and will
 * cause us to do a composite/joined cset-vs-baseline-vs-pendingtree
 * diff.  if not set, we do a simple baseline-vs-pendingtree diff.
 */
static void _sg_pendingtree__diff(SG_context* pCtx, SG_pendingtree * pPendingTree,
								  const char * szHidArbitraryCSet,		/* optional */
								  SG_treediff2 ** ppTreeDiff)
{
	SG_ERR_CHECK_RETURN(  SG_pendingtree__diff__shared(pCtx, pPendingTree,
														szHidArbitraryCSet,
														NULL,
														ppTreeDiff)  );
}

void SG_pendingtree__do_diff__keep_clean_objects(SG_context* pCtx, SG_pendingtree * pPendingTree,
														 const char * szHidArbitraryCSet,		/* optional */
														 SG_treediff2 * pTreeDiff_cache_owner,	/* optional */
														 SG_treediff2 ** ppTreeDiff)
{
	SG_ERR_CHECK_RETURN(  _sg_pendingtree__do_diff(pCtx, pPendingTree,
												   SG_FALSE,
												   szHidArbitraryCSet,
												   pTreeDiff_cache_owner,
												   ppTreeDiff)  );
}

void SG_pendingtree__do_diff__keep_clean_objects__reload(SG_context* pCtx, SG_pendingtree * pPendingTree,
														 const char * szHidArbitraryCSet,		/* optional */
														 SG_treediff2 * pTreeDiff_cache_owner,	/* optional */
														 SG_treediff2 ** ppTreeDiff)
{
	SG_ERR_CHECK_RETURN(  SG_pendingtree__do_diff__keep_clean_objects(pCtx, pPendingTree,
																	  szHidArbitraryCSet,
																	  pTreeDiff_cache_owner,
																	  ppTreeDiff)  );
	SG_ERR_CHECK_RETURN(  SG_pendingtree__reload_pendingtree_from_vfile(pCtx, pPendingTree)  );
}

static void _sg_pendingtree__do_diff__unload_clean_objects(SG_context* pCtx, SG_pendingtree * pPendingTree,
														   const char * szHidArbitraryCSet,		/* optional */
														   SG_treediff2 * pTreeDiff_cache_owner,	/* optional */
														   SG_treediff2 ** ppTreeDiff)
{
	SG_ERR_CHECK_RETURN(  _sg_pendingtree__do_diff(pCtx, pPendingTree,
												   SG_TRUE,
												   szHidArbitraryCSet,
												   pTreeDiff_cache_owner,
												   ppTreeDiff)  );
}

//////////////////////////////////////////////////////////////////

/**
 * take 0, 1, or 2 cset HIDs and create a SG_treediff2 that reflects
 * the changes (committed or pending) between
 * [] 2 csets or
 * [] between 1 cset and the working copy or
 * [] between the baseline and the working copy.
 *
 * WARNING: we DO NOT take any of the --include/--exclude file-spec stuff.
 * We only compute status on a complete view of the tree and not on pre-filtered
 * views of the tree.  (We MIGHT be able to in the future, but not now.)  When we
 * are called with 0 or 1 CSET HID (and need to reference the WD/pendingtree),
 * we cause a sg_pendingtree__scan_dir() to occur; if that is filtered, then when
 * the diff gets computed, we'll be looking at an un-filtered treenode vs a filtered
 * WD which may cause us to report the filtered-out things as missing.  Furthermore,
 * when we are comparing 2 CSET HIDs, we won't have any filtering.  So in all 3
 * cases we would need to hack the treenodes to match the filtering.  SO INSTEAD,
 * we do not do any filtering here while computing the diff/status.  Our caller can
 * use SG_treediff2__filter__using_file_spec() do the dirty work on our results.
 */
void SG_pendingtree__diff_or_status(SG_context * pCtx, SG_pendingtree * pPendingTree,
									const char * psz_cset_1,
									const char * psz_cset_2,
									SG_treediff2 ** ppTreeDiff)
{
	SG_treediff2 * pTreeDiff = NULL;
	SG_bool bHave1 = ((psz_cset_1 != NULL) && (*psz_cset_1 != 0));
	SG_bool bHave2 = ((psz_cset_2 != NULL) && (*psz_cset_2 != 0));

	SG_NULLARGCHECK_RETURN(pPendingTree);
	SG_NULLARGCHECK_RETURN(ppTreeDiff);

	if (!bHave1)
	{
		SG_ASSERT(  (!bHave2)  );

		// baseline vs pendingtree (WD)
		SG_ERR_CHECK(  _sg_pendingtree__diff(pCtx, pPendingTree, NULL, &pTreeDiff)  );

		// Write out any timestamps/HIDs that we may have had to compute.
		// This is independent of whether our caller will eventually save
		// the pendingtree.
		if (pPendingTree->pTSC)
			SG_ERR_CHECK(  SG_timestamp_cache__save(pCtx, pPendingTree->pTSC)  );
	}
	else if (!bHave2)
	{
		// arbitrary-cset vs pendingtree (WD)
		SG_ERR_CHECK(  _sg_pendingtree__diff(pCtx, pPendingTree, psz_cset_1, &pTreeDiff)  );

		// Write out any timestamps/HIDs that we may have had to compute.
		// This is independent of whether our caller will eventually save
		// the pendingtree.
		if (pPendingTree->pTSC)
			SG_ERR_CHECK(  SG_timestamp_cache__save(pCtx, pPendingTree->pTSC)  );
	}
	else
	{
		// cset vs cset
		SG_ERR_CHECK(  SG_treediff2__alloc(pCtx, pPendingTree->pRepo,&pTreeDiff)  );

		// TODO 4/14/10 BUGBUG a cset-vs-cset diff does not need a WD.  we should not have
		// TODO 4/14/10 BUGBUG have to set one here -- and nothing in the diff should use it.
		SG_ERR_CHECK(  SG_treediff2__set_wd_top(pCtx, pTreeDiff,pPendingTree->pPathWorkingDirectoryTop)  );
		SG_ERR_CHECK(  SG_treediff2__compare_cset_vs_cset(pCtx, pTreeDiff, psz_cset_1, psz_cset_2)  );
	}

	*ppTreeDiff = pTreeDiff;
	return;

fail:
    SG_TREEDIFF2_NULLFREE(pCtx, pTreeDiff);
}

//////////////////////////////////////////////////////////////////

void SG_pendingtree__get_gid_from_local_path(SG_context* pCtx, SG_pendingtree* pTree,
											 const SG_pathname* pLocalPathPathname, char** ppszgid)
{
	SG_bool bRoot = SG_FALSE;
	SG_pathname* ppnParentDir = NULL;
	struct sg_ptnode* ptnParentDir = NULL;
	struct sg_ptnode* ptnItem = NULL;
	SG_string* pstrItemName = NULL;

	SG_NULLARGCHECK_RETURN(pTree);
	SG_NULLARGCHECK_RETURN(pLocalPathPathname);

	/* Check if this is the pendingtree's root */
	SG_ERR_CHECK(  SG_pathname__equal_ignoring_final_slashes(pCtx, pLocalPathPathname, pTree->pPathWorkingDirectoryTop, &bRoot)  );
	if (bRoot == SG_TRUE)
	{
		/* Find the root node directly */
		SG_ERR_CHECK(  sg_pendingtree__local_path_to_ptnode(pCtx, pTree, pLocalPathPathname, SG_TRUE, &ptnItem)  );
	}
	else
	{
		/* We want the parent directory of this item.  So copy the path and remove the last component */
		SG_ERR_CHECK(  SG_PATHNAME__ALLOC__COPY(pCtx, &ppnParentDir, pLocalPathPathname)  );
		SG_ERR_CHECK(  SG_pathname__remove_last(pCtx, ppnParentDir)  );

		/* Find the ptnodes */
		SG_ERR_CHECK(  sg_pendingtree__local_path_to_ptnode(pCtx, pTree, ppnParentDir, SG_TRUE, &ptnParentDir)  );
		if (!ptnParentDir)
		{
			SG_ERR_THROW2(  SG_ERR_ITEM_NOT_UNDER_VERSION_CONTROL,
							(pCtx, "Pathname '%s' is not under version control.", SG_pathname__sz(ppnParentDir))  );
		}

		/* Get the base name from the original path */
		SG_ERR_CHECK(  SG_pathname__get_last(pCtx, pLocalPathPathname, &pstrItemName)  );

		/* leave the item in both places, one as a phantom */

		SG_ERR_CHECK(  sg_ptnode__find_entry_by_name_not_recursive__active(pCtx, ptnParentDir, SG_string__sz(pstrItemName), &ptnItem)  );
		if (!ptnItem)
		{
			SG_ERR_THROW2(  SG_ERR_ITEM_NOT_UNDER_VERSION_CONTROL,
							(pCtx, "Pathname '%s' is not under version control.", SG_pathname__sz(pLocalPathPathname))  );
		}
	}

	SG_ERR_CHECK(  SG_STRDUP(pCtx, ptnItem->pszgid, ppszgid)  );

fail:
	SG_PATHNAME_NULLFREE(pCtx, ppnParentDir);
	SG_STRING_NULLFREE(pCtx, pstrItemName);
	return;
}

//////////////////////////////////////////////////////////////////

void SG_pendingtree__get_warnings(SG_context* pCtx,
								  SG_pendingtree* pPendingTree,
								  SG_varray** ppResult)
{
    SG_NULLARGCHECK_RETURN(pPendingTree);
	SG_NULLARGCHECK_RETURN(ppResult);

    *ppResult = pPendingTree->pvaWarnings;
}

void SG_pendingtree__steal_warnings(SG_context* pCtx,
									SG_pendingtree* pPendingTree,
									SG_varray** ppResult)
{
    SG_NULLARGCHECK_RETURN(pPendingTree);
	SG_NULLARGCHECK_RETURN(ppResult);

    *ppResult = pPendingTree->pvaWarnings;
	pPendingTree->pvaWarnings = NULL;
}

void SG_pendingtree__get_port_settings(SG_context * pCtx,
									   SG_pendingtree * pPendingTree,
									   SG_utf8_converter ** ppConverter,
									   SG_varray ** ppvaWarnings,
									   SG_portability_flags * pPortMask)
{
	SG_NULLARGCHECK_RETURN(pPendingTree);

	if (ppConverter)
		*ppConverter = pPendingTree->pConverter;

	if (ppvaWarnings)
		*ppvaWarnings = pPendingTree->pvaWarnings;

	if (pPortMask)
		*pPortMask = pPendingTree->portMask;
}

//////////////////////////////////////////////////////////////////

/**
 * build the pendingtree and dump it to a treediff and an enhanced vhash (like
 * we write to the vfile, but also include extra debug info).  we export both
 * of these at the same time so that the transient GIDs created for FOUND entries
 * match.
 */
void SG_pendingtree_debug__export(SG_context * pCtx, SG_pathname * pPathWorkingDir,
								  SG_bool bProcessIgnores,		// if true do filtering using "ignores" localsettings
								  SG_treediff2 ** ppTreeDiff,
								  SG_vhash ** ppvhExport)
{
	SG_pendingtree * pPendingTree = NULL;
	SG_treediff2 * pTreeDiff = NULL;
	SG_vhash* pvh = NULL;
	SG_file_spec* pFilespec = NULL;

	SG_ERR_CHECK(  SG_PENDINGTREE__ALLOC(pCtx,pPathWorkingDir,&pPendingTree)  );
	SG_ERR_CHECK(  SG_FILE_SPEC__ALLOC(pCtx, &pFilespec)  );

	if (bProcessIgnores == SG_FALSE)
	{
		SG_ERR_CHECK(  SG_file_spec__modify_flags(pCtx, pFilespec, SG_FILE_SPEC__NO_IGNORES, 0u, NULL, NULL)  );
	}

	// scan directory.  strip out clean stuff.  build treediff of baseline-vs-working-directory

	SG_ERR_CHECK(  _sg_pendingtree__do_diff__unload_clean_objects(pCtx,pPendingTree, NULL, NULL, &pTreeDiff)  );
	SG_ERR_CHECK(  SG_treediff2__file_spec_filter(pCtx, pTreeDiff,
												  0, NULL,
												  SG_TRUE,
												  pFilespec)  );

	// convert ptnodes into full vhash (with debug info)

	if (pPendingTree->pSuperRoot)
	{
		pPendingTree->bDebugExport = SG_TRUE;
		SG_ERR_CHECK(  sg_pendingtree__to_vhash(pCtx, pPendingTree, &pvh)  );

		/* we don't change the saved pending changeset here. */
		SG_PTNODE_NULLFREE(pCtx, pPendingTree->pSuperRoot);
	}

	SG_FILE_SPEC_NULLFREE(pCtx, pFilespec);
	SG_PENDINGTREE_NULLFREE(pCtx,pPendingTree);

	*ppTreeDiff = pTreeDiff;
	*ppvhExport = pvh;
	return;

fail:
	// TODO see note [X] at top of file
	SG_FILE_SPEC_NULLFREE(pCtx, pFilespec);
	SG_PENDINGTREE_NULLFREE(pCtx,pPendingTree);
	SG_TREEDIFF2_NULLFREE(pCtx, pTreeDiff);
	SG_VHASH_NULLFREE(pCtx, pvh);
}

#if defined(DEBUG)
/**
 * dump an existing pendingtree to the console.
 */
void SG_pendingtree_debug__dump_existing_to_console(SG_context * pCtx, SG_pendingtree * pPendingTree)
{
	SG_vhash * pvh = NULL;
	SG_bool bDebug = pPendingTree->bDebugExport;

	SG_NULLARGCHECK_RETURN(pPendingTree);

	pPendingTree->bDebugExport = SG_TRUE;
	if (pPendingTree->pSuperRoot)
		SG_ERR_IGNORE(  sg_pendingtree__to_vhash(pCtx,pPendingTree,&pvh)  );
	pPendingTree->bDebugExport = bDebug;

	if (pvh)
		SG_ERR_IGNORE(  SG_vhash_debug__dump_to_console(pCtx,pvh)  );

	SG_VHASH_NULLFREE(pCtx, pvh);
}
#endif

void SG_pendingtree__get_gids_for_paths(SG_context * pCtx, SG_pendingtree * pPendingTree, SG_uint32 count_args, const char ** paszArgs,  SG_stringarray* ppReturnGIDs, SG_bool * bRootIsOneOfThem)
{
	SG_pathname * pPathName = NULL;
	struct sg_ptnode * ptnCurrentItem = NULL;
	struct sg_ptnode * ptn = NULL;
	SG_pathname * pPathDir = NULL;
	SG_string *pstrName = NULL;
	SG_uint32 i = 0;

	*bRootIsOneOfThem = SG_FALSE;
	SG_ERR_CHECK(  SG_pathname__alloc(pCtx, &pPathName)  );

	for (i = 0; i < count_args; i++)
	{
		/* First get a pathname for the item */
		SG_ERR_CHECK(  SG_pathname__set__from_sz(pCtx, pPathName, paszArgs[i])  );
		SG_ERR_IGNORE(  sg_pendingtree__local_path_to_ptnode(pCtx, pPendingTree, pPathName, SG_FALSE, &ptnCurrentItem)  );
		if (ptnCurrentItem == NULL)
		{
			/* We want the parent directory of this item.  So copy the path and remove the last component */
			SG_ERR_CHECK(  SG_PATHNAME__ALLOC__COPY(pCtx, &pPathDir, pPathName)  );
			SG_ERR_CHECK(  SG_pathname__remove_last(pCtx, pPathDir)  );

			/* Find the directory's ptnode */
			SG_ERR_CHECK(  sg_pendingtree__local_path_to_ptnode(pCtx, pPendingTree, pPathDir, SG_TRUE, &ptn)  );

			/* Get the base name from the original path */
			SG_ERR_CHECK(  SG_pathname__get_last(pCtx, pPathName, &pstrName)  );

			/* Find the ptnode */
			SG_ERR_CHECK(  sg_ptnode__find_entry_by_name_not_recursive__active(pCtx, ptn, SG_string__sz(pstrName), &ptnCurrentItem)  );
			if (SG_string__length_in_bytes(pstrName) == 0)
				ptnCurrentItem = ptn;
			if (!ptnCurrentItem)
				SG_ERR_THROW2(  SG_ERR_NOT_FOUND,
								(pCtx,"Looking for '%s' in the pending tree node for '%s'",
								 SG_string__sz(pstrName),
								 SG_pathname__sz(pPathDir))  );
			SG_PATHNAME_NULLFREE(pCtx, pPathDir);
			SG_STRING_NULLFREE(pCtx, pstrName);
		}
		if (ptnCurrentItem->pCurrentParent->pCurrentParent == NULL)
			*bRootIsOneOfThem = SG_TRUE;
		//SG_ERR_CHECK(  sg_pendingtree__local_path_to_ptnode(pCtx, pPendingTree, pPathName, SG_TRUE, &ptnCurrentItem)  );
		SG_ERR_CHECK(  SG_stringarray__add(pCtx, ppReturnGIDs, ptnCurrentItem->pszgid)  );
	}
	SG_PATHNAME_NULLFREE(pCtx, pPathName);
	SG_PATHNAME_NULLFREE(pCtx, pPathDir);
	SG_STRING_NULLFREE(pCtx, pstrName);
	return;
fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathName);
	SG_PATHNAME_NULLFREE(pCtx, pPathDir);
	SG_STRING_NULLFREE(pCtx, pstrName);
	return;

}

//////////////////////////////////////////////////////////////////

void SG_pendingtree_debug__get_wd_dot_json_stats(SG_context * pCtx,
												 const SG_pathname * pPathWorkingDir,
												 SG_uint32 * pNrParents,
												 SG_bool * pbHavePending)
{
	SG_pathname * pPathActualTop = NULL;
	SG_pathname * pPath = NULL;
	SG_string * pStrRepoDescriptorName = NULL;
	char * pszGidAnchorDirectory = NULL;
	SG_vhash * pvh_wd = NULL;
	SG_varray * pva_parents;	// we do not own this
	SG_bool bFound;
	
	SG_ERR_CHECK(  SG_workingdir__find_mapping(pCtx, pPathWorkingDir, &pPathActualTop, &pStrRepoDescriptorName, &pszGidAnchorDirectory)  );
	SG_ERR_CHECK(  SG_workingdir__get_drawer_path(pCtx, pPathActualTop, &pPath)  );
	SG_ERR_CHECK(  SG_pathname__append__from_sz(pCtx, pPath, "wd.json")  );
	SG_ERR_CHECK(  SG_vfile__slurp(pCtx, pPath, &pvh_wd)  );

	*pNrParents = 0;
	*pbHavePending = SG_FALSE;

	if (!pvh_wd)
		goto done;

	// "pending" is only present when there is dirt.
	SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh_wd, "pending", pbHavePending)  );

	// "parents" should always be present.  (return zero if not)
	SG_ERR_CHECK(  SG_vhash__has(pCtx, pvh_wd, "parents", &bFound)  );
	if (bFound)
	{
		SG_ERR_CHECK(  SG_vhash__get__varray(pCtx, pvh_wd, "parents", &pva_parents)  );
		SG_ERR_CHECK(  SG_varray__count(pCtx, pva_parents, pNrParents)  );
	}

done:
fail:
	SG_VHASH_NULLFREE(pCtx, pvh_wd);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_PATHNAME_NULLFREE(pCtx, pPathActualTop);
	SG_STRING_NULLFREE(pCtx, pStrRepoDescriptorName);
	SG_NULLFREE(pCtx, pszGidAnchorDirectory);
}

//////////////////////////////////////////////////////////////////

/**
 * HACK Un-add an entry from the pendingtree BUT DO NOT REMOVE IT FROM DISK.
 * HACK This is like "vv revert foo" following a "vv add foo", but
 * HACK without all the baggage of the full REVERT mechanism.  I need this
 * HACK routine in the middle of the MERGE code to get rid of a temporary
 * HACK directory.  As such, I have lots of open changes in the in-memory
 * HACK pendingtree and since the REVERT code assumes that it is in complete
 * HACK control of the pendingtree, I don't want to have to save my dirty
 * HACK version so that REVERT can load it and do one operation.
 */
void SG_pendingtree__hack__unadd_dont_save_pendingtree(SG_context * pCtx, SG_pendingtree * pPendingTree, const char * pszGid, const SG_pathname * pPath_Input)
{
	SG_pathname * pPathDir = NULL;
	SG_string * pStringEntryname = NULL;
	struct sg_ptnode * ptnDir;
	struct sg_ptnode * ptnItem;
	struct sg_ptnode * ptnItem2;

	// split input pathname into <dir> / <entryname>.
	// get the ptnode of <dir>.
	// get the ptnode for the item itself.

	SG_ERR_CHECK(  SG_pathname__get_last(pCtx, pPath_Input, &pStringEntryname)  );
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__COPY(pCtx, &pPathDir, pPath_Input)  );
	SG_ERR_CHECK(  SG_pathname__remove_last(pCtx, pPathDir)  );
	SG_ERR_CHECK(  sg_pendingtree__local_path_to_ptnode(pCtx, pPendingTree, pPathDir, SG_FALSE, &ptnDir)  );
	SG_ERR_CHECK(  sg_ptnode__find_entry_by_name_not_recursive__active(pCtx, ptnDir, SG_string__sz(pStringEntryname), &ptnItem)  );

	if (!ptnItem)
		SG_ERR_THROW2(  SG_ERR_NOT_FOUND,
						(pCtx, "Un-adding %s from pendingtree.", SG_pathname__sz(pPath_Input))  );

	SG_ASSERT(  (ptnItem->pCurrentParent == ptnDir)  );
	SG_ASSERT(  (strcmp(pszGid, ptnItem->pszgid) == 0)  );
	SG_ASSERT(  (ptnItem->pBaselineEntry == NULL)  );

	// I only needed this function to un-add a temporary directory that
	// I only added to help MERGE be a lot easier.
	//
	// TODO I don't know if we need this or not but sg_ptnode__remove_safely()
	// TODO has stuff to prevent a REMOVE if the item has been RENAMED, MOVED,
	// TODO and etc.  For now, I'm going to say no.

	if (ptnItem->type == SG_TREENODEENTRY_TYPE_DIRECTORY)
	{
		// see if the
		SG_uint32 nrChildren = 0;

		if (ptnItem->prbItems)
			SG_ERR_CHECK(  SG_rbtree__count(pCtx, ptnItem->prbItems, &nrChildren)  );

		if (nrChildren != 0)
			SG_ERR_THROW2(  SG_ERR_INVALIDARG,
							(pCtx, "Cannot un-add directory %s because pendingtree is not empty.", SG_pathname__sz(pPath_Input))  );
	}

	SG_ERR_CHECK(  SG_rbtree__remove__with_assoc(pCtx, ptnDir->prbItems, pszGid, (void **)&ptnItem2)  );
	SG_ASSERT(  (ptnItem2 == ptnItem)  );
	SG_PTNODE_NULLFREE(pCtx, ptnItem2);

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathDir);
	SG_STRING_NULLFREE(pCtx, pStringEntryname);
}

/*
 * HACK Un-add an entry from the pendingtree BUT DO NOT REMOVE IT FROM DISK.
 * HACK This is like "vv revert foo" following a "vv add foo", but
 * HACK without all the baggage of the full REVERT mechanism.  I need this
 * HACK routine in the middle of the UPDATE code to get rid of the ptnode for
 * HACK a temporary directory.  As such, I have lots of open changes in the in-memory
 * HACK pendingtree and since the REVERT code assumes that it is in complete
 * HACK control of the pendingtree, I don't want to have to save my dirty
 * HACK version so that REVERT can load it and do one operation.
 */ 
static void sg_pendingtree__hack__unadd_dont_save__by_ptn(SG_context * pCtx,
														  struct sg_ptnode * ptn)
{
	SG_string * pStringRepoPath = NULL;
	struct sg_ptnode * ptn_OWNED = NULL;

#if TRACE_UPDATE
	SG_ERR_CHECK(  sg_ptnode__get_repo_path(pCtx, ptn, &pStringRepoPath)  );
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "sg_pendingtree__unadd__by_ptn: %s\n",
							   SG_string__sz(pStringRepoPath))  );
#endif

	if (ptn->type == SG_TREENODEENTRY_TYPE_DIRECTORY)
	{
		SG_uint32 nrChildren = 0;

		if (ptn->prbItems)
			SG_ERR_CHECK(  SG_rbtree__count(pCtx, ptn->prbItems, &nrChildren)  );
		if (nrChildren != 0)
		{
			if (!pStringRepoPath)
				SG_ERR_CHECK(  sg_ptnode__get_repo_path(pCtx, ptn, &pStringRepoPath)  );

			SG_ERR_THROW2(  SG_ERR_INVALIDARG,
							(pCtx, "Cannot un-add directory because pendingtree is not empty: %s",
							 SG_string__sz(pStringRepoPath))  );
		}
	}

	SG_ERR_CHECK(  SG_rbtree__remove__with_assoc(pCtx, ptn->pCurrentParent->prbItems, ptn->pszgid, (void **)&ptn_OWNED)  );
	SG_ASSERT(  (ptn_OWNED == ptn)  );

fail:
	SG_PTNODE_NULLFREE(pCtx, ptn_OWNED);
	SG_STRING_NULLFREE(pCtx, pStringRepoPath);
}

//////////////////////////////////////////////////////////////////

void SG_pendingtree__get_timestamp_cache(SG_context * pCtx,
										 SG_pendingtree * pPendingTree,
										 SG_timestamp_cache ** ppTSC)
{

	SG_NULLARGCHECK_RETURN(pPendingTree);

	if (!pPendingTree->pTSC)
		SG_ERR_CHECK_RETURN(  SG_timestamp_cache__allocate_and_load(pCtx, &pPendingTree->pTSC, pPendingTree->pPathWorkingDirectoryTop)  );

	if (ppTSC)
		*ppTSC = pPendingTree->pTSC;
}

//////////////////////////////////////////////////////////////////

#include "sg_pendingtree__private_remove_safely2.h"

//////////////////////////////////////////////////////////////////

void SG_pendingtree__find_gids_for_locking__must_be_files(
    SG_context* pCtx,
    SG_pendingtree* pPendingTree,
    const SG_pathname* pPathRelativeTo,
    SG_uint32 count_items, 
    const char** paszItems,
    SG_vhash** ppvh_gids
    )
{
	SG_uint32 i;
	struct sg_ptnode* ptn = NULL;
	struct sg_ptnode* ptnItem = NULL;
	SG_string* pstrName = NULL;
	SG_pathname* pPathDir = NULL;
	SG_pathname* pPathLocal = NULL;
	SG_repo* pRepo = NULL;
    SG_vhash* pvh_result = NULL;

	SG_ERR_CHECK(  SG_pendingtree__get_repo(pCtx, pPendingTree, &pRepo)  );
    SG_ERR_CHECK(  SG_vhash__alloc(pCtx, &pvh_result)  );

    for (i=0; i<count_items; i++)
    {
        /* First get a pathname for the item */
        SG_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx, &pPathLocal, pPathRelativeTo, paszItems[i])  );

        /* We want the parent directory of this item.  So copy the path and remove the last component */
        SG_ERR_CHECK(  SG_PATHNAME__ALLOC__COPY(pCtx, &pPathDir, pPathLocal)  );
        SG_ERR_CHECK(  SG_pathname__remove_last(pCtx, pPathDir)  );

        /* Find the directory's ptnode */
        SG_ERR_CHECK(  sg_pendingtree__local_path_to_ptnode(pCtx, pPendingTree, pPathDir, SG_TRUE, &ptn)  );

        /* Get the base name from the original path */
        SG_ERR_CHECK(  SG_pathname__get_last(pCtx, pPathLocal, &pstrName)  );

        /* Find the ptnode */
        SG_ERR_CHECK(  sg_ptnode__find_entry_by_name_not_recursive__active(pCtx, ptn, SG_string__sz(pstrName), &ptnItem)  );
        if (!ptnItem)
        {
            SG_ERR_THROW2(  SG_ERR_NOT_FOUND,
                        (pCtx,"Looking for '%s' in the pending tree node for '%s'",
                         SG_string__sz(pstrName),
                         SG_pathname__sz(pPathDir))  );
        }

        if (SG_TREENODEENTRY_TYPE_REGULAR_FILE != ptnItem->type)
        {
            SG_ERR_THROW2(  SG_ERR_VC_LOCK_FILES_ONLY,
                        (pCtx, "%s",
                         paszItems[i])  );
        }

        if (!ptnItem->pBaselineEntry)
        {
            SG_ERR_THROW2(  SG_ERR_VC_LOCK_NOT_COMMITTED_YET,
                        (pCtx, "%s",
                         paszItems[i])  );
        }

        SG_ERR_CHECK(  SG_vhash__add__null(pCtx, pvh_result, ptnItem->pszgid)  );

        SG_STRING_NULLFREE(pCtx, pstrName);
        SG_PATHNAME_NULLFREE(pCtx, pPathDir);
        SG_PATHNAME_NULLFREE(pCtx, pPathLocal);
    }

    *ppvh_gids = pvh_result;
    pvh_result = NULL;

fail:
    SG_VHASH_NULLFREE(pCtx, pvh_result);
	SG_STRING_NULLFREE(pCtx, pstrName);
	SG_PATHNAME_NULLFREE(pCtx, pPathDir);
	SG_PATHNAME_NULLFREE(pCtx, pPathLocal);
}

//////////////////////////////////////////////////////////////////

/**
 * Do an official 'vv remove' on an entry in the WD that is
 * currently/temporarily parked in the parking lot.
 *
 * This can happen during a MERGE when the OTHER branch
 * deleted "foo" and created a new "foo" (with a different GID)
 * in the directory and the original file was unchanged in the
 * BASELINE branch.  Because of the transient name collision
 * and because MERGE needs to be operation-order-independent when
 * it applies the merge, the baseline version of the file gets
 * PARKED in the parking lot rather than being immediately deleted.
 * MERGE then populates the directory with the new version of the
 * file (the one with the new GID), and then it removes the
 * baseline version from the disk and *SETS THE _SAVE_FLAG_DELETED*
 * on the PARKED PTNODE in the pendingtree.  The entry is actually
 * removed from version control during the COMMIT.
 *
 * Note: The MOVE to the parking lot causes the PTNODE for the
 * entry in the original directory to be marked "phantom" (inactive)
 * node (so that other code will know that there was something here).
 * The MOVE also creates a PTNODE for the entry in the parking lot.
 * (And a PTNODE for the parking lot directory itself is temporarily
 * added to project.)
 *
 * After the MERGE is complete, we "un-add" the temporary parking
 * lot PTNODE (so it never actually appears in the commit).  But
 * for this to happen, it must be empty (or we'd orphan some nodes).
 * 
 * However, having the entry "active" in this "Parked PTNODE" is a
 * problem when we get ready to un-add the temporary parking lot PTNODE
 * because it cannot hold the only reference to an entry because the
 * entry needs to exist (but in a deleted state) until the COMMIT.
 * 
 * This routine is aware of this and deletes the file from parking
 * lot directory on disk and updates *BOTH* PTNODES.  It effectively
 * MOVES the entry back to the original directory (by changing the
 * _SAVED_FLAG_MOVED_AWAY to _SAVED_FLAG_DELETED in the INACTIVE
 * PTNODE and ACTUALLY deleting the PARKED PTNODE.
 *
 * WARNING: You probably don't want to call this.  You probably
 * WARNING: want the normal version of SG_pendingtree__remove().
 * 
 */
void SG_pendingtree__remove_parked__dont_save_pendingtree(SG_context * pCtx,
														  SG_pendingtree * pPendingTree,
														  const SG_pathname * pPathParked,
														  const char * pszGidEntry)
{
	struct sg_ptnode * ptnParkingLotEntry_Owned = NULL;
	struct sg_ptnode * ptnParkingLotEntry = NULL;	// we do not own this
	struct sg_ptnode * ptnParkingLotDir = NULL;	// we do not own this
	struct sg_ptnode * ptnMovedFromDir = NULL;		// we do not own this
	struct sg_ptnode * ptnInactiveEntry = NULL;	// we do not own this
	const char * pszParkedName;
	const char * pszOldName;
	SG_timestamp_cache * pTSC = NULL;		// we do not own this

	// find the PTNODE for the currently ACTIVE version of the entry
	// (which should be in the parking lot).

	SG_ERR_CHECK(  sg_pendingtree__find_by_gid(pCtx, pPendingTree, pszGidEntry, &ptnParkingLotEntry)  );
	SG_ASSERT(  (strcmp(ptnParkingLotEntry->pszgid, pszGidEntry) == 0)  );
	SG_ERR_CHECK(  sg_ptnode__get_name(pCtx, ptnParkingLotEntry, &pszParkedName)  );

	// get the PTNODE for the containing directory in the parking lot.

	ptnParkingLotDir = ptnParkingLotEntry->pCurrentParent;

	// parking the entry may do a rename, so fetch the original name
	// before we parked it.  (this is probably the name that was a
	// member of a collision of some kind.)

	SG_ERR_CHECK(  sg_ptnode__get_old_name(pCtx, ptnParkingLotEntry, &pszOldName)  );

	// fetch the PTNODE of the directory it was moved from.  and
	// find the INACTIVE (phantom) PTNODE that MOVE left behind.

	SG_ERR_CHECK(  sg_pendingtree__find_by_gid(pCtx, pPendingTree, ptnParkingLotEntry->pszgidMovedFrom, &ptnMovedFromDir)  );
	SG_ERR_CHECK(  sg_ptnode__find_entry_by_name_not_recursive__inactive(pCtx, ptnMovedFromDir, pszOldName, &ptnInactiveEntry)  );
	SG_ASSERT(  (strcmp(ptnInactiveEntry->pszgid, pszGidEntry) == 0)  );

#if 0 && defined(DEBUG)
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   ("RemoveParked: [pathParked %s]\n"
								"              [gidEntry %s]\n"
								"              [gidParkingLotDir %s]\n"
								"                  [currentName %s]\n"
								"                  [ptnSavedFlags %d]\n"
								"                  [ptnTempFlags %d]\n"
								"              [gidMovedFrom %s]\n"
								"                  [oldName %s]\n"
								"                  [ptnSavedFlags %d]\n"
								"                  [ptnTempFlags %d]\n"),
							   SG_pathname__sz(pPathParked),
							   pszGidEntry,
							   ptnParkingLotDir->pszgid,
							   pszParkedName,
							   (SG_uint32)ptnParkingLotEntry->saved_flags,
							   (SG_uint32)ptnParkingLotEntry->temp_flags,
							   ptnMovedFromDir->pszgid,
							   pszOldName,
							   (SG_uint32)ptnInactiveEntry->saved_flags,
							   (SG_uint32)ptnInactiveEntry->temp_flags)  );
#endif

	SG_ASSERT(  (ptnInactiveEntry->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY)  );
	ptnInactiveEntry->saved_flags &= ~sg_PTNODE_SAVED_FLAG_MOVED_AWAY;
	ptnInactiveEntry->saved_flags |=  sg_PTNODE_SAVED_FLAG_DELETED;

	SG_ERR_CHECK(  SG_rbtree__remove__with_assoc(pCtx, ptnParkingLotDir->prbItems, pszGidEntry, (void **)&ptnParkingLotEntry_Owned)  );
	SG_ASSERT(  (ptnParkingLotEntry_Owned == ptnParkingLotEntry)  );
	
	switch (ptnParkingLotEntry->type)
	{
	case SG_TREENODEENTRY_TYPE_SUBMODULE:
		SG_ERR_CHECK(  SG_workingdir__move_submodule_to_graveyard(pCtx,
																  pPendingTree->pPathWorkingDirectoryTop,
																  pPathParked,
																  ptnParkingLotDir->pszgid,
																  NULL)  );
		break;

	case SG_TREENODEENTRY_TYPE_DIRECTORY:
		SG_ERR_CHECK(  SG_fsobj__rmdir__pathname(pCtx, pPathParked)  );
		break;

	case SG_TREENODEENTRY_TYPE_REGULAR_FILE:
	case SG_TREENODEENTRY_TYPE_SYMLINK:
		SG_ERR_CHECK(  SG_fsobj__remove__pathname(pCtx, pPathParked)  );
		break;

	default:
		SG_ERR_THROW2(  SG_ERR_NOTIMPLEMENTED,
						(pCtx,
						 "RemoveParked with tneType %d: %s",
						 ptnParkingLotEntry->type,
						 SG_pathname__sz(pPathParked))  );
		break;
	}

	SG_ERR_CHECK(  SG_pendingtree__get_timestamp_cache(pCtx, pPendingTree, &pTSC)  );
	SG_ERR_CHECK(  SG_timestamp_cache__remove(pCtx, pTSC, pszGidEntry)  );

fail:
	SG_PTNODE_NULLFREE(pCtx, ptnParkingLotEntry_Owned);
}

//////////////////////////////////////////////////////////////////
// TODO 2010/08/18 Should these walk routines be #ifdef DEBUG ?

void _walk_down_ptnode(SG_context * pCtx, SG_pendingtree * pPendingTree, struct sg_ptnode* ptnode)
{
	SG_rbtree_iterator * prbt_it = NULL;
	SG_bool bContinue = SG_FALSE;
	struct sg_ptnode * ptnode_Sub = NULL;
	if (SG_TREENODEENTRY_TYPE_DIRECTORY == ptnode->type)
	{
		SG_ERR_CHECK(  sg_ptnode__load_dir_entries(pCtx, ptnode)  );
		SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &prbt_it, ptnode->prbItems, &bContinue, NULL, (void**)&ptnode_Sub)  );
		while (bContinue)
		{
			SG_ERR_CHECK(  _walk_down_ptnode(pCtx, pPendingTree, ptnode_Sub)  );
			SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, prbt_it, &bContinue, NULL, (void**)&ptnode_Sub)  );
		}
	}
fail:
	SG_RBTREE_ITERATOR_NULLFREE(pCtx, prbt_it);
	return;
}

//This function is a debugging utility to just see how long it takes to load every pendingtree node.
void SG_pendingtree__walk_tree(SG_context * pCtx, SG_pendingtree * pPendingTree)
{
	SG_ERR_CHECK( _walk_down_ptnode(pCtx, pPendingTree, pPendingTree->pSuperRoot)  );
fail:
	return;
}

//////////////////////////////////////////////////////////////////

static void sg_ptnode__set_disposable_hid(SG_context * pCtx,
										  struct sg_ptnode * ptn,
										  const char * pszHid)
{
#if TRACE_PENDINGTREE
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "__set_disposable_hid: [gid %s] [hid %s]\n",
							   ptn->pszgid,
							   pszHid)  );
#endif

	SG_ERR_CHECK_RETURN(  SG_strpool__add__sz(pCtx, ptn->pPendingTree->pStrPool, pszHid, &ptn->pszHidDisposable)  );
}

void SG_pendingtree__set_disposable_hid(SG_context * pCtx,
										SG_pendingtree * pPendingTree,
										const char * pszGid,
										const char * pszHid)
{
	struct sg_ptnode * ptn = NULL;	// we do not own this

	sg_pendingtree__find_by_gid(pCtx, pPendingTree, pszGid, &ptn);
	if (SG_context__has_err(pCtx))
	{
		if (SG_context__err_equals(pCtx, SG_ERR_NOT_FOUND))
		{
#if TRACE_PENDINGTREE
			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
									   "__set_disposable_hid: did not find ptnode for %s\n",
									   pszGid)  );
#endif
			// not worth dying for.  just don't set it.
			SG_context__err_reset(pCtx);
			return;
		}

		SG_ERR_RETHROW;
	}

	SG_ERR_CHECK(  sg_ptnode__set_disposable_hid(pCtx, ptn, pszHid)  );

fail:
	return;
}

static void sg_ptnode__set_disposable_dir(SG_context * pCtx,
										  struct sg_ptnode * ptn)
{
#if TRACE_PENDINGTREE
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "__set_disposable_dir: [gid %s]\n",
							   ptn->pszgid)  );
#else
	SG_UNUSED( pCtx );
#endif

	ptn->bDisposableDir = SG_TRUE;
}

void SG_pendingtree__set_disposable_dir(SG_context * pCtx,
										SG_pendingtree * pPendingTree,
										const char * pszGid)
{
	struct sg_ptnode * ptn = NULL;	// we do not own this

	sg_pendingtree__find_by_gid(pCtx, pPendingTree, pszGid, &ptn);
	if (SG_context__has_err(pCtx))
	{
		if (SG_context__err_equals(pCtx, SG_ERR_NOT_FOUND))
		{
#if TRACE_PENDINGTREE
			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
									   "__set_disposable_dir: did not find ptnode for %s\n",
									   pszGid)  );
#endif
			// not worth dying for.  just don't set it.
			SG_context__err_reset(pCtx);
			return;
		}

		SG_ERR_RETHROW;
	}

	SG_ERR_CHECK(  sg_ptnode__set_disposable_dir(pCtx, ptn)  );

fail:
	return;
}

//////////////////////////////////////////////////////////////////

static void sg_ptnode__set_sticky_dir(SG_context * pCtx,
									  struct sg_ptnode * ptn)
{
#if TRACE_PENDINGTREE
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "__set_sticky_dir: [gid %s]\n",
							   ptn->pszgid)  );
#else
	SG_UNUSED( pCtx );
#endif

	ptn->bStickyDir = SG_TRUE;
}

void SG_pendingtree__hack__set_sticky_dir(SG_context * pCtx,
										  SG_pendingtree * pPendingTree,
										  const char * pszGid)
{
	struct sg_ptnode * ptn = NULL;	// we do not own this

	sg_pendingtree__find_by_gid(pCtx, pPendingTree, pszGid, &ptn);
	if (SG_context__has_err(pCtx))
	{
		if (SG_context__err_equals(pCtx, SG_ERR_NOT_FOUND))
		{
#if TRACE_PENDINGTREE
			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
									   "__set_sticky_dir: did not find ptnode for %s\n",
									   pszGid)  );
#endif
			// not worth dying for.  just don't set it.
			SG_context__err_reset(pCtx);
			return;
		}

		SG_ERR_RETHROW;
	}

	SG_ERR_CHECK(  sg_ptnode__set_sticky_dir(pCtx, ptn)  );

fail:
	return;
}

static SG_rbtree_foreach_callback _sg_ptnode__clear_all_sticky_bits_cb;

static void _sg_ptnode__clear_all_sticky_bits_cb(SG_context * pCtx,
												 const char * pszKey,
												 void * pVoid_ptn,
												 void * pVoid_data)
{
	struct sg_ptnode * ptn = (struct sg_ptnode *)pVoid_ptn;

	SG_UNUSED( pszKey );
	SG_UNUSED( pVoid_data );

	if (!ptn)
		return;

	if (ptn->type != SG_TREENODEENTRY_TYPE_DIRECTORY)
		return;
	
	if (!ptn->prbItems)
		return;

	ptn->bStickyDir = SG_FALSE;
	
	SG_ERR_CHECK_RETURN(  SG_rbtree__foreach(pCtx,
											 ptn->prbItems,
											 _sg_ptnode__clear_all_sticky_bits_cb,
											 NULL)  );
}

void SG_pendingtree__hack__clear_all_sticky_bits(SG_context * pCtx,
												 SG_pendingtree * pPendingTree)
{
	SG_NULLARGCHECK_RETURN( pPendingTree );

	if (!pPendingTree->pSuperRoot)
		return;
	
	SG_ERR_CHECK_RETURN(  SG_rbtree__foreach(pCtx,
											 pPendingTree->pSuperRoot->prbItems,
											 _sg_ptnode__clear_all_sticky_bits_cb,
											 NULL)  );
}
