/*
Copyright 2010-2011 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 *
 * @file sg_pendingtree__private_revert.h
 *
 * @details
 *
 */

// TODO investigate all of the _CHECKMYNAME flags that we set here.

//////////////////////////////////////////////////////////////////

#ifndef H_SG_PENDINGTREE__PRIVATE_REVERT_H
#define H_SG_PENDINGTREE__PRIVATE_REVERT_H

BEGIN_EXTERN_C;

//////////////////////////////////////////////////////////////////

struct _pt_revert_data
{
	SG_pendingtree *		pPendingTree;		// we do not own this.
	SG_treediff2 *			pTreeDiff;			// we own this.   (treediff between baseline and WD)

	SG_inv_dirs *			pInvDirs;			// we own this.
	SG_vector *				pVecCycles;			// vec[struct sg_ptnode *] of moved dirs marked PARKED_FOR_CYCLE    we do not own these.

	SG_uint32				nrDisposableDirs;
	SG_vector *				pVecDisposablePathsToDelete;	// vec[SG_pathname *]		-- we will own these
	SG_vector *				pVecDisposablePTNodesToDelete;	// vec[struct sg_ptnode *]	-- we do not own these

	const SG_vector *		pvec_use_submodules;			// vec[ hints ] -- we do not own this.
	SG_varray *				pvaMissingDetails;				// array of submodule problems -- we do not own this
	SG_vector *				pvec_queued_subs;				// vec[ struct _queued_subs * ] -- we own these

	enum
	{
		_PT_REVERT_PASS__SCAN_AND_BIND,
		_PT_REVERT_PASS__CHECK_FOR_CONFLICTS,
		_PT_REVERT_PASS__PRECOMPUTE_SUBMODULE_STATE,
		_PT_REVERT_PASS__APPLY_STRUCTURAL_CHANGES,
	}						pass;

	SG_bool					bNoBackups;			// don't litter WD with ~sg00~ files for dirty files.
	SG_bool					bNoSubmodules;		// don't dive into submodules and try to revert their contents

};

//////////////////////////////////////////////////////////////////

static void _pt_revert__revert_attrbits(SG_context * pCtx,
										SG_UNUSED_PARAM( struct _pt_revert_data * prd ),
										const SG_pathname * pPathAbsolute,
										struct sg_ptnode * ptn,
										const char * pszReason);

static void _pt_revert__revert_xattrs(SG_context * pCtx,
									  struct _pt_revert_data * prd,
									  const SG_pathname * pPathAbsolute,
									  struct sg_ptnode * ptn,
									  const char * pszReason);

static void _pt_revert__is_disposable_file_symlink(SG_context * pCtx,
												   struct sg_ptnode * ptn,
												   SG_bool * pbIsDisposable);

static void _pt_revert__is_disposable_submodule(SG_context * pCtx,
												   struct sg_ptnode * ptn,
												   SG_bool * pbIsDisposable);

//////////////////////////////////////////////////////////////////

struct _queued_subs
{
	const SG_treediff2_ObjectData *		pOD;				// we do not own this
	struct sg_ptnode *					ptn;				// we do not own this
	SG_pathname *						pPathGraveyard;		// (optional) we own this
};

static void _queued_subs__alloc(SG_context * pCtx,
								struct _queued_subs ** ppQS,
								const SG_treediff2_ObjectData * pOD,
								struct sg_ptnode * ptn,
								SG_pathname ** ppPathGraveyard)	// we steal this
{
	struct _queued_subs * pQS;

	SG_ERR_CHECK_RETURN(  SG_alloc1(pCtx, pQS)  );

	pQS->pOD = pOD;
	pQS->ptn = ptn;

	if (ppPathGraveyard)
	{
		pQS->pPathGraveyard = *ppPathGraveyard;
		*ppPathGraveyard = NULL;
	}
	else
	{
		pQS->pPathGraveyard = NULL;
	}

	*ppQS = pQS;
}

static void _queued_subs__free(SG_context * pCtx,
							   struct _queued_subs * pQS)
{
	if (!pQS)
		return;

	SG_PATHNAME_NULLFREE(pCtx, pQS->pPathGraveyard);
	SG_NULLFREE(pCtx, pQS);
}

//////////////////////////////////////////////////////////////////

/**
 * This submodule was modified (with currently pended dirt *AND/OR*
 * it not references a different baseline than the outer is expecting).
 * To fully REVERT it from the outer's point of view, we must REVERT
 * the pending dirt within *and* UPDATE it to the desired cset.
 *
 * PRECOMPUTE whether we are able to do a REVERT and/or UPDATE.
 *
 */
static void _pt_revert__precompute_modified_submodule(SG_context * pCtx,
													  struct _pt_revert_data * prd,
													  const SG_treediff2_ObjectData * pOD,
													  struct sg_ptnode * ptn)
{
	SG_string * pStringRepoPath = NULL;
	SG_pathname * pPathAbsolute = NULL;
	struct _queued_subs * pQS = NULL;
	SG_uint32 nrBefore = 0;
	SG_uint32 nrAfter = 0;
	SG_bool bDirtyPending = SG_FALSE;
	SG_bool bDirtyReference = SG_FALSE;

#if TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "precompute_update_modified_submodule: [gid %s]\n",
							   ptn->pszgid)  );
#endif

	SG_ERR_CHECK(  SG_varray__count(pCtx, prd->pvaMissingDetails, &nrBefore)  );

	SG_ERR_CHECK(  sg_ptnode__get_repo_path(pCtx, ptn, &pStringRepoPath)  );
	SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path(pCtx,
																		 prd->pPendingTree->pPathWorkingDirectoryTop,
																		 SG_string__sz(pStringRepoPath),
																		 &pPathAbsolute)  );
	SG_ASSERT(  (ptn->pTSM)  );
	SG_ERR_CHECK(  sg_ptnode__is_submodule_changed_or_dirty(pCtx, ptn, &bDirtyReference, &bDirtyPending)  );

#if TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "precompute_update_modified_submodule: [bDR %d][bDP %d]: %s\n",
							   bDirtyReference, bDirtyPending,
							   SG_pathname__sz(pPathAbsolute))  );
#endif

	if (bDirtyPending)
	{
		// see if we can revert just the pending dirt back to the
		// sub's current baseline.
		// 
		// we do not pass bNoSubmodules, because it is an all-or-nothing
		// thing.  it is not a depth thing.

		SG_ERR_CHECK(  SG_treenode_submodule__precompute_revert(pCtx,
																pPathAbsolute,
																ptn->pTSM,	// only used for vtable
																prd->pvec_use_submodules,
																prd->bNoBackups,
																prd->pvaMissingDetails)  );
		SG_ERR_CHECK(  SG_varray__count(pCtx, prd->pvaMissingDetails, &nrAfter)  );
		if (nrAfter > nrBefore)			// some pending dirt could not be resolved
			goto done;
	}

	if (bDirtyReference)
	{
		// TODO 2011/03/31 See the warning below starting with "ACK!".
	}

	// we can do the REVERT *IFF* we did not append anything to prd->pvaMissingDetails.

	if (!prd->pvec_queued_subs)
		SG_ERR_CHECK(  SG_VECTOR__ALLOC(pCtx, &prd->pvec_queued_subs, 10)  );
	SG_ERR_CHECK(  _queued_subs__alloc(pCtx, &pQS, pOD, ptn, NULL)  );
	SG_ERR_CHECK(  SG_vector__append(pCtx, prd->pvec_queued_subs, pQS, NULL)  );
	pQS = NULL;		// vector owns it now

done:

#if TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "precompute_update_modified_submodule: [gid %s] [nrMissingDetails %d]\n",
							   ptn->pszgid, (nrAfter - nrBefore))  );
#endif
	;

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathAbsolute);
	SG_STRING_NULLFREE(pCtx, pStringRepoPath);
	_queued_subs__free(pCtx, pQS);
}

static SG_treediff2_foreach_callback _pt_revert__precompute_modified_submodules;

static void _pt_revert__precompute_modified_submodules(SG_context * pCtx,
													   SG_UNUSED_PARAM( SG_treediff2 * pTreeDiff ),
													   const char * pszGidObject,
													   const SG_treediff2_ObjectData * pOD,
													   void * pVoidData)
{
	struct _pt_revert_data * prd = (struct _pt_revert_data *)pVoidData;
	struct sg_ptnode * ptn = NULL;
	SG_diffstatus_flags dsFlags;
	SG_treenode_entry_type tneType;

	SG_UNUSED( pTreeDiff );

	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_tneType(pCtx, pOD, &tneType)  );
	if (tneType == SG_TREENODEENTRY_TYPE_SUBMODULE)
	{
		SG_ERR_CHECK(  SG_treediff2__ObjectData__get_dsFlags(pCtx,pOD,&dsFlags)  );
		if (dsFlags & SG_DIFFSTATUS_FLAGS__MODIFIED)
		{
			SG_ERR_CHECK(  sg_pendingtree__find_by_gid(pCtx,prd->pPendingTree,pszGidObject,&ptn)  );
			SG_ERR_CHECK(  _pt_revert__precompute_modified_submodule(pCtx, prd, pOD, ptn)  );
		}
	}

fail:
	return;
}

/**
 * This submodule was modified (with currently pended dirt *AND/OR*
 * it not references a different baseline than the outer is expecting).
 * To fully REVERT it from the outer's point of view, we must REVERT
 * the pending dirt within *and* UPDATE it to the desired cset.
 *
 */
static void _pt_revert__modified_submodule(SG_context * pCtx,
										   struct _pt_revert_data * prd,
										   struct _queued_subs * pQS)
{
	SG_string * pStringRepoPath = NULL;
	SG_pathname * pPathAbsolute = NULL;
	SG_treenode_submodule_instance_data_update * pTSIDU = NULL;
	SG_treenode_submodule * pTSM_goal = NULL;
	SG_bool bDirtyPending = SG_FALSE;
	SG_bool bDirtyReference = SG_FALSE;
	SG_bool bIsWD = SG_FALSE;

#if TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "update_modified_submodule: [gid %s]\n",
							   pQS->ptn->pszgid)  );
#endif

	SG_ASSERT(  (prd->bNoSubmodules == SG_FALSE)  );

	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_old_tsm(pCtx, pQS->pOD, &pTSM_goal)  );

	SG_ERR_CHECK(  sg_ptnode__get_repo_path(pCtx, pQS->ptn, &pStringRepoPath)  );
	SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path(pCtx,
																		 prd->pPendingTree->pPathWorkingDirectoryTop,
																		 SG_string__sz(pStringRepoPath),
																		 &pPathAbsolute)  );
	SG_ASSERT(  (pQS->ptn->pTSM)  );
	SG_ERR_CHECK(  sg_ptnode__is_submodule_changed_or_dirty(pCtx, pQS->ptn, &bDirtyReference, &bDirtyPending)  );

#if TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "precompute_update_modified_submodule: [bDR %d][bDP %d]: %s\n",
							   bDirtyReference, bDirtyPending,
							   SG_pathname__sz(pPathAbsolute))  );
#endif

	if (   bDirtyPending
		   && bDirtyReference
		   && (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
		   && ((prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT) == 0))
	{
		// TODO 2011/04/04 The whole log-it then do-it model needs to be
		// TODO            re-thought.  it breaks down here because we 
		// TODO            may have to do 2 steps that could each fail
		// TODO            and that are out of our control.

		SG_ERR_CHECK(  _pt_log__submodule_revert(pCtx, prd->pPendingTree,
												 pQS->ptn->pszgid,
												 SG_pathname__sz(pPathAbsolute),
												 pQS->ptn->pTSM,	// use ptn's pTSM for log message
												 "REVERT_AND_UPDATE_SUBMODULE")  );
	}
	else
	{
		if (bDirtyPending)
		{
			// dive in and do a 'vv revert --all' it back to the sub's
			// current baseline.

			if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
				SG_ERR_CHECK(  _pt_log__submodule_revert(pCtx, prd->pPendingTree,
														 pQS->ptn->pszgid,
														 SG_pathname__sz(pPathAbsolute),
														 pQS->ptn->pTSM,
														 "REVERT_SUBMODULE")  );
			if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
			{
				SG_ERR_CHECK(  SG_treenode_submodule__apply_revert(pCtx,
																   pPathAbsolute,
																   pQS->ptn->pTSM,	// only used for vtable
																   prd->pvec_use_submodules,
																   prd->bNoBackups)  );
					
				// recompute/sniff the submodule to see if there were
				// any problems or dirt that couldn't be reverted.
				// and to get a fresh copy of the pTSM.

				SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pQS->ptn->pTSM);
				SG_ERR_CHECK(  SG_TREENODE_SUBMODULE__SNIFF_DIRECTORY(pCtx,
																	  pPathAbsolute,
																	  pQS->ptn->pPendingTree->pRepo,
																	  &bIsWD,
																	  &pQS->ptn->pTSM)  );
				SG_ERR_CHECK(  SG_treenode_submodule__is_dirty(pCtx, pQS->ptn->pTSM, &bDirtyPending)  );
				if (bDirtyPending)
				{
					// if still dirty after the revert, don't do step 2.
					if (bDirtyReference && (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT))
						SG_ERR_CHECK(  _pt_log__submodule_revert(pCtx, prd->pPendingTree,
																 pQS->ptn->pszgid,
																 SG_pathname__sz(pPathAbsolute),
																 pQS->ptn->pTSM,	// use ptn's pTSM for log message
																 "SKIPPING_UPDATE_OF_SUBMODULE_DUE_TO_EXISTING_DIRT")  );
				}
			}
		}

		if (bDirtyReference && !bDirtyPending)
		{
			// dive in and do a 'vv update --rev r' so that we get
			// a clean update to the cset that we think it should
			// have.

			if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
				SG_ERR_CHECK(  _pt_log__submodule_update(pCtx, prd->pPendingTree,
														 pQS->ptn->pszgid,
														 SG_pathname__sz(pPathAbsolute),
														 pQS->ptn->pTSM,
														 pTSM_goal,
														 "UPDATE_SUBMODULE")  );
			if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
			{
				SG_ERR_CHECK(  SG_treenode_submodule__precompute_update(pCtx,
																		pPathAbsolute,
																		pQS->ptn->pTSM,
																		pTSM_goal,
																		prd->pvec_use_submodules,
																		prd->pvaMissingDetails,
																		&pTSIDU)  );
				SG_ASSERT(  (pTSIDU != NULL)  );
				SG_ERR_CHECK(  SG_treenode_submodule__apply_update(pCtx,
																   pPathAbsolute,
																   pQS->ptn->pTSM,
																   pTSM_goal,
																   pTSIDU)  );
			}
		}
	}
	
fail:
	if (pTSIDU)
		SG_ERR_IGNORE(  (*pTSIDU->pfn_dtor)(pCtx, pTSIDU)  );
	SG_PATHNAME_NULLFREE(pCtx, pPathAbsolute);
	SG_STRING_NULLFREE(pCtx, pStringRepoPath);
	SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pTSM_goal);
}

//////////////////////////////////////////////////////////////////

/**
 * this submodule was deleted in the pendingtree and is
 * to be reverted.  so we need to recreate it.  this is
 * probably a full CHECKOUT (which would re-create the
 * directory and make it appear clean).
 *
 * However, if when the user deleted the submodule, we
 * moved it to the graveyard rather than just deleting
 * it, we could try to recover it.  If we do that we may
 * have to dive in and REVERT and/or UPDATE it.
 *
 * Note: The reason we try to fetch the sub from the
 * graveyard is that it will already have the correct
 * binding (repo-id ==> repo-name).  If we can't find
 * the sub in the graveyard, we'll have to use the
 * regular checkout reverse lookup and that may cause
 * the user to have to give us --usesub hints.
 *
 */
static void _pt_revert__precompute_recreate_deleted_submodule(SG_context * pCtx,
															  struct _pt_revert_data * prd,
															  const SG_treediff2_ObjectData * pOD,
															  struct sg_ptnode * ptn)
{
	SG_pathname * pPathInGraveyard = NULL;
	SG_pathname * pPath_XXXXXX = NULL;
	SG_treenode_submodule * pTSM_live = NULL;
	SG_treenode_submodule * pTSM_baseline = NULL;
	SG_vhash * pvhCheckoutInstanceData_XXXXXX = NULL;
	SG_treenode_submodule_instance_data_update * pTSIDU_XXXXXX = NULL;
	struct _queued_subs * pQS = NULL;
	SG_uint32 nrBefore = 0;
	SG_uint32 nrAfter = 0;
	SG_bool bFound = SG_FALSE;

#if TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "precompute_recreate_deleted_submodule: [gid %s]\n",
							   ptn->pszgid)  );
#endif

	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_old_tsm(pCtx, pOD, &pTSM_baseline)  );

	SG_ERR_CHECK(  SG_varray__count(pCtx, prd->pvaMissingDetails, &nrBefore)  );

	SG_ERR_CHECK(  SG_workingdir__find_submodule_in_graveyard(pCtx,
															  ptn->pPendingTree->pPathWorkingDirectoryTop,
															  ptn->pszgid,
															  &bFound,
															  &pPathInGraveyard)  );
	if (bFound)
	{
#if TRACE_REVERT
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "precompute_recreate_deleted_submodule: found in graveyard: %s\n",
								   SG_pathname__sz(pPathInGraveyard))  );
#endif

		// We found a matching submodule in the graveyard.
		// We ***ASSUME*** that this directory corresponds
		// to the one that was recently "deleted".  So we
		// should be able to just move it back and pretend
		// like nothing happened.
		//
		// TODO 2011/03/31 There is a risk that this directory
		// TODO            is left-over crud from a previous
		// TODO            delete that they somehow cancelled
		// TODO            or an update to a rev where the sub
		// TODO            didn't exist and then the user (somehow)
		// TODO            came back to this cset and (somehow)
		// TODO            caused a delete that didn't create a
		// TODO            graveyard entry.
		// TODO
		// TODO            so the content of this submodule
		// TODO            directory might not reflect the one
		// TODO            they had before the delete.
		// TODO
		// TODO            Add the reference cset hid of the
		// TODO            outer to the .json control file that
		// TODO            we keep in the graveyard.  That would
		// TODO            eliminate this concern.
		//
		// We don't actually do the move into the WD yet
		// because we are still in a precompute phase, so
		// we record the pathname of the directory that we
		// found in pQS for later.

		if (prd->bNoSubmodules)
		{
			// They don't want us to dive *into* submodules
			// an revert them.  This is command option is
			// usually thought of when looking at modified
			// submodules where the user doesn't want us to
			// dive in and revert/update the contents.
			//
			// But were here, we're going to recover a submodule
			// from the graveyard.  We allow this option here
			// so that they can recover from the delete, but not
			// lose their work within.
			//
			// *** This is a bit of a crapshoot since we're not
			// *** 100% sure that this is the instance of the
			// *** submodule directory that they deleted.
		}
		else
		{
			// Dive into the submodule and try to REVERT it
			// to a clean state and/or do an UPDATE to get
			// it to refer to the cset that the outer thinks
			// it should be.

			const char * pszHid_baseline = NULL;
			const char * pszHid_graveyard = NULL;
			SG_bool bDirtyPending = SG_FALSE;
			SG_bool bIsWD = SG_FALSE;
			SG_bool bDirtyReference = SG_FALSE;

			// TODO 2011/03/31 i'm going to do the dirty testing the hard way
			// TODO            because the ptn->pTSM may not be set and if it
			// TODO            reflects the deleted status that this GID/item
			// TODO            has in the WD.  And I want to get the dirtyness
			// TODO            of this thing we found in the graveyard.

			SG_ERR_CHECK(  SG_TREENODE_SUBMODULE__SNIFF_DIRECTORY(pCtx,
																  pPathInGraveyard,
																  ptn->pPendingTree->pRepo,
																  &bIsWD,
																  &pTSM_live)  );
			SG_ASSERT(  (bIsWD)  );
			SG_ERR_CHECK(  SG_treenode_submodule__is_dirty(pCtx, pTSM_live, &bDirtyPending)  );

			SG_ERR_CHECK(  SG_treenode_entry__get_hid_blob(pCtx,
														   ptn->pBaselineEntry,
														   &pszHid_baseline)  );
			SG_ERR_CHECK(  SG_treenode_submodule__get_id_ref(pCtx,
															 pTSM_live,
															 &pszHid_graveyard)  );
			bDirtyReference = (strcmp(pszHid_graveyard, pszHid_baseline) != 0);
			
#if TRACE_REVERT
			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
									   "precompute_recreate_deleted_submodule: [pending dirt %d][changed cset %d]\n",
									   bDirtyPending,
									   bDirtyReference)  );
#endif

			if (bDirtyPending)
			{
				// see if we can revert just the pending dirt back to the
				// sub's current baseline.
				// 
				// we do not pass bNoSubmodules, because it is an all-or-nothing
				// thing.  it is not a depth thing.

				SG_ERR_CHECK(  SG_treenode_submodule__precompute_revert(pCtx,
																		pPathInGraveyard,
																		pTSM_live,
																		prd->pvec_use_submodules,
																		prd->bNoBackups,
																		prd->pvaMissingDetails)  );
				SG_ERR_CHECK(  SG_varray__count(pCtx, prd->pvaMissingDetails, &nrAfter)  );
				if (nrAfter > nrBefore)			// some pending dirt could not be resolved
					goto done;

				if (bDirtyReference)
				{
					// so in theory, if we *had* actually done the REVERT, we
					// would now want to UPDATE the sub from its current baseline
					// back to the cset that the outer thinks it should have.
					//
					// TODO 2011/03/31 ACK! I can't precompute this!
					// TODO
					// TODO            If we assume that a clean update between
					// TODO            any 2 csets is always possible, we could
					// TODO            just assume that the update could be done.
					// TODO            But the user could have done a 'vv deleterepo'
					// TODO            on a sub deeply nested within this submdoule
					// TODO            and this update would fail.
					// TODO
					// TODO            An _update_precompute could catch this,
					// TODO            but if I call it now we might get false
					// TODO            positives because we didn't actually do
					// TODO            the revert above.
					// TODO
					// TODO            Ah, but I could just do a _checkout_precompute
					// TODO            on the desired cset and see if all the
					// TODO            required repos/csets are present, but that
					// TODO            may stumble if we can't uniquely map
					// TODO            repo-id==>repo-names for all of the nested
					// TODO            subs within this one.  (They'd have to give
					// TODO            a full set of --usesub args for all possible
					// TODO            issues (rather than just for the ones that
					// TODO            actually need to be created in this recovered
					// TODO            WD).)  This seems wrong too.
					// TODO
					// TODO            For now, let's assume that if they did at
					// TODO            one time have the submodule in the WD of
					// TODO            the outer (such that this graveyard recovery
					// TODO            is well-defined) that they haven't done
					// TODO            something like 'vv deleterepo' and that all
					// TODO            will be well.
				}
			}
			else if (bDirtyReference)
			{
				SG_ERR_CHECK(  SG_treenode_submodule__precompute_update(pCtx,
																		pPathInGraveyard,
																		pTSM_live,
																		pTSM_baseline,
																		prd->pvec_use_submodules,
																		prd->pvaMissingDetails,
																		&pTSIDU_XXXXXX)  );
				SG_ERR_CHECK(  SG_varray__count(pCtx, prd->pvaMissingDetails, &nrAfter)  );
				if (nrAfter > nrBefore)			// clean update could not be performed ?
					goto done;
			}
		}
	}
	else
	{
		const char * pszRepoPath;

		// We did not find a candidate submodule WD in the graveyard.
		// We need to do a regular CHECKOUT for this submodule.
		// (And the --no-submodules flag doesn't matter since we're
		// going to have to do a fresh checkout.)
		// 
		// Precompute it now and see if there are any issues.
		//
		// We get the REPO-PATH of the item and build an absolute
		// path; THIS IS FOR DEBUG/LOGGING PURPOSES ONLY.  Since
		// we haven't done any WD manipulation yet, this is the path
		// to the submodule in the current/dirty outer WD; it may or
		// may not match the path where the submodule will be when
		// we finish the revert.
		//
		// Also, disregard the returned checkout-instance data.  We'll
		// rebuild this later if/when we actually do the checkout when
		// we have valid pathnames.

		SG_ERR_CHECK(  SG_treediff2__ObjectData__get_repo_path(pCtx, pOD, &pszRepoPath)  );
		SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path(pCtx,
																			 ptn->pPendingTree->pPathWorkingDirectoryTop,
																			 pszRepoPath,
																			 &pPath_XXXXXX)  );

		SG_ERR_CHECK(  SG_treenode_submodule__precompute_checkout(pCtx,
																  pPath_XXXXXX,
																  pTSM_baseline,
																  SG_FALSE, // do not pass down bNoSubmodules
																  prd->pvec_use_submodules,
																  prd->pvaMissingDetails,
																  &pvhCheckoutInstanceData_XXXXXX)  );
		SG_ERR_CHECK(  SG_varray__count(pCtx, prd->pvaMissingDetails, &nrAfter)  );
		if (nrAfter > nrBefore)
			goto done;
	}

	// we can do the REVERT *IFF* we did not append anything to prd->pvaMissingDetails.

	// TODO 2011/04/05 Consider putting pTSM_live or pTSM_baseline in the pQS in _precompute
	// TODO            and just using it here.  (Question: the pathname inside the pTSM created
	// TODO            in _precompute may be different from the one we have now, right?)

	if (!prd->pvec_queued_subs)
		SG_ERR_CHECK(  SG_VECTOR__ALLOC(pCtx, &prd->pvec_queued_subs, 10)  );
	SG_ERR_CHECK(  _queued_subs__alloc(pCtx, &pQS, pOD, ptn, &pPathInGraveyard)  );
	SG_ERR_CHECK(  SG_vector__append(pCtx, prd->pvec_queued_subs, pQS, NULL)  );
	pQS = NULL;		// vector owns it now

done:
	// the reason that we do not throw now is so that we can do the precompute on all of
	// the peer submodules and eventually do one big throw with (hopefully) all of the
	// details that the user will need (--usesub hints and/or clone/pull requests) to
	// make to let us retry.

#if TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "precompute_recreate_deleted_submodule: [gid %s] [nrMissingDetails %d]\n",
							   ptn->pszgid, (nrAfter - nrBefore))  );
#endif
	;

fail:
	if (pTSIDU_XXXXXX)
		SG_ERR_IGNORE(  (*pTSIDU_XXXXXX->pfn_dtor)(pCtx, pTSIDU_XXXXXX)  );
	SG_PATHNAME_NULLFREE(pCtx, pPath_XXXXXX);
	SG_VHASH_NULLFREE(pCtx, pvhCheckoutInstanceData_XXXXXX);
	SG_PATHNAME_NULLFREE(pCtx, pPathInGraveyard);
	SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pTSM_live);
	SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pTSM_baseline);
}

static void _pt_revert__recreate_deleted_submodule(SG_context * pCtx,
												   struct _pt_revert_data * prd,
												   struct _queued_subs * pQS)
{
	SG_string * pStringFinalRepoPath = NULL;
	SG_pathname * pPathFinalDestination = NULL;
	SG_treenode_submodule * pTSM_live = NULL;
	SG_treenode_submodule * pTSM_baseline = NULL;
	SG_vhash * pvhCheckoutInstanceData = NULL;
	SG_treenode_submodule_instance_data_update * pTSIDU = NULL;
	SG_bool bAsIs = SG_TRUE;

	SG_ERR_CHECK(  sg_ptnode__get_repo_path(pCtx, pQS->ptn, &pStringFinalRepoPath)  );
	SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path(pCtx,
																		 prd->pPendingTree->pPathWorkingDirectoryTop,
																		 SG_string__sz(pStringFinalRepoPath),
																		 &pPathFinalDestination)  );

#if TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "recreate_deleted_submodule: [gid %s] %s\n",
							   pQS->ptn->pszgid,
							   SG_pathname__sz(pPathFinalDestination))  );
#endif

	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_old_tsm(pCtx, pQS->pOD, &pTSM_baseline)  );

	if (pQS->pPathGraveyard)
	{
		const char * pszHid_baseline = NULL;
		const char * pszHid_graveyard = NULL;
		SG_bool bDirtyPending = SG_FALSE;
		SG_bool bDirtyReference = SG_FALSE;
		SG_bool bIsWD = SG_FALSE;

		// recover zombie submodule from graveyard.

		// TODO 2011/04/05 Consider putting pTSM_live in the pQS in _precompute
		// TODO            and just using it here.

		SG_ERR_CHECK(  SG_TREENODE_SUBMODULE__SNIFF_DIRECTORY(pCtx,
															  pQS->pPathGraveyard,
															  pQS->ptn->pPendingTree->pRepo,
															  &bIsWD,
															  &pTSM_live)  );
		SG_ASSERT(  (bIsWD)  );
		SG_ERR_CHECK(  SG_treenode_submodule__is_dirty(pCtx, pTSM_live, &bDirtyPending)  );

		SG_ERR_CHECK(  SG_treenode_entry__get_hid_blob(pCtx,
													   pQS->ptn->pBaselineEntry,
													   &pszHid_baseline)  );
		SG_ERR_CHECK(  SG_treenode_submodule__get_id_ref(pCtx,
														 pTSM_live,
														 &pszHid_graveyard)  );
		bDirtyReference = (strcmp(pszHid_graveyard, pszHid_baseline) != 0);

		if (prd->bNoSubmodules)
		{
			// leave the recovered submodule as-is.

			bAsIs = SG_TRUE;
		}
		else
		{
			// while it is still in the graveyard, dive into it
			// and try to revert/update it to meet our needs.

			if (   bDirtyPending
				&& bDirtyReference
				&& (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
				&& ((prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT) == 0))
			{
				// TODO 2011/04/04 The whole log-it then do-it model needs to be
				// TODO            re-thought.  it breaks down here because we 
				// TODO            may have to do 2 steps that could each fail
				// TODO            and that are out of our control.
				// TODO
				// TODO            if (bDirtyPending && bDirtyReference) then
				// TODO            we need to dive in and do a 'vv revert --all'
				// TODO            on the sub *and then* dive in and do a 'vv update --rev r'
				// TODO            on it.  when we are actually doing it, if the first
				// TODO            step fails, we don't attempt the second because that
				// TODO            would just compound the problems (and cause a dirty
				// TODO            update (which is a different thing entirely)).
				// TODO
				// TODO            So with the current model, I can't get a 'vv revert --all --test'
				// TODO            to behave exactly like a 'vv revert --all --verbose'
				// TODO            because I can't tell whether the revert will actually
				// TODO            fail.  (We've predicted that it won't, but who knows
				// TODO            what screw cases are lurking.)
				// TODO
				// TODO            So just log a combined message and don't try to be
				// TODO            exact here.

				SG_ERR_CHECK(  _pt_log__submodule_revert(pCtx, prd->pPendingTree,
														 pQS->ptn->pszgid,
														 SG_pathname__sz(pQS->pPathGraveyard),
														 pTSM_baseline,
														 "REVERT_AND_UPDATE_RECOVERED_SUBMODULE")  );
			}
			else
			{
				if (bDirtyPending)
				{
					// dive in and do a 'vv revert --all' it back to the sub's
					// current baseline.

					if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
						SG_ERR_CHECK(  _pt_log__submodule_revert(pCtx, prd->pPendingTree,
																 pQS->ptn->pszgid,
																 SG_pathname__sz(pQS->pPathGraveyard),
																 pTSM_live,
																 "REVERT_RECOVERED_SUBMODULE")  );
					if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
					{
						SG_ERR_CHECK(  SG_treenode_submodule__apply_revert(pCtx,
																		   pQS->pPathGraveyard,
																		   pTSM_live,	// only used for vtable
																		   prd->pvec_use_submodules,
																		   prd->bNoBackups)  );
						bAsIs = SG_FALSE;
					
						// recompute/sniff the submodule to see if there were
						// any problems or dirt that couldn't be reverted.
						// and to get a fresh copy of the pTSM.

						SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pTSM_live);
						SG_ERR_CHECK(  SG_TREENODE_SUBMODULE__SNIFF_DIRECTORY(pCtx,
																			  pQS->pPathGraveyard,
																			  pQS->ptn->pPendingTree->pRepo,
																			  &bIsWD,
																			  &pTSM_live)  );
						SG_ERR_CHECK(  SG_treenode_submodule__is_dirty(pCtx, pTSM_live, &bDirtyPending)  );
						if (bDirtyPending)
						{
							// if still dirty after the revert, don't do step 2.
							if (bDirtyReference && (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT))
								SG_ERR_CHECK(  _pt_log__submodule_revert(pCtx, prd->pPendingTree,
																		 pQS->ptn->pszgid,
																		 SG_pathname__sz(pQS->pPathGraveyard),
																		 pTSM_live,
																		 "SKIPPING_UPDATE_OF_RECOVERED_SUBMODULE_DUE_TO_EXISTING_DIRT")  );
						}
					}
				}

				if (bDirtyReference && !bDirtyPending)
				{
					// dive in and do a 'vv update --rev r' so that we get
					// a clean update to the cset that we think it should
					// have.

					if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
						SG_ERR_CHECK(  _pt_log__submodule_update(pCtx, prd->pPendingTree,
																 pQS->ptn->pszgid,
																 SG_pathname__sz(pQS->pPathGraveyard),
																 pTSM_live,
																 pTSM_baseline,
																 "UPDATE_RECOVERED_SUBMODULE")  );
					if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
					{
						SG_ERR_CHECK(  SG_treenode_submodule__precompute_update(pCtx,
																				pQS->pPathGraveyard,
																				pTSM_live,
																				pTSM_baseline,
																				prd->pvec_use_submodules,
																				prd->pvaMissingDetails,
																				&pTSIDU)  );
						SG_ASSERT(  (pTSIDU != NULL)  );
						SG_ERR_CHECK(  SG_treenode_submodule__apply_update(pCtx,
																		   pQS->pPathGraveyard,
																		   pTSM_live,
																		   pTSM_baseline,
																		   pTSIDU)  );
						bAsIs = SG_FALSE;

#if 0 // TODO 2011/04/04 do we really need this?   THIS CODE HAS NEVER BEEN TESTED.
						// recompute/sniff the submodule to see if there were
						// any problems or dirt that couldn't be reverted.
						// and to get a fresh copy of the pTSM.

						SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pTSM_live);
						SG_ERR_CHECK(  SG_TREENODE_SUBMODULE__SNIFF_DIRECTORY(pCtx,
																			  pQS->pPathGraveyard,
																			  pQS->ptn->pPendingTree->pRepo,
																			  &bIsWD,
																			  &pTSM_live)  );
						
						// if the UPDATE was able to get the submodule back
						// to our expected baseline, we can clear the current-hid
						// so that it doesn't look dirty.  if not, update the current-hid
						// to the live value of the submodule.

						const char * pszHidNew = NULL;
						const char * pszHidOld = NULL;

						SG_ERR_CHECK(  sg_ptnode__get_old_hid(pCtx, pQS->ptn, &pszHidOld)  );
						SG_ERR_CHECK(  SG_treenode_submodule__get_id_ref(pCtx, pTSM_live, &pszHidNew)  );
						bDirtyReference = (strcmp(pszHidNew, pszHidOld) != 0);
						if (bDirtyReference)
							SG_ERR_CHECK(  SG_strpool__add__sz(pCtx, pQS->ptn->pPendingTree->pStrPool, pszHidNew, &pQS->ptn->pszCurrentHid)  );
						else
							pQS->ptn->pszCurrentHid = NULL;

						pQS->ptn->pszHidDisposable = NULL;		// TODO 2011/04/04 This probably doesn't matter.
#endif
					}
				}
			}
		}

		// now that we have either reverted/updated (or not) the
		// contents of the submodule, move it into place in the
		// outer's WD.

		if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
			SG_ERR_CHECK(  _pt_log__submodule_revert(pCtx, prd->pPendingTree,
													 pQS->ptn->pszgid,
													 SG_pathname__sz(pPathFinalDestination),
													 pTSM_baseline,
													 ((bAsIs)
													  ? "RECOVER_AS_IS_SUBMODULE_FROM_GRAVEYARD"
													  : "RECOVER_REVERTED_SUBMODULE_FROM_GRAVEYARD"))  );
		if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
		{
			SG_ERR_CHECK(  SG_fsobj__move__pathname_pathname(pCtx,
															 pQS->pPathGraveyard,
															 pPathFinalDestination)  );
		}

		SG_ERR_CHECK(  _pt_revert__revert_attrbits(pCtx,prd,pPathFinalDestination,pQS->ptn,"RESTORE")  );
		SG_ERR_CHECK(  _pt_revert__revert_xattrs(pCtx,prd,pPathFinalDestination,pQS->ptn,"RESTORE")  );
	}
	else
	{
		// We did not find a candidate submodule WD in the graveyard.
		// We need to do a regular CHECKOUT for this submodule.

		if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
			SG_ERR_CHECK(  _pt_log__submodule_checkout(pCtx,
													   prd->pPendingTree,
													   pQS->ptn->pszgid,
													   SG_pathname__sz(pPathFinalDestination),
													   pTSM_baseline,
													   "RECHECKOUT_DELETED_SUBMODULE")  );

		if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
		{
			SG_ERR_CHECK(  SG_treenode_submodule__precompute_checkout(pCtx,
																	  pPathFinalDestination,
																	  pTSM_baseline,
																	  SG_FALSE, // do not pass down bNoSubmodules
																	  prd->pvec_use_submodules,
																	  prd->pvaMissingDetails,
																	  &pvhCheckoutInstanceData)  );
		
			SG_ERR_CHECK(  SG_treenode_submodule__apply_checkout(pCtx,
																 pPathFinalDestination,
																 pTSM_baseline,
																 pvhCheckoutInstanceData)  );
		}
	}

fail:
	if (pTSIDU)
		SG_ERR_IGNORE(  (*pTSIDU->pfn_dtor)(pCtx, pTSIDU)  );
	SG_VHASH_NULLFREE(pCtx, pvhCheckoutInstanceData);
	SG_PATHNAME_NULLFREE(pCtx, pPathFinalDestination);
	SG_STRING_NULLFREE(pCtx, pStringFinalRepoPath);
	SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pTSM_live);
	SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pTSM_baseline);
}

//////////////////////////////////////////////////////////////////

/**
 * This submodule was LOST from the WD and is being reverted.
 * So we need to recreate it from scratch.  I don't think it
 * is correct to search the graveyard for it (since the user
 * didn't ask us to delete it, so we didn't put it there).
 *
 * Just see if we can do a fresh checkout using the info available.
 *
 */
static void _pt_revert__precompute_recreate_lost_submodule(SG_context * pCtx,
														   struct _pt_revert_data * prd,
														   const SG_treediff2_ObjectData * pOD,
														   struct sg_ptnode * ptn)
{
	SG_pathname * pPath_XXXXXX = NULL;
	SG_vhash * pvhCheckoutInstanceData_XXXXXX = NULL;
	SG_treenode_submodule * pTSM_baseline = NULL;
	struct _queued_subs * pQS = NULL;
	const char * pszRepoPath;
	SG_uint32 nrBefore = 0;
	SG_uint32 nrAfter = 0;
	SG_diffstatus_flags dsFlags;

	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_dsFlags(pCtx, pOD, &dsFlags)  );

	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_old_tsm(pCtx, pOD, &pTSM_baseline)  );
	if (!pTSM_baseline)
	{
		// The entry DID NOT have exist in the baseline.
		// This could happen when you do a:
		//     date > foo
		//     vv add foo
		//     /bin/rm foo
		//     vv revert foo
		// That is an ADDED-but-deleted gets labeled as a LOST
		// item so the REVERT just needs to remove the ptnode.

#if TRACE_REVERT
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "precompute_recreate_lost_submodule: [gid %s] [dsFlags 0x%x] not in baseline\n",
								   ptn->pszgid, dsFlags)  );
#endif

		SG_ASSERT(  (dsFlags & SG_DIFFSTATUS_FLAGS__HACK__LOST_ADDED)  );
	}
	else
	{
#if TRACE_REVERT
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "precompute_recreate_lost_submodule: [gid %s] [dsFlags 0x%x]\n",
								   ptn->pszgid, dsFlags)  );
#endif

		SG_ERR_CHECK(  SG_varray__count(pCtx, prd->pvaMissingDetails, &nrBefore)  );

		// We get the REPO-PATH of the item and build an absolute
		// path; THIS IS FOR DEBUG/LOGGING PURPOSES ONLY.  Since
		// we haven't done any WD manipulation yet, this is the path
		// to the submodule in the current/dirty outer WD; it may or
		// may not match the path where the submodule will be when
		// we finish the revert.
		// 
		// Also, disregard the returned checkout-instance data.  We'll
		// rebuild this later if/when we actually do the checkout when
		// we have valid pathnames.

		SG_ERR_CHECK(  SG_treediff2__ObjectData__get_repo_path(pCtx, pOD, &pszRepoPath)  );
		SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path(pCtx,
																			 ptn->pPendingTree->pPathWorkingDirectoryTop,
																			 pszRepoPath,
																			 &pPath_XXXXXX)  );
		SG_ERR_CHECK(  SG_treenode_submodule__precompute_checkout(pCtx,
																  pPath_XXXXXX,
																  pTSM_baseline,
																  SG_FALSE, // do not pass down bNoSubmodules
																  prd->pvec_use_submodules,
																  prd->pvaMissingDetails,
																  &pvhCheckoutInstanceData_XXXXXX)  );
		SG_ERR_CHECK(  SG_varray__count(pCtx, prd->pvaMissingDetails, &nrAfter)  );
		if (nrAfter > nrBefore)
			goto done;

		// we can do the REVERT *IFF* we did not append anything to prd->pvaMissingDetails.
	}

	// TODO 2011/04/06 Consider putting pTSM_baseline in the pQS.

	if (!prd->pvec_queued_subs)
		SG_ERR_CHECK(  SG_VECTOR__ALLOC(pCtx, &prd->pvec_queued_subs, 10)  );
	SG_ERR_CHECK(  _queued_subs__alloc(pCtx, &pQS, pOD, ptn, NULL)  );
	SG_ERR_CHECK(  SG_vector__append(pCtx, prd->pvec_queued_subs, pQS, NULL)  );
	pQS = NULL;		// vector owns it now

done:
	// the reason that we do not throw now is so that we can do the precompute on all of
	// the peer submodules and eventually do one big throw with (hopefully) all of the
	// details that the user will need (--usesub hints and/or clone/pull requests) to
	// make to let us retry.

#if TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "precompute_recreate_lost_submodule: [gid %s] [nrMissingDetails %d]\n",
							   ptn->pszgid, (nrAfter - nrBefore))  );
#endif
	;

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPath_XXXXXX);
	SG_VHASH_NULLFREE(pCtx, pvhCheckoutInstanceData_XXXXXX);
	SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pTSM_baseline);
}

static void _pt_revert__recreate_lost_submodule(SG_context * pCtx,
												struct _pt_revert_data * prd,
												struct _queued_subs * pQS)
{
	SG_string * pStringFinalRepoPath = NULL;
	SG_pathname * pPathFinalDestination = NULL;
	SG_vhash * pvhCheckoutInstanceData = NULL;
	SG_treenode_submodule * pTSM_baseline = NULL;
	struct sg_ptnode * ptn_owned = NULL;
	SG_diffstatus_flags dsFlags;
	SG_timestamp_cache * pTSC = NULL;		// we do not own this

	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_dsFlags(pCtx,pQS->pOD,&dsFlags)  );

	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_old_tsm(pCtx, pQS->pOD, &pTSM_baseline)  );
	if (!pTSM_baseline)
	{
#if TRACE_REVERT
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "_pt_revert__restore_lost_submodule: [gid %s] no name [dsFlags 0x%x].\n",
								   pQS->ptn->pszgid, dsFlags)  );
#endif
		SG_ASSERT(  (dsFlags & SG_DIFFSTATUS_FLAGS__HACK__LOST_ADDED)  );

		if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
			SG_ERR_CHECK(  _pt_log__unadd_item(pCtx,prd->pPendingTree,
											   pQS->ptn->pszgid,
											   "UNADD")  );

		// TODO 2011/04/07 Shouldn't we guard this stuff with "if __DO_IT..." ?
		// TODO 2011/04/11 Also, submodules shouldn't have a TSC entry.

		SG_ERR_CHECK(  SG_pendingtree__get_timestamp_cache(pCtx, pQS->ptn->pPendingTree, &pTSC)  );
		SG_ERR_CHECK(  SG_timestamp_cache__remove(pCtx, pTSC, pQS->ptn->pszgid)  );
		SG_ERR_CHECK(  SG_rbtree__remove__with_assoc(pCtx,pQS->ptn->pCurrentParent->prbItems,pQS->ptn->pszgid,(void **)&ptn_owned)  );
		SG_ASSERT(  (ptn_owned == pQS->ptn)  );
		SG_PTNODE_NULLFREE(pCtx,ptn_owned);
	}
	else
	{
		// We need to do a regular CHECKOUT to re-create this LOST submodule.

		SG_ERR_CHECK(  sg_ptnode__get_repo_path(pCtx, pQS->ptn, &pStringFinalRepoPath)  );
		SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path(pCtx,
																			 prd->pPendingTree->pPathWorkingDirectoryTop,
																			 SG_string__sz(pStringFinalRepoPath),
																			 &pPathFinalDestination)  );

#if TRACE_REVERT
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "recreate_lost_submodule: [gid %s] [dsFlags 0x%x] %s\n",
								   pQS->ptn->pszgid, dsFlags,
								   SG_pathname__sz(pPathFinalDestination))  );
#endif

		if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
			SG_ERR_CHECK(  _pt_log__submodule_checkout(pCtx,
													   prd->pPendingTree,
													   pQS->ptn->pszgid,
													   SG_pathname__sz(pPathFinalDestination),
													   pTSM_baseline,
													   "RECHECKOUT_LOST_SUBMODULE")  );

		if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
		{
			SG_ERR_CHECK(  SG_treenode_submodule__precompute_checkout(pCtx,
																	  pPathFinalDestination,
																	  pTSM_baseline,
																	  SG_FALSE, // do not pass down bNoSubmodules
																	  prd->pvec_use_submodules,
																	  prd->pvaMissingDetails,
																	  &pvhCheckoutInstanceData)  );
		
			SG_ERR_CHECK(  SG_treenode_submodule__apply_checkout(pCtx,
																 pPathFinalDestination,
																 pTSM_baseline,
																 pvhCheckoutInstanceData)  );
		}
	}

fail:
	SG_VHASH_NULLFREE(pCtx, pvhCheckoutInstanceData);
	SG_PATHNAME_NULLFREE(pCtx, pPathFinalDestination);
	SG_STRING_NULLFREE(pCtx, pStringFinalRepoPath);
	SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pTSM_baseline);
	SG_PTNODE_NULLFREE(pCtx,ptn_owned);
}

//////////////////////////////////////////////////////////////////

static void _pt_revert__bind_right_side_of_diff_to_source(SG_context * pCtx,
									struct _pt_revert_data * prd,
									const char * pszGidObject,
									const SG_treediff2_ObjectData * pOD,
									SG_bool bActive)
{
	const char * pszGidObject_CurrentParent;
	const SG_treediff2_ObjectData * pOD_CurrentParent;
	const char * pszRepoPath_CurrentParent_CurrentPath;
	const char * pszEntryName_Current;
	SG_treenode_entry_type tneType;
	SG_bool bFound;

	// REVERT is modeled on making a diff(baseline,wd) and then UNDOing {all, some} of the changes.
	//        So the "baseline" is on the left and the "wd" is on the right.
	//        So in this context, the CURRENT means the current WD and Ndx_Net in the treediff.
	//        The given entry is present in the WD in the current parent.
	//
	// The entry is present in the "current" part of the TreeDiff/ObjectData.
	// This is the right side of the diff and reflects the current state in
	// the working directory.
	//
	// Get the "current" values and set the "source" fields in _inv_dir/_inv_entry.

	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_name(pCtx,pOD,&pszEntryName_Current)  );

	// Get gid-object of the CURRENT parent directory and the repo-path of that directory
	// *as they currently exist* in the CURRENT version of the tree and on disk.

	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_parent_gid(pCtx,pOD,&pszGidObject_CurrentParent)  );
	SG_ERR_CHECK(  SG_treediff2__get_ObjectData(pCtx,prd->pTreeDiff,pszGidObject_CurrentParent,&bFound,&pOD_CurrentParent)  );
	SG_ASSERT(  (bFound)  );	// TODO think about reverting the initial commit which would un-add "@/"....
	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_repo_path(pCtx,pOD_CurrentParent,&pszRepoPath_CurrentParent_CurrentPath)  );

	SG_ERR_CHECK_RETURN(  SG_treediff2__ObjectData__get_tneType(pCtx, pOD, &tneType)  );

#if TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
							   "_pt_revert__bind_right_side_of_diff_to_source: [gid %s] claming source [active %d][%s][%s] [tneType %d]\n",
							   pszGidObject,
							   bActive,
							   pszRepoPath_CurrentParent_CurrentPath,
							   pszEntryName_Current,
							   tneType)  );
#endif

	SG_ERR_CHECK(  SG_inv_dirs__bind_source(pCtx,prd->pInvDirs,
											pszGidObject_CurrentParent,pszRepoPath_CurrentParent_CurrentPath,
											pszGidObject,pszEntryName_Current,
											(tneType == SG_TREENODEENTRY_TYPE_DIRECTORY),
											(void *)pOD,
											bActive,
											NULL,NULL)  );

fail:
	return;
}

static void _pt_revert__bind_left_side_of_diff_to_target(SG_context * pCtx,
									struct _pt_revert_data * prd,
									const char * pszGidObject,
									const SG_treediff2_ObjectData * pOD,
									SG_bool bActive)
{
	const char * pszGidObject_OriginalParent;
	const SG_treediff2_ObjectData * pOD_OriginalParent;
	const char * pszRepoPath_OriginalParent_CurrentPath;
	const char * pszEntryName_Original;
	SG_bool bFound;

	// REVERT is modeled on making a diff(baseline,wd) and then UNDOing {all, some} of the changes.
	//        So the "baseline" is on the left and the "wd" is on the right.
	//        So in this context, the ORIGINAL means the baseline and Ndx_Orig in the treediff.
	//        The given entry needs to exist after everything is done, but it may or may not
	//        exist in the current WD and it may or may not be in the same directory and the
	//        final directory may or may not exist in the current WD and may or may not have
	//        the same pathname in the result.
	//
	// The entry is present in the "original" part of the TreeDiff/ObjectData.
	// This is the left side of the diff and reflects the original/baseline
	// state.
	//
	// Get the "original" values and set the "target" fields in _inv_dir/_inv_entry.

	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_old_name(pCtx,pOD,&pszEntryName_Original)  );
	if (!pszEntryName_Original || !*pszEntryName_Original)
	{
		// The entry DID NOT have a name in the baseline.
		// This could happen when you do a:
		//     date > foo
		//     vv add foo
		//     /bin/rm foo
		//     vv revert foo
		// That is an ADDED-but-deleted gets labeled as a LOST
		// item so the REVERT just needs to remove the ptnode.
		// Anyway, for our purposes here, there is no name for
		// the left side to claim.

#if TRACE_REVERT
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "_pt_revert__bind_left_side_of_diff_to_target: [gid %s] no name to claim in target [active %d]\n",
								   pszGidObject,
								   bActive)  );
#endif
		return;
	}

	// Get gid-object of the ORIGINAL parent directory and the CURRENT repo-path of that directory
	// *as they exist* in the current version of the tree and on disk.  (this is necessary so
	// that we can find the directory on disk when one or more of its parents were renamed/moved.)

	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_moved_from_gid(pCtx,pOD,&pszGidObject_OriginalParent)  );
	SG_ERR_CHECK(  SG_treediff2__get_ObjectData(pCtx,prd->pTreeDiff,pszGidObject_OriginalParent,&bFound,&pOD_OriginalParent)  );
	SG_ASSERT_RELEASE_RETURN( (bFound) );
	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_repo_path(pCtx,pOD_OriginalParent,&pszRepoPath_OriginalParent_CurrentPath)  );

#if TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
							   "_pt_revert__bind_left_side_of_diff_to_target: [gid %s] claming target [active %d][%s][%s]\n",
							   pszGidObject,
							   bActive,
							   pszRepoPath_OriginalParent_CurrentPath,
							   pszEntryName_Original)  );
#endif

	SG_ERR_CHECK(  SG_inv_dirs__bind_target(pCtx,prd->pInvDirs,
											pszGidObject_OriginalParent,
											pszGidObject,pszEntryName_Original,(void *)pOD,
											bActive,
											NULL,NULL)  );

fail:
	return;
}

//////////////////////////////////////////////////////////////////

/**
 * Since we parked entries that have collisions in the swap directory,
 * we cannot use the normal sg_ptnode__get_repo_path() and the
 * repo_path --> absolute_path code.  rather, we need something that
 * computes the "transient absolute path".  for example, if we have:
 *     @/dir1/dir2/dir3/dir4/dir5/dir6/file1
 *     @/dirA/dirB/fileA
 * and the user does:
 *     vv rm @/dir1/dir2/dir3/dir4/dir5
 *     vv mv @/dir1/dir2/dir3 @/dir1/dir2/xxx
 *     vv mv @/dirA @/dir1/dir2/dir3
 * then we have an collision/order problem on the name @/dir3.
 * so we move dir3 --> <swap> as we juggle everything and revert
 * the changes.  When we try to un-delete dir5 (and everything
 * within it) (and depending on the random gids involved), dir3
 * may or may not have been handled:
 * [1] it could refer to dirA, if we got to dir5 before dirA
 * [2] it could be missing, if dirA has already been restored
 * [3] it could refer to the original dir3, if it has already been
 *     restored.
 *
 * So we need to be able to say the "transient absolute pathname"
 * for dir5 is <swap>/dir4/dir5 in cases [1] and [2] and
 * <wd>/dir1/dir2/dir3/dir4/dir5 in case [3].
 */
static void _pt_revert__get_transient_path(SG_context * pCtx,
										   struct _pt_revert_data * prd,
										   struct sg_ptnode * ptn,
										   SG_pathname ** ppPath)
{
	SG_pathname * pPath_Allocated = NULL;
	SG_inv_entry * pInvEntry = NULL;
	const char * pszEntryName = NULL;
	const SG_pathname * pPathParked = NULL;		// we do not own this
	SG_bool bFound, bExists;

	if (!ptn->pszgid || !ptn->pszgid[0])
	{
		// ptn is the super-root
		SG_ERR_CHECK_RETURN(  SG_PATHNAME__ALLOC__COPY(pCtx,ppPath,prd->pPendingTree->pPathWorkingDirectoryTop)  );
		return;
	}

	SG_ERR_CHECK_RETURN(  SG_inv_dirs__fetch_source_binding(pCtx,prd->pInvDirs,ptn->pszgid,&bFound,&pInvEntry,NULL)  );
	if (bFound)
		SG_ERR_CHECK(  SG_inv_entry__get_parked_path__ref(pCtx, pInvEntry, &pPathParked)  );
	if (pPathParked)
	{
		// if the swap path still exists on disk, then we haven't restored it into the working-directory yet.
		//
		// TODO 3/23/10 does this "if exists" cause our answer to change when doing "--test" only vs
		// TODO 3/23/10 when we are actually doing it and have "--verbose" turned on?
		// TODO
		// TODO 3/25/10 when __process_structual_changes does (or pretends to do) the work on an
		// TODO 3/25/10 pInvEntry, should we free pParkedPath -- because in theory it is no longer
		// TODO 3/25/10 parked.  that would eliminate the discrepancy between --test and --verbose
		// TODO 3/25/10 runs, right?  It also help us split the REVERT code into separate
		// TODO 3/25/10 build-the-list-of-stuff-to-do and do-the-list steps.

		SG_ERR_CHECK_RETURN(  SG_fsobj__exists__pathname(pCtx,pPathParked,&bExists,NULL,NULL)  );
		if (bExists)
		{
			SG_ERR_CHECK_RETURN(  SG_PATHNAME__ALLOC__COPY(pCtx,ppPath,pPathParked)  );
			return;
		}
	}

	// TODO 3/23/10 should the __get_name, strcmp, __append__from_sz all be under the "if pCurrentParent" ?
	// TODO 3/25/10 OR should we just ASSERT(ptn->pCurrentParent) ?

	if (ptn->pCurrentParent)
		SG_ERR_CHECK(  _pt_revert__get_transient_path(pCtx,prd,ptn->pCurrentParent,&pPath_Allocated)  );

	SG_ERR_CHECK(  sg_ptnode__get_name(pCtx,ptn,&pszEntryName)  );	// get current or fallback to baseline
	if (pszEntryName && (strcmp(pszEntryName,"@") != 0))
		SG_ERR_CHECK(  SG_pathname__append__from_sz(pCtx,pPath_Allocated,pszEntryName)  );

	*ppPath = pPath_Allocated;
	return;

fail:
	SG_PATHNAME_NULLFREE(pCtx,pPath_Allocated);
}

//////////////////////////////////////////////////////////////////

static void _pt_revert__backup_entry_for_un_add_or_un_find(SG_context * pCtx,
														   struct _pt_revert_data * prd,
														   const char * pszGidObject,
														   const SG_treediff2_ObjectData * pOD,
														   struct sg_ptnode * ptn,
														   SG_bool bActive,
														   SG_bool * pbWasDeleted)
{
	SG_pathname * pPathTrans = NULL;
	SG_pathname * pPathBackup = NULL;
	SG_pathname * pPathMyTrans = NULL;
	SG_inv_entry * pInvEntry = NULL;
	const char * pszEntryName_Current = NULL;
	const SG_pathname * pPathParked = NULL;		// we do not own this
	SG_inv_entry_flags inv_entry_flags;
	SG_bool bIsDisposable = SG_FALSE;
	SG_bool bFound;
	SG_bool bWasDeleted = SG_FALSE;

	SG_ERR_CHECK(  SG_inv_dirs__fetch_source_binding(pCtx,prd->pInvDirs,pszGidObject,&bFound,&pInvEntry,NULL)  );
	SG_ASSERT(  (bFound)  );
	SG_ERR_CHECK(  SG_inv_entry__get_flags(pCtx, pInvEntry, &inv_entry_flags)  );

	// in one sense an un-add means that we just want to
	// revert an uncommited 'vv add' and prevent version
	// control from managing this entry.  (and normally
	// the entry would/should have FOUND status.)
	//
	// HOWEVER, if *we* added the entry as part of a MERGE
	// (say it was added in the other branch), then we
	// really *don't* want it lying around afterwards.
	// (unless they have edited it after the merge.)
	//
	// so if we know that the content was created/generated
	// by automatic means (and can be recreated again if
	// necessary), then we delete it.

	switch (ptn->type)
	{
	default:
		SG_ERR_THROW(  SG_ERR_NOTIMPLEMENTED  );
		break;

	case SG_TREENODEENTRY_TYPE_DIRECTORY:
		// if MERGE marked the directory as disposable, we might be
		// able to delete it (in a later pass) rather than giving it
		// a ~sg00~ name or leaving it as "found" trash.  for now,
		// go ahead and park/unpark it as if we're going to keep it.
		if (ptn->bDisposableDir)
			prd->nrDisposableDirs++;
		bWasDeleted = SG_FALSE;
		break;

	case SG_TREENODEENTRY_TYPE_SUBMODULE:
		// we DO NOT --no-submodules here because we consider
		// the add/delete of a submodule to be OUTSIDE the
		// submodule and the --no-submodules flag to refer to
		// the contents OF the submodule.
		SG_ERR_CHECK(  _pt_revert__is_disposable_submodule(pCtx, ptn, &bIsDisposable)  );
#if TRACE_REVERT
		SG_ERR_CHECK(  SG_console(pCtx, SG_CS_STDERR,
								  "TODO in _pt_revert__un_add for submodule...: [saved flags %d][temp flags %d][disposable %d][active %d]\n",
								  ptn->saved_flags, ptn->temp_flags, bIsDisposable, bActive)  );
#endif
		if (bIsDisposable)
		{
			// if it was ADDED by MERGE and hasn't been modified since
			// the merge, then we say it is DISPOSABLE.

			if (bActive)
			{
				// when DISPOSABLE and ACTIVELY reverted, we can just
				// delete it rather than worrying about a backup name.
				// 
				// TODO 2011/03/28 for now as a safety thing, we move
				// TODO            it to the submodule graveyard rather
				// TODO            than deleting it.

				bWasDeleted = SG_TRUE;

				if (inv_entry_flags & SG_INV_ENTRY_FLAGS__PARKED_FOR_SWAP)
				{
					SG_ERR_CHECK(  SG_inv_entry__get_parked_path__ref(pCtx, pInvEntry, &pPathParked)  );
#if TRACE_REVERT
					SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
											   "_pt_revert__backup_entry: dispose of parked submodule: %s\n",
											   SG_pathname__sz(pPathParked))  );
#endif
					if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
						SG_ERR_CHECK(  _pt_log__remove_item(pCtx,
															prd->pPendingTree,
															ptn->pszgid,
															SG_pathname__sz(pPathParked),
															"REMOVE_PARKED_DISPOSABLE_SUBMODULE")  );
					if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
						SG_ERR_CHECK(  SG_workingdir__move_submodule_to_graveyard(pCtx,
																				  ptn->pPendingTree->pPathWorkingDirectoryTop,
																				  pPathParked,
																				  ptn->pszgid,
																				  NULL)  );
				}
				else
				{
					// we're not parked, but a parent directory might be, so
					// we can't do normal repo-path==>absolute-path.
					SG_ERR_CHECK(  _pt_revert__get_transient_path(pCtx,
																  prd,
																  ptn,
																  &pPathMyTrans)  );
#if TRACE_REVERT
					SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
											   "_pt_revert__backup_entry: dispose of non-parked submodule: %s\n",
											   SG_pathname__sz(pPathMyTrans))  );
#endif
					if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
						SG_ERR_CHECK(  _pt_log__remove_item(pCtx,
															prd->pPendingTree,
															ptn->pszgid,
															SG_pathname__sz(pPathMyTrans),
															"REMOVE_DISPOSABLE_SUBMODULE")  );
					if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
						SG_ERR_CHECK(  SG_workingdir__move_submodule_to_graveyard(pCtx,
																				  ptn->pPendingTree->pPathWorkingDirectoryTop,
																				  pPathMyTrans,
																				  ptn->pszgid,
																				  NULL)  );
				}

				goto done;
			}
			else
			{
				// disposable, but not actively being reverted, so we
				// can't delete it.  let the backup code take care of
				// any name contention/collisions.
			}
		}
		else
		{
			// not disposable.  active or not active.  let the backup
			// code take care of any name contention/collisons.
		}
		break;

	case SG_TREENODEENTRY_TYPE_REGULAR_FILE:
	case SG_TREENODEENTRY_TYPE_SYMLINK:
		SG_ERR_CHECK(  _pt_revert__is_disposable_file_symlink(pCtx, ptn, &bIsDisposable)  );
#if TRACE_REVERT
		SG_ERR_CHECK(  SG_console(pCtx, SG_CS_STDERR,
								  "TODO in _pt_revert__un_add for file/symlink...: [saved flags %d][temp flags %d][disposable %d][active %d]\n",
								  ptn->saved_flags, ptn->temp_flags, bIsDisposable, bActive)  );
#endif
		if (bIsDisposable)
		{
			if (bActive)
			{
				bWasDeleted = SG_TRUE;

				if (inv_entry_flags & SG_INV_ENTRY_FLAGS__PARKED_FOR_SWAP)
				{
					SG_ERR_CHECK(  SG_inv_entry__get_parked_path__ref(pCtx, pInvEntry, &pPathParked)  );
#if TRACE_REVERT
					SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
											   "_pt_revert__backup_entry: dispose of parked: %s\n",
											   SG_pathname__sz(pPathParked))  );
#endif
					if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
						SG_ERR_CHECK(  _pt_log__remove_item(pCtx,
															prd->pPendingTree,
															ptn->pszgid,
															SG_pathname__sz(pPathParked),
															"REMOVE_PARKED_DISPOSABLE")  );
					if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
						SG_ERR_CHECK(  SG_fsobj__remove__pathname(pCtx, pPathParked)  );
				}
				else
				{
					// we're not parked, but a parent directory might be, so
					// we can't do normal repo-path==>absolute-path.
					SG_ERR_CHECK(  _pt_revert__get_transient_path(pCtx,
																  prd,
																  ptn,
																  &pPathMyTrans)  );
#if TRACE_REVERT
					SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
											   "_pt_revert__backup_entry: dispose of non-parked: %s\n",
											   SG_pathname__sz(pPathMyTrans))  );
#endif
					if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
						SG_ERR_CHECK(  _pt_log__remove_item(pCtx,
															prd->pPendingTree,
															ptn->pszgid,
															SG_pathname__sz(pPathMyTrans),
															"REMOVE_DISPOSABLE")  );
					if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
						SG_ERR_CHECK(  SG_fsobj__remove__pathname(pCtx, pPathMyTrans)  );
				}

				goto done;
			}
			else
			{
			}
		}
		else
		{
		}
		break;
	}

	// not generated or they have edited it afterwards.
	
	// back it up if there is entryname contention (for both un-add and un-find).
	// 
	// we don't care about PARKED_FOR_CYCLE.  we only need that flag so that we
	// can safely get from the SOURCE tree layout to the TARGET tree layout without
	// worrying about ordering problems as we un-wind the moves; there is no issue
	// about collisions and/or backing up stuff.
	//
	// we DO care about PARKED_FOR_SWAP.  this flag means that there is contention
	// for a particular ENTRYNAME that we need to address.

	if (inv_entry_flags & SG_INV_ENTRY_FLAGS__PARKED_FOR_SWAP)
	{
		SG_ERR_CHECK(  SG_inv_entry__get_parked_path__ref(pCtx, pInvEntry, &pPathParked)  );
		SG_ASSERT(  (pPathParked)  );

		// someone else wanted our entryname so we moved the source to swap.
		// currently, both parties must be active for this to happen.  see
		// _inv_entry__check_for_swap_cb() and TODO [4] for other possibilities.

		SG_ASSERT(  (inv_entry_flags & SG_INV_ENTRY_FLAGS__ACTIVE_SOURCE)  );
		// i need to relax the assert on an active target to include a transient
		// portability collision.  there's a hidden assert here that there is another
		// entry pInvEntry_j in the table who's entryname is "equivalent" to ours
		// and has the (__ACTIVE_TARGET & __PORT_TRANSIENT) bits set.
		SG_ASSERT(  (inv_entry_flags & (SG_INV_ENTRY_FLAGS__ACTIVE_TARGET | SG_INV_ENTRY_FLAGS__PORT_TRANSIENT))  );

		// get the current disk path of the SOURCE (aka TARGET) parent directory.
		// we can't use the normal repo-path-->absolute-path code because an
		// ancestor might be parked in swap.

		SG_ERR_CHECK(  _pt_revert__get_transient_path(pCtx,prd,ptn->pCurrentParent,&pPathTrans)  );

		// we can't use the normal sg_pendingtree__clear_obstacle() because we
		// might not have seen it yet in the loop that called us (and therefore
		// the other entry hasn't renamed itself to the contested entryname).
		//
		// we can't use the normal sg_pendingtree__rename_to_backup() because
		// the swap path is not in the result directory and __rename_to_backup()
		// wants to do a local rename within a directory.
		//
		//////////////////////////////////////////////////////////////////
		// This case happens when you do something like:
		// 
		// $ vv init foo1 .
		// $ date > d
		// $ vv add d
		// $ vv commit -m "added d"
		// $ vv rename d d1
		// $ date > d
		// $ vv status
		// Renamed:  @/d1
		//      # was d
		//   Found:  @/d
		// $ vv revert --all --verbose
		//
		// Where we want to un-rename d1 ==> d, but there is an unknown item
		// using the entryname d.
		//
		// One could say that found d has nothing to do with the original d
		// so we should give the found d an new unique name so that we can
		// complete the revert and have clean status (WRT the things that
		// are already under version control).  This is different from a
		// regular backup (where we are just trying to save the contents of
		// a dirty file before we restore the original content).
		//
		// SO WE DO NOT USE prd->bNoBackups HERE.

		SG_ERR_CHECK(  SG_treediff2__ObjectData__get_name(pCtx,pOD,&pszEntryName_Current)  );
		SG_ERR_CHECK(  SG_pathname__append__from_sz(pCtx,pPathTrans,pszEntryName_Current)  );
		SG_ERR_CHECK(  SG_workingdir__calc_backup_found_name(pCtx,pPathTrans,&pPathBackup)  );

#if TRACE_REVERT
		SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
								   ("_pt_revert__backup_entry: backing-up [%s]\n"
									"                                  as [%s]\n"),
								   SG_pathname__sz(pPathParked),
								   SG_pathname__sz(pPathBackup))  );
#endif
		if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
			SG_ERR_CHECK(  _pt_log__move(pCtx,prd->pPendingTree,
										 ptn->pszgid,
										 SG_pathname__sz(pPathParked),
										 SG_pathname__sz(pPathBackup),
										 "NAME_CONTENTION")  );
		if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
			SG_ERR_CHECK(  SG_fsobj__move__pathname_pathname(pCtx,pPathParked,pPathBackup)  );
	}
	else
	{
		// no one else wanted our entryname in the directory.  so we have nothing
		// to do.  we may remove the ptnode (for ADDS), but we don't need to do
		// anything the directory.
	}

done:
	*pbWasDeleted = bWasDeleted;

fail:
	SG_PATHNAME_NULLFREE(pCtx,pPathTrans);
	SG_PATHNAME_NULLFREE(pCtx,pPathMyTrans);
	SG_PATHNAME_NULLFREE(pCtx,pPathBackup);
}

//////////////////////////////////////////////////////////////////

static void _pt_revert__restore_entry_for_un_delete_or_un_lost(SG_context * pCtx,
															   struct _pt_revert_data * prd,
															   struct sg_ptnode * ptn)
{
	SG_pathname * pPathTrans = NULL;
	SG_file * pFile = NULL;
	SG_string * pStringLink = NULL;
	SG_byte * pBytes = NULL;
	const char * pszEntryName_old = NULL;
	struct sg_ptnode * ptnDesiredParent = NULL;
#ifdef DEBUG
	SG_bool bFound;
#endif

	// we must assume that any entryname collisions were properly handled by
	// the SOURCE side and parked in swap; that is, we can blindly restore the
	// missing entries without collision/order problems.
	//
	// we also assume that _undelete_skeleton() has done its job and that our
	// parent directory exists on disk (either in the WD or somewhere in the swap).

	// we fetch the old-name because we want the TARGET version.  for DELETES, this should be
	// the same as the SOURCE version (because we don't allow DELETE+RENAME); for LOSTS, they
	// could be different, but that doesn't matter.

	SG_ERR_CHECK(  sg_ptnode__get_old_name(pCtx, ptn, &pszEntryName_old)  );
	if (!pszEntryName_old || !*pszEntryName_old)
	{
		// This entry DID NOT have a name in the baseline.
		// This could happen when you do a:
		//     date > foo
		//     vv add foo
		//     /bin/rm foo
		//     vv revert foo
		// That is an ADDED-but-deleted gets labeled as a LOST
		// item so the REVERT just needs to remove the ptnode
		// and not mess with the WD.

#if TRACE_REVERT
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   "_pt_revert__restore_entry_for_un_delete_or_un_lost: [gid %s] no name.\n",
								   ptn->pszgid)  );
#endif

		return;
	}

	// we cannot use sg_ptnode__get() because we still cannot use repo-path-->absolute
	// yet (because an ancestor directory (with a higher gid) may still be in the swap).
	// we also don't want the recursive stuff.

	// get the current disk path of the DESIRED parent directory.  (because we are after the swap
	// parking and the restore has not finished yet, we can't use the normal repo-path-->absolute-path
	// code because an ancestor directory might still be parked in swap.

	if (ptn->pszgidMovedFrom)
	{
		// pending-tree allows LOST+MOVE.  (currently, pending-tree does not allow DELETE+MOVE,
		// but that doesn't matter for our purposes here.)  but i'm going to assert that we have
		// a LOST for now.

		SG_ASSERT(  (ptn->temp_flags & sg_PTNODE_TEMP_FLAG_IMPLICIT_DELETE)  );

		// we have a +MOVE, so we need choose the TARGET parent directory where we were moved from.
		// This is a little more lookup than the normal case.

		SG_ERR_CHECK(  sg_pendingtree__find_by_gid(pCtx,prd->pPendingTree,ptn->pszgidMovedFrom,&ptnDesiredParent)  );

		// TODO do we need to assert that the TARGET parent will exist when we're done?
		// TODO that is, is it possible for the TARGET parent to be INACTIVE (excluded
		// TODO from the RESTORE) and was DELETED/LOST?
	}
	else
	{
		// otherwise the SOURCE parent and the TARGET parent are the same (not MOVED).

		ptnDesiredParent = ptn->pCurrentParent;
	}

	SG_ERR_CHECK(  _pt_revert__get_transient_path(pCtx,prd,ptnDesiredParent,&pPathTrans)  );

#if defined(DEBUG)
	if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
	{
		// if all is right, our DESIRED parent directory will already exist.
		SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx,pPathTrans,&bFound,NULL,NULL)  );
		if (!bFound)
			SG_ERR_THROW2(  SG_ERR_ASSERT,
							(pCtx, "_pt_revert__restore_for_un_delete_or_un_lost: [pathTrans %s] not found.",
							 SG_pathname__sz(pPathTrans))  );
	}
#endif

	// because we were not present in the SOURCE (dirty WD) version of the tree, we did not claim a
	// SOURCE-entryname and thus cannot be parked in swap.  furthermore, since we claimed the
	// TARGET-(baseline)-entryname, we own it cleanly and can create the directory without having to
	// worry about backup names or collisions.

	SG_ERR_CHECK(  SG_pathname__append__from_sz(pCtx, pPathTrans, pszEntryName_old)  );

	if (ptn->type == SG_TREENODEENTRY_TYPE_DIRECTORY)
	{
		// _undelete_skeleton() should already have done the mkdir.

#if defined(DEBUG)
		if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
		{
			SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx,pPathTrans,&bFound,NULL,NULL)  );
			SG_ASSERT(  (bFound)  );
		}
#endif
		// restore the ATTR bits and XATTRS that were saved with the entry.

		SG_ERR_CHECK(  _pt_revert__revert_attrbits(pCtx,prd,pPathTrans,ptn,"RESTORE")  );
		SG_ERR_CHECK(  _pt_revert__revert_xattrs(pCtx,prd,pPathTrans,ptn,"RESTORE")  );
	}
	else if (ptn->type == SG_TREENODEENTRY_TYPE_REGULAR_FILE)
	{
		const char* pszidhid = NULL;

		SG_ERR_CHECK(  SG_treenode_entry__get_hid_blob(pCtx, ptn->pBaselineEntry, &pszidhid)  );

#if TRACE_REVERT
		SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
								   ("_pt_revert__restore_entry: restoring file contents [gid %s]\n"
									"                                                   [hid %s]\n"
									"                                                to %s\n"),
								   ptn->pszgid,pszidhid,SG_pathname__sz(pPathTrans))  );
#endif
		if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
			SG_ERR_CHECK(  _pt_log__restore_file(pCtx,prd->pPendingTree,
												 ptn->pszgid,
												 SG_pathname__sz(pPathTrans),
												 pszidhid,
												 "RESTORE")  );
		if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
		{
			SG_ERR_CHECK(  SG_file__open__pathname(pCtx, pPathTrans, SG_FILE_WRONLY | SG_FILE_CREATE_NEW, 0644, &pFile)  );
			SG_ERR_CHECK(  SG_repo__fetch_blob_into_file(pCtx, ptn->pPendingTree->pRepo, pszidhid,  pFile, NULL)  );
			SG_FILE_NULLCLOSE(pCtx, pFile);
		}

		// restore the ATTR bits and XATTRS that were saved with the entry.

		SG_ERR_CHECK(  _pt_revert__revert_attrbits(pCtx,prd,pPathTrans,ptn,"RESTORE")  );
		SG_ERR_CHECK(  _pt_revert__revert_xattrs(pCtx,prd,pPathTrans,ptn,"RESTORE")  );
	}
	else if (ptn->type == SG_TREENODEENTRY_TYPE_SYMLINK)
	{
		const char* pszidhid = NULL;
		SG_uint64 iLenBytes = 0;

		SG_ERR_CHECK(  SG_treenode_entry__get_hid_blob(pCtx, ptn->pBaselineEntry, &pszidhid)  );

#if TRACE_REVERT
		SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
								   ("_pt_revert__restore_entry: restoring symlink [gid %s]\n"
									"                                             [hid %s]\n"
									"                                          to %s\n"),
								   ptn->pszgid,pszidhid,SG_pathname__sz(pPathTrans))  );
#endif

		SG_ERR_CHECK(  SG_repo__fetch_blob_into_memory(pCtx, ptn->pPendingTree->pRepo, pszidhid, &pBytes, &iLenBytes)  );
		SG_ERR_CHECK(  SG_STRING__ALLOC__BUF_LEN(pCtx, &pStringLink, pBytes, (SG_uint32) iLenBytes)  );
		SG_NULLFREE(pCtx, pBytes);

		if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
			SG_ERR_CHECK(  _pt_log__restore_symlink(pCtx,prd->pPendingTree,
													ptn->pszgid,
													SG_pathname__sz(pPathTrans),
													SG_string__sz(pStringLink),
													"RESTORE")  );
		if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
			SG_ERR_CHECK(  SG_fsobj__symlink(pCtx, pStringLink, pPathTrans)  );

		// restore the ATTR bits and XATTRS that were saved with the entry.

		SG_ERR_CHECK(  _pt_revert__revert_attrbits(pCtx,prd,pPathTrans,ptn,"RESTORE")  );
		SG_ERR_CHECK(  _pt_revert__revert_xattrs(pCtx,prd,pPathTrans,ptn,"RESTORE")  );

		SG_STRING_NULLFREE(pCtx, pStringLink);
	}
	else if (ptn->type == SG_TREENODEENTRY_TYPE_SUBMODULE)
	{
	}
	else
	{
		SG_ERR_THROW(  SG_ERR_INVALID_PENDINGTREE_OBJECT_TYPE  );
	}

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPathTrans);
	SG_FILE_NULLCLOSE(pCtx, pFile);
	SG_STRING_NULLFREE(pCtx, pStringLink);
	SG_NULLFREE(pCtx, pBytes);
}

//////////////////////////////////////////////////////////////////

/**
 * See if the directory with the given GID is a parent/ancestor of the given ptnode.
 * Also, return a flag indicating whether there were ANY moves to a parent/ancestor.
 *
 * ptn_Start contains the current ptnode.  We search the chain of parents back to
 * the root directory to see if one of the ancestors has the requested GID.
 *
 * TODO we could probably make this faster by passing in the start-ptnode and the
 * TODO desired-ptnode and just doing pointer compares -- rather than string
 * TODO compares on the GIDs.  But that makes a few assumptions that I'm not ready
 * TODO to make.
 */
static void _pt_revert__is_ancestor_in_tree(SG_context * pCtx,
											struct _pt_revert_data * prd,
											struct sg_ptnode * ptn_Start,
											const char * pszGidObject_Desired,
											SG_bool * pbFoundAncestor,
											SG_bool * pbHadMoves)
{
	struct sg_ptnode * ptn_Parent = NULL;
	SG_bool bActive;
	SG_bool bMoved;

	SG_ASSERT(  (ptn_Start->type == SG_TREENODEENTRY_TYPE_DIRECTORY)  );

	bActive = ((ptn_Start->temp_flags & sg_PTNODE_TEMP_FLAG_REVERTING) != 0);
	bMoved  = (bActive  &&  (ptn_Start->pszgidMovedFrom));

	if (bMoved)
		SG_ERR_CHECK(  sg_pendingtree__find_by_gid(pCtx,prd->pPendingTree,ptn_Start->pszgidMovedFrom,&ptn_Parent)  );
	else
		ptn_Parent = ptn_Start->pCurrentParent;

	if (!ptn_Parent  ||  !ptn_Parent->pszgid)
	{
		*pbFoundAncestor = SG_FALSE;
		*pbHadMoves = bMoved;
	}
	else if (strcmp(ptn_Parent->pszgid,pszGidObject_Desired) == 0)
	{
		*pbFoundAncestor = SG_TRUE;
		*pbHadMoves = bMoved;
	}
	else
	{
		SG_bool bFoundAncestor;
		SG_bool bAncestorHadMoves;

		SG_ERR_CHECK(  _pt_revert__is_ancestor_in_tree(pCtx,prd,ptn_Parent,pszGidObject_Desired,&bFoundAncestor,&bAncestorHadMoves)  );

		*pbFoundAncestor = bFoundAncestor;
		*pbHadMoves = (bMoved | bAncestorHadMoves);
	}

fail:
	return;
}

//////////////////////////////////////////////////////////////////

/**
 * the entry appears in the currently-dirty working-directory, but not in the baseline.
 *
 * we only claim the "source"; that is, we only "own" this entryname in the current
 * view of the tree -- after the REVERT is complete we won't own the entryname.
 *
 * Ideally, the "revert of an add" should just forget about the entry (causing
 * it to appear as a "found" item after everything is finished).
 *
 * BUT, if another entry owns this entryname in the original baseline (and it is
 * being reverted too, then we need to forget about it *and* move it to a backup name.
 * For example, if they did:
 *     vv mv foo bar
 *     vv add foo
 * then we need to do:
 *     mv foo foo~sg00~
 *     mv bar foo
 *
 * TODO in this example, we have chosen to give the un-added file the ~sg00~ backup
 * TODO name and let the foo-->bar file have the original foo name after the un-rename.
 * TODO this might be an arbitrary choice -- perhaps we should give them both ~sg*~ names...
 */
static void _pt_revert__revert_added(SG_context * pCtx,
									 struct _pt_revert_data * prd,
									 const char * pszGidObject,
									 const SG_treediff2_ObjectData * pOD,
									 SG_diffstatus_flags dsFlags,
									 struct sg_ptnode * ptn)
{
	struct sg_ptnode * ptn_owned = NULL;
	SG_bool bActive = ((ptn->temp_flags & sg_PTNODE_TEMP_FLAG_REVERTING) != 0);

#if !defined(DEBUG)
    SG_UNUSED(dsFlags);
#endif

	switch (prd->pass)
	{
	case _PT_REVERT_PASS__SCAN_AND_BIND:
		{
			// we DO NOT filter out the non-reverting changes.  we want a complete
			// picture of entryname ownership for ALL of the structural changes.
			// but we do mark the ones that will be reverted as "active".

			// since we build a simple (non-composite) treediff, we should only see
			// plain ADDS (ADD+RENAME and/or ADD+MOVE can only occur in a composite
			// treediff).
			SG_ASSERT(  ((dsFlags & (SG_DIFFSTATUS_FLAGS__RENAMED | SG_DIFFSTATUS_FLAGS__MOVED)) == 0)  );

			SG_ERR_CHECK(  _pt_revert__bind_right_side_of_diff_to_source(pCtx,prd,pszGidObject,pOD,bActive)  );
		}
		break;

	case _PT_REVERT_PASS__CHECK_FOR_CONFLICTS:
		{
			// un-ADDing this entry will not require us to create an entry on the disk.
			// we already asserted that the parent directory cannot change (via a MOVE
			// or RENAME).  so there is nothing to do here.
		}
		break;

//	case _PT_REVERT_PASS__PRECOMPUTE_SUBMODULE_STATE:
//		{
//			if (ptn->type == SG_TREENODEENTRY_TYPE_SUBMODULE)
//			{
//				// since we can always un-add submodule (by either
//				// reverting it to a found status or (in the case of
//				// a disposable one (added by merge))), we don't need
//				// a precompute step.  just let __APPLY_ handle it the
//				// next time our state machine is called.
//			}
//		}
//		break;

	case _PT_REVERT_PASS__APPLY_STRUCTURAL_CHANGES:
		{
			// un-ADD this entry from the pendingtree.  afterwards, it MAY appear as a
			// FOUND entry or we may back-it-up using the ~sg00~ trick or if disposable
			// we may actually delete it from the WD.
			//
			// whatever we do, unless we know that the file was generated, we should *try* to
			// avoid deleting the user's file.

			SG_bool bWasDeleted = SG_FALSE;

			SG_ERR_CHECK(  _pt_revert__backup_entry_for_un_add_or_un_find(pCtx,prd,pszGidObject,pOD,ptn,bActive,&bWasDeleted)  );

			if (bActive)
			{
				SG_timestamp_cache * pTSC = NULL;		// we do not own this

				// remove this entry from pendingtree so that it will look like a FOUND item after we're done.

#if TRACE_REVERT
				SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
										   "_pt_revert__revert_added: un-adding [%s] [was-deleted %d]\n",
										   pszGidObject, bWasDeleted)  );
#endif

				if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
					SG_ERR_CHECK(  _pt_log__unadd_item(pCtx,prd->pPendingTree,
													   ptn->pszgid,
													   "UNADD")  );
				
				// since the entry will become a "found" rather than "added" node, we
				// get rid of its timestamp cache entry (knowing that scan-dir doesn't
				// create entries for "found" items).
				//
				// TODO 2011/03/08 Don't we only need to update the timestamp cache if ptn->type is _REGULAR_FILE ?
				//
				// TODO 2011/04/05 Shouldn't we guard this stuff with "if __DO_IT..." ?

				SG_ERR_CHECK(  SG_pendingtree__get_timestamp_cache(pCtx, ptn->pPendingTree, &pTSC)  );
				SG_ERR_CHECK(  SG_timestamp_cache__remove(pCtx, pTSC, ptn->pszgid)  );

				if (bWasDeleted)
				{
					// TODO 2011/05/09 See W0168 (do we need to delete any ISSUES from MERGE/RESOLVE)?
					SG_ERR_CHECK(  SG_rbtree__remove__with_assoc(pCtx,ptn->pCurrentParent->prbItems,pszGidObject,(void **)&ptn_owned)  );
					SG_ASSERT(  (ptn_owned == ptn)  );
					SG_PTNODE_NULLFREE(pCtx,ptn_owned);
				}
				else
				{
					// XREF[A]:
					// in the original REVERT code, an un-ADD removes the PTNODE.  This causes
					// sg_pendingtree__find_by_gid() to return NULL and causes us problems.
					// this is best seen when you delete (implicitly or explicitly) multiple
					// levels of directories and want to find something nested inside it.
					// so we just change the __ADDED to a __FOUND by changing the temp_flags.
					// see also XREF[B] and XREF[C].
					ptn->temp_flags |= sg_PTNODE_TEMP_FLAG_IMPLICIT_ADD;
				}
			}
		}
		break;

	default:
		break;
	}

fail:
	SG_PTNODE_NULLFREE(pCtx,ptn_owned);
}

/**
 * the entry does not appear in the currently-dirty working-directory, but it does appear
 * in the baseline.
 *
 * the REVERT should re-create this entry from data in the repo.
 *
 * we only claim the original entryname (in the original parent directory) in the "target".
 * we do not claim anything in the "source".
 *
 * TODO if a directory is deleted that contains a bunch of stuff,  the contents within the
 * TODO directory are marked "IMPLICIT" and deleted too.  So, each of them should appear
 * TODO in the pendingtree too, right?  So, we shouldn't have to dive on a directory because
 * TODO each of the things within will get processed in turn.  (Now we may have to do some
 * TODO magic to get the parent to get processed before the children....)
 */
static void _pt_revert__revert_deleted(SG_context * pCtx,
									   struct _pt_revert_data * prd,
									   const char * pszGidObject,
									   const SG_treediff2_ObjectData * pOD,
									   SG_diffstatus_flags dsFlags,
									   struct sg_ptnode * ptn)
{
	SG_bool bActive = ((ptn->temp_flags & sg_PTNODE_TEMP_FLAG_REVERTING) != 0);

#if !defined(DEBUG)
    SG_UNUSED(dsFlags);
#endif

	switch (prd->pass)
	{
	case _PT_REVERT_PASS__SCAN_AND_BIND:
		{
			// we DO NOT filter out the non-reverting changes.  we want a complete
			// picture of entryname ownership for ALL of the structural changes.
			// but we do mark the ones that will be reverted as "active".

			// TODO the user should not be able to do a RENAME+DELETE, so both the "original"
			// TODO and "current" entrynames should be the same, so the following assert
			// TODO is probably not necessary.
			SG_ASSERT(  ((dsFlags & SG_DIFFSTATUS_FLAGS__RENAMED) == 0)  );

			// TODO likewise, they should not be able to do a MOVE+DELETE, so both the "original"
			// TODO and "current" parent directories should be the same.  so the following assert
			// TODO is also probably not necessary.
			SG_ASSERT(  ((dsFlags & SG_DIFFSTATUS_FLAGS__MOVED) == 0)  );

			SG_ERR_CHECK(  _pt_revert__bind_left_side_of_diff_to_target(pCtx,prd,pszGidObject,pOD,bActive)  );
		}
		break;

	case _PT_REVERT_PASS__CHECK_FOR_CONFLICTS:
		{
			SG_ASSERT(  (ptn->pCurrentParent)  );		// the user-root directory could not have been deleted

			// if we are un-deleting, this entry make sure that the parent directory is/will be present;
			// that is, if the the parent was also deleted/lost make sure that it WILL be restored too.
			// (we can't just check to see if the parent directory exists because we haven't started
			// any of the restores yet.)

			if (bActive)
			{
				// TODO we are assuming that we still don't allow a moved entry to be deleted in the same commit,
				// TODO so we don't have to worry about which parent in the following.
				SG_bool bActiveParent = ((ptn->pCurrentParent->temp_flags & sg_PTNODE_TEMP_FLAG_REVERTING) != 0);

				if (!bActiveParent)
				{
					const char * pszGid_Parent;
					const SG_treediff2_ObjectData * pOD_Parent;
					SG_bool bFound_Parent;
					SG_diffstatus_flags dsFlags_Parent;

					SG_ERR_CHECK(  SG_treediff2__ObjectData__get_parent_gid(pCtx,pOD,&pszGid_Parent)  );
					SG_ASSERT(  (strcmp(pszGid_Parent,ptn->pCurrentParent->pszgid) == 0)  );

					SG_ERR_CHECK(  SG_treediff2__get_ObjectData(pCtx,prd->pTreeDiff,pszGid_Parent,&bFound_Parent,&pOD_Parent)  );
					SG_ASSERT(  (bFound_Parent)  );		// parent dir should also be in treediff because dir contents changed
					SG_ERR_CHECK(  SG_treediff2__ObjectData__get_dsFlags(pCtx,pOD_Parent,&dsFlags_Parent)  );

					if (dsFlags_Parent & (SG_DIFFSTATUS_FLAGS__DELETED | SG_DIFFSTATUS_FLAGS__LOST))
						SG_ERR_THROW(  SG_ERR_PENDINGTREE_PARTIAL_CHANGE_COLLISION  );
				}
			}
		}
		break;

	case _PT_REVERT_PASS__PRECOMPUTE_SUBMODULE_STATE:
		{
			if (bActive)
			{
				if (ptn->type == SG_TREENODEENTRY_TYPE_SUBMODULE)
				{
					SG_ERR_CHECK(  _pt_revert__precompute_recreate_deleted_submodule(pCtx, prd, pOD, ptn)  );
				}
			}
		}
		break;

	case _PT_REVERT_PASS__APPLY_STRUCTURAL_CHANGES:
		{
			// un-delete entry.  restore it from the repo to the target directory.

			if (bActive)
			{
				if (ptn->type != SG_TREENODEENTRY_TYPE_SUBMODULE)
				{
					SG_ERR_CHECK(  _pt_revert__restore_entry_for_un_delete_or_un_lost(pCtx,prd,ptn)  );

					ptn->saved_flags &= ~sg_PTNODE_SAVED_FLAG_DELETED;
					ptn->bDeletedAfterAdded = SG_FALSE;
					ptn->temp_flags |= sg_PTNODE_TEMP_FLAG_CHECKMYNAME;
				}
				else
				{
					// submodules are deferred until step [8.6]
				}
			}
		}
		break;

	default:
		break;
	}

fail:
	return;
}

/**
 * the entry was FOUND in the currently-dirty working-directory but it was not ADDED
 * (and also not in the baseline).  it is just a random file/symlink/directory that
 * got created in the directory; we know nothing about it.
 *
 * REVERT has nothing to do here.  Just like when we REVERT and ADDED file, we leave
 * it in the directory so that it will appear as a FOUND entry.  However, we do do
 * the swap/backup trick (like for ADDED files), so that another entry that needs
 * this entryname can have it.
 *
 * TODO is it appropriate to move the found item to a ~sg*~ name when another
 * TODO entry (that we are tracking) needs the entryname?  I mean, we really don't
 * TODO know anything about this entry -- we just found it in the directory.
 */
static void _pt_revert__revert_found(SG_context * pCtx,
									 struct _pt_revert_data * prd,
									 const char * pszGidObject,
									 const SG_treediff2_ObjectData * pOD,
									 SG_diffstatus_flags dsFlags,
									 struct sg_ptnode * ptn)
{
	SG_bool bActive = ((ptn->temp_flags & sg_PTNODE_TEMP_FLAG_REVERTING) != 0);

#if !defined(DEBUG)
    SG_UNUSED(dsFlags);
#endif

	switch (prd->pass)
	{
	case _PT_REVERT_PASS__SCAN_AND_BIND:
		{
			// we DO NOT filter out the non-reverting changes.  we want a complete
			// picture of entryname ownership for ALL of the structural changes.
			// but we do mark the ones that will be reverted as "active".

			// we found an entry in a particular directory.  that's where it is and that's
			// all we know about it.  so it can't have been moved/renamed.
			SG_ASSERT(  ((dsFlags & (SG_DIFFSTATUS_FLAGS__RENAMED | SG_DIFFSTATUS_FLAGS__MOVED)) == 0)  );

			SG_ERR_CHECK(  _pt_revert__bind_right_side_of_diff_to_source(pCtx,prd,pszGidObject,pOD,bActive)  );
		}
		break;

	case _PT_REVERT_PASS__CHECK_FOR_CONFLICTS:
		{
			// there is nothing for us to do for a FOUND entry.
		}
		break;

//	case _PT_REVERT_PASS__PRECOMPUTE_SUBMODULE_STATE:
//		{
//			if (ptn->type == SG_TREENODEENTRY_TYPE_SUBMODULE)
//			{
//				// since we can always un-add submodule (by either
//				// reverting it to a found status or (in the case of
//				// a disposable one (added by merge))), we don't need
//				// a queued step.  just let __APPLY_ handle it the
//				// next time our state machine is called.
//			}
//		}
//		break;

	case _PT_REVERT_PASS__APPLY_STRUCTURAL_CHANGES:
		{
			// we don't really "un-find" it.  that's just plain silly.  we leave it
			// in the directory so that it will still have FOUND status after we're
			// finished.  just like we do for un-ADDED entries.

			SG_bool bWasDeleted = SG_FALSE;

			SG_ERR_CHECK(  _pt_revert__backup_entry_for_un_add_or_un_find(pCtx,prd,pszGidObject,pOD,ptn,bActive,&bWasDeleted)  );
			SG_ASSERT(  (bWasDeleted == SG_FALSE)  );

			// bActive doesn't really matter here because the entry wasn't registered.

#if TRACE_REVERT
			SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
									   "_pt_revert__revert_found: leaving [%s]\n",
									   pszGidObject)  );
#endif
		}
		break;

	default:
		break;
	}

fail:
	return;
}

/**
 * the entry cannot be found in the currently-dirty working-directory, but it did appear
 * in the baseline.  (they probably moved/deleted it without telling us about it.
 *
 * the REVERT should re-create it as it was.
 *
 * note that they could have done a "vv move" or "vv rename" on it *AND THEN*
 * done a "/bin/rm" on it.  in this cse, we'd have LOST+RENAME or LOST+MOVE (or
 * LOST+RENAME+MOVE).  but none of that matters because we want the orginal name/location
 * as it appeared in the baseline.
 */
static void _pt_revert__revert_lost(SG_context * pCtx,
									struct _pt_revert_data * prd,
									const char * pszGidObject,
									const SG_treediff2_ObjectData * pOD,
									SG_diffstatus_flags dsFlags,
									struct sg_ptnode * ptn)
{
	struct sg_ptnode * ptn_owned = NULL;
	SG_bool bActive = ((ptn->temp_flags & sg_PTNODE_TEMP_FLAG_REVERTING) != 0);

	switch (prd->pass)
	{
	case _PT_REVERT_PASS__SCAN_AND_BIND:
		{
			// we DO NOT filter out the non-reverting changes.  we want a complete
			// picture of entryname ownership for ALL of the structural changes.
			// but we do mark the ones that will be reverted as "active".

			SG_ERR_CHECK(  _pt_revert__bind_left_side_of_diff_to_target(pCtx,prd,pszGidObject,pOD,bActive)  );
		}
		break;

	case _PT_REVERT_PASS__CHECK_FOR_CONFLICTS:
		{
			// we are reverting a LOST entry so we need to restore it.  make sure that the parent directory
			// is present; that is, if the the parent was also deleted/lost make sure that it WILL be restored too.
			// (we can't just check to see if the parent directory exists because we haven't started
			// any of the restores yet.)  we have to be careful here because we could have LOST+MOVED.

			if (bActive)
			{
				const char * pszGid_Parent = NULL;
				struct sg_ptnode * ptn_Parent = NULL;
				SG_bool bActive_Parent;

				if (dsFlags & SG_DIFFSTATUS_FLAGS__MOVED)			// LOST+MOVED
					SG_ERR_CHECK(  SG_treediff2__ObjectData__get_moved_from_gid(pCtx,pOD,&pszGid_Parent)  );
				else
					SG_ERR_CHECK(  SG_treediff2__ObjectData__get_parent_gid(pCtx,pOD,&pszGid_Parent)  );
				SG_ERR_CHECK(  sg_pendingtree__find_by_gid(pCtx,prd->pPendingTree,pszGid_Parent,&ptn_Parent)  );

				bActive_Parent = ((ptn_Parent->temp_flags & sg_PTNODE_TEMP_FLAG_REVERTING) != 0);
				if (bActive_Parent)
				{
					// if the original parent is active, it will be restored/reverted too.  so we can assume that
					// it will (eventually) be as it was so that this entry can be placed in it.
				}
				else
				{
					// the original parent is not active, it will not be restored/reverted.  make sure that that
					// directory is currently present in the tree.

					const SG_treediff2_ObjectData * pOD_Parent;
					SG_bool bFound_Parent;
					SG_diffstatus_flags dsFlags_Parent;

					SG_ERR_CHECK(  SG_treediff2__get_ObjectData(pCtx,prd->pTreeDiff,pszGid_Parent,&bFound_Parent,&pOD_Parent)  );
					SG_ASSERT(  (bFound_Parent)  );		// parent dir should also be in treediff because dir contents changed
					SG_ERR_CHECK(  SG_treediff2__ObjectData__get_dsFlags(pCtx,pOD_Parent,&dsFlags_Parent)  );

					if (dsFlags_Parent & (SG_DIFFSTATUS_FLAGS__DELETED | SG_DIFFSTATUS_FLAGS__LOST))
						SG_ERR_THROW(  SG_ERR_PENDINGTREE_PARTIAL_CHANGE_COLLISION  );
				}
			}
		}
		break;

	case _PT_REVERT_PASS__PRECOMPUTE_SUBMODULE_STATE:
		{
			if (bActive)
			{
				if (ptn->type == SG_TREENODEENTRY_TYPE_SUBMODULE)
				{
					SG_ERR_CHECK(  _pt_revert__precompute_recreate_lost_submodule(pCtx, prd, pOD, ptn)  );
				}
			}
		}
		break;

	case _PT_REVERT_PASS__APPLY_STRUCTURAL_CHANGES:
		{
			// un-lose the entry.  restore it from the repo to the target directory.

			if (bActive)
			{
				// we defer restoring lost submodules until [8.6]
				// TODO 2011/04/05 Should we also defer the moved-from ptnode mgmt?

				if (ptn->type != SG_TREENODEENTRY_TYPE_SUBMODULE)
					SG_ERR_CHECK(  _pt_revert__restore_entry_for_un_delete_or_un_lost(pCtx,prd,ptn)  );

				if (ptn->pszgidMovedFrom)
				{
					// LOST+MOVED.  so in addition to clearing various flags, we need to move this ptn
					// back to the TARGET parent and delete the phantom/moved-away ptnode.  (just like
					// in __revert_moved().

					struct sg_ptnode * ptn_MovedFromParent = NULL;
					struct sg_ptnode * ptnMe = NULL;
					struct sg_ptnode * ptnPhantomMe = NULL;

					// remove ptnode from our current parent

					SG_ERR_CHECK(  SG_rbtree__remove__with_assoc(pCtx,ptn->pCurrentParent->prbItems,pszGidObject,(void **)&ptnMe)  );
					SG_ASSERT(  (ptnMe == ptn)  );

					// remove the old phantom ptnode from our old parent

					SG_ERR_CHECK(  sg_pendingtree__find_by_gid(pCtx,prd->pPendingTree,ptn->pszgidMovedFrom,&ptn_MovedFromParent)  );
					SG_ERR_CHECK(  SG_rbtree__remove__with_assoc(pCtx,ptn_MovedFromParent->prbItems,pszGidObject,(void **)&ptnPhantomMe)  );
					SG_PTNODE_NULLFREE(pCtx,ptnPhantomMe);

					// add this ptnode back to the old parent

					ptn->pszgidMovedFrom = NULL;		// the un-MOVE
					SG_ERR_CHECK(  sg_ptnode__add_entry(pCtx,ptn_MovedFromParent,ptn)  );
				}

				ptn->pszCurrentName = NULL;				// the un-RENAME (if necessary)

				if (ptn->type == SG_TREENODEENTRY_TYPE_SUBMODULE)
				{
					// we defer submodules until [8.6].
				}
				else
				{
					// update ptnode now for non-submodule.

					const char * pszEntryName_old = NULL;

					SG_ERR_CHECK(  sg_ptnode__get_old_name(pCtx, ptn, &pszEntryName_old)  );
					if (pszEntryName_old && *pszEntryName_old)
					{
						// a real revert.  so update flags.

						ptn->saved_flags &= ~sg_PTNODE_SAVED_FLAG_DELETED;
						ptn->temp_flags &= ~sg_PTNODE_TEMP_FLAG_IMPLICIT_DELETE;
						ptn->temp_flags |= sg_PTNODE_TEMP_FLAG_CHECKMYNAME;
					}
					else
					{
						// This entry DID NOT have a name in the baseline.
						// This could happen when you do a:
						//     date > foo
						//     vv add foo
						//     /bin/rm foo
						//     vv revert foo
						// That is an ADDED-but-deleted gets labeled as a LOST
						// item so the REVERT just needs to remove the ptnode
						// and not mess with the WD.  This code should look like
						// the stuff in __revert_added().

						SG_timestamp_cache * pTSC = NULL;		// we do not own this

#if TRACE_REVERT
						SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
												   "_pt_revert__restore_lost: [gid %s] no name.\n",
												   ptn->pszgid)  );
#endif
						SG_ASSERT(  (dsFlags & SG_DIFFSTATUS_FLAGS__HACK__LOST_ADDED)  );

						if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
							SG_ERR_CHECK(  _pt_log__unadd_item(pCtx,prd->pPendingTree,
															   ptn->pszgid,
															   "UNADD")  );

						// TODO 2011/04/07 Shouldn't we guard this stuff with "if __DO_IT..." ?

						SG_ERR_CHECK(  SG_pendingtree__get_timestamp_cache(pCtx, ptn->pPendingTree, &pTSC)  );
						SG_ERR_CHECK(  SG_timestamp_cache__remove(pCtx, pTSC, ptn->pszgid)  );
						SG_ERR_CHECK(  SG_rbtree__remove__with_assoc(pCtx,ptn->pCurrentParent->prbItems,ptn->pszgid,(void **)&ptn_owned)  );
						SG_ASSERT(  (ptn_owned == ptn)  );
						SG_PTNODE_NULLFREE(pCtx,ptn_owned);
					}
				}
			}
		}
		break;

	default:
		break;
	}

fail:
	SG_PTNODE_NULLFREE(pCtx,ptn_owned);
}

/**
 * The item was "ADDED by MERGE" and then "DELETED by USER".
 * So what should REVERT do?
 *
 * I'm going to assume that REVERT should undo both steps.
 * Ideally, we should have a command line flag to let them
 * pick:
 * [a] revert the delete to get back to what merge left in the directory.
 * [b] do [a] and then revert the add to get back to a clean WD.
 *
 */ 
static void _pt_revert__revert_cancel_added(SG_context * pCtx,
											struct _pt_revert_data * prd,
											const char * pszGidObject,
											const SG_treediff2_ObjectData * pOD,
											SG_diffstatus_flags dsFlags,
											struct sg_ptnode * ptn)
{
	struct sg_ptnode * ptn_owned = NULL;
	SG_bool bActive = ((ptn->temp_flags & sg_PTNODE_TEMP_FLAG_REVERTING) != 0);

	SG_UNUSED( pOD );
	SG_UNUSED( dsFlags );

	switch (prd->pass)
	{
	case _PT_REVERT_PASS__SCAN_AND_BIND:
		// item is not present in either the baseline nor the WD,
		// so we don't claim the entryname in either source or target.
		break;

	case _PT_REVERT_PASS__CHECK_FOR_CONFLICTS:
		// nothing to do here
		break;

	case _PT_REVERT_PASS__PRECOMPUTE_SUBMODULE_STATE:
		// nothing to do here
		break;

	case _PT_REVERT_PASS__APPLY_STRUCTURAL_CHANGES:
		if (bActive)
		{
#if TRACE_REVERT
			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
									   "RevertCancelAdded: [gid %s] %s\n",
									   pszGidObject,
									   ((ptn->pszCurrentName) ? ptn->pszCurrentName : "(null)"))  );
#endif

			// TODO 2011/05/09 See W0168 (do we need to delete any ISSUES from MERGE/RESOLVE)?
			SG_ERR_CHECK(  SG_rbtree__remove__with_assoc(pCtx,
														 ptn->pCurrentParent->prbItems,
														 pszGidObject,
														 (void **)&ptn_owned)  );
			SG_ASSERT(  (ptn_owned == ptn)  );
			SG_PTNODE_NULLFREE(pCtx,ptn_owned);
		}
		
		break;

	default:
		break;
	}

fail:
	SG_PTNODE_NULLFREE(pCtx,ptn_owned);
}

/**
 * the entry was RENAMED (but not MOVED) in the currently-dirty working-direcory
 * relative to the original baseline.   since there was no MOVE, both entrynames
 * are in the same directory.
 *
 * the REVERT should undo the rename.
 */
static void _pt_revert__revert_renamed(SG_context * pCtx,
									   struct _pt_revert_data * prd,
									   const char * pszGidObject,
									   const SG_treediff2_ObjectData * pOD,
									   SG_diffstatus_flags dsFlags,
									   struct sg_ptnode * ptn)
{
	const char * pszEntryName_Original;
	SG_pathname * pPathTransSource = NULL;
	SG_pathname * pPathTransTarget = NULL;
#ifdef DEBUG
	SG_bool bFound;
#endif
	SG_bool bActive = ((ptn->temp_flags & sg_PTNODE_TEMP_FLAG_REVERTING) != 0);

#if !defined(DEBUG)
    SG_UNUSED(dsFlags);
#endif

	SG_ASSERT(  ((dsFlags & SG_DIFFSTATUS_FLAGS__ADDED) == 0)  );
	SG_ASSERT(  ((dsFlags & SG_DIFFSTATUS_FLAGS__DELETED) == 0)  );
	SG_ASSERT(  ((dsFlags & SG_DIFFSTATUS_FLAGS__LOST) == 0)  );
	SG_ASSERT(  ((dsFlags & SG_DIFFSTATUS_FLAGS__FOUND) == 0)  );
	SG_ASSERT(  ((dsFlags & SG_DIFFSTATUS_FLAGS__MOVED) == 0)  );	// we do RENAME-only, see __revert_moved()

	switch (prd->pass)
	{
	case _PT_REVERT_PASS__SCAN_AND_BIND:
		{
			// we DO NOT filter out the non-reverting changes.  we want a complete
			// picture of entryname ownership for ALL of the structural changes.
			// but we do mark the ones that will be reverted as "active".

			// claim the CURRENT entryname as the "source".
			// claim the ORIGINAL entryname as the "target".

			SG_ERR_CHECK(  _pt_revert__bind_right_side_of_diff_to_source(pCtx,prd,pszGidObject,pOD,bActive)  );
			SG_ERR_CHECK(  _pt_revert__bind_left_side_of_diff_to_target(pCtx,prd,pszGidObject,pOD,bActive)  );
		}
		break;

	case _PT_REVERT_PASS__CHECK_FOR_CONFLICTS:
		{
			// to un-RENAME an entry, we just need to change the name within the current
			// directory.  that directory must exist on disk (either in the original repo-path
			// location or in a swap path) and it should exist in both the before and after
			// views of tree -- so the parent dir cannot have been ADDED/FOUND/DELETED/LOST.  and so
			// we should be able to do the job without any conflicts because of the parent.
			//
			// note: _inv_entry__check_for_swap_cb() already checked that we have clear title
			// to our original (aka TARGET) entryname.  so there are no name collisions.
		}
		break;

//	case _PT_REVERT_PASS__PRECOMPUTE_SUBMODULE_STATE:
//		{
//			if (ptn->type == SG_TREENODEENTRY_TYPE_SUBMODULE)
//			{
//			}
//		}
//		break;

	case _PT_REVERT_PASS__APPLY_STRUCTURAL_CHANGES:
		{
			if (bActive)
			{
				// because we are after the swap parking -and- the restore has not finished yet,
				// we can't use the normal repo-path-->absolute-path code because an ancestor
				// directory might still be parked in swap.

				// get the current disk path of the current SOURCE entry.  this may be in the
				// current SOURCE parent directory -or- it may have been moved to swap.

				SG_ERR_CHECK(  _pt_revert__get_transient_path(pCtx,prd,ptn,&pPathTransSource)  );

				// get the current disk path of the SOURCE (aka TARGET) parent directory and
				// append the TARGET entryname.  the resulting pathname should not exist on disk.

				SG_ERR_CHECK(  _pt_revert__get_transient_path(pCtx,prd,ptn->pCurrentParent,&pPathTransTarget)  );
#if defined(DEBUG)
				if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
				{
					SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx,pPathTransTarget,&bFound,NULL,NULL)  );
					SG_ASSERT(  (bFound)  );
				}
#endif
				SG_ERR_CHECK(  SG_treediff2__ObjectData__get_old_name(pCtx,pOD,&pszEntryName_Original)  );
				SG_ERR_CHECK(  SG_pathname__append__from_sz(pCtx,pPathTransTarget,pszEntryName_Original)  );
#if defined(DEBUG)
				if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
				{
					SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx,pPathTransTarget,&bFound,NULL,NULL)  );
//					SG_ASSERT(  (!bFound)  );
					if (bFound)
						SG_ERR_THROW2(  SG_ERR_NOTIMPLEMENTED,
										(pCtx,
										 "_pt_revert__revert_rename: deal with collisions caused by case-insensitive filesystems [pPathTransTarget %s]",
										 SG_pathname__sz(pPathTransTarget))  );
				}
#endif

#if TRACE_REVERT
				SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
										   ("_pt_revert__revert_rename: restoring entry [gid %s]\n"
											"                     renamed path (source) %s\n"
											"                    original path (target) %s\n"),
										   pszGidObject,
										   SG_pathname__sz(pPathTransSource),
										   SG_pathname__sz(pPathTransTarget))  );
#endif
				if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
					SG_ERR_CHECK(  _pt_log__move(pCtx,prd->pPendingTree,
												 ptn->pszgid,
												 SG_pathname__sz(pPathTransSource),
												 SG_pathname__sz(pPathTransTarget),
												 "UNDO_RENAME")  );
				if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
					SG_ERR_CHECK(  SG_fsobj__move__pathname_pathname(pCtx,pPathTransSource,pPathTransTarget)  );

				SG_ERR_CHECK(  sg_ptnode__alter_name(pCtx, prd->pPendingTree, ptn, NULL)  );
			}
		}
		break;

	default:
		break;
	}

fail:
	SG_PATHNAME_NULLFREE(pCtx,pPathTransSource);
	SG_PATHNAME_NULLFREE(pCtx,pPathTransTarget);
}

//////////////////////////////////////////////////////////////////

static void _pt_revert__revert_moved__apply(SG_context * pCtx,
											struct _pt_revert_data * prd,
											struct sg_ptnode * ptn)
{
	const char * pszEntryName_Original;
	struct sg_ptnode * ptn_MovedFromParent;
	struct sg_ptnode * ptnMe;
	struct sg_ptnode * ptnPhantomMe = NULL;
	SG_pathname * pPathTransSource = NULL;
	SG_pathname * pPathTransTarget = NULL;
#ifdef DEBUG
	SG_bool bFound;
#endif

	// because we are after the swap parking -and- the restore has not finished yet,
	// we can't use the normal repo-path-->absolute-path code because an ancestor
	// directory might still be parked in swap.

	// get the current disk path of the current SOURCE entry.  this may be in the
	// current SOURCE parent directory -or- it may have been moved to swap.

	SG_ERR_CHECK(  _pt_revert__get_transient_path(pCtx,prd,ptn,&pPathTransSource)  );

	// get the ptnode of the moved-from parent.  we cannot use __find_by_gid() right
	// now (because the tree may be temporarily broken (disconnected, cyclic, or both)).

	SG_ASSERT(  (ptn->ptnPhantomMe)  );
	ptn_MovedFromParent = ptn->ptnPhantomMe->pCurrentParent;
	SG_ASSERT(  (ptn_MovedFromParent)  );

	// get the current disk path of the TARGET ( != SOURCE) parent directory and
	// append the TARGET entryname (this covers MOVES and MOVES+RENAME).

	SG_ERR_CHECK(  _pt_revert__get_transient_path(pCtx,prd,ptn_MovedFromParent,&pPathTransTarget)  );
#if defined(DEBUG)
	if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
	{
		SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx,pPathTransTarget,&bFound,NULL,NULL)  );
		SG_ASSERT(  (bFound)  );
	}
#endif

	SG_ERR_CHECK(  sg_ptnode__get_old_name(pCtx,ptn,&pszEntryName_Original)  );
	SG_ERR_CHECK(  SG_pathname__append__from_sz(pCtx,pPathTransTarget,pszEntryName_Original)  );
#if defined(DEBUG)
	if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
	{
		SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx,pPathTransTarget,&bFound,NULL,NULL)  );
//		SG_ASSERT(  (!bFound)  );
		if (bFound)
			SG_ERR_THROW2(  SG_ERR_NOTIMPLEMENTED,
							(pCtx,
							 "_pt_revert__revert_move: deal with collisions caused by case-insensitive filesystems [pPathTransTarget %s]",
							 SG_pathname__sz(pPathTransTarget))  );
	}
#endif

#if TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
							   ("_pt_revert__revert_move: restoring entry [gid %s]\n"
								"                     moved path (source) %s\n"
								"                  original path (target) %s\n"),
							   ptn->pszgid,
							   SG_pathname__sz(pPathTransSource),
							   SG_pathname__sz(pPathTransTarget))  );
#endif
	if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
		SG_ERR_CHECK(  _pt_log__move(pCtx,prd->pPendingTree,
									 ptn->pszgid,
									 SG_pathname__sz(pPathTransSource),
									 SG_pathname__sz(pPathTransTarget),
									 "UNDO_MOVE")  );
	if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
		SG_ERR_CHECK(  SG_fsobj__move__pathname_pathname(pCtx,pPathTransSource,pPathTransTarget)  );

	// remove ptnode from our current parent

	SG_ERR_CHECK(  SG_rbtree__remove__with_assoc(pCtx,ptn->pCurrentParent->prbItems,ptn->pszgid,(void **)&ptnMe)  );
	SG_ASSERT(  (ptnMe == ptn)  );
#if TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
							   "_pt_revert__revert_move: removing child [parent %s][child %s][child saved %d][child temp %d]\n",
							   ptn->pCurrentParent->pszgid,
							   ptn->pszgid,ptn->saved_flags,ptn->temp_flags)  );
#endif

	// remove the old phantom ptnode from our old parent

	SG_ERR_CHECK(  SG_rbtree__remove__with_assoc(pCtx,ptn_MovedFromParent->prbItems,ptn->pszgid,(void **)&ptnPhantomMe)  );
	SG_ASSERT(  (ptnPhantomMe == ptn->ptnPhantomMe)  );
#if TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
							   "_pt_revert__revert_move: removing child [parent %s][child %s][child saved %d][child temp %d]\n",
							   ptn_MovedFromParent->pszgid,
							   ptnPhantomMe->pszgid,ptnPhantomMe->saved_flags,ptnPhantomMe->temp_flags)  );
#endif
	SG_ASSERT(  (ptnPhantomMe->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY)  );
	SG_PTNODE_NULLFREE(pCtx,ptnPhantomMe);
	ptn->ptnPhantomMe = NULL;

	// add this ptnode back to the old parent

	ptn->pszgidMovedFrom = NULL;		// the un-MOVE
	SG_ERR_CHECK(  sg_ptnode__add_entry(pCtx,ptn_MovedFromParent,ptn)  );

	SG_ERR_CHECK(  sg_ptnode__alter_name(pCtx, prd->pPendingTree, ptn, NULL)  );

fail:
	SG_PATHNAME_NULLFREE(pCtx,pPathTransSource);
	SG_PATHNAME_NULLFREE(pCtx,pPathTransTarget);
}

//////////////////////////////////////////////////////////////////

/**
 * the entry was MOVED (and possibly RENAMED) in the currently-dirty working-direcory
 * relative to its name/location in the original baseline.
 *
 * handle the REVERT of MOVE or MOVE+RENAME actions.
 */
static void _pt_revert__revert_moved(SG_context * pCtx,
									 struct _pt_revert_data * prd,
									 const char * pszGidObject,
									 const SG_treediff2_ObjectData * pOD,
									 SG_diffstatus_flags dsFlags,
									 struct sg_ptnode * ptn)
{
	SG_bool bActive = ((ptn->temp_flags & sg_PTNODE_TEMP_FLAG_REVERTING) != 0);

#if !defined(DEBUG)
    SG_UNUSED(dsFlags);
#endif

	SG_ASSERT(  ((dsFlags & SG_DIFFSTATUS_FLAGS__ADDED) == 0)  );
	SG_ASSERT(  ((dsFlags & SG_DIFFSTATUS_FLAGS__DELETED) == 0)  );
	SG_ASSERT(  ((dsFlags & SG_DIFFSTATUS_FLAGS__LOST) == 0)  );
	SG_ASSERT(  ((dsFlags & SG_DIFFSTATUS_FLAGS__FOUND) == 0)  );

	switch (prd->pass)
	{
	case _PT_REVERT_PASS__SCAN_AND_BIND:
		{
			// we DO NOT filter out the non-reverting changes.  we want a complete
			// picture of entryname ownership for ALL of the structural changes.
			// but we do mark the ones that will be reverted as "active".

			// claim the CURRENT entryname in the CURRENT directory as the "source".
			// claim the ORIGINAL entryname in the ORIGINAL directory as the "target".

			SG_ERR_CHECK(  _pt_revert__bind_right_side_of_diff_to_source(pCtx,prd,pszGidObject,pOD,bActive)  );
			SG_ERR_CHECK(  _pt_revert__bind_left_side_of_diff_to_target(pCtx,prd,pszGidObject,pOD,bActive)  );
			return;
		}

	case _PT_REVERT_PASS__CHECK_FOR_CONFLICTS:
		{
			// to undo a MOVE we need to ensure that the original (aka TARGET) parent
			// directory WILL exist after everything is finished.

			if (bActive)
			{
				struct sg_ptnode * ptn_MovedFromParent = NULL;
				SG_bool bActive_MovedFromParent;
				SG_bool bFoundPhantomMe;

				// get the ptnode of the moved-from parent.
				SG_ERR_CHECK(  sg_pendingtree__find_by_gid(pCtx,prd->pPendingTree,ptn->pszgidMovedFrom,&ptn_MovedFromParent)  );

				bActive_MovedFromParent = ((ptn_MovedFromParent->temp_flags & sg_PTNODE_TEMP_FLAG_REVERTING) != 0);
				if (bActive_MovedFromParent)
				{
					// if the original parent is active, it will be restored/reverted too.  so we can assume that
					// it will (eventually) be as it was so that this entry can be placed in it.
				}
				else
				{
					// the original parent is not active, it will not be restored/reverted.  make sure that that
					// directory is currently present in the tree.

					const SG_treediff2_ObjectData * pOD_MovedFromParent;
					SG_bool bFound_MovedFromParent;
					SG_diffstatus_flags dsFlags_MovedFromParent;

					SG_ERR_CHECK(  SG_treediff2__get_ObjectData(pCtx,prd->pTreeDiff,ptn->pszgidMovedFrom,&bFound_MovedFromParent,&pOD_MovedFromParent)  );
					SG_ASSERT(  (bFound_MovedFromParent)  );		// parent dir should also be in treediff because dir contents changed
					SG_ERR_CHECK(  SG_treediff2__ObjectData__get_dsFlags(pCtx,pOD_MovedFromParent,&dsFlags_MovedFromParent)  );

					if (dsFlags_MovedFromParent & (SG_DIFFSTATUS_FLAGS__DELETED | SG_DIFFSTATUS_FLAGS__LOST))
						SG_ERR_THROW(  SG_ERR_PENDINGTREE_PARTIAL_CHANGE_COLLISION  );
				}

				// find the phantom/moved-away ptnode for this ptnode in the moved-from parent.
				// temporarily store this PTNODE POINTER in our ptnode so that we will have it
				// when get ready to apply the changes.  this is necessary because in some screwcases
				// (when we need to use the PARKED_FOR_CYCLE code), as we apply the changes, the ptnode
				// tree can become temporarily disconnected or cyclic.  besides being distrubing, it
				// causes __find_by_gid() to fail.

				SG_ERR_CHECK(  SG_rbtree__find(pCtx,ptn_MovedFromParent->prbItems,pszGidObject,&bFoundPhantomMe,(void *)&ptn->ptnPhantomMe)  );
				SG_ASSERT(  (bFoundPhantomMe)  );
				SG_ASSERT(  (ptn->ptnPhantomMe->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY)  );

				if (ptn->type == SG_TREENODEENTRY_TYPE_DIRECTORY)
				{
					// for moved directories, if the directory has a moved parent/ancestor then it is possible
					// for us to get pathname cycles.  in theory, if everything is being reverted, then both the
					// SOURCE and TARGET views of the tree should be consistent (and cycle-free).  but for a
					// partial revert, one of the ancestor moves could be omitted and we might stumble into the
					// case where we want to try to un-move this directory and make it a parent/ancestor of itself.

					SG_bool bIsAncestor, bAncestorHadMoves;

					SG_ERR_CHECK(  _pt_revert__is_ancestor_in_tree(pCtx,prd,ptn,pszGidObject,&bIsAncestor,&bAncestorHadMoves)  );
					if (bIsAncestor)
						SG_ERR_THROW(  SG_ERR_PENDINGTREE_PARTIAL_CHANGE_COLLISION  );

					// even when there are no cycles (in the SOURCE view -or- in the TARGET view), we still may
					// have problems un-winding all of the changes -- because of a transient cycle.  for example,
					// if we have:
					//      @/dir_1/dir_2/dir_3/dir_4/
					//      @/dir_1_dir_5/
					// and do:
					//      mv dir_1/dir_2/dir_3 dir_1/dir_5/
					//      mv dir_1/dir_2       dir_1/dir_5/dir_3/dir_4/
					// giving:
					//      @/dir_1/dir_5/dir_3/dir_4/dir_2/
					// (so 2 and 3 are tangled)
					//
					// to undo these moves, we either need to know the ordering (have to undo dir_2 *before* dir_3
					// or we get an error in the filesystem) --or-- we have to park one or both of them so that we
					// can undo them without worrying about the order.
					//
					// i'm going to do the latter and force park the entry.  this will produce some false positives
					// because we're only going to ask if ANY ancestor was moved.  it may be the case that the
					// ancestor's move may be harmless and not tangled with our move (but then again, it could be some
					// convoluted permutation that cause a cycle of cycles)).  but that doesn't matter, we REGISTER
					// to park this directory knowing that we can unpark them all later.

					if (bAncestorHadMoves)
					{
						SG_inv_entry * pInvEntry = NULL;
						SG_bool bFound;

						SG_ERR_CHECK(  SG_inv_dirs__fetch_source_binding(pCtx,prd->pInvDirs,pszGidObject,&bFound,&pInvEntry,NULL)  );
						SG_ASSERT(  (bFound)  );
						SG_ERR_CHECK_RETURN(  SG_inv_entry__create_parked_pathname(pCtx,pInvEntry,SG_INV_ENTRY_FLAGS__PARKED_FOR_CYCLE)  );

						// add this ptnode to the list of potential cycles so that we can defer it until [7b].

						if (!prd->pVecCycles)
							SG_ERR_CHECK(  SG_VECTOR__ALLOC(pCtx,&prd->pVecCycles,10)  );
						SG_ERR_CHECK(  SG_vector__append(pCtx,prd->pVecCycles,ptn,NULL)  );
						ptn->temp_flags |= sg_PTNODE_TEMP_FLAG_PARKED_FOR_CYCLE;
					}
				}
			}
		}
		break;

//	case _PT_REVERT_PASS__PRECOMPUTE_SUBMODULE_STATE:
//		{
//			if (ptn->type == SG_TREENODEENTRY_TYPE_SUBMODULE)
//			{
//			}
//		}
//		break;

	case _PT_REVERT_PASS__APPLY_STRUCTURAL_CHANGES:
		{
			if (bActive)
			{
				// defer the parked-for-cycle ones to step [7b].
				if ((ptn->temp_flags & sg_PTNODE_TEMP_FLAG_PARKED_FOR_CYCLE) == 0)
					SG_ERR_CHECK(  _pt_revert__revert_moved__apply(pCtx,prd,ptn)  );
			}
		}
		break;

	default:
		break;
	}

fail:
	return;
}

//////////////////////////////////////////////////////////////////

static SG_treediff2_foreach_callback _pt_revert__process_structural_changes;

static void _pt_revert__process_structural_changes(SG_context * pCtx,
												   SG_UNUSED_PARAM( SG_treediff2 * pTreeDiff ),
												   const char * pszGidObject,
												   const SG_treediff2_ObjectData * pOD,
												   void * pVoidData)
{
	struct _pt_revert_data * prd = (struct _pt_revert_data *)pVoidData;
	struct sg_ptnode * ptn = NULL;
	SG_diffstatus_flags dsFlags;

	SG_UNUSED( pTreeDiff );

#if 0
	SG_ERR_CHECK(  SG_treediff2_debug__dump_od_to_console(pCtx, pOD)  );
#endif

	// TODO XREF[C] see XREF[A]
	SG_ERR_CHECK(  sg_pendingtree__find_by_gid(pCtx,prd->pPendingTree,pszGidObject,&ptn)  );
	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_dsFlags(pCtx,pOD,&dsFlags)  );

#if TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
							   "ProcessStructuralChanges: [pass %d][gid %s][dsFlags 0x%x][ptn s flags %d][ptn t flags %d]\n",
							   prd->pass,pszGidObject,dsFlags,ptn->saved_flags,ptn->temp_flags)  );
#endif

	switch (dsFlags & SG_DIFFSTATUS_FLAGS__MASK_MUTUALLY_EXCLUSIVE)
	{
	case SG_DIFFSTATUS_FLAGS__ADDED:
		SG_ERR_CHECK(  _pt_revert__revert_added(pCtx,prd,pszGidObject,pOD,dsFlags,ptn)  );
		return;

	case SG_DIFFSTATUS_FLAGS__DELETED:
		SG_ASSERT(  (ptn->bDeletedAfterAdded == ((dsFlags & SG_DIFFSTATUS_FLAGS__HACK__DELETED_ADDED) != 0))  );
		if (ptn->bDeletedAfterAdded)
			SG_ERR_CHECK(  _pt_revert__revert_cancel_added(pCtx,prd,pszGidObject,pOD,dsFlags,ptn)  );
		else
			SG_ERR_CHECK(  _pt_revert__revert_deleted(pCtx,prd,pszGidObject,pOD,dsFlags,ptn)  );
		return;

	case SG_DIFFSTATUS_FLAGS__FOUND:
		SG_ERR_CHECK(  _pt_revert__revert_found(pCtx,prd,pszGidObject,pOD,dsFlags,ptn)  );
		return;

	case SG_DIFFSTATUS_FLAGS__LOST:
		SG_ERR_CHECK(  _pt_revert__revert_lost(pCtx,prd,pszGidObject,pOD,dsFlags,ptn)  );
		return;

	default:		// __MODIFIED or __ZERO

		// __MODIFIED or UNMODIFIED (and LOST) entries can have __RENAMED and/or __MOVED (because we are
		// assuming a simple (non-composite) treediff.

		if (dsFlags & SG_DIFFSTATUS_FLAGS__MOVED)			// MOVED or MOVED+RENAMED
		{
			SG_ERR_CHECK(  _pt_revert__revert_moved(pCtx,prd,pszGidObject,pOD,dsFlags,ptn)  );
			return;
		}
		else if (dsFlags & SG_DIFFSTATUS_FLAGS__RENAMED)	// only RENAMED
		{
			SG_ERR_CHECK(  _pt_revert__revert_renamed(pCtx,prd,pszGidObject,pOD,dsFlags,ptn)  );
			return;
		}

		// there were no structural changes.
		//
		// the only thing left are __MODIFIED, __CHANGED_XATTRS, and __CHANGED_ATTRBITS.
		// these don't matter because they have the SAME entryname and the SAME parent
		// directory in both the SOURCE and TARGET views of the tree -- just like all of
		// the unmodified entries that do not appear in the treediff at all.
		//
		// we DO NOT claim anything for them (in either side).
		//
		// That is, we DO NOT put them into the inverted tree/index.  We DO NOT cause the
		// entryname to appear in the _inv_dir.prbEntries map.  BUT, it MAY get loaded if
		// the _inv_dir code needs to scan the directory.

		return;
	}

fail:
	return;
}

//////////////////////////////////////////////////////////////////

/**
 * Use the hierarchial view of the pendingtree to do a depth-first walk and move
 * all entries that need to be parked/swapped to their associated "parked" pathname.
 * We have to do this depth-first because as soon as we start this the whole
 * "repo-path --> absolute-path" stuff is invalid.
 *
 * For example, if we have:
 *      @/dir1/file1
 *      @/dir1/file2
 *      @/dir2/x
 * And the user does:
 *      vv rename ./dir1/file1 ./dir1/ftemp
 *      vv rename ./dir1/file2 ./dir1/file1
 *      vv rename ./dir1/ftemp ./dir1/file2
 *      vv rename ./dir1       ./dtemp
 *      vv rename ./dir2       ./dir1
 *      vv rename ./dtemp      ./dir2
 * Then the treediff sees:
 *      file1 --> file2
 *      file2 --> file1
 *      dir1  --> dir2
 *      dir2  --> dir1
 * Both of these files and both of these directories will have swap pathnames.
 * We must park the files before we park the directories, because parking the
 * directory also moves the contents of the directory.
 *
 * TODO 03/19/10 See about using the new pInvDirs->prbQueueForParking queue
 * TODO 03/19/10 to drive the move rather than this dive thru the entire
 * TODO 03/19/10 pendingtree.
 */
static void _pt_revert__move_to_parking_lot(SG_context * pCtx, struct _pt_revert_data * prd, const struct sg_ptnode * ptn, SG_uint32 depth)
{
	SG_rbtree_iterator * pIter = NULL;
	SG_pathname * pPath_in_wd = NULL;
	SG_bool bFound;

	// if this ptnode is the phantom part of a move, we ignore it.
	// all activity happens in the move-destination ptnode.

	if (ptn->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY)
		return;

	// dive first so that we can still use the repo-path-->absolute-path code.

	if (ptn->type == SG_TREENODEENTRY_TYPE_DIRECTORY)
	{
		if (ptn->prbItems)
		{
			const char * pszKey;
			const struct sg_ptnode * ptnSub;

			SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx,&pIter,ptn->prbItems,&bFound,&pszKey,(void **)&ptnSub)  );
			while (bFound)
			{
				SG_ERR_CHECK(  _pt_revert__move_to_parking_lot(pCtx,prd,ptnSub,depth+1)  );
				SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx,pIter,&bFound,&pszKey,(void **)&ptnSub)  );
			}
			SG_RBTREE_ITERATOR_NULLFREE(pCtx,pIter);
		}
	}

	if (ptn->pszgid)			// the super-root doesn't have a gid.
	{
		SG_inv_entry * pInvEntry = NULL;
		const SG_treediff2_ObjectData * pOD = NULL;
		const SG_pathname * pPathParked = NULL;		// we do not own this
		SG_inv_entry_flags inv_entry_flags;

		SG_ERR_CHECK(  SG_inv_dirs__fetch_source_binding(pCtx,prd->pInvDirs,ptn->pszgid,&bFound,&pInvEntry,(void **)&pOD)  );
		if (bFound)
			SG_ERR_CHECK(  SG_inv_entry__get_parked_path__ref(pCtx, pInvEntry, &pPathParked)  );
		if (pPathParked)
		{
			const char * pszRepoPath;

			SG_ASSERT(  (ptn->pCurrentParent)  );		// actual root @/ cannot be moved/renamed and so should never need swapping
			SG_ASSERT(  (pOD)  );

			SG_ERR_CHECK(  SG_inv_entry__get_flags(pCtx, pInvEntry, &inv_entry_flags)  );
			SG_ASSERT(  (inv_entry_flags & SG_INV_ENTRY_FLAGS__PARKED__MASK)  );

			SG_ERR_CHECK(  SG_treediff2__ObjectData__get_repo_path(pCtx,pOD,&pszRepoPath)  );
			SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path(pCtx,
																				 prd->pPendingTree->pPathWorkingDirectoryTop,
																				 pszRepoPath,
																				 &pPath_in_wd)  );
#if TRACE_REVERT
			SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
									   ("_pt_revert__move_to_parking_lot: %*c [reason %x][%s]\n"
										"                                 %*c wd   [%s]\n"
										"                                 %*c park [%s]\n"),
									   depth*4,' ',inv_entry_flags,pszRepoPath,
									   depth*4,' ',SG_pathname__sz(pPath_in_wd),
									   depth*4,' ',SG_pathname__sz(pPathParked))  );
#endif
			if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
				SG_ERR_CHECK(  _pt_log__move(pCtx,prd->pPendingTree,
											 ptn->pszgid,
											 SG_pathname__sz(pPath_in_wd),
											 SG_pathname__sz(pPathParked),
											 "PARK_FOR_SWAP")  );
			if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
				SG_ERR_CHECK(  SG_fsobj__move__pathname_pathname(pCtx,pPath_in_wd,pPathParked)  );

			SG_PATHNAME_NULLFREE(pCtx,pPath_in_wd);
		}
	}

	return;

fail:
	SG_RBTREE_ITERATOR_NULLFREE(pCtx,pIter);
	SG_PATHNAME_NULLFREE(pCtx,pPath_in_wd);
}

static void _pt_revert__undelete_skeleton(SG_context * pCtx, struct _pt_revert_data * prd, const struct sg_ptnode * ptn, SG_uint32 depth)
{
	//////////////////////////////////////////////////////////////////
	// TODO 3/25/10 Investigate if it would be quicker to use the linear
	// TODO 3/25/10 treediff and build a queue of needed skeleton dirs
	// TODO 3/25/10 and then create them in sorted order.  Like the UPDATE
	// TODO 3/25/10 code does.  See _pt_update__do__full_{6,6_1,6_2}().
	//////////////////////////////////////////////////////////////////

	SG_rbtree_iterator * pIter = NULL;
	SG_pathname * pPathTrans = NULL;
	struct sg_ptnode * ptnDesiredParent = NULL;
	SG_bool bFound;

	SG_ASSERT(  (ptn->type == SG_TREENODEENTRY_TYPE_DIRECTORY)  );

	// if this ptnode is the phantom part of a move, we ignore it.
	// all activity happens in the move-destination ptnode.

	if (ptn->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY)
		return;

	if (ptn->saved_flags & sg_PTNODE_SAVED_FLAG_DELETED)
	{
		// this node is either DELETED or LOST.  and in the case of LOST it could
		// be LOST, LOST+MOVED, LOST+RENAMED, or LOST+MOVED+RENAMED.

		SG_bool bActive = ((ptn->temp_flags & sg_PTNODE_TEMP_FLAG_REVERTING) != 0);

		if (bActive)
		{
			const char * pszEntryName_old;

			// get the current disk path of the DESIRED parent directory.  (because we are after the swap
			// parking and before the restore, we can't use the normal repo-path-->absolute-path code
			// because an ancestor might be parked in swap.

			if (ptn->pszgidMovedFrom)
			{
				// pending-tree allows LOST+MOVE.  (currently, pending-tree does not allow DELETE+MOVE,
				// but that doesn't matter for our purposes here.)  but i'm going to assert that we have
				// a LOST for now.

				SG_ASSERT(  (ptn->temp_flags & sg_PTNODE_TEMP_FLAG_IMPLICIT_DELETE)  );

				// we have a +MOVE, so we need choose the TARGET parent directory where we were moved from.
				// This is a little more lookup than the normal case.

				SG_ERR_CHECK(  sg_pendingtree__find_by_gid(pCtx,prd->pPendingTree,ptn->pszgidMovedFrom,&ptnDesiredParent)  );
			}
			else
			{
				// otherwise the SOURCE parent and the TARGET parent are the same (not MOVED).

				ptnDesiredParent = ptn->pCurrentParent;
			}

			SG_ERR_CHECK(  _pt_revert__get_transient_path(pCtx,prd,ptnDesiredParent,&pPathTrans)  );

#if defined(DEBUG)
			if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
			{
				// if all is right, our DESIRED parent directory will already exist.
				SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx,pPathTrans,&bFound,NULL,NULL)  );
				SG_ASSERT(  (bFound)  );
			}
#endif

			// because we were not present in the SOURCE version of the tree, we did not claim a
			// SOURCE-entryname and thus cannot be parked in swap.  furthermore, since we claimed the
			// TARGET-entryname, we own it cleanly and can create the directory without having to
			// worry about backup names or collisions.
			//
			// we fetch the old-name because we want the TARGET version.  for DELETES, this should be
			// the same as the SOURCE version (because we don't allow DELETE+RENAME); for LOSTS, they
			// could be different, but that doesn't matter.

			SG_ERR_CHECK(  sg_ptnode__get_old_name(pCtx, ptn, &pszEntryName_old)  );
			if (pszEntryName_old && *pszEntryName_old)
			{
				SG_ERR_CHECK(  SG_pathname__append__from_sz(pCtx, pPathTrans, pszEntryName_old)  );

#if TRACE_REVERT
				SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
										   ("_pt_revert__undelete_skeleton: %*c [%s]\n"),
										   depth*4,' ',SG_pathname__sz(pPathTrans))  );
#endif
				if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
					SG_ERR_CHECK(  _pt_log__mkdir(pCtx,prd->pPendingTree,
												  ptn->pszgid,
												  SG_pathname__sz(pPathTrans),
												  "SKELETON")  );
				if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
				{
					// TODO this can throw SG_ERR_DIR_ALREADY_EXISTS -- this should not happen because
					// TODO we already determined that it was missing (hence the DELETED or LOST status),
					// TODO so i'm going to just let this throw.
					//
					// TODO do we need to call SG_pathname__remove_final_slash() on this?
					SG_ERR_CHECK(  SG_fsobj__mkdir__pathname(pCtx, pPathTrans)  );

					// NOTE: We do not bother to restore attrbits and xattrs on this directory at this point.
					// NOTE: Nor reset the _FLAG_DELETED bit on the ptnode.
					// NOTE: We save these for _pt_revert__process_structural_change() to deal with so that we
					// NOTE: can maybe use the same code as we do for un-deleting files/symlinks.
				}
			}
			else
			{
				// This directory DID NOT have a name in the baseline.
				// This could happen when you do a:
				//     date > foo
				//     vv add foo
				//     /bin/rm foo
				//     vv revert foo
				// That is an ADDED-but-deleted gets labeled as a LOST
				// item so the REVERT just needs to remove the ptnode.
				// Anyway, for our purposes here, there is no need to
				// create the directory.

#if TRACE_REVERT
				SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
										   "_pt_revert__undelete_skeleton: [gid %s] no name.\n",
										   ptn->pszgid)  );
#endif

				// TODO 2011/04/07 Should we assert that ptn->prbItems is empty ?

			}

			SG_PATHNAME_NULLFREE(pCtx, pPathTrans);
		}
	}

	// TODO do i need to call sg_ptnode__load_dir_entries()?  it was called in sg_ptnode__get()
	// TODO that i was looking at as i wrote this, but i don't think it is necessary since we've
	// TODO already done the treediff.  so if ptn->prbItems is NULL and you don't think it should
	// TODO be, then maybe we do.

	if (ptn->prbItems)
	{
		const char * pszKey;
		const struct sg_ptnode * ptnSub;

		SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx,&pIter,ptn->prbItems,&bFound,&pszKey,(void **)&ptnSub)  );
		while (bFound)
		{
			if (ptnSub->type == SG_TREENODEENTRY_TYPE_DIRECTORY)
				SG_ERR_CHECK(  _pt_revert__undelete_skeleton(pCtx,prd,ptnSub,depth+1)  );
			SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx,pIter,&bFound,&pszKey,(void **)&ptnSub)  );
		}
		SG_RBTREE_ITERATOR_NULLFREE(pCtx,pIter);
	}

fail:
	SG_RBTREE_ITERATOR_NULLFREE(pCtx,pIter);
	SG_PATHNAME_NULLFREE(pCtx, pPathTrans);
}

//////////////////////////////////////////////////////////////////

static void _pt_revert__revert_attrbits(SG_context * pCtx,
										struct _pt_revert_data * prd,
										const SG_pathname * pPathAbsolute,
										struct sg_ptnode * ptn,
										const char * pszReason)
{
	SG_int64 attrBitsBaseline;

	SG_ERR_CHECK(  SG_treenode_entry__get_attribute_bits(pCtx,ptn->pBaselineEntry,&attrBitsBaseline)  );

#if TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
							   ("_pt_revert__revert_attrbits: setting ATTRBITS to [0x%x]\n"
								"                                             for %s\n"),
							   ((SG_uint32)attrBitsBaseline),
							   SG_pathname__sz(pPathAbsolute))  );
#endif
	if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
		SG_ERR_CHECK(  _pt_log__attrbits(pCtx,prd->pPendingTree,
										 ptn->pszgid,
										 SG_pathname__sz(pPathAbsolute),
										 attrBitsBaseline,
										 pszReason)  );
	if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
		SG_ERR_CHECK(  SG_attributes__bits__apply(pCtx,pPathAbsolute,attrBitsBaseline)  );

	ptn->iCurrentAttributeBits = attrBitsBaseline;

fail:
	return;
}

static void _pt_revert__revert_xattrs(SG_context * pCtx,
									  struct _pt_revert_data * prd,
									  const SG_pathname * pPathAbsolute,
									  struct sg_ptnode * ptn,
									  const char * pszReason)
{
	SG_vhash * pvhBaselineXAttrs = NULL;
	const char * pszHidBaselineXAttrs;

    SG_ERR_CHECK(  SG_treenode_entry__get_hid_xattrs(pCtx, ptn->pBaselineEntry, &pszHidBaselineXAttrs)  );

#if TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
							   ("_pt_revert__revert_xattrs: settings XATTRs to [hid %s]\n"
								"                                          for %s\n"),
							   ((pszHidBaselineXAttrs) ? pszHidBaselineXAttrs : "(null)"),
							   SG_pathname__sz(pPathAbsolute))  );
#endif

    if (pszHidBaselineXAttrs)
        SG_ERR_CHECK(  SG_repo__fetch_vhash(pCtx, prd->pPendingTree->pRepo,pszHidBaselineXAttrs,&pvhBaselineXAttrs)  );

	if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
		SG_ERR_CHECK(  _pt_log__xattrs(pCtx,prd->pPendingTree,
									   ptn->pszgid,
									   SG_pathname__sz(pPathAbsolute),
									   pszHidBaselineXAttrs,
									   pvhBaselineXAttrs,
									   pszReason)  );
#ifdef SG_BUILD_FLAG_FEATURE_XATTR
	if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
		SG_ERR_CHECK(  SG_attributes__xattrs__apply(pCtx, pPathAbsolute, pvhBaselineXAttrs, prd->pPendingTree->pRepo)  );
#else
	// TODO 3/29/10 What to do on NON-XATTR platforms?
#endif

	if (pszHidBaselineXAttrs)
		SG_ERR_CHECK(  SG_strpool__add__sz(pCtx, prd->pPendingTree->pStrPool, pszHidBaselineXAttrs, &ptn->pszHidXattrs)  );
	else
		ptn->pszHidXattrs = NULL;

fail:
    SG_VHASH_NULLFREE(pCtx, pvhBaselineXAttrs);
}

//////////////////////////////////////////////////////////////////

/**
 * the contents of the entry were not changes.  we may or may not have some XATTR and/or ATTRBITS
 * changes to restore.
 */
static void _pt_revert__revert_unmodified(SG_context * pCtx,
										  struct _pt_revert_data * prd,
										  SG_UNUSED_PARAM( const char * pszGidObject ),
										  SG_UNUSED_PARAM( const SG_treediff2_ObjectData * pOD ),
										  SG_diffstatus_flags dsFlags,
										  struct sg_ptnode * ptn)
{
	SG_string * pStringRepoPath = NULL;
	SG_pathname * pPathAbsolute = NULL;

	SG_UNUSED( pszGidObject );
	SG_UNUSED( pOD );

	if ((dsFlags & (SG_DIFFSTATUS_FLAGS__CHANGED_ATTRBITS | SG_DIFFSTATUS_FLAGS__CHANGED_XATTRS)) == 0)
		return;

	// see note about needing to use sg_ptnode__get_repo_path() rather than SG_treediff2__.

	SG_ERR_CHECK(  sg_ptnode__get_repo_path(pCtx,ptn,&pStringRepoPath)  );
	SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path(pCtx,
																		 prd->pPendingTree->pPathWorkingDirectoryTop,
																		 SG_string__sz(pStringRepoPath),
																		 &pPathAbsolute)  );

	if (dsFlags & SG_DIFFSTATUS_FLAGS__CHANGED_ATTRBITS)
		SG_ERR_CHECK(  _pt_revert__revert_attrbits(pCtx,prd,pPathAbsolute,ptn,"UNDO_CHANGE")  );

	if (dsFlags & SG_DIFFSTATUS_FLAGS__CHANGED_XATTRS)
		SG_ERR_CHECK(  _pt_revert__revert_xattrs(pCtx,prd,pPathAbsolute,ptn,"UNDO_CHANGE")  );

	// TODO I'm going to argue that we don't need to update the timestamp cache for these.

fail:
	SG_STRING_NULLFREE(pCtx,pStringRepoPath);
	SG_PATHNAME_NULLFREE(pCtx,pPathAbsolute);
}

/**
 * I'm going to say that a file/symlink is disposable
 * if its current content matches something that we
 * know that we created/generated.  Such as something
 * we fetched from the other branch in a MERGE or
 * something that we generated thru auto-merged (when
 * we *KNOW* that the tool was fully automatic).
 */
static void _pt_revert__is_disposable_file_symlink(SG_context * pCtx,
												   struct sg_ptnode * ptn,
												   SG_bool * pbIsDisposable)
{
	SG_bool bIsDisposable = SG_FALSE;

	switch (ptn->type)
	{
	default:
		SG_ERR_THROW(  SG_ERR_NOTIMPLEMENTED  );
		break;

	case SG_TREENODEENTRY_TYPE_REGULAR_FILE:
	case SG_TREENODEENTRY_TYPE_SYMLINK:
		if (ptn->pszHidDisposable && *ptn->pszHidDisposable)
		{
			// because we did a scan-dir (inside do_diff) at the start
			// of the revert, we already have the current/live HID of
			// the content.

			const char * pszHidCurrentOrBaseline;

			SG_ERR_CHECK(  sg_ptnode__get_hid(pCtx, ptn, &pszHidCurrentOrBaseline)  );
			bIsDisposable = (strcmp(ptn->pszHidDisposable, pszHidCurrentOrBaseline) == 0);
		}
		break;
	}

#if TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "_pt_revert__is_disposable_file_symlink: entry %s disposable: %s\n",
							   ((bIsDisposable) ? "is" : "is not"),
							   ptn->pszgid)  );
#endif

	*pbIsDisposable = bIsDisposable;

fail:
	;
}

/**
 * A 'disposable' submodule is one that was created/generated
 * (such as by MERGE when folding in new items created in the
 * other branch) **AND** that has not be modified since the
 * merge.  (and is in theory, completely re-constructable.)
 */
static void _pt_revert__is_disposable_submodule(SG_context * pCtx,
												   struct sg_ptnode * ptn,
												   SG_bool * pbIsDisposable)
{
	SG_bool bIsDisposable = SG_FALSE;
	SG_bool bDirtyPending = SG_FALSE;
	SG_bool bMatch = SG_FALSE;

	switch (ptn->type)
	{
	default:
		SG_ERR_THROW(  SG_ERR_NOTIMPLEMENTED  );
		break;

	case SG_TREENODEENTRY_TYPE_SUBMODULE:
		if (ptn->pszHidDisposable && *ptn->pszHidDisposable)
		{
			// because we did a scan-dir (inside do_diff) at the start
			// of the revert, we already have the current/live HID of
			// the content.

			const char * pszHidCurrentOrBaseline;

			// we don't pass bDirtyReference because the submodule looks
			// like an ADDED node and since there isn't a baseline hid,
			// it will always be true.  we use pszHidDisposable as the
			// baseline reference and compare that with the current hid.

			SG_ERR_CHECK(  sg_ptnode__is_submodule_changed_or_dirty(pCtx, ptn,
																	NULL,
																	&bDirtyPending)  );
			
			SG_ERR_CHECK(  sg_ptnode__get_hid(pCtx, ptn, &pszHidCurrentOrBaseline)  );
			bMatch = (strcmp(ptn->pszHidDisposable, pszHidCurrentOrBaseline) == 0);

			bIsDisposable = (bMatch && !bDirtyPending);
		}
		break;
	}

#if TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "_pt_revert__is_disposable_submodule: [dirty_p %d][match %d][disposable %d]: %s\n",
							   bDirtyPending, bMatch, bIsDisposable,
							   ptn->pszgid)  );
#endif

	*pbIsDisposable = bIsDisposable;

fail:
	;
			
}

//////////////////////////////////////////////////////////////////

static void _pt_revert__revert_modified__file_symlink(SG_context * pCtx,
													  struct _pt_revert_data * prd,
													  struct sg_ptnode * ptn);

/**
 * the contents of the entry have been changed and we want to revert it.
 *
 * for files and (maybe) symlinks, we want to backup the entry before restoring the contents from
 * the repo.  (because we don't want to blindly delete the user's only copy of something that they
 * changed.)
 *
 * Note: we also get a __MODIFIED bit for directories, but this is just noise (indicating that
 * something *within* the directory changed.  We ignore the __MODIFIED bit for them.
 *
 * We also need to restore the original XATTRs and ATTRBITS.
 */
static void _pt_revert__revert_modified(SG_context * pCtx,
										struct _pt_revert_data * prd,
										const char * pszGidObject,
										const SG_treediff2_ObjectData * pOD,
										SG_diffstatus_flags dsFlags,
										struct sg_ptnode * ptn)
{
	switch (ptn->type)
	{
	default:
		SG_ERR_THROW2(  SG_ERR_NOTIMPLEMENTED,
						(pCtx, "TODO unknown tneType: %d", ptn->type)  );
		break;

	case SG_TREENODEENTRY_TYPE_DIRECTORY:
		{
			// if MERGE marked this directory as disposable, then we
			// might be able to just delete the directory on the revert
			// rather than either giving it a ~sg00~ name or leaving it
			// as "found" trash in the WD.
			//
			// BUT we cannot actually do anything about it in this pass;
			// (because we are called in GID order rather than tree depth)
			// we must wait until a later pass.  But we can count them to
			// see if we actually need that later pass.
			if (ptn->bDisposableDir)
				prd->nrDisposableDirs++;

			// regardless of what eventually happens (delete vs backup),
			// reset the ATTRs/XATTRs.
			SG_ERR_CHECK(  _pt_revert__revert_unmodified(pCtx,prd,pszGidObject,pOD,dsFlags,ptn)  );
			ptn->pszCurrentHid = NULL;
		}
		break;

	case SG_TREENODEENTRY_TYPE_SUBMODULE:
		// This routine was called in step [8].  I want to defer
		// this until step [8.6] so that we complete the revert
		// of everything in the outer WD before we do any diving
		// into submodules.
		break;

	case SG_TREENODEENTRY_TYPE_REGULAR_FILE:
	case SG_TREENODEENTRY_TYPE_SYMLINK:
		SG_ERR_CHECK(  _pt_revert__revert_modified__file_symlink(pCtx, prd, ptn)  );
		break;
	}

fail:
	;
}

static void _pt_revert__revert_modified__file_symlink(SG_context * pCtx,
													  struct _pt_revert_data * prd,
													  struct sg_ptnode * ptn)
{
	SG_string * pStringRepoPath = NULL;
	SG_pathname * pPathAbsolute = NULL;
	SG_pathname * pPathBackup = NULL;
	SG_file * pFile = NULL;
	SG_string * pStringLink = NULL;
	SG_byte * pBytes = NULL;
	const char * pszHidBaseline;
	SG_bool bIsDisposable = SG_FALSE;

	// see note about needing to use sg_ptnode__get_repo_path() rather than SG_treediff2__.

	SG_ERR_CHECK(  sg_ptnode__get_repo_path(pCtx,ptn,&pStringRepoPath)  );
	SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path(pCtx,
																		 prd->pPendingTree->pPathWorkingDirectoryTop,
																		 SG_string__sz(pStringRepoPath),
																		 &pPathAbsolute)  );

	SG_ERR_CHECK(  SG_treenode_entry__get_hid_blob(pCtx, ptn->pBaselineEntry, &pszHidBaseline)  );

	// see if the dirty file/symlink matches something that we
	// created/generated.  if so, it is disposable and we don't
	// need to back it up before we restore/overwrite it.

	SG_ERR_CHECK(  _pt_revert__is_disposable_file_symlink(pCtx, ptn, &bIsDisposable)  );
	if (bIsDisposable)
	{
		if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
			SG_ERR_CHECK(  _pt_log__remove_item(pCtx,
												prd->pPendingTree,
												ptn->pszgid,
												SG_pathname__sz(pPathAbsolute),
												"DISCARDING_DISPOSABLE_MODIFIED_ITEM_BEFORE_RESTORE")  );
		if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
		{
#if TRACE_REVERT
			SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
									   ("_pt_revert__revert_modified: repo-path [%s]\n"
										"                          disposing of [%s]\n"),
									   SG_string__sz(pStringRepoPath),
									   SG_pathname__sz(pPathAbsolute))  );
#endif
			SG_ERR_CHECK(  SG_fsobj__remove__pathname(pCtx, pPathAbsolute)  );
		}
	}
	else if (prd->bNoBackups)
	{
		if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
			SG_ERR_CHECK(  _pt_log__remove_item(pCtx,
												prd->pPendingTree,
												ptn->pszgid,
												SG_pathname__sz(pPathAbsolute),
												"NO_BACKUP_OF_MODIFIED_ITEM_BEFORE_RESTORE")  );
		if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
		{
#if TRACE_REVERT
			SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
									   ("_pt_revert__revert_modified: repo-path [%s]\n"
										"                             no-backup [%s]\n"),
									   SG_string__sz(pStringRepoPath),
									   SG_pathname__sz(pPathAbsolute))  );
#endif
			SG_ERR_CHECK(  SG_fsobj__remove__pathname(pCtx, pPathAbsolute)  );
		}
	}
	else
	{
		SG_ERR_CHECK(  SG_workingdir__calc_backup_name(pCtx,pPathAbsolute,&pPathBackup)  );

#if TRACE_REVERT
		SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
								   ("_pt_revert__revert_modified: repo-path [%s]\n"
									"                            backing-up [%s]\n"
									"                                    as [%s]\n"),
								   SG_string__sz(pStringRepoPath),
								   SG_pathname__sz(pPathAbsolute),
								   SG_pathname__sz(pPathBackup))  );
#endif
		if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
			SG_ERR_CHECK(  _pt_log__move(pCtx,prd->pPendingTree,
										 ptn->pszgid,
										 SG_pathname__sz(pPathAbsolute),
										 SG_pathname__sz(pPathBackup),
										 "BACKUP_BEFORE_RESTORE")  );
		if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
			SG_ERR_CHECK(  SG_fsobj__move__pathname_pathname(pCtx,pPathAbsolute,pPathBackup)  );
	}

	// restore the file/symlink from the repo.

#if TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
							   ("_pt_revert__revert_modified: restoring [gid %s]\n"
								"                                       [hid %s]\n"
								"                                    to %s\n"),
							   ptn->pszgid,pszHidBaseline,SG_pathname__sz(pPathAbsolute))  );
#endif

	if (ptn->type == SG_TREENODEENTRY_TYPE_REGULAR_FILE)
	{
		if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
			SG_ERR_CHECK(  _pt_log__restore_file(pCtx,prd->pPendingTree,
												 ptn->pszgid,
												 SG_pathname__sz(pPathAbsolute),
												 pszHidBaseline,
												 "RESTORE")  );
		if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
		{
			SG_timestamp_cache * pTSC = NULL;		// we do not own this
			SG_fsobj_stat stat_now_current;

			SG_ERR_CHECK(  SG_file__open__pathname(pCtx,
												   pPathAbsolute,
												   SG_FILE_WRONLY | SG_FILE_OPEN_OR_CREATE | SG_FILE_TRUNC,
												   0644,
												   &pFile)  );
			SG_ERR_CHECK(  SG_repo__fetch_blob_into_file(pCtx, ptn->pPendingTree->pRepo, pszHidBaseline, pFile, NULL)  );
			SG_FILE_NULLCLOSE(pCtx, pFile);

			// update the timestamp cache for the restored file.

			SG_ERR_CHECK(  SG_fsobj__stat__pathname(pCtx, pPathAbsolute, &stat_now_current)  );
			SG_ERR_CHECK(  SG_pendingtree__get_timestamp_cache(pCtx, ptn->pPendingTree, &pTSC)  );
			SG_ERR_CHECK(  SG_timestamp_cache__add(pCtx, pTSC, ptn->pszgid, stat_now_current.mtime_ms, stat_now_current.size, pszHidBaseline)  );
		}

		ptn->pszCurrentHid = NULL;
		ptn->pszHidDisposable = NULL;
	}
	else
	{
		SG_uint64 iLenBytes = 0;

		SG_ERR_CHECK(  SG_repo__fetch_blob_into_memory(pCtx, ptn->pPendingTree->pRepo, pszHidBaseline, &pBytes, &iLenBytes)  );
		SG_ERR_CHECK(  SG_STRING__ALLOC__BUF_LEN(pCtx, &pStringLink, pBytes, (SG_uint32) iLenBytes)  );
		SG_NULLFREE(pCtx, pBytes);

		if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
			SG_ERR_CHECK(  _pt_log__restore_symlink(pCtx,prd->pPendingTree,
													ptn->pszgid,
													SG_pathname__sz(pPathAbsolute),
													SG_string__sz(pStringLink),
													"RESTORE")  );
		if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
			SG_ERR_CHECK(  SG_fsobj__symlink(pCtx, pStringLink, pPathAbsolute)  );

		ptn->pszCurrentHid = NULL;
		ptn->pszHidDisposable = NULL;

		SG_STRING_NULLFREE(pCtx, pStringLink);
	}

	// since we just create the entry on the disk, we always apply the ATTRBITS and XATTRS as
	// they were recorded in the repo.

	SG_ERR_CHECK(  _pt_revert__revert_attrbits(pCtx,prd,pPathAbsolute,ptn,"RESTORE")  );

	SG_ERR_CHECK(  _pt_revert__revert_xattrs(pCtx,prd,pPathAbsolute,ptn,"RESTORE")  );

fail:
	SG_STRING_NULLFREE(pCtx, pStringRepoPath);
	SG_PATHNAME_NULLFREE(pCtx, pPathAbsolute);
	SG_PATHNAME_NULLFREE(pCtx, pPathBackup);
	SG_FILE_NULLCLOSE(pCtx, pFile);
	SG_STRING_NULLFREE(pCtx, pStringLink);
	SG_NULLFREE(pCtx, pBytes);
}

//////////////////////////////////////////////////////////////////

static SG_vector_foreach_callback _pt_revert__process_cyclic_directory_moves;

static void _pt_revert__process_cyclic_directory_moves(SG_context * pCtx, SG_UNUSED_PARAM( SG_uint32 ndx ), void * pVoidAssoc, void * pVoidData)
{
	struct sg_ptnode * ptn = (struct sg_ptnode *)pVoidAssoc;
	struct _pt_revert_data * prd = (struct _pt_revert_data *)pVoidData;

	SG_UNUSED( ndx );

	SG_ASSERT(  (ptn->temp_flags & sg_PTNODE_TEMP_FLAG_PARKED_FOR_CYCLE)  );

#if TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
							   "_pt_revert__process_cyclic_directory_moves: handling deferred revert for [gid %s]\n",
							   ptn->pszgid)  );
#endif

	SG_ERR_CHECK(  _pt_revert__revert_moved__apply(pCtx,prd,ptn)  );

fail:
	return;
}

//////////////////////////////////////////////////////////////////

/**
 * Process any non-structural changes on the entry.  This includes restoring modified files/symlinks
 * and any changes to XATTR or ATTRBITS.
 *
 * We have to be careful with the repo-path here.   Because we have already handled the MOVES/RENAMES/etc,
 * the entry now has its ORIGINAL entryname and is in its ORIGINAL directory in the WD on disk.  so we
 * cannot use the SG_treediff2__ObjectData__get_repo_path() because it computes values based upon the
 * CURRENT settings.  Use the PTNODE version because (in theory) the PTNODES have been updated during the
 * earlier steps in the REVERT.  (The TreeDiff has not.)
 */
static SG_treediff2_foreach_callback _pt_revert__process_nonstructural_changes;

static void _pt_revert__process_nonstructural_changes(SG_context * pCtx,
													  SG_UNUSED_PARAM( SG_treediff2 * pTreeDiff ),
													  const char * pszGidObject,
													  const SG_treediff2_ObjectData * pOD,
													  void * pVoidData)
{
	struct _pt_revert_data * prd = (struct _pt_revert_data *)pVoidData;
	struct sg_ptnode * ptn = NULL;
	SG_diffstatus_flags dsFlags;
	SG_treenode_entry_type tneType;

	SG_UNUSED( pTreeDiff );

	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_tneType(pCtx, pOD, &tneType)  );
	if (tneType == SG_TREENODEENTRY_TYPE_SUBMODULE)
	{
		// we get called from step [8] where we take care of the contents of files/symlinks.
		// but we need to defer the contents of submodules to [8.6].
		return;
	}
	
	SG_ERR_CHECK(  SG_treediff2__ObjectData__get_dsFlags(pCtx,pOD,&dsFlags)  );
	switch (dsFlags & SG_DIFFSTATUS_FLAGS__MASK_MUTUALLY_EXCLUSIVE)
	{
	case SG_DIFFSTATUS_FLAGS__ADDED:		// TODO XREF[B]: ptn may no longer exist (see XREF[A])
	case SG_DIFFSTATUS_FLAGS__DELETED:
	case SG_DIFFSTATUS_FLAGS__FOUND:
	case SG_DIFFSTATUS_FLAGS__LOST:
		SG_ASSERT(  ((dsFlags & (SG_DIFFSTATUS_FLAGS__CHANGED_XATTRS | SG_DIFFSTATUS_FLAGS__CHANGED_ATTRBITS)) == 0)  );
		return;

	default:
		break;
	}

	SG_ERR_CHECK(  sg_pendingtree__find_by_gid(pCtx,prd->pPendingTree,pszGidObject,&ptn)  );

#if TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
							   "ProcessContentChanges: [gid %s][dsFlags 0x%x][ptn s flags %d][ptn t flags %d]\n",
							   pszGidObject,dsFlags,ptn->saved_flags,ptn->temp_flags)  );
#endif

	if (!(ptn->temp_flags & sg_PTNODE_TEMP_FLAG_REVERTING))
		return;

	if (dsFlags & SG_DIFFSTATUS_FLAGS__MODIFIED)
		SG_ERR_CHECK(  _pt_revert__revert_modified(pCtx,prd,pszGidObject,pOD,dsFlags,ptn)  );
	else
		SG_ERR_CHECK(  _pt_revert__revert_unmodified(pCtx,prd,pszGidObject,pOD,dsFlags,ptn)  );

fail:
	return;
}

//////////////////////////////////////////////////////////////////

static void _pt_revert__identify_disposable_dirs(SG_context * pCtx,
												 struct _pt_revert_data * prd,
												 struct sg_ptnode * ptn,
												 SG_bool * pbShouldBeDeleted)
{
	struct sg_ptnode * ptn_k;
	const char * pszGid_k;
	SG_rbtree_iterator * pIter = NULL;
	SG_string * pStrRepoPath = NULL;
	SG_pathname * pPath = NULL;
	SG_uint32 nrItemsToBeDeleted;
	SG_uint32 nrItemsRemaining;
	SG_uint32 nrItems;
	SG_bool b;

	*pbShouldBeDeleted = SG_FALSE;

	if (!ptn)
		return;
	if (ptn->type != SG_TREENODEENTRY_TYPE_DIRECTORY)
		return;
	if (ptn->saved_flags & sg_PTNODE_SAVED_FLAG_MOVED_AWAY)
		return;
	if (!ptn->prbItems)
		return;

	// dive first and try to delete disposable subdirectories.
	// we can't actually delete the items from prbItems during
	// an iteration.

	nrItemsToBeDeleted = 0;
	SG_ERR_CHECK(  SG_rbtree__iterator__first(pCtx, &pIter, ptn->prbItems, &b, &pszGid_k, (void **)&ptn_k)  );
	while (b)
	{
		SG_bool bShouldBeDeleted_k;

		SG_ERR_CHECK(  _pt_revert__identify_disposable_dirs(pCtx, prd, ptn_k, &bShouldBeDeleted_k)  );
		if (bShouldBeDeleted_k)
			nrItemsToBeDeleted++;

		SG_ERR_CHECK(  SG_rbtree__iterator__next(pCtx, pIter, &b, &pszGid_k, (void **)&ptn_k)  );
	}

	if (ptn->bDisposableDir)
	{
		// we have a candidate disposable directory.
		// since this code runs after all of the structural changes are
		// made, we should be able to use the regular pathname.
		// 
		// TODO 2011/03/08 confirm this when only a --test is run.

		SG_ERR_CHECK(  sg_ptnode__get_repo_path(pCtx, ptn, &pStrRepoPath)  );
		SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path(pCtx,
																			 prd->pPendingTree->pPathWorkingDirectoryTop,
																			 SG_string__sz(pStrRepoPath),
																			 &pPath)  );

		// see if the user created any trash in it that would prevent
		// us from actually disposing of it.  we can't just use the
		// actual prbItems count because we haven't actually deleted
		// the files or ptnodes yet.

		SG_ERR_CHECK(  SG_rbtree__count(pCtx, ptn->prbItems, &nrItems)  );
		nrItemsRemaining = (nrItems - nrItemsToBeDeleted);

#if TRACE_REVERT
		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR, "Candidate disposable directory: [nrItemsRemaining %d] %s\n",
								   nrItemsRemaining, SG_pathname__sz(pPath))  );
#endif

		if (nrItemsRemaining == 0)
		{
			// this directory is disposable.  add it to the queue of things
			// to delete.  we can't delete it now because our caller is iterating
			// on the parent ptnode's prbItems.

			SG_ERR_CHECK(  SG_vector__append(pCtx, prd->pVecDisposablePathsToDelete, (void *)pPath, NULL)  );
			pPath = NULL;		// vector owns this now
			SG_ERR_CHECK(  SG_vector__append(pCtx, prd->pVecDisposablePTNodesToDelete, (void *)ptn, NULL)  );

			*pbShouldBeDeleted = SG_TRUE;
		}
		else
		{
			// we computed that there is dirt in the directory so we should
			// not delete it.
			//
			// TODO 2011/03/08 Should we give it a ~sg00~ name so that it is
			// TODO            ignored or should we leave it as it so that the
			// TODO            user is more likely to see it ?  I'm not sure
			// TODO            there is a right answer.
			// TODO
			// TODO            Note that it may already have a ~sg00~ name
			// TODO            if there was an entryname collision and it lost.
		}
	}

fail:
	SG_STRING_NULLFREE(pCtx, pStrRepoPath);
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_RBTREE_ITERATOR_NULLFREE(pCtx, pIter);
}

static void _pt_revert__delete_queued_disposable_dirs(SG_context * pCtx,
													  struct _pt_revert_data * prd)
{
	SG_uint32 k, kLimit;
	struct sg_ptnode * ptn_owned = NULL;

	SG_ERR_CHECK(  SG_vector__length(pCtx, prd->pVecDisposablePathsToDelete, &kLimit)  );
	for (k=0; k<kLimit; k++)
	{
		SG_pathname * pPath_k;		// we do not own this
		struct sg_ptnode * ptn_k;	// we do not own this

		SG_ERR_CHECK(  SG_vector__get(pCtx, prd->pVecDisposablePathsToDelete, k, (void **)&pPath_k)  );
		SG_ERR_CHECK(  SG_vector__get(pCtx, prd->pVecDisposablePTNodesToDelete, k, (void **)&ptn_k)  );

		if (prd->pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)
			SG_ERR_CHECK(  _pt_log__remove_item(pCtx,
												prd->pPendingTree,
												ptn_k->pszgid,
												SG_pathname__sz(pPath_k),
												"REMOVE_DISPOSABLE_DIRECTORY")  );
		if (prd->pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
		{
			// TODO 2011/03/08 should we let this quietly fail (in case there
			// TODO            is stuff in the directory that we don't know about) ?

			SG_ERR_CHECK(  SG_fsobj__rmdir__pathname(pCtx, pPath_k)  );
		}

		SG_ERR_CHECK(  SG_rbtree__remove__with_assoc(pCtx, ptn_k->pCurrentParent->prbItems,
													 ptn_k->pszgid, (void **)&ptn_owned)  );
		SG_ASSERT(  (ptn_owned == ptn_k)  );
		SG_PTNODE_NULLFREE(pCtx, ptn_owned);
	}

fail:
	;
}

//////////////////////////////////////////////////////////////////

/**
 * Stuff that should have been been done in step [7] __APPLY_STRUCTURAL_CHANGES
 * phase (for DELETED and LOST submodules) and step [8] non-structural changes
 * (for MODIFIED/DIRTY submodules) that we deferred until last.
 *
 */
static void _pt_revert__handle_queued_subs(SG_context * pCtx,
										   struct _pt_revert_data * prd)
{
	SG_uint32 k, count;

	SG_ERR_CHECK(  SG_vector__length(pCtx, prd->pvec_queued_subs, &count)  );

#if TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "HandleQueuedSubs: [count %d]\n",
							   count)  );
#endif

	for (k=0; k<count; k++)
	{
		struct _queued_subs * pQS;
		SG_diffstatus_flags dsFlags;
		
		SG_ERR_CHECK(  SG_vector__get(pCtx, prd->pvec_queued_subs, k, (void **)&pQS)  );
		SG_ERR_CHECK(  SG_treediff2__ObjectData__get_dsFlags(pCtx, pQS->pOD, &dsFlags)  );

#if TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "HandleQueuedSubs: [%d]: [dsFlags 0x%x] %s\n",
							   k,
							   dsFlags,
							   pQS->ptn->pszgid)  );
#endif

		switch (dsFlags & SG_DIFFSTATUS_FLAGS__MASK_MUTUALLY_EXCLUSIVE)
		{
		case SG_DIFFSTATUS_FLAGS__MODIFIED:
			SG_ERR_CHECK(  _pt_revert__modified_submodule(pCtx, prd, pQS)  );
			break;

		case SG_DIFFSTATUS_FLAGS__DELETED:
			SG_ERR_CHECK(  _pt_revert__recreate_deleted_submodule(pCtx, prd, pQS)  );

			pQS->ptn->saved_flags &= ~sg_PTNODE_SAVED_FLAG_DELETED;
			pQS->ptn->bDeletedAfterAdded = SG_FALSE;
			pQS->ptn->temp_flags |= sg_PTNODE_TEMP_FLAG_CHECKMYNAME;
			break;

		case SG_DIFFSTATUS_FLAGS__LOST:
			SG_ERR_CHECK(  _pt_revert__recreate_lost_submodule(pCtx, prd, pQS)  );

			pQS->ptn->saved_flags &= ~sg_PTNODE_SAVED_FLAG_DELETED;
			pQS->ptn->temp_flags &= ~sg_PTNODE_TEMP_FLAG_IMPLICIT_DELETE;
			pQS->ptn->temp_flags |= sg_PTNODE_TEMP_FLAG_CHECKMYNAME;
			break;

		default:
			SG_ERR_THROW( SG_ERR_NOTIMPLEMENTED );
			break;
		}
	}

#if TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
							   "HandleQueuedSubs: Finished.\n")  );
#endif

fail:
	;
}

//////////////////////////////////////////////////////////////////

void SG_pendingtree__revert(SG_context* pCtx,
							SG_pendingtree* pPendingTree,
							SG_pendingtree_action_log_enum eActionLog,
							SG_uint32 count_items, const char** paszItems,
							SG_bool bRecursive,
							SG_bool bNoSubmodules,
							SG_bool bNoBackups,
							const SG_vector * pvec_use_submodules,
							const SG_file_spec* pFilespec,
							SG_bool bPrecomputeOnly,
							SG_varray * pvaMissingDetails)
{
	// We DO NOT need/use the IGNORES list because we are not scanning.

	struct _pt_revert_data rd;
	SG_uint32 count_changes;
	SG_bool bDirty;
	SG_uint32 nrParents;
	const SG_varray * pva_wd_parents = NULL;
	SG_uint32 nrParked;
	SG_file_spec* pInternalFilespec = NULL;
	SG_pathname * pPathMergeTempDir = NULL;
	SG_uint32 nrMissingDetails = 0;
	SG_bool bHasMergeTempDir = SG_FALSE;
	SG_portability_flags flags_logged;

	memset(&rd,0,sizeof(rd));

	SG_NULLARGCHECK_RETURN(pPendingTree);
	SG_ARGCHECK_RETURN(  ((   count_items==0)   || (   paszItems != NULL)), count_items     );
	SG_ARGCHECK_RETURN(  ((bRecursive==SG_TRUE) || (count_items > 0)),      bRecursive      );	// if not recursive, require at least one item.

	SG_ERR_CHECK(  SG_FILE_SPEC__ALLOC__COPY(pCtx, pFilespec, &pInternalFilespec)  );

#if TRACE_REVERT
	{
		SG_uint32 k;

		SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
								   ("SG_pendingtree__revert:\n"
									"\t[eActionLog %d]\n"
									"\t[bRecursive %d]\n"
									"\t[bNoSubmodules %d]\n"
									"\t[bNoBackups %d]\n"
									"\t[count_items %d]\n"),
								   eActionLog,
								   bRecursive,
								   bNoSubmodules,
								   bNoBackups,
								   count_items)  );
		for (k=0; k<count_items; k++)
			SG_ERR_IGNORE(  SG_console(pCtx, SG_CS_STDERR,
									   ("\t\titem[%d]: %s\n"),
									   k,paszItems[k])  );
	}
#endif

	SG_ERR_CHECK(  SG_pendingtree__get_wd_parents__ref(pCtx, pPendingTree, &pva_wd_parents)  );
	SG_ERR_CHECK(  SG_varray__count(pCtx, pva_wd_parents, &nrParents)  );
	SG_ASSERT(  (nrParents >= 1)  );

	// for now we disallow partial-reverts when we have a merge.  there are
	// various reasons to allow this and to disallow this.  for example, what
	// does it mean?  for an individual file that was auto-merged and then
	// edited by the user, do we revert it to the auto-merge-result or to the
	// version in one of the input csets and if so, which one?

	if (nrParents > 1)
	{
		SG_bool bHasIncludeExclude = SG_FALSE;
		SG_bool bHaveArgs = SG_FALSE;
		
		SG_ERR_CHECK(  SG_file_spec__has_pattern(pCtx, pInternalFilespec, SG_FILE_SPEC__PATTERN__INCLUDE | SG_FILE_SPEC__PATTERN__EXCLUDE, &bHasIncludeExclude)  );

		bHaveArgs = ((count_items > 0) || (bHasIncludeExclude == SG_TRUE) || (!bRecursive));
		if (bHaveArgs)
			SG_ERR_THROW(  SG_ERR_CANNOT_PARTIAL_REVERT_AFTER_MERGE  );
	}

	rd.bNoBackups = bNoBackups;
	rd.bNoSubmodules = bNoSubmodules;
	rd.pvec_use_submodules = pvec_use_submodules;
	rd.pvaMissingDetails = pvaMissingDetails;
	rd.pPendingTree = pPendingTree;

	// eActionLog indicates whether we should actually do the work, just log what
	// should be done, or both.  We set it here for the body of the REVERT.  We
	// don't really anticipate it being changed/set differently by a series of
	// commands.  That is, the log should either be a complete list of what SHOULD
	// be done -OR- a complete list of what we DID do, but not a mingling of the 2.
	SG_ASSERT(  (rd.pPendingTree->eActionLog == SG_PT_ACTION__ZERO) || (rd.pPendingTree->eActionLog == eActionLog)  );
	rd.pPendingTree->eActionLog = eActionLog;

	SG_ERR_CHECK(  SG_INV_DIRS__ALLOC(pCtx, pPendingTree->pPathWorkingDirectoryTop, "revert", &rd.pInvDirs)  );
	// defer allocation of rd.pVecCycles until we need it.

	// Compute diff between the baseline and the WD.  this lets us know if the WD is clean or dirty.
	// The BASELINE will be on the left and the WD (baseline+dirt) will be on the right.  REVERT is
	// then a {complete, partial} UNDO.
	//
	// We DO NOT use the __file_spec_filter on this treediff; we want the complete diff.  The --include
	// --exclude stuff gets handled by __mark_all / __mark_some later.  (handwave)

	SG_ERR_CHECK(  _sg_pendingtree__do_diff__unload_clean_objects(pCtx,pPendingTree, NULL, NULL, &rd.pTreeDiff)  );

	SG_ERR_CHECK(  SG_treediff2__count_changes(pCtx,rd.pTreeDiff,&count_changes)  );
	if (count_changes == 0)
		goto done;

#if TRACE_REVERT
	{
		SG_vhash * pvh = NULL;

		SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
								   "SG_pendingtree__revert: [count_changes %d] [wd %s]\n",
								   count_changes,
								   SG_pathname__sz(pPendingTree->pPathWorkingDirectoryTop))  );
		SG_ERR_IGNORE(  SG_treediff2_debug__dump_to_console(pCtx,rd.pTreeDiff,SG_FALSE)  );
		pPendingTree->bDebugExport = SG_TRUE;
		SG_ERR_IGNORE(  sg_pendingtree__to_vhash(pCtx,pPendingTree,&pvh)  );
		pPendingTree->bDebugExport = SG_FALSE;
		SG_ERR_IGNORE(  SG_vhash_debug__dump_to_console(pCtx,pvh)  );
		SG_VHASH_NULLFREE(pCtx,pvh);
	}
#endif

	if (count_items > 0)
	{
		SG_ERR_CHECK(  sg_ptnode__maybe_mark_items_for_revert(pCtx,
															  pPendingTree,
															  count_items, paszItems,
															  bRecursive,
															  pInternalFilespec)  );

		/* check the tree to see if the user is trying to revert only
		 * the target half of a move operation.  If so, fail.  We
		 * cannot allow this case because it can result in a situation
		 * where the same object (identified by its gid) exists in two
		 * places in the tree at the same time. */

		// TODO Jeff says: 3/5/10 Is this _split_moves stuff still needed?
		// TODO            I mean, how could they have requested only half
		// TODO            of the move?  The phantom node is not addressable,
		// TODO            right?

		SG_ERR_CHECK(  sg_ptnode__fail_if_any_split_moves(pCtx, pPendingTree->pSuperRoot, sg_PTNODE_TEMP_FLAG_REVERTING)  );
	}
	else					// revert everything
	{
		SG_ERR_CHECK(  sg_ptnode__maybe_mark_ptnode_for_revert__recursive(pCtx,
																		  pPendingTree->pSuperRoot,
																		  bRecursive,
																		  pInternalFilespec)  );
	}

	//////////////////////////////////////////////////////////////////
	// We process the structural differences in PASSES.  We have to do this because there
	// may be order-dependencies that we don't know (and don't want to know) about.
	// So we do things in passes so that we can eliminate the dependencies.
	//
	// For example, if they did something like:
	//     vv mv foo bar
	//     vv mv xxx foo
	// then we might have a problem depending the order that we choose
	// to revert them.  We don't know the order that the user performed the operations
	// on disk originally; all we know is the pendingtree/treediff of everything that
	// *is currently* different from the baseline.  Also, because our rbtrees are
	// keyed by gid-object (which are quite random), we might get different results
	// each time we are run.
	//
	// [1] Scan each difference and build an "inverted tree" that maps
	//         (directory-gid-object, entryname) --> { entry-gid-object_source, entry-gid-object_target }
	//         Based upon the type of difference in the treediff, claim the (directory-gid-object, entryname)
	//         in the SOURCE or TARGET version of the tree.
	//
	//         NOTE: Terminology: For REVERT, we call the current/dirty version the "SOURCE"
	//               and the desired clean/baseline version the "TARGET".
	//
	//         NOTE: The SG_treediff2 is a flat (1-level, non-hierarchial) list of changes.
	//               The "inverted tree (_inv_dirs, _inv_dir, _inv_entry)" is a flat (2-level,
	//               {dir, entry}) view of the structural changes.
	//               The pendingtree (SG_pendingtree, sg_ptnode) is a hierarchial view of the
	//               "interesting" portion of the tree (as it currently exists (SOURCE)).

	rd.pass = _PT_REVERT_PASS__SCAN_AND_BIND;
	SG_ERR_CHECK(  SG_treediff2__foreach(pCtx,rd.pTreeDiff,_pt_revert__process_structural_changes,&rd)  );

#if 0 && TRACE_REVERT
	SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,"between [1] and [2]\n")  );
	SG_ERR_IGNORE(  SG_inv_dirs_debug__dump_to_console(pCtx,rd.pInvDirs)  );
#endif

	// [2] Identify entrynames that are claimed by different entries in the 2 trees and create a "swap"
	//         pathname in the parking lot.
	//
	//         If there is a conflict that cannot be resolved, stop now.  This can happen, for example,
	//         during a PARTIAL REVERT if we are requested to undo foo-->bar but not xxx-->foo because
	//         there would be 2 foo's in the directory.  TODO see TODO labeled [4] in __check_for_swap_cb.
	//
	//         The __check_for_portability code is used here to detect POTENTIAL TRANSIENT COLLISIONS
	//         so that __check_for_swaps code can register temp/swap pathnames for both REAL and
	//         POTENTIAL TRANSIENT COLLISIONS.

	SG_ERR_CHECK(  SG_inv_dirs__check_for_portability(pCtx,rd.pInvDirs,rd.pPendingTree)  );
	SG_ERR_CHECK(  SG_inv_dirs__check_for_swaps(pCtx,rd.pInvDirs)  );

	// [3] See if there are any other types of conflicts due to a PARTIAL REVERT.  For example,
	//     if they did:
	//         vv mv @/dir1/dir2/dir3/file1 @/file1
	//         vv rm @/dir1
	//     and they only want to revert the mv and not the (recursive) rmdir.  Then we won't have
	//     the "dir3" directory in the WD to un-move "file1".
	//
	//     If there is a conflict such as the one above that cannot be resolved, stop now.

	rd.pass = _PT_REVERT_PASS__CHECK_FOR_CONFLICTS;
	SG_ERR_CHECK(  SG_treediff2__foreach(pCtx,rd.pTreeDiff,_pt_revert__process_structural_changes,&rd)  );

	// [4] We need to check for POTENTIAL NON-TRANSIENT COLLISIONS because of portability issues, do it now
	//     *before* we have made any changes to the WD.
	//
	//     For example, if they did something like:
	//         vv mv foo bar
	//         vv mv xxx FOO
	//     and then only want to revert foo --> bar
	//     then the REVERT will have a problem on a case-insensitive filesystem.
	//
	//     The __check_for_portability code in step [2] actually did 2 computations:
	//     [a] the potential transient collisions;
	//     [b] the potential non-transient (final) collisions.
	//     So all we need to do here is see if there were and of type [b] and complain.

	SG_ERR_CHECK(  SG_inv_dirs__get_portability_warnings_observed(pCtx, rd.pInvDirs, &flags_logged)  );
	if (flags_logged)
		SG_ERR_THROW(  SG_ERR_PORTABILITY_WARNINGS  );

	// [4.5] See if there are any submodules that we need to (re)create
	//       or dive into and revert too.  this suffers the same problems
	//       as a nested checkout/update.
	//       (1) if the user has an uncomitted delete of a submodule, then
	//           this revert needs to redo the CHECKOUT (or try to recover
	//           it from the graveyard (where we probably put it if they
	//           did the delete properly)).
	//       (2) if it wasn't deleted, but there is dirt within the submodule
	//           (and --no-submodules was NOT given) we need to dive in and
	//           REVERT the dirt *and* maybe do an UPDATE the to the originally
	//           referenced cset (so that the outer sees it as clean).  this
	//           update step could cause sub-sub modules to be re-created.
	//
	//       So, we have to be able to do a precompute check using the
	//       pvec_use_submodule hints and/or throw when we don't have
	//       sufficient infomation (like checkout/update do).

	// [4.5.1] Inspect DELETED and LOST submodules
	rd.pass = _PT_REVERT_PASS__PRECOMPUTE_SUBMODULE_STATE;
	SG_ERR_CHECK(  SG_treediff2__foreach(pCtx,rd.pTreeDiff,_pt_revert__process_structural_changes,&rd)  );

	// [4.5.2] Inspect MODIFIED submodules *IFF* we are supposed to dive into them.
	if (rd.bNoSubmodules == SG_FALSE)
		SG_ERR_CHECK(  SG_treediff2__foreach(pCtx,rd.pTreeDiff,_pt_revert__precompute_modified_submodules,&rd)  );

	// [4.5.9] If one of the submodules had issues, we report them and stop.
	SG_ERR_CHECK(  SG_varray__count(pCtx, rd.pvaMissingDetails, &nrMissingDetails)  );
	if (nrMissingDetails > 0)
		goto fail;

	// [4.9] If they said "precompute-only" we stop here.  This is a hack
	//       because I don't want to cut this function into __precompute
	//       and __apply parts and pass instance data around.  I'm not sure
	//       whether this style or the other style (used in checkout and update)
	//       is better or worse or just different.
	//
	//       NOTE That a precompute-only run DOES NOT update the LOG.  Here we
	//       just do enough to get the pvaMissingDetails populated.
	//
	// TODO 2011/04/04 I really need to refactor all of this, but I
	// TODO            don't have time (or brain cells) today.

	if (bPrecomputeOnly)
		goto fail;

	//////////////////////////////////////////////////////////////////
	// at this point we start making changes to the WD.
	//////////////////////////////////////////////////////////////////

	// TODO 2010/05/20 This function is too big and it is misleading.  Each of
	// TODO            functions that we call in [5]..[9] all have to run when
	// TODO            either __DO_IT or __LOG_IT -- even though it looks like
	// TODO            we are always going to make changes.  Convert the way we
	// TODO            build the log to use SG_wd_plan to build a plan and then
	// TODO            optionally execute it -- rather than intertwining them
	// TODO            as I have here.  Convert caller to expect a WD_Plan rather
	// TODO            than this LOG format.

	// [5] Use the hierarchial view to do a depth-first walk and move the entries that have a "parked/swap"
	//         pathname to their new/temporary location.  We have to do this depth-first because the
	//         repo-path --> absolute-path stuff is temporarily invalid while we have things are parked.

	SG_ERR_CHECK(  SG_inv_dirs__get_parking_stats(pCtx, rd.pInvDirs, &nrParked, NULL, NULL)  );
	if (nrParked > 0)
		SG_ERR_CHECK(  _pt_revert__move_to_parking_lot(pCtx,&rd,rd.pPendingTree->pSuperRoot,0)  );

#if TRACE_REVERT
	if (rd.pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
	{
		SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,"after __move_to_parking_lot: [nrParked %d]\n",nrParked)  );
		SG_ERR_IGNORE(  SG_inv_dirs_debug__dump_to_console(pCtx,rd.pInvDirs)  );
	}
#endif

	// [6] Use the hierarchial view to identify the directories that were deleted/lost and re-build
	//         the skeleton of WD.  We need to do this first in case there are things which were moved
	//         out of a directory before the directory was deleted.  This pass should just re-create
	//         the directories; we'll worry about attrbits/xattrs/etc later.
	//         We create the directories either in the WD-absolute-path or a transient-path under
	//         a parent in parked in swap.

	SG_ERR_CHECK(  _pt_revert__undelete_skeleton(pCtx,&rd,rd.pPendingTree->pSuperRoot,0)  );

	// [7] Use the treediff view to actually perform the structural changes.  This will reshape
	//         the WD by UNDO-ing all of the RENAMES, MOVES, DELETES, and ADDS.  This will operate
	//         within the WD and pull everything that was parked back into the tree.  We do this
	//         in 2 parts:
	//         [7a] RENAMES, MOVES (excluding cyclic directory MOVES), DELETES, and ADDS.
	//         [7b] cyclic directory MOVES (because the tree is kinda broken during this).
	//         In theory, when this step is finished, the parking lot should be be empty.

#if TRACE_REVERT
	if (rd.pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
	{
		SG_vhash * pvh = NULL;

		SG_ERR_IGNORE(  SG_console(pCtx,SG_CS_STDERR,
								   "SG_pendingtree__revert: after undelete_skeleton and before apply_structural_changes\n")  );
		pPendingTree->bDebugExport = SG_TRUE;
		SG_ERR_IGNORE(  sg_pendingtree__to_vhash(pCtx,pPendingTree,&pvh)  );
		pPendingTree->bDebugExport = SG_FALSE;
		SG_ERR_IGNORE(  SG_vhash_debug__dump_to_console(pCtx,pvh)  );
		SG_VHASH_NULLFREE(pCtx,pvh);
	}
#endif

	rd.pass = _PT_REVERT_PASS__APPLY_STRUCTURAL_CHANGES;
	SG_ERR_CHECK(  SG_treediff2__foreach(pCtx,rd.pTreeDiff,_pt_revert__process_structural_changes,&rd)  );

	if (rd.pVecCycles)
		SG_ERR_CHECK(  SG_vector__foreach(pCtx,rd.pVecCycles,_pt_revert__process_cyclic_directory_moves,&rd)  );

#if defined(DEBUG)
	if (rd.pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
		SG_ERR_IGNORE(  SG_inv_dirs_debug__verify_parking_lot_empty(pCtx,rd.pInvDirs)  );
#endif

	// [8] Use the treediff view to UNDO changes *to* entries.  That is, we fetch the original
	//         contents of MODIFIED files and symlinks.  We also UNDO any XATTR or ATTRBIT changes
	//         that were made.

	SG_ERR_CHECK(  SG_treediff2__foreach(pCtx,rd.pTreeDiff,_pt_revert__process_nonstructural_changes,&rd)  );

	// [8.5] See if there were any "disposable" directories (created by MERGE)
	//       that are now empty and that we could/should delete.

	if (rd.nrDisposableDirs > 0)
	{
		SG_bool bShouldBeDeleted;

		SG_ERR_CHECK(  SG_VECTOR__ALLOC(pCtx, &rd.pVecDisposablePathsToDelete, rd.nrDisposableDirs)  );
		SG_ERR_CHECK(  SG_VECTOR__ALLOC(pCtx, &rd.pVecDisposablePTNodesToDelete, rd.nrDisposableDirs)  );

		SG_ERR_CHECK(  _pt_revert__identify_disposable_dirs(pCtx, &rd, rd.pPendingTree->pSuperRoot, &bShouldBeDeleted)  );
		SG_ERR_CHECK(  _pt_revert__delete_queued_disposable_dirs(pCtx, &rd)  );
	}

	// [8.6] Deal with any submodules as late as possible (in case they
	//       throw something we didn't expect we will already have the
	//       WD almost ready).
	//
	//       TODO 2011/04/01 Should we write out the pendingtree before
	//       TODO            we do this --or-- capture all of this in a
	//       TODO            TRY BLOCK and let it continue ?

	if (rd.pvec_queued_subs)
		SG_ERR_CHECK(  _pt_revert__handle_queued_subs(pCtx, &rd)  );

done:
	// [9] All is done.  if __DO_IT, we need to cleanup the pendingtree
	//         in memory and save it to disk, so that it matches any
	//         remaining dirt in the working directory.  if not __DO_IT
	//         (and we only did a __LOG_IT), then we didn't modify the
	//         working directory, but have probably trashed the in
	//         memory pendingtree, so invalidate as much of it as we
	//         can without actually deleting the pendingtree pointer.

	if (rd.pPendingTree->eActionLog & SG_PT_ACTION__DO_IT)
	{
		if (nrParents > 1)
		{
			// we can only get "sticky" bits set by MERGE and
			// we only allow "REVERT --ALL" after a MERGE, so
			// if we get here we need to clear the sticky bits.
			// (because having them set keeps the pendingtree
			// from collapsing to the sparsest thing possible
			// (which after a revert --all should be an empty
			// tree)).

			SG_ERR_CHECK(  SG_pendingtree__hack__clear_all_sticky_bits(pCtx, rd.pPendingTree)  );
		}
		
		/* we need to undo all the IMPLICIT entries now.  if we don't,
		 * this revert function will accidentally perform an addremove
		 * function. */

		SG_ERR_CHECK(  sg_ptnode__undo_implicit(pCtx, pPendingTree->pSuperRoot)  );
		SG_ERR_CHECK(  sg_ptnode__unload_dir_entries_if_not_dirty(pCtx, pPendingTree->pSuperRoot, &bDirty)  );

		if (!(pPendingTree->pSuperRoot->prbItems))
			SG_PTNODE_NULLFREE(pCtx, pPendingTree->pSuperRoot);

		if (nrParents > 1)
		{
			const char * psz_hid_parent_0;

			// We were looking at an "uncommitted merge".
			//
			// Since we currently only allow "REVERT --ALL" after a MERGE,
			// we should always have an empty pendingtree *BUT* with the
			// new PINNED ptnodes with hid_disposable values, we may still
			// have some ptnodes, but they should all have 0 flags.
			//
			// Also because of the "REVERT --ALL" constraint, we should
			// reset the parents and clear the list of issues.  Later, when
			// we do allow PARTIAL REVERTS, we'll need to figure out which
			// or all of the parents to keep and which of the issues to
			// "CANCEL" (which is stronger than a "MARK RESOLVED").

			SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pva_wd_parents, 0, &psz_hid_parent_0)  );
			SG_ERR_CHECK(  SG_pendingtree__set_single_wd_parent(pCtx, pPendingTree, psz_hid_parent_0)  );

			SG_ERR_CHECK(  SG_pendingtree__clear_wd_issues(pCtx, pPendingTree)  );
			SG_ERR_CHECK(  SG_pendingtree__set_wd_choices(pCtx, pPendingTree, NULL)  );

			// Likewise, clear the MERGE-TEMP-DIR associated with the pending merge.
			// This is the temp dir that MERGE created and left laying around for
			// RESOLVE and COMMIT to use.  It has nothing to do with anything needed
			// by REVERT.

			SG_ERR_CHECK(  SG_pendingtree__get_mrg_temp_dir(pCtx, pPendingTree, &bHasMergeTempDir, &pPathMergeTempDir)  );
			if (bHasMergeTempDir)
				SG_ERR_CHECK(  SG_pendingtree__set_mrg_temp_dir(pCtx, pPendingTree, NULL)  );
		}

		SG_ERR_CHECK(  SG_pendingtree__save(pCtx, pPendingTree)  );

		if (bHasMergeTempDir)
		{
			// After we get the revert completely finalized (and don't care
			// so much about an error), we'll try to delete the merge-temp-dir
			// from the disk.  this is pretty much trash now.  if it fails,
			// it is no big deal.

			SG_ERR_IGNORE(  SG_fsobj__rmdir_recursive__pathname2(pCtx, pPathMergeTempDir, SG_TRUE)  );
			SG_PATHNAME_NULLFREE(pCtx, pPathMergeTempDir);
		}
	}
	else
	{
		SG_ASSERT(  (rd.pPendingTree->eActionLog & SG_PT_ACTION__LOG_IT)  );

		/* we don't change the saved pending changeset here. */
		SG_PTNODE_NULLFREE(pCtx, pPendingTree->pSuperRoot);
	}

fail:
	SG_VECTOR_NULLFREE_WITH_ASSOC(pCtx, rd.pvec_queued_subs, (SG_free_callback *)_queued_subs__free);
	SG_VECTOR_NULLFREE_WITH_ASSOC(pCtx, rd.pVecDisposablePathsToDelete, (SG_free_callback *)SG_pathname__free);
	SG_VECTOR_NULLFREE(pCtx, rd.pVecDisposablePTNodesToDelete);
	SG_TREEDIFF2_NULLFREE(pCtx, rd.pTreeDiff);
	SG_INV_DIRS_NULLFREE(pCtx, rd.pInvDirs);
	SG_VECTOR_NULLFREE(pCtx, rd.pVecCycles);
	SG_FILE_SPEC_NULLFREE(pCtx, pInternalFilespec);
	SG_PATHNAME_NULLFREE(pCtx, pPathMergeTempDir);
}

//////////////////////////////////////////////////////////////////

END_EXTERN_C;

#endif//H_SG_PENDINGTREE__PRIVATE_REVERT_H
