/*
Copyright 2010 SourceGear, LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include <sg.h>


/*
**
** Types
**
*/

/**
 * A single state value associated with a choice.
 * Values are generally created from the changesets involved in the merge.
 * Values are also created from the WD state value if necessary.
 * For states whose values are mergeable, values can also be created by merging two other values.
 * Values that are the result of the merge may be invalid (have no value), if the merge failed for some reason.
 */
struct _sg_resolve__value
{
	// identification
	SG_resolve__choice* pChoice;          //< The choice that owns this value.
	char*               szLabel;          //< The label/name for this value.

	// value
	// If all of these are NULL, see the "changeset" and "merge" sections for what it means in each case.
	SG_variant*         pVariant;         //< The actual data value.
	                                      //< Type of the variant depends on the state of the choice that owns the value (see below this struct definition).
	                                      //< For mergeable values, this is the HID of a blob.
	                                      //< NULL for values that don't have a variant/value (failed merges).
	SG_pathname*        pVariantFile;     //< Absolute path to a file that contains the actual data value.
	                                      //< Only used for mergeable values, where it contains the blob contents.
	                                      //< NULL for non-mergeable values.
	char*               szVariantSummary; //< More human-readable summary of the data stored in pVariant.
	                                      //< Used for values where the actual data is not easily comprehensible or available in a file.
	                                      //< NULL for values that don't require it (see below under Types for pVariant).

	// miscellaneous
	SG_uint32           uChildren;        //< Number of other values that list this one as a parent.

	// changeset
	// These are only used for values created from changesets.
	// The WD is considered a changeset, and uses these fields.
	// If a changeset value has all NULL variant data, then it's a WD value for a state other than EXISTENCE whose item doesn't exist in the WD.
	// In that case, szChangeset should be SG_RESOLVE__HID__WORKING, and szRepoPath should be NULL.
	const char*         szChangeset;      //< The HID of the changeset associated with this value.
	                                      //< Points at one of: a HID in the resolve's pParentHids, the item's szAncestorHid, or SG_RESOLVE__HID__WORKING.

	// merge
	// These are only used for values created by merging other values.
	// If a merge value has all NULL variant data, then it's the result of a failed merge.
	// Use iMergeToolResult to get more detail about the nature of the failure.
	SG_resolve__value*  pBaselineParent;  //< The baseline parent value that was merged into this one.
	                                      //< Points at one of the other values in the same choice.
	SG_resolve__value*  pOtherParent;     //< The other parent value that was merged into this one.
	                                      //< Points at one of the other values in the same choice.
	SG_bool             bAutoMerge;       //< If SG_TRUE, this value was generated automatically by merge.
	                                      //< If SG_FALSE, it was generated manually by the user with resolve.
	char*               szMergeTool;      //< The name of the merge tool used to create this node from its parents.
	SG_int32            iMergeToolResult; //< The status result when szMergeTool was run.
	                                      //< One of SG_MERGETOOL__STATUS__*
};
/*
 * Types for pVariant:
 *
 * EXISTENCE (bool) True if the item exists, false if it doesn't.
 * NAME (string) The name of the item.
 * LOCATION (string) The GID of the item's parent folder (szVariantSummary will contain the parent's repo path).
 * ATTRIBUTES (int) The item's attribute bits.
 * XATTRIBUTES (string) The HID of the blob containing the item's extended attributes, or a copy of SG_MRG_TOKEN__NO_XATTR if it has none.
 * CONTENTS/FILE (string) The HID of the blob containing the item's contents (pVariantFile will indicate an actual file with the contents).
 * CONTENTS/DIRECTORY (not possible)
 * CONTENTS/SYMLINK (string) The target of the symlink item.
 * CONTENTS/SUBMODULE (string) The HID of the blob containing the submodule's data (szVariantSummary will contain a summary of the submodule).
 */

/**
 * Data about a single choice on an item.
 * Each choice regards one particular state on the item.
 * So the combination of an item and state uniquely identify a choice.
 */
struct _sg_resolve__choice
{
	SG_resolve__item*                pItem;           //< The item that owns this choice.
	SG_resolve__state                eState;          //< The conflicting state that the choice is about on the item.
	SG_mrg_cset_entry_conflict_flags eConflictCauses; //< Conflict flag(s) that caused this choice to be created.
	SG_bool                          bCollisionCause; //< Whether or not a collision helped cause this choice.
	SG_pathname*                     pTempPath;       //< Absolute path where temporary files related to the choice are stored.
	                                                  //< Might be NULL for choices that don't require temporary files.
	SG_vector*                       pValues;         //< The set of values associated with the choice.
	                                                  //< This is ordered such that a value's parents always come BEFORE it in the list.
	                                                  //< Vector contains SG_resolve__value* values.
	SG_vector*                       pLeafValues;     //< List of the current leaf values after considering completed merges.
	                                                  //< NULL if the choice is non-mergeable (because it will always equal pValues).
	                                                  //< Vector contains SG_resolve__value* values.
	                                                  //< Points into pValues.
	SG_resolve__value*               pResult;         //< The resolved value for the choice.
	                                                  //< NULL if the choice is unresolved.
	                                                  //< Points into pValues.

	// only used when bCollisionCause == SG_TRUE
	// NULL otherwise
	SG_vector*                       pCollisionItems; //< List of other items that are involved in the collision causing this choice.
	                                                  //< Vector contains SG_resolve__item* values.
	                                                  //< Points to other items in the same resolve.

	// only used when eConflictCauses includes SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__MOVES_CAUSED_PATH_CYCLE
	// NULL otherwise
	SG_vector*                       pCycleItems;     //< List of other items that are involved in the cycle causing this choice.
	                                                  //< Vector contains SG_resolve__item* values.
	                                                  //< Points to other items in the same resolve.
	char*                            szCycleHint;     //< A hint provided by merge about the choice's cycle.
	                                                  //< Currently this is a cycling repo path that merge wanted to create, but couldn't.
};

/**
 * Data about a single item with choices on it.
 * The item is resolved if all of its choices have been resolved.
 */
struct _sg_resolve__item
{
	SG_resolve*            pResolve;                           //< The resolve that owns this item.
	char*                  szGid;                              //< The item's unique GID.
	SG_treenode_entry_type eType;                              //< The item's type.
	SG_diffstatus_flags    eStatus;                            //< The item's current WD status.
	char*                  szAncestorHid;                      //< HID of the ancestor changeset used to merge this item.
	                                                           //< NULL if the item has no common ancestor (only currently known to occur in add/add collisions).
	SG_vhash*              pRepoPaths;                         //< Maps changeset HIDs to the item's repo path in that changeset.
	                                                           //< Contains data for szAncestorHid, SG_RESOLVE__HID__WORKING, and all HIDs in resolve's pParentHids.
	                                                           //< If one of those HIDs doesn't exist in the hash, the item didn't exist in that changeset.
	char*                  szMergedName;                       //< Name of the item as it was left by merge.
	char*                  szMergedContentHid;                 //< HID of the item's content as it was left by merge.
	SG_int64               iMergedAttributes;                  //< Attributes of the item as it was left by merge.
	char*                  szMergedXAttrsHid;                  //< HID of the item's extended attributes as it was left by merge.
	char*                  szOriginalName;                     //< If the item was assigned a temporary name by merge,
	                                                           //< then this is the name we'll revert it back to once it's resolved.
	                                                           //< Will be NULL if the item had different names in different parent changesets.
	                                                           //< In that case, reverting back to an original name will be unnecessary,
	                                                           //< because the user will have to choose a name to resolve the NAME choice that it will have.
	char*                  szOriginalLocation;                 //< Same as szOriginalName, but for the original location.
	                                                           //< Unlike original name, this isn't necessary for reverting a temporary location.
	                                                           //< Instead, it's used to "unmove" an item that we need to delete, if the user moved it.
	                                                           //< The "unmove" is necessary because pendingtree won't delete something that's been moved.
	SG_resolve__choice*    aChoices[SG_RESOLVE__STATE__COUNT]; //< The set of choices on the item, indexed by the state in conflict.
	                                                           //< Indices for states that are not in conflict are NULL.
};

/**
 * All of the resolve data about a single working directory.
 * This is the basic data structure used by the resolve system.
 * All other resolve data is stored within it somewhere.
 */
struct _sg_resolve
{
	SG_pendingtree* pPendingTree; //< The pendingtree this resolve data was loaded from.
	SG_varray*      pParentHids;  //< HIDs of all parent changesets involved in the merge.
	                              //< The first HID is the baseline, per pendingtree convention.
	SG_vector*      pItems;       //< The set of items that are/were in conflict.
	                              //< Vector contains SG_resolve__item* values.
};


/*
**
** Globals
**
*/

/**
 * Maps state values to info about them.
 *
 * Note: The order of this array must match the order of the SG_resolve__state enumeration.
 */
static const struct
{
	const char* szName;
	const char* szDescription;
}
gaStateData[] =
{
	{ "existence", "Merge couldn't decide if the item should exist." },
	{ "name",      "Merge couldn't choose a name for the item." },
	{ "location",  "Merge couldn't choose a location for the item." },
	{ "attrs",     "Merge couldn't choose file system attributes for the item." },
	{ "xattrs",    "Merge couldn't choose extended attributes for the item." },
	{ "contents",  "Merge couldn't generate the item's contents." },
	{ "invalid",   "This is an invalid type of conflict." },
};

/**
 * Maps the causes of conflicts/choices to information about them.
 */
static const struct
{
	SG_mrg_cset_entry_conflict_flags eConflictCauses;
	SG_bool                          bCollisionCause;
	const char*                      szName;
	const char*                      szDescription;
}
gaCauseData[] =
{
	{ SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DELETE_VS_MOVE,            SG_FALSE, "Delete/Move",   "Item deleted in one branch and moved in another." },
	{ SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DELETE_VS_RENAME,          SG_FALSE, "Delete/Rename", "Item deleted in one branch and renamed in another." },
	{ SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DELETE_VS_ATTRBITS,        SG_FALSE, "Delete/Attr",   "Item deleted in one branch and attributes changed in another." },
	{ SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DELETE_VS_XATTR,           SG_FALSE, "Delete/Xattr",  "Item deleted in one branch and attributes changed in another." },
	{ SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DELETE_VS_SYMLINK_EDIT,    SG_FALSE, "Delete/Edit",   "Symlink deleted in one branch and changed in another." },
	{ SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DELETE_VS_SUBMODULE_EDIT,  SG_FALSE, "Delete/Edit",   "Submodule deleted in one branch and changed in another." },
	{ SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DELETE_VS_FILE_EDIT,       SG_FALSE, "Delete/Edit",   "File deleted in one branch and changed in another." },
	{ SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_MOVE,            SG_FALSE, "Move/Move",     "Item moved to different places in different branches." },
	{ SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_RENAME,          SG_FALSE, "Rename/Rename", "Item renamed to different names in different branches." },
	{ SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_ATTRBITS,        SG_FALSE, "Attr/Attr",     "Changes to item's attributes in different branches conflict." },
	{ SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_XATTR,           SG_FALSE, "Xattr/Xattr",   "Changes to item's attributes in different branches conflict." },
	{ SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_SYMLINK_EDIT,    SG_FALSE, "Edit/Edit",     "Changes to symlink's target in different branches conflict." },
	{ SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_SUBMODULE_EDIT,  SG_FALSE, "Edit/Edit",     "Changes to submodule's target in different branches conflict." },
	{ SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__MASK, SG_FALSE, "Edit/Edit",     "Changes to item's contents in different branches conflict." },
	{ SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DELETE_CAUSED_ORPHAN,      SG_FALSE, "Orphaned Item", "Item added to a folder that was deleted in another branch." },
	{ SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__MOVES_CAUSED_PATH_CYCLE,   SG_FALSE, "Path Cycle",    "Item moved to a folder that is its own child in another branch." },
	{ SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__ZERO,                      SG_TRUE,  "Collision",     "Item's name matches a different item's name in another branch." },
};

/**
 * Maps SG_resolve__state values to the conflict flags associated with them.
 */
static const SG_uint32 gaConflictFlagMasks[SG_RESOLVE__STATE__COUNT] =
{
	// SG_RESOLVE__STATE__EXISTENCE
	  SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DELETE_VS_MOVE
	| SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DELETE_VS_RENAME
	| SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DELETE_VS_ATTRBITS
	| SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DELETE_VS_XATTR
	| SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DELETE_VS_SYMLINK_EDIT
	| SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DELETE_VS_SUBMODULE_EDIT
	| SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DELETE_VS_FILE_EDIT
	| SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DELETE_CAUSED_ORPHAN,

	// SG_RESOLVE__STATE__NAME
	  SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_RENAME,

	// SG_RESOLVE__STATE__LOCATION
	  SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_MOVE
	| SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__MOVES_CAUSED_PATH_CYCLE,

	// SG_RESOLVE__STATE__ATTRIBUTES
	  SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_ATTRBITS,

	// SG_RESOLVE__STATE__XATTRIBUTES
	  SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_XATTR,

	// SG_RESOLVE__STATE__CONTENTS
	  SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__MASK
	| SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_SYMLINK_EDIT
	| SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_SUBMODULE_EDIT
};

/**
 * Maps SG_resolve__state values to the collision flags associated with them.
 */
static const SG_bool gaCollisionMasks[SG_RESOLVE__STATE__COUNT] =
{
	// SG_RESOLVE__STATE__EXISTENCE
	SG_TRUE,

	// SG_RESOLVE__STATE__NAME
	SG_TRUE,

	// SG_RESOLVE__STATE__LOCATION
	SG_TRUE,

	// SG_RESOLVE__STATE__ATTRIBUTES
	SG_FALSE,

	// SG_RESOLVE__STATE__XATTRIBUTES
	SG_FALSE,

	// SG_RESOLVE__STATE__CONTENTS
	SG_FALSE,
};

/**
 * Maps SG_treenode_entry_type values to the conflict flags denoting
 * conflicts that are mergeable for that type.
 */
static const SG_uint32 gaMergeableMasks[SG_RESOLVE__STATE__COUNT] =
{
	// none
	0,

	// SG_TREENODEENTRY_TYPE_REGULAR_FILE
	SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__MASK,

	// SG_TREENODEENTRY_TYPE_DIRECTORY
	0,

	// SG_TREENODEENTRY_TYPE_SYMLINK
	0,

	// SG_TREENODEENTRY_TYPE_SUBMODULE
	0,
};

/**
 * Maps SG_resolve__state values to the values we use to store them on disk.
 *
 * This allows the enumeration values to change without invalidating any data
 * that is stored in wd.json files, as long as this array is updated to match
 * the new values.
 */
static const char* gaSerializedStateValues[SG_RESOLVE__STATE__COUNT] =
{
	"existence",
	"name",
	"location",
	"attributes",
	"xattributes",
	"contents",
};

/*
 * Well known value labels.
 *
 * Note: Change these with care, since they're stored on disk in wd.json files.
 *       A new version loading a wd.json with old values will cause it to not
 *       be able to find values.
 */
const char* SG_RESOLVE__LABEL__ANCESTOR = "ancestor";
const char* SG_RESOLVE__LABEL__BASELINE = "baseline";
const char* SG_RESOLVE__LABEL__OTHER    = "other";
const char* SG_RESOLVE__LABEL__WORKING  = "working";
const char* SG_RESOLVE__LABEL__MERGE    = "merge";

/**
 * String used as the HID of the working directory / pendingtree.
 */
const char* SG_RESOLVE__HID__WORKING = "";

/**
 * Format string used to create actual labels from base strings and a counter.
 */
static const char* gszIndexedLabelFormat = "%s%d";

/**
 * Format strings for creating temporary filenames for values.
 */
static const char* gszFilenameFormat_Merge   = "%s~M%u~";
static const char* gszFilenameFormat_Working = "%s~W%u~";

/*
 * Keys used in wd.json to store merge information.
 */
static const char* gszIssueKey_Gid                       = "gid";
static const char* gszIssueKey_Type                      = "tne_type";
static const char* gszIssueKey_ConflictFlags             = "conflict_flags";
static const char* gszIssueKey_Collision                 = "collision_flags";
static const char* gszIssueKey_TempPath                  = "repopath_tempdir";
static const char* gszIssueKey_Inputs                    = "input";
static const char* gszIssueKey_Inputs_RepoPath           = "repopath";
static const char* gszIssueKey_Inputs_Name               = "entryname";
static const char* gszIssueKey_Inputs_Location           = "gid_parent";
static const char* gszIssueKey_Inputs_Attributes         = "attrbits";
static const char* gszIssueKey_Inputs_XAttributes        = "hid_xattrs";
static const char* gszIssueKey_Inputs_Contents           = "hid";
static const char* gszIssueKey_Output                    = "output";
static const char* gszIssueKey_Output_Name               = "entryname";
static const char* gszIssueKey_Output_Attributes         = "attrbits";
static const char* gszIssueKey_Output_XAttributes        = "hid_xattrs";
static const char* gszIssueKey_Output_Contents           = "hid";
static const char* gszIssueKey_File_AncestorTempRepoPath = "repopath_tempfile_ancestor";
static const char* gszIssueKey_File_BaselineTempRepoPath = "repopath_tempfile_baseline";
static const char* gszIssueKey_File_OtherTempRepoPath    = "repopath_tempfile_other";
static const char* gszIssueKey_File_MergeTempRepoPath    = "repopath_tempfile_result";
static const char* gszIssueKey_File_AutomergeTool        = "automerge_mergetool";
static const char* gszIssueKey_Collision_Items           = "collisions";
static const char* gszIssueKey_Cycle_Items               = "conflict_path_cycle_others";
static const char* gszIssueKey_Cycle_Hint                = "conflict_path_cycle_hint";

/*
 * Keys used in wd.json to store resolve information.
 */
static const char* gszChoiceKey_Result    = "result";
static const char* gszChoiceKey_Values    = "values";
static const char* gszValueKey_Label      = "label";
static const char* gszValueKey_Value      = "value";
static const char* gszValueKey_ValueFile  = "valuefile";
static const char* gszValueKey_Changeset  = "changeset";
static const char* gszValueKey_Baseline   = "baseline";
static const char* gszValueKey_Other      = "other";
static const char* gszValueKey_Tool       = "tool";
static const char* gszValueKey_ToolResult = "tool_result";


/*
**
** Internal Functions - Helpers
**
*/

/**
 * Allocates a variant and sets its value to a bool.
 */
static void _variant__alloc__bool(
	SG_context*  pCtx,      //< [in] [out] Error and context info.
	SG_variant** ppVariant, //< [out] The allocated variant.
	SG_bool      bValue     //< [in] The value to set to the variant.
	)
{
	SG_variant* pVariant = NULL;

	SG_NULLARGCHECK(ppVariant);

	SG_ERR_CHECK(  SG_alloc1(pCtx, pVariant)  );

	pVariant->type = SG_VARIANT_TYPE_BOOL;
	pVariant->v.val_bool = bValue;

	*ppVariant = pVariant;
	pVariant = NULL;

fail:
	SG_VARIANT_NULLFREE(pCtx, pVariant);
	return;
}

/**
 * Allocates a variant and sets its value to an int.
 */
static void _variant__alloc__int64(
	SG_context*  pCtx,      //< [in] [out] Error and context info.
	SG_variant** ppVariant, //< [out] The allocated variant.
	SG_int64     iValue     //< [in] The value to set to the variant.
	)
{
	SG_variant* pVariant = NULL;

	SG_NULLARGCHECK(ppVariant);

	SG_ERR_CHECK(  SG_alloc1(pCtx, pVariant)  );

	pVariant->type = SG_VARIANT_TYPE_INT64;
	pVariant->v.val_int64 = iValue;

	*ppVariant = pVariant;
	pVariant = NULL;

fail:
	SG_VARIANT_NULLFREE(pCtx, pVariant);
	return;
}

/**
 * Allocates a variant and sets its value to a copy of a string.
 */
static void _variant__alloc__sz(
	SG_context*  pCtx,      //< [in] [out] Error and context info.
	SG_variant** ppVariant, //< [out] The allocated variant.
	const char*  szValue    //< [in] The string to copy and set to the variant.
	)
{
	SG_variant* pVariant  = NULL;
	char*       szMyValue = NULL;

	SG_NULLARGCHECK(ppVariant);
	SG_NULLARGCHECK(szValue);

	SG_ERR_CHECK(  SG_strdup(pCtx, szValue, &szMyValue)  );
	SG_ERR_CHECK(  SG_alloc1(pCtx, pVariant)  );

	pVariant->type = SG_VARIANT_TYPE_SZ;
	pVariant->v.val_sz = szMyValue;
	szMyValue = NULL;

	*ppVariant = pVariant;
	pVariant = NULL;

fail:
	SG_VARIANT_NULLFREE(pCtx, pVariant);
	SG_NULLFREE(pCtx, szMyValue);
	return;
}

/**
 * Allocates a variant and sets its value to the contents of a blob.
 */
static void _variant__alloc__blob(
	SG_context*  pCtx,      //< [in] [out] Error and context info.
	SG_variant** ppVariant, //< [out] The allocated variant.
	SG_repo*     pRepo,     //< [in] The repo to look up the blob in.
	const char*  szHid      //< [in] The HID of the blob to look up.
	)
{
	SG_byte*    pBuffer   = NULL;
	SG_uint64   uSize     = 0u;
	SG_variant* pVariant  = NULL;
	SG_string*  sContents = NULL;

	SG_NULLARGCHECK(ppVariant);
	SG_NULLARGCHECK(pRepo);
	SG_NULLARGCHECK(szHid);

	// retrieve the HID contents from the repo
	SG_ERR_CHECK(  SG_repo__fetch_blob_into_memory(pCtx, pRepo, szHid, &pBuffer, &uSize)  );
	SG_ERR_CHECK(  SG_STRING__ALLOC__BUF_LEN(pCtx, &sContents, pBuffer, (SG_uint32)uSize)  );

	// allocate the variant
	SG_ERR_CHECK(  SG_alloc1(pCtx, pVariant)  );
	pVariant->type = SG_VARIANT_TYPE_SZ;
	SG_ERR_CHECK(  SG_string__sizzle(pCtx, &sContents, (SG_byte**)&pVariant->v.val_sz, NULL)  );

	// return the variant
	*ppVariant = pVariant;
	pVariant = NULL;

fail:
	SG_NULLFREE(pCtx, pBuffer);
	SG_VARIANT_NULLFREE(pCtx, pVariant);
	SG_STRING_NULLFREE(pCtx, sContents);
	return;
}

/**
 * Loads a given submodule and retrieves a summary description of it.
 */
static void _submodule__load_summary(
	SG_context*  pCtx,     //< [in] [out] Error and context info.
	const char*  szHid,    //< [in] The HID of the submodule to get a summary of.
	SG_repo*     pRepo,    //< [in] The repo to look up the submodule in.
	char**       ppSummary //< [out] A summary of the submodule.
	)
{
	SG_treenode_submodule* pSubmodule = NULL;

	SG_NULLARGCHECK(szHid);
	SG_NULLARGCHECK(pRepo);
	SG_NULLARGCHECK(ppSummary);

	// retrieve the HID contents from the repo as a submodule
	SG_ERR_CHECK(  SG_treenode_submodule__load_from_repo(pCtx, pRepo, szHid, &pSubmodule)  );

	// get a summary of the submodule
	SG_ERR_CHECK(  SG_treenode_submodule__describe_for_status(pCtx, pSubmodule, ppSummary)  );

fail:
	SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pSubmodule);
	return;
}

/**
 * Generates an indexed label using the next available index.
 */
static void _generate_indexed_label(
	SG_context*         pCtx,    //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice, //< [in] The choice that will own the value we're generating a label for.
	const char*         szBase,  //< [in] The base label name to add the next available index to.
	SG_string**         ppLabel  //< [out] The generated label, using the next available index.
	)
{
	SG_string*         sLabel     = NULL;
	SG_resolve__value* pTempValue = NULL;
	SG_uint32          uIndex     = 1u;
	SG_bool            bUsed      = SG_TRUE;

	// keep looking up values with the given base label and an ever increasing index
	// until we find an index that's not yet in use on the given choice
	while (bUsed == SG_TRUE)
	{
		// safe to cast away the choice's const here, because:
		// 1) we know that this function won't alter it
		// 2) it's only const to begin with to explicitly indicate an input instead of an output
		SG_ERR_CHECK(  SG_resolve__choice__check_value__label(pCtx, pChoice, szBase, uIndex, &pTempValue)  );
		if (pTempValue == NULL)
		{
			bUsed = SG_FALSE;
		}
		else
		{
			uIndex += 1u;
		}
	}

	// create a label from the given base and next available index we found
	if (uIndex <= 1u)
	{
		// labels with index 1 should actually not have a numeric suffix
		// This makes for a friendlier user experience in the 99% of cases
		// where there is only one value with any given base name.
		SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &sLabel, szBase)  );
	}
	else
	{
		// labels with a higher index need to have the index appended as a suffix
		SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &sLabel)  );
		SG_ERR_CHECK(  SG_string__sprintf(pCtx, sLabel, gszIndexedLabelFormat, szBase, uIndex)  );
	}

	// return the generated label
	*ppLabel = sLabel;
	sLabel = NULL;

fail:
	SG_STRING_NULLFREE(pCtx, sLabel);
	return;
}

/**
 * Generates a temporary filename for a mergeable value.
 * It is assumed that this file will be in its choice's temporary path.
 */
static void _generate_temp_filename(
	SG_context*         pCtx,     //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice,  //< [in] The choice that owns the value we're generating a filename for.
	const char*         szFormat, //< [in] Format string to use for the filename.
	                              //<      Should contain a %s (the label) and a %u (a counter), in that order.
	                              //<      See gszFilenameFormat_*
	const char*         szLabel,  //< [in] The label for the value we're generating a filename for.
	SG_pathname**       ppPath    //< [out] Full pathname including generated temporary filename.
	)
{
	SG_string*   pFilename = NULL;
	SG_pathname* pPath     = NULL;
	SG_uint32    uCount    = 0u;
	SG_bool      bUsed     = SG_TRUE;

	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(szLabel);

	// keep generating filenames with an ever increasing counter
	// until we find one that doesn't exist yet
	SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &pFilename)  );
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC(pCtx, &pPath)  );
	while (bUsed == SG_TRUE)
	{
		SG_ERR_CHECK(  SG_string__sprintf(pCtx, pFilename, szFormat, szLabel, uCount)  );
		SG_ERR_CHECK(  SG_pathname__set__from_pathname(pCtx, pPath, pChoice->pTempPath)  );
		SG_ERR_CHECK(  SG_pathname__append__from_string(pCtx, pPath, pFilename)  );
		SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pPath, &bUsed, NULL, NULL)  );
		uCount += 1u;
	}

	if (ppPath != NULL)
	{
		*ppPath = pPath;
		pPath = NULL;
	}

fail:
	SG_PATHNAME_NULLFREE(pCtx, pPath);
	SG_STRING_NULLFREE(pCtx, pFilename);
	return;
}

/**
 * Gets the absolute path of an item by GID from a pendingtree.
 * This just combines getting the repo path with converting it to an absolute path.
 */
static void _pendingtree__get_absolute_path(
	SG_context*     pCtx,         //< [in] [out] Error and context info.
	SG_pendingtree* pPendingTree, //< [in] The pendingtree to look up the GID in.
	const char*     szGid,        //< [in] The GID to look up.
	SG_pathname**   ppPath        //< [out] Absolute path to the identified item.
	)
{
	SG_string* sRepoPath = NULL;

	SG_ERR_CHECK(  SG_pendingtree__get_repo_path(pCtx, pPendingTree, szGid, &sRepoPath)  );
	SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path2(pCtx, pPendingTree, SG_string__sz(sRepoPath), ppPath)  );

fail:
	SG_STRING_NULLFREE(pCtx, sRepoPath);
	return;
}

/**
 * Gets a repo path by GID from a pendingtree,
 * and returns it as a char* rather than an SG_string*.
 */
static void _pendingtree__get_repo_path(
	SG_context*     pCtx,         //< [in] [out] Error and context info.
	SG_pendingtree* pPendingTree, //< [in] The pendingtree to look up the GID in.
	const char*     szGid,        //< [in] The GID to look up.
	char**          ppRepoPath    //< [out] Repo path to the identified item.
	)
{
	SG_string* sRepoPath = NULL;

	SG_ERR_CHECK(  SG_pendingtree__get_repo_path(pCtx, pPendingTree, szGid, &sRepoPath)  );
	SG_ERR_CHECK(  SG_string__sizzle(pCtx, &sRepoPath, (SG_byte**)ppRepoPath, NULL)  );

fail:
	SG_STRING_NULLFREE(pCtx, sRepoPath);
	return;
}

/**
 * Renames an item to a unique backup name.
 */
static void _pendingtree__rename_temp(
	SG_context*        pCtx,         //< [in] [out] Error and context info.
	SG_pendingtree*    pPendingTree, //< [in] The pendingtree to rename an item in.
	const SG_pathname* pPath,        //< [in] Absolute path of the item to rename.
	SG_pathname**      ppTempPath    //< [out] Absolute path to the renamed item.
	)
{
	SG_pathname* pTempPath = NULL;
	SG_string*   sTempName = NULL;

	SG_ERR_CHECK(  SG_workingdir__calc_backup_name(pCtx, pPath, &pTempPath)  );
	SG_ERR_CHECK(  SG_pathname__get_last(pCtx, pTempPath, &sTempName)  );
	SG_ERR_CHECK(  SG_pendingtree__rename_dont_save_pendingtree(pCtx, pPendingTree, pPath, SG_string__sz(sTempName), SG_FALSE)  );

	*ppTempPath = pTempPath;
	pTempPath = NULL;

fail:
	SG_PATHNAME_NULLFREE(pCtx, pTempPath);
	SG_STRING_NULLFREE(pCtx, sTempName);
}

/**
 * This function is a total hack to make it sort of possible to have pendingtree
 * delete a moved and/or renamed item.  Pendingtree won't allow that, so this
 * function "unmoves" and/or "unrenames" the item back to its original location
 * and then deletes it, which pendingtree will allow.
 */
static void _pendingtree__revert_delete(
	SG_context*        pCtx,              //< [in] [out] Error and context info.
	SG_pendingtree*    pPendingTree,      //< [in] The pendingtree to remove an item from.
	const SG_pathname* pPath,             //< [in] The absolute path of the item to remove.
	const char*        szOriginalName,    //< [in] The name the item had in the pendingtree's baseline.
	const char*        szOriginalLocation //< [in] The GID of the item's parent in the pendingtree's baseline.
	)
{
	SG_pathname* pCurrentLocation  = NULL;
	SG_string*   sCurrentName      = NULL;
	SG_pathname* pOriginalPath     = NULL;
	SG_bool      bExists           = SG_FALSE;
	SG_pathname* pUnmovePath       = NULL;
	SG_pathname* pUnmoveTempPath   = NULL;
	SG_string*   sUnmoveTempName   = NULL;
	SG_pathname* pUnrenamePath     = NULL;
	SG_pathname* pUnrenameTempPath = NULL;
	SG_string*   sUnrenameTempName = NULL;
	const char*  szRevertedPath    = NULL;

	// figure out the item's current location and name
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__COPY(pCtx, &pCurrentLocation, pPath)  );
	SG_ERR_CHECK(  SG_pathname__get_last(pCtx, pCurrentLocation, &sCurrentName)  );
	SG_ERR_CHECK(  SG_pathname__remove_last(pCtx, pCurrentLocation)  );

	// convert the item's original location from GID to absolute path
	SG_ERR_CHECK(  _pendingtree__get_absolute_path(pCtx, pPendingTree, szOriginalLocation, &pOriginalPath)  );

	// if the current location isn't the original one, unmove the item
	if (SG_stricmp(SG_pathname__sz(pCurrentLocation), SG_pathname__sz(pOriginalPath)) != 0)
	{
		// if there's already a file at the target, give it a temporary name
		SG_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx, &pUnmovePath, pOriginalPath, SG_string__sz(sCurrentName))  );
		SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pUnmovePath, &bExists, NULL, NULL)  );
		if (bExists == SG_TRUE)
		{
			SG_ERR_CHECK(  _pendingtree__rename_temp(pCtx, pPendingTree, pUnmovePath, &pUnmoveTempPath)  );
		}

		// unmove the item
		SG_ERR_CHECK(  SG_pendingtree__move_dont_save_pendingtree(pCtx, pPendingTree, pPath, pOriginalPath, SG_FALSE)  );
		SG_ERR_CHECK(  SG_pathname__set__from_pathname(pCtx, pCurrentLocation, pOriginalPath)  );
	}

	// if the current name isn't the original one, unrename the item
	if (SG_stricmp(SG_string__sz(sCurrentName), szOriginalName) != 0)
	{
		// if there's already a file at the target, give it a temporary name
		SG_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx, &pUnrenamePath, pCurrentLocation, szOriginalName)  );
		SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pUnrenamePath, &bExists, NULL, NULL)  );
		if (bExists == SG_TRUE)
		{
			SG_ERR_CHECK(  _pendingtree__rename_temp(pCtx, pPendingTree, pUnrenamePath, &pUnrenameTempPath)  );
		}

		// unrename the item
		SG_ERR_CHECK(  SG_pathname__append__from_string(pCtx, pCurrentLocation, sCurrentName)  );
		SG_ERR_CHECK(  SG_pendingtree__rename_dont_save_pendingtree(pCtx, pPendingTree, pCurrentLocation, szOriginalName, SG_FALSE)  );
		SG_ERR_CHECK(  SG_string__set__sz(pCtx, sCurrentName, szOriginalName)  );
		SG_ERR_CHECK(  SG_pathname__remove_last(pCtx, pCurrentLocation)  );
	}

	// if we had to rename an existing file to unmove the item, now we can replace it
	if (pUnmoveTempPath != NULL)
	{
		SG_ERR_CHECK(  SG_pathname__get_last(pCtx, pUnmovePath, &sUnmoveTempName)  );
		SG_ERR_CHECK(  SG_pendingtree__rename_dont_save_pendingtree(pCtx, pPendingTree, pUnmoveTempPath, SG_string__sz(sUnmoveTempName), SG_FALSE)  );
	}

	// try removing the item again now that it's back to its original location and name
	SG_ERR_CHECK(  SG_pathname__append__from_string(pCtx, pCurrentLocation, sCurrentName)  );
	szRevertedPath = SG_pathname__sz(pCurrentLocation);
	SG_ERR_CHECK(  SG_pendingtree__remove_dont_save_pendingtree(pCtx, pPendingTree, 1u, &szRevertedPath, gpBlankFilespec, SG_FALSE, SG_TRUE, SG_TRUE)  );
	szRevertedPath = NULL;
	SG_ERR_CHECK(  SG_pathname__remove_last(pCtx, pCurrentLocation)  );

	// if we had to rename an existing file to unrename the item, now we can replace it
	if (pUnrenameTempPath != NULL)
	{
		SG_ERR_CHECK(  SG_pathname__get_last(pCtx, pUnrenamePath, &sUnrenameTempName)  );
		SG_ERR_CHECK(  SG_pendingtree__rename_dont_save_pendingtree(pCtx, pPendingTree, pUnrenameTempPath, SG_string__sz(sUnrenameTempName), SG_FALSE)  );
	}

fail:
	SG_PATHNAME_NULLFREE(pCtx, pCurrentLocation);
	SG_STRING_NULLFREE(pCtx, sCurrentName);
	SG_PATHNAME_NULLFREE(pCtx, pOriginalPath);
	SG_PATHNAME_NULLFREE(pCtx, pUnmovePath);
	SG_PATHNAME_NULLFREE(pCtx, pUnmoveTempPath);
	SG_STRING_NULLFREE(pCtx, sUnmoveTempName);
	SG_PATHNAME_NULLFREE(pCtx, pUnrenamePath);
	SG_PATHNAME_NULLFREE(pCtx, pUnrenameTempPath);
	SG_STRING_NULLFREE(pCtx, sUnrenameTempName);
	return;
}


/*
**
** Internal Functions - Value
**
*/

/**
 * Frees a value.
 */
static void _value__free(
	SG_context*        pCtx,  //< [in] [out] Error and context info.
	SG_resolve__value* pValue //< [in] The value to free.
	)
{
	if (pValue == NULL)
	{
		return;
	}

	SG_NULLFREE(pCtx, pValue->szLabel);
	SG_VARIANT_NULLFREE(pCtx, pValue->pVariant);
	SG_PATHNAME_NULLFREE(pCtx, pValue->pVariantFile);
	SG_NULLFREE(pCtx, pValue->szVariantSummary);
	SG_NULLFREE(pCtx, pValue->szMergeTool);
	pValue->pChoice          = NULL;
	pValue->uChildren        = 0u;
	pValue->szChangeset      = NULL;
	pValue->pBaselineParent  = NULL;
	pValue->pOtherParent     = NULL;
	pValue->bAutoMerge       = SG_FALSE;
	pValue->iMergeToolResult = 0;
	SG_NULLFREE(pCtx, pValue);
}

/**
 * NULLFREE wrapper around _value__free.
 */
#define _VALUE__NULLFREE(pCtx, pValue) _sg_generic_nullfree(pCtx, pValue, _value__free)

/**
 * Compares two values for equality of their variant data.
 */
static void _value__equal_variants(
	SG_context*        pCtx,   //< [in] [out] Error and context info.
	SG_resolve__value* pLeft,  //< [in] Left-hand value to compare.
	SG_resolve__value* pRight, //< [in] Right-hand value to compare.
	SG_bool*           pEqual  //< [out] True if the values have equivalent variant data, false otherwise.
	)
{
	SG_bool bEqual = SG_TRUE;

	SG_NULLARGCHECK(pLeft);
	SG_NULLARGCHECK(pRight);
	SG_NULLARGCHECK(pEqual);

	if (
		   (pLeft->pVariant == NULL || pRight->pVariant == NULL)
		&& (pLeft->pVariant != pRight->pVariant)
		)
	{
		// one variant is NULL and the other isn't
		bEqual = SG_FALSE;
	}
	else
	{
		// both variants are non-NULL, compare them
		SG_ERR_CHECK(  SG_variant__equal(pCtx, pLeft->pVariant, pRight->pVariant, &bEqual)  );
	}

	// return whatever we ended up with
	*pEqual = bEqual;

fail:
	return;
}

/**
 * Allocates a blank value structure.
 */
static void _value__alloc__blank(
	SG_context*         pCtx,    //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice, //< [in] The choice that will own the new value.
	SG_resolve__value** ppValue  //< [out] The allocated value.
	)
{
	SG_resolve__value* pValue = NULL;

	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(ppValue);

	SG_ERR_CHECK(  SG_alloc1(pCtx, pValue)  );

	pValue->pChoice          = pChoice;
	pValue->szLabel          = NULL;
	pValue->pVariant         = NULL;
	pValue->pVariantFile     = NULL;
	pValue->szVariantSummary = NULL;
	pValue->uChildren        = 0u;
	pValue->szChangeset      = NULL;
	pValue->pBaselineParent  = NULL;
	pValue->pOtherParent     = NULL;
	pValue->bAutoMerge       = SG_FALSE;
	pValue->szMergeTool      = NULL;
	pValue->iMergeToolResult = 0;

	*ppValue = pValue;
	pValue = NULL;

fail:
	_VALUE__NULLFREE(pCtx, pValue);
}

/**
 * Allocates a value structure that is associated with a changeset.
 */
static void _value__alloc__changeset(
	SG_context*         pCtx,           //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice,        //< [in] The choice that will own the value.
	const char*         szLabel,        //< [in] The label to assign to the value.
	                                    //<      The value will create its own copy of this.
	const SG_vhash*     pInputs,        //< [in] The inputs data from the merge issue to look up variant values in.
	const char*         szTempRepoPath, //< [in] Repo path to the temporary file to associate with the value.
	                                    //<      The file usually contains the full value of what's in pVariant,
	                                    //<      like the contents of a file whose HID is in pVariant.
	                                    //<      Pass NULL if the value has no associated temporary file.
	const char*         szChangeset,    //< [in] The changeset to associate with the value.
	                                    //<      The value stores this pointer directly, but does not assume ownership.
	SG_resolve__value** ppValue         //< [out] The allocated value.
	)
{
	SG_resolve__value* pValue = NULL;

	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(szLabel);
	SG_NULLARGCHECK(pInputs);
	SG_NULLARGCHECK(szChangeset);
	SG_NULLARGCHECK(ppValue);

	// allocate a blank value
	SG_ERR_CHECK(  _value__alloc__blank(pCtx, pChoice, &pValue)  );

	// we need to handle EXISTENCE choices somewhat separately
	if (pChoice->eState == SG_RESOLVE__STATE__EXISTENCE)
	{
		// the existence value is just whether or not we have input data for this changeset
		SG_bool bValue = SG_FALSE;
		SG_ERR_CHECK(  SG_vhash__has(pCtx, pInputs, szChangeset, &bValue)  );
		SG_ERR_CHECK(  _variant__alloc__bool(pCtx, &pValue->pVariant, bValue)  );
	}
	else
	{
		SG_vhash* pInput = NULL;

		// check if we have input data for this changeset
		SG_ERR_CHECK(  SG_vhash__check__vhash(pCtx, pInputs, szChangeset, &pInput)  );
		if (pInput == NULL)
		{
			// no input data, so there's no value to create
			_VALUE__NULLFREE(pCtx, pValue);
		}
		else
		{
			// we found some input data, go ahead and parse our variant data out of it
			switch (pChoice->eState)
			{
			case SG_RESOLVE__STATE__NAME:
				{
					const char* szValue = NULL;
					SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pInput, gszIssueKey_Inputs_Name, &szValue)  );
					SG_ERR_CHECK(  _variant__alloc__sz(pCtx, &pValue->pVariant, szValue)  );
					break;
				}
			case SG_RESOLVE__STATE__LOCATION:
				{
					const char* szValue = NULL;
					SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pInput, gszIssueKey_Inputs_Location, &szValue)  );
					SG_ERR_CHECK(  _variant__alloc__sz(pCtx, &pValue->pVariant, szValue)  );
					SG_ERR_CHECK(  _pendingtree__get_repo_path(pCtx, pChoice->pItem->pResolve->pPendingTree, pValue->pVariant->v.val_sz, &pValue->szVariantSummary)  );
					break;
				}
			case SG_RESOLVE__STATE__ATTRIBUTES:
				{
					SG_int64  iValue = 0;
					SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pInput, gszIssueKey_Inputs_Attributes, &iValue)  );
					SG_ERR_CHECK(  _variant__alloc__int64(pCtx, &pValue->pVariant, iValue)  );
					break;
				}
			case SG_RESOLVE__STATE__XATTRIBUTES:
				{
					const char* szValue = NULL;
					SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pInput, gszIssueKey_Inputs_XAttributes, &szValue)  );
					SG_ERR_CHECK(  _variant__alloc__sz(pCtx, &pValue->pVariant, szValue)  );
					break;
				}
			case SG_RESOLVE__STATE__CONTENTS:
				{
					switch (pChoice->pItem->eType)
					{
					case SG_TREENODEENTRY_TYPE_REGULAR_FILE:
						{
							const char* szValue = NULL;
							SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pInput, gszIssueKey_Inputs_Contents, &szValue)  );
							SG_ERR_CHECK(  _variant__alloc__sz(pCtx, &pValue->pVariant, szValue)  );
							break;
						}
					case SG_TREENODEENTRY_TYPE_DIRECTORY:
						{
							SG_ERR_THROW2(SG_ERR_NOTIMPLEMENTED, (pCtx, "Cannot resolve a directory contents conflict because they should never occur."));
							break;
						}
					case SG_TREENODEENTRY_TYPE_SYMLINK:
						{
							SG_repo*    pRepo   = NULL;
							const char* szValue = NULL;
							SG_ERR_CHECK(  SG_pendingtree__get_repo(pCtx, pChoice->pItem->pResolve->pPendingTree, &pRepo)  );
							SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pInput, gszIssueKey_Inputs_Contents, &szValue)  );
							SG_ERR_CHECK(  _variant__alloc__blob(pCtx, &pValue->pVariant, pRepo, szValue)  );
							break;
						}
					case SG_TREENODEENTRY_TYPE_SUBMODULE:
						{
							const char* szValue = NULL;
							SG_repo*    pRepo   = NULL;
							SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pInput, gszIssueKey_Inputs_Contents, &szValue)  );
							SG_ERR_CHECK(  _variant__alloc__sz(pCtx, &pValue->pVariant, szValue)  );
							SG_ERR_CHECK(  SG_pendingtree__get_repo(pCtx, pChoice->pItem->pResolve->pPendingTree, &pRepo)  );
							SG_ERR_CHECK(  _submodule__load_summary(pCtx, szValue, pRepo, &pValue->szVariantSummary)  );
							break;
						}
					default:
						{
							SG_ERR_THROW2(SG_ERR_NOTIMPLEMENTED, (pCtx, "Unknown node type: %i", pChoice->pItem->eType));
							break;
						}
					}
					break;
				}
			default:
				{
					SG_ERR_THROW2(SG_ERR_NOTIMPLEMENTED, (pCtx, "Unknown resolve item state: %i", pChoice->eState));
					break;
				}
			}
		}
	}

	// if we still have a value, fill in the rest of it
	if (pValue != NULL)
	{
		// duplicate the given label
		SG_ERR_CHECK(  SG_strdup(pCtx, szLabel, &pValue->szLabel)  );

		// if a temp repo path was supplied, translate it into an absolute path
		if (szTempRepoPath != NULL)
		{
			SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path2(pCtx, pChoice->pItem->pResolve->pPendingTree, szTempRepoPath, &pValue->pVariantFile)  );
		}

		// store the given changeset
		pValue->szChangeset = szChangeset;
	}

	// return the value
	*ppValue = pValue;
	pValue = NULL;

fail:
	_VALUE__NULLFREE(pCtx, pValue);
	return;
}

/**
 * Allocates a value structure that is associated with the working directory.
 */
static void _value__alloc__working(
	SG_context*                    pCtx,    //< [in] [out] Error and context info.
	SG_resolve__choice*            pChoice, //< [in] The choice that will own the value.
	const SG_treediff2_ObjectData* pDiff,   //< [in] Diff to look up current state data in.
	                                        //<      NULL if the item is unchanged from the baseline.
	const SG_vhash*                pInput,  //< [in] The "input" that contains the item's status in the baseline changeset.
	                                        //<      Used to retrieve the working value if pDiff is NULL.
	SG_resolve__value**            ppValue  //< [out] The allocated value.
	)
{
	const char*        szRepoPath    = NULL;
	SG_resolve__value* pValue        = NULL;
	SG_string*         sLabel        = NULL;
	SG_pathname*       pAbsolutePath = NULL;
	SG_string*         sTarget       = NULL;

	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(ppValue);

	// allocate a blank value
	SG_ERR_CHECK(  _value__alloc__blank(pCtx, pChoice, &pValue)  );

	// use the standard "working" HID
	pValue->szChangeset = SG_RESOLVE__HID__WORKING;

	// look up the item's current repo path
	SG_ERR_CHECK(  SG_resolve__item__get_repo_path__working(pCtx, pChoice->pItem, &szRepoPath)  );

	// generate a working label
	SG_ERR_CHECK(  _generate_indexed_label(pCtx, pChoice, SG_RESOLVE__LABEL__WORKING, &sLabel)  );
	SG_ERR_CHECK(  SG_string__sizzle(pCtx, &sLabel, (SG_byte**)&pValue->szLabel, NULL)  );
	SG_ASSERT(sLabel == NULL);

	// look up the appropriate value in the diff
	if (pChoice->eState == SG_RESOLVE__STATE__EXISTENCE)
	{
		// we only care about whether or not the item currently exists
		// and we already know that without checking the diff
		SG_ERR_CHECK(  _variant__alloc__bool(pCtx, &pValue->pVariant, szRepoPath == NULL ? SG_FALSE : SG_TRUE)  );
	}
	else if (szRepoPath != NULL)
	{
		// look up the appropriate data from the diff
		switch (pChoice->eState)
		{
		case SG_RESOLVE__STATE__NAME:
			{
				const char* szValue = NULL;

				if (pDiff != NULL)
				{
					SG_ERR_CHECK(  SG_treediff2__ObjectData__get_name(pCtx, pDiff, &szValue)  );
				}
				else if (pInput != NULL)
				{
					SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pInput, gszIssueKey_Inputs_Name, &szValue)  );
				}

				SG_ERR_CHECK(  _variant__alloc__sz(pCtx, &pValue->pVariant, szValue)  );
			}
			break;
		case SG_RESOLVE__STATE__LOCATION:
			{
				const char* szValue = NULL;

				if (pDiff != NULL)
				{
					SG_ERR_CHECK(  SG_treediff2__ObjectData__get_parent_gid(pCtx, pDiff, &szValue)  );
				}
				else if (pInput != NULL)
				{
					SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pInput, gszIssueKey_Inputs_Location, &szValue)  );
				}

				SG_ERR_CHECK(  _variant__alloc__sz(pCtx, &pValue->pVariant, szValue)  );
				SG_ERR_CHECK(  _pendingtree__get_repo_path(pCtx, pChoice->pItem->pResolve->pPendingTree, pValue->pVariant->v.val_sz, &pValue->szVariantSummary)  );
			}
			break;
		case SG_RESOLVE__STATE__ATTRIBUTES:
			{
				SG_int64 iValue = 0;

				if (pDiff != NULL)
				{
					SG_ERR_CHECK(  SG_treediff2__ObjectData__get_attrbits(pCtx, pDiff, &iValue)  );
				}
				else if (pInput != NULL)
				{
					SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pInput, gszIssueKey_Inputs_Attributes, &iValue)  );
				}

				SG_ERR_CHECK(  _variant__alloc__int64(pCtx, &pValue->pVariant, iValue)  );
			}
			break;
		case SG_RESOLVE__STATE__XATTRIBUTES:
			{
				const char* szValue = NULL;

				if (pDiff != NULL)
				{
					SG_ERR_CHECK(  SG_treediff2__ObjectData__get_xattrs(pCtx, pDiff, &szValue)  );
				}
				else if (pInput != NULL)
				{
					SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pInput, gszIssueKey_Inputs_XAttributes, &szValue)  );
				}

				if (szValue == NULL)
				{
					szValue = SG_MRG_TOKEN__NO_XATTR;
				}

				SG_ERR_CHECK(  _variant__alloc__sz(pCtx, &pValue->pVariant, szValue)  );
			}
			break;
		case SG_RESOLVE__STATE__CONTENTS:
			{
				switch (pChoice->pItem->eType)
				{
				case SG_TREENODEENTRY_TYPE_REGULAR_FILE:
					{
						const char* szValue = NULL;

						if (pDiff != NULL)
						{
							SG_ERR_CHECK(  SG_treediff2__ObjectData__get_content_hid(pCtx, pDiff, &szValue)  );
						}
						else if (pInput != NULL)
						{
							SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pInput, gszIssueKey_Inputs_Contents, &szValue)  );
						}

						SG_ERR_CHECK(  _variant__alloc__sz(pCtx, &pValue->pVariant, szValue)  );
						SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path2(pCtx, pChoice->pItem->pResolve->pPendingTree, szRepoPath, &pValue->pVariantFile)  );

						break;
					}
				case SG_TREENODEENTRY_TYPE_DIRECTORY:
					{
						SG_ERR_THROW2(SG_ERR_NOTIMPLEMENTED, (pCtx, "Cannot resolve a directory contents conflict because they should never occur."));
						break;
					}
				case SG_TREENODEENTRY_TYPE_SYMLINK:
					{
						// Note: Can't look up the symlink's content HID in the repo
						//       because the working value probably isn't in the repo.
						SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path2(pCtx, pChoice->pItem->pResolve->pPendingTree, szRepoPath, &pAbsolutePath)  );
						SG_ERR_CHECK(  SG_fsobj__readlink(pCtx, pAbsolutePath, &sTarget)  );
						SG_ERR_CHECK(  _variant__alloc__sz(pCtx, &pValue->pVariant, SG_string__sz(sTarget))  );

						break;
					}
				case SG_TREENODEENTRY_TYPE_SUBMODULE:
					{
						const char* szValue = NULL;
						SG_repo*    pRepo   = NULL;

						if (pDiff != NULL)
						{
							SG_ERR_CHECK(  SG_treediff2__ObjectData__get_content_hid(pCtx, pDiff, &szValue)  );
						}
						else if (pInput != NULL)
						{
							SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pInput, gszIssueKey_Inputs_Contents, &szValue)  );
						}

						SG_ERR_CHECK(  _variant__alloc__sz(pCtx, &pValue->pVariant, szValue)  );
						SG_ERR_CHECK(  SG_pendingtree__get_repo(pCtx, pChoice->pItem->pResolve->pPendingTree, &pRepo)  );
						SG_ERR_CHECK(  _submodule__load_summary(pCtx, szValue, pRepo, &pValue->szVariantSummary)  );

						break;
					}
				default:
					{
						SG_ERR_THROW2(SG_ERR_NOTIMPLEMENTED, (pCtx, "Unknown node type: %i", pChoice->pItem->eType));
						break;
					}
				}
			}
			break;
		default:
			{
				SG_ERR_THROW2(SG_ERR_NOTIMPLEMENTED, (pCtx, "Unknown resolve item state: %i", pChoice->eState));
			}
			break;
		}
	}
	else
	{
		// the item doesn't exist and this is not an EXISTENCE conflict
		// so for now we can't provide a working value for this conflict
		_VALUE__NULLFREE(pCtx, pValue);
	}

	// return the value
	*ppValue = pValue;
	pValue = NULL;

fail:
	SG_STRING_NULLFREE(pCtx, sLabel);
	_VALUE__NULLFREE(pCtx, pValue);
	SG_PATHNAME_NULLFREE(pCtx, pAbsolutePath);
	SG_STRING_NULLFREE(pCtx, sTarget);
	return;
}

/**
 * Allocates a new value by loading it from a resolve hash (created using _value__build_resolve_vhash).
 */
static void _value__alloc__saved(
	SG_context*         pCtx,     //< [in] [out] Error and context info.
	const SG_vhash*     pHash,    //< [in] Hash to load the value from.
	SG_resolve__choice* pChoice,  //< [in] The choice that will own the value.
	SG_resolve__value** ppValue   //< [out] The allocated value.
	)
{
	SG_resolve__value* pValue   = NULL;
	SG_bool            bBool    = SG_FALSE;
	const char*        szString = NULL;
	SG_variant*        pVariant = NULL;
	SG_int64           iInt     = 0;

	SG_NULLARGCHECK(pHash);
	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(ppValue);

	// start with a blank value
	SG_ERR_CHECK(  _value__alloc__blank(pCtx, pChoice, &pValue)  );

	// read the label
	SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pHash, gszValueKey_Label, &szString)  );
	if (szString != NULL)
	{
		SG_ERR_CHECK(  SG_strdup(pCtx, szString, &pValue->szLabel)  );
		szString = NULL;
	}

	// read the variant
	SG_ERR_CHECK(  SG_vhash__check__variant(pCtx, pHash, gszValueKey_Value, &pVariant)  );
	if (pVariant != NULL)
	{
		SG_ERR_CHECK(  SG_variant__copy(pCtx, pVariant, &pValue->pVariant)  );
		pVariant = NULL;
	}

	// read the variant file
	SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pHash, gszValueKey_ValueFile, &szString)  );
	if (szString != NULL)
	{
		SG_ERR_CHECK(  SG_PATHNAME__ALLOC__PATHNAME_SZ(pCtx, &pValue->pVariantFile, pChoice->pTempPath, szString)  );
		szString = NULL;
	}

	// calculate the variant summary
	if (pValue->pChoice->eState == SG_RESOLVE__STATE__LOCATION)
	{
		SG_ERR_CHECK(  _pendingtree__get_repo_path(pCtx, pChoice->pItem->pResolve->pPendingTree, pValue->pVariant->v.val_sz, &pValue->szVariantSummary)  );
	}
	else if (
		   pValue->pChoice->eState == SG_RESOLVE__STATE__CONTENTS
		&& pValue->pChoice->pItem->eType == SG_TREENODEENTRY_TYPE_SUBMODULE
		)
	{
		SG_repo* pRepo = NULL;
		SG_ERR_CHECK(  SG_pendingtree__get_repo(pCtx, pValue->pChoice->pItem->pResolve->pPendingTree, &pRepo)  );
		SG_ERR_CHECK(  _submodule__load_summary(pCtx, pValue->pVariant->v.val_sz, pRepo, &pValue->szVariantSummary)  );
	}

	// read the changeset data
	SG_ERR_CHECK(  SG_vhash__has(pCtx, pHash, gszValueKey_Changeset, &bBool)  );
	if (bBool == SG_TRUE)
	{
		// read the changeset
		SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pHash, gszValueKey_Changeset, &szString)  );

		// the value doesn't own its changeset string, it points at someone else's
		// we have to find the right one to point at
		if (SG_stricmp__null(szString, pChoice->pItem->szAncestorHid) == 0)
		{
			// it matches the item's ancestor HID, point there
			pValue->szChangeset = pChoice->pItem->szAncestorHid;
		}
		else if (SG_stricmp(szString, SG_RESOLVE__HID__WORKING) == 0)
		{
			// it matches the working HID, point there
			pValue->szChangeset = SG_RESOLVE__HID__WORKING;
		}
		else
		{
			SG_bool   bFound = SG_FALSE;
			SG_uint32 uIndex = 0u;

			// look for it in the resolve's list of parents
			SG_ERR_CHECK(  SG_varray__find__sz(pCtx, pChoice->pItem->pResolve->pParentHids, szString, &bFound, &uIndex)  );
			if (bFound == SG_TRUE)
			{
				// found it, point into the list
				SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pChoice->pItem->pResolve->pParentHids, uIndex, &pValue->szChangeset)  );
			}
			else
			{
				SG_ERR_THROW2(SG_ERR_RESOLVE_INVALID_SAVE_DATA, (pCtx, "Invalid saved resolve data: unknown changeset: %s", szString));
			}
		}
		szString = NULL;
	}

	// read the merge data
	SG_ERR_CHECK(  SG_vhash__has(pCtx, pHash, gszValueKey_Baseline, &bBool)  );
	if (bBool == SG_TRUE)
	{
		// read the baseline's label and find a pointer to it
		SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pHash, gszValueKey_Baseline, &szString)  );
		SG_ERR_CHECK(  SG_resolve__choice__get_value__label(pCtx, pChoice, szString, 0u, &pValue->pBaselineParent)  );
		szString = NULL;

		// read the other's label and find a pointer to it
		SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pHash, gszValueKey_Other, &szString)  );
		SG_ERR_CHECK(  SG_resolve__choice__get_value__label(pCtx, pChoice, szString, 0u, &pValue->pOtherParent)  );
		szString = NULL;

		// update the child count of the value's parents
		pValue->pBaselineParent->uChildren += 1u;
		pValue->pOtherParent->uChildren += 1u;

		// we only save values that were manually merged
		// so any loaded values must be the same
		pValue->bAutoMerge = SG_FALSE;

		// read the merge tool name
		SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pHash, gszValueKey_Tool, &szString)  );
		SG_ERR_CHECK(  SG_strdup(pCtx, szString, &pValue->szMergeTool)  );
		szString = NULL;

		// read the merge tool result
		SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pHash, gszValueKey_ToolResult, &iInt)  );
		pValue->iMergeToolResult = (SG_int32)iInt;
		iInt = 0;
	}

	// return the value
	*ppValue = pValue;
	pValue = NULL;

fail:
	_VALUE__NULLFREE(pCtx, pValue);
	return;
}

/**
 * Creates and populates a new SG_resolve__value from a merge result.
 */
static void _value__alloc__merge(
	SG_context*         pCtx,        //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice,     //< [in] The choice that will own the value.
	SG_string**         ppLabel,     //< [in] The new value's label.
	                                 //<      We steal this pointer and NULL the caller's.
	SG_pathname**       ppTempFile,  //< [in] The new value's absolute temporary pathname.
	                                 //<      We steal this pointer and NULL the caller's.
	                                 //<      May be NULL if the merge failed.
	SG_repo*            pRepo,       //< [in] Repo to use to calculate the HID of the temp file.
	                                 //<      Ignored if ppTempFile is NULL.
	SG_resolve__value*  pBaseline,   //< [in] The new value's baseline parent.
	SG_resolve__value*  pOther,      //< [in] The new value's other parent.
	SG_bool             bAutomatic,  //< [in] Whether or not the value was created automatically by merge,
	                                 //<      as opposed to manually by resolve.
	char**              ppTool,      //< [in] The merge tool used to create the new value.
	                                 //<      We steal this pointer and NULL the caller's.
	SG_int32            iToolResult, //< [in] The result returned by the specified tool.
	SG_resolve__value** ppValue      //< [out] The new value.
	                                 //<       The pointers we stole now belong to it.
	)
{
	SG_resolve__value* pValue = NULL;
	SG_file*           pFile  = NULL;
	char*              szHid  = NULL;

	SG_NULLARGCHECK(ppLabel);
	SG_NULLARGCHECK(pBaseline);
	SG_NULLARGCHECK(pOther);
	SG_NULLARGCHECK(ppTool);
	SG_NULLARGCHECK(ppValue);

	SG_ERR_CHECK(  _value__alloc__blank(pCtx, pChoice, &pValue)  );

	SG_ERR_CHECK(  SG_string__sizzle(pCtx, ppLabel, (SG_byte**)&pValue->szLabel, NULL)  );
	SG_ASSERT(*ppLabel == NULL);

	if (ppTempFile != NULL)
	{
		// store the filename
		pValue->pVariantFile = *ppTempFile;
		*ppTempFile = NULL;

		// open the temp file with write permission so that we know that nobody
		// else is writing to the file while we're computing its HID
		SG_ERR_CHECK(  SG_file__open__pathname(pCtx, pValue->pVariantFile, SG_FILE_OPEN_EXISTING | SG_FILE_RDWR, 0, &pFile)  );

		// compute the temp file's HID
		SG_ERR_CHECK(  SG_repo__alloc_compute_hash__from_file(pCtx, pRepo, pFile, &szHid)  );
		SG_ERR_CHECK(  _variant__alloc__sz(pCtx, &pValue->pVariant, szHid)  );

		// close the file
		SG_ERR_CHECK(  SG_file__close(pCtx, &pFile)  );
		SG_ASSERT(pFile == NULL);
	}

	pValue->pBaselineParent = pBaseline;
	pValue->pOtherParent    = pOther;
	pValue->bAutoMerge      = bAutomatic;

	pValue->szMergeTool = *ppTool;
	*ppTool = NULL;

	pValue->iMergeToolResult = iToolResult;

	pBaseline->uChildren += 1u;
	pOther->uChildren    += 1u;

	*ppValue = pValue;
	pValue = NULL;

fail:
	SG_PATHNAME_NULLFREE(pCtx, *ppTempFile);
	SG_FILE_NULLCLOSE(pCtx, pFile);
	SG_NULLFREE(pCtx, szHid);
	SG_NULLFREE(pCtx, *ppTool);
	_VALUE__NULLFREE(pCtx, pValue);
	return;
}

/**
 * Creates a hash that contains the value data that is unique to resolve.
 * Data derived from merge issues or pendingtree is not included.
 * If the given value can be entirely constructed from merge/pendingtree data,
 * then no hash is generated at all.
 */
static void _value__build_resolve_vhash(
	SG_context*        pCtx,   //< [in] [out] Error and context data.
	SG_resolve__value* pValue, //< [in] The value to build a vhash from.
	SG_vhash**         ppHash  //< [out] The built vhash, or NULL if none is required.
	)
{
	SG_vhash* pHash = NULL;

	SG_NULLARGCHECK(pValue);
	SG_NULLARGCHECK(ppHash);

	// There are two kinds of values that we need to save out.. any other values
	// can be reconstructed next time from merge issue and/or pendingtree data.
	// 1) Merge results that were manual (generated by resolve rather than merge).
	// 2) Working directory values that were merged into another value.
	//    Working values that haven't been merged are unnecessary, because we'll just
	//    create an updated working value to replace it next time we allocate resolve data.
	//    Ones that have been merged need to be kept because next time we allocate
	//    resolve data we'll need to have both the data from the time it was merged
	//    and a new working value from the time of the allocation.
	if (
		// test for (1)
		(
			   pValue->pBaselineParent != NULL
			&& pValue->pOtherParent    != NULL
			&& pValue->bAutoMerge      == SG_FALSE
		)
		||
		// test for (2)
		(
			   pValue->szChangeset == SG_RESOLVE__HID__WORKING
			&& pValue->uChildren   >  0u
		)
		)
	{
		const char* szFilename = NULL;

		// allocate the vhash
		SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pHash)  );

		// label
		SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pHash, gszValueKey_Label, pValue->szLabel)  );

		// value
		if (pValue->pVariant != NULL)
		{
			SG_ERR_CHECK(  SG_vhash__addcopy__variant(pCtx, pHash, gszValueKey_Value, pValue->pVariant)  );
		}

		// value file
		// Only save the filename part of the variant file.
		// we'll add the path back on when we load it.
		// This is necessary because the absolute path may
		// change between save and load.
		if (pValue->pVariantFile != NULL)
		{
			SG_ERR_CHECK(  SG_pathname__filename(pCtx, pValue->pVariantFile, &szFilename)  );
			SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pHash, gszValueKey_ValueFile, szFilename)  );
		}

		// changeset data
		if (pValue->szChangeset != NULL)
		{
			SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pHash, gszValueKey_Changeset, pValue->szChangeset)  );
		}

		// merge data
		if (pValue->pBaselineParent != NULL && pValue->pOtherParent != NULL)
		{
			SG_ASSERT(pValue->szMergeTool != NULL);
			SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pHash, gszValueKey_Baseline, pValue->pBaselineParent->szLabel)  );
			SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pHash, gszValueKey_Other, pValue->pOtherParent->szLabel)  );
			SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pHash, gszValueKey_Tool, pValue->szMergeTool)  );
			SG_ERR_CHECK(  SG_vhash__add__int64(pCtx, pHash, gszValueKey_ToolResult, pValue->iMergeToolResult)  );
		}
	}

	// return the hash
	*ppHash = pHash;
	pHash = NULL;

fail:
	SG_VHASH_NULLFREE(pCtx, pHash);
	return;
}

/**
 * Finds a changeset to reasonably associate with a value.
 *
 * If the value isn't directly associated with a changeset and is a merge result,
 * then its parents are searched recursively until a changeset is found.
 */
static void _value__find_changeset(
	SG_context*        pCtx,       //< [in] [out] Error and context info.
	SG_resolve__value* pValue,     //< [in] The value to find a changeset for.
	const char**       ppChangeset //< [out] The changeset found.
	)
{
	SG_NULLARGCHECK(pValue);

	if (pValue->szChangeset != NULL)
	{
		*ppChangeset = pValue->szChangeset;
	}
	else if (pValue->pBaselineParent != NULL)
	{
		SG_ERR_CHECK(  _value__find_changeset(pCtx, pValue->pBaselineParent, ppChangeset)  );
	}
	else if (pValue->pOtherParent != NULL)
	{
		SG_ERR_CHECK(  _value__find_changeset(pCtx, pValue->pOtherParent, ppChangeset)  );
	}
	else
	{
		*ppChangeset = NULL;

		// It shouldn't be possible to not find a changeset for a value.
		// All values that are merged should have an ancestry that ultimately
		// leads back to values that are directly associated with a changeset,
		// and values that are neither directly associated with a changeset or
		// merged should not be possible.
		SG_ASSERT(SG_FALSE);
	}

fail:
	return;
}

/**
 * An SG_free_callback that frees values.
 */
static void _value__sg_action__free(
	SG_context* pCtx, //< [in] [out] Error and context info.
	void*       pData //< [in] The SG_resolve__value to free.
	)
{
	SG_resolve__value* pValue = (SG_resolve__value*)pData;
	_VALUE__NULLFREE(pCtx, pValue);
}

/**
 * An SG_vector_predicate that matches an SG_resolve__value's
 * label against a given label.
 */
static void _value__vector_predicate__match_label(
	SG_context* pCtx,      //< [in] [out] Error and context info.
	SG_uint32   uIndex,    //< [in] The index of the value being matched.
	void*       pVal,      //< [in] The SG_resolve__value being matched.
	void*       pUserData, //< [in] The string label (const char*) to match against.
	SG_bool*    pMatch     //< [out] Whether or not the value's label matches the given one.
	)
{
	SG_resolve__value* pValue       = (SG_resolve__value*)pVal;
	const char*        szMatchLabel = (const char*)pUserData;
	const char*        szValueLabel = NULL;

	SG_UNUSED(uIndex);
	SG_NULLARGCHECK(pVal);
	SG_NULLARGCHECK(pUserData);
	SG_NULLARGCHECK(pMatch);

	SG_ERR_CHECK(  SG_resolve__value__get_label(pCtx, pValue, &szValueLabel)  );

	if (SG_stricmp(szMatchLabel, szValueLabel) == 0)
	{
		*pMatch = SG_TRUE;
	}
	else
	{
		*pMatch = SG_FALSE;
	}

fail:
	return;
}

/**
 * An SG_vector_predicate that matches an SG_resolve__value's
 * changeset HID against the given HID.
 */
static void _value__vector_predicate__match_changeset(
	SG_context* pCtx,      //< [in] [out] Error and context info.
	SG_uint32   uIndex,    //< [in] The index of the value being matched.
	void*       pVal,      //< [in] The SG_resolve__value being matched.
	void*       pUserData, //< [in] The changeset hid string (const char*) to match against.
	SG_bool*    pMatch     //< [out] Whether or not the value's changeset HID matches the given one.
	)
{
	SG_resolve__value* pValue           = (SG_resolve__value*)pVal;
	const char*        szMatchChangeset = (const char*)pUserData;
	SG_bool            bValueChangeset  = SG_FALSE;
	const char*        szValueChangeset = NULL;

	SG_UNUSED(uIndex);
	SG_NULLARGCHECK(pVal);
	SG_NULLARGCHECK(pUserData);
	SG_NULLARGCHECK(pMatch);

	SG_ERR_CHECK(  SG_resolve__value__get_changeset(pCtx, pValue, &bValueChangeset, &szValueChangeset)  );

	if (bValueChangeset == SG_TRUE && SG_stricmp(szMatchChangeset, szValueChangeset) == 0)
	{
		*pMatch = SG_TRUE;
	}
	else
	{
		*pMatch = SG_FALSE;
	}

fail:
	return;
}

/**
 * An SG_vector_predicate that matches either of an
 * SG_resolve__value's parent pointers.
 */
static void _value__vector_predicate__match_parents(
	SG_context* pCtx,      //< [in] [out] Error and context info.
	SG_uint32   uIndex,    //< [in] The index of the value being matched.
	void*       pVal,      //< [in] The current SG_resolve__value being matched.
	void*       pUserData, //< [in] The SG_resolve__value whose parents will match.
	SG_bool*    pMatch     //< [out] Whether or not the current value is either of the specified value's parents.
	)
{
	SG_resolve__value* pCurrentValue   = (SG_resolve__value*)pVal;
	SG_resolve__value* pChildValue     = (SG_resolve__value*)pUserData;
	SG_bool            bMerged         = SG_FALSE;
	SG_resolve__value* pBaselineParent = NULL;
	SG_resolve__value* pOtherParent    = NULL;

	SG_UNUSED(uIndex);
	SG_NULLARGCHECK(pVal);
	SG_NULLARGCHECK(pUserData);
	SG_NULLARGCHECK(pMatch);

	SG_ERR_CHECK(  SG_resolve__value__get_merge(pCtx, pChildValue, &bMerged, &pBaselineParent, &pOtherParent, NULL, NULL, NULL)  );

	if (
		bMerged == SG_TRUE
		&& (
			   pCurrentValue == pBaselineParent
			|| pCurrentValue == pOtherParent
		)
	)
	{
		*pMatch = SG_TRUE;
	}
	else
	{
		*pMatch = SG_FALSE;
	}

fail:
	return;
}

/**
 * An SG_vector_predicate that matches an SG_resolve__value's data.
 */
static void _value__vector_predicate__match_data(
	SG_context* pCtx,      //< [in] [out] Error and context info.
	SG_uint32   uIndex,    //< [in] The index of the value being matched.
	void*       pVal,      //< [in] The current SG_resolve__value being matched.
	void*       pUserData, //< [in] The SG_resolve__value whose data should be matched against.
	SG_bool*    pMatch     //< [out] Whether or not the current value's data matches the given value's.
	)
{
	SG_resolve__value* pCurrentValue = (SG_resolve__value*)pVal;
	SG_resolve__value* pMatchValue   = (SG_resolve__value*)pUserData;

	SG_UNUSED(uIndex);
	SG_NULLARGCHECK(pVal);
	SG_NULLARGCHECK(pUserData);
	SG_NULLARGCHECK(pMatch);

	// we can't use SG_variant__equal in the case of strings
	// because we want a case-insensitive comparison,
	// but it will work fine for other cases
	if (
		   pCurrentValue->pVariant->type == SG_VARIANT_TYPE_SZ
		&& pMatchValue->pVariant->type   == SG_VARIANT_TYPE_SZ
		)
	{
		*pMatch = SG_stricmp(pCurrentValue->pVariant->v.val_sz, pMatchValue->pVariant->v.val_sz) == 0 ? SG_TRUE : SG_FALSE;
	}
	else
	{
		SG_ERR_CHECK(  SG_variant__equal(pCtx, pCurrentValue->pVariant, pMatchValue->pVariant, pMatch)  );
	}

fail:
	return;
}

/**
 * An SG_resolve__value__callback that adds each value to a varray as a vhash.
 */
static void _value__resolve_action__add_resolve_vhash_to_varray(
	SG_context*        pCtx,      //< [in] [out] Error and context info.
	SG_resolve__value* pValue,    //< [in] The current value.
	void*              pUserData, //< [in] The SG_varray to add the value to.
	SG_bool*           pContinue  //< [out] Whether or not to continue iteration.
	)
{
	SG_varray* pValues = (SG_varray*)pUserData;
	SG_vhash*  pHash   = NULL;

	SG_NULLARGCHECK(pValue);
	SG_NULLARGCHECK(pUserData);
	SG_NULLARGCHECK(pContinue);

	// get a hash from the value
	SG_ERR_CHECK(  _value__build_resolve_vhash(pCtx, pValue, &pHash)  );

	// if we got a hash back, add it to the supplied array of values
	// (the array takes ownership of the hash)
	if (pHash != NULL)
	{
		SG_ERR_CHECK(  SG_varray__append__vhash(pCtx, pValues, &pHash)  );
		pHash = NULL;
	}

	// continue iterating
	*pContinue = SG_TRUE;

fail:
	SG_VHASH_NULLFREE(pCtx, pHash);
	return;
}

/**
 * Type of user data required by _value__vector_action__call_resolve_action.
 */
typedef struct
{
	SG_resolve__value__callback* fCallback; //< The value callback to call.
	void*                        pUserData; //< The user data to pass to the callback.
	SG_bool                      bContinue; //< Whether or not to actually call the callback.
}
_value__vector_action__call_resolve_action__data;

/**
 * An SG_vector_foreach_callback that calls an SG_resolve__value__callback.
 */
static void _value__vector_action__call_resolve_action(
	SG_context* pCtx,     //< [in] [out] Error and context info.
	SG_uint32   uIndex,   //< [in] The index of the item.
	void*       pVal,     //< [in] The value of the item.
	void*       pUserData //< [in] A populated _value__vector_action__call_resolve_action structure.
	)
{
	SG_resolve__value*                                pValue = (SG_resolve__value*)pVal;
	_value__vector_action__call_resolve_action__data* pArgs  = (_value__vector_action__call_resolve_action__data*)pUserData;

	SG_UNUSED(uIndex);
	SG_NULLARGCHECK(pVal);
	SG_NULLARGCHECK(pUserData);

	if (pArgs->bContinue == SG_TRUE)
	{
		SG_ERR_CHECK(  pArgs->fCallback(pCtx, pValue, pArgs->pUserData, &pArgs->bContinue)  );
	}

fail:
	return;
}


/*
**
** Internal Functions - Choice
**
*/

/**
 * Frees a choice.
 */
static void _choice__free(
	SG_context*         pCtx,   //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice //< [in] The choice to free.
	)
{
	if (pChoice == NULL)
	{
		return;
	}

	SG_PATHNAME_NULLFREE(pCtx, pChoice->pTempPath);
	SG_VECTOR_NULLFREE_WITH_ASSOC(pCtx, pChoice->pValues, _value__sg_action__free);
	SG_VECTOR_NULLFREE(pCtx, pChoice->pLeafValues);
	SG_VECTOR_NULLFREE(pCtx, pChoice->pCollisionItems);
	SG_VECTOR_NULLFREE(pCtx, pChoice->pCycleItems);
	SG_NULLFREE(pCtx, pChoice->szCycleHint);
	pChoice->pItem           = NULL;
	pChoice->eState          = SG_RESOLVE__STATE__COUNT;
	pChoice->eConflictCauses = SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__ZERO;
	pChoice->bCollisionCause = SG_FALSE;
	pChoice->pResult         = NULL;
	SG_NULLFREE(pCtx, pChoice);
}

/**
 * NULLFREE wrapper around _choice__free.
 */
#define _CHOICE__NULLFREE(pCtx, pChoice) _sg_generic_nullfree(pCtx, pChoice, _choice__free)

/**
 * Updates a choice's list of leaves by taking a newly merged value into consideration.
 *
 * The value's parents are removed from the list of leaves.  If either parent was
 * in the list of leaves, then the value itself is added as a leaf.
 */
static void _choice__update_leaves(
	SG_context*         pCtx,      //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice,   //< [in] The choice to update the leaves of.
	SG_resolve__value*  pValue,    //< [in] The newly merged value.
	                               //<      Must already list pChoice as its owner.
	SG_bool             bOverwrite //< [in] If true, remove any existing leaf with the same label as the new value.
	                               //<      Use this when potentially replacing an old leaf value with a new one.
	)
{
	SG_uint32 uParentLeaves = 0u;

	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pValue);
	SG_ARGCHECK(pValue->pChoice == pChoice, pValue);

	// if we're potentially overwriting an old leaf value,
	// then remove any existing leaf with the same label as the new one
	if (bOverwrite == SG_TRUE)
	{
		SG_ERR_CHECK(  SG_vector__remove__if(pCtx, pChoice->pLeafValues, _value__vector_predicate__match_label, (void*)pValue->szLabel, NULL, NULL)  );
	}

	// remove any leaf values that are the parents of the new value
	SG_ERR_CHECK(  SG_vector__remove__if(pCtx, pChoice->pLeafValues, _value__vector_predicate__match_parents, (void*)pValue, &uParentLeaves, NULL)  );

	// if any of the values parents were leaves, then add it to the list
	if (uParentLeaves > 0u)
	{
		SG_ERR_CHECK(  SG_vector__append(pCtx, pChoice->pLeafValues, (void*)pValue, NULL)  );
	}

fail:
	return;
}

/**
 * Gets the only leaf in a choice.
 */
static void _choice__get_only_leaf(
	SG_context*         pCtx,    //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice, //< [in] The choice to get the only leaf from.
	SG_resolve__value** ppLeaf   //< [out] The only leaf value in the choice.
	                             //<       NULL if the choice doesn't have exactly one leaf value.
	)
{
	SG_uint32          uCount = 0u;
	SG_resolve__value* pLeaf  = NULL;

	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(ppLeaf);

	SG_ERR_CHECK(  SG_vector__length(pCtx, pChoice->pLeafValues, &uCount)  );
	if (uCount == 1u)
	{
		SG_ERR_CHECK(  SG_vector__get(pCtx, pChoice->pLeafValues, 0u, (void**)&pLeaf)  );
	}

	*ppLeaf = pLeaf;
	pLeaf = NULL;

fail:
	return;
}

/**
 * Checks to see if all of the values in a vector (presumably from a choice)
 * all have equivalent variant data.  If this occurs, the choice is unnecessary.
 */
static void _choice__values_equivalent(
	SG_context* pCtx,    //< [in] [out] Error and context info.
	SG_vector*  pValues, //< [in] Values to compare for equivalency.
	SG_bool*    pEqual   //< [out] True if all values in the vector are equivalent, false otherwise.
	)
{
	SG_bool   bEqual     = SG_TRUE;
	SG_uint32 uCount     = 0u;
	SG_uint32 uLeftIndex = 0u;

	SG_NULLARGCHECK(pValues);
	SG_NULLARGCHECK(pEqual);

	// run through each value in the vector
	// compare each value to all of the values after it
	SG_ERR_CHECK(  SG_vector__length(pCtx, pValues, &uCount)  );
	for (uLeftIndex = 0u; uLeftIndex < uCount - 1u; ++uLeftIndex)
	{
		SG_uint32 uRightIndex = 0u;

		// run through each value after the current one
		// and compare it to the current one
		for (uRightIndex = uLeftIndex + 1u; uRightIndex < uCount; ++uRightIndex)
		{
			SG_resolve__value* pLeftValue   = NULL;
			SG_resolve__value* pRightValue  = NULL;

			// get the two values to compare from the vector
			SG_ERR_CHECK(  SG_vector__get(pCtx, pValues, uLeftIndex, (void**)&pLeftValue)  );
			SG_ERR_CHECK(  SG_vector__get(pCtx, pValues, uRightIndex, (void**)&pRightValue)  );

			// compare the values
			SG_ERR_CHECK(  _value__equal_variants(pCtx, pLeftValue, pRightValue, &bEqual)  );

			// if they're not equal, then we can skip the rest of the checks
			if (bEqual == SG_FALSE)
			{
				break;
			}
		}

		// if any were not equal, then we can skip the rest of the checks
		if (bEqual == SG_FALSE)
		{
			break;
		}
	}

	// return whatever we ended up with
	*pEqual = bEqual;

fail:
	return;
}

/**
 * Allocates all of a choice's values that are specified in a merge issue.
 */
static void _choice__alloc__changeset_values(
	SG_context*         pCtx,   //< [in] [out] Error and context info.
	const SG_vhash*     pIssue, //< [in] The issue we're allocating from.
	SG_resolve__choice* pChoice //< [in] The choice we're allocating within.
	)
{
	SG_vhash*          pInputs                = NULL;
	SG_resolve__value* pValue                 = NULL;
	const char*        szAncestorTempRepoPath = NULL;
	const char*        szBaselineTempRepoPath = NULL;
	const char*        szOtherTempRepoPath    = NULL;
	SG_string*         sLabel                 = NULL;
	SG_uint32          uCount                 = 0u;
	SG_uint32          uIndex                 = 0u;

	// check the arguments
	SG_NULLARGCHECK(pIssue);
	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pChoice->pValues);
	SG_NULLARGCHECK(pChoice->pLeafValues);

	// get the list of input changeset data from the issue
	SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pIssue, gszIssueKey_Inputs, &pInputs)  );

	// get the number of parents we have
	SG_ERR_CHECK(  SG_varray__count(pCtx, pChoice->pItem->pResolve->pParentHids, &uCount)  );

	// for CONTENTS/FILE values, get the temp paths from the issue
	// Note: CONTENTS/FILE merge issues only support 2-way merges.
	//       So we'll always have exactly two parents, one baseline and one other.
	if (pChoice->eState == SG_RESOLVE__STATE__CONTENTS && pChoice->pItem->eType == SG_TREENODEENTRY_TYPE_REGULAR_FILE)
	{
		SG_ASSERT(uCount == 2u);
		SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pIssue, gszIssueKey_File_AncestorTempRepoPath, &szAncestorTempRepoPath)  );
		SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pIssue, gszIssueKey_File_BaselineTempRepoPath, &szBaselineTempRepoPath)  );
		SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pIssue, gszIssueKey_File_OtherTempRepoPath,    &szOtherTempRepoPath)  );
	}

	// create a value for the ancestor changeset, if the item has one
	if (pChoice->pItem->szAncestorHid != NULL)
	{
		SG_ERR_CHECK(  _value__alloc__changeset(pCtx, pChoice, SG_RESOLVE__LABEL__ANCESTOR, pInputs, szAncestorTempRepoPath, pChoice->pItem->szAncestorHid, &pValue)  );

		// we'd only get a NULL value back if the ancestor HID wasn't found in pInputs
		// but if we have an ancestor HID at all (which we already checked for)
		// then we originally found it in pInputs
		SG_ASSERT(pValue != NULL);

		SG_ERR_CHECK(  SG_vector__append(pCtx, pChoice->pValues, (void*)pValue, NULL)  );
		pValue = NULL;
	}

	// create a value for each parent changeset
	for (uIndex = 0u; uIndex < uCount; ++uIndex)
	{
		const char* szChangeset    = NULL;
		const char* szLabel        = NULL;
		const char* szTempRepoPath = NULL;

		// get the HID of the current parent
		SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pChoice->pItem->pResolve->pParentHids, uIndex, &szChangeset)  );

		// configure the label and temp repo path appropriately
		if (uIndex == 0u)
		{
			// the first parent is always the baseline
			szTempRepoPath = szBaselineTempRepoPath;
			szLabel        = SG_RESOLVE__LABEL__BASELINE;
		}
		else
		{
			// parents beyond the first are indexed "others"
			// Note: The temp repo path is not indexed, because the only time
			//       szOtherTempRepoPath is non-NULL is for a CONTENTS/FILE value
			//       in which case we only support a single "other".
			szTempRepoPath = szOtherTempRepoPath;
			SG_ERR_CHECK(  _generate_indexed_label(pCtx, pChoice, SG_RESOLVE__LABEL__OTHER, &sLabel)  );
			szLabel = SG_string__sz(sLabel);
		}

		// create the value
		SG_ERR_CHECK(  _value__alloc__changeset(pCtx, pChoice, szLabel, pInputs, szTempRepoPath, szChangeset, &pValue)  );
		szLabel = NULL;
		SG_STRING_NULLFREE(pCtx, sLabel);

		// if we got a value, append it to the choice's lists
		if (pValue != NULL)
		{
			SG_ERR_CHECK(  SG_vector__append(pCtx, pChoice->pLeafValues, (void*)pValue, NULL)  );
			SG_ERR_CHECK(  SG_vector__append(pCtx, pChoice->pValues, (void*)pValue, NULL)  );
			pValue = NULL;
		}
	}

fail:
	_VALUE__NULLFREE(pCtx, pValue);
	SG_STRING_NULLFREE(pCtx, sLabel);
	return;
}

/**
 * Allocates a value to represent an automerge result, if there is one.
 *
 * Note: Right now the only conflict that automerges is a divergent file edit.
 *       Right now, merge issues only support data for 2-way file edit merges.
 *       Therefore, this function currently never creates more than one automerge value.
 *       Once N-way file edit merges are supported, it could conceivably create a whole cascade of automerge values.
 */
static void _choice__alloc__automerge_values(
	SG_context*         pCtx,   //< [in] [out] Error and context info.
	const SG_vhash*     pIssue, //< [in] The issue we're allocating from.
	SG_resolve__choice* pChoice //< [in] The choice we're allocating within.
	)
{
	SG_string*         sLabel    = NULL;
	SG_resolve__value* pValue    = NULL;
	SG_pathname*       pTempFile = NULL;
	char*              szTool    = NULL;

	SG_NULLARGCHECK(pIssue);
	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pChoice->pValues);
	SG_NULLARGCHECK(pChoice->pLeafValues);

	// no automerge value will exist unless the issue contains a divergent file edit
	if ((pChoice->eConflictCauses & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__MASK) != 0)
	{
		SG_int32 eToolResult = SG_MERGETOOL__STATUS__MUST_BE_LAST;

		// check exactly what happened when the automerge was run
		if ((pChoice->eConflictCauses & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__AUTO_CONFLICT) != 0)
		{
			eToolResult = SG_MERGETOOL__STATUS__CONFLICT;
		}
		else if ((pChoice->eConflictCauses & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__AUTO_ERROR) != 0)
		{
			eToolResult = SG_MERGETOOL__STATUS__ERROR;
		}
		else if ((pChoice->eConflictCauses & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__AUTO_OK) != 0)
		{
			eToolResult = SG_MERGETOOL__STATUS__SUCCESSFUL;
		}
		// other DIVERGENT_FILE_EDIT__* flags indicate that no automerge was attempted

		// create the value, if necessary
		if (eToolResult != SG_MERGETOOL__STATUS__MUST_BE_LAST)
		{
			SG_uint32          uCount         = 0u;
			SG_resolve__value* pBaseline      = NULL;
			SG_resolve__value* pOther         = NULL;
			const char*        szTempRepoPath = NULL;
			const char*        szTempTool     = NULL;
			SG_repo*           pRepo          = NULL;

			// get the number of values in the choice's vector
			SG_ERR_CHECK(  SG_vector__length(pCtx, pChoice->pValues, &uCount)  );
			SG_ARGCHECK(uCount == 3u, pValues);

			// get the baseline and other values
			SG_ERR_CHECK(  SG_vector__get(pCtx, pChoice->pValues, 1u, (void**)&pBaseline)  );
			SG_ERR_CHECK(  SG_vector__get(pCtx, pChoice->pValues, 2u, (void**)&pOther)  );

			// build a label
			SG_ERR_CHECK(  _generate_indexed_label(pCtx, pChoice, SG_RESOLVE__LABEL__MERGE, &sLabel)  );

			// get the temp filename from the merge issue
			// convert it from a repo path to an absolute one
			SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pIssue, gszIssueKey_File_MergeTempRepoPath, &szTempRepoPath)  );
			SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path2(pCtx, pChoice->pItem->pResolve->pPendingTree, szTempRepoPath, &pTempFile)  );

			// get the tool name from the merge issue and duplicate it
			SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pIssue, gszIssueKey_File_AutomergeTool, &szTempTool)  );
			SG_ERR_CHECK(  SG_strdup(pCtx, szTempTool, &szTool)  );

			// allocate the value
			SG_ERR_CHECK(  SG_pendingtree__get_repo(pCtx, pChoice->pItem->pResolve->pPendingTree, &pRepo)  );
			SG_ERR_CHECK(  _value__alloc__merge(pCtx, pChoice, &sLabel, &pTempFile, pRepo, pBaseline, pOther, SG_TRUE, &szTool, eToolResult, &pValue)  );

			// if the value represents a successful merge
			// then it replaces its parents in the list of leaves
			if (eToolResult == SG_MERGETOOL__STATUS__SUCCESSFUL)
			{
				SG_ERR_CHECK(  _choice__update_leaves(pCtx, pChoice, pValue, SG_FALSE)  );
			}

			// add the new value to the vector
			SG_ERR_CHECK(  SG_vector__append(pCtx, pChoice->pValues, (void*)pValue, NULL)  );
			pValue = NULL;
		}
	}

fail:
	SG_STRING_NULLFREE(pCtx, sLabel);
	SG_PATHNAME_NULLFREE(pCtx, pTempFile);
	SG_NULLFREE(pCtx, szTool);
	_VALUE__NULLFREE(pCtx, pValue);
	return;
}

/**
 * Allocates all of a choice's values that are stored in saved choice data.
 */
static void _choice__alloc__saved_values(
	SG_context*         pCtx,   //< [in] [out] Error and context info.
	const SG_vhash*     pSaved, //< [in] The saved choice data to load values from.
	SG_resolve__choice* pChoice //< [in] The choice we're allocating within.
	)
{
	SG_varray*         pValues   = NULL;
	SG_resolve__value* pValue    = NULL;
	SG_bool            bOwnValue = SG_FALSE;

	SG_NULLARGCHECK(pSaved);
	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pChoice->pValues);
	SG_NULLARGCHECK(pChoice->pLeafValues);

	// get the list of values from the saved data and iterate through them
	SG_ERR_CHECK(  SG_vhash__check__varray(pCtx, pSaved, gszChoiceKey_Values, &pValues)  );
	if (pValues != NULL)
	{
		SG_uint32 uCount = 0u;
		SG_uint32 uIndex = 0u;

		SG_ERR_CHECK(  SG_varray__count(pCtx, pValues, &uCount)  );
		for (uIndex = 0u; uIndex < uCount; ++uIndex)
		{
			SG_vhash* pHash  = NULL;

			// get the current value hash
			SG_ERR_CHECK(  SG_varray__get__vhash(pCtx, pValues, uIndex, &pHash)  );

			// load a value from the hash
			SG_ERR_CHECK(  _value__alloc__saved(pCtx, pHash, pChoice, &pValue)  );
			bOwnValue = SG_TRUE;

			// add the value to the choice's list
			SG_ERR_CHECK(  SG_vector__append(pCtx, pChoice->pValues, (void*)pValue, NULL)  );
			bOwnValue = SG_FALSE;

			// if the value has parents, then we need to update the choice's leaves
			SG_ASSERT(pValue->pBaselineParent == NULL || pValue->pOtherParent != NULL);
			if (pValue->pBaselineParent != NULL && pValue->pOtherParent != NULL)
			{
				SG_ERR_CHECK(  _choice__update_leaves(pCtx, pChoice, pValue, SG_FALSE)  );
			}
		}
	}

fail:
	if (bOwnValue == SG_TRUE)
	{
		_VALUE__NULLFREE(pCtx, pValue);
	}
	return;
}

/**
 * Allocates a choice's working value.
 */
static void _choice__alloc__working_values(
	SG_context*                    pCtx,   //< [in] [out] Error and context info.
	const SG_vhash*                pIssue, //< [in] The issue we're allocating a choice from.
	const SG_treediff2_ObjectData* pDiff,  //< [in] Diff object to read current item data from.
	                                       //<      NULL if the item is unchanged from the baseline.
	SG_resolve__choice*            pChoice //< [in] The choice we're allocating within.
	)
{
	SG_resolve__value* pValue = NULL;
	SG_vhash*          pInput = NULL;

	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pChoice->pValues);
	SG_NULLARGCHECK(pChoice->pLeafValues);

	// If the item doesn't appear in the diff
	// then its current value is the same that it was in the baseline.
	// Look up the state of the item as it occurred in the baseline
	// so that we can get the current value from there instead.
	if (pDiff == NULL)
	{
		const char* szBaselineHid  = NULL;
		SG_vhash*   pInputs        = NULL;

		SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pChoice->pItem->pResolve->pParentHids, 0u, &szBaselineHid)  );
		SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pIssue, gszIssueKey_Inputs, &pInputs)  );
		SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pInputs, szBaselineHid, &pInput)  );
	}

	// allocate a working value
	SG_ERR_CHECK(  _value__alloc__working(pCtx, pChoice, pDiff, pInput, &pValue)  );
	if (pValue != NULL)
	{
		// check if we need to do anything with the choice's leaves
		if (
			((pChoice->eConflictCauses & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__TBD) != 0)
			|| ((pChoice->eConflictCauses & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__DIVERGENT_FILE_EDIT__NO_RULE) != 0)
			)
		{
			// If the working value was not created by an automerge,
			// and the choice would end up being mergeable,
			// then the value effectively came from the baseline's value.
			// This case occurs when merge ran into a situation where it
			// would like to merge two values but couldn't for some reason,
			// usually because no rule specifying a mergetool is defined.
			// Therefore, if the working value is not from an automerge
			// we want to replace the baseline value with the working value
			// in the list of leaves.  That way the user will only need to
			// run a single merge to get the baseline, other, and working
			// changes (since the working value started as the baseline value,
			// the working value contains the baseline changes by default).
			// If the choice is non-mergeable, the entire list of leaves will
			// be going away anyway, so it doesn't hurt in that case.
			SG_resolve__value* pBaseline = NULL;

			// find the baseline value and remove it from the leaves
			SG_ERR_CHECK(  SG_resolve__choice__get_value__label(pCtx, pChoice, SG_RESOLVE__LABEL__BASELINE, 0u, &pBaseline)  );
			SG_ERR_CHECK(  SG_vector__remove__value(pCtx, pChoice->pLeafValues, pBaseline, NULL, NULL)  );

			// add the working value to the leaves
			SG_ERR_CHECK(  SG_vector__append(pCtx, pChoice->pLeafValues, (void*)pValue, NULL)  );
		}
		else
		{
			// If the working value's data HID is different than any other value's
			// then it should be a leaf, because it contains changes that the
			// user has made directly in the working directory that aren't
			// accounted for in any of the other values.
			void* pMatch = NULL;

			// check if any other values have the working value's data
			SG_ERR_CHECK(  SG_vector__find__first(pCtx, pChoice->pValues, _value__vector_predicate__match_data, pValue, NULL, &pMatch)  );
			if (pMatch == NULL)
			{
				// didn't find any, add the working value to the leaves
				SG_ERR_CHECK(  SG_vector__append(pCtx, pChoice->pLeafValues, (void*)pValue, NULL)  );
			}
		}

		// append the working value to the choice's value list
		SG_ERR_CHECK(  SG_vector__append(pCtx, pChoice->pValues, (void*)pValue, NULL)  );
		pValue = NULL;
	}

fail:
	_VALUE__NULLFREE(pCtx, pValue);
	return;
}

/**
 * Allocates and populates a choice of the given type from an issue, if necessary.
 */
static void _choice__alloc(
	SG_context*                    pCtx,    //< [in] [out] Error and context info.
	const SG_vhash*                pIssue,  //< [in] The issue we're allocating from.
	const SG_treediff2_ObjectData* pDiff,   //< [in] Diff to read current item data from.
	                                        //<      NULL if the item doesn't appear in the WD or baseline changesets.
	const SG_vhash*                pSaved,  //< [in] Saved resolve data about the choice.
	                                        //<      May be NULL if there is no saved data for the choice.
	SG_resolve__item*              pItem,   //< [in] The item that will own the choice.
	SG_resolve__state              eState,  //< [in] The type of choice to create.
	SG_resolve__choice**           ppChoice //< [out] The created choice, or NULL if none is necessary.
	)
{
	SG_resolve__choice* pChoice         = NULL;
	SG_int64            iConflictFlags  = 0;
	SG_uint32           uConflictFlags  = 0u;
	SG_uint32           uConflictCauses = 0u;
	SG_bool             bCollision      = SG_FALSE;
	SG_bool             bCollisionCause = SG_FALSE;

	SG_NULLARGCHECK(pIssue);
	SG_NULLARGCHECK(pItem);
	SG_NULLARGCHECK(ppChoice);

	// look up the issue data we need to determine cause
	SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pIssue, gszIssueKey_ConflictFlags, &iConflictFlags)  );
	uConflictFlags = (SG_uint32)iConflictFlags;
	SG_ERR_CHECK(  SG_vhash__get__bool(pCtx, pIssue, gszIssueKey_Collision, &bCollision)  );

	// check if there are any causes for this type of choice
	uConflictCauses = uConflictFlags & gaConflictFlagMasks[eState];
	bCollisionCause = (bCollision == SG_TRUE && gaCollisionMasks[eState] == SG_TRUE) ? SG_TRUE : SG_FALSE;

	// create a choice if there is a cause for it
	if (uConflictCauses != 0u || bCollisionCause == SG_TRUE)
	{
		SG_bool     bExists           = SG_FALSE;
		SG_bool     bValuesEquivalent = SG_FALSE;
		const char* szResultLabel     = NULL;

		// allocate the choice and set its basic info
		SG_ERR_CHECK(  SG_alloc1(pCtx, pChoice)  );
		pChoice->pItem           = pItem;
		pChoice->eState          = eState;
		pChoice->eConflictCauses = uConflictCauses;
		pChoice->bCollisionCause = bCollisionCause;
		pChoice->pTempPath       = NULL;
		pChoice->pValues         = NULL;
		pChoice->pLeafValues     = NULL;
		pChoice->pResult         = NULL;
		pChoice->pCollisionItems = NULL;
		pChoice->pCycleItems     = NULL;
		pChoice->szCycleHint     = NULL;

		// if the issue has a path for temporary files, retrieve it
		SG_ERR_CHECK(  SG_vhash__has(pCtx, pIssue, gszIssueKey_TempPath, &bExists)  );
		if (bExists == SG_TRUE)
		{
			const char* szTempPath = NULL;

			SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pIssue, gszIssueKey_TempPath, &szTempPath)  );
			SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path2(pCtx, pItem->pResolve->pPendingTree, szTempPath, &pChoice->pTempPath)  );
		}

		// if there's a cycle involved and the issue contains a hint, retrieve it
		SG_ERR_CHECK(  SG_vhash__has(pCtx, pIssue, gszIssueKey_Cycle_Hint, &bExists)  );
		if (bExists == SG_TRUE && ((uConflictCauses & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__MOVES_CAUSED_PATH_CYCLE) != 0))
		{
			const char* szCycleHint = NULL;

			SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pIssue, gszIssueKey_Cycle_Hint, &szCycleHint)  );
			SG_ERR_CHECK(  SG_strdup(pCtx, szCycleHint, &pChoice->szCycleHint)  );
		}

		// allocate our value vectors
		SG_ERR_CHECK(  SG_VECTOR__ALLOC(pCtx, &pChoice->pValues, 4u)  );
		SG_ERR_CHECK(  SG_VECTOR__ALLOC(pCtx, &pChoice->pLeafValues, 4u)  );

		// create values associated with changesets, as specified in the merge issue
		SG_ERR_CHECK(  _choice__alloc__changeset_values(pCtx, pIssue, pChoice)  );

		// check if all the values created so far have the same data, in which case the choice is unnecessary
		// Note that we're doing this while only the changeset values have been
		// allocated, because those are the only ones we want to check.
		SG_ERR_CHECK(  _choice__values_equivalent(pCtx, pChoice->pValues, &bValuesEquivalent)  );
		if (bValuesEquivalent == SG_TRUE)
		{
			// This should only ever occur with collisions, because they can be caused by
			// add/add situations or various combinations of moves and renames.  Therefore, we
			// need to try setting up each potential choice (existence, location, and name) and then
			// abort here if it turns out that the choice is unnecessary, which would happen if the
			// collision wasn't actually caused by a particular action (for instance, the location
			// choice would be unneeded if the collision was caused entirely by renames, and we'll
			// detect that here because the attempted location choice will have all equivalent values).
			SG_ASSERT(bCollisionCause == SG_TRUE);
			_CHOICE__NULLFREE(pCtx, pChoice);
		}
		else
		{
			// create any values that were automerged
			SG_ERR_CHECK(  _choice__alloc__automerge_values(pCtx, pIssue, pChoice)  );

			// if we have saved resolve data, load the values from it
			if (pSaved != NULL)
			{
				SG_ERR_CHECK(  _choice__alloc__saved_values(pCtx, pSaved, pChoice)  );
			}

			// create a value for the working directory
			// Note: This has to come after we allocate saved values in order for the
			//       leaf nodes to be properly kept up to date.
			SG_ERR_CHECK(  _choice__alloc__working_values(pCtx, pIssue, pDiff, pChoice)  );

			// if we have a result value label saved, get a pointer to the actual value
			// Note: This has to come after we allocate the working value(s), because
			//       the saved result label might refer to a working value that's not
			//       saved.
			if (pSaved != NULL)
			{
				SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pSaved, gszChoiceKey_Result, &szResultLabel)  );
				if (szResultLabel != NULL)
				{
					SG_ERR_CHECK(  SG_resolve__choice__get_value__label(pCtx, pChoice, szResultLabel, 0u, &pChoice->pResult)  );
				}
			}

			// check if the choice is mergeable
			if ((uConflictCauses & gaMergeableMasks[pItem->eType]) == 0)
			{
				// this choice isn't mergeable, free the leaf values
				// a NULL leaf value list indicates later that this choice isn't mergeable
				SG_VECTOR_NULLFREE(pCtx, pChoice->pLeafValues);
				SG_ASSERT(pChoice->pLeafValues == NULL);
			}
			else if (pChoice->pResult == NULL)
			{
				// Check if we should mark the choice as already resolved.
				// We'll do that if the choice has a single leaf value that was from an automerge.
				// What this would mean is that "merge" did the resolve work for us.
				SG_resolve__value* pLeaf = NULL;

				// get the single leaf and see if it's an automerge
				SG_ERR_CHECK(  _choice__get_only_leaf(pCtx, pChoice, &pLeaf)  );
				if (pLeaf != NULL && pLeaf->bAutoMerge == SG_TRUE)
				{
					// that leaf will be our resolution
					pChoice->pResult = pLeaf;
				}
			}

			// if the choice is mergeable and has already been resolved
			// then we no longer care to track what its leaves are
			// because that information is only helpful to arrive at a result
			if (pChoice->pLeafValues != NULL && pChoice->pResult != NULL)
			{
				// clear the vector of leaves, but don't free it
				// a NULL vector would mean a non-mergeable choice
				// we want to know the choice is mergeable
				// we just don't want it to have leaves anymore
				SG_ERR_CHECK(  SG_vector__clear(pCtx, pChoice->pLeafValues)  );
			}
		}
	}

	// return the choice
	*ppChoice = pChoice;
	pChoice = NULL;

fail:
	_CHOICE__NULLFREE(pCtx, pChoice);
	return;
}

/**
 * Checks if a choice is resolved based strictly on its own data.
 * It doesn't take its item into account.
 */
void _choice__get_resolved(
	SG_context*         pCtx,     //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice,  //< [in] The choice to get the resolved state of.
	SG_bool*            pResolved //< [out] Whether or not the choice is resolved.
	)
{
	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pResolved);

	if (pChoice->pResult == NULL)
	{
		*pResolved = SG_FALSE;
	}
	else
	{
		*pResolved = SG_TRUE;
	}

fail:
	return;
}

/**
 * Creates a hash that contains the choice data that is unique to resolve.
 * Data derived from merge issues or pendingtree is not included.
 * If the given choice can be entirely constructed from merge/pendingtree data,
 * then no hash is generated at all.
 */
static void _choice__build_resolve_vhash(
	SG_context*         pCtx,    //< [in] [out] Error and context data.
	SG_resolve__choice* pChoice, //< [in] The choice to build a vhash from.
	SG_vhash**          ppHash   //< [out] The built vhash, or NULL if none is required.
	)
{
	SG_uint32  uCount  = 0u;
	SG_vhash*  pHash   = NULL;
	SG_varray* pValues = NULL;

	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(ppHash);

	// if there are any values for this choice, build a hash for it
	SG_ERR_CHECK(  SG_vector__length(pCtx, pChoice->pValues, &uCount)  );
	if (uCount > 0u)
	{
		// allocate a vhash to store the choice's properties in
		SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pHash)  );

		// add the choice's basic properties to the hash
		if (pChoice->pResult != NULL)
		{
			SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pHash, gszChoiceKey_Result, pChoice->pResult->szLabel)  );
		}

		// add the choice's values to a varray
		SG_ERR_CHECK(  SG_VARRAY__ALLOC(pCtx, &pValues)  );
		SG_ERR_CHECK(  SG_resolve__choice__foreach_value(pCtx, pChoice, SG_FALSE, _value__resolve_action__add_resolve_vhash_to_varray, (void*)pValues)  );

		// if any values were saved to the array, add it to the hash (the hash takes ownership of the array)
		SG_ERR_CHECK(  SG_varray__count(pCtx, pValues, &uCount)  );
		if (uCount > 0u)
		{
			SG_ERR_CHECK(  SG_vhash__add__varray(pCtx, pHash, gszChoiceKey_Values, &pValues)  );
			SG_ASSERT(pValues == NULL);
		}
	}

	// return the hash
	*ppHash = pHash;
	pHash = NULL;

fail:
	SG_VHASH_NULLFREE(pCtx, pHash);
	SG_VARRAY_NULLFREE(pCtx, pValues);
	return;
}

/**
 * An SG_resolve__choice__callback that builds a vhash out of each choice
 * and adds it to a given vhash using the choice's state as the key.
 */
static void _choice__resolve_action__add_resolve_vhash_to_vhash(
	SG_context*         pCtx,      //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice,   //< [in] The current choice.
	void*               pUserData, //< [in] The SG_vhash to add the choice's vhash to.
	SG_bool*            pContinue  //< [out] Whether or not to continue iteration.
	)
{
	SG_vhash* pChoices = (SG_vhash*)pUserData;
	SG_vhash* pHash    = NULL;

	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pUserData);
	SG_NULLARGCHECK(pContinue);

	// build a hash from the choice
	SG_ERR_CHECK(  _choice__build_resolve_vhash(pCtx, pChoice, &pHash)  );

	// if we got a hash back, add it to the supplied hash of choices
	// (the hash takes ownership of the other hash)
	if (pHash != NULL)
	{
		SG_ERR_CHECK(  SG_vhash__add__vhash(pCtx, pChoices, gaSerializedStateValues[pChoice->eState], &pHash)  );
		SG_ASSERT(pHash == NULL);
	}

	// continue iterating
	*pContinue = SG_TRUE;

fail:
	SG_VHASH_NULLFREE(pCtx, pHash);
	return;
}

/**
 * An SG_resolve__choice__callback that increments a counter.
 */
static void _choice__resolve_action__increment_count(
	SG_context*         pCtx,      //< [in] [out] Error and context info.
	SG_resolve__choice* pChoice,   //< [in] The current choice.
	void*               pUserData, //< [in] Pointer to an SG_uint32 to increment.
	SG_bool*            pContinue  //< [out] Whether or not to continue iteration.
	)
{
	SG_uint32* pCounter = (SG_uint32*)pUserData;

	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pUserData);
	SG_NULLARGCHECK(pContinue);

	*pCounter += 1u;

fail:
	return;
}

/**
 * An SG_free_callback that frees choices.
 *//*
static void _choice__sg_action__free(
	SG_context* pCtx, //< [in] [out] Error and context info.
	void*       pData //< [in] The SG_resolve__choice to free.
	)
{
	SG_resolve__choice* pChoice = (SG_resolve__choice*)pData;
	_CHOICE__NULLFREE(pCtx, pChoice);
}
*/

/*
**
** Internal Functions - Item
**
*/

/**
 * Frees an SG_resolve__item.
 */
static void _item__free(
	SG_context*       pCtx, //< [in] [out] Error and context info.
	SG_resolve__item* pItem //< [in] The item to free.
	)
{
	SG_uint32 uIndex = 0u;

	if (pItem == NULL)
	{
		return;
	}

	SG_NULLFREE(pCtx, pItem->szGid);
	SG_NULLFREE(pCtx, pItem->szAncestorHid);
	SG_VHASH_NULLFREE(pCtx, pItem->pRepoPaths);
	SG_NULLFREE(pCtx, pItem->szMergedName);
	SG_NULLFREE(pCtx, pItem->szMergedContentHid);
	SG_NULLFREE(pCtx, pItem->szMergedXAttrsHid);
	SG_NULLFREE(pCtx, pItem->szOriginalName);
	SG_NULLFREE(pCtx, pItem->szOriginalLocation);
	for (uIndex = 0u; uIndex < SG_RESOLVE__STATE__COUNT; ++uIndex)
	{
		_CHOICE__NULLFREE(pCtx, pItem->aChoices[uIndex]);
	}
	pItem->pResolve          = NULL;
	pItem->eType             = SG_TREENODEENTRY_TYPE__INVALID;
	pItem->iMergedAttributes = 0;
	SG_NULLFREE(pCtx, pItem);
}

/**
 * NULLFREE wrapper around _item__free.
 */
#define _ITEM__NULLFREE(pCtx, pItem) _sg_generic_nullfree(pCtx, pItem, _item__free)

/**
 * Allocates an item from a merge issue.
 */
static void _item__alloc(
	SG_context*         pCtx,     //< [in] [out] Error and context info.
	const SG_vhash*     pIssue,   //< [in] The issue we're allocating from.
	SG_treediff2*       pDiff,    //< [in] A diff between the pendingtree and baseline parent.
	SG_resolve*         pResolve, //< [in] The resolve that will own the item.
	SG_resolve__item**  ppItem    //< [out] The created item.
	)
{
	SG_resolve__item*              pItem       = NULL;
	const char*                    szGid       = NULL;
	SG_int64                       iType       = 0;
	SG_bool                        bDiffObject = SG_FALSE;
	const SG_treediff2_ObjectData* pDiffObject = NULL;
	SG_vhash*                      pInputs     = NULL;
	SG_uint32                      uCount      = 0u;
	SG_uint32                      uIndex      = 0u;
	SG_vhash*                      pOutput     = NULL;
	const char*                    szName      = NULL;
	const char*                    szHid       = NULL;
	SG_string*                     sRepoPath   = NULL;
	const SG_vhash*                pSavedItems = NULL;
	SG_vhash*                      pSavedItem  = NULL;

	SG_NULLARGCHECK(pIssue);
	SG_NULLARGCHECK(pResolve);
	SG_NULLARGCHECK(ppItem);

	// allocate the item
	SG_ERR_CHECK(  SG_alloc1(pCtx, pItem)  );
	pItem->pResolve           = pResolve;
	pItem->szGid              = NULL;
	pItem->eType              = SG_TREENODEENTRY_TYPE__INVALID;
	pItem->eStatus            = SG_DIFFSTATUS_FLAGS__ZERO;
	pItem->szAncestorHid      = NULL;
	pItem->pRepoPaths         = NULL;
	pItem->szMergedName       = NULL;
	pItem->szMergedContentHid = NULL;
	pItem->iMergedAttributes  = 0;
	pItem->szMergedXAttrsHid  = NULL;
	pItem->szOriginalName     = NULL;
	pItem->szOriginalLocation = NULL;
	for (uIndex = 0u; uIndex < SG_RESOLVE__STATE__COUNT; ++uIndex)
	{
		pItem->aChoices[uIndex] = NULL;
	}

	// get the item's GID and type from the issue
	SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pIssue, gszIssueKey_Gid, &szGid)  );
	SG_ERR_CHECK(  SG_strdup(pCtx, szGid, &pItem->szGid)  );
	SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pIssue, gszIssueKey_Type, &iType)  );
	pItem->eType = (SG_treenode_entry_type)iType;

	// find this item's data in the diff
	SG_ERR_CHECK(  SG_treediff2__get_ObjectData(pCtx, pDiff, pItem->szGid, &bDiffObject, &pDiffObject)  );
	SG_ASSERT(bDiffObject == SG_TRUE || pDiffObject == NULL);

	// if the item appears in the diff, get its diff flags
	if (pDiffObject != NULL)
	{
		SG_ERR_CHECK(  SG_treediff2__ObjectData__get_dsFlags(pCtx, pDiffObject, &pItem->eStatus)  );
	}

	// allocate a hash for the repo paths
	SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pItem->pRepoPaths)  );

	// gather item-wide data from each of the issue's inputs
	SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pIssue, gszIssueKey_Inputs, &pInputs)  );
	SG_ERR_CHECK(  SG_vhash__count(pCtx, pInputs, &uCount)  );
	for (uIndex = 0u; uIndex < uCount; ++uIndex)
	{
		const char* szHid      = NULL;
		SG_vhash*   pInput     = NULL;
		const char* szRepoPath = NULL;
		const char* szName     = NULL;
		const char* szLocation = NULL;
		SG_bool     bParent    = SG_FALSE;

		// get the current input, containing the state of the item in one of the parent changesets
		SG_ERR_CHECK(  SG_vhash__get_nth_pair__vhash(pCtx, pInputs, uIndex, &szHid, &pInput)  );

		// get the repo path from this input and add it to the item's hash
		SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pInput, gszIssueKey_Inputs_RepoPath, &szRepoPath)  );
		SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pItem->pRepoPaths, szHid, szRepoPath)  );

		// check if the current HID appears in the resolve's list of parent HIDs
		// if it doesn't, we'll assume it's the ancestor
		SG_ERR_CHECK(  SG_varray__find__sz(pCtx, pResolve->pParentHids, szHid, &bParent, NULL)  );
		if (bParent == SG_FALSE)
		{
			// we should never find two HIDs that aren't parents
			SG_ASSERT(pItem->szAncestorHid == NULL);
			SG_ERR_CHECK(  SG_strdup(pCtx, szHid, &pItem->szAncestorHid)  );
		}
		else
		{
			// get the name the item had in this parent
			// The goal is to find the common name this item had in all parents it existed in.
			// This is the name that we'll eventually revert it back to if merge assigned the item a temporary name.
			// If the item had different names in different parents, there will be a NAME choice on the item
			// and the user will decide which name to use when resolving that choice, so we won't have to revert.
			SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pInput, gszIssueKey_Inputs_Name, &szName)  );
			if (pItem->szOriginalName == NULL && szName != NULL)
			{
				// we haven't found the item's original name yet and it had a name in this parent
				// so we'll use that as the original name for now
				SG_ERR_CHECK(  SG_strdup(pCtx, szName, &pItem->szOriginalName)  );
			}
			else if (
				   pItem->szOriginalName != NULL
				&& szName != NULL
				&& strcmp(pItem->szOriginalName, szName) != 0
				)
			{
				// the name from this parent doesn't match the name we already had from another parent
				// When this occurs, we should end up with a NAME choice later on
				// meaning we should not need to know the item's original value.
				// So we'll set it back to NULL, and the code that does the revert
				// (which should not be necessary in this case) will blow up if it
				// finds that this is NULL.
				SG_NULLFREE(pCtx, pItem->szOriginalName);
				SG_ASSERT(pItem->szOriginalName == NULL);
			}

			// get the location the item had in this parent
			// this uses the same logic as above for finding the original name, see there for details
			SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pInput, gszIssueKey_Inputs_Location, &szLocation)  );
			if (pItem->szOriginalLocation == NULL && szLocation != NULL)
			{
				SG_ERR_CHECK(  SG_strdup(pCtx, szLocation, &pItem->szOriginalLocation)  );
			}
			else if (
				   pItem->szOriginalLocation != NULL
				&& szLocation != NULL
				&& strcmp(pItem->szOriginalLocation, szLocation) != 0
				)
			{
				SG_NULLFREE(pCtx, pItem->szOriginalLocation);
				SG_ASSERT(pItem->szOriginalLocation == NULL);
			}
		}
	}

	// gather data from the item's merge output
	SG_ERR_CHECK(  SG_vhash__get__vhash(pCtx, pIssue, gszIssueKey_Output, &pOutput)  );
	SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pOutput, gszIssueKey_Output_Name, &szName)  );
	SG_ERR_CHECK(  SG_strdup(pCtx, szName, &pItem->szMergedName)  );
	SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pOutput, gszIssueKey_Output_Contents, &szHid)  );
	SG_ERR_CHECK(  SG_strdup(pCtx, szHid, &pItem->szMergedContentHid)  );
	SG_ERR_CHECK(  SG_vhash__get__int64(pCtx, pOutput, gszIssueKey_Output_Attributes, &pItem->iMergedAttributes)  );
	SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pOutput, gszIssueKey_Output_XAttributes, &szHid)  );
	SG_ERR_CHECK(  SG_strdup(pCtx, szHid, &pItem->szMergedXAttrsHid)  );

	// get the item's current repo path
	if ((pItem->eStatus & SG_DIFFSTATUS_FLAGS__DELETED) == 0)
	{
		SG_ERR_CHECK(  SG_pendingtree__get_repo_path(pCtx, pResolve->pPendingTree, szGid, &sRepoPath)  );
		SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pItem->pRepoPaths, SG_RESOLVE__HID__WORKING, SG_string__sz(sRepoPath))  );
	}

	// find any saved data we have for this item
	SG_ERR_CHECK(  SG_pendingtree__get_wd_choices__ref(pCtx, pResolve->pPendingTree, NULL, &pSavedItems)  );
	if (pSavedItems != NULL)
	{
		SG_ERR_CHECK(  SG_vhash__check__vhash(pCtx, pSavedItems, pItem->szGid, &pSavedItem)  );
	}

	// run through each possible type of choice and allocate one if necessary
	for (uIndex = 0u; uIndex < SG_RESOLVE__STATE__COUNT; ++uIndex)
	{
		SG_vhash* pSavedChoice = NULL;

		// check if we have any saved data for this choice
		if (pSavedItem != NULL)
		{
			SG_ERR_CHECK(  SG_vhash__check__vhash(pCtx, pSavedItem, gaSerializedStateValues[uIndex], &pSavedChoice)  );
		}

		// allocate the choice
		SG_ERR_CHECK(  _choice__alloc(pCtx, pIssue, pDiffObject, pSavedChoice, pItem, (SG_resolve__state)uIndex, &(pItem->aChoices[uIndex]))  );
	}

	// return the new item
	*ppItem = pItem;
	pItem = NULL;

fail:
	SG_STRING_NULLFREE(pCtx, sRepoPath);
	_ITEM__NULLFREE(pCtx, pItem);
	return;
}

/**
 * Checks if an item is resolved based ONLY on its own data.
 * It doesn't take its choices or resolve-wide data into account.
 */
static void _item__get_resolved(
	SG_context*       pCtx,      //< [in] [out] Error and context info.
	SG_resolve__item* pItem,     //< [in] The item to get the resolved state of.
	SG_bool*          pResolved, //< [out] True if the item considers itself resolved, false otherwise.
	SG_bool*          pDeleted   //< [out] If pResolved is set to true, this indicates whether or not
	                             //<       the resolution is due to deletion.  Currently if pResolved
	                             //<       is true this will always be set to true, because deletion
	                             //<       is currently the only way for an item to be resolved at the
	                             //<       item level (without taking its choices into account).  If
	                             //<       pResolved is set to false, this is ignored.
	)
{
	SG_bool bDeleted = SG_FALSE;

	SG_UNUSED(pCtx);
	SG_NULLARGCHECK(pItem);
	SG_NULLARGCHECK(pResolved);

	// Note: This check is currently commented out because we can't currently
	//       consider DELETED (or LOST or anything else) items as resolved.  To do
	//       so currently would require our having a good opportunity to update the
	//       old resolve data in pendingtree to mark such conflicts resolved.  The only
	//       real time to do this is during SG_resolve__alloc, but we can't actually
	//       modify and save the pendingtree at that point without invalidating it.
	//       That still wouldn't even be perfect, because you should theoretically be
	//       able to delete all conflicting items and them commit without ever running
	//       a resolve, and therefore never giving SG_resolve a chance to mark the
	//       conflicts resolved.  The real solution here is to remove the old
	//       resolve data entirely and update the rest of the code base to get
	//       conflict data from SG_resolve instead, but that is better left for later,
	//       especially since the need to consider a deleted item resolved is rare.
	// Note: We may someday want to consider LOST items resolved as well as DELETED ones.
	//       This avoids problems where the user deletes the item from the file system
	//       even though it has conflicts, and then cannot resolve them without restoring it,
	//       which they may not actually be able to do, leaving them unable to commit.
	//       The problem with considering LOST items resolved is that a LOST item
	//       might also indicate a MOVE or a RENAME that was done on the file system
	//       without informing us.  In those cases it's not appropriate to consider
	//       the item resolved just because it's name and/or path are different.  We
	//       want the user to see the resolve accept error, restore the file, do a
	//       proper move/rename via sglib, and then be able to resolve its conflicts
	//       normally (or resolve them before doing the move/rename).
//	SG_ERR_CHECK(  SG_resolve__item__has_status(pCtx, pItem, SG_DIFFSTATUS_FLAGS__DELETED, &bDeleted)  );

	// our resolved state currently equals our deleted state
	if (bDeleted == SG_TRUE)
	{
		*pResolved = SG_TRUE;
	}
	else
	{
		*pResolved = SG_FALSE;
	}

	// if we're resolved and they want the deleted state, return that
	if (*pResolved == SG_TRUE && pDeleted != NULL)
	{
		*pDeleted = bDeleted;
	}

fail:
	return;
}

/**
 * Creates a hash that contains the item data that is unique to resolve.
 * Data derived from merge issues or pendingtree is not included.
 * If the given item can be entirely constructed from merge/pendingtree data,
 * then no hash is generated at all.
 */
static void _item__build_resolve_vhash(
	SG_context*       pCtx,  //< [in] [out] Error and context data.
	SG_resolve__item* pItem, //< [in] The item to build a vhash from.
	SG_vhash**        ppHash //< [out] The built vhash, or NULL if none is required.
	)
{
	SG_uint32 uCount = 0u;
	SG_vhash* pHash  = NULL;
	SG_uint32 uIndex = 0u;

	SG_NULLARGCHECK(pItem);
	SG_NULLARGCHECK(ppHash);

	// check how many choices the item has
	for (uIndex = 0u; uIndex < SG_NrElements(pItem->aChoices); ++uIndex)
	{
		if (pItem->aChoices[uIndex] != NULL)
		{
			uCount += 1u;
		}
	}

	// if we have any choices, then create a hash for the item
	if (uCount > 0u)
	{
		// allocate a vhash to store the item's properties in
		SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pHash)  );

		// run through each choice in the item
		for (uIndex = 0u; uIndex < SG_NrElements(pItem->aChoices); ++uIndex)
		{
			// get the current choice
			SG_resolve__choice* pChoice = pItem->aChoices[uIndex];

			// if there is a choice for this state, add its hash to our hash
			if (pChoice != NULL)
			{
				SG_bool bContinue = SG_TRUE;

				SG_ERR_CHECK(  _choice__resolve_action__add_resolve_vhash_to_vhash(pCtx, pChoice, (void*)pHash, &bContinue)  );
			}
		}
	}

	// return the hash
	*ppHash = pHash;
	pHash = NULL;

fail:
	SG_VHASH_NULLFREE(pCtx, pHash);
	return;
}

/**
 * If the item is resolved, updates the old resolve data to mark it resolved there as well.
 * TODO: Remove this function once the old resolve code is gone.
 */
static void _item__update_old_issue_status(
	SG_context*       pCtx, //< [in] [out] Error and context info.
	SG_resolve__item* pItem //< [in] The item to update the old resolve status of.
	)
{
	SG_bool          bResolved = SG_FALSE;
	SG_bool          bIssues   = SG_FALSE;
	const SG_varray* pIssues   = NULL;
	SG_uint32        uCount    = 0u;
	SG_uint32        uIndex    = 0u;
	SG_vhash*        pIssue    = NULL;

	SG_NULLARGCHECK(pItem);

	// check if the item is resolved
	SG_ERR_CHECK(  SG_resolve__item__get_resolved(pCtx, pItem, &bResolved, NULL)  );
	if (bResolved == SG_TRUE)
	{
		// find the issue that corresponds to this item
		SG_ERR_CHECK(  SG_pendingtree__get_wd_issues__ref(pCtx, pItem->pResolve->pPendingTree, &bIssues, &pIssues)  );
		SG_ASSERT(bIssues == SG_TRUE);
		SG_ERR_CHECK(  SG_varray__count(pCtx, pIssues, &uCount)  );
		for (uIndex = 0u; uIndex < uCount; ++uIndex)
		{
			const char* szGid = NULL;

			SG_ERR_CHECK(  SG_varray__get__vhash(pCtx, pIssues, uIndex, &pIssue)  );
			SG_ERR_CHECK(  SG_vhash__get__sz(pCtx, pIssue, gszIssueKey_Gid, &szGid)  );
			if (SG_stricmp(pItem->szGid, szGid) == 0)
			{
				break;
			}
		}

		// we should always find an issue, otherwise where did the item come from?
		SG_ASSERT(pIssue != NULL);

		// update the status
		SG_ERR_CHECK(  SG_pendingtree__set_wd_issue_status__dont_save_pendingtree(pCtx, pItem->pResolve->pPendingTree, pIssue, SG_ISSUE_STATUS__MARKED_RESOLVED)  );
	}

fail:
	return;
}

/**
 * Iterates over each choice in an item.
 */
static void _item__foreach_choice(
	SG_context*                   pCtx,            //< [in] [out] Error and context info.
	SG_resolve__item*             pItem,           //< [in] The item to iterate choices on.
	SG_bool                       bOnlyUnresolved, //< [in] Whether or not only unresolved choices should be iterated.
	SG_resolve__choice__callback* fCallback,       //< [in] The callback to call with each choice.
	void*                         pUserData,       //< [in] User data to pass to the callback.
	SG_bool*                      pContinue        //< [in] [out] Whether or not to continue iteration.
	)
{
	SG_int32 iIndex = 0;

	SG_NULLARGCHECK(pItem);
	SG_NULLARGCHECK(fCallback);
	SG_NULLARGCHECK(pContinue);

	for (iIndex = 0; iIndex < SG_RESOLVE__STATE__COUNT; ++iIndex)
	{
		SG_resolve__choice* pChoice = pItem->aChoices[iIndex];

		if (pChoice == NULL)
		{
			continue;
		}

		if (bOnlyUnresolved == SG_TRUE && pChoice->pResult != NULL)
		{
			continue;
		}

		SG_ERR_CHECK(  fCallback(pCtx, pChoice, pUserData, pContinue)  );
		
		if (*pContinue == SG_FALSE)
		{
			break;
		}
	}

fail:
	return;
}

/**
 * An SG_resolve__item__callback that builds a vhash out of each item
 * and adds it to a given vhash using the item's GID as the key.
 */
static void _item__resolve_action__add_resolve_vhash_to_vhash(
	SG_context*       pCtx,      //< [in] [out] Error and context info.
	SG_resolve__item* pItem,     //< [in] The current item.
	void*             pUserData, //< [in] The SG_vhash to add the item's vhash to.
	SG_bool*          pContinue  //< [out] Whether or not to continue iteration.
	)
{
	SG_vhash* pItems = (SG_vhash*)pUserData;
	SG_vhash* pHash  = NULL;

	SG_NULLARGCHECK(pItem);
	SG_NULLARGCHECK(pUserData);
	SG_NULLARGCHECK(pContinue);

	// build a hash from the item
	SG_ERR_CHECK(  _item__build_resolve_vhash(pCtx, pItem, &pHash)  );

	// if we got a hash back, add it to the supplied hash of items
	// (the hash takes ownership of the other hash)
	if (pHash != NULL)
	{
		SG_ERR_CHECK(  SG_vhash__add__vhash(pCtx, pItems, pItem->szGid, &pHash)  );
		SG_ASSERT(pHash == NULL);
	}

	// continue iterating
	*pContinue = SG_TRUE;

fail:
	SG_VHASH_NULLFREE(pCtx, pHash);
	return;
}

/**
 * An SG_free_callback that frees SG_resolve__item values.
 */
static void _item__sg_action__free(
	SG_context* pCtx, //< [in] [out] Error and context info.
	void*       pData //< [in] The SG_resolve__item to free.
	)
{
	SG_resolve__item* pItem  = (SG_resolve__item*)pData;
	_ITEM__NULLFREE(pCtx, pItem);
}

/**
 * An SG_vector_predicate that matches an SG_resolve__item's
 * szGid against the string given in the user data.
 */
static void _item__vector_predicate__match_gid(
	SG_context* pCtx,      //< [in] [out] Error and context info.
	SG_uint32   uIndex,    //< [in] The index of the item being matched.
	void*       pValue,    //< [in] The item being matched.
	void*       pUserData, //< [in] The string to match the item's GID against.
	SG_bool*    pMatch     //< [out] Whether or not the item's GID matches.
	)
{
	SG_resolve__item* pItem      = (SG_resolve__item*)pValue;
	const char*       szMatchGid = (const char*)pUserData;
	const char*       szItemGid  = NULL;

	SG_UNUSED(uIndex);
	SG_NULLARGCHECK(pValue);
	SG_NULLARGCHECK(pUserData);
	SG_NULLARGCHECK(pMatch);

	SG_ERR_CHECK(  SG_resolve__item__get_gid(pCtx, pItem, &szItemGid)  );

	if (SG_stricmp(szMatchGid, szItemGid) == 0)
	{
		*pMatch = SG_TRUE;
	}
	else
	{
		*pMatch = SG_FALSE;
	}

fail:
	return;
}

/**
 * An SG_vector_predicate that matches an SG_resolve__item's
 * working repo path against the string given in the user data.
 */
static void _item__vector_predicate__match_working_path(
	SG_context* pCtx,      //< [in] [out] Error and context info.
	SG_uint32   uIndex,    //< [in] The index of the item being matched.
	void*       pValue,    //< [in] The item being matched.
	void*       pUserData, //< [in] The string to match the item's path against.
	SG_bool*    pMatch     //< [out] Whether or not the item's path matches.
	)
{
	SG_resolve__item* pItem           = (SG_resolve__item*)pValue;
	const char*       szMatchRepoPath = (const char*)pUserData;
	const char*       szItemRepoPath  = NULL;
	SG_string*        sMatchRepoPath  = NULL;
	SG_string*        sItemRepoPath   = NULL;
	SG_int32          iResult         = 0;

	SG_UNUSED(uIndex);
	SG_NULLARGCHECK(pValue);
	SG_NULLARGCHECK(pUserData);
	SG_NULLARGCHECK(pMatch);

	SG_ERR_CHECK(  SG_resolve__item__get_repo_path__working(pCtx, pItem, &szItemRepoPath)  );
	if (szItemRepoPath != NULL)
	{
		SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &sMatchRepoPath, szMatchRepoPath)  );
		SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &sItemRepoPath, szItemRepoPath)  );
		SG_ERR_CHECK(  SG_repopath__remove_final_slash(pCtx, sMatchRepoPath)  );
		SG_ERR_CHECK(  SG_repopath__remove_final_slash(pCtx, sItemRepoPath)  );
		SG_ERR_CHECK(  SG_repopath__compare(pCtx, SG_string__sz(sMatchRepoPath), SG_string__sz(sItemRepoPath), &iResult)  );
	}

	if (szItemRepoPath != NULL && iResult == 0)
	{
		*pMatch = SG_TRUE;
	}
	else
	{
		*pMatch = SG_FALSE;
	}

fail:
	SG_STRING_NULLFREE(pCtx, sMatchRepoPath);
	SG_STRING_NULLFREE(pCtx, sItemRepoPath);
	return;
}

/**
 * An SG_vector_predicate that matches an SG_resolve__item's
 * resolve status against the given status.
 */
static void _item__vector_predicate__match_resolved(
	SG_context* pCtx,      //< [in] [out] Error and context info.
	SG_uint32   uIndex,    //< [in] The index of the item being matched.
	void*       pValue,    //< [in] The item being matched.
	void*       pUserData, //< [in] Pointer to a bool resolved status to match.
	SG_bool*    pMatch     //< [out] Whether or not the item's resolved status matches.
	)
{
	SG_resolve__item* pItem          = (SG_resolve__item*)pValue;
	SG_bool*          pMatchResolved = (SG_bool*)pUserData;
	SG_bool           bItemResolved  = SG_FALSE;

	SG_UNUSED(uIndex);
	SG_NULLARGCHECK(pValue);
	SG_NULLARGCHECK(pUserData);
	SG_NULLARGCHECK(pMatch);

	SG_ERR_CHECK(  SG_resolve__item__get_resolved(pCtx, pItem, &bItemResolved, NULL)  );

	if (*pMatchResolved == bItemResolved)
	{
		*pMatch = SG_TRUE;
	}
	else
	{
		*pMatch = SG_FALSE;
	}

fail:
	return;
}

/**
 * Type of user data required by _item__vector_action__call_resolve_action.
 */
typedef struct
{
	SG_resolve__item__callback* fCallback; //< The item callback to call.
	void*                       pUserData; //< The user data to pass to the callback.
	SG_bool                     bContinue; //< Whether or not to actually call the callback.
}
_item__vector_action__call_resolve_action__data;

/**
 * An SG_vector_foreach_callback that calls an SG_resolve__item__callback.
 */
static void _item__vector_action__call_resolve_action(
	SG_context* pCtx,     //< [in] [out] Error and context info.
	SG_uint32   uIndex,   //< [in] The index of the item.
	void*       pValue,   //< [in] The item.
	void*       pUserData //< [in] An _item__vector_action__call_resolve_action__data structure.
	)
{
	SG_resolve__item*                                pItem = (SG_resolve__item*)pValue;
	_item__vector_action__call_resolve_action__data* pArgs = (_item__vector_action__call_resolve_action__data*)pUserData;

	SG_UNUSED(uIndex);
	SG_NULLARGCHECK(pValue);
	SG_NULLARGCHECK(pUserData);

	if (pArgs->bContinue == SG_TRUE)
	{
		SG_ERR_CHECK(  pArgs->fCallback(pCtx, pItem, pArgs->pUserData, &pArgs->bContinue)  );
	}

fail:
	return;
}

/**
 * Type of user data required by _item__vector_action__foreach_choice.
 */
typedef struct
{
	SG_bool                       bOnlyUnresolved; //< Whether or not we should only iterate over unresolved choices.
	SG_resolve__choice__callback* fCallback;       //< The choice callback to call on each choice.
	void*                         pUserData;       //< The user data to pass to the callback.
	SG_bool                       bContinue;       //< Whether or not to actually call the callback.
}
_item__vector_action__foreach_choice__data;

/**
 * An SG_vector_foreach_callback that iterates over the item's choices.
 */
static void _item__vector_action__foreach_choice(
	SG_context* pCtx,     //< [in] [out] Error and context info.
	SG_uint32   uIndex,   //< [in] The index of the item.
	void*       pValue,   //< [in] The item.
	void*       pUserData //< [in] An _item__vector_action__foreach_choice__data structure.
	)
{
	SG_resolve__item*                           pItem = (SG_resolve__item*)pValue;
	_item__vector_action__foreach_choice__data* pArgs = (_item__vector_action__foreach_choice__data*)pUserData;

	SG_UNUSED(uIndex);
	SG_NULLARGCHECK(pValue);
	SG_NULLARGCHECK(pUserData);

	if (pArgs->bContinue == SG_TRUE)
	{
		SG_ERR_CHECK(  _item__foreach_choice(pCtx, pItem, pArgs->bOnlyUnresolved, pArgs->fCallback, pArgs->pUserData, &pArgs->bContinue)  );
	}

fail:
	return;
}

/**
 * Iterates a list of related item GIDs, finds the corresponding item for each
 * in an item vector, and adds each found item to a vector.
 */
static void _item__gather_items(
	SG_context*       pCtx,    //< [in] [out] Error and context info.
	SG_varray*        pGids,   //< [in] An array of item GIDs that are related to pItem.
	SG_vector*        pItems,  //< [in] An SG_vector of items to search for related ones by GID.
	SG_resolve__item* pItem,   //< [in] An item to ignore if found in the list of GIDs.
	                           //<      This item will never be added to either destination.
	SG_vector*        pRelated //< [in] A vector to add the related items to.
	)
{
	SG_uint32 uCount = 0u;
	SG_uint32 uIndex = 0u;

	SG_NULLARGCHECK(pGids);
	SG_NULLARGCHECK(pItems);
	SG_NULLARGCHECK(pItem);
	SG_NULLARGCHECK(pRelated);

	// run through each GID involved in the collision
	SG_ERR_CHECK(  SG_varray__count(pCtx, pGids, &uCount)  );
	for (uIndex = 0u; uIndex < uCount; ++uIndex)
	{
		const char*       szGid        = NULL;
		SG_resolve__item* pRelatedItem = NULL;

		// get the current GID
		SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pGids, uIndex, &szGid)  );

		// find the item with this GID in the search vector
		SG_ERR_CHECK(  SG_vector__find__first(pCtx, pItems, _item__vector_predicate__match_gid, (void*)szGid, NULL, (void**)&pRelatedItem)  );
		SG_ASSERT(pRelatedItem != NULL);

		// add the found item to the destination vector(s)
		// unless we found ourselves (the current item may be in the collision list as well)
		if (pRelatedItem != pItem)
		{
			SG_ERR_CHECK(  SG_vector__append(pCtx, pRelated, (void*)pRelatedItem, NULL)  );
		}
	}

fail:
	return;
}

/**
 * Finds other items related to this one and stores them in its appropriate choices.
 */
static void _item__gather_related_items(
	SG_context*       pCtx,   //< [in] [out] Error and context info.
	SG_resolve__item* pItem,  //< [in] The item to gather related items into.
	const SG_vhash*   pIssue, //< [in] The merge issue to gather related items from.
	SG_vector*        pItems  //< [in] An SG_vector of items to search for related ones.
	)
{
	SG_varray* pGids = NULL;

	SG_NULLARGCHECK(pItem);
	SG_NULLARGCHECK(pIssue);
	SG_NULLARGCHECK(pItems);

	// If we have an EXISTENCE choice caused by a collision,
	// find the other items in the collision.
	if (
		   (pItem->aChoices[SG_RESOLVE__STATE__EXISTENCE] != NULL)
		&& (pItem->aChoices[SG_RESOLVE__STATE__EXISTENCE]->bCollisionCause == SG_TRUE)
	)
	{
		SG_uint32 uCount = 0u;

		SG_ERR_CHECK(  SG_vhash__get__varray(pCtx, pIssue, gszIssueKey_Collision_Items, &pGids)  );
		SG_ERR_CHECK(  SG_varray__count(pCtx, pGids, &uCount)  );
		SG_ERR_CHECK(  SG_VECTOR__ALLOC(pCtx, &pItem->aChoices[SG_RESOLVE__STATE__EXISTENCE]->pCollisionItems, uCount)  );
		SG_ERR_CHECK(  _item__gather_items(pCtx, pGids, pItems, pItem, pItem->aChoices[SG_RESOLVE__STATE__EXISTENCE]->pCollisionItems)  );
	}

	// If we have a NAME choice caused by a collision,
	// find the other items in the collision.
	if (
		   (pItem->aChoices[SG_RESOLVE__STATE__NAME] != NULL)
		&& (pItem->aChoices[SG_RESOLVE__STATE__NAME]->bCollisionCause == SG_TRUE)
	)
	{
		SG_uint32 uCount = 0u;

		SG_ERR_CHECK(  SG_vhash__get__varray(pCtx, pIssue, gszIssueKey_Collision_Items, &pGids)  );
		SG_ERR_CHECK(  SG_varray__count(pCtx, pGids, &uCount)  );
		SG_ERR_CHECK(  SG_VECTOR__ALLOC(pCtx, &pItem->aChoices[SG_RESOLVE__STATE__NAME]->pCollisionItems, uCount)  );
		SG_ERR_CHECK(  _item__gather_items(pCtx, pGids, pItems, pItem, pItem->aChoices[SG_RESOLVE__STATE__NAME]->pCollisionItems)  );
	}

	// If we have a LOCATION choice caused by a collision,
	// find the other items in the collision.
	if (
		   (pItem->aChoices[SG_RESOLVE__STATE__LOCATION] != NULL)
		&& (pItem->aChoices[SG_RESOLVE__STATE__LOCATION]->bCollisionCause == SG_TRUE)
	)
	{
		SG_uint32 uCount = 0u;

		SG_ERR_CHECK(  SG_vhash__get__varray(pCtx, pIssue, gszIssueKey_Collision_Items, &pGids)  );
		SG_ERR_CHECK(  SG_varray__count(pCtx, pGids, &uCount)  );
		SG_ERR_CHECK(  SG_VECTOR__ALLOC(pCtx, &pItem->aChoices[SG_RESOLVE__STATE__LOCATION]->pCollisionItems, uCount)  );
		SG_ERR_CHECK(  _item__gather_items(pCtx, pGids, pItems, pItem, pItem->aChoices[SG_RESOLVE__STATE__LOCATION]->pCollisionItems)  );
	}

	// If we have a LOCATION choice caused by a path cycle,
	// find the other items in the cycle.
	if (
		   (pItem->aChoices[SG_RESOLVE__STATE__LOCATION] != NULL)
		&& ((pItem->aChoices[SG_RESOLVE__STATE__LOCATION]->eConflictCauses & SG_MRG_CSET_ENTRY_CONFLICT_FLAGS__MOVES_CAUSED_PATH_CYCLE) != 0)
	)
	{
		SG_uint32 uCount = 0u;

		SG_ERR_CHECK(  SG_vhash__get__varray(pCtx, pIssue, gszIssueKey_Cycle_Items, &pGids)  );
		SG_ERR_CHECK(  SG_varray__count(pCtx, pGids, &uCount)  );
		SG_ERR_CHECK(  SG_VECTOR__ALLOC(pCtx, &pItem->aChoices[SG_RESOLVE__STATE__LOCATION]->pCycleItems, uCount)  );
		SG_ERR_CHECK(  _item__gather_items(pCtx, pGids, pItems, pItem, pItem->aChoices[SG_RESOLVE__STATE__LOCATION]->pCycleItems)  );
	}

fail:
	return;
}


/*
**
** Internal Functions - Merge
**
*/

/**
 * If a given value represents a file in the working directory, that
 * file is copied into the choice's temporary path, if it's not already there.
 * The value's file reference (pVariantFile) is updated accordingly.
 */
static void _merge__copy_working_file(
	SG_context*         pCtx,  //< [in] [out] Error and context info.
	SG_resolve__value*  pValue //< [in] [out] The value whose file should be copied.
	)
{
	SG_pathname* pVariantPath = NULL;
	SG_bool      bEqual       = SG_FALSE;
	SG_pathname* pTargetPath  = NULL;
	SG_pathname* pOldPath     = NULL;

	SG_NULLARGCHECK(pValue);
	SG_NULLARGCHECK(pValue->pVariantFile);
	SG_ARGCHECK(pValue->szChangeset == SG_RESOLVE__HID__WORKING, pValue->szChangeset);

	// allocate a copy of the value's path
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__COPY(pCtx, &pVariantPath, pValue->pVariantFile)  );

	// chop off the last component of the copied path
	SG_ERR_CHECK(  SG_pathname__remove_last(pCtx, pVariantPath)  );

	// if it's now equal to the choice's temp path
	// then the value's file is already in the choice's temp directory
	SG_ERR_CHECK(  SG_pathname__equal_ignoring_final_slashes(pCtx, pVariantPath, pValue->pChoice->pTempPath, &bEqual)  );
	if (bEqual == SG_FALSE)
	{
		// the value's file is somewhere else, we'll need to copy it
		SG_fsobj_stat cStat;

		// generate a temporary filename to copy the file to
		SG_ERR_CHECK(  _generate_temp_filename(pCtx, pValue->pChoice, gszFilenameFormat_Working, SG_RESOLVE__LABEL__WORKING, &pTargetPath)  );

		// copy the file
		SG_ERR_CHECK(  SG_fsobj__stat__pathname(pCtx, pValue->pVariantFile, &cStat)  );
		SG_ERR_CHECK(  SG_fsobj__copy_file(pCtx, pValue->pVariantFile, pTargetPath, cStat.perms)  );

		// update the value to point to the copied file
		// we'll take ownership of the value's old path
		pOldPath = pValue->pVariantFile;
		pValue->pVariantFile = pTargetPath;
		pTargetPath = NULL;
	}

fail:
	SG_PATHNAME_NULLFREE(pCtx, pVariantPath);
	SG_PATHNAME_NULLFREE(pCtx, pTargetPath);
	SG_PATHNAME_NULLFREE(pCtx, pOldPath);
	return;
}

/**
 * Selects and invokes a mergetool to run on two given values in a choice.
 */
static void _merge__run_tool(
	SG_context*         pCtx,            //< [in] [out] Error and context info.
	SG_resolve__value*  pAncestor,       //< [in] The value to use as the common ancestor.
	SG_resolve__value*  pBaseline,       //< [in] The value to use as the baseline parent.
	SG_resolve__value*  pOther,          //< [in] The value to use as the other parent.
	const char*         szResultLabel,   //< [in] Label to use for the result file.
	const SG_pathname*  pResultPath,     //< [in] Full temp path to use for the result file.
	const char*         szSuggestedTool, //< [in] A suggested tool name to use, or NULL if it doesn't matter.
	SG_repo*            pRepo,           //< [in] The repo to look for mergetools in, or NULL to not consider repo-level tools.
	SG_int32*           pStatus,         //< [out] The exit status of the mergetool.
	                                     //<       One of SG_MERGETOOL__STATUS__*
	char**              ppToolName       //< [out] Name of the merge tool that was used.
	                                     //<       NULL if the merge was unsuccessful.
	)
{
	const char*  szAncestorRepoPath = NULL;
	SG_filetool* pTool              = NULL;
	const char*  szToolName         = NULL;
	SG_string*   sAncestorLabel     = NULL;
	SG_string*   sBaselineLabel     = NULL;
	SG_string*   sOtherLabel        = NULL;
	SG_string*   sResultLabel       = NULL;
	SG_int32     iStatus            = SG_MERGETOOL__STATUS__MUST_BE_LAST;

	SG_NULLARGCHECK(pAncestor);
	SG_NULLARGCHECK(pAncestor->pVariantFile);
	SG_NULLARGCHECK(pBaseline);
	SG_NULLARGCHECK(pBaseline->pVariantFile);
	SG_NULLARGCHECK(pOther);
	SG_NULLARGCHECK(pOther->pVariantFile);
	SG_NULLARGCHECK(szResultLabel);
	SG_NULLARGCHECK(pResultPath);
	SG_NULLARGCHECK(pStatus);
	SG_ARGCHECK(pAncestor->pChoice == pBaseline->pChoice, pAncestor+pBaseline);
	SG_ARGCHECK(pAncestor->pChoice == pOther->pChoice, pAncestor+pOther);

	// find an appropriate mergetool for the ancestor file
	SG_ERR_CHECK(  SG_resolve__item__get_repo_path__ancestor(pCtx, pAncestor->pChoice->pItem, &szAncestorRepoPath)  );
	SG_ERR_CHECK(  SG_mergetool__select(pCtx, SG_MERGETOOL__CONTEXT__RESOLVE, szAncestorRepoPath, pAncestor->pVariantFile, szSuggestedTool, pRepo, &pTool)  );
	if (pTool == NULL)
	{
		SG_ERR_THROW2(SG_ERR_RESOLVE_TOOL_NOT_FOUND, (pCtx, "No merge tool found for ancestor value: %s", pAncestor->szLabel));
	}
	SG_ERR_CHECK(  SG_filetool__get_name(pCtx, pTool, &szToolName)  );

	// build all the paths and labels we need
	SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &sAncestorLabel, pAncestor->szLabel)  );
	SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &sBaselineLabel, pBaseline->szLabel)  );
	SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &sOtherLabel, pOther->szLabel)  );
	SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &sResultLabel, szResultLabel)  );

	// invoke the tool
	SG_ERR_CHECK(  SG_mergetool__invoke(pCtx, pTool, pAncestor->pVariantFile, sAncestorLabel, pBaseline->pVariantFile, sBaselineLabel, pOther->pVariantFile, sOtherLabel, pResultPath, sResultLabel, &iStatus)  );

	// return the results
	SG_ERR_CHECK(  SG_strdup(pCtx, szToolName, ppToolName)  );
	*pStatus = iStatus;

fail:
	SG_FILETOOL_NULLFREE(pCtx, pTool);
	SG_STRING_NULLFREE(pCtx, sAncestorLabel);
	SG_STRING_NULLFREE(pCtx, sBaselineLabel);
	SG_STRING_NULLFREE(pCtx, sOtherLabel);
	SG_STRING_NULLFREE(pCtx, sResultLabel);
	return;
}

/**
 * Adds/updates a value in its choice's list of values.
 * If a value with the same label already exists, it is replaced.
 * Otherwise, the new value is appended.
 */
static void _merge__update_values(
	SG_context*        pCtx,  //< [in] [out] Error and context info.
	SG_resolve__value* pValue //< [in] The value to add to its choice's list.
	)
{
	SG_resolve__choice* pChoice   = pValue->pChoice;
	SG_uint32           uIndex    = 0u;
	SG_resolve__value*  pOldValue = NULL;

	SG_NULLARGCHECK(pValue);

	// look up any value that already exists at the same label
	SG_ERR_CHECK(  SG_vector__find__first(pCtx, pChoice->pValues, _value__vector_predicate__match_label, (void*)pValue->szLabel, &uIndex, (void**)&pOldValue)  );
	if (pOldValue != NULL)
	{
		// make sure it's a manually generated value
		// this should have been verified before we were called
		SG_ASSERT(pOldValue->szChangeset != NULL || pOldValue->bAutoMerge == SG_FALSE);

		// make sure it's not the choice's current result
		// this should also have been verified before we were called
		SG_ASSERT(pOldValue != pChoice->pResult);

		// remove the old value so that it is effectively replaced with the new one
		// Note: We are removing the old value and appending the new one,
		//       rather than simply overwriting the value in its old position.
		//       This is to ensure that a value's parents always come before
		//       it in the list of values.
		SG_ERR_CHECK(  SG_vector__remove__index(pCtx, pChoice->pValues, uIndex, _value__sg_action__free)  );
		pOldValue = NULL; // old value was freed by the remove
	}

	// append the new value
	SG_ERR_CHECK(  SG_vector__append(pCtx, pChoice->pValues, (void*)pValue, NULL)  );

fail:
	return;
}


/*
**
** Internal Functions - Accept
**
*/

/**
 * Sets the existence of a file to a given value.
 */
static void _accept__existence(
	SG_context*        pCtx,         //< [in] [out] Error and context info.
	SG_pendingtree*    pPendingTree, //< [in] The pendingtree to potentially add/remove the file in.
	const SG_pathname* pPath,        //< [in] The path of the file to manipulate.
	                                 //<      NULL if the file doesn't currently exist.
	SG_bool            bValue,       //< [in] True if the file should exist, false if it should not.
	SG_resolve__item*  pItem         //< [in] The item that we're accepting a value on.
	)
{
	SG_NULLARGCHECK(pPendingTree);
	SG_ARGCHECK(bValue == SG_TRUE || bValue == SG_FALSE, bValue);

	// check what to do
	if (pPath != NULL && bValue == SG_FALSE)
	{
		const char* szPath  = SG_pathname__sz(pPath);
		SG_bool     bRevert = SG_FALSE;

		// the item exists and is no longer supposed to, delete it
		SG_ERR_TRY(  SG_pendingtree__remove_dont_save_pendingtree(pCtx, pPendingTree, 1u, &szPath, gpBlankFilespec, SG_FALSE, SG_TRUE, SG_TRUE)  )
		SG_ERR_CATCH_SET(SG_ERR_CANNOT_REMOVE_SAFELY, bRevert, SG_TRUE)
		SG_ERR_CATCH_END;

		// If we couldn't remove it safely, it's because pendingtree currently
		// won't remove something that's renamed or moved (though hopefully that
		// will change in the future).  Unfortunately revert won't help us here
		// because it won't allow a partial revert after a merge.  So we'll have
		// to "revert" it ourselves by renaming/moving it back to its original
		// name/location.  Then we'll try the remove again.  Since this can be
		// an odd and surprisingly complex procedure, it's in a helper function.
		// Once pendingtree can properly deal with removing a renamed/moved file,
		// this call (and the function itself) can go away.
		if (bRevert == SG_TRUE)
		{
			SG_ERR_CHECK(  _pendingtree__revert_delete(pCtx, pPendingTree, pPath, pItem->szOriginalName, pItem->szOriginalLocation)  );
		}
	}
	else if (pPath == NULL && bValue == SG_TRUE)
	{
		// It's not really possible to recover an item that has been deleted
		// since the merge occurred, which is the only way to end up here.  I
		// tried going down this road a little, but it only leads to half-baked
		// attempts at restoring an item as best we can.  What the user probably
		// wants, though, is the file as it existed just before they deleted it,
		// and that information simply doesn't exist anymore.  They might have
		// made changes to it since the merge, and even if they didn't then we
		// still may not have all the information we need to reconstruct it
		// from parent changesets and/or merge output or whatever.  It becomes
		// especially hairy when talking about restoring directories, which may
		// have had changes within them somewhere.
		SG_ERR_THROW2(SG_ERR_NOT_FOUND, (pCtx, "Unable to reconstruct an item that has been deleted since merging."));
	}
	// in other cases, the item's current state already matches the given desired state
	// so there's no work to do

fail:
	return;
}

/**
 * Sets the name of a file to a given value.
 */
static void _accept__name(
	SG_context*        pCtx,         //< [in] [out] Error and context info.
	SG_pendingtree*    pPendingTree, //< [in] The pendingtree to potentially rename the file in.
	const SG_pathname* pPath,        //< [in] The path of the file to manipulate.
	const char*        szValue       //< [in] The name to rename the file to.
	)
{
	SG_string* sOldName = NULL;

	SG_NULLARGCHECK(pPendingTree);
	SG_NULLARGCHECK(pPath);
	SG_NULLARGCHECK(szValue);

	// check if the item already has the correct name
	// this will always occur when the working value is chosen
	SG_ERR_CHECK(  SG_pathname__get_last(pCtx, pPath, &sOldName)  );
	if (SG_stricmp(szValue, SG_string__sz(sOldName)) != 0)
	{
		// rename the item to the new name
		SG_ERR_CHECK(  SG_pendingtree__rename_dont_save_pendingtree(pCtx, pPendingTree, pPath, szValue, SG_FALSE)  );
	}

fail:
	SG_STRING_NULLFREE(pCtx, sOldName);
	return;
}

/**
 * Sets the location of a file to a given folder.
 */
static void _accept__location(
	SG_context*        pCtx,         //< [in] [out] Error and context info.
	SG_pendingtree*    pPendingTree, //< [in] The pendingtree to potentially move the file in.
	const SG_pathname* pPath,        //< [in] The path of the file to manipulate.
	const char*        szValue       //< [in] The repo path of the folder to move the item into.
	)
{
	SG_pathname* pTargetLocalPath   = NULL;
	SG_pathname* pCurrentParentPath = NULL;
	SG_bool      bEqual             = SG_FALSE;

	SG_NULLARGCHECK(pPendingTree);
	SG_NULLARGCHECK(pPath);
	SG_NULLARGCHECK(szValue);

	// convert the target repo path to an absolute one
	SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path2(pCtx, pPendingTree, szValue, &pTargetLocalPath)  );

	// check if the current path already equals the target path
	// this will always occur when the working value is chosen
	SG_ERR_CHECK(  SG_PATHNAME__ALLOC__COPY(pCtx, &pCurrentParentPath, pPath)  );
	SG_ERR_CHECK(  SG_pathname__remove_last(pCtx, pCurrentParentPath)  );
	SG_ERR_CHECK(  SG_pathname__equal_ignoring_final_slashes(pCtx, pCurrentParentPath, pTargetLocalPath, &bEqual)  );
	if (bEqual == SG_FALSE)
	{
		// move the item to its new path
		// Note: This will fail if another item is already at the target path.
		//       We could tell it to overwrite by passing true for bForce, but I don't think that's a good idea for now.
		SG_ERR_CHECK(  SG_pendingtree__move_dont_save_pendingtree(pCtx, pPendingTree, pPath, pTargetLocalPath, SG_FALSE)  );
	}

fail:
	SG_PATHNAME_NULLFREE(pCtx, pTargetLocalPath);
	SG_PATHNAME_NULLFREE(pCtx, pCurrentParentPath);
	return;
}

/**
 * Sets the attribute bits of a file to a given value.
 */
static void _accept__attributes(
	SG_context*        pCtx,  //< [in] [out] Error and context info.
	const SG_pathname* pPath, //< [in] The path of the file to manipulate.
	SG_int32           iValue //< [in] The attribute bits to set on the file.
	)
{
	SG_NULLARGCHECK(pPath);

	// change the file's attributes
	SG_ERR_CHECK(  SG_fsobj__chmod__pathname(pCtx, pPath, iValue)  );

fail:
	return;
}

/**
 * Sets the extended attributes of a file to the attributes in a given blob.
 */
static void _accept__xattributes(
	SG_context*        pCtx,   //< [in] [out] Error and context info.
	const SG_pathname* pPath,  //< [in] The path of the file to manipulate.
	const char*        szValue //< [in] The HID of the blob containing extended attributes to set on the item.
	)
{
	// TODO: Implement this.
	//       Problems:
	//       1) fsobj__xattr__* functions don't exist on Windows.
	//          Add stubs that do nothing?
	//       2) Don't know the format of blobs that contain extended attribute data.
	SG_UNUSED(pCtx);
	SG_UNUSED(pPath);
	SG_UNUSED(szValue);
}

/**
 * Sets the contents of a file to the contents of another given file.
 */
static void _accept__file(
	SG_context*        pCtx,  //< [in] [out] Error and context info.
	const SG_pathname* pPath, //< [in] The path of the file to manipulate.
	const SG_pathname* pValue //< [in] The path of the file whose contents should be copied.
	)
{
	SG_bool       bEqual = SG_FALSE;
	SG_fsobj_stat cStat;

	SG_NULLARGCHECK(pPath);
	SG_NULLARGCHECK(pValue);

	// check if the item is already at the target path
	// this should always occur when the working value is being accepted
	SG_ERR_CHECK(  SG_pathname__equal_ignoring_final_slashes(pCtx, pValue, pPath, &bEqual)  );
	if (bEqual == SG_FALSE)
	{
		// stat the file to get its current attributes, we need to preserve them
		// They may have been set by merge when the conflict was originally created
		// or they may have been set by resolve when handling an attribute conflict.
		// If there's an attribute conflict still pending, then the result of that
		// will correctly override whatever we set here now.
		SG_ERR_CHECK(  SG_fsobj__stat__pathname(pCtx, pPath, &cStat)  );

		// copy the file to the given path
		SG_ERR_CHECK(  SG_fsobj__copy_file(pCtx, pValue, pPath, cStat.perms)  );
	}

fail:
	return;
}

/**
 * Sets the target of a symlink to a given target.
 */
static void _accept__symlink(
	SG_context*        pCtx,   //< [in] [out] Error and context info.
	const SG_pathname* pPath,  //< [in] The path of the symlink to manipulate.
	const char*        szValue //< [in] The target to set the symlink to.
	)
{
	SG_bool    bExists        = SG_FALSE;
	SG_string* sCurrentTarget = NULL;
	SG_string* sNewTarget     = NULL;

	SG_NULLARGCHECK(pPath);
	SG_NULLARGCHECK(szValue);

	// try to read the target from the current symlink
	SG_ERR_CHECK(  SG_fsobj__exists__pathname(pCtx, pPath, &bExists, NULL, NULL)  );
	if (bExists == SG_TRUE)
	{
		SG_ERR_IGNORE(  SG_fsobj__readlink(pCtx, pPath, &sCurrentTarget)  );
	}
	SG_ASSERT(bExists == SG_TRUE || sCurrentTarget == NULL);

	// check if the symlink already exists with the requested target
	// this should always occur if the working value is being accepted
	if (
		   sCurrentTarget == NULL
		|| SG_stricmp(SG_string__sz(sCurrentTarget), szValue) != 0
		)
	{
		// if the symlink already exists, delete it so we can make a new one
		if (bExists == SG_TRUE)
		{
			SG_ERR_CHECK(  SG_fsobj__remove__pathname(pCtx, pPath)  );
		}

		// create the new symlink
		SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &sNewTarget, szValue)  );
		SG_ERR_CHECK(  SG_fsobj__symlink(pCtx, sNewTarget, pPath)  );
	}

fail:
	SG_STRING_NULLFREE(pCtx, sCurrentTarget);
	SG_STRING_NULLFREE(pCtx, sNewTarget);
	return;
}

/**
 * Sets the target of a submodule to a given target.
 */
static void _accept__submodule(
	SG_context*        pCtx,    //< [in] [out] Error and context info.
	const SG_pathname* pPath,   //< [in] The path of the submodule to manipulate.
	const char*        szValue, //< [in] The HID of a submodule blob to accept.
	SG_repo*           pRepo    //< [in] The repo to load submodules from.
	)
{
	SG_bool                                     bBool           = SG_FALSE;
	SG_treenode_submodule*                      pPathSubmodule  = NULL;
	SG_treenode_submodule*                      pValueSubmodule = NULL;
	SG_varray*                                  pMissingDetails = NULL;
	SG_treenode_submodule_instance_data_update* pUpdateData     = NULL;

	SG_NULLARGCHECK(pPath);
	SG_NULLARGCHECK(szValue);

	// load the submodule referred to by the path
	SG_ERR_CHECK(  SG_treenode_submodule__sniff_directory(pCtx, pPath, pRepo, &bBool, &pPathSubmodule)  );
	if (bBool == SG_FALSE)
	{
		SG_ERR_THROW2(SG_ERR_RESOLVE_INVALID_TARGET, (pCtx, "Cannot accept a submodule to a path that is not already a submodule: %s", SG_pathname__sz(pPath)));
	}

	// load the submodule referred to by the value
	SG_ERR_CHECK(  SG_TREENODE_SUBMODULE__LOAD_FROM_REPO(pCtx, pRepo, szValue, &pValueSubmodule)  );

	// check if the submodules are comparable
	SG_ERR_CHECK(  SG_treenode_submodule__are_comparable(pCtx, pPathSubmodule, pValueSubmodule, &bBool)  );
	if (bBool == SG_TRUE)
	{
		// check if the submodules are already equal
		SG_ERR_CHECK(  SG_treenode_submodule__are_same_hid(pCtx, pPathSubmodule, pValueSubmodule, &bBool)  );
		if (bBool == SG_FALSE)
		{
			// check if the submodule already in the path is dirty
			SG_ERR_CHECK(  SG_treenode_submodule__is_dirty(pCtx, pPathSubmodule, &bBool)  );
			if (bBool == SG_FALSE)
			{
				// update the submodule in the path to the submodule specified by the value
				SG_ERR_CHECK(  SG_VARRAY__ALLOC(pCtx, &pMissingDetails)  );
				SG_ERR_CHECK(  SG_treenode_submodule__precompute_update(pCtx, pPath, pPathSubmodule, pValueSubmodule, NULL, pMissingDetails, &pUpdateData)  );
				SG_ERR_CHECK(  SG_treenode_submodule__apply_update(pCtx, pPath, pPathSubmodule, pValueSubmodule, pUpdateData)  );
			}
			else
			{
				SG_ERR_THROW2(SG_ERR_RESOLVE_INVALID_TARGET, (pCtx, "Cannot resolve a submodule path that is currently dirty: %s", SG_pathname__sz(pPath)));
			}
		}
	}
	else
	{
		SG_ERR_THROW2(SG_ERR_RESOLVE_INVALID_VALUE, (pCtx, "Accepted submodule value is unrelated to submodule being resolved: %s", SG_pathname__sz(pPath)));
	}

fail:
	SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pPathSubmodule);
	SG_TREENODE_SUBMODULE_NULLFREE(pCtx, pValueSubmodule);
	SG_VARRAY_NULLFREE(pCtx, pMissingDetails);
	if (pUpdateData != NULL)
	{
		SG_ERR_IGNORE(  SG_treenode_submodule_instance_data_update__free(pCtx, pUpdateData)  );
	}
	return;
}


/*
**
** Public Functions
**
*/

void SG_resolve__state__get_info(
	SG_context*       pCtx,
	SG_resolve__state eState,
	const char**      ppName,
	const char**      ppDescription
	)
{
	SG_ARGCHECK(eState < SG_RESOLVE__STATE__COUNT, eState);

	if (ppName != NULL)
	{
		*ppName = gaStateData[eState].szName;
	}
	if (ppDescription != NULL)
	{
		*ppDescription = gaStateData[eState].szDescription;
	}

fail:
	return;
}

void SG_resolve__state__check_name(
	SG_context*        pCtx,
	const char*        szName,
	SG_resolve__state* pState
	)
{
	SG_uint32 uIndex = 0u;

	SG_NULLARGCHECK(pState);

	for (uIndex = 0u; uIndex < SG_RESOLVE__STATE__COUNT; ++uIndex)
	{
		if (SG_stricmp(szName, gaStateData[uIndex].szName) == 0)
		{
			break;
		}
	}

	*pState = (SG_resolve__state)uIndex;

fail:
	return;
}

void SG_resolve__state__parse_name(
	SG_context*        pCtx,
	const char*        szName,
	SG_resolve__state* pState
	)
{
	SG_resolve__state eState = SG_RESOLVE__STATE__COUNT;

	SG_NULLARGCHECK(pState);

	SG_ERR_CHECK(  SG_resolve__state__check_name(pCtx, szName, &eState)  );
	if (eState == SG_RESOLVE__STATE__COUNT)
	{
		SG_ERR_THROW2(SG_ERR_NOT_FOUND, (pCtx, "conflict state by name: %s", szName));
	}

	*pState = eState;

fail:
	return;
}

void SG_resolve__alloc(
	SG_context*     pCtx,
	SG_pendingtree* pPendingTree,
	SG_resolve**    ppResolve
	)
{
	SG_resolve*      pResolve = NULL;
	SG_treediff2*    pDiff    = NULL;
	SG_bool          bIssues  = SG_FALSE;
	const SG_varray* pIssues  = NULL;

	SG_NULLARGCHECK(pPendingTree);
	SG_NULLARGCHECK(ppResolve);

	// allocate the resolve structure and setup its basic data
	SG_ERR_CHECK(  SG_alloc1(pCtx, pResolve)  );
	pResolve->pPendingTree = pPendingTree;
	pResolve->pParentHids  = NULL;
	pResolve->pItems       = NULL;

	// run a diff between the pendingtree and baseline
	// we'll use the diff to get the current working state of conflicting items
	// Note: This must come before we retrieve other data that points into
	//       the pendingtree because this call causes it to be re-loaded.
	SG_ERR_CHECK(  SG_pendingtree__do_diff__keep_clean_objects__reload(pCtx, pPendingTree, NULL, NULL, &pDiff)  );

	// retrieve parent HIDs from the pendingtree
	SG_ERR_CHECK(  SG_pendingtree__get_wd_parents(pCtx, pPendingTree, &pResolve->pParentHids)  );

	// retrieve the merge issues from the pendingtree
	// this will return NULL if there aren't any
	SG_ERR_CHECK(  SG_pendingtree__get_wd_issues__ref(pCtx, pPendingTree, &bIssues, &pIssues)  );

	// process any issues
	if (pIssues != NULL)
	{
		SG_uint32 uCount = 0u;
		SG_uint32 uIndex = 0u;

		// allocate space for the appropriate number of items
		SG_ERR_CHECK(  SG_varray__count(pCtx, pIssues, &uCount)  );
		SG_ERR_CHECK(  SG_VECTOR__ALLOC(pCtx, &pResolve->pItems, uCount)  );

		// run through the list of issues and build an item for each
		for (uIndex = 0u; uIndex < uCount; ++uIndex)
		{
			SG_resolve__item* pItem  = NULL;
			SG_vhash*         pIssue = NULL;

			SG_ERR_CHECK(  SG_varray__get__vhash(pCtx, pIssues, uIndex, &pIssue)  );
			SG_ERR_CHECK(  _item__alloc(pCtx, pIssue, pDiff, pResolve, &pItem)  );
			SG_ERR_CHECK(  SG_vector__append(pCtx, pResolve->pItems, (void*)pItem, NULL)  );
		}

		// Now that all the items are created,
		// have each item gather its related items.
		// Note: we're relying on pIssues and pResolve->pItems to be parallel.
		for (uIndex = 0u; uIndex < uCount; ++uIndex)
		{
			SG_resolve__item* pItem = NULL;
			SG_vhash*         pIssue = NULL;

			SG_ERR_CHECK(  SG_vector__get(pCtx, pResolve->pItems, uIndex, (void**)&pItem)  );
			SG_ERR_CHECK(  SG_varray__get__vhash(pCtx, pIssues, uIndex, &pIssue)  );
			SG_ERR_CHECK(  _item__gather_related_items(pCtx, pItem, pIssue, pResolve->pItems)  );
		}
	}
	else
	{
		// there weren't any issues, allocate an empty item list
		SG_ERR_CHECK(  SG_VECTOR__ALLOC(pCtx, &pResolve->pItems, 0u)  );
	}

	// return the resolve
	*ppResolve = pResolve;
	pResolve = NULL;

fail:
	SG_RESOLVE_NULLFREE(pCtx, pResolve);
	SG_TREEDIFF2_NULLFREE(pCtx, pDiff);
	return;
}

void SG_resolve__free(
	SG_context* pCtx,
	SG_resolve* pResolve
	)
{
	if (pResolve == NULL)
	{
		return;
	}

	SG_VARRAY_NULLFREE(pCtx, pResolve->pParentHids);
	SG_VECTOR_NULLFREE_WITH_ASSOC(pCtx, pResolve->pItems, _item__sg_action__free);
	pResolve->pPendingTree = NULL;
	SG_NULLFREE(pCtx, pResolve);
}

void SG_resolve__get_unresolved_count(
	SG_context*     pCtx,
	SG_resolve*     pResolve,
	SG_pendingtree* pPendingTree,
	SG_uint32*      pCount
	)
{
	SG_uint32 uCount = 0u;

	SG_ARGCHECK(!SG_BOTH_NULL_OR_NOT(pResolve, pPendingTree), pResolve+pPendingTree);
	SG_NULLARGCHECK(pCount);

	// if they didn't pass a resolve, alloc one from the pendingtree
	if (pResolve == NULL)
	{
		SG_ERR_CHECK(  SG_resolve__alloc(pCtx, pPendingTree, &pResolve)  );
	}
	SG_ASSERT(pResolve != NULL);

	// count up how many unresolved choices we have
	SG_ERR_CHECK(  SG_resolve__foreach_choice(pCtx, pResolve, SG_TRUE, _choice__resolve_action__increment_count, &uCount)  );

	// return the count
	*pCount = uCount;

fail:
	// we own the resolve if they passed a pendingtree instead
	if (pPendingTree != NULL)
	{
		SG_RESOLVE_NULLFREE(pCtx, pResolve);
	}
	return;
}

void SG_resolve__save(
	SG_context* pCtx,
	SG_resolve* pResolve,
	SG_bool     bSavePendingTree
	)
{
	SG_vhash* pHash = NULL;

	SG_NULLARGCHECK(pResolve);

	// allocate a hash to save all the items to
	SG_ERR_CHECK(  SG_VHASH__ALLOC(pCtx, &pHash)  );

	// run through each item in the resolve and add it to the hash
	SG_ERR_CHECK(  SG_resolve__foreach_item(pCtx, pResolve, SG_FALSE, _item__resolve_action__add_resolve_vhash_to_vhash, (void*)pHash)  );

	// store the hash to the pendingtree (it steals the pointer)
	SG_ERR_CHECK(  SG_pendingtree__set_wd_choices(pCtx, pResolve->pPendingTree, &pHash)  );
	SG_ASSERT(pHash == NULL);

	// if they also want to save the pendingtree, do that
	if (bSavePendingTree == SG_TRUE)
	{
		SG_ERR_CHECK(  SG_pendingtree__save(pCtx, pResolve->pPendingTree)  );
	}

fail:
	SG_VHASH_NULLFREE(pCtx, pHash);
	return;
}

void SG_resolve__diff_values(
	SG_context*        pCtx,
	SG_resolve__value* pLeft,
	SG_resolve__value* pRight,
	const char*        szSuggestedTool
	)
{
	SG_resolve__choice* pChoice          = pLeft->pChoice;
	SG_bool             bMergeable       = SG_FALSE;
	SG_repo*            pRepo            = NULL;
	const char*         szLeftChangeset  = NULL;
	const char*         szRightChangeset = NULL;
	const char*         szLeftRepoPath   = NULL;
	const char*         szRightRepoPath  = NULL;
	SG_filetool*        pLeftTool        = NULL;
	SG_filetool*        pRightTool       = NULL;
	const char*         szLeftToolName   = NULL;
	const char*         szRightToolName  = NULL;
	SG_filetool*        pDiffTool        = NULL;
	SG_string*          pLeftLabel       = NULL;
	SG_string*          pRightLabel      = NULL;

	SG_NULLARGCHECK(pLeft);
	SG_NULLARGCHECK(pRight);
	SG_ARGCHECK(pLeft->pChoice == pRight->pChoice, pLeft+pRight);

	// make sure this is a mergeable choice
	SG_ERR_CHECK(  SG_resolve__choice__get_mergeable(pCtx, pChoice, &bMergeable)  );
	SG_ARGCHECK(bMergeable == SG_TRUE, pLeft|pRight);
	SG_ASSERT(pLeft->pVariantFile != NULL && pRight->pVariantFile != NULL);

	// get the repo from the pendingtree
	SG_ERR_CHECK(  SG_pendingtree__get_repo(pCtx, pChoice->pItem->pResolve->pPendingTree, &pRepo)  );

	// select an appropriate diff tool for the left value
	SG_ERR_CHECK(  _value__find_changeset(pCtx, pLeft, &szLeftChangeset)  );
	SG_ERR_CHECK(  SG_resolve__item__get_repo_path(pCtx, pChoice->pItem, szLeftChangeset, &szLeftRepoPath)  );
	SG_ERR_CHECK(  SG_difftool__select(pCtx, szLeftRepoPath, pLeft->pVariantFile, szSuggestedTool, pRepo, &pLeftTool)  );
	if (pLeftTool == NULL)
	{
		SG_ERR_THROW2(SG_ERR_RESOLVE_TOOL_NOT_FOUND, (pCtx, "No diff tool found for left value: %s", pLeft->szLabel));
	}

	// select an appropriate diff tool for the right value
	SG_ERR_CHECK(  _value__find_changeset(pCtx, pRight, &szRightChangeset)  );
	SG_ERR_CHECK(  SG_resolve__item__get_repo_path(pCtx, pChoice->pItem, szRightChangeset, &szRightRepoPath)  );
	SG_ERR_CHECK(  SG_difftool__select(pCtx, szRightRepoPath, pRight->pVariantFile, szSuggestedTool, pRepo, &pRightTool)  );
	if (pRightTool == NULL)
	{
		SG_ERR_THROW2(SG_ERR_RESOLVE_TOOL_NOT_FOUND, (pCtx, "No diff tool found for right value: %s", pRight->szLabel));
	}

	// get the names of both diff tools and decide which to use
	SG_ERR_CHECK(  SG_filetool__get_name(pCtx, pLeftTool,  &szLeftToolName)  );
	SG_ERR_CHECK(  SG_filetool__get_name(pCtx, pRightTool, &szRightToolName)  );
	if (strcmp(szLeftToolName, szRightToolName) == 0)
	{
		// both tools are the same, just pick one
		pDiffTool = pLeftTool;
	}
	else if (strcmp(szLeftToolName, szSuggestedTool) == 0)
	{
		// different tools, but left is the suggested one, use it
		pDiffTool = pLeftTool;
	}
	else if (strcmp(szRightToolName, szSuggestedTool) == 0)
	{
		// different tools, but right is the suggested one, use it
		pDiffTool = pRightTool;
	}
	else
	{
		// no other criteria to use.. just arbitrarily pick one
		pDiffTool = pLeftTool;
	}

	// invoke the selected tool.
	// 
	// TODO 2011/06/23 Andy, I'm passing SG_FALSE for bRightSideIsWritable because
	// TODO            I'm thinking that this code is just to display the differences
	// TODO            between any combination of the possible inputs.  We DO NOT
	// TODO            want to make this a gateway to let them edit the right side.
	// TODO            That kind of work happens elsewhere.  Please let me know if
	// TODO            we want to re-think this.  --Jeff
	SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &pLeftLabel, pLeft->szLabel)  );
	SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &pRightLabel, pRight->szLabel)  );
	SG_ERR_CHECK(  SG_difftool__invoke(pCtx, pDiffTool, pLeft->pVariantFile, pLeftLabel, pRight->pVariantFile, pRightLabel, SG_FALSE)  );
	pDiffTool = NULL;

fail:
	SG_FILETOOL_NULLFREE(pCtx, pLeftTool);
	SG_FILETOOL_NULLFREE(pCtx, pRightTool);
	SG_STRING_NULLFREE(pCtx, pLeftLabel);
	SG_STRING_NULLFREE(pCtx, pRightLabel);
	return;
}

void SG_resolve__merge_values(
	SG_context*         pCtx,
	SG_resolve__value*  pBaseline,
	SG_resolve__value*  pOther,
	const char*         szSuggestedTool,
	const char*         szLabel,
	SG_int32*           pStatus,
	SG_resolve__value** ppResult
	)
{
	SG_resolve__choice* pChoice         = pBaseline->pChoice;
	SG_bool             bMergeable      = SG_FALSE;
	SG_resolve__value*  pAncestor       = NULL;
	SG_string*          sResultLabel    = NULL;
	SG_pathname*        pResultTempFile = NULL;
	SG_repo*            pRepo           = NULL;
	SG_int32            iStatus         = SG_MERGETOOL__STATUS__MUST_BE_LAST;
	char*               szToolName      = NULL;
	SG_resolve__value*  pResult         = NULL;
	SG_bool             bOwnResult      = SG_FALSE;

	SG_NULLARGCHECK(pBaseline);
	SG_NULLARGCHECK(pOther);
	SG_ARGCHECK(pBaseline->pChoice == pOther->pChoice, pBaseline+pOther);

	// make sure this is a mergeable choice
	SG_ERR_CHECK(  SG_resolve__choice__get_mergeable(pCtx, pChoice, &bMergeable)  );
	SG_ARGCHECK(bMergeable == SG_TRUE, pBaseline|pOther);

	// if they supplied a label, make sure it will work
	if (szLabel != NULL)
	{
		SG_resolve__value* pTempValue = NULL;

		SG_ERR_CHECK(  SG_resolve__choice__check_value__label(pCtx, pChoice, szLabel, 0u, &pTempValue)  );
		if (pTempValue != NULL)
		{
			SG_bool bOverwritable = SG_FALSE;

			SG_ERR_CHECK(  SG_resolve__value__has_flag(pCtx, pTempValue, SG_RESOLVE__VALUE__OVERWRITABLE, &bOverwritable)  );
			if (bOverwritable == SG_FALSE)
			{
				SG_ERR_THROW2(SG_ERR_RESOLVE_INVALID_VALUE, (pCtx, "The specified value already exists and cannot be overwritten."));
			}
		}
	}

	// find our common ancestor value
	SG_ERR_CHECK(  SG_resolve__choice__get_value__label(pCtx, pChoice, SG_RESOLVE__LABEL__ANCESTOR, 0u, &pAncestor)  );

	// copy any parent files that are in the working copy into our temp folder
	if (pBaseline->szChangeset == SG_RESOLVE__HID__WORKING)
	{
		SG_ERR_CHECK(  _merge__copy_working_file(pCtx, pBaseline)  );
	}
	if (pOther->szChangeset == SG_RESOLVE__HID__WORKING)
	{
		SG_ERR_CHECK(  _merge__copy_working_file(pCtx, pOther)  );
	}

	// generate a label for the merge result
	if (szLabel == NULL)
	{
		SG_ERR_CHECK(  _generate_indexed_label(pCtx, pChoice, SG_RESOLVE__LABEL__MERGE, &sResultLabel)  );
	}
	else
	{
		SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &sResultLabel, szLabel)  );
	}

	// generate a temp filename for the merge result
	SG_ERR_CHECK(  _generate_temp_filename(pCtx, pChoice, gszFilenameFormat_Merge, SG_string__sz(sResultLabel), &pResultTempFile)  );

	// run a mergetool to merge the values into a result
	SG_ERR_CHECK(  SG_pendingtree__get_repo(pCtx, pChoice->pItem->pResolve->pPendingTree, &pRepo)  );
	SG_ERR_CHECK(  _merge__run_tool(pCtx, pAncestor, pBaseline, pOther, SG_string__sz(sResultLabel), pResultTempFile, szSuggestedTool, pRepo, &iStatus, &szToolName)  );

	// check the tool's output
	if (
		   iStatus == SG_MERGETOOL__STATUS__SUCCESSFUL
		|| iStatus == SG_MERGETOOL__STATUS__CONFLICT
		|| iStatus == SG_MERGETOOL__STATUS__ERROR
		)
	{
		// create a new value from the results
		SG_ERR_CHECK(  _value__alloc__merge(pCtx, pChoice, &sResultLabel, &pResultTempFile, pRepo, pBaseline, pOther, SG_FALSE, &szToolName, iStatus, &pResult)  );
		bOwnResult = SG_TRUE;

		// add/update the value to/in the choice's list of values
		SG_ERR_CHECK(  _merge__update_values(pCtx, pResult)  );
		bOwnResult = SG_FALSE;

		// if the merge was successful, update the list of leaf values
		if (iStatus == SG_MERGETOOL__STATUS__SUCCESSFUL)
		{
			SG_resolve__value* pLeaf = NULL;

			// add/overwrite the result in the leaves and remove its parents
			SG_ERR_CHECK(  _choice__update_leaves(pCtx, pChoice, pResult, SG_TRUE)  );

			// if the new value is the last leaf value, accept it automatically
			SG_ERR_CHECK(  _choice__get_only_leaf(pCtx, pChoice, &pLeaf)  );
			if (pLeaf == pResult)
			{
				SG_ERR_CHECK(  SG_resolve__accept_value(pCtx, pResult)  );
			}
		}
	}

	// return the results
	if (pStatus != NULL)
	{
		*pStatus = iStatus;
	}
	if (ppResult != NULL)
	{
		*ppResult = pResult;
		pResult = NULL;
	}

fail:
	SG_STRING_NULLFREE(pCtx, sResultLabel);
	SG_PATHNAME_NULLFREE(pCtx, pResultTempFile);
	SG_NULLFREE(pCtx, szToolName);
	if (bOwnResult == SG_TRUE)
	{
		_VALUE__NULLFREE(pCtx, pResult);
	}
	return;
}

void SG_resolve__accept_value(
	SG_context*        pCtx,
	SG_resolve__value* pValue
	)
{
	SG_resolve__choice* pChoice       = pValue->pChoice;
	SG_resolve__item*   pItem         = pChoice->pItem;
	SG_resolve*         pResolve      = pItem->pResolve;
	SG_pendingtree*     pPendingTree  = pResolve->pPendingTree;
	SG_pathname*        pAbsolutePath = NULL;
	SG_string*          sRepoPath     = NULL;
	SG_string*          sName         = NULL;

	SG_NULLARGCHECK(pValue);

	// check that the value can actually be used as a resolution
	if (pValue->pVariant == NULL && pValue->pVariantFile == NULL)
	{
		// this value is the result of a failed merge
		SG_ASSERT(pValue->pBaselineParent != NULL && pValue->pOtherParent != NULL);
		SG_ASSERT(pValue->iMergeToolResult != SG_MERGETOOL__STATUS__SUCCESSFUL);
		SG_ERR_THROW2(SG_ERR_RESOLVE_INVALID_VALUE, (pCtx, "Cannot accept a value that resulted from a failed merge: %s", pValue->szLabel));
	}
	else if (
		   (pItem->eStatus & (SG_DIFFSTATUS_FLAGS__DELETED | SG_DIFFSTATUS_FLAGS__LOST)) != 0
		&& pChoice->eState != SG_RESOLVE__STATE__EXISTENCE
		&& pValue->pVariant->v.val_bool != SG_FALSE
		)
	{
		// can't accept a value on an item that doesn't currently exist
		// unless it's an EXISTENCE choice and we're accepting a FALSE value
		SG_ERR_THROW2(SG_ERR_NOT_FOUND, (pCtx, "Cannot accept a value on an item that doesn't currently exist."));
	}
	else if (pChoice->pResult == pValue)
	{
		// the choice has already been resolved with this value
		SG_ERR_THROW2(SG_ERR_RESOLVE_ALREADY_RESOLVED, (pCtx, "The conflict has already been resolved using the given value."));
	}
	else
	{
		const char* szRepoPath      = NULL;
		SG_bool     bRevertTempName = SG_FALSE;

		// find the item's path in the WD
		SG_ERR_CHECK(  SG_resolve__item__get_repo_path__working(pCtx, pItem, &szRepoPath)  );
		if (szRepoPath != NULL)
		{
			SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path2(pCtx, pPendingTree, szRepoPath, &pAbsolutePath)  );
		}

		// TODO: Calculate the item's full absolute destination path.
		//       Then make sure there's nothing already there.
		//       If there is, don't allow the value to be accepted.
		//       Note: For LOCATION choices, remember to take into account
		//             that the item will need to be reverted to its original
		//             name.  In that case, it will need BOTH its temporary name
		//             in its new location AND its original name in its new
		//             location.  If either is blocked, abort the accept.

		// the action we need to take on the file depends on what type of choice this is
		if (pChoice->eState == SG_RESOLVE__STATE__EXISTENCE)
		{
			// accept the existence value on the item's file
			SG_ERR_CHECK(  _accept__existence(pCtx, pPendingTree, pAbsolutePath, pValue->pVariant->v.val_bool, pItem)  );
			if (pValue->pVariant->v.val_bool == SG_TRUE)
			{
				// if the accepted value is existence,
				// we need to revert the file's temporary name
				bRevertTempName = SG_TRUE;
			}
		}
		else
		{
			// for any non-EXISTENCE case, make sure the item exists before acting
			// Note: This will currently always fail if the item doesn't already exist (pAbsolutePath == NULL).
			//       It will throw the correct failure message for this case, though.
			//       Some day maybe we'll figure out a way to have it not fail in that case.
			SG_ERR_CHECK(  _accept__existence(pCtx, pPendingTree, pAbsolutePath, SG_TRUE, pItem)  );

			// if we didn't have an absolute path before, because the item didn't exist
			// calculate it now that the item exists again
			if (pAbsolutePath == NULL)
			{
				SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path2(pCtx, pPendingTree, szRepoPath, &pAbsolutePath)  );
			}

			// make appropriate changes to the item based on the accepted value
			switch (pChoice->eState)
			{
			case SG_RESOLVE__STATE__NAME:
				{
					SG_ERR_CHECK(  _accept__name(pCtx, pPendingTree, pAbsolutePath, pValue->pVariant->v.val_sz)  );
					break;
				}
			case SG_RESOLVE__STATE__LOCATION:
				{
					SG_ERR_CHECK(  _accept__location(pCtx, pPendingTree, pAbsolutePath, pValue->szVariantSummary)  );
					bRevertTempName = SG_TRUE;
					break;
				}
			case SG_RESOLVE__STATE__ATTRIBUTES:
				{
					SG_ERR_CHECK(  _accept__attributes(pCtx, pAbsolutePath, (SG_int32)pValue->pVariant->v.val_int64)  );
					break;
				}
			case SG_RESOLVE__STATE__XATTRIBUTES:
				{
					SG_ERR_CHECK(  _accept__xattributes(pCtx, pAbsolutePath, pValue->pVariant->v.val_sz)  );
					break;
				}
			case SG_RESOLVE__STATE__CONTENTS:
				{
					switch (pItem->eType)
					{
					case SG_TREENODEENTRY_TYPE_REGULAR_FILE:
						{
							SG_ERR_CHECK(  _accept__file(pCtx, pAbsolutePath, pValue->pVariantFile)  );
							break;
						}
					case SG_TREENODEENTRY_TYPE_DIRECTORY:
						{
							SG_ERR_THROW2(SG_ERR_NOTIMPLEMENTED, (pCtx, "Cannot resolve a directory contents conflict because they should never occur."));
							break;
						}
					case SG_TREENODEENTRY_TYPE_SYMLINK:
						{
							SG_ERR_CHECK(  _accept__symlink(pCtx, pAbsolutePath, pValue->pVariant->v.val_sz)  );
							break;
						}
					case SG_TREENODEENTRY_TYPE_SUBMODULE:
						{
							SG_repo* pRepo = NULL;
							SG_ERR_CHECK(  SG_pendingtree__get_repo(pCtx, pPendingTree, &pRepo)  );
							SG_ERR_CHECK(  _accept__submodule(pCtx, pAbsolutePath, pValue->pVariant->v.val_sz, pRepo)  );
							break;
						}
					default:
						{
							SG_ERR_THROW2(SG_ERR_NOTIMPLEMENTED, (pCtx, "Unknown node type: %i", pItem->eType));
							break;
						}
					}
					break;
				}
			default:
				{
					SG_ERR_THROW2(SG_ERR_NOTIMPLEMENTED, (pCtx, "Unknown resolve item state: %i", pChoice->eState));
					break;
				}
			}
		}

		// paths/names might have changed during acceptance
		// so make sure we don't try to keep using them
		SG_PATHNAME_NULLFREE(pCtx, pAbsolutePath);
		SG_STRING_NULLFREE(pCtx, sRepoPath);
		SG_STRING_NULLFREE(pCtx, sName);

		// check a few cases where we don't really want to revert the name
		if (bRevertTempName == SG_TRUE)
		{
			// If there's a NAME choice on the same item then we never want to revert the temp name.
			// We'll let the user choose the name (or stick with what they already chose).
			if (pItem->aChoices[SG_RESOLVE__STATE__NAME] != NULL)
			{
				bRevertTempName = SG_FALSE;
			}

			// calculate the item's current name, if we haven't already
			if (sName == NULL)
			{
				if (sRepoPath == NULL)
				{
					SG_ERR_CHECK(  SG_pendingtree__get_repo_path(pCtx, pPendingTree, pItem->szGid, &sRepoPath)  );
				}
				SG_ERR_CHECK(  SG_repopath__get_last(pCtx, sRepoPath, &sName)  );
			}

			// If the item's name isn't still the temp name that merge assigned to it,
			// then the user has renamed it manually since the merge, and we shouldn't touch it.
			if (SG_stricmp(SG_string__sz(sName), pItem->szMergedName) != 0)
			{
				bRevertTempName = SG_FALSE;
			}
		}

		// if we need to revert the item's temporary name, do that
		if (bRevertTempName == SG_TRUE)
		{
			// calculate the item's current absolute path, if we haven't already
			if (pAbsolutePath == NULL)
			{
				if (sRepoPath == NULL)
				{
					SG_ERR_CHECK(  SG_pendingtree__get_repo_path(pCtx, pPendingTree, pItem->szGid, &sRepoPath)  );
				}
				SG_ERR_CHECK(  SG_workingdir__construct_absolute_path_from_repo_path2(pCtx, pPendingTree, SG_string__sz(sRepoPath), &pAbsolutePath)  );
			}

			// rename the item back to its original name
			SG_ERR_CHECK(  SG_pendingtree__rename_dont_save_pendingtree(pCtx, pPendingTree, pAbsolutePath, pItem->szOriginalName, SG_FALSE)  );
		}

		// store the value as the choice's result
		pChoice->pResult = pValue;

		// update the old resolve status on the issue that corresponds to this item
		// TODO: remove this once the old resolve status data is gone
		SG_ERR_CHECK(  _item__update_old_issue_status(pCtx, pItem)  );
	}

fail:
	SG_PATHNAME_NULLFREE(pCtx, pAbsolutePath);
	SG_STRING_NULLFREE(pCtx, sRepoPath);
	SG_STRING_NULLFREE(pCtx, sName);
	return;
}

void SG_resolve__get_pendingtree(
	SG_context*      pCtx,
	SG_resolve*      pResolve,
	SG_pendingtree** ppPendingTree
	)
{
	SG_NULLARGCHECK(pResolve);
	SG_NULLARGCHECK(ppPendingTree);

	*ppPendingTree = pResolve->pPendingTree;

fail:
	return;
}

void SG_resolve__check_item__path(
	SG_context*        pCtx,
	SG_resolve*        pResolve,
	const char*        szRepoPath,
	SG_resolve__item** ppItem
	)
{
	SG_resolve__item* pItem = NULL;

	SG_NULLARGCHECK(pResolve);
	SG_NULLARGCHECK(szRepoPath);
	SG_NULLARGCHECK(ppItem);

	SG_ERR_CHECK(  SG_vector__find__first(pCtx, pResolve->pItems, _item__vector_predicate__match_working_path, (void*)szRepoPath, NULL, (void**)&pItem)  );

	*ppItem = pItem;
	pItem = NULL;

fail:
	return;
}

void SG_resolve__get_item__path(
	SG_context*        pCtx,
	SG_resolve*        pResolve,
	const char*        szRepoPath,
	SG_resolve__item** ppItem
	)
{
	SG_resolve__item* pItem = NULL;

	SG_NULLARGCHECK(pResolve);
	SG_NULLARGCHECK(szRepoPath);
	SG_NULLARGCHECK(ppItem);

	SG_ERR_CHECK(  SG_resolve__check_item__path(pCtx, pResolve, szRepoPath, &pItem)  );
	if (pItem == NULL)
	{
		SG_ERR_THROW2(SG_ERR_NOT_FOUND, (pCtx, "resolve item by repo path: %s", szRepoPath));
	}

	*ppItem = pItem;
	pItem = NULL;

fail:
	return;
}

void SG_resolve__check_item__gid(
	SG_context*        pCtx,
	SG_resolve*        pResolve,
	const char*        szGid,
	SG_resolve__item** ppItem
	)
{
	SG_resolve__item* pItem = NULL;

	SG_NULLARGCHECK(pResolve);
	SG_NULLARGCHECK(szGid);
	SG_NULLARGCHECK(ppItem);

	SG_ERR_CHECK(  SG_vector__find__first(pCtx, pResolve->pItems, _item__vector_predicate__match_gid, (void*)szGid, NULL, (void**)&pItem)  );

	*ppItem = pItem;
	pItem = NULL;

fail:
	return;
}

void SG_resolve__get_item__gid(
	SG_context*        pCtx,
	SG_resolve*        pResolve,
	const char*        szGid,
	SG_resolve__item** ppItem
	)
{
	SG_resolve__item* pItem = NULL;

	SG_NULLARGCHECK(pResolve);
	SG_NULLARGCHECK(szGid);
	SG_NULLARGCHECK(ppItem);

	SG_ERR_CHECK(  SG_resolve__check_item__gid(pCtx, pResolve, szGid, &pItem)  );
	if (pItem == NULL)
	{
		SG_ERR_THROW2(SG_ERR_NOT_FOUND, (pCtx, "resolve item by GID: %s", szGid));
	}

	*ppItem = pItem;
	pItem = NULL;

fail:
	return;
}

void SG_resolve__foreach_item(
	SG_context*                 pCtx,
	SG_resolve*                 pResolve,
	SG_bool                     bOnlyUnresolved,
	SG_resolve__item__callback* fCallback,
	void*                       pUserData
	)
{
	_item__vector_action__call_resolve_action__data cData;

	SG_NULLARGCHECK(pResolve);
	SG_NULLARGCHECK(fCallback);

	cData.fCallback = fCallback;
	cData.pUserData = pUserData;
	cData.bContinue = SG_TRUE;

	if (bOnlyUnresolved == SG_TRUE)
	{
		SG_bool bResolved = SG_FALSE;
		SG_ERR_CHECK(  SG_vector__find__all(pCtx, pResolve->pItems, _item__vector_predicate__match_resolved, &bResolved, _item__vector_action__call_resolve_action, &cData)  );
	}
	else
	{
		SG_ERR_CHECK(  SG_vector__foreach(pCtx, pResolve->pItems, _item__vector_action__call_resolve_action, &cData)  );
	}

fail:
	return;
}

void SG_resolve__foreach_choice(
	SG_context*                   pCtx,
	SG_resolve*                   pResolve,
	SG_bool                       bOnlyUnresolved,
	SG_resolve__choice__callback* fCallback,
	void*                         pUserData
	)
{
	_item__vector_action__foreach_choice__data cData;

	SG_NULLARGCHECK(pResolve);
	SG_NULLARGCHECK(fCallback);

	cData.bOnlyUnresolved = bOnlyUnresolved;
	cData.fCallback       = fCallback;
	cData.pUserData       = pUserData;
	cData.bContinue       = SG_TRUE;

	if (bOnlyUnresolved == SG_TRUE)
	{
		SG_bool bResolved = SG_FALSE;
		SG_ERR_CHECK(  SG_vector__find__all(pCtx, pResolve->pItems, _item__vector_predicate__match_resolved, &bResolved, _item__vector_action__foreach_choice, &cData)  );
	}
	else
	{
		SG_ERR_CHECK(  SG_vector__foreach(pCtx, pResolve->pItems, _item__vector_action__foreach_choice, &cData)  );
	}

fail:
	return;
}

void SG_resolve__item__get_resolve(
	SG_context*       pCtx,
	SG_resolve__item* pItem,
	SG_resolve**      ppResolve
	)
{
	SG_NULLARGCHECK(pItem);
	SG_NULLARGCHECK(ppResolve);

	*ppResolve = pItem->pResolve;

fail:
	return;
}

void SG_resolve__item__get_gid(
	SG_context*       pCtx,
	SG_resolve__item* pItem,
	const char**      ppGid
	)
{
	SG_NULLARGCHECK(pItem);
	SG_NULLARGCHECK(ppGid);

	*ppGid = pItem->szGid;

fail:
	return;
}

void SG_resolve__item__get_type(
	SG_context*             pCtx,
	SG_resolve__item*       pItem,
	SG_treenode_entry_type* pType
	)
{
	SG_NULLARGCHECK(pItem);
	SG_NULLARGCHECK(pType);

	*pType = pItem->eType;

fail:
	return;
}

void SG_resolve__item__get_status(
	SG_context*          pCtx,
	SG_resolve__item*    pItem,
	SG_diffstatus_flags* pStatus
	)
{
	SG_NULLARGCHECK(pItem);
	SG_NULLARGCHECK(pStatus);

	*pStatus = pItem->eStatus;

fail:
	return;
}

void SG_resolve__item__has_status(
	SG_context*         pCtx,
	SG_resolve__item*   pItem,
	SG_diffstatus_flags eStatus,
	SG_bool*            pHasStatus
	)
{
	SG_NULLARGCHECK(pItem);
	SG_NULLARGCHECK(pHasStatus);

	if ((pItem->eStatus & eStatus) == eStatus)
	{
		*pHasStatus = SG_TRUE;
	}
	else
	{
		*pHasStatus = SG_FALSE;
	}

fail:
	return;
}

void SG_resolve__item__get_ancestor(
	SG_context*       pCtx,
	SG_resolve__item* pItem,
	const char**      ppAncestorHid
	)
{
	SG_NULLARGCHECK(pItem);
	SG_NULLARGCHECK(ppAncestorHid);

	*ppAncestorHid = pItem->szAncestorHid;

fail:
	return;
}

void SG_resolve__item__get_repo_path(
	SG_context*       pCtx,
	SG_resolve__item* pItem,
	const char*       szChangeset,
	const char**      ppRepoPath
	)
{
	const char* szRepoPath = NULL;

	SG_NULLARGCHECK(pItem);
	SG_NULLARGCHECK(szChangeset);
	SG_NULLARGCHECK(ppRepoPath);

	SG_ERR_CHECK(  SG_vhash__check__sz(pCtx, pItem->pRepoPaths, szChangeset, &szRepoPath)  );

	*ppRepoPath = szRepoPath;
	szRepoPath = NULL;

fail:
	return;
}

void SG_resolve__item__get_repo_path__working(
	SG_context*       pCtx,
	SG_resolve__item* pItem,
	const char**      ppRepoPath
	)
{
	SG_NULLARGCHECK(pItem);
	SG_NULLARGCHECK(ppRepoPath);

	SG_ERR_CHECK(  SG_resolve__item__get_repo_path(pCtx, pItem, SG_RESOLVE__HID__WORKING, ppRepoPath)  );

fail:
	return;
}

void SG_resolve__item__get_repo_path__ancestor(
	SG_context*       pCtx,
	SG_resolve__item* pItem,
	const char**      ppRepoPath
	)
{
	SG_NULLARGCHECK(pItem);
	SG_NULLARGCHECK(ppRepoPath);

	if (pItem->szAncestorHid == NULL)
	{
		*ppRepoPath = NULL;
	}
	else
	{
		SG_ERR_CHECK(  SG_resolve__item__get_repo_path(pCtx, pItem, pItem->szAncestorHid, ppRepoPath)  );
	}

fail:
	return;
}

void SG_resolve__item__get_repo_path__baseline(
	SG_context*       pCtx,
	SG_resolve__item* pItem,
	const char**      ppRepoPath
	)
{
	const char* szHid = NULL;

	SG_NULLARGCHECK(pItem);
	SG_NULLARGCHECK(ppRepoPath);

	SG_ERR_CHECK(  SG_varray__get__sz(pCtx, pItem->pResolve->pParentHids, 0u, &szHid)  );
	SG_ERR_CHECK(  SG_resolve__item__get_repo_path(pCtx, pItem, szHid, ppRepoPath)  );

fail:
	return;
}

void SG_resolve__item__get_repo_path__other(
	SG_context*       pCtx,
	SG_resolve__item* pItem,
	SG_uint32         uIndex,
	const char**      ppRepoPath
	)
{
	const char* szHid = NULL;

	SG_NULLARGCHECK(pItem);
	SG_NULLARGCHECK(ppRepoPath);
	SG_ARGCHECK(uIndex > 0u, uIndex);

	SG_ERR_TRY(  SG_varray__get__sz(pCtx, pItem->pResolve->pParentHids, uIndex, &szHid)  );
	SG_ERR_CATCH_REPLACE(SG_ERR_VARRAY_INDEX_OUT_OF_RANGE, SG_ERR_NOT_FOUND);
	SG_ERR_CATCH_END;

	SG_ERR_CHECK(  SG_resolve__item__get_repo_path(pCtx, pItem, szHid, ppRepoPath)  );

fail:
	return;
}

void SG_resolve__item__check_choice__state(
	SG_context*          pCtx,
	SG_resolve__item*    pItem,
	SG_resolve__state    eState,
	SG_resolve__choice** ppChoice
	)
{
	SG_NULLARGCHECK(pItem);
	SG_NULLARGCHECK(ppChoice);

	*ppChoice = pItem->aChoices[eState];

fail:
	return;
}

void SG_resolve__item__get_choice__state(
	SG_context*          pCtx,
	SG_resolve__item*    pItem,
	SG_resolve__state    eState,
	SG_resolve__choice** ppChoice
	)
{
	SG_resolve__choice* pChoice = NULL;

	SG_NULLARGCHECK(pItem);
	SG_NULLARGCHECK(ppChoice);

	SG_ERR_CHECK(  SG_resolve__item__check_choice__state(pCtx, pItem, eState, &pChoice)  );
	if (pChoice == NULL)
	{
		SG_ERR_THROW2(SG_ERR_NOT_FOUND, (pCtx, "resolve choice by item/state: %s/%d", pItem->szGid, eState));
	}

	*ppChoice = pChoice;
	pChoice = NULL;

fail:
	return;
}

void SG_resolve__item__get_resolved(
	SG_context*       pCtx,
	SG_resolve__item* pItem,
	SG_bool*          pResolved,
	SG_bool*          pDeleted
	)
{
	SG_bool bResolved = SG_FALSE;
	SG_bool bDeleted  = SG_FALSE;

	SG_UNUSED(pCtx);
	SG_NULLARGCHECK(pItem);
	SG_NULLARGCHECK(pResolved);

	// check if the item is resolved at an item-wide level
	SG_ERR_CHECK(  _item__get_resolved(pCtx, pItem, &bResolved, &bDeleted)  );
	if (bResolved == SG_FALSE)
	{
		SG_uint32 uIndex = 0u;

		// item itself isn't resolved, check if all its choices are resolved
		bResolved = SG_TRUE;
		for (uIndex = 0u; uIndex < SG_RESOLVE__STATE__COUNT; ++uIndex)
		{
			SG_resolve__choice* pChoice = pItem->aChoices[uIndex];

			// check if the choice is resolved
			if (pChoice != NULL)
			{
				SG_ERR_CHECK(  _choice__get_resolved(pCtx, pChoice, &bResolved)  );

				// This is a temporary hack.  We want pDeleted to return SG_TRUE if
				// the item was resolved via an EXISTENCE choice accepting an SG_FALSE
				// value.  Once we support pDeleted returning true for ALL causes of
				// the item being deleted, this can go away and _item__get_resolved
				// will start returning the proper value.
				if (
					   uIndex == SG_RESOLVE__STATE__EXISTENCE
					&& pChoice->pResult != NULL
					&& pChoice->pResult->pVariant->v.val_bool == SG_FALSE
					)
				{
					bDeleted = SG_TRUE;
				}
			}

			// if this choice isn't resolved, the entire item is unresolved
			if (bResolved == SG_FALSE)
			{
				break;
			}
		}
	}

	// return the results
	*pResolved = bResolved;
	if (bResolved == SG_TRUE && pDeleted != NULL)
	{
		*pDeleted = bDeleted;
	}

fail:
	return;
}

void SG_resolve__item__foreach_choice(
	SG_context*                   pCtx,
	SG_resolve__item*             pItem,
	SG_bool                       bOnlyUnresolved,
	SG_resolve__choice__callback* fCallback,
	void*                         pUserData
	)
{
	SG_bool bContinue = SG_TRUE;

	SG_NULLARGCHECK(pItem);
	SG_NULLARGCHECK(fCallback);

	SG_ERR_CHECK(  _item__foreach_choice(pCtx, pItem, bOnlyUnresolved, fCallback, pUserData, &bContinue)  );

fail:
	return;
}

void SG_resolve__choice__get_item(
	SG_context*         pCtx,
	SG_resolve__choice* pChoice,
	SG_resolve__item**  ppItem
	)
{
	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(ppItem);

	*ppItem = pChoice->pItem;

fail:
	return;
}

void SG_resolve__choice__get_state(
	SG_context*         pCtx,
	SG_resolve__choice* pChoice,
	SG_resolve__state*  pState
	)
{
	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pState);

	*pState = pChoice->eState;

fail:
	return;
}

void SG_resolve__choice__get_causes(
	SG_context*                       pCtx,
	SG_resolve__choice*               pChoice,
	SG_mrg_cset_entry_conflict_flags* pConflicts,
	SG_bool*                          pCollision,
	SG_vhash**                        ppDescriptions
	)
{
	SG_vhash* pDescriptions = NULL;

	SG_NULLARGCHECK(pChoice);

	if (pConflicts != NULL)
	{
		*pConflicts = pChoice->eConflictCauses;
	}
	if (pCollision != NULL)
	{
		*pCollision = pChoice->bCollisionCause;
	}
	if (ppDescriptions != NULL)
	{
		SG_uint32 uIndex = 0u;

		SG_VHASH__ALLOC(pCtx, &pDescriptions);

		// run through each possible cause
		for (uIndex = 0u; uIndex < SG_NrElements(gaCauseData); ++uIndex)
		{
			// if this choice has the current cause
			// then add the cause's name/description to the hash
			if (
				   ((pChoice->eConflictCauses & gaCauseData[uIndex].eConflictCauses) != 0)
				|| (
					   pChoice->bCollisionCause == SG_TRUE
					&& gaCauseData[uIndex].bCollisionCause == SG_TRUE
					)
				)
			{
				SG_ERR_CHECK(  SG_vhash__add__string__sz(pCtx, pDescriptions, gaCauseData[uIndex].szName, gaCauseData[uIndex].szDescription)  );
			}
		}

		// return the generated hash
		*ppDescriptions = pDescriptions;
		pDescriptions = NULL;
	}

fail:
	SG_VHASH_NULLFREE(pCtx, pDescriptions);
	return;
}

void SG_resolve__choice__get_mergeable(
	SG_context*         pCtx,
	SG_resolve__choice* pChoice,
	SG_bool*            pMergeable
	)
{
	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pMergeable);

	if (pChoice->pLeafValues != NULL)
	{
		*pMergeable = SG_TRUE;
	}
	else
	{
		*pMergeable = SG_FALSE;
	}

fail:
	return;
}

void SG_resolve__choice__get_resolved(
	SG_context*         pCtx,
	SG_resolve__choice* pChoice,
	SG_bool*            pResolved,
	SG_bool*            pDeleted
	)
{
	SG_bool bResolved = SG_FALSE;
	SG_bool bDeleted  = SG_FALSE;

	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(pResolved);

	// check if our item is resolved at an item-wide level
	SG_ERR_CHECK(  _item__get_resolved(pCtx, pChoice->pItem, &bResolved, &bDeleted)  );

	// This is a temporary hack.  We want pDeleted to return SG_TRUE if
	// the item was resolved via an EXISTENCE choice accepting an SG_FALSE
	// value.  Once we support pDeleted returning true for ALL causes of
	// the item being deleted, this can go away and _item__get_resolved
	// will start returning the proper value.
	if (bResolved == SG_FALSE)
	{
		SG_resolve__choice* pExistenceChoice = pChoice->pItem->aChoices[SG_RESOLVE__STATE__EXISTENCE];
		if (
			   pExistenceChoice != NULL
			&& pExistenceChoice->pResult != NULL
			&& pExistenceChoice->pResult->pVariant->v.val_bool == SG_FALSE
			)
		{
			bResolved = SG_TRUE;
			bDeleted  = SG_TRUE;
		}
	}

	// check if this choice specifically is resolved
	if (bResolved == SG_FALSE)
	{
		SG_ERR_CHECK(  _choice__get_resolved(pCtx, pChoice, &bResolved)  );
	}

	// return the results
	*pResolved = bResolved;
	if (bResolved == SG_TRUE && pDeleted != NULL)
	{
		*pDeleted = bDeleted;
	}

fail:
	return;
}

void SG_resolve__choice__foreach_colliding_item(
	SG_context*                 pCtx,
	SG_resolve__choice*         pChoice,
	SG_resolve__item__callback* fCallback,
	void*                       pUserData
	)
{
	_item__vector_action__call_resolve_action__data cData;

	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(fCallback);

	if (pChoice->pCollisionItems == NULL)
	{
		return;
	}

	cData.fCallback = fCallback;
	cData.pUserData = pUserData;
	cData.bContinue = SG_TRUE;
	SG_ERR_CHECK(  SG_vector__foreach(pCtx, pChoice->pCollisionItems, _item__vector_action__call_resolve_action, &cData)  );

fail:
	return;
}

void SG_resolve__choice__get_cycle_info(
	SG_context*         pCtx,
	SG_resolve__choice* pChoice,
	const char**        ppPath
	)
{
	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(ppPath);

	*ppPath = pChoice->szCycleHint;

fail:
	return;
}

void SG_resolve__choice__foreach_cycling_item(
	SG_context*                 pCtx,
	SG_resolve__choice*         pChoice,
	SG_resolve__item__callback* fCallback,
	void*                       pUserData
	)
{
	_item__vector_action__call_resolve_action__data cData;

	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(fCallback);

	if (pChoice->pCycleItems == NULL)
	{
		return;
	}

	cData.fCallback = fCallback;
	cData.pUserData = pUserData;
	cData.bContinue = SG_TRUE;
	SG_ERR_CHECK(  SG_vector__foreach(pCtx, pChoice->pCycleItems, _item__vector_action__call_resolve_action, &cData)  );

fail:
	return;
}

void SG_resolve__choice__check_value__label(
	SG_context*         pCtx,
	SG_resolve__choice* pChoice,
	const char*         szLabel,
	SG_int32            iIndex,
	SG_resolve__value** ppValue
	)
{
	SG_resolve__value* pValue = NULL;
	SG_string*         sLabel = NULL;

	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(szLabel);
	SG_NULLARGCHECK(ppValue);

	// if they specified a negative index, convert it to a positive one
	if (iIndex < 0)
	{
		// figure out what the highest valid index for this label is
		SG_int32 iMaxIndex = 0;
		do
		{
			++iMaxIndex;
			SG_ERR_CHECK(  SG_resolve__choice__check_value__label(pCtx, pChoice, szLabel, iMaxIndex, &pValue)  );
		}
		while (pValue != NULL);

		// convert the negative index into the equivalent positive one
		iIndex = iMaxIndex + iIndex + 1;
		// if we ended up negative again, then the index was out of range
		// we'll check for that later
	}

	// if they specified a valid index, build the actual label
	// We ignore 0 because that signifies that the label is literal.
	// We ignore 1 because the first value in an indexed series doesn't
	// actually have a suffix of 1.  This makes for a friendlier user
	// experience in the 99% of cases where there is only a single value
	// with any given base name.
	if (iIndex > 1)
	{
		SG_ERR_CHECK(  SG_STRING__ALLOC(pCtx, &sLabel)  );
		SG_ERR_CHECK(  SG_string__sprintf(pCtx, sLabel, gszIndexedLabelFormat, szLabel, iIndex)  );
		szLabel = SG_string__sz(sLabel);
	}

	// find the value with the actual label
	// if our index is negative at this point, then it's out of range
	if (iIndex >= 0)
	{
		SG_ERR_CHECK(  SG_vector__find__first(pCtx, pChoice->pValues, _value__vector_predicate__match_label, (void*)szLabel, NULL, (void**)&pValue)  );
	}

	// return whatever we found
	*ppValue = pValue;
	pValue = NULL;

fail:
	SG_STRING_NULLFREE(pCtx, sLabel);
	return;
}

void SG_resolve__choice__get_value__label(
	SG_context*         pCtx,
	SG_resolve__choice* pChoice,
	const char*         szLabel,
	SG_int32            iIndex,
	SG_resolve__value** ppValue
	)
{
	SG_resolve__value* pValue = NULL;

	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(szLabel);
	SG_NULLARGCHECK(ppValue);

	SG_ERR_CHECK(  SG_resolve__choice__check_value__label(pCtx, pChoice, szLabel, iIndex, &pValue)  );
	if (pValue == NULL)
	{
		SG_ERR_THROW2(SG_ERR_NOT_FOUND, (pCtx, "resolve choice by value/index: %s/%d", szLabel, iIndex));
	}

	*ppValue = pValue;
	pValue = NULL;

fail:
	return;
}

void SG_resolve__choice__check_value__changeset(
	SG_context*         pCtx,
	SG_resolve__choice* pChoice,
	const char*         szHid,
	SG_resolve__value** ppValue
	)
{
	SG_resolve__value* pValue = NULL;

	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(szHid);
	SG_ARGCHECK(szHid != SG_RESOLVE__HID__WORKING, szHid);
	SG_NULLARGCHECK(ppValue);

	SG_ERR_CHECK(  SG_vector__find__first(pCtx, pChoice->pValues, _value__vector_predicate__match_changeset, (void*)szHid, NULL, (void**)&pValue)  );

	*ppValue = pValue;
	pValue = NULL;

fail:
	return;
}

void SG_resolve__choice__get_value__changeset(
	SG_context*         pCtx,
	SG_resolve__choice* pChoice,
	const char*         szHid,
	SG_resolve__value** ppValue
	)
{
	SG_resolve__value* pValue = NULL;

	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(szHid);
	SG_NULLARGCHECK(ppValue);

	SG_ERR_CHECK(  SG_resolve__choice__check_value__changeset(pCtx, pChoice, szHid, &pValue)  );
	if (pValue == NULL)
	{
		SG_ERR_THROW2(SG_ERR_NOT_FOUND, (pCtx, "resolve value by choice/changeset: %s", szHid));
	}

	*ppValue = pValue;
	pValue = NULL;

fail:
	return;
}

void SG_resolve__choice__get_value__accepted(
	SG_context*         pCtx,
	SG_resolve__choice* pChoice,
	SG_resolve__value** ppValue
	)
{
	SG_UNUSED(pCtx);
	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(ppValue);

	*ppValue = pChoice->pResult;

fail:
	return;
}

void SG_resolve__choice__foreach_value(
	SG_context*                  pCtx,
	SG_resolve__choice*          pChoice,
	SG_bool                      bOnlyLeaves,
	SG_resolve__value__callback* fCallback,
	void*                        pUserData
	)
{
	SG_vector*                                       pVector = NULL;
	_value__vector_action__call_resolve_action__data cData;

	SG_NULLARGCHECK(pChoice);
	SG_NULLARGCHECK(fCallback);

	cData.fCallback = fCallback;
	cData.pUserData = pUserData;
	cData.bContinue = SG_TRUE;

	if (bOnlyLeaves == SG_TRUE)
	{
		pVector = pChoice->pLeafValues;
	}
	else
	{
		pVector = pChoice->pValues;
	}

	SG_ERR_CHECK(  SG_vector__foreach(pCtx, pVector, _value__vector_action__call_resolve_action, &cData)  );

fail:
	return;
}

void SG_resolve__value__get_choice(
	SG_context*          pCtx,
	SG_resolve__value*   pValue,
	SG_resolve__choice** ppChoice
	)
{
	SG_NULLARGCHECK(pValue);
	SG_NULLARGCHECK(ppChoice);

	*ppChoice = pValue->pChoice;

fail:
	return;
}

void SG_resolve__value__get_label(
	SG_context*        pCtx,
	SG_resolve__value* pValue,
	const char**       ppLabel
	)
{
	SG_NULLARGCHECK(pValue);
	SG_NULLARGCHECK(ppLabel);

	*ppLabel = pValue->szLabel;

fail:
	return;
}

void SG_resolve__value__get_data__bool(
	SG_context*        pCtx,
	SG_resolve__value* pValue,
	SG_bool*           pData
	)
{
	SG_NULLARGCHECK(pValue);
	SG_NULLARGCHECK(pData);

	SG_ERR_TRY(  SG_variant__get__bool(pCtx, pValue->pVariant, pData)  );
	SG_ERR_CATCH_REPLACE(SG_ERR_VARIANT_INVALIDTYPE, SG_ERR_USAGE);
	SG_ERR_CATCH_END;

fail:
	return;
}

void SG_resolve__value__get_data__int64(
	SG_context*        pCtx,
	SG_resolve__value* pValue,
	SG_int64*          pData
	)
{
	SG_NULLARGCHECK(pValue);
	SG_NULLARGCHECK(pData);

	// I would call SG_variant__get__int64 here, but at the moment it
	// attempts an implicit conversion from a string, if the variant
	// happens to be a string (and a value's data could very possibly
	// be a string).  Until that's fixed, it's safer here to just check
	// the variant's internal data manually.
	if (pValue->pVariant->type != SG_VARIANT_TYPE_INT64)
	{
		SG_ERR_THROW(SG_ERR_USAGE);
	}
	*pData = pValue->pVariant->v.val_int64;

fail:
	return;
}

void SG_resolve__value__get_data__sz(
	SG_context*        pCtx,
	SG_resolve__value* pValue,
	const char**       ppData
	)
{
	SG_NULLARGCHECK(pValue);
	SG_NULLARGCHECK(ppData);

	SG_ERR_TRY(  SG_variant__get__sz(pCtx, pValue->pVariant, ppData)  );
	SG_ERR_CATCH_REPLACE(SG_ERR_VARIANT_INVALIDTYPE, SG_ERR_USAGE);
	SG_ERR_CATCH_END;

fail:
	return;
}

void SG_resolve__value__get_summary(
	SG_context*        pCtx,
	SG_resolve__value* pValue,
	SG_string**        ppSummary
	)
{
	SG_string* sSummary = NULL;

	SG_NULLARGCHECK(pValue);
	SG_NULLARGCHECK(ppSummary);

	switch (pValue->pChoice->eState)
	{
	case SG_RESOLVE__STATE__EXISTENCE:
		{
			SG_ERR_CHECK(  SG_string__alloc__format(pCtx, &sSummary, "%s", pValue->pVariant->v.val_bool == SG_FALSE ? "deleted" : "existing"));
			break;
		}
	case SG_RESOLVE__STATE__NAME:
		{
			SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &sSummary, pValue->pVariant->v.val_sz)  );
			break;
		}
	case SG_RESOLVE__STATE__LOCATION:
		{
			SG_ERR_CHECK(  SG_pendingtree__get_repo_path(pCtx, pValue->pChoice->pItem->pResolve->pPendingTree, pValue->pVariant->v.val_sz, &sSummary)  );
			break;
		}
	case SG_RESOLVE__STATE__ATTRIBUTES:
		{
			SG_uint64 uValue = 0u;
			SG_ERR_CHECK(  SG_variant__get__uint64(pCtx, pValue->pVariant, &uValue)  );
			SG_ERR_CHECK(  SG_string__alloc__format(pCtx, &sSummary, "%o", (SG_uint32)uValue)  );
			break;
		}
	case SG_RESOLVE__STATE__XATTRIBUTES:
		{
			// no good way to summarize extended attributes without just using the HID
			break;
		}
	case SG_RESOLVE__STATE__CONTENTS:
		{
			switch (pValue->pChoice->pItem->eType)
			{
			case SG_TREENODEENTRY_TYPE_REGULAR_FILE:
				{
					if (pValue->pBaselineParent != NULL && pValue->pOtherParent != NULL)
					{
						SG_ERR_CHECK(  SG_string__alloc__format(pCtx, &sSummary, "%s merged from '%s' and '%s' with '%s'", pValue->bAutoMerge == SG_TRUE ? "automatically" : "manually", pValue->pBaselineParent->szLabel, pValue->pOtherParent->szLabel, pValue->szMergeTool)  );
					}
					// no good way to summarize non merged file contents without just using the HID
					break;
				}
			case SG_TREENODEENTRY_TYPE_DIRECTORY:
				{
					SG_ERR_THROW2(SG_ERR_NOTIMPLEMENTED, (pCtx, "Cannot resolve a directory contents conflict because they should never occur."));
					break;
				}
			case SG_TREENODEENTRY_TYPE_SYMLINK:
				{
					SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &sSummary, pValue->pVariant->v.val_sz)  );
					break;
				}
			case SG_TREENODEENTRY_TYPE_SUBMODULE:
				{
					SG_ERR_CHECK(  SG_STRING__ALLOC__SZ(pCtx, &sSummary, pValue->szVariantSummary)  );
					break;
				}
			default:
				{
					SG_ERR_THROW2(SG_ERR_NOTIMPLEMENTED, (pCtx, "Unknown node type: %i", pValue->pChoice->pItem->eType));
					break;
				}
			}
			break;
		}
	default:
		{
			SG_ERR_THROW2(SG_ERR_NOTIMPLEMENTED, (pCtx, "Unknown resolve item state: %i", pValue->pChoice->eState));
			break;
		}
	}

	*ppSummary = sSummary;
	sSummary = NULL;

fail:
	SG_STRING_NULLFREE(pCtx, sSummary);
	return;
}

void SG_resolve__value__get_flags(
	SG_context*        pCtx,
	SG_resolve__value* pValue,
	SG_uint32*         pFlags
	)
{
	SG_uint32 uFlags = 0u;
	SG_bool   bLeaf  = SG_FALSE;

	SG_NULLARGCHECK(pValue);
	SG_NULLARGCHECK(pFlags);

	// check if the value is a changeset or a merge
	// and whether or not it was automatically created
	if (pValue->szChangeset != NULL)
	{
		uFlags |= SG_RESOLVE__VALUE__CHANGESET;
		uFlags |= SG_RESOLVE__VALUE__AUTOMATIC;
	}
	else if (pValue->pBaselineParent != NULL && pValue->pOtherParent != NULL)
	{
		uFlags |= SG_RESOLVE__VALUE__MERGE;
		if (pValue->bAutoMerge == SG_TRUE)
		{
			uFlags |= SG_RESOLVE__VALUE__AUTOMATIC;
		}
	}

	// check if the value is its choice's result
	if (pValue == pValue->pChoice->pResult)
	{
		uFlags |= SG_RESOLVE__VALUE__RESULT;
	}

	// check if the value can be overwritten
	// meaning it's not automatic and it's not a result
	if ((uFlags & (SG_RESOLVE__VALUE__AUTOMATIC | SG_RESOLVE__VALUE__RESULT)) == 0)
	{
		uFlags |= SG_RESOLVE__VALUE__OVERWRITABLE;
	}

	// check if the value is a merge leaf
	if (pValue->pChoice->pLeafValues != NULL)
	{
		SG_ERR_CHECK(  SG_vector__contains(pCtx, pValue->pChoice->pLeafValues, (void*)pValue, &bLeaf)  );
		if (bLeaf == SG_TRUE)
		{
			uFlags |= SG_RESOLVE__VALUE__LEAF;
		}
	}

	// return the flags
	*pFlags = uFlags;

fail:
	return;
}

void SG_resolve__value__has_flag(
	SG_context*        pCtx,
	SG_resolve__value* pValue,
	SG_uint32          uFlag,
	SG_bool*           pHasFlag
	)
{
	SG_uint32 uFlags = 0u;

	SG_NULLARGCHECK(pValue);
	SG_NULLARGCHECK(pHasFlag);

	SG_ERR_CHECK(  SG_resolve__value__get_flags(pCtx, pValue, &uFlags)  );
	if ((uFlags & uFlag) == uFlag)
	{
		*pHasFlag = SG_TRUE;
	}
	else
	{
		*pHasFlag = SG_FALSE;
	}

fail:
	return;
}

void SG_resolve__value__get_changeset(
	SG_context*        pCtx,
	SG_resolve__value* pValue,
	SG_bool*           pHasChangeset,
	const char**       ppChangeset
	)
{
	SG_NULLARGCHECK(pValue);
	SG_NULLARGCHECK(pHasChangeset);

	if (pValue->szChangeset == NULL)
	{
		*pHasChangeset = SG_FALSE;
	}
	else
	{
		*pHasChangeset = SG_TRUE;
		if (ppChangeset != NULL)
		{
			*ppChangeset = pValue->szChangeset;
		}
	}

fail:
	return;
}

void SG_resolve__value__get_merge(
	SG_context*         pCtx,
	SG_resolve__value*  pValue,
	SG_bool*            pWasMerged,
	SG_resolve__value** ppBaseline,
	SG_resolve__value** ppOther,
	SG_bool*            pAutomatic,
	const char**        ppTool,
	SG_int32*           pToolResult
	)
{
	SG_NULLARGCHECK(pValue);
	SG_NULLARGCHECK(pWasMerged);

	if (pValue->pBaselineParent == NULL || pValue->pOtherParent == NULL)
	{
		*pWasMerged = SG_FALSE;
	}
	else
	{
		*pWasMerged = SG_TRUE;
		if (ppBaseline != NULL)
		{
			*ppBaseline = pValue->pBaselineParent;
		}
		if (ppOther != NULL)
		{
			*ppOther = pValue->pOtherParent;
		}
		if (pAutomatic != NULL)
		{
			*pAutomatic = pValue->bAutoMerge;
		}
		if (ppTool != NULL)
		{
			*ppTool = pValue->szMergeTool;
		}
		if (pToolResult != NULL)
		{
			*pToolResult = pValue->iMergeToolResult;
		}
	}

fail:
	return;
}
